/* Automatically @generated by tree-sitter v0.25.4 (726dcd1e872149d95de581589fc408fb8ea9cb0b) */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 2015
#define LARGE_STATE_COUNT 455
#define SYMBOL_COUNT 360
#define ALIAS_COUNT 3
#define TOKEN_COUNT 161
#define EXTERNAL_TOKEN_COUNT 0
#define FIELD_COUNT 39
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 131
#define SUPERTYPE_COUNT 7

enum ts_symbol_identifiers {
  sym_identifier = 1,
  aux_sym_preproc_include_token1 = 2,
  aux_sym_preproc_include_token2 = 3,
  aux_sym_preproc_def_token1 = 4,
  anon_sym_LPAREN = 5,
  anon_sym_DOT_DOT_DOT = 6,
  anon_sym_COMMA = 7,
  anon_sym_RPAREN = 8,
  aux_sym_preproc_if_token1 = 9,
  anon_sym_LF = 10,
  aux_sym_preproc_if_token2 = 11,
  aux_sym_preproc_ifdef_token1 = 12,
  aux_sym_preproc_ifdef_token2 = 13,
  aux_sym_preproc_else_token1 = 14,
  aux_sym_preproc_elif_token1 = 15,
  aux_sym_preproc_elifdef_token1 = 16,
  aux_sym_preproc_elifdef_token2 = 17,
  sym_preproc_arg = 18,
  sym_preproc_directive = 19,
  anon_sym_LPAREN2 = 20,
  anon_sym_defined = 21,
  anon_sym_BANG = 22,
  anon_sym_TILDE = 23,
  anon_sym_DASH = 24,
  anon_sym_PLUS = 25,
  anon_sym_STAR = 26,
  anon_sym_SLASH = 27,
  anon_sym_PERCENT = 28,
  anon_sym_PIPE_PIPE = 29,
  anon_sym_AMP_AMP = 30,
  anon_sym_PIPE = 31,
  anon_sym_CARET = 32,
  anon_sym_AMP = 33,
  anon_sym_EQ_EQ = 34,
  anon_sym_BANG_EQ = 35,
  anon_sym_GT = 36,
  anon_sym_GT_EQ = 37,
  anon_sym_LT_EQ = 38,
  anon_sym_LT = 39,
  anon_sym_LT_LT = 40,
  anon_sym_GT_GT = 41,
  anon_sym_SEMI = 42,
  anon_sym___extension__ = 43,
  anon_sym_typedef = 44,
  anon_sym_extern = 45,
  anon_sym___attribute__ = 46,
  anon_sym___attribute = 47,
  anon_sym_COLON_COLON = 48,
  anon_sym_LBRACK_LBRACK = 49,
  anon_sym_RBRACK_RBRACK = 50,
  anon_sym___declspec = 51,
  anon_sym___based = 52,
  anon_sym___cdecl = 53,
  anon_sym___clrcall = 54,
  anon_sym___stdcall = 55,
  anon_sym___fastcall = 56,
  anon_sym___thiscall = 57,
  anon_sym___vectorcall = 58,
  sym_ms_restrict_modifier = 59,
  sym_ms_unsigned_ptr_modifier = 60,
  sym_ms_signed_ptr_modifier = 61,
  anon_sym__unaligned = 62,
  anon_sym___unaligned = 63,
  anon_sym_LBRACE = 64,
  anon_sym_RBRACE = 65,
  anon_sym_signed = 66,
  anon_sym_unsigned = 67,
  anon_sym_long = 68,
  anon_sym_short = 69,
  anon_sym_LBRACK = 70,
  anon_sym_static = 71,
  anon_sym_RBRACK = 72,
  anon_sym_EQ = 73,
  anon_sym_auto = 74,
  anon_sym_register = 75,
  anon_sym_inline = 76,
  anon_sym___inline = 77,
  anon_sym___inline__ = 78,
  anon_sym___forceinline = 79,
  anon_sym_thread_local = 80,
  anon_sym___thread = 81,
  anon_sym_const = 82,
  anon_sym_constexpr = 83,
  anon_sym_volatile = 84,
  anon_sym_restrict = 85,
  anon_sym___restrict__ = 86,
  anon_sym__Atomic = 87,
  anon_sym__Noreturn = 88,
  anon_sym_noreturn = 89,
  anon_sym__Nonnull = 90,
  anon_sym_alignas = 91,
  anon_sym__Alignas = 92,
  sym_primitive_type = 93,
  anon_sym_enum = 94,
  anon_sym_COLON = 95,
  anon_sym_struct = 96,
  anon_sym_union = 97,
  anon_sym_if = 98,
  anon_sym_else = 99,
  anon_sym_switch = 100,
  anon_sym_case = 101,
  anon_sym_default = 102,
  anon_sym_while = 103,
  anon_sym_do = 104,
  anon_sym_for = 105,
  anon_sym_return = 106,
  anon_sym_break = 107,
  anon_sym_continue = 108,
  anon_sym_goto = 109,
  anon_sym___try = 110,
  anon_sym___except = 111,
  anon_sym___finally = 112,
  anon_sym___leave = 113,
  anon_sym_QMARK = 114,
  anon_sym_STAR_EQ = 115,
  anon_sym_SLASH_EQ = 116,
  anon_sym_PERCENT_EQ = 117,
  anon_sym_PLUS_EQ = 118,
  anon_sym_DASH_EQ = 119,
  anon_sym_LT_LT_EQ = 120,
  anon_sym_GT_GT_EQ = 121,
  anon_sym_AMP_EQ = 122,
  anon_sym_CARET_EQ = 123,
  anon_sym_PIPE_EQ = 124,
  anon_sym_DASH_DASH = 125,
  anon_sym_PLUS_PLUS = 126,
  anon_sym_sizeof = 127,
  anon_sym___alignof__ = 128,
  anon_sym___alignof = 129,
  anon_sym__alignof = 130,
  anon_sym_alignof = 131,
  anon_sym__Alignof = 132,
  anon_sym_offsetof = 133,
  anon_sym__Generic = 134,
  anon_sym_asm = 135,
  anon_sym___asm__ = 136,
  anon_sym___asm = 137,
  anon_sym___volatile__ = 138,
  anon_sym_DOT = 139,
  anon_sym_DASH_GT = 140,
  sym_number_literal = 141,
  anon_sym_L_SQUOTE = 142,
  anon_sym_u_SQUOTE = 143,
  anon_sym_U_SQUOTE = 144,
  anon_sym_u8_SQUOTE = 145,
  anon_sym_SQUOTE = 146,
  aux_sym_char_literal_token1 = 147,
  anon_sym_L_DQUOTE = 148,
  anon_sym_u_DQUOTE = 149,
  anon_sym_U_DQUOTE = 150,
  anon_sym_u8_DQUOTE = 151,
  anon_sym_DQUOTE = 152,
  aux_sym_string_literal_token1 = 153,
  sym_escape_sequence = 154,
  sym_system_lib_string = 155,
  sym_true = 156,
  sym_false = 157,
  anon_sym_NULL = 158,
  anon_sym_nullptr = 159,
  sym_comment = 160,
  sym_translation_unit = 161,
  sym__top_level_item = 162,
  sym__block_item = 163,
  sym_preproc_include = 164,
  sym_preproc_def = 165,
  sym_preproc_function_def = 166,
  sym_preproc_params = 167,
  sym_preproc_call = 168,
  sym_preproc_if = 169,
  sym_preproc_ifdef = 170,
  sym_preproc_else = 171,
  sym_preproc_elif = 172,
  sym_preproc_elifdef = 173,
  sym_preproc_if_in_field_declaration_list = 174,
  sym_preproc_ifdef_in_field_declaration_list = 175,
  sym_preproc_else_in_field_declaration_list = 176,
  sym_preproc_elif_in_field_declaration_list = 177,
  sym_preproc_elifdef_in_field_declaration_list = 178,
  sym_preproc_if_in_enumerator_list = 179,
  sym_preproc_ifdef_in_enumerator_list = 180,
  sym_preproc_else_in_enumerator_list = 181,
  sym_preproc_elif_in_enumerator_list = 182,
  sym_preproc_elifdef_in_enumerator_list = 183,
  sym_preproc_if_in_enumerator_list_no_comma = 184,
  sym_preproc_ifdef_in_enumerator_list_no_comma = 185,
  sym_preproc_else_in_enumerator_list_no_comma = 186,
  sym_preproc_elif_in_enumerator_list_no_comma = 187,
  sym_preproc_elifdef_in_enumerator_list_no_comma = 188,
  sym__preproc_expression = 189,
  sym_preproc_parenthesized_expression = 190,
  sym_preproc_defined = 191,
  sym_preproc_unary_expression = 192,
  sym_preproc_call_expression = 193,
  sym_preproc_argument_list = 194,
  sym_preproc_binary_expression = 195,
  sym_function_definition = 196,
  sym__old_style_function_definition = 197,
  sym_declaration = 198,
  sym_type_definition = 199,
  sym__type_definition_type = 200,
  sym__type_definition_declarators = 201,
  sym__declaration_modifiers = 202,
  sym__declaration_specifiers = 203,
  sym_linkage_specification = 204,
  sym_attribute_specifier = 205,
  sym_attribute = 206,
  sym_attribute_declaration = 207,
  sym_ms_declspec_modifier = 208,
  sym_ms_based_modifier = 209,
  sym_ms_call_modifier = 210,
  sym_ms_unaligned_ptr_modifier = 211,
  sym_ms_pointer_modifier = 212,
  sym_declaration_list = 213,
  sym__declarator = 214,
  sym__declaration_declarator = 215,
  sym__field_declarator = 216,
  sym__type_declarator = 217,
  sym__abstract_declarator = 218,
  sym_parenthesized_declarator = 219,
  sym_parenthesized_field_declarator = 220,
  sym_parenthesized_type_declarator = 221,
  sym_abstract_parenthesized_declarator = 222,
  sym_attributed_declarator = 223,
  sym_attributed_field_declarator = 224,
  sym_attributed_type_declarator = 225,
  sym_pointer_declarator = 226,
  sym_pointer_field_declarator = 227,
  sym_pointer_type_declarator = 228,
  sym_abstract_pointer_declarator = 229,
  sym_function_declarator = 230,
  sym__function_declaration_declarator = 231,
  sym_function_field_declarator = 232,
  sym_function_type_declarator = 233,
  sym_abstract_function_declarator = 234,
  sym__old_style_function_declarator = 235,
  sym_array_declarator = 236,
  sym_array_field_declarator = 237,
  sym_array_type_declarator = 238,
  sym_abstract_array_declarator = 239,
  sym_init_declarator = 240,
  sym_compound_statement = 241,
  sym_storage_class_specifier = 242,
  sym_type_qualifier = 243,
  sym_alignas_qualifier = 244,
  sym_type_specifier = 245,
  sym_sized_type_specifier = 246,
  sym_enum_specifier = 247,
  sym_enumerator_list = 248,
  sym_struct_specifier = 249,
  sym_union_specifier = 250,
  sym_field_declaration_list = 251,
  sym__field_declaration_list_item = 252,
  sym_field_declaration = 253,
  sym__field_declaration_declarator = 254,
  sym_bitfield_clause = 255,
  sym_enumerator = 256,
  sym_variadic_parameter = 257,
  sym_parameter_list = 258,
  sym__old_style_parameter_list = 259,
  sym_parameter_declaration = 260,
  sym_attributed_statement = 261,
  sym_statement = 262,
  sym__top_level_statement = 263,
  sym_labeled_statement = 264,
  sym__top_level_expression_statement = 265,
  sym_expression_statement = 266,
  sym_if_statement = 267,
  sym_else_clause = 268,
  sym_switch_statement = 269,
  sym_case_statement = 270,
  sym_while_statement = 271,
  sym_do_statement = 272,
  sym_for_statement = 273,
  sym__for_statement_body = 274,
  sym_return_statement = 275,
  sym_break_statement = 276,
  sym_continue_statement = 277,
  sym_goto_statement = 278,
  sym_seh_try_statement = 279,
  sym_seh_except_clause = 280,
  sym_seh_finally_clause = 281,
  sym_seh_leave_statement = 282,
  sym_expression = 283,
  sym__string = 284,
  sym_comma_expression = 285,
  sym_conditional_expression = 286,
  sym_assignment_expression = 287,
  sym_pointer_expression = 288,
  sym_unary_expression = 289,
  sym_binary_expression = 290,
  sym_update_expression = 291,
  sym_cast_expression = 292,
  sym_type_descriptor = 293,
  sym_sizeof_expression = 294,
  sym_alignof_expression = 295,
  sym_offsetof_expression = 296,
  sym_generic_expression = 297,
  sym_subscript_expression = 298,
  sym_call_expression = 299,
  sym_gnu_asm_expression = 300,
  sym_gnu_asm_qualifier = 301,
  sym_gnu_asm_output_operand_list = 302,
  sym_gnu_asm_output_operand = 303,
  sym_gnu_asm_input_operand_list = 304,
  sym_gnu_asm_input_operand = 305,
  sym_gnu_asm_clobber_list = 306,
  sym_gnu_asm_goto_list = 307,
  sym_extension_expression = 308,
  sym_argument_list = 309,
  sym_field_expression = 310,
  sym_compound_literal_expression = 311,
  sym_parenthesized_expression = 312,
  sym_initializer_list = 313,
  sym_initializer_pair = 314,
  sym_subscript_designator = 315,
  sym_subscript_range_designator = 316,
  sym_field_designator = 317,
  sym_char_literal = 318,
  sym_concatenated_string = 319,
  sym_string_literal = 320,
  sym_null = 321,
  sym__empty_declaration = 322,
  sym_macro_type_specifier = 323,
  aux_sym_translation_unit_repeat1 = 324,
  aux_sym_preproc_params_repeat1 = 325,
  aux_sym_preproc_if_repeat1 = 326,
  aux_sym_preproc_if_in_field_declaration_list_repeat1 = 327,
  aux_sym_preproc_if_in_enumerator_list_repeat1 = 328,
  aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1 = 329,
  aux_sym_preproc_argument_list_repeat1 = 330,
  aux_sym__old_style_function_definition_repeat1 = 331,
  aux_sym_declaration_repeat1 = 332,
  aux_sym_type_definition_repeat1 = 333,
  aux_sym__type_definition_type_repeat1 = 334,
  aux_sym__type_definition_declarators_repeat1 = 335,
  aux_sym__declaration_specifiers_repeat1 = 336,
  aux_sym_attribute_declaration_repeat1 = 337,
  aux_sym_attributed_declarator_repeat1 = 338,
  aux_sym_pointer_declarator_repeat1 = 339,
  aux_sym_function_declarator_repeat1 = 340,
  aux_sym_array_declarator_repeat1 = 341,
  aux_sym_sized_type_specifier_repeat1 = 342,
  aux_sym_enumerator_list_repeat1 = 343,
  aux_sym__field_declaration_declarator_repeat1 = 344,
  aux_sym_parameter_list_repeat1 = 345,
  aux_sym__old_style_parameter_list_repeat1 = 346,
  aux_sym_case_statement_repeat1 = 347,
  aux_sym_generic_expression_repeat1 = 348,
  aux_sym_gnu_asm_expression_repeat1 = 349,
  aux_sym_gnu_asm_output_operand_list_repeat1 = 350,
  aux_sym_gnu_asm_input_operand_list_repeat1 = 351,
  aux_sym_gnu_asm_clobber_list_repeat1 = 352,
  aux_sym_gnu_asm_goto_list_repeat1 = 353,
  aux_sym_argument_list_repeat1 = 354,
  aux_sym_initializer_list_repeat1 = 355,
  aux_sym_initializer_pair_repeat1 = 356,
  aux_sym_char_literal_repeat1 = 357,
  aux_sym_concatenated_string_repeat1 = 358,
  aux_sym_string_literal_repeat1 = 359,
  alias_sym_field_identifier = 360,
  alias_sym_statement_identifier = 361,
  alias_sym_type_identifier = 362,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [aux_sym_preproc_include_token1] = "#include",
  [aux_sym_preproc_include_token2] = "preproc_include_token2",
  [aux_sym_preproc_def_token1] = "#define",
  [anon_sym_LPAREN] = "(",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_COMMA] = ",",
  [anon_sym_RPAREN] = ")",
  [aux_sym_preproc_if_token1] = "#if",
  [anon_sym_LF] = "\n",
  [aux_sym_preproc_if_token2] = "#endif",
  [aux_sym_preproc_ifdef_token1] = "#ifdef",
  [aux_sym_preproc_ifdef_token2] = "#ifndef",
  [aux_sym_preproc_else_token1] = "#else",
  [aux_sym_preproc_elif_token1] = "#elif",
  [aux_sym_preproc_elifdef_token1] = "#elifdef",
  [aux_sym_preproc_elifdef_token2] = "#elifndef",
  [sym_preproc_arg] = "preproc_arg",
  [sym_preproc_directive] = "preproc_directive",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_defined] = "defined",
  [anon_sym_BANG] = "!",
  [anon_sym_TILDE] = "~",
  [anon_sym_DASH] = "-",
  [anon_sym_PLUS] = "+",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_LT] = "<",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_SEMI] = ";",
  [anon_sym___extension__] = "__extension__",
  [anon_sym_typedef] = "typedef",
  [anon_sym_extern] = "extern",
  [anon_sym___attribute__] = "__attribute__",
  [anon_sym___attribute] = "__attribute",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_LBRACK_LBRACK] = "[[",
  [anon_sym_RBRACK_RBRACK] = "]]",
  [anon_sym___declspec] = "__declspec",
  [anon_sym___based] = "__based",
  [anon_sym___cdecl] = "__cdecl",
  [anon_sym___clrcall] = "__clrcall",
  [anon_sym___stdcall] = "__stdcall",
  [anon_sym___fastcall] = "__fastcall",
  [anon_sym___thiscall] = "__thiscall",
  [anon_sym___vectorcall] = "__vectorcall",
  [sym_ms_restrict_modifier] = "ms_restrict_modifier",
  [sym_ms_unsigned_ptr_modifier] = "ms_unsigned_ptr_modifier",
  [sym_ms_signed_ptr_modifier] = "ms_signed_ptr_modifier",
  [anon_sym__unaligned] = "_unaligned",
  [anon_sym___unaligned] = "__unaligned",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_signed] = "signed",
  [anon_sym_unsigned] = "unsigned",
  [anon_sym_long] = "long",
  [anon_sym_short] = "short",
  [anon_sym_LBRACK] = "[",
  [anon_sym_static] = "static",
  [anon_sym_RBRACK] = "]",
  [anon_sym_EQ] = "=",
  [anon_sym_auto] = "auto",
  [anon_sym_register] = "register",
  [anon_sym_inline] = "inline",
  [anon_sym___inline] = "__inline",
  [anon_sym___inline__] = "__inline__",
  [anon_sym___forceinline] = "__forceinline",
  [anon_sym_thread_local] = "thread_local",
  [anon_sym___thread] = "__thread",
  [anon_sym_const] = "const",
  [anon_sym_constexpr] = "constexpr",
  [anon_sym_volatile] = "volatile",
  [anon_sym_restrict] = "restrict",
  [anon_sym___restrict__] = "__restrict__",
  [anon_sym__Atomic] = "_Atomic",
  [anon_sym__Noreturn] = "_Noreturn",
  [anon_sym_noreturn] = "noreturn",
  [anon_sym__Nonnull] = "_Nonnull",
  [anon_sym_alignas] = "alignas",
  [anon_sym__Alignas] = "_Alignas",
  [sym_primitive_type] = "primitive_type",
  [anon_sym_enum] = "enum",
  [anon_sym_COLON] = ":",
  [anon_sym_struct] = "struct",
  [anon_sym_union] = "union",
  [anon_sym_if] = "if",
  [anon_sym_else] = "else",
  [anon_sym_switch] = "switch",
  [anon_sym_case] = "case",
  [anon_sym_default] = "default",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_for] = "for",
  [anon_sym_return] = "return",
  [anon_sym_break] = "break",
  [anon_sym_continue] = "continue",
  [anon_sym_goto] = "goto",
  [anon_sym___try] = "__try",
  [anon_sym___except] = "__except",
  [anon_sym___finally] = "__finally",
  [anon_sym___leave] = "__leave",
  [anon_sym_QMARK] = "\?",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_sizeof] = "sizeof",
  [anon_sym___alignof__] = "__alignof__",
  [anon_sym___alignof] = "__alignof",
  [anon_sym__alignof] = "_alignof",
  [anon_sym_alignof] = "alignof",
  [anon_sym__Alignof] = "_Alignof",
  [anon_sym_offsetof] = "offsetof",
  [anon_sym__Generic] = "_Generic",
  [anon_sym_asm] = "asm",
  [anon_sym___asm__] = "__asm__",
  [anon_sym___asm] = "__asm",
  [anon_sym___volatile__] = "__volatile__",
  [anon_sym_DOT] = ".",
  [anon_sym_DASH_GT] = "->",
  [sym_number_literal] = "number_literal",
  [anon_sym_L_SQUOTE] = "L'",
  [anon_sym_u_SQUOTE] = "u'",
  [anon_sym_U_SQUOTE] = "U'",
  [anon_sym_u8_SQUOTE] = "u8'",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_char_literal_token1] = "character",
  [anon_sym_L_DQUOTE] = "L\"",
  [anon_sym_u_DQUOTE] = "u\"",
  [anon_sym_U_DQUOTE] = "U\"",
  [anon_sym_u8_DQUOTE] = "u8\"",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_literal_token1] = "string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_system_lib_string] = "system_lib_string",
  [sym_true] = "true",
  [sym_false] = "false",
  [anon_sym_NULL] = "NULL",
  [anon_sym_nullptr] = "nullptr",
  [sym_comment] = "comment",
  [sym_translation_unit] = "translation_unit",
  [sym__top_level_item] = "_top_level_item",
  [sym__block_item] = "_block_item",
  [sym_preproc_include] = "preproc_include",
  [sym_preproc_def] = "preproc_def",
  [sym_preproc_function_def] = "preproc_function_def",
  [sym_preproc_params] = "preproc_params",
  [sym_preproc_call] = "preproc_call",
  [sym_preproc_if] = "preproc_if",
  [sym_preproc_ifdef] = "preproc_ifdef",
  [sym_preproc_else] = "preproc_else",
  [sym_preproc_elif] = "preproc_elif",
  [sym_preproc_elifdef] = "preproc_elifdef",
  [sym_preproc_if_in_field_declaration_list] = "preproc_if",
  [sym_preproc_ifdef_in_field_declaration_list] = "preproc_ifdef",
  [sym_preproc_else_in_field_declaration_list] = "preproc_else",
  [sym_preproc_elif_in_field_declaration_list] = "preproc_elif",
  [sym_preproc_elifdef_in_field_declaration_list] = "preproc_elifdef",
  [sym_preproc_if_in_enumerator_list] = "preproc_if",
  [sym_preproc_ifdef_in_enumerator_list] = "preproc_ifdef",
  [sym_preproc_else_in_enumerator_list] = "preproc_else",
  [sym_preproc_elif_in_enumerator_list] = "preproc_elif",
  [sym_preproc_elifdef_in_enumerator_list] = "preproc_elifdef",
  [sym_preproc_if_in_enumerator_list_no_comma] = "preproc_if",
  [sym_preproc_ifdef_in_enumerator_list_no_comma] = "preproc_ifdef",
  [sym_preproc_else_in_enumerator_list_no_comma] = "preproc_else",
  [sym_preproc_elif_in_enumerator_list_no_comma] = "preproc_elif",
  [sym_preproc_elifdef_in_enumerator_list_no_comma] = "preproc_elifdef",
  [sym__preproc_expression] = "_preproc_expression",
  [sym_preproc_parenthesized_expression] = "parenthesized_expression",
  [sym_preproc_defined] = "preproc_defined",
  [sym_preproc_unary_expression] = "unary_expression",
  [sym_preproc_call_expression] = "call_expression",
  [sym_preproc_argument_list] = "argument_list",
  [sym_preproc_binary_expression] = "binary_expression",
  [sym_function_definition] = "function_definition",
  [sym__old_style_function_definition] = "function_definition",
  [sym_declaration] = "declaration",
  [sym_type_definition] = "type_definition",
  [sym__type_definition_type] = "_type_definition_type",
  [sym__type_definition_declarators] = "_type_definition_declarators",
  [sym__declaration_modifiers] = "_declaration_modifiers",
  [sym__declaration_specifiers] = "_declaration_specifiers",
  [sym_linkage_specification] = "linkage_specification",
  [sym_attribute_specifier] = "attribute_specifier",
  [sym_attribute] = "attribute",
  [sym_attribute_declaration] = "attribute_declaration",
  [sym_ms_declspec_modifier] = "ms_declspec_modifier",
  [sym_ms_based_modifier] = "ms_based_modifier",
  [sym_ms_call_modifier] = "ms_call_modifier",
  [sym_ms_unaligned_ptr_modifier] = "ms_unaligned_ptr_modifier",
  [sym_ms_pointer_modifier] = "ms_pointer_modifier",
  [sym_declaration_list] = "declaration_list",
  [sym__declarator] = "_declarator",
  [sym__declaration_declarator] = "_declaration_declarator",
  [sym__field_declarator] = "_field_declarator",
  [sym__type_declarator] = "_type_declarator",
  [sym__abstract_declarator] = "_abstract_declarator",
  [sym_parenthesized_declarator] = "parenthesized_declarator",
  [sym_parenthesized_field_declarator] = "parenthesized_declarator",
  [sym_parenthesized_type_declarator] = "parenthesized_declarator",
  [sym_abstract_parenthesized_declarator] = "abstract_parenthesized_declarator",
  [sym_attributed_declarator] = "attributed_declarator",
  [sym_attributed_field_declarator] = "attributed_declarator",
  [sym_attributed_type_declarator] = "attributed_declarator",
  [sym_pointer_declarator] = "pointer_declarator",
  [sym_pointer_field_declarator] = "pointer_declarator",
  [sym_pointer_type_declarator] = "pointer_declarator",
  [sym_abstract_pointer_declarator] = "abstract_pointer_declarator",
  [sym_function_declarator] = "function_declarator",
  [sym__function_declaration_declarator] = "function_declarator",
  [sym_function_field_declarator] = "function_declarator",
  [sym_function_type_declarator] = "function_declarator",
  [sym_abstract_function_declarator] = "abstract_function_declarator",
  [sym__old_style_function_declarator] = "function_declarator",
  [sym_array_declarator] = "array_declarator",
  [sym_array_field_declarator] = "array_declarator",
  [sym_array_type_declarator] = "array_declarator",
  [sym_abstract_array_declarator] = "abstract_array_declarator",
  [sym_init_declarator] = "init_declarator",
  [sym_compound_statement] = "compound_statement",
  [sym_storage_class_specifier] = "storage_class_specifier",
  [sym_type_qualifier] = "type_qualifier",
  [sym_alignas_qualifier] = "alignas_qualifier",
  [sym_type_specifier] = "type_specifier",
  [sym_sized_type_specifier] = "sized_type_specifier",
  [sym_enum_specifier] = "enum_specifier",
  [sym_enumerator_list] = "enumerator_list",
  [sym_struct_specifier] = "struct_specifier",
  [sym_union_specifier] = "union_specifier",
  [sym_field_declaration_list] = "field_declaration_list",
  [sym__field_declaration_list_item] = "_field_declaration_list_item",
  [sym_field_declaration] = "field_declaration",
  [sym__field_declaration_declarator] = "_field_declaration_declarator",
  [sym_bitfield_clause] = "bitfield_clause",
  [sym_enumerator] = "enumerator",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_parameter_list] = "parameter_list",
  [sym__old_style_parameter_list] = "parameter_list",
  [sym_parameter_declaration] = "parameter_declaration",
  [sym_attributed_statement] = "attributed_statement",
  [sym_statement] = "statement",
  [sym__top_level_statement] = "_top_level_statement",
  [sym_labeled_statement] = "labeled_statement",
  [sym__top_level_expression_statement] = "expression_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_if_statement] = "if_statement",
  [sym_else_clause] = "else_clause",
  [sym_switch_statement] = "switch_statement",
  [sym_case_statement] = "case_statement",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_for_statement] = "for_statement",
  [sym__for_statement_body] = "_for_statement_body",
  [sym_return_statement] = "return_statement",
  [sym_break_statement] = "break_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_goto_statement] = "goto_statement",
  [sym_seh_try_statement] = "seh_try_statement",
  [sym_seh_except_clause] = "seh_except_clause",
  [sym_seh_finally_clause] = "seh_finally_clause",
  [sym_seh_leave_statement] = "seh_leave_statement",
  [sym_expression] = "expression",
  [sym__string] = "_string",
  [sym_comma_expression] = "comma_expression",
  [sym_conditional_expression] = "conditional_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_pointer_expression] = "pointer_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_update_expression] = "update_expression",
  [sym_cast_expression] = "cast_expression",
  [sym_type_descriptor] = "type_descriptor",
  [sym_sizeof_expression] = "sizeof_expression",
  [sym_alignof_expression] = "alignof_expression",
  [sym_offsetof_expression] = "offsetof_expression",
  [sym_generic_expression] = "generic_expression",
  [sym_subscript_expression] = "subscript_expression",
  [sym_call_expression] = "call_expression",
  [sym_gnu_asm_expression] = "gnu_asm_expression",
  [sym_gnu_asm_qualifier] = "gnu_asm_qualifier",
  [sym_gnu_asm_output_operand_list] = "gnu_asm_output_operand_list",
  [sym_gnu_asm_output_operand] = "gnu_asm_output_operand",
  [sym_gnu_asm_input_operand_list] = "gnu_asm_input_operand_list",
  [sym_gnu_asm_input_operand] = "gnu_asm_input_operand",
  [sym_gnu_asm_clobber_list] = "gnu_asm_clobber_list",
  [sym_gnu_asm_goto_list] = "gnu_asm_goto_list",
  [sym_extension_expression] = "extension_expression",
  [sym_argument_list] = "argument_list",
  [sym_field_expression] = "field_expression",
  [sym_compound_literal_expression] = "compound_literal_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_initializer_list] = "initializer_list",
  [sym_initializer_pair] = "initializer_pair",
  [sym_subscript_designator] = "subscript_designator",
  [sym_subscript_range_designator] = "subscript_range_designator",
  [sym_field_designator] = "field_designator",
  [sym_char_literal] = "char_literal",
  [sym_concatenated_string] = "concatenated_string",
  [sym_string_literal] = "string_literal",
  [sym_null] = "null",
  [sym__empty_declaration] = "_empty_declaration",
  [sym_macro_type_specifier] = "macro_type_specifier",
  [aux_sym_translation_unit_repeat1] = "translation_unit_repeat1",
  [aux_sym_preproc_params_repeat1] = "preproc_params_repeat1",
  [aux_sym_preproc_if_repeat1] = "preproc_if_repeat1",
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = "preproc_if_in_field_declaration_list_repeat1",
  [aux_sym_preproc_if_in_enumerator_list_repeat1] = "preproc_if_in_enumerator_list_repeat1",
  [aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1] = "preproc_if_in_enumerator_list_no_comma_repeat1",
  [aux_sym_preproc_argument_list_repeat1] = "preproc_argument_list_repeat1",
  [aux_sym__old_style_function_definition_repeat1] = "_old_style_function_definition_repeat1",
  [aux_sym_declaration_repeat1] = "declaration_repeat1",
  [aux_sym_type_definition_repeat1] = "type_definition_repeat1",
  [aux_sym__type_definition_type_repeat1] = "_type_definition_type_repeat1",
  [aux_sym__type_definition_declarators_repeat1] = "_type_definition_declarators_repeat1",
  [aux_sym__declaration_specifiers_repeat1] = "_declaration_specifiers_repeat1",
  [aux_sym_attribute_declaration_repeat1] = "attribute_declaration_repeat1",
  [aux_sym_attributed_declarator_repeat1] = "attributed_declarator_repeat1",
  [aux_sym_pointer_declarator_repeat1] = "pointer_declarator_repeat1",
  [aux_sym_function_declarator_repeat1] = "function_declarator_repeat1",
  [aux_sym_array_declarator_repeat1] = "array_declarator_repeat1",
  [aux_sym_sized_type_specifier_repeat1] = "sized_type_specifier_repeat1",
  [aux_sym_enumerator_list_repeat1] = "enumerator_list_repeat1",
  [aux_sym__field_declaration_declarator_repeat1] = "_field_declaration_declarator_repeat1",
  [aux_sym_parameter_list_repeat1] = "parameter_list_repeat1",
  [aux_sym__old_style_parameter_list_repeat1] = "_old_style_parameter_list_repeat1",
  [aux_sym_case_statement_repeat1] = "case_statement_repeat1",
  [aux_sym_generic_expression_repeat1] = "generic_expression_repeat1",
  [aux_sym_gnu_asm_expression_repeat1] = "gnu_asm_expression_repeat1",
  [aux_sym_gnu_asm_output_operand_list_repeat1] = "gnu_asm_output_operand_list_repeat1",
  [aux_sym_gnu_asm_input_operand_list_repeat1] = "gnu_asm_input_operand_list_repeat1",
  [aux_sym_gnu_asm_clobber_list_repeat1] = "gnu_asm_clobber_list_repeat1",
  [aux_sym_gnu_asm_goto_list_repeat1] = "gnu_asm_goto_list_repeat1",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym_initializer_list_repeat1] = "initializer_list_repeat1",
  [aux_sym_initializer_pair_repeat1] = "initializer_pair_repeat1",
  [aux_sym_char_literal_repeat1] = "char_literal_repeat1",
  [aux_sym_concatenated_string_repeat1] = "concatenated_string_repeat1",
  [aux_sym_string_literal_repeat1] = "string_literal_repeat1",
  [alias_sym_field_identifier] = "field_identifier",
  [alias_sym_statement_identifier] = "statement_identifier",
  [alias_sym_type_identifier] = "type_identifier",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [aux_sym_preproc_include_token1] = aux_sym_preproc_include_token1,
  [aux_sym_preproc_include_token2] = aux_sym_preproc_include_token2,
  [aux_sym_preproc_def_token1] = aux_sym_preproc_def_token1,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [aux_sym_preproc_if_token1] = aux_sym_preproc_if_token1,
  [anon_sym_LF] = anon_sym_LF,
  [aux_sym_preproc_if_token2] = aux_sym_preproc_if_token2,
  [aux_sym_preproc_ifdef_token1] = aux_sym_preproc_ifdef_token1,
  [aux_sym_preproc_ifdef_token2] = aux_sym_preproc_ifdef_token2,
  [aux_sym_preproc_else_token1] = aux_sym_preproc_else_token1,
  [aux_sym_preproc_elif_token1] = aux_sym_preproc_elif_token1,
  [aux_sym_preproc_elifdef_token1] = aux_sym_preproc_elifdef_token1,
  [aux_sym_preproc_elifdef_token2] = aux_sym_preproc_elifdef_token2,
  [sym_preproc_arg] = sym_preproc_arg,
  [sym_preproc_directive] = sym_preproc_directive,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_defined] = anon_sym_defined,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym___extension__] = anon_sym___extension__,
  [anon_sym_typedef] = anon_sym_typedef,
  [anon_sym_extern] = anon_sym_extern,
  [anon_sym___attribute__] = anon_sym___attribute__,
  [anon_sym___attribute] = anon_sym___attribute,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_LBRACK_LBRACK] = anon_sym_LBRACK_LBRACK,
  [anon_sym_RBRACK_RBRACK] = anon_sym_RBRACK_RBRACK,
  [anon_sym___declspec] = anon_sym___declspec,
  [anon_sym___based] = anon_sym___based,
  [anon_sym___cdecl] = anon_sym___cdecl,
  [anon_sym___clrcall] = anon_sym___clrcall,
  [anon_sym___stdcall] = anon_sym___stdcall,
  [anon_sym___fastcall] = anon_sym___fastcall,
  [anon_sym___thiscall] = anon_sym___thiscall,
  [anon_sym___vectorcall] = anon_sym___vectorcall,
  [sym_ms_restrict_modifier] = sym_ms_restrict_modifier,
  [sym_ms_unsigned_ptr_modifier] = sym_ms_unsigned_ptr_modifier,
  [sym_ms_signed_ptr_modifier] = sym_ms_signed_ptr_modifier,
  [anon_sym__unaligned] = anon_sym__unaligned,
  [anon_sym___unaligned] = anon_sym___unaligned,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_signed] = anon_sym_signed,
  [anon_sym_unsigned] = anon_sym_unsigned,
  [anon_sym_long] = anon_sym_long,
  [anon_sym_short] = anon_sym_short,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_static] = anon_sym_static,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_auto] = anon_sym_auto,
  [anon_sym_register] = anon_sym_register,
  [anon_sym_inline] = anon_sym_inline,
  [anon_sym___inline] = anon_sym___inline,
  [anon_sym___inline__] = anon_sym___inline__,
  [anon_sym___forceinline] = anon_sym___forceinline,
  [anon_sym_thread_local] = anon_sym_thread_local,
  [anon_sym___thread] = anon_sym___thread,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_constexpr] = anon_sym_constexpr,
  [anon_sym_volatile] = anon_sym_volatile,
  [anon_sym_restrict] = anon_sym_restrict,
  [anon_sym___restrict__] = anon_sym___restrict__,
  [anon_sym__Atomic] = anon_sym__Atomic,
  [anon_sym__Noreturn] = anon_sym__Noreturn,
  [anon_sym_noreturn] = anon_sym_noreturn,
  [anon_sym__Nonnull] = anon_sym__Nonnull,
  [anon_sym_alignas] = anon_sym_alignas,
  [anon_sym__Alignas] = anon_sym__Alignas,
  [sym_primitive_type] = sym_primitive_type,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_union] = anon_sym_union,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_switch] = anon_sym_switch,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_goto] = anon_sym_goto,
  [anon_sym___try] = anon_sym___try,
  [anon_sym___except] = anon_sym___except,
  [anon_sym___finally] = anon_sym___finally,
  [anon_sym___leave] = anon_sym___leave,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_sizeof] = anon_sym_sizeof,
  [anon_sym___alignof__] = anon_sym___alignof__,
  [anon_sym___alignof] = anon_sym___alignof,
  [anon_sym__alignof] = anon_sym__alignof,
  [anon_sym_alignof] = anon_sym_alignof,
  [anon_sym__Alignof] = anon_sym__Alignof,
  [anon_sym_offsetof] = anon_sym_offsetof,
  [anon_sym__Generic] = anon_sym__Generic,
  [anon_sym_asm] = anon_sym_asm,
  [anon_sym___asm__] = anon_sym___asm__,
  [anon_sym___asm] = anon_sym___asm,
  [anon_sym___volatile__] = anon_sym___volatile__,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [sym_number_literal] = sym_number_literal,
  [anon_sym_L_SQUOTE] = anon_sym_L_SQUOTE,
  [anon_sym_u_SQUOTE] = anon_sym_u_SQUOTE,
  [anon_sym_U_SQUOTE] = anon_sym_U_SQUOTE,
  [anon_sym_u8_SQUOTE] = anon_sym_u8_SQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_char_literal_token1] = aux_sym_char_literal_token1,
  [anon_sym_L_DQUOTE] = anon_sym_L_DQUOTE,
  [anon_sym_u_DQUOTE] = anon_sym_u_DQUOTE,
  [anon_sym_U_DQUOTE] = anon_sym_U_DQUOTE,
  [anon_sym_u8_DQUOTE] = anon_sym_u8_DQUOTE,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_literal_token1] = aux_sym_string_literal_token1,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_system_lib_string] = sym_system_lib_string,
  [sym_true] = sym_true,
  [sym_false] = sym_false,
  [anon_sym_NULL] = anon_sym_NULL,
  [anon_sym_nullptr] = anon_sym_nullptr,
  [sym_comment] = sym_comment,
  [sym_translation_unit] = sym_translation_unit,
  [sym__top_level_item] = sym__top_level_item,
  [sym__block_item] = sym__block_item,
  [sym_preproc_include] = sym_preproc_include,
  [sym_preproc_def] = sym_preproc_def,
  [sym_preproc_function_def] = sym_preproc_function_def,
  [sym_preproc_params] = sym_preproc_params,
  [sym_preproc_call] = sym_preproc_call,
  [sym_preproc_if] = sym_preproc_if,
  [sym_preproc_ifdef] = sym_preproc_ifdef,
  [sym_preproc_else] = sym_preproc_else,
  [sym_preproc_elif] = sym_preproc_elif,
  [sym_preproc_elifdef] = sym_preproc_elifdef,
  [sym_preproc_if_in_field_declaration_list] = sym_preproc_if,
  [sym_preproc_ifdef_in_field_declaration_list] = sym_preproc_ifdef,
  [sym_preproc_else_in_field_declaration_list] = sym_preproc_else,
  [sym_preproc_elif_in_field_declaration_list] = sym_preproc_elif,
  [sym_preproc_elifdef_in_field_declaration_list] = sym_preproc_elifdef,
  [sym_preproc_if_in_enumerator_list] = sym_preproc_if,
  [sym_preproc_ifdef_in_enumerator_list] = sym_preproc_ifdef,
  [sym_preproc_else_in_enumerator_list] = sym_preproc_else,
  [sym_preproc_elif_in_enumerator_list] = sym_preproc_elif,
  [sym_preproc_elifdef_in_enumerator_list] = sym_preproc_elifdef,
  [sym_preproc_if_in_enumerator_list_no_comma] = sym_preproc_if,
  [sym_preproc_ifdef_in_enumerator_list_no_comma] = sym_preproc_ifdef,
  [sym_preproc_else_in_enumerator_list_no_comma] = sym_preproc_else,
  [sym_preproc_elif_in_enumerator_list_no_comma] = sym_preproc_elif,
  [sym_preproc_elifdef_in_enumerator_list_no_comma] = sym_preproc_elifdef,
  [sym__preproc_expression] = sym__preproc_expression,
  [sym_preproc_parenthesized_expression] = sym_parenthesized_expression,
  [sym_preproc_defined] = sym_preproc_defined,
  [sym_preproc_unary_expression] = sym_unary_expression,
  [sym_preproc_call_expression] = sym_call_expression,
  [sym_preproc_argument_list] = sym_argument_list,
  [sym_preproc_binary_expression] = sym_binary_expression,
  [sym_function_definition] = sym_function_definition,
  [sym__old_style_function_definition] = sym_function_definition,
  [sym_declaration] = sym_declaration,
  [sym_type_definition] = sym_type_definition,
  [sym__type_definition_type] = sym__type_definition_type,
  [sym__type_definition_declarators] = sym__type_definition_declarators,
  [sym__declaration_modifiers] = sym__declaration_modifiers,
  [sym__declaration_specifiers] = sym__declaration_specifiers,
  [sym_linkage_specification] = sym_linkage_specification,
  [sym_attribute_specifier] = sym_attribute_specifier,
  [sym_attribute] = sym_attribute,
  [sym_attribute_declaration] = sym_attribute_declaration,
  [sym_ms_declspec_modifier] = sym_ms_declspec_modifier,
  [sym_ms_based_modifier] = sym_ms_based_modifier,
  [sym_ms_call_modifier] = sym_ms_call_modifier,
  [sym_ms_unaligned_ptr_modifier] = sym_ms_unaligned_ptr_modifier,
  [sym_ms_pointer_modifier] = sym_ms_pointer_modifier,
  [sym_declaration_list] = sym_declaration_list,
  [sym__declarator] = sym__declarator,
  [sym__declaration_declarator] = sym__declaration_declarator,
  [sym__field_declarator] = sym__field_declarator,
  [sym__type_declarator] = sym__type_declarator,
  [sym__abstract_declarator] = sym__abstract_declarator,
  [sym_parenthesized_declarator] = sym_parenthesized_declarator,
  [sym_parenthesized_field_declarator] = sym_parenthesized_declarator,
  [sym_parenthesized_type_declarator] = sym_parenthesized_declarator,
  [sym_abstract_parenthesized_declarator] = sym_abstract_parenthesized_declarator,
  [sym_attributed_declarator] = sym_attributed_declarator,
  [sym_attributed_field_declarator] = sym_attributed_declarator,
  [sym_attributed_type_declarator] = sym_attributed_declarator,
  [sym_pointer_declarator] = sym_pointer_declarator,
  [sym_pointer_field_declarator] = sym_pointer_declarator,
  [sym_pointer_type_declarator] = sym_pointer_declarator,
  [sym_abstract_pointer_declarator] = sym_abstract_pointer_declarator,
  [sym_function_declarator] = sym_function_declarator,
  [sym__function_declaration_declarator] = sym_function_declarator,
  [sym_function_field_declarator] = sym_function_declarator,
  [sym_function_type_declarator] = sym_function_declarator,
  [sym_abstract_function_declarator] = sym_abstract_function_declarator,
  [sym__old_style_function_declarator] = sym_function_declarator,
  [sym_array_declarator] = sym_array_declarator,
  [sym_array_field_declarator] = sym_array_declarator,
  [sym_array_type_declarator] = sym_array_declarator,
  [sym_abstract_array_declarator] = sym_abstract_array_declarator,
  [sym_init_declarator] = sym_init_declarator,
  [sym_compound_statement] = sym_compound_statement,
  [sym_storage_class_specifier] = sym_storage_class_specifier,
  [sym_type_qualifier] = sym_type_qualifier,
  [sym_alignas_qualifier] = sym_alignas_qualifier,
  [sym_type_specifier] = sym_type_specifier,
  [sym_sized_type_specifier] = sym_sized_type_specifier,
  [sym_enum_specifier] = sym_enum_specifier,
  [sym_enumerator_list] = sym_enumerator_list,
  [sym_struct_specifier] = sym_struct_specifier,
  [sym_union_specifier] = sym_union_specifier,
  [sym_field_declaration_list] = sym_field_declaration_list,
  [sym__field_declaration_list_item] = sym__field_declaration_list_item,
  [sym_field_declaration] = sym_field_declaration,
  [sym__field_declaration_declarator] = sym__field_declaration_declarator,
  [sym_bitfield_clause] = sym_bitfield_clause,
  [sym_enumerator] = sym_enumerator,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_parameter_list] = sym_parameter_list,
  [sym__old_style_parameter_list] = sym_parameter_list,
  [sym_parameter_declaration] = sym_parameter_declaration,
  [sym_attributed_statement] = sym_attributed_statement,
  [sym_statement] = sym_statement,
  [sym__top_level_statement] = sym__top_level_statement,
  [sym_labeled_statement] = sym_labeled_statement,
  [sym__top_level_expression_statement] = sym_expression_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_if_statement] = sym_if_statement,
  [sym_else_clause] = sym_else_clause,
  [sym_switch_statement] = sym_switch_statement,
  [sym_case_statement] = sym_case_statement,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_for_statement] = sym_for_statement,
  [sym__for_statement_body] = sym__for_statement_body,
  [sym_return_statement] = sym_return_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_goto_statement] = sym_goto_statement,
  [sym_seh_try_statement] = sym_seh_try_statement,
  [sym_seh_except_clause] = sym_seh_except_clause,
  [sym_seh_finally_clause] = sym_seh_finally_clause,
  [sym_seh_leave_statement] = sym_seh_leave_statement,
  [sym_expression] = sym_expression,
  [sym__string] = sym__string,
  [sym_comma_expression] = sym_comma_expression,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_pointer_expression] = sym_pointer_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_update_expression] = sym_update_expression,
  [sym_cast_expression] = sym_cast_expression,
  [sym_type_descriptor] = sym_type_descriptor,
  [sym_sizeof_expression] = sym_sizeof_expression,
  [sym_alignof_expression] = sym_alignof_expression,
  [sym_offsetof_expression] = sym_offsetof_expression,
  [sym_generic_expression] = sym_generic_expression,
  [sym_subscript_expression] = sym_subscript_expression,
  [sym_call_expression] = sym_call_expression,
  [sym_gnu_asm_expression] = sym_gnu_asm_expression,
  [sym_gnu_asm_qualifier] = sym_gnu_asm_qualifier,
  [sym_gnu_asm_output_operand_list] = sym_gnu_asm_output_operand_list,
  [sym_gnu_asm_output_operand] = sym_gnu_asm_output_operand,
  [sym_gnu_asm_input_operand_list] = sym_gnu_asm_input_operand_list,
  [sym_gnu_asm_input_operand] = sym_gnu_asm_input_operand,
  [sym_gnu_asm_clobber_list] = sym_gnu_asm_clobber_list,
  [sym_gnu_asm_goto_list] = sym_gnu_asm_goto_list,
  [sym_extension_expression] = sym_extension_expression,
  [sym_argument_list] = sym_argument_list,
  [sym_field_expression] = sym_field_expression,
  [sym_compound_literal_expression] = sym_compound_literal_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_initializer_list] = sym_initializer_list,
  [sym_initializer_pair] = sym_initializer_pair,
  [sym_subscript_designator] = sym_subscript_designator,
  [sym_subscript_range_designator] = sym_subscript_range_designator,
  [sym_field_designator] = sym_field_designator,
  [sym_char_literal] = sym_char_literal,
  [sym_concatenated_string] = sym_concatenated_string,
  [sym_string_literal] = sym_string_literal,
  [sym_null] = sym_null,
  [sym__empty_declaration] = sym__empty_declaration,
  [sym_macro_type_specifier] = sym_macro_type_specifier,
  [aux_sym_translation_unit_repeat1] = aux_sym_translation_unit_repeat1,
  [aux_sym_preproc_params_repeat1] = aux_sym_preproc_params_repeat1,
  [aux_sym_preproc_if_repeat1] = aux_sym_preproc_if_repeat1,
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = aux_sym_preproc_if_in_field_declaration_list_repeat1,
  [aux_sym_preproc_if_in_enumerator_list_repeat1] = aux_sym_preproc_if_in_enumerator_list_repeat1,
  [aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1] = aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
  [aux_sym_preproc_argument_list_repeat1] = aux_sym_preproc_argument_list_repeat1,
  [aux_sym__old_style_function_definition_repeat1] = aux_sym__old_style_function_definition_repeat1,
  [aux_sym_declaration_repeat1] = aux_sym_declaration_repeat1,
  [aux_sym_type_definition_repeat1] = aux_sym_type_definition_repeat1,
  [aux_sym__type_definition_type_repeat1] = aux_sym__type_definition_type_repeat1,
  [aux_sym__type_definition_declarators_repeat1] = aux_sym__type_definition_declarators_repeat1,
  [aux_sym__declaration_specifiers_repeat1] = aux_sym__declaration_specifiers_repeat1,
  [aux_sym_attribute_declaration_repeat1] = aux_sym_attribute_declaration_repeat1,
  [aux_sym_attributed_declarator_repeat1] = aux_sym_attributed_declarator_repeat1,
  [aux_sym_pointer_declarator_repeat1] = aux_sym_pointer_declarator_repeat1,
  [aux_sym_function_declarator_repeat1] = aux_sym_function_declarator_repeat1,
  [aux_sym_array_declarator_repeat1] = aux_sym_array_declarator_repeat1,
  [aux_sym_sized_type_specifier_repeat1] = aux_sym_sized_type_specifier_repeat1,
  [aux_sym_enumerator_list_repeat1] = aux_sym_enumerator_list_repeat1,
  [aux_sym__field_declaration_declarator_repeat1] = aux_sym__field_declaration_declarator_repeat1,
  [aux_sym_parameter_list_repeat1] = aux_sym_parameter_list_repeat1,
  [aux_sym__old_style_parameter_list_repeat1] = aux_sym__old_style_parameter_list_repeat1,
  [aux_sym_case_statement_repeat1] = aux_sym_case_statement_repeat1,
  [aux_sym_generic_expression_repeat1] = aux_sym_generic_expression_repeat1,
  [aux_sym_gnu_asm_expression_repeat1] = aux_sym_gnu_asm_expression_repeat1,
  [aux_sym_gnu_asm_output_operand_list_repeat1] = aux_sym_gnu_asm_output_operand_list_repeat1,
  [aux_sym_gnu_asm_input_operand_list_repeat1] = aux_sym_gnu_asm_input_operand_list_repeat1,
  [aux_sym_gnu_asm_clobber_list_repeat1] = aux_sym_gnu_asm_clobber_list_repeat1,
  [aux_sym_gnu_asm_goto_list_repeat1] = aux_sym_gnu_asm_goto_list_repeat1,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym_initializer_list_repeat1] = aux_sym_initializer_list_repeat1,
  [aux_sym_initializer_pair_repeat1] = aux_sym_initializer_pair_repeat1,
  [aux_sym_char_literal_repeat1] = aux_sym_char_literal_repeat1,
  [aux_sym_concatenated_string_repeat1] = aux_sym_concatenated_string_repeat1,
  [aux_sym_string_literal_repeat1] = aux_sym_string_literal_repeat1,
  [alias_sym_field_identifier] = alias_sym_field_identifier,
  [alias_sym_statement_identifier] = alias_sym_statement_identifier,
  [alias_sym_type_identifier] = alias_sym_type_identifier,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_preproc_include_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_include_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_def_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_if_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LF] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_if_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_ifdef_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_ifdef_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_else_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elif_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elifdef_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_preproc_elifdef_token2] = {
    .visible = true,
    .named = false,
  },
  [sym_preproc_arg] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_directive] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_defined] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___extension__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typedef] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extern] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___attribute__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___attribute] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___declspec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___based] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___cdecl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___clrcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___stdcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___fastcall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___thiscall] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___vectorcall] = {
    .visible = true,
    .named = false,
  },
  [sym_ms_restrict_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_unsigned_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_signed_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym__unaligned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___unaligned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_signed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsigned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_long] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_short] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_auto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_register] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___inline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___inline__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___forceinline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_thread_local] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___thread] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_constexpr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_volatile] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_restrict] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___restrict__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Atomic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Noreturn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_noreturn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Nonnull] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alignas] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Alignas] = {
    .visible = true,
    .named = false,
  },
  [sym_primitive_type] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_union] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_switch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_goto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___except] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___leave] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sizeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___alignof__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Alignof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_offsetof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__Generic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_asm] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___asm__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___asm] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___volatile__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_number_literal] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_L_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_char_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_L_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_system_lib_string] = {
    .visible = true,
    .named = true,
  },
  [sym_true] = {
    .visible = true,
    .named = true,
  },
  [sym_false] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_NULL] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nullptr] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_translation_unit] = {
    .visible = true,
    .named = true,
  },
  [sym__top_level_item] = {
    .visible = false,
    .named = true,
  },
  [sym__block_item] = {
    .visible = false,
    .named = true,
  },
  [sym_preproc_include] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_def] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_function_def] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_params] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_call] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef_in_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef_in_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef_in_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_if_in_enumerator_list_no_comma] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_ifdef_in_enumerator_list_no_comma] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_else_in_enumerator_list_no_comma] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elif_in_enumerator_list_no_comma] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_elifdef_in_enumerator_list_no_comma] = {
    .visible = true,
    .named = true,
  },
  [sym__preproc_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_preproc_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_defined] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_preproc_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__type_definition_type] = {
    .visible = false,
    .named = true,
  },
  [sym__type_definition_declarators] = {
    .visible = false,
    .named = true,
  },
  [sym__declaration_modifiers] = {
    .visible = false,
    .named = true,
  },
  [sym__declaration_specifiers] = {
    .visible = false,
    .named = true,
  },
  [sym_linkage_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_declspec_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_based_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_call_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_unaligned_ptr_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ms_pointer_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__declaration_declarator] = {
    .visible = false,
    .named = true,
  },
  [sym__field_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__type_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__abstract_declarator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym__function_declaration_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_field_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_init_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_storage_class_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_alignas_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_specifier] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_sized_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_enumerator_list] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_union_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__field_declaration_list_item] = {
    .visible = false,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__field_declaration_declarator] = {
    .visible = false,
    .named = true,
  },
  [sym_bitfield_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_enumerator] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym__old_style_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_attributed_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__top_level_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__top_level_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_statement_body] = {
    .visible = false,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_seh_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_seh_except_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_seh_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_seh_leave_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__string] = {
    .visible = false,
    .named = true,
  },
  [sym_comma_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_descriptor] = {
    .visible = true,
    .named = true,
  },
  [sym_sizeof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_alignof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_offsetof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_output_operand_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_output_operand] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_input_operand_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_input_operand] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_clobber_list] = {
    .visible = true,
    .named = true,
  },
  [sym_gnu_asm_goto_list] = {
    .visible = true,
    .named = true,
  },
  [sym_extension_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_literal_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_initializer_list] = {
    .visible = true,
    .named = true,
  },
  [sym_initializer_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_range_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_field_designator] = {
    .visible = true,
    .named = true,
  },
  [sym_char_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_concatenated_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_null] = {
    .visible = true,
    .named = true,
  },
  [sym__empty_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_macro_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_translation_unit_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_params_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_field_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_enumerator_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_preproc_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__old_style_function_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_definition_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_definition_declarators_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__declaration_specifiers_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attributed_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pointer_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sized_type_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enumerator_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__field_declaration_declarator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__old_style_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_output_operand_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_input_operand_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_clobber_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_gnu_asm_goto_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_initializer_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_initializer_pair_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_char_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_concatenated_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_statement_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alternative = 1,
  field_argument = 2,
  field_arguments = 3,
  field_assembly_code = 4,
  field_body = 5,
  field_clobbers = 6,
  field_condition = 7,
  field_consequence = 8,
  field_constraint = 9,
  field_declarator = 10,
  field_designator = 11,
  field_directive = 12,
  field_end = 13,
  field_field = 14,
  field_filter = 15,
  field_function = 16,
  field_goto_labels = 17,
  field_index = 18,
  field_initializer = 19,
  field_input_operands = 20,
  field_label = 21,
  field_left = 22,
  field_member = 23,
  field_name = 24,
  field_operand = 25,
  field_operator = 26,
  field_output_operands = 27,
  field_parameters = 28,
  field_path = 29,
  field_prefix = 30,
  field_register = 31,
  field_right = 32,
  field_size = 33,
  field_start = 34,
  field_symbol = 35,
  field_type = 36,
  field_underlying_type = 37,
  field_update = 38,
  field_value = 39,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_assembly_code] = "assembly_code",
  [field_body] = "body",
  [field_clobbers] = "clobbers",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_constraint] = "constraint",
  [field_declarator] = "declarator",
  [field_designator] = "designator",
  [field_directive] = "directive",
  [field_end] = "end",
  [field_field] = "field",
  [field_filter] = "filter",
  [field_function] = "function",
  [field_goto_labels] = "goto_labels",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_input_operands] = "input_operands",
  [field_label] = "label",
  [field_left] = "left",
  [field_member] = "member",
  [field_name] = "name",
  [field_operand] = "operand",
  [field_operator] = "operator",
  [field_output_operands] = "output_operands",
  [field_parameters] = "parameters",
  [field_path] = "path",
  [field_prefix] = "prefix",
  [field_register] = "register",
  [field_right] = "right",
  [field_size] = "size",
  [field_start] = "start",
  [field_symbol] = "symbol",
  [field_type] = "type",
  [field_underlying_type] = "underlying_type",
  [field_update] = "update",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 3},
  [3] = {.index = 3, .length = 1},
  [4] = {.index = 4, .length = 1},
  [5] = {.index = 5, .length = 2},
  [6] = {.index = 7, .length = 1},
  [7] = {.index = 8, .length = 1},
  [8] = {.index = 9, .length = 1},
  [9] = {.index = 10, .length = 1},
  [10] = {.index = 3, .length = 1},
  [11] = {.index = 11, .length = 2},
  [12] = {.index = 13, .length = 2},
  [13] = {.index = 15, .length = 2},
  [14] = {.index = 17, .length = 1},
  [15] = {.index = 17, .length = 1},
  [16] = {.index = 18, .length = 1},
  [17] = {.index = 8, .length = 1},
  [18] = {.index = 19, .length = 2},
  [19] = {.index = 21, .length = 2},
  [20] = {.index = 23, .length = 1},
  [22] = {.index = 24, .length = 1},
  [23] = {.index = 25, .length = 2},
  [24] = {.index = 27, .length = 2},
  [25] = {.index = 29, .length = 1},
  [26] = {.index = 30, .length = 1},
  [27] = {.index = 31, .length = 2},
  [28] = {.index = 33, .length = 2},
  [29] = {.index = 35, .length = 1},
  [30] = {.index = 36, .length = 3},
  [31] = {.index = 39, .length = 1},
  [32] = {.index = 40, .length = 1},
  [33] = {.index = 41, .length = 3},
  [34] = {.index = 44, .length = 2},
  [35] = {.index = 46, .length = 2},
  [36] = {.index = 48, .length = 3},
  [37] = {.index = 51, .length = 1},
  [38] = {.index = 51, .length = 1},
  [39] = {.index = 52, .length = 2},
  [40] = {.index = 54, .length = 2},
  [41] = {.index = 56, .length = 1},
  [42] = {.index = 57, .length = 2},
  [43] = {.index = 59, .length = 1},
  [44] = {.index = 60, .length = 2},
  [45] = {.index = 62, .length = 2},
  [46] = {.index = 64, .length = 2},
  [47] = {.index = 66, .length = 2},
  [48] = {.index = 68, .length = 2},
  [49] = {.index = 70, .length = 2},
  [50] = {.index = 72, .length = 2},
  [51] = {.index = 74, .length = 2},
  [53] = {.index = 76, .length = 2},
  [54] = {.index = 78, .length = 1},
  [55] = {.index = 79, .length = 1},
  [56] = {.index = 80, .length = 3},
  [57] = {.index = 83, .length = 1},
  [58] = {.index = 84, .length = 1},
  [59] = {.index = 85, .length = 2},
  [60] = {.index = 87, .length = 1},
  [61] = {.index = 88, .length = 3},
  [62] = {.index = 91, .length = 3},
  [63] = {.index = 94, .length = 2},
  [64] = {.index = 96, .length = 3},
  [65] = {.index = 99, .length = 2},
  [66] = {.index = 101, .length = 5},
  [67] = {.index = 106, .length = 3},
  [68] = {.index = 109, .length = 2},
  [69] = {.index = 111, .length = 2},
  [70] = {.index = 113, .length = 3},
  [71] = {.index = 116, .length = 2},
  [72] = {.index = 118, .length = 2},
  [73] = {.index = 120, .length = 1},
  [74] = {.index = 121, .length = 2},
  [75] = {.index = 123, .length = 2},
  [76] = {.index = 125, .length = 2},
  [77] = {.index = 127, .length = 3},
  [78] = {.index = 130, .length = 2},
  [79] = {.index = 132, .length = 2},
  [80] = {.index = 134, .length = 2},
  [81] = {.index = 136, .length = 1},
  [82] = {.index = 137, .length = 2},
  [83] = {.index = 139, .length = 2},
  [84] = {.index = 141, .length = 4},
  [85] = {.index = 145, .length = 1},
  [86] = {.index = 146, .length = 2},
  [87] = {.index = 148, .length = 1},
  [88] = {.index = 149, .length = 1},
  [89] = {.index = 150, .length = 4},
  [90] = {.index = 154, .length = 4},
  [91] = {.index = 158, .length = 2},
  [92] = {.index = 160, .length = 2},
  [93] = {.index = 162, .length = 3},
  [94] = {.index = 165, .length = 5},
  [95] = {.index = 170, .length = 3},
  [96] = {.index = 173, .length = 2},
  [97] = {.index = 175, .length = 1},
  [99] = {.index = 176, .length = 2},
  [100] = {.index = 178, .length = 2},
  [101] = {.index = 180, .length = 2},
  [102] = {.index = 182, .length = 3},
  [103] = {.index = 185, .length = 2},
  [104] = {.index = 187, .length = 2},
  [105] = {.index = 189, .length = 2},
  [106] = {.index = 191, .length = 2},
  [107] = {.index = 193, .length = 3},
  [108] = {.index = 196, .length = 2},
  [109] = {.index = 198, .length = 1},
  [110] = {.index = 199, .length = 5},
  [111] = {.index = 204, .length = 2},
  [112] = {.index = 206, .length = 3},
  [113] = {.index = 209, .length = 2},
  [114] = {.index = 209, .length = 2},
  [115] = {.index = 211, .length = 3},
  [116] = {.index = 214, .length = 2},
  [117] = {.index = 216, .length = 1},
  [118] = {.index = 217, .length = 4},
  [119] = {.index = 221, .length = 3},
  [120] = {.index = 224, .length = 2},
  [121] = {.index = 226, .length = 2},
  [122] = {.index = 35, .length = 1},
  [123] = {.index = 228, .length = 5},
  [124] = {.index = 233, .length = 4},
  [125] = {.index = 237, .length = 2},
  [126] = {.index = 239, .length = 2},
  [127] = {.index = 241, .length = 2},
  [128] = {.index = 243, .length = 5},
  [129] = {.index = 248, .length = 2},
  [130] = {.index = 250, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_body, 0, .inherited = true},
    {field_declarator, 0, .inherited = true},
    {field_type, 0, .inherited = true},
  [3] =
    {field_type, 0},
  [4] =
    {field_directive, 0},
  [5] =
    {field_argument, 1},
    {field_operator, 0},
  [7] =
    {field_name, 0},
  [8] =
    {field_name, 1},
  [9] =
    {field_body, 1},
  [10] =
    {field_value, 1},
  [11] =
    {field_declarator, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
  [13] =
    {field_argument, 0},
    {field_operator, 1},
  [15] =
    {field_arguments, 1},
    {field_function, 0},
  [17] =
    {field_type, 1},
  [18] =
    {field_path, 1},
  [19] =
    {field_argument, 1},
    {field_directive, 0},
  [21] =
    {field_declarator, 1},
    {field_type, 0},
  [23] =
    {field_parameters, 0},
  [24] =
    {field_declarator, 0},
  [25] =
    {field_body, 2},
    {field_value, 1},
  [27] =
    {field_body, 2},
    {field_name, 1},
  [29] =
    {field_name, 2},
  [30] =
    {field_body, 2},
  [31] =
    {field_condition, 1},
    {field_consequence, 2},
  [33] =
    {field_body, 2},
    {field_condition, 1},
  [35] =
    {field_label, 1},
  [36] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [39] =
    {field_label, 0},
  [40] =
    {field_declarator, 1},
  [41] =
    {field_body, 2},
    {field_declarator, 1},
    {field_type, 0, .inherited = true},
  [44] =
    {field_declarator, 0},
    {field_parameters, 1},
  [46] =
    {field_declarator, 1},
    {field_type, 0, .inherited = true},
  [48] =
    {field_argument, 0},
    {field_field, 2},
    {field_operator, 1},
  [51] =
    {field_type, 2},
  [52] =
    {field_name, 1},
    {field_value, 2},
  [54] =
    {field_name, 1},
    {field_parameters, 2},
  [56] =
    {field_condition, 1},
  [57] =
    {field_alternative, 2},
    {field_name, 1},
  [59] =
    {field_type, 0, .inherited = true},
  [60] =
    {field_declarator, 2},
    {field_type, 0},
  [62] =
    {field_left, 0},
    {field_right, 2},
  [64] =
    {field_type, 1},
    {field_value, 3},
  [66] =
    {field_declarator, 2},
    {field_type, 1},
  [68] =
    {field_declarator, 2, .inherited = true},
    {field_type, 1, .inherited = true},
  [70] =
    {field_declarator, 0},
    {field_declarator, 1, .inherited = true},
  [72] =
    {field_name, 2},
    {field_prefix, 0},
  [74] =
    {field_name, 1},
    {field_underlying_type, 3},
  [76] =
    {field_body, 3},
    {field_name, 2},
  [78] =
    {field_name, 3},
  [79] =
    {field_body, 3},
  [80] =
    {field_alternative, 3},
    {field_condition, 1},
    {field_consequence, 2},
  [83] =
    {field_initializer, 0},
  [84] =
    {field_assembly_code, 2},
  [85] =
    {field_name, 0},
    {field_type, 2},
  [87] =
    {field_declarator, 2},
  [88] =
    {field_body, 3},
    {field_declarator, 2},
    {field_type, 0, .inherited = true},
  [91] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_type, 0, .inherited = true},
  [94] =
    {field_declarator, 0},
    {field_value, 2},
  [96] =
    {field_declarator, 1},
    {field_declarator, 2, .inherited = true},
    {field_type, 0, .inherited = true},
  [99] =
    {field_declarator, 0, .inherited = true},
    {field_declarator, 1, .inherited = true},
  [101] =
    {field_body, 3},
    {field_declarator, 1},
    {field_declarator, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [106] =
    {field_body, 3},
    {field_declarator, 2},
    {field_type, 1, .inherited = true},
  [109] =
    {field_argument, 0},
    {field_index, 2},
  [111] =
    {field_alternative, 3},
    {field_condition, 0},
  [113] =
    {field_name, 1},
    {field_parameters, 2},
    {field_value, 3},
  [116] =
    {field_alternative, 3},
    {field_condition, 1},
  [118] =
    {field_alternative, 3},
    {field_name, 1},
  [120] =
    {field_size, 1},
  [121] =
    {field_declarator, 3},
    {field_type, 1},
  [123] =
    {field_declarator, 3, .inherited = true},
    {field_type, 2, .inherited = true},
  [125] =
    {field_name, 0},
    {field_value, 2},
  [127] =
    {field_body, 4},
    {field_name, 1},
    {field_underlying_type, 3},
  [130] =
    {field_declarator, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [132] =
    {field_body, 4},
    {field_name, 3},
  [134] =
    {field_body, 1},
    {field_condition, 3},
  [136] =
    {field_update, 2},
  [137] =
    {field_initializer, 0},
    {field_update, 2},
  [139] =
    {field_condition, 1},
    {field_initializer, 0},
  [141] =
    {field_body, 4},
    {field_condition, 2, .inherited = true},
    {field_initializer, 2, .inherited = true},
    {field_update, 2, .inherited = true},
  [145] =
    {field_operand, 1},
  [146] =
    {field_assembly_code, 2},
    {field_output_operands, 3},
  [148] =
    {field_assembly_code, 3},
  [149] =
    {field_declarator, 3},
  [150] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_declarator, 3},
    {field_type, 0, .inherited = true},
  [154] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_declarator, 3, .inherited = true},
    {field_type, 0, .inherited = true},
  [158] =
    {field_declarator, 0},
    {field_size, 2},
  [160] =
    {field_declarator, 1},
    {field_declarator, 2},
  [162] =
    {field_body, 4},
    {field_declarator, 3},
    {field_type, 1, .inherited = true},
  [165] =
    {field_body, 4},
    {field_declarator, 2},
    {field_declarator, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_type, 1, .inherited = true},
  [170] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequence, 2},
  [173] =
    {field_alternative, 4},
    {field_condition, 1},
  [175] =
    {field_size, 2},
  [176] =
    {field_body, 2},
    {field_filter, 1},
  [178] =
    {field_declarator, 0},
    {field_declarator, 2, .inherited = true},
  [180] =
    {field_condition, 1},
    {field_update, 3},
  [182] =
    {field_condition, 1},
    {field_initializer, 0},
    {field_update, 3},
  [185] =
    {field_initializer, 0},
    {field_update, 3},
  [187] =
    {field_condition, 2},
    {field_initializer, 0},
  [189] =
    {field_member, 4},
    {field_type, 2},
  [191] =
    {field_operand, 1},
    {field_operand, 2, .inherited = true},
  [193] =
    {field_assembly_code, 2},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [196] =
    {field_assembly_code, 3},
    {field_output_operands, 4},
  [198] =
    {field_declarator, 4},
  [199] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_declarator, 3},
    {field_declarator, 4, .inherited = true},
    {field_type, 0, .inherited = true},
  [204] =
    {field_declarator, 0},
    {field_size, 3},
  [206] =
    {field_declarator, 1},
    {field_declarator, 2},
    {field_declarator, 3},
  [209] =
    {field_designator, 0},
    {field_value, 2},
  [211] =
    {field_condition, 2},
    {field_initializer, 0},
    {field_update, 4},
  [214] =
    {field_operand, 0, .inherited = true},
    {field_operand, 1, .inherited = true},
  [216] =
    {field_register, 1},
  [217] =
    {field_assembly_code, 2},
    {field_clobbers, 5},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [221] =
    {field_assembly_code, 3},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [224] =
    {field_constraint, 0},
    {field_value, 2},
  [226] =
    {field_register, 1},
    {field_register, 2, .inherited = true},
  [228] =
    {field_assembly_code, 2},
    {field_clobbers, 5},
    {field_goto_labels, 6},
    {field_input_operands, 4},
    {field_output_operands, 3},
  [233] =
    {field_assembly_code, 3},
    {field_clobbers, 6},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [237] =
    {field_end, 3},
    {field_start, 1},
  [239] =
    {field_register, 0, .inherited = true},
    {field_register, 1, .inherited = true},
  [241] =
    {field_label, 1},
    {field_label, 2, .inherited = true},
  [243] =
    {field_assembly_code, 3},
    {field_clobbers, 6},
    {field_goto_labels, 7},
    {field_input_operands, 5},
    {field_output_operands, 4},
  [248] =
    {field_label, 0, .inherited = true},
    {field_label, 1, .inherited = true},
  [250] =
    {field_constraint, 3},
    {field_symbol, 1},
    {field_value, 5},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_type_identifier,
  },
  [7] = {
    [1] = alias_sym_type_identifier,
  },
  [10] = {
    [0] = alias_sym_type_identifier,
  },
  [15] = {
    [1] = alias_sym_type_identifier,
  },
  [21] = {
    [0] = sym_primitive_type,
  },
  [24] = {
    [1] = alias_sym_type_identifier,
  },
  [25] = {
    [2] = alias_sym_type_identifier,
  },
  [29] = {
    [1] = alias_sym_statement_identifier,
  },
  [31] = {
    [0] = alias_sym_statement_identifier,
  },
  [36] = {
    [2] = alias_sym_field_identifier,
  },
  [38] = {
    [2] = alias_sym_type_identifier,
  },
  [51] = {
    [1] = alias_sym_type_identifier,
  },
  [52] = {
    [0] = alias_sym_field_identifier,
  },
  [53] = {
    [2] = alias_sym_type_identifier,
  },
  [54] = {
    [3] = alias_sym_type_identifier,
  },
  [77] = {
    [1] = alias_sym_type_identifier,
  },
  [79] = {
    [3] = alias_sym_type_identifier,
  },
  [98] = {
    [1] = alias_sym_field_identifier,
  },
  [105] = {
    [4] = alias_sym_field_identifier,
  },
  [113] = {
    [0] = alias_sym_field_identifier,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 2,
  [11] = 4,
  [12] = 3,
  [13] = 5,
  [14] = 2,
  [15] = 4,
  [16] = 3,
  [17] = 5,
  [18] = 2,
  [19] = 4,
  [20] = 3,
  [21] = 5,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 22,
  [27] = 23,
  [28] = 25,
  [29] = 29,
  [30] = 29,
  [31] = 22,
  [32] = 32,
  [33] = 23,
  [34] = 25,
  [35] = 29,
  [36] = 36,
  [37] = 32,
  [38] = 23,
  [39] = 25,
  [40] = 32,
  [41] = 29,
  [42] = 32,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 46,
  [51] = 45,
  [52] = 49,
  [53] = 45,
  [54] = 47,
  [55] = 48,
  [56] = 46,
  [57] = 47,
  [58] = 45,
  [59] = 46,
  [60] = 49,
  [61] = 49,
  [62] = 48,
  [63] = 48,
  [64] = 47,
  [65] = 65,
  [66] = 65,
  [67] = 46,
  [68] = 65,
  [69] = 45,
  [70] = 65,
  [71] = 47,
  [72] = 65,
  [73] = 49,
  [74] = 48,
  [75] = 75,
  [76] = 75,
  [77] = 75,
  [78] = 75,
  [79] = 75,
  [80] = 80,
  [81] = 81,
  [82] = 82,
  [83] = 83,
  [84] = 84,
  [85] = 85,
  [86] = 86,
  [87] = 82,
  [88] = 88,
  [89] = 88,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 122,
  [123] = 123,
  [124] = 124,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 138,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 143,
  [144] = 144,
  [145] = 145,
  [146] = 80,
  [147] = 80,
  [148] = 80,
  [149] = 90,
  [150] = 83,
  [151] = 88,
  [152] = 97,
  [153] = 84,
  [154] = 81,
  [155] = 85,
  [156] = 91,
  [157] = 110,
  [158] = 92,
  [159] = 93,
  [160] = 111,
  [161] = 108,
  [162] = 112,
  [163] = 109,
  [164] = 105,
  [165] = 108,
  [166] = 90,
  [167] = 114,
  [168] = 115,
  [169] = 117,
  [170] = 113,
  [171] = 109,
  [172] = 86,
  [173] = 82,
  [174] = 118,
  [175] = 94,
  [176] = 83,
  [177] = 88,
  [178] = 81,
  [179] = 86,
  [180] = 109,
  [181] = 105,
  [182] = 108,
  [183] = 114,
  [184] = 115,
  [185] = 117,
  [186] = 82,
  [187] = 95,
  [188] = 86,
  [189] = 98,
  [190] = 99,
  [191] = 105,
  [192] = 118,
  [193] = 98,
  [194] = 99,
  [195] = 104,
  [196] = 106,
  [197] = 107,
  [198] = 90,
  [199] = 118,
  [200] = 116,
  [201] = 97,
  [202] = 84,
  [203] = 85,
  [204] = 91,
  [205] = 116,
  [206] = 96,
  [207] = 100,
  [208] = 101,
  [209] = 103,
  [210] = 110,
  [211] = 111,
  [212] = 112,
  [213] = 113,
  [214] = 93,
  [215] = 94,
  [216] = 95,
  [217] = 102,
  [218] = 98,
  [219] = 99,
  [220] = 101,
  [221] = 103,
  [222] = 104,
  [223] = 106,
  [224] = 107,
  [225] = 101,
  [226] = 103,
  [227] = 110,
  [228] = 111,
  [229] = 112,
  [230] = 113,
  [231] = 116,
  [232] = 97,
  [233] = 84,
  [234] = 85,
  [235] = 91,
  [236] = 92,
  [237] = 93,
  [238] = 94,
  [239] = 95,
  [240] = 96,
  [241] = 100,
  [242] = 102,
  [243] = 96,
  [244] = 81,
  [245] = 83,
  [246] = 115,
  [247] = 104,
  [248] = 117,
  [249] = 106,
  [250] = 107,
  [251] = 100,
  [252] = 114,
  [253] = 102,
  [254] = 92,
  [255] = 121,
  [256] = 145,
  [257] = 143,
  [258] = 120,
  [259] = 121,
  [260] = 122,
  [261] = 124,
  [262] = 125,
  [263] = 126,
  [264] = 123,
  [265] = 128,
  [266] = 127,
  [267] = 119,
  [268] = 131,
  [269] = 137,
  [270] = 270,
  [271] = 129,
  [272] = 130,
  [273] = 132,
  [274] = 133,
  [275] = 134,
  [276] = 135,
  [277] = 136,
  [278] = 138,
  [279] = 139,
  [280] = 140,
  [281] = 141,
  [282] = 131,
  [283] = 142,
  [284] = 144,
  [285] = 145,
  [286] = 120,
  [287] = 122,
  [288] = 124,
  [289] = 123,
  [290] = 128,
  [291] = 127,
  [292] = 119,
  [293] = 129,
  [294] = 130,
  [295] = 132,
  [296] = 134,
  [297] = 135,
  [298] = 136,
  [299] = 138,
  [300] = 139,
  [301] = 140,
  [302] = 141,
  [303] = 142,
  [304] = 143,
  [305] = 121,
  [306] = 125,
  [307] = 126,
  [308] = 133,
  [309] = 142,
  [310] = 144,
  [311] = 137,
  [312] = 270,
  [313] = 313,
  [314] = 313,
  [315] = 122,
  [316] = 316,
  [317] = 128,
  [318] = 127,
  [319] = 119,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 129,
  [325] = 130,
  [326] = 132,
  [327] = 133,
  [328] = 134,
  [329] = 135,
  [330] = 144,
  [331] = 136,
  [332] = 138,
  [333] = 323,
  [334] = 139,
  [335] = 140,
  [336] = 141,
  [337] = 313,
  [338] = 338,
  [339] = 124,
  [340] = 340,
  [341] = 125,
  [342] = 342,
  [343] = 340,
  [344] = 316,
  [345] = 338,
  [346] = 321,
  [347] = 323,
  [348] = 340,
  [349] = 316,
  [350] = 338,
  [351] = 126,
  [352] = 321,
  [353] = 323,
  [354] = 123,
  [355] = 342,
  [356] = 356,
  [357] = 340,
  [358] = 316,
  [359] = 338,
  [360] = 338,
  [361] = 145,
  [362] = 340,
  [363] = 143,
  [364] = 321,
  [365] = 323,
  [366] = 316,
  [367] = 342,
  [368] = 120,
  [369] = 342,
  [370] = 370,
  [371] = 342,
  [372] = 313,
  [373] = 321,
  [374] = 374,
  [375] = 374,
  [376] = 374,
  [377] = 374,
  [378] = 378,
  [379] = 378,
  [380] = 380,
  [381] = 378,
  [382] = 382,
  [383] = 270,
  [384] = 378,
  [385] = 378,
  [386] = 382,
  [387] = 382,
  [388] = 382,
  [389] = 382,
  [390] = 378,
  [391] = 382,
  [392] = 392,
  [393] = 392,
  [394] = 394,
  [395] = 395,
  [396] = 270,
  [397] = 270,
  [398] = 80,
  [399] = 399,
  [400] = 400,
  [401] = 400,
  [402] = 400,
  [403] = 400,
  [404] = 400,
  [405] = 400,
  [406] = 400,
  [407] = 400,
  [408] = 408,
  [409] = 270,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 380,
  [423] = 423,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 429,
  [431] = 426,
  [432] = 432,
  [433] = 427,
  [434] = 428,
  [435] = 435,
  [436] = 426,
  [437] = 427,
  [438] = 428,
  [439] = 429,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 443,
  [444] = 444,
  [445] = 444,
  [446] = 446,
  [447] = 447,
  [448] = 448,
  [449] = 444,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 452,
  [454] = 451,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 456,
  [463] = 463,
  [464] = 458,
  [465] = 465,
  [466] = 456,
  [467] = 467,
  [468] = 458,
  [469] = 469,
  [470] = 470,
  [471] = 470,
  [472] = 456,
  [473] = 456,
  [474] = 470,
  [475] = 458,
  [476] = 476,
  [477] = 477,
  [478] = 470,
  [479] = 479,
  [480] = 470,
  [481] = 481,
  [482] = 470,
  [483] = 456,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 484,
  [489] = 489,
  [490] = 490,
  [491] = 484,
  [492] = 492,
  [493] = 484,
  [494] = 494,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 499,
  [500] = 500,
  [501] = 501,
  [502] = 501,
  [503] = 503,
  [504] = 504,
  [505] = 505,
  [506] = 506,
  [507] = 505,
  [508] = 505,
  [509] = 501,
  [510] = 504,
  [511] = 511,
  [512] = 512,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 499,
  [518] = 518,
  [519] = 519,
  [520] = 506,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 506,
  [525] = 501,
  [526] = 521,
  [527] = 410,
  [528] = 528,
  [529] = 505,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 512,
  [534] = 534,
  [535] = 500,
  [536] = 536,
  [537] = 521,
  [538] = 536,
  [539] = 513,
  [540] = 534,
  [541] = 532,
  [542] = 542,
  [543] = 514,
  [544] = 515,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 513,
  [549] = 516,
  [550] = 532,
  [551] = 519,
  [552] = 534,
  [553] = 500,
  [554] = 554,
  [555] = 554,
  [556] = 536,
  [557] = 534,
  [558] = 500,
  [559] = 559,
  [560] = 512,
  [561] = 513,
  [562] = 562,
  [563] = 536,
  [564] = 522,
  [565] = 514,
  [566] = 559,
  [567] = 512,
  [568] = 513,
  [569] = 514,
  [570] = 515,
  [571] = 516,
  [572] = 499,
  [573] = 518,
  [574] = 519,
  [575] = 506,
  [576] = 521,
  [577] = 499,
  [578] = 501,
  [579] = 505,
  [580] = 512,
  [581] = 514,
  [582] = 515,
  [583] = 516,
  [584] = 499,
  [585] = 518,
  [586] = 519,
  [587] = 506,
  [588] = 521,
  [589] = 501,
  [590] = 505,
  [591] = 532,
  [592] = 534,
  [593] = 500,
  [594] = 536,
  [595] = 518,
  [596] = 559,
  [597] = 522,
  [598] = 515,
  [599] = 522,
  [600] = 512,
  [601] = 522,
  [602] = 513,
  [603] = 514,
  [604] = 532,
  [605] = 554,
  [606] = 534,
  [607] = 500,
  [608] = 516,
  [609] = 536,
  [610] = 515,
  [611] = 516,
  [612] = 499,
  [613] = 518,
  [614] = 518,
  [615] = 519,
  [616] = 506,
  [617] = 521,
  [618] = 519,
  [619] = 532,
  [620] = 620,
  [621] = 621,
  [622] = 622,
  [623] = 623,
  [624] = 624,
  [625] = 625,
  [626] = 626,
  [627] = 627,
  [628] = 628,
  [629] = 629,
  [630] = 630,
  [631] = 631,
  [632] = 632,
  [633] = 633,
  [634] = 634,
  [635] = 635,
  [636] = 636,
  [637] = 637,
  [638] = 638,
  [639] = 639,
  [640] = 640,
  [641] = 640,
  [642] = 639,
  [643] = 639,
  [644] = 640,
  [645] = 640,
  [646] = 639,
  [647] = 647,
  [648] = 648,
  [649] = 649,
  [650] = 650,
  [651] = 651,
  [652] = 651,
  [653] = 653,
  [654] = 654,
  [655] = 654,
  [656] = 395,
  [657] = 651,
  [658] = 654,
  [659] = 394,
  [660] = 654,
  [661] = 651,
  [662] = 662,
  [663] = 663,
  [664] = 653,
  [665] = 665,
  [666] = 666,
  [667] = 667,
  [668] = 668,
  [669] = 669,
  [670] = 670,
  [671] = 671,
  [672] = 672,
  [673] = 673,
  [674] = 674,
  [675] = 675,
  [676] = 676,
  [677] = 677,
  [678] = 678,
  [679] = 679,
  [680] = 680,
  [681] = 681,
  [682] = 682,
  [683] = 666,
  [684] = 684,
  [685] = 676,
  [686] = 666,
  [687] = 676,
  [688] = 666,
  [689] = 676,
  [690] = 690,
  [691] = 691,
  [692] = 692,
  [693] = 693,
  [694] = 694,
  [695] = 695,
  [696] = 696,
  [697] = 697,
  [698] = 698,
  [699] = 699,
  [700] = 700,
  [701] = 701,
  [702] = 702,
  [703] = 703,
  [704] = 704,
  [705] = 705,
  [706] = 706,
  [707] = 694,
  [708] = 708,
  [709] = 709,
  [710] = 710,
  [711] = 700,
  [712] = 712,
  [713] = 663,
  [714] = 625,
  [715] = 625,
  [716] = 716,
  [717] = 394,
  [718] = 395,
  [719] = 719,
  [720] = 720,
  [721] = 721,
  [722] = 722,
  [723] = 723,
  [724] = 724,
  [725] = 725,
  [726] = 726,
  [727] = 727,
  [728] = 728,
  [729] = 729,
  [730] = 730,
  [731] = 731,
  [732] = 732,
  [733] = 733,
  [734] = 734,
  [735] = 735,
  [736] = 736,
  [737] = 737,
  [738] = 738,
  [739] = 739,
  [740] = 740,
  [741] = 741,
  [742] = 742,
  [743] = 743,
  [744] = 744,
  [745] = 745,
  [746] = 746,
  [747] = 747,
  [748] = 748,
  [749] = 749,
  [750] = 750,
  [751] = 751,
  [752] = 752,
  [753] = 753,
  [754] = 754,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 758,
  [759] = 759,
  [760] = 760,
  [761] = 761,
  [762] = 762,
  [763] = 763,
  [764] = 764,
  [765] = 765,
  [766] = 766,
  [767] = 767,
  [768] = 768,
  [769] = 769,
  [770] = 770,
  [771] = 771,
  [772] = 772,
  [773] = 773,
  [774] = 774,
  [775] = 775,
  [776] = 776,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 780,
  [781] = 781,
  [782] = 782,
  [783] = 778,
  [784] = 784,
  [785] = 785,
  [786] = 778,
  [787] = 142,
  [788] = 788,
  [789] = 782,
  [790] = 778,
  [791] = 143,
  [792] = 121,
  [793] = 778,
  [794] = 794,
  [795] = 125,
  [796] = 782,
  [797] = 126,
  [798] = 133,
  [799] = 799,
  [800] = 800,
  [801] = 782,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 811,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 806,
  [816] = 809,
  [817] = 817,
  [818] = 810,
  [819] = 811,
  [820] = 820,
  [821] = 821,
  [822] = 812,
  [823] = 807,
  [824] = 824,
  [825] = 817,
  [826] = 805,
  [827] = 625,
  [828] = 828,
  [829] = 813,
  [830] = 830,
  [831] = 808,
  [832] = 828,
  [833] = 830,
  [834] = 814,
  [835] = 675,
  [836] = 836,
  [837] = 675,
  [838] = 788,
  [839] = 636,
  [840] = 784,
  [841] = 627,
  [842] = 777,
  [843] = 775,
  [844] = 779,
  [845] = 780,
  [846] = 781,
  [847] = 800,
  [848] = 848,
  [849] = 802,
  [850] = 803,
  [851] = 776,
  [852] = 777,
  [853] = 804,
  [854] = 785,
  [855] = 799,
  [856] = 775,
  [857] = 779,
  [858] = 780,
  [859] = 781,
  [860] = 142,
  [861] = 143,
  [862] = 802,
  [863] = 125,
  [864] = 126,
  [865] = 133,
  [866] = 142,
  [867] = 143,
  [868] = 121,
  [869] = 125,
  [870] = 126,
  [871] = 133,
  [872] = 800,
  [873] = 873,
  [874] = 874,
  [875] = 804,
  [876] = 785,
  [877] = 873,
  [878] = 799,
  [879] = 803,
  [880] = 776,
  [881] = 873,
  [882] = 873,
  [883] = 883,
  [884] = 121,
  [885] = 702,
  [886] = 886,
  [887] = 887,
  [888] = 697,
  [889] = 699,
  [890] = 890,
  [891] = 828,
  [892] = 830,
  [893] = 805,
  [894] = 807,
  [895] = 808,
  [896] = 809,
  [897] = 810,
  [898] = 811,
  [899] = 812,
  [900] = 813,
  [901] = 814,
  [902] = 817,
  [903] = 903,
  [904] = 904,
  [905] = 691,
  [906] = 692,
  [907] = 701,
  [908] = 908,
  [909] = 806,
  [910] = 693,
  [911] = 675,
  [912] = 912,
  [913] = 102,
  [914] = 95,
  [915] = 103,
  [916] = 698,
  [917] = 696,
  [918] = 110,
  [919] = 101,
  [920] = 111,
  [921] = 94,
  [922] = 112,
  [923] = 113,
  [924] = 924,
  [925] = 93,
  [926] = 926,
  [927] = 927,
  [928] = 928,
  [929] = 111,
  [930] = 112,
  [931] = 113,
  [932] = 101,
  [933] = 103,
  [934] = 94,
  [935] = 95,
  [936] = 110,
  [937] = 102,
  [938] = 93,
  [939] = 939,
  [940] = 940,
  [941] = 941,
  [942] = 942,
  [943] = 943,
  [944] = 944,
  [945] = 945,
  [946] = 927,
  [947] = 926,
  [948] = 948,
  [949] = 949,
  [950] = 950,
  [951] = 951,
  [952] = 952,
  [953] = 953,
  [954] = 950,
  [955] = 952,
  [956] = 948,
  [957] = 949,
  [958] = 958,
  [959] = 805,
  [960] = 810,
  [961] = 812,
  [962] = 813,
  [963] = 828,
  [964] = 830,
  [965] = 805,
  [966] = 806,
  [967] = 807,
  [968] = 808,
  [969] = 809,
  [970] = 810,
  [971] = 811,
  [972] = 812,
  [973] = 813,
  [974] = 814,
  [975] = 830,
  [976] = 806,
  [977] = 807,
  [978] = 814,
  [979] = 808,
  [980] = 817,
  [981] = 817,
  [982] = 828,
  [983] = 983,
  [984] = 809,
  [985] = 811,
  [986] = 986,
  [987] = 987,
  [988] = 988,
  [989] = 989,
  [990] = 990,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 993,
  [995] = 995,
  [996] = 996,
  [997] = 992,
  [998] = 998,
  [999] = 993,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 992,
  [1003] = 993,
  [1004] = 992,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1006,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 836,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 824,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 1017,
  [1019] = 1019,
  [1020] = 813,
  [1021] = 874,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 814,
  [1025] = 1025,
  [1026] = 704,
  [1027] = 1027,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 1031,
  [1032] = 1032,
  [1033] = 1031,
  [1034] = 1034,
  [1035] = 1028,
  [1036] = 817,
  [1037] = 1037,
  [1038] = 1038,
  [1039] = 1039,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 1043,
  [1044] = 712,
  [1045] = 1019,
  [1046] = 1046,
  [1047] = 1019,
  [1048] = 1048,
  [1049] = 1049,
  [1050] = 848,
  [1051] = 828,
  [1052] = 1029,
  [1053] = 983,
  [1054] = 830,
  [1055] = 1028,
  [1056] = 1028,
  [1057] = 1029,
  [1058] = 805,
  [1059] = 806,
  [1060] = 807,
  [1061] = 808,
  [1062] = 809,
  [1063] = 810,
  [1064] = 811,
  [1065] = 1019,
  [1066] = 812,
  [1067] = 1019,
  [1068] = 883,
  [1069] = 1019,
  [1070] = 1029,
  [1071] = 1071,
  [1072] = 1072,
  [1073] = 1073,
  [1074] = 1074,
  [1075] = 1075,
  [1076] = 1072,
  [1077] = 1077,
  [1078] = 1078,
  [1079] = 1079,
  [1080] = 1080,
  [1081] = 1081,
  [1082] = 1072,
  [1083] = 1083,
  [1084] = 1080,
  [1085] = 1072,
  [1086] = 1086,
  [1087] = 1087,
  [1088] = 1072,
  [1089] = 1089,
  [1090] = 1090,
  [1091] = 1091,
  [1092] = 1092,
  [1093] = 1093,
  [1094] = 1094,
  [1095] = 1095,
  [1096] = 1096,
  [1097] = 1097,
  [1098] = 1098,
  [1099] = 1099,
  [1100] = 1100,
  [1101] = 1101,
  [1102] = 1102,
  [1103] = 1103,
  [1104] = 1104,
  [1105] = 1105,
  [1106] = 1106,
  [1107] = 1102,
  [1108] = 1108,
  [1109] = 1104,
  [1110] = 1110,
  [1111] = 1105,
  [1112] = 712,
  [1113] = 750,
  [1114] = 704,
  [1115] = 1115,
  [1116] = 1116,
  [1117] = 1117,
  [1118] = 1118,
  [1119] = 1119,
  [1120] = 1120,
  [1121] = 1121,
  [1122] = 764,
  [1123] = 767,
  [1124] = 741,
  [1125] = 765,
  [1126] = 742,
  [1127] = 743,
  [1128] = 1128,
  [1129] = 744,
  [1130] = 745,
  [1131] = 736,
  [1132] = 1128,
  [1133] = 1128,
  [1134] = 762,
  [1135] = 737,
  [1136] = 750,
  [1137] = 754,
  [1138] = 739,
  [1139] = 1128,
  [1140] = 1140,
  [1141] = 1121,
  [1142] = 1140,
  [1143] = 1143,
  [1144] = 1143,
  [1145] = 1143,
  [1146] = 1146,
  [1147] = 1140,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 1146,
  [1151] = 1140,
  [1152] = 1149,
  [1153] = 1149,
  [1154] = 1143,
  [1155] = 1140,
  [1156] = 1140,
  [1157] = 1146,
  [1158] = 824,
  [1159] = 1159,
  [1160] = 1160,
  [1161] = 1159,
  [1162] = 1162,
  [1163] = 1163,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1167,
  [1168] = 1168,
  [1169] = 1169,
  [1170] = 1168,
  [1171] = 1169,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1175,
  [1176] = 1176,
  [1177] = 1177,
  [1178] = 1164,
  [1179] = 1179,
  [1180] = 1160,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 1183,
  [1185] = 1185,
  [1186] = 1186,
  [1187] = 1174,
  [1188] = 1188,
  [1189] = 1173,
  [1190] = 1190,
  [1191] = 1191,
  [1192] = 1190,
  [1193] = 1193,
  [1194] = 1163,
  [1195] = 1195,
  [1196] = 1196,
  [1197] = 1159,
  [1198] = 1188,
  [1199] = 1186,
  [1200] = 1186,
  [1201] = 1201,
  [1202] = 1162,
  [1203] = 1203,
  [1204] = 1204,
  [1205] = 1186,
  [1206] = 1172,
  [1207] = 1207,
  [1208] = 1208,
  [1209] = 1209,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 690,
  [1213] = 1213,
  [1214] = 1214,
  [1215] = 1181,
  [1216] = 1196,
  [1217] = 1201,
  [1218] = 1207,
  [1219] = 1209,
  [1220] = 1210,
  [1221] = 1211,
  [1222] = 1191,
  [1223] = 1204,
  [1224] = 1179,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1193,
  [1228] = 1228,
  [1229] = 1229,
  [1230] = 991,
  [1231] = 1231,
  [1232] = 1214,
  [1233] = 1165,
  [1234] = 1234,
  [1235] = 1235,
  [1236] = 690,
  [1237] = 986,
  [1238] = 1208,
  [1239] = 1001,
  [1240] = 1240,
  [1241] = 1234,
  [1242] = 998,
  [1243] = 1240,
  [1244] = 1244,
  [1245] = 1245,
  [1246] = 1228,
  [1247] = 1234,
  [1248] = 1240,
  [1249] = 1228,
  [1250] = 1234,
  [1251] = 1240,
  [1252] = 1203,
  [1253] = 1253,
  [1254] = 1254,
  [1255] = 1255,
  [1256] = 1256,
  [1257] = 1257,
  [1258] = 1253,
  [1259] = 1253,
  [1260] = 1260,
  [1261] = 1253,
  [1262] = 1262,
  [1263] = 1263,
  [1264] = 1264,
  [1265] = 1265,
  [1266] = 1265,
  [1267] = 1267,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1267,
  [1271] = 1267,
  [1272] = 1267,
  [1273] = 1265,
  [1274] = 1265,
  [1275] = 1275,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1275,
  [1279] = 1275,
  [1280] = 1280,
  [1281] = 1254,
  [1282] = 1282,
  [1283] = 1283,
  [1284] = 1284,
  [1285] = 1285,
  [1286] = 1286,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1290,
  [1291] = 1291,
  [1292] = 1292,
  [1293] = 1293,
  [1294] = 1294,
  [1295] = 1295,
  [1296] = 1296,
  [1297] = 1297,
  [1298] = 1295,
  [1299] = 1299,
  [1300] = 1300,
  [1301] = 1301,
  [1302] = 1302,
  [1303] = 1303,
  [1304] = 1304,
  [1305] = 1305,
  [1306] = 1306,
  [1307] = 1295,
  [1308] = 1308,
  [1309] = 1309,
  [1310] = 1302,
  [1311] = 1295,
  [1312] = 1302,
  [1313] = 1313,
  [1314] = 1295,
  [1315] = 1295,
  [1316] = 1302,
  [1317] = 1317,
  [1318] = 1318,
  [1319] = 1319,
  [1320] = 1320,
  [1321] = 1321,
  [1322] = 1322,
  [1323] = 1323,
  [1324] = 1324,
  [1325] = 1268,
  [1326] = 1326,
  [1327] = 1327,
  [1328] = 1328,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1264,
  [1332] = 1332,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1335,
  [1336] = 1262,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1340,
  [1341] = 1341,
  [1342] = 1342,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1345,
  [1346] = 1346,
  [1347] = 1347,
  [1348] = 1344,
  [1349] = 1344,
  [1350] = 1344,
  [1351] = 1351,
  [1352] = 1352,
  [1353] = 1353,
  [1354] = 1354,
  [1355] = 1294,
  [1356] = 1356,
  [1357] = 1357,
  [1358] = 1358,
  [1359] = 1292,
  [1360] = 1360,
  [1361] = 1356,
  [1362] = 1356,
  [1363] = 1363,
  [1364] = 1356,
  [1365] = 1365,
  [1366] = 1291,
  [1367] = 1293,
  [1368] = 1368,
  [1369] = 1369,
  [1370] = 1370,
  [1371] = 1371,
  [1372] = 1372,
  [1373] = 1373,
  [1374] = 1374,
  [1375] = 1375,
  [1376] = 1372,
  [1377] = 1377,
  [1378] = 1378,
  [1379] = 1371,
  [1380] = 1372,
  [1381] = 1381,
  [1382] = 1382,
  [1383] = 1373,
  [1384] = 1371,
  [1385] = 1385,
  [1386] = 1386,
  [1387] = 1372,
  [1388] = 1373,
  [1389] = 1389,
  [1390] = 1371,
  [1391] = 1391,
  [1392] = 1373,
  [1393] = 1393,
  [1394] = 1394,
  [1395] = 1395,
  [1396] = 1396,
  [1397] = 1397,
  [1398] = 1398,
  [1399] = 1399,
  [1400] = 1400,
  [1401] = 1401,
  [1402] = 1402,
  [1403] = 1403,
  [1404] = 1404,
  [1405] = 1401,
  [1406] = 1403,
  [1407] = 1407,
  [1408] = 1408,
  [1409] = 1409,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 1399,
  [1413] = 1403,
  [1414] = 1414,
  [1415] = 1415,
  [1416] = 1399,
  [1417] = 1417,
  [1418] = 1401,
  [1419] = 1401,
  [1420] = 1420,
  [1421] = 1421,
  [1422] = 1403,
  [1423] = 1423,
  [1424] = 1424,
  [1425] = 1425,
  [1426] = 1426,
  [1427] = 1399,
  [1428] = 1428,
  [1429] = 1429,
  [1430] = 1430,
  [1431] = 1431,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1435,
  [1436] = 1436,
  [1437] = 1437,
  [1438] = 1434,
  [1439] = 1437,
  [1440] = 1434,
  [1441] = 1441,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 1434,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 1437,
  [1448] = 1434,
  [1449] = 1449,
  [1450] = 1434,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1437,
  [1456] = 1456,
  [1457] = 1457,
  [1458] = 1437,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1463,
  [1464] = 1464,
  [1465] = 1437,
  [1466] = 1466,
  [1467] = 1467,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1470,
  [1471] = 1471,
  [1472] = 1472,
  [1473] = 1473,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1476,
  [1477] = 1477,
  [1478] = 1478,
  [1479] = 1479,
  [1480] = 1480,
  [1481] = 1481,
  [1482] = 1482,
  [1483] = 1483,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1490,
  [1491] = 1491,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 1494,
  [1495] = 1495,
  [1496] = 1496,
  [1497] = 1492,
  [1498] = 1492,
  [1499] = 1499,
  [1500] = 1500,
  [1501] = 1495,
  [1502] = 1493,
  [1503] = 1496,
  [1504] = 1504,
  [1505] = 1492,
  [1506] = 1506,
  [1507] = 1495,
  [1508] = 1496,
  [1509] = 1509,
  [1510] = 1493,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1496,
  [1514] = 1514,
  [1515] = 1493,
  [1516] = 1516,
  [1517] = 1495,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1519,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1524,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1524,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1531,
  [1533] = 1533,
  [1534] = 1524,
  [1535] = 1535,
  [1536] = 1521,
  [1537] = 1537,
  [1538] = 1520,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1542,
  [1544] = 1524,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1520,
  [1548] = 1542,
  [1549] = 1549,
  [1550] = 1542,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1542,
  [1557] = 1531,
  [1558] = 1524,
  [1559] = 1524,
  [1560] = 1560,
  [1561] = 1521,
  [1562] = 1519,
  [1563] = 1563,
  [1564] = 1564,
  [1565] = 1565,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1571,
  [1572] = 1572,
  [1573] = 1573,
  [1574] = 1570,
  [1575] = 1575,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1578,
  [1579] = 1579,
  [1580] = 1579,
  [1581] = 1581,
  [1582] = 1582,
  [1583] = 1583,
  [1584] = 1584,
  [1585] = 1570,
  [1586] = 1586,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1590,
  [1591] = 1591,
  [1592] = 1575,
  [1593] = 1578,
  [1594] = 1594,
  [1595] = 1595,
  [1596] = 1596,
  [1597] = 1597,
  [1598] = 1598,
  [1599] = 1567,
  [1600] = 1600,
  [1601] = 1601,
  [1602] = 1567,
  [1603] = 1603,
  [1604] = 1579,
  [1605] = 1575,
  [1606] = 1575,
  [1607] = 1607,
  [1608] = 1581,
  [1609] = 1582,
  [1610] = 1610,
  [1611] = 1583,
  [1612] = 1612,
  [1613] = 1581,
  [1614] = 1589,
  [1615] = 1590,
  [1616] = 1591,
  [1617] = 1582,
  [1618] = 1618,
  [1619] = 1583,
  [1620] = 1595,
  [1621] = 1621,
  [1622] = 1622,
  [1623] = 1623,
  [1624] = 1581,
  [1625] = 1582,
  [1626] = 1583,
  [1627] = 1589,
  [1628] = 1590,
  [1629] = 1591,
  [1630] = 1595,
  [1631] = 1631,
  [1632] = 1575,
  [1633] = 1581,
  [1634] = 1582,
  [1635] = 1583,
  [1636] = 1589,
  [1637] = 1590,
  [1638] = 1591,
  [1639] = 1595,
  [1640] = 1420,
  [1641] = 1618,
  [1642] = 1642,
  [1643] = 1643,
  [1644] = 1644,
  [1645] = 1645,
  [1646] = 1646,
  [1647] = 1647,
  [1648] = 1648,
  [1649] = 1649,
  [1650] = 1589,
  [1651] = 1590,
  [1652] = 1591,
  [1653] = 1653,
  [1654] = 1572,
  [1655] = 1578,
  [1656] = 1656,
  [1657] = 1657,
  [1658] = 1595,
  [1659] = 1618,
  [1660] = 1660,
  [1661] = 1581,
  [1662] = 1589,
  [1663] = 1590,
  [1664] = 1582,
  [1665] = 1583,
  [1666] = 1564,
  [1667] = 1621,
  [1668] = 1668,
  [1669] = 1669,
  [1670] = 1670,
  [1671] = 1671,
  [1672] = 1672,
  [1673] = 1595,
  [1674] = 1564,
  [1675] = 1567,
  [1676] = 1676,
  [1677] = 1591,
  [1678] = 1575,
  [1679] = 1679,
  [1680] = 1679,
  [1681] = 1681,
  [1682] = 1682,
  [1683] = 1683,
  [1684] = 1684,
  [1685] = 1679,
  [1686] = 1686,
  [1687] = 1684,
  [1688] = 1688,
  [1689] = 1689,
  [1690] = 1690,
  [1691] = 1681,
  [1692] = 1684,
  [1693] = 1693,
  [1694] = 1694,
  [1695] = 1695,
  [1696] = 1684,
  [1697] = 1697,
  [1698] = 1698,
  [1699] = 1699,
  [1700] = 1698,
  [1701] = 1701,
  [1702] = 1702,
  [1703] = 1703,
  [1704] = 1679,
  [1705] = 1684,
  [1706] = 1698,
  [1707] = 1707,
  [1708] = 1690,
  [1709] = 1703,
  [1710] = 1684,
  [1711] = 1711,
  [1712] = 1712,
  [1713] = 1713,
  [1714] = 1714,
  [1715] = 1679,
  [1716] = 1716,
  [1717] = 1684,
  [1718] = 1718,
  [1719] = 1719,
  [1720] = 1681,
  [1721] = 1693,
  [1722] = 1719,
  [1723] = 1679,
  [1724] = 1724,
  [1725] = 1724,
  [1726] = 1726,
  [1727] = 1727,
  [1728] = 1698,
  [1729] = 1714,
  [1730] = 1698,
  [1731] = 1697,
  [1732] = 1732,
  [1733] = 1711,
  [1734] = 1734,
  [1735] = 1735,
  [1736] = 1703,
  [1737] = 1703,
  [1738] = 1719,
  [1739] = 1681,
  [1740] = 1693,
  [1741] = 1711,
  [1742] = 1724,
  [1743] = 1716,
  [1744] = 1744,
  [1745] = 1714,
  [1746] = 1697,
  [1747] = 1690,
  [1748] = 1748,
  [1749] = 1693,
  [1750] = 1679,
  [1751] = 1724,
  [1752] = 1714,
  [1753] = 1697,
  [1754] = 1754,
  [1755] = 1681,
  [1756] = 1697,
  [1757] = 1711,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 1703,
  [1761] = 1724,
  [1762] = 1690,
  [1763] = 1763,
  [1764] = 1764,
  [1765] = 1765,
  [1766] = 1711,
  [1767] = 1767,
  [1768] = 1768,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1771,
  [1772] = 1772,
  [1773] = 1773,
  [1774] = 1774,
  [1775] = 1768,
  [1776] = 1776,
  [1777] = 1777,
  [1778] = 1769,
  [1779] = 1779,
  [1780] = 1780,
  [1781] = 1781,
  [1782] = 1782,
  [1783] = 998,
  [1784] = 1780,
  [1785] = 1769,
  [1786] = 1786,
  [1787] = 1787,
  [1788] = 1788,
  [1789] = 1769,
  [1790] = 1790,
  [1791] = 1791,
  [1792] = 1773,
  [1793] = 1779,
  [1794] = 1794,
  [1795] = 986,
  [1796] = 991,
  [1797] = 1797,
  [1798] = 1780,
  [1799] = 1799,
  [1800] = 1800,
  [1801] = 1801,
  [1802] = 1802,
  [1803] = 1803,
  [1804] = 1804,
  [1805] = 1769,
  [1806] = 1773,
  [1807] = 1807,
  [1808] = 1780,
  [1809] = 1809,
  [1810] = 1810,
  [1811] = 1811,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 1814,
  [1815] = 1815,
  [1816] = 1774,
  [1817] = 1807,
  [1818] = 1818,
  [1819] = 1819,
  [1820] = 1820,
  [1821] = 1821,
  [1822] = 1822,
  [1823] = 1823,
  [1824] = 1801,
  [1825] = 1779,
  [1826] = 1780,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1829,
  [1830] = 1830,
  [1831] = 1809,
  [1832] = 1832,
  [1833] = 1833,
  [1834] = 1834,
  [1835] = 1780,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1838,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1841,
  [1842] = 1842,
  [1843] = 1843,
  [1844] = 1840,
  [1845] = 1845,
  [1846] = 1846,
  [1847] = 1773,
  [1848] = 1848,
  [1849] = 1807,
  [1850] = 1842,
  [1851] = 1807,
  [1852] = 1852,
  [1853] = 1853,
  [1854] = 1845,
  [1855] = 1818,
  [1856] = 1779,
  [1857] = 1857,
  [1858] = 1858,
  [1859] = 1859,
  [1860] = 1860,
  [1861] = 1861,
  [1862] = 1768,
  [1863] = 1863,
  [1864] = 1864,
  [1865] = 1865,
  [1866] = 1866,
  [1867] = 1803,
  [1868] = 1801,
  [1869] = 1852,
  [1870] = 1865,
  [1871] = 1809,
  [1872] = 1872,
  [1873] = 1873,
  [1874] = 1874,
  [1875] = 1875,
  [1876] = 1859,
  [1877] = 1773,
  [1878] = 1818,
  [1879] = 1810,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1857,
  [1884] = 1774,
  [1885] = 1776,
  [1886] = 1827,
  [1887] = 1768,
  [1888] = 1888,
  [1889] = 1889,
  [1890] = 1819,
  [1891] = 1891,
  [1892] = 1863,
  [1893] = 1893,
  [1894] = 1894,
  [1895] = 1895,
  [1896] = 1896,
  [1897] = 1873,
  [1898] = 1898,
  [1899] = 1899,
  [1900] = 1900,
  [1901] = 1772,
  [1902] = 1810,
  [1903] = 1811,
  [1904] = 1804,
  [1905] = 1905,
  [1906] = 1857,
  [1907] = 1773,
  [1908] = 1908,
  [1909] = 1909,
  [1910] = 1801,
  [1911] = 1819,
  [1912] = 1912,
  [1913] = 1913,
  [1914] = 627,
  [1915] = 1822,
  [1916] = 1916,
  [1917] = 1803,
  [1918] = 1788,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 1863,
  [1922] = 1861,
  [1923] = 1780,
  [1924] = 1859,
  [1925] = 1840,
  [1926] = 1861,
  [1927] = 1927,
  [1928] = 1928,
  [1929] = 1837,
  [1930] = 1768,
  [1931] = 1931,
  [1932] = 1840,
  [1933] = 1842,
  [1934] = 1861,
  [1935] = 1810,
  [1936] = 1811,
  [1937] = 1937,
  [1938] = 1881,
  [1939] = 1001,
  [1940] = 1779,
  [1941] = 1819,
  [1942] = 1804,
  [1943] = 1943,
  [1944] = 1944,
  [1945] = 1858,
  [1946] = 1946,
  [1947] = 636,
  [1948] = 1880,
  [1949] = 1949,
  [1950] = 1842,
  [1951] = 1865,
  [1952] = 1952,
  [1953] = 1858,
  [1954] = 1954,
  [1955] = 1857,
  [1956] = 1881,
  [1957] = 1957,
  [1958] = 1810,
  [1959] = 1881,
  [1960] = 1811,
  [1961] = 1769,
  [1962] = 1810,
  [1963] = 1810,
  [1964] = 1818,
  [1965] = 1965,
  [1966] = 1891,
  [1967] = 1967,
  [1968] = 1880,
  [1969] = 1898,
  [1970] = 1946,
  [1971] = 1971,
  [1972] = 1972,
  [1973] = 1973,
  [1974] = 1974,
  [1975] = 1779,
  [1976] = 1852,
  [1977] = 1804,
  [1978] = 1880,
  [1979] = 1979,
  [1980] = 1888,
  [1981] = 1981,
  [1982] = 1891,
  [1983] = 1858,
  [1984] = 1900,
  [1985] = 1985,
  [1986] = 1986,
  [1987] = 1865,
  [1988] = 1988,
  [1989] = 1809,
  [1990] = 1990,
  [1991] = 1873,
  [1992] = 1946,
  [1993] = 1803,
  [1994] = 1994,
  [1995] = 1888,
  [1996] = 1768,
  [1997] = 1769,
  [1998] = 1861,
  [1999] = 1809,
  [2000] = 2000,
  [2001] = 1773,
  [2002] = 2002,
  [2003] = 1888,
  [2004] = 1881,
  [2005] = 1774,
  [2006] = 2006,
  [2007] = 1822,
  [2008] = 1946,
  [2009] = 2009,
  [2010] = 1788,
  [2011] = 1946,
  [2012] = 2012,
  [2013] = 2013,
  [2014] = 1809,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__abstract_declarator,
  sym__declarator,
  sym__field_declarator,
  sym__type_declarator,
  sym_expression,
  sym_statement,
  sym_type_specifier,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__abstract_declarator] = {.index = 0, .length = 4},
  [sym__declarator] = {.index = 4, .length = 6},
  [sym__field_declarator] = {.index = 10, .length = 18},
  [sym__type_declarator] = {.index = 28, .length = 19},
  [sym_expression] = {.index = 47, .length = 26},
  [sym_statement] = {.index = 73, .length = 16},
  [sym_type_specifier] = {.index = 89, .length = 7},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_abstract_array_declarator,
    sym_abstract_function_declarator,
    sym_abstract_parenthesized_declarator,
    sym_abstract_pointer_declarator,
  [4] =
    sym_array_declarator,
    sym_attributed_declarator,
    sym_function_declarator,
    sym_identifier,
    sym_parenthesized_declarator,
    sym_pointer_declarator,
  [10] =
    alias_sym_field_identifier,
    sym__function_declaration_declarator,
    sym__old_style_function_declarator,
    sym_array_declarator,
    sym_array_field_declarator,
    sym_array_type_declarator,
    sym_attributed_declarator,
    sym_attributed_field_declarator,
    sym_attributed_type_declarator,
    sym_function_declarator,
    sym_function_field_declarator,
    sym_function_type_declarator,
    sym_parenthesized_declarator,
    sym_parenthesized_field_declarator,
    sym_parenthesized_type_declarator,
    sym_pointer_declarator,
    sym_pointer_field_declarator,
    sym_pointer_type_declarator,
  [28] =
    alias_sym_type_identifier,
    sym__function_declaration_declarator,
    sym__old_style_function_declarator,
    sym_array_declarator,
    sym_array_field_declarator,
    sym_array_type_declarator,
    sym_attributed_declarator,
    sym_attributed_field_declarator,
    sym_attributed_type_declarator,
    sym_function_declarator,
    sym_function_field_declarator,
    sym_function_type_declarator,
    sym_parenthesized_declarator,
    sym_parenthesized_field_declarator,
    sym_parenthesized_type_declarator,
    sym_pointer_declarator,
    sym_pointer_field_declarator,
    sym_pointer_type_declarator,
    sym_primitive_type,
  [47] =
    sym_alignof_expression,
    sym_assignment_expression,
    sym_binary_expression,
    sym_call_expression,
    sym_cast_expression,
    sym_char_literal,
    sym_compound_literal_expression,
    sym_concatenated_string,
    sym_conditional_expression,
    sym_extension_expression,
    sym_false,
    sym_field_expression,
    sym_generic_expression,
    sym_gnu_asm_expression,
    sym_identifier,
    sym_null,
    sym_number_literal,
    sym_offsetof_expression,
    sym_parenthesized_expression,
    sym_pointer_expression,
    sym_sizeof_expression,
    sym_string_literal,
    sym_subscript_expression,
    sym_true,
    sym_unary_expression,
    sym_update_expression,
  [73] =
    sym_attributed_statement,
    sym_break_statement,
    sym_case_statement,
    sym_compound_statement,
    sym_continue_statement,
    sym_do_statement,
    sym_expression_statement,
    sym_for_statement,
    sym_goto_statement,
    sym_if_statement,
    sym_labeled_statement,
    sym_return_statement,
    sym_seh_leave_statement,
    sym_seh_try_statement,
    sym_switch_statement,
    sym_while_statement,
  [89] =
    alias_sym_type_identifier,
    sym_enum_specifier,
    sym_macro_type_specifier,
    sym_primitive_type,
    sym_sized_type_specifier,
    sym_struct_specifier,
    sym_union_specifier,
};

static const TSCharacterRange sym_number_literal_character_set_13[] = {
  {'0', '9'}, {'B', 'B'}, {'D', 'D'}, {'F', 'F'}, {'L', 'L'}, {'U', 'U'}, {'W', 'W'}, {'b', 'b'},
  {'d', 'd'}, {'f', 'f'}, {'l', 'l'}, {'u', 'u'}, {'w', 'w'},
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'$', '$'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba},
  {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374},
  {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a},
  {0x66e, 0x66f}, {0x671, 0x6d3}, {0x6d5, 0x6d5}, {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710},
  {0x712, 0x72f}, {0x74d, 0x7a5}, {0x7b1, 0x7b1}, {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a},
  {0x824, 0x824}, {0x828, 0x828}, {0x840, 0x858}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939},
  {0x93d, 0x93d}, {0x950, 0x950}, {0x958, 0x961}, {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0},
  {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bd, 0x9bd}, {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc},
  {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa72, 0xa74}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabd, 0xabd}, {0xad0, 0xad0}, {0xae0, 0xae1}, {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30},
  {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3d, 0xb3d}, {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbd0, 0xbd0}, {0xc05, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc61}, {0xc80, 0xc80}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd},
  {0xcdd, 0xcde}, {0xce0, 0xce1}, {0xcf1, 0xcf2}, {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e},
  {0xd54, 0xd56}, {0xd5f, 0xd61}, {0xd7a, 0xd7f}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xe01, 0xe30}, {0xe32, 0xe32}, {0xe40, 0xe46}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5},
  {0xea7, 0xeb0}, {0xeb2, 0xeb2}, {0xebd, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47},
  {0xf49, 0xf6c}, {0xf88, 0xf8c}, {0x1000, 0x102a}, {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066},
  {0x106e, 0x1070}, {0x1075, 0x1081}, {0x108e, 0x108e}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248},
  {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5},
  {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711},
  {0x171f, 0x1731}, {0x1740, 0x1751}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878},
  {0x1880, 0x18a8}, {0x18aa, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9},
  {0x1a00, 0x1a16}, {0x1a20, 0x1a54}, {0x1aa7, 0x1aa7}, {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5},
  {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f}, {0x1c5a, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3},
  {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa}, {0x1d00, 0x1dbf}, {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f},
  {0x2090, 0x209c}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126},
  {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee},
  {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6},
  {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007},
  {0x3021, 0x3029}, {0x3031, 0x3035}, {0x3038, 0x303c}, {0x3041, 0x3096}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f},
  {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f},
  {0xa62a, 0xa62b}, {0xa640, 0xa66e}, {0xa67f, 0xa69d}, {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1},
  {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa801}, {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3},
  {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa8fe}, {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf},
  {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef}, {0xa9fa, 0xa9fe}, {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a},
  {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1}, {0xaab5, 0xaab6}, {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea},
  {0xaaf2, 0xaaf4}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69},
  {0xab70, 0xabe2}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17},
  {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1},
  {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77},
  {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xff9d}, {0xffa0, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a},
  {0x10350, 0x10375}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb},
  {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1},
  {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0},
  {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876},
  {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00},
  {0x10a10, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35},
  {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81},
  {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126},
  {0x11144, 0x11144}, {0x11147, 0x11147}, {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc},
  {0x11200, 0x11211}, {0x11213, 0x1122b}, {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8},
  {0x112b0, 0x112de}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d},
  {0x11350, 0x11350}, {0x1135d, 0x11361}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1},
  {0x113d3, 0x113d3}, {0x11400, 0x11434}, {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae},
  {0x115d8, 0x115db}, {0x11600, 0x1162f}, {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b},
  {0x118a0, 0x118df}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941},
  {0x119a0, 0x119a7}, {0x119aa, 0x119d0}, {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50},
  {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f},
  {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98},
  {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02}, {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543},
  {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e},
  {0x16a70, 0x16abe}, {0x16ad0, 0x16aed}, {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f50, 0x16f50}, {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f},
  {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a},
  {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5},
  {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788},
  {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad}, {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739},
  {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'$', '$'}, {'0', '9'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5},
  {0xb7, 0xb7}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec},
  {0x2ee, 0x2ee}, {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1},
  {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd},
  {0x5bf, 0x5bf}, {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669},
  {0x66e, 0x6d3}, {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5},
  {0x7fa, 0x7fa}, {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1},
  {0x8e3, 0x963}, {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2},
  {0x9b6, 0x9b9}, {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1},
  {0x9fc, 0x9fc}, {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33},
  {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c},
  {0xa5e, 0xa5e}, {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3},
  {0xab5, 0xab9}, {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff},
  {0xb01, 0xb03}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44},
  {0xb47, 0xb48}, {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83},
  {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa},
  {0xbae, 0xbb9}, {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c},
  {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a},
  {0xc5d, 0xc5d}, {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3},
  {0xcb5, 0xcb9}, {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef},
  {0xcf1, 0xcf3}, {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63},
  {0xd66, 0xd6f}, {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6},
  {0xdca, 0xdca}, {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e},
  {0xe50, 0xe59}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4},
  {0xec6, 0xec6}, {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35},
  {0xf37, 0xf37}, {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6},
  {0x1000, 0x1049}, {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d},
  {0x1250, 0x1256}, {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be},
  {0x12c0, 0x12c0}, {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371},
  {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8},
  {0x1700, 0x1715}, {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7},
  {0x17dc, 0x17dd}, {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e},
  {0x1920, 0x192b}, {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b},
  {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c},
  {0x1b50, 0x1b59}, {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba},
  {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4},
  {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040},
  {0x2054, 0x2054}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102},
  {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf},
  {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d},
  {0xa67f, 0xa6f1}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827},
  {0xa82c, 0xa82c}, {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953},
  {0xa960, 0xa97c}, {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76},
  {0xaa7a, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26},
  {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c},
  {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7},
  {0xfdf0, 0xfdf9}, {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77},
  {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a},
  {0xff65, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a},
  {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0},
  {0x102e0, 0x102e0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5},
  {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a},
  {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736},
  {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835},
  {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915},
  {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35},
  {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55},
  {0x10b60, 0x10b72}, {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65},
  {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27},
  {0x10f30, 0x10f50}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2},
  {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4},
  {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288},
  {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310},
  {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350},
  {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5},
  {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459},
  {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640},
  {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739},
  {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935},
  {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e},
  {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36},
  {0x11c38, 0x11c40}, {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36},
  {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91},
  {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0},
  {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646},
  {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4},
  {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f},
  {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5},
  {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155},
  {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9},
  {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244},
  {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb},
  {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544},
  {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734},
  {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36},
  {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006},
  {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d},
  {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb},
  {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f},
  {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42},
  {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59},
  {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77},
  {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(121);
      ADVANCE_MAP(
        '!', 188,
        '"', 287,
        '#', 75,
        '%', 205,
        '&', 214,
        '\'', 278,
        '(', 125,
        ')', 128,
        '*', 201,
        '+', 196,
        ',', 127,
        '-', 191,
        '.', 254,
        '/', 203,
        '0', 260,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        'L', 302,
        'U', 304,
        '[', 234,
        '\\', 2,
        ']', 235,
        '^', 211,
        'u', 306,
        '{', 231,
        '|', 208,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(119);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(43);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(43);
      if (lookahead == '\r') SKIP(1);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(46);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(46);
      if (lookahead == '\r') SKIP(3);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(45);
      END_STATE();
    case 6:
      if (lookahead == '\n') SKIP(45);
      if (lookahead == '\r') SKIP(5);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(47);
      END_STATE();
    case 8:
      if (lookahead == '\n') SKIP(47);
      if (lookahead == '\r') SKIP(7);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 9:
      if (lookahead == '\n') SKIP(49);
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(49);
      if (lookahead == '\r') SKIP(9);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(53);
      END_STATE();
    case 12:
      if (lookahead == '\n') SKIP(53);
      if (lookahead == '\r') SKIP(11);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 13:
      if (lookahead == '\n') SKIP(52);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(52);
      if (lookahead == '\r') SKIP(13);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(57);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(57);
      if (lookahead == '\r') SKIP(15);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(50);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(50);
      if (lookahead == '\r') SKIP(17);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(51);
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(51);
      if (lookahead == '\r') SKIP(19);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(23);
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(23);
      if (lookahead == '\r') SKIP(21);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '\n', 130,
        '!', 68,
        '%', 204,
        '&', 213,
        '(', 186,
        '*', 200,
        '+', 195,
        '-', 190,
        '/', 202,
        '<', 222,
        '=', 69,
        '>', 218,
      );
      if (lookahead == '\\') SKIP(22);
      if (lookahead == '^') ADVANCE(210);
      if (lookahead == '|') ADVANCE(209);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(48);
      END_STATE();
    case 25:
      if (lookahead == '\n') SKIP(48);
      if (lookahead == '\r') SKIP(24);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 26:
      if (lookahead == '\n') SKIP(56);
      END_STATE();
    case 27:
      if (lookahead == '\n') SKIP(56);
      if (lookahead == '\r') SKIP(26);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(54);
      END_STATE();
    case 29:
      if (lookahead == '\n') SKIP(54);
      if (lookahead == '\r') SKIP(28);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 30:
      if (lookahead == '\n') SKIP(58);
      if (lookahead == '\'') ADVANCE(278);
      if (lookahead == '/') ADVANCE(281);
      if (lookahead == '\\') ADVANCE(280);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(282);
      if (lookahead != 0) ADVANCE(279);
      END_STATE();
    case 31:
      if (lookahead == '\n') ADVANCE(294);
      if (lookahead == '\r') ADVANCE(293);
      if (lookahead == 'U') ADVANCE(117);
      if (lookahead == 'u') ADVANCE(109);
      if (lookahead == 'x') ADVANCE(103);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(296);
      if (lookahead != 0) ADVANCE(293);
      END_STATE();
    case 32:
      if (lookahead == '\n') SKIP(55);
      if (lookahead == '"') ADVANCE(287);
      if (lookahead == '/') ADVANCE(288);
      if (lookahead == '\\') ADVANCE(31);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(291);
      if (lookahead != 0) ADVANCE(292);
      END_STATE();
    case 33:
      if (lookahead == '\n') ADVANCE(123);
      if (lookahead == '\r') ADVANCE(37);
      if (lookahead == '(') ADVANCE(125);
      if (lookahead == '/') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(146);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(66);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 34:
      if (lookahead == '\n') ADVANCE(123);
      if (lookahead == '\r') ADVANCE(37);
      if (lookahead == '/') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(146);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(66);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 35:
      if (lookahead == '\n') ADVANCE(123);
      if (lookahead == '\r') ADVANCE(36);
      if (lookahead == '(') ADVANCE(186);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') SKIP(39);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(59);
      END_STATE();
    case 36:
      if (lookahead == '\n') ADVANCE(123);
      if (lookahead == '(') ADVANCE(186);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') SKIP(39);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(59);
      END_STATE();
    case 37:
      if (lookahead == '\n') ADVANCE(123);
      if (lookahead == '/') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(146);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 38:
      if (lookahead == '\n') SKIP(59);
      END_STATE();
    case 39:
      if (lookahead == '\n') SKIP(59);
      if (lookahead == '\r') SKIP(38);
      END_STATE();
    case 40:
      if (lookahead == '\n') SKIP(44);
      END_STATE();
    case 41:
      if (lookahead == '\n') SKIP(44);
      if (lookahead == '\r') SKIP(40);
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 42:
      if (lookahead == '\r') ADVANCE(323);
      if (lookahead == '\\') ADVANCE(317);
      if (lookahead != 0) ADVANCE(322);
      END_STATE();
    case 43:
      ADVANCE_MAP(
        '!', 188,
        '"', 287,
        '#', 75,
        '%', 205,
        '&', 214,
        '\'', 278,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 196,
        ',', 127,
        '-', 191,
        '.', 254,
        '/', 203,
        '0', 260,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        'L', 302,
        'U', 304,
        '[', 234,
        '\\', 2,
        ']', 235,
        '^', 211,
        'u', 306,
        '{', 231,
        '|', 208,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(43);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        '!', 188,
        '"', 287,
        '#', 83,
        '%', 205,
        '&', 214,
        '\'', 278,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 196,
        ',', 127,
        '-', 191,
        '.', 254,
        '/', 203,
        '0', 260,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        'L', 302,
        'U', 304,
        '[', 233,
        '\\', 41,
        ']', 235,
        '^', 211,
        'u', 306,
        '{', 231,
        '|', 208,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(44);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 45:
      ADVANCE_MAP(
        '!', 187,
        '"', 287,
        '#', 75,
        '&', 212,
        '\'', 278,
        '(', 186,
        '*', 200,
        '+', 197,
        ',', 127,
        '-', 192,
        '.', 96,
        '/', 60,
        '0', 260,
        ':', 67,
        ';', 227,
        'L', 302,
        'U', 304,
        '[', 73,
        '\\', 6,
        ']', 74,
        'u', 306,
        '{', 231,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 46:
      ADVANCE_MAP(
        '!', 187,
        '"', 287,
        '#', 79,
        '&', 212,
        '\'', 278,
        '(', 186,
        ')', 128,
        '*', 200,
        '+', 197,
        ',', 127,
        '-', 192,
        '.', 255,
        '/', 60,
        '0', 260,
        ':', 238,
        ';', 227,
        '=', 236,
        'L', 302,
        'U', 304,
        '[', 234,
        '\\', 4,
        ']', 235,
        'u', 306,
        '{', 231,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(46);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 47:
      ADVANCE_MAP(
        '!', 187,
        '"', 287,
        '#', 77,
        '&', 212,
        '\'', 278,
        '(', 186,
        '*', 200,
        '+', 197,
        '-', 192,
        '.', 96,
        '/', 60,
        '0', 260,
        ';', 227,
        'L', 302,
        'U', 304,
        '[', 73,
        '\\', 8,
        'u', 306,
        '{', 231,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(47);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 48:
      ADVANCE_MAP(
        '!', 187,
        '\'', 278,
        '(', 186,
        ')', 128,
        '+', 199,
        '-', 194,
        '.', 96,
        '/', 60,
        '0', 260,
        'L', 310,
        'U', 311,
        '\\', 25,
        'u', 312,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(48);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 49:
      ADVANCE_MAP(
        '!', 68,
        '"', 287,
        '#', 83,
        '%', 205,
        '&', 214,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 198,
        ',', 127,
        '-', 193,
        '.', 253,
        '/', 203,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        'L', 303,
        'U', 305,
        '[', 234,
        '\\', 10,
        ']', 235,
        '^', 211,
        'u', 307,
        '|', 208,
        '}', 232,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(49);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 50:
      ADVANCE_MAP(
        '!', 68,
        '#', 83,
        '%', 205,
        '&', 214,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 198,
        ',', 127,
        '-', 193,
        '.', 253,
        '/', 203,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        '[', 234,
        '\\', 18,
        ']', 235,
        '^', 211,
        '{', 231,
        '|', 208,
        '}', 232,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(50);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '!', 68,
        '#', 83,
        '%', 205,
        '&', 214,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 198,
        ',', 127,
        '-', 193,
        '.', 252,
        '/', 203,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        '[', 233,
        '\\', 20,
        ']', 74,
        '^', 211,
        '|', 208,
        '}', 232,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '!', 68,
        '#', 80,
        '%', 204,
        '&', 213,
        '(', 186,
        ')', 128,
        '*', 200,
        '+', 195,
        ',', 127,
        '-', 190,
        '/', 202,
        ';', 227,
        '<', 222,
        '=', 237,
        '>', 218,
        '[', 234,
        '\\', 14,
        '^', 210,
        '{', 231,
        '|', 209,
        '}', 232,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 53:
      ADVANCE_MAP(
        '!', 68,
        '#', 76,
        '%', 205,
        '&', 214,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 198,
        ',', 127,
        '-', 193,
        '.', 253,
        '/', 203,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        '[', 234,
        '\\', 12,
        '^', 211,
        '{', 231,
        '|', 208,
        '}', 232,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(53);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 54:
      if (lookahead == '"') ADVANCE(287);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '<') ADVANCE(70);
      if (lookahead == 'L') ADVANCE(303);
      if (lookahead == 'U') ADVANCE(305);
      if (lookahead == '\\') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(307);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 55:
      if (lookahead == '"') ADVANCE(287);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') ADVANCE(31);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      END_STATE();
    case 56:
      if (lookahead == '#') ADVANCE(93);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') ADVANCE(27);
      if (lookahead == '}') ADVANCE(232);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 57:
      if (lookahead == '#') ADVANCE(78);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '[') ADVANCE(73);
      if (lookahead == '\\') ADVANCE(16);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 58:
      if (lookahead == '\'') ADVANCE(278);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') ADVANCE(31);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(58);
      END_STATE();
    case 59:
      if (lookahead == '(') ADVANCE(186);
      if (lookahead == '/') ADVANCE(60);
      if (lookahead == '\\') SKIP(39);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(59);
      END_STATE();
    case 60:
      if (lookahead == '*') ADVANCE(62);
      if (lookahead == '/') ADVANCE(322);
      END_STATE();
    case 61:
      if (lookahead == '*') ADVANCE(61);
      if (lookahead == '/') ADVANCE(315);
      if (lookahead != 0) ADVANCE(62);
      END_STATE();
    case 62:
      if (lookahead == '*') ADVANCE(61);
      if (lookahead != 0) ADVANCE(62);
      END_STATE();
    case 63:
      if (lookahead == '*') ADVANCE(61);
      if (lookahead != 0) ADVANCE(144);
      END_STATE();
    case 64:
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(258);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(268);
      END_STATE();
    case 65:
      if (lookahead == '.') ADVANCE(126);
      END_STATE();
    case 66:
      if (lookahead == '/') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(146);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 67:
      if (lookahead == ':') ADVANCE(228);
      END_STATE();
    case 68:
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 69:
      if (lookahead == '=') ADVANCE(215);
      END_STATE();
    case 70:
      if (lookahead == '>') ADVANCE(300);
      if (lookahead == '\\') ADVANCE(71);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(70);
      END_STATE();
    case 71:
      if (lookahead == '>') ADVANCE(301);
      if (lookahead == '\\') ADVANCE(71);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(70);
      END_STATE();
    case 72:
      if (lookahead == 'U') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(108);
      END_STATE();
    case 73:
      if (lookahead == '[') ADVANCE(229);
      END_STATE();
    case 74:
      if (lookahead == ']') ADVANCE(230);
      END_STATE();
    case 75:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'e') ADVANCE(180);
      if (lookahead == 'i') ADVANCE(168);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(75);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 76:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'e') ADVANCE(180);
      if (lookahead == 'i') ADVANCE(169);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(76);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 77:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'e') ADVANCE(182);
      if (lookahead == 'i') ADVANCE(168);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(77);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 78:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'e') ADVANCE(182);
      if (lookahead == 'i') ADVANCE(169);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(78);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 79:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'i') ADVANCE(168);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(79);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 80:
      if (lookahead == 'd') ADVANCE(160);
      if (lookahead == 'i') ADVANCE(169);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(80);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 81:
      if (lookahead == 'd') ADVANCE(92);
      END_STATE();
    case 82:
      if (lookahead == 'd') ADVANCE(86);
      END_STATE();
    case 83:
      if (lookahead == 'e') ADVANCE(94);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(83);
      END_STATE();
    case 84:
      if (lookahead == 'e') ADVANCE(135);
      END_STATE();
    case 85:
      if (lookahead == 'e') ADVANCE(89);
      END_STATE();
    case 86:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 87:
      if (lookahead == 'f') ADVANCE(137);
      END_STATE();
    case 88:
      if (lookahead == 'f') ADVANCE(131);
      END_STATE();
    case 89:
      if (lookahead == 'f') ADVANCE(139);
      END_STATE();
    case 90:
      if (lookahead == 'f') ADVANCE(141);
      END_STATE();
    case 91:
      if (lookahead == 'i') ADVANCE(87);
      if (lookahead == 's') ADVANCE(84);
      END_STATE();
    case 92:
      if (lookahead == 'i') ADVANCE(88);
      END_STATE();
    case 93:
      if (lookahead == 'i') ADVANCE(169);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 94:
      if (lookahead == 'l') ADVANCE(91);
      if (lookahead == 'n') ADVANCE(81);
      END_STATE();
    case 95:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 96:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(257);
      END_STATE();
    case 97:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(268);
      END_STATE();
    case 98:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(314);
      END_STATE();
    case 99:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(265);
      END_STATE();
    case 100:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(268);
      END_STATE();
    case 101:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(263);
      END_STATE();
    case 102:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(293);
      END_STATE();
    case 103:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(299);
      END_STATE();
    case 104:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(98);
      END_STATE();
    case 105:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(102);
      END_STATE();
    case 106:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(104);
      END_STATE();
    case 107:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(105);
      END_STATE();
    case 108:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(106);
      END_STATE();
    case 109:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(107);
      END_STATE();
    case 110:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(108);
      END_STATE();
    case 111:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(109);
      END_STATE();
    case 112:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(110);
      END_STATE();
    case 113:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(111);
      END_STATE();
    case 114:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(112);
      END_STATE();
    case 115:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(113);
      END_STATE();
    case 116:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(114);
      END_STATE();
    case 117:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(115);
      END_STATE();
    case 118:
      if (lookahead != 0 &&
          lookahead != '*') ADVANCE(153);
      END_STATE();
    case 119:
      if (eof) ADVANCE(121);
      ADVANCE_MAP(
        '!', 188,
        '"', 287,
        '#', 75,
        '%', 205,
        '&', 214,
        '\'', 278,
        '(', 186,
        ')', 128,
        '*', 201,
        '+', 196,
        ',', 127,
        '-', 191,
        '.', 254,
        '/', 203,
        '0', 260,
        ':', 238,
        ';', 227,
        '<', 221,
        '=', 237,
        '>', 217,
        '?', 239,
        'L', 302,
        'U', 304,
        '[', 234,
        '\\', 2,
        ']', 235,
        '^', 211,
        'u', 306,
        '{', 231,
        '|', 208,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(119);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 120:
      if (eof) ADVANCE(121);
      ADVANCE_MAP(
        '!', 187,
        '"', 287,
        '#', 79,
        '&', 212,
        '\'', 278,
        '(', 186,
        ')', 128,
        '*', 200,
        '+', 197,
        ',', 127,
        '-', 192,
        '.', 255,
        '/', 60,
        '0', 260,
        ':', 238,
        ';', 227,
        '=', 236,
        'L', 302,
        'U', 304,
        '[', 234,
        '\\', 4,
        ']', 235,
        'u', 306,
        '{', 231,
        '}', 232,
        '~', 189,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(120);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_identifier_character_set_1, 687, lookahead)) ADVANCE(314);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(aux_sym_preproc_include_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(aux_sym_preproc_include_token2);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(aux_sym_preproc_def_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(aux_sym_preproc_if_token1);
      if (lookahead == 'd') ADVANCE(164);
      if (lookahead == 'n') ADVANCE(158);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_LF);
      if (lookahead == '\n') ADVANCE(130);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(aux_sym_preproc_if_token2);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(aux_sym_preproc_if_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(aux_sym_preproc_ifdef_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(aux_sym_preproc_else_token1);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(aux_sym_preproc_else_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      if (lookahead == 'd') ADVANCE(85);
      if (lookahead == 'n') ADVANCE(82);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(aux_sym_preproc_elif_token1);
      if (lookahead == 'd') ADVANCE(166);
      if (lookahead == 'n') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token1);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token2);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(aux_sym_preproc_elifdef_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(62);
      if (lookahead == '*') ADVANCE(143);
      if (lookahead == '/') ADVANCE(315);
      if (lookahead == '\\') ADVANCE(149);
      if (lookahead != 0) ADVANCE(144);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(62);
      if (lookahead == '*') ADVANCE(143);
      if (lookahead == '/') ADVANCE(63);
      if (lookahead == '\\') ADVANCE(149);
      if (lookahead != 0) ADVANCE(144);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') ADVANCE(322);
      if (lookahead == '\r') ADVANCE(316);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(318);
      if (lookahead != 0) ADVANCE(320);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') SKIP(66);
      if (lookahead == '\r') ADVANCE(147);
      if (lookahead == '/') ADVANCE(118);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\n') SKIP(66);
      if (lookahead == '/') ADVANCE(118);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(154);
      if (lookahead == '/') ADVANCE(118);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(152);
      if (lookahead == '*') ADVANCE(143);
      if (lookahead == '/') ADVANCE(63);
      if (lookahead == '\\') ADVANCE(149);
      if (lookahead != 0) ADVANCE(144);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '\r') ADVANCE(321);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(318);
      if (lookahead != 0) ADVANCE(320);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '*') ADVANCE(144);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '*') ADVANCE(143);
      if (lookahead == '/') ADVANCE(63);
      if (lookahead == '\\') ADVANCE(149);
      if (lookahead != 0) ADVANCE(144);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '/') ADVANCE(118);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(sym_preproc_arg);
      if (lookahead == '/') ADVANCE(118);
      if (lookahead == '\\') ADVANCE(148);
      if (lookahead != 0) ADVANCE(153);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'c') ADVANCE(181);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(179);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(163);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(165);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'd') ADVANCE(167);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(170);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(122);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(173);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(174);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(175);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'e') ADVANCE(176);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(129);
      if (lookahead == 'n') ADVANCE(155);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(177);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(132);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(134);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'f') ADVANCE(142);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(171);
      if (lookahead == 's') ADVANCE(161);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'i') ADVANCE(172);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'l') ADVANCE(178);
      if (lookahead == 'n') ADVANCE(156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'l') ADVANCE(184);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'n') ADVANCE(156);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'n') ADVANCE(162);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (lookahead == 'u') ADVANCE(157);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(sym_preproc_directive);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(185);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(250);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (lookahead == '=') ADVANCE(244);
      if (lookahead == '>') ADVANCE(256);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(250);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(250);
      if (lookahead == '=') ADVANCE(244);
      if (lookahead == '>') ADVANCE(256);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(251);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (lookahead == '=') ADVANCE(243);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(251);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(251);
      if (lookahead == '=') ADVANCE(243);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(260);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(240);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(62);
      if (lookahead == '/') ADVANCE(322);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(62);
      if (lookahead == '/') ADVANCE(322);
      if (lookahead == '=') ADVANCE(241);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(242);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(249);
      if (lookahead == '|') ADVANCE(206);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(206);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(248);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(207);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(207);
      if (lookahead == '=') ADVANCE(247);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(219);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(219);
      if (lookahead == '>') ADVANCE(225);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(224);
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(223);
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(245);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(246);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_LBRACK_LBRACK);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_RBRACK_RBRACK);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '[') ADVANCE(229);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(215);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(65);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(65);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(257);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(257);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(sym_number_literal);
      if (lookahead == '\'') ADVANCE(96);
      if (lookahead == 'E' ||
          lookahead == 'P' ||
          lookahead == 'e' ||
          lookahead == 'p') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(257);
      if (set_contains(sym_number_literal_character_set_13, 13, lookahead)) ADVANCE(273);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 97,
        '.', 271,
        'B', 267,
        'b', 267,
        'E', 266,
        'e', 266,
        'P', 270,
        'p', 270,
        'X', 100,
        'x', 100,
        'A', 268,
        'C', 268,
        'a', 268,
        'c', 268,
        'D', 268,
        'F', 268,
        'd', 268,
        'f', 268,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 97,
        '.', 271,
        'E', 266,
        'e', 266,
        'P', 270,
        'p', 270,
        'A', 268,
        'C', 268,
        'a', 268,
        'c', 268,
        'B', 268,
        'D', 268,
        'F', 268,
        'b', 268,
        'd', 268,
        'f', 268,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 95,
        '.', 271,
        'B', 269,
        'b', 269,
        'X', 64,
        'x', 64,
        'E', 270,
        'P', 270,
        'e', 270,
        'p', 270,
        'D', 273,
        'F', 273,
        'L', 273,
        'U', 273,
        'W', 273,
        'd', 273,
        'f', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 95,
        '.', 271,
        'B', 272,
        'b', 272,
        'X', 100,
        'x', 100,
        'E', 270,
        'P', 270,
        'e', 270,
        'p', 270,
        'D', 273,
        'F', 273,
        'L', 273,
        'U', 273,
        'W', 273,
        'd', 273,
        'f', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(262);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_number_literal);
      if (lookahead == '\'') ADVANCE(95);
      if (lookahead == '.') ADVANCE(271);
      if (lookahead == 'E' ||
          lookahead == 'P' ||
          lookahead == 'e' ||
          lookahead == 'p') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_number_literal_character_set_13, 13, lookahead)) ADVANCE(273);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 101,
        'B', 263,
        'D', 263,
        'F', 263,
        'b', 263,
        'd', 263,
        'f', 263,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'E') ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(263);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 99,
        '+', 101,
        '-', 101,
        'E', 264,
        'e', 264,
        'P', 270,
        'p', 270,
        'B', 265,
        'D', 265,
        'F', 265,
        'b', 265,
        'd', 265,
        'f', 265,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'C') ||
          ('a' <= lookahead && lookahead <= 'c')) ADVANCE(265);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 99,
        'E', 264,
        'e', 264,
        'P', 270,
        'p', 270,
        'B', 265,
        'D', 265,
        'F', 265,
        'b', 265,
        'd', 265,
        'f', 265,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'C') ||
          ('a' <= lookahead && lookahead <= 'c')) ADVANCE(265);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 100,
        '.', 271,
        '+', 101,
        '-', 101,
        'E', 266,
        'e', 266,
        'P', 270,
        'p', 270,
        'B', 268,
        'D', 268,
        'F', 268,
        'b', 268,
        'd', 268,
        'f', 268,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'C') ||
          ('a' <= lookahead && lookahead <= 'c')) ADVANCE(268);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 100,
        '.', 271,
        'E', 266,
        'e', 266,
        'P', 270,
        'p', 270,
        'A', 268,
        'C', 268,
        'a', 268,
        'c', 268,
        'B', 268,
        'D', 268,
        'F', 268,
        'b', 268,
        'd', 268,
        'f', 268,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '\'', 100,
        '.', 271,
        'E', 266,
        'e', 266,
        'P', 270,
        'p', 270,
        'B', 268,
        'D', 268,
        'F', 268,
        'b', 268,
        'd', 268,
        'f', 268,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'C') ||
          ('a' <= lookahead && lookahead <= 'c')) ADVANCE(268);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(sym_number_literal);
      if (lookahead == '.') ADVANCE(96);
      if (lookahead == '0') ADVANCE(261);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_number_literal_character_set_13, 13, lookahead)) ADVANCE(273);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        '+', 101,
        '-', 101,
        'B', 263,
        'D', 263,
        'F', 263,
        'b', 263,
        'd', 263,
        'f', 263,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'E') ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(263);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        'E', 264,
        'e', 264,
        'P', 270,
        'p', 270,
        'B', 265,
        'D', 265,
        'F', 265,
        'b', 265,
        'd', 265,
        'f', 265,
        'L', 273,
        'U', 273,
        'W', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'C') ||
          ('a' <= lookahead && lookahead <= 'c')) ADVANCE(265);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(sym_number_literal);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(262);
      if (set_contains(sym_number_literal_character_set_13, 13, lookahead)) ADVANCE(273);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(sym_number_literal);
      ADVANCE_MAP(
        'B', 273,
        'D', 273,
        'F', 273,
        'L', 273,
        'U', 273,
        'W', 273,
        'b', 273,
        'd', 273,
        'f', 273,
        'l', 273,
        'u', 273,
        'w', 273,
      );
      END_STATE();
    case 274:
      ACCEPT_TOKEN(anon_sym_L_SQUOTE);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(anon_sym_u_SQUOTE);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(anon_sym_U_SQUOTE);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_u8_SQUOTE);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '\n') ADVANCE(294);
      if (lookahead == '\r') ADVANCE(293);
      if (lookahead == 'U') ADVANCE(117);
      if (lookahead == 'u') ADVANCE(109);
      if (lookahead == 'x') ADVANCE(103);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(296);
      if (lookahead != 0) ADVANCE(293);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '*') ADVANCE(62);
      if (lookahead == '/') ADVANCE(322);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(aux_sym_char_literal_token1);
      if (lookahead == '\\') ADVANCE(31);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(anon_sym_L_DQUOTE);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(anon_sym_u_DQUOTE);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_U_DQUOTE);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_u8_DQUOTE);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(290);
      if (lookahead == '/') ADVANCE(292);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(292);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(289);
      if (lookahead == '/') ADVANCE(292);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(290);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '*') ADVANCE(289);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(290);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead == '/') ADVANCE(288);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(291);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(292);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(292);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(31);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(293);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(295);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(293);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(297);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(298);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(sym_system_lib_string);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(sym_system_lib_string);
      if (lookahead == '>') ADVANCE(300);
      if (lookahead == '\\') ADVANCE(71);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(70);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(283);
      if (lookahead == '\'') ADVANCE(274);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(283);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(285);
      if (lookahead == '\'') ADVANCE(276);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(285);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(284);
      if (lookahead == '\'') ADVANCE(275);
      if (lookahead == '8') ADVANCE(308);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(284);
      if (lookahead == '8') ADVANCE(309);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(286);
      if (lookahead == '\'') ADVANCE(277);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(286);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(274);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(276);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(275);
      if (lookahead == '8') ADVANCE(313);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\'') ADVANCE(277);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(72);
      if (set_contains(sym_identifier_character_set_2, 802, lookahead)) ADVANCE(314);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(322);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(150);
      if (lookahead != 0) ADVANCE(320);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(323);
      if (lookahead == '\\') ADVANCE(317);
      if (lookahead != 0) ADVANCE(322);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\r') ADVANCE(321);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(318);
      if (lookahead != 0) ADVANCE(320);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '*') ADVANCE(322);
      if (lookahead == '\\') ADVANCE(145);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(320);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(150);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(320);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\\') ADVANCE(150);
      if (lookahead != 0) ADVANCE(320);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\\') ADVANCE(42);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(322);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\\') ADVANCE(42);
      if (lookahead != 0) ADVANCE(322);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == 'F') ADVANCE(1);
      if (lookahead == 'N') ADVANCE(2);
      if (lookahead == 'T') ADVANCE(3);
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '_') ADVANCE(5);
      if (lookahead == 'a') ADVANCE(6);
      if (lookahead == 'b') ADVANCE(7);
      if (lookahead == 'c') ADVANCE(8);
      if (lookahead == 'd') ADVANCE(9);
      if (lookahead == 'e') ADVANCE(10);
      if (lookahead == 'f') ADVANCE(11);
      if (lookahead == 'g') ADVANCE(12);
      if (lookahead == 'i') ADVANCE(13);
      if (lookahead == 'l') ADVANCE(14);
      if (lookahead == 'm') ADVANCE(15);
      if (lookahead == 'n') ADVANCE(16);
      if (lookahead == 'o') ADVANCE(17);
      if (lookahead == 'p') ADVANCE(18);
      if (lookahead == 'r') ADVANCE(19);
      if (lookahead == 's') ADVANCE(20);
      if (lookahead == 't') ADVANCE(21);
      if (lookahead == 'u') ADVANCE(22);
      if (lookahead == 'v') ADVANCE(23);
      if (lookahead == 'w') ADVANCE(24);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'A') ADVANCE(25);
      END_STATE();
    case 2:
      if (lookahead == 'U') ADVANCE(26);
      END_STATE();
    case 3:
      if (lookahead == 'R') ADVANCE(27);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(0);
      if (lookahead == '\r') SKIP(28);
      END_STATE();
    case 5:
      if (lookahead == 'A') ADVANCE(29);
      if (lookahead == 'G') ADVANCE(30);
      if (lookahead == 'N') ADVANCE(31);
      if (lookahead == '_') ADVANCE(32);
      if (lookahead == 'a') ADVANCE(33);
      if (lookahead == 'u') ADVANCE(34);
      END_STATE();
    case 6:
      if (lookahead == 'l') ADVANCE(35);
      if (lookahead == 's') ADVANCE(36);
      if (lookahead == 'u') ADVANCE(37);
      END_STATE();
    case 7:
      if (lookahead == 'o') ADVANCE(38);
      if (lookahead == 'r') ADVANCE(39);
      END_STATE();
    case 8:
      if (lookahead == 'a') ADVANCE(40);
      if (lookahead == 'h') ADVANCE(41);
      if (lookahead == 'o') ADVANCE(42);
      END_STATE();
    case 9:
      if (lookahead == 'e') ADVANCE(43);
      if (lookahead == 'o') ADVANCE(44);
      END_STATE();
    case 10:
      if (lookahead == 'l') ADVANCE(45);
      if (lookahead == 'n') ADVANCE(46);
      if (lookahead == 'x') ADVANCE(47);
      END_STATE();
    case 11:
      if (lookahead == 'a') ADVANCE(48);
      if (lookahead == 'l') ADVANCE(49);
      if (lookahead == 'o') ADVANCE(50);
      END_STATE();
    case 12:
      if (lookahead == 'o') ADVANCE(51);
      END_STATE();
    case 13:
      if (lookahead == 'f') ADVANCE(52);
      if (lookahead == 'n') ADVANCE(53);
      END_STATE();
    case 14:
      if (lookahead == 'o') ADVANCE(54);
      END_STATE();
    case 15:
      if (lookahead == 'a') ADVANCE(55);
      END_STATE();
    case 16:
      if (lookahead == 'o') ADVANCE(56);
      if (lookahead == 'u') ADVANCE(57);
      END_STATE();
    case 17:
      if (lookahead == 'f') ADVANCE(58);
      END_STATE();
    case 18:
      if (lookahead == 't') ADVANCE(59);
      END_STATE();
    case 19:
      if (lookahead == 'e') ADVANCE(60);
      END_STATE();
    case 20:
      if (lookahead == 'h') ADVANCE(61);
      if (lookahead == 'i') ADVANCE(62);
      if (lookahead == 's') ADVANCE(63);
      if (lookahead == 't') ADVANCE(64);
      if (lookahead == 'w') ADVANCE(65);
      END_STATE();
    case 21:
      if (lookahead == 'h') ADVANCE(66);
      if (lookahead == 'r') ADVANCE(67);
      if (lookahead == 'y') ADVANCE(68);
      END_STATE();
    case 22:
      if (lookahead == 'i') ADVANCE(69);
      if (lookahead == 'n') ADVANCE(70);
      END_STATE();
    case 23:
      if (lookahead == 'o') ADVANCE(71);
      END_STATE();
    case 24:
      if (lookahead == 'h') ADVANCE(72);
      END_STATE();
    case 25:
      if (lookahead == 'L') ADVANCE(73);
      END_STATE();
    case 26:
      if (lookahead == 'L') ADVANCE(74);
      END_STATE();
    case 27:
      if (lookahead == 'U') ADVANCE(75);
      END_STATE();
    case 28:
      if (lookahead == '\n') SKIP(0);
      END_STATE();
    case 29:
      if (lookahead == 'l') ADVANCE(76);
      if (lookahead == 't') ADVANCE(77);
      END_STATE();
    case 30:
      if (lookahead == 'e') ADVANCE(78);
      END_STATE();
    case 31:
      if (lookahead == 'o') ADVANCE(79);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        'a', 80,
        'b', 81,
        'c', 82,
        'd', 83,
        'e', 84,
        'f', 85,
        'i', 86,
        'l', 87,
        'r', 88,
        's', 89,
        't', 90,
        'u', 91,
        'v', 92,
      );
      END_STATE();
    case 33:
      if (lookahead == 'l') ADVANCE(93);
      END_STATE();
    case 34:
      if (lookahead == 'n') ADVANCE(94);
      END_STATE();
    case 35:
      if (lookahead == 'i') ADVANCE(95);
      END_STATE();
    case 36:
      if (lookahead == 'm') ADVANCE(96);
      END_STATE();
    case 37:
      if (lookahead == 't') ADVANCE(97);
      END_STATE();
    case 38:
      if (lookahead == 'o') ADVANCE(98);
      END_STATE();
    case 39:
      if (lookahead == 'e') ADVANCE(99);
      END_STATE();
    case 40:
      if (lookahead == 's') ADVANCE(100);
      END_STATE();
    case 41:
      if (lookahead == 'a') ADVANCE(101);
      END_STATE();
    case 42:
      if (lookahead == 'n') ADVANCE(102);
      END_STATE();
    case 43:
      if (lookahead == 'f') ADVANCE(103);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'u') ADVANCE(104);
      END_STATE();
    case 45:
      if (lookahead == 's') ADVANCE(105);
      END_STATE();
    case 46:
      if (lookahead == 'u') ADVANCE(106);
      END_STATE();
    case 47:
      if (lookahead == 't') ADVANCE(107);
      END_STATE();
    case 48:
      if (lookahead == 'l') ADVANCE(108);
      END_STATE();
    case 49:
      if (lookahead == 'o') ADVANCE(109);
      END_STATE();
    case 50:
      if (lookahead == 'r') ADVANCE(110);
      END_STATE();
    case 51:
      if (lookahead == 't') ADVANCE(111);
      END_STATE();
    case 52:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 53:
      if (lookahead == 'l') ADVANCE(112);
      if (lookahead == 't') ADVANCE(113);
      END_STATE();
    case 54:
      if (lookahead == 'n') ADVANCE(114);
      END_STATE();
    case 55:
      if (lookahead == 'x') ADVANCE(115);
      END_STATE();
    case 56:
      if (lookahead == 'r') ADVANCE(116);
      END_STATE();
    case 57:
      if (lookahead == 'l') ADVANCE(117);
      END_STATE();
    case 58:
      if (lookahead == 'f') ADVANCE(118);
      END_STATE();
    case 59:
      if (lookahead == 'r') ADVANCE(119);
      END_STATE();
    case 60:
      if (lookahead == 'g') ADVANCE(120);
      if (lookahead == 's') ADVANCE(121);
      if (lookahead == 't') ADVANCE(122);
      END_STATE();
    case 61:
      if (lookahead == 'o') ADVANCE(123);
      END_STATE();
    case 62:
      if (lookahead == 'g') ADVANCE(124);
      if (lookahead == 'z') ADVANCE(125);
      END_STATE();
    case 63:
      if (lookahead == 'i') ADVANCE(126);
      END_STATE();
    case 64:
      if (lookahead == 'a') ADVANCE(127);
      if (lookahead == 'r') ADVANCE(128);
      END_STATE();
    case 65:
      if (lookahead == 'i') ADVANCE(129);
      END_STATE();
    case 66:
      if (lookahead == 'r') ADVANCE(130);
      END_STATE();
    case 67:
      if (lookahead == 'u') ADVANCE(131);
      END_STATE();
    case 68:
      if (lookahead == 'p') ADVANCE(132);
      END_STATE();
    case 69:
      if (lookahead == 'n') ADVANCE(133);
      END_STATE();
    case 70:
      if (lookahead == 'i') ADVANCE(134);
      if (lookahead == 's') ADVANCE(135);
      END_STATE();
    case 71:
      if (lookahead == 'i') ADVANCE(136);
      if (lookahead == 'l') ADVANCE(137);
      END_STATE();
    case 72:
      if (lookahead == 'i') ADVANCE(138);
      END_STATE();
    case 73:
      if (lookahead == 'S') ADVANCE(139);
      END_STATE();
    case 74:
      if (lookahead == 'L') ADVANCE(140);
      END_STATE();
    case 75:
      if (lookahead == 'E') ADVANCE(141);
      END_STATE();
    case 76:
      if (lookahead == 'i') ADVANCE(142);
      END_STATE();
    case 77:
      if (lookahead == 'o') ADVANCE(143);
      END_STATE();
    case 78:
      if (lookahead == 'n') ADVANCE(144);
      END_STATE();
    case 79:
      if (lookahead == 'n') ADVANCE(145);
      if (lookahead == 'r') ADVANCE(146);
      END_STATE();
    case 80:
      if (lookahead == 'l') ADVANCE(147);
      if (lookahead == 's') ADVANCE(148);
      if (lookahead == 't') ADVANCE(149);
      END_STATE();
    case 81:
      if (lookahead == 'a') ADVANCE(150);
      END_STATE();
    case 82:
      if (lookahead == 'd') ADVANCE(151);
      if (lookahead == 'l') ADVANCE(152);
      END_STATE();
    case 83:
      if (lookahead == 'e') ADVANCE(153);
      END_STATE();
    case 84:
      if (lookahead == 'x') ADVANCE(154);
      END_STATE();
    case 85:
      if (lookahead == 'a') ADVANCE(155);
      if (lookahead == 'i') ADVANCE(156);
      if (lookahead == 'o') ADVANCE(157);
      END_STATE();
    case 86:
      if (lookahead == 'n') ADVANCE(158);
      END_STATE();
    case 87:
      if (lookahead == 'e') ADVANCE(159);
      END_STATE();
    case 88:
      if (lookahead == 'e') ADVANCE(160);
      END_STATE();
    case 89:
      if (lookahead == 'p') ADVANCE(161);
      if (lookahead == 't') ADVANCE(162);
      END_STATE();
    case 90:
      if (lookahead == 'h') ADVANCE(163);
      if (lookahead == 'r') ADVANCE(164);
      END_STATE();
    case 91:
      if (lookahead == 'n') ADVANCE(165);
      if (lookahead == 'p') ADVANCE(166);
      END_STATE();
    case 92:
      if (lookahead == 'e') ADVANCE(167);
      if (lookahead == 'o') ADVANCE(168);
      END_STATE();
    case 93:
      if (lookahead == 'i') ADVANCE(169);
      END_STATE();
    case 94:
      if (lookahead == 'a') ADVANCE(170);
      END_STATE();
    case 95:
      if (lookahead == 'g') ADVANCE(171);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_asm);
      END_STATE();
    case 97:
      if (lookahead == 'o') ADVANCE(172);
      END_STATE();
    case 98:
      if (lookahead == 'l') ADVANCE(173);
      END_STATE();
    case 99:
      if (lookahead == 'a') ADVANCE(174);
      END_STATE();
    case 100:
      if (lookahead == 'e') ADVANCE(175);
      END_STATE();
    case 101:
      if (lookahead == 'r') ADVANCE(176);
      END_STATE();
    case 102:
      if (lookahead == 's') ADVANCE(177);
      if (lookahead == 't') ADVANCE(178);
      END_STATE();
    case 103:
      if (lookahead == 'a') ADVANCE(179);
      if (lookahead == 'i') ADVANCE(180);
      END_STATE();
    case 104:
      if (lookahead == 'b') ADVANCE(181);
      END_STATE();
    case 105:
      if (lookahead == 'e') ADVANCE(182);
      END_STATE();
    case 106:
      if (lookahead == 'm') ADVANCE(183);
      END_STATE();
    case 107:
      if (lookahead == 'e') ADVANCE(184);
      END_STATE();
    case 108:
      if (lookahead == 's') ADVANCE(185);
      END_STATE();
    case 109:
      if (lookahead == 'a') ADVANCE(186);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 111:
      if (lookahead == 'o') ADVANCE(187);
      END_STATE();
    case 112:
      if (lookahead == 'i') ADVANCE(188);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(sym_primitive_type);
      if (lookahead == '1') ADVANCE(189);
      if (lookahead == '3') ADVANCE(190);
      if (lookahead == '6') ADVANCE(191);
      if (lookahead == '8') ADVANCE(192);
      if (lookahead == 'p') ADVANCE(193);
      END_STATE();
    case 114:
      if (lookahead == 'g') ADVANCE(194);
      END_STATE();
    case 115:
      if (lookahead == '_') ADVANCE(195);
      END_STATE();
    case 116:
      if (lookahead == 'e') ADVANCE(196);
      END_STATE();
    case 117:
      if (lookahead == 'l') ADVANCE(197);
      END_STATE();
    case 118:
      if (lookahead == 's') ADVANCE(198);
      END_STATE();
    case 119:
      if (lookahead == 'd') ADVANCE(199);
      END_STATE();
    case 120:
      if (lookahead == 'i') ADVANCE(200);
      END_STATE();
    case 121:
      if (lookahead == 't') ADVANCE(201);
      END_STATE();
    case 122:
      if (lookahead == 'u') ADVANCE(202);
      END_STATE();
    case 123:
      if (lookahead == 'r') ADVANCE(203);
      END_STATE();
    case 124:
      if (lookahead == 'n') ADVANCE(204);
      END_STATE();
    case 125:
      if (lookahead == 'e') ADVANCE(205);
      END_STATE();
    case 126:
      if (lookahead == 'z') ADVANCE(206);
      END_STATE();
    case 127:
      if (lookahead == 't') ADVANCE(207);
      END_STATE();
    case 128:
      if (lookahead == 'u') ADVANCE(208);
      END_STATE();
    case 129:
      if (lookahead == 't') ADVANCE(209);
      END_STATE();
    case 130:
      if (lookahead == 'e') ADVANCE(210);
      END_STATE();
    case 131:
      if (lookahead == 'e') ADVANCE(141);
      END_STATE();
    case 132:
      if (lookahead == 'e') ADVANCE(211);
      END_STATE();
    case 133:
      if (lookahead == 't') ADVANCE(212);
      END_STATE();
    case 134:
      if (lookahead == 'o') ADVANCE(213);
      END_STATE();
    case 135:
      if (lookahead == 'i') ADVANCE(214);
      END_STATE();
    case 136:
      if (lookahead == 'd') ADVANCE(173);
      END_STATE();
    case 137:
      if (lookahead == 'a') ADVANCE(215);
      END_STATE();
    case 138:
      if (lookahead == 'l') ADVANCE(216);
      END_STATE();
    case 139:
      if (lookahead == 'E') ADVANCE(217);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_NULL);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(sym_true);
      END_STATE();
    case 142:
      if (lookahead == 'g') ADVANCE(218);
      END_STATE();
    case 143:
      if (lookahead == 'm') ADVANCE(219);
      END_STATE();
    case 144:
      if (lookahead == 'e') ADVANCE(220);
      END_STATE();
    case 145:
      if (lookahead == 'n') ADVANCE(221);
      END_STATE();
    case 146:
      if (lookahead == 'e') ADVANCE(222);
      END_STATE();
    case 147:
      if (lookahead == 'i') ADVANCE(223);
      END_STATE();
    case 148:
      if (lookahead == 'm') ADVANCE(224);
      END_STATE();
    case 149:
      if (lookahead == 't') ADVANCE(225);
      END_STATE();
    case 150:
      if (lookahead == 's') ADVANCE(226);
      END_STATE();
    case 151:
      if (lookahead == 'e') ADVANCE(227);
      END_STATE();
    case 152:
      if (lookahead == 'r') ADVANCE(228);
      END_STATE();
    case 153:
      if (lookahead == 'c') ADVANCE(229);
      END_STATE();
    case 154:
      if (lookahead == 'c') ADVANCE(230);
      if (lookahead == 't') ADVANCE(231);
      END_STATE();
    case 155:
      if (lookahead == 's') ADVANCE(232);
      END_STATE();
    case 156:
      if (lookahead == 'n') ADVANCE(233);
      END_STATE();
    case 157:
      if (lookahead == 'r') ADVANCE(234);
      END_STATE();
    case 158:
      if (lookahead == 'l') ADVANCE(235);
      END_STATE();
    case 159:
      if (lookahead == 'a') ADVANCE(236);
      END_STATE();
    case 160:
      if (lookahead == 's') ADVANCE(237);
      END_STATE();
    case 161:
      if (lookahead == 't') ADVANCE(238);
      END_STATE();
    case 162:
      if (lookahead == 'd') ADVANCE(239);
      END_STATE();
    case 163:
      if (lookahead == 'i') ADVANCE(240);
      if (lookahead == 'r') ADVANCE(241);
      END_STATE();
    case 164:
      if (lookahead == 'y') ADVANCE(242);
      END_STATE();
    case 165:
      if (lookahead == 'a') ADVANCE(243);
      END_STATE();
    case 166:
      if (lookahead == 't') ADVANCE(244);
      END_STATE();
    case 167:
      if (lookahead == 'c') ADVANCE(245);
      END_STATE();
    case 168:
      if (lookahead == 'l') ADVANCE(246);
      END_STATE();
    case 169:
      if (lookahead == 'g') ADVANCE(247);
      END_STATE();
    case 170:
      if (lookahead == 'l') ADVANCE(248);
      END_STATE();
    case 171:
      if (lookahead == 'n') ADVANCE(249);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_auto);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(sym_primitive_type);
      END_STATE();
    case 174:
      if (lookahead == 'k') ADVANCE(250);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(sym_primitive_type);
      if (lookahead == '1') ADVANCE(251);
      if (lookahead == '3') ADVANCE(252);
      if (lookahead == '6') ADVANCE(253);
      if (lookahead == '8') ADVANCE(254);
      if (lookahead == 'p') ADVANCE(255);
      END_STATE();
    case 177:
      if (lookahead == 't') ADVANCE(256);
      END_STATE();
    case 178:
      if (lookahead == 'i') ADVANCE(257);
      END_STATE();
    case 179:
      if (lookahead == 'u') ADVANCE(258);
      END_STATE();
    case 180:
      if (lookahead == 'n') ADVANCE(259);
      END_STATE();
    case 181:
      if (lookahead == 'l') ADVANCE(260);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 184:
      if (lookahead == 'r') ADVANCE(261);
      END_STATE();
    case 185:
      if (lookahead == 'e') ADVANCE(217);
      END_STATE();
    case 186:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_goto);
      END_STATE();
    case 188:
      if (lookahead == 'n') ADVANCE(262);
      END_STATE();
    case 189:
      if (lookahead == '6') ADVANCE(263);
      END_STATE();
    case 190:
      if (lookahead == '2') ADVANCE(264);
      END_STATE();
    case 191:
      if (lookahead == '4') ADVANCE(265);
      END_STATE();
    case 192:
      if (lookahead == '_') ADVANCE(266);
      END_STATE();
    case 193:
      if (lookahead == 't') ADVANCE(267);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_long);
      END_STATE();
    case 195:
      if (lookahead == 'a') ADVANCE(268);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(269);
      END_STATE();
    case 197:
      if (lookahead == 'p') ADVANCE(270);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(271);
      END_STATE();
    case 199:
      if (lookahead == 'i') ADVANCE(272);
      END_STATE();
    case 200:
      if (lookahead == 's') ADVANCE(273);
      END_STATE();
    case 201:
      if (lookahead == 'r') ADVANCE(274);
      END_STATE();
    case 202:
      if (lookahead == 'r') ADVANCE(275);
      END_STATE();
    case 203:
      if (lookahead == 't') ADVANCE(276);
      END_STATE();
    case 204:
      if (lookahead == 'e') ADVANCE(277);
      END_STATE();
    case 205:
      if (lookahead == '_') ADVANCE(278);
      if (lookahead == 'o') ADVANCE(279);
      END_STATE();
    case 206:
      if (lookahead == 'e') ADVANCE(280);
      END_STATE();
    case 207:
      if (lookahead == 'i') ADVANCE(281);
      END_STATE();
    case 208:
      if (lookahead == 'c') ADVANCE(282);
      END_STATE();
    case 209:
      if (lookahead == 'c') ADVANCE(283);
      END_STATE();
    case 210:
      if (lookahead == 'a') ADVANCE(284);
      END_STATE();
    case 211:
      if (lookahead == 'd') ADVANCE(285);
      END_STATE();
    case 212:
      if (lookahead == '1') ADVANCE(286);
      if (lookahead == '3') ADVANCE(287);
      if (lookahead == '6') ADVANCE(288);
      if (lookahead == '8') ADVANCE(289);
      if (lookahead == 'p') ADVANCE(290);
      END_STATE();
    case 213:
      if (lookahead == 'n') ADVANCE(291);
      END_STATE();
    case 214:
      if (lookahead == 'g') ADVANCE(292);
      END_STATE();
    case 215:
      if (lookahead == 't') ADVANCE(293);
      END_STATE();
    case 216:
      if (lookahead == 'e') ADVANCE(294);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(sym_false);
      END_STATE();
    case 218:
      if (lookahead == 'n') ADVANCE(295);
      END_STATE();
    case 219:
      if (lookahead == 'i') ADVANCE(296);
      END_STATE();
    case 220:
      if (lookahead == 'r') ADVANCE(297);
      END_STATE();
    case 221:
      if (lookahead == 'u') ADVANCE(298);
      END_STATE();
    case 222:
      if (lookahead == 't') ADVANCE(299);
      END_STATE();
    case 223:
      if (lookahead == 'g') ADVANCE(300);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym___asm);
      if (lookahead == '_') ADVANCE(301);
      END_STATE();
    case 225:
      if (lookahead == 'r') ADVANCE(302);
      END_STATE();
    case 226:
      if (lookahead == 'e') ADVANCE(303);
      END_STATE();
    case 227:
      if (lookahead == 'c') ADVANCE(304);
      END_STATE();
    case 228:
      if (lookahead == 'c') ADVANCE(305);
      END_STATE();
    case 229:
      if (lookahead == 'l') ADVANCE(306);
      END_STATE();
    case 230:
      if (lookahead == 'e') ADVANCE(307);
      END_STATE();
    case 231:
      if (lookahead == 'e') ADVANCE(308);
      END_STATE();
    case 232:
      if (lookahead == 't') ADVANCE(309);
      END_STATE();
    case 233:
      if (lookahead == 'a') ADVANCE(310);
      END_STATE();
    case 234:
      if (lookahead == 'c') ADVANCE(311);
      END_STATE();
    case 235:
      if (lookahead == 'i') ADVANCE(312);
      END_STATE();
    case 236:
      if (lookahead == 'v') ADVANCE(313);
      END_STATE();
    case 237:
      if (lookahead == 't') ADVANCE(314);
      END_STATE();
    case 238:
      if (lookahead == 'r') ADVANCE(315);
      END_STATE();
    case 239:
      if (lookahead == 'c') ADVANCE(316);
      END_STATE();
    case 240:
      if (lookahead == 's') ADVANCE(317);
      END_STATE();
    case 241:
      if (lookahead == 'e') ADVANCE(318);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym___try);
      END_STATE();
    case 243:
      if (lookahead == 'l') ADVANCE(319);
      END_STATE();
    case 244:
      if (lookahead == 'r') ADVANCE(320);
      END_STATE();
    case 245:
      if (lookahead == 't') ADVANCE(321);
      END_STATE();
    case 246:
      if (lookahead == 'a') ADVANCE(322);
      END_STATE();
    case 247:
      if (lookahead == 'n') ADVANCE(323);
      END_STATE();
    case 248:
      if (lookahead == 'i') ADVANCE(324);
      END_STATE();
    case 249:
      if (lookahead == 'a') ADVANCE(325);
      if (lookahead == 'o') ADVANCE(326);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 251:
      if (lookahead == '6') ADVANCE(327);
      END_STATE();
    case 252:
      if (lookahead == '2') ADVANCE(328);
      END_STATE();
    case 253:
      if (lookahead == '4') ADVANCE(329);
      END_STATE();
    case 254:
      if (lookahead == '_') ADVANCE(330);
      END_STATE();
    case 255:
      if (lookahead == 't') ADVANCE(331);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_const);
      if (lookahead == 'e') ADVANCE(332);
      END_STATE();
    case 257:
      if (lookahead == 'n') ADVANCE(333);
      END_STATE();
    case 258:
      if (lookahead == 'l') ADVANCE(334);
      END_STATE();
    case 259:
      if (lookahead == 'e') ADVANCE(335);
      END_STATE();
    case 260:
      if (lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 261:
      if (lookahead == 'n') ADVANCE(336);
      END_STATE();
    case 262:
      if (lookahead == 'e') ADVANCE(337);
      END_STATE();
    case 263:
      if (lookahead == '_') ADVANCE(338);
      END_STATE();
    case 264:
      if (lookahead == '_') ADVANCE(339);
      END_STATE();
    case 265:
      if (lookahead == '_') ADVANCE(340);
      END_STATE();
    case 266:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 267:
      if (lookahead == 'r') ADVANCE(341);
      END_STATE();
    case 268:
      if (lookahead == 'l') ADVANCE(342);
      END_STATE();
    case 269:
      if (lookahead == 'u') ADVANCE(343);
      END_STATE();
    case 270:
      if (lookahead == 't') ADVANCE(344);
      END_STATE();
    case 271:
      if (lookahead == 't') ADVANCE(345);
      END_STATE();
    case 272:
      if (lookahead == 'f') ADVANCE(346);
      END_STATE();
    case 273:
      if (lookahead == 't') ADVANCE(347);
      END_STATE();
    case 274:
      if (lookahead == 'i') ADVANCE(348);
      END_STATE();
    case 275:
      if (lookahead == 'n') ADVANCE(349);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(anon_sym_short);
      END_STATE();
    case 277:
      if (lookahead == 'd') ADVANCE(350);
      END_STATE();
    case 278:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 279:
      if (lookahead == 'f') ADVANCE(351);
      END_STATE();
    case 280:
      if (lookahead == '_') ADVANCE(352);
      END_STATE();
    case 281:
      if (lookahead == 'c') ADVANCE(353);
      END_STATE();
    case 282:
      if (lookahead == 't') ADVANCE(354);
      END_STATE();
    case 283:
      if (lookahead == 'h') ADVANCE(355);
      END_STATE();
    case 284:
      if (lookahead == 'd') ADVANCE(356);
      END_STATE();
    case 285:
      if (lookahead == 'e') ADVANCE(357);
      END_STATE();
    case 286:
      if (lookahead == '6') ADVANCE(358);
      END_STATE();
    case 287:
      if (lookahead == '2') ADVANCE(359);
      END_STATE();
    case 288:
      if (lookahead == '4') ADVANCE(360);
      END_STATE();
    case 289:
      if (lookahead == '_') ADVANCE(361);
      END_STATE();
    case 290:
      if (lookahead == 't') ADVANCE(362);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(anon_sym_union);
      END_STATE();
    case 292:
      if (lookahead == 'n') ADVANCE(363);
      END_STATE();
    case 293:
      if (lookahead == 'i') ADVANCE(364);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 295:
      if (lookahead == 'a') ADVANCE(365);
      if (lookahead == 'o') ADVANCE(366);
      END_STATE();
    case 296:
      if (lookahead == 'c') ADVANCE(367);
      END_STATE();
    case 297:
      if (lookahead == 'i') ADVANCE(368);
      END_STATE();
    case 298:
      if (lookahead == 'l') ADVANCE(369);
      END_STATE();
    case 299:
      if (lookahead == 'u') ADVANCE(370);
      END_STATE();
    case 300:
      if (lookahead == 'n') ADVANCE(371);
      END_STATE();
    case 301:
      if (lookahead == '_') ADVANCE(372);
      END_STATE();
    case 302:
      if (lookahead == 'i') ADVANCE(373);
      END_STATE();
    case 303:
      if (lookahead == 'd') ADVANCE(374);
      END_STATE();
    case 304:
      if (lookahead == 'l') ADVANCE(375);
      END_STATE();
    case 305:
      if (lookahead == 'a') ADVANCE(376);
      END_STATE();
    case 306:
      if (lookahead == 's') ADVANCE(377);
      END_STATE();
    case 307:
      if (lookahead == 'p') ADVANCE(378);
      END_STATE();
    case 308:
      if (lookahead == 'n') ADVANCE(379);
      END_STATE();
    case 309:
      if (lookahead == 'c') ADVANCE(380);
      END_STATE();
    case 310:
      if (lookahead == 'l') ADVANCE(381);
      END_STATE();
    case 311:
      if (lookahead == 'e') ADVANCE(382);
      END_STATE();
    case 312:
      if (lookahead == 'n') ADVANCE(383);
      END_STATE();
    case 313:
      if (lookahead == 'e') ADVANCE(384);
      END_STATE();
    case 314:
      if (lookahead == 'r') ADVANCE(385);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_ms_signed_ptr_modifier);
      END_STATE();
    case 316:
      if (lookahead == 'a') ADVANCE(386);
      END_STATE();
    case 317:
      if (lookahead == 'c') ADVANCE(387);
      END_STATE();
    case 318:
      if (lookahead == 'a') ADVANCE(388);
      END_STATE();
    case 319:
      if (lookahead == 'i') ADVANCE(389);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_ms_unsigned_ptr_modifier);
      END_STATE();
    case 321:
      if (lookahead == 'o') ADVANCE(390);
      END_STATE();
    case 322:
      if (lookahead == 't') ADVANCE(391);
      END_STATE();
    case 323:
      if (lookahead == 'o') ADVANCE(392);
      END_STATE();
    case 324:
      if (lookahead == 'g') ADVANCE(393);
      END_STATE();
    case 325:
      if (lookahead == 's') ADVANCE(394);
      END_STATE();
    case 326:
      if (lookahead == 'f') ADVANCE(395);
      END_STATE();
    case 327:
      if (lookahead == '_') ADVANCE(396);
      END_STATE();
    case 328:
      if (lookahead == '_') ADVANCE(397);
      END_STATE();
    case 329:
      if (lookahead == '_') ADVANCE(398);
      END_STATE();
    case 330:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 331:
      if (lookahead == 'r') ADVANCE(399);
      END_STATE();
    case 332:
      if (lookahead == 'x') ADVANCE(400);
      END_STATE();
    case 333:
      if (lookahead == 'u') ADVANCE(401);
      END_STATE();
    case 334:
      if (lookahead == 't') ADVANCE(402);
      END_STATE();
    case 335:
      if (lookahead == 'd') ADVANCE(403);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(anon_sym_extern);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(anon_sym_inline);
      END_STATE();
    case 338:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 339:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 340:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 341:
      if (lookahead == '_') ADVANCE(404);
      END_STATE();
    case 342:
      if (lookahead == 'i') ADVANCE(405);
      END_STATE();
    case 343:
      if (lookahead == 'r') ADVANCE(406);
      END_STATE();
    case 344:
      if (lookahead == 'r') ADVANCE(407);
      END_STATE();
    case 345:
      if (lookahead == 'o') ADVANCE(408);
      END_STATE();
    case 346:
      if (lookahead == 'f') ADVANCE(409);
      END_STATE();
    case 347:
      if (lookahead == 'e') ADVANCE(410);
      END_STATE();
    case 348:
      if (lookahead == 'c') ADVANCE(411);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(anon_sym_signed);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(anon_sym_sizeof);
      END_STATE();
    case 352:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(anon_sym_switch);
      END_STATE();
    case 356:
      if (lookahead == '_') ADVANCE(412);
      END_STATE();
    case 357:
      if (lookahead == 'f') ADVANCE(413);
      END_STATE();
    case 358:
      if (lookahead == '_') ADVANCE(414);
      END_STATE();
    case 359:
      if (lookahead == '_') ADVANCE(415);
      END_STATE();
    case 360:
      if (lookahead == '_') ADVANCE(416);
      END_STATE();
    case 361:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 362:
      if (lookahead == 'r') ADVANCE(417);
      END_STATE();
    case 363:
      if (lookahead == 'e') ADVANCE(418);
      END_STATE();
    case 364:
      if (lookahead == 'l') ADVANCE(419);
      END_STATE();
    case 365:
      if (lookahead == 's') ADVANCE(420);
      END_STATE();
    case 366:
      if (lookahead == 'f') ADVANCE(421);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(anon_sym__Atomic);
      END_STATE();
    case 368:
      if (lookahead == 'c') ADVANCE(422);
      END_STATE();
    case 369:
      if (lookahead == 'l') ADVANCE(423);
      END_STATE();
    case 370:
      if (lookahead == 'r') ADVANCE(424);
      END_STATE();
    case 371:
      if (lookahead == 'o') ADVANCE(425);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(anon_sym___asm__);
      END_STATE();
    case 373:
      if (lookahead == 'b') ADVANCE(426);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(anon_sym___based);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(anon_sym___cdecl);
      END_STATE();
    case 376:
      if (lookahead == 'l') ADVANCE(427);
      END_STATE();
    case 377:
      if (lookahead == 'p') ADVANCE(428);
      END_STATE();
    case 378:
      if (lookahead == 't') ADVANCE(429);
      END_STATE();
    case 379:
      if (lookahead == 's') ADVANCE(430);
      END_STATE();
    case 380:
      if (lookahead == 'a') ADVANCE(431);
      END_STATE();
    case 381:
      if (lookahead == 'l') ADVANCE(432);
      END_STATE();
    case 382:
      if (lookahead == 'i') ADVANCE(433);
      END_STATE();
    case 383:
      if (lookahead == 'e') ADVANCE(434);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(anon_sym___leave);
      END_STATE();
    case 385:
      if (lookahead == 'i') ADVANCE(435);
      END_STATE();
    case 386:
      if (lookahead == 'l') ADVANCE(436);
      END_STATE();
    case 387:
      if (lookahead == 'a') ADVANCE(437);
      END_STATE();
    case 388:
      if (lookahead == 'd') ADVANCE(438);
      END_STATE();
    case 389:
      if (lookahead == 'g') ADVANCE(439);
      END_STATE();
    case 390:
      if (lookahead == 'r') ADVANCE(440);
      END_STATE();
    case 391:
      if (lookahead == 'i') ADVANCE(441);
      END_STATE();
    case 392:
      if (lookahead == 'f') ADVANCE(442);
      END_STATE();
    case 393:
      if (lookahead == 'n') ADVANCE(443);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(anon_sym_alignas);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(anon_sym_alignof);
      END_STATE();
    case 396:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 397:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 398:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 399:
      if (lookahead == '_') ADVANCE(444);
      END_STATE();
    case 400:
      if (lookahead == 'p') ADVANCE(445);
      END_STATE();
    case 401:
      if (lookahead == 'e') ADVANCE(446);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(anon_sym_defined);
      END_STATE();
    case 404:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 405:
      if (lookahead == 'g') ADVANCE(447);
      END_STATE();
    case 406:
      if (lookahead == 'n') ADVANCE(448);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(anon_sym_nullptr);
      if (lookahead == '_') ADVANCE(449);
      END_STATE();
    case 408:
      if (lookahead == 'f') ADVANCE(450);
      END_STATE();
    case 409:
      if (lookahead == '_') ADVANCE(451);
      END_STATE();
    case 410:
      if (lookahead == 'r') ADVANCE(452);
      END_STATE();
    case 411:
      if (lookahead == 't') ADVANCE(453);
      END_STATE();
    case 412:
      if (lookahead == 'l') ADVANCE(454);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(anon_sym_typedef);
      END_STATE();
    case 414:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 415:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 416:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 417:
      if (lookahead == '_') ADVANCE(455);
      END_STATE();
    case 418:
      if (lookahead == 'd') ADVANCE(456);
      END_STATE();
    case 419:
      if (lookahead == 'e') ADVANCE(457);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(anon_sym__Alignas);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(anon_sym__Alignof);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(anon_sym__Generic);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(anon_sym__Nonnull);
      END_STATE();
    case 424:
      if (lookahead == 'n') ADVANCE(458);
      END_STATE();
    case 425:
      if (lookahead == 'f') ADVANCE(459);
      END_STATE();
    case 426:
      if (lookahead == 'u') ADVANCE(460);
      END_STATE();
    case 427:
      if (lookahead == 'l') ADVANCE(461);
      END_STATE();
    case 428:
      if (lookahead == 'e') ADVANCE(462);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(anon_sym___except);
      END_STATE();
    case 430:
      if (lookahead == 'i') ADVANCE(463);
      END_STATE();
    case 431:
      if (lookahead == 'l') ADVANCE(464);
      END_STATE();
    case 432:
      if (lookahead == 'y') ADVANCE(465);
      END_STATE();
    case 433:
      if (lookahead == 'n') ADVANCE(466);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(anon_sym___inline);
      if (lookahead == '_') ADVANCE(467);
      END_STATE();
    case 435:
      if (lookahead == 'c') ADVANCE(468);
      END_STATE();
    case 436:
      if (lookahead == 'l') ADVANCE(469);
      END_STATE();
    case 437:
      if (lookahead == 'l') ADVANCE(470);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(anon_sym___thread);
      END_STATE();
    case 439:
      if (lookahead == 'n') ADVANCE(471);
      END_STATE();
    case 440:
      if (lookahead == 'c') ADVANCE(472);
      END_STATE();
    case 441:
      if (lookahead == 'l') ADVANCE(473);
      END_STATE();
    case 442:
      ACCEPT_TOKEN(anon_sym__alignof);
      END_STATE();
    case 443:
      if (lookahead == 'e') ADVANCE(474);
      END_STATE();
    case 444:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 445:
      if (lookahead == 'r') ADVANCE(475);
      END_STATE();
    case 446:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 447:
      if (lookahead == 'n') ADVANCE(476);
      END_STATE();
    case 448:
      ACCEPT_TOKEN(anon_sym_noreturn);
      END_STATE();
    case 449:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 450:
      ACCEPT_TOKEN(anon_sym_offsetof);
      END_STATE();
    case 451:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 452:
      ACCEPT_TOKEN(anon_sym_register);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(anon_sym_restrict);
      END_STATE();
    case 454:
      if (lookahead == 'o') ADVANCE(477);
      END_STATE();
    case 455:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(anon_sym_unsigned);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(anon_sym_volatile);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(anon_sym__Noreturn);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(anon_sym___alignof);
      if (lookahead == '_') ADVANCE(478);
      END_STATE();
    case 460:
      if (lookahead == 't') ADVANCE(479);
      END_STATE();
    case 461:
      ACCEPT_TOKEN(anon_sym___clrcall);
      END_STATE();
    case 462:
      if (lookahead == 'c') ADVANCE(480);
      END_STATE();
    case 463:
      if (lookahead == 'o') ADVANCE(481);
      END_STATE();
    case 464:
      if (lookahead == 'l') ADVANCE(482);
      END_STATE();
    case 465:
      ACCEPT_TOKEN(anon_sym___finally);
      END_STATE();
    case 466:
      if (lookahead == 'l') ADVANCE(483);
      END_STATE();
    case 467:
      if (lookahead == '_') ADVANCE(484);
      END_STATE();
    case 468:
      if (lookahead == 't') ADVANCE(485);
      END_STATE();
    case 469:
      ACCEPT_TOKEN(anon_sym___stdcall);
      END_STATE();
    case 470:
      if (lookahead == 'l') ADVANCE(486);
      END_STATE();
    case 471:
      if (lookahead == 'e') ADVANCE(487);
      END_STATE();
    case 472:
      if (lookahead == 'a') ADVANCE(488);
      END_STATE();
    case 473:
      if (lookahead == 'e') ADVANCE(489);
      END_STATE();
    case 474:
      if (lookahead == 'd') ADVANCE(490);
      END_STATE();
    case 475:
      ACCEPT_TOKEN(anon_sym_constexpr);
      END_STATE();
    case 476:
      if (lookahead == '_') ADVANCE(491);
      END_STATE();
    case 477:
      if (lookahead == 'c') ADVANCE(492);
      END_STATE();
    case 478:
      if (lookahead == '_') ADVANCE(493);
      END_STATE();
    case 479:
      if (lookahead == 'e') ADVANCE(494);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(anon_sym___declspec);
      END_STATE();
    case 481:
      if (lookahead == 'n') ADVANCE(495);
      END_STATE();
    case 482:
      ACCEPT_TOKEN(anon_sym___fastcall);
      END_STATE();
    case 483:
      if (lookahead == 'i') ADVANCE(496);
      END_STATE();
    case 484:
      ACCEPT_TOKEN(anon_sym___inline__);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(sym_ms_restrict_modifier);
      if (lookahead == '_') ADVANCE(497);
      END_STATE();
    case 486:
      ACCEPT_TOKEN(anon_sym___thiscall);
      END_STATE();
    case 487:
      if (lookahead == 'd') ADVANCE(498);
      END_STATE();
    case 488:
      if (lookahead == 'l') ADVANCE(499);
      END_STATE();
    case 489:
      if (lookahead == '_') ADVANCE(500);
      END_STATE();
    case 490:
      ACCEPT_TOKEN(anon_sym__unaligned);
      END_STATE();
    case 491:
      if (lookahead == 't') ADVANCE(173);
      END_STATE();
    case 492:
      if (lookahead == 'a') ADVANCE(501);
      END_STATE();
    case 493:
      ACCEPT_TOKEN(anon_sym___alignof__);
      END_STATE();
    case 494:
      ACCEPT_TOKEN(anon_sym___attribute);
      if (lookahead == '_') ADVANCE(502);
      END_STATE();
    case 495:
      if (lookahead == '_') ADVANCE(503);
      END_STATE();
    case 496:
      if (lookahead == 'n') ADVANCE(504);
      END_STATE();
    case 497:
      if (lookahead == '_') ADVANCE(505);
      END_STATE();
    case 498:
      ACCEPT_TOKEN(anon_sym___unaligned);
      END_STATE();
    case 499:
      if (lookahead == 'l') ADVANCE(506);
      END_STATE();
    case 500:
      if (lookahead == '_') ADVANCE(507);
      END_STATE();
    case 501:
      if (lookahead == 'l') ADVANCE(508);
      END_STATE();
    case 502:
      if (lookahead == '_') ADVANCE(509);
      END_STATE();
    case 503:
      if (lookahead == '_') ADVANCE(510);
      END_STATE();
    case 504:
      if (lookahead == 'e') ADVANCE(511);
      END_STATE();
    case 505:
      ACCEPT_TOKEN(anon_sym___restrict__);
      END_STATE();
    case 506:
      ACCEPT_TOKEN(anon_sym___vectorcall);
      END_STATE();
    case 507:
      ACCEPT_TOKEN(anon_sym___volatile__);
      END_STATE();
    case 508:
      ACCEPT_TOKEN(anon_sym_thread_local);
      END_STATE();
    case 509:
      ACCEPT_TOKEN(anon_sym___attribute__);
      END_STATE();
    case 510:
      ACCEPT_TOKEN(anon_sym___extension__);
      END_STATE();
    case 511:
      ACCEPT_TOKEN(anon_sym___forceinline);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0},
  [1] = {.lex_state = 120},
  [2] = {.lex_state = 45},
  [3] = {.lex_state = 45},
  [4] = {.lex_state = 45},
  [5] = {.lex_state = 45},
  [6] = {.lex_state = 45},
  [7] = {.lex_state = 45},
  [8] = {.lex_state = 45},
  [9] = {.lex_state = 45},
  [10] = {.lex_state = 45},
  [11] = {.lex_state = 45},
  [12] = {.lex_state = 45},
  [13] = {.lex_state = 45},
  [14] = {.lex_state = 45},
  [15] = {.lex_state = 45},
  [16] = {.lex_state = 45},
  [17] = {.lex_state = 45},
  [18] = {.lex_state = 45},
  [19] = {.lex_state = 45},
  [20] = {.lex_state = 45},
  [21] = {.lex_state = 45},
  [22] = {.lex_state = 45},
  [23] = {.lex_state = 120},
  [24] = {.lex_state = 47},
  [25] = {.lex_state = 120},
  [26] = {.lex_state = 47},
  [27] = {.lex_state = 120},
  [28] = {.lex_state = 120},
  [29] = {.lex_state = 120},
  [30] = {.lex_state = 120},
  [31] = {.lex_state = 120},
  [32] = {.lex_state = 120},
  [33] = {.lex_state = 120},
  [34] = {.lex_state = 120},
  [35] = {.lex_state = 120},
  [36] = {.lex_state = 47},
  [37] = {.lex_state = 120},
  [38] = {.lex_state = 120},
  [39] = {.lex_state = 120},
  [40] = {.lex_state = 120},
  [41] = {.lex_state = 120},
  [42] = {.lex_state = 120},
  [43] = {.lex_state = 120},
  [44] = {.lex_state = 120},
  [45] = {.lex_state = 45},
  [46] = {.lex_state = 45},
  [47] = {.lex_state = 45},
  [48] = {.lex_state = 45},
  [49] = {.lex_state = 45},
  [50] = {.lex_state = 120},
  [51] = {.lex_state = 120},
  [52] = {.lex_state = 120},
  [53] = {.lex_state = 120},
  [54] = {.lex_state = 120},
  [55] = {.lex_state = 120},
  [56] = {.lex_state = 120},
  [57] = {.lex_state = 47},
  [58] = {.lex_state = 47},
  [59] = {.lex_state = 47},
  [60] = {.lex_state = 47},
  [61] = {.lex_state = 120},
  [62] = {.lex_state = 47},
  [63] = {.lex_state = 120},
  [64] = {.lex_state = 120},
  [65] = {.lex_state = 120},
  [66] = {.lex_state = 120},
  [67] = {.lex_state = 120},
  [68] = {.lex_state = 120},
  [69] = {.lex_state = 120},
  [70] = {.lex_state = 120},
  [71] = {.lex_state = 120},
  [72] = {.lex_state = 120},
  [73] = {.lex_state = 120},
  [74] = {.lex_state = 120},
  [75] = {.lex_state = 120},
  [76] = {.lex_state = 120},
  [77] = {.lex_state = 120},
  [78] = {.lex_state = 120},
  [79] = {.lex_state = 120},
  [80] = {.lex_state = 45},
  [81] = {.lex_state = 45},
  [82] = {.lex_state = 45},
  [83] = {.lex_state = 45},
  [84] = {.lex_state = 45},
  [85] = {.lex_state = 45},
  [86] = {.lex_state = 45},
  [87] = {.lex_state = 120},
  [88] = {.lex_state = 45},
  [89] = {.lex_state = 120},
  [90] = {.lex_state = 45},
  [91] = {.lex_state = 45},
  [92] = {.lex_state = 45},
  [93] = {.lex_state = 45},
  [94] = {.lex_state = 45},
  [95] = {.lex_state = 45},
  [96] = {.lex_state = 45},
  [97] = {.lex_state = 45},
  [98] = {.lex_state = 45},
  [99] = {.lex_state = 45},
  [100] = {.lex_state = 45},
  [101] = {.lex_state = 45},
  [102] = {.lex_state = 45},
  [103] = {.lex_state = 45},
  [104] = {.lex_state = 45},
  [105] = {.lex_state = 45},
  [106] = {.lex_state = 45},
  [107] = {.lex_state = 45},
  [108] = {.lex_state = 45},
  [109] = {.lex_state = 45},
  [110] = {.lex_state = 45},
  [111] = {.lex_state = 45},
  [112] = {.lex_state = 45},
  [113] = {.lex_state = 45},
  [114] = {.lex_state = 45},
  [115] = {.lex_state = 45},
  [116] = {.lex_state = 45},
  [117] = {.lex_state = 45},
  [118] = {.lex_state = 45},
  [119] = {.lex_state = 45},
  [120] = {.lex_state = 45},
  [121] = {.lex_state = 45},
  [122] = {.lex_state = 45},
  [123] = {.lex_state = 45},
  [124] = {.lex_state = 45},
  [125] = {.lex_state = 45},
  [126] = {.lex_state = 45},
  [127] = {.lex_state = 45},
  [128] = {.lex_state = 45},
  [129] = {.lex_state = 45},
  [130] = {.lex_state = 45},
  [131] = {.lex_state = 45},
  [132] = {.lex_state = 45},
  [133] = {.lex_state = 45},
  [134] = {.lex_state = 45},
  [135] = {.lex_state = 45},
  [136] = {.lex_state = 45},
  [137] = {.lex_state = 45},
  [138] = {.lex_state = 45},
  [139] = {.lex_state = 45},
  [140] = {.lex_state = 45},
  [141] = {.lex_state = 45},
  [142] = {.lex_state = 45},
  [143] = {.lex_state = 45},
  [144] = {.lex_state = 45},
  [145] = {.lex_state = 45},
  [146] = {.lex_state = 120},
  [147] = {.lex_state = 120},
  [148] = {.lex_state = 47},
  [149] = {.lex_state = 47},
  [150] = {.lex_state = 120},
  [151] = {.lex_state = 120},
  [152] = {.lex_state = 120},
  [153] = {.lex_state = 120},
  [154] = {.lex_state = 120},
  [155] = {.lex_state = 120},
  [156] = {.lex_state = 120},
  [157] = {.lex_state = 120},
  [158] = {.lex_state = 120},
  [159] = {.lex_state = 120},
  [160] = {.lex_state = 120},
  [161] = {.lex_state = 120},
  [162] = {.lex_state = 120},
  [163] = {.lex_state = 120},
  [164] = {.lex_state = 120},
  [165] = {.lex_state = 120},
  [166] = {.lex_state = 120},
  [167] = {.lex_state = 120},
  [168] = {.lex_state = 120},
  [169] = {.lex_state = 120},
  [170] = {.lex_state = 120},
  [171] = {.lex_state = 120},
  [172] = {.lex_state = 120},
  [173] = {.lex_state = 120},
  [174] = {.lex_state = 120},
  [175] = {.lex_state = 120},
  [176] = {.lex_state = 47},
  [177] = {.lex_state = 47},
  [178] = {.lex_state = 47},
  [179] = {.lex_state = 120},
  [180] = {.lex_state = 47},
  [181] = {.lex_state = 47},
  [182] = {.lex_state = 47},
  [183] = {.lex_state = 47},
  [184] = {.lex_state = 47},
  [185] = {.lex_state = 47},
  [186] = {.lex_state = 47},
  [187] = {.lex_state = 120},
  [188] = {.lex_state = 47},
  [189] = {.lex_state = 120},
  [190] = {.lex_state = 120},
  [191] = {.lex_state = 120},
  [192] = {.lex_state = 47},
  [193] = {.lex_state = 47},
  [194] = {.lex_state = 47},
  [195] = {.lex_state = 47},
  [196] = {.lex_state = 47},
  [197] = {.lex_state = 47},
  [198] = {.lex_state = 120},
  [199] = {.lex_state = 120},
  [200] = {.lex_state = 47},
  [201] = {.lex_state = 47},
  [202] = {.lex_state = 47},
  [203] = {.lex_state = 47},
  [204] = {.lex_state = 47},
  [205] = {.lex_state = 120},
  [206] = {.lex_state = 47},
  [207] = {.lex_state = 47},
  [208] = {.lex_state = 47},
  [209] = {.lex_state = 47},
  [210] = {.lex_state = 47},
  [211] = {.lex_state = 47},
  [212] = {.lex_state = 47},
  [213] = {.lex_state = 47},
  [214] = {.lex_state = 47},
  [215] = {.lex_state = 47},
  [216] = {.lex_state = 47},
  [217] = {.lex_state = 47},
  [218] = {.lex_state = 120},
  [219] = {.lex_state = 120},
  [220] = {.lex_state = 120},
  [221] = {.lex_state = 120},
  [222] = {.lex_state = 120},
  [223] = {.lex_state = 120},
  [224] = {.lex_state = 120},
  [225] = {.lex_state = 120},
  [226] = {.lex_state = 120},
  [227] = {.lex_state = 120},
  [228] = {.lex_state = 120},
  [229] = {.lex_state = 120},
  [230] = {.lex_state = 120},
  [231] = {.lex_state = 120},
  [232] = {.lex_state = 120},
  [233] = {.lex_state = 120},
  [234] = {.lex_state = 120},
  [235] = {.lex_state = 120},
  [236] = {.lex_state = 120},
  [237] = {.lex_state = 120},
  [238] = {.lex_state = 120},
  [239] = {.lex_state = 120},
  [240] = {.lex_state = 120},
  [241] = {.lex_state = 120},
  [242] = {.lex_state = 120},
  [243] = {.lex_state = 120},
  [244] = {.lex_state = 120},
  [245] = {.lex_state = 120},
  [246] = {.lex_state = 120},
  [247] = {.lex_state = 120},
  [248] = {.lex_state = 120},
  [249] = {.lex_state = 120},
  [250] = {.lex_state = 120},
  [251] = {.lex_state = 120},
  [252] = {.lex_state = 120},
  [253] = {.lex_state = 120},
  [254] = {.lex_state = 47},
  [255] = {.lex_state = 120},
  [256] = {.lex_state = 120},
  [257] = {.lex_state = 120},
  [258] = {.lex_state = 120},
  [259] = {.lex_state = 120},
  [260] = {.lex_state = 120},
  [261] = {.lex_state = 120},
  [262] = {.lex_state = 120},
  [263] = {.lex_state = 120},
  [264] = {.lex_state = 120},
  [265] = {.lex_state = 120},
  [266] = {.lex_state = 120},
  [267] = {.lex_state = 120},
  [268] = {.lex_state = 120},
  [269] = {.lex_state = 120},
  [270] = {.lex_state = 44},
  [271] = {.lex_state = 120},
  [272] = {.lex_state = 120},
  [273] = {.lex_state = 120},
  [274] = {.lex_state = 120},
  [275] = {.lex_state = 120},
  [276] = {.lex_state = 120},
  [277] = {.lex_state = 120},
  [278] = {.lex_state = 120},
  [279] = {.lex_state = 120},
  [280] = {.lex_state = 120},
  [281] = {.lex_state = 120},
  [282] = {.lex_state = 47},
  [283] = {.lex_state = 120},
  [284] = {.lex_state = 47},
  [285] = {.lex_state = 47},
  [286] = {.lex_state = 47},
  [287] = {.lex_state = 47},
  [288] = {.lex_state = 47},
  [289] = {.lex_state = 47},
  [290] = {.lex_state = 47},
  [291] = {.lex_state = 47},
  [292] = {.lex_state = 47},
  [293] = {.lex_state = 47},
  [294] = {.lex_state = 47},
  [295] = {.lex_state = 47},
  [296] = {.lex_state = 47},
  [297] = {.lex_state = 47},
  [298] = {.lex_state = 47},
  [299] = {.lex_state = 47},
  [300] = {.lex_state = 47},
  [301] = {.lex_state = 47},
  [302] = {.lex_state = 47},
  [303] = {.lex_state = 47},
  [304] = {.lex_state = 47},
  [305] = {.lex_state = 47},
  [306] = {.lex_state = 47},
  [307] = {.lex_state = 47},
  [308] = {.lex_state = 47},
  [309] = {.lex_state = 120},
  [310] = {.lex_state = 120},
  [311] = {.lex_state = 47},
  [312] = {.lex_state = 44},
  [313] = {.lex_state = 120},
  [314] = {.lex_state = 120},
  [315] = {.lex_state = 120},
  [316] = {.lex_state = 120},
  [317] = {.lex_state = 120},
  [318] = {.lex_state = 120},
  [319] = {.lex_state = 120},
  [320] = {.lex_state = 120},
  [321] = {.lex_state = 120},
  [322] = {.lex_state = 120},
  [323] = {.lex_state = 120},
  [324] = {.lex_state = 120},
  [325] = {.lex_state = 120},
  [326] = {.lex_state = 120},
  [327] = {.lex_state = 120},
  [328] = {.lex_state = 120},
  [329] = {.lex_state = 120},
  [330] = {.lex_state = 120},
  [331] = {.lex_state = 120},
  [332] = {.lex_state = 120},
  [333] = {.lex_state = 120},
  [334] = {.lex_state = 120},
  [335] = {.lex_state = 120},
  [336] = {.lex_state = 120},
  [337] = {.lex_state = 120},
  [338] = {.lex_state = 120},
  [339] = {.lex_state = 120},
  [340] = {.lex_state = 120},
  [341] = {.lex_state = 120},
  [342] = {.lex_state = 120},
  [343] = {.lex_state = 120},
  [344] = {.lex_state = 120},
  [345] = {.lex_state = 120},
  [346] = {.lex_state = 120},
  [347] = {.lex_state = 120},
  [348] = {.lex_state = 120},
  [349] = {.lex_state = 120},
  [350] = {.lex_state = 120},
  [351] = {.lex_state = 120},
  [352] = {.lex_state = 120},
  [353] = {.lex_state = 120},
  [354] = {.lex_state = 120},
  [355] = {.lex_state = 120},
  [356] = {.lex_state = 120},
  [357] = {.lex_state = 120},
  [358] = {.lex_state = 120},
  [359] = {.lex_state = 120},
  [360] = {.lex_state = 120},
  [361] = {.lex_state = 120},
  [362] = {.lex_state = 120},
  [363] = {.lex_state = 120},
  [364] = {.lex_state = 120},
  [365] = {.lex_state = 120},
  [366] = {.lex_state = 120},
  [367] = {.lex_state = 120},
  [368] = {.lex_state = 120},
  [369] = {.lex_state = 120},
  [370] = {.lex_state = 120},
  [371] = {.lex_state = 120},
  [372] = {.lex_state = 120},
  [373] = {.lex_state = 120},
  [374] = {.lex_state = 120},
  [375] = {.lex_state = 120},
  [376] = {.lex_state = 120},
  [377] = {.lex_state = 120},
  [378] = {.lex_state = 120},
  [379] = {.lex_state = 120},
  [380] = {.lex_state = 120},
  [381] = {.lex_state = 120},
  [382] = {.lex_state = 120},
  [383] = {.lex_state = 44},
  [384] = {.lex_state = 120},
  [385] = {.lex_state = 120},
  [386] = {.lex_state = 120},
  [387] = {.lex_state = 120},
  [388] = {.lex_state = 120},
  [389] = {.lex_state = 120},
  [390] = {.lex_state = 120},
  [391] = {.lex_state = 120},
  [392] = {.lex_state = 120},
  [393] = {.lex_state = 120},
  [394] = {.lex_state = 120},
  [395] = {.lex_state = 120},
  [396] = {.lex_state = 44},
  [397] = {.lex_state = 44},
  [398] = {.lex_state = 120},
  [399] = {.lex_state = 120},
  [400] = {.lex_state = 49},
  [401] = {.lex_state = 49},
  [402] = {.lex_state = 49},
  [403] = {.lex_state = 49},
  [404] = {.lex_state = 49},
  [405] = {.lex_state = 49},
  [406] = {.lex_state = 49},
  [407] = {.lex_state = 49},
  [408] = {.lex_state = 49},
  [409] = {.lex_state = 44},
  [410] = {.lex_state = 49},
  [411] = {.lex_state = 53},
  [412] = {.lex_state = 120},
  [413] = {.lex_state = 120},
  [414] = {.lex_state = 120},
  [415] = {.lex_state = 120},
  [416] = {.lex_state = 120},
  [417] = {.lex_state = 120},
  [418] = {.lex_state = 120},
  [419] = {.lex_state = 120},
  [420] = {.lex_state = 120},
  [421] = {.lex_state = 120},
  [422] = {.lex_state = 120},
  [423] = {.lex_state = 53},
  [424] = {.lex_state = 53},
  [425] = {.lex_state = 53},
  [426] = {.lex_state = 53},
  [427] = {.lex_state = 53},
  [428] = {.lex_state = 53},
  [429] = {.lex_state = 53},
  [430] = {.lex_state = 53},
  [431] = {.lex_state = 53},
  [432] = {.lex_state = 120},
  [433] = {.lex_state = 53},
  [434] = {.lex_state = 53},
  [435] = {.lex_state = 53},
  [436] = {.lex_state = 53},
  [437] = {.lex_state = 53},
  [438] = {.lex_state = 53},
  [439] = {.lex_state = 53},
  [440] = {.lex_state = 53},
  [441] = {.lex_state = 120},
  [442] = {.lex_state = 120},
  [443] = {.lex_state = 120},
  [444] = {.lex_state = 53},
  [445] = {.lex_state = 52},
  [446] = {.lex_state = 57},
  [447] = {.lex_state = 52},
  [448] = {.lex_state = 57},
  [449] = {.lex_state = 57},
  [450] = {.lex_state = 52},
  [451] = {.lex_state = 120},
  [452] = {.lex_state = 120},
  [453] = {.lex_state = 120},
  [454] = {.lex_state = 120},
  [455] = {.lex_state = 120},
  [456] = {.lex_state = 120},
  [457] = {.lex_state = 120},
  [458] = {.lex_state = 120},
  [459] = {.lex_state = 120},
  [460] = {.lex_state = 120},
  [461] = {.lex_state = 120},
  [462] = {.lex_state = 120},
  [463] = {.lex_state = 120},
  [464] = {.lex_state = 120},
  [465] = {.lex_state = 120},
  [466] = {.lex_state = 120},
  [467] = {.lex_state = 120},
  [468] = {.lex_state = 120},
  [469] = {.lex_state = 120},
  [470] = {.lex_state = 120},
  [471] = {.lex_state = 120},
  [472] = {.lex_state = 120},
  [473] = {.lex_state = 120},
  [474] = {.lex_state = 120},
  [475] = {.lex_state = 120},
  [476] = {.lex_state = 120},
  [477] = {.lex_state = 120},
  [478] = {.lex_state = 120},
  [479] = {.lex_state = 120},
  [480] = {.lex_state = 120},
  [481] = {.lex_state = 120},
  [482] = {.lex_state = 120},
  [483] = {.lex_state = 120},
  [484] = {.lex_state = 53},
  [485] = {.lex_state = 120},
  [486] = {.lex_state = 120},
  [487] = {.lex_state = 120},
  [488] = {.lex_state = 53},
  [489] = {.lex_state = 120},
  [490] = {.lex_state = 120},
  [491] = {.lex_state = 53},
  [492] = {.lex_state = 120},
  [493] = {.lex_state = 53},
  [494] = {.lex_state = 120},
  [495] = {.lex_state = 120},
  [496] = {.lex_state = 120},
  [497] = {.lex_state = 120},
  [498] = {.lex_state = 120},
  [499] = {.lex_state = 120},
  [500] = {.lex_state = 120},
  [501] = {.lex_state = 120},
  [502] = {.lex_state = 120},
  [503] = {.lex_state = 120},
  [504] = {.lex_state = 120},
  [505] = {.lex_state = 120},
  [506] = {.lex_state = 120},
  [507] = {.lex_state = 120},
  [508] = {.lex_state = 120},
  [509] = {.lex_state = 120},
  [510] = {.lex_state = 120},
  [511] = {.lex_state = 120},
  [512] = {.lex_state = 120},
  [513] = {.lex_state = 120},
  [514] = {.lex_state = 120},
  [515] = {.lex_state = 120},
  [516] = {.lex_state = 120},
  [517] = {.lex_state = 120},
  [518] = {.lex_state = 120},
  [519] = {.lex_state = 120},
  [520] = {.lex_state = 120},
  [521] = {.lex_state = 120},
  [522] = {.lex_state = 120},
  [523] = {.lex_state = 120},
  [524] = {.lex_state = 120},
  [525] = {.lex_state = 120},
  [526] = {.lex_state = 120},
  [527] = {.lex_state = 49},
  [528] = {.lex_state = 120},
  [529] = {.lex_state = 120},
  [530] = {.lex_state = 120},
  [531] = {.lex_state = 120},
  [532] = {.lex_state = 120},
  [533] = {.lex_state = 120},
  [534] = {.lex_state = 120},
  [535] = {.lex_state = 120},
  [536] = {.lex_state = 120},
  [537] = {.lex_state = 120},
  [538] = {.lex_state = 120},
  [539] = {.lex_state = 120},
  [540] = {.lex_state = 120},
  [541] = {.lex_state = 120},
  [542] = {.lex_state = 120},
  [543] = {.lex_state = 120},
  [544] = {.lex_state = 120},
  [545] = {.lex_state = 120},
  [546] = {.lex_state = 120},
  [547] = {.lex_state = 120},
  [548] = {.lex_state = 120},
  [549] = {.lex_state = 120},
  [550] = {.lex_state = 120},
  [551] = {.lex_state = 120},
  [552] = {.lex_state = 120},
  [553] = {.lex_state = 120},
  [554] = {.lex_state = 120},
  [555] = {.lex_state = 120},
  [556] = {.lex_state = 120},
  [557] = {.lex_state = 120},
  [558] = {.lex_state = 120},
  [559] = {.lex_state = 120},
  [560] = {.lex_state = 120},
  [561] = {.lex_state = 120},
  [562] = {.lex_state = 120},
  [563] = {.lex_state = 120},
  [564] = {.lex_state = 120},
  [565] = {.lex_state = 120},
  [566] = {.lex_state = 120},
  [567] = {.lex_state = 120},
  [568] = {.lex_state = 120},
  [569] = {.lex_state = 120},
  [570] = {.lex_state = 120},
  [571] = {.lex_state = 120},
  [572] = {.lex_state = 120},
  [573] = {.lex_state = 120},
  [574] = {.lex_state = 120},
  [575] = {.lex_state = 120},
  [576] = {.lex_state = 120},
  [577] = {.lex_state = 120},
  [578] = {.lex_state = 120},
  [579] = {.lex_state = 120},
  [580] = {.lex_state = 120},
  [581] = {.lex_state = 120},
  [582] = {.lex_state = 120},
  [583] = {.lex_state = 120},
  [584] = {.lex_state = 120},
  [585] = {.lex_state = 120},
  [586] = {.lex_state = 120},
  [587] = {.lex_state = 120},
  [588] = {.lex_state = 120},
  [589] = {.lex_state = 120},
  [590] = {.lex_state = 120},
  [591] = {.lex_state = 120},
  [592] = {.lex_state = 120},
  [593] = {.lex_state = 120},
  [594] = {.lex_state = 120},
  [595] = {.lex_state = 120},
  [596] = {.lex_state = 120},
  [597] = {.lex_state = 120},
  [598] = {.lex_state = 120},
  [599] = {.lex_state = 120},
  [600] = {.lex_state = 120},
  [601] = {.lex_state = 120},
  [602] = {.lex_state = 120},
  [603] = {.lex_state = 120},
  [604] = {.lex_state = 120},
  [605] = {.lex_state = 120},
  [606] = {.lex_state = 120},
  [607] = {.lex_state = 120},
  [608] = {.lex_state = 120},
  [609] = {.lex_state = 120},
  [610] = {.lex_state = 120},
  [611] = {.lex_state = 120},
  [612] = {.lex_state = 120},
  [613] = {.lex_state = 120},
  [614] = {.lex_state = 120},
  [615] = {.lex_state = 120},
  [616] = {.lex_state = 120},
  [617] = {.lex_state = 120},
  [618] = {.lex_state = 120},
  [619] = {.lex_state = 120},
  [620] = {.lex_state = 49},
  [621] = {.lex_state = 49},
  [622] = {.lex_state = 49},
  [623] = {.lex_state = 53},
  [624] = {.lex_state = 53},
  [625] = {.lex_state = 49},
  [626] = {.lex_state = 53},
  [627] = {.lex_state = 49},
  [628] = {.lex_state = 50},
  [629] = {.lex_state = 50},
  [630] = {.lex_state = 50},
  [631] = {.lex_state = 50},
  [632] = {.lex_state = 50},
  [633] = {.lex_state = 50},
  [634] = {.lex_state = 50},
  [635] = {.lex_state = 50},
  [636] = {.lex_state = 49},
  [637] = {.lex_state = 50},
  [638] = {.lex_state = 50},
  [639] = {.lex_state = 53},
  [640] = {.lex_state = 53},
  [641] = {.lex_state = 53},
  [642] = {.lex_state = 53},
  [643] = {.lex_state = 53},
  [644] = {.lex_state = 53},
  [645] = {.lex_state = 53},
  [646] = {.lex_state = 53},
  [647] = {.lex_state = 53},
  [648] = {.lex_state = 53},
  [649] = {.lex_state = 53},
  [650] = {.lex_state = 53},
  [651] = {.lex_state = 53},
  [652] = {.lex_state = 53},
  [653] = {.lex_state = 120},
  [654] = {.lex_state = 53},
  [655] = {.lex_state = 53},
  [656] = {.lex_state = 120},
  [657] = {.lex_state = 53},
  [658] = {.lex_state = 53},
  [659] = {.lex_state = 120},
  [660] = {.lex_state = 53},
  [661] = {.lex_state = 53},
  [662] = {.lex_state = 120},
  [663] = {.lex_state = 49},
  [664] = {.lex_state = 50},
  [665] = {.lex_state = 53},
  [666] = {.lex_state = 53},
  [667] = {.lex_state = 50},
  [668] = {.lex_state = 50},
  [669] = {.lex_state = 50},
  [670] = {.lex_state = 50},
  [671] = {.lex_state = 50},
  [672] = {.lex_state = 50},
  [673] = {.lex_state = 50},
  [674] = {.lex_state = 50},
  [675] = {.lex_state = 50},
  [676] = {.lex_state = 53},
  [677] = {.lex_state = 50},
  [678] = {.lex_state = 50},
  [679] = {.lex_state = 50},
  [680] = {.lex_state = 50},
  [681] = {.lex_state = 50},
  [682] = {.lex_state = 50},
  [683] = {.lex_state = 53},
  [684] = {.lex_state = 50},
  [685] = {.lex_state = 53},
  [686] = {.lex_state = 53},
  [687] = {.lex_state = 53},
  [688] = {.lex_state = 53},
  [689] = {.lex_state = 53},
  [690] = {.lex_state = 50},
  [691] = {.lex_state = 50},
  [692] = {.lex_state = 50},
  [693] = {.lex_state = 51},
  [694] = {.lex_state = 120},
  [695] = {.lex_state = 53},
  [696] = {.lex_state = 51},
  [697] = {.lex_state = 50},
  [698] = {.lex_state = 51},
  [699] = {.lex_state = 50},
  [700] = {.lex_state = 120},
  [701] = {.lex_state = 50},
  [702] = {.lex_state = 50},
  [703] = {.lex_state = 53},
  [704] = {.lex_state = 53},
  [705] = {.lex_state = 53},
  [706] = {.lex_state = 53},
  [707] = {.lex_state = 53},
  [708] = {.lex_state = 53},
  [709] = {.lex_state = 53},
  [710] = {.lex_state = 53},
  [711] = {.lex_state = 53},
  [712] = {.lex_state = 53},
  [713] = {.lex_state = 49},
  [714] = {.lex_state = 49},
  [715] = {.lex_state = 49},
  [716] = {.lex_state = 53},
  [717] = {.lex_state = 53},
  [718] = {.lex_state = 53},
  [719] = {.lex_state = 53},
  [720] = {.lex_state = 53},
  [721] = {.lex_state = 53},
  [722] = {.lex_state = 53},
  [723] = {.lex_state = 53},
  [724] = {.lex_state = 53},
  [725] = {.lex_state = 53},
  [726] = {.lex_state = 53},
  [727] = {.lex_state = 53},
  [728] = {.lex_state = 53},
  [729] = {.lex_state = 53},
  [730] = {.lex_state = 53},
  [731] = {.lex_state = 53},
  [732] = {.lex_state = 53},
  [733] = {.lex_state = 53},
  [734] = {.lex_state = 53},
  [735] = {.lex_state = 53},
  [736] = {.lex_state = 53},
  [737] = {.lex_state = 53},
  [738] = {.lex_state = 53},
  [739] = {.lex_state = 53},
  [740] = {.lex_state = 53},
  [741] = {.lex_state = 53},
  [742] = {.lex_state = 53},
  [743] = {.lex_state = 53},
  [744] = {.lex_state = 53},
  [745] = {.lex_state = 53},
  [746] = {.lex_state = 53},
  [747] = {.lex_state = 53},
  [748] = {.lex_state = 53},
  [749] = {.lex_state = 53},
  [750] = {.lex_state = 53},
  [751] = {.lex_state = 53},
  [752] = {.lex_state = 53},
  [753] = {.lex_state = 53},
  [754] = {.lex_state = 53},
  [755] = {.lex_state = 53},
  [756] = {.lex_state = 53},
  [757] = {.lex_state = 53},
  [758] = {.lex_state = 53},
  [759] = {.lex_state = 53},
  [760] = {.lex_state = 53},
  [761] = {.lex_state = 53},
  [762] = {.lex_state = 53},
  [763] = {.lex_state = 53},
  [764] = {.lex_state = 53},
  [765] = {.lex_state = 53},
  [766] = {.lex_state = 53},
  [767] = {.lex_state = 53},
  [768] = {.lex_state = 53},
  [769] = {.lex_state = 53},
  [770] = {.lex_state = 53},
  [771] = {.lex_state = 53},
  [772] = {.lex_state = 53},
  [773] = {.lex_state = 53},
  [774] = {.lex_state = 53},
  [775] = {.lex_state = 53},
  [776] = {.lex_state = 53},
  [777] = {.lex_state = 53},
  [778] = {.lex_state = 49},
  [779] = {.lex_state = 53},
  [780] = {.lex_state = 53},
  [781] = {.lex_state = 53},
  [782] = {.lex_state = 53},
  [783] = {.lex_state = 49},
  [784] = {.lex_state = 53},
  [785] = {.lex_state = 53},
  [786] = {.lex_state = 49},
  [787] = {.lex_state = 53},
  [788] = {.lex_state = 53},
  [789] = {.lex_state = 53},
  [790] = {.lex_state = 49},
  [791] = {.lex_state = 53},
  [792] = {.lex_state = 53},
  [793] = {.lex_state = 49},
  [794] = {.lex_state = 49},
  [795] = {.lex_state = 53},
  [796] = {.lex_state = 53},
  [797] = {.lex_state = 53},
  [798] = {.lex_state = 53},
  [799] = {.lex_state = 53},
  [800] = {.lex_state = 53},
  [801] = {.lex_state = 53},
  [802] = {.lex_state = 53},
  [803] = {.lex_state = 53},
  [804] = {.lex_state = 53},
  [805] = {.lex_state = 50},
  [806] = {.lex_state = 50},
  [807] = {.lex_state = 50},
  [808] = {.lex_state = 50},
  [809] = {.lex_state = 50},
  [810] = {.lex_state = 50},
  [811] = {.lex_state = 50},
  [812] = {.lex_state = 50},
  [813] = {.lex_state = 50},
  [814] = {.lex_state = 50},
  [815] = {.lex_state = 53},
  [816] = {.lex_state = 53},
  [817] = {.lex_state = 50},
  [818] = {.lex_state = 53},
  [819] = {.lex_state = 53},
  [820] = {.lex_state = 53},
  [821] = {.lex_state = 53},
  [822] = {.lex_state = 53},
  [823] = {.lex_state = 53},
  [824] = {.lex_state = 53},
  [825] = {.lex_state = 53},
  [826] = {.lex_state = 53},
  [827] = {.lex_state = 49},
  [828] = {.lex_state = 53},
  [829] = {.lex_state = 53},
  [830] = {.lex_state = 53},
  [831] = {.lex_state = 53},
  [832] = {.lex_state = 50},
  [833] = {.lex_state = 50},
  [834] = {.lex_state = 53},
  [835] = {.lex_state = 50},
  [836] = {.lex_state = 53},
  [837] = {.lex_state = 53},
  [838] = {.lex_state = 53},
  [839] = {.lex_state = 53},
  [840] = {.lex_state = 53},
  [841] = {.lex_state = 53},
  [842] = {.lex_state = 52},
  [843] = {.lex_state = 52},
  [844] = {.lex_state = 52},
  [845] = {.lex_state = 52},
  [846] = {.lex_state = 52},
  [847] = {.lex_state = 57},
  [848] = {.lex_state = 53},
  [849] = {.lex_state = 57},
  [850] = {.lex_state = 57},
  [851] = {.lex_state = 57},
  [852] = {.lex_state = 57},
  [853] = {.lex_state = 57},
  [854] = {.lex_state = 57},
  [855] = {.lex_state = 57},
  [856] = {.lex_state = 57},
  [857] = {.lex_state = 57},
  [858] = {.lex_state = 57},
  [859] = {.lex_state = 57},
  [860] = {.lex_state = 52},
  [861] = {.lex_state = 52},
  [862] = {.lex_state = 52},
  [863] = {.lex_state = 52},
  [864] = {.lex_state = 52},
  [865] = {.lex_state = 52},
  [866] = {.lex_state = 57},
  [867] = {.lex_state = 57},
  [868] = {.lex_state = 57},
  [869] = {.lex_state = 57},
  [870] = {.lex_state = 57},
  [871] = {.lex_state = 57},
  [872] = {.lex_state = 52},
  [873] = {.lex_state = 49},
  [874] = {.lex_state = 53},
  [875] = {.lex_state = 52},
  [876] = {.lex_state = 52},
  [877] = {.lex_state = 49},
  [878] = {.lex_state = 52},
  [879] = {.lex_state = 52},
  [880] = {.lex_state = 52},
  [881] = {.lex_state = 49},
  [882] = {.lex_state = 49},
  [883] = {.lex_state = 53},
  [884] = {.lex_state = 52},
  [885] = {.lex_state = 49},
  [886] = {.lex_state = 53},
  [887] = {.lex_state = 53},
  [888] = {.lex_state = 49},
  [889] = {.lex_state = 49},
  [890] = {.lex_state = 53},
  [891] = {.lex_state = 49},
  [892] = {.lex_state = 49},
  [893] = {.lex_state = 49},
  [894] = {.lex_state = 49},
  [895] = {.lex_state = 49},
  [896] = {.lex_state = 49},
  [897] = {.lex_state = 49},
  [898] = {.lex_state = 49},
  [899] = {.lex_state = 49},
  [900] = {.lex_state = 49},
  [901] = {.lex_state = 49},
  [902] = {.lex_state = 49},
  [903] = {.lex_state = 53},
  [904] = {.lex_state = 53},
  [905] = {.lex_state = 49},
  [906] = {.lex_state = 49},
  [907] = {.lex_state = 49},
  [908] = {.lex_state = 53},
  [909] = {.lex_state = 49},
  [910] = {.lex_state = 49},
  [911] = {.lex_state = 49},
  [912] = {.lex_state = 53},
  [913] = {.lex_state = 120},
  [914] = {.lex_state = 120},
  [915] = {.lex_state = 120},
  [916] = {.lex_state = 49},
  [917] = {.lex_state = 49},
  [918] = {.lex_state = 120},
  [919] = {.lex_state = 120},
  [920] = {.lex_state = 120},
  [921] = {.lex_state = 120},
  [922] = {.lex_state = 120},
  [923] = {.lex_state = 120},
  [924] = {.lex_state = 53},
  [925] = {.lex_state = 120},
  [926] = {.lex_state = 53},
  [927] = {.lex_state = 53},
  [928] = {.lex_state = 49},
  [929] = {.lex_state = 53},
  [930] = {.lex_state = 53},
  [931] = {.lex_state = 53},
  [932] = {.lex_state = 53},
  [933] = {.lex_state = 53},
  [934] = {.lex_state = 53},
  [935] = {.lex_state = 53},
  [936] = {.lex_state = 53},
  [937] = {.lex_state = 53},
  [938] = {.lex_state = 53},
  [939] = {.lex_state = 53},
  [940] = {.lex_state = 53},
  [941] = {.lex_state = 53},
  [942] = {.lex_state = 53},
  [943] = {.lex_state = 53},
  [944] = {.lex_state = 53},
  [945] = {.lex_state = 53},
  [946] = {.lex_state = 53},
  [947] = {.lex_state = 53},
  [948] = {.lex_state = 53},
  [949] = {.lex_state = 53},
  [950] = {.lex_state = 53},
  [951] = {.lex_state = 53},
  [952] = {.lex_state = 53},
  [953] = {.lex_state = 53},
  [954] = {.lex_state = 53},
  [955] = {.lex_state = 53},
  [956] = {.lex_state = 53},
  [957] = {.lex_state = 53},
  [958] = {.lex_state = 53},
  [959] = {.lex_state = 53},
  [960] = {.lex_state = 53},
  [961] = {.lex_state = 53},
  [962] = {.lex_state = 53},
  [963] = {.lex_state = 50},
  [964] = {.lex_state = 50},
  [965] = {.lex_state = 50},
  [966] = {.lex_state = 50},
  [967] = {.lex_state = 50},
  [968] = {.lex_state = 50},
  [969] = {.lex_state = 50},
  [970] = {.lex_state = 50},
  [971] = {.lex_state = 50},
  [972] = {.lex_state = 50},
  [973] = {.lex_state = 50},
  [974] = {.lex_state = 50},
  [975] = {.lex_state = 53},
  [976] = {.lex_state = 53},
  [977] = {.lex_state = 53},
  [978] = {.lex_state = 53},
  [979] = {.lex_state = 53},
  [980] = {.lex_state = 53},
  [981] = {.lex_state = 50},
  [982] = {.lex_state = 53},
  [983] = {.lex_state = 50},
  [984] = {.lex_state = 53},
  [985] = {.lex_state = 53},
  [986] = {.lex_state = 52},
  [987] = {.lex_state = 53},
  [988] = {.lex_state = 53},
  [989] = {.lex_state = 53},
  [990] = {.lex_state = 53},
  [991] = {.lex_state = 52},
  [992] = {.lex_state = 53},
  [993] = {.lex_state = 53},
  [994] = {.lex_state = 53},
  [995] = {.lex_state = 53},
  [996] = {.lex_state = 53},
  [997] = {.lex_state = 53},
  [998] = {.lex_state = 52},
  [999] = {.lex_state = 53},
  [1000] = {.lex_state = 53},
  [1001] = {.lex_state = 52},
  [1002] = {.lex_state = 53},
  [1003] = {.lex_state = 53},
  [1004] = {.lex_state = 53},
  [1005] = {.lex_state = 53},
  [1006] = {.lex_state = 53},
  [1007] = {.lex_state = 53},
  [1008] = {.lex_state = 53},
  [1009] = {.lex_state = 53},
  [1010] = {.lex_state = 53},
  [1011] = {.lex_state = 53},
  [1012] = {.lex_state = 49},
  [1013] = {.lex_state = 53},
  [1014] = {.lex_state = 53},
  [1015] = {.lex_state = 49},
  [1016] = {.lex_state = 49},
  [1017] = {.lex_state = 49},
  [1018] = {.lex_state = 49},
  [1019] = {.lex_state = 49},
  [1020] = {.lex_state = 49},
  [1021] = {.lex_state = 53},
  [1022] = {.lex_state = 49},
  [1023] = {.lex_state = 49},
  [1024] = {.lex_state = 49},
  [1025] = {.lex_state = 49},
  [1026] = {.lex_state = 53},
  [1027] = {.lex_state = 49},
  [1028] = {.lex_state = 49},
  [1029] = {.lex_state = 49},
  [1030] = {.lex_state = 49},
  [1031] = {.lex_state = 49},
  [1032] = {.lex_state = 49},
  [1033] = {.lex_state = 49},
  [1034] = {.lex_state = 49},
  [1035] = {.lex_state = 49},
  [1036] = {.lex_state = 49},
  [1037] = {.lex_state = 49},
  [1038] = {.lex_state = 49},
  [1039] = {.lex_state = 49},
  [1040] = {.lex_state = 49},
  [1041] = {.lex_state = 49},
  [1042] = {.lex_state = 49},
  [1043] = {.lex_state = 49},
  [1044] = {.lex_state = 53},
  [1045] = {.lex_state = 49},
  [1046] = {.lex_state = 49},
  [1047] = {.lex_state = 49},
  [1048] = {.lex_state = 49},
  [1049] = {.lex_state = 49},
  [1050] = {.lex_state = 53},
  [1051] = {.lex_state = 49},
  [1052] = {.lex_state = 49},
  [1053] = {.lex_state = 49},
  [1054] = {.lex_state = 49},
  [1055] = {.lex_state = 49},
  [1056] = {.lex_state = 49},
  [1057] = {.lex_state = 49},
  [1058] = {.lex_state = 49},
  [1059] = {.lex_state = 49},
  [1060] = {.lex_state = 49},
  [1061] = {.lex_state = 49},
  [1062] = {.lex_state = 49},
  [1063] = {.lex_state = 49},
  [1064] = {.lex_state = 49},
  [1065] = {.lex_state = 49},
  [1066] = {.lex_state = 49},
  [1067] = {.lex_state = 49},
  [1068] = {.lex_state = 53},
  [1069] = {.lex_state = 49},
  [1070] = {.lex_state = 49},
  [1071] = {.lex_state = 49},
  [1072] = {.lex_state = 49},
  [1073] = {.lex_state = 49},
  [1074] = {.lex_state = 53},
  [1075] = {.lex_state = 49},
  [1076] = {.lex_state = 49},
  [1077] = {.lex_state = 53},
  [1078] = {.lex_state = 49},
  [1079] = {.lex_state = 53},
  [1080] = {.lex_state = 49},
  [1081] = {.lex_state = 49},
  [1082] = {.lex_state = 49},
  [1083] = {.lex_state = 49},
  [1084] = {.lex_state = 49},
  [1085] = {.lex_state = 49},
  [1086] = {.lex_state = 53},
  [1087] = {.lex_state = 49},
  [1088] = {.lex_state = 49},
  [1089] = {.lex_state = 49},
  [1090] = {.lex_state = 49},
  [1091] = {.lex_state = 49},
  [1092] = {.lex_state = 49},
  [1093] = {.lex_state = 53},
  [1094] = {.lex_state = 49},
  [1095] = {.lex_state = 49},
  [1096] = {.lex_state = 49},
  [1097] = {.lex_state = 49},
  [1098] = {.lex_state = 53},
  [1099] = {.lex_state = 49},
  [1100] = {.lex_state = 49},
  [1101] = {.lex_state = 49},
  [1102] = {.lex_state = 53},
  [1103] = {.lex_state = 49},
  [1104] = {.lex_state = 53},
  [1105] = {.lex_state = 53},
  [1106] = {.lex_state = 53},
  [1107] = {.lex_state = 53},
  [1108] = {.lex_state = 53},
  [1109] = {.lex_state = 53},
  [1110] = {.lex_state = 53},
  [1111] = {.lex_state = 53},
  [1112] = {.lex_state = 53},
  [1113] = {.lex_state = 53},
  [1114] = {.lex_state = 53},
  [1115] = {.lex_state = 53},
  [1116] = {.lex_state = 53},
  [1117] = {.lex_state = 53},
  [1118] = {.lex_state = 53},
  [1119] = {.lex_state = 53},
  [1120] = {.lex_state = 53},
  [1121] = {.lex_state = 52},
  [1122] = {.lex_state = 53},
  [1123] = {.lex_state = 53},
  [1124] = {.lex_state = 53},
  [1125] = {.lex_state = 53},
  [1126] = {.lex_state = 53},
  [1127] = {.lex_state = 53},
  [1128] = {.lex_state = 53},
  [1129] = {.lex_state = 53},
  [1130] = {.lex_state = 53},
  [1131] = {.lex_state = 53},
  [1132] = {.lex_state = 53},
  [1133] = {.lex_state = 53},
  [1134] = {.lex_state = 53},
  [1135] = {.lex_state = 53},
  [1136] = {.lex_state = 53},
  [1137] = {.lex_state = 53},
  [1138] = {.lex_state = 53},
  [1139] = {.lex_state = 53},
  [1140] = {.lex_state = 53},
  [1141] = {.lex_state = 23},
  [1142] = {.lex_state = 53},
  [1143] = {.lex_state = 53},
  [1144] = {.lex_state = 53},
  [1145] = {.lex_state = 53},
  [1146] = {.lex_state = 48},
  [1147] = {.lex_state = 53},
  [1148] = {.lex_state = 53},
  [1149] = {.lex_state = 52},
  [1150] = {.lex_state = 48},
  [1151] = {.lex_state = 53},
  [1152] = {.lex_state = 52},
  [1153] = {.lex_state = 52},
  [1154] = {.lex_state = 53},
  [1155] = {.lex_state = 53},
  [1156] = {.lex_state = 53},
  [1157] = {.lex_state = 48},
  [1158] = {.lex_state = 53},
  [1159] = {.lex_state = 48},
  [1160] = {.lex_state = 48},
  [1161] = {.lex_state = 48},
  [1162] = {.lex_state = 48},
  [1163] = {.lex_state = 48},
  [1164] = {.lex_state = 48},
  [1165] = {.lex_state = 52},
  [1166] = {.lex_state = 48},
  [1167] = {.lex_state = 48},
  [1168] = {.lex_state = 48},
  [1169] = {.lex_state = 48},
  [1170] = {.lex_state = 48},
  [1171] = {.lex_state = 48},
  [1172] = {.lex_state = 48},
  [1173] = {.lex_state = 48},
  [1174] = {.lex_state = 48},
  [1175] = {.lex_state = 48},
  [1176] = {.lex_state = 48},
  [1177] = {.lex_state = 48},
  [1178] = {.lex_state = 48},
  [1179] = {.lex_state = 52},
  [1180] = {.lex_state = 48},
  [1181] = {.lex_state = 52},
  [1182] = {.lex_state = 52},
  [1183] = {.lex_state = 48},
  [1184] = {.lex_state = 48},
  [1185] = {.lex_state = 48},
  [1186] = {.lex_state = 48},
  [1187] = {.lex_state = 48},
  [1188] = {.lex_state = 48},
  [1189] = {.lex_state = 48},
  [1190] = {.lex_state = 48},
  [1191] = {.lex_state = 52},
  [1192] = {.lex_state = 48},
  [1193] = {.lex_state = 52},
  [1194] = {.lex_state = 48},
  [1195] = {.lex_state = 48},
  [1196] = {.lex_state = 52},
  [1197] = {.lex_state = 48},
  [1198] = {.lex_state = 48},
  [1199] = {.lex_state = 48},
  [1200] = {.lex_state = 48},
  [1201] = {.lex_state = 52},
  [1202] = {.lex_state = 48},
  [1203] = {.lex_state = 52},
  [1204] = {.lex_state = 52},
  [1205] = {.lex_state = 48},
  [1206] = {.lex_state = 48},
  [1207] = {.lex_state = 52},
  [1208] = {.lex_state = 52},
  [1209] = {.lex_state = 52},
  [1210] = {.lex_state = 52},
  [1211] = {.lex_state = 52},
  [1212] = {.lex_state = 52},
  [1213] = {.lex_state = 48},
  [1214] = {.lex_state = 52},
  [1215] = {.lex_state = 23},
  [1216] = {.lex_state = 23},
  [1217] = {.lex_state = 23},
  [1218] = {.lex_state = 23},
  [1219] = {.lex_state = 23},
  [1220] = {.lex_state = 23},
  [1221] = {.lex_state = 23},
  [1222] = {.lex_state = 23},
  [1223] = {.lex_state = 23},
  [1224] = {.lex_state = 23},
  [1225] = {.lex_state = 23},
  [1226] = {.lex_state = 23},
  [1227] = {.lex_state = 23},
  [1228] = {.lex_state = 23},
  [1229] = {.lex_state = 23},
  [1230] = {.lex_state = 23},
  [1231] = {.lex_state = 23},
  [1232] = {.lex_state = 52},
  [1233] = {.lex_state = 23},
  [1234] = {.lex_state = 23},
  [1235] = {.lex_state = 23},
  [1236] = {.lex_state = 23},
  [1237] = {.lex_state = 23},
  [1238] = {.lex_state = 23},
  [1239] = {.lex_state = 23},
  [1240] = {.lex_state = 53},
  [1241] = {.lex_state = 23},
  [1242] = {.lex_state = 23},
  [1243] = {.lex_state = 53},
  [1244] = {.lex_state = 23},
  [1245] = {.lex_state = 23},
  [1246] = {.lex_state = 23},
  [1247] = {.lex_state = 23},
  [1248] = {.lex_state = 53},
  [1249] = {.lex_state = 23},
  [1250] = {.lex_state = 23},
  [1251] = {.lex_state = 53},
  [1252] = {.lex_state = 23},
  [1253] = {.lex_state = 53},
  [1254] = {.lex_state = 53},
  [1255] = {.lex_state = 53},
  [1256] = {.lex_state = 53},
  [1257] = {.lex_state = 53},
  [1258] = {.lex_state = 53},
  [1259] = {.lex_state = 53},
  [1260] = {.lex_state = 53},
  [1261] = {.lex_state = 53},
  [1262] = {.lex_state = 53},
  [1263] = {.lex_state = 53},
  [1264] = {.lex_state = 53},
  [1265] = {.lex_state = 53},
  [1266] = {.lex_state = 53},
  [1267] = {.lex_state = 53},
  [1268] = {.lex_state = 53},
  [1269] = {.lex_state = 53},
  [1270] = {.lex_state = 53},
  [1271] = {.lex_state = 53},
  [1272] = {.lex_state = 53},
  [1273] = {.lex_state = 53},
  [1274] = {.lex_state = 53},
  [1275] = {.lex_state = 53},
  [1276] = {.lex_state = 53},
  [1277] = {.lex_state = 53},
  [1278] = {.lex_state = 53},
  [1279] = {.lex_state = 53},
  [1280] = {.lex_state = 53},
  [1281] = {.lex_state = 53},
  [1282] = {.lex_state = 53},
  [1283] = {.lex_state = 50},
  [1284] = {.lex_state = 53},
  [1285] = {.lex_state = 50},
  [1286] = {.lex_state = 50},
  [1287] = {.lex_state = 50},
  [1288] = {.lex_state = 53},
  [1289] = {.lex_state = 53},
  [1290] = {.lex_state = 53},
  [1291] = {.lex_state = 53},
  [1292] = {.lex_state = 53},
  [1293] = {.lex_state = 53},
  [1294] = {.lex_state = 53},
  [1295] = {.lex_state = 53},
  [1296] = {.lex_state = 56},
  [1297] = {.lex_state = 53},
  [1298] = {.lex_state = 53},
  [1299] = {.lex_state = 53},
  [1300] = {.lex_state = 53},
  [1301] = {.lex_state = 53},
  [1302] = {.lex_state = 53},
  [1303] = {.lex_state = 53},
  [1304] = {.lex_state = 53},
  [1305] = {.lex_state = 53},
  [1306] = {.lex_state = 53},
  [1307] = {.lex_state = 53},
  [1308] = {.lex_state = 56},
  [1309] = {.lex_state = 53},
  [1310] = {.lex_state = 53},
  [1311] = {.lex_state = 53},
  [1312] = {.lex_state = 53},
  [1313] = {.lex_state = 53},
  [1314] = {.lex_state = 53},
  [1315] = {.lex_state = 53},
  [1316] = {.lex_state = 53},
  [1317] = {.lex_state = 53},
  [1318] = {.lex_state = 53},
  [1319] = {.lex_state = 53},
  [1320] = {.lex_state = 53},
  [1321] = {.lex_state = 53},
  [1322] = {.lex_state = 53},
  [1323] = {.lex_state = 53},
  [1324] = {.lex_state = 50},
  [1325] = {.lex_state = 53},
  [1326] = {.lex_state = 53},
  [1327] = {.lex_state = 50},
  [1328] = {.lex_state = 49},
  [1329] = {.lex_state = 50},
  [1330] = {.lex_state = 50},
  [1331] = {.lex_state = 53},
  [1332] = {.lex_state = 56},
  [1333] = {.lex_state = 53},
  [1334] = {.lex_state = 50},
  [1335] = {.lex_state = 53},
  [1336] = {.lex_state = 53},
  [1337] = {.lex_state = 53},
  [1338] = {.lex_state = 53},
  [1339] = {.lex_state = 50},
  [1340] = {.lex_state = 50},
  [1341] = {.lex_state = 53},
  [1342] = {.lex_state = 50},
  [1343] = {.lex_state = 50},
  [1344] = {.lex_state = 53},
  [1345] = {.lex_state = 50},
  [1346] = {.lex_state = 50},
  [1347] = {.lex_state = 50},
  [1348] = {.lex_state = 53},
  [1349] = {.lex_state = 53},
  [1350] = {.lex_state = 53},
  [1351] = {.lex_state = 50},
  [1352] = {.lex_state = 53},
  [1353] = {.lex_state = 53},
  [1354] = {.lex_state = 50},
  [1355] = {.lex_state = 53},
  [1356] = {.lex_state = 120},
  [1357] = {.lex_state = 53},
  [1358] = {.lex_state = 53},
  [1359] = {.lex_state = 53},
  [1360] = {.lex_state = 0},
  [1361] = {.lex_state = 120},
  [1362] = {.lex_state = 120},
  [1363] = {.lex_state = 53},
  [1364] = {.lex_state = 120},
  [1365] = {.lex_state = 53},
  [1366] = {.lex_state = 53},
  [1367] = {.lex_state = 53},
  [1368] = {.lex_state = 0},
  [1369] = {.lex_state = 53},
  [1370] = {.lex_state = 53},
  [1371] = {.lex_state = 120},
  [1372] = {.lex_state = 120},
  [1373] = {.lex_state = 54},
  [1374] = {.lex_state = 53},
  [1375] = {.lex_state = 120},
  [1376] = {.lex_state = 120},
  [1377] = {.lex_state = 53},
  [1378] = {.lex_state = 53},
  [1379] = {.lex_state = 120},
  [1380] = {.lex_state = 120},
  [1381] = {.lex_state = 49},
  [1382] = {.lex_state = 53},
  [1383] = {.lex_state = 54},
  [1384] = {.lex_state = 120},
  [1385] = {.lex_state = 53},
  [1386] = {.lex_state = 53},
  [1387] = {.lex_state = 120},
  [1388] = {.lex_state = 54},
  [1389] = {.lex_state = 49},
  [1390] = {.lex_state = 120},
  [1391] = {.lex_state = 53},
  [1392] = {.lex_state = 54},
  [1393] = {.lex_state = 53},
  [1394] = {.lex_state = 53},
  [1395] = {.lex_state = 49},
  [1396] = {.lex_state = 53},
  [1397] = {.lex_state = 53},
  [1398] = {.lex_state = 53},
  [1399] = {.lex_state = 120},
  [1400] = {.lex_state = 53},
  [1401] = {.lex_state = 120},
  [1402] = {.lex_state = 53},
  [1403] = {.lex_state = 120},
  [1404] = {.lex_state = 53},
  [1405] = {.lex_state = 120},
  [1406] = {.lex_state = 120},
  [1407] = {.lex_state = 53},
  [1408] = {.lex_state = 53},
  [1409] = {.lex_state = 50},
  [1410] = {.lex_state = 53},
  [1411] = {.lex_state = 53},
  [1412] = {.lex_state = 120},
  [1413] = {.lex_state = 120},
  [1414] = {.lex_state = 50},
  [1415] = {.lex_state = 53},
  [1416] = {.lex_state = 120},
  [1417] = {.lex_state = 53},
  [1418] = {.lex_state = 120},
  [1419] = {.lex_state = 120},
  [1420] = {.lex_state = 50},
  [1421] = {.lex_state = 0},
  [1422] = {.lex_state = 120},
  [1423] = {.lex_state = 53},
  [1424] = {.lex_state = 53},
  [1425] = {.lex_state = 0},
  [1426] = {.lex_state = 53},
  [1427] = {.lex_state = 120},
  [1428] = {.lex_state = 53},
  [1429] = {.lex_state = 53},
  [1430] = {.lex_state = 120},
  [1431] = {.lex_state = 53},
  [1432] = {.lex_state = 53},
  [1433] = {.lex_state = 120},
  [1434] = {.lex_state = 53},
  [1435] = {.lex_state = 53},
  [1436] = {.lex_state = 53},
  [1437] = {.lex_state = 53},
  [1438] = {.lex_state = 53},
  [1439] = {.lex_state = 53},
  [1440] = {.lex_state = 53},
  [1441] = {.lex_state = 120},
  [1442] = {.lex_state = 120},
  [1443] = {.lex_state = 53},
  [1444] = {.lex_state = 53},
  [1445] = {.lex_state = 0},
  [1446] = {.lex_state = 120},
  [1447] = {.lex_state = 53},
  [1448] = {.lex_state = 53},
  [1449] = {.lex_state = 53},
  [1450] = {.lex_state = 53},
  [1451] = {.lex_state = 50},
  [1452] = {.lex_state = 53},
  [1453] = {.lex_state = 53},
  [1454] = {.lex_state = 53},
  [1455] = {.lex_state = 53},
  [1456] = {.lex_state = 53},
  [1457] = {.lex_state = 120},
  [1458] = {.lex_state = 53},
  [1459] = {.lex_state = 53},
  [1460] = {.lex_state = 120},
  [1461] = {.lex_state = 0},
  [1462] = {.lex_state = 53},
  [1463] = {.lex_state = 53},
  [1464] = {.lex_state = 53},
  [1465] = {.lex_state = 53},
  [1466] = {.lex_state = 0},
  [1467] = {.lex_state = 56},
  [1468] = {.lex_state = 56},
  [1469] = {.lex_state = 120},
  [1470] = {.lex_state = 120},
  [1471] = {.lex_state = 56},
  [1472] = {.lex_state = 56},
  [1473] = {.lex_state = 56},
  [1474] = {.lex_state = 120},
  [1475] = {.lex_state = 56},
  [1476] = {.lex_state = 50},
  [1477] = {.lex_state = 53},
  [1478] = {.lex_state = 53},
  [1479] = {.lex_state = 56},
  [1480] = {.lex_state = 56},
  [1481] = {.lex_state = 53},
  [1482] = {.lex_state = 0},
  [1483] = {.lex_state = 120},
  [1484] = {.lex_state = 53},
  [1485] = {.lex_state = 56},
  [1486] = {.lex_state = 56},
  [1487] = {.lex_state = 120},
  [1488] = {.lex_state = 56},
  [1489] = {.lex_state = 53},
  [1490] = {.lex_state = 0},
  [1491] = {.lex_state = 53},
  [1492] = {.lex_state = 53},
  [1493] = {.lex_state = 53},
  [1494] = {.lex_state = 50},
  [1495] = {.lex_state = 53},
  [1496] = {.lex_state = 53},
  [1497] = {.lex_state = 53},
  [1498] = {.lex_state = 53},
  [1499] = {.lex_state = 53},
  [1500] = {.lex_state = 53},
  [1501] = {.lex_state = 53},
  [1502] = {.lex_state = 53},
  [1503] = {.lex_state = 53},
  [1504] = {.lex_state = 53},
  [1505] = {.lex_state = 53},
  [1506] = {.lex_state = 53},
  [1507] = {.lex_state = 53},
  [1508] = {.lex_state = 53},
  [1509] = {.lex_state = 53},
  [1510] = {.lex_state = 53},
  [1511] = {.lex_state = 53},
  [1512] = {.lex_state = 45},
  [1513] = {.lex_state = 53},
  [1514] = {.lex_state = 53},
  [1515] = {.lex_state = 53},
  [1516] = {.lex_state = 53},
  [1517] = {.lex_state = 53},
  [1518] = {.lex_state = 53},
  [1519] = {.lex_state = 30},
  [1520] = {.lex_state = 32},
  [1521] = {.lex_state = 53},
  [1522] = {.lex_state = 30},
  [1523] = {.lex_state = 50},
  [1524] = {.lex_state = 33},
  [1525] = {.lex_state = 33},
  [1526] = {.lex_state = 50},
  [1527] = {.lex_state = 120},
  [1528] = {.lex_state = 33},
  [1529] = {.lex_state = 0},
  [1530] = {.lex_state = 30},
  [1531] = {.lex_state = 32},
  [1532] = {.lex_state = 32},
  [1533] = {.lex_state = 0},
  [1534] = {.lex_state = 33},
  [1535] = {.lex_state = 53},
  [1536] = {.lex_state = 53},
  [1537] = {.lex_state = 0},
  [1538] = {.lex_state = 32},
  [1539] = {.lex_state = 0},
  [1540] = {.lex_state = 0},
  [1541] = {.lex_state = 0},
  [1542] = {.lex_state = 53},
  [1543] = {.lex_state = 53},
  [1544] = {.lex_state = 33},
  [1545] = {.lex_state = 45},
  [1546] = {.lex_state = 50},
  [1547] = {.lex_state = 32},
  [1548] = {.lex_state = 53},
  [1549] = {.lex_state = 0},
  [1550] = {.lex_state = 53},
  [1551] = {.lex_state = 120},
  [1552] = {.lex_state = 0},
  [1553] = {.lex_state = 0},
  [1554] = {.lex_state = 32},
  [1555] = {.lex_state = 0},
  [1556] = {.lex_state = 53},
  [1557] = {.lex_state = 32},
  [1558] = {.lex_state = 33},
  [1559] = {.lex_state = 33},
  [1560] = {.lex_state = 50},
  [1561] = {.lex_state = 53},
  [1562] = {.lex_state = 30},
  [1563] = {.lex_state = 0},
  [1564] = {.lex_state = 53},
  [1565] = {.lex_state = 0},
  [1566] = {.lex_state = 0},
  [1567] = {.lex_state = 35},
  [1568] = {.lex_state = 0},
  [1569] = {.lex_state = 0},
  [1570] = {.lex_state = 45},
  [1571] = {.lex_state = 0},
  [1572] = {.lex_state = 0},
  [1573] = {.lex_state = 0},
  [1574] = {.lex_state = 45},
  [1575] = {.lex_state = 0},
  [1576] = {.lex_state = 0},
  [1577] = {.lex_state = 0},
  [1578] = {.lex_state = 0},
  [1579] = {.lex_state = 45},
  [1580] = {.lex_state = 45},
  [1581] = {.lex_state = 0},
  [1582] = {.lex_state = 0},
  [1583] = {.lex_state = 0},
  [1584] = {.lex_state = 0},
  [1585] = {.lex_state = 45},
  [1586] = {.lex_state = 0},
  [1587] = {.lex_state = 0},
  [1588] = {.lex_state = 0},
  [1589] = {.lex_state = 0},
  [1590] = {.lex_state = 0},
  [1591] = {.lex_state = 0},
  [1592] = {.lex_state = 0},
  [1593] = {.lex_state = 0},
  [1594] = {.lex_state = 0},
  [1595] = {.lex_state = 0},
  [1596] = {.lex_state = 0},
  [1597] = {.lex_state = 0},
  [1598] = {.lex_state = 0},
  [1599] = {.lex_state = 35},
  [1600] = {.lex_state = 0},
  [1601] = {.lex_state = 0},
  [1602] = {.lex_state = 35},
  [1603] = {.lex_state = 0},
  [1604] = {.lex_state = 45},
  [1605] = {.lex_state = 0},
  [1606] = {.lex_state = 0},
  [1607] = {.lex_state = 45},
  [1608] = {.lex_state = 0},
  [1609] = {.lex_state = 0},
  [1610] = {.lex_state = 0},
  [1611] = {.lex_state = 0},
  [1612] = {.lex_state = 0},
  [1613] = {.lex_state = 0},
  [1614] = {.lex_state = 0},
  [1615] = {.lex_state = 0},
  [1616] = {.lex_state = 0},
  [1617] = {.lex_state = 0},
  [1618] = {.lex_state = 30},
  [1619] = {.lex_state = 0},
  [1620] = {.lex_state = 0},
  [1621] = {.lex_state = 0},
  [1622] = {.lex_state = 0},
  [1623] = {.lex_state = 0},
  [1624] = {.lex_state = 0},
  [1625] = {.lex_state = 0},
  [1626] = {.lex_state = 0},
  [1627] = {.lex_state = 0},
  [1628] = {.lex_state = 0},
  [1629] = {.lex_state = 0},
  [1630] = {.lex_state = 0},
  [1631] = {.lex_state = 0},
  [1632] = {.lex_state = 0},
  [1633] = {.lex_state = 0},
  [1634] = {.lex_state = 0},
  [1635] = {.lex_state = 0},
  [1636] = {.lex_state = 0},
  [1637] = {.lex_state = 0},
  [1638] = {.lex_state = 0},
  [1639] = {.lex_state = 0},
  [1640] = {.lex_state = 0},
  [1641] = {.lex_state = 30},
  [1642] = {.lex_state = 0},
  [1643] = {.lex_state = 0},
  [1644] = {.lex_state = 53},
  [1645] = {.lex_state = 0},
  [1646] = {.lex_state = 53},
  [1647] = {.lex_state = 0},
  [1648] = {.lex_state = 0},
  [1649] = {.lex_state = 53},
  [1650] = {.lex_state = 0},
  [1651] = {.lex_state = 0},
  [1652] = {.lex_state = 0},
  [1653] = {.lex_state = 53},
  [1654] = {.lex_state = 0},
  [1655] = {.lex_state = 0},
  [1656] = {.lex_state = 0},
  [1657] = {.lex_state = 0},
  [1658] = {.lex_state = 0},
  [1659] = {.lex_state = 30},
  [1660] = {.lex_state = 0},
  [1661] = {.lex_state = 0},
  [1662] = {.lex_state = 0},
  [1663] = {.lex_state = 0},
  [1664] = {.lex_state = 0},
  [1665] = {.lex_state = 0},
  [1666] = {.lex_state = 53},
  [1667] = {.lex_state = 0},
  [1668] = {.lex_state = 0},
  [1669] = {.lex_state = 0},
  [1670] = {.lex_state = 53},
  [1671] = {.lex_state = 0},
  [1672] = {.lex_state = 0},
  [1673] = {.lex_state = 0},
  [1674] = {.lex_state = 53},
  [1675] = {.lex_state = 35},
  [1676] = {.lex_state = 0},
  [1677] = {.lex_state = 0},
  [1678] = {.lex_state = 0},
  [1679] = {.lex_state = 34},
  [1680] = {.lex_state = 34},
  [1681] = {.lex_state = 120},
  [1682] = {.lex_state = 34},
  [1683] = {.lex_state = 0},
  [1684] = {.lex_state = 34},
  [1685] = {.lex_state = 34},
  [1686] = {.lex_state = 0},
  [1687] = {.lex_state = 34},
  [1688] = {.lex_state = 0},
  [1689] = {.lex_state = 120},
  [1690] = {.lex_state = 0},
  [1691] = {.lex_state = 120},
  [1692] = {.lex_state = 34},
  [1693] = {.lex_state = 120},
  [1694] = {.lex_state = 45},
  [1695] = {.lex_state = 0},
  [1696] = {.lex_state = 34},
  [1697] = {.lex_state = 120},
  [1698] = {.lex_state = 0},
  [1699] = {.lex_state = 45},
  [1700] = {.lex_state = 0},
  [1701] = {.lex_state = 0},
  [1702] = {.lex_state = 53},
  [1703] = {.lex_state = 0},
  [1704] = {.lex_state = 34},
  [1705] = {.lex_state = 34},
  [1706] = {.lex_state = 0},
  [1707] = {.lex_state = 53},
  [1708] = {.lex_state = 0},
  [1709] = {.lex_state = 0},
  [1710] = {.lex_state = 34},
  [1711] = {.lex_state = 120},
  [1712] = {.lex_state = 0},
  [1713] = {.lex_state = 120},
  [1714] = {.lex_state = 120},
  [1715] = {.lex_state = 34},
  [1716] = {.lex_state = 53},
  [1717] = {.lex_state = 34},
  [1718] = {.lex_state = 0},
  [1719] = {.lex_state = 53},
  [1720] = {.lex_state = 120},
  [1721] = {.lex_state = 120},
  [1722] = {.lex_state = 53},
  [1723] = {.lex_state = 34},
  [1724] = {.lex_state = 0},
  [1725] = {.lex_state = 0},
  [1726] = {.lex_state = 34},
  [1727] = {.lex_state = 34},
  [1728] = {.lex_state = 0},
  [1729] = {.lex_state = 120},
  [1730] = {.lex_state = 0},
  [1731] = {.lex_state = 120},
  [1732] = {.lex_state = 0},
  [1733] = {.lex_state = 120},
  [1734] = {.lex_state = 0},
  [1735] = {.lex_state = 45},
  [1736] = {.lex_state = 0},
  [1737] = {.lex_state = 0},
  [1738] = {.lex_state = 53},
  [1739] = {.lex_state = 120},
  [1740] = {.lex_state = 120},
  [1741] = {.lex_state = 120},
  [1742] = {.lex_state = 0},
  [1743] = {.lex_state = 53},
  [1744] = {.lex_state = 0},
  [1745] = {.lex_state = 120},
  [1746] = {.lex_state = 120},
  [1747] = {.lex_state = 0},
  [1748] = {.lex_state = 0},
  [1749] = {.lex_state = 120},
  [1750] = {.lex_state = 34},
  [1751] = {.lex_state = 0},
  [1752] = {.lex_state = 120},
  [1753] = {.lex_state = 120},
  [1754] = {.lex_state = 0},
  [1755] = {.lex_state = 120},
  [1756] = {.lex_state = 120},
  [1757] = {.lex_state = 120},
  [1758] = {.lex_state = 0},
  [1759] = {.lex_state = 0},
  [1760] = {.lex_state = 0},
  [1761] = {.lex_state = 0},
  [1762] = {.lex_state = 0},
  [1763] = {.lex_state = 53},
  [1764] = {.lex_state = 0},
  [1765] = {.lex_state = 0},
  [1766] = {.lex_state = 120},
  [1767] = {.lex_state = 44},
  [1768] = {.lex_state = 0},
  [1769] = {.lex_state = 35},
  [1770] = {.lex_state = 0},
  [1771] = {.lex_state = 0},
  [1772] = {.lex_state = 0},
  [1773] = {.lex_state = 35},
  [1774] = {.lex_state = 35},
  [1775] = {.lex_state = 0},
  [1776] = {.lex_state = 0},
  [1777] = {.lex_state = 0},
  [1778] = {.lex_state = 35},
  [1779] = {.lex_state = 0},
  [1780] = {.lex_state = 35},
  [1781] = {.lex_state = 0},
  [1782] = {.lex_state = 0},
  [1783] = {.lex_state = 35},
  [1784] = {.lex_state = 35},
  [1785] = {.lex_state = 35},
  [1786] = {.lex_state = 120},
  [1787] = {.lex_state = 44},
  [1788] = {.lex_state = 44},
  [1789] = {.lex_state = 35},
  [1790] = {.lex_state = 53},
  [1791] = {.lex_state = 44},
  [1792] = {.lex_state = 35},
  [1793] = {.lex_state = 0},
  [1794] = {.lex_state = 53},
  [1795] = {.lex_state = 35},
  [1796] = {.lex_state = 35},
  [1797] = {.lex_state = 53},
  [1798] = {.lex_state = 35},
  [1799] = {.lex_state = 44},
  [1800] = {.lex_state = 0},
  [1801] = {.lex_state = 0},
  [1802] = {.lex_state = 0},
  [1803] = {.lex_state = 44},
  [1804] = {.lex_state = 0},
  [1805] = {.lex_state = 35},
  [1806] = {.lex_state = 35},
  [1807] = {.lex_state = 0},
  [1808] = {.lex_state = 35},
  [1809] = {.lex_state = 0},
  [1810] = {.lex_state = 53},
  [1811] = {.lex_state = 53},
  [1812] = {.lex_state = 0},
  [1813] = {.lex_state = 0},
  [1814] = {.lex_state = 0},
  [1815] = {.lex_state = 53},
  [1816] = {.lex_state = 35},
  [1817] = {.lex_state = 0},
  [1818] = {.lex_state = 0},
  [1819] = {.lex_state = 53},
  [1820] = {.lex_state = 0},
  [1821] = {.lex_state = 44},
  [1822] = {.lex_state = 44},
  [1823] = {.lex_state = 0},
  [1824] = {.lex_state = 0},
  [1825] = {.lex_state = 0},
  [1826] = {.lex_state = 35},
  [1827] = {.lex_state = 53},
  [1828] = {.lex_state = 44},
  [1829] = {.lex_state = 0},
  [1830] = {.lex_state = 44},
  [1831] = {.lex_state = 0},
  [1832] = {.lex_state = 53},
  [1833] = {.lex_state = 0},
  [1834] = {.lex_state = 0},
  [1835] = {.lex_state = 35},
  [1836] = {.lex_state = 44},
  [1837] = {.lex_state = 0},
  [1838] = {.lex_state = 44},
  [1839] = {.lex_state = 53},
  [1840] = {.lex_state = 44},
  [1841] = {.lex_state = 53},
  [1842] = {.lex_state = 44},
  [1843] = {.lex_state = 0},
  [1844] = {.lex_state = 44},
  [1845] = {.lex_state = 53},
  [1846] = {.lex_state = 120},
  [1847] = {.lex_state = 35},
  [1848] = {.lex_state = 120},
  [1849] = {.lex_state = 0},
  [1850] = {.lex_state = 44},
  [1851] = {.lex_state = 0},
  [1852] = {.lex_state = 0},
  [1853] = {.lex_state = 44},
  [1854] = {.lex_state = 53},
  [1855] = {.lex_state = 0},
  [1856] = {.lex_state = 0},
  [1857] = {.lex_state = 0},
  [1858] = {.lex_state = 0},
  [1859] = {.lex_state = 53},
  [1860] = {.lex_state = 44},
  [1861] = {.lex_state = 0},
  [1862] = {.lex_state = 0},
  [1863] = {.lex_state = 44},
  [1864] = {.lex_state = 53},
  [1865] = {.lex_state = 0},
  [1866] = {.lex_state = 0},
  [1867] = {.lex_state = 44},
  [1868] = {.lex_state = 0},
  [1869] = {.lex_state = 0},
  [1870] = {.lex_state = 0},
  [1871] = {.lex_state = 0},
  [1872] = {.lex_state = 44},
  [1873] = {.lex_state = 0},
  [1874] = {.lex_state = 0},
  [1875] = {.lex_state = 0},
  [1876] = {.lex_state = 53},
  [1877] = {.lex_state = 35},
  [1878] = {.lex_state = 0},
  [1879] = {.lex_state = 53},
  [1880] = {.lex_state = 44},
  [1881] = {.lex_state = 0},
  [1882] = {.lex_state = 120},
  [1883] = {.lex_state = 0},
  [1884] = {.lex_state = 35},
  [1885] = {.lex_state = 0},
  [1886] = {.lex_state = 53},
  [1887] = {.lex_state = 0},
  [1888] = {.lex_state = 53},
  [1889] = {.lex_state = 0},
  [1890] = {.lex_state = 53},
  [1891] = {.lex_state = 44},
  [1892] = {.lex_state = 44},
  [1893] = {.lex_state = 53},
  [1894] = {.lex_state = 44},
  [1895] = {.lex_state = 44},
  [1896] = {.lex_state = 0},
  [1897] = {.lex_state = 0},
  [1898] = {.lex_state = 120},
  [1899] = {.lex_state = 120},
  [1900] = {.lex_state = 0},
  [1901] = {.lex_state = 0},
  [1902] = {.lex_state = 53},
  [1903] = {.lex_state = 53},
  [1904] = {.lex_state = 0},
  [1905] = {.lex_state = 44},
  [1906] = {.lex_state = 0},
  [1907] = {.lex_state = 35},
  [1908] = {.lex_state = 44},
  [1909] = {.lex_state = 44},
  [1910] = {.lex_state = 0},
  [1911] = {.lex_state = 53},
  [1912] = {.lex_state = 120},
  [1913] = {.lex_state = 44},
  [1914] = {.lex_state = 35},
  [1915] = {.lex_state = 44},
  [1916] = {.lex_state = 44},
  [1917] = {.lex_state = 44},
  [1918] = {.lex_state = 44},
  [1919] = {.lex_state = 0},
  [1920] = {.lex_state = 0},
  [1921] = {.lex_state = 44},
  [1922] = {.lex_state = 0},
  [1923] = {.lex_state = 35},
  [1924] = {.lex_state = 53},
  [1925] = {.lex_state = 44},
  [1926] = {.lex_state = 0},
  [1927] = {.lex_state = 44},
  [1928] = {.lex_state = 53},
  [1929] = {.lex_state = 0},
  [1930] = {.lex_state = 0},
  [1931] = {.lex_state = 0},
  [1932] = {.lex_state = 44},
  [1933] = {.lex_state = 44},
  [1934] = {.lex_state = 0},
  [1935] = {.lex_state = 53},
  [1936] = {.lex_state = 53},
  [1937] = {.lex_state = 0},
  [1938] = {.lex_state = 0},
  [1939] = {.lex_state = 35},
  [1940] = {.lex_state = 0},
  [1941] = {.lex_state = 53},
  [1942] = {.lex_state = 0},
  [1943] = {.lex_state = 0},
  [1944] = {.lex_state = 0},
  [1945] = {.lex_state = 0},
  [1946] = {.lex_state = 120},
  [1947] = {.lex_state = 35},
  [1948] = {.lex_state = 44},
  [1949] = {.lex_state = 44},
  [1950] = {.lex_state = 44},
  [1951] = {.lex_state = 0},
  [1952] = {.lex_state = 53},
  [1953] = {.lex_state = 0},
  [1954] = {.lex_state = 0},
  [1955] = {.lex_state = 0},
  [1956] = {.lex_state = 0},
  [1957] = {.lex_state = 0},
  [1958] = {.lex_state = 53},
  [1959] = {.lex_state = 0},
  [1960] = {.lex_state = 53},
  [1961] = {.lex_state = 35},
  [1962] = {.lex_state = 53},
  [1963] = {.lex_state = 53},
  [1964] = {.lex_state = 0},
  [1965] = {.lex_state = 53},
  [1966] = {.lex_state = 44},
  [1967] = {.lex_state = 0},
  [1968] = {.lex_state = 44},
  [1969] = {.lex_state = 120},
  [1970] = {.lex_state = 120},
  [1971] = {.lex_state = 0},
  [1972] = {.lex_state = 120},
  [1973] = {.lex_state = 0},
  [1974] = {.lex_state = 53},
  [1975] = {.lex_state = 0},
  [1976] = {.lex_state = 0},
  [1977] = {.lex_state = 0},
  [1978] = {.lex_state = 44},
  [1979] = {.lex_state = 44},
  [1980] = {.lex_state = 53},
  [1981] = {.lex_state = 0},
  [1982] = {.lex_state = 44},
  [1983] = {.lex_state = 0},
  [1984] = {.lex_state = 0},
  [1985] = {.lex_state = 53},
  [1986] = {.lex_state = 0},
  [1987] = {.lex_state = 0},
  [1988] = {.lex_state = 0},
  [1989] = {.lex_state = 0},
  [1990] = {.lex_state = 0},
  [1991] = {.lex_state = 0},
  [1992] = {.lex_state = 120},
  [1993] = {.lex_state = 44},
  [1994] = {.lex_state = 53},
  [1995] = {.lex_state = 53},
  [1996] = {.lex_state = 0},
  [1997] = {.lex_state = 35},
  [1998] = {.lex_state = 0},
  [1999] = {.lex_state = 0},
  [2000] = {.lex_state = 0},
  [2001] = {.lex_state = 35},
  [2002] = {.lex_state = 0},
  [2003] = {.lex_state = 53},
  [2004] = {.lex_state = 0},
  [2005] = {.lex_state = 35},
  [2006] = {.lex_state = 44},
  [2007] = {.lex_state = 44},
  [2008] = {.lex_state = 120},
  [2009] = {.lex_state = 44},
  [2010] = {.lex_state = 44},
  [2011] = {.lex_state = 120},
  [2012] = {.lex_state = 120},
  [2013] = {.lex_state = 120},
  [2014] = {.lex_state = 0},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [aux_sym_preproc_include_token1] = ACTIONS(1),
    [aux_sym_preproc_def_token1] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [aux_sym_preproc_if_token1] = ACTIONS(1),
    [aux_sym_preproc_if_token2] = ACTIONS(1),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1),
    [aux_sym_preproc_else_token1] = ACTIONS(1),
    [aux_sym_preproc_elif_token1] = ACTIONS(1),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1),
    [sym_preproc_directive] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_defined] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym___extension__] = ACTIONS(1),
    [anon_sym_typedef] = ACTIONS(1),
    [anon_sym_extern] = ACTIONS(1),
    [anon_sym___attribute__] = ACTIONS(1),
    [anon_sym___attribute] = ACTIONS(1),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1),
    [anon_sym___declspec] = ACTIONS(1),
    [anon_sym___based] = ACTIONS(1),
    [anon_sym___cdecl] = ACTIONS(1),
    [anon_sym___clrcall] = ACTIONS(1),
    [anon_sym___stdcall] = ACTIONS(1),
    [anon_sym___fastcall] = ACTIONS(1),
    [anon_sym___thiscall] = ACTIONS(1),
    [anon_sym___vectorcall] = ACTIONS(1),
    [sym_ms_restrict_modifier] = ACTIONS(1),
    [sym_ms_unsigned_ptr_modifier] = ACTIONS(1),
    [sym_ms_signed_ptr_modifier] = ACTIONS(1),
    [anon_sym__unaligned] = ACTIONS(1),
    [anon_sym___unaligned] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_signed] = ACTIONS(1),
    [anon_sym_unsigned] = ACTIONS(1),
    [anon_sym_long] = ACTIONS(1),
    [anon_sym_short] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_auto] = ACTIONS(1),
    [anon_sym_register] = ACTIONS(1),
    [anon_sym_inline] = ACTIONS(1),
    [anon_sym___inline] = ACTIONS(1),
    [anon_sym___inline__] = ACTIONS(1),
    [anon_sym___forceinline] = ACTIONS(1),
    [anon_sym_thread_local] = ACTIONS(1),
    [anon_sym___thread] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_constexpr] = ACTIONS(1),
    [anon_sym_volatile] = ACTIONS(1),
    [anon_sym_restrict] = ACTIONS(1),
    [anon_sym___restrict__] = ACTIONS(1),
    [anon_sym__Atomic] = ACTIONS(1),
    [anon_sym__Noreturn] = ACTIONS(1),
    [anon_sym_noreturn] = ACTIONS(1),
    [anon_sym__Nonnull] = ACTIONS(1),
    [anon_sym_alignas] = ACTIONS(1),
    [anon_sym__Alignas] = ACTIONS(1),
    [sym_primitive_type] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_union] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_switch] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_goto] = ACTIONS(1),
    [anon_sym___try] = ACTIONS(1),
    [anon_sym___except] = ACTIONS(1),
    [anon_sym___finally] = ACTIONS(1),
    [anon_sym___leave] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_sizeof] = ACTIONS(1),
    [anon_sym___alignof__] = ACTIONS(1),
    [anon_sym___alignof] = ACTIONS(1),
    [anon_sym__alignof] = ACTIONS(1),
    [anon_sym_alignof] = ACTIONS(1),
    [anon_sym__Alignof] = ACTIONS(1),
    [anon_sym_offsetof] = ACTIONS(1),
    [anon_sym__Generic] = ACTIONS(1),
    [anon_sym_asm] = ACTIONS(1),
    [anon_sym___asm__] = ACTIONS(1),
    [anon_sym___asm] = ACTIONS(1),
    [anon_sym___volatile__] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [sym_number_literal] = ACTIONS(1),
    [anon_sym_L_SQUOTE] = ACTIONS(1),
    [anon_sym_u_SQUOTE] = ACTIONS(1),
    [anon_sym_U_SQUOTE] = ACTIONS(1),
    [anon_sym_u8_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_L_DQUOTE] = ACTIONS(1),
    [anon_sym_u_DQUOTE] = ACTIONS(1),
    [anon_sym_U_DQUOTE] = ACTIONS(1),
    [anon_sym_u8_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [sym_true] = ACTIONS(1),
    [sym_false] = ACTIONS(1),
    [anon_sym_NULL] = ACTIONS(1),
    [anon_sym_nullptr] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(1)] = {
    [sym_translation_unit] = STATE(1820),
    [sym__top_level_item] = STATE(44),
    [sym_preproc_include] = STATE(44),
    [sym_preproc_def] = STATE(44),
    [sym_preproc_function_def] = STATE(44),
    [sym_preproc_call] = STATE(44),
    [sym_preproc_if] = STATE(44),
    [sym_preproc_ifdef] = STATE(44),
    [sym_function_definition] = STATE(44),
    [sym__old_style_function_definition] = STATE(320),
    [sym_declaration] = STATE(44),
    [sym_type_definition] = STATE(44),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1128),
    [sym_linkage_specification] = STATE(44),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(683),
    [sym_compound_statement] = STATE(44),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(796),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(356),
    [sym__top_level_statement] = STATE(44),
    [sym_labeled_statement] = STATE(44),
    [sym__top_level_expression_statement] = STATE(44),
    [sym_if_statement] = STATE(44),
    [sym_switch_statement] = STATE(44),
    [sym_case_statement] = STATE(44),
    [sym_while_statement] = STATE(44),
    [sym_do_statement] = STATE(44),
    [sym_for_statement] = STATE(44),
    [sym_return_statement] = STATE(44),
    [sym_break_statement] = STATE(44),
    [sym_continue_statement] = STATE(44),
    [sym_goto_statement] = STATE(44),
    [sym_expression] = STATE(1101),
    [sym__string] = STATE(1103),
    [sym_conditional_expression] = STATE(1103),
    [sym_assignment_expression] = STATE(1103),
    [sym_pointer_expression] = STATE(928),
    [sym_unary_expression] = STATE(1103),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(1103),
    [sym_cast_expression] = STATE(1103),
    [sym_sizeof_expression] = STATE(1103),
    [sym_alignof_expression] = STATE(1103),
    [sym_offsetof_expression] = STATE(1103),
    [sym_generic_expression] = STATE(1103),
    [sym_subscript_expression] = STATE(928),
    [sym_call_expression] = STATE(928),
    [sym_gnu_asm_expression] = STATE(1103),
    [sym_extension_expression] = STATE(1103),
    [sym_field_expression] = STATE(928),
    [sym_compound_literal_expression] = STATE(1103),
    [sym_parenthesized_expression] = STATE(928),
    [sym_char_literal] = STATE(1103),
    [sym_concatenated_string] = STATE(1103),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(1103),
    [sym__empty_declaration] = STATE(44),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_translation_unit_repeat1] = STATE(44),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [aux_sym_preproc_include_token1] = ACTIONS(9),
    [aux_sym_preproc_def_token1] = ACTIONS(11),
    [aux_sym_preproc_if_token1] = ACTIONS(13),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(15),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(15),
    [sym_preproc_directive] = ACTIONS(17),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(27),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(33),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(95),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(101),
    [sym_false] = ACTIONS(101),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(2)] = {
    [sym__block_item] = STATE(3),
    [sym_preproc_include] = STATE(3),
    [sym_preproc_def] = STATE(3),
    [sym_preproc_function_def] = STATE(3),
    [sym_preproc_call] = STATE(3),
    [sym_preproc_if] = STATE(3),
    [sym_preproc_ifdef] = STATE(3),
    [sym_preproc_else] = STATE(1993),
    [sym_preproc_elif] = STATE(1993),
    [sym_preproc_elifdef] = STATE(1993),
    [sym_function_definition] = STATE(3),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(3),
    [sym_type_definition] = STATE(3),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(3),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(3),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(3),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(3),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(113),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(3)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1950),
    [sym_preproc_elif] = STATE(1950),
    [sym_preproc_elifdef] = STATE(1950),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(165),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(4)] = {
    [sym__block_item] = STATE(5),
    [sym_preproc_include] = STATE(5),
    [sym_preproc_def] = STATE(5),
    [sym_preproc_function_def] = STATE(5),
    [sym_preproc_call] = STATE(5),
    [sym_preproc_if] = STATE(5),
    [sym_preproc_ifdef] = STATE(5),
    [sym_preproc_else] = STATE(1932),
    [sym_preproc_elif] = STATE(1932),
    [sym_preproc_elifdef] = STATE(1932),
    [sym_function_definition] = STATE(5),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(5),
    [sym_type_definition] = STATE(5),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(5),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(5),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(5),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(5),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(167),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(5)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1978),
    [sym_preproc_elif] = STATE(1978),
    [sym_preproc_elifdef] = STATE(1978),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(169),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(6)] = {
    [sym__block_item] = STATE(8),
    [sym_preproc_include] = STATE(8),
    [sym_preproc_def] = STATE(8),
    [sym_preproc_function_def] = STATE(8),
    [sym_preproc_call] = STATE(8),
    [sym_preproc_if] = STATE(8),
    [sym_preproc_ifdef] = STATE(8),
    [sym_preproc_else] = STATE(2009),
    [sym_preproc_elif] = STATE(2009),
    [sym_preproc_elifdef] = STATE(2009),
    [sym_function_definition] = STATE(8),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(8),
    [sym_type_definition] = STATE(8),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(8),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(8),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(8),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(8),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(171),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(7)] = {
    [sym__block_item] = STATE(9),
    [sym_preproc_include] = STATE(9),
    [sym_preproc_def] = STATE(9),
    [sym_preproc_function_def] = STATE(9),
    [sym_preproc_call] = STATE(9),
    [sym_preproc_if] = STATE(9),
    [sym_preproc_ifdef] = STATE(9),
    [sym_preproc_else] = STATE(1791),
    [sym_preproc_elif] = STATE(1791),
    [sym_preproc_elifdef] = STATE(1791),
    [sym_function_definition] = STATE(9),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(9),
    [sym_type_definition] = STATE(9),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(9),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(9),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(9),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(9),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(173),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(8)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(2006),
    [sym_preproc_elif] = STATE(2006),
    [sym_preproc_elifdef] = STATE(2006),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(175),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(9)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1979),
    [sym_preproc_elif] = STATE(1979),
    [sym_preproc_elifdef] = STATE(1979),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(177),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(10)] = {
    [sym__block_item] = STATE(12),
    [sym_preproc_include] = STATE(12),
    [sym_preproc_def] = STATE(12),
    [sym_preproc_function_def] = STATE(12),
    [sym_preproc_call] = STATE(12),
    [sym_preproc_if] = STATE(12),
    [sym_preproc_ifdef] = STATE(12),
    [sym_preproc_else] = STATE(1803),
    [sym_preproc_elif] = STATE(1803),
    [sym_preproc_elifdef] = STATE(1803),
    [sym_function_definition] = STATE(12),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(12),
    [sym_type_definition] = STATE(12),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(12),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(12),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(12),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(12),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(179),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(11)] = {
    [sym__block_item] = STATE(13),
    [sym_preproc_include] = STATE(13),
    [sym_preproc_def] = STATE(13),
    [sym_preproc_function_def] = STATE(13),
    [sym_preproc_call] = STATE(13),
    [sym_preproc_if] = STATE(13),
    [sym_preproc_ifdef] = STATE(13),
    [sym_preproc_else] = STATE(1840),
    [sym_preproc_elif] = STATE(1840),
    [sym_preproc_elifdef] = STATE(1840),
    [sym_function_definition] = STATE(13),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(13),
    [sym_type_definition] = STATE(13),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(13),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(13),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(13),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(13),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(181),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(12)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1842),
    [sym_preproc_elif] = STATE(1842),
    [sym_preproc_elifdef] = STATE(1842),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(183),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(13)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1880),
    [sym_preproc_elif] = STATE(1880),
    [sym_preproc_elifdef] = STATE(1880),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(185),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(14)] = {
    [sym__block_item] = STATE(16),
    [sym_preproc_include] = STATE(16),
    [sym_preproc_def] = STATE(16),
    [sym_preproc_function_def] = STATE(16),
    [sym_preproc_call] = STATE(16),
    [sym_preproc_if] = STATE(16),
    [sym_preproc_ifdef] = STATE(16),
    [sym_preproc_else] = STATE(1917),
    [sym_preproc_elif] = STATE(1917),
    [sym_preproc_elifdef] = STATE(1917),
    [sym_function_definition] = STATE(16),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(16),
    [sym_type_definition] = STATE(16),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(16),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(16),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(16),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(16),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(187),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(15)] = {
    [sym__block_item] = STATE(17),
    [sym_preproc_include] = STATE(17),
    [sym_preproc_def] = STATE(17),
    [sym_preproc_function_def] = STATE(17),
    [sym_preproc_call] = STATE(17),
    [sym_preproc_if] = STATE(17),
    [sym_preproc_ifdef] = STATE(17),
    [sym_preproc_else] = STATE(1844),
    [sym_preproc_elif] = STATE(1844),
    [sym_preproc_elifdef] = STATE(1844),
    [sym_function_definition] = STATE(17),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(17),
    [sym_type_definition] = STATE(17),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(17),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(17),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(17),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(17),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(189),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(16)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1850),
    [sym_preproc_elif] = STATE(1850),
    [sym_preproc_elifdef] = STATE(1850),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(191),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(17)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1948),
    [sym_preproc_elif] = STATE(1948),
    [sym_preproc_elifdef] = STATE(1948),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(193),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(18)] = {
    [sym__block_item] = STATE(20),
    [sym_preproc_include] = STATE(20),
    [sym_preproc_def] = STATE(20),
    [sym_preproc_function_def] = STATE(20),
    [sym_preproc_call] = STATE(20),
    [sym_preproc_if] = STATE(20),
    [sym_preproc_ifdef] = STATE(20),
    [sym_preproc_else] = STATE(1867),
    [sym_preproc_elif] = STATE(1867),
    [sym_preproc_elifdef] = STATE(1867),
    [sym_function_definition] = STATE(20),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(20),
    [sym_type_definition] = STATE(20),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(20),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(20),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(20),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(20),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(195),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(19)] = {
    [sym__block_item] = STATE(21),
    [sym_preproc_include] = STATE(21),
    [sym_preproc_def] = STATE(21),
    [sym_preproc_function_def] = STATE(21),
    [sym_preproc_call] = STATE(21),
    [sym_preproc_if] = STATE(21),
    [sym_preproc_ifdef] = STATE(21),
    [sym_preproc_else] = STATE(1925),
    [sym_preproc_elif] = STATE(1925),
    [sym_preproc_elifdef] = STATE(1925),
    [sym_function_definition] = STATE(21),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(21),
    [sym_type_definition] = STATE(21),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(21),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(21),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(21),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(21),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(197),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(20)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1933),
    [sym_preproc_elif] = STATE(1933),
    [sym_preproc_elifdef] = STATE(1933),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(199),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(21)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_preproc_else] = STATE(1968),
    [sym_preproc_elif] = STATE(1968),
    [sym_preproc_elifdef] = STATE(1968),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(105),
    [aux_sym_preproc_include_token1] = ACTIONS(107),
    [aux_sym_preproc_def_token1] = ACTIONS(109),
    [aux_sym_preproc_if_token1] = ACTIONS(111),
    [aux_sym_preproc_if_token2] = ACTIONS(201),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(115),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(115),
    [aux_sym_preproc_else_token1] = ACTIONS(117),
    [aux_sym_preproc_elif_token1] = ACTIONS(119),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(121),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(121),
    [sym_preproc_directive] = ACTIONS(123),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(131),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(22)] = {
    [sym__block_item] = STATE(22),
    [sym_preproc_include] = STATE(22),
    [sym_preproc_def] = STATE(22),
    [sym_preproc_function_def] = STATE(22),
    [sym_preproc_call] = STATE(22),
    [sym_preproc_if] = STATE(22),
    [sym_preproc_ifdef] = STATE(22),
    [sym_function_definition] = STATE(22),
    [sym__old_style_function_definition] = STATE(131),
    [sym_declaration] = STATE(22),
    [sym_type_definition] = STATE(22),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1132),
    [sym_linkage_specification] = STATE(22),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(686),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(782),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(137),
    [sym_statement] = STATE(22),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(22),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(22),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(203),
    [aux_sym_preproc_include_token1] = ACTIONS(206),
    [aux_sym_preproc_def_token1] = ACTIONS(209),
    [aux_sym_preproc_if_token1] = ACTIONS(212),
    [aux_sym_preproc_if_token2] = ACTIONS(215),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(217),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(217),
    [aux_sym_preproc_else_token1] = ACTIONS(215),
    [aux_sym_preproc_elif_token1] = ACTIONS(215),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(215),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(215),
    [sym_preproc_directive] = ACTIONS(220),
    [anon_sym_LPAREN2] = ACTIONS(223),
    [anon_sym_BANG] = ACTIONS(226),
    [anon_sym_TILDE] = ACTIONS(226),
    [anon_sym_DASH] = ACTIONS(229),
    [anon_sym_PLUS] = ACTIONS(229),
    [anon_sym_STAR] = ACTIONS(232),
    [anon_sym_AMP] = ACTIONS(232),
    [anon_sym_SEMI] = ACTIONS(235),
    [anon_sym___extension__] = ACTIONS(238),
    [anon_sym_typedef] = ACTIONS(241),
    [anon_sym_extern] = ACTIONS(244),
    [anon_sym___attribute__] = ACTIONS(247),
    [anon_sym___attribute] = ACTIONS(247),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(250),
    [anon_sym___declspec] = ACTIONS(253),
    [anon_sym___cdecl] = ACTIONS(256),
    [anon_sym___clrcall] = ACTIONS(256),
    [anon_sym___stdcall] = ACTIONS(256),
    [anon_sym___fastcall] = ACTIONS(256),
    [anon_sym___thiscall] = ACTIONS(256),
    [anon_sym___vectorcall] = ACTIONS(256),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_signed] = ACTIONS(262),
    [anon_sym_unsigned] = ACTIONS(262),
    [anon_sym_long] = ACTIONS(262),
    [anon_sym_short] = ACTIONS(262),
    [anon_sym_static] = ACTIONS(265),
    [anon_sym_auto] = ACTIONS(265),
    [anon_sym_register] = ACTIONS(265),
    [anon_sym_inline] = ACTIONS(265),
    [anon_sym___inline] = ACTIONS(265),
    [anon_sym___inline__] = ACTIONS(265),
    [anon_sym___forceinline] = ACTIONS(265),
    [anon_sym_thread_local] = ACTIONS(265),
    [anon_sym___thread] = ACTIONS(265),
    [anon_sym_const] = ACTIONS(268),
    [anon_sym_constexpr] = ACTIONS(268),
    [anon_sym_volatile] = ACTIONS(268),
    [anon_sym_restrict] = ACTIONS(268),
    [anon_sym___restrict__] = ACTIONS(268),
    [anon_sym__Atomic] = ACTIONS(268),
    [anon_sym__Noreturn] = ACTIONS(268),
    [anon_sym_noreturn] = ACTIONS(268),
    [anon_sym__Nonnull] = ACTIONS(268),
    [anon_sym_alignas] = ACTIONS(271),
    [anon_sym__Alignas] = ACTIONS(271),
    [sym_primitive_type] = ACTIONS(274),
    [anon_sym_enum] = ACTIONS(277),
    [anon_sym_struct] = ACTIONS(280),
    [anon_sym_union] = ACTIONS(283),
    [anon_sym_if] = ACTIONS(286),
    [anon_sym_switch] = ACTIONS(289),
    [anon_sym_case] = ACTIONS(292),
    [anon_sym_default] = ACTIONS(295),
    [anon_sym_while] = ACTIONS(298),
    [anon_sym_do] = ACTIONS(301),
    [anon_sym_for] = ACTIONS(304),
    [anon_sym_return] = ACTIONS(307),
    [anon_sym_break] = ACTIONS(310),
    [anon_sym_continue] = ACTIONS(313),
    [anon_sym_goto] = ACTIONS(316),
    [anon_sym___try] = ACTIONS(319),
    [anon_sym___leave] = ACTIONS(322),
    [anon_sym_DASH_DASH] = ACTIONS(325),
    [anon_sym_PLUS_PLUS] = ACTIONS(325),
    [anon_sym_sizeof] = ACTIONS(328),
    [anon_sym___alignof__] = ACTIONS(331),
    [anon_sym___alignof] = ACTIONS(331),
    [anon_sym__alignof] = ACTIONS(331),
    [anon_sym_alignof] = ACTIONS(331),
    [anon_sym__Alignof] = ACTIONS(331),
    [anon_sym_offsetof] = ACTIONS(334),
    [anon_sym__Generic] = ACTIONS(337),
    [anon_sym_asm] = ACTIONS(340),
    [anon_sym___asm__] = ACTIONS(340),
    [anon_sym___asm] = ACTIONS(340),
    [sym_number_literal] = ACTIONS(343),
    [anon_sym_L_SQUOTE] = ACTIONS(346),
    [anon_sym_u_SQUOTE] = ACTIONS(346),
    [anon_sym_U_SQUOTE] = ACTIONS(346),
    [anon_sym_u8_SQUOTE] = ACTIONS(346),
    [anon_sym_SQUOTE] = ACTIONS(346),
    [anon_sym_L_DQUOTE] = ACTIONS(349),
    [anon_sym_u_DQUOTE] = ACTIONS(349),
    [anon_sym_U_DQUOTE] = ACTIONS(349),
    [anon_sym_u8_DQUOTE] = ACTIONS(349),
    [anon_sym_DQUOTE] = ACTIONS(349),
    [sym_true] = ACTIONS(352),
    [sym_false] = ACTIONS(352),
    [anon_sym_NULL] = ACTIONS(355),
    [anon_sym_nullptr] = ACTIONS(355),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(23)] = {
    [sym__block_item] = STATE(25),
    [sym_preproc_include] = STATE(25),
    [sym_preproc_def] = STATE(25),
    [sym_preproc_function_def] = STATE(25),
    [sym_preproc_call] = STATE(25),
    [sym_preproc_if] = STATE(25),
    [sym_preproc_ifdef] = STATE(25),
    [sym_function_definition] = STATE(25),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(25),
    [sym_type_definition] = STATE(25),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(25),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(25),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(25),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(25),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(380),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(24)] = {
    [sym__block_item] = STATE(36),
    [sym_preproc_include] = STATE(36),
    [sym_preproc_def] = STATE(36),
    [sym_preproc_function_def] = STATE(36),
    [sym_preproc_call] = STATE(36),
    [sym_preproc_if] = STATE(36),
    [sym_preproc_ifdef] = STATE(36),
    [sym_function_definition] = STATE(36),
    [sym__old_style_function_definition] = STATE(282),
    [sym_declaration] = STATE(36),
    [sym_type_definition] = STATE(36),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1133),
    [sym_linkage_specification] = STATE(36),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(688),
    [sym_compound_statement] = STATE(178),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(789),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(311),
    [sym_statement] = STATE(36),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(36),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(36),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(408),
    [aux_sym_preproc_include_token1] = ACTIONS(410),
    [aux_sym_preproc_def_token1] = ACTIONS(412),
    [aux_sym_preproc_if_token1] = ACTIONS(414),
    [aux_sym_preproc_if_token2] = ACTIONS(416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(418),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(418),
    [sym_preproc_directive] = ACTIONS(420),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(428),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(25)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(458),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(26)] = {
    [sym__block_item] = STATE(26),
    [sym_preproc_include] = STATE(26),
    [sym_preproc_def] = STATE(26),
    [sym_preproc_function_def] = STATE(26),
    [sym_preproc_call] = STATE(26),
    [sym_preproc_if] = STATE(26),
    [sym_preproc_ifdef] = STATE(26),
    [sym_function_definition] = STATE(26),
    [sym__old_style_function_definition] = STATE(282),
    [sym_declaration] = STATE(26),
    [sym_type_definition] = STATE(26),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1133),
    [sym_linkage_specification] = STATE(26),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(688),
    [sym_compound_statement] = STATE(178),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(789),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(311),
    [sym_statement] = STATE(26),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(26),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(26),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(460),
    [aux_sym_preproc_include_token1] = ACTIONS(463),
    [aux_sym_preproc_def_token1] = ACTIONS(466),
    [aux_sym_preproc_if_token1] = ACTIONS(469),
    [aux_sym_preproc_if_token2] = ACTIONS(215),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(472),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(472),
    [sym_preproc_directive] = ACTIONS(475),
    [anon_sym_LPAREN2] = ACTIONS(223),
    [anon_sym_BANG] = ACTIONS(226),
    [anon_sym_TILDE] = ACTIONS(226),
    [anon_sym_DASH] = ACTIONS(229),
    [anon_sym_PLUS] = ACTIONS(229),
    [anon_sym_STAR] = ACTIONS(232),
    [anon_sym_AMP] = ACTIONS(232),
    [anon_sym_SEMI] = ACTIONS(478),
    [anon_sym___extension__] = ACTIONS(481),
    [anon_sym_typedef] = ACTIONS(484),
    [anon_sym_extern] = ACTIONS(487),
    [anon_sym___attribute__] = ACTIONS(247),
    [anon_sym___attribute] = ACTIONS(247),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(250),
    [anon_sym___declspec] = ACTIONS(253),
    [anon_sym___cdecl] = ACTIONS(256),
    [anon_sym___clrcall] = ACTIONS(256),
    [anon_sym___stdcall] = ACTIONS(256),
    [anon_sym___fastcall] = ACTIONS(256),
    [anon_sym___thiscall] = ACTIONS(256),
    [anon_sym___vectorcall] = ACTIONS(256),
    [anon_sym_LBRACE] = ACTIONS(490),
    [anon_sym_signed] = ACTIONS(262),
    [anon_sym_unsigned] = ACTIONS(262),
    [anon_sym_long] = ACTIONS(262),
    [anon_sym_short] = ACTIONS(262),
    [anon_sym_static] = ACTIONS(265),
    [anon_sym_auto] = ACTIONS(265),
    [anon_sym_register] = ACTIONS(265),
    [anon_sym_inline] = ACTIONS(265),
    [anon_sym___inline] = ACTIONS(265),
    [anon_sym___inline__] = ACTIONS(265),
    [anon_sym___forceinline] = ACTIONS(265),
    [anon_sym_thread_local] = ACTIONS(265),
    [anon_sym___thread] = ACTIONS(265),
    [anon_sym_const] = ACTIONS(268),
    [anon_sym_constexpr] = ACTIONS(268),
    [anon_sym_volatile] = ACTIONS(268),
    [anon_sym_restrict] = ACTIONS(268),
    [anon_sym___restrict__] = ACTIONS(268),
    [anon_sym__Atomic] = ACTIONS(268),
    [anon_sym__Noreturn] = ACTIONS(268),
    [anon_sym_noreturn] = ACTIONS(268),
    [anon_sym__Nonnull] = ACTIONS(268),
    [anon_sym_alignas] = ACTIONS(271),
    [anon_sym__Alignas] = ACTIONS(271),
    [sym_primitive_type] = ACTIONS(274),
    [anon_sym_enum] = ACTIONS(277),
    [anon_sym_struct] = ACTIONS(280),
    [anon_sym_union] = ACTIONS(283),
    [anon_sym_if] = ACTIONS(493),
    [anon_sym_switch] = ACTIONS(496),
    [anon_sym_case] = ACTIONS(499),
    [anon_sym_default] = ACTIONS(502),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(508),
    [anon_sym_for] = ACTIONS(511),
    [anon_sym_return] = ACTIONS(514),
    [anon_sym_break] = ACTIONS(517),
    [anon_sym_continue] = ACTIONS(520),
    [anon_sym_goto] = ACTIONS(523),
    [anon_sym___try] = ACTIONS(526),
    [anon_sym___leave] = ACTIONS(529),
    [anon_sym_DASH_DASH] = ACTIONS(325),
    [anon_sym_PLUS_PLUS] = ACTIONS(325),
    [anon_sym_sizeof] = ACTIONS(328),
    [anon_sym___alignof__] = ACTIONS(331),
    [anon_sym___alignof] = ACTIONS(331),
    [anon_sym__alignof] = ACTIONS(331),
    [anon_sym_alignof] = ACTIONS(331),
    [anon_sym__Alignof] = ACTIONS(331),
    [anon_sym_offsetof] = ACTIONS(334),
    [anon_sym__Generic] = ACTIONS(337),
    [anon_sym_asm] = ACTIONS(340),
    [anon_sym___asm__] = ACTIONS(340),
    [anon_sym___asm] = ACTIONS(340),
    [sym_number_literal] = ACTIONS(343),
    [anon_sym_L_SQUOTE] = ACTIONS(346),
    [anon_sym_u_SQUOTE] = ACTIONS(346),
    [anon_sym_U_SQUOTE] = ACTIONS(346),
    [anon_sym_u8_SQUOTE] = ACTIONS(346),
    [anon_sym_SQUOTE] = ACTIONS(346),
    [anon_sym_L_DQUOTE] = ACTIONS(349),
    [anon_sym_u_DQUOTE] = ACTIONS(349),
    [anon_sym_U_DQUOTE] = ACTIONS(349),
    [anon_sym_u8_DQUOTE] = ACTIONS(349),
    [anon_sym_DQUOTE] = ACTIONS(349),
    [sym_true] = ACTIONS(352),
    [sym_false] = ACTIONS(352),
    [anon_sym_NULL] = ACTIONS(355),
    [anon_sym_nullptr] = ACTIONS(355),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(27)] = {
    [sym__block_item] = STATE(28),
    [sym_preproc_include] = STATE(28),
    [sym_preproc_def] = STATE(28),
    [sym_preproc_function_def] = STATE(28),
    [sym_preproc_call] = STATE(28),
    [sym_preproc_if] = STATE(28),
    [sym_preproc_ifdef] = STATE(28),
    [sym_function_definition] = STATE(28),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(28),
    [sym_type_definition] = STATE(28),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(28),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(28),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(28),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(28),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(532),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(28)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(534),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(29)] = {
    [sym__block_item] = STATE(40),
    [sym_preproc_include] = STATE(40),
    [sym_preproc_def] = STATE(40),
    [sym_preproc_function_def] = STATE(40),
    [sym_preproc_call] = STATE(40),
    [sym_preproc_if] = STATE(40),
    [sym_preproc_ifdef] = STATE(40),
    [sym_function_definition] = STATE(40),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(40),
    [sym_type_definition] = STATE(40),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(40),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(40),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(40),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(40),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(536),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(30)] = {
    [sym__block_item] = STATE(32),
    [sym_preproc_include] = STATE(32),
    [sym_preproc_def] = STATE(32),
    [sym_preproc_function_def] = STATE(32),
    [sym_preproc_call] = STATE(32),
    [sym_preproc_if] = STATE(32),
    [sym_preproc_ifdef] = STATE(32),
    [sym_function_definition] = STATE(32),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(32),
    [sym_type_definition] = STATE(32),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(32),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(32),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(32),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(32),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(538),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(31)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(540),
    [aux_sym_preproc_include_token1] = ACTIONS(543),
    [aux_sym_preproc_def_token1] = ACTIONS(546),
    [aux_sym_preproc_if_token1] = ACTIONS(549),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(552),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(552),
    [sym_preproc_directive] = ACTIONS(555),
    [anon_sym_LPAREN2] = ACTIONS(223),
    [anon_sym_BANG] = ACTIONS(226),
    [anon_sym_TILDE] = ACTIONS(226),
    [anon_sym_DASH] = ACTIONS(229),
    [anon_sym_PLUS] = ACTIONS(229),
    [anon_sym_STAR] = ACTIONS(232),
    [anon_sym_AMP] = ACTIONS(232),
    [anon_sym_SEMI] = ACTIONS(558),
    [anon_sym___extension__] = ACTIONS(561),
    [anon_sym_typedef] = ACTIONS(564),
    [anon_sym_extern] = ACTIONS(567),
    [anon_sym___attribute__] = ACTIONS(247),
    [anon_sym___attribute] = ACTIONS(247),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(250),
    [anon_sym___declspec] = ACTIONS(253),
    [anon_sym___cdecl] = ACTIONS(256),
    [anon_sym___clrcall] = ACTIONS(256),
    [anon_sym___stdcall] = ACTIONS(256),
    [anon_sym___fastcall] = ACTIONS(256),
    [anon_sym___thiscall] = ACTIONS(256),
    [anon_sym___vectorcall] = ACTIONS(256),
    [anon_sym_LBRACE] = ACTIONS(570),
    [anon_sym_RBRACE] = ACTIONS(573),
    [anon_sym_signed] = ACTIONS(262),
    [anon_sym_unsigned] = ACTIONS(262),
    [anon_sym_long] = ACTIONS(262),
    [anon_sym_short] = ACTIONS(262),
    [anon_sym_static] = ACTIONS(265),
    [anon_sym_auto] = ACTIONS(265),
    [anon_sym_register] = ACTIONS(265),
    [anon_sym_inline] = ACTIONS(265),
    [anon_sym___inline] = ACTIONS(265),
    [anon_sym___inline__] = ACTIONS(265),
    [anon_sym___forceinline] = ACTIONS(265),
    [anon_sym_thread_local] = ACTIONS(265),
    [anon_sym___thread] = ACTIONS(265),
    [anon_sym_const] = ACTIONS(268),
    [anon_sym_constexpr] = ACTIONS(268),
    [anon_sym_volatile] = ACTIONS(268),
    [anon_sym_restrict] = ACTIONS(268),
    [anon_sym___restrict__] = ACTIONS(268),
    [anon_sym__Atomic] = ACTIONS(268),
    [anon_sym__Noreturn] = ACTIONS(268),
    [anon_sym_noreturn] = ACTIONS(268),
    [anon_sym__Nonnull] = ACTIONS(268),
    [anon_sym_alignas] = ACTIONS(271),
    [anon_sym__Alignas] = ACTIONS(271),
    [sym_primitive_type] = ACTIONS(274),
    [anon_sym_enum] = ACTIONS(277),
    [anon_sym_struct] = ACTIONS(280),
    [anon_sym_union] = ACTIONS(283),
    [anon_sym_if] = ACTIONS(575),
    [anon_sym_switch] = ACTIONS(578),
    [anon_sym_case] = ACTIONS(581),
    [anon_sym_default] = ACTIONS(584),
    [anon_sym_while] = ACTIONS(587),
    [anon_sym_do] = ACTIONS(590),
    [anon_sym_for] = ACTIONS(593),
    [anon_sym_return] = ACTIONS(596),
    [anon_sym_break] = ACTIONS(599),
    [anon_sym_continue] = ACTIONS(602),
    [anon_sym_goto] = ACTIONS(605),
    [anon_sym___try] = ACTIONS(608),
    [anon_sym___leave] = ACTIONS(611),
    [anon_sym_DASH_DASH] = ACTIONS(325),
    [anon_sym_PLUS_PLUS] = ACTIONS(325),
    [anon_sym_sizeof] = ACTIONS(328),
    [anon_sym___alignof__] = ACTIONS(331),
    [anon_sym___alignof] = ACTIONS(331),
    [anon_sym__alignof] = ACTIONS(331),
    [anon_sym_alignof] = ACTIONS(331),
    [anon_sym__Alignof] = ACTIONS(331),
    [anon_sym_offsetof] = ACTIONS(334),
    [anon_sym__Generic] = ACTIONS(337),
    [anon_sym_asm] = ACTIONS(340),
    [anon_sym___asm__] = ACTIONS(340),
    [anon_sym___asm] = ACTIONS(340),
    [sym_number_literal] = ACTIONS(343),
    [anon_sym_L_SQUOTE] = ACTIONS(346),
    [anon_sym_u_SQUOTE] = ACTIONS(346),
    [anon_sym_U_SQUOTE] = ACTIONS(346),
    [anon_sym_u8_SQUOTE] = ACTIONS(346),
    [anon_sym_SQUOTE] = ACTIONS(346),
    [anon_sym_L_DQUOTE] = ACTIONS(349),
    [anon_sym_u_DQUOTE] = ACTIONS(349),
    [anon_sym_U_DQUOTE] = ACTIONS(349),
    [anon_sym_u8_DQUOTE] = ACTIONS(349),
    [anon_sym_DQUOTE] = ACTIONS(349),
    [sym_true] = ACTIONS(352),
    [sym_false] = ACTIONS(352),
    [anon_sym_NULL] = ACTIONS(355),
    [anon_sym_nullptr] = ACTIONS(355),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(32)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(614),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(33)] = {
    [sym__block_item] = STATE(34),
    [sym_preproc_include] = STATE(34),
    [sym_preproc_def] = STATE(34),
    [sym_preproc_function_def] = STATE(34),
    [sym_preproc_call] = STATE(34),
    [sym_preproc_if] = STATE(34),
    [sym_preproc_ifdef] = STATE(34),
    [sym_function_definition] = STATE(34),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(34),
    [sym_type_definition] = STATE(34),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(34),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(34),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(34),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(34),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(616),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(34)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(618),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(35)] = {
    [sym__block_item] = STATE(37),
    [sym_preproc_include] = STATE(37),
    [sym_preproc_def] = STATE(37),
    [sym_preproc_function_def] = STATE(37),
    [sym_preproc_call] = STATE(37),
    [sym_preproc_if] = STATE(37),
    [sym_preproc_ifdef] = STATE(37),
    [sym_function_definition] = STATE(37),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(37),
    [sym_type_definition] = STATE(37),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(37),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(37),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(37),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(37),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(620),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(36)] = {
    [sym__block_item] = STATE(26),
    [sym_preproc_include] = STATE(26),
    [sym_preproc_def] = STATE(26),
    [sym_preproc_function_def] = STATE(26),
    [sym_preproc_call] = STATE(26),
    [sym_preproc_if] = STATE(26),
    [sym_preproc_ifdef] = STATE(26),
    [sym_function_definition] = STATE(26),
    [sym__old_style_function_definition] = STATE(282),
    [sym_declaration] = STATE(26),
    [sym_type_definition] = STATE(26),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1133),
    [sym_linkage_specification] = STATE(26),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(688),
    [sym_compound_statement] = STATE(178),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(789),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(311),
    [sym_statement] = STATE(26),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(26),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(26),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(408),
    [aux_sym_preproc_include_token1] = ACTIONS(410),
    [aux_sym_preproc_def_token1] = ACTIONS(412),
    [aux_sym_preproc_if_token1] = ACTIONS(414),
    [aux_sym_preproc_if_token2] = ACTIONS(622),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(418),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(418),
    [sym_preproc_directive] = ACTIONS(420),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(428),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(37)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(624),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(38)] = {
    [sym__block_item] = STATE(39),
    [sym_preproc_include] = STATE(39),
    [sym_preproc_def] = STATE(39),
    [sym_preproc_function_def] = STATE(39),
    [sym_preproc_call] = STATE(39),
    [sym_preproc_if] = STATE(39),
    [sym_preproc_ifdef] = STATE(39),
    [sym_function_definition] = STATE(39),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(39),
    [sym_type_definition] = STATE(39),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(39),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(39),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(39),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(39),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(626),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(39)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(628),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(40)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(630),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(41)] = {
    [sym__block_item] = STATE(42),
    [sym_preproc_include] = STATE(42),
    [sym_preproc_def] = STATE(42),
    [sym_preproc_function_def] = STATE(42),
    [sym_preproc_call] = STATE(42),
    [sym_preproc_if] = STATE(42),
    [sym_preproc_ifdef] = STATE(42),
    [sym_function_definition] = STATE(42),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(42),
    [sym_type_definition] = STATE(42),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(42),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(42),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(42),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(42),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(632),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(42)] = {
    [sym__block_item] = STATE(31),
    [sym_preproc_include] = STATE(31),
    [sym_preproc_def] = STATE(31),
    [sym_preproc_function_def] = STATE(31),
    [sym_preproc_call] = STATE(31),
    [sym_preproc_if] = STATE(31),
    [sym_preproc_ifdef] = STATE(31),
    [sym_function_definition] = STATE(31),
    [sym__old_style_function_definition] = STATE(268),
    [sym_declaration] = STATE(31),
    [sym_type_definition] = STATE(31),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1139),
    [sym_linkage_specification] = STATE(31),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(666),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(801),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(269),
    [sym_statement] = STATE(31),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym__empty_declaration] = STATE(31),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_repeat1] = STATE(31),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(358),
    [aux_sym_preproc_include_token1] = ACTIONS(360),
    [aux_sym_preproc_def_token1] = ACTIONS(362),
    [aux_sym_preproc_if_token1] = ACTIONS(364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(366),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(366),
    [sym_preproc_directive] = ACTIONS(368),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(376),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(634),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(43)] = {
    [sym__top_level_item] = STATE(43),
    [sym_preproc_include] = STATE(43),
    [sym_preproc_def] = STATE(43),
    [sym_preproc_function_def] = STATE(43),
    [sym_preproc_call] = STATE(43),
    [sym_preproc_if] = STATE(43),
    [sym_preproc_ifdef] = STATE(43),
    [sym_function_definition] = STATE(43),
    [sym__old_style_function_definition] = STATE(320),
    [sym_declaration] = STATE(43),
    [sym_type_definition] = STATE(43),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1128),
    [sym_linkage_specification] = STATE(43),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(683),
    [sym_compound_statement] = STATE(43),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(796),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(356),
    [sym__top_level_statement] = STATE(43),
    [sym_labeled_statement] = STATE(43),
    [sym__top_level_expression_statement] = STATE(43),
    [sym_if_statement] = STATE(43),
    [sym_switch_statement] = STATE(43),
    [sym_case_statement] = STATE(43),
    [sym_while_statement] = STATE(43),
    [sym_do_statement] = STATE(43),
    [sym_for_statement] = STATE(43),
    [sym_return_statement] = STATE(43),
    [sym_break_statement] = STATE(43),
    [sym_continue_statement] = STATE(43),
    [sym_goto_statement] = STATE(43),
    [sym_expression] = STATE(1101),
    [sym__string] = STATE(1103),
    [sym_conditional_expression] = STATE(1103),
    [sym_assignment_expression] = STATE(1103),
    [sym_pointer_expression] = STATE(928),
    [sym_unary_expression] = STATE(1103),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(1103),
    [sym_cast_expression] = STATE(1103),
    [sym_sizeof_expression] = STATE(1103),
    [sym_alignof_expression] = STATE(1103),
    [sym_offsetof_expression] = STATE(1103),
    [sym_generic_expression] = STATE(1103),
    [sym_subscript_expression] = STATE(928),
    [sym_call_expression] = STATE(928),
    [sym_gnu_asm_expression] = STATE(1103),
    [sym_extension_expression] = STATE(1103),
    [sym_field_expression] = STATE(928),
    [sym_compound_literal_expression] = STATE(1103),
    [sym_parenthesized_expression] = STATE(928),
    [sym_char_literal] = STATE(1103),
    [sym_concatenated_string] = STATE(1103),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(1103),
    [sym__empty_declaration] = STATE(43),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_translation_unit_repeat1] = STATE(43),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [ts_builtin_sym_end] = ACTIONS(636),
    [sym_identifier] = ACTIONS(638),
    [aux_sym_preproc_include_token1] = ACTIONS(641),
    [aux_sym_preproc_def_token1] = ACTIONS(644),
    [aux_sym_preproc_if_token1] = ACTIONS(647),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(650),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(650),
    [sym_preproc_directive] = ACTIONS(653),
    [anon_sym_LPAREN2] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(659),
    [anon_sym_TILDE] = ACTIONS(659),
    [anon_sym_DASH] = ACTIONS(662),
    [anon_sym_PLUS] = ACTIONS(662),
    [anon_sym_STAR] = ACTIONS(665),
    [anon_sym_AMP] = ACTIONS(665),
    [anon_sym_SEMI] = ACTIONS(668),
    [anon_sym___extension__] = ACTIONS(671),
    [anon_sym_typedef] = ACTIONS(674),
    [anon_sym_extern] = ACTIONS(677),
    [anon_sym___attribute__] = ACTIONS(680),
    [anon_sym___attribute] = ACTIONS(680),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(683),
    [anon_sym___declspec] = ACTIONS(686),
    [anon_sym___cdecl] = ACTIONS(689),
    [anon_sym___clrcall] = ACTIONS(689),
    [anon_sym___stdcall] = ACTIONS(689),
    [anon_sym___fastcall] = ACTIONS(689),
    [anon_sym___thiscall] = ACTIONS(689),
    [anon_sym___vectorcall] = ACTIONS(689),
    [anon_sym_LBRACE] = ACTIONS(692),
    [anon_sym_signed] = ACTIONS(695),
    [anon_sym_unsigned] = ACTIONS(695),
    [anon_sym_long] = ACTIONS(695),
    [anon_sym_short] = ACTIONS(695),
    [anon_sym_static] = ACTIONS(698),
    [anon_sym_auto] = ACTIONS(698),
    [anon_sym_register] = ACTIONS(698),
    [anon_sym_inline] = ACTIONS(698),
    [anon_sym___inline] = ACTIONS(698),
    [anon_sym___inline__] = ACTIONS(698),
    [anon_sym___forceinline] = ACTIONS(698),
    [anon_sym_thread_local] = ACTIONS(698),
    [anon_sym___thread] = ACTIONS(698),
    [anon_sym_const] = ACTIONS(701),
    [anon_sym_constexpr] = ACTIONS(701),
    [anon_sym_volatile] = ACTIONS(701),
    [anon_sym_restrict] = ACTIONS(701),
    [anon_sym___restrict__] = ACTIONS(701),
    [anon_sym__Atomic] = ACTIONS(701),
    [anon_sym__Noreturn] = ACTIONS(701),
    [anon_sym_noreturn] = ACTIONS(701),
    [anon_sym__Nonnull] = ACTIONS(701),
    [anon_sym_alignas] = ACTIONS(704),
    [anon_sym__Alignas] = ACTIONS(704),
    [sym_primitive_type] = ACTIONS(707),
    [anon_sym_enum] = ACTIONS(710),
    [anon_sym_struct] = ACTIONS(713),
    [anon_sym_union] = ACTIONS(716),
    [anon_sym_if] = ACTIONS(719),
    [anon_sym_switch] = ACTIONS(722),
    [anon_sym_case] = ACTIONS(725),
    [anon_sym_default] = ACTIONS(728),
    [anon_sym_while] = ACTIONS(731),
    [anon_sym_do] = ACTIONS(734),
    [anon_sym_for] = ACTIONS(737),
    [anon_sym_return] = ACTIONS(740),
    [anon_sym_break] = ACTIONS(743),
    [anon_sym_continue] = ACTIONS(746),
    [anon_sym_goto] = ACTIONS(749),
    [anon_sym_DASH_DASH] = ACTIONS(752),
    [anon_sym_PLUS_PLUS] = ACTIONS(752),
    [anon_sym_sizeof] = ACTIONS(755),
    [anon_sym___alignof__] = ACTIONS(758),
    [anon_sym___alignof] = ACTIONS(758),
    [anon_sym__alignof] = ACTIONS(758),
    [anon_sym_alignof] = ACTIONS(758),
    [anon_sym__Alignof] = ACTIONS(758),
    [anon_sym_offsetof] = ACTIONS(761),
    [anon_sym__Generic] = ACTIONS(764),
    [anon_sym_asm] = ACTIONS(767),
    [anon_sym___asm__] = ACTIONS(767),
    [anon_sym___asm] = ACTIONS(767),
    [sym_number_literal] = ACTIONS(770),
    [anon_sym_L_SQUOTE] = ACTIONS(773),
    [anon_sym_u_SQUOTE] = ACTIONS(773),
    [anon_sym_U_SQUOTE] = ACTIONS(773),
    [anon_sym_u8_SQUOTE] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(773),
    [anon_sym_L_DQUOTE] = ACTIONS(776),
    [anon_sym_u_DQUOTE] = ACTIONS(776),
    [anon_sym_U_DQUOTE] = ACTIONS(776),
    [anon_sym_u8_DQUOTE] = ACTIONS(776),
    [anon_sym_DQUOTE] = ACTIONS(776),
    [sym_true] = ACTIONS(779),
    [sym_false] = ACTIONS(779),
    [anon_sym_NULL] = ACTIONS(782),
    [anon_sym_nullptr] = ACTIONS(782),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(44)] = {
    [sym__top_level_item] = STATE(43),
    [sym_preproc_include] = STATE(43),
    [sym_preproc_def] = STATE(43),
    [sym_preproc_function_def] = STATE(43),
    [sym_preproc_call] = STATE(43),
    [sym_preproc_if] = STATE(43),
    [sym_preproc_ifdef] = STATE(43),
    [sym_function_definition] = STATE(43),
    [sym__old_style_function_definition] = STATE(320),
    [sym_declaration] = STATE(43),
    [sym_type_definition] = STATE(43),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1128),
    [sym_linkage_specification] = STATE(43),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(683),
    [sym_compound_statement] = STATE(43),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(796),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(356),
    [sym__top_level_statement] = STATE(43),
    [sym_labeled_statement] = STATE(43),
    [sym__top_level_expression_statement] = STATE(43),
    [sym_if_statement] = STATE(43),
    [sym_switch_statement] = STATE(43),
    [sym_case_statement] = STATE(43),
    [sym_while_statement] = STATE(43),
    [sym_do_statement] = STATE(43),
    [sym_for_statement] = STATE(43),
    [sym_return_statement] = STATE(43),
    [sym_break_statement] = STATE(43),
    [sym_continue_statement] = STATE(43),
    [sym_goto_statement] = STATE(43),
    [sym_expression] = STATE(1101),
    [sym__string] = STATE(1103),
    [sym_conditional_expression] = STATE(1103),
    [sym_assignment_expression] = STATE(1103),
    [sym_pointer_expression] = STATE(928),
    [sym_unary_expression] = STATE(1103),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(1103),
    [sym_cast_expression] = STATE(1103),
    [sym_sizeof_expression] = STATE(1103),
    [sym_alignof_expression] = STATE(1103),
    [sym_offsetof_expression] = STATE(1103),
    [sym_generic_expression] = STATE(1103),
    [sym_subscript_expression] = STATE(928),
    [sym_call_expression] = STATE(928),
    [sym_gnu_asm_expression] = STATE(1103),
    [sym_extension_expression] = STATE(1103),
    [sym_field_expression] = STATE(928),
    [sym_compound_literal_expression] = STATE(1103),
    [sym_parenthesized_expression] = STATE(928),
    [sym_char_literal] = STATE(1103),
    [sym_concatenated_string] = STATE(1103),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(1103),
    [sym__empty_declaration] = STATE(43),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_translation_unit_repeat1] = STATE(43),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [ts_builtin_sym_end] = ACTIONS(785),
    [sym_identifier] = ACTIONS(7),
    [aux_sym_preproc_include_token1] = ACTIONS(9),
    [aux_sym_preproc_def_token1] = ACTIONS(11),
    [aux_sym_preproc_if_token1] = ACTIONS(13),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(15),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(15),
    [sym_preproc_directive] = ACTIONS(17),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(27),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(33),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(95),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(101),
    [sym_false] = ACTIONS(101),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(45)] = {
    [sym_declaration] = STATE(46),
    [sym_type_definition] = STATE(46),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(46),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(46),
    [sym_labeled_statement] = STATE(46),
    [sym_expression_statement] = STATE(46),
    [sym_if_statement] = STATE(46),
    [sym_switch_statement] = STATE(46),
    [sym_while_statement] = STATE(46),
    [sym_do_statement] = STATE(46),
    [sym_for_statement] = STATE(46),
    [sym_return_statement] = STATE(46),
    [sym_break_statement] = STATE(46),
    [sym_continue_statement] = STATE(46),
    [sym_goto_statement] = STATE(46),
    [sym_seh_try_statement] = STATE(46),
    [sym_seh_leave_statement] = STATE(46),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(46),
    [sym_identifier] = ACTIONS(787),
    [aux_sym_preproc_include_token1] = ACTIONS(789),
    [aux_sym_preproc_def_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token2] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(789),
    [aux_sym_preproc_else_token1] = ACTIONS(789),
    [aux_sym_preproc_elif_token1] = ACTIONS(789),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(789),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(789),
    [sym_preproc_directive] = ACTIONS(789),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(789),
    [anon_sym___clrcall] = ACTIONS(789),
    [anon_sym___stdcall] = ACTIONS(789),
    [anon_sym___fastcall] = ACTIONS(789),
    [anon_sym___thiscall] = ACTIONS(789),
    [anon_sym___vectorcall] = ACTIONS(789),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_else] = ACTIONS(789),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(789),
    [anon_sym_default] = ACTIONS(789),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(46)] = {
    [sym_declaration] = STATE(48),
    [sym_type_definition] = STATE(48),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(48),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(48),
    [sym_labeled_statement] = STATE(48),
    [sym_expression_statement] = STATE(48),
    [sym_if_statement] = STATE(48),
    [sym_switch_statement] = STATE(48),
    [sym_while_statement] = STATE(48),
    [sym_do_statement] = STATE(48),
    [sym_for_statement] = STATE(48),
    [sym_return_statement] = STATE(48),
    [sym_break_statement] = STATE(48),
    [sym_continue_statement] = STATE(48),
    [sym_goto_statement] = STATE(48),
    [sym_seh_try_statement] = STATE(48),
    [sym_seh_leave_statement] = STATE(48),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(48),
    [sym_identifier] = ACTIONS(787),
    [aux_sym_preproc_include_token1] = ACTIONS(791),
    [aux_sym_preproc_def_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token2] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(791),
    [aux_sym_preproc_else_token1] = ACTIONS(791),
    [aux_sym_preproc_elif_token1] = ACTIONS(791),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(791),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(791),
    [sym_preproc_directive] = ACTIONS(791),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(791),
    [anon_sym___clrcall] = ACTIONS(791),
    [anon_sym___stdcall] = ACTIONS(791),
    [anon_sym___fastcall] = ACTIONS(791),
    [anon_sym___thiscall] = ACTIONS(791),
    [anon_sym___vectorcall] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_else] = ACTIONS(791),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(791),
    [anon_sym_default] = ACTIONS(791),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(47)] = {
    [sym_declaration] = STATE(49),
    [sym_type_definition] = STATE(49),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(49),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(49),
    [sym_labeled_statement] = STATE(49),
    [sym_expression_statement] = STATE(49),
    [sym_if_statement] = STATE(49),
    [sym_switch_statement] = STATE(49),
    [sym_while_statement] = STATE(49),
    [sym_do_statement] = STATE(49),
    [sym_for_statement] = STATE(49),
    [sym_return_statement] = STATE(49),
    [sym_break_statement] = STATE(49),
    [sym_continue_statement] = STATE(49),
    [sym_goto_statement] = STATE(49),
    [sym_seh_try_statement] = STATE(49),
    [sym_seh_leave_statement] = STATE(49),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(49),
    [sym_identifier] = ACTIONS(787),
    [aux_sym_preproc_include_token1] = ACTIONS(793),
    [aux_sym_preproc_def_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token2] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(793),
    [aux_sym_preproc_else_token1] = ACTIONS(793),
    [aux_sym_preproc_elif_token1] = ACTIONS(793),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(793),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(793),
    [sym_preproc_directive] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(793),
    [anon_sym___clrcall] = ACTIONS(793),
    [anon_sym___stdcall] = ACTIONS(793),
    [anon_sym___fastcall] = ACTIONS(793),
    [anon_sym___thiscall] = ACTIONS(793),
    [anon_sym___vectorcall] = ACTIONS(793),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_else] = ACTIONS(793),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(793),
    [anon_sym_default] = ACTIONS(793),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(48)] = {
    [sym_declaration] = STATE(48),
    [sym_type_definition] = STATE(48),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(48),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(48),
    [sym_labeled_statement] = STATE(48),
    [sym_expression_statement] = STATE(48),
    [sym_if_statement] = STATE(48),
    [sym_switch_statement] = STATE(48),
    [sym_while_statement] = STATE(48),
    [sym_do_statement] = STATE(48),
    [sym_for_statement] = STATE(48),
    [sym_return_statement] = STATE(48),
    [sym_break_statement] = STATE(48),
    [sym_continue_statement] = STATE(48),
    [sym_goto_statement] = STATE(48),
    [sym_seh_try_statement] = STATE(48),
    [sym_seh_leave_statement] = STATE(48),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(48),
    [sym_identifier] = ACTIONS(795),
    [aux_sym_preproc_include_token1] = ACTIONS(798),
    [aux_sym_preproc_def_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token2] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(798),
    [aux_sym_preproc_else_token1] = ACTIONS(798),
    [aux_sym_preproc_elif_token1] = ACTIONS(798),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(798),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(798),
    [sym_preproc_directive] = ACTIONS(798),
    [anon_sym_LPAREN2] = ACTIONS(800),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_TILDE] = ACTIONS(803),
    [anon_sym_DASH] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(809),
    [anon_sym_AMP] = ACTIONS(809),
    [anon_sym_SEMI] = ACTIONS(812),
    [anon_sym___extension__] = ACTIONS(815),
    [anon_sym_typedef] = ACTIONS(818),
    [anon_sym_extern] = ACTIONS(821),
    [anon_sym___attribute__] = ACTIONS(824),
    [anon_sym___attribute] = ACTIONS(824),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(827),
    [anon_sym___declspec] = ACTIONS(830),
    [anon_sym___cdecl] = ACTIONS(798),
    [anon_sym___clrcall] = ACTIONS(798),
    [anon_sym___stdcall] = ACTIONS(798),
    [anon_sym___fastcall] = ACTIONS(798),
    [anon_sym___thiscall] = ACTIONS(798),
    [anon_sym___vectorcall] = ACTIONS(798),
    [anon_sym_LBRACE] = ACTIONS(833),
    [anon_sym_signed] = ACTIONS(836),
    [anon_sym_unsigned] = ACTIONS(836),
    [anon_sym_long] = ACTIONS(836),
    [anon_sym_short] = ACTIONS(836),
    [anon_sym_static] = ACTIONS(821),
    [anon_sym_auto] = ACTIONS(821),
    [anon_sym_register] = ACTIONS(821),
    [anon_sym_inline] = ACTIONS(821),
    [anon_sym___inline] = ACTIONS(821),
    [anon_sym___inline__] = ACTIONS(821),
    [anon_sym___forceinline] = ACTIONS(821),
    [anon_sym_thread_local] = ACTIONS(821),
    [anon_sym___thread] = ACTIONS(821),
    [anon_sym_const] = ACTIONS(839),
    [anon_sym_constexpr] = ACTIONS(839),
    [anon_sym_volatile] = ACTIONS(839),
    [anon_sym_restrict] = ACTIONS(839),
    [anon_sym___restrict__] = ACTIONS(839),
    [anon_sym__Atomic] = ACTIONS(839),
    [anon_sym__Noreturn] = ACTIONS(839),
    [anon_sym_noreturn] = ACTIONS(839),
    [anon_sym__Nonnull] = ACTIONS(839),
    [anon_sym_alignas] = ACTIONS(842),
    [anon_sym__Alignas] = ACTIONS(842),
    [sym_primitive_type] = ACTIONS(845),
    [anon_sym_enum] = ACTIONS(848),
    [anon_sym_struct] = ACTIONS(851),
    [anon_sym_union] = ACTIONS(854),
    [anon_sym_if] = ACTIONS(857),
    [anon_sym_else] = ACTIONS(798),
    [anon_sym_switch] = ACTIONS(860),
    [anon_sym_case] = ACTIONS(798),
    [anon_sym_default] = ACTIONS(798),
    [anon_sym_while] = ACTIONS(863),
    [anon_sym_do] = ACTIONS(866),
    [anon_sym_for] = ACTIONS(869),
    [anon_sym_return] = ACTIONS(872),
    [anon_sym_break] = ACTIONS(875),
    [anon_sym_continue] = ACTIONS(878),
    [anon_sym_goto] = ACTIONS(881),
    [anon_sym___try] = ACTIONS(884),
    [anon_sym___leave] = ACTIONS(887),
    [anon_sym_DASH_DASH] = ACTIONS(890),
    [anon_sym_PLUS_PLUS] = ACTIONS(890),
    [anon_sym_sizeof] = ACTIONS(893),
    [anon_sym___alignof__] = ACTIONS(896),
    [anon_sym___alignof] = ACTIONS(896),
    [anon_sym__alignof] = ACTIONS(896),
    [anon_sym_alignof] = ACTIONS(896),
    [anon_sym__Alignof] = ACTIONS(896),
    [anon_sym_offsetof] = ACTIONS(899),
    [anon_sym__Generic] = ACTIONS(902),
    [anon_sym_asm] = ACTIONS(905),
    [anon_sym___asm__] = ACTIONS(905),
    [anon_sym___asm] = ACTIONS(905),
    [sym_number_literal] = ACTIONS(908),
    [anon_sym_L_SQUOTE] = ACTIONS(911),
    [anon_sym_u_SQUOTE] = ACTIONS(911),
    [anon_sym_U_SQUOTE] = ACTIONS(911),
    [anon_sym_u8_SQUOTE] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(911),
    [anon_sym_L_DQUOTE] = ACTIONS(914),
    [anon_sym_u_DQUOTE] = ACTIONS(914),
    [anon_sym_U_DQUOTE] = ACTIONS(914),
    [anon_sym_u8_DQUOTE] = ACTIONS(914),
    [anon_sym_DQUOTE] = ACTIONS(914),
    [sym_true] = ACTIONS(917),
    [sym_false] = ACTIONS(917),
    [anon_sym_NULL] = ACTIONS(920),
    [anon_sym_nullptr] = ACTIONS(920),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(49)] = {
    [sym_declaration] = STATE(48),
    [sym_type_definition] = STATE(48),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(48),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(48),
    [sym_labeled_statement] = STATE(48),
    [sym_expression_statement] = STATE(48),
    [sym_if_statement] = STATE(48),
    [sym_switch_statement] = STATE(48),
    [sym_while_statement] = STATE(48),
    [sym_do_statement] = STATE(48),
    [sym_for_statement] = STATE(48),
    [sym_return_statement] = STATE(48),
    [sym_break_statement] = STATE(48),
    [sym_continue_statement] = STATE(48),
    [sym_goto_statement] = STATE(48),
    [sym_seh_try_statement] = STATE(48),
    [sym_seh_leave_statement] = STATE(48),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(48),
    [sym_identifier] = ACTIONS(787),
    [aux_sym_preproc_include_token1] = ACTIONS(923),
    [aux_sym_preproc_def_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token2] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(923),
    [aux_sym_preproc_else_token1] = ACTIONS(923),
    [aux_sym_preproc_elif_token1] = ACTIONS(923),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(923),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(923),
    [sym_preproc_directive] = ACTIONS(923),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(127),
    [anon_sym_typedef] = ACTIONS(129),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(923),
    [anon_sym___clrcall] = ACTIONS(923),
    [anon_sym___stdcall] = ACTIONS(923),
    [anon_sym___fastcall] = ACTIONS(923),
    [anon_sym___thiscall] = ACTIONS(923),
    [anon_sym___vectorcall] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_else] = ACTIONS(923),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(923),
    [anon_sym_default] = ACTIONS(923),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(50)] = {
    [sym_declaration] = STATE(55),
    [sym_type_definition] = STATE(55),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(55),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(55),
    [sym_labeled_statement] = STATE(55),
    [sym_expression_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_return_statement] = STATE(55),
    [sym_break_statement] = STATE(55),
    [sym_continue_statement] = STATE(55),
    [sym_goto_statement] = STATE(55),
    [sym_seh_try_statement] = STATE(55),
    [sym_seh_leave_statement] = STATE(55),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(55),
    [sym_identifier] = ACTIONS(925),
    [aux_sym_preproc_include_token1] = ACTIONS(791),
    [aux_sym_preproc_def_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(791),
    [sym_preproc_directive] = ACTIONS(791),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(791),
    [anon_sym___clrcall] = ACTIONS(791),
    [anon_sym___stdcall] = ACTIONS(791),
    [anon_sym___fastcall] = ACTIONS(791),
    [anon_sym___thiscall] = ACTIONS(791),
    [anon_sym___vectorcall] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(927),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_else] = ACTIONS(791),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(791),
    [anon_sym_default] = ACTIONS(791),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(51)] = {
    [sym_declaration] = STATE(56),
    [sym_type_definition] = STATE(56),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(56),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(56),
    [sym_labeled_statement] = STATE(56),
    [sym_expression_statement] = STATE(56),
    [sym_if_statement] = STATE(56),
    [sym_switch_statement] = STATE(56),
    [sym_while_statement] = STATE(56),
    [sym_do_statement] = STATE(56),
    [sym_for_statement] = STATE(56),
    [sym_return_statement] = STATE(56),
    [sym_break_statement] = STATE(56),
    [sym_continue_statement] = STATE(56),
    [sym_goto_statement] = STATE(56),
    [sym_seh_try_statement] = STATE(56),
    [sym_seh_leave_statement] = STATE(56),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(56),
    [ts_builtin_sym_end] = ACTIONS(929),
    [sym_identifier] = ACTIONS(931),
    [aux_sym_preproc_include_token1] = ACTIONS(789),
    [aux_sym_preproc_def_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(789),
    [sym_preproc_directive] = ACTIONS(789),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(789),
    [anon_sym___clrcall] = ACTIONS(789),
    [anon_sym___stdcall] = ACTIONS(789),
    [anon_sym___fastcall] = ACTIONS(789),
    [anon_sym___thiscall] = ACTIONS(789),
    [anon_sym___vectorcall] = ACTIONS(789),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_else] = ACTIONS(789),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(789),
    [anon_sym_default] = ACTIONS(789),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(52)] = {
    [sym_declaration] = STATE(55),
    [sym_type_definition] = STATE(55),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(55),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(55),
    [sym_labeled_statement] = STATE(55),
    [sym_expression_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_return_statement] = STATE(55),
    [sym_break_statement] = STATE(55),
    [sym_continue_statement] = STATE(55),
    [sym_goto_statement] = STATE(55),
    [sym_seh_try_statement] = STATE(55),
    [sym_seh_leave_statement] = STATE(55),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(55),
    [sym_identifier] = ACTIONS(925),
    [aux_sym_preproc_include_token1] = ACTIONS(923),
    [aux_sym_preproc_def_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(923),
    [sym_preproc_directive] = ACTIONS(923),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(923),
    [anon_sym___clrcall] = ACTIONS(923),
    [anon_sym___stdcall] = ACTIONS(923),
    [anon_sym___fastcall] = ACTIONS(923),
    [anon_sym___thiscall] = ACTIONS(923),
    [anon_sym___vectorcall] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(939),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_else] = ACTIONS(923),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(923),
    [anon_sym_default] = ACTIONS(923),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(53)] = {
    [sym_declaration] = STATE(50),
    [sym_type_definition] = STATE(50),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(50),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(50),
    [sym_labeled_statement] = STATE(50),
    [sym_expression_statement] = STATE(50),
    [sym_if_statement] = STATE(50),
    [sym_switch_statement] = STATE(50),
    [sym_while_statement] = STATE(50),
    [sym_do_statement] = STATE(50),
    [sym_for_statement] = STATE(50),
    [sym_return_statement] = STATE(50),
    [sym_break_statement] = STATE(50),
    [sym_continue_statement] = STATE(50),
    [sym_goto_statement] = STATE(50),
    [sym_seh_try_statement] = STATE(50),
    [sym_seh_leave_statement] = STATE(50),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(50),
    [sym_identifier] = ACTIONS(925),
    [aux_sym_preproc_include_token1] = ACTIONS(789),
    [aux_sym_preproc_def_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(789),
    [sym_preproc_directive] = ACTIONS(789),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(789),
    [anon_sym___clrcall] = ACTIONS(789),
    [anon_sym___stdcall] = ACTIONS(789),
    [anon_sym___fastcall] = ACTIONS(789),
    [anon_sym___thiscall] = ACTIONS(789),
    [anon_sym___vectorcall] = ACTIONS(789),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(929),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_else] = ACTIONS(789),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(789),
    [anon_sym_default] = ACTIONS(789),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(54)] = {
    [sym_declaration] = STATE(61),
    [sym_type_definition] = STATE(61),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(61),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(61),
    [sym_labeled_statement] = STATE(61),
    [sym_expression_statement] = STATE(61),
    [sym_if_statement] = STATE(61),
    [sym_switch_statement] = STATE(61),
    [sym_while_statement] = STATE(61),
    [sym_do_statement] = STATE(61),
    [sym_for_statement] = STATE(61),
    [sym_return_statement] = STATE(61),
    [sym_break_statement] = STATE(61),
    [sym_continue_statement] = STATE(61),
    [sym_goto_statement] = STATE(61),
    [sym_seh_try_statement] = STATE(61),
    [sym_seh_leave_statement] = STATE(61),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(61),
    [ts_builtin_sym_end] = ACTIONS(941),
    [sym_identifier] = ACTIONS(931),
    [aux_sym_preproc_include_token1] = ACTIONS(793),
    [aux_sym_preproc_def_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(793),
    [sym_preproc_directive] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(793),
    [anon_sym___clrcall] = ACTIONS(793),
    [anon_sym___stdcall] = ACTIONS(793),
    [anon_sym___fastcall] = ACTIONS(793),
    [anon_sym___thiscall] = ACTIONS(793),
    [anon_sym___vectorcall] = ACTIONS(793),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_else] = ACTIONS(793),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(793),
    [anon_sym_default] = ACTIONS(793),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(55)] = {
    [sym_declaration] = STATE(55),
    [sym_type_definition] = STATE(55),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(55),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(55),
    [sym_labeled_statement] = STATE(55),
    [sym_expression_statement] = STATE(55),
    [sym_if_statement] = STATE(55),
    [sym_switch_statement] = STATE(55),
    [sym_while_statement] = STATE(55),
    [sym_do_statement] = STATE(55),
    [sym_for_statement] = STATE(55),
    [sym_return_statement] = STATE(55),
    [sym_break_statement] = STATE(55),
    [sym_continue_statement] = STATE(55),
    [sym_goto_statement] = STATE(55),
    [sym_seh_try_statement] = STATE(55),
    [sym_seh_leave_statement] = STATE(55),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(55),
    [sym_identifier] = ACTIONS(943),
    [aux_sym_preproc_include_token1] = ACTIONS(798),
    [aux_sym_preproc_def_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(798),
    [sym_preproc_directive] = ACTIONS(798),
    [anon_sym_LPAREN2] = ACTIONS(800),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_TILDE] = ACTIONS(803),
    [anon_sym_DASH] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(809),
    [anon_sym_AMP] = ACTIONS(809),
    [anon_sym_SEMI] = ACTIONS(946),
    [anon_sym___extension__] = ACTIONS(949),
    [anon_sym_typedef] = ACTIONS(952),
    [anon_sym_extern] = ACTIONS(821),
    [anon_sym___attribute__] = ACTIONS(824),
    [anon_sym___attribute] = ACTIONS(824),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(827),
    [anon_sym___declspec] = ACTIONS(830),
    [anon_sym___cdecl] = ACTIONS(798),
    [anon_sym___clrcall] = ACTIONS(798),
    [anon_sym___stdcall] = ACTIONS(798),
    [anon_sym___fastcall] = ACTIONS(798),
    [anon_sym___thiscall] = ACTIONS(798),
    [anon_sym___vectorcall] = ACTIONS(798),
    [anon_sym_LBRACE] = ACTIONS(955),
    [anon_sym_RBRACE] = ACTIONS(958),
    [anon_sym_signed] = ACTIONS(836),
    [anon_sym_unsigned] = ACTIONS(836),
    [anon_sym_long] = ACTIONS(836),
    [anon_sym_short] = ACTIONS(836),
    [anon_sym_static] = ACTIONS(821),
    [anon_sym_auto] = ACTIONS(821),
    [anon_sym_register] = ACTIONS(821),
    [anon_sym_inline] = ACTIONS(821),
    [anon_sym___inline] = ACTIONS(821),
    [anon_sym___inline__] = ACTIONS(821),
    [anon_sym___forceinline] = ACTIONS(821),
    [anon_sym_thread_local] = ACTIONS(821),
    [anon_sym___thread] = ACTIONS(821),
    [anon_sym_const] = ACTIONS(839),
    [anon_sym_constexpr] = ACTIONS(839),
    [anon_sym_volatile] = ACTIONS(839),
    [anon_sym_restrict] = ACTIONS(839),
    [anon_sym___restrict__] = ACTIONS(839),
    [anon_sym__Atomic] = ACTIONS(839),
    [anon_sym__Noreturn] = ACTIONS(839),
    [anon_sym_noreturn] = ACTIONS(839),
    [anon_sym__Nonnull] = ACTIONS(839),
    [anon_sym_alignas] = ACTIONS(842),
    [anon_sym__Alignas] = ACTIONS(842),
    [sym_primitive_type] = ACTIONS(845),
    [anon_sym_enum] = ACTIONS(848),
    [anon_sym_struct] = ACTIONS(851),
    [anon_sym_union] = ACTIONS(854),
    [anon_sym_if] = ACTIONS(960),
    [anon_sym_else] = ACTIONS(798),
    [anon_sym_switch] = ACTIONS(963),
    [anon_sym_case] = ACTIONS(798),
    [anon_sym_default] = ACTIONS(798),
    [anon_sym_while] = ACTIONS(966),
    [anon_sym_do] = ACTIONS(969),
    [anon_sym_for] = ACTIONS(972),
    [anon_sym_return] = ACTIONS(975),
    [anon_sym_break] = ACTIONS(978),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_goto] = ACTIONS(984),
    [anon_sym___try] = ACTIONS(987),
    [anon_sym___leave] = ACTIONS(990),
    [anon_sym_DASH_DASH] = ACTIONS(890),
    [anon_sym_PLUS_PLUS] = ACTIONS(890),
    [anon_sym_sizeof] = ACTIONS(893),
    [anon_sym___alignof__] = ACTIONS(896),
    [anon_sym___alignof] = ACTIONS(896),
    [anon_sym__alignof] = ACTIONS(896),
    [anon_sym_alignof] = ACTIONS(896),
    [anon_sym__Alignof] = ACTIONS(896),
    [anon_sym_offsetof] = ACTIONS(899),
    [anon_sym__Generic] = ACTIONS(902),
    [anon_sym_asm] = ACTIONS(905),
    [anon_sym___asm__] = ACTIONS(905),
    [anon_sym___asm] = ACTIONS(905),
    [sym_number_literal] = ACTIONS(908),
    [anon_sym_L_SQUOTE] = ACTIONS(911),
    [anon_sym_u_SQUOTE] = ACTIONS(911),
    [anon_sym_U_SQUOTE] = ACTIONS(911),
    [anon_sym_u8_SQUOTE] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(911),
    [anon_sym_L_DQUOTE] = ACTIONS(914),
    [anon_sym_u_DQUOTE] = ACTIONS(914),
    [anon_sym_U_DQUOTE] = ACTIONS(914),
    [anon_sym_u8_DQUOTE] = ACTIONS(914),
    [anon_sym_DQUOTE] = ACTIONS(914),
    [sym_true] = ACTIONS(917),
    [sym_false] = ACTIONS(917),
    [anon_sym_NULL] = ACTIONS(920),
    [anon_sym_nullptr] = ACTIONS(920),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(56)] = {
    [sym_declaration] = STATE(63),
    [sym_type_definition] = STATE(63),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(63),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(63),
    [sym_labeled_statement] = STATE(63),
    [sym_expression_statement] = STATE(63),
    [sym_if_statement] = STATE(63),
    [sym_switch_statement] = STATE(63),
    [sym_while_statement] = STATE(63),
    [sym_do_statement] = STATE(63),
    [sym_for_statement] = STATE(63),
    [sym_return_statement] = STATE(63),
    [sym_break_statement] = STATE(63),
    [sym_continue_statement] = STATE(63),
    [sym_goto_statement] = STATE(63),
    [sym_seh_try_statement] = STATE(63),
    [sym_seh_leave_statement] = STATE(63),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(63),
    [ts_builtin_sym_end] = ACTIONS(927),
    [sym_identifier] = ACTIONS(931),
    [aux_sym_preproc_include_token1] = ACTIONS(791),
    [aux_sym_preproc_def_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(791),
    [sym_preproc_directive] = ACTIONS(791),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(791),
    [anon_sym___clrcall] = ACTIONS(791),
    [anon_sym___stdcall] = ACTIONS(791),
    [anon_sym___fastcall] = ACTIONS(791),
    [anon_sym___thiscall] = ACTIONS(791),
    [anon_sym___vectorcall] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_else] = ACTIONS(791),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(791),
    [anon_sym_default] = ACTIONS(791),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(57)] = {
    [sym_declaration] = STATE(60),
    [sym_type_definition] = STATE(60),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(60),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(60),
    [sym_labeled_statement] = STATE(60),
    [sym_expression_statement] = STATE(60),
    [sym_if_statement] = STATE(60),
    [sym_switch_statement] = STATE(60),
    [sym_while_statement] = STATE(60),
    [sym_do_statement] = STATE(60),
    [sym_for_statement] = STATE(60),
    [sym_return_statement] = STATE(60),
    [sym_break_statement] = STATE(60),
    [sym_continue_statement] = STATE(60),
    [sym_goto_statement] = STATE(60),
    [sym_seh_try_statement] = STATE(60),
    [sym_seh_leave_statement] = STATE(60),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(60),
    [sym_identifier] = ACTIONS(993),
    [aux_sym_preproc_include_token1] = ACTIONS(793),
    [aux_sym_preproc_def_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token2] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(793),
    [sym_preproc_directive] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(793),
    [anon_sym___clrcall] = ACTIONS(793),
    [anon_sym___stdcall] = ACTIONS(793),
    [anon_sym___fastcall] = ACTIONS(793),
    [anon_sym___thiscall] = ACTIONS(793),
    [anon_sym___vectorcall] = ACTIONS(793),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_else] = ACTIONS(793),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(793),
    [anon_sym_default] = ACTIONS(793),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(58)] = {
    [sym_declaration] = STATE(59),
    [sym_type_definition] = STATE(59),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(59),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(59),
    [sym_labeled_statement] = STATE(59),
    [sym_expression_statement] = STATE(59),
    [sym_if_statement] = STATE(59),
    [sym_switch_statement] = STATE(59),
    [sym_while_statement] = STATE(59),
    [sym_do_statement] = STATE(59),
    [sym_for_statement] = STATE(59),
    [sym_return_statement] = STATE(59),
    [sym_break_statement] = STATE(59),
    [sym_continue_statement] = STATE(59),
    [sym_goto_statement] = STATE(59),
    [sym_seh_try_statement] = STATE(59),
    [sym_seh_leave_statement] = STATE(59),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(59),
    [sym_identifier] = ACTIONS(993),
    [aux_sym_preproc_include_token1] = ACTIONS(789),
    [aux_sym_preproc_def_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token1] = ACTIONS(789),
    [aux_sym_preproc_if_token2] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(789),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(789),
    [sym_preproc_directive] = ACTIONS(789),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(789),
    [anon_sym___clrcall] = ACTIONS(789),
    [anon_sym___stdcall] = ACTIONS(789),
    [anon_sym___fastcall] = ACTIONS(789),
    [anon_sym___thiscall] = ACTIONS(789),
    [anon_sym___vectorcall] = ACTIONS(789),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_else] = ACTIONS(789),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(789),
    [anon_sym_default] = ACTIONS(789),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(59)] = {
    [sym_declaration] = STATE(62),
    [sym_type_definition] = STATE(62),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(62),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(62),
    [sym_labeled_statement] = STATE(62),
    [sym_expression_statement] = STATE(62),
    [sym_if_statement] = STATE(62),
    [sym_switch_statement] = STATE(62),
    [sym_while_statement] = STATE(62),
    [sym_do_statement] = STATE(62),
    [sym_for_statement] = STATE(62),
    [sym_return_statement] = STATE(62),
    [sym_break_statement] = STATE(62),
    [sym_continue_statement] = STATE(62),
    [sym_goto_statement] = STATE(62),
    [sym_seh_try_statement] = STATE(62),
    [sym_seh_leave_statement] = STATE(62),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(62),
    [sym_identifier] = ACTIONS(993),
    [aux_sym_preproc_include_token1] = ACTIONS(791),
    [aux_sym_preproc_def_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token1] = ACTIONS(791),
    [aux_sym_preproc_if_token2] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(791),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(791),
    [sym_preproc_directive] = ACTIONS(791),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(791),
    [anon_sym___clrcall] = ACTIONS(791),
    [anon_sym___stdcall] = ACTIONS(791),
    [anon_sym___fastcall] = ACTIONS(791),
    [anon_sym___thiscall] = ACTIONS(791),
    [anon_sym___vectorcall] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_else] = ACTIONS(791),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(791),
    [anon_sym_default] = ACTIONS(791),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(60)] = {
    [sym_declaration] = STATE(62),
    [sym_type_definition] = STATE(62),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(62),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(62),
    [sym_labeled_statement] = STATE(62),
    [sym_expression_statement] = STATE(62),
    [sym_if_statement] = STATE(62),
    [sym_switch_statement] = STATE(62),
    [sym_while_statement] = STATE(62),
    [sym_do_statement] = STATE(62),
    [sym_for_statement] = STATE(62),
    [sym_return_statement] = STATE(62),
    [sym_break_statement] = STATE(62),
    [sym_continue_statement] = STATE(62),
    [sym_goto_statement] = STATE(62),
    [sym_seh_try_statement] = STATE(62),
    [sym_seh_leave_statement] = STATE(62),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(62),
    [sym_identifier] = ACTIONS(993),
    [aux_sym_preproc_include_token1] = ACTIONS(923),
    [aux_sym_preproc_def_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token2] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(923),
    [sym_preproc_directive] = ACTIONS(923),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(424),
    [anon_sym_typedef] = ACTIONS(426),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(923),
    [anon_sym___clrcall] = ACTIONS(923),
    [anon_sym___stdcall] = ACTIONS(923),
    [anon_sym___fastcall] = ACTIONS(923),
    [anon_sym___thiscall] = ACTIONS(923),
    [anon_sym___vectorcall] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_else] = ACTIONS(923),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(923),
    [anon_sym_default] = ACTIONS(923),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(61)] = {
    [sym_declaration] = STATE(63),
    [sym_type_definition] = STATE(63),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(63),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(63),
    [sym_labeled_statement] = STATE(63),
    [sym_expression_statement] = STATE(63),
    [sym_if_statement] = STATE(63),
    [sym_switch_statement] = STATE(63),
    [sym_while_statement] = STATE(63),
    [sym_do_statement] = STATE(63),
    [sym_for_statement] = STATE(63),
    [sym_return_statement] = STATE(63),
    [sym_break_statement] = STATE(63),
    [sym_continue_statement] = STATE(63),
    [sym_goto_statement] = STATE(63),
    [sym_seh_try_statement] = STATE(63),
    [sym_seh_leave_statement] = STATE(63),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(63),
    [ts_builtin_sym_end] = ACTIONS(939),
    [sym_identifier] = ACTIONS(931),
    [aux_sym_preproc_include_token1] = ACTIONS(923),
    [aux_sym_preproc_def_token1] = ACTIONS(923),
    [aux_sym_preproc_if_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(923),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(923),
    [sym_preproc_directive] = ACTIONS(923),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(923),
    [anon_sym___clrcall] = ACTIONS(923),
    [anon_sym___stdcall] = ACTIONS(923),
    [anon_sym___fastcall] = ACTIONS(923),
    [anon_sym___thiscall] = ACTIONS(923),
    [anon_sym___vectorcall] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_else] = ACTIONS(923),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(923),
    [anon_sym_default] = ACTIONS(923),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(62)] = {
    [sym_declaration] = STATE(62),
    [sym_type_definition] = STATE(62),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(62),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(62),
    [sym_labeled_statement] = STATE(62),
    [sym_expression_statement] = STATE(62),
    [sym_if_statement] = STATE(62),
    [sym_switch_statement] = STATE(62),
    [sym_while_statement] = STATE(62),
    [sym_do_statement] = STATE(62),
    [sym_for_statement] = STATE(62),
    [sym_return_statement] = STATE(62),
    [sym_break_statement] = STATE(62),
    [sym_continue_statement] = STATE(62),
    [sym_goto_statement] = STATE(62),
    [sym_seh_try_statement] = STATE(62),
    [sym_seh_leave_statement] = STATE(62),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(62),
    [sym_identifier] = ACTIONS(995),
    [aux_sym_preproc_include_token1] = ACTIONS(798),
    [aux_sym_preproc_def_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token2] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(798),
    [sym_preproc_directive] = ACTIONS(798),
    [anon_sym_LPAREN2] = ACTIONS(800),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_TILDE] = ACTIONS(803),
    [anon_sym_DASH] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(809),
    [anon_sym_AMP] = ACTIONS(809),
    [anon_sym_SEMI] = ACTIONS(998),
    [anon_sym___extension__] = ACTIONS(1001),
    [anon_sym_typedef] = ACTIONS(1004),
    [anon_sym_extern] = ACTIONS(821),
    [anon_sym___attribute__] = ACTIONS(824),
    [anon_sym___attribute] = ACTIONS(824),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(827),
    [anon_sym___declspec] = ACTIONS(830),
    [anon_sym___cdecl] = ACTIONS(798),
    [anon_sym___clrcall] = ACTIONS(798),
    [anon_sym___stdcall] = ACTIONS(798),
    [anon_sym___fastcall] = ACTIONS(798),
    [anon_sym___thiscall] = ACTIONS(798),
    [anon_sym___vectorcall] = ACTIONS(798),
    [anon_sym_LBRACE] = ACTIONS(1007),
    [anon_sym_signed] = ACTIONS(836),
    [anon_sym_unsigned] = ACTIONS(836),
    [anon_sym_long] = ACTIONS(836),
    [anon_sym_short] = ACTIONS(836),
    [anon_sym_static] = ACTIONS(821),
    [anon_sym_auto] = ACTIONS(821),
    [anon_sym_register] = ACTIONS(821),
    [anon_sym_inline] = ACTIONS(821),
    [anon_sym___inline] = ACTIONS(821),
    [anon_sym___inline__] = ACTIONS(821),
    [anon_sym___forceinline] = ACTIONS(821),
    [anon_sym_thread_local] = ACTIONS(821),
    [anon_sym___thread] = ACTIONS(821),
    [anon_sym_const] = ACTIONS(839),
    [anon_sym_constexpr] = ACTIONS(839),
    [anon_sym_volatile] = ACTIONS(839),
    [anon_sym_restrict] = ACTIONS(839),
    [anon_sym___restrict__] = ACTIONS(839),
    [anon_sym__Atomic] = ACTIONS(839),
    [anon_sym__Noreturn] = ACTIONS(839),
    [anon_sym_noreturn] = ACTIONS(839),
    [anon_sym__Nonnull] = ACTIONS(839),
    [anon_sym_alignas] = ACTIONS(842),
    [anon_sym__Alignas] = ACTIONS(842),
    [sym_primitive_type] = ACTIONS(845),
    [anon_sym_enum] = ACTIONS(848),
    [anon_sym_struct] = ACTIONS(851),
    [anon_sym_union] = ACTIONS(854),
    [anon_sym_if] = ACTIONS(1010),
    [anon_sym_else] = ACTIONS(798),
    [anon_sym_switch] = ACTIONS(1013),
    [anon_sym_case] = ACTIONS(798),
    [anon_sym_default] = ACTIONS(798),
    [anon_sym_while] = ACTIONS(1016),
    [anon_sym_do] = ACTIONS(1019),
    [anon_sym_for] = ACTIONS(1022),
    [anon_sym_return] = ACTIONS(1025),
    [anon_sym_break] = ACTIONS(1028),
    [anon_sym_continue] = ACTIONS(1031),
    [anon_sym_goto] = ACTIONS(1034),
    [anon_sym___try] = ACTIONS(1037),
    [anon_sym___leave] = ACTIONS(1040),
    [anon_sym_DASH_DASH] = ACTIONS(890),
    [anon_sym_PLUS_PLUS] = ACTIONS(890),
    [anon_sym_sizeof] = ACTIONS(893),
    [anon_sym___alignof__] = ACTIONS(896),
    [anon_sym___alignof] = ACTIONS(896),
    [anon_sym__alignof] = ACTIONS(896),
    [anon_sym_alignof] = ACTIONS(896),
    [anon_sym__Alignof] = ACTIONS(896),
    [anon_sym_offsetof] = ACTIONS(899),
    [anon_sym__Generic] = ACTIONS(902),
    [anon_sym_asm] = ACTIONS(905),
    [anon_sym___asm__] = ACTIONS(905),
    [anon_sym___asm] = ACTIONS(905),
    [sym_number_literal] = ACTIONS(908),
    [anon_sym_L_SQUOTE] = ACTIONS(911),
    [anon_sym_u_SQUOTE] = ACTIONS(911),
    [anon_sym_U_SQUOTE] = ACTIONS(911),
    [anon_sym_u8_SQUOTE] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(911),
    [anon_sym_L_DQUOTE] = ACTIONS(914),
    [anon_sym_u_DQUOTE] = ACTIONS(914),
    [anon_sym_U_DQUOTE] = ACTIONS(914),
    [anon_sym_u8_DQUOTE] = ACTIONS(914),
    [anon_sym_DQUOTE] = ACTIONS(914),
    [sym_true] = ACTIONS(917),
    [sym_false] = ACTIONS(917),
    [anon_sym_NULL] = ACTIONS(920),
    [anon_sym_nullptr] = ACTIONS(920),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(63)] = {
    [sym_declaration] = STATE(63),
    [sym_type_definition] = STATE(63),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(63),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(63),
    [sym_labeled_statement] = STATE(63),
    [sym_expression_statement] = STATE(63),
    [sym_if_statement] = STATE(63),
    [sym_switch_statement] = STATE(63),
    [sym_while_statement] = STATE(63),
    [sym_do_statement] = STATE(63),
    [sym_for_statement] = STATE(63),
    [sym_return_statement] = STATE(63),
    [sym_break_statement] = STATE(63),
    [sym_continue_statement] = STATE(63),
    [sym_goto_statement] = STATE(63),
    [sym_seh_try_statement] = STATE(63),
    [sym_seh_leave_statement] = STATE(63),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(63),
    [ts_builtin_sym_end] = ACTIONS(958),
    [sym_identifier] = ACTIONS(1043),
    [aux_sym_preproc_include_token1] = ACTIONS(798),
    [aux_sym_preproc_def_token1] = ACTIONS(798),
    [aux_sym_preproc_if_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(798),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(798),
    [sym_preproc_directive] = ACTIONS(798),
    [anon_sym_LPAREN2] = ACTIONS(800),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_TILDE] = ACTIONS(803),
    [anon_sym_DASH] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(809),
    [anon_sym_AMP] = ACTIONS(809),
    [anon_sym_SEMI] = ACTIONS(1046),
    [anon_sym___extension__] = ACTIONS(1049),
    [anon_sym_typedef] = ACTIONS(1052),
    [anon_sym_extern] = ACTIONS(821),
    [anon_sym___attribute__] = ACTIONS(824),
    [anon_sym___attribute] = ACTIONS(824),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(827),
    [anon_sym___declspec] = ACTIONS(830),
    [anon_sym___cdecl] = ACTIONS(798),
    [anon_sym___clrcall] = ACTIONS(798),
    [anon_sym___stdcall] = ACTIONS(798),
    [anon_sym___fastcall] = ACTIONS(798),
    [anon_sym___thiscall] = ACTIONS(798),
    [anon_sym___vectorcall] = ACTIONS(798),
    [anon_sym_LBRACE] = ACTIONS(1055),
    [anon_sym_signed] = ACTIONS(836),
    [anon_sym_unsigned] = ACTIONS(836),
    [anon_sym_long] = ACTIONS(836),
    [anon_sym_short] = ACTIONS(836),
    [anon_sym_static] = ACTIONS(821),
    [anon_sym_auto] = ACTIONS(821),
    [anon_sym_register] = ACTIONS(821),
    [anon_sym_inline] = ACTIONS(821),
    [anon_sym___inline] = ACTIONS(821),
    [anon_sym___inline__] = ACTIONS(821),
    [anon_sym___forceinline] = ACTIONS(821),
    [anon_sym_thread_local] = ACTIONS(821),
    [anon_sym___thread] = ACTIONS(821),
    [anon_sym_const] = ACTIONS(839),
    [anon_sym_constexpr] = ACTIONS(839),
    [anon_sym_volatile] = ACTIONS(839),
    [anon_sym_restrict] = ACTIONS(839),
    [anon_sym___restrict__] = ACTIONS(839),
    [anon_sym__Atomic] = ACTIONS(839),
    [anon_sym__Noreturn] = ACTIONS(839),
    [anon_sym_noreturn] = ACTIONS(839),
    [anon_sym__Nonnull] = ACTIONS(839),
    [anon_sym_alignas] = ACTIONS(842),
    [anon_sym__Alignas] = ACTIONS(842),
    [sym_primitive_type] = ACTIONS(845),
    [anon_sym_enum] = ACTIONS(848),
    [anon_sym_struct] = ACTIONS(851),
    [anon_sym_union] = ACTIONS(854),
    [anon_sym_if] = ACTIONS(1058),
    [anon_sym_else] = ACTIONS(798),
    [anon_sym_switch] = ACTIONS(1061),
    [anon_sym_case] = ACTIONS(798),
    [anon_sym_default] = ACTIONS(798),
    [anon_sym_while] = ACTIONS(1064),
    [anon_sym_do] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1070),
    [anon_sym_return] = ACTIONS(1073),
    [anon_sym_break] = ACTIONS(1076),
    [anon_sym_continue] = ACTIONS(1079),
    [anon_sym_goto] = ACTIONS(1082),
    [anon_sym___try] = ACTIONS(1085),
    [anon_sym___leave] = ACTIONS(1088),
    [anon_sym_DASH_DASH] = ACTIONS(890),
    [anon_sym_PLUS_PLUS] = ACTIONS(890),
    [anon_sym_sizeof] = ACTIONS(893),
    [anon_sym___alignof__] = ACTIONS(896),
    [anon_sym___alignof] = ACTIONS(896),
    [anon_sym__alignof] = ACTIONS(896),
    [anon_sym_alignof] = ACTIONS(896),
    [anon_sym__Alignof] = ACTIONS(896),
    [anon_sym_offsetof] = ACTIONS(899),
    [anon_sym__Generic] = ACTIONS(902),
    [anon_sym_asm] = ACTIONS(905),
    [anon_sym___asm__] = ACTIONS(905),
    [anon_sym___asm] = ACTIONS(905),
    [sym_number_literal] = ACTIONS(908),
    [anon_sym_L_SQUOTE] = ACTIONS(911),
    [anon_sym_u_SQUOTE] = ACTIONS(911),
    [anon_sym_U_SQUOTE] = ACTIONS(911),
    [anon_sym_u8_SQUOTE] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(911),
    [anon_sym_L_DQUOTE] = ACTIONS(914),
    [anon_sym_u_DQUOTE] = ACTIONS(914),
    [anon_sym_U_DQUOTE] = ACTIONS(914),
    [anon_sym_u8_DQUOTE] = ACTIONS(914),
    [anon_sym_DQUOTE] = ACTIONS(914),
    [sym_true] = ACTIONS(917),
    [sym_false] = ACTIONS(917),
    [anon_sym_NULL] = ACTIONS(920),
    [anon_sym_nullptr] = ACTIONS(920),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(64)] = {
    [sym_declaration] = STATE(52),
    [sym_type_definition] = STATE(52),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(52),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(52),
    [sym_labeled_statement] = STATE(52),
    [sym_expression_statement] = STATE(52),
    [sym_if_statement] = STATE(52),
    [sym_switch_statement] = STATE(52),
    [sym_while_statement] = STATE(52),
    [sym_do_statement] = STATE(52),
    [sym_for_statement] = STATE(52),
    [sym_return_statement] = STATE(52),
    [sym_break_statement] = STATE(52),
    [sym_continue_statement] = STATE(52),
    [sym_goto_statement] = STATE(52),
    [sym_seh_try_statement] = STATE(52),
    [sym_seh_leave_statement] = STATE(52),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(52),
    [sym_identifier] = ACTIONS(925),
    [aux_sym_preproc_include_token1] = ACTIONS(793),
    [aux_sym_preproc_def_token1] = ACTIONS(793),
    [aux_sym_preproc_if_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(793),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(793),
    [sym_preproc_directive] = ACTIONS(793),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(372),
    [anon_sym_typedef] = ACTIONS(374),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(793),
    [anon_sym___clrcall] = ACTIONS(793),
    [anon_sym___stdcall] = ACTIONS(793),
    [anon_sym___fastcall] = ACTIONS(793),
    [anon_sym___thiscall] = ACTIONS(793),
    [anon_sym___vectorcall] = ACTIONS(793),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_RBRACE] = ACTIONS(941),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_else] = ACTIONS(793),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(793),
    [anon_sym_default] = ACTIONS(793),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(65)] = {
    [sym_declaration] = STATE(190),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(154),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(190),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(931),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(66)] = {
    [sym_declaration] = STATE(219),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1140),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(219),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(925),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(67)] = {
    [sym_declaration] = STATE(74),
    [sym_type_definition] = STATE(74),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(74),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(74),
    [sym_labeled_statement] = STATE(74),
    [sym_expression_statement] = STATE(74),
    [sym_if_statement] = STATE(74),
    [sym_switch_statement] = STATE(74),
    [sym_while_statement] = STATE(74),
    [sym_do_statement] = STATE(74),
    [sym_for_statement] = STATE(74),
    [sym_return_statement] = STATE(74),
    [sym_break_statement] = STATE(74),
    [sym_continue_statement] = STATE(74),
    [sym_goto_statement] = STATE(74),
    [sym_seh_try_statement] = STATE(74),
    [sym_seh_leave_statement] = STATE(74),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(74),
    [sym_identifier] = ACTIONS(1093),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_else] = ACTIONS(791),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(68)] = {
    [sym_declaration] = STATE(99),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1142),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(81),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(99),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(787),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(69)] = {
    [sym_declaration] = STATE(67),
    [sym_type_definition] = STATE(67),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(67),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(67),
    [sym_labeled_statement] = STATE(67),
    [sym_expression_statement] = STATE(67),
    [sym_if_statement] = STATE(67),
    [sym_switch_statement] = STATE(67),
    [sym_while_statement] = STATE(67),
    [sym_do_statement] = STATE(67),
    [sym_for_statement] = STATE(67),
    [sym_return_statement] = STATE(67),
    [sym_break_statement] = STATE(67),
    [sym_continue_statement] = STATE(67),
    [sym_goto_statement] = STATE(67),
    [sym_seh_try_statement] = STATE(67),
    [sym_seh_leave_statement] = STATE(67),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(67),
    [sym_identifier] = ACTIONS(1093),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_else] = ACTIONS(789),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(70)] = {
    [sym_declaration] = STATE(190),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(244),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(190),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1093),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(71)] = {
    [sym_declaration] = STATE(73),
    [sym_type_definition] = STATE(73),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(73),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(73),
    [sym_labeled_statement] = STATE(73),
    [sym_expression_statement] = STATE(73),
    [sym_if_statement] = STATE(73),
    [sym_switch_statement] = STATE(73),
    [sym_while_statement] = STATE(73),
    [sym_do_statement] = STATE(73),
    [sym_for_statement] = STATE(73),
    [sym_return_statement] = STATE(73),
    [sym_break_statement] = STATE(73),
    [sym_continue_statement] = STATE(73),
    [sym_goto_statement] = STATE(73),
    [sym_seh_try_statement] = STATE(73),
    [sym_seh_leave_statement] = STATE(73),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(73),
    [sym_identifier] = ACTIONS(1093),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_else] = ACTIONS(793),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(72)] = {
    [sym_declaration] = STATE(194),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1147),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(178),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(194),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(73)] = {
    [sym_declaration] = STATE(74),
    [sym_type_definition] = STATE(74),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(74),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(74),
    [sym_labeled_statement] = STATE(74),
    [sym_expression_statement] = STATE(74),
    [sym_if_statement] = STATE(74),
    [sym_switch_statement] = STATE(74),
    [sym_while_statement] = STATE(74),
    [sym_do_statement] = STATE(74),
    [sym_for_statement] = STATE(74),
    [sym_return_statement] = STATE(74),
    [sym_break_statement] = STATE(74),
    [sym_continue_statement] = STATE(74),
    [sym_goto_statement] = STATE(74),
    [sym_seh_try_statement] = STATE(74),
    [sym_seh_leave_statement] = STATE(74),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(74),
    [sym_identifier] = ACTIONS(1093),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(29),
    [anon_sym_typedef] = ACTIONS(31),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(37),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_else] = ACTIONS(923),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(74)] = {
    [sym_declaration] = STATE(74),
    [sym_type_definition] = STATE(74),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1155),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(399),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_compound_statement] = STATE(74),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_attributed_statement] = STATE(74),
    [sym_labeled_statement] = STATE(74),
    [sym_expression_statement] = STATE(74),
    [sym_if_statement] = STATE(74),
    [sym_switch_statement] = STATE(74),
    [sym_while_statement] = STATE(74),
    [sym_do_statement] = STATE(74),
    [sym_for_statement] = STATE(74),
    [sym_return_statement] = STATE(74),
    [sym_break_statement] = STATE(74),
    [sym_continue_statement] = STATE(74),
    [sym_goto_statement] = STATE(74),
    [sym_seh_try_statement] = STATE(74),
    [sym_seh_leave_statement] = STATE(74),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [aux_sym_case_statement_repeat1] = STATE(74),
    [sym_identifier] = ACTIONS(1107),
    [anon_sym_LPAREN2] = ACTIONS(800),
    [anon_sym_BANG] = ACTIONS(803),
    [anon_sym_TILDE] = ACTIONS(803),
    [anon_sym_DASH] = ACTIONS(806),
    [anon_sym_PLUS] = ACTIONS(806),
    [anon_sym_STAR] = ACTIONS(809),
    [anon_sym_AMP] = ACTIONS(809),
    [anon_sym_SEMI] = ACTIONS(946),
    [anon_sym___extension__] = ACTIONS(1049),
    [anon_sym_typedef] = ACTIONS(1052),
    [anon_sym_extern] = ACTIONS(821),
    [anon_sym___attribute__] = ACTIONS(824),
    [anon_sym___attribute] = ACTIONS(824),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(827),
    [anon_sym___declspec] = ACTIONS(830),
    [anon_sym_LBRACE] = ACTIONS(1055),
    [anon_sym_signed] = ACTIONS(836),
    [anon_sym_unsigned] = ACTIONS(836),
    [anon_sym_long] = ACTIONS(836),
    [anon_sym_short] = ACTIONS(836),
    [anon_sym_static] = ACTIONS(821),
    [anon_sym_auto] = ACTIONS(821),
    [anon_sym_register] = ACTIONS(821),
    [anon_sym_inline] = ACTIONS(821),
    [anon_sym___inline] = ACTIONS(821),
    [anon_sym___inline__] = ACTIONS(821),
    [anon_sym___forceinline] = ACTIONS(821),
    [anon_sym_thread_local] = ACTIONS(821),
    [anon_sym___thread] = ACTIONS(821),
    [anon_sym_const] = ACTIONS(839),
    [anon_sym_constexpr] = ACTIONS(839),
    [anon_sym_volatile] = ACTIONS(839),
    [anon_sym_restrict] = ACTIONS(839),
    [anon_sym___restrict__] = ACTIONS(839),
    [anon_sym__Atomic] = ACTIONS(839),
    [anon_sym__Noreturn] = ACTIONS(839),
    [anon_sym_noreturn] = ACTIONS(839),
    [anon_sym__Nonnull] = ACTIONS(839),
    [anon_sym_alignas] = ACTIONS(842),
    [anon_sym__Alignas] = ACTIONS(842),
    [sym_primitive_type] = ACTIONS(845),
    [anon_sym_enum] = ACTIONS(848),
    [anon_sym_struct] = ACTIONS(851),
    [anon_sym_union] = ACTIONS(854),
    [anon_sym_if] = ACTIONS(1110),
    [anon_sym_else] = ACTIONS(798),
    [anon_sym_switch] = ACTIONS(1061),
    [anon_sym_while] = ACTIONS(1113),
    [anon_sym_do] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1116),
    [anon_sym_return] = ACTIONS(1073),
    [anon_sym_break] = ACTIONS(1076),
    [anon_sym_continue] = ACTIONS(1079),
    [anon_sym_goto] = ACTIONS(1082),
    [anon_sym___try] = ACTIONS(1119),
    [anon_sym___leave] = ACTIONS(990),
    [anon_sym_DASH_DASH] = ACTIONS(890),
    [anon_sym_PLUS_PLUS] = ACTIONS(890),
    [anon_sym_sizeof] = ACTIONS(893),
    [anon_sym___alignof__] = ACTIONS(896),
    [anon_sym___alignof] = ACTIONS(896),
    [anon_sym__alignof] = ACTIONS(896),
    [anon_sym_alignof] = ACTIONS(896),
    [anon_sym__Alignof] = ACTIONS(896),
    [anon_sym_offsetof] = ACTIONS(899),
    [anon_sym__Generic] = ACTIONS(902),
    [anon_sym_asm] = ACTIONS(905),
    [anon_sym___asm__] = ACTIONS(905),
    [anon_sym___asm] = ACTIONS(905),
    [sym_number_literal] = ACTIONS(908),
    [anon_sym_L_SQUOTE] = ACTIONS(911),
    [anon_sym_u_SQUOTE] = ACTIONS(911),
    [anon_sym_U_SQUOTE] = ACTIONS(911),
    [anon_sym_u8_SQUOTE] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(911),
    [anon_sym_L_DQUOTE] = ACTIONS(914),
    [anon_sym_u_DQUOTE] = ACTIONS(914),
    [anon_sym_U_DQUOTE] = ACTIONS(914),
    [anon_sym_u8_DQUOTE] = ACTIONS(914),
    [anon_sym_DQUOTE] = ACTIONS(914),
    [sym_true] = ACTIONS(917),
    [sym_false] = ACTIONS(917),
    [anon_sym_NULL] = ACTIONS(920),
    [anon_sym_nullptr] = ACTIONS(920),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(75)] = {
    [sym_declaration] = STATE(455),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1151),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__for_statement_body] = STATE(1934),
    [sym_expression] = STATE(1042),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1800),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(76)] = {
    [sym_declaration] = STATE(455),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1151),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__for_statement_body] = STATE(1861),
    [sym_expression] = STATE(1042),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1800),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(77)] = {
    [sym_declaration] = STATE(455),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1151),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__for_statement_body] = STATE(1926),
    [sym_expression] = STATE(1042),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1800),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(78)] = {
    [sym_declaration] = STATE(455),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1151),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__for_statement_body] = STATE(1922),
    [sym_expression] = STATE(1042),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1800),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(79)] = {
    [sym_declaration] = STATE(455),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1151),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__for_statement_body] = STATE(1998),
    [sym_expression] = STATE(1042),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1800),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(80)] = {
    [sym_else_clause] = STATE(107),
    [sym_identifier] = ACTIONS(1128),
    [aux_sym_preproc_include_token1] = ACTIONS(1128),
    [aux_sym_preproc_def_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token2] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1128),
    [aux_sym_preproc_else_token1] = ACTIONS(1128),
    [aux_sym_preproc_elif_token1] = ACTIONS(1128),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1128),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1128),
    [sym_preproc_directive] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1130),
    [anon_sym_AMP] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [anon_sym___extension__] = ACTIONS(1128),
    [anon_sym_typedef] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1128),
    [anon_sym___attribute__] = ACTIONS(1128),
    [anon_sym___attribute] = ACTIONS(1128),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1130),
    [anon_sym___declspec] = ACTIONS(1128),
    [anon_sym___cdecl] = ACTIONS(1128),
    [anon_sym___clrcall] = ACTIONS(1128),
    [anon_sym___stdcall] = ACTIONS(1128),
    [anon_sym___fastcall] = ACTIONS(1128),
    [anon_sym___thiscall] = ACTIONS(1128),
    [anon_sym___vectorcall] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_signed] = ACTIONS(1128),
    [anon_sym_unsigned] = ACTIONS(1128),
    [anon_sym_long] = ACTIONS(1128),
    [anon_sym_short] = ACTIONS(1128),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_auto] = ACTIONS(1128),
    [anon_sym_register] = ACTIONS(1128),
    [anon_sym_inline] = ACTIONS(1128),
    [anon_sym___inline] = ACTIONS(1128),
    [anon_sym___inline__] = ACTIONS(1128),
    [anon_sym___forceinline] = ACTIONS(1128),
    [anon_sym_thread_local] = ACTIONS(1128),
    [anon_sym___thread] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_constexpr] = ACTIONS(1128),
    [anon_sym_volatile] = ACTIONS(1128),
    [anon_sym_restrict] = ACTIONS(1128),
    [anon_sym___restrict__] = ACTIONS(1128),
    [anon_sym__Atomic] = ACTIONS(1128),
    [anon_sym__Noreturn] = ACTIONS(1128),
    [anon_sym_noreturn] = ACTIONS(1128),
    [anon_sym__Nonnull] = ACTIONS(1128),
    [anon_sym_alignas] = ACTIONS(1128),
    [anon_sym__Alignas] = ACTIONS(1128),
    [sym_primitive_type] = ACTIONS(1128),
    [anon_sym_enum] = ACTIONS(1128),
    [anon_sym_struct] = ACTIONS(1128),
    [anon_sym_union] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1132),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_case] = ACTIONS(1128),
    [anon_sym_default] = ACTIONS(1128),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_goto] = ACTIONS(1128),
    [anon_sym___try] = ACTIONS(1128),
    [anon_sym___leave] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1128),
    [anon_sym___alignof] = ACTIONS(1128),
    [anon_sym__alignof] = ACTIONS(1128),
    [anon_sym_alignof] = ACTIONS(1128),
    [anon_sym__Alignof] = ACTIONS(1128),
    [anon_sym_offsetof] = ACTIONS(1128),
    [anon_sym__Generic] = ACTIONS(1128),
    [anon_sym_asm] = ACTIONS(1128),
    [anon_sym___asm__] = ACTIONS(1128),
    [anon_sym___asm] = ACTIONS(1128),
    [sym_number_literal] = ACTIONS(1130),
    [anon_sym_L_SQUOTE] = ACTIONS(1130),
    [anon_sym_u_SQUOTE] = ACTIONS(1130),
    [anon_sym_U_SQUOTE] = ACTIONS(1130),
    [anon_sym_u8_SQUOTE] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_L_DQUOTE] = ACTIONS(1130),
    [anon_sym_u_DQUOTE] = ACTIONS(1130),
    [anon_sym_U_DQUOTE] = ACTIONS(1130),
    [anon_sym_u8_DQUOTE] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [anon_sym_NULL] = ACTIONS(1128),
    [anon_sym_nullptr] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(81)] = {
    [sym_identifier] = ACTIONS(1134),
    [aux_sym_preproc_include_token1] = ACTIONS(1134),
    [aux_sym_preproc_def_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token2] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1134),
    [aux_sym_preproc_else_token1] = ACTIONS(1134),
    [aux_sym_preproc_elif_token1] = ACTIONS(1134),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1134),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1134),
    [sym_preproc_directive] = ACTIONS(1134),
    [anon_sym_LPAREN2] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_TILDE] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1134),
    [anon_sym_PLUS] = ACTIONS(1134),
    [anon_sym_STAR] = ACTIONS(1136),
    [anon_sym_AMP] = ACTIONS(1136),
    [anon_sym_SEMI] = ACTIONS(1136),
    [anon_sym___extension__] = ACTIONS(1134),
    [anon_sym_typedef] = ACTIONS(1134),
    [anon_sym_extern] = ACTIONS(1134),
    [anon_sym___attribute__] = ACTIONS(1134),
    [anon_sym___attribute] = ACTIONS(1134),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1136),
    [anon_sym___declspec] = ACTIONS(1134),
    [anon_sym___cdecl] = ACTIONS(1134),
    [anon_sym___clrcall] = ACTIONS(1134),
    [anon_sym___stdcall] = ACTIONS(1134),
    [anon_sym___fastcall] = ACTIONS(1134),
    [anon_sym___thiscall] = ACTIONS(1134),
    [anon_sym___vectorcall] = ACTIONS(1134),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_signed] = ACTIONS(1134),
    [anon_sym_unsigned] = ACTIONS(1134),
    [anon_sym_long] = ACTIONS(1134),
    [anon_sym_short] = ACTIONS(1134),
    [anon_sym_static] = ACTIONS(1134),
    [anon_sym_auto] = ACTIONS(1134),
    [anon_sym_register] = ACTIONS(1134),
    [anon_sym_inline] = ACTIONS(1134),
    [anon_sym___inline] = ACTIONS(1134),
    [anon_sym___inline__] = ACTIONS(1134),
    [anon_sym___forceinline] = ACTIONS(1134),
    [anon_sym_thread_local] = ACTIONS(1134),
    [anon_sym___thread] = ACTIONS(1134),
    [anon_sym_const] = ACTIONS(1134),
    [anon_sym_constexpr] = ACTIONS(1134),
    [anon_sym_volatile] = ACTIONS(1134),
    [anon_sym_restrict] = ACTIONS(1134),
    [anon_sym___restrict__] = ACTIONS(1134),
    [anon_sym__Atomic] = ACTIONS(1134),
    [anon_sym__Noreturn] = ACTIONS(1134),
    [anon_sym_noreturn] = ACTIONS(1134),
    [anon_sym__Nonnull] = ACTIONS(1134),
    [anon_sym_alignas] = ACTIONS(1134),
    [anon_sym__Alignas] = ACTIONS(1134),
    [sym_primitive_type] = ACTIONS(1134),
    [anon_sym_enum] = ACTIONS(1134),
    [anon_sym_struct] = ACTIONS(1134),
    [anon_sym_union] = ACTIONS(1134),
    [anon_sym_if] = ACTIONS(1134),
    [anon_sym_else] = ACTIONS(1134),
    [anon_sym_switch] = ACTIONS(1134),
    [anon_sym_case] = ACTIONS(1134),
    [anon_sym_default] = ACTIONS(1134),
    [anon_sym_while] = ACTIONS(1134),
    [anon_sym_do] = ACTIONS(1134),
    [anon_sym_for] = ACTIONS(1134),
    [anon_sym_return] = ACTIONS(1134),
    [anon_sym_break] = ACTIONS(1134),
    [anon_sym_continue] = ACTIONS(1134),
    [anon_sym_goto] = ACTIONS(1134),
    [anon_sym___try] = ACTIONS(1134),
    [anon_sym___leave] = ACTIONS(1134),
    [anon_sym_DASH_DASH] = ACTIONS(1136),
    [anon_sym_PLUS_PLUS] = ACTIONS(1136),
    [anon_sym_sizeof] = ACTIONS(1134),
    [anon_sym___alignof__] = ACTIONS(1134),
    [anon_sym___alignof] = ACTIONS(1134),
    [anon_sym__alignof] = ACTIONS(1134),
    [anon_sym_alignof] = ACTIONS(1134),
    [anon_sym__Alignof] = ACTIONS(1134),
    [anon_sym_offsetof] = ACTIONS(1134),
    [anon_sym__Generic] = ACTIONS(1134),
    [anon_sym_asm] = ACTIONS(1134),
    [anon_sym___asm__] = ACTIONS(1134),
    [anon_sym___asm] = ACTIONS(1134),
    [sym_number_literal] = ACTIONS(1136),
    [anon_sym_L_SQUOTE] = ACTIONS(1136),
    [anon_sym_u_SQUOTE] = ACTIONS(1136),
    [anon_sym_U_SQUOTE] = ACTIONS(1136),
    [anon_sym_u8_SQUOTE] = ACTIONS(1136),
    [anon_sym_SQUOTE] = ACTIONS(1136),
    [anon_sym_L_DQUOTE] = ACTIONS(1136),
    [anon_sym_u_DQUOTE] = ACTIONS(1136),
    [anon_sym_U_DQUOTE] = ACTIONS(1136),
    [anon_sym_u8_DQUOTE] = ACTIONS(1136),
    [anon_sym_DQUOTE] = ACTIONS(1136),
    [sym_true] = ACTIONS(1134),
    [sym_false] = ACTIONS(1134),
    [anon_sym_NULL] = ACTIONS(1134),
    [anon_sym_nullptr] = ACTIONS(1134),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(82)] = {
    [sym_identifier] = ACTIONS(1138),
    [aux_sym_preproc_include_token1] = ACTIONS(1138),
    [aux_sym_preproc_def_token1] = ACTIONS(1138),
    [aux_sym_preproc_if_token1] = ACTIONS(1138),
    [aux_sym_preproc_if_token2] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1138),
    [aux_sym_preproc_else_token1] = ACTIONS(1138),
    [aux_sym_preproc_elif_token1] = ACTIONS(1138),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1138),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1138),
    [sym_preproc_directive] = ACTIONS(1138),
    [anon_sym_LPAREN2] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_TILDE] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1138),
    [anon_sym_PLUS] = ACTIONS(1138),
    [anon_sym_STAR] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(1140),
    [anon_sym_SEMI] = ACTIONS(1140),
    [anon_sym___extension__] = ACTIONS(1138),
    [anon_sym_typedef] = ACTIONS(1138),
    [anon_sym_extern] = ACTIONS(1138),
    [anon_sym___attribute__] = ACTIONS(1138),
    [anon_sym___attribute] = ACTIONS(1138),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1140),
    [anon_sym___declspec] = ACTIONS(1138),
    [anon_sym___cdecl] = ACTIONS(1138),
    [anon_sym___clrcall] = ACTIONS(1138),
    [anon_sym___stdcall] = ACTIONS(1138),
    [anon_sym___fastcall] = ACTIONS(1138),
    [anon_sym___thiscall] = ACTIONS(1138),
    [anon_sym___vectorcall] = ACTIONS(1138),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_signed] = ACTIONS(1138),
    [anon_sym_unsigned] = ACTIONS(1138),
    [anon_sym_long] = ACTIONS(1138),
    [anon_sym_short] = ACTIONS(1138),
    [anon_sym_static] = ACTIONS(1138),
    [anon_sym_auto] = ACTIONS(1138),
    [anon_sym_register] = ACTIONS(1138),
    [anon_sym_inline] = ACTIONS(1138),
    [anon_sym___inline] = ACTIONS(1138),
    [anon_sym___inline__] = ACTIONS(1138),
    [anon_sym___forceinline] = ACTIONS(1138),
    [anon_sym_thread_local] = ACTIONS(1138),
    [anon_sym___thread] = ACTIONS(1138),
    [anon_sym_const] = ACTIONS(1138),
    [anon_sym_constexpr] = ACTIONS(1138),
    [anon_sym_volatile] = ACTIONS(1138),
    [anon_sym_restrict] = ACTIONS(1138),
    [anon_sym___restrict__] = ACTIONS(1138),
    [anon_sym__Atomic] = ACTIONS(1138),
    [anon_sym__Noreturn] = ACTIONS(1138),
    [anon_sym_noreturn] = ACTIONS(1138),
    [anon_sym__Nonnull] = ACTIONS(1138),
    [anon_sym_alignas] = ACTIONS(1138),
    [anon_sym__Alignas] = ACTIONS(1138),
    [sym_primitive_type] = ACTIONS(1138),
    [anon_sym_enum] = ACTIONS(1138),
    [anon_sym_struct] = ACTIONS(1138),
    [anon_sym_union] = ACTIONS(1138),
    [anon_sym_if] = ACTIONS(1138),
    [anon_sym_else] = ACTIONS(1138),
    [anon_sym_switch] = ACTIONS(1138),
    [anon_sym_case] = ACTIONS(1138),
    [anon_sym_default] = ACTIONS(1138),
    [anon_sym_while] = ACTIONS(1138),
    [anon_sym_do] = ACTIONS(1138),
    [anon_sym_for] = ACTIONS(1138),
    [anon_sym_return] = ACTIONS(1138),
    [anon_sym_break] = ACTIONS(1138),
    [anon_sym_continue] = ACTIONS(1138),
    [anon_sym_goto] = ACTIONS(1138),
    [anon_sym___try] = ACTIONS(1138),
    [anon_sym___leave] = ACTIONS(1138),
    [anon_sym_DASH_DASH] = ACTIONS(1140),
    [anon_sym_PLUS_PLUS] = ACTIONS(1140),
    [anon_sym_sizeof] = ACTIONS(1138),
    [anon_sym___alignof__] = ACTIONS(1138),
    [anon_sym___alignof] = ACTIONS(1138),
    [anon_sym__alignof] = ACTIONS(1138),
    [anon_sym_alignof] = ACTIONS(1138),
    [anon_sym__Alignof] = ACTIONS(1138),
    [anon_sym_offsetof] = ACTIONS(1138),
    [anon_sym__Generic] = ACTIONS(1138),
    [anon_sym_asm] = ACTIONS(1138),
    [anon_sym___asm__] = ACTIONS(1138),
    [anon_sym___asm] = ACTIONS(1138),
    [sym_number_literal] = ACTIONS(1140),
    [anon_sym_L_SQUOTE] = ACTIONS(1140),
    [anon_sym_u_SQUOTE] = ACTIONS(1140),
    [anon_sym_U_SQUOTE] = ACTIONS(1140),
    [anon_sym_u8_SQUOTE] = ACTIONS(1140),
    [anon_sym_SQUOTE] = ACTIONS(1140),
    [anon_sym_L_DQUOTE] = ACTIONS(1140),
    [anon_sym_u_DQUOTE] = ACTIONS(1140),
    [anon_sym_U_DQUOTE] = ACTIONS(1140),
    [anon_sym_u8_DQUOTE] = ACTIONS(1140),
    [anon_sym_DQUOTE] = ACTIONS(1140),
    [sym_true] = ACTIONS(1138),
    [sym_false] = ACTIONS(1138),
    [anon_sym_NULL] = ACTIONS(1138),
    [anon_sym_nullptr] = ACTIONS(1138),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(83)] = {
    [sym_identifier] = ACTIONS(1142),
    [aux_sym_preproc_include_token1] = ACTIONS(1142),
    [aux_sym_preproc_def_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token2] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1142),
    [aux_sym_preproc_else_token1] = ACTIONS(1142),
    [aux_sym_preproc_elif_token1] = ACTIONS(1142),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1142),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1142),
    [sym_preproc_directive] = ACTIONS(1142),
    [anon_sym_LPAREN2] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_TILDE] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1142),
    [anon_sym_PLUS] = ACTIONS(1142),
    [anon_sym_STAR] = ACTIONS(1144),
    [anon_sym_AMP] = ACTIONS(1144),
    [anon_sym_SEMI] = ACTIONS(1144),
    [anon_sym___extension__] = ACTIONS(1142),
    [anon_sym_typedef] = ACTIONS(1142),
    [anon_sym_extern] = ACTIONS(1142),
    [anon_sym___attribute__] = ACTIONS(1142),
    [anon_sym___attribute] = ACTIONS(1142),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1144),
    [anon_sym___declspec] = ACTIONS(1142),
    [anon_sym___cdecl] = ACTIONS(1142),
    [anon_sym___clrcall] = ACTIONS(1142),
    [anon_sym___stdcall] = ACTIONS(1142),
    [anon_sym___fastcall] = ACTIONS(1142),
    [anon_sym___thiscall] = ACTIONS(1142),
    [anon_sym___vectorcall] = ACTIONS(1142),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_signed] = ACTIONS(1142),
    [anon_sym_unsigned] = ACTIONS(1142),
    [anon_sym_long] = ACTIONS(1142),
    [anon_sym_short] = ACTIONS(1142),
    [anon_sym_static] = ACTIONS(1142),
    [anon_sym_auto] = ACTIONS(1142),
    [anon_sym_register] = ACTIONS(1142),
    [anon_sym_inline] = ACTIONS(1142),
    [anon_sym___inline] = ACTIONS(1142),
    [anon_sym___inline__] = ACTIONS(1142),
    [anon_sym___forceinline] = ACTIONS(1142),
    [anon_sym_thread_local] = ACTIONS(1142),
    [anon_sym___thread] = ACTIONS(1142),
    [anon_sym_const] = ACTIONS(1142),
    [anon_sym_constexpr] = ACTIONS(1142),
    [anon_sym_volatile] = ACTIONS(1142),
    [anon_sym_restrict] = ACTIONS(1142),
    [anon_sym___restrict__] = ACTIONS(1142),
    [anon_sym__Atomic] = ACTIONS(1142),
    [anon_sym__Noreturn] = ACTIONS(1142),
    [anon_sym_noreturn] = ACTIONS(1142),
    [anon_sym__Nonnull] = ACTIONS(1142),
    [anon_sym_alignas] = ACTIONS(1142),
    [anon_sym__Alignas] = ACTIONS(1142),
    [sym_primitive_type] = ACTIONS(1142),
    [anon_sym_enum] = ACTIONS(1142),
    [anon_sym_struct] = ACTIONS(1142),
    [anon_sym_union] = ACTIONS(1142),
    [anon_sym_if] = ACTIONS(1142),
    [anon_sym_else] = ACTIONS(1142),
    [anon_sym_switch] = ACTIONS(1142),
    [anon_sym_case] = ACTIONS(1142),
    [anon_sym_default] = ACTIONS(1142),
    [anon_sym_while] = ACTIONS(1142),
    [anon_sym_do] = ACTIONS(1142),
    [anon_sym_for] = ACTIONS(1142),
    [anon_sym_return] = ACTIONS(1142),
    [anon_sym_break] = ACTIONS(1142),
    [anon_sym_continue] = ACTIONS(1142),
    [anon_sym_goto] = ACTIONS(1142),
    [anon_sym___try] = ACTIONS(1142),
    [anon_sym___leave] = ACTIONS(1142),
    [anon_sym_DASH_DASH] = ACTIONS(1144),
    [anon_sym_PLUS_PLUS] = ACTIONS(1144),
    [anon_sym_sizeof] = ACTIONS(1142),
    [anon_sym___alignof__] = ACTIONS(1142),
    [anon_sym___alignof] = ACTIONS(1142),
    [anon_sym__alignof] = ACTIONS(1142),
    [anon_sym_alignof] = ACTIONS(1142),
    [anon_sym__Alignof] = ACTIONS(1142),
    [anon_sym_offsetof] = ACTIONS(1142),
    [anon_sym__Generic] = ACTIONS(1142),
    [anon_sym_asm] = ACTIONS(1142),
    [anon_sym___asm__] = ACTIONS(1142),
    [anon_sym___asm] = ACTIONS(1142),
    [sym_number_literal] = ACTIONS(1144),
    [anon_sym_L_SQUOTE] = ACTIONS(1144),
    [anon_sym_u_SQUOTE] = ACTIONS(1144),
    [anon_sym_U_SQUOTE] = ACTIONS(1144),
    [anon_sym_u8_SQUOTE] = ACTIONS(1144),
    [anon_sym_SQUOTE] = ACTIONS(1144),
    [anon_sym_L_DQUOTE] = ACTIONS(1144),
    [anon_sym_u_DQUOTE] = ACTIONS(1144),
    [anon_sym_U_DQUOTE] = ACTIONS(1144),
    [anon_sym_u8_DQUOTE] = ACTIONS(1144),
    [anon_sym_DQUOTE] = ACTIONS(1144),
    [sym_true] = ACTIONS(1142),
    [sym_false] = ACTIONS(1142),
    [anon_sym_NULL] = ACTIONS(1142),
    [anon_sym_nullptr] = ACTIONS(1142),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(84)] = {
    [sym_identifier] = ACTIONS(1146),
    [aux_sym_preproc_include_token1] = ACTIONS(1146),
    [aux_sym_preproc_def_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token2] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1146),
    [aux_sym_preproc_else_token1] = ACTIONS(1146),
    [aux_sym_preproc_elif_token1] = ACTIONS(1146),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1146),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1146),
    [sym_preproc_directive] = ACTIONS(1146),
    [anon_sym_LPAREN2] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_TILDE] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1146),
    [anon_sym_PLUS] = ACTIONS(1146),
    [anon_sym_STAR] = ACTIONS(1148),
    [anon_sym_AMP] = ACTIONS(1148),
    [anon_sym_SEMI] = ACTIONS(1148),
    [anon_sym___extension__] = ACTIONS(1146),
    [anon_sym_typedef] = ACTIONS(1146),
    [anon_sym_extern] = ACTIONS(1146),
    [anon_sym___attribute__] = ACTIONS(1146),
    [anon_sym___attribute] = ACTIONS(1146),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1148),
    [anon_sym___declspec] = ACTIONS(1146),
    [anon_sym___cdecl] = ACTIONS(1146),
    [anon_sym___clrcall] = ACTIONS(1146),
    [anon_sym___stdcall] = ACTIONS(1146),
    [anon_sym___fastcall] = ACTIONS(1146),
    [anon_sym___thiscall] = ACTIONS(1146),
    [anon_sym___vectorcall] = ACTIONS(1146),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_signed] = ACTIONS(1146),
    [anon_sym_unsigned] = ACTIONS(1146),
    [anon_sym_long] = ACTIONS(1146),
    [anon_sym_short] = ACTIONS(1146),
    [anon_sym_static] = ACTIONS(1146),
    [anon_sym_auto] = ACTIONS(1146),
    [anon_sym_register] = ACTIONS(1146),
    [anon_sym_inline] = ACTIONS(1146),
    [anon_sym___inline] = ACTIONS(1146),
    [anon_sym___inline__] = ACTIONS(1146),
    [anon_sym___forceinline] = ACTIONS(1146),
    [anon_sym_thread_local] = ACTIONS(1146),
    [anon_sym___thread] = ACTIONS(1146),
    [anon_sym_const] = ACTIONS(1146),
    [anon_sym_constexpr] = ACTIONS(1146),
    [anon_sym_volatile] = ACTIONS(1146),
    [anon_sym_restrict] = ACTIONS(1146),
    [anon_sym___restrict__] = ACTIONS(1146),
    [anon_sym__Atomic] = ACTIONS(1146),
    [anon_sym__Noreturn] = ACTIONS(1146),
    [anon_sym_noreturn] = ACTIONS(1146),
    [anon_sym__Nonnull] = ACTIONS(1146),
    [anon_sym_alignas] = ACTIONS(1146),
    [anon_sym__Alignas] = ACTIONS(1146),
    [sym_primitive_type] = ACTIONS(1146),
    [anon_sym_enum] = ACTIONS(1146),
    [anon_sym_struct] = ACTIONS(1146),
    [anon_sym_union] = ACTIONS(1146),
    [anon_sym_if] = ACTIONS(1146),
    [anon_sym_else] = ACTIONS(1146),
    [anon_sym_switch] = ACTIONS(1146),
    [anon_sym_case] = ACTIONS(1146),
    [anon_sym_default] = ACTIONS(1146),
    [anon_sym_while] = ACTIONS(1146),
    [anon_sym_do] = ACTIONS(1146),
    [anon_sym_for] = ACTIONS(1146),
    [anon_sym_return] = ACTIONS(1146),
    [anon_sym_break] = ACTIONS(1146),
    [anon_sym_continue] = ACTIONS(1146),
    [anon_sym_goto] = ACTIONS(1146),
    [anon_sym___try] = ACTIONS(1146),
    [anon_sym___leave] = ACTIONS(1146),
    [anon_sym_DASH_DASH] = ACTIONS(1148),
    [anon_sym_PLUS_PLUS] = ACTIONS(1148),
    [anon_sym_sizeof] = ACTIONS(1146),
    [anon_sym___alignof__] = ACTIONS(1146),
    [anon_sym___alignof] = ACTIONS(1146),
    [anon_sym__alignof] = ACTIONS(1146),
    [anon_sym_alignof] = ACTIONS(1146),
    [anon_sym__Alignof] = ACTIONS(1146),
    [anon_sym_offsetof] = ACTIONS(1146),
    [anon_sym__Generic] = ACTIONS(1146),
    [anon_sym_asm] = ACTIONS(1146),
    [anon_sym___asm__] = ACTIONS(1146),
    [anon_sym___asm] = ACTIONS(1146),
    [sym_number_literal] = ACTIONS(1148),
    [anon_sym_L_SQUOTE] = ACTIONS(1148),
    [anon_sym_u_SQUOTE] = ACTIONS(1148),
    [anon_sym_U_SQUOTE] = ACTIONS(1148),
    [anon_sym_u8_SQUOTE] = ACTIONS(1148),
    [anon_sym_SQUOTE] = ACTIONS(1148),
    [anon_sym_L_DQUOTE] = ACTIONS(1148),
    [anon_sym_u_DQUOTE] = ACTIONS(1148),
    [anon_sym_U_DQUOTE] = ACTIONS(1148),
    [anon_sym_u8_DQUOTE] = ACTIONS(1148),
    [anon_sym_DQUOTE] = ACTIONS(1148),
    [sym_true] = ACTIONS(1146),
    [sym_false] = ACTIONS(1146),
    [anon_sym_NULL] = ACTIONS(1146),
    [anon_sym_nullptr] = ACTIONS(1146),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(85)] = {
    [sym_identifier] = ACTIONS(1150),
    [aux_sym_preproc_include_token1] = ACTIONS(1150),
    [aux_sym_preproc_def_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token2] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1150),
    [aux_sym_preproc_else_token1] = ACTIONS(1150),
    [aux_sym_preproc_elif_token1] = ACTIONS(1150),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1150),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1150),
    [sym_preproc_directive] = ACTIONS(1150),
    [anon_sym_LPAREN2] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_TILDE] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1150),
    [anon_sym_PLUS] = ACTIONS(1150),
    [anon_sym_STAR] = ACTIONS(1152),
    [anon_sym_AMP] = ACTIONS(1152),
    [anon_sym_SEMI] = ACTIONS(1152),
    [anon_sym___extension__] = ACTIONS(1150),
    [anon_sym_typedef] = ACTIONS(1150),
    [anon_sym_extern] = ACTIONS(1150),
    [anon_sym___attribute__] = ACTIONS(1150),
    [anon_sym___attribute] = ACTIONS(1150),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1152),
    [anon_sym___declspec] = ACTIONS(1150),
    [anon_sym___cdecl] = ACTIONS(1150),
    [anon_sym___clrcall] = ACTIONS(1150),
    [anon_sym___stdcall] = ACTIONS(1150),
    [anon_sym___fastcall] = ACTIONS(1150),
    [anon_sym___thiscall] = ACTIONS(1150),
    [anon_sym___vectorcall] = ACTIONS(1150),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_signed] = ACTIONS(1150),
    [anon_sym_unsigned] = ACTIONS(1150),
    [anon_sym_long] = ACTIONS(1150),
    [anon_sym_short] = ACTIONS(1150),
    [anon_sym_static] = ACTIONS(1150),
    [anon_sym_auto] = ACTIONS(1150),
    [anon_sym_register] = ACTIONS(1150),
    [anon_sym_inline] = ACTIONS(1150),
    [anon_sym___inline] = ACTIONS(1150),
    [anon_sym___inline__] = ACTIONS(1150),
    [anon_sym___forceinline] = ACTIONS(1150),
    [anon_sym_thread_local] = ACTIONS(1150),
    [anon_sym___thread] = ACTIONS(1150),
    [anon_sym_const] = ACTIONS(1150),
    [anon_sym_constexpr] = ACTIONS(1150),
    [anon_sym_volatile] = ACTIONS(1150),
    [anon_sym_restrict] = ACTIONS(1150),
    [anon_sym___restrict__] = ACTIONS(1150),
    [anon_sym__Atomic] = ACTIONS(1150),
    [anon_sym__Noreturn] = ACTIONS(1150),
    [anon_sym_noreturn] = ACTIONS(1150),
    [anon_sym__Nonnull] = ACTIONS(1150),
    [anon_sym_alignas] = ACTIONS(1150),
    [anon_sym__Alignas] = ACTIONS(1150),
    [sym_primitive_type] = ACTIONS(1150),
    [anon_sym_enum] = ACTIONS(1150),
    [anon_sym_struct] = ACTIONS(1150),
    [anon_sym_union] = ACTIONS(1150),
    [anon_sym_if] = ACTIONS(1150),
    [anon_sym_else] = ACTIONS(1150),
    [anon_sym_switch] = ACTIONS(1150),
    [anon_sym_case] = ACTIONS(1150),
    [anon_sym_default] = ACTIONS(1150),
    [anon_sym_while] = ACTIONS(1150),
    [anon_sym_do] = ACTIONS(1150),
    [anon_sym_for] = ACTIONS(1150),
    [anon_sym_return] = ACTIONS(1150),
    [anon_sym_break] = ACTIONS(1150),
    [anon_sym_continue] = ACTIONS(1150),
    [anon_sym_goto] = ACTIONS(1150),
    [anon_sym___try] = ACTIONS(1150),
    [anon_sym___leave] = ACTIONS(1150),
    [anon_sym_DASH_DASH] = ACTIONS(1152),
    [anon_sym_PLUS_PLUS] = ACTIONS(1152),
    [anon_sym_sizeof] = ACTIONS(1150),
    [anon_sym___alignof__] = ACTIONS(1150),
    [anon_sym___alignof] = ACTIONS(1150),
    [anon_sym__alignof] = ACTIONS(1150),
    [anon_sym_alignof] = ACTIONS(1150),
    [anon_sym__Alignof] = ACTIONS(1150),
    [anon_sym_offsetof] = ACTIONS(1150),
    [anon_sym__Generic] = ACTIONS(1150),
    [anon_sym_asm] = ACTIONS(1150),
    [anon_sym___asm__] = ACTIONS(1150),
    [anon_sym___asm] = ACTIONS(1150),
    [sym_number_literal] = ACTIONS(1152),
    [anon_sym_L_SQUOTE] = ACTIONS(1152),
    [anon_sym_u_SQUOTE] = ACTIONS(1152),
    [anon_sym_U_SQUOTE] = ACTIONS(1152),
    [anon_sym_u8_SQUOTE] = ACTIONS(1152),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_L_DQUOTE] = ACTIONS(1152),
    [anon_sym_u_DQUOTE] = ACTIONS(1152),
    [anon_sym_U_DQUOTE] = ACTIONS(1152),
    [anon_sym_u8_DQUOTE] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1152),
    [sym_true] = ACTIONS(1150),
    [sym_false] = ACTIONS(1150),
    [anon_sym_NULL] = ACTIONS(1150),
    [anon_sym_nullptr] = ACTIONS(1150),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(86)] = {
    [sym_identifier] = ACTIONS(1154),
    [aux_sym_preproc_include_token1] = ACTIONS(1154),
    [aux_sym_preproc_def_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token2] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1154),
    [aux_sym_preproc_else_token1] = ACTIONS(1154),
    [aux_sym_preproc_elif_token1] = ACTIONS(1154),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1154),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1154),
    [sym_preproc_directive] = ACTIONS(1154),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1156),
    [anon_sym_AMP] = ACTIONS(1156),
    [anon_sym_SEMI] = ACTIONS(1156),
    [anon_sym___extension__] = ACTIONS(1154),
    [anon_sym_typedef] = ACTIONS(1154),
    [anon_sym_extern] = ACTIONS(1154),
    [anon_sym___attribute__] = ACTIONS(1154),
    [anon_sym___attribute] = ACTIONS(1154),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1156),
    [anon_sym___declspec] = ACTIONS(1154),
    [anon_sym___cdecl] = ACTIONS(1154),
    [anon_sym___clrcall] = ACTIONS(1154),
    [anon_sym___stdcall] = ACTIONS(1154),
    [anon_sym___fastcall] = ACTIONS(1154),
    [anon_sym___thiscall] = ACTIONS(1154),
    [anon_sym___vectorcall] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_signed] = ACTIONS(1154),
    [anon_sym_unsigned] = ACTIONS(1154),
    [anon_sym_long] = ACTIONS(1154),
    [anon_sym_short] = ACTIONS(1154),
    [anon_sym_static] = ACTIONS(1154),
    [anon_sym_auto] = ACTIONS(1154),
    [anon_sym_register] = ACTIONS(1154),
    [anon_sym_inline] = ACTIONS(1154),
    [anon_sym___inline] = ACTIONS(1154),
    [anon_sym___inline__] = ACTIONS(1154),
    [anon_sym___forceinline] = ACTIONS(1154),
    [anon_sym_thread_local] = ACTIONS(1154),
    [anon_sym___thread] = ACTIONS(1154),
    [anon_sym_const] = ACTIONS(1154),
    [anon_sym_constexpr] = ACTIONS(1154),
    [anon_sym_volatile] = ACTIONS(1154),
    [anon_sym_restrict] = ACTIONS(1154),
    [anon_sym___restrict__] = ACTIONS(1154),
    [anon_sym__Atomic] = ACTIONS(1154),
    [anon_sym__Noreturn] = ACTIONS(1154),
    [anon_sym_noreturn] = ACTIONS(1154),
    [anon_sym__Nonnull] = ACTIONS(1154),
    [anon_sym_alignas] = ACTIONS(1154),
    [anon_sym__Alignas] = ACTIONS(1154),
    [sym_primitive_type] = ACTIONS(1154),
    [anon_sym_enum] = ACTIONS(1154),
    [anon_sym_struct] = ACTIONS(1154),
    [anon_sym_union] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_else] = ACTIONS(1154),
    [anon_sym_switch] = ACTIONS(1154),
    [anon_sym_case] = ACTIONS(1154),
    [anon_sym_default] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_break] = ACTIONS(1154),
    [anon_sym_continue] = ACTIONS(1154),
    [anon_sym_goto] = ACTIONS(1154),
    [anon_sym___try] = ACTIONS(1154),
    [anon_sym___leave] = ACTIONS(1154),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [anon_sym_sizeof] = ACTIONS(1154),
    [anon_sym___alignof__] = ACTIONS(1154),
    [anon_sym___alignof] = ACTIONS(1154),
    [anon_sym__alignof] = ACTIONS(1154),
    [anon_sym_alignof] = ACTIONS(1154),
    [anon_sym__Alignof] = ACTIONS(1154),
    [anon_sym_offsetof] = ACTIONS(1154),
    [anon_sym__Generic] = ACTIONS(1154),
    [anon_sym_asm] = ACTIONS(1154),
    [anon_sym___asm__] = ACTIONS(1154),
    [anon_sym___asm] = ACTIONS(1154),
    [sym_number_literal] = ACTIONS(1156),
    [anon_sym_L_SQUOTE] = ACTIONS(1156),
    [anon_sym_u_SQUOTE] = ACTIONS(1156),
    [anon_sym_U_SQUOTE] = ACTIONS(1156),
    [anon_sym_u8_SQUOTE] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_L_DQUOTE] = ACTIONS(1156),
    [anon_sym_u_DQUOTE] = ACTIONS(1156),
    [anon_sym_U_DQUOTE] = ACTIONS(1156),
    [anon_sym_u8_DQUOTE] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [anon_sym_NULL] = ACTIONS(1154),
    [anon_sym_nullptr] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(87)] = {
    [ts_builtin_sym_end] = ACTIONS(1140),
    [sym_identifier] = ACTIONS(1138),
    [aux_sym_preproc_include_token1] = ACTIONS(1138),
    [aux_sym_preproc_def_token1] = ACTIONS(1138),
    [anon_sym_COMMA] = ACTIONS(1140),
    [anon_sym_RPAREN] = ACTIONS(1140),
    [aux_sym_preproc_if_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1138),
    [sym_preproc_directive] = ACTIONS(1138),
    [anon_sym_LPAREN2] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_TILDE] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1138),
    [anon_sym_PLUS] = ACTIONS(1138),
    [anon_sym_STAR] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(1140),
    [anon_sym_SEMI] = ACTIONS(1140),
    [anon_sym___extension__] = ACTIONS(1138),
    [anon_sym_typedef] = ACTIONS(1138),
    [anon_sym_extern] = ACTIONS(1138),
    [anon_sym___attribute__] = ACTIONS(1138),
    [anon_sym___attribute] = ACTIONS(1138),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1140),
    [anon_sym___declspec] = ACTIONS(1138),
    [anon_sym___cdecl] = ACTIONS(1138),
    [anon_sym___clrcall] = ACTIONS(1138),
    [anon_sym___stdcall] = ACTIONS(1138),
    [anon_sym___fastcall] = ACTIONS(1138),
    [anon_sym___thiscall] = ACTIONS(1138),
    [anon_sym___vectorcall] = ACTIONS(1138),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_signed] = ACTIONS(1138),
    [anon_sym_unsigned] = ACTIONS(1138),
    [anon_sym_long] = ACTIONS(1138),
    [anon_sym_short] = ACTIONS(1138),
    [anon_sym_static] = ACTIONS(1138),
    [anon_sym_auto] = ACTIONS(1138),
    [anon_sym_register] = ACTIONS(1138),
    [anon_sym_inline] = ACTIONS(1138),
    [anon_sym___inline] = ACTIONS(1138),
    [anon_sym___inline__] = ACTIONS(1138),
    [anon_sym___forceinline] = ACTIONS(1138),
    [anon_sym_thread_local] = ACTIONS(1138),
    [anon_sym___thread] = ACTIONS(1138),
    [anon_sym_const] = ACTIONS(1138),
    [anon_sym_constexpr] = ACTIONS(1138),
    [anon_sym_volatile] = ACTIONS(1138),
    [anon_sym_restrict] = ACTIONS(1138),
    [anon_sym___restrict__] = ACTIONS(1138),
    [anon_sym__Atomic] = ACTIONS(1138),
    [anon_sym__Noreturn] = ACTIONS(1138),
    [anon_sym_noreturn] = ACTIONS(1138),
    [anon_sym__Nonnull] = ACTIONS(1138),
    [anon_sym_alignas] = ACTIONS(1138),
    [anon_sym__Alignas] = ACTIONS(1138),
    [sym_primitive_type] = ACTIONS(1138),
    [anon_sym_enum] = ACTIONS(1138),
    [anon_sym_struct] = ACTIONS(1138),
    [anon_sym_union] = ACTIONS(1138),
    [anon_sym_if] = ACTIONS(1138),
    [anon_sym_else] = ACTIONS(1138),
    [anon_sym_switch] = ACTIONS(1138),
    [anon_sym_case] = ACTIONS(1138),
    [anon_sym_default] = ACTIONS(1138),
    [anon_sym_while] = ACTIONS(1138),
    [anon_sym_do] = ACTIONS(1138),
    [anon_sym_for] = ACTIONS(1138),
    [anon_sym_return] = ACTIONS(1138),
    [anon_sym_break] = ACTIONS(1138),
    [anon_sym_continue] = ACTIONS(1138),
    [anon_sym_goto] = ACTIONS(1138),
    [anon_sym___try] = ACTIONS(1138),
    [anon_sym___except] = ACTIONS(1138),
    [anon_sym___finally] = ACTIONS(1138),
    [anon_sym___leave] = ACTIONS(1138),
    [anon_sym_DASH_DASH] = ACTIONS(1140),
    [anon_sym_PLUS_PLUS] = ACTIONS(1140),
    [anon_sym_sizeof] = ACTIONS(1138),
    [anon_sym___alignof__] = ACTIONS(1138),
    [anon_sym___alignof] = ACTIONS(1138),
    [anon_sym__alignof] = ACTIONS(1138),
    [anon_sym_alignof] = ACTIONS(1138),
    [anon_sym__Alignof] = ACTIONS(1138),
    [anon_sym_offsetof] = ACTIONS(1138),
    [anon_sym__Generic] = ACTIONS(1138),
    [anon_sym_asm] = ACTIONS(1138),
    [anon_sym___asm__] = ACTIONS(1138),
    [anon_sym___asm] = ACTIONS(1138),
    [sym_number_literal] = ACTIONS(1140),
    [anon_sym_L_SQUOTE] = ACTIONS(1140),
    [anon_sym_u_SQUOTE] = ACTIONS(1140),
    [anon_sym_U_SQUOTE] = ACTIONS(1140),
    [anon_sym_u8_SQUOTE] = ACTIONS(1140),
    [anon_sym_SQUOTE] = ACTIONS(1140),
    [anon_sym_L_DQUOTE] = ACTIONS(1140),
    [anon_sym_u_DQUOTE] = ACTIONS(1140),
    [anon_sym_U_DQUOTE] = ACTIONS(1140),
    [anon_sym_u8_DQUOTE] = ACTIONS(1140),
    [anon_sym_DQUOTE] = ACTIONS(1140),
    [sym_true] = ACTIONS(1138),
    [sym_false] = ACTIONS(1138),
    [anon_sym_NULL] = ACTIONS(1138),
    [anon_sym_nullptr] = ACTIONS(1138),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(88)] = {
    [sym_identifier] = ACTIONS(1158),
    [aux_sym_preproc_include_token1] = ACTIONS(1158),
    [aux_sym_preproc_def_token1] = ACTIONS(1158),
    [aux_sym_preproc_if_token1] = ACTIONS(1158),
    [aux_sym_preproc_if_token2] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1158),
    [aux_sym_preproc_else_token1] = ACTIONS(1158),
    [aux_sym_preproc_elif_token1] = ACTIONS(1158),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1158),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1158),
    [sym_preproc_directive] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_TILDE] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_STAR] = ACTIONS(1160),
    [anon_sym_AMP] = ACTIONS(1160),
    [anon_sym_SEMI] = ACTIONS(1160),
    [anon_sym___extension__] = ACTIONS(1158),
    [anon_sym_typedef] = ACTIONS(1158),
    [anon_sym_extern] = ACTIONS(1158),
    [anon_sym___attribute__] = ACTIONS(1158),
    [anon_sym___attribute] = ACTIONS(1158),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1160),
    [anon_sym___declspec] = ACTIONS(1158),
    [anon_sym___cdecl] = ACTIONS(1158),
    [anon_sym___clrcall] = ACTIONS(1158),
    [anon_sym___stdcall] = ACTIONS(1158),
    [anon_sym___fastcall] = ACTIONS(1158),
    [anon_sym___thiscall] = ACTIONS(1158),
    [anon_sym___vectorcall] = ACTIONS(1158),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_signed] = ACTIONS(1158),
    [anon_sym_unsigned] = ACTIONS(1158),
    [anon_sym_long] = ACTIONS(1158),
    [anon_sym_short] = ACTIONS(1158),
    [anon_sym_static] = ACTIONS(1158),
    [anon_sym_auto] = ACTIONS(1158),
    [anon_sym_register] = ACTIONS(1158),
    [anon_sym_inline] = ACTIONS(1158),
    [anon_sym___inline] = ACTIONS(1158),
    [anon_sym___inline__] = ACTIONS(1158),
    [anon_sym___forceinline] = ACTIONS(1158),
    [anon_sym_thread_local] = ACTIONS(1158),
    [anon_sym___thread] = ACTIONS(1158),
    [anon_sym_const] = ACTIONS(1158),
    [anon_sym_constexpr] = ACTIONS(1158),
    [anon_sym_volatile] = ACTIONS(1158),
    [anon_sym_restrict] = ACTIONS(1158),
    [anon_sym___restrict__] = ACTIONS(1158),
    [anon_sym__Atomic] = ACTIONS(1158),
    [anon_sym__Noreturn] = ACTIONS(1158),
    [anon_sym_noreturn] = ACTIONS(1158),
    [anon_sym__Nonnull] = ACTIONS(1158),
    [anon_sym_alignas] = ACTIONS(1158),
    [anon_sym__Alignas] = ACTIONS(1158),
    [sym_primitive_type] = ACTIONS(1158),
    [anon_sym_enum] = ACTIONS(1158),
    [anon_sym_struct] = ACTIONS(1158),
    [anon_sym_union] = ACTIONS(1158),
    [anon_sym_if] = ACTIONS(1158),
    [anon_sym_else] = ACTIONS(1158),
    [anon_sym_switch] = ACTIONS(1158),
    [anon_sym_case] = ACTIONS(1158),
    [anon_sym_default] = ACTIONS(1158),
    [anon_sym_while] = ACTIONS(1158),
    [anon_sym_do] = ACTIONS(1158),
    [anon_sym_for] = ACTIONS(1158),
    [anon_sym_return] = ACTIONS(1158),
    [anon_sym_break] = ACTIONS(1158),
    [anon_sym_continue] = ACTIONS(1158),
    [anon_sym_goto] = ACTIONS(1158),
    [anon_sym___try] = ACTIONS(1158),
    [anon_sym___leave] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1160),
    [anon_sym_PLUS_PLUS] = ACTIONS(1160),
    [anon_sym_sizeof] = ACTIONS(1158),
    [anon_sym___alignof__] = ACTIONS(1158),
    [anon_sym___alignof] = ACTIONS(1158),
    [anon_sym__alignof] = ACTIONS(1158),
    [anon_sym_alignof] = ACTIONS(1158),
    [anon_sym__Alignof] = ACTIONS(1158),
    [anon_sym_offsetof] = ACTIONS(1158),
    [anon_sym__Generic] = ACTIONS(1158),
    [anon_sym_asm] = ACTIONS(1158),
    [anon_sym___asm__] = ACTIONS(1158),
    [anon_sym___asm] = ACTIONS(1158),
    [sym_number_literal] = ACTIONS(1160),
    [anon_sym_L_SQUOTE] = ACTIONS(1160),
    [anon_sym_u_SQUOTE] = ACTIONS(1160),
    [anon_sym_U_SQUOTE] = ACTIONS(1160),
    [anon_sym_u8_SQUOTE] = ACTIONS(1160),
    [anon_sym_SQUOTE] = ACTIONS(1160),
    [anon_sym_L_DQUOTE] = ACTIONS(1160),
    [anon_sym_u_DQUOTE] = ACTIONS(1160),
    [anon_sym_U_DQUOTE] = ACTIONS(1160),
    [anon_sym_u8_DQUOTE] = ACTIONS(1160),
    [anon_sym_DQUOTE] = ACTIONS(1160),
    [sym_true] = ACTIONS(1158),
    [sym_false] = ACTIONS(1158),
    [anon_sym_NULL] = ACTIONS(1158),
    [anon_sym_nullptr] = ACTIONS(1158),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(89)] = {
    [ts_builtin_sym_end] = ACTIONS(1160),
    [sym_identifier] = ACTIONS(1158),
    [aux_sym_preproc_include_token1] = ACTIONS(1158),
    [aux_sym_preproc_def_token1] = ACTIONS(1158),
    [anon_sym_COMMA] = ACTIONS(1160),
    [anon_sym_RPAREN] = ACTIONS(1160),
    [aux_sym_preproc_if_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1158),
    [sym_preproc_directive] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_TILDE] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_STAR] = ACTIONS(1160),
    [anon_sym_AMP] = ACTIONS(1160),
    [anon_sym_SEMI] = ACTIONS(1160),
    [anon_sym___extension__] = ACTIONS(1158),
    [anon_sym_typedef] = ACTIONS(1158),
    [anon_sym_extern] = ACTIONS(1158),
    [anon_sym___attribute__] = ACTIONS(1158),
    [anon_sym___attribute] = ACTIONS(1158),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1160),
    [anon_sym___declspec] = ACTIONS(1158),
    [anon_sym___cdecl] = ACTIONS(1158),
    [anon_sym___clrcall] = ACTIONS(1158),
    [anon_sym___stdcall] = ACTIONS(1158),
    [anon_sym___fastcall] = ACTIONS(1158),
    [anon_sym___thiscall] = ACTIONS(1158),
    [anon_sym___vectorcall] = ACTIONS(1158),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_signed] = ACTIONS(1158),
    [anon_sym_unsigned] = ACTIONS(1158),
    [anon_sym_long] = ACTIONS(1158),
    [anon_sym_short] = ACTIONS(1158),
    [anon_sym_static] = ACTIONS(1158),
    [anon_sym_auto] = ACTIONS(1158),
    [anon_sym_register] = ACTIONS(1158),
    [anon_sym_inline] = ACTIONS(1158),
    [anon_sym___inline] = ACTIONS(1158),
    [anon_sym___inline__] = ACTIONS(1158),
    [anon_sym___forceinline] = ACTIONS(1158),
    [anon_sym_thread_local] = ACTIONS(1158),
    [anon_sym___thread] = ACTIONS(1158),
    [anon_sym_const] = ACTIONS(1158),
    [anon_sym_constexpr] = ACTIONS(1158),
    [anon_sym_volatile] = ACTIONS(1158),
    [anon_sym_restrict] = ACTIONS(1158),
    [anon_sym___restrict__] = ACTIONS(1158),
    [anon_sym__Atomic] = ACTIONS(1158),
    [anon_sym__Noreturn] = ACTIONS(1158),
    [anon_sym_noreturn] = ACTIONS(1158),
    [anon_sym__Nonnull] = ACTIONS(1158),
    [anon_sym_alignas] = ACTIONS(1158),
    [anon_sym__Alignas] = ACTIONS(1158),
    [sym_primitive_type] = ACTIONS(1158),
    [anon_sym_enum] = ACTIONS(1158),
    [anon_sym_struct] = ACTIONS(1158),
    [anon_sym_union] = ACTIONS(1158),
    [anon_sym_if] = ACTIONS(1158),
    [anon_sym_else] = ACTIONS(1158),
    [anon_sym_switch] = ACTIONS(1158),
    [anon_sym_case] = ACTIONS(1158),
    [anon_sym_default] = ACTIONS(1158),
    [anon_sym_while] = ACTIONS(1158),
    [anon_sym_do] = ACTIONS(1158),
    [anon_sym_for] = ACTIONS(1158),
    [anon_sym_return] = ACTIONS(1158),
    [anon_sym_break] = ACTIONS(1158),
    [anon_sym_continue] = ACTIONS(1158),
    [anon_sym_goto] = ACTIONS(1158),
    [anon_sym___try] = ACTIONS(1158),
    [anon_sym___except] = ACTIONS(1158),
    [anon_sym___finally] = ACTIONS(1158),
    [anon_sym___leave] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1160),
    [anon_sym_PLUS_PLUS] = ACTIONS(1160),
    [anon_sym_sizeof] = ACTIONS(1158),
    [anon_sym___alignof__] = ACTIONS(1158),
    [anon_sym___alignof] = ACTIONS(1158),
    [anon_sym__alignof] = ACTIONS(1158),
    [anon_sym_alignof] = ACTIONS(1158),
    [anon_sym__Alignof] = ACTIONS(1158),
    [anon_sym_offsetof] = ACTIONS(1158),
    [anon_sym__Generic] = ACTIONS(1158),
    [anon_sym_asm] = ACTIONS(1158),
    [anon_sym___asm__] = ACTIONS(1158),
    [anon_sym___asm] = ACTIONS(1158),
    [sym_number_literal] = ACTIONS(1160),
    [anon_sym_L_SQUOTE] = ACTIONS(1160),
    [anon_sym_u_SQUOTE] = ACTIONS(1160),
    [anon_sym_U_SQUOTE] = ACTIONS(1160),
    [anon_sym_u8_SQUOTE] = ACTIONS(1160),
    [anon_sym_SQUOTE] = ACTIONS(1160),
    [anon_sym_L_DQUOTE] = ACTIONS(1160),
    [anon_sym_u_DQUOTE] = ACTIONS(1160),
    [anon_sym_U_DQUOTE] = ACTIONS(1160),
    [anon_sym_u8_DQUOTE] = ACTIONS(1160),
    [anon_sym_DQUOTE] = ACTIONS(1160),
    [sym_true] = ACTIONS(1158),
    [sym_false] = ACTIONS(1158),
    [anon_sym_NULL] = ACTIONS(1158),
    [anon_sym_nullptr] = ACTIONS(1158),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(90)] = {
    [sym_identifier] = ACTIONS(1162),
    [aux_sym_preproc_include_token1] = ACTIONS(1162),
    [aux_sym_preproc_def_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token2] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1162),
    [aux_sym_preproc_else_token1] = ACTIONS(1162),
    [aux_sym_preproc_elif_token1] = ACTIONS(1162),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1162),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1162),
    [sym_preproc_directive] = ACTIONS(1162),
    [anon_sym_LPAREN2] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1162),
    [anon_sym_PLUS] = ACTIONS(1162),
    [anon_sym_STAR] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym___extension__] = ACTIONS(1162),
    [anon_sym_typedef] = ACTIONS(1162),
    [anon_sym_extern] = ACTIONS(1162),
    [anon_sym___attribute__] = ACTIONS(1162),
    [anon_sym___attribute] = ACTIONS(1162),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1164),
    [anon_sym___declspec] = ACTIONS(1162),
    [anon_sym___cdecl] = ACTIONS(1162),
    [anon_sym___clrcall] = ACTIONS(1162),
    [anon_sym___stdcall] = ACTIONS(1162),
    [anon_sym___fastcall] = ACTIONS(1162),
    [anon_sym___thiscall] = ACTIONS(1162),
    [anon_sym___vectorcall] = ACTIONS(1162),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_signed] = ACTIONS(1162),
    [anon_sym_unsigned] = ACTIONS(1162),
    [anon_sym_long] = ACTIONS(1162),
    [anon_sym_short] = ACTIONS(1162),
    [anon_sym_static] = ACTIONS(1162),
    [anon_sym_auto] = ACTIONS(1162),
    [anon_sym_register] = ACTIONS(1162),
    [anon_sym_inline] = ACTIONS(1162),
    [anon_sym___inline] = ACTIONS(1162),
    [anon_sym___inline__] = ACTIONS(1162),
    [anon_sym___forceinline] = ACTIONS(1162),
    [anon_sym_thread_local] = ACTIONS(1162),
    [anon_sym___thread] = ACTIONS(1162),
    [anon_sym_const] = ACTIONS(1162),
    [anon_sym_constexpr] = ACTIONS(1162),
    [anon_sym_volatile] = ACTIONS(1162),
    [anon_sym_restrict] = ACTIONS(1162),
    [anon_sym___restrict__] = ACTIONS(1162),
    [anon_sym__Atomic] = ACTIONS(1162),
    [anon_sym__Noreturn] = ACTIONS(1162),
    [anon_sym_noreturn] = ACTIONS(1162),
    [anon_sym__Nonnull] = ACTIONS(1162),
    [anon_sym_alignas] = ACTIONS(1162),
    [anon_sym__Alignas] = ACTIONS(1162),
    [sym_primitive_type] = ACTIONS(1162),
    [anon_sym_enum] = ACTIONS(1162),
    [anon_sym_struct] = ACTIONS(1162),
    [anon_sym_union] = ACTIONS(1162),
    [anon_sym_if] = ACTIONS(1162),
    [anon_sym_else] = ACTIONS(1162),
    [anon_sym_switch] = ACTIONS(1162),
    [anon_sym_case] = ACTIONS(1162),
    [anon_sym_default] = ACTIONS(1162),
    [anon_sym_while] = ACTIONS(1162),
    [anon_sym_do] = ACTIONS(1162),
    [anon_sym_for] = ACTIONS(1162),
    [anon_sym_return] = ACTIONS(1162),
    [anon_sym_break] = ACTIONS(1162),
    [anon_sym_continue] = ACTIONS(1162),
    [anon_sym_goto] = ACTIONS(1162),
    [anon_sym___try] = ACTIONS(1162),
    [anon_sym___leave] = ACTIONS(1162),
    [anon_sym_DASH_DASH] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1164),
    [anon_sym_sizeof] = ACTIONS(1162),
    [anon_sym___alignof__] = ACTIONS(1162),
    [anon_sym___alignof] = ACTIONS(1162),
    [anon_sym__alignof] = ACTIONS(1162),
    [anon_sym_alignof] = ACTIONS(1162),
    [anon_sym__Alignof] = ACTIONS(1162),
    [anon_sym_offsetof] = ACTIONS(1162),
    [anon_sym__Generic] = ACTIONS(1162),
    [anon_sym_asm] = ACTIONS(1162),
    [anon_sym___asm__] = ACTIONS(1162),
    [anon_sym___asm] = ACTIONS(1162),
    [sym_number_literal] = ACTIONS(1164),
    [anon_sym_L_SQUOTE] = ACTIONS(1164),
    [anon_sym_u_SQUOTE] = ACTIONS(1164),
    [anon_sym_U_SQUOTE] = ACTIONS(1164),
    [anon_sym_u8_SQUOTE] = ACTIONS(1164),
    [anon_sym_SQUOTE] = ACTIONS(1164),
    [anon_sym_L_DQUOTE] = ACTIONS(1164),
    [anon_sym_u_DQUOTE] = ACTIONS(1164),
    [anon_sym_U_DQUOTE] = ACTIONS(1164),
    [anon_sym_u8_DQUOTE] = ACTIONS(1164),
    [anon_sym_DQUOTE] = ACTIONS(1164),
    [sym_true] = ACTIONS(1162),
    [sym_false] = ACTIONS(1162),
    [anon_sym_NULL] = ACTIONS(1162),
    [anon_sym_nullptr] = ACTIONS(1162),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(91)] = {
    [sym_identifier] = ACTIONS(1166),
    [aux_sym_preproc_include_token1] = ACTIONS(1166),
    [aux_sym_preproc_def_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token2] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1166),
    [aux_sym_preproc_else_token1] = ACTIONS(1166),
    [aux_sym_preproc_elif_token1] = ACTIONS(1166),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1166),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1166),
    [sym_preproc_directive] = ACTIONS(1166),
    [anon_sym_LPAREN2] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_TILDE] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1166),
    [anon_sym_PLUS] = ACTIONS(1166),
    [anon_sym_STAR] = ACTIONS(1168),
    [anon_sym_AMP] = ACTIONS(1168),
    [anon_sym_SEMI] = ACTIONS(1168),
    [anon_sym___extension__] = ACTIONS(1166),
    [anon_sym_typedef] = ACTIONS(1166),
    [anon_sym_extern] = ACTIONS(1166),
    [anon_sym___attribute__] = ACTIONS(1166),
    [anon_sym___attribute] = ACTIONS(1166),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1168),
    [anon_sym___declspec] = ACTIONS(1166),
    [anon_sym___cdecl] = ACTIONS(1166),
    [anon_sym___clrcall] = ACTIONS(1166),
    [anon_sym___stdcall] = ACTIONS(1166),
    [anon_sym___fastcall] = ACTIONS(1166),
    [anon_sym___thiscall] = ACTIONS(1166),
    [anon_sym___vectorcall] = ACTIONS(1166),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_signed] = ACTIONS(1166),
    [anon_sym_unsigned] = ACTIONS(1166),
    [anon_sym_long] = ACTIONS(1166),
    [anon_sym_short] = ACTIONS(1166),
    [anon_sym_static] = ACTIONS(1166),
    [anon_sym_auto] = ACTIONS(1166),
    [anon_sym_register] = ACTIONS(1166),
    [anon_sym_inline] = ACTIONS(1166),
    [anon_sym___inline] = ACTIONS(1166),
    [anon_sym___inline__] = ACTIONS(1166),
    [anon_sym___forceinline] = ACTIONS(1166),
    [anon_sym_thread_local] = ACTIONS(1166),
    [anon_sym___thread] = ACTIONS(1166),
    [anon_sym_const] = ACTIONS(1166),
    [anon_sym_constexpr] = ACTIONS(1166),
    [anon_sym_volatile] = ACTIONS(1166),
    [anon_sym_restrict] = ACTIONS(1166),
    [anon_sym___restrict__] = ACTIONS(1166),
    [anon_sym__Atomic] = ACTIONS(1166),
    [anon_sym__Noreturn] = ACTIONS(1166),
    [anon_sym_noreturn] = ACTIONS(1166),
    [anon_sym__Nonnull] = ACTIONS(1166),
    [anon_sym_alignas] = ACTIONS(1166),
    [anon_sym__Alignas] = ACTIONS(1166),
    [sym_primitive_type] = ACTIONS(1166),
    [anon_sym_enum] = ACTIONS(1166),
    [anon_sym_struct] = ACTIONS(1166),
    [anon_sym_union] = ACTIONS(1166),
    [anon_sym_if] = ACTIONS(1166),
    [anon_sym_else] = ACTIONS(1166),
    [anon_sym_switch] = ACTIONS(1166),
    [anon_sym_case] = ACTIONS(1166),
    [anon_sym_default] = ACTIONS(1166),
    [anon_sym_while] = ACTIONS(1166),
    [anon_sym_do] = ACTIONS(1166),
    [anon_sym_for] = ACTIONS(1166),
    [anon_sym_return] = ACTIONS(1166),
    [anon_sym_break] = ACTIONS(1166),
    [anon_sym_continue] = ACTIONS(1166),
    [anon_sym_goto] = ACTIONS(1166),
    [anon_sym___try] = ACTIONS(1166),
    [anon_sym___leave] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_PLUS_PLUS] = ACTIONS(1168),
    [anon_sym_sizeof] = ACTIONS(1166),
    [anon_sym___alignof__] = ACTIONS(1166),
    [anon_sym___alignof] = ACTIONS(1166),
    [anon_sym__alignof] = ACTIONS(1166),
    [anon_sym_alignof] = ACTIONS(1166),
    [anon_sym__Alignof] = ACTIONS(1166),
    [anon_sym_offsetof] = ACTIONS(1166),
    [anon_sym__Generic] = ACTIONS(1166),
    [anon_sym_asm] = ACTIONS(1166),
    [anon_sym___asm__] = ACTIONS(1166),
    [anon_sym___asm] = ACTIONS(1166),
    [sym_number_literal] = ACTIONS(1168),
    [anon_sym_L_SQUOTE] = ACTIONS(1168),
    [anon_sym_u_SQUOTE] = ACTIONS(1168),
    [anon_sym_U_SQUOTE] = ACTIONS(1168),
    [anon_sym_u8_SQUOTE] = ACTIONS(1168),
    [anon_sym_SQUOTE] = ACTIONS(1168),
    [anon_sym_L_DQUOTE] = ACTIONS(1168),
    [anon_sym_u_DQUOTE] = ACTIONS(1168),
    [anon_sym_U_DQUOTE] = ACTIONS(1168),
    [anon_sym_u8_DQUOTE] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1168),
    [sym_true] = ACTIONS(1166),
    [sym_false] = ACTIONS(1166),
    [anon_sym_NULL] = ACTIONS(1166),
    [anon_sym_nullptr] = ACTIONS(1166),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(92)] = {
    [sym_identifier] = ACTIONS(1170),
    [aux_sym_preproc_include_token1] = ACTIONS(1170),
    [aux_sym_preproc_def_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token2] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1170),
    [aux_sym_preproc_else_token1] = ACTIONS(1170),
    [aux_sym_preproc_elif_token1] = ACTIONS(1170),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1170),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1170),
    [sym_preproc_directive] = ACTIONS(1170),
    [anon_sym_LPAREN2] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_TILDE] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1172),
    [anon_sym_AMP] = ACTIONS(1172),
    [anon_sym_SEMI] = ACTIONS(1172),
    [anon_sym___extension__] = ACTIONS(1170),
    [anon_sym_typedef] = ACTIONS(1170),
    [anon_sym_extern] = ACTIONS(1170),
    [anon_sym___attribute__] = ACTIONS(1170),
    [anon_sym___attribute] = ACTIONS(1170),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1172),
    [anon_sym___declspec] = ACTIONS(1170),
    [anon_sym___cdecl] = ACTIONS(1170),
    [anon_sym___clrcall] = ACTIONS(1170),
    [anon_sym___stdcall] = ACTIONS(1170),
    [anon_sym___fastcall] = ACTIONS(1170),
    [anon_sym___thiscall] = ACTIONS(1170),
    [anon_sym___vectorcall] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_signed] = ACTIONS(1170),
    [anon_sym_unsigned] = ACTIONS(1170),
    [anon_sym_long] = ACTIONS(1170),
    [anon_sym_short] = ACTIONS(1170),
    [anon_sym_static] = ACTIONS(1170),
    [anon_sym_auto] = ACTIONS(1170),
    [anon_sym_register] = ACTIONS(1170),
    [anon_sym_inline] = ACTIONS(1170),
    [anon_sym___inline] = ACTIONS(1170),
    [anon_sym___inline__] = ACTIONS(1170),
    [anon_sym___forceinline] = ACTIONS(1170),
    [anon_sym_thread_local] = ACTIONS(1170),
    [anon_sym___thread] = ACTIONS(1170),
    [anon_sym_const] = ACTIONS(1170),
    [anon_sym_constexpr] = ACTIONS(1170),
    [anon_sym_volatile] = ACTIONS(1170),
    [anon_sym_restrict] = ACTIONS(1170),
    [anon_sym___restrict__] = ACTIONS(1170),
    [anon_sym__Atomic] = ACTIONS(1170),
    [anon_sym__Noreturn] = ACTIONS(1170),
    [anon_sym_noreturn] = ACTIONS(1170),
    [anon_sym__Nonnull] = ACTIONS(1170),
    [anon_sym_alignas] = ACTIONS(1170),
    [anon_sym__Alignas] = ACTIONS(1170),
    [sym_primitive_type] = ACTIONS(1170),
    [anon_sym_enum] = ACTIONS(1170),
    [anon_sym_struct] = ACTIONS(1170),
    [anon_sym_union] = ACTIONS(1170),
    [anon_sym_if] = ACTIONS(1170),
    [anon_sym_else] = ACTIONS(1170),
    [anon_sym_switch] = ACTIONS(1170),
    [anon_sym_case] = ACTIONS(1170),
    [anon_sym_default] = ACTIONS(1170),
    [anon_sym_while] = ACTIONS(1170),
    [anon_sym_do] = ACTIONS(1170),
    [anon_sym_for] = ACTIONS(1170),
    [anon_sym_return] = ACTIONS(1170),
    [anon_sym_break] = ACTIONS(1170),
    [anon_sym_continue] = ACTIONS(1170),
    [anon_sym_goto] = ACTIONS(1170),
    [anon_sym___try] = ACTIONS(1170),
    [anon_sym___leave] = ACTIONS(1170),
    [anon_sym_DASH_DASH] = ACTIONS(1172),
    [anon_sym_PLUS_PLUS] = ACTIONS(1172),
    [anon_sym_sizeof] = ACTIONS(1170),
    [anon_sym___alignof__] = ACTIONS(1170),
    [anon_sym___alignof] = ACTIONS(1170),
    [anon_sym__alignof] = ACTIONS(1170),
    [anon_sym_alignof] = ACTIONS(1170),
    [anon_sym__Alignof] = ACTIONS(1170),
    [anon_sym_offsetof] = ACTIONS(1170),
    [anon_sym__Generic] = ACTIONS(1170),
    [anon_sym_asm] = ACTIONS(1170),
    [anon_sym___asm__] = ACTIONS(1170),
    [anon_sym___asm] = ACTIONS(1170),
    [sym_number_literal] = ACTIONS(1172),
    [anon_sym_L_SQUOTE] = ACTIONS(1172),
    [anon_sym_u_SQUOTE] = ACTIONS(1172),
    [anon_sym_U_SQUOTE] = ACTIONS(1172),
    [anon_sym_u8_SQUOTE] = ACTIONS(1172),
    [anon_sym_SQUOTE] = ACTIONS(1172),
    [anon_sym_L_DQUOTE] = ACTIONS(1172),
    [anon_sym_u_DQUOTE] = ACTIONS(1172),
    [anon_sym_U_DQUOTE] = ACTIONS(1172),
    [anon_sym_u8_DQUOTE] = ACTIONS(1172),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [sym_true] = ACTIONS(1170),
    [sym_false] = ACTIONS(1170),
    [anon_sym_NULL] = ACTIONS(1170),
    [anon_sym_nullptr] = ACTIONS(1170),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(93)] = {
    [sym_identifier] = ACTIONS(1174),
    [aux_sym_preproc_include_token1] = ACTIONS(1174),
    [aux_sym_preproc_def_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token2] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1174),
    [aux_sym_preproc_else_token1] = ACTIONS(1174),
    [aux_sym_preproc_elif_token1] = ACTIONS(1174),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1174),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1174),
    [sym_preproc_directive] = ACTIONS(1174),
    [anon_sym_LPAREN2] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_TILDE] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1174),
    [anon_sym_PLUS] = ACTIONS(1174),
    [anon_sym_STAR] = ACTIONS(1176),
    [anon_sym_AMP] = ACTIONS(1176),
    [anon_sym_SEMI] = ACTIONS(1176),
    [anon_sym___extension__] = ACTIONS(1174),
    [anon_sym_typedef] = ACTIONS(1174),
    [anon_sym_extern] = ACTIONS(1174),
    [anon_sym___attribute__] = ACTIONS(1174),
    [anon_sym___attribute] = ACTIONS(1174),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1176),
    [anon_sym___declspec] = ACTIONS(1174),
    [anon_sym___cdecl] = ACTIONS(1174),
    [anon_sym___clrcall] = ACTIONS(1174),
    [anon_sym___stdcall] = ACTIONS(1174),
    [anon_sym___fastcall] = ACTIONS(1174),
    [anon_sym___thiscall] = ACTIONS(1174),
    [anon_sym___vectorcall] = ACTIONS(1174),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_signed] = ACTIONS(1174),
    [anon_sym_unsigned] = ACTIONS(1174),
    [anon_sym_long] = ACTIONS(1174),
    [anon_sym_short] = ACTIONS(1174),
    [anon_sym_static] = ACTIONS(1174),
    [anon_sym_auto] = ACTIONS(1174),
    [anon_sym_register] = ACTIONS(1174),
    [anon_sym_inline] = ACTIONS(1174),
    [anon_sym___inline] = ACTIONS(1174),
    [anon_sym___inline__] = ACTIONS(1174),
    [anon_sym___forceinline] = ACTIONS(1174),
    [anon_sym_thread_local] = ACTIONS(1174),
    [anon_sym___thread] = ACTIONS(1174),
    [anon_sym_const] = ACTIONS(1174),
    [anon_sym_constexpr] = ACTIONS(1174),
    [anon_sym_volatile] = ACTIONS(1174),
    [anon_sym_restrict] = ACTIONS(1174),
    [anon_sym___restrict__] = ACTIONS(1174),
    [anon_sym__Atomic] = ACTIONS(1174),
    [anon_sym__Noreturn] = ACTIONS(1174),
    [anon_sym_noreturn] = ACTIONS(1174),
    [anon_sym__Nonnull] = ACTIONS(1174),
    [anon_sym_alignas] = ACTIONS(1174),
    [anon_sym__Alignas] = ACTIONS(1174),
    [sym_primitive_type] = ACTIONS(1174),
    [anon_sym_enum] = ACTIONS(1174),
    [anon_sym_struct] = ACTIONS(1174),
    [anon_sym_union] = ACTIONS(1174),
    [anon_sym_if] = ACTIONS(1174),
    [anon_sym_else] = ACTIONS(1174),
    [anon_sym_switch] = ACTIONS(1174),
    [anon_sym_case] = ACTIONS(1174),
    [anon_sym_default] = ACTIONS(1174),
    [anon_sym_while] = ACTIONS(1174),
    [anon_sym_do] = ACTIONS(1174),
    [anon_sym_for] = ACTIONS(1174),
    [anon_sym_return] = ACTIONS(1174),
    [anon_sym_break] = ACTIONS(1174),
    [anon_sym_continue] = ACTIONS(1174),
    [anon_sym_goto] = ACTIONS(1174),
    [anon_sym___try] = ACTIONS(1174),
    [anon_sym___leave] = ACTIONS(1174),
    [anon_sym_DASH_DASH] = ACTIONS(1176),
    [anon_sym_PLUS_PLUS] = ACTIONS(1176),
    [anon_sym_sizeof] = ACTIONS(1174),
    [anon_sym___alignof__] = ACTIONS(1174),
    [anon_sym___alignof] = ACTIONS(1174),
    [anon_sym__alignof] = ACTIONS(1174),
    [anon_sym_alignof] = ACTIONS(1174),
    [anon_sym__Alignof] = ACTIONS(1174),
    [anon_sym_offsetof] = ACTIONS(1174),
    [anon_sym__Generic] = ACTIONS(1174),
    [anon_sym_asm] = ACTIONS(1174),
    [anon_sym___asm__] = ACTIONS(1174),
    [anon_sym___asm] = ACTIONS(1174),
    [sym_number_literal] = ACTIONS(1176),
    [anon_sym_L_SQUOTE] = ACTIONS(1176),
    [anon_sym_u_SQUOTE] = ACTIONS(1176),
    [anon_sym_U_SQUOTE] = ACTIONS(1176),
    [anon_sym_u8_SQUOTE] = ACTIONS(1176),
    [anon_sym_SQUOTE] = ACTIONS(1176),
    [anon_sym_L_DQUOTE] = ACTIONS(1176),
    [anon_sym_u_DQUOTE] = ACTIONS(1176),
    [anon_sym_U_DQUOTE] = ACTIONS(1176),
    [anon_sym_u8_DQUOTE] = ACTIONS(1176),
    [anon_sym_DQUOTE] = ACTIONS(1176),
    [sym_true] = ACTIONS(1174),
    [sym_false] = ACTIONS(1174),
    [anon_sym_NULL] = ACTIONS(1174),
    [anon_sym_nullptr] = ACTIONS(1174),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(94)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token2] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [aux_sym_preproc_else_token1] = ACTIONS(1178),
    [aux_sym_preproc_elif_token1] = ACTIONS(1178),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(95)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token2] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [aux_sym_preproc_else_token1] = ACTIONS(1178),
    [aux_sym_preproc_elif_token1] = ACTIONS(1178),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(96)] = {
    [sym_identifier] = ACTIONS(1182),
    [aux_sym_preproc_include_token1] = ACTIONS(1182),
    [aux_sym_preproc_def_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token2] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1182),
    [aux_sym_preproc_else_token1] = ACTIONS(1182),
    [aux_sym_preproc_elif_token1] = ACTIONS(1182),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1182),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1182),
    [sym_preproc_directive] = ACTIONS(1182),
    [anon_sym_LPAREN2] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_TILDE] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1182),
    [anon_sym_PLUS] = ACTIONS(1182),
    [anon_sym_STAR] = ACTIONS(1184),
    [anon_sym_AMP] = ACTIONS(1184),
    [anon_sym_SEMI] = ACTIONS(1184),
    [anon_sym___extension__] = ACTIONS(1182),
    [anon_sym_typedef] = ACTIONS(1182),
    [anon_sym_extern] = ACTIONS(1182),
    [anon_sym___attribute__] = ACTIONS(1182),
    [anon_sym___attribute] = ACTIONS(1182),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1184),
    [anon_sym___declspec] = ACTIONS(1182),
    [anon_sym___cdecl] = ACTIONS(1182),
    [anon_sym___clrcall] = ACTIONS(1182),
    [anon_sym___stdcall] = ACTIONS(1182),
    [anon_sym___fastcall] = ACTIONS(1182),
    [anon_sym___thiscall] = ACTIONS(1182),
    [anon_sym___vectorcall] = ACTIONS(1182),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_signed] = ACTIONS(1182),
    [anon_sym_unsigned] = ACTIONS(1182),
    [anon_sym_long] = ACTIONS(1182),
    [anon_sym_short] = ACTIONS(1182),
    [anon_sym_static] = ACTIONS(1182),
    [anon_sym_auto] = ACTIONS(1182),
    [anon_sym_register] = ACTIONS(1182),
    [anon_sym_inline] = ACTIONS(1182),
    [anon_sym___inline] = ACTIONS(1182),
    [anon_sym___inline__] = ACTIONS(1182),
    [anon_sym___forceinline] = ACTIONS(1182),
    [anon_sym_thread_local] = ACTIONS(1182),
    [anon_sym___thread] = ACTIONS(1182),
    [anon_sym_const] = ACTIONS(1182),
    [anon_sym_constexpr] = ACTIONS(1182),
    [anon_sym_volatile] = ACTIONS(1182),
    [anon_sym_restrict] = ACTIONS(1182),
    [anon_sym___restrict__] = ACTIONS(1182),
    [anon_sym__Atomic] = ACTIONS(1182),
    [anon_sym__Noreturn] = ACTIONS(1182),
    [anon_sym_noreturn] = ACTIONS(1182),
    [anon_sym__Nonnull] = ACTIONS(1182),
    [anon_sym_alignas] = ACTIONS(1182),
    [anon_sym__Alignas] = ACTIONS(1182),
    [sym_primitive_type] = ACTIONS(1182),
    [anon_sym_enum] = ACTIONS(1182),
    [anon_sym_struct] = ACTIONS(1182),
    [anon_sym_union] = ACTIONS(1182),
    [anon_sym_if] = ACTIONS(1182),
    [anon_sym_else] = ACTIONS(1182),
    [anon_sym_switch] = ACTIONS(1182),
    [anon_sym_case] = ACTIONS(1182),
    [anon_sym_default] = ACTIONS(1182),
    [anon_sym_while] = ACTIONS(1182),
    [anon_sym_do] = ACTIONS(1182),
    [anon_sym_for] = ACTIONS(1182),
    [anon_sym_return] = ACTIONS(1182),
    [anon_sym_break] = ACTIONS(1182),
    [anon_sym_continue] = ACTIONS(1182),
    [anon_sym_goto] = ACTIONS(1182),
    [anon_sym___try] = ACTIONS(1182),
    [anon_sym___leave] = ACTIONS(1182),
    [anon_sym_DASH_DASH] = ACTIONS(1184),
    [anon_sym_PLUS_PLUS] = ACTIONS(1184),
    [anon_sym_sizeof] = ACTIONS(1182),
    [anon_sym___alignof__] = ACTIONS(1182),
    [anon_sym___alignof] = ACTIONS(1182),
    [anon_sym__alignof] = ACTIONS(1182),
    [anon_sym_alignof] = ACTIONS(1182),
    [anon_sym__Alignof] = ACTIONS(1182),
    [anon_sym_offsetof] = ACTIONS(1182),
    [anon_sym__Generic] = ACTIONS(1182),
    [anon_sym_asm] = ACTIONS(1182),
    [anon_sym___asm__] = ACTIONS(1182),
    [anon_sym___asm] = ACTIONS(1182),
    [sym_number_literal] = ACTIONS(1184),
    [anon_sym_L_SQUOTE] = ACTIONS(1184),
    [anon_sym_u_SQUOTE] = ACTIONS(1184),
    [anon_sym_U_SQUOTE] = ACTIONS(1184),
    [anon_sym_u8_SQUOTE] = ACTIONS(1184),
    [anon_sym_SQUOTE] = ACTIONS(1184),
    [anon_sym_L_DQUOTE] = ACTIONS(1184),
    [anon_sym_u_DQUOTE] = ACTIONS(1184),
    [anon_sym_U_DQUOTE] = ACTIONS(1184),
    [anon_sym_u8_DQUOTE] = ACTIONS(1184),
    [anon_sym_DQUOTE] = ACTIONS(1184),
    [sym_true] = ACTIONS(1182),
    [sym_false] = ACTIONS(1182),
    [anon_sym_NULL] = ACTIONS(1182),
    [anon_sym_nullptr] = ACTIONS(1182),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(97)] = {
    [sym_identifier] = ACTIONS(1186),
    [aux_sym_preproc_include_token1] = ACTIONS(1186),
    [aux_sym_preproc_def_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token2] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1186),
    [aux_sym_preproc_else_token1] = ACTIONS(1186),
    [aux_sym_preproc_elif_token1] = ACTIONS(1186),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1186),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1186),
    [sym_preproc_directive] = ACTIONS(1186),
    [anon_sym_LPAREN2] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1186),
    [anon_sym_PLUS] = ACTIONS(1186),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym___extension__] = ACTIONS(1186),
    [anon_sym_typedef] = ACTIONS(1186),
    [anon_sym_extern] = ACTIONS(1186),
    [anon_sym___attribute__] = ACTIONS(1186),
    [anon_sym___attribute] = ACTIONS(1186),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1188),
    [anon_sym___declspec] = ACTIONS(1186),
    [anon_sym___cdecl] = ACTIONS(1186),
    [anon_sym___clrcall] = ACTIONS(1186),
    [anon_sym___stdcall] = ACTIONS(1186),
    [anon_sym___fastcall] = ACTIONS(1186),
    [anon_sym___thiscall] = ACTIONS(1186),
    [anon_sym___vectorcall] = ACTIONS(1186),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_signed] = ACTIONS(1186),
    [anon_sym_unsigned] = ACTIONS(1186),
    [anon_sym_long] = ACTIONS(1186),
    [anon_sym_short] = ACTIONS(1186),
    [anon_sym_static] = ACTIONS(1186),
    [anon_sym_auto] = ACTIONS(1186),
    [anon_sym_register] = ACTIONS(1186),
    [anon_sym_inline] = ACTIONS(1186),
    [anon_sym___inline] = ACTIONS(1186),
    [anon_sym___inline__] = ACTIONS(1186),
    [anon_sym___forceinline] = ACTIONS(1186),
    [anon_sym_thread_local] = ACTIONS(1186),
    [anon_sym___thread] = ACTIONS(1186),
    [anon_sym_const] = ACTIONS(1186),
    [anon_sym_constexpr] = ACTIONS(1186),
    [anon_sym_volatile] = ACTIONS(1186),
    [anon_sym_restrict] = ACTIONS(1186),
    [anon_sym___restrict__] = ACTIONS(1186),
    [anon_sym__Atomic] = ACTIONS(1186),
    [anon_sym__Noreturn] = ACTIONS(1186),
    [anon_sym_noreturn] = ACTIONS(1186),
    [anon_sym__Nonnull] = ACTIONS(1186),
    [anon_sym_alignas] = ACTIONS(1186),
    [anon_sym__Alignas] = ACTIONS(1186),
    [sym_primitive_type] = ACTIONS(1186),
    [anon_sym_enum] = ACTIONS(1186),
    [anon_sym_struct] = ACTIONS(1186),
    [anon_sym_union] = ACTIONS(1186),
    [anon_sym_if] = ACTIONS(1186),
    [anon_sym_else] = ACTIONS(1186),
    [anon_sym_switch] = ACTIONS(1186),
    [anon_sym_case] = ACTIONS(1186),
    [anon_sym_default] = ACTIONS(1186),
    [anon_sym_while] = ACTIONS(1186),
    [anon_sym_do] = ACTIONS(1186),
    [anon_sym_for] = ACTIONS(1186),
    [anon_sym_return] = ACTIONS(1186),
    [anon_sym_break] = ACTIONS(1186),
    [anon_sym_continue] = ACTIONS(1186),
    [anon_sym_goto] = ACTIONS(1186),
    [anon_sym___try] = ACTIONS(1186),
    [anon_sym___leave] = ACTIONS(1186),
    [anon_sym_DASH_DASH] = ACTIONS(1188),
    [anon_sym_PLUS_PLUS] = ACTIONS(1188),
    [anon_sym_sizeof] = ACTIONS(1186),
    [anon_sym___alignof__] = ACTIONS(1186),
    [anon_sym___alignof] = ACTIONS(1186),
    [anon_sym__alignof] = ACTIONS(1186),
    [anon_sym_alignof] = ACTIONS(1186),
    [anon_sym__Alignof] = ACTIONS(1186),
    [anon_sym_offsetof] = ACTIONS(1186),
    [anon_sym__Generic] = ACTIONS(1186),
    [anon_sym_asm] = ACTIONS(1186),
    [anon_sym___asm__] = ACTIONS(1186),
    [anon_sym___asm] = ACTIONS(1186),
    [sym_number_literal] = ACTIONS(1188),
    [anon_sym_L_SQUOTE] = ACTIONS(1188),
    [anon_sym_u_SQUOTE] = ACTIONS(1188),
    [anon_sym_U_SQUOTE] = ACTIONS(1188),
    [anon_sym_u8_SQUOTE] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_L_DQUOTE] = ACTIONS(1188),
    [anon_sym_u_DQUOTE] = ACTIONS(1188),
    [anon_sym_U_DQUOTE] = ACTIONS(1188),
    [anon_sym_u8_DQUOTE] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [sym_true] = ACTIONS(1186),
    [sym_false] = ACTIONS(1186),
    [anon_sym_NULL] = ACTIONS(1186),
    [anon_sym_nullptr] = ACTIONS(1186),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(98)] = {
    [sym_identifier] = ACTIONS(1190),
    [aux_sym_preproc_include_token1] = ACTIONS(1190),
    [aux_sym_preproc_def_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token2] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1190),
    [aux_sym_preproc_else_token1] = ACTIONS(1190),
    [aux_sym_preproc_elif_token1] = ACTIONS(1190),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1190),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1190),
    [sym_preproc_directive] = ACTIONS(1190),
    [anon_sym_LPAREN2] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_TILDE] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1192),
    [anon_sym_AMP] = ACTIONS(1192),
    [anon_sym_SEMI] = ACTIONS(1192),
    [anon_sym___extension__] = ACTIONS(1190),
    [anon_sym_typedef] = ACTIONS(1190),
    [anon_sym_extern] = ACTIONS(1190),
    [anon_sym___attribute__] = ACTIONS(1190),
    [anon_sym___attribute] = ACTIONS(1190),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1192),
    [anon_sym___declspec] = ACTIONS(1190),
    [anon_sym___cdecl] = ACTIONS(1190),
    [anon_sym___clrcall] = ACTIONS(1190),
    [anon_sym___stdcall] = ACTIONS(1190),
    [anon_sym___fastcall] = ACTIONS(1190),
    [anon_sym___thiscall] = ACTIONS(1190),
    [anon_sym___vectorcall] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_signed] = ACTIONS(1190),
    [anon_sym_unsigned] = ACTIONS(1190),
    [anon_sym_long] = ACTIONS(1190),
    [anon_sym_short] = ACTIONS(1190),
    [anon_sym_static] = ACTIONS(1190),
    [anon_sym_auto] = ACTIONS(1190),
    [anon_sym_register] = ACTIONS(1190),
    [anon_sym_inline] = ACTIONS(1190),
    [anon_sym___inline] = ACTIONS(1190),
    [anon_sym___inline__] = ACTIONS(1190),
    [anon_sym___forceinline] = ACTIONS(1190),
    [anon_sym_thread_local] = ACTIONS(1190),
    [anon_sym___thread] = ACTIONS(1190),
    [anon_sym_const] = ACTIONS(1190),
    [anon_sym_constexpr] = ACTIONS(1190),
    [anon_sym_volatile] = ACTIONS(1190),
    [anon_sym_restrict] = ACTIONS(1190),
    [anon_sym___restrict__] = ACTIONS(1190),
    [anon_sym__Atomic] = ACTIONS(1190),
    [anon_sym__Noreturn] = ACTIONS(1190),
    [anon_sym_noreturn] = ACTIONS(1190),
    [anon_sym__Nonnull] = ACTIONS(1190),
    [anon_sym_alignas] = ACTIONS(1190),
    [anon_sym__Alignas] = ACTIONS(1190),
    [sym_primitive_type] = ACTIONS(1190),
    [anon_sym_enum] = ACTIONS(1190),
    [anon_sym_struct] = ACTIONS(1190),
    [anon_sym_union] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_else] = ACTIONS(1190),
    [anon_sym_switch] = ACTIONS(1190),
    [anon_sym_case] = ACTIONS(1190),
    [anon_sym_default] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_break] = ACTIONS(1190),
    [anon_sym_continue] = ACTIONS(1190),
    [anon_sym_goto] = ACTIONS(1190),
    [anon_sym___try] = ACTIONS(1190),
    [anon_sym___leave] = ACTIONS(1190),
    [anon_sym_DASH_DASH] = ACTIONS(1192),
    [anon_sym_PLUS_PLUS] = ACTIONS(1192),
    [anon_sym_sizeof] = ACTIONS(1190),
    [anon_sym___alignof__] = ACTIONS(1190),
    [anon_sym___alignof] = ACTIONS(1190),
    [anon_sym__alignof] = ACTIONS(1190),
    [anon_sym_alignof] = ACTIONS(1190),
    [anon_sym__Alignof] = ACTIONS(1190),
    [anon_sym_offsetof] = ACTIONS(1190),
    [anon_sym__Generic] = ACTIONS(1190),
    [anon_sym_asm] = ACTIONS(1190),
    [anon_sym___asm__] = ACTIONS(1190),
    [anon_sym___asm] = ACTIONS(1190),
    [sym_number_literal] = ACTIONS(1192),
    [anon_sym_L_SQUOTE] = ACTIONS(1192),
    [anon_sym_u_SQUOTE] = ACTIONS(1192),
    [anon_sym_U_SQUOTE] = ACTIONS(1192),
    [anon_sym_u8_SQUOTE] = ACTIONS(1192),
    [anon_sym_SQUOTE] = ACTIONS(1192),
    [anon_sym_L_DQUOTE] = ACTIONS(1192),
    [anon_sym_u_DQUOTE] = ACTIONS(1192),
    [anon_sym_U_DQUOTE] = ACTIONS(1192),
    [anon_sym_u8_DQUOTE] = ACTIONS(1192),
    [anon_sym_DQUOTE] = ACTIONS(1192),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [anon_sym_NULL] = ACTIONS(1190),
    [anon_sym_nullptr] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(99)] = {
    [sym_identifier] = ACTIONS(1194),
    [aux_sym_preproc_include_token1] = ACTIONS(1194),
    [aux_sym_preproc_def_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token2] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1194),
    [aux_sym_preproc_else_token1] = ACTIONS(1194),
    [aux_sym_preproc_elif_token1] = ACTIONS(1194),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1194),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1194),
    [sym_preproc_directive] = ACTIONS(1194),
    [anon_sym_LPAREN2] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_TILDE] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1196),
    [anon_sym_AMP] = ACTIONS(1196),
    [anon_sym_SEMI] = ACTIONS(1196),
    [anon_sym___extension__] = ACTIONS(1194),
    [anon_sym_typedef] = ACTIONS(1194),
    [anon_sym_extern] = ACTIONS(1194),
    [anon_sym___attribute__] = ACTIONS(1194),
    [anon_sym___attribute] = ACTIONS(1194),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1196),
    [anon_sym___declspec] = ACTIONS(1194),
    [anon_sym___cdecl] = ACTIONS(1194),
    [anon_sym___clrcall] = ACTIONS(1194),
    [anon_sym___stdcall] = ACTIONS(1194),
    [anon_sym___fastcall] = ACTIONS(1194),
    [anon_sym___thiscall] = ACTIONS(1194),
    [anon_sym___vectorcall] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_signed] = ACTIONS(1194),
    [anon_sym_unsigned] = ACTIONS(1194),
    [anon_sym_long] = ACTIONS(1194),
    [anon_sym_short] = ACTIONS(1194),
    [anon_sym_static] = ACTIONS(1194),
    [anon_sym_auto] = ACTIONS(1194),
    [anon_sym_register] = ACTIONS(1194),
    [anon_sym_inline] = ACTIONS(1194),
    [anon_sym___inline] = ACTIONS(1194),
    [anon_sym___inline__] = ACTIONS(1194),
    [anon_sym___forceinline] = ACTIONS(1194),
    [anon_sym_thread_local] = ACTIONS(1194),
    [anon_sym___thread] = ACTIONS(1194),
    [anon_sym_const] = ACTIONS(1194),
    [anon_sym_constexpr] = ACTIONS(1194),
    [anon_sym_volatile] = ACTIONS(1194),
    [anon_sym_restrict] = ACTIONS(1194),
    [anon_sym___restrict__] = ACTIONS(1194),
    [anon_sym__Atomic] = ACTIONS(1194),
    [anon_sym__Noreturn] = ACTIONS(1194),
    [anon_sym_noreturn] = ACTIONS(1194),
    [anon_sym__Nonnull] = ACTIONS(1194),
    [anon_sym_alignas] = ACTIONS(1194),
    [anon_sym__Alignas] = ACTIONS(1194),
    [sym_primitive_type] = ACTIONS(1194),
    [anon_sym_enum] = ACTIONS(1194),
    [anon_sym_struct] = ACTIONS(1194),
    [anon_sym_union] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_else] = ACTIONS(1194),
    [anon_sym_switch] = ACTIONS(1194),
    [anon_sym_case] = ACTIONS(1194),
    [anon_sym_default] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_break] = ACTIONS(1194),
    [anon_sym_continue] = ACTIONS(1194),
    [anon_sym_goto] = ACTIONS(1194),
    [anon_sym___try] = ACTIONS(1194),
    [anon_sym___leave] = ACTIONS(1194),
    [anon_sym_DASH_DASH] = ACTIONS(1196),
    [anon_sym_PLUS_PLUS] = ACTIONS(1196),
    [anon_sym_sizeof] = ACTIONS(1194),
    [anon_sym___alignof__] = ACTIONS(1194),
    [anon_sym___alignof] = ACTIONS(1194),
    [anon_sym__alignof] = ACTIONS(1194),
    [anon_sym_alignof] = ACTIONS(1194),
    [anon_sym__Alignof] = ACTIONS(1194),
    [anon_sym_offsetof] = ACTIONS(1194),
    [anon_sym__Generic] = ACTIONS(1194),
    [anon_sym_asm] = ACTIONS(1194),
    [anon_sym___asm__] = ACTIONS(1194),
    [anon_sym___asm] = ACTIONS(1194),
    [sym_number_literal] = ACTIONS(1196),
    [anon_sym_L_SQUOTE] = ACTIONS(1196),
    [anon_sym_u_SQUOTE] = ACTIONS(1196),
    [anon_sym_U_SQUOTE] = ACTIONS(1196),
    [anon_sym_u8_SQUOTE] = ACTIONS(1196),
    [anon_sym_SQUOTE] = ACTIONS(1196),
    [anon_sym_L_DQUOTE] = ACTIONS(1196),
    [anon_sym_u_DQUOTE] = ACTIONS(1196),
    [anon_sym_U_DQUOTE] = ACTIONS(1196),
    [anon_sym_u8_DQUOTE] = ACTIONS(1196),
    [anon_sym_DQUOTE] = ACTIONS(1196),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [anon_sym_NULL] = ACTIONS(1194),
    [anon_sym_nullptr] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(100)] = {
    [sym_identifier] = ACTIONS(1198),
    [aux_sym_preproc_include_token1] = ACTIONS(1198),
    [aux_sym_preproc_def_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token2] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1198),
    [aux_sym_preproc_else_token1] = ACTIONS(1198),
    [aux_sym_preproc_elif_token1] = ACTIONS(1198),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1198),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1198),
    [sym_preproc_directive] = ACTIONS(1198),
    [anon_sym_LPAREN2] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1198),
    [anon_sym_PLUS] = ACTIONS(1198),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym___extension__] = ACTIONS(1198),
    [anon_sym_typedef] = ACTIONS(1198),
    [anon_sym_extern] = ACTIONS(1198),
    [anon_sym___attribute__] = ACTIONS(1198),
    [anon_sym___attribute] = ACTIONS(1198),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1200),
    [anon_sym___declspec] = ACTIONS(1198),
    [anon_sym___cdecl] = ACTIONS(1198),
    [anon_sym___clrcall] = ACTIONS(1198),
    [anon_sym___stdcall] = ACTIONS(1198),
    [anon_sym___fastcall] = ACTIONS(1198),
    [anon_sym___thiscall] = ACTIONS(1198),
    [anon_sym___vectorcall] = ACTIONS(1198),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_signed] = ACTIONS(1198),
    [anon_sym_unsigned] = ACTIONS(1198),
    [anon_sym_long] = ACTIONS(1198),
    [anon_sym_short] = ACTIONS(1198),
    [anon_sym_static] = ACTIONS(1198),
    [anon_sym_auto] = ACTIONS(1198),
    [anon_sym_register] = ACTIONS(1198),
    [anon_sym_inline] = ACTIONS(1198),
    [anon_sym___inline] = ACTIONS(1198),
    [anon_sym___inline__] = ACTIONS(1198),
    [anon_sym___forceinline] = ACTIONS(1198),
    [anon_sym_thread_local] = ACTIONS(1198),
    [anon_sym___thread] = ACTIONS(1198),
    [anon_sym_const] = ACTIONS(1198),
    [anon_sym_constexpr] = ACTIONS(1198),
    [anon_sym_volatile] = ACTIONS(1198),
    [anon_sym_restrict] = ACTIONS(1198),
    [anon_sym___restrict__] = ACTIONS(1198),
    [anon_sym__Atomic] = ACTIONS(1198),
    [anon_sym__Noreturn] = ACTIONS(1198),
    [anon_sym_noreturn] = ACTIONS(1198),
    [anon_sym__Nonnull] = ACTIONS(1198),
    [anon_sym_alignas] = ACTIONS(1198),
    [anon_sym__Alignas] = ACTIONS(1198),
    [sym_primitive_type] = ACTIONS(1198),
    [anon_sym_enum] = ACTIONS(1198),
    [anon_sym_struct] = ACTIONS(1198),
    [anon_sym_union] = ACTIONS(1198),
    [anon_sym_if] = ACTIONS(1198),
    [anon_sym_else] = ACTIONS(1198),
    [anon_sym_switch] = ACTIONS(1198),
    [anon_sym_case] = ACTIONS(1198),
    [anon_sym_default] = ACTIONS(1198),
    [anon_sym_while] = ACTIONS(1198),
    [anon_sym_do] = ACTIONS(1198),
    [anon_sym_for] = ACTIONS(1198),
    [anon_sym_return] = ACTIONS(1198),
    [anon_sym_break] = ACTIONS(1198),
    [anon_sym_continue] = ACTIONS(1198),
    [anon_sym_goto] = ACTIONS(1198),
    [anon_sym___try] = ACTIONS(1198),
    [anon_sym___leave] = ACTIONS(1198),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_sizeof] = ACTIONS(1198),
    [anon_sym___alignof__] = ACTIONS(1198),
    [anon_sym___alignof] = ACTIONS(1198),
    [anon_sym__alignof] = ACTIONS(1198),
    [anon_sym_alignof] = ACTIONS(1198),
    [anon_sym__Alignof] = ACTIONS(1198),
    [anon_sym_offsetof] = ACTIONS(1198),
    [anon_sym__Generic] = ACTIONS(1198),
    [anon_sym_asm] = ACTIONS(1198),
    [anon_sym___asm__] = ACTIONS(1198),
    [anon_sym___asm] = ACTIONS(1198),
    [sym_number_literal] = ACTIONS(1200),
    [anon_sym_L_SQUOTE] = ACTIONS(1200),
    [anon_sym_u_SQUOTE] = ACTIONS(1200),
    [anon_sym_U_SQUOTE] = ACTIONS(1200),
    [anon_sym_u8_SQUOTE] = ACTIONS(1200),
    [anon_sym_SQUOTE] = ACTIONS(1200),
    [anon_sym_L_DQUOTE] = ACTIONS(1200),
    [anon_sym_u_DQUOTE] = ACTIONS(1200),
    [anon_sym_U_DQUOTE] = ACTIONS(1200),
    [anon_sym_u8_DQUOTE] = ACTIONS(1200),
    [anon_sym_DQUOTE] = ACTIONS(1200),
    [sym_true] = ACTIONS(1198),
    [sym_false] = ACTIONS(1198),
    [anon_sym_NULL] = ACTIONS(1198),
    [anon_sym_nullptr] = ACTIONS(1198),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(101)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token2] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [aux_sym_preproc_else_token1] = ACTIONS(1202),
    [aux_sym_preproc_elif_token1] = ACTIONS(1202),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(102)] = {
    [sym_identifier] = ACTIONS(1206),
    [aux_sym_preproc_include_token1] = ACTIONS(1206),
    [aux_sym_preproc_def_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token2] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1206),
    [aux_sym_preproc_else_token1] = ACTIONS(1206),
    [aux_sym_preproc_elif_token1] = ACTIONS(1206),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1206),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1206),
    [sym_preproc_directive] = ACTIONS(1206),
    [anon_sym_LPAREN2] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_TILDE] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1208),
    [anon_sym_AMP] = ACTIONS(1208),
    [anon_sym_SEMI] = ACTIONS(1208),
    [anon_sym___extension__] = ACTIONS(1206),
    [anon_sym_typedef] = ACTIONS(1206),
    [anon_sym_extern] = ACTIONS(1206),
    [anon_sym___attribute__] = ACTIONS(1206),
    [anon_sym___attribute] = ACTIONS(1206),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1208),
    [anon_sym___declspec] = ACTIONS(1206),
    [anon_sym___cdecl] = ACTIONS(1206),
    [anon_sym___clrcall] = ACTIONS(1206),
    [anon_sym___stdcall] = ACTIONS(1206),
    [anon_sym___fastcall] = ACTIONS(1206),
    [anon_sym___thiscall] = ACTIONS(1206),
    [anon_sym___vectorcall] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_signed] = ACTIONS(1206),
    [anon_sym_unsigned] = ACTIONS(1206),
    [anon_sym_long] = ACTIONS(1206),
    [anon_sym_short] = ACTIONS(1206),
    [anon_sym_static] = ACTIONS(1206),
    [anon_sym_auto] = ACTIONS(1206),
    [anon_sym_register] = ACTIONS(1206),
    [anon_sym_inline] = ACTIONS(1206),
    [anon_sym___inline] = ACTIONS(1206),
    [anon_sym___inline__] = ACTIONS(1206),
    [anon_sym___forceinline] = ACTIONS(1206),
    [anon_sym_thread_local] = ACTIONS(1206),
    [anon_sym___thread] = ACTIONS(1206),
    [anon_sym_const] = ACTIONS(1206),
    [anon_sym_constexpr] = ACTIONS(1206),
    [anon_sym_volatile] = ACTIONS(1206),
    [anon_sym_restrict] = ACTIONS(1206),
    [anon_sym___restrict__] = ACTIONS(1206),
    [anon_sym__Atomic] = ACTIONS(1206),
    [anon_sym__Noreturn] = ACTIONS(1206),
    [anon_sym_noreturn] = ACTIONS(1206),
    [anon_sym__Nonnull] = ACTIONS(1206),
    [anon_sym_alignas] = ACTIONS(1206),
    [anon_sym__Alignas] = ACTIONS(1206),
    [sym_primitive_type] = ACTIONS(1206),
    [anon_sym_enum] = ACTIONS(1206),
    [anon_sym_struct] = ACTIONS(1206),
    [anon_sym_union] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1206),
    [anon_sym_else] = ACTIONS(1206),
    [anon_sym_switch] = ACTIONS(1206),
    [anon_sym_case] = ACTIONS(1206),
    [anon_sym_default] = ACTIONS(1206),
    [anon_sym_while] = ACTIONS(1206),
    [anon_sym_do] = ACTIONS(1206),
    [anon_sym_for] = ACTIONS(1206),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_break] = ACTIONS(1206),
    [anon_sym_continue] = ACTIONS(1206),
    [anon_sym_goto] = ACTIONS(1206),
    [anon_sym___try] = ACTIONS(1206),
    [anon_sym___leave] = ACTIONS(1206),
    [anon_sym_DASH_DASH] = ACTIONS(1208),
    [anon_sym_PLUS_PLUS] = ACTIONS(1208),
    [anon_sym_sizeof] = ACTIONS(1206),
    [anon_sym___alignof__] = ACTIONS(1206),
    [anon_sym___alignof] = ACTIONS(1206),
    [anon_sym__alignof] = ACTIONS(1206),
    [anon_sym_alignof] = ACTIONS(1206),
    [anon_sym__Alignof] = ACTIONS(1206),
    [anon_sym_offsetof] = ACTIONS(1206),
    [anon_sym__Generic] = ACTIONS(1206),
    [anon_sym_asm] = ACTIONS(1206),
    [anon_sym___asm__] = ACTIONS(1206),
    [anon_sym___asm] = ACTIONS(1206),
    [sym_number_literal] = ACTIONS(1208),
    [anon_sym_L_SQUOTE] = ACTIONS(1208),
    [anon_sym_u_SQUOTE] = ACTIONS(1208),
    [anon_sym_U_SQUOTE] = ACTIONS(1208),
    [anon_sym_u8_SQUOTE] = ACTIONS(1208),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_L_DQUOTE] = ACTIONS(1208),
    [anon_sym_u_DQUOTE] = ACTIONS(1208),
    [anon_sym_U_DQUOTE] = ACTIONS(1208),
    [anon_sym_u8_DQUOTE] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [sym_true] = ACTIONS(1206),
    [sym_false] = ACTIONS(1206),
    [anon_sym_NULL] = ACTIONS(1206),
    [anon_sym_nullptr] = ACTIONS(1206),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(103)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token2] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [aux_sym_preproc_else_token1] = ACTIONS(1202),
    [aux_sym_preproc_elif_token1] = ACTIONS(1202),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(104)] = {
    [sym_identifier] = ACTIONS(1210),
    [aux_sym_preproc_include_token1] = ACTIONS(1210),
    [aux_sym_preproc_def_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token2] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1210),
    [aux_sym_preproc_else_token1] = ACTIONS(1210),
    [aux_sym_preproc_elif_token1] = ACTIONS(1210),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1210),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1210),
    [sym_preproc_directive] = ACTIONS(1210),
    [anon_sym_LPAREN2] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1210),
    [anon_sym_PLUS] = ACTIONS(1210),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym___extension__] = ACTIONS(1210),
    [anon_sym_typedef] = ACTIONS(1210),
    [anon_sym_extern] = ACTIONS(1210),
    [anon_sym___attribute__] = ACTIONS(1210),
    [anon_sym___attribute] = ACTIONS(1210),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1212),
    [anon_sym___declspec] = ACTIONS(1210),
    [anon_sym___cdecl] = ACTIONS(1210),
    [anon_sym___clrcall] = ACTIONS(1210),
    [anon_sym___stdcall] = ACTIONS(1210),
    [anon_sym___fastcall] = ACTIONS(1210),
    [anon_sym___thiscall] = ACTIONS(1210),
    [anon_sym___vectorcall] = ACTIONS(1210),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_signed] = ACTIONS(1210),
    [anon_sym_unsigned] = ACTIONS(1210),
    [anon_sym_long] = ACTIONS(1210),
    [anon_sym_short] = ACTIONS(1210),
    [anon_sym_static] = ACTIONS(1210),
    [anon_sym_auto] = ACTIONS(1210),
    [anon_sym_register] = ACTIONS(1210),
    [anon_sym_inline] = ACTIONS(1210),
    [anon_sym___inline] = ACTIONS(1210),
    [anon_sym___inline__] = ACTIONS(1210),
    [anon_sym___forceinline] = ACTIONS(1210),
    [anon_sym_thread_local] = ACTIONS(1210),
    [anon_sym___thread] = ACTIONS(1210),
    [anon_sym_const] = ACTIONS(1210),
    [anon_sym_constexpr] = ACTIONS(1210),
    [anon_sym_volatile] = ACTIONS(1210),
    [anon_sym_restrict] = ACTIONS(1210),
    [anon_sym___restrict__] = ACTIONS(1210),
    [anon_sym__Atomic] = ACTIONS(1210),
    [anon_sym__Noreturn] = ACTIONS(1210),
    [anon_sym_noreturn] = ACTIONS(1210),
    [anon_sym__Nonnull] = ACTIONS(1210),
    [anon_sym_alignas] = ACTIONS(1210),
    [anon_sym__Alignas] = ACTIONS(1210),
    [sym_primitive_type] = ACTIONS(1210),
    [anon_sym_enum] = ACTIONS(1210),
    [anon_sym_struct] = ACTIONS(1210),
    [anon_sym_union] = ACTIONS(1210),
    [anon_sym_if] = ACTIONS(1210),
    [anon_sym_else] = ACTIONS(1210),
    [anon_sym_switch] = ACTIONS(1210),
    [anon_sym_case] = ACTIONS(1210),
    [anon_sym_default] = ACTIONS(1210),
    [anon_sym_while] = ACTIONS(1210),
    [anon_sym_do] = ACTIONS(1210),
    [anon_sym_for] = ACTIONS(1210),
    [anon_sym_return] = ACTIONS(1210),
    [anon_sym_break] = ACTIONS(1210),
    [anon_sym_continue] = ACTIONS(1210),
    [anon_sym_goto] = ACTIONS(1210),
    [anon_sym___try] = ACTIONS(1210),
    [anon_sym___leave] = ACTIONS(1210),
    [anon_sym_DASH_DASH] = ACTIONS(1212),
    [anon_sym_PLUS_PLUS] = ACTIONS(1212),
    [anon_sym_sizeof] = ACTIONS(1210),
    [anon_sym___alignof__] = ACTIONS(1210),
    [anon_sym___alignof] = ACTIONS(1210),
    [anon_sym__alignof] = ACTIONS(1210),
    [anon_sym_alignof] = ACTIONS(1210),
    [anon_sym__Alignof] = ACTIONS(1210),
    [anon_sym_offsetof] = ACTIONS(1210),
    [anon_sym__Generic] = ACTIONS(1210),
    [anon_sym_asm] = ACTIONS(1210),
    [anon_sym___asm__] = ACTIONS(1210),
    [anon_sym___asm] = ACTIONS(1210),
    [sym_number_literal] = ACTIONS(1212),
    [anon_sym_L_SQUOTE] = ACTIONS(1212),
    [anon_sym_u_SQUOTE] = ACTIONS(1212),
    [anon_sym_U_SQUOTE] = ACTIONS(1212),
    [anon_sym_u8_SQUOTE] = ACTIONS(1212),
    [anon_sym_SQUOTE] = ACTIONS(1212),
    [anon_sym_L_DQUOTE] = ACTIONS(1212),
    [anon_sym_u_DQUOTE] = ACTIONS(1212),
    [anon_sym_U_DQUOTE] = ACTIONS(1212),
    [anon_sym_u8_DQUOTE] = ACTIONS(1212),
    [anon_sym_DQUOTE] = ACTIONS(1212),
    [sym_true] = ACTIONS(1210),
    [sym_false] = ACTIONS(1210),
    [anon_sym_NULL] = ACTIONS(1210),
    [anon_sym_nullptr] = ACTIONS(1210),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(105)] = {
    [sym_identifier] = ACTIONS(1214),
    [aux_sym_preproc_include_token1] = ACTIONS(1214),
    [aux_sym_preproc_def_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token2] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1214),
    [aux_sym_preproc_else_token1] = ACTIONS(1214),
    [aux_sym_preproc_elif_token1] = ACTIONS(1214),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1214),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1214),
    [sym_preproc_directive] = ACTIONS(1214),
    [anon_sym_LPAREN2] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_TILDE] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1214),
    [anon_sym_PLUS] = ACTIONS(1214),
    [anon_sym_STAR] = ACTIONS(1216),
    [anon_sym_AMP] = ACTIONS(1216),
    [anon_sym_SEMI] = ACTIONS(1216),
    [anon_sym___extension__] = ACTIONS(1214),
    [anon_sym_typedef] = ACTIONS(1214),
    [anon_sym_extern] = ACTIONS(1214),
    [anon_sym___attribute__] = ACTIONS(1214),
    [anon_sym___attribute] = ACTIONS(1214),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1216),
    [anon_sym___declspec] = ACTIONS(1214),
    [anon_sym___cdecl] = ACTIONS(1214),
    [anon_sym___clrcall] = ACTIONS(1214),
    [anon_sym___stdcall] = ACTIONS(1214),
    [anon_sym___fastcall] = ACTIONS(1214),
    [anon_sym___thiscall] = ACTIONS(1214),
    [anon_sym___vectorcall] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_signed] = ACTIONS(1214),
    [anon_sym_unsigned] = ACTIONS(1214),
    [anon_sym_long] = ACTIONS(1214),
    [anon_sym_short] = ACTIONS(1214),
    [anon_sym_static] = ACTIONS(1214),
    [anon_sym_auto] = ACTIONS(1214),
    [anon_sym_register] = ACTIONS(1214),
    [anon_sym_inline] = ACTIONS(1214),
    [anon_sym___inline] = ACTIONS(1214),
    [anon_sym___inline__] = ACTIONS(1214),
    [anon_sym___forceinline] = ACTIONS(1214),
    [anon_sym_thread_local] = ACTIONS(1214),
    [anon_sym___thread] = ACTIONS(1214),
    [anon_sym_const] = ACTIONS(1214),
    [anon_sym_constexpr] = ACTIONS(1214),
    [anon_sym_volatile] = ACTIONS(1214),
    [anon_sym_restrict] = ACTIONS(1214),
    [anon_sym___restrict__] = ACTIONS(1214),
    [anon_sym__Atomic] = ACTIONS(1214),
    [anon_sym__Noreturn] = ACTIONS(1214),
    [anon_sym_noreturn] = ACTIONS(1214),
    [anon_sym__Nonnull] = ACTIONS(1214),
    [anon_sym_alignas] = ACTIONS(1214),
    [anon_sym__Alignas] = ACTIONS(1214),
    [sym_primitive_type] = ACTIONS(1214),
    [anon_sym_enum] = ACTIONS(1214),
    [anon_sym_struct] = ACTIONS(1214),
    [anon_sym_union] = ACTIONS(1214),
    [anon_sym_if] = ACTIONS(1214),
    [anon_sym_else] = ACTIONS(1214),
    [anon_sym_switch] = ACTIONS(1214),
    [anon_sym_case] = ACTIONS(1214),
    [anon_sym_default] = ACTIONS(1214),
    [anon_sym_while] = ACTIONS(1214),
    [anon_sym_do] = ACTIONS(1214),
    [anon_sym_for] = ACTIONS(1214),
    [anon_sym_return] = ACTIONS(1214),
    [anon_sym_break] = ACTIONS(1214),
    [anon_sym_continue] = ACTIONS(1214),
    [anon_sym_goto] = ACTIONS(1214),
    [anon_sym___try] = ACTIONS(1214),
    [anon_sym___leave] = ACTIONS(1214),
    [anon_sym_DASH_DASH] = ACTIONS(1216),
    [anon_sym_PLUS_PLUS] = ACTIONS(1216),
    [anon_sym_sizeof] = ACTIONS(1214),
    [anon_sym___alignof__] = ACTIONS(1214),
    [anon_sym___alignof] = ACTIONS(1214),
    [anon_sym__alignof] = ACTIONS(1214),
    [anon_sym_alignof] = ACTIONS(1214),
    [anon_sym__Alignof] = ACTIONS(1214),
    [anon_sym_offsetof] = ACTIONS(1214),
    [anon_sym__Generic] = ACTIONS(1214),
    [anon_sym_asm] = ACTIONS(1214),
    [anon_sym___asm__] = ACTIONS(1214),
    [anon_sym___asm] = ACTIONS(1214),
    [sym_number_literal] = ACTIONS(1216),
    [anon_sym_L_SQUOTE] = ACTIONS(1216),
    [anon_sym_u_SQUOTE] = ACTIONS(1216),
    [anon_sym_U_SQUOTE] = ACTIONS(1216),
    [anon_sym_u8_SQUOTE] = ACTIONS(1216),
    [anon_sym_SQUOTE] = ACTIONS(1216),
    [anon_sym_L_DQUOTE] = ACTIONS(1216),
    [anon_sym_u_DQUOTE] = ACTIONS(1216),
    [anon_sym_U_DQUOTE] = ACTIONS(1216),
    [anon_sym_u8_DQUOTE] = ACTIONS(1216),
    [anon_sym_DQUOTE] = ACTIONS(1216),
    [sym_true] = ACTIONS(1214),
    [sym_false] = ACTIONS(1214),
    [anon_sym_NULL] = ACTIONS(1214),
    [anon_sym_nullptr] = ACTIONS(1214),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(106)] = {
    [sym_identifier] = ACTIONS(1218),
    [aux_sym_preproc_include_token1] = ACTIONS(1218),
    [aux_sym_preproc_def_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token2] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1218),
    [aux_sym_preproc_else_token1] = ACTIONS(1218),
    [aux_sym_preproc_elif_token1] = ACTIONS(1218),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1218),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1218),
    [sym_preproc_directive] = ACTIONS(1218),
    [anon_sym_LPAREN2] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_TILDE] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1218),
    [anon_sym_PLUS] = ACTIONS(1218),
    [anon_sym_STAR] = ACTIONS(1220),
    [anon_sym_AMP] = ACTIONS(1220),
    [anon_sym_SEMI] = ACTIONS(1220),
    [anon_sym___extension__] = ACTIONS(1218),
    [anon_sym_typedef] = ACTIONS(1218),
    [anon_sym_extern] = ACTIONS(1218),
    [anon_sym___attribute__] = ACTIONS(1218),
    [anon_sym___attribute] = ACTIONS(1218),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1220),
    [anon_sym___declspec] = ACTIONS(1218),
    [anon_sym___cdecl] = ACTIONS(1218),
    [anon_sym___clrcall] = ACTIONS(1218),
    [anon_sym___stdcall] = ACTIONS(1218),
    [anon_sym___fastcall] = ACTIONS(1218),
    [anon_sym___thiscall] = ACTIONS(1218),
    [anon_sym___vectorcall] = ACTIONS(1218),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_signed] = ACTIONS(1218),
    [anon_sym_unsigned] = ACTIONS(1218),
    [anon_sym_long] = ACTIONS(1218),
    [anon_sym_short] = ACTIONS(1218),
    [anon_sym_static] = ACTIONS(1218),
    [anon_sym_auto] = ACTIONS(1218),
    [anon_sym_register] = ACTIONS(1218),
    [anon_sym_inline] = ACTIONS(1218),
    [anon_sym___inline] = ACTIONS(1218),
    [anon_sym___inline__] = ACTIONS(1218),
    [anon_sym___forceinline] = ACTIONS(1218),
    [anon_sym_thread_local] = ACTIONS(1218),
    [anon_sym___thread] = ACTIONS(1218),
    [anon_sym_const] = ACTIONS(1218),
    [anon_sym_constexpr] = ACTIONS(1218),
    [anon_sym_volatile] = ACTIONS(1218),
    [anon_sym_restrict] = ACTIONS(1218),
    [anon_sym___restrict__] = ACTIONS(1218),
    [anon_sym__Atomic] = ACTIONS(1218),
    [anon_sym__Noreturn] = ACTIONS(1218),
    [anon_sym_noreturn] = ACTIONS(1218),
    [anon_sym__Nonnull] = ACTIONS(1218),
    [anon_sym_alignas] = ACTIONS(1218),
    [anon_sym__Alignas] = ACTIONS(1218),
    [sym_primitive_type] = ACTIONS(1218),
    [anon_sym_enum] = ACTIONS(1218),
    [anon_sym_struct] = ACTIONS(1218),
    [anon_sym_union] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1218),
    [anon_sym_else] = ACTIONS(1218),
    [anon_sym_switch] = ACTIONS(1218),
    [anon_sym_case] = ACTIONS(1218),
    [anon_sym_default] = ACTIONS(1218),
    [anon_sym_while] = ACTIONS(1218),
    [anon_sym_do] = ACTIONS(1218),
    [anon_sym_for] = ACTIONS(1218),
    [anon_sym_return] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1218),
    [anon_sym_continue] = ACTIONS(1218),
    [anon_sym_goto] = ACTIONS(1218),
    [anon_sym___try] = ACTIONS(1218),
    [anon_sym___leave] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_PLUS_PLUS] = ACTIONS(1220),
    [anon_sym_sizeof] = ACTIONS(1218),
    [anon_sym___alignof__] = ACTIONS(1218),
    [anon_sym___alignof] = ACTIONS(1218),
    [anon_sym__alignof] = ACTIONS(1218),
    [anon_sym_alignof] = ACTIONS(1218),
    [anon_sym__Alignof] = ACTIONS(1218),
    [anon_sym_offsetof] = ACTIONS(1218),
    [anon_sym__Generic] = ACTIONS(1218),
    [anon_sym_asm] = ACTIONS(1218),
    [anon_sym___asm__] = ACTIONS(1218),
    [anon_sym___asm] = ACTIONS(1218),
    [sym_number_literal] = ACTIONS(1220),
    [anon_sym_L_SQUOTE] = ACTIONS(1220),
    [anon_sym_u_SQUOTE] = ACTIONS(1220),
    [anon_sym_U_SQUOTE] = ACTIONS(1220),
    [anon_sym_u8_SQUOTE] = ACTIONS(1220),
    [anon_sym_SQUOTE] = ACTIONS(1220),
    [anon_sym_L_DQUOTE] = ACTIONS(1220),
    [anon_sym_u_DQUOTE] = ACTIONS(1220),
    [anon_sym_U_DQUOTE] = ACTIONS(1220),
    [anon_sym_u8_DQUOTE] = ACTIONS(1220),
    [anon_sym_DQUOTE] = ACTIONS(1220),
    [sym_true] = ACTIONS(1218),
    [sym_false] = ACTIONS(1218),
    [anon_sym_NULL] = ACTIONS(1218),
    [anon_sym_nullptr] = ACTIONS(1218),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(107)] = {
    [sym_identifier] = ACTIONS(1222),
    [aux_sym_preproc_include_token1] = ACTIONS(1222),
    [aux_sym_preproc_def_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token2] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1222),
    [aux_sym_preproc_else_token1] = ACTIONS(1222),
    [aux_sym_preproc_elif_token1] = ACTIONS(1222),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1222),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1222),
    [sym_preproc_directive] = ACTIONS(1222),
    [anon_sym_LPAREN2] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1222),
    [anon_sym_PLUS] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym___extension__] = ACTIONS(1222),
    [anon_sym_typedef] = ACTIONS(1222),
    [anon_sym_extern] = ACTIONS(1222),
    [anon_sym___attribute__] = ACTIONS(1222),
    [anon_sym___attribute] = ACTIONS(1222),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1224),
    [anon_sym___declspec] = ACTIONS(1222),
    [anon_sym___cdecl] = ACTIONS(1222),
    [anon_sym___clrcall] = ACTIONS(1222),
    [anon_sym___stdcall] = ACTIONS(1222),
    [anon_sym___fastcall] = ACTIONS(1222),
    [anon_sym___thiscall] = ACTIONS(1222),
    [anon_sym___vectorcall] = ACTIONS(1222),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_signed] = ACTIONS(1222),
    [anon_sym_unsigned] = ACTIONS(1222),
    [anon_sym_long] = ACTIONS(1222),
    [anon_sym_short] = ACTIONS(1222),
    [anon_sym_static] = ACTIONS(1222),
    [anon_sym_auto] = ACTIONS(1222),
    [anon_sym_register] = ACTIONS(1222),
    [anon_sym_inline] = ACTIONS(1222),
    [anon_sym___inline] = ACTIONS(1222),
    [anon_sym___inline__] = ACTIONS(1222),
    [anon_sym___forceinline] = ACTIONS(1222),
    [anon_sym_thread_local] = ACTIONS(1222),
    [anon_sym___thread] = ACTIONS(1222),
    [anon_sym_const] = ACTIONS(1222),
    [anon_sym_constexpr] = ACTIONS(1222),
    [anon_sym_volatile] = ACTIONS(1222),
    [anon_sym_restrict] = ACTIONS(1222),
    [anon_sym___restrict__] = ACTIONS(1222),
    [anon_sym__Atomic] = ACTIONS(1222),
    [anon_sym__Noreturn] = ACTIONS(1222),
    [anon_sym_noreturn] = ACTIONS(1222),
    [anon_sym__Nonnull] = ACTIONS(1222),
    [anon_sym_alignas] = ACTIONS(1222),
    [anon_sym__Alignas] = ACTIONS(1222),
    [sym_primitive_type] = ACTIONS(1222),
    [anon_sym_enum] = ACTIONS(1222),
    [anon_sym_struct] = ACTIONS(1222),
    [anon_sym_union] = ACTIONS(1222),
    [anon_sym_if] = ACTIONS(1222),
    [anon_sym_else] = ACTIONS(1222),
    [anon_sym_switch] = ACTIONS(1222),
    [anon_sym_case] = ACTIONS(1222),
    [anon_sym_default] = ACTIONS(1222),
    [anon_sym_while] = ACTIONS(1222),
    [anon_sym_do] = ACTIONS(1222),
    [anon_sym_for] = ACTIONS(1222),
    [anon_sym_return] = ACTIONS(1222),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_goto] = ACTIONS(1222),
    [anon_sym___try] = ACTIONS(1222),
    [anon_sym___leave] = ACTIONS(1222),
    [anon_sym_DASH_DASH] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1224),
    [anon_sym_sizeof] = ACTIONS(1222),
    [anon_sym___alignof__] = ACTIONS(1222),
    [anon_sym___alignof] = ACTIONS(1222),
    [anon_sym__alignof] = ACTIONS(1222),
    [anon_sym_alignof] = ACTIONS(1222),
    [anon_sym__Alignof] = ACTIONS(1222),
    [anon_sym_offsetof] = ACTIONS(1222),
    [anon_sym__Generic] = ACTIONS(1222),
    [anon_sym_asm] = ACTIONS(1222),
    [anon_sym___asm__] = ACTIONS(1222),
    [anon_sym___asm] = ACTIONS(1222),
    [sym_number_literal] = ACTIONS(1224),
    [anon_sym_L_SQUOTE] = ACTIONS(1224),
    [anon_sym_u_SQUOTE] = ACTIONS(1224),
    [anon_sym_U_SQUOTE] = ACTIONS(1224),
    [anon_sym_u8_SQUOTE] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_L_DQUOTE] = ACTIONS(1224),
    [anon_sym_u_DQUOTE] = ACTIONS(1224),
    [anon_sym_U_DQUOTE] = ACTIONS(1224),
    [anon_sym_u8_DQUOTE] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [sym_true] = ACTIONS(1222),
    [sym_false] = ACTIONS(1222),
    [anon_sym_NULL] = ACTIONS(1222),
    [anon_sym_nullptr] = ACTIONS(1222),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(108)] = {
    [sym_identifier] = ACTIONS(1226),
    [aux_sym_preproc_include_token1] = ACTIONS(1226),
    [aux_sym_preproc_def_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token2] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1226),
    [aux_sym_preproc_else_token1] = ACTIONS(1226),
    [aux_sym_preproc_elif_token1] = ACTIONS(1226),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1226),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1226),
    [sym_preproc_directive] = ACTIONS(1226),
    [anon_sym_LPAREN2] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1226),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym___extension__] = ACTIONS(1226),
    [anon_sym_typedef] = ACTIONS(1226),
    [anon_sym_extern] = ACTIONS(1226),
    [anon_sym___attribute__] = ACTIONS(1226),
    [anon_sym___attribute] = ACTIONS(1226),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1228),
    [anon_sym___declspec] = ACTIONS(1226),
    [anon_sym___cdecl] = ACTIONS(1226),
    [anon_sym___clrcall] = ACTIONS(1226),
    [anon_sym___stdcall] = ACTIONS(1226),
    [anon_sym___fastcall] = ACTIONS(1226),
    [anon_sym___thiscall] = ACTIONS(1226),
    [anon_sym___vectorcall] = ACTIONS(1226),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_signed] = ACTIONS(1226),
    [anon_sym_unsigned] = ACTIONS(1226),
    [anon_sym_long] = ACTIONS(1226),
    [anon_sym_short] = ACTIONS(1226),
    [anon_sym_static] = ACTIONS(1226),
    [anon_sym_auto] = ACTIONS(1226),
    [anon_sym_register] = ACTIONS(1226),
    [anon_sym_inline] = ACTIONS(1226),
    [anon_sym___inline] = ACTIONS(1226),
    [anon_sym___inline__] = ACTIONS(1226),
    [anon_sym___forceinline] = ACTIONS(1226),
    [anon_sym_thread_local] = ACTIONS(1226),
    [anon_sym___thread] = ACTIONS(1226),
    [anon_sym_const] = ACTIONS(1226),
    [anon_sym_constexpr] = ACTIONS(1226),
    [anon_sym_volatile] = ACTIONS(1226),
    [anon_sym_restrict] = ACTIONS(1226),
    [anon_sym___restrict__] = ACTIONS(1226),
    [anon_sym__Atomic] = ACTIONS(1226),
    [anon_sym__Noreturn] = ACTIONS(1226),
    [anon_sym_noreturn] = ACTIONS(1226),
    [anon_sym__Nonnull] = ACTIONS(1226),
    [anon_sym_alignas] = ACTIONS(1226),
    [anon_sym__Alignas] = ACTIONS(1226),
    [sym_primitive_type] = ACTIONS(1226),
    [anon_sym_enum] = ACTIONS(1226),
    [anon_sym_struct] = ACTIONS(1226),
    [anon_sym_union] = ACTIONS(1226),
    [anon_sym_if] = ACTIONS(1226),
    [anon_sym_else] = ACTIONS(1226),
    [anon_sym_switch] = ACTIONS(1226),
    [anon_sym_case] = ACTIONS(1226),
    [anon_sym_default] = ACTIONS(1226),
    [anon_sym_while] = ACTIONS(1226),
    [anon_sym_do] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(1226),
    [anon_sym_return] = ACTIONS(1226),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_goto] = ACTIONS(1226),
    [anon_sym___try] = ACTIONS(1226),
    [anon_sym___leave] = ACTIONS(1226),
    [anon_sym_DASH_DASH] = ACTIONS(1228),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [anon_sym_sizeof] = ACTIONS(1226),
    [anon_sym___alignof__] = ACTIONS(1226),
    [anon_sym___alignof] = ACTIONS(1226),
    [anon_sym__alignof] = ACTIONS(1226),
    [anon_sym_alignof] = ACTIONS(1226),
    [anon_sym__Alignof] = ACTIONS(1226),
    [anon_sym_offsetof] = ACTIONS(1226),
    [anon_sym__Generic] = ACTIONS(1226),
    [anon_sym_asm] = ACTIONS(1226),
    [anon_sym___asm__] = ACTIONS(1226),
    [anon_sym___asm] = ACTIONS(1226),
    [sym_number_literal] = ACTIONS(1228),
    [anon_sym_L_SQUOTE] = ACTIONS(1228),
    [anon_sym_u_SQUOTE] = ACTIONS(1228),
    [anon_sym_U_SQUOTE] = ACTIONS(1228),
    [anon_sym_u8_SQUOTE] = ACTIONS(1228),
    [anon_sym_SQUOTE] = ACTIONS(1228),
    [anon_sym_L_DQUOTE] = ACTIONS(1228),
    [anon_sym_u_DQUOTE] = ACTIONS(1228),
    [anon_sym_U_DQUOTE] = ACTIONS(1228),
    [anon_sym_u8_DQUOTE] = ACTIONS(1228),
    [anon_sym_DQUOTE] = ACTIONS(1228),
    [sym_true] = ACTIONS(1226),
    [sym_false] = ACTIONS(1226),
    [anon_sym_NULL] = ACTIONS(1226),
    [anon_sym_nullptr] = ACTIONS(1226),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(109)] = {
    [sym_identifier] = ACTIONS(1230),
    [aux_sym_preproc_include_token1] = ACTIONS(1230),
    [aux_sym_preproc_def_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token2] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1230),
    [aux_sym_preproc_else_token1] = ACTIONS(1230),
    [aux_sym_preproc_elif_token1] = ACTIONS(1230),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1230),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1230),
    [sym_preproc_directive] = ACTIONS(1230),
    [anon_sym_LPAREN2] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1230),
    [anon_sym_PLUS] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym___extension__] = ACTIONS(1230),
    [anon_sym_typedef] = ACTIONS(1230),
    [anon_sym_extern] = ACTIONS(1230),
    [anon_sym___attribute__] = ACTIONS(1230),
    [anon_sym___attribute] = ACTIONS(1230),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1232),
    [anon_sym___declspec] = ACTIONS(1230),
    [anon_sym___cdecl] = ACTIONS(1230),
    [anon_sym___clrcall] = ACTIONS(1230),
    [anon_sym___stdcall] = ACTIONS(1230),
    [anon_sym___fastcall] = ACTIONS(1230),
    [anon_sym___thiscall] = ACTIONS(1230),
    [anon_sym___vectorcall] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_signed] = ACTIONS(1230),
    [anon_sym_unsigned] = ACTIONS(1230),
    [anon_sym_long] = ACTIONS(1230),
    [anon_sym_short] = ACTIONS(1230),
    [anon_sym_static] = ACTIONS(1230),
    [anon_sym_auto] = ACTIONS(1230),
    [anon_sym_register] = ACTIONS(1230),
    [anon_sym_inline] = ACTIONS(1230),
    [anon_sym___inline] = ACTIONS(1230),
    [anon_sym___inline__] = ACTIONS(1230),
    [anon_sym___forceinline] = ACTIONS(1230),
    [anon_sym_thread_local] = ACTIONS(1230),
    [anon_sym___thread] = ACTIONS(1230),
    [anon_sym_const] = ACTIONS(1230),
    [anon_sym_constexpr] = ACTIONS(1230),
    [anon_sym_volatile] = ACTIONS(1230),
    [anon_sym_restrict] = ACTIONS(1230),
    [anon_sym___restrict__] = ACTIONS(1230),
    [anon_sym__Atomic] = ACTIONS(1230),
    [anon_sym__Noreturn] = ACTIONS(1230),
    [anon_sym_noreturn] = ACTIONS(1230),
    [anon_sym__Nonnull] = ACTIONS(1230),
    [anon_sym_alignas] = ACTIONS(1230),
    [anon_sym__Alignas] = ACTIONS(1230),
    [sym_primitive_type] = ACTIONS(1230),
    [anon_sym_enum] = ACTIONS(1230),
    [anon_sym_struct] = ACTIONS(1230),
    [anon_sym_union] = ACTIONS(1230),
    [anon_sym_if] = ACTIONS(1230),
    [anon_sym_else] = ACTIONS(1230),
    [anon_sym_switch] = ACTIONS(1230),
    [anon_sym_case] = ACTIONS(1230),
    [anon_sym_default] = ACTIONS(1230),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_break] = ACTIONS(1230),
    [anon_sym_continue] = ACTIONS(1230),
    [anon_sym_goto] = ACTIONS(1230),
    [anon_sym___try] = ACTIONS(1230),
    [anon_sym___leave] = ACTIONS(1230),
    [anon_sym_DASH_DASH] = ACTIONS(1232),
    [anon_sym_PLUS_PLUS] = ACTIONS(1232),
    [anon_sym_sizeof] = ACTIONS(1230),
    [anon_sym___alignof__] = ACTIONS(1230),
    [anon_sym___alignof] = ACTIONS(1230),
    [anon_sym__alignof] = ACTIONS(1230),
    [anon_sym_alignof] = ACTIONS(1230),
    [anon_sym__Alignof] = ACTIONS(1230),
    [anon_sym_offsetof] = ACTIONS(1230),
    [anon_sym__Generic] = ACTIONS(1230),
    [anon_sym_asm] = ACTIONS(1230),
    [anon_sym___asm__] = ACTIONS(1230),
    [anon_sym___asm] = ACTIONS(1230),
    [sym_number_literal] = ACTIONS(1232),
    [anon_sym_L_SQUOTE] = ACTIONS(1232),
    [anon_sym_u_SQUOTE] = ACTIONS(1232),
    [anon_sym_U_SQUOTE] = ACTIONS(1232),
    [anon_sym_u8_SQUOTE] = ACTIONS(1232),
    [anon_sym_SQUOTE] = ACTIONS(1232),
    [anon_sym_L_DQUOTE] = ACTIONS(1232),
    [anon_sym_u_DQUOTE] = ACTIONS(1232),
    [anon_sym_U_DQUOTE] = ACTIONS(1232),
    [anon_sym_u8_DQUOTE] = ACTIONS(1232),
    [anon_sym_DQUOTE] = ACTIONS(1232),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [anon_sym_NULL] = ACTIONS(1230),
    [anon_sym_nullptr] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(110)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token2] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [aux_sym_preproc_else_token1] = ACTIONS(1234),
    [aux_sym_preproc_elif_token1] = ACTIONS(1234),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(111)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token2] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [aux_sym_preproc_else_token1] = ACTIONS(1234),
    [aux_sym_preproc_elif_token1] = ACTIONS(1234),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(112)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token2] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [aux_sym_preproc_else_token1] = ACTIONS(1238),
    [aux_sym_preproc_elif_token1] = ACTIONS(1238),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(113)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token2] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [aux_sym_preproc_else_token1] = ACTIONS(1238),
    [aux_sym_preproc_elif_token1] = ACTIONS(1238),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(114)] = {
    [sym_identifier] = ACTIONS(1242),
    [aux_sym_preproc_include_token1] = ACTIONS(1242),
    [aux_sym_preproc_def_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token2] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1242),
    [aux_sym_preproc_else_token1] = ACTIONS(1242),
    [aux_sym_preproc_elif_token1] = ACTIONS(1242),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1242),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1242),
    [sym_preproc_directive] = ACTIONS(1242),
    [anon_sym_LPAREN2] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_TILDE] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1242),
    [anon_sym_PLUS] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(1244),
    [anon_sym_AMP] = ACTIONS(1244),
    [anon_sym_SEMI] = ACTIONS(1244),
    [anon_sym___extension__] = ACTIONS(1242),
    [anon_sym_typedef] = ACTIONS(1242),
    [anon_sym_extern] = ACTIONS(1242),
    [anon_sym___attribute__] = ACTIONS(1242),
    [anon_sym___attribute] = ACTIONS(1242),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1244),
    [anon_sym___declspec] = ACTIONS(1242),
    [anon_sym___cdecl] = ACTIONS(1242),
    [anon_sym___clrcall] = ACTIONS(1242),
    [anon_sym___stdcall] = ACTIONS(1242),
    [anon_sym___fastcall] = ACTIONS(1242),
    [anon_sym___thiscall] = ACTIONS(1242),
    [anon_sym___vectorcall] = ACTIONS(1242),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_signed] = ACTIONS(1242),
    [anon_sym_unsigned] = ACTIONS(1242),
    [anon_sym_long] = ACTIONS(1242),
    [anon_sym_short] = ACTIONS(1242),
    [anon_sym_static] = ACTIONS(1242),
    [anon_sym_auto] = ACTIONS(1242),
    [anon_sym_register] = ACTIONS(1242),
    [anon_sym_inline] = ACTIONS(1242),
    [anon_sym___inline] = ACTIONS(1242),
    [anon_sym___inline__] = ACTIONS(1242),
    [anon_sym___forceinline] = ACTIONS(1242),
    [anon_sym_thread_local] = ACTIONS(1242),
    [anon_sym___thread] = ACTIONS(1242),
    [anon_sym_const] = ACTIONS(1242),
    [anon_sym_constexpr] = ACTIONS(1242),
    [anon_sym_volatile] = ACTIONS(1242),
    [anon_sym_restrict] = ACTIONS(1242),
    [anon_sym___restrict__] = ACTIONS(1242),
    [anon_sym__Atomic] = ACTIONS(1242),
    [anon_sym__Noreturn] = ACTIONS(1242),
    [anon_sym_noreturn] = ACTIONS(1242),
    [anon_sym__Nonnull] = ACTIONS(1242),
    [anon_sym_alignas] = ACTIONS(1242),
    [anon_sym__Alignas] = ACTIONS(1242),
    [sym_primitive_type] = ACTIONS(1242),
    [anon_sym_enum] = ACTIONS(1242),
    [anon_sym_struct] = ACTIONS(1242),
    [anon_sym_union] = ACTIONS(1242),
    [anon_sym_if] = ACTIONS(1242),
    [anon_sym_else] = ACTIONS(1242),
    [anon_sym_switch] = ACTIONS(1242),
    [anon_sym_case] = ACTIONS(1242),
    [anon_sym_default] = ACTIONS(1242),
    [anon_sym_while] = ACTIONS(1242),
    [anon_sym_do] = ACTIONS(1242),
    [anon_sym_for] = ACTIONS(1242),
    [anon_sym_return] = ACTIONS(1242),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_goto] = ACTIONS(1242),
    [anon_sym___try] = ACTIONS(1242),
    [anon_sym___leave] = ACTIONS(1242),
    [anon_sym_DASH_DASH] = ACTIONS(1244),
    [anon_sym_PLUS_PLUS] = ACTIONS(1244),
    [anon_sym_sizeof] = ACTIONS(1242),
    [anon_sym___alignof__] = ACTIONS(1242),
    [anon_sym___alignof] = ACTIONS(1242),
    [anon_sym__alignof] = ACTIONS(1242),
    [anon_sym_alignof] = ACTIONS(1242),
    [anon_sym__Alignof] = ACTIONS(1242),
    [anon_sym_offsetof] = ACTIONS(1242),
    [anon_sym__Generic] = ACTIONS(1242),
    [anon_sym_asm] = ACTIONS(1242),
    [anon_sym___asm__] = ACTIONS(1242),
    [anon_sym___asm] = ACTIONS(1242),
    [sym_number_literal] = ACTIONS(1244),
    [anon_sym_L_SQUOTE] = ACTIONS(1244),
    [anon_sym_u_SQUOTE] = ACTIONS(1244),
    [anon_sym_U_SQUOTE] = ACTIONS(1244),
    [anon_sym_u8_SQUOTE] = ACTIONS(1244),
    [anon_sym_SQUOTE] = ACTIONS(1244),
    [anon_sym_L_DQUOTE] = ACTIONS(1244),
    [anon_sym_u_DQUOTE] = ACTIONS(1244),
    [anon_sym_U_DQUOTE] = ACTIONS(1244),
    [anon_sym_u8_DQUOTE] = ACTIONS(1244),
    [anon_sym_DQUOTE] = ACTIONS(1244),
    [sym_true] = ACTIONS(1242),
    [sym_false] = ACTIONS(1242),
    [anon_sym_NULL] = ACTIONS(1242),
    [anon_sym_nullptr] = ACTIONS(1242),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(115)] = {
    [sym_identifier] = ACTIONS(1246),
    [aux_sym_preproc_include_token1] = ACTIONS(1246),
    [aux_sym_preproc_def_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token2] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1246),
    [aux_sym_preproc_else_token1] = ACTIONS(1246),
    [aux_sym_preproc_elif_token1] = ACTIONS(1246),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1246),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1246),
    [sym_preproc_directive] = ACTIONS(1246),
    [anon_sym_LPAREN2] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_TILDE] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1246),
    [anon_sym_PLUS] = ACTIONS(1246),
    [anon_sym_STAR] = ACTIONS(1248),
    [anon_sym_AMP] = ACTIONS(1248),
    [anon_sym_SEMI] = ACTIONS(1248),
    [anon_sym___extension__] = ACTIONS(1246),
    [anon_sym_typedef] = ACTIONS(1246),
    [anon_sym_extern] = ACTIONS(1246),
    [anon_sym___attribute__] = ACTIONS(1246),
    [anon_sym___attribute] = ACTIONS(1246),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1248),
    [anon_sym___declspec] = ACTIONS(1246),
    [anon_sym___cdecl] = ACTIONS(1246),
    [anon_sym___clrcall] = ACTIONS(1246),
    [anon_sym___stdcall] = ACTIONS(1246),
    [anon_sym___fastcall] = ACTIONS(1246),
    [anon_sym___thiscall] = ACTIONS(1246),
    [anon_sym___vectorcall] = ACTIONS(1246),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_signed] = ACTIONS(1246),
    [anon_sym_unsigned] = ACTIONS(1246),
    [anon_sym_long] = ACTIONS(1246),
    [anon_sym_short] = ACTIONS(1246),
    [anon_sym_static] = ACTIONS(1246),
    [anon_sym_auto] = ACTIONS(1246),
    [anon_sym_register] = ACTIONS(1246),
    [anon_sym_inline] = ACTIONS(1246),
    [anon_sym___inline] = ACTIONS(1246),
    [anon_sym___inline__] = ACTIONS(1246),
    [anon_sym___forceinline] = ACTIONS(1246),
    [anon_sym_thread_local] = ACTIONS(1246),
    [anon_sym___thread] = ACTIONS(1246),
    [anon_sym_const] = ACTIONS(1246),
    [anon_sym_constexpr] = ACTIONS(1246),
    [anon_sym_volatile] = ACTIONS(1246),
    [anon_sym_restrict] = ACTIONS(1246),
    [anon_sym___restrict__] = ACTIONS(1246),
    [anon_sym__Atomic] = ACTIONS(1246),
    [anon_sym__Noreturn] = ACTIONS(1246),
    [anon_sym_noreturn] = ACTIONS(1246),
    [anon_sym__Nonnull] = ACTIONS(1246),
    [anon_sym_alignas] = ACTIONS(1246),
    [anon_sym__Alignas] = ACTIONS(1246),
    [sym_primitive_type] = ACTIONS(1246),
    [anon_sym_enum] = ACTIONS(1246),
    [anon_sym_struct] = ACTIONS(1246),
    [anon_sym_union] = ACTIONS(1246),
    [anon_sym_if] = ACTIONS(1246),
    [anon_sym_else] = ACTIONS(1246),
    [anon_sym_switch] = ACTIONS(1246),
    [anon_sym_case] = ACTIONS(1246),
    [anon_sym_default] = ACTIONS(1246),
    [anon_sym_while] = ACTIONS(1246),
    [anon_sym_do] = ACTIONS(1246),
    [anon_sym_for] = ACTIONS(1246),
    [anon_sym_return] = ACTIONS(1246),
    [anon_sym_break] = ACTIONS(1246),
    [anon_sym_continue] = ACTIONS(1246),
    [anon_sym_goto] = ACTIONS(1246),
    [anon_sym___try] = ACTIONS(1246),
    [anon_sym___leave] = ACTIONS(1246),
    [anon_sym_DASH_DASH] = ACTIONS(1248),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [anon_sym_sizeof] = ACTIONS(1246),
    [anon_sym___alignof__] = ACTIONS(1246),
    [anon_sym___alignof] = ACTIONS(1246),
    [anon_sym__alignof] = ACTIONS(1246),
    [anon_sym_alignof] = ACTIONS(1246),
    [anon_sym__Alignof] = ACTIONS(1246),
    [anon_sym_offsetof] = ACTIONS(1246),
    [anon_sym__Generic] = ACTIONS(1246),
    [anon_sym_asm] = ACTIONS(1246),
    [anon_sym___asm__] = ACTIONS(1246),
    [anon_sym___asm] = ACTIONS(1246),
    [sym_number_literal] = ACTIONS(1248),
    [anon_sym_L_SQUOTE] = ACTIONS(1248),
    [anon_sym_u_SQUOTE] = ACTIONS(1248),
    [anon_sym_U_SQUOTE] = ACTIONS(1248),
    [anon_sym_u8_SQUOTE] = ACTIONS(1248),
    [anon_sym_SQUOTE] = ACTIONS(1248),
    [anon_sym_L_DQUOTE] = ACTIONS(1248),
    [anon_sym_u_DQUOTE] = ACTIONS(1248),
    [anon_sym_U_DQUOTE] = ACTIONS(1248),
    [anon_sym_u8_DQUOTE] = ACTIONS(1248),
    [anon_sym_DQUOTE] = ACTIONS(1248),
    [sym_true] = ACTIONS(1246),
    [sym_false] = ACTIONS(1246),
    [anon_sym_NULL] = ACTIONS(1246),
    [anon_sym_nullptr] = ACTIONS(1246),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(116)] = {
    [sym_identifier] = ACTIONS(1250),
    [aux_sym_preproc_include_token1] = ACTIONS(1250),
    [aux_sym_preproc_def_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token2] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1250),
    [aux_sym_preproc_else_token1] = ACTIONS(1250),
    [aux_sym_preproc_elif_token1] = ACTIONS(1250),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1250),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1250),
    [sym_preproc_directive] = ACTIONS(1250),
    [anon_sym_LPAREN2] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1250),
    [anon_sym_PLUS] = ACTIONS(1250),
    [anon_sym_STAR] = ACTIONS(1252),
    [anon_sym_AMP] = ACTIONS(1252),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym___extension__] = ACTIONS(1250),
    [anon_sym_typedef] = ACTIONS(1250),
    [anon_sym_extern] = ACTIONS(1250),
    [anon_sym___attribute__] = ACTIONS(1250),
    [anon_sym___attribute] = ACTIONS(1250),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1252),
    [anon_sym___declspec] = ACTIONS(1250),
    [anon_sym___cdecl] = ACTIONS(1250),
    [anon_sym___clrcall] = ACTIONS(1250),
    [anon_sym___stdcall] = ACTIONS(1250),
    [anon_sym___fastcall] = ACTIONS(1250),
    [anon_sym___thiscall] = ACTIONS(1250),
    [anon_sym___vectorcall] = ACTIONS(1250),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_signed] = ACTIONS(1250),
    [anon_sym_unsigned] = ACTIONS(1250),
    [anon_sym_long] = ACTIONS(1250),
    [anon_sym_short] = ACTIONS(1250),
    [anon_sym_static] = ACTIONS(1250),
    [anon_sym_auto] = ACTIONS(1250),
    [anon_sym_register] = ACTIONS(1250),
    [anon_sym_inline] = ACTIONS(1250),
    [anon_sym___inline] = ACTIONS(1250),
    [anon_sym___inline__] = ACTIONS(1250),
    [anon_sym___forceinline] = ACTIONS(1250),
    [anon_sym_thread_local] = ACTIONS(1250),
    [anon_sym___thread] = ACTIONS(1250),
    [anon_sym_const] = ACTIONS(1250),
    [anon_sym_constexpr] = ACTIONS(1250),
    [anon_sym_volatile] = ACTIONS(1250),
    [anon_sym_restrict] = ACTIONS(1250),
    [anon_sym___restrict__] = ACTIONS(1250),
    [anon_sym__Atomic] = ACTIONS(1250),
    [anon_sym__Noreturn] = ACTIONS(1250),
    [anon_sym_noreturn] = ACTIONS(1250),
    [anon_sym__Nonnull] = ACTIONS(1250),
    [anon_sym_alignas] = ACTIONS(1250),
    [anon_sym__Alignas] = ACTIONS(1250),
    [sym_primitive_type] = ACTIONS(1250),
    [anon_sym_enum] = ACTIONS(1250),
    [anon_sym_struct] = ACTIONS(1250),
    [anon_sym_union] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(1250),
    [anon_sym_else] = ACTIONS(1250),
    [anon_sym_switch] = ACTIONS(1250),
    [anon_sym_case] = ACTIONS(1250),
    [anon_sym_default] = ACTIONS(1250),
    [anon_sym_while] = ACTIONS(1250),
    [anon_sym_do] = ACTIONS(1250),
    [anon_sym_for] = ACTIONS(1250),
    [anon_sym_return] = ACTIONS(1250),
    [anon_sym_break] = ACTIONS(1250),
    [anon_sym_continue] = ACTIONS(1250),
    [anon_sym_goto] = ACTIONS(1250),
    [anon_sym___try] = ACTIONS(1250),
    [anon_sym___leave] = ACTIONS(1250),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_sizeof] = ACTIONS(1250),
    [anon_sym___alignof__] = ACTIONS(1250),
    [anon_sym___alignof] = ACTIONS(1250),
    [anon_sym__alignof] = ACTIONS(1250),
    [anon_sym_alignof] = ACTIONS(1250),
    [anon_sym__Alignof] = ACTIONS(1250),
    [anon_sym_offsetof] = ACTIONS(1250),
    [anon_sym__Generic] = ACTIONS(1250),
    [anon_sym_asm] = ACTIONS(1250),
    [anon_sym___asm__] = ACTIONS(1250),
    [anon_sym___asm] = ACTIONS(1250),
    [sym_number_literal] = ACTIONS(1252),
    [anon_sym_L_SQUOTE] = ACTIONS(1252),
    [anon_sym_u_SQUOTE] = ACTIONS(1252),
    [anon_sym_U_SQUOTE] = ACTIONS(1252),
    [anon_sym_u8_SQUOTE] = ACTIONS(1252),
    [anon_sym_SQUOTE] = ACTIONS(1252),
    [anon_sym_L_DQUOTE] = ACTIONS(1252),
    [anon_sym_u_DQUOTE] = ACTIONS(1252),
    [anon_sym_U_DQUOTE] = ACTIONS(1252),
    [anon_sym_u8_DQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1252),
    [sym_true] = ACTIONS(1250),
    [sym_false] = ACTIONS(1250),
    [anon_sym_NULL] = ACTIONS(1250),
    [anon_sym_nullptr] = ACTIONS(1250),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(117)] = {
    [sym_identifier] = ACTIONS(1254),
    [aux_sym_preproc_include_token1] = ACTIONS(1254),
    [aux_sym_preproc_def_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token2] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1254),
    [aux_sym_preproc_else_token1] = ACTIONS(1254),
    [aux_sym_preproc_elif_token1] = ACTIONS(1254),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1254),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1254),
    [sym_preproc_directive] = ACTIONS(1254),
    [anon_sym_LPAREN2] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_TILDE] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_STAR] = ACTIONS(1256),
    [anon_sym_AMP] = ACTIONS(1256),
    [anon_sym_SEMI] = ACTIONS(1256),
    [anon_sym___extension__] = ACTIONS(1254),
    [anon_sym_typedef] = ACTIONS(1254),
    [anon_sym_extern] = ACTIONS(1254),
    [anon_sym___attribute__] = ACTIONS(1254),
    [anon_sym___attribute] = ACTIONS(1254),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1256),
    [anon_sym___declspec] = ACTIONS(1254),
    [anon_sym___cdecl] = ACTIONS(1254),
    [anon_sym___clrcall] = ACTIONS(1254),
    [anon_sym___stdcall] = ACTIONS(1254),
    [anon_sym___fastcall] = ACTIONS(1254),
    [anon_sym___thiscall] = ACTIONS(1254),
    [anon_sym___vectorcall] = ACTIONS(1254),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_signed] = ACTIONS(1254),
    [anon_sym_unsigned] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_auto] = ACTIONS(1254),
    [anon_sym_register] = ACTIONS(1254),
    [anon_sym_inline] = ACTIONS(1254),
    [anon_sym___inline] = ACTIONS(1254),
    [anon_sym___inline__] = ACTIONS(1254),
    [anon_sym___forceinline] = ACTIONS(1254),
    [anon_sym_thread_local] = ACTIONS(1254),
    [anon_sym___thread] = ACTIONS(1254),
    [anon_sym_const] = ACTIONS(1254),
    [anon_sym_constexpr] = ACTIONS(1254),
    [anon_sym_volatile] = ACTIONS(1254),
    [anon_sym_restrict] = ACTIONS(1254),
    [anon_sym___restrict__] = ACTIONS(1254),
    [anon_sym__Atomic] = ACTIONS(1254),
    [anon_sym__Noreturn] = ACTIONS(1254),
    [anon_sym_noreturn] = ACTIONS(1254),
    [anon_sym__Nonnull] = ACTIONS(1254),
    [anon_sym_alignas] = ACTIONS(1254),
    [anon_sym__Alignas] = ACTIONS(1254),
    [sym_primitive_type] = ACTIONS(1254),
    [anon_sym_enum] = ACTIONS(1254),
    [anon_sym_struct] = ACTIONS(1254),
    [anon_sym_union] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_goto] = ACTIONS(1254),
    [anon_sym___try] = ACTIONS(1254),
    [anon_sym___leave] = ACTIONS(1254),
    [anon_sym_DASH_DASH] = ACTIONS(1256),
    [anon_sym_PLUS_PLUS] = ACTIONS(1256),
    [anon_sym_sizeof] = ACTIONS(1254),
    [anon_sym___alignof__] = ACTIONS(1254),
    [anon_sym___alignof] = ACTIONS(1254),
    [anon_sym__alignof] = ACTIONS(1254),
    [anon_sym_alignof] = ACTIONS(1254),
    [anon_sym__Alignof] = ACTIONS(1254),
    [anon_sym_offsetof] = ACTIONS(1254),
    [anon_sym__Generic] = ACTIONS(1254),
    [anon_sym_asm] = ACTIONS(1254),
    [anon_sym___asm__] = ACTIONS(1254),
    [anon_sym___asm] = ACTIONS(1254),
    [sym_number_literal] = ACTIONS(1256),
    [anon_sym_L_SQUOTE] = ACTIONS(1256),
    [anon_sym_u_SQUOTE] = ACTIONS(1256),
    [anon_sym_U_SQUOTE] = ACTIONS(1256),
    [anon_sym_u8_SQUOTE] = ACTIONS(1256),
    [anon_sym_SQUOTE] = ACTIONS(1256),
    [anon_sym_L_DQUOTE] = ACTIONS(1256),
    [anon_sym_u_DQUOTE] = ACTIONS(1256),
    [anon_sym_U_DQUOTE] = ACTIONS(1256),
    [anon_sym_u8_DQUOTE] = ACTIONS(1256),
    [anon_sym_DQUOTE] = ACTIONS(1256),
    [sym_true] = ACTIONS(1254),
    [sym_false] = ACTIONS(1254),
    [anon_sym_NULL] = ACTIONS(1254),
    [anon_sym_nullptr] = ACTIONS(1254),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(118)] = {
    [sym_identifier] = ACTIONS(1258),
    [aux_sym_preproc_include_token1] = ACTIONS(1258),
    [aux_sym_preproc_def_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token2] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1258),
    [aux_sym_preproc_else_token1] = ACTIONS(1258),
    [aux_sym_preproc_elif_token1] = ACTIONS(1258),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1258),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1258),
    [sym_preproc_directive] = ACTIONS(1258),
    [anon_sym_LPAREN2] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1258),
    [anon_sym_PLUS] = ACTIONS(1258),
    [anon_sym_STAR] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym___extension__] = ACTIONS(1258),
    [anon_sym_typedef] = ACTIONS(1258),
    [anon_sym_extern] = ACTIONS(1258),
    [anon_sym___attribute__] = ACTIONS(1258),
    [anon_sym___attribute] = ACTIONS(1258),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1260),
    [anon_sym___declspec] = ACTIONS(1258),
    [anon_sym___cdecl] = ACTIONS(1258),
    [anon_sym___clrcall] = ACTIONS(1258),
    [anon_sym___stdcall] = ACTIONS(1258),
    [anon_sym___fastcall] = ACTIONS(1258),
    [anon_sym___thiscall] = ACTIONS(1258),
    [anon_sym___vectorcall] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_signed] = ACTIONS(1258),
    [anon_sym_unsigned] = ACTIONS(1258),
    [anon_sym_long] = ACTIONS(1258),
    [anon_sym_short] = ACTIONS(1258),
    [anon_sym_static] = ACTIONS(1258),
    [anon_sym_auto] = ACTIONS(1258),
    [anon_sym_register] = ACTIONS(1258),
    [anon_sym_inline] = ACTIONS(1258),
    [anon_sym___inline] = ACTIONS(1258),
    [anon_sym___inline__] = ACTIONS(1258),
    [anon_sym___forceinline] = ACTIONS(1258),
    [anon_sym_thread_local] = ACTIONS(1258),
    [anon_sym___thread] = ACTIONS(1258),
    [anon_sym_const] = ACTIONS(1258),
    [anon_sym_constexpr] = ACTIONS(1258),
    [anon_sym_volatile] = ACTIONS(1258),
    [anon_sym_restrict] = ACTIONS(1258),
    [anon_sym___restrict__] = ACTIONS(1258),
    [anon_sym__Atomic] = ACTIONS(1258),
    [anon_sym__Noreturn] = ACTIONS(1258),
    [anon_sym_noreturn] = ACTIONS(1258),
    [anon_sym__Nonnull] = ACTIONS(1258),
    [anon_sym_alignas] = ACTIONS(1258),
    [anon_sym__Alignas] = ACTIONS(1258),
    [sym_primitive_type] = ACTIONS(1258),
    [anon_sym_enum] = ACTIONS(1258),
    [anon_sym_struct] = ACTIONS(1258),
    [anon_sym_union] = ACTIONS(1258),
    [anon_sym_if] = ACTIONS(1258),
    [anon_sym_else] = ACTIONS(1258),
    [anon_sym_switch] = ACTIONS(1258),
    [anon_sym_case] = ACTIONS(1258),
    [anon_sym_default] = ACTIONS(1258),
    [anon_sym_while] = ACTIONS(1258),
    [anon_sym_do] = ACTIONS(1258),
    [anon_sym_for] = ACTIONS(1258),
    [anon_sym_return] = ACTIONS(1258),
    [anon_sym_break] = ACTIONS(1258),
    [anon_sym_continue] = ACTIONS(1258),
    [anon_sym_goto] = ACTIONS(1258),
    [anon_sym___try] = ACTIONS(1258),
    [anon_sym___leave] = ACTIONS(1258),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_sizeof] = ACTIONS(1258),
    [anon_sym___alignof__] = ACTIONS(1258),
    [anon_sym___alignof] = ACTIONS(1258),
    [anon_sym__alignof] = ACTIONS(1258),
    [anon_sym_alignof] = ACTIONS(1258),
    [anon_sym__Alignof] = ACTIONS(1258),
    [anon_sym_offsetof] = ACTIONS(1258),
    [anon_sym__Generic] = ACTIONS(1258),
    [anon_sym_asm] = ACTIONS(1258),
    [anon_sym___asm__] = ACTIONS(1258),
    [anon_sym___asm] = ACTIONS(1258),
    [sym_number_literal] = ACTIONS(1260),
    [anon_sym_L_SQUOTE] = ACTIONS(1260),
    [anon_sym_u_SQUOTE] = ACTIONS(1260),
    [anon_sym_U_SQUOTE] = ACTIONS(1260),
    [anon_sym_u8_SQUOTE] = ACTIONS(1260),
    [anon_sym_SQUOTE] = ACTIONS(1260),
    [anon_sym_L_DQUOTE] = ACTIONS(1260),
    [anon_sym_u_DQUOTE] = ACTIONS(1260),
    [anon_sym_U_DQUOTE] = ACTIONS(1260),
    [anon_sym_u8_DQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1260),
    [sym_true] = ACTIONS(1258),
    [sym_false] = ACTIONS(1258),
    [anon_sym_NULL] = ACTIONS(1258),
    [anon_sym_nullptr] = ACTIONS(1258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(119)] = {
    [sym_identifier] = ACTIONS(1262),
    [aux_sym_preproc_include_token1] = ACTIONS(1262),
    [aux_sym_preproc_def_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token2] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1262),
    [aux_sym_preproc_else_token1] = ACTIONS(1262),
    [aux_sym_preproc_elif_token1] = ACTIONS(1262),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1262),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1262),
    [sym_preproc_directive] = ACTIONS(1262),
    [anon_sym_LPAREN2] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_TILDE] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_STAR] = ACTIONS(1264),
    [anon_sym_AMP] = ACTIONS(1264),
    [anon_sym_SEMI] = ACTIONS(1264),
    [anon_sym___extension__] = ACTIONS(1262),
    [anon_sym_typedef] = ACTIONS(1262),
    [anon_sym_extern] = ACTIONS(1262),
    [anon_sym___attribute__] = ACTIONS(1262),
    [anon_sym___attribute] = ACTIONS(1262),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1264),
    [anon_sym___declspec] = ACTIONS(1262),
    [anon_sym___cdecl] = ACTIONS(1262),
    [anon_sym___clrcall] = ACTIONS(1262),
    [anon_sym___stdcall] = ACTIONS(1262),
    [anon_sym___fastcall] = ACTIONS(1262),
    [anon_sym___thiscall] = ACTIONS(1262),
    [anon_sym___vectorcall] = ACTIONS(1262),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_signed] = ACTIONS(1262),
    [anon_sym_unsigned] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_auto] = ACTIONS(1262),
    [anon_sym_register] = ACTIONS(1262),
    [anon_sym_inline] = ACTIONS(1262),
    [anon_sym___inline] = ACTIONS(1262),
    [anon_sym___inline__] = ACTIONS(1262),
    [anon_sym___forceinline] = ACTIONS(1262),
    [anon_sym_thread_local] = ACTIONS(1262),
    [anon_sym___thread] = ACTIONS(1262),
    [anon_sym_const] = ACTIONS(1262),
    [anon_sym_constexpr] = ACTIONS(1262),
    [anon_sym_volatile] = ACTIONS(1262),
    [anon_sym_restrict] = ACTIONS(1262),
    [anon_sym___restrict__] = ACTIONS(1262),
    [anon_sym__Atomic] = ACTIONS(1262),
    [anon_sym__Noreturn] = ACTIONS(1262),
    [anon_sym_noreturn] = ACTIONS(1262),
    [anon_sym__Nonnull] = ACTIONS(1262),
    [anon_sym_alignas] = ACTIONS(1262),
    [anon_sym__Alignas] = ACTIONS(1262),
    [sym_primitive_type] = ACTIONS(1262),
    [anon_sym_enum] = ACTIONS(1262),
    [anon_sym_struct] = ACTIONS(1262),
    [anon_sym_union] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_goto] = ACTIONS(1262),
    [anon_sym___try] = ACTIONS(1262),
    [anon_sym___leave] = ACTIONS(1262),
    [anon_sym_DASH_DASH] = ACTIONS(1264),
    [anon_sym_PLUS_PLUS] = ACTIONS(1264),
    [anon_sym_sizeof] = ACTIONS(1262),
    [anon_sym___alignof__] = ACTIONS(1262),
    [anon_sym___alignof] = ACTIONS(1262),
    [anon_sym__alignof] = ACTIONS(1262),
    [anon_sym_alignof] = ACTIONS(1262),
    [anon_sym__Alignof] = ACTIONS(1262),
    [anon_sym_offsetof] = ACTIONS(1262),
    [anon_sym__Generic] = ACTIONS(1262),
    [anon_sym_asm] = ACTIONS(1262),
    [anon_sym___asm__] = ACTIONS(1262),
    [anon_sym___asm] = ACTIONS(1262),
    [sym_number_literal] = ACTIONS(1264),
    [anon_sym_L_SQUOTE] = ACTIONS(1264),
    [anon_sym_u_SQUOTE] = ACTIONS(1264),
    [anon_sym_U_SQUOTE] = ACTIONS(1264),
    [anon_sym_u8_SQUOTE] = ACTIONS(1264),
    [anon_sym_SQUOTE] = ACTIONS(1264),
    [anon_sym_L_DQUOTE] = ACTIONS(1264),
    [anon_sym_u_DQUOTE] = ACTIONS(1264),
    [anon_sym_U_DQUOTE] = ACTIONS(1264),
    [anon_sym_u8_DQUOTE] = ACTIONS(1264),
    [anon_sym_DQUOTE] = ACTIONS(1264),
    [sym_true] = ACTIONS(1262),
    [sym_false] = ACTIONS(1262),
    [anon_sym_NULL] = ACTIONS(1262),
    [anon_sym_nullptr] = ACTIONS(1262),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(120)] = {
    [sym_identifier] = ACTIONS(1266),
    [aux_sym_preproc_include_token1] = ACTIONS(1266),
    [aux_sym_preproc_def_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token2] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1266),
    [aux_sym_preproc_else_token1] = ACTIONS(1266),
    [aux_sym_preproc_elif_token1] = ACTIONS(1266),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1266),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1266),
    [sym_preproc_directive] = ACTIONS(1266),
    [anon_sym_LPAREN2] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1266),
    [anon_sym_PLUS] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1268),
    [anon_sym_AMP] = ACTIONS(1268),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym___extension__] = ACTIONS(1266),
    [anon_sym_typedef] = ACTIONS(1266),
    [anon_sym_extern] = ACTIONS(1266),
    [anon_sym___attribute__] = ACTIONS(1266),
    [anon_sym___attribute] = ACTIONS(1266),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1268),
    [anon_sym___declspec] = ACTIONS(1266),
    [anon_sym___cdecl] = ACTIONS(1266),
    [anon_sym___clrcall] = ACTIONS(1266),
    [anon_sym___stdcall] = ACTIONS(1266),
    [anon_sym___fastcall] = ACTIONS(1266),
    [anon_sym___thiscall] = ACTIONS(1266),
    [anon_sym___vectorcall] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_signed] = ACTIONS(1266),
    [anon_sym_unsigned] = ACTIONS(1266),
    [anon_sym_long] = ACTIONS(1266),
    [anon_sym_short] = ACTIONS(1266),
    [anon_sym_static] = ACTIONS(1266),
    [anon_sym_auto] = ACTIONS(1266),
    [anon_sym_register] = ACTIONS(1266),
    [anon_sym_inline] = ACTIONS(1266),
    [anon_sym___inline] = ACTIONS(1266),
    [anon_sym___inline__] = ACTIONS(1266),
    [anon_sym___forceinline] = ACTIONS(1266),
    [anon_sym_thread_local] = ACTIONS(1266),
    [anon_sym___thread] = ACTIONS(1266),
    [anon_sym_const] = ACTIONS(1266),
    [anon_sym_constexpr] = ACTIONS(1266),
    [anon_sym_volatile] = ACTIONS(1266),
    [anon_sym_restrict] = ACTIONS(1266),
    [anon_sym___restrict__] = ACTIONS(1266),
    [anon_sym__Atomic] = ACTIONS(1266),
    [anon_sym__Noreturn] = ACTIONS(1266),
    [anon_sym_noreturn] = ACTIONS(1266),
    [anon_sym__Nonnull] = ACTIONS(1266),
    [anon_sym_alignas] = ACTIONS(1266),
    [anon_sym__Alignas] = ACTIONS(1266),
    [sym_primitive_type] = ACTIONS(1266),
    [anon_sym_enum] = ACTIONS(1266),
    [anon_sym_struct] = ACTIONS(1266),
    [anon_sym_union] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_switch] = ACTIONS(1266),
    [anon_sym_case] = ACTIONS(1266),
    [anon_sym_default] = ACTIONS(1266),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_break] = ACTIONS(1266),
    [anon_sym_continue] = ACTIONS(1266),
    [anon_sym_goto] = ACTIONS(1266),
    [anon_sym___try] = ACTIONS(1266),
    [anon_sym___leave] = ACTIONS(1266),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_sizeof] = ACTIONS(1266),
    [anon_sym___alignof__] = ACTIONS(1266),
    [anon_sym___alignof] = ACTIONS(1266),
    [anon_sym__alignof] = ACTIONS(1266),
    [anon_sym_alignof] = ACTIONS(1266),
    [anon_sym__Alignof] = ACTIONS(1266),
    [anon_sym_offsetof] = ACTIONS(1266),
    [anon_sym__Generic] = ACTIONS(1266),
    [anon_sym_asm] = ACTIONS(1266),
    [anon_sym___asm__] = ACTIONS(1266),
    [anon_sym___asm] = ACTIONS(1266),
    [sym_number_literal] = ACTIONS(1268),
    [anon_sym_L_SQUOTE] = ACTIONS(1268),
    [anon_sym_u_SQUOTE] = ACTIONS(1268),
    [anon_sym_U_SQUOTE] = ACTIONS(1268),
    [anon_sym_u8_SQUOTE] = ACTIONS(1268),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_L_DQUOTE] = ACTIONS(1268),
    [anon_sym_u_DQUOTE] = ACTIONS(1268),
    [anon_sym_U_DQUOTE] = ACTIONS(1268),
    [anon_sym_u8_DQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1268),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [anon_sym_NULL] = ACTIONS(1266),
    [anon_sym_nullptr] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(121)] = {
    [sym_identifier] = ACTIONS(1270),
    [aux_sym_preproc_include_token1] = ACTIONS(1270),
    [aux_sym_preproc_def_token1] = ACTIONS(1270),
    [aux_sym_preproc_if_token1] = ACTIONS(1270),
    [aux_sym_preproc_if_token2] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1270),
    [aux_sym_preproc_else_token1] = ACTIONS(1270),
    [aux_sym_preproc_elif_token1] = ACTIONS(1270),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1270),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1270),
    [sym_preproc_directive] = ACTIONS(1270),
    [anon_sym_LPAREN2] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_TILDE] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_AMP] = ACTIONS(1272),
    [anon_sym_SEMI] = ACTIONS(1272),
    [anon_sym___extension__] = ACTIONS(1270),
    [anon_sym_typedef] = ACTIONS(1270),
    [anon_sym_extern] = ACTIONS(1270),
    [anon_sym___attribute__] = ACTIONS(1270),
    [anon_sym___attribute] = ACTIONS(1270),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1272),
    [anon_sym___declspec] = ACTIONS(1270),
    [anon_sym___cdecl] = ACTIONS(1270),
    [anon_sym___clrcall] = ACTIONS(1270),
    [anon_sym___stdcall] = ACTIONS(1270),
    [anon_sym___fastcall] = ACTIONS(1270),
    [anon_sym___thiscall] = ACTIONS(1270),
    [anon_sym___vectorcall] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_signed] = ACTIONS(1270),
    [anon_sym_unsigned] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_auto] = ACTIONS(1270),
    [anon_sym_register] = ACTIONS(1270),
    [anon_sym_inline] = ACTIONS(1270),
    [anon_sym___inline] = ACTIONS(1270),
    [anon_sym___inline__] = ACTIONS(1270),
    [anon_sym___forceinline] = ACTIONS(1270),
    [anon_sym_thread_local] = ACTIONS(1270),
    [anon_sym___thread] = ACTIONS(1270),
    [anon_sym_const] = ACTIONS(1270),
    [anon_sym_constexpr] = ACTIONS(1270),
    [anon_sym_volatile] = ACTIONS(1270),
    [anon_sym_restrict] = ACTIONS(1270),
    [anon_sym___restrict__] = ACTIONS(1270),
    [anon_sym__Atomic] = ACTIONS(1270),
    [anon_sym__Noreturn] = ACTIONS(1270),
    [anon_sym_noreturn] = ACTIONS(1270),
    [anon_sym__Nonnull] = ACTIONS(1270),
    [anon_sym_alignas] = ACTIONS(1270),
    [anon_sym__Alignas] = ACTIONS(1270),
    [sym_primitive_type] = ACTIONS(1270),
    [anon_sym_enum] = ACTIONS(1270),
    [anon_sym_struct] = ACTIONS(1270),
    [anon_sym_union] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_goto] = ACTIONS(1270),
    [anon_sym___try] = ACTIONS(1270),
    [anon_sym___leave] = ACTIONS(1270),
    [anon_sym_DASH_DASH] = ACTIONS(1272),
    [anon_sym_PLUS_PLUS] = ACTIONS(1272),
    [anon_sym_sizeof] = ACTIONS(1270),
    [anon_sym___alignof__] = ACTIONS(1270),
    [anon_sym___alignof] = ACTIONS(1270),
    [anon_sym__alignof] = ACTIONS(1270),
    [anon_sym_alignof] = ACTIONS(1270),
    [anon_sym__Alignof] = ACTIONS(1270),
    [anon_sym_offsetof] = ACTIONS(1270),
    [anon_sym__Generic] = ACTIONS(1270),
    [anon_sym_asm] = ACTIONS(1270),
    [anon_sym___asm__] = ACTIONS(1270),
    [anon_sym___asm] = ACTIONS(1270),
    [sym_number_literal] = ACTIONS(1272),
    [anon_sym_L_SQUOTE] = ACTIONS(1272),
    [anon_sym_u_SQUOTE] = ACTIONS(1272),
    [anon_sym_U_SQUOTE] = ACTIONS(1272),
    [anon_sym_u8_SQUOTE] = ACTIONS(1272),
    [anon_sym_SQUOTE] = ACTIONS(1272),
    [anon_sym_L_DQUOTE] = ACTIONS(1272),
    [anon_sym_u_DQUOTE] = ACTIONS(1272),
    [anon_sym_U_DQUOTE] = ACTIONS(1272),
    [anon_sym_u8_DQUOTE] = ACTIONS(1272),
    [anon_sym_DQUOTE] = ACTIONS(1272),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [anon_sym_NULL] = ACTIONS(1270),
    [anon_sym_nullptr] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(122)] = {
    [sym_identifier] = ACTIONS(1274),
    [aux_sym_preproc_include_token1] = ACTIONS(1274),
    [aux_sym_preproc_def_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token2] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1274),
    [aux_sym_preproc_else_token1] = ACTIONS(1274),
    [aux_sym_preproc_elif_token1] = ACTIONS(1274),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1274),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1274),
    [sym_preproc_directive] = ACTIONS(1274),
    [anon_sym_LPAREN2] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_TILDE] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1274),
    [anon_sym_PLUS] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1276),
    [anon_sym_AMP] = ACTIONS(1276),
    [anon_sym_SEMI] = ACTIONS(1276),
    [anon_sym___extension__] = ACTIONS(1274),
    [anon_sym_typedef] = ACTIONS(1274),
    [anon_sym_extern] = ACTIONS(1274),
    [anon_sym___attribute__] = ACTIONS(1274),
    [anon_sym___attribute] = ACTIONS(1274),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1276),
    [anon_sym___declspec] = ACTIONS(1274),
    [anon_sym___cdecl] = ACTIONS(1274),
    [anon_sym___clrcall] = ACTIONS(1274),
    [anon_sym___stdcall] = ACTIONS(1274),
    [anon_sym___fastcall] = ACTIONS(1274),
    [anon_sym___thiscall] = ACTIONS(1274),
    [anon_sym___vectorcall] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_signed] = ACTIONS(1274),
    [anon_sym_unsigned] = ACTIONS(1274),
    [anon_sym_long] = ACTIONS(1274),
    [anon_sym_short] = ACTIONS(1274),
    [anon_sym_static] = ACTIONS(1274),
    [anon_sym_auto] = ACTIONS(1274),
    [anon_sym_register] = ACTIONS(1274),
    [anon_sym_inline] = ACTIONS(1274),
    [anon_sym___inline] = ACTIONS(1274),
    [anon_sym___inline__] = ACTIONS(1274),
    [anon_sym___forceinline] = ACTIONS(1274),
    [anon_sym_thread_local] = ACTIONS(1274),
    [anon_sym___thread] = ACTIONS(1274),
    [anon_sym_const] = ACTIONS(1274),
    [anon_sym_constexpr] = ACTIONS(1274),
    [anon_sym_volatile] = ACTIONS(1274),
    [anon_sym_restrict] = ACTIONS(1274),
    [anon_sym___restrict__] = ACTIONS(1274),
    [anon_sym__Atomic] = ACTIONS(1274),
    [anon_sym__Noreturn] = ACTIONS(1274),
    [anon_sym_noreturn] = ACTIONS(1274),
    [anon_sym__Nonnull] = ACTIONS(1274),
    [anon_sym_alignas] = ACTIONS(1274),
    [anon_sym__Alignas] = ACTIONS(1274),
    [sym_primitive_type] = ACTIONS(1274),
    [anon_sym_enum] = ACTIONS(1274),
    [anon_sym_struct] = ACTIONS(1274),
    [anon_sym_union] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1274),
    [anon_sym_switch] = ACTIONS(1274),
    [anon_sym_case] = ACTIONS(1274),
    [anon_sym_default] = ACTIONS(1274),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_break] = ACTIONS(1274),
    [anon_sym_continue] = ACTIONS(1274),
    [anon_sym_goto] = ACTIONS(1274),
    [anon_sym___try] = ACTIONS(1274),
    [anon_sym___leave] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_PLUS_PLUS] = ACTIONS(1276),
    [anon_sym_sizeof] = ACTIONS(1274),
    [anon_sym___alignof__] = ACTIONS(1274),
    [anon_sym___alignof] = ACTIONS(1274),
    [anon_sym__alignof] = ACTIONS(1274),
    [anon_sym_alignof] = ACTIONS(1274),
    [anon_sym__Alignof] = ACTIONS(1274),
    [anon_sym_offsetof] = ACTIONS(1274),
    [anon_sym__Generic] = ACTIONS(1274),
    [anon_sym_asm] = ACTIONS(1274),
    [anon_sym___asm__] = ACTIONS(1274),
    [anon_sym___asm] = ACTIONS(1274),
    [sym_number_literal] = ACTIONS(1276),
    [anon_sym_L_SQUOTE] = ACTIONS(1276),
    [anon_sym_u_SQUOTE] = ACTIONS(1276),
    [anon_sym_U_SQUOTE] = ACTIONS(1276),
    [anon_sym_u8_SQUOTE] = ACTIONS(1276),
    [anon_sym_SQUOTE] = ACTIONS(1276),
    [anon_sym_L_DQUOTE] = ACTIONS(1276),
    [anon_sym_u_DQUOTE] = ACTIONS(1276),
    [anon_sym_U_DQUOTE] = ACTIONS(1276),
    [anon_sym_u8_DQUOTE] = ACTIONS(1276),
    [anon_sym_DQUOTE] = ACTIONS(1276),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [anon_sym_NULL] = ACTIONS(1274),
    [anon_sym_nullptr] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(123)] = {
    [sym_identifier] = ACTIONS(1278),
    [aux_sym_preproc_include_token1] = ACTIONS(1278),
    [aux_sym_preproc_def_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token2] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1278),
    [aux_sym_preproc_else_token1] = ACTIONS(1278),
    [aux_sym_preproc_elif_token1] = ACTIONS(1278),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1278),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1278),
    [sym_preproc_directive] = ACTIONS(1278),
    [anon_sym_LPAREN2] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1278),
    [anon_sym_PLUS] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1280),
    [anon_sym___extension__] = ACTIONS(1278),
    [anon_sym_typedef] = ACTIONS(1278),
    [anon_sym_extern] = ACTIONS(1278),
    [anon_sym___attribute__] = ACTIONS(1278),
    [anon_sym___attribute] = ACTIONS(1278),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1280),
    [anon_sym___declspec] = ACTIONS(1278),
    [anon_sym___cdecl] = ACTIONS(1278),
    [anon_sym___clrcall] = ACTIONS(1278),
    [anon_sym___stdcall] = ACTIONS(1278),
    [anon_sym___fastcall] = ACTIONS(1278),
    [anon_sym___thiscall] = ACTIONS(1278),
    [anon_sym___vectorcall] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_signed] = ACTIONS(1278),
    [anon_sym_unsigned] = ACTIONS(1278),
    [anon_sym_long] = ACTIONS(1278),
    [anon_sym_short] = ACTIONS(1278),
    [anon_sym_static] = ACTIONS(1278),
    [anon_sym_auto] = ACTIONS(1278),
    [anon_sym_register] = ACTIONS(1278),
    [anon_sym_inline] = ACTIONS(1278),
    [anon_sym___inline] = ACTIONS(1278),
    [anon_sym___inline__] = ACTIONS(1278),
    [anon_sym___forceinline] = ACTIONS(1278),
    [anon_sym_thread_local] = ACTIONS(1278),
    [anon_sym___thread] = ACTIONS(1278),
    [anon_sym_const] = ACTIONS(1278),
    [anon_sym_constexpr] = ACTIONS(1278),
    [anon_sym_volatile] = ACTIONS(1278),
    [anon_sym_restrict] = ACTIONS(1278),
    [anon_sym___restrict__] = ACTIONS(1278),
    [anon_sym__Atomic] = ACTIONS(1278),
    [anon_sym__Noreturn] = ACTIONS(1278),
    [anon_sym_noreturn] = ACTIONS(1278),
    [anon_sym__Nonnull] = ACTIONS(1278),
    [anon_sym_alignas] = ACTIONS(1278),
    [anon_sym__Alignas] = ACTIONS(1278),
    [sym_primitive_type] = ACTIONS(1278),
    [anon_sym_enum] = ACTIONS(1278),
    [anon_sym_struct] = ACTIONS(1278),
    [anon_sym_union] = ACTIONS(1278),
    [anon_sym_if] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1278),
    [anon_sym_case] = ACTIONS(1278),
    [anon_sym_default] = ACTIONS(1278),
    [anon_sym_while] = ACTIONS(1278),
    [anon_sym_do] = ACTIONS(1278),
    [anon_sym_for] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1278),
    [anon_sym_break] = ACTIONS(1278),
    [anon_sym_continue] = ACTIONS(1278),
    [anon_sym_goto] = ACTIONS(1278),
    [anon_sym___try] = ACTIONS(1278),
    [anon_sym___leave] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1280),
    [anon_sym_sizeof] = ACTIONS(1278),
    [anon_sym___alignof__] = ACTIONS(1278),
    [anon_sym___alignof] = ACTIONS(1278),
    [anon_sym__alignof] = ACTIONS(1278),
    [anon_sym_alignof] = ACTIONS(1278),
    [anon_sym__Alignof] = ACTIONS(1278),
    [anon_sym_offsetof] = ACTIONS(1278),
    [anon_sym__Generic] = ACTIONS(1278),
    [anon_sym_asm] = ACTIONS(1278),
    [anon_sym___asm__] = ACTIONS(1278),
    [anon_sym___asm] = ACTIONS(1278),
    [sym_number_literal] = ACTIONS(1280),
    [anon_sym_L_SQUOTE] = ACTIONS(1280),
    [anon_sym_u_SQUOTE] = ACTIONS(1280),
    [anon_sym_U_SQUOTE] = ACTIONS(1280),
    [anon_sym_u8_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_L_DQUOTE] = ACTIONS(1280),
    [anon_sym_u_DQUOTE] = ACTIONS(1280),
    [anon_sym_U_DQUOTE] = ACTIONS(1280),
    [anon_sym_u8_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [sym_true] = ACTIONS(1278),
    [sym_false] = ACTIONS(1278),
    [anon_sym_NULL] = ACTIONS(1278),
    [anon_sym_nullptr] = ACTIONS(1278),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(124)] = {
    [sym_identifier] = ACTIONS(1282),
    [aux_sym_preproc_include_token1] = ACTIONS(1282),
    [aux_sym_preproc_def_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token2] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1282),
    [aux_sym_preproc_else_token1] = ACTIONS(1282),
    [aux_sym_preproc_elif_token1] = ACTIONS(1282),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1282),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1282),
    [sym_preproc_directive] = ACTIONS(1282),
    [anon_sym_LPAREN2] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_TILDE] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1282),
    [anon_sym_PLUS] = ACTIONS(1282),
    [anon_sym_STAR] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym___extension__] = ACTIONS(1282),
    [anon_sym_typedef] = ACTIONS(1282),
    [anon_sym_extern] = ACTIONS(1282),
    [anon_sym___attribute__] = ACTIONS(1282),
    [anon_sym___attribute] = ACTIONS(1282),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1284),
    [anon_sym___declspec] = ACTIONS(1282),
    [anon_sym___cdecl] = ACTIONS(1282),
    [anon_sym___clrcall] = ACTIONS(1282),
    [anon_sym___stdcall] = ACTIONS(1282),
    [anon_sym___fastcall] = ACTIONS(1282),
    [anon_sym___thiscall] = ACTIONS(1282),
    [anon_sym___vectorcall] = ACTIONS(1282),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_signed] = ACTIONS(1282),
    [anon_sym_unsigned] = ACTIONS(1282),
    [anon_sym_long] = ACTIONS(1282),
    [anon_sym_short] = ACTIONS(1282),
    [anon_sym_static] = ACTIONS(1282),
    [anon_sym_auto] = ACTIONS(1282),
    [anon_sym_register] = ACTIONS(1282),
    [anon_sym_inline] = ACTIONS(1282),
    [anon_sym___inline] = ACTIONS(1282),
    [anon_sym___inline__] = ACTIONS(1282),
    [anon_sym___forceinline] = ACTIONS(1282),
    [anon_sym_thread_local] = ACTIONS(1282),
    [anon_sym___thread] = ACTIONS(1282),
    [anon_sym_const] = ACTIONS(1282),
    [anon_sym_constexpr] = ACTIONS(1282),
    [anon_sym_volatile] = ACTIONS(1282),
    [anon_sym_restrict] = ACTIONS(1282),
    [anon_sym___restrict__] = ACTIONS(1282),
    [anon_sym__Atomic] = ACTIONS(1282),
    [anon_sym__Noreturn] = ACTIONS(1282),
    [anon_sym_noreturn] = ACTIONS(1282),
    [anon_sym__Nonnull] = ACTIONS(1282),
    [anon_sym_alignas] = ACTIONS(1282),
    [anon_sym__Alignas] = ACTIONS(1282),
    [sym_primitive_type] = ACTIONS(1282),
    [anon_sym_enum] = ACTIONS(1282),
    [anon_sym_struct] = ACTIONS(1282),
    [anon_sym_union] = ACTIONS(1282),
    [anon_sym_if] = ACTIONS(1282),
    [anon_sym_switch] = ACTIONS(1282),
    [anon_sym_case] = ACTIONS(1282),
    [anon_sym_default] = ACTIONS(1282),
    [anon_sym_while] = ACTIONS(1282),
    [anon_sym_do] = ACTIONS(1282),
    [anon_sym_for] = ACTIONS(1282),
    [anon_sym_return] = ACTIONS(1282),
    [anon_sym_break] = ACTIONS(1282),
    [anon_sym_continue] = ACTIONS(1282),
    [anon_sym_goto] = ACTIONS(1282),
    [anon_sym___try] = ACTIONS(1282),
    [anon_sym___leave] = ACTIONS(1282),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_sizeof] = ACTIONS(1282),
    [anon_sym___alignof__] = ACTIONS(1282),
    [anon_sym___alignof] = ACTIONS(1282),
    [anon_sym__alignof] = ACTIONS(1282),
    [anon_sym_alignof] = ACTIONS(1282),
    [anon_sym__Alignof] = ACTIONS(1282),
    [anon_sym_offsetof] = ACTIONS(1282),
    [anon_sym__Generic] = ACTIONS(1282),
    [anon_sym_asm] = ACTIONS(1282),
    [anon_sym___asm__] = ACTIONS(1282),
    [anon_sym___asm] = ACTIONS(1282),
    [sym_number_literal] = ACTIONS(1284),
    [anon_sym_L_SQUOTE] = ACTIONS(1284),
    [anon_sym_u_SQUOTE] = ACTIONS(1284),
    [anon_sym_U_SQUOTE] = ACTIONS(1284),
    [anon_sym_u8_SQUOTE] = ACTIONS(1284),
    [anon_sym_SQUOTE] = ACTIONS(1284),
    [anon_sym_L_DQUOTE] = ACTIONS(1284),
    [anon_sym_u_DQUOTE] = ACTIONS(1284),
    [anon_sym_U_DQUOTE] = ACTIONS(1284),
    [anon_sym_u8_DQUOTE] = ACTIONS(1284),
    [anon_sym_DQUOTE] = ACTIONS(1284),
    [sym_true] = ACTIONS(1282),
    [sym_false] = ACTIONS(1282),
    [anon_sym_NULL] = ACTIONS(1282),
    [anon_sym_nullptr] = ACTIONS(1282),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(125)] = {
    [sym_identifier] = ACTIONS(1286),
    [aux_sym_preproc_include_token1] = ACTIONS(1286),
    [aux_sym_preproc_def_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token2] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1286),
    [aux_sym_preproc_else_token1] = ACTIONS(1286),
    [aux_sym_preproc_elif_token1] = ACTIONS(1286),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1286),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1286),
    [sym_preproc_directive] = ACTIONS(1286),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_TILDE] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1288),
    [anon_sym_AMP] = ACTIONS(1288),
    [anon_sym_SEMI] = ACTIONS(1288),
    [anon_sym___extension__] = ACTIONS(1286),
    [anon_sym_typedef] = ACTIONS(1286),
    [anon_sym_extern] = ACTIONS(1286),
    [anon_sym___attribute__] = ACTIONS(1286),
    [anon_sym___attribute] = ACTIONS(1286),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1288),
    [anon_sym___declspec] = ACTIONS(1286),
    [anon_sym___cdecl] = ACTIONS(1286),
    [anon_sym___clrcall] = ACTIONS(1286),
    [anon_sym___stdcall] = ACTIONS(1286),
    [anon_sym___fastcall] = ACTIONS(1286),
    [anon_sym___thiscall] = ACTIONS(1286),
    [anon_sym___vectorcall] = ACTIONS(1286),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_signed] = ACTIONS(1286),
    [anon_sym_unsigned] = ACTIONS(1286),
    [anon_sym_long] = ACTIONS(1286),
    [anon_sym_short] = ACTIONS(1286),
    [anon_sym_static] = ACTIONS(1286),
    [anon_sym_auto] = ACTIONS(1286),
    [anon_sym_register] = ACTIONS(1286),
    [anon_sym_inline] = ACTIONS(1286),
    [anon_sym___inline] = ACTIONS(1286),
    [anon_sym___inline__] = ACTIONS(1286),
    [anon_sym___forceinline] = ACTIONS(1286),
    [anon_sym_thread_local] = ACTIONS(1286),
    [anon_sym___thread] = ACTIONS(1286),
    [anon_sym_const] = ACTIONS(1286),
    [anon_sym_constexpr] = ACTIONS(1286),
    [anon_sym_volatile] = ACTIONS(1286),
    [anon_sym_restrict] = ACTIONS(1286),
    [anon_sym___restrict__] = ACTIONS(1286),
    [anon_sym__Atomic] = ACTIONS(1286),
    [anon_sym__Noreturn] = ACTIONS(1286),
    [anon_sym_noreturn] = ACTIONS(1286),
    [anon_sym__Nonnull] = ACTIONS(1286),
    [anon_sym_alignas] = ACTIONS(1286),
    [anon_sym__Alignas] = ACTIONS(1286),
    [sym_primitive_type] = ACTIONS(1286),
    [anon_sym_enum] = ACTIONS(1286),
    [anon_sym_struct] = ACTIONS(1286),
    [anon_sym_union] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(1286),
    [anon_sym_switch] = ACTIONS(1286),
    [anon_sym_case] = ACTIONS(1286),
    [anon_sym_default] = ACTIONS(1286),
    [anon_sym_while] = ACTIONS(1286),
    [anon_sym_do] = ACTIONS(1286),
    [anon_sym_for] = ACTIONS(1286),
    [anon_sym_return] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(1286),
    [anon_sym_continue] = ACTIONS(1286),
    [anon_sym_goto] = ACTIONS(1286),
    [anon_sym___try] = ACTIONS(1286),
    [anon_sym___leave] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1288),
    [anon_sym_PLUS_PLUS] = ACTIONS(1288),
    [anon_sym_sizeof] = ACTIONS(1286),
    [anon_sym___alignof__] = ACTIONS(1286),
    [anon_sym___alignof] = ACTIONS(1286),
    [anon_sym__alignof] = ACTIONS(1286),
    [anon_sym_alignof] = ACTIONS(1286),
    [anon_sym__Alignof] = ACTIONS(1286),
    [anon_sym_offsetof] = ACTIONS(1286),
    [anon_sym__Generic] = ACTIONS(1286),
    [anon_sym_asm] = ACTIONS(1286),
    [anon_sym___asm__] = ACTIONS(1286),
    [anon_sym___asm] = ACTIONS(1286),
    [sym_number_literal] = ACTIONS(1288),
    [anon_sym_L_SQUOTE] = ACTIONS(1288),
    [anon_sym_u_SQUOTE] = ACTIONS(1288),
    [anon_sym_U_SQUOTE] = ACTIONS(1288),
    [anon_sym_u8_SQUOTE] = ACTIONS(1288),
    [anon_sym_SQUOTE] = ACTIONS(1288),
    [anon_sym_L_DQUOTE] = ACTIONS(1288),
    [anon_sym_u_DQUOTE] = ACTIONS(1288),
    [anon_sym_U_DQUOTE] = ACTIONS(1288),
    [anon_sym_u8_DQUOTE] = ACTIONS(1288),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [sym_true] = ACTIONS(1286),
    [sym_false] = ACTIONS(1286),
    [anon_sym_NULL] = ACTIONS(1286),
    [anon_sym_nullptr] = ACTIONS(1286),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(126)] = {
    [sym_identifier] = ACTIONS(1290),
    [aux_sym_preproc_include_token1] = ACTIONS(1290),
    [aux_sym_preproc_def_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token2] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1290),
    [aux_sym_preproc_else_token1] = ACTIONS(1290),
    [aux_sym_preproc_elif_token1] = ACTIONS(1290),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1290),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1290),
    [sym_preproc_directive] = ACTIONS(1290),
    [anon_sym_LPAREN2] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym___extension__] = ACTIONS(1290),
    [anon_sym_typedef] = ACTIONS(1290),
    [anon_sym_extern] = ACTIONS(1290),
    [anon_sym___attribute__] = ACTIONS(1290),
    [anon_sym___attribute] = ACTIONS(1290),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1292),
    [anon_sym___declspec] = ACTIONS(1290),
    [anon_sym___cdecl] = ACTIONS(1290),
    [anon_sym___clrcall] = ACTIONS(1290),
    [anon_sym___stdcall] = ACTIONS(1290),
    [anon_sym___fastcall] = ACTIONS(1290),
    [anon_sym___thiscall] = ACTIONS(1290),
    [anon_sym___vectorcall] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_signed] = ACTIONS(1290),
    [anon_sym_unsigned] = ACTIONS(1290),
    [anon_sym_long] = ACTIONS(1290),
    [anon_sym_short] = ACTIONS(1290),
    [anon_sym_static] = ACTIONS(1290),
    [anon_sym_auto] = ACTIONS(1290),
    [anon_sym_register] = ACTIONS(1290),
    [anon_sym_inline] = ACTIONS(1290),
    [anon_sym___inline] = ACTIONS(1290),
    [anon_sym___inline__] = ACTIONS(1290),
    [anon_sym___forceinline] = ACTIONS(1290),
    [anon_sym_thread_local] = ACTIONS(1290),
    [anon_sym___thread] = ACTIONS(1290),
    [anon_sym_const] = ACTIONS(1290),
    [anon_sym_constexpr] = ACTIONS(1290),
    [anon_sym_volatile] = ACTIONS(1290),
    [anon_sym_restrict] = ACTIONS(1290),
    [anon_sym___restrict__] = ACTIONS(1290),
    [anon_sym__Atomic] = ACTIONS(1290),
    [anon_sym__Noreturn] = ACTIONS(1290),
    [anon_sym_noreturn] = ACTIONS(1290),
    [anon_sym__Nonnull] = ACTIONS(1290),
    [anon_sym_alignas] = ACTIONS(1290),
    [anon_sym__Alignas] = ACTIONS(1290),
    [sym_primitive_type] = ACTIONS(1290),
    [anon_sym_enum] = ACTIONS(1290),
    [anon_sym_struct] = ACTIONS(1290),
    [anon_sym_union] = ACTIONS(1290),
    [anon_sym_if] = ACTIONS(1290),
    [anon_sym_switch] = ACTIONS(1290),
    [anon_sym_case] = ACTIONS(1290),
    [anon_sym_default] = ACTIONS(1290),
    [anon_sym_while] = ACTIONS(1290),
    [anon_sym_do] = ACTIONS(1290),
    [anon_sym_for] = ACTIONS(1290),
    [anon_sym_return] = ACTIONS(1290),
    [anon_sym_break] = ACTIONS(1290),
    [anon_sym_continue] = ACTIONS(1290),
    [anon_sym_goto] = ACTIONS(1290),
    [anon_sym___try] = ACTIONS(1290),
    [anon_sym___leave] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_sizeof] = ACTIONS(1290),
    [anon_sym___alignof__] = ACTIONS(1290),
    [anon_sym___alignof] = ACTIONS(1290),
    [anon_sym__alignof] = ACTIONS(1290),
    [anon_sym_alignof] = ACTIONS(1290),
    [anon_sym__Alignof] = ACTIONS(1290),
    [anon_sym_offsetof] = ACTIONS(1290),
    [anon_sym__Generic] = ACTIONS(1290),
    [anon_sym_asm] = ACTIONS(1290),
    [anon_sym___asm__] = ACTIONS(1290),
    [anon_sym___asm] = ACTIONS(1290),
    [sym_number_literal] = ACTIONS(1292),
    [anon_sym_L_SQUOTE] = ACTIONS(1292),
    [anon_sym_u_SQUOTE] = ACTIONS(1292),
    [anon_sym_U_SQUOTE] = ACTIONS(1292),
    [anon_sym_u8_SQUOTE] = ACTIONS(1292),
    [anon_sym_SQUOTE] = ACTIONS(1292),
    [anon_sym_L_DQUOTE] = ACTIONS(1292),
    [anon_sym_u_DQUOTE] = ACTIONS(1292),
    [anon_sym_U_DQUOTE] = ACTIONS(1292),
    [anon_sym_u8_DQUOTE] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [sym_true] = ACTIONS(1290),
    [sym_false] = ACTIONS(1290),
    [anon_sym_NULL] = ACTIONS(1290),
    [anon_sym_nullptr] = ACTIONS(1290),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(127)] = {
    [sym_identifier] = ACTIONS(1294),
    [aux_sym_preproc_include_token1] = ACTIONS(1294),
    [aux_sym_preproc_def_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token2] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1294),
    [aux_sym_preproc_else_token1] = ACTIONS(1294),
    [aux_sym_preproc_elif_token1] = ACTIONS(1294),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1294),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1294),
    [sym_preproc_directive] = ACTIONS(1294),
    [anon_sym_LPAREN2] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym___extension__] = ACTIONS(1294),
    [anon_sym_typedef] = ACTIONS(1294),
    [anon_sym_extern] = ACTIONS(1294),
    [anon_sym___attribute__] = ACTIONS(1294),
    [anon_sym___attribute] = ACTIONS(1294),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1296),
    [anon_sym___declspec] = ACTIONS(1294),
    [anon_sym___cdecl] = ACTIONS(1294),
    [anon_sym___clrcall] = ACTIONS(1294),
    [anon_sym___stdcall] = ACTIONS(1294),
    [anon_sym___fastcall] = ACTIONS(1294),
    [anon_sym___thiscall] = ACTIONS(1294),
    [anon_sym___vectorcall] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_signed] = ACTIONS(1294),
    [anon_sym_unsigned] = ACTIONS(1294),
    [anon_sym_long] = ACTIONS(1294),
    [anon_sym_short] = ACTIONS(1294),
    [anon_sym_static] = ACTIONS(1294),
    [anon_sym_auto] = ACTIONS(1294),
    [anon_sym_register] = ACTIONS(1294),
    [anon_sym_inline] = ACTIONS(1294),
    [anon_sym___inline] = ACTIONS(1294),
    [anon_sym___inline__] = ACTIONS(1294),
    [anon_sym___forceinline] = ACTIONS(1294),
    [anon_sym_thread_local] = ACTIONS(1294),
    [anon_sym___thread] = ACTIONS(1294),
    [anon_sym_const] = ACTIONS(1294),
    [anon_sym_constexpr] = ACTIONS(1294),
    [anon_sym_volatile] = ACTIONS(1294),
    [anon_sym_restrict] = ACTIONS(1294),
    [anon_sym___restrict__] = ACTIONS(1294),
    [anon_sym__Atomic] = ACTIONS(1294),
    [anon_sym__Noreturn] = ACTIONS(1294),
    [anon_sym_noreturn] = ACTIONS(1294),
    [anon_sym__Nonnull] = ACTIONS(1294),
    [anon_sym_alignas] = ACTIONS(1294),
    [anon_sym__Alignas] = ACTIONS(1294),
    [sym_primitive_type] = ACTIONS(1294),
    [anon_sym_enum] = ACTIONS(1294),
    [anon_sym_struct] = ACTIONS(1294),
    [anon_sym_union] = ACTIONS(1294),
    [anon_sym_if] = ACTIONS(1294),
    [anon_sym_switch] = ACTIONS(1294),
    [anon_sym_case] = ACTIONS(1294),
    [anon_sym_default] = ACTIONS(1294),
    [anon_sym_while] = ACTIONS(1294),
    [anon_sym_do] = ACTIONS(1294),
    [anon_sym_for] = ACTIONS(1294),
    [anon_sym_return] = ACTIONS(1294),
    [anon_sym_break] = ACTIONS(1294),
    [anon_sym_continue] = ACTIONS(1294),
    [anon_sym_goto] = ACTIONS(1294),
    [anon_sym___try] = ACTIONS(1294),
    [anon_sym___leave] = ACTIONS(1294),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_sizeof] = ACTIONS(1294),
    [anon_sym___alignof__] = ACTIONS(1294),
    [anon_sym___alignof] = ACTIONS(1294),
    [anon_sym__alignof] = ACTIONS(1294),
    [anon_sym_alignof] = ACTIONS(1294),
    [anon_sym__Alignof] = ACTIONS(1294),
    [anon_sym_offsetof] = ACTIONS(1294),
    [anon_sym__Generic] = ACTIONS(1294),
    [anon_sym_asm] = ACTIONS(1294),
    [anon_sym___asm__] = ACTIONS(1294),
    [anon_sym___asm] = ACTIONS(1294),
    [sym_number_literal] = ACTIONS(1296),
    [anon_sym_L_SQUOTE] = ACTIONS(1296),
    [anon_sym_u_SQUOTE] = ACTIONS(1296),
    [anon_sym_U_SQUOTE] = ACTIONS(1296),
    [anon_sym_u8_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_L_DQUOTE] = ACTIONS(1296),
    [anon_sym_u_DQUOTE] = ACTIONS(1296),
    [anon_sym_U_DQUOTE] = ACTIONS(1296),
    [anon_sym_u8_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [sym_true] = ACTIONS(1294),
    [sym_false] = ACTIONS(1294),
    [anon_sym_NULL] = ACTIONS(1294),
    [anon_sym_nullptr] = ACTIONS(1294),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(128)] = {
    [sym_identifier] = ACTIONS(1298),
    [aux_sym_preproc_include_token1] = ACTIONS(1298),
    [aux_sym_preproc_def_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token2] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1298),
    [aux_sym_preproc_else_token1] = ACTIONS(1298),
    [aux_sym_preproc_elif_token1] = ACTIONS(1298),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1298),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1298),
    [sym_preproc_directive] = ACTIONS(1298),
    [anon_sym_LPAREN2] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1298),
    [anon_sym_PLUS] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym___extension__] = ACTIONS(1298),
    [anon_sym_typedef] = ACTIONS(1298),
    [anon_sym_extern] = ACTIONS(1298),
    [anon_sym___attribute__] = ACTIONS(1298),
    [anon_sym___attribute] = ACTIONS(1298),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1300),
    [anon_sym___declspec] = ACTIONS(1298),
    [anon_sym___cdecl] = ACTIONS(1298),
    [anon_sym___clrcall] = ACTIONS(1298),
    [anon_sym___stdcall] = ACTIONS(1298),
    [anon_sym___fastcall] = ACTIONS(1298),
    [anon_sym___thiscall] = ACTIONS(1298),
    [anon_sym___vectorcall] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_signed] = ACTIONS(1298),
    [anon_sym_unsigned] = ACTIONS(1298),
    [anon_sym_long] = ACTIONS(1298),
    [anon_sym_short] = ACTIONS(1298),
    [anon_sym_static] = ACTIONS(1298),
    [anon_sym_auto] = ACTIONS(1298),
    [anon_sym_register] = ACTIONS(1298),
    [anon_sym_inline] = ACTIONS(1298),
    [anon_sym___inline] = ACTIONS(1298),
    [anon_sym___inline__] = ACTIONS(1298),
    [anon_sym___forceinline] = ACTIONS(1298),
    [anon_sym_thread_local] = ACTIONS(1298),
    [anon_sym___thread] = ACTIONS(1298),
    [anon_sym_const] = ACTIONS(1298),
    [anon_sym_constexpr] = ACTIONS(1298),
    [anon_sym_volatile] = ACTIONS(1298),
    [anon_sym_restrict] = ACTIONS(1298),
    [anon_sym___restrict__] = ACTIONS(1298),
    [anon_sym__Atomic] = ACTIONS(1298),
    [anon_sym__Noreturn] = ACTIONS(1298),
    [anon_sym_noreturn] = ACTIONS(1298),
    [anon_sym__Nonnull] = ACTIONS(1298),
    [anon_sym_alignas] = ACTIONS(1298),
    [anon_sym__Alignas] = ACTIONS(1298),
    [sym_primitive_type] = ACTIONS(1298),
    [anon_sym_enum] = ACTIONS(1298),
    [anon_sym_struct] = ACTIONS(1298),
    [anon_sym_union] = ACTIONS(1298),
    [anon_sym_if] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1298),
    [anon_sym_case] = ACTIONS(1298),
    [anon_sym_default] = ACTIONS(1298),
    [anon_sym_while] = ACTIONS(1298),
    [anon_sym_do] = ACTIONS(1298),
    [anon_sym_for] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1298),
    [anon_sym_break] = ACTIONS(1298),
    [anon_sym_continue] = ACTIONS(1298),
    [anon_sym_goto] = ACTIONS(1298),
    [anon_sym___try] = ACTIONS(1298),
    [anon_sym___leave] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_sizeof] = ACTIONS(1298),
    [anon_sym___alignof__] = ACTIONS(1298),
    [anon_sym___alignof] = ACTIONS(1298),
    [anon_sym__alignof] = ACTIONS(1298),
    [anon_sym_alignof] = ACTIONS(1298),
    [anon_sym__Alignof] = ACTIONS(1298),
    [anon_sym_offsetof] = ACTIONS(1298),
    [anon_sym__Generic] = ACTIONS(1298),
    [anon_sym_asm] = ACTIONS(1298),
    [anon_sym___asm__] = ACTIONS(1298),
    [anon_sym___asm] = ACTIONS(1298),
    [sym_number_literal] = ACTIONS(1300),
    [anon_sym_L_SQUOTE] = ACTIONS(1300),
    [anon_sym_u_SQUOTE] = ACTIONS(1300),
    [anon_sym_U_SQUOTE] = ACTIONS(1300),
    [anon_sym_u8_SQUOTE] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_L_DQUOTE] = ACTIONS(1300),
    [anon_sym_u_DQUOTE] = ACTIONS(1300),
    [anon_sym_U_DQUOTE] = ACTIONS(1300),
    [anon_sym_u8_DQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [sym_true] = ACTIONS(1298),
    [sym_false] = ACTIONS(1298),
    [anon_sym_NULL] = ACTIONS(1298),
    [anon_sym_nullptr] = ACTIONS(1298),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(129)] = {
    [sym_identifier] = ACTIONS(1302),
    [aux_sym_preproc_include_token1] = ACTIONS(1302),
    [aux_sym_preproc_def_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token2] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1302),
    [aux_sym_preproc_else_token1] = ACTIONS(1302),
    [aux_sym_preproc_elif_token1] = ACTIONS(1302),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1302),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1302),
    [sym_preproc_directive] = ACTIONS(1302),
    [anon_sym_LPAREN2] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym___extension__] = ACTIONS(1302),
    [anon_sym_typedef] = ACTIONS(1302),
    [anon_sym_extern] = ACTIONS(1302),
    [anon_sym___attribute__] = ACTIONS(1302),
    [anon_sym___attribute] = ACTIONS(1302),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1304),
    [anon_sym___declspec] = ACTIONS(1302),
    [anon_sym___cdecl] = ACTIONS(1302),
    [anon_sym___clrcall] = ACTIONS(1302),
    [anon_sym___stdcall] = ACTIONS(1302),
    [anon_sym___fastcall] = ACTIONS(1302),
    [anon_sym___thiscall] = ACTIONS(1302),
    [anon_sym___vectorcall] = ACTIONS(1302),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_signed] = ACTIONS(1302),
    [anon_sym_unsigned] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_auto] = ACTIONS(1302),
    [anon_sym_register] = ACTIONS(1302),
    [anon_sym_inline] = ACTIONS(1302),
    [anon_sym___inline] = ACTIONS(1302),
    [anon_sym___inline__] = ACTIONS(1302),
    [anon_sym___forceinline] = ACTIONS(1302),
    [anon_sym_thread_local] = ACTIONS(1302),
    [anon_sym___thread] = ACTIONS(1302),
    [anon_sym_const] = ACTIONS(1302),
    [anon_sym_constexpr] = ACTIONS(1302),
    [anon_sym_volatile] = ACTIONS(1302),
    [anon_sym_restrict] = ACTIONS(1302),
    [anon_sym___restrict__] = ACTIONS(1302),
    [anon_sym__Atomic] = ACTIONS(1302),
    [anon_sym__Noreturn] = ACTIONS(1302),
    [anon_sym_noreturn] = ACTIONS(1302),
    [anon_sym__Nonnull] = ACTIONS(1302),
    [anon_sym_alignas] = ACTIONS(1302),
    [anon_sym__Alignas] = ACTIONS(1302),
    [sym_primitive_type] = ACTIONS(1302),
    [anon_sym_enum] = ACTIONS(1302),
    [anon_sym_struct] = ACTIONS(1302),
    [anon_sym_union] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_goto] = ACTIONS(1302),
    [anon_sym___try] = ACTIONS(1302),
    [anon_sym___leave] = ACTIONS(1302),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_sizeof] = ACTIONS(1302),
    [anon_sym___alignof__] = ACTIONS(1302),
    [anon_sym___alignof] = ACTIONS(1302),
    [anon_sym__alignof] = ACTIONS(1302),
    [anon_sym_alignof] = ACTIONS(1302),
    [anon_sym__Alignof] = ACTIONS(1302),
    [anon_sym_offsetof] = ACTIONS(1302),
    [anon_sym__Generic] = ACTIONS(1302),
    [anon_sym_asm] = ACTIONS(1302),
    [anon_sym___asm__] = ACTIONS(1302),
    [anon_sym___asm] = ACTIONS(1302),
    [sym_number_literal] = ACTIONS(1304),
    [anon_sym_L_SQUOTE] = ACTIONS(1304),
    [anon_sym_u_SQUOTE] = ACTIONS(1304),
    [anon_sym_U_SQUOTE] = ACTIONS(1304),
    [anon_sym_u8_SQUOTE] = ACTIONS(1304),
    [anon_sym_SQUOTE] = ACTIONS(1304),
    [anon_sym_L_DQUOTE] = ACTIONS(1304),
    [anon_sym_u_DQUOTE] = ACTIONS(1304),
    [anon_sym_U_DQUOTE] = ACTIONS(1304),
    [anon_sym_u8_DQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [sym_true] = ACTIONS(1302),
    [sym_false] = ACTIONS(1302),
    [anon_sym_NULL] = ACTIONS(1302),
    [anon_sym_nullptr] = ACTIONS(1302),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(130)] = {
    [sym_identifier] = ACTIONS(1306),
    [aux_sym_preproc_include_token1] = ACTIONS(1306),
    [aux_sym_preproc_def_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token2] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1306),
    [aux_sym_preproc_else_token1] = ACTIONS(1306),
    [aux_sym_preproc_elif_token1] = ACTIONS(1306),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1306),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1306),
    [sym_preproc_directive] = ACTIONS(1306),
    [anon_sym_LPAREN2] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym___extension__] = ACTIONS(1306),
    [anon_sym_typedef] = ACTIONS(1306),
    [anon_sym_extern] = ACTIONS(1306),
    [anon_sym___attribute__] = ACTIONS(1306),
    [anon_sym___attribute] = ACTIONS(1306),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1308),
    [anon_sym___declspec] = ACTIONS(1306),
    [anon_sym___cdecl] = ACTIONS(1306),
    [anon_sym___clrcall] = ACTIONS(1306),
    [anon_sym___stdcall] = ACTIONS(1306),
    [anon_sym___fastcall] = ACTIONS(1306),
    [anon_sym___thiscall] = ACTIONS(1306),
    [anon_sym___vectorcall] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_signed] = ACTIONS(1306),
    [anon_sym_unsigned] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_auto] = ACTIONS(1306),
    [anon_sym_register] = ACTIONS(1306),
    [anon_sym_inline] = ACTIONS(1306),
    [anon_sym___inline] = ACTIONS(1306),
    [anon_sym___inline__] = ACTIONS(1306),
    [anon_sym___forceinline] = ACTIONS(1306),
    [anon_sym_thread_local] = ACTIONS(1306),
    [anon_sym___thread] = ACTIONS(1306),
    [anon_sym_const] = ACTIONS(1306),
    [anon_sym_constexpr] = ACTIONS(1306),
    [anon_sym_volatile] = ACTIONS(1306),
    [anon_sym_restrict] = ACTIONS(1306),
    [anon_sym___restrict__] = ACTIONS(1306),
    [anon_sym__Atomic] = ACTIONS(1306),
    [anon_sym__Noreturn] = ACTIONS(1306),
    [anon_sym_noreturn] = ACTIONS(1306),
    [anon_sym__Nonnull] = ACTIONS(1306),
    [anon_sym_alignas] = ACTIONS(1306),
    [anon_sym__Alignas] = ACTIONS(1306),
    [sym_primitive_type] = ACTIONS(1306),
    [anon_sym_enum] = ACTIONS(1306),
    [anon_sym_struct] = ACTIONS(1306),
    [anon_sym_union] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_goto] = ACTIONS(1306),
    [anon_sym___try] = ACTIONS(1306),
    [anon_sym___leave] = ACTIONS(1306),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_sizeof] = ACTIONS(1306),
    [anon_sym___alignof__] = ACTIONS(1306),
    [anon_sym___alignof] = ACTIONS(1306),
    [anon_sym__alignof] = ACTIONS(1306),
    [anon_sym_alignof] = ACTIONS(1306),
    [anon_sym__Alignof] = ACTIONS(1306),
    [anon_sym_offsetof] = ACTIONS(1306),
    [anon_sym__Generic] = ACTIONS(1306),
    [anon_sym_asm] = ACTIONS(1306),
    [anon_sym___asm__] = ACTIONS(1306),
    [anon_sym___asm] = ACTIONS(1306),
    [sym_number_literal] = ACTIONS(1308),
    [anon_sym_L_SQUOTE] = ACTIONS(1308),
    [anon_sym_u_SQUOTE] = ACTIONS(1308),
    [anon_sym_U_SQUOTE] = ACTIONS(1308),
    [anon_sym_u8_SQUOTE] = ACTIONS(1308),
    [anon_sym_SQUOTE] = ACTIONS(1308),
    [anon_sym_L_DQUOTE] = ACTIONS(1308),
    [anon_sym_u_DQUOTE] = ACTIONS(1308),
    [anon_sym_U_DQUOTE] = ACTIONS(1308),
    [anon_sym_u8_DQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_true] = ACTIONS(1306),
    [sym_false] = ACTIONS(1306),
    [anon_sym_NULL] = ACTIONS(1306),
    [anon_sym_nullptr] = ACTIONS(1306),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(131)] = {
    [sym_identifier] = ACTIONS(1310),
    [aux_sym_preproc_include_token1] = ACTIONS(1310),
    [aux_sym_preproc_def_token1] = ACTIONS(1310),
    [aux_sym_preproc_if_token1] = ACTIONS(1310),
    [aux_sym_preproc_if_token2] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1310),
    [aux_sym_preproc_else_token1] = ACTIONS(1310),
    [aux_sym_preproc_elif_token1] = ACTIONS(1310),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1310),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1310),
    [sym_preproc_directive] = ACTIONS(1310),
    [anon_sym_LPAREN2] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym___extension__] = ACTIONS(1310),
    [anon_sym_typedef] = ACTIONS(1310),
    [anon_sym_extern] = ACTIONS(1310),
    [anon_sym___attribute__] = ACTIONS(1310),
    [anon_sym___attribute] = ACTIONS(1310),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1312),
    [anon_sym___declspec] = ACTIONS(1310),
    [anon_sym___cdecl] = ACTIONS(1310),
    [anon_sym___clrcall] = ACTIONS(1310),
    [anon_sym___stdcall] = ACTIONS(1310),
    [anon_sym___fastcall] = ACTIONS(1310),
    [anon_sym___thiscall] = ACTIONS(1310),
    [anon_sym___vectorcall] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_signed] = ACTIONS(1310),
    [anon_sym_unsigned] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_auto] = ACTIONS(1310),
    [anon_sym_register] = ACTIONS(1310),
    [anon_sym_inline] = ACTIONS(1310),
    [anon_sym___inline] = ACTIONS(1310),
    [anon_sym___inline__] = ACTIONS(1310),
    [anon_sym___forceinline] = ACTIONS(1310),
    [anon_sym_thread_local] = ACTIONS(1310),
    [anon_sym___thread] = ACTIONS(1310),
    [anon_sym_const] = ACTIONS(1310),
    [anon_sym_constexpr] = ACTIONS(1310),
    [anon_sym_volatile] = ACTIONS(1310),
    [anon_sym_restrict] = ACTIONS(1310),
    [anon_sym___restrict__] = ACTIONS(1310),
    [anon_sym__Atomic] = ACTIONS(1310),
    [anon_sym__Noreturn] = ACTIONS(1310),
    [anon_sym_noreturn] = ACTIONS(1310),
    [anon_sym__Nonnull] = ACTIONS(1310),
    [anon_sym_alignas] = ACTIONS(1310),
    [anon_sym__Alignas] = ACTIONS(1310),
    [sym_primitive_type] = ACTIONS(1310),
    [anon_sym_enum] = ACTIONS(1310),
    [anon_sym_struct] = ACTIONS(1310),
    [anon_sym_union] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_goto] = ACTIONS(1310),
    [anon_sym___try] = ACTIONS(1310),
    [anon_sym___leave] = ACTIONS(1310),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_sizeof] = ACTIONS(1310),
    [anon_sym___alignof__] = ACTIONS(1310),
    [anon_sym___alignof] = ACTIONS(1310),
    [anon_sym__alignof] = ACTIONS(1310),
    [anon_sym_alignof] = ACTIONS(1310),
    [anon_sym__Alignof] = ACTIONS(1310),
    [anon_sym_offsetof] = ACTIONS(1310),
    [anon_sym__Generic] = ACTIONS(1310),
    [anon_sym_asm] = ACTIONS(1310),
    [anon_sym___asm__] = ACTIONS(1310),
    [anon_sym___asm] = ACTIONS(1310),
    [sym_number_literal] = ACTIONS(1312),
    [anon_sym_L_SQUOTE] = ACTIONS(1312),
    [anon_sym_u_SQUOTE] = ACTIONS(1312),
    [anon_sym_U_SQUOTE] = ACTIONS(1312),
    [anon_sym_u8_SQUOTE] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_L_DQUOTE] = ACTIONS(1312),
    [anon_sym_u_DQUOTE] = ACTIONS(1312),
    [anon_sym_U_DQUOTE] = ACTIONS(1312),
    [anon_sym_u8_DQUOTE] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [sym_true] = ACTIONS(1310),
    [sym_false] = ACTIONS(1310),
    [anon_sym_NULL] = ACTIONS(1310),
    [anon_sym_nullptr] = ACTIONS(1310),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(132)] = {
    [sym_identifier] = ACTIONS(1314),
    [aux_sym_preproc_include_token1] = ACTIONS(1314),
    [aux_sym_preproc_def_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token2] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1314),
    [aux_sym_preproc_else_token1] = ACTIONS(1314),
    [aux_sym_preproc_elif_token1] = ACTIONS(1314),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1314),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1314),
    [sym_preproc_directive] = ACTIONS(1314),
    [anon_sym_LPAREN2] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1314),
    [anon_sym_PLUS] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym___extension__] = ACTIONS(1314),
    [anon_sym_typedef] = ACTIONS(1314),
    [anon_sym_extern] = ACTIONS(1314),
    [anon_sym___attribute__] = ACTIONS(1314),
    [anon_sym___attribute] = ACTIONS(1314),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1316),
    [anon_sym___declspec] = ACTIONS(1314),
    [anon_sym___cdecl] = ACTIONS(1314),
    [anon_sym___clrcall] = ACTIONS(1314),
    [anon_sym___stdcall] = ACTIONS(1314),
    [anon_sym___fastcall] = ACTIONS(1314),
    [anon_sym___thiscall] = ACTIONS(1314),
    [anon_sym___vectorcall] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_signed] = ACTIONS(1314),
    [anon_sym_unsigned] = ACTIONS(1314),
    [anon_sym_long] = ACTIONS(1314),
    [anon_sym_short] = ACTIONS(1314),
    [anon_sym_static] = ACTIONS(1314),
    [anon_sym_auto] = ACTIONS(1314),
    [anon_sym_register] = ACTIONS(1314),
    [anon_sym_inline] = ACTIONS(1314),
    [anon_sym___inline] = ACTIONS(1314),
    [anon_sym___inline__] = ACTIONS(1314),
    [anon_sym___forceinline] = ACTIONS(1314),
    [anon_sym_thread_local] = ACTIONS(1314),
    [anon_sym___thread] = ACTIONS(1314),
    [anon_sym_const] = ACTIONS(1314),
    [anon_sym_constexpr] = ACTIONS(1314),
    [anon_sym_volatile] = ACTIONS(1314),
    [anon_sym_restrict] = ACTIONS(1314),
    [anon_sym___restrict__] = ACTIONS(1314),
    [anon_sym__Atomic] = ACTIONS(1314),
    [anon_sym__Noreturn] = ACTIONS(1314),
    [anon_sym_noreturn] = ACTIONS(1314),
    [anon_sym__Nonnull] = ACTIONS(1314),
    [anon_sym_alignas] = ACTIONS(1314),
    [anon_sym__Alignas] = ACTIONS(1314),
    [sym_primitive_type] = ACTIONS(1314),
    [anon_sym_enum] = ACTIONS(1314),
    [anon_sym_struct] = ACTIONS(1314),
    [anon_sym_union] = ACTIONS(1314),
    [anon_sym_if] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1314),
    [anon_sym_case] = ACTIONS(1314),
    [anon_sym_default] = ACTIONS(1314),
    [anon_sym_while] = ACTIONS(1314),
    [anon_sym_do] = ACTIONS(1314),
    [anon_sym_for] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1314),
    [anon_sym_break] = ACTIONS(1314),
    [anon_sym_continue] = ACTIONS(1314),
    [anon_sym_goto] = ACTIONS(1314),
    [anon_sym___try] = ACTIONS(1314),
    [anon_sym___leave] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_sizeof] = ACTIONS(1314),
    [anon_sym___alignof__] = ACTIONS(1314),
    [anon_sym___alignof] = ACTIONS(1314),
    [anon_sym__alignof] = ACTIONS(1314),
    [anon_sym_alignof] = ACTIONS(1314),
    [anon_sym__Alignof] = ACTIONS(1314),
    [anon_sym_offsetof] = ACTIONS(1314),
    [anon_sym__Generic] = ACTIONS(1314),
    [anon_sym_asm] = ACTIONS(1314),
    [anon_sym___asm__] = ACTIONS(1314),
    [anon_sym___asm] = ACTIONS(1314),
    [sym_number_literal] = ACTIONS(1316),
    [anon_sym_L_SQUOTE] = ACTIONS(1316),
    [anon_sym_u_SQUOTE] = ACTIONS(1316),
    [anon_sym_U_SQUOTE] = ACTIONS(1316),
    [anon_sym_u8_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_L_DQUOTE] = ACTIONS(1316),
    [anon_sym_u_DQUOTE] = ACTIONS(1316),
    [anon_sym_U_DQUOTE] = ACTIONS(1316),
    [anon_sym_u8_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [sym_true] = ACTIONS(1314),
    [sym_false] = ACTIONS(1314),
    [anon_sym_NULL] = ACTIONS(1314),
    [anon_sym_nullptr] = ACTIONS(1314),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(133)] = {
    [sym_identifier] = ACTIONS(1318),
    [aux_sym_preproc_include_token1] = ACTIONS(1318),
    [aux_sym_preproc_def_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token2] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1318),
    [aux_sym_preproc_else_token1] = ACTIONS(1318),
    [aux_sym_preproc_elif_token1] = ACTIONS(1318),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1318),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1318),
    [sym_preproc_directive] = ACTIONS(1318),
    [anon_sym_LPAREN2] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym___extension__] = ACTIONS(1318),
    [anon_sym_typedef] = ACTIONS(1318),
    [anon_sym_extern] = ACTIONS(1318),
    [anon_sym___attribute__] = ACTIONS(1318),
    [anon_sym___attribute] = ACTIONS(1318),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1320),
    [anon_sym___declspec] = ACTIONS(1318),
    [anon_sym___cdecl] = ACTIONS(1318),
    [anon_sym___clrcall] = ACTIONS(1318),
    [anon_sym___stdcall] = ACTIONS(1318),
    [anon_sym___fastcall] = ACTIONS(1318),
    [anon_sym___thiscall] = ACTIONS(1318),
    [anon_sym___vectorcall] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_signed] = ACTIONS(1318),
    [anon_sym_unsigned] = ACTIONS(1318),
    [anon_sym_long] = ACTIONS(1318),
    [anon_sym_short] = ACTIONS(1318),
    [anon_sym_static] = ACTIONS(1318),
    [anon_sym_auto] = ACTIONS(1318),
    [anon_sym_register] = ACTIONS(1318),
    [anon_sym_inline] = ACTIONS(1318),
    [anon_sym___inline] = ACTIONS(1318),
    [anon_sym___inline__] = ACTIONS(1318),
    [anon_sym___forceinline] = ACTIONS(1318),
    [anon_sym_thread_local] = ACTIONS(1318),
    [anon_sym___thread] = ACTIONS(1318),
    [anon_sym_const] = ACTIONS(1318),
    [anon_sym_constexpr] = ACTIONS(1318),
    [anon_sym_volatile] = ACTIONS(1318),
    [anon_sym_restrict] = ACTIONS(1318),
    [anon_sym___restrict__] = ACTIONS(1318),
    [anon_sym__Atomic] = ACTIONS(1318),
    [anon_sym__Noreturn] = ACTIONS(1318),
    [anon_sym_noreturn] = ACTIONS(1318),
    [anon_sym__Nonnull] = ACTIONS(1318),
    [anon_sym_alignas] = ACTIONS(1318),
    [anon_sym__Alignas] = ACTIONS(1318),
    [sym_primitive_type] = ACTIONS(1318),
    [anon_sym_enum] = ACTIONS(1318),
    [anon_sym_struct] = ACTIONS(1318),
    [anon_sym_union] = ACTIONS(1318),
    [anon_sym_if] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1318),
    [anon_sym_case] = ACTIONS(1318),
    [anon_sym_default] = ACTIONS(1318),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_break] = ACTIONS(1318),
    [anon_sym_continue] = ACTIONS(1318),
    [anon_sym_goto] = ACTIONS(1318),
    [anon_sym___try] = ACTIONS(1318),
    [anon_sym___leave] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_sizeof] = ACTIONS(1318),
    [anon_sym___alignof__] = ACTIONS(1318),
    [anon_sym___alignof] = ACTIONS(1318),
    [anon_sym__alignof] = ACTIONS(1318),
    [anon_sym_alignof] = ACTIONS(1318),
    [anon_sym__Alignof] = ACTIONS(1318),
    [anon_sym_offsetof] = ACTIONS(1318),
    [anon_sym__Generic] = ACTIONS(1318),
    [anon_sym_asm] = ACTIONS(1318),
    [anon_sym___asm__] = ACTIONS(1318),
    [anon_sym___asm] = ACTIONS(1318),
    [sym_number_literal] = ACTIONS(1320),
    [anon_sym_L_SQUOTE] = ACTIONS(1320),
    [anon_sym_u_SQUOTE] = ACTIONS(1320),
    [anon_sym_U_SQUOTE] = ACTIONS(1320),
    [anon_sym_u8_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_L_DQUOTE] = ACTIONS(1320),
    [anon_sym_u_DQUOTE] = ACTIONS(1320),
    [anon_sym_U_DQUOTE] = ACTIONS(1320),
    [anon_sym_u8_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [anon_sym_NULL] = ACTIONS(1318),
    [anon_sym_nullptr] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(134)] = {
    [sym_identifier] = ACTIONS(1322),
    [aux_sym_preproc_include_token1] = ACTIONS(1322),
    [aux_sym_preproc_def_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token2] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1322),
    [aux_sym_preproc_else_token1] = ACTIONS(1322),
    [aux_sym_preproc_elif_token1] = ACTIONS(1322),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1322),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1322),
    [sym_preproc_directive] = ACTIONS(1322),
    [anon_sym_LPAREN2] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1322),
    [anon_sym_PLUS] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym___extension__] = ACTIONS(1322),
    [anon_sym_typedef] = ACTIONS(1322),
    [anon_sym_extern] = ACTIONS(1322),
    [anon_sym___attribute__] = ACTIONS(1322),
    [anon_sym___attribute] = ACTIONS(1322),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1324),
    [anon_sym___declspec] = ACTIONS(1322),
    [anon_sym___cdecl] = ACTIONS(1322),
    [anon_sym___clrcall] = ACTIONS(1322),
    [anon_sym___stdcall] = ACTIONS(1322),
    [anon_sym___fastcall] = ACTIONS(1322),
    [anon_sym___thiscall] = ACTIONS(1322),
    [anon_sym___vectorcall] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_signed] = ACTIONS(1322),
    [anon_sym_unsigned] = ACTIONS(1322),
    [anon_sym_long] = ACTIONS(1322),
    [anon_sym_short] = ACTIONS(1322),
    [anon_sym_static] = ACTIONS(1322),
    [anon_sym_auto] = ACTIONS(1322),
    [anon_sym_register] = ACTIONS(1322),
    [anon_sym_inline] = ACTIONS(1322),
    [anon_sym___inline] = ACTIONS(1322),
    [anon_sym___inline__] = ACTIONS(1322),
    [anon_sym___forceinline] = ACTIONS(1322),
    [anon_sym_thread_local] = ACTIONS(1322),
    [anon_sym___thread] = ACTIONS(1322),
    [anon_sym_const] = ACTIONS(1322),
    [anon_sym_constexpr] = ACTIONS(1322),
    [anon_sym_volatile] = ACTIONS(1322),
    [anon_sym_restrict] = ACTIONS(1322),
    [anon_sym___restrict__] = ACTIONS(1322),
    [anon_sym__Atomic] = ACTIONS(1322),
    [anon_sym__Noreturn] = ACTIONS(1322),
    [anon_sym_noreturn] = ACTIONS(1322),
    [anon_sym__Nonnull] = ACTIONS(1322),
    [anon_sym_alignas] = ACTIONS(1322),
    [anon_sym__Alignas] = ACTIONS(1322),
    [sym_primitive_type] = ACTIONS(1322),
    [anon_sym_enum] = ACTIONS(1322),
    [anon_sym_struct] = ACTIONS(1322),
    [anon_sym_union] = ACTIONS(1322),
    [anon_sym_if] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1322),
    [anon_sym_case] = ACTIONS(1322),
    [anon_sym_default] = ACTIONS(1322),
    [anon_sym_while] = ACTIONS(1322),
    [anon_sym_do] = ACTIONS(1322),
    [anon_sym_for] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1322),
    [anon_sym_break] = ACTIONS(1322),
    [anon_sym_continue] = ACTIONS(1322),
    [anon_sym_goto] = ACTIONS(1322),
    [anon_sym___try] = ACTIONS(1322),
    [anon_sym___leave] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_sizeof] = ACTIONS(1322),
    [anon_sym___alignof__] = ACTIONS(1322),
    [anon_sym___alignof] = ACTIONS(1322),
    [anon_sym__alignof] = ACTIONS(1322),
    [anon_sym_alignof] = ACTIONS(1322),
    [anon_sym__Alignof] = ACTIONS(1322),
    [anon_sym_offsetof] = ACTIONS(1322),
    [anon_sym__Generic] = ACTIONS(1322),
    [anon_sym_asm] = ACTIONS(1322),
    [anon_sym___asm__] = ACTIONS(1322),
    [anon_sym___asm] = ACTIONS(1322),
    [sym_number_literal] = ACTIONS(1324),
    [anon_sym_L_SQUOTE] = ACTIONS(1324),
    [anon_sym_u_SQUOTE] = ACTIONS(1324),
    [anon_sym_U_SQUOTE] = ACTIONS(1324),
    [anon_sym_u8_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_L_DQUOTE] = ACTIONS(1324),
    [anon_sym_u_DQUOTE] = ACTIONS(1324),
    [anon_sym_U_DQUOTE] = ACTIONS(1324),
    [anon_sym_u8_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [sym_true] = ACTIONS(1322),
    [sym_false] = ACTIONS(1322),
    [anon_sym_NULL] = ACTIONS(1322),
    [anon_sym_nullptr] = ACTIONS(1322),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(135)] = {
    [sym_identifier] = ACTIONS(1326),
    [aux_sym_preproc_include_token1] = ACTIONS(1326),
    [aux_sym_preproc_def_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token2] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1326),
    [aux_sym_preproc_else_token1] = ACTIONS(1326),
    [aux_sym_preproc_elif_token1] = ACTIONS(1326),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1326),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1326),
    [sym_preproc_directive] = ACTIONS(1326),
    [anon_sym_LPAREN2] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1326),
    [anon_sym_PLUS] = ACTIONS(1326),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym___extension__] = ACTIONS(1326),
    [anon_sym_typedef] = ACTIONS(1326),
    [anon_sym_extern] = ACTIONS(1326),
    [anon_sym___attribute__] = ACTIONS(1326),
    [anon_sym___attribute] = ACTIONS(1326),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1328),
    [anon_sym___declspec] = ACTIONS(1326),
    [anon_sym___cdecl] = ACTIONS(1326),
    [anon_sym___clrcall] = ACTIONS(1326),
    [anon_sym___stdcall] = ACTIONS(1326),
    [anon_sym___fastcall] = ACTIONS(1326),
    [anon_sym___thiscall] = ACTIONS(1326),
    [anon_sym___vectorcall] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_signed] = ACTIONS(1326),
    [anon_sym_unsigned] = ACTIONS(1326),
    [anon_sym_long] = ACTIONS(1326),
    [anon_sym_short] = ACTIONS(1326),
    [anon_sym_static] = ACTIONS(1326),
    [anon_sym_auto] = ACTIONS(1326),
    [anon_sym_register] = ACTIONS(1326),
    [anon_sym_inline] = ACTIONS(1326),
    [anon_sym___inline] = ACTIONS(1326),
    [anon_sym___inline__] = ACTIONS(1326),
    [anon_sym___forceinline] = ACTIONS(1326),
    [anon_sym_thread_local] = ACTIONS(1326),
    [anon_sym___thread] = ACTIONS(1326),
    [anon_sym_const] = ACTIONS(1326),
    [anon_sym_constexpr] = ACTIONS(1326),
    [anon_sym_volatile] = ACTIONS(1326),
    [anon_sym_restrict] = ACTIONS(1326),
    [anon_sym___restrict__] = ACTIONS(1326),
    [anon_sym__Atomic] = ACTIONS(1326),
    [anon_sym__Noreturn] = ACTIONS(1326),
    [anon_sym_noreturn] = ACTIONS(1326),
    [anon_sym__Nonnull] = ACTIONS(1326),
    [anon_sym_alignas] = ACTIONS(1326),
    [anon_sym__Alignas] = ACTIONS(1326),
    [sym_primitive_type] = ACTIONS(1326),
    [anon_sym_enum] = ACTIONS(1326),
    [anon_sym_struct] = ACTIONS(1326),
    [anon_sym_union] = ACTIONS(1326),
    [anon_sym_if] = ACTIONS(1326),
    [anon_sym_switch] = ACTIONS(1326),
    [anon_sym_case] = ACTIONS(1326),
    [anon_sym_default] = ACTIONS(1326),
    [anon_sym_while] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1326),
    [anon_sym_for] = ACTIONS(1326),
    [anon_sym_return] = ACTIONS(1326),
    [anon_sym_break] = ACTIONS(1326),
    [anon_sym_continue] = ACTIONS(1326),
    [anon_sym_goto] = ACTIONS(1326),
    [anon_sym___try] = ACTIONS(1326),
    [anon_sym___leave] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_sizeof] = ACTIONS(1326),
    [anon_sym___alignof__] = ACTIONS(1326),
    [anon_sym___alignof] = ACTIONS(1326),
    [anon_sym__alignof] = ACTIONS(1326),
    [anon_sym_alignof] = ACTIONS(1326),
    [anon_sym__Alignof] = ACTIONS(1326),
    [anon_sym_offsetof] = ACTIONS(1326),
    [anon_sym__Generic] = ACTIONS(1326),
    [anon_sym_asm] = ACTIONS(1326),
    [anon_sym___asm__] = ACTIONS(1326),
    [anon_sym___asm] = ACTIONS(1326),
    [sym_number_literal] = ACTIONS(1328),
    [anon_sym_L_SQUOTE] = ACTIONS(1328),
    [anon_sym_u_SQUOTE] = ACTIONS(1328),
    [anon_sym_U_SQUOTE] = ACTIONS(1328),
    [anon_sym_u8_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_L_DQUOTE] = ACTIONS(1328),
    [anon_sym_u_DQUOTE] = ACTIONS(1328),
    [anon_sym_U_DQUOTE] = ACTIONS(1328),
    [anon_sym_u8_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [sym_true] = ACTIONS(1326),
    [sym_false] = ACTIONS(1326),
    [anon_sym_NULL] = ACTIONS(1326),
    [anon_sym_nullptr] = ACTIONS(1326),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(136)] = {
    [sym_identifier] = ACTIONS(1330),
    [aux_sym_preproc_include_token1] = ACTIONS(1330),
    [aux_sym_preproc_def_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token2] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1330),
    [aux_sym_preproc_else_token1] = ACTIONS(1330),
    [aux_sym_preproc_elif_token1] = ACTIONS(1330),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1330),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1330),
    [sym_preproc_directive] = ACTIONS(1330),
    [anon_sym_LPAREN2] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1330),
    [anon_sym_PLUS] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym___extension__] = ACTIONS(1330),
    [anon_sym_typedef] = ACTIONS(1330),
    [anon_sym_extern] = ACTIONS(1330),
    [anon_sym___attribute__] = ACTIONS(1330),
    [anon_sym___attribute] = ACTIONS(1330),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1332),
    [anon_sym___declspec] = ACTIONS(1330),
    [anon_sym___cdecl] = ACTIONS(1330),
    [anon_sym___clrcall] = ACTIONS(1330),
    [anon_sym___stdcall] = ACTIONS(1330),
    [anon_sym___fastcall] = ACTIONS(1330),
    [anon_sym___thiscall] = ACTIONS(1330),
    [anon_sym___vectorcall] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_signed] = ACTIONS(1330),
    [anon_sym_unsigned] = ACTIONS(1330),
    [anon_sym_long] = ACTIONS(1330),
    [anon_sym_short] = ACTIONS(1330),
    [anon_sym_static] = ACTIONS(1330),
    [anon_sym_auto] = ACTIONS(1330),
    [anon_sym_register] = ACTIONS(1330),
    [anon_sym_inline] = ACTIONS(1330),
    [anon_sym___inline] = ACTIONS(1330),
    [anon_sym___inline__] = ACTIONS(1330),
    [anon_sym___forceinline] = ACTIONS(1330),
    [anon_sym_thread_local] = ACTIONS(1330),
    [anon_sym___thread] = ACTIONS(1330),
    [anon_sym_const] = ACTIONS(1330),
    [anon_sym_constexpr] = ACTIONS(1330),
    [anon_sym_volatile] = ACTIONS(1330),
    [anon_sym_restrict] = ACTIONS(1330),
    [anon_sym___restrict__] = ACTIONS(1330),
    [anon_sym__Atomic] = ACTIONS(1330),
    [anon_sym__Noreturn] = ACTIONS(1330),
    [anon_sym_noreturn] = ACTIONS(1330),
    [anon_sym__Nonnull] = ACTIONS(1330),
    [anon_sym_alignas] = ACTIONS(1330),
    [anon_sym__Alignas] = ACTIONS(1330),
    [sym_primitive_type] = ACTIONS(1330),
    [anon_sym_enum] = ACTIONS(1330),
    [anon_sym_struct] = ACTIONS(1330),
    [anon_sym_union] = ACTIONS(1330),
    [anon_sym_if] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1330),
    [anon_sym_case] = ACTIONS(1330),
    [anon_sym_default] = ACTIONS(1330),
    [anon_sym_while] = ACTIONS(1330),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_for] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1330),
    [anon_sym_break] = ACTIONS(1330),
    [anon_sym_continue] = ACTIONS(1330),
    [anon_sym_goto] = ACTIONS(1330),
    [anon_sym___try] = ACTIONS(1330),
    [anon_sym___leave] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_sizeof] = ACTIONS(1330),
    [anon_sym___alignof__] = ACTIONS(1330),
    [anon_sym___alignof] = ACTIONS(1330),
    [anon_sym__alignof] = ACTIONS(1330),
    [anon_sym_alignof] = ACTIONS(1330),
    [anon_sym__Alignof] = ACTIONS(1330),
    [anon_sym_offsetof] = ACTIONS(1330),
    [anon_sym__Generic] = ACTIONS(1330),
    [anon_sym_asm] = ACTIONS(1330),
    [anon_sym___asm__] = ACTIONS(1330),
    [anon_sym___asm] = ACTIONS(1330),
    [sym_number_literal] = ACTIONS(1332),
    [anon_sym_L_SQUOTE] = ACTIONS(1332),
    [anon_sym_u_SQUOTE] = ACTIONS(1332),
    [anon_sym_U_SQUOTE] = ACTIONS(1332),
    [anon_sym_u8_SQUOTE] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_L_DQUOTE] = ACTIONS(1332),
    [anon_sym_u_DQUOTE] = ACTIONS(1332),
    [anon_sym_U_DQUOTE] = ACTIONS(1332),
    [anon_sym_u8_DQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [sym_true] = ACTIONS(1330),
    [sym_false] = ACTIONS(1330),
    [anon_sym_NULL] = ACTIONS(1330),
    [anon_sym_nullptr] = ACTIONS(1330),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(137)] = {
    [sym_identifier] = ACTIONS(1334),
    [aux_sym_preproc_include_token1] = ACTIONS(1334),
    [aux_sym_preproc_def_token1] = ACTIONS(1334),
    [aux_sym_preproc_if_token1] = ACTIONS(1334),
    [aux_sym_preproc_if_token2] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1334),
    [aux_sym_preproc_else_token1] = ACTIONS(1334),
    [aux_sym_preproc_elif_token1] = ACTIONS(1334),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1334),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1334),
    [sym_preproc_directive] = ACTIONS(1334),
    [anon_sym_LPAREN2] = ACTIONS(1337),
    [anon_sym_BANG] = ACTIONS(1337),
    [anon_sym_TILDE] = ACTIONS(1337),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_STAR] = ACTIONS(1337),
    [anon_sym_AMP] = ACTIONS(1337),
    [anon_sym_SEMI] = ACTIONS(1337),
    [anon_sym___extension__] = ACTIONS(1334),
    [anon_sym_typedef] = ACTIONS(1334),
    [anon_sym_extern] = ACTIONS(1334),
    [anon_sym___attribute__] = ACTIONS(1334),
    [anon_sym___attribute] = ACTIONS(1334),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1337),
    [anon_sym___declspec] = ACTIONS(1334),
    [anon_sym___cdecl] = ACTIONS(1334),
    [anon_sym___clrcall] = ACTIONS(1334),
    [anon_sym___stdcall] = ACTIONS(1334),
    [anon_sym___fastcall] = ACTIONS(1334),
    [anon_sym___thiscall] = ACTIONS(1334),
    [anon_sym___vectorcall] = ACTIONS(1334),
    [anon_sym_LBRACE] = ACTIONS(1337),
    [anon_sym_signed] = ACTIONS(1334),
    [anon_sym_unsigned] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_auto] = ACTIONS(1334),
    [anon_sym_register] = ACTIONS(1334),
    [anon_sym_inline] = ACTIONS(1334),
    [anon_sym___inline] = ACTIONS(1334),
    [anon_sym___inline__] = ACTIONS(1334),
    [anon_sym___forceinline] = ACTIONS(1334),
    [anon_sym_thread_local] = ACTIONS(1334),
    [anon_sym___thread] = ACTIONS(1334),
    [anon_sym_const] = ACTIONS(1334),
    [anon_sym_constexpr] = ACTIONS(1334),
    [anon_sym_volatile] = ACTIONS(1334),
    [anon_sym_restrict] = ACTIONS(1334),
    [anon_sym___restrict__] = ACTIONS(1334),
    [anon_sym__Atomic] = ACTIONS(1334),
    [anon_sym__Noreturn] = ACTIONS(1334),
    [anon_sym_noreturn] = ACTIONS(1334),
    [anon_sym__Nonnull] = ACTIONS(1334),
    [anon_sym_alignas] = ACTIONS(1334),
    [anon_sym__Alignas] = ACTIONS(1334),
    [sym_primitive_type] = ACTIONS(1334),
    [anon_sym_enum] = ACTIONS(1334),
    [anon_sym_struct] = ACTIONS(1334),
    [anon_sym_union] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_goto] = ACTIONS(1334),
    [anon_sym___try] = ACTIONS(1334),
    [anon_sym___leave] = ACTIONS(1334),
    [anon_sym_DASH_DASH] = ACTIONS(1337),
    [anon_sym_PLUS_PLUS] = ACTIONS(1337),
    [anon_sym_sizeof] = ACTIONS(1334),
    [anon_sym___alignof__] = ACTIONS(1334),
    [anon_sym___alignof] = ACTIONS(1334),
    [anon_sym__alignof] = ACTIONS(1334),
    [anon_sym_alignof] = ACTIONS(1334),
    [anon_sym__Alignof] = ACTIONS(1334),
    [anon_sym_offsetof] = ACTIONS(1334),
    [anon_sym__Generic] = ACTIONS(1334),
    [anon_sym_asm] = ACTIONS(1334),
    [anon_sym___asm__] = ACTIONS(1334),
    [anon_sym___asm] = ACTIONS(1334),
    [sym_number_literal] = ACTIONS(1337),
    [anon_sym_L_SQUOTE] = ACTIONS(1337),
    [anon_sym_u_SQUOTE] = ACTIONS(1337),
    [anon_sym_U_SQUOTE] = ACTIONS(1337),
    [anon_sym_u8_SQUOTE] = ACTIONS(1337),
    [anon_sym_SQUOTE] = ACTIONS(1337),
    [anon_sym_L_DQUOTE] = ACTIONS(1337),
    [anon_sym_u_DQUOTE] = ACTIONS(1337),
    [anon_sym_U_DQUOTE] = ACTIONS(1337),
    [anon_sym_u8_DQUOTE] = ACTIONS(1337),
    [anon_sym_DQUOTE] = ACTIONS(1337),
    [sym_true] = ACTIONS(1334),
    [sym_false] = ACTIONS(1334),
    [anon_sym_NULL] = ACTIONS(1334),
    [anon_sym_nullptr] = ACTIONS(1334),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(138)] = {
    [sym_identifier] = ACTIONS(1340),
    [aux_sym_preproc_include_token1] = ACTIONS(1340),
    [aux_sym_preproc_def_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token2] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1340),
    [aux_sym_preproc_else_token1] = ACTIONS(1340),
    [aux_sym_preproc_elif_token1] = ACTIONS(1340),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1340),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1340),
    [sym_preproc_directive] = ACTIONS(1340),
    [anon_sym_LPAREN2] = ACTIONS(1342),
    [anon_sym_BANG] = ACTIONS(1342),
    [anon_sym_TILDE] = ACTIONS(1342),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1342),
    [anon_sym_AMP] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1342),
    [anon_sym___extension__] = ACTIONS(1340),
    [anon_sym_typedef] = ACTIONS(1340),
    [anon_sym_extern] = ACTIONS(1340),
    [anon_sym___attribute__] = ACTIONS(1340),
    [anon_sym___attribute] = ACTIONS(1340),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1342),
    [anon_sym___declspec] = ACTIONS(1340),
    [anon_sym___cdecl] = ACTIONS(1340),
    [anon_sym___clrcall] = ACTIONS(1340),
    [anon_sym___stdcall] = ACTIONS(1340),
    [anon_sym___fastcall] = ACTIONS(1340),
    [anon_sym___thiscall] = ACTIONS(1340),
    [anon_sym___vectorcall] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1342),
    [anon_sym_signed] = ACTIONS(1340),
    [anon_sym_unsigned] = ACTIONS(1340),
    [anon_sym_long] = ACTIONS(1340),
    [anon_sym_short] = ACTIONS(1340),
    [anon_sym_static] = ACTIONS(1340),
    [anon_sym_auto] = ACTIONS(1340),
    [anon_sym_register] = ACTIONS(1340),
    [anon_sym_inline] = ACTIONS(1340),
    [anon_sym___inline] = ACTIONS(1340),
    [anon_sym___inline__] = ACTIONS(1340),
    [anon_sym___forceinline] = ACTIONS(1340),
    [anon_sym_thread_local] = ACTIONS(1340),
    [anon_sym___thread] = ACTIONS(1340),
    [anon_sym_const] = ACTIONS(1340),
    [anon_sym_constexpr] = ACTIONS(1340),
    [anon_sym_volatile] = ACTIONS(1340),
    [anon_sym_restrict] = ACTIONS(1340),
    [anon_sym___restrict__] = ACTIONS(1340),
    [anon_sym__Atomic] = ACTIONS(1340),
    [anon_sym__Noreturn] = ACTIONS(1340),
    [anon_sym_noreturn] = ACTIONS(1340),
    [anon_sym__Nonnull] = ACTIONS(1340),
    [anon_sym_alignas] = ACTIONS(1340),
    [anon_sym__Alignas] = ACTIONS(1340),
    [sym_primitive_type] = ACTIONS(1340),
    [anon_sym_enum] = ACTIONS(1340),
    [anon_sym_struct] = ACTIONS(1340),
    [anon_sym_union] = ACTIONS(1340),
    [anon_sym_if] = ACTIONS(1340),
    [anon_sym_switch] = ACTIONS(1340),
    [anon_sym_case] = ACTIONS(1340),
    [anon_sym_default] = ACTIONS(1340),
    [anon_sym_while] = ACTIONS(1340),
    [anon_sym_do] = ACTIONS(1340),
    [anon_sym_for] = ACTIONS(1340),
    [anon_sym_return] = ACTIONS(1340),
    [anon_sym_break] = ACTIONS(1340),
    [anon_sym_continue] = ACTIONS(1340),
    [anon_sym_goto] = ACTIONS(1340),
    [anon_sym___try] = ACTIONS(1340),
    [anon_sym___leave] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1342),
    [anon_sym_PLUS_PLUS] = ACTIONS(1342),
    [anon_sym_sizeof] = ACTIONS(1340),
    [anon_sym___alignof__] = ACTIONS(1340),
    [anon_sym___alignof] = ACTIONS(1340),
    [anon_sym__alignof] = ACTIONS(1340),
    [anon_sym_alignof] = ACTIONS(1340),
    [anon_sym__Alignof] = ACTIONS(1340),
    [anon_sym_offsetof] = ACTIONS(1340),
    [anon_sym__Generic] = ACTIONS(1340),
    [anon_sym_asm] = ACTIONS(1340),
    [anon_sym___asm__] = ACTIONS(1340),
    [anon_sym___asm] = ACTIONS(1340),
    [sym_number_literal] = ACTIONS(1342),
    [anon_sym_L_SQUOTE] = ACTIONS(1342),
    [anon_sym_u_SQUOTE] = ACTIONS(1342),
    [anon_sym_U_SQUOTE] = ACTIONS(1342),
    [anon_sym_u8_SQUOTE] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_L_DQUOTE] = ACTIONS(1342),
    [anon_sym_u_DQUOTE] = ACTIONS(1342),
    [anon_sym_U_DQUOTE] = ACTIONS(1342),
    [anon_sym_u8_DQUOTE] = ACTIONS(1342),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [sym_true] = ACTIONS(1340),
    [sym_false] = ACTIONS(1340),
    [anon_sym_NULL] = ACTIONS(1340),
    [anon_sym_nullptr] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(139)] = {
    [sym_identifier] = ACTIONS(1344),
    [aux_sym_preproc_include_token1] = ACTIONS(1344),
    [aux_sym_preproc_def_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token2] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1344),
    [aux_sym_preproc_else_token1] = ACTIONS(1344),
    [aux_sym_preproc_elif_token1] = ACTIONS(1344),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1344),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1344),
    [sym_preproc_directive] = ACTIONS(1344),
    [anon_sym_LPAREN2] = ACTIONS(1346),
    [anon_sym_BANG] = ACTIONS(1346),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1346),
    [anon_sym_AMP] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1346),
    [anon_sym___extension__] = ACTIONS(1344),
    [anon_sym_typedef] = ACTIONS(1344),
    [anon_sym_extern] = ACTIONS(1344),
    [anon_sym___attribute__] = ACTIONS(1344),
    [anon_sym___attribute] = ACTIONS(1344),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1346),
    [anon_sym___declspec] = ACTIONS(1344),
    [anon_sym___cdecl] = ACTIONS(1344),
    [anon_sym___clrcall] = ACTIONS(1344),
    [anon_sym___stdcall] = ACTIONS(1344),
    [anon_sym___fastcall] = ACTIONS(1344),
    [anon_sym___thiscall] = ACTIONS(1344),
    [anon_sym___vectorcall] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [anon_sym_signed] = ACTIONS(1344),
    [anon_sym_unsigned] = ACTIONS(1344),
    [anon_sym_long] = ACTIONS(1344),
    [anon_sym_short] = ACTIONS(1344),
    [anon_sym_static] = ACTIONS(1344),
    [anon_sym_auto] = ACTIONS(1344),
    [anon_sym_register] = ACTIONS(1344),
    [anon_sym_inline] = ACTIONS(1344),
    [anon_sym___inline] = ACTIONS(1344),
    [anon_sym___inline__] = ACTIONS(1344),
    [anon_sym___forceinline] = ACTIONS(1344),
    [anon_sym_thread_local] = ACTIONS(1344),
    [anon_sym___thread] = ACTIONS(1344),
    [anon_sym_const] = ACTIONS(1344),
    [anon_sym_constexpr] = ACTIONS(1344),
    [anon_sym_volatile] = ACTIONS(1344),
    [anon_sym_restrict] = ACTIONS(1344),
    [anon_sym___restrict__] = ACTIONS(1344),
    [anon_sym__Atomic] = ACTIONS(1344),
    [anon_sym__Noreturn] = ACTIONS(1344),
    [anon_sym_noreturn] = ACTIONS(1344),
    [anon_sym__Nonnull] = ACTIONS(1344),
    [anon_sym_alignas] = ACTIONS(1344),
    [anon_sym__Alignas] = ACTIONS(1344),
    [sym_primitive_type] = ACTIONS(1344),
    [anon_sym_enum] = ACTIONS(1344),
    [anon_sym_struct] = ACTIONS(1344),
    [anon_sym_union] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_switch] = ACTIONS(1344),
    [anon_sym_case] = ACTIONS(1344),
    [anon_sym_default] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_break] = ACTIONS(1344),
    [anon_sym_continue] = ACTIONS(1344),
    [anon_sym_goto] = ACTIONS(1344),
    [anon_sym___try] = ACTIONS(1344),
    [anon_sym___leave] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1346),
    [anon_sym_PLUS_PLUS] = ACTIONS(1346),
    [anon_sym_sizeof] = ACTIONS(1344),
    [anon_sym___alignof__] = ACTIONS(1344),
    [anon_sym___alignof] = ACTIONS(1344),
    [anon_sym__alignof] = ACTIONS(1344),
    [anon_sym_alignof] = ACTIONS(1344),
    [anon_sym__Alignof] = ACTIONS(1344),
    [anon_sym_offsetof] = ACTIONS(1344),
    [anon_sym__Generic] = ACTIONS(1344),
    [anon_sym_asm] = ACTIONS(1344),
    [anon_sym___asm__] = ACTIONS(1344),
    [anon_sym___asm] = ACTIONS(1344),
    [sym_number_literal] = ACTIONS(1346),
    [anon_sym_L_SQUOTE] = ACTIONS(1346),
    [anon_sym_u_SQUOTE] = ACTIONS(1346),
    [anon_sym_U_SQUOTE] = ACTIONS(1346),
    [anon_sym_u8_SQUOTE] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_L_DQUOTE] = ACTIONS(1346),
    [anon_sym_u_DQUOTE] = ACTIONS(1346),
    [anon_sym_U_DQUOTE] = ACTIONS(1346),
    [anon_sym_u8_DQUOTE] = ACTIONS(1346),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [anon_sym_NULL] = ACTIONS(1344),
    [anon_sym_nullptr] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(140)] = {
    [sym_identifier] = ACTIONS(1348),
    [aux_sym_preproc_include_token1] = ACTIONS(1348),
    [aux_sym_preproc_def_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token2] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1348),
    [aux_sym_preproc_else_token1] = ACTIONS(1348),
    [aux_sym_preproc_elif_token1] = ACTIONS(1348),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1348),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1348),
    [sym_preproc_directive] = ACTIONS(1348),
    [anon_sym_LPAREN2] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym___extension__] = ACTIONS(1348),
    [anon_sym_typedef] = ACTIONS(1348),
    [anon_sym_extern] = ACTIONS(1348),
    [anon_sym___attribute__] = ACTIONS(1348),
    [anon_sym___attribute] = ACTIONS(1348),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1350),
    [anon_sym___declspec] = ACTIONS(1348),
    [anon_sym___cdecl] = ACTIONS(1348),
    [anon_sym___clrcall] = ACTIONS(1348),
    [anon_sym___stdcall] = ACTIONS(1348),
    [anon_sym___fastcall] = ACTIONS(1348),
    [anon_sym___thiscall] = ACTIONS(1348),
    [anon_sym___vectorcall] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_signed] = ACTIONS(1348),
    [anon_sym_unsigned] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_auto] = ACTIONS(1348),
    [anon_sym_register] = ACTIONS(1348),
    [anon_sym_inline] = ACTIONS(1348),
    [anon_sym___inline] = ACTIONS(1348),
    [anon_sym___inline__] = ACTIONS(1348),
    [anon_sym___forceinline] = ACTIONS(1348),
    [anon_sym_thread_local] = ACTIONS(1348),
    [anon_sym___thread] = ACTIONS(1348),
    [anon_sym_const] = ACTIONS(1348),
    [anon_sym_constexpr] = ACTIONS(1348),
    [anon_sym_volatile] = ACTIONS(1348),
    [anon_sym_restrict] = ACTIONS(1348),
    [anon_sym___restrict__] = ACTIONS(1348),
    [anon_sym__Atomic] = ACTIONS(1348),
    [anon_sym__Noreturn] = ACTIONS(1348),
    [anon_sym_noreturn] = ACTIONS(1348),
    [anon_sym__Nonnull] = ACTIONS(1348),
    [anon_sym_alignas] = ACTIONS(1348),
    [anon_sym__Alignas] = ACTIONS(1348),
    [sym_primitive_type] = ACTIONS(1348),
    [anon_sym_enum] = ACTIONS(1348),
    [anon_sym_struct] = ACTIONS(1348),
    [anon_sym_union] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_goto] = ACTIONS(1348),
    [anon_sym___try] = ACTIONS(1348),
    [anon_sym___leave] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_sizeof] = ACTIONS(1348),
    [anon_sym___alignof__] = ACTIONS(1348),
    [anon_sym___alignof] = ACTIONS(1348),
    [anon_sym__alignof] = ACTIONS(1348),
    [anon_sym_alignof] = ACTIONS(1348),
    [anon_sym__Alignof] = ACTIONS(1348),
    [anon_sym_offsetof] = ACTIONS(1348),
    [anon_sym__Generic] = ACTIONS(1348),
    [anon_sym_asm] = ACTIONS(1348),
    [anon_sym___asm__] = ACTIONS(1348),
    [anon_sym___asm] = ACTIONS(1348),
    [sym_number_literal] = ACTIONS(1350),
    [anon_sym_L_SQUOTE] = ACTIONS(1350),
    [anon_sym_u_SQUOTE] = ACTIONS(1350),
    [anon_sym_U_SQUOTE] = ACTIONS(1350),
    [anon_sym_u8_SQUOTE] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_L_DQUOTE] = ACTIONS(1350),
    [anon_sym_u_DQUOTE] = ACTIONS(1350),
    [anon_sym_U_DQUOTE] = ACTIONS(1350),
    [anon_sym_u8_DQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [sym_true] = ACTIONS(1348),
    [sym_false] = ACTIONS(1348),
    [anon_sym_NULL] = ACTIONS(1348),
    [anon_sym_nullptr] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(141)] = {
    [sym_identifier] = ACTIONS(1352),
    [aux_sym_preproc_include_token1] = ACTIONS(1352),
    [aux_sym_preproc_def_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token2] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1352),
    [aux_sym_preproc_else_token1] = ACTIONS(1352),
    [aux_sym_preproc_elif_token1] = ACTIONS(1352),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1352),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1352),
    [sym_preproc_directive] = ACTIONS(1352),
    [anon_sym_LPAREN2] = ACTIONS(1354),
    [anon_sym_BANG] = ACTIONS(1354),
    [anon_sym_TILDE] = ACTIONS(1354),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1354),
    [anon_sym_AMP] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1354),
    [anon_sym___extension__] = ACTIONS(1352),
    [anon_sym_typedef] = ACTIONS(1352),
    [anon_sym_extern] = ACTIONS(1352),
    [anon_sym___attribute__] = ACTIONS(1352),
    [anon_sym___attribute] = ACTIONS(1352),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1354),
    [anon_sym___declspec] = ACTIONS(1352),
    [anon_sym___cdecl] = ACTIONS(1352),
    [anon_sym___clrcall] = ACTIONS(1352),
    [anon_sym___stdcall] = ACTIONS(1352),
    [anon_sym___fastcall] = ACTIONS(1352),
    [anon_sym___thiscall] = ACTIONS(1352),
    [anon_sym___vectorcall] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1354),
    [anon_sym_signed] = ACTIONS(1352),
    [anon_sym_unsigned] = ACTIONS(1352),
    [anon_sym_long] = ACTIONS(1352),
    [anon_sym_short] = ACTIONS(1352),
    [anon_sym_static] = ACTIONS(1352),
    [anon_sym_auto] = ACTIONS(1352),
    [anon_sym_register] = ACTIONS(1352),
    [anon_sym_inline] = ACTIONS(1352),
    [anon_sym___inline] = ACTIONS(1352),
    [anon_sym___inline__] = ACTIONS(1352),
    [anon_sym___forceinline] = ACTIONS(1352),
    [anon_sym_thread_local] = ACTIONS(1352),
    [anon_sym___thread] = ACTIONS(1352),
    [anon_sym_const] = ACTIONS(1352),
    [anon_sym_constexpr] = ACTIONS(1352),
    [anon_sym_volatile] = ACTIONS(1352),
    [anon_sym_restrict] = ACTIONS(1352),
    [anon_sym___restrict__] = ACTIONS(1352),
    [anon_sym__Atomic] = ACTIONS(1352),
    [anon_sym__Noreturn] = ACTIONS(1352),
    [anon_sym_noreturn] = ACTIONS(1352),
    [anon_sym__Nonnull] = ACTIONS(1352),
    [anon_sym_alignas] = ACTIONS(1352),
    [anon_sym__Alignas] = ACTIONS(1352),
    [sym_primitive_type] = ACTIONS(1352),
    [anon_sym_enum] = ACTIONS(1352),
    [anon_sym_struct] = ACTIONS(1352),
    [anon_sym_union] = ACTIONS(1352),
    [anon_sym_if] = ACTIONS(1352),
    [anon_sym_switch] = ACTIONS(1352),
    [anon_sym_case] = ACTIONS(1352),
    [anon_sym_default] = ACTIONS(1352),
    [anon_sym_while] = ACTIONS(1352),
    [anon_sym_do] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_break] = ACTIONS(1352),
    [anon_sym_continue] = ACTIONS(1352),
    [anon_sym_goto] = ACTIONS(1352),
    [anon_sym___try] = ACTIONS(1352),
    [anon_sym___leave] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1354),
    [anon_sym_PLUS_PLUS] = ACTIONS(1354),
    [anon_sym_sizeof] = ACTIONS(1352),
    [anon_sym___alignof__] = ACTIONS(1352),
    [anon_sym___alignof] = ACTIONS(1352),
    [anon_sym__alignof] = ACTIONS(1352),
    [anon_sym_alignof] = ACTIONS(1352),
    [anon_sym__Alignof] = ACTIONS(1352),
    [anon_sym_offsetof] = ACTIONS(1352),
    [anon_sym__Generic] = ACTIONS(1352),
    [anon_sym_asm] = ACTIONS(1352),
    [anon_sym___asm__] = ACTIONS(1352),
    [anon_sym___asm] = ACTIONS(1352),
    [sym_number_literal] = ACTIONS(1354),
    [anon_sym_L_SQUOTE] = ACTIONS(1354),
    [anon_sym_u_SQUOTE] = ACTIONS(1354),
    [anon_sym_U_SQUOTE] = ACTIONS(1354),
    [anon_sym_u8_SQUOTE] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_L_DQUOTE] = ACTIONS(1354),
    [anon_sym_u_DQUOTE] = ACTIONS(1354),
    [anon_sym_U_DQUOTE] = ACTIONS(1354),
    [anon_sym_u8_DQUOTE] = ACTIONS(1354),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [sym_true] = ACTIONS(1352),
    [sym_false] = ACTIONS(1352),
    [anon_sym_NULL] = ACTIONS(1352),
    [anon_sym_nullptr] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(142)] = {
    [sym_identifier] = ACTIONS(1356),
    [aux_sym_preproc_include_token1] = ACTIONS(1356),
    [aux_sym_preproc_def_token1] = ACTIONS(1356),
    [aux_sym_preproc_if_token1] = ACTIONS(1356),
    [aux_sym_preproc_if_token2] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1356),
    [aux_sym_preproc_else_token1] = ACTIONS(1356),
    [aux_sym_preproc_elif_token1] = ACTIONS(1356),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1356),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1356),
    [sym_preproc_directive] = ACTIONS(1356),
    [anon_sym_LPAREN2] = ACTIONS(1358),
    [anon_sym_BANG] = ACTIONS(1358),
    [anon_sym_TILDE] = ACTIONS(1358),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1358),
    [anon_sym_AMP] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1358),
    [anon_sym___extension__] = ACTIONS(1356),
    [anon_sym_typedef] = ACTIONS(1356),
    [anon_sym_extern] = ACTIONS(1356),
    [anon_sym___attribute__] = ACTIONS(1356),
    [anon_sym___attribute] = ACTIONS(1356),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1358),
    [anon_sym___declspec] = ACTIONS(1356),
    [anon_sym___cdecl] = ACTIONS(1356),
    [anon_sym___clrcall] = ACTIONS(1356),
    [anon_sym___stdcall] = ACTIONS(1356),
    [anon_sym___fastcall] = ACTIONS(1356),
    [anon_sym___thiscall] = ACTIONS(1356),
    [anon_sym___vectorcall] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1358),
    [anon_sym_signed] = ACTIONS(1356),
    [anon_sym_unsigned] = ACTIONS(1356),
    [anon_sym_long] = ACTIONS(1356),
    [anon_sym_short] = ACTIONS(1356),
    [anon_sym_static] = ACTIONS(1356),
    [anon_sym_auto] = ACTIONS(1356),
    [anon_sym_register] = ACTIONS(1356),
    [anon_sym_inline] = ACTIONS(1356),
    [anon_sym___inline] = ACTIONS(1356),
    [anon_sym___inline__] = ACTIONS(1356),
    [anon_sym___forceinline] = ACTIONS(1356),
    [anon_sym_thread_local] = ACTIONS(1356),
    [anon_sym___thread] = ACTIONS(1356),
    [anon_sym_const] = ACTIONS(1356),
    [anon_sym_constexpr] = ACTIONS(1356),
    [anon_sym_volatile] = ACTIONS(1356),
    [anon_sym_restrict] = ACTIONS(1356),
    [anon_sym___restrict__] = ACTIONS(1356),
    [anon_sym__Atomic] = ACTIONS(1356),
    [anon_sym__Noreturn] = ACTIONS(1356),
    [anon_sym_noreturn] = ACTIONS(1356),
    [anon_sym__Nonnull] = ACTIONS(1356),
    [anon_sym_alignas] = ACTIONS(1356),
    [anon_sym__Alignas] = ACTIONS(1356),
    [sym_primitive_type] = ACTIONS(1356),
    [anon_sym_enum] = ACTIONS(1356),
    [anon_sym_struct] = ACTIONS(1356),
    [anon_sym_union] = ACTIONS(1356),
    [anon_sym_if] = ACTIONS(1356),
    [anon_sym_switch] = ACTIONS(1356),
    [anon_sym_case] = ACTIONS(1356),
    [anon_sym_default] = ACTIONS(1356),
    [anon_sym_while] = ACTIONS(1356),
    [anon_sym_do] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1356),
    [anon_sym_return] = ACTIONS(1356),
    [anon_sym_break] = ACTIONS(1356),
    [anon_sym_continue] = ACTIONS(1356),
    [anon_sym_goto] = ACTIONS(1356),
    [anon_sym___try] = ACTIONS(1356),
    [anon_sym___leave] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1358),
    [anon_sym_PLUS_PLUS] = ACTIONS(1358),
    [anon_sym_sizeof] = ACTIONS(1356),
    [anon_sym___alignof__] = ACTIONS(1356),
    [anon_sym___alignof] = ACTIONS(1356),
    [anon_sym__alignof] = ACTIONS(1356),
    [anon_sym_alignof] = ACTIONS(1356),
    [anon_sym__Alignof] = ACTIONS(1356),
    [anon_sym_offsetof] = ACTIONS(1356),
    [anon_sym__Generic] = ACTIONS(1356),
    [anon_sym_asm] = ACTIONS(1356),
    [anon_sym___asm__] = ACTIONS(1356),
    [anon_sym___asm] = ACTIONS(1356),
    [sym_number_literal] = ACTIONS(1358),
    [anon_sym_L_SQUOTE] = ACTIONS(1358),
    [anon_sym_u_SQUOTE] = ACTIONS(1358),
    [anon_sym_U_SQUOTE] = ACTIONS(1358),
    [anon_sym_u8_SQUOTE] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_L_DQUOTE] = ACTIONS(1358),
    [anon_sym_u_DQUOTE] = ACTIONS(1358),
    [anon_sym_U_DQUOTE] = ACTIONS(1358),
    [anon_sym_u8_DQUOTE] = ACTIONS(1358),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [sym_true] = ACTIONS(1356),
    [sym_false] = ACTIONS(1356),
    [anon_sym_NULL] = ACTIONS(1356),
    [anon_sym_nullptr] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(143)] = {
    [sym_identifier] = ACTIONS(1360),
    [aux_sym_preproc_include_token1] = ACTIONS(1360),
    [aux_sym_preproc_def_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token2] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1360),
    [aux_sym_preproc_else_token1] = ACTIONS(1360),
    [aux_sym_preproc_elif_token1] = ACTIONS(1360),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1360),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1360),
    [sym_preproc_directive] = ACTIONS(1360),
    [anon_sym_LPAREN2] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_PLUS] = ACTIONS(1360),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym___extension__] = ACTIONS(1360),
    [anon_sym_typedef] = ACTIONS(1360),
    [anon_sym_extern] = ACTIONS(1360),
    [anon_sym___attribute__] = ACTIONS(1360),
    [anon_sym___attribute] = ACTIONS(1360),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1362),
    [anon_sym___declspec] = ACTIONS(1360),
    [anon_sym___cdecl] = ACTIONS(1360),
    [anon_sym___clrcall] = ACTIONS(1360),
    [anon_sym___stdcall] = ACTIONS(1360),
    [anon_sym___fastcall] = ACTIONS(1360),
    [anon_sym___thiscall] = ACTIONS(1360),
    [anon_sym___vectorcall] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_signed] = ACTIONS(1360),
    [anon_sym_unsigned] = ACTIONS(1360),
    [anon_sym_long] = ACTIONS(1360),
    [anon_sym_short] = ACTIONS(1360),
    [anon_sym_static] = ACTIONS(1360),
    [anon_sym_auto] = ACTIONS(1360),
    [anon_sym_register] = ACTIONS(1360),
    [anon_sym_inline] = ACTIONS(1360),
    [anon_sym___inline] = ACTIONS(1360),
    [anon_sym___inline__] = ACTIONS(1360),
    [anon_sym___forceinline] = ACTIONS(1360),
    [anon_sym_thread_local] = ACTIONS(1360),
    [anon_sym___thread] = ACTIONS(1360),
    [anon_sym_const] = ACTIONS(1360),
    [anon_sym_constexpr] = ACTIONS(1360),
    [anon_sym_volatile] = ACTIONS(1360),
    [anon_sym_restrict] = ACTIONS(1360),
    [anon_sym___restrict__] = ACTIONS(1360),
    [anon_sym__Atomic] = ACTIONS(1360),
    [anon_sym__Noreturn] = ACTIONS(1360),
    [anon_sym_noreturn] = ACTIONS(1360),
    [anon_sym__Nonnull] = ACTIONS(1360),
    [anon_sym_alignas] = ACTIONS(1360),
    [anon_sym__Alignas] = ACTIONS(1360),
    [sym_primitive_type] = ACTIONS(1360),
    [anon_sym_enum] = ACTIONS(1360),
    [anon_sym_struct] = ACTIONS(1360),
    [anon_sym_union] = ACTIONS(1360),
    [anon_sym_if] = ACTIONS(1360),
    [anon_sym_switch] = ACTIONS(1360),
    [anon_sym_case] = ACTIONS(1360),
    [anon_sym_default] = ACTIONS(1360),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_do] = ACTIONS(1360),
    [anon_sym_for] = ACTIONS(1360),
    [anon_sym_return] = ACTIONS(1360),
    [anon_sym_break] = ACTIONS(1360),
    [anon_sym_continue] = ACTIONS(1360),
    [anon_sym_goto] = ACTIONS(1360),
    [anon_sym___try] = ACTIONS(1360),
    [anon_sym___leave] = ACTIONS(1360),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_sizeof] = ACTIONS(1360),
    [anon_sym___alignof__] = ACTIONS(1360),
    [anon_sym___alignof] = ACTIONS(1360),
    [anon_sym__alignof] = ACTIONS(1360),
    [anon_sym_alignof] = ACTIONS(1360),
    [anon_sym__Alignof] = ACTIONS(1360),
    [anon_sym_offsetof] = ACTIONS(1360),
    [anon_sym__Generic] = ACTIONS(1360),
    [anon_sym_asm] = ACTIONS(1360),
    [anon_sym___asm__] = ACTIONS(1360),
    [anon_sym___asm] = ACTIONS(1360),
    [sym_number_literal] = ACTIONS(1362),
    [anon_sym_L_SQUOTE] = ACTIONS(1362),
    [anon_sym_u_SQUOTE] = ACTIONS(1362),
    [anon_sym_U_SQUOTE] = ACTIONS(1362),
    [anon_sym_u8_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_L_DQUOTE] = ACTIONS(1362),
    [anon_sym_u_DQUOTE] = ACTIONS(1362),
    [anon_sym_U_DQUOTE] = ACTIONS(1362),
    [anon_sym_u8_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [sym_true] = ACTIONS(1360),
    [sym_false] = ACTIONS(1360),
    [anon_sym_NULL] = ACTIONS(1360),
    [anon_sym_nullptr] = ACTIONS(1360),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(144)] = {
    [sym_identifier] = ACTIONS(1364),
    [aux_sym_preproc_include_token1] = ACTIONS(1364),
    [aux_sym_preproc_def_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token2] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1364),
    [aux_sym_preproc_else_token1] = ACTIONS(1364),
    [aux_sym_preproc_elif_token1] = ACTIONS(1364),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1364),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1364),
    [sym_preproc_directive] = ACTIONS(1364),
    [anon_sym_LPAREN2] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_PLUS] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym___extension__] = ACTIONS(1364),
    [anon_sym_typedef] = ACTIONS(1364),
    [anon_sym_extern] = ACTIONS(1364),
    [anon_sym___attribute__] = ACTIONS(1364),
    [anon_sym___attribute] = ACTIONS(1364),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1366),
    [anon_sym___declspec] = ACTIONS(1364),
    [anon_sym___cdecl] = ACTIONS(1364),
    [anon_sym___clrcall] = ACTIONS(1364),
    [anon_sym___stdcall] = ACTIONS(1364),
    [anon_sym___fastcall] = ACTIONS(1364),
    [anon_sym___thiscall] = ACTIONS(1364),
    [anon_sym___vectorcall] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_signed] = ACTIONS(1364),
    [anon_sym_unsigned] = ACTIONS(1364),
    [anon_sym_long] = ACTIONS(1364),
    [anon_sym_short] = ACTIONS(1364),
    [anon_sym_static] = ACTIONS(1364),
    [anon_sym_auto] = ACTIONS(1364),
    [anon_sym_register] = ACTIONS(1364),
    [anon_sym_inline] = ACTIONS(1364),
    [anon_sym___inline] = ACTIONS(1364),
    [anon_sym___inline__] = ACTIONS(1364),
    [anon_sym___forceinline] = ACTIONS(1364),
    [anon_sym_thread_local] = ACTIONS(1364),
    [anon_sym___thread] = ACTIONS(1364),
    [anon_sym_const] = ACTIONS(1364),
    [anon_sym_constexpr] = ACTIONS(1364),
    [anon_sym_volatile] = ACTIONS(1364),
    [anon_sym_restrict] = ACTIONS(1364),
    [anon_sym___restrict__] = ACTIONS(1364),
    [anon_sym__Atomic] = ACTIONS(1364),
    [anon_sym__Noreturn] = ACTIONS(1364),
    [anon_sym_noreturn] = ACTIONS(1364),
    [anon_sym__Nonnull] = ACTIONS(1364),
    [anon_sym_alignas] = ACTIONS(1364),
    [anon_sym__Alignas] = ACTIONS(1364),
    [sym_primitive_type] = ACTIONS(1364),
    [anon_sym_enum] = ACTIONS(1364),
    [anon_sym_struct] = ACTIONS(1364),
    [anon_sym_union] = ACTIONS(1364),
    [anon_sym_if] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1364),
    [anon_sym_case] = ACTIONS(1364),
    [anon_sym_default] = ACTIONS(1364),
    [anon_sym_while] = ACTIONS(1364),
    [anon_sym_do] = ACTIONS(1364),
    [anon_sym_for] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1364),
    [anon_sym_break] = ACTIONS(1364),
    [anon_sym_continue] = ACTIONS(1364),
    [anon_sym_goto] = ACTIONS(1364),
    [anon_sym___try] = ACTIONS(1364),
    [anon_sym___leave] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1366),
    [anon_sym_sizeof] = ACTIONS(1364),
    [anon_sym___alignof__] = ACTIONS(1364),
    [anon_sym___alignof] = ACTIONS(1364),
    [anon_sym__alignof] = ACTIONS(1364),
    [anon_sym_alignof] = ACTIONS(1364),
    [anon_sym__Alignof] = ACTIONS(1364),
    [anon_sym_offsetof] = ACTIONS(1364),
    [anon_sym__Generic] = ACTIONS(1364),
    [anon_sym_asm] = ACTIONS(1364),
    [anon_sym___asm__] = ACTIONS(1364),
    [anon_sym___asm] = ACTIONS(1364),
    [sym_number_literal] = ACTIONS(1366),
    [anon_sym_L_SQUOTE] = ACTIONS(1366),
    [anon_sym_u_SQUOTE] = ACTIONS(1366),
    [anon_sym_U_SQUOTE] = ACTIONS(1366),
    [anon_sym_u8_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_L_DQUOTE] = ACTIONS(1366),
    [anon_sym_u_DQUOTE] = ACTIONS(1366),
    [anon_sym_U_DQUOTE] = ACTIONS(1366),
    [anon_sym_u8_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [sym_true] = ACTIONS(1364),
    [sym_false] = ACTIONS(1364),
    [anon_sym_NULL] = ACTIONS(1364),
    [anon_sym_nullptr] = ACTIONS(1364),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(145)] = {
    [sym_identifier] = ACTIONS(1368),
    [aux_sym_preproc_include_token1] = ACTIONS(1368),
    [aux_sym_preproc_def_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token2] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1368),
    [aux_sym_preproc_else_token1] = ACTIONS(1368),
    [aux_sym_preproc_elif_token1] = ACTIONS(1368),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1368),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1368),
    [sym_preproc_directive] = ACTIONS(1368),
    [anon_sym_LPAREN2] = ACTIONS(1370),
    [anon_sym_BANG] = ACTIONS(1370),
    [anon_sym_TILDE] = ACTIONS(1370),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_PLUS] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1370),
    [anon_sym_AMP] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1370),
    [anon_sym___extension__] = ACTIONS(1368),
    [anon_sym_typedef] = ACTIONS(1368),
    [anon_sym_extern] = ACTIONS(1368),
    [anon_sym___attribute__] = ACTIONS(1368),
    [anon_sym___attribute] = ACTIONS(1368),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1370),
    [anon_sym___declspec] = ACTIONS(1368),
    [anon_sym___cdecl] = ACTIONS(1368),
    [anon_sym___clrcall] = ACTIONS(1368),
    [anon_sym___stdcall] = ACTIONS(1368),
    [anon_sym___fastcall] = ACTIONS(1368),
    [anon_sym___thiscall] = ACTIONS(1368),
    [anon_sym___vectorcall] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1370),
    [anon_sym_signed] = ACTIONS(1368),
    [anon_sym_unsigned] = ACTIONS(1368),
    [anon_sym_long] = ACTIONS(1368),
    [anon_sym_short] = ACTIONS(1368),
    [anon_sym_static] = ACTIONS(1368),
    [anon_sym_auto] = ACTIONS(1368),
    [anon_sym_register] = ACTIONS(1368),
    [anon_sym_inline] = ACTIONS(1368),
    [anon_sym___inline] = ACTIONS(1368),
    [anon_sym___inline__] = ACTIONS(1368),
    [anon_sym___forceinline] = ACTIONS(1368),
    [anon_sym_thread_local] = ACTIONS(1368),
    [anon_sym___thread] = ACTIONS(1368),
    [anon_sym_const] = ACTIONS(1368),
    [anon_sym_constexpr] = ACTIONS(1368),
    [anon_sym_volatile] = ACTIONS(1368),
    [anon_sym_restrict] = ACTIONS(1368),
    [anon_sym___restrict__] = ACTIONS(1368),
    [anon_sym__Atomic] = ACTIONS(1368),
    [anon_sym__Noreturn] = ACTIONS(1368),
    [anon_sym_noreturn] = ACTIONS(1368),
    [anon_sym__Nonnull] = ACTIONS(1368),
    [anon_sym_alignas] = ACTIONS(1368),
    [anon_sym__Alignas] = ACTIONS(1368),
    [sym_primitive_type] = ACTIONS(1368),
    [anon_sym_enum] = ACTIONS(1368),
    [anon_sym_struct] = ACTIONS(1368),
    [anon_sym_union] = ACTIONS(1368),
    [anon_sym_if] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1368),
    [anon_sym_case] = ACTIONS(1368),
    [anon_sym_default] = ACTIONS(1368),
    [anon_sym_while] = ACTIONS(1368),
    [anon_sym_do] = ACTIONS(1368),
    [anon_sym_for] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1368),
    [anon_sym_break] = ACTIONS(1368),
    [anon_sym_continue] = ACTIONS(1368),
    [anon_sym_goto] = ACTIONS(1368),
    [anon_sym___try] = ACTIONS(1368),
    [anon_sym___leave] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1370),
    [anon_sym_PLUS_PLUS] = ACTIONS(1370),
    [anon_sym_sizeof] = ACTIONS(1368),
    [anon_sym___alignof__] = ACTIONS(1368),
    [anon_sym___alignof] = ACTIONS(1368),
    [anon_sym__alignof] = ACTIONS(1368),
    [anon_sym_alignof] = ACTIONS(1368),
    [anon_sym__Alignof] = ACTIONS(1368),
    [anon_sym_offsetof] = ACTIONS(1368),
    [anon_sym__Generic] = ACTIONS(1368),
    [anon_sym_asm] = ACTIONS(1368),
    [anon_sym___asm__] = ACTIONS(1368),
    [anon_sym___asm] = ACTIONS(1368),
    [sym_number_literal] = ACTIONS(1370),
    [anon_sym_L_SQUOTE] = ACTIONS(1370),
    [anon_sym_u_SQUOTE] = ACTIONS(1370),
    [anon_sym_U_SQUOTE] = ACTIONS(1370),
    [anon_sym_u8_SQUOTE] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_L_DQUOTE] = ACTIONS(1370),
    [anon_sym_u_DQUOTE] = ACTIONS(1370),
    [anon_sym_U_DQUOTE] = ACTIONS(1370),
    [anon_sym_u8_DQUOTE] = ACTIONS(1370),
    [anon_sym_DQUOTE] = ACTIONS(1370),
    [sym_true] = ACTIONS(1368),
    [sym_false] = ACTIONS(1368),
    [anon_sym_NULL] = ACTIONS(1368),
    [anon_sym_nullptr] = ACTIONS(1368),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(146)] = {
    [sym_else_clause] = STATE(250),
    [ts_builtin_sym_end] = ACTIONS(1130),
    [sym_identifier] = ACTIONS(1128),
    [aux_sym_preproc_include_token1] = ACTIONS(1128),
    [aux_sym_preproc_def_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1128),
    [sym_preproc_directive] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1130),
    [anon_sym_AMP] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [anon_sym___extension__] = ACTIONS(1128),
    [anon_sym_typedef] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1128),
    [anon_sym___attribute__] = ACTIONS(1128),
    [anon_sym___attribute] = ACTIONS(1128),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1130),
    [anon_sym___declspec] = ACTIONS(1128),
    [anon_sym___cdecl] = ACTIONS(1128),
    [anon_sym___clrcall] = ACTIONS(1128),
    [anon_sym___stdcall] = ACTIONS(1128),
    [anon_sym___fastcall] = ACTIONS(1128),
    [anon_sym___thiscall] = ACTIONS(1128),
    [anon_sym___vectorcall] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_signed] = ACTIONS(1128),
    [anon_sym_unsigned] = ACTIONS(1128),
    [anon_sym_long] = ACTIONS(1128),
    [anon_sym_short] = ACTIONS(1128),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_auto] = ACTIONS(1128),
    [anon_sym_register] = ACTIONS(1128),
    [anon_sym_inline] = ACTIONS(1128),
    [anon_sym___inline] = ACTIONS(1128),
    [anon_sym___inline__] = ACTIONS(1128),
    [anon_sym___forceinline] = ACTIONS(1128),
    [anon_sym_thread_local] = ACTIONS(1128),
    [anon_sym___thread] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_constexpr] = ACTIONS(1128),
    [anon_sym_volatile] = ACTIONS(1128),
    [anon_sym_restrict] = ACTIONS(1128),
    [anon_sym___restrict__] = ACTIONS(1128),
    [anon_sym__Atomic] = ACTIONS(1128),
    [anon_sym__Noreturn] = ACTIONS(1128),
    [anon_sym_noreturn] = ACTIONS(1128),
    [anon_sym__Nonnull] = ACTIONS(1128),
    [anon_sym_alignas] = ACTIONS(1128),
    [anon_sym__Alignas] = ACTIONS(1128),
    [sym_primitive_type] = ACTIONS(1128),
    [anon_sym_enum] = ACTIONS(1128),
    [anon_sym_struct] = ACTIONS(1128),
    [anon_sym_union] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1372),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_case] = ACTIONS(1128),
    [anon_sym_default] = ACTIONS(1128),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_goto] = ACTIONS(1128),
    [anon_sym___try] = ACTIONS(1128),
    [anon_sym___leave] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1128),
    [anon_sym___alignof] = ACTIONS(1128),
    [anon_sym__alignof] = ACTIONS(1128),
    [anon_sym_alignof] = ACTIONS(1128),
    [anon_sym__Alignof] = ACTIONS(1128),
    [anon_sym_offsetof] = ACTIONS(1128),
    [anon_sym__Generic] = ACTIONS(1128),
    [anon_sym_asm] = ACTIONS(1128),
    [anon_sym___asm__] = ACTIONS(1128),
    [anon_sym___asm] = ACTIONS(1128),
    [sym_number_literal] = ACTIONS(1130),
    [anon_sym_L_SQUOTE] = ACTIONS(1130),
    [anon_sym_u_SQUOTE] = ACTIONS(1130),
    [anon_sym_U_SQUOTE] = ACTIONS(1130),
    [anon_sym_u8_SQUOTE] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_L_DQUOTE] = ACTIONS(1130),
    [anon_sym_u_DQUOTE] = ACTIONS(1130),
    [anon_sym_U_DQUOTE] = ACTIONS(1130),
    [anon_sym_u8_DQUOTE] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [anon_sym_NULL] = ACTIONS(1128),
    [anon_sym_nullptr] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(147)] = {
    [sym_else_clause] = STATE(224),
    [sym_identifier] = ACTIONS(1128),
    [aux_sym_preproc_include_token1] = ACTIONS(1128),
    [aux_sym_preproc_def_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1128),
    [sym_preproc_directive] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1130),
    [anon_sym_AMP] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [anon_sym___extension__] = ACTIONS(1128),
    [anon_sym_typedef] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1128),
    [anon_sym___attribute__] = ACTIONS(1128),
    [anon_sym___attribute] = ACTIONS(1128),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1130),
    [anon_sym___declspec] = ACTIONS(1128),
    [anon_sym___cdecl] = ACTIONS(1128),
    [anon_sym___clrcall] = ACTIONS(1128),
    [anon_sym___stdcall] = ACTIONS(1128),
    [anon_sym___fastcall] = ACTIONS(1128),
    [anon_sym___thiscall] = ACTIONS(1128),
    [anon_sym___vectorcall] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_RBRACE] = ACTIONS(1130),
    [anon_sym_signed] = ACTIONS(1128),
    [anon_sym_unsigned] = ACTIONS(1128),
    [anon_sym_long] = ACTIONS(1128),
    [anon_sym_short] = ACTIONS(1128),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_auto] = ACTIONS(1128),
    [anon_sym_register] = ACTIONS(1128),
    [anon_sym_inline] = ACTIONS(1128),
    [anon_sym___inline] = ACTIONS(1128),
    [anon_sym___inline__] = ACTIONS(1128),
    [anon_sym___forceinline] = ACTIONS(1128),
    [anon_sym_thread_local] = ACTIONS(1128),
    [anon_sym___thread] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_constexpr] = ACTIONS(1128),
    [anon_sym_volatile] = ACTIONS(1128),
    [anon_sym_restrict] = ACTIONS(1128),
    [anon_sym___restrict__] = ACTIONS(1128),
    [anon_sym__Atomic] = ACTIONS(1128),
    [anon_sym__Noreturn] = ACTIONS(1128),
    [anon_sym_noreturn] = ACTIONS(1128),
    [anon_sym__Nonnull] = ACTIONS(1128),
    [anon_sym_alignas] = ACTIONS(1128),
    [anon_sym__Alignas] = ACTIONS(1128),
    [sym_primitive_type] = ACTIONS(1128),
    [anon_sym_enum] = ACTIONS(1128),
    [anon_sym_struct] = ACTIONS(1128),
    [anon_sym_union] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_case] = ACTIONS(1128),
    [anon_sym_default] = ACTIONS(1128),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_goto] = ACTIONS(1128),
    [anon_sym___try] = ACTIONS(1128),
    [anon_sym___leave] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1128),
    [anon_sym___alignof] = ACTIONS(1128),
    [anon_sym__alignof] = ACTIONS(1128),
    [anon_sym_alignof] = ACTIONS(1128),
    [anon_sym__Alignof] = ACTIONS(1128),
    [anon_sym_offsetof] = ACTIONS(1128),
    [anon_sym__Generic] = ACTIONS(1128),
    [anon_sym_asm] = ACTIONS(1128),
    [anon_sym___asm__] = ACTIONS(1128),
    [anon_sym___asm] = ACTIONS(1128),
    [sym_number_literal] = ACTIONS(1130),
    [anon_sym_L_SQUOTE] = ACTIONS(1130),
    [anon_sym_u_SQUOTE] = ACTIONS(1130),
    [anon_sym_U_SQUOTE] = ACTIONS(1130),
    [anon_sym_u8_SQUOTE] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_L_DQUOTE] = ACTIONS(1130),
    [anon_sym_u_DQUOTE] = ACTIONS(1130),
    [anon_sym_U_DQUOTE] = ACTIONS(1130),
    [anon_sym_u8_DQUOTE] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [anon_sym_NULL] = ACTIONS(1128),
    [anon_sym_nullptr] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(148)] = {
    [sym_else_clause] = STATE(197),
    [sym_identifier] = ACTIONS(1128),
    [aux_sym_preproc_include_token1] = ACTIONS(1128),
    [aux_sym_preproc_def_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token1] = ACTIONS(1128),
    [aux_sym_preproc_if_token2] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1128),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1128),
    [sym_preproc_directive] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1130),
    [anon_sym_AMP] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [anon_sym___extension__] = ACTIONS(1128),
    [anon_sym_typedef] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1128),
    [anon_sym___attribute__] = ACTIONS(1128),
    [anon_sym___attribute] = ACTIONS(1128),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1130),
    [anon_sym___declspec] = ACTIONS(1128),
    [anon_sym___cdecl] = ACTIONS(1128),
    [anon_sym___clrcall] = ACTIONS(1128),
    [anon_sym___stdcall] = ACTIONS(1128),
    [anon_sym___fastcall] = ACTIONS(1128),
    [anon_sym___thiscall] = ACTIONS(1128),
    [anon_sym___vectorcall] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_signed] = ACTIONS(1128),
    [anon_sym_unsigned] = ACTIONS(1128),
    [anon_sym_long] = ACTIONS(1128),
    [anon_sym_short] = ACTIONS(1128),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_auto] = ACTIONS(1128),
    [anon_sym_register] = ACTIONS(1128),
    [anon_sym_inline] = ACTIONS(1128),
    [anon_sym___inline] = ACTIONS(1128),
    [anon_sym___inline__] = ACTIONS(1128),
    [anon_sym___forceinline] = ACTIONS(1128),
    [anon_sym_thread_local] = ACTIONS(1128),
    [anon_sym___thread] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_constexpr] = ACTIONS(1128),
    [anon_sym_volatile] = ACTIONS(1128),
    [anon_sym_restrict] = ACTIONS(1128),
    [anon_sym___restrict__] = ACTIONS(1128),
    [anon_sym__Atomic] = ACTIONS(1128),
    [anon_sym__Noreturn] = ACTIONS(1128),
    [anon_sym_noreturn] = ACTIONS(1128),
    [anon_sym__Nonnull] = ACTIONS(1128),
    [anon_sym_alignas] = ACTIONS(1128),
    [anon_sym__Alignas] = ACTIONS(1128),
    [sym_primitive_type] = ACTIONS(1128),
    [anon_sym_enum] = ACTIONS(1128),
    [anon_sym_struct] = ACTIONS(1128),
    [anon_sym_union] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_case] = ACTIONS(1128),
    [anon_sym_default] = ACTIONS(1128),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_goto] = ACTIONS(1128),
    [anon_sym___try] = ACTIONS(1128),
    [anon_sym___leave] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1128),
    [anon_sym___alignof] = ACTIONS(1128),
    [anon_sym__alignof] = ACTIONS(1128),
    [anon_sym_alignof] = ACTIONS(1128),
    [anon_sym__Alignof] = ACTIONS(1128),
    [anon_sym_offsetof] = ACTIONS(1128),
    [anon_sym__Generic] = ACTIONS(1128),
    [anon_sym_asm] = ACTIONS(1128),
    [anon_sym___asm__] = ACTIONS(1128),
    [anon_sym___asm] = ACTIONS(1128),
    [sym_number_literal] = ACTIONS(1130),
    [anon_sym_L_SQUOTE] = ACTIONS(1130),
    [anon_sym_u_SQUOTE] = ACTIONS(1130),
    [anon_sym_U_SQUOTE] = ACTIONS(1130),
    [anon_sym_u8_SQUOTE] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_L_DQUOTE] = ACTIONS(1130),
    [anon_sym_u_DQUOTE] = ACTIONS(1130),
    [anon_sym_U_DQUOTE] = ACTIONS(1130),
    [anon_sym_u8_DQUOTE] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [anon_sym_NULL] = ACTIONS(1128),
    [anon_sym_nullptr] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(149)] = {
    [sym_identifier] = ACTIONS(1162),
    [aux_sym_preproc_include_token1] = ACTIONS(1162),
    [aux_sym_preproc_def_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token2] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1162),
    [sym_preproc_directive] = ACTIONS(1162),
    [anon_sym_LPAREN2] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1162),
    [anon_sym_PLUS] = ACTIONS(1162),
    [anon_sym_STAR] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym___extension__] = ACTIONS(1162),
    [anon_sym_typedef] = ACTIONS(1162),
    [anon_sym_extern] = ACTIONS(1162),
    [anon_sym___attribute__] = ACTIONS(1162),
    [anon_sym___attribute] = ACTIONS(1162),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1164),
    [anon_sym___declspec] = ACTIONS(1162),
    [anon_sym___cdecl] = ACTIONS(1162),
    [anon_sym___clrcall] = ACTIONS(1162),
    [anon_sym___stdcall] = ACTIONS(1162),
    [anon_sym___fastcall] = ACTIONS(1162),
    [anon_sym___thiscall] = ACTIONS(1162),
    [anon_sym___vectorcall] = ACTIONS(1162),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_signed] = ACTIONS(1162),
    [anon_sym_unsigned] = ACTIONS(1162),
    [anon_sym_long] = ACTIONS(1162),
    [anon_sym_short] = ACTIONS(1162),
    [anon_sym_static] = ACTIONS(1162),
    [anon_sym_auto] = ACTIONS(1162),
    [anon_sym_register] = ACTIONS(1162),
    [anon_sym_inline] = ACTIONS(1162),
    [anon_sym___inline] = ACTIONS(1162),
    [anon_sym___inline__] = ACTIONS(1162),
    [anon_sym___forceinline] = ACTIONS(1162),
    [anon_sym_thread_local] = ACTIONS(1162),
    [anon_sym___thread] = ACTIONS(1162),
    [anon_sym_const] = ACTIONS(1162),
    [anon_sym_constexpr] = ACTIONS(1162),
    [anon_sym_volatile] = ACTIONS(1162),
    [anon_sym_restrict] = ACTIONS(1162),
    [anon_sym___restrict__] = ACTIONS(1162),
    [anon_sym__Atomic] = ACTIONS(1162),
    [anon_sym__Noreturn] = ACTIONS(1162),
    [anon_sym_noreturn] = ACTIONS(1162),
    [anon_sym__Nonnull] = ACTIONS(1162),
    [anon_sym_alignas] = ACTIONS(1162),
    [anon_sym__Alignas] = ACTIONS(1162),
    [sym_primitive_type] = ACTIONS(1162),
    [anon_sym_enum] = ACTIONS(1162),
    [anon_sym_struct] = ACTIONS(1162),
    [anon_sym_union] = ACTIONS(1162),
    [anon_sym_if] = ACTIONS(1162),
    [anon_sym_else] = ACTIONS(1162),
    [anon_sym_switch] = ACTIONS(1162),
    [anon_sym_case] = ACTIONS(1162),
    [anon_sym_default] = ACTIONS(1162),
    [anon_sym_while] = ACTIONS(1162),
    [anon_sym_do] = ACTIONS(1162),
    [anon_sym_for] = ACTIONS(1162),
    [anon_sym_return] = ACTIONS(1162),
    [anon_sym_break] = ACTIONS(1162),
    [anon_sym_continue] = ACTIONS(1162),
    [anon_sym_goto] = ACTIONS(1162),
    [anon_sym___try] = ACTIONS(1162),
    [anon_sym___leave] = ACTIONS(1162),
    [anon_sym_DASH_DASH] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1164),
    [anon_sym_sizeof] = ACTIONS(1162),
    [anon_sym___alignof__] = ACTIONS(1162),
    [anon_sym___alignof] = ACTIONS(1162),
    [anon_sym__alignof] = ACTIONS(1162),
    [anon_sym_alignof] = ACTIONS(1162),
    [anon_sym__Alignof] = ACTIONS(1162),
    [anon_sym_offsetof] = ACTIONS(1162),
    [anon_sym__Generic] = ACTIONS(1162),
    [anon_sym_asm] = ACTIONS(1162),
    [anon_sym___asm__] = ACTIONS(1162),
    [anon_sym___asm] = ACTIONS(1162),
    [sym_number_literal] = ACTIONS(1164),
    [anon_sym_L_SQUOTE] = ACTIONS(1164),
    [anon_sym_u_SQUOTE] = ACTIONS(1164),
    [anon_sym_U_SQUOTE] = ACTIONS(1164),
    [anon_sym_u8_SQUOTE] = ACTIONS(1164),
    [anon_sym_SQUOTE] = ACTIONS(1164),
    [anon_sym_L_DQUOTE] = ACTIONS(1164),
    [anon_sym_u_DQUOTE] = ACTIONS(1164),
    [anon_sym_U_DQUOTE] = ACTIONS(1164),
    [anon_sym_u8_DQUOTE] = ACTIONS(1164),
    [anon_sym_DQUOTE] = ACTIONS(1164),
    [sym_true] = ACTIONS(1162),
    [sym_false] = ACTIONS(1162),
    [anon_sym_NULL] = ACTIONS(1162),
    [anon_sym_nullptr] = ACTIONS(1162),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(150)] = {
    [ts_builtin_sym_end] = ACTIONS(1144),
    [sym_identifier] = ACTIONS(1142),
    [aux_sym_preproc_include_token1] = ACTIONS(1142),
    [aux_sym_preproc_def_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1142),
    [sym_preproc_directive] = ACTIONS(1142),
    [anon_sym_LPAREN2] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_TILDE] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1142),
    [anon_sym_PLUS] = ACTIONS(1142),
    [anon_sym_STAR] = ACTIONS(1144),
    [anon_sym_AMP] = ACTIONS(1144),
    [anon_sym_SEMI] = ACTIONS(1144),
    [anon_sym___extension__] = ACTIONS(1142),
    [anon_sym_typedef] = ACTIONS(1142),
    [anon_sym_extern] = ACTIONS(1142),
    [anon_sym___attribute__] = ACTIONS(1142),
    [anon_sym___attribute] = ACTIONS(1142),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1144),
    [anon_sym___declspec] = ACTIONS(1142),
    [anon_sym___cdecl] = ACTIONS(1142),
    [anon_sym___clrcall] = ACTIONS(1142),
    [anon_sym___stdcall] = ACTIONS(1142),
    [anon_sym___fastcall] = ACTIONS(1142),
    [anon_sym___thiscall] = ACTIONS(1142),
    [anon_sym___vectorcall] = ACTIONS(1142),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_signed] = ACTIONS(1142),
    [anon_sym_unsigned] = ACTIONS(1142),
    [anon_sym_long] = ACTIONS(1142),
    [anon_sym_short] = ACTIONS(1142),
    [anon_sym_static] = ACTIONS(1142),
    [anon_sym_auto] = ACTIONS(1142),
    [anon_sym_register] = ACTIONS(1142),
    [anon_sym_inline] = ACTIONS(1142),
    [anon_sym___inline] = ACTIONS(1142),
    [anon_sym___inline__] = ACTIONS(1142),
    [anon_sym___forceinline] = ACTIONS(1142),
    [anon_sym_thread_local] = ACTIONS(1142),
    [anon_sym___thread] = ACTIONS(1142),
    [anon_sym_const] = ACTIONS(1142),
    [anon_sym_constexpr] = ACTIONS(1142),
    [anon_sym_volatile] = ACTIONS(1142),
    [anon_sym_restrict] = ACTIONS(1142),
    [anon_sym___restrict__] = ACTIONS(1142),
    [anon_sym__Atomic] = ACTIONS(1142),
    [anon_sym__Noreturn] = ACTIONS(1142),
    [anon_sym_noreturn] = ACTIONS(1142),
    [anon_sym__Nonnull] = ACTIONS(1142),
    [anon_sym_alignas] = ACTIONS(1142),
    [anon_sym__Alignas] = ACTIONS(1142),
    [sym_primitive_type] = ACTIONS(1142),
    [anon_sym_enum] = ACTIONS(1142),
    [anon_sym_struct] = ACTIONS(1142),
    [anon_sym_union] = ACTIONS(1142),
    [anon_sym_if] = ACTIONS(1142),
    [anon_sym_else] = ACTIONS(1142),
    [anon_sym_switch] = ACTIONS(1142),
    [anon_sym_case] = ACTIONS(1142),
    [anon_sym_default] = ACTIONS(1142),
    [anon_sym_while] = ACTIONS(1142),
    [anon_sym_do] = ACTIONS(1142),
    [anon_sym_for] = ACTIONS(1142),
    [anon_sym_return] = ACTIONS(1142),
    [anon_sym_break] = ACTIONS(1142),
    [anon_sym_continue] = ACTIONS(1142),
    [anon_sym_goto] = ACTIONS(1142),
    [anon_sym___try] = ACTIONS(1142),
    [anon_sym___leave] = ACTIONS(1142),
    [anon_sym_DASH_DASH] = ACTIONS(1144),
    [anon_sym_PLUS_PLUS] = ACTIONS(1144),
    [anon_sym_sizeof] = ACTIONS(1142),
    [anon_sym___alignof__] = ACTIONS(1142),
    [anon_sym___alignof] = ACTIONS(1142),
    [anon_sym__alignof] = ACTIONS(1142),
    [anon_sym_alignof] = ACTIONS(1142),
    [anon_sym__Alignof] = ACTIONS(1142),
    [anon_sym_offsetof] = ACTIONS(1142),
    [anon_sym__Generic] = ACTIONS(1142),
    [anon_sym_asm] = ACTIONS(1142),
    [anon_sym___asm__] = ACTIONS(1142),
    [anon_sym___asm] = ACTIONS(1142),
    [sym_number_literal] = ACTIONS(1144),
    [anon_sym_L_SQUOTE] = ACTIONS(1144),
    [anon_sym_u_SQUOTE] = ACTIONS(1144),
    [anon_sym_U_SQUOTE] = ACTIONS(1144),
    [anon_sym_u8_SQUOTE] = ACTIONS(1144),
    [anon_sym_SQUOTE] = ACTIONS(1144),
    [anon_sym_L_DQUOTE] = ACTIONS(1144),
    [anon_sym_u_DQUOTE] = ACTIONS(1144),
    [anon_sym_U_DQUOTE] = ACTIONS(1144),
    [anon_sym_u8_DQUOTE] = ACTIONS(1144),
    [anon_sym_DQUOTE] = ACTIONS(1144),
    [sym_true] = ACTIONS(1142),
    [sym_false] = ACTIONS(1142),
    [anon_sym_NULL] = ACTIONS(1142),
    [anon_sym_nullptr] = ACTIONS(1142),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(151)] = {
    [sym_identifier] = ACTIONS(1158),
    [aux_sym_preproc_include_token1] = ACTIONS(1158),
    [aux_sym_preproc_def_token1] = ACTIONS(1158),
    [aux_sym_preproc_if_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1158),
    [sym_preproc_directive] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_TILDE] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_STAR] = ACTIONS(1160),
    [anon_sym_AMP] = ACTIONS(1160),
    [anon_sym_SEMI] = ACTIONS(1160),
    [anon_sym___extension__] = ACTIONS(1158),
    [anon_sym_typedef] = ACTIONS(1158),
    [anon_sym_extern] = ACTIONS(1158),
    [anon_sym___attribute__] = ACTIONS(1158),
    [anon_sym___attribute] = ACTIONS(1158),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1160),
    [anon_sym___declspec] = ACTIONS(1158),
    [anon_sym___cdecl] = ACTIONS(1158),
    [anon_sym___clrcall] = ACTIONS(1158),
    [anon_sym___stdcall] = ACTIONS(1158),
    [anon_sym___fastcall] = ACTIONS(1158),
    [anon_sym___thiscall] = ACTIONS(1158),
    [anon_sym___vectorcall] = ACTIONS(1158),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_RBRACE] = ACTIONS(1160),
    [anon_sym_signed] = ACTIONS(1158),
    [anon_sym_unsigned] = ACTIONS(1158),
    [anon_sym_long] = ACTIONS(1158),
    [anon_sym_short] = ACTIONS(1158),
    [anon_sym_static] = ACTIONS(1158),
    [anon_sym_auto] = ACTIONS(1158),
    [anon_sym_register] = ACTIONS(1158),
    [anon_sym_inline] = ACTIONS(1158),
    [anon_sym___inline] = ACTIONS(1158),
    [anon_sym___inline__] = ACTIONS(1158),
    [anon_sym___forceinline] = ACTIONS(1158),
    [anon_sym_thread_local] = ACTIONS(1158),
    [anon_sym___thread] = ACTIONS(1158),
    [anon_sym_const] = ACTIONS(1158),
    [anon_sym_constexpr] = ACTIONS(1158),
    [anon_sym_volatile] = ACTIONS(1158),
    [anon_sym_restrict] = ACTIONS(1158),
    [anon_sym___restrict__] = ACTIONS(1158),
    [anon_sym__Atomic] = ACTIONS(1158),
    [anon_sym__Noreturn] = ACTIONS(1158),
    [anon_sym_noreturn] = ACTIONS(1158),
    [anon_sym__Nonnull] = ACTIONS(1158),
    [anon_sym_alignas] = ACTIONS(1158),
    [anon_sym__Alignas] = ACTIONS(1158),
    [sym_primitive_type] = ACTIONS(1158),
    [anon_sym_enum] = ACTIONS(1158),
    [anon_sym_struct] = ACTIONS(1158),
    [anon_sym_union] = ACTIONS(1158),
    [anon_sym_if] = ACTIONS(1158),
    [anon_sym_else] = ACTIONS(1158),
    [anon_sym_switch] = ACTIONS(1158),
    [anon_sym_case] = ACTIONS(1158),
    [anon_sym_default] = ACTIONS(1158),
    [anon_sym_while] = ACTIONS(1158),
    [anon_sym_do] = ACTIONS(1158),
    [anon_sym_for] = ACTIONS(1158),
    [anon_sym_return] = ACTIONS(1158),
    [anon_sym_break] = ACTIONS(1158),
    [anon_sym_continue] = ACTIONS(1158),
    [anon_sym_goto] = ACTIONS(1158),
    [anon_sym___try] = ACTIONS(1158),
    [anon_sym___leave] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1160),
    [anon_sym_PLUS_PLUS] = ACTIONS(1160),
    [anon_sym_sizeof] = ACTIONS(1158),
    [anon_sym___alignof__] = ACTIONS(1158),
    [anon_sym___alignof] = ACTIONS(1158),
    [anon_sym__alignof] = ACTIONS(1158),
    [anon_sym_alignof] = ACTIONS(1158),
    [anon_sym__Alignof] = ACTIONS(1158),
    [anon_sym_offsetof] = ACTIONS(1158),
    [anon_sym__Generic] = ACTIONS(1158),
    [anon_sym_asm] = ACTIONS(1158),
    [anon_sym___asm__] = ACTIONS(1158),
    [anon_sym___asm] = ACTIONS(1158),
    [sym_number_literal] = ACTIONS(1160),
    [anon_sym_L_SQUOTE] = ACTIONS(1160),
    [anon_sym_u_SQUOTE] = ACTIONS(1160),
    [anon_sym_U_SQUOTE] = ACTIONS(1160),
    [anon_sym_u8_SQUOTE] = ACTIONS(1160),
    [anon_sym_SQUOTE] = ACTIONS(1160),
    [anon_sym_L_DQUOTE] = ACTIONS(1160),
    [anon_sym_u_DQUOTE] = ACTIONS(1160),
    [anon_sym_U_DQUOTE] = ACTIONS(1160),
    [anon_sym_u8_DQUOTE] = ACTIONS(1160),
    [anon_sym_DQUOTE] = ACTIONS(1160),
    [sym_true] = ACTIONS(1158),
    [sym_false] = ACTIONS(1158),
    [anon_sym_NULL] = ACTIONS(1158),
    [anon_sym_nullptr] = ACTIONS(1158),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(152)] = {
    [ts_builtin_sym_end] = ACTIONS(1188),
    [sym_identifier] = ACTIONS(1186),
    [aux_sym_preproc_include_token1] = ACTIONS(1186),
    [aux_sym_preproc_def_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1186),
    [sym_preproc_directive] = ACTIONS(1186),
    [anon_sym_LPAREN2] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1186),
    [anon_sym_PLUS] = ACTIONS(1186),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym___extension__] = ACTIONS(1186),
    [anon_sym_typedef] = ACTIONS(1186),
    [anon_sym_extern] = ACTIONS(1186),
    [anon_sym___attribute__] = ACTIONS(1186),
    [anon_sym___attribute] = ACTIONS(1186),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1188),
    [anon_sym___declspec] = ACTIONS(1186),
    [anon_sym___cdecl] = ACTIONS(1186),
    [anon_sym___clrcall] = ACTIONS(1186),
    [anon_sym___stdcall] = ACTIONS(1186),
    [anon_sym___fastcall] = ACTIONS(1186),
    [anon_sym___thiscall] = ACTIONS(1186),
    [anon_sym___vectorcall] = ACTIONS(1186),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_signed] = ACTIONS(1186),
    [anon_sym_unsigned] = ACTIONS(1186),
    [anon_sym_long] = ACTIONS(1186),
    [anon_sym_short] = ACTIONS(1186),
    [anon_sym_static] = ACTIONS(1186),
    [anon_sym_auto] = ACTIONS(1186),
    [anon_sym_register] = ACTIONS(1186),
    [anon_sym_inline] = ACTIONS(1186),
    [anon_sym___inline] = ACTIONS(1186),
    [anon_sym___inline__] = ACTIONS(1186),
    [anon_sym___forceinline] = ACTIONS(1186),
    [anon_sym_thread_local] = ACTIONS(1186),
    [anon_sym___thread] = ACTIONS(1186),
    [anon_sym_const] = ACTIONS(1186),
    [anon_sym_constexpr] = ACTIONS(1186),
    [anon_sym_volatile] = ACTIONS(1186),
    [anon_sym_restrict] = ACTIONS(1186),
    [anon_sym___restrict__] = ACTIONS(1186),
    [anon_sym__Atomic] = ACTIONS(1186),
    [anon_sym__Noreturn] = ACTIONS(1186),
    [anon_sym_noreturn] = ACTIONS(1186),
    [anon_sym__Nonnull] = ACTIONS(1186),
    [anon_sym_alignas] = ACTIONS(1186),
    [anon_sym__Alignas] = ACTIONS(1186),
    [sym_primitive_type] = ACTIONS(1186),
    [anon_sym_enum] = ACTIONS(1186),
    [anon_sym_struct] = ACTIONS(1186),
    [anon_sym_union] = ACTIONS(1186),
    [anon_sym_if] = ACTIONS(1186),
    [anon_sym_else] = ACTIONS(1186),
    [anon_sym_switch] = ACTIONS(1186),
    [anon_sym_case] = ACTIONS(1186),
    [anon_sym_default] = ACTIONS(1186),
    [anon_sym_while] = ACTIONS(1186),
    [anon_sym_do] = ACTIONS(1186),
    [anon_sym_for] = ACTIONS(1186),
    [anon_sym_return] = ACTIONS(1186),
    [anon_sym_break] = ACTIONS(1186),
    [anon_sym_continue] = ACTIONS(1186),
    [anon_sym_goto] = ACTIONS(1186),
    [anon_sym___try] = ACTIONS(1186),
    [anon_sym___leave] = ACTIONS(1186),
    [anon_sym_DASH_DASH] = ACTIONS(1188),
    [anon_sym_PLUS_PLUS] = ACTIONS(1188),
    [anon_sym_sizeof] = ACTIONS(1186),
    [anon_sym___alignof__] = ACTIONS(1186),
    [anon_sym___alignof] = ACTIONS(1186),
    [anon_sym__alignof] = ACTIONS(1186),
    [anon_sym_alignof] = ACTIONS(1186),
    [anon_sym__Alignof] = ACTIONS(1186),
    [anon_sym_offsetof] = ACTIONS(1186),
    [anon_sym__Generic] = ACTIONS(1186),
    [anon_sym_asm] = ACTIONS(1186),
    [anon_sym___asm__] = ACTIONS(1186),
    [anon_sym___asm] = ACTIONS(1186),
    [sym_number_literal] = ACTIONS(1188),
    [anon_sym_L_SQUOTE] = ACTIONS(1188),
    [anon_sym_u_SQUOTE] = ACTIONS(1188),
    [anon_sym_U_SQUOTE] = ACTIONS(1188),
    [anon_sym_u8_SQUOTE] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_L_DQUOTE] = ACTIONS(1188),
    [anon_sym_u_DQUOTE] = ACTIONS(1188),
    [anon_sym_U_DQUOTE] = ACTIONS(1188),
    [anon_sym_u8_DQUOTE] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [sym_true] = ACTIONS(1186),
    [sym_false] = ACTIONS(1186),
    [anon_sym_NULL] = ACTIONS(1186),
    [anon_sym_nullptr] = ACTIONS(1186),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(153)] = {
    [sym_identifier] = ACTIONS(1146),
    [aux_sym_preproc_include_token1] = ACTIONS(1146),
    [aux_sym_preproc_def_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1146),
    [sym_preproc_directive] = ACTIONS(1146),
    [anon_sym_LPAREN2] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_TILDE] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1146),
    [anon_sym_PLUS] = ACTIONS(1146),
    [anon_sym_STAR] = ACTIONS(1148),
    [anon_sym_AMP] = ACTIONS(1148),
    [anon_sym_SEMI] = ACTIONS(1148),
    [anon_sym___extension__] = ACTIONS(1146),
    [anon_sym_typedef] = ACTIONS(1146),
    [anon_sym_extern] = ACTIONS(1146),
    [anon_sym___attribute__] = ACTIONS(1146),
    [anon_sym___attribute] = ACTIONS(1146),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1148),
    [anon_sym___declspec] = ACTIONS(1146),
    [anon_sym___cdecl] = ACTIONS(1146),
    [anon_sym___clrcall] = ACTIONS(1146),
    [anon_sym___stdcall] = ACTIONS(1146),
    [anon_sym___fastcall] = ACTIONS(1146),
    [anon_sym___thiscall] = ACTIONS(1146),
    [anon_sym___vectorcall] = ACTIONS(1146),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_RBRACE] = ACTIONS(1148),
    [anon_sym_signed] = ACTIONS(1146),
    [anon_sym_unsigned] = ACTIONS(1146),
    [anon_sym_long] = ACTIONS(1146),
    [anon_sym_short] = ACTIONS(1146),
    [anon_sym_static] = ACTIONS(1146),
    [anon_sym_auto] = ACTIONS(1146),
    [anon_sym_register] = ACTIONS(1146),
    [anon_sym_inline] = ACTIONS(1146),
    [anon_sym___inline] = ACTIONS(1146),
    [anon_sym___inline__] = ACTIONS(1146),
    [anon_sym___forceinline] = ACTIONS(1146),
    [anon_sym_thread_local] = ACTIONS(1146),
    [anon_sym___thread] = ACTIONS(1146),
    [anon_sym_const] = ACTIONS(1146),
    [anon_sym_constexpr] = ACTIONS(1146),
    [anon_sym_volatile] = ACTIONS(1146),
    [anon_sym_restrict] = ACTIONS(1146),
    [anon_sym___restrict__] = ACTIONS(1146),
    [anon_sym__Atomic] = ACTIONS(1146),
    [anon_sym__Noreturn] = ACTIONS(1146),
    [anon_sym_noreturn] = ACTIONS(1146),
    [anon_sym__Nonnull] = ACTIONS(1146),
    [anon_sym_alignas] = ACTIONS(1146),
    [anon_sym__Alignas] = ACTIONS(1146),
    [sym_primitive_type] = ACTIONS(1146),
    [anon_sym_enum] = ACTIONS(1146),
    [anon_sym_struct] = ACTIONS(1146),
    [anon_sym_union] = ACTIONS(1146),
    [anon_sym_if] = ACTIONS(1146),
    [anon_sym_else] = ACTIONS(1146),
    [anon_sym_switch] = ACTIONS(1146),
    [anon_sym_case] = ACTIONS(1146),
    [anon_sym_default] = ACTIONS(1146),
    [anon_sym_while] = ACTIONS(1146),
    [anon_sym_do] = ACTIONS(1146),
    [anon_sym_for] = ACTIONS(1146),
    [anon_sym_return] = ACTIONS(1146),
    [anon_sym_break] = ACTIONS(1146),
    [anon_sym_continue] = ACTIONS(1146),
    [anon_sym_goto] = ACTIONS(1146),
    [anon_sym___try] = ACTIONS(1146),
    [anon_sym___leave] = ACTIONS(1146),
    [anon_sym_DASH_DASH] = ACTIONS(1148),
    [anon_sym_PLUS_PLUS] = ACTIONS(1148),
    [anon_sym_sizeof] = ACTIONS(1146),
    [anon_sym___alignof__] = ACTIONS(1146),
    [anon_sym___alignof] = ACTIONS(1146),
    [anon_sym__alignof] = ACTIONS(1146),
    [anon_sym_alignof] = ACTIONS(1146),
    [anon_sym__Alignof] = ACTIONS(1146),
    [anon_sym_offsetof] = ACTIONS(1146),
    [anon_sym__Generic] = ACTIONS(1146),
    [anon_sym_asm] = ACTIONS(1146),
    [anon_sym___asm__] = ACTIONS(1146),
    [anon_sym___asm] = ACTIONS(1146),
    [sym_number_literal] = ACTIONS(1148),
    [anon_sym_L_SQUOTE] = ACTIONS(1148),
    [anon_sym_u_SQUOTE] = ACTIONS(1148),
    [anon_sym_U_SQUOTE] = ACTIONS(1148),
    [anon_sym_u8_SQUOTE] = ACTIONS(1148),
    [anon_sym_SQUOTE] = ACTIONS(1148),
    [anon_sym_L_DQUOTE] = ACTIONS(1148),
    [anon_sym_u_DQUOTE] = ACTIONS(1148),
    [anon_sym_U_DQUOTE] = ACTIONS(1148),
    [anon_sym_u8_DQUOTE] = ACTIONS(1148),
    [anon_sym_DQUOTE] = ACTIONS(1148),
    [sym_true] = ACTIONS(1146),
    [sym_false] = ACTIONS(1146),
    [anon_sym_NULL] = ACTIONS(1146),
    [anon_sym_nullptr] = ACTIONS(1146),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(154)] = {
    [ts_builtin_sym_end] = ACTIONS(1136),
    [sym_identifier] = ACTIONS(1134),
    [aux_sym_preproc_include_token1] = ACTIONS(1134),
    [aux_sym_preproc_def_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1134),
    [sym_preproc_directive] = ACTIONS(1134),
    [anon_sym_LPAREN2] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_TILDE] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1134),
    [anon_sym_PLUS] = ACTIONS(1134),
    [anon_sym_STAR] = ACTIONS(1136),
    [anon_sym_AMP] = ACTIONS(1136),
    [anon_sym_SEMI] = ACTIONS(1136),
    [anon_sym___extension__] = ACTIONS(1134),
    [anon_sym_typedef] = ACTIONS(1134),
    [anon_sym_extern] = ACTIONS(1134),
    [anon_sym___attribute__] = ACTIONS(1134),
    [anon_sym___attribute] = ACTIONS(1134),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1136),
    [anon_sym___declspec] = ACTIONS(1134),
    [anon_sym___cdecl] = ACTIONS(1134),
    [anon_sym___clrcall] = ACTIONS(1134),
    [anon_sym___stdcall] = ACTIONS(1134),
    [anon_sym___fastcall] = ACTIONS(1134),
    [anon_sym___thiscall] = ACTIONS(1134),
    [anon_sym___vectorcall] = ACTIONS(1134),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_signed] = ACTIONS(1134),
    [anon_sym_unsigned] = ACTIONS(1134),
    [anon_sym_long] = ACTIONS(1134),
    [anon_sym_short] = ACTIONS(1134),
    [anon_sym_static] = ACTIONS(1134),
    [anon_sym_auto] = ACTIONS(1134),
    [anon_sym_register] = ACTIONS(1134),
    [anon_sym_inline] = ACTIONS(1134),
    [anon_sym___inline] = ACTIONS(1134),
    [anon_sym___inline__] = ACTIONS(1134),
    [anon_sym___forceinline] = ACTIONS(1134),
    [anon_sym_thread_local] = ACTIONS(1134),
    [anon_sym___thread] = ACTIONS(1134),
    [anon_sym_const] = ACTIONS(1134),
    [anon_sym_constexpr] = ACTIONS(1134),
    [anon_sym_volatile] = ACTIONS(1134),
    [anon_sym_restrict] = ACTIONS(1134),
    [anon_sym___restrict__] = ACTIONS(1134),
    [anon_sym__Atomic] = ACTIONS(1134),
    [anon_sym__Noreturn] = ACTIONS(1134),
    [anon_sym_noreturn] = ACTIONS(1134),
    [anon_sym__Nonnull] = ACTIONS(1134),
    [anon_sym_alignas] = ACTIONS(1134),
    [anon_sym__Alignas] = ACTIONS(1134),
    [sym_primitive_type] = ACTIONS(1134),
    [anon_sym_enum] = ACTIONS(1134),
    [anon_sym_struct] = ACTIONS(1134),
    [anon_sym_union] = ACTIONS(1134),
    [anon_sym_if] = ACTIONS(1134),
    [anon_sym_else] = ACTIONS(1134),
    [anon_sym_switch] = ACTIONS(1134),
    [anon_sym_case] = ACTIONS(1134),
    [anon_sym_default] = ACTIONS(1134),
    [anon_sym_while] = ACTIONS(1134),
    [anon_sym_do] = ACTIONS(1134),
    [anon_sym_for] = ACTIONS(1134),
    [anon_sym_return] = ACTIONS(1134),
    [anon_sym_break] = ACTIONS(1134),
    [anon_sym_continue] = ACTIONS(1134),
    [anon_sym_goto] = ACTIONS(1134),
    [anon_sym___try] = ACTIONS(1134),
    [anon_sym___leave] = ACTIONS(1134),
    [anon_sym_DASH_DASH] = ACTIONS(1136),
    [anon_sym_PLUS_PLUS] = ACTIONS(1136),
    [anon_sym_sizeof] = ACTIONS(1134),
    [anon_sym___alignof__] = ACTIONS(1134),
    [anon_sym___alignof] = ACTIONS(1134),
    [anon_sym__alignof] = ACTIONS(1134),
    [anon_sym_alignof] = ACTIONS(1134),
    [anon_sym__Alignof] = ACTIONS(1134),
    [anon_sym_offsetof] = ACTIONS(1134),
    [anon_sym__Generic] = ACTIONS(1134),
    [anon_sym_asm] = ACTIONS(1134),
    [anon_sym___asm__] = ACTIONS(1134),
    [anon_sym___asm] = ACTIONS(1134),
    [sym_number_literal] = ACTIONS(1136),
    [anon_sym_L_SQUOTE] = ACTIONS(1136),
    [anon_sym_u_SQUOTE] = ACTIONS(1136),
    [anon_sym_U_SQUOTE] = ACTIONS(1136),
    [anon_sym_u8_SQUOTE] = ACTIONS(1136),
    [anon_sym_SQUOTE] = ACTIONS(1136),
    [anon_sym_L_DQUOTE] = ACTIONS(1136),
    [anon_sym_u_DQUOTE] = ACTIONS(1136),
    [anon_sym_U_DQUOTE] = ACTIONS(1136),
    [anon_sym_u8_DQUOTE] = ACTIONS(1136),
    [anon_sym_DQUOTE] = ACTIONS(1136),
    [sym_true] = ACTIONS(1134),
    [sym_false] = ACTIONS(1134),
    [anon_sym_NULL] = ACTIONS(1134),
    [anon_sym_nullptr] = ACTIONS(1134),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(155)] = {
    [ts_builtin_sym_end] = ACTIONS(1152),
    [sym_identifier] = ACTIONS(1150),
    [aux_sym_preproc_include_token1] = ACTIONS(1150),
    [aux_sym_preproc_def_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1150),
    [sym_preproc_directive] = ACTIONS(1150),
    [anon_sym_LPAREN2] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_TILDE] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1150),
    [anon_sym_PLUS] = ACTIONS(1150),
    [anon_sym_STAR] = ACTIONS(1152),
    [anon_sym_AMP] = ACTIONS(1152),
    [anon_sym_SEMI] = ACTIONS(1152),
    [anon_sym___extension__] = ACTIONS(1150),
    [anon_sym_typedef] = ACTIONS(1150),
    [anon_sym_extern] = ACTIONS(1150),
    [anon_sym___attribute__] = ACTIONS(1150),
    [anon_sym___attribute] = ACTIONS(1150),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1152),
    [anon_sym___declspec] = ACTIONS(1150),
    [anon_sym___cdecl] = ACTIONS(1150),
    [anon_sym___clrcall] = ACTIONS(1150),
    [anon_sym___stdcall] = ACTIONS(1150),
    [anon_sym___fastcall] = ACTIONS(1150),
    [anon_sym___thiscall] = ACTIONS(1150),
    [anon_sym___vectorcall] = ACTIONS(1150),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_signed] = ACTIONS(1150),
    [anon_sym_unsigned] = ACTIONS(1150),
    [anon_sym_long] = ACTIONS(1150),
    [anon_sym_short] = ACTIONS(1150),
    [anon_sym_static] = ACTIONS(1150),
    [anon_sym_auto] = ACTIONS(1150),
    [anon_sym_register] = ACTIONS(1150),
    [anon_sym_inline] = ACTIONS(1150),
    [anon_sym___inline] = ACTIONS(1150),
    [anon_sym___inline__] = ACTIONS(1150),
    [anon_sym___forceinline] = ACTIONS(1150),
    [anon_sym_thread_local] = ACTIONS(1150),
    [anon_sym___thread] = ACTIONS(1150),
    [anon_sym_const] = ACTIONS(1150),
    [anon_sym_constexpr] = ACTIONS(1150),
    [anon_sym_volatile] = ACTIONS(1150),
    [anon_sym_restrict] = ACTIONS(1150),
    [anon_sym___restrict__] = ACTIONS(1150),
    [anon_sym__Atomic] = ACTIONS(1150),
    [anon_sym__Noreturn] = ACTIONS(1150),
    [anon_sym_noreturn] = ACTIONS(1150),
    [anon_sym__Nonnull] = ACTIONS(1150),
    [anon_sym_alignas] = ACTIONS(1150),
    [anon_sym__Alignas] = ACTIONS(1150),
    [sym_primitive_type] = ACTIONS(1150),
    [anon_sym_enum] = ACTIONS(1150),
    [anon_sym_struct] = ACTIONS(1150),
    [anon_sym_union] = ACTIONS(1150),
    [anon_sym_if] = ACTIONS(1150),
    [anon_sym_else] = ACTIONS(1150),
    [anon_sym_switch] = ACTIONS(1150),
    [anon_sym_case] = ACTIONS(1150),
    [anon_sym_default] = ACTIONS(1150),
    [anon_sym_while] = ACTIONS(1150),
    [anon_sym_do] = ACTIONS(1150),
    [anon_sym_for] = ACTIONS(1150),
    [anon_sym_return] = ACTIONS(1150),
    [anon_sym_break] = ACTIONS(1150),
    [anon_sym_continue] = ACTIONS(1150),
    [anon_sym_goto] = ACTIONS(1150),
    [anon_sym___try] = ACTIONS(1150),
    [anon_sym___leave] = ACTIONS(1150),
    [anon_sym_DASH_DASH] = ACTIONS(1152),
    [anon_sym_PLUS_PLUS] = ACTIONS(1152),
    [anon_sym_sizeof] = ACTIONS(1150),
    [anon_sym___alignof__] = ACTIONS(1150),
    [anon_sym___alignof] = ACTIONS(1150),
    [anon_sym__alignof] = ACTIONS(1150),
    [anon_sym_alignof] = ACTIONS(1150),
    [anon_sym__Alignof] = ACTIONS(1150),
    [anon_sym_offsetof] = ACTIONS(1150),
    [anon_sym__Generic] = ACTIONS(1150),
    [anon_sym_asm] = ACTIONS(1150),
    [anon_sym___asm__] = ACTIONS(1150),
    [anon_sym___asm] = ACTIONS(1150),
    [sym_number_literal] = ACTIONS(1152),
    [anon_sym_L_SQUOTE] = ACTIONS(1152),
    [anon_sym_u_SQUOTE] = ACTIONS(1152),
    [anon_sym_U_SQUOTE] = ACTIONS(1152),
    [anon_sym_u8_SQUOTE] = ACTIONS(1152),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_L_DQUOTE] = ACTIONS(1152),
    [anon_sym_u_DQUOTE] = ACTIONS(1152),
    [anon_sym_U_DQUOTE] = ACTIONS(1152),
    [anon_sym_u8_DQUOTE] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1152),
    [sym_true] = ACTIONS(1150),
    [sym_false] = ACTIONS(1150),
    [anon_sym_NULL] = ACTIONS(1150),
    [anon_sym_nullptr] = ACTIONS(1150),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(156)] = {
    [ts_builtin_sym_end] = ACTIONS(1168),
    [sym_identifier] = ACTIONS(1166),
    [aux_sym_preproc_include_token1] = ACTIONS(1166),
    [aux_sym_preproc_def_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1166),
    [sym_preproc_directive] = ACTIONS(1166),
    [anon_sym_LPAREN2] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_TILDE] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1166),
    [anon_sym_PLUS] = ACTIONS(1166),
    [anon_sym_STAR] = ACTIONS(1168),
    [anon_sym_AMP] = ACTIONS(1168),
    [anon_sym_SEMI] = ACTIONS(1168),
    [anon_sym___extension__] = ACTIONS(1166),
    [anon_sym_typedef] = ACTIONS(1166),
    [anon_sym_extern] = ACTIONS(1166),
    [anon_sym___attribute__] = ACTIONS(1166),
    [anon_sym___attribute] = ACTIONS(1166),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1168),
    [anon_sym___declspec] = ACTIONS(1166),
    [anon_sym___cdecl] = ACTIONS(1166),
    [anon_sym___clrcall] = ACTIONS(1166),
    [anon_sym___stdcall] = ACTIONS(1166),
    [anon_sym___fastcall] = ACTIONS(1166),
    [anon_sym___thiscall] = ACTIONS(1166),
    [anon_sym___vectorcall] = ACTIONS(1166),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_signed] = ACTIONS(1166),
    [anon_sym_unsigned] = ACTIONS(1166),
    [anon_sym_long] = ACTIONS(1166),
    [anon_sym_short] = ACTIONS(1166),
    [anon_sym_static] = ACTIONS(1166),
    [anon_sym_auto] = ACTIONS(1166),
    [anon_sym_register] = ACTIONS(1166),
    [anon_sym_inline] = ACTIONS(1166),
    [anon_sym___inline] = ACTIONS(1166),
    [anon_sym___inline__] = ACTIONS(1166),
    [anon_sym___forceinline] = ACTIONS(1166),
    [anon_sym_thread_local] = ACTIONS(1166),
    [anon_sym___thread] = ACTIONS(1166),
    [anon_sym_const] = ACTIONS(1166),
    [anon_sym_constexpr] = ACTIONS(1166),
    [anon_sym_volatile] = ACTIONS(1166),
    [anon_sym_restrict] = ACTIONS(1166),
    [anon_sym___restrict__] = ACTIONS(1166),
    [anon_sym__Atomic] = ACTIONS(1166),
    [anon_sym__Noreturn] = ACTIONS(1166),
    [anon_sym_noreturn] = ACTIONS(1166),
    [anon_sym__Nonnull] = ACTIONS(1166),
    [anon_sym_alignas] = ACTIONS(1166),
    [anon_sym__Alignas] = ACTIONS(1166),
    [sym_primitive_type] = ACTIONS(1166),
    [anon_sym_enum] = ACTIONS(1166),
    [anon_sym_struct] = ACTIONS(1166),
    [anon_sym_union] = ACTIONS(1166),
    [anon_sym_if] = ACTIONS(1166),
    [anon_sym_else] = ACTIONS(1166),
    [anon_sym_switch] = ACTIONS(1166),
    [anon_sym_case] = ACTIONS(1166),
    [anon_sym_default] = ACTIONS(1166),
    [anon_sym_while] = ACTIONS(1166),
    [anon_sym_do] = ACTIONS(1166),
    [anon_sym_for] = ACTIONS(1166),
    [anon_sym_return] = ACTIONS(1166),
    [anon_sym_break] = ACTIONS(1166),
    [anon_sym_continue] = ACTIONS(1166),
    [anon_sym_goto] = ACTIONS(1166),
    [anon_sym___try] = ACTIONS(1166),
    [anon_sym___leave] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_PLUS_PLUS] = ACTIONS(1168),
    [anon_sym_sizeof] = ACTIONS(1166),
    [anon_sym___alignof__] = ACTIONS(1166),
    [anon_sym___alignof] = ACTIONS(1166),
    [anon_sym__alignof] = ACTIONS(1166),
    [anon_sym_alignof] = ACTIONS(1166),
    [anon_sym__Alignof] = ACTIONS(1166),
    [anon_sym_offsetof] = ACTIONS(1166),
    [anon_sym__Generic] = ACTIONS(1166),
    [anon_sym_asm] = ACTIONS(1166),
    [anon_sym___asm__] = ACTIONS(1166),
    [anon_sym___asm] = ACTIONS(1166),
    [sym_number_literal] = ACTIONS(1168),
    [anon_sym_L_SQUOTE] = ACTIONS(1168),
    [anon_sym_u_SQUOTE] = ACTIONS(1168),
    [anon_sym_U_SQUOTE] = ACTIONS(1168),
    [anon_sym_u8_SQUOTE] = ACTIONS(1168),
    [anon_sym_SQUOTE] = ACTIONS(1168),
    [anon_sym_L_DQUOTE] = ACTIONS(1168),
    [anon_sym_u_DQUOTE] = ACTIONS(1168),
    [anon_sym_U_DQUOTE] = ACTIONS(1168),
    [anon_sym_u8_DQUOTE] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1168),
    [sym_true] = ACTIONS(1166),
    [sym_false] = ACTIONS(1166),
    [anon_sym_NULL] = ACTIONS(1166),
    [anon_sym_nullptr] = ACTIONS(1166),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(157)] = {
    [ts_builtin_sym_end] = ACTIONS(1236),
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(158)] = {
    [ts_builtin_sym_end] = ACTIONS(1172),
    [sym_identifier] = ACTIONS(1170),
    [aux_sym_preproc_include_token1] = ACTIONS(1170),
    [aux_sym_preproc_def_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1170),
    [sym_preproc_directive] = ACTIONS(1170),
    [anon_sym_LPAREN2] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_TILDE] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1172),
    [anon_sym_AMP] = ACTIONS(1172),
    [anon_sym_SEMI] = ACTIONS(1172),
    [anon_sym___extension__] = ACTIONS(1170),
    [anon_sym_typedef] = ACTIONS(1170),
    [anon_sym_extern] = ACTIONS(1170),
    [anon_sym___attribute__] = ACTIONS(1170),
    [anon_sym___attribute] = ACTIONS(1170),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1172),
    [anon_sym___declspec] = ACTIONS(1170),
    [anon_sym___cdecl] = ACTIONS(1170),
    [anon_sym___clrcall] = ACTIONS(1170),
    [anon_sym___stdcall] = ACTIONS(1170),
    [anon_sym___fastcall] = ACTIONS(1170),
    [anon_sym___thiscall] = ACTIONS(1170),
    [anon_sym___vectorcall] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_signed] = ACTIONS(1170),
    [anon_sym_unsigned] = ACTIONS(1170),
    [anon_sym_long] = ACTIONS(1170),
    [anon_sym_short] = ACTIONS(1170),
    [anon_sym_static] = ACTIONS(1170),
    [anon_sym_auto] = ACTIONS(1170),
    [anon_sym_register] = ACTIONS(1170),
    [anon_sym_inline] = ACTIONS(1170),
    [anon_sym___inline] = ACTIONS(1170),
    [anon_sym___inline__] = ACTIONS(1170),
    [anon_sym___forceinline] = ACTIONS(1170),
    [anon_sym_thread_local] = ACTIONS(1170),
    [anon_sym___thread] = ACTIONS(1170),
    [anon_sym_const] = ACTIONS(1170),
    [anon_sym_constexpr] = ACTIONS(1170),
    [anon_sym_volatile] = ACTIONS(1170),
    [anon_sym_restrict] = ACTIONS(1170),
    [anon_sym___restrict__] = ACTIONS(1170),
    [anon_sym__Atomic] = ACTIONS(1170),
    [anon_sym__Noreturn] = ACTIONS(1170),
    [anon_sym_noreturn] = ACTIONS(1170),
    [anon_sym__Nonnull] = ACTIONS(1170),
    [anon_sym_alignas] = ACTIONS(1170),
    [anon_sym__Alignas] = ACTIONS(1170),
    [sym_primitive_type] = ACTIONS(1170),
    [anon_sym_enum] = ACTIONS(1170),
    [anon_sym_struct] = ACTIONS(1170),
    [anon_sym_union] = ACTIONS(1170),
    [anon_sym_if] = ACTIONS(1170),
    [anon_sym_else] = ACTIONS(1170),
    [anon_sym_switch] = ACTIONS(1170),
    [anon_sym_case] = ACTIONS(1170),
    [anon_sym_default] = ACTIONS(1170),
    [anon_sym_while] = ACTIONS(1170),
    [anon_sym_do] = ACTIONS(1170),
    [anon_sym_for] = ACTIONS(1170),
    [anon_sym_return] = ACTIONS(1170),
    [anon_sym_break] = ACTIONS(1170),
    [anon_sym_continue] = ACTIONS(1170),
    [anon_sym_goto] = ACTIONS(1170),
    [anon_sym___try] = ACTIONS(1170),
    [anon_sym___leave] = ACTIONS(1170),
    [anon_sym_DASH_DASH] = ACTIONS(1172),
    [anon_sym_PLUS_PLUS] = ACTIONS(1172),
    [anon_sym_sizeof] = ACTIONS(1170),
    [anon_sym___alignof__] = ACTIONS(1170),
    [anon_sym___alignof] = ACTIONS(1170),
    [anon_sym__alignof] = ACTIONS(1170),
    [anon_sym_alignof] = ACTIONS(1170),
    [anon_sym__Alignof] = ACTIONS(1170),
    [anon_sym_offsetof] = ACTIONS(1170),
    [anon_sym__Generic] = ACTIONS(1170),
    [anon_sym_asm] = ACTIONS(1170),
    [anon_sym___asm__] = ACTIONS(1170),
    [anon_sym___asm] = ACTIONS(1170),
    [sym_number_literal] = ACTIONS(1172),
    [anon_sym_L_SQUOTE] = ACTIONS(1172),
    [anon_sym_u_SQUOTE] = ACTIONS(1172),
    [anon_sym_U_SQUOTE] = ACTIONS(1172),
    [anon_sym_u8_SQUOTE] = ACTIONS(1172),
    [anon_sym_SQUOTE] = ACTIONS(1172),
    [anon_sym_L_DQUOTE] = ACTIONS(1172),
    [anon_sym_u_DQUOTE] = ACTIONS(1172),
    [anon_sym_U_DQUOTE] = ACTIONS(1172),
    [anon_sym_u8_DQUOTE] = ACTIONS(1172),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [sym_true] = ACTIONS(1170),
    [sym_false] = ACTIONS(1170),
    [anon_sym_NULL] = ACTIONS(1170),
    [anon_sym_nullptr] = ACTIONS(1170),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(159)] = {
    [ts_builtin_sym_end] = ACTIONS(1176),
    [sym_identifier] = ACTIONS(1174),
    [aux_sym_preproc_include_token1] = ACTIONS(1174),
    [aux_sym_preproc_def_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1174),
    [sym_preproc_directive] = ACTIONS(1174),
    [anon_sym_LPAREN2] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_TILDE] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1174),
    [anon_sym_PLUS] = ACTIONS(1174),
    [anon_sym_STAR] = ACTIONS(1176),
    [anon_sym_AMP] = ACTIONS(1176),
    [anon_sym_SEMI] = ACTIONS(1176),
    [anon_sym___extension__] = ACTIONS(1174),
    [anon_sym_typedef] = ACTIONS(1174),
    [anon_sym_extern] = ACTIONS(1174),
    [anon_sym___attribute__] = ACTIONS(1174),
    [anon_sym___attribute] = ACTIONS(1174),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1176),
    [anon_sym___declspec] = ACTIONS(1174),
    [anon_sym___cdecl] = ACTIONS(1174),
    [anon_sym___clrcall] = ACTIONS(1174),
    [anon_sym___stdcall] = ACTIONS(1174),
    [anon_sym___fastcall] = ACTIONS(1174),
    [anon_sym___thiscall] = ACTIONS(1174),
    [anon_sym___vectorcall] = ACTIONS(1174),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_signed] = ACTIONS(1174),
    [anon_sym_unsigned] = ACTIONS(1174),
    [anon_sym_long] = ACTIONS(1174),
    [anon_sym_short] = ACTIONS(1174),
    [anon_sym_static] = ACTIONS(1174),
    [anon_sym_auto] = ACTIONS(1174),
    [anon_sym_register] = ACTIONS(1174),
    [anon_sym_inline] = ACTIONS(1174),
    [anon_sym___inline] = ACTIONS(1174),
    [anon_sym___inline__] = ACTIONS(1174),
    [anon_sym___forceinline] = ACTIONS(1174),
    [anon_sym_thread_local] = ACTIONS(1174),
    [anon_sym___thread] = ACTIONS(1174),
    [anon_sym_const] = ACTIONS(1174),
    [anon_sym_constexpr] = ACTIONS(1174),
    [anon_sym_volatile] = ACTIONS(1174),
    [anon_sym_restrict] = ACTIONS(1174),
    [anon_sym___restrict__] = ACTIONS(1174),
    [anon_sym__Atomic] = ACTIONS(1174),
    [anon_sym__Noreturn] = ACTIONS(1174),
    [anon_sym_noreturn] = ACTIONS(1174),
    [anon_sym__Nonnull] = ACTIONS(1174),
    [anon_sym_alignas] = ACTIONS(1174),
    [anon_sym__Alignas] = ACTIONS(1174),
    [sym_primitive_type] = ACTIONS(1174),
    [anon_sym_enum] = ACTIONS(1174),
    [anon_sym_struct] = ACTIONS(1174),
    [anon_sym_union] = ACTIONS(1174),
    [anon_sym_if] = ACTIONS(1174),
    [anon_sym_else] = ACTIONS(1174),
    [anon_sym_switch] = ACTIONS(1174),
    [anon_sym_case] = ACTIONS(1174),
    [anon_sym_default] = ACTIONS(1174),
    [anon_sym_while] = ACTIONS(1174),
    [anon_sym_do] = ACTIONS(1174),
    [anon_sym_for] = ACTIONS(1174),
    [anon_sym_return] = ACTIONS(1174),
    [anon_sym_break] = ACTIONS(1174),
    [anon_sym_continue] = ACTIONS(1174),
    [anon_sym_goto] = ACTIONS(1174),
    [anon_sym___try] = ACTIONS(1174),
    [anon_sym___leave] = ACTIONS(1174),
    [anon_sym_DASH_DASH] = ACTIONS(1176),
    [anon_sym_PLUS_PLUS] = ACTIONS(1176),
    [anon_sym_sizeof] = ACTIONS(1174),
    [anon_sym___alignof__] = ACTIONS(1174),
    [anon_sym___alignof] = ACTIONS(1174),
    [anon_sym__alignof] = ACTIONS(1174),
    [anon_sym_alignof] = ACTIONS(1174),
    [anon_sym__Alignof] = ACTIONS(1174),
    [anon_sym_offsetof] = ACTIONS(1174),
    [anon_sym__Generic] = ACTIONS(1174),
    [anon_sym_asm] = ACTIONS(1174),
    [anon_sym___asm__] = ACTIONS(1174),
    [anon_sym___asm] = ACTIONS(1174),
    [sym_number_literal] = ACTIONS(1176),
    [anon_sym_L_SQUOTE] = ACTIONS(1176),
    [anon_sym_u_SQUOTE] = ACTIONS(1176),
    [anon_sym_U_SQUOTE] = ACTIONS(1176),
    [anon_sym_u8_SQUOTE] = ACTIONS(1176),
    [anon_sym_SQUOTE] = ACTIONS(1176),
    [anon_sym_L_DQUOTE] = ACTIONS(1176),
    [anon_sym_u_DQUOTE] = ACTIONS(1176),
    [anon_sym_U_DQUOTE] = ACTIONS(1176),
    [anon_sym_u8_DQUOTE] = ACTIONS(1176),
    [anon_sym_DQUOTE] = ACTIONS(1176),
    [sym_true] = ACTIONS(1174),
    [sym_false] = ACTIONS(1174),
    [anon_sym_NULL] = ACTIONS(1174),
    [anon_sym_nullptr] = ACTIONS(1174),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(160)] = {
    [ts_builtin_sym_end] = ACTIONS(1236),
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(161)] = {
    [ts_builtin_sym_end] = ACTIONS(1228),
    [sym_identifier] = ACTIONS(1226),
    [aux_sym_preproc_include_token1] = ACTIONS(1226),
    [aux_sym_preproc_def_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1226),
    [sym_preproc_directive] = ACTIONS(1226),
    [anon_sym_LPAREN2] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1226),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym___extension__] = ACTIONS(1226),
    [anon_sym_typedef] = ACTIONS(1226),
    [anon_sym_extern] = ACTIONS(1226),
    [anon_sym___attribute__] = ACTIONS(1226),
    [anon_sym___attribute] = ACTIONS(1226),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1228),
    [anon_sym___declspec] = ACTIONS(1226),
    [anon_sym___cdecl] = ACTIONS(1226),
    [anon_sym___clrcall] = ACTIONS(1226),
    [anon_sym___stdcall] = ACTIONS(1226),
    [anon_sym___fastcall] = ACTIONS(1226),
    [anon_sym___thiscall] = ACTIONS(1226),
    [anon_sym___vectorcall] = ACTIONS(1226),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_signed] = ACTIONS(1226),
    [anon_sym_unsigned] = ACTIONS(1226),
    [anon_sym_long] = ACTIONS(1226),
    [anon_sym_short] = ACTIONS(1226),
    [anon_sym_static] = ACTIONS(1226),
    [anon_sym_auto] = ACTIONS(1226),
    [anon_sym_register] = ACTIONS(1226),
    [anon_sym_inline] = ACTIONS(1226),
    [anon_sym___inline] = ACTIONS(1226),
    [anon_sym___inline__] = ACTIONS(1226),
    [anon_sym___forceinline] = ACTIONS(1226),
    [anon_sym_thread_local] = ACTIONS(1226),
    [anon_sym___thread] = ACTIONS(1226),
    [anon_sym_const] = ACTIONS(1226),
    [anon_sym_constexpr] = ACTIONS(1226),
    [anon_sym_volatile] = ACTIONS(1226),
    [anon_sym_restrict] = ACTIONS(1226),
    [anon_sym___restrict__] = ACTIONS(1226),
    [anon_sym__Atomic] = ACTIONS(1226),
    [anon_sym__Noreturn] = ACTIONS(1226),
    [anon_sym_noreturn] = ACTIONS(1226),
    [anon_sym__Nonnull] = ACTIONS(1226),
    [anon_sym_alignas] = ACTIONS(1226),
    [anon_sym__Alignas] = ACTIONS(1226),
    [sym_primitive_type] = ACTIONS(1226),
    [anon_sym_enum] = ACTIONS(1226),
    [anon_sym_struct] = ACTIONS(1226),
    [anon_sym_union] = ACTIONS(1226),
    [anon_sym_if] = ACTIONS(1226),
    [anon_sym_else] = ACTIONS(1226),
    [anon_sym_switch] = ACTIONS(1226),
    [anon_sym_case] = ACTIONS(1226),
    [anon_sym_default] = ACTIONS(1226),
    [anon_sym_while] = ACTIONS(1226),
    [anon_sym_do] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(1226),
    [anon_sym_return] = ACTIONS(1226),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_goto] = ACTIONS(1226),
    [anon_sym___try] = ACTIONS(1226),
    [anon_sym___leave] = ACTIONS(1226),
    [anon_sym_DASH_DASH] = ACTIONS(1228),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [anon_sym_sizeof] = ACTIONS(1226),
    [anon_sym___alignof__] = ACTIONS(1226),
    [anon_sym___alignof] = ACTIONS(1226),
    [anon_sym__alignof] = ACTIONS(1226),
    [anon_sym_alignof] = ACTIONS(1226),
    [anon_sym__Alignof] = ACTIONS(1226),
    [anon_sym_offsetof] = ACTIONS(1226),
    [anon_sym__Generic] = ACTIONS(1226),
    [anon_sym_asm] = ACTIONS(1226),
    [anon_sym___asm__] = ACTIONS(1226),
    [anon_sym___asm] = ACTIONS(1226),
    [sym_number_literal] = ACTIONS(1228),
    [anon_sym_L_SQUOTE] = ACTIONS(1228),
    [anon_sym_u_SQUOTE] = ACTIONS(1228),
    [anon_sym_U_SQUOTE] = ACTIONS(1228),
    [anon_sym_u8_SQUOTE] = ACTIONS(1228),
    [anon_sym_SQUOTE] = ACTIONS(1228),
    [anon_sym_L_DQUOTE] = ACTIONS(1228),
    [anon_sym_u_DQUOTE] = ACTIONS(1228),
    [anon_sym_U_DQUOTE] = ACTIONS(1228),
    [anon_sym_u8_DQUOTE] = ACTIONS(1228),
    [anon_sym_DQUOTE] = ACTIONS(1228),
    [sym_true] = ACTIONS(1226),
    [sym_false] = ACTIONS(1226),
    [anon_sym_NULL] = ACTIONS(1226),
    [anon_sym_nullptr] = ACTIONS(1226),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(162)] = {
    [ts_builtin_sym_end] = ACTIONS(1240),
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(163)] = {
    [sym_identifier] = ACTIONS(1230),
    [aux_sym_preproc_include_token1] = ACTIONS(1230),
    [aux_sym_preproc_def_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1230),
    [sym_preproc_directive] = ACTIONS(1230),
    [anon_sym_LPAREN2] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1230),
    [anon_sym_PLUS] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym___extension__] = ACTIONS(1230),
    [anon_sym_typedef] = ACTIONS(1230),
    [anon_sym_extern] = ACTIONS(1230),
    [anon_sym___attribute__] = ACTIONS(1230),
    [anon_sym___attribute] = ACTIONS(1230),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1232),
    [anon_sym___declspec] = ACTIONS(1230),
    [anon_sym___cdecl] = ACTIONS(1230),
    [anon_sym___clrcall] = ACTIONS(1230),
    [anon_sym___stdcall] = ACTIONS(1230),
    [anon_sym___fastcall] = ACTIONS(1230),
    [anon_sym___thiscall] = ACTIONS(1230),
    [anon_sym___vectorcall] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_RBRACE] = ACTIONS(1232),
    [anon_sym_signed] = ACTIONS(1230),
    [anon_sym_unsigned] = ACTIONS(1230),
    [anon_sym_long] = ACTIONS(1230),
    [anon_sym_short] = ACTIONS(1230),
    [anon_sym_static] = ACTIONS(1230),
    [anon_sym_auto] = ACTIONS(1230),
    [anon_sym_register] = ACTIONS(1230),
    [anon_sym_inline] = ACTIONS(1230),
    [anon_sym___inline] = ACTIONS(1230),
    [anon_sym___inline__] = ACTIONS(1230),
    [anon_sym___forceinline] = ACTIONS(1230),
    [anon_sym_thread_local] = ACTIONS(1230),
    [anon_sym___thread] = ACTIONS(1230),
    [anon_sym_const] = ACTIONS(1230),
    [anon_sym_constexpr] = ACTIONS(1230),
    [anon_sym_volatile] = ACTIONS(1230),
    [anon_sym_restrict] = ACTIONS(1230),
    [anon_sym___restrict__] = ACTIONS(1230),
    [anon_sym__Atomic] = ACTIONS(1230),
    [anon_sym__Noreturn] = ACTIONS(1230),
    [anon_sym_noreturn] = ACTIONS(1230),
    [anon_sym__Nonnull] = ACTIONS(1230),
    [anon_sym_alignas] = ACTIONS(1230),
    [anon_sym__Alignas] = ACTIONS(1230),
    [sym_primitive_type] = ACTIONS(1230),
    [anon_sym_enum] = ACTIONS(1230),
    [anon_sym_struct] = ACTIONS(1230),
    [anon_sym_union] = ACTIONS(1230),
    [anon_sym_if] = ACTIONS(1230),
    [anon_sym_else] = ACTIONS(1230),
    [anon_sym_switch] = ACTIONS(1230),
    [anon_sym_case] = ACTIONS(1230),
    [anon_sym_default] = ACTIONS(1230),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_break] = ACTIONS(1230),
    [anon_sym_continue] = ACTIONS(1230),
    [anon_sym_goto] = ACTIONS(1230),
    [anon_sym___try] = ACTIONS(1230),
    [anon_sym___leave] = ACTIONS(1230),
    [anon_sym_DASH_DASH] = ACTIONS(1232),
    [anon_sym_PLUS_PLUS] = ACTIONS(1232),
    [anon_sym_sizeof] = ACTIONS(1230),
    [anon_sym___alignof__] = ACTIONS(1230),
    [anon_sym___alignof] = ACTIONS(1230),
    [anon_sym__alignof] = ACTIONS(1230),
    [anon_sym_alignof] = ACTIONS(1230),
    [anon_sym__Alignof] = ACTIONS(1230),
    [anon_sym_offsetof] = ACTIONS(1230),
    [anon_sym__Generic] = ACTIONS(1230),
    [anon_sym_asm] = ACTIONS(1230),
    [anon_sym___asm__] = ACTIONS(1230),
    [anon_sym___asm] = ACTIONS(1230),
    [sym_number_literal] = ACTIONS(1232),
    [anon_sym_L_SQUOTE] = ACTIONS(1232),
    [anon_sym_u_SQUOTE] = ACTIONS(1232),
    [anon_sym_U_SQUOTE] = ACTIONS(1232),
    [anon_sym_u8_SQUOTE] = ACTIONS(1232),
    [anon_sym_SQUOTE] = ACTIONS(1232),
    [anon_sym_L_DQUOTE] = ACTIONS(1232),
    [anon_sym_u_DQUOTE] = ACTIONS(1232),
    [anon_sym_U_DQUOTE] = ACTIONS(1232),
    [anon_sym_u8_DQUOTE] = ACTIONS(1232),
    [anon_sym_DQUOTE] = ACTIONS(1232),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [anon_sym_NULL] = ACTIONS(1230),
    [anon_sym_nullptr] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(164)] = {
    [sym_identifier] = ACTIONS(1214),
    [aux_sym_preproc_include_token1] = ACTIONS(1214),
    [aux_sym_preproc_def_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1214),
    [sym_preproc_directive] = ACTIONS(1214),
    [anon_sym_LPAREN2] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_TILDE] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1214),
    [anon_sym_PLUS] = ACTIONS(1214),
    [anon_sym_STAR] = ACTIONS(1216),
    [anon_sym_AMP] = ACTIONS(1216),
    [anon_sym_SEMI] = ACTIONS(1216),
    [anon_sym___extension__] = ACTIONS(1214),
    [anon_sym_typedef] = ACTIONS(1214),
    [anon_sym_extern] = ACTIONS(1214),
    [anon_sym___attribute__] = ACTIONS(1214),
    [anon_sym___attribute] = ACTIONS(1214),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1216),
    [anon_sym___declspec] = ACTIONS(1214),
    [anon_sym___cdecl] = ACTIONS(1214),
    [anon_sym___clrcall] = ACTIONS(1214),
    [anon_sym___stdcall] = ACTIONS(1214),
    [anon_sym___fastcall] = ACTIONS(1214),
    [anon_sym___thiscall] = ACTIONS(1214),
    [anon_sym___vectorcall] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_RBRACE] = ACTIONS(1216),
    [anon_sym_signed] = ACTIONS(1214),
    [anon_sym_unsigned] = ACTIONS(1214),
    [anon_sym_long] = ACTIONS(1214),
    [anon_sym_short] = ACTIONS(1214),
    [anon_sym_static] = ACTIONS(1214),
    [anon_sym_auto] = ACTIONS(1214),
    [anon_sym_register] = ACTIONS(1214),
    [anon_sym_inline] = ACTIONS(1214),
    [anon_sym___inline] = ACTIONS(1214),
    [anon_sym___inline__] = ACTIONS(1214),
    [anon_sym___forceinline] = ACTIONS(1214),
    [anon_sym_thread_local] = ACTIONS(1214),
    [anon_sym___thread] = ACTIONS(1214),
    [anon_sym_const] = ACTIONS(1214),
    [anon_sym_constexpr] = ACTIONS(1214),
    [anon_sym_volatile] = ACTIONS(1214),
    [anon_sym_restrict] = ACTIONS(1214),
    [anon_sym___restrict__] = ACTIONS(1214),
    [anon_sym__Atomic] = ACTIONS(1214),
    [anon_sym__Noreturn] = ACTIONS(1214),
    [anon_sym_noreturn] = ACTIONS(1214),
    [anon_sym__Nonnull] = ACTIONS(1214),
    [anon_sym_alignas] = ACTIONS(1214),
    [anon_sym__Alignas] = ACTIONS(1214),
    [sym_primitive_type] = ACTIONS(1214),
    [anon_sym_enum] = ACTIONS(1214),
    [anon_sym_struct] = ACTIONS(1214),
    [anon_sym_union] = ACTIONS(1214),
    [anon_sym_if] = ACTIONS(1214),
    [anon_sym_else] = ACTIONS(1214),
    [anon_sym_switch] = ACTIONS(1214),
    [anon_sym_case] = ACTIONS(1214),
    [anon_sym_default] = ACTIONS(1214),
    [anon_sym_while] = ACTIONS(1214),
    [anon_sym_do] = ACTIONS(1214),
    [anon_sym_for] = ACTIONS(1214),
    [anon_sym_return] = ACTIONS(1214),
    [anon_sym_break] = ACTIONS(1214),
    [anon_sym_continue] = ACTIONS(1214),
    [anon_sym_goto] = ACTIONS(1214),
    [anon_sym___try] = ACTIONS(1214),
    [anon_sym___leave] = ACTIONS(1214),
    [anon_sym_DASH_DASH] = ACTIONS(1216),
    [anon_sym_PLUS_PLUS] = ACTIONS(1216),
    [anon_sym_sizeof] = ACTIONS(1214),
    [anon_sym___alignof__] = ACTIONS(1214),
    [anon_sym___alignof] = ACTIONS(1214),
    [anon_sym__alignof] = ACTIONS(1214),
    [anon_sym_alignof] = ACTIONS(1214),
    [anon_sym__Alignof] = ACTIONS(1214),
    [anon_sym_offsetof] = ACTIONS(1214),
    [anon_sym__Generic] = ACTIONS(1214),
    [anon_sym_asm] = ACTIONS(1214),
    [anon_sym___asm__] = ACTIONS(1214),
    [anon_sym___asm] = ACTIONS(1214),
    [sym_number_literal] = ACTIONS(1216),
    [anon_sym_L_SQUOTE] = ACTIONS(1216),
    [anon_sym_u_SQUOTE] = ACTIONS(1216),
    [anon_sym_U_SQUOTE] = ACTIONS(1216),
    [anon_sym_u8_SQUOTE] = ACTIONS(1216),
    [anon_sym_SQUOTE] = ACTIONS(1216),
    [anon_sym_L_DQUOTE] = ACTIONS(1216),
    [anon_sym_u_DQUOTE] = ACTIONS(1216),
    [anon_sym_U_DQUOTE] = ACTIONS(1216),
    [anon_sym_u8_DQUOTE] = ACTIONS(1216),
    [anon_sym_DQUOTE] = ACTIONS(1216),
    [sym_true] = ACTIONS(1214),
    [sym_false] = ACTIONS(1214),
    [anon_sym_NULL] = ACTIONS(1214),
    [anon_sym_nullptr] = ACTIONS(1214),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(165)] = {
    [sym_identifier] = ACTIONS(1226),
    [aux_sym_preproc_include_token1] = ACTIONS(1226),
    [aux_sym_preproc_def_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1226),
    [sym_preproc_directive] = ACTIONS(1226),
    [anon_sym_LPAREN2] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1226),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym___extension__] = ACTIONS(1226),
    [anon_sym_typedef] = ACTIONS(1226),
    [anon_sym_extern] = ACTIONS(1226),
    [anon_sym___attribute__] = ACTIONS(1226),
    [anon_sym___attribute] = ACTIONS(1226),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1228),
    [anon_sym___declspec] = ACTIONS(1226),
    [anon_sym___cdecl] = ACTIONS(1226),
    [anon_sym___clrcall] = ACTIONS(1226),
    [anon_sym___stdcall] = ACTIONS(1226),
    [anon_sym___fastcall] = ACTIONS(1226),
    [anon_sym___thiscall] = ACTIONS(1226),
    [anon_sym___vectorcall] = ACTIONS(1226),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_RBRACE] = ACTIONS(1228),
    [anon_sym_signed] = ACTIONS(1226),
    [anon_sym_unsigned] = ACTIONS(1226),
    [anon_sym_long] = ACTIONS(1226),
    [anon_sym_short] = ACTIONS(1226),
    [anon_sym_static] = ACTIONS(1226),
    [anon_sym_auto] = ACTIONS(1226),
    [anon_sym_register] = ACTIONS(1226),
    [anon_sym_inline] = ACTIONS(1226),
    [anon_sym___inline] = ACTIONS(1226),
    [anon_sym___inline__] = ACTIONS(1226),
    [anon_sym___forceinline] = ACTIONS(1226),
    [anon_sym_thread_local] = ACTIONS(1226),
    [anon_sym___thread] = ACTIONS(1226),
    [anon_sym_const] = ACTIONS(1226),
    [anon_sym_constexpr] = ACTIONS(1226),
    [anon_sym_volatile] = ACTIONS(1226),
    [anon_sym_restrict] = ACTIONS(1226),
    [anon_sym___restrict__] = ACTIONS(1226),
    [anon_sym__Atomic] = ACTIONS(1226),
    [anon_sym__Noreturn] = ACTIONS(1226),
    [anon_sym_noreturn] = ACTIONS(1226),
    [anon_sym__Nonnull] = ACTIONS(1226),
    [anon_sym_alignas] = ACTIONS(1226),
    [anon_sym__Alignas] = ACTIONS(1226),
    [sym_primitive_type] = ACTIONS(1226),
    [anon_sym_enum] = ACTIONS(1226),
    [anon_sym_struct] = ACTIONS(1226),
    [anon_sym_union] = ACTIONS(1226),
    [anon_sym_if] = ACTIONS(1226),
    [anon_sym_else] = ACTIONS(1226),
    [anon_sym_switch] = ACTIONS(1226),
    [anon_sym_case] = ACTIONS(1226),
    [anon_sym_default] = ACTIONS(1226),
    [anon_sym_while] = ACTIONS(1226),
    [anon_sym_do] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(1226),
    [anon_sym_return] = ACTIONS(1226),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_goto] = ACTIONS(1226),
    [anon_sym___try] = ACTIONS(1226),
    [anon_sym___leave] = ACTIONS(1226),
    [anon_sym_DASH_DASH] = ACTIONS(1228),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [anon_sym_sizeof] = ACTIONS(1226),
    [anon_sym___alignof__] = ACTIONS(1226),
    [anon_sym___alignof] = ACTIONS(1226),
    [anon_sym__alignof] = ACTIONS(1226),
    [anon_sym_alignof] = ACTIONS(1226),
    [anon_sym__Alignof] = ACTIONS(1226),
    [anon_sym_offsetof] = ACTIONS(1226),
    [anon_sym__Generic] = ACTIONS(1226),
    [anon_sym_asm] = ACTIONS(1226),
    [anon_sym___asm__] = ACTIONS(1226),
    [anon_sym___asm] = ACTIONS(1226),
    [sym_number_literal] = ACTIONS(1228),
    [anon_sym_L_SQUOTE] = ACTIONS(1228),
    [anon_sym_u_SQUOTE] = ACTIONS(1228),
    [anon_sym_U_SQUOTE] = ACTIONS(1228),
    [anon_sym_u8_SQUOTE] = ACTIONS(1228),
    [anon_sym_SQUOTE] = ACTIONS(1228),
    [anon_sym_L_DQUOTE] = ACTIONS(1228),
    [anon_sym_u_DQUOTE] = ACTIONS(1228),
    [anon_sym_U_DQUOTE] = ACTIONS(1228),
    [anon_sym_u8_DQUOTE] = ACTIONS(1228),
    [anon_sym_DQUOTE] = ACTIONS(1228),
    [sym_true] = ACTIONS(1226),
    [sym_false] = ACTIONS(1226),
    [anon_sym_NULL] = ACTIONS(1226),
    [anon_sym_nullptr] = ACTIONS(1226),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(166)] = {
    [ts_builtin_sym_end] = ACTIONS(1164),
    [sym_identifier] = ACTIONS(1162),
    [aux_sym_preproc_include_token1] = ACTIONS(1162),
    [aux_sym_preproc_def_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1162),
    [sym_preproc_directive] = ACTIONS(1162),
    [anon_sym_LPAREN2] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1162),
    [anon_sym_PLUS] = ACTIONS(1162),
    [anon_sym_STAR] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym___extension__] = ACTIONS(1162),
    [anon_sym_typedef] = ACTIONS(1162),
    [anon_sym_extern] = ACTIONS(1162),
    [anon_sym___attribute__] = ACTIONS(1162),
    [anon_sym___attribute] = ACTIONS(1162),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1164),
    [anon_sym___declspec] = ACTIONS(1162),
    [anon_sym___cdecl] = ACTIONS(1162),
    [anon_sym___clrcall] = ACTIONS(1162),
    [anon_sym___stdcall] = ACTIONS(1162),
    [anon_sym___fastcall] = ACTIONS(1162),
    [anon_sym___thiscall] = ACTIONS(1162),
    [anon_sym___vectorcall] = ACTIONS(1162),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_signed] = ACTIONS(1162),
    [anon_sym_unsigned] = ACTIONS(1162),
    [anon_sym_long] = ACTIONS(1162),
    [anon_sym_short] = ACTIONS(1162),
    [anon_sym_static] = ACTIONS(1162),
    [anon_sym_auto] = ACTIONS(1162),
    [anon_sym_register] = ACTIONS(1162),
    [anon_sym_inline] = ACTIONS(1162),
    [anon_sym___inline] = ACTIONS(1162),
    [anon_sym___inline__] = ACTIONS(1162),
    [anon_sym___forceinline] = ACTIONS(1162),
    [anon_sym_thread_local] = ACTIONS(1162),
    [anon_sym___thread] = ACTIONS(1162),
    [anon_sym_const] = ACTIONS(1162),
    [anon_sym_constexpr] = ACTIONS(1162),
    [anon_sym_volatile] = ACTIONS(1162),
    [anon_sym_restrict] = ACTIONS(1162),
    [anon_sym___restrict__] = ACTIONS(1162),
    [anon_sym__Atomic] = ACTIONS(1162),
    [anon_sym__Noreturn] = ACTIONS(1162),
    [anon_sym_noreturn] = ACTIONS(1162),
    [anon_sym__Nonnull] = ACTIONS(1162),
    [anon_sym_alignas] = ACTIONS(1162),
    [anon_sym__Alignas] = ACTIONS(1162),
    [sym_primitive_type] = ACTIONS(1162),
    [anon_sym_enum] = ACTIONS(1162),
    [anon_sym_struct] = ACTIONS(1162),
    [anon_sym_union] = ACTIONS(1162),
    [anon_sym_if] = ACTIONS(1162),
    [anon_sym_else] = ACTIONS(1162),
    [anon_sym_switch] = ACTIONS(1162),
    [anon_sym_case] = ACTIONS(1162),
    [anon_sym_default] = ACTIONS(1162),
    [anon_sym_while] = ACTIONS(1162),
    [anon_sym_do] = ACTIONS(1162),
    [anon_sym_for] = ACTIONS(1162),
    [anon_sym_return] = ACTIONS(1162),
    [anon_sym_break] = ACTIONS(1162),
    [anon_sym_continue] = ACTIONS(1162),
    [anon_sym_goto] = ACTIONS(1162),
    [anon_sym___try] = ACTIONS(1162),
    [anon_sym___leave] = ACTIONS(1162),
    [anon_sym_DASH_DASH] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1164),
    [anon_sym_sizeof] = ACTIONS(1162),
    [anon_sym___alignof__] = ACTIONS(1162),
    [anon_sym___alignof] = ACTIONS(1162),
    [anon_sym__alignof] = ACTIONS(1162),
    [anon_sym_alignof] = ACTIONS(1162),
    [anon_sym__Alignof] = ACTIONS(1162),
    [anon_sym_offsetof] = ACTIONS(1162),
    [anon_sym__Generic] = ACTIONS(1162),
    [anon_sym_asm] = ACTIONS(1162),
    [anon_sym___asm__] = ACTIONS(1162),
    [anon_sym___asm] = ACTIONS(1162),
    [sym_number_literal] = ACTIONS(1164),
    [anon_sym_L_SQUOTE] = ACTIONS(1164),
    [anon_sym_u_SQUOTE] = ACTIONS(1164),
    [anon_sym_U_SQUOTE] = ACTIONS(1164),
    [anon_sym_u8_SQUOTE] = ACTIONS(1164),
    [anon_sym_SQUOTE] = ACTIONS(1164),
    [anon_sym_L_DQUOTE] = ACTIONS(1164),
    [anon_sym_u_DQUOTE] = ACTIONS(1164),
    [anon_sym_U_DQUOTE] = ACTIONS(1164),
    [anon_sym_u8_DQUOTE] = ACTIONS(1164),
    [anon_sym_DQUOTE] = ACTIONS(1164),
    [sym_true] = ACTIONS(1162),
    [sym_false] = ACTIONS(1162),
    [anon_sym_NULL] = ACTIONS(1162),
    [anon_sym_nullptr] = ACTIONS(1162),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(167)] = {
    [sym_identifier] = ACTIONS(1242),
    [aux_sym_preproc_include_token1] = ACTIONS(1242),
    [aux_sym_preproc_def_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1242),
    [sym_preproc_directive] = ACTIONS(1242),
    [anon_sym_LPAREN2] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_TILDE] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1242),
    [anon_sym_PLUS] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(1244),
    [anon_sym_AMP] = ACTIONS(1244),
    [anon_sym_SEMI] = ACTIONS(1244),
    [anon_sym___extension__] = ACTIONS(1242),
    [anon_sym_typedef] = ACTIONS(1242),
    [anon_sym_extern] = ACTIONS(1242),
    [anon_sym___attribute__] = ACTIONS(1242),
    [anon_sym___attribute] = ACTIONS(1242),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1244),
    [anon_sym___declspec] = ACTIONS(1242),
    [anon_sym___cdecl] = ACTIONS(1242),
    [anon_sym___clrcall] = ACTIONS(1242),
    [anon_sym___stdcall] = ACTIONS(1242),
    [anon_sym___fastcall] = ACTIONS(1242),
    [anon_sym___thiscall] = ACTIONS(1242),
    [anon_sym___vectorcall] = ACTIONS(1242),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_RBRACE] = ACTIONS(1244),
    [anon_sym_signed] = ACTIONS(1242),
    [anon_sym_unsigned] = ACTIONS(1242),
    [anon_sym_long] = ACTIONS(1242),
    [anon_sym_short] = ACTIONS(1242),
    [anon_sym_static] = ACTIONS(1242),
    [anon_sym_auto] = ACTIONS(1242),
    [anon_sym_register] = ACTIONS(1242),
    [anon_sym_inline] = ACTIONS(1242),
    [anon_sym___inline] = ACTIONS(1242),
    [anon_sym___inline__] = ACTIONS(1242),
    [anon_sym___forceinline] = ACTIONS(1242),
    [anon_sym_thread_local] = ACTIONS(1242),
    [anon_sym___thread] = ACTIONS(1242),
    [anon_sym_const] = ACTIONS(1242),
    [anon_sym_constexpr] = ACTIONS(1242),
    [anon_sym_volatile] = ACTIONS(1242),
    [anon_sym_restrict] = ACTIONS(1242),
    [anon_sym___restrict__] = ACTIONS(1242),
    [anon_sym__Atomic] = ACTIONS(1242),
    [anon_sym__Noreturn] = ACTIONS(1242),
    [anon_sym_noreturn] = ACTIONS(1242),
    [anon_sym__Nonnull] = ACTIONS(1242),
    [anon_sym_alignas] = ACTIONS(1242),
    [anon_sym__Alignas] = ACTIONS(1242),
    [sym_primitive_type] = ACTIONS(1242),
    [anon_sym_enum] = ACTIONS(1242),
    [anon_sym_struct] = ACTIONS(1242),
    [anon_sym_union] = ACTIONS(1242),
    [anon_sym_if] = ACTIONS(1242),
    [anon_sym_else] = ACTIONS(1242),
    [anon_sym_switch] = ACTIONS(1242),
    [anon_sym_case] = ACTIONS(1242),
    [anon_sym_default] = ACTIONS(1242),
    [anon_sym_while] = ACTIONS(1242),
    [anon_sym_do] = ACTIONS(1242),
    [anon_sym_for] = ACTIONS(1242),
    [anon_sym_return] = ACTIONS(1242),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_goto] = ACTIONS(1242),
    [anon_sym___try] = ACTIONS(1242),
    [anon_sym___leave] = ACTIONS(1242),
    [anon_sym_DASH_DASH] = ACTIONS(1244),
    [anon_sym_PLUS_PLUS] = ACTIONS(1244),
    [anon_sym_sizeof] = ACTIONS(1242),
    [anon_sym___alignof__] = ACTIONS(1242),
    [anon_sym___alignof] = ACTIONS(1242),
    [anon_sym__alignof] = ACTIONS(1242),
    [anon_sym_alignof] = ACTIONS(1242),
    [anon_sym__Alignof] = ACTIONS(1242),
    [anon_sym_offsetof] = ACTIONS(1242),
    [anon_sym__Generic] = ACTIONS(1242),
    [anon_sym_asm] = ACTIONS(1242),
    [anon_sym___asm__] = ACTIONS(1242),
    [anon_sym___asm] = ACTIONS(1242),
    [sym_number_literal] = ACTIONS(1244),
    [anon_sym_L_SQUOTE] = ACTIONS(1244),
    [anon_sym_u_SQUOTE] = ACTIONS(1244),
    [anon_sym_U_SQUOTE] = ACTIONS(1244),
    [anon_sym_u8_SQUOTE] = ACTIONS(1244),
    [anon_sym_SQUOTE] = ACTIONS(1244),
    [anon_sym_L_DQUOTE] = ACTIONS(1244),
    [anon_sym_u_DQUOTE] = ACTIONS(1244),
    [anon_sym_U_DQUOTE] = ACTIONS(1244),
    [anon_sym_u8_DQUOTE] = ACTIONS(1244),
    [anon_sym_DQUOTE] = ACTIONS(1244),
    [sym_true] = ACTIONS(1242),
    [sym_false] = ACTIONS(1242),
    [anon_sym_NULL] = ACTIONS(1242),
    [anon_sym_nullptr] = ACTIONS(1242),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(168)] = {
    [ts_builtin_sym_end] = ACTIONS(1248),
    [sym_identifier] = ACTIONS(1246),
    [aux_sym_preproc_include_token1] = ACTIONS(1246),
    [aux_sym_preproc_def_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1246),
    [sym_preproc_directive] = ACTIONS(1246),
    [anon_sym_LPAREN2] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_TILDE] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1246),
    [anon_sym_PLUS] = ACTIONS(1246),
    [anon_sym_STAR] = ACTIONS(1248),
    [anon_sym_AMP] = ACTIONS(1248),
    [anon_sym_SEMI] = ACTIONS(1248),
    [anon_sym___extension__] = ACTIONS(1246),
    [anon_sym_typedef] = ACTIONS(1246),
    [anon_sym_extern] = ACTIONS(1246),
    [anon_sym___attribute__] = ACTIONS(1246),
    [anon_sym___attribute] = ACTIONS(1246),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1248),
    [anon_sym___declspec] = ACTIONS(1246),
    [anon_sym___cdecl] = ACTIONS(1246),
    [anon_sym___clrcall] = ACTIONS(1246),
    [anon_sym___stdcall] = ACTIONS(1246),
    [anon_sym___fastcall] = ACTIONS(1246),
    [anon_sym___thiscall] = ACTIONS(1246),
    [anon_sym___vectorcall] = ACTIONS(1246),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_signed] = ACTIONS(1246),
    [anon_sym_unsigned] = ACTIONS(1246),
    [anon_sym_long] = ACTIONS(1246),
    [anon_sym_short] = ACTIONS(1246),
    [anon_sym_static] = ACTIONS(1246),
    [anon_sym_auto] = ACTIONS(1246),
    [anon_sym_register] = ACTIONS(1246),
    [anon_sym_inline] = ACTIONS(1246),
    [anon_sym___inline] = ACTIONS(1246),
    [anon_sym___inline__] = ACTIONS(1246),
    [anon_sym___forceinline] = ACTIONS(1246),
    [anon_sym_thread_local] = ACTIONS(1246),
    [anon_sym___thread] = ACTIONS(1246),
    [anon_sym_const] = ACTIONS(1246),
    [anon_sym_constexpr] = ACTIONS(1246),
    [anon_sym_volatile] = ACTIONS(1246),
    [anon_sym_restrict] = ACTIONS(1246),
    [anon_sym___restrict__] = ACTIONS(1246),
    [anon_sym__Atomic] = ACTIONS(1246),
    [anon_sym__Noreturn] = ACTIONS(1246),
    [anon_sym_noreturn] = ACTIONS(1246),
    [anon_sym__Nonnull] = ACTIONS(1246),
    [anon_sym_alignas] = ACTIONS(1246),
    [anon_sym__Alignas] = ACTIONS(1246),
    [sym_primitive_type] = ACTIONS(1246),
    [anon_sym_enum] = ACTIONS(1246),
    [anon_sym_struct] = ACTIONS(1246),
    [anon_sym_union] = ACTIONS(1246),
    [anon_sym_if] = ACTIONS(1246),
    [anon_sym_else] = ACTIONS(1246),
    [anon_sym_switch] = ACTIONS(1246),
    [anon_sym_case] = ACTIONS(1246),
    [anon_sym_default] = ACTIONS(1246),
    [anon_sym_while] = ACTIONS(1246),
    [anon_sym_do] = ACTIONS(1246),
    [anon_sym_for] = ACTIONS(1246),
    [anon_sym_return] = ACTIONS(1246),
    [anon_sym_break] = ACTIONS(1246),
    [anon_sym_continue] = ACTIONS(1246),
    [anon_sym_goto] = ACTIONS(1246),
    [anon_sym___try] = ACTIONS(1246),
    [anon_sym___leave] = ACTIONS(1246),
    [anon_sym_DASH_DASH] = ACTIONS(1248),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [anon_sym_sizeof] = ACTIONS(1246),
    [anon_sym___alignof__] = ACTIONS(1246),
    [anon_sym___alignof] = ACTIONS(1246),
    [anon_sym__alignof] = ACTIONS(1246),
    [anon_sym_alignof] = ACTIONS(1246),
    [anon_sym__Alignof] = ACTIONS(1246),
    [anon_sym_offsetof] = ACTIONS(1246),
    [anon_sym__Generic] = ACTIONS(1246),
    [anon_sym_asm] = ACTIONS(1246),
    [anon_sym___asm__] = ACTIONS(1246),
    [anon_sym___asm] = ACTIONS(1246),
    [sym_number_literal] = ACTIONS(1248),
    [anon_sym_L_SQUOTE] = ACTIONS(1248),
    [anon_sym_u_SQUOTE] = ACTIONS(1248),
    [anon_sym_U_SQUOTE] = ACTIONS(1248),
    [anon_sym_u8_SQUOTE] = ACTIONS(1248),
    [anon_sym_SQUOTE] = ACTIONS(1248),
    [anon_sym_L_DQUOTE] = ACTIONS(1248),
    [anon_sym_u_DQUOTE] = ACTIONS(1248),
    [anon_sym_U_DQUOTE] = ACTIONS(1248),
    [anon_sym_u8_DQUOTE] = ACTIONS(1248),
    [anon_sym_DQUOTE] = ACTIONS(1248),
    [sym_true] = ACTIONS(1246),
    [sym_false] = ACTIONS(1246),
    [anon_sym_NULL] = ACTIONS(1246),
    [anon_sym_nullptr] = ACTIONS(1246),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(169)] = {
    [ts_builtin_sym_end] = ACTIONS(1256),
    [sym_identifier] = ACTIONS(1254),
    [aux_sym_preproc_include_token1] = ACTIONS(1254),
    [aux_sym_preproc_def_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1254),
    [sym_preproc_directive] = ACTIONS(1254),
    [anon_sym_LPAREN2] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_TILDE] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_STAR] = ACTIONS(1256),
    [anon_sym_AMP] = ACTIONS(1256),
    [anon_sym_SEMI] = ACTIONS(1256),
    [anon_sym___extension__] = ACTIONS(1254),
    [anon_sym_typedef] = ACTIONS(1254),
    [anon_sym_extern] = ACTIONS(1254),
    [anon_sym___attribute__] = ACTIONS(1254),
    [anon_sym___attribute] = ACTIONS(1254),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1256),
    [anon_sym___declspec] = ACTIONS(1254),
    [anon_sym___cdecl] = ACTIONS(1254),
    [anon_sym___clrcall] = ACTIONS(1254),
    [anon_sym___stdcall] = ACTIONS(1254),
    [anon_sym___fastcall] = ACTIONS(1254),
    [anon_sym___thiscall] = ACTIONS(1254),
    [anon_sym___vectorcall] = ACTIONS(1254),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_signed] = ACTIONS(1254),
    [anon_sym_unsigned] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_auto] = ACTIONS(1254),
    [anon_sym_register] = ACTIONS(1254),
    [anon_sym_inline] = ACTIONS(1254),
    [anon_sym___inline] = ACTIONS(1254),
    [anon_sym___inline__] = ACTIONS(1254),
    [anon_sym___forceinline] = ACTIONS(1254),
    [anon_sym_thread_local] = ACTIONS(1254),
    [anon_sym___thread] = ACTIONS(1254),
    [anon_sym_const] = ACTIONS(1254),
    [anon_sym_constexpr] = ACTIONS(1254),
    [anon_sym_volatile] = ACTIONS(1254),
    [anon_sym_restrict] = ACTIONS(1254),
    [anon_sym___restrict__] = ACTIONS(1254),
    [anon_sym__Atomic] = ACTIONS(1254),
    [anon_sym__Noreturn] = ACTIONS(1254),
    [anon_sym_noreturn] = ACTIONS(1254),
    [anon_sym__Nonnull] = ACTIONS(1254),
    [anon_sym_alignas] = ACTIONS(1254),
    [anon_sym__Alignas] = ACTIONS(1254),
    [sym_primitive_type] = ACTIONS(1254),
    [anon_sym_enum] = ACTIONS(1254),
    [anon_sym_struct] = ACTIONS(1254),
    [anon_sym_union] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_goto] = ACTIONS(1254),
    [anon_sym___try] = ACTIONS(1254),
    [anon_sym___leave] = ACTIONS(1254),
    [anon_sym_DASH_DASH] = ACTIONS(1256),
    [anon_sym_PLUS_PLUS] = ACTIONS(1256),
    [anon_sym_sizeof] = ACTIONS(1254),
    [anon_sym___alignof__] = ACTIONS(1254),
    [anon_sym___alignof] = ACTIONS(1254),
    [anon_sym__alignof] = ACTIONS(1254),
    [anon_sym_alignof] = ACTIONS(1254),
    [anon_sym__Alignof] = ACTIONS(1254),
    [anon_sym_offsetof] = ACTIONS(1254),
    [anon_sym__Generic] = ACTIONS(1254),
    [anon_sym_asm] = ACTIONS(1254),
    [anon_sym___asm__] = ACTIONS(1254),
    [anon_sym___asm] = ACTIONS(1254),
    [sym_number_literal] = ACTIONS(1256),
    [anon_sym_L_SQUOTE] = ACTIONS(1256),
    [anon_sym_u_SQUOTE] = ACTIONS(1256),
    [anon_sym_U_SQUOTE] = ACTIONS(1256),
    [anon_sym_u8_SQUOTE] = ACTIONS(1256),
    [anon_sym_SQUOTE] = ACTIONS(1256),
    [anon_sym_L_DQUOTE] = ACTIONS(1256),
    [anon_sym_u_DQUOTE] = ACTIONS(1256),
    [anon_sym_U_DQUOTE] = ACTIONS(1256),
    [anon_sym_u8_DQUOTE] = ACTIONS(1256),
    [anon_sym_DQUOTE] = ACTIONS(1256),
    [sym_true] = ACTIONS(1254),
    [sym_false] = ACTIONS(1254),
    [anon_sym_NULL] = ACTIONS(1254),
    [anon_sym_nullptr] = ACTIONS(1254),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(170)] = {
    [ts_builtin_sym_end] = ACTIONS(1240),
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(171)] = {
    [ts_builtin_sym_end] = ACTIONS(1232),
    [sym_identifier] = ACTIONS(1230),
    [aux_sym_preproc_include_token1] = ACTIONS(1230),
    [aux_sym_preproc_def_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1230),
    [sym_preproc_directive] = ACTIONS(1230),
    [anon_sym_LPAREN2] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1230),
    [anon_sym_PLUS] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym___extension__] = ACTIONS(1230),
    [anon_sym_typedef] = ACTIONS(1230),
    [anon_sym_extern] = ACTIONS(1230),
    [anon_sym___attribute__] = ACTIONS(1230),
    [anon_sym___attribute] = ACTIONS(1230),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1232),
    [anon_sym___declspec] = ACTIONS(1230),
    [anon_sym___cdecl] = ACTIONS(1230),
    [anon_sym___clrcall] = ACTIONS(1230),
    [anon_sym___stdcall] = ACTIONS(1230),
    [anon_sym___fastcall] = ACTIONS(1230),
    [anon_sym___thiscall] = ACTIONS(1230),
    [anon_sym___vectorcall] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_signed] = ACTIONS(1230),
    [anon_sym_unsigned] = ACTIONS(1230),
    [anon_sym_long] = ACTIONS(1230),
    [anon_sym_short] = ACTIONS(1230),
    [anon_sym_static] = ACTIONS(1230),
    [anon_sym_auto] = ACTIONS(1230),
    [anon_sym_register] = ACTIONS(1230),
    [anon_sym_inline] = ACTIONS(1230),
    [anon_sym___inline] = ACTIONS(1230),
    [anon_sym___inline__] = ACTIONS(1230),
    [anon_sym___forceinline] = ACTIONS(1230),
    [anon_sym_thread_local] = ACTIONS(1230),
    [anon_sym___thread] = ACTIONS(1230),
    [anon_sym_const] = ACTIONS(1230),
    [anon_sym_constexpr] = ACTIONS(1230),
    [anon_sym_volatile] = ACTIONS(1230),
    [anon_sym_restrict] = ACTIONS(1230),
    [anon_sym___restrict__] = ACTIONS(1230),
    [anon_sym__Atomic] = ACTIONS(1230),
    [anon_sym__Noreturn] = ACTIONS(1230),
    [anon_sym_noreturn] = ACTIONS(1230),
    [anon_sym__Nonnull] = ACTIONS(1230),
    [anon_sym_alignas] = ACTIONS(1230),
    [anon_sym__Alignas] = ACTIONS(1230),
    [sym_primitive_type] = ACTIONS(1230),
    [anon_sym_enum] = ACTIONS(1230),
    [anon_sym_struct] = ACTIONS(1230),
    [anon_sym_union] = ACTIONS(1230),
    [anon_sym_if] = ACTIONS(1230),
    [anon_sym_else] = ACTIONS(1230),
    [anon_sym_switch] = ACTIONS(1230),
    [anon_sym_case] = ACTIONS(1230),
    [anon_sym_default] = ACTIONS(1230),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_break] = ACTIONS(1230),
    [anon_sym_continue] = ACTIONS(1230),
    [anon_sym_goto] = ACTIONS(1230),
    [anon_sym___try] = ACTIONS(1230),
    [anon_sym___leave] = ACTIONS(1230),
    [anon_sym_DASH_DASH] = ACTIONS(1232),
    [anon_sym_PLUS_PLUS] = ACTIONS(1232),
    [anon_sym_sizeof] = ACTIONS(1230),
    [anon_sym___alignof__] = ACTIONS(1230),
    [anon_sym___alignof] = ACTIONS(1230),
    [anon_sym__alignof] = ACTIONS(1230),
    [anon_sym_alignof] = ACTIONS(1230),
    [anon_sym__Alignof] = ACTIONS(1230),
    [anon_sym_offsetof] = ACTIONS(1230),
    [anon_sym__Generic] = ACTIONS(1230),
    [anon_sym_asm] = ACTIONS(1230),
    [anon_sym___asm__] = ACTIONS(1230),
    [anon_sym___asm] = ACTIONS(1230),
    [sym_number_literal] = ACTIONS(1232),
    [anon_sym_L_SQUOTE] = ACTIONS(1232),
    [anon_sym_u_SQUOTE] = ACTIONS(1232),
    [anon_sym_U_SQUOTE] = ACTIONS(1232),
    [anon_sym_u8_SQUOTE] = ACTIONS(1232),
    [anon_sym_SQUOTE] = ACTIONS(1232),
    [anon_sym_L_DQUOTE] = ACTIONS(1232),
    [anon_sym_u_DQUOTE] = ACTIONS(1232),
    [anon_sym_U_DQUOTE] = ACTIONS(1232),
    [anon_sym_u8_DQUOTE] = ACTIONS(1232),
    [anon_sym_DQUOTE] = ACTIONS(1232),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [anon_sym_NULL] = ACTIONS(1230),
    [anon_sym_nullptr] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(172)] = {
    [ts_builtin_sym_end] = ACTIONS(1156),
    [sym_identifier] = ACTIONS(1154),
    [aux_sym_preproc_include_token1] = ACTIONS(1154),
    [aux_sym_preproc_def_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1154),
    [sym_preproc_directive] = ACTIONS(1154),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1156),
    [anon_sym_AMP] = ACTIONS(1156),
    [anon_sym_SEMI] = ACTIONS(1156),
    [anon_sym___extension__] = ACTIONS(1154),
    [anon_sym_typedef] = ACTIONS(1154),
    [anon_sym_extern] = ACTIONS(1154),
    [anon_sym___attribute__] = ACTIONS(1154),
    [anon_sym___attribute] = ACTIONS(1154),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1156),
    [anon_sym___declspec] = ACTIONS(1154),
    [anon_sym___cdecl] = ACTIONS(1154),
    [anon_sym___clrcall] = ACTIONS(1154),
    [anon_sym___stdcall] = ACTIONS(1154),
    [anon_sym___fastcall] = ACTIONS(1154),
    [anon_sym___thiscall] = ACTIONS(1154),
    [anon_sym___vectorcall] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_signed] = ACTIONS(1154),
    [anon_sym_unsigned] = ACTIONS(1154),
    [anon_sym_long] = ACTIONS(1154),
    [anon_sym_short] = ACTIONS(1154),
    [anon_sym_static] = ACTIONS(1154),
    [anon_sym_auto] = ACTIONS(1154),
    [anon_sym_register] = ACTIONS(1154),
    [anon_sym_inline] = ACTIONS(1154),
    [anon_sym___inline] = ACTIONS(1154),
    [anon_sym___inline__] = ACTIONS(1154),
    [anon_sym___forceinline] = ACTIONS(1154),
    [anon_sym_thread_local] = ACTIONS(1154),
    [anon_sym___thread] = ACTIONS(1154),
    [anon_sym_const] = ACTIONS(1154),
    [anon_sym_constexpr] = ACTIONS(1154),
    [anon_sym_volatile] = ACTIONS(1154),
    [anon_sym_restrict] = ACTIONS(1154),
    [anon_sym___restrict__] = ACTIONS(1154),
    [anon_sym__Atomic] = ACTIONS(1154),
    [anon_sym__Noreturn] = ACTIONS(1154),
    [anon_sym_noreturn] = ACTIONS(1154),
    [anon_sym__Nonnull] = ACTIONS(1154),
    [anon_sym_alignas] = ACTIONS(1154),
    [anon_sym__Alignas] = ACTIONS(1154),
    [sym_primitive_type] = ACTIONS(1154),
    [anon_sym_enum] = ACTIONS(1154),
    [anon_sym_struct] = ACTIONS(1154),
    [anon_sym_union] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_else] = ACTIONS(1154),
    [anon_sym_switch] = ACTIONS(1154),
    [anon_sym_case] = ACTIONS(1154),
    [anon_sym_default] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_break] = ACTIONS(1154),
    [anon_sym_continue] = ACTIONS(1154),
    [anon_sym_goto] = ACTIONS(1154),
    [anon_sym___try] = ACTIONS(1154),
    [anon_sym___leave] = ACTIONS(1154),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [anon_sym_sizeof] = ACTIONS(1154),
    [anon_sym___alignof__] = ACTIONS(1154),
    [anon_sym___alignof] = ACTIONS(1154),
    [anon_sym__alignof] = ACTIONS(1154),
    [anon_sym_alignof] = ACTIONS(1154),
    [anon_sym__Alignof] = ACTIONS(1154),
    [anon_sym_offsetof] = ACTIONS(1154),
    [anon_sym__Generic] = ACTIONS(1154),
    [anon_sym_asm] = ACTIONS(1154),
    [anon_sym___asm__] = ACTIONS(1154),
    [anon_sym___asm] = ACTIONS(1154),
    [sym_number_literal] = ACTIONS(1156),
    [anon_sym_L_SQUOTE] = ACTIONS(1156),
    [anon_sym_u_SQUOTE] = ACTIONS(1156),
    [anon_sym_U_SQUOTE] = ACTIONS(1156),
    [anon_sym_u8_SQUOTE] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_L_DQUOTE] = ACTIONS(1156),
    [anon_sym_u_DQUOTE] = ACTIONS(1156),
    [anon_sym_U_DQUOTE] = ACTIONS(1156),
    [anon_sym_u8_DQUOTE] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [anon_sym_NULL] = ACTIONS(1154),
    [anon_sym_nullptr] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(173)] = {
    [sym_identifier] = ACTIONS(1138),
    [aux_sym_preproc_include_token1] = ACTIONS(1138),
    [aux_sym_preproc_def_token1] = ACTIONS(1138),
    [aux_sym_preproc_if_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1138),
    [sym_preproc_directive] = ACTIONS(1138),
    [anon_sym_LPAREN2] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_TILDE] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1138),
    [anon_sym_PLUS] = ACTIONS(1138),
    [anon_sym_STAR] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(1140),
    [anon_sym_SEMI] = ACTIONS(1140),
    [anon_sym___extension__] = ACTIONS(1138),
    [anon_sym_typedef] = ACTIONS(1138),
    [anon_sym_extern] = ACTIONS(1138),
    [anon_sym___attribute__] = ACTIONS(1138),
    [anon_sym___attribute] = ACTIONS(1138),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1140),
    [anon_sym___declspec] = ACTIONS(1138),
    [anon_sym___cdecl] = ACTIONS(1138),
    [anon_sym___clrcall] = ACTIONS(1138),
    [anon_sym___stdcall] = ACTIONS(1138),
    [anon_sym___fastcall] = ACTIONS(1138),
    [anon_sym___thiscall] = ACTIONS(1138),
    [anon_sym___vectorcall] = ACTIONS(1138),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_RBRACE] = ACTIONS(1140),
    [anon_sym_signed] = ACTIONS(1138),
    [anon_sym_unsigned] = ACTIONS(1138),
    [anon_sym_long] = ACTIONS(1138),
    [anon_sym_short] = ACTIONS(1138),
    [anon_sym_static] = ACTIONS(1138),
    [anon_sym_auto] = ACTIONS(1138),
    [anon_sym_register] = ACTIONS(1138),
    [anon_sym_inline] = ACTIONS(1138),
    [anon_sym___inline] = ACTIONS(1138),
    [anon_sym___inline__] = ACTIONS(1138),
    [anon_sym___forceinline] = ACTIONS(1138),
    [anon_sym_thread_local] = ACTIONS(1138),
    [anon_sym___thread] = ACTIONS(1138),
    [anon_sym_const] = ACTIONS(1138),
    [anon_sym_constexpr] = ACTIONS(1138),
    [anon_sym_volatile] = ACTIONS(1138),
    [anon_sym_restrict] = ACTIONS(1138),
    [anon_sym___restrict__] = ACTIONS(1138),
    [anon_sym__Atomic] = ACTIONS(1138),
    [anon_sym__Noreturn] = ACTIONS(1138),
    [anon_sym_noreturn] = ACTIONS(1138),
    [anon_sym__Nonnull] = ACTIONS(1138),
    [anon_sym_alignas] = ACTIONS(1138),
    [anon_sym__Alignas] = ACTIONS(1138),
    [sym_primitive_type] = ACTIONS(1138),
    [anon_sym_enum] = ACTIONS(1138),
    [anon_sym_struct] = ACTIONS(1138),
    [anon_sym_union] = ACTIONS(1138),
    [anon_sym_if] = ACTIONS(1138),
    [anon_sym_else] = ACTIONS(1138),
    [anon_sym_switch] = ACTIONS(1138),
    [anon_sym_case] = ACTIONS(1138),
    [anon_sym_default] = ACTIONS(1138),
    [anon_sym_while] = ACTIONS(1138),
    [anon_sym_do] = ACTIONS(1138),
    [anon_sym_for] = ACTIONS(1138),
    [anon_sym_return] = ACTIONS(1138),
    [anon_sym_break] = ACTIONS(1138),
    [anon_sym_continue] = ACTIONS(1138),
    [anon_sym_goto] = ACTIONS(1138),
    [anon_sym___try] = ACTIONS(1138),
    [anon_sym___leave] = ACTIONS(1138),
    [anon_sym_DASH_DASH] = ACTIONS(1140),
    [anon_sym_PLUS_PLUS] = ACTIONS(1140),
    [anon_sym_sizeof] = ACTIONS(1138),
    [anon_sym___alignof__] = ACTIONS(1138),
    [anon_sym___alignof] = ACTIONS(1138),
    [anon_sym__alignof] = ACTIONS(1138),
    [anon_sym_alignof] = ACTIONS(1138),
    [anon_sym__Alignof] = ACTIONS(1138),
    [anon_sym_offsetof] = ACTIONS(1138),
    [anon_sym__Generic] = ACTIONS(1138),
    [anon_sym_asm] = ACTIONS(1138),
    [anon_sym___asm__] = ACTIONS(1138),
    [anon_sym___asm] = ACTIONS(1138),
    [sym_number_literal] = ACTIONS(1140),
    [anon_sym_L_SQUOTE] = ACTIONS(1140),
    [anon_sym_u_SQUOTE] = ACTIONS(1140),
    [anon_sym_U_SQUOTE] = ACTIONS(1140),
    [anon_sym_u8_SQUOTE] = ACTIONS(1140),
    [anon_sym_SQUOTE] = ACTIONS(1140),
    [anon_sym_L_DQUOTE] = ACTIONS(1140),
    [anon_sym_u_DQUOTE] = ACTIONS(1140),
    [anon_sym_U_DQUOTE] = ACTIONS(1140),
    [anon_sym_u8_DQUOTE] = ACTIONS(1140),
    [anon_sym_DQUOTE] = ACTIONS(1140),
    [sym_true] = ACTIONS(1138),
    [sym_false] = ACTIONS(1138),
    [anon_sym_NULL] = ACTIONS(1138),
    [anon_sym_nullptr] = ACTIONS(1138),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(174)] = {
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1258),
    [aux_sym_preproc_include_token1] = ACTIONS(1258),
    [aux_sym_preproc_def_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1258),
    [sym_preproc_directive] = ACTIONS(1258),
    [anon_sym_LPAREN2] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1258),
    [anon_sym_PLUS] = ACTIONS(1258),
    [anon_sym_STAR] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym___extension__] = ACTIONS(1258),
    [anon_sym_typedef] = ACTIONS(1258),
    [anon_sym_extern] = ACTIONS(1258),
    [anon_sym___attribute__] = ACTIONS(1258),
    [anon_sym___attribute] = ACTIONS(1258),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1260),
    [anon_sym___declspec] = ACTIONS(1258),
    [anon_sym___cdecl] = ACTIONS(1258),
    [anon_sym___clrcall] = ACTIONS(1258),
    [anon_sym___stdcall] = ACTIONS(1258),
    [anon_sym___fastcall] = ACTIONS(1258),
    [anon_sym___thiscall] = ACTIONS(1258),
    [anon_sym___vectorcall] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_signed] = ACTIONS(1258),
    [anon_sym_unsigned] = ACTIONS(1258),
    [anon_sym_long] = ACTIONS(1258),
    [anon_sym_short] = ACTIONS(1258),
    [anon_sym_static] = ACTIONS(1258),
    [anon_sym_auto] = ACTIONS(1258),
    [anon_sym_register] = ACTIONS(1258),
    [anon_sym_inline] = ACTIONS(1258),
    [anon_sym___inline] = ACTIONS(1258),
    [anon_sym___inline__] = ACTIONS(1258),
    [anon_sym___forceinline] = ACTIONS(1258),
    [anon_sym_thread_local] = ACTIONS(1258),
    [anon_sym___thread] = ACTIONS(1258),
    [anon_sym_const] = ACTIONS(1258),
    [anon_sym_constexpr] = ACTIONS(1258),
    [anon_sym_volatile] = ACTIONS(1258),
    [anon_sym_restrict] = ACTIONS(1258),
    [anon_sym___restrict__] = ACTIONS(1258),
    [anon_sym__Atomic] = ACTIONS(1258),
    [anon_sym__Noreturn] = ACTIONS(1258),
    [anon_sym_noreturn] = ACTIONS(1258),
    [anon_sym__Nonnull] = ACTIONS(1258),
    [anon_sym_alignas] = ACTIONS(1258),
    [anon_sym__Alignas] = ACTIONS(1258),
    [sym_primitive_type] = ACTIONS(1258),
    [anon_sym_enum] = ACTIONS(1258),
    [anon_sym_struct] = ACTIONS(1258),
    [anon_sym_union] = ACTIONS(1258),
    [anon_sym_if] = ACTIONS(1258),
    [anon_sym_else] = ACTIONS(1258),
    [anon_sym_switch] = ACTIONS(1258),
    [anon_sym_case] = ACTIONS(1258),
    [anon_sym_default] = ACTIONS(1258),
    [anon_sym_while] = ACTIONS(1258),
    [anon_sym_do] = ACTIONS(1258),
    [anon_sym_for] = ACTIONS(1258),
    [anon_sym_return] = ACTIONS(1258),
    [anon_sym_break] = ACTIONS(1258),
    [anon_sym_continue] = ACTIONS(1258),
    [anon_sym_goto] = ACTIONS(1258),
    [anon_sym___try] = ACTIONS(1258),
    [anon_sym___leave] = ACTIONS(1258),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_sizeof] = ACTIONS(1258),
    [anon_sym___alignof__] = ACTIONS(1258),
    [anon_sym___alignof] = ACTIONS(1258),
    [anon_sym__alignof] = ACTIONS(1258),
    [anon_sym_alignof] = ACTIONS(1258),
    [anon_sym__Alignof] = ACTIONS(1258),
    [anon_sym_offsetof] = ACTIONS(1258),
    [anon_sym__Generic] = ACTIONS(1258),
    [anon_sym_asm] = ACTIONS(1258),
    [anon_sym___asm__] = ACTIONS(1258),
    [anon_sym___asm] = ACTIONS(1258),
    [sym_number_literal] = ACTIONS(1260),
    [anon_sym_L_SQUOTE] = ACTIONS(1260),
    [anon_sym_u_SQUOTE] = ACTIONS(1260),
    [anon_sym_U_SQUOTE] = ACTIONS(1260),
    [anon_sym_u8_SQUOTE] = ACTIONS(1260),
    [anon_sym_SQUOTE] = ACTIONS(1260),
    [anon_sym_L_DQUOTE] = ACTIONS(1260),
    [anon_sym_u_DQUOTE] = ACTIONS(1260),
    [anon_sym_U_DQUOTE] = ACTIONS(1260),
    [anon_sym_u8_DQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1260),
    [sym_true] = ACTIONS(1258),
    [sym_false] = ACTIONS(1258),
    [anon_sym_NULL] = ACTIONS(1258),
    [anon_sym_nullptr] = ACTIONS(1258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(175)] = {
    [ts_builtin_sym_end] = ACTIONS(1180),
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(176)] = {
    [sym_identifier] = ACTIONS(1142),
    [aux_sym_preproc_include_token1] = ACTIONS(1142),
    [aux_sym_preproc_def_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token2] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1142),
    [sym_preproc_directive] = ACTIONS(1142),
    [anon_sym_LPAREN2] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_TILDE] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1142),
    [anon_sym_PLUS] = ACTIONS(1142),
    [anon_sym_STAR] = ACTIONS(1144),
    [anon_sym_AMP] = ACTIONS(1144),
    [anon_sym_SEMI] = ACTIONS(1144),
    [anon_sym___extension__] = ACTIONS(1142),
    [anon_sym_typedef] = ACTIONS(1142),
    [anon_sym_extern] = ACTIONS(1142),
    [anon_sym___attribute__] = ACTIONS(1142),
    [anon_sym___attribute] = ACTIONS(1142),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1144),
    [anon_sym___declspec] = ACTIONS(1142),
    [anon_sym___cdecl] = ACTIONS(1142),
    [anon_sym___clrcall] = ACTIONS(1142),
    [anon_sym___stdcall] = ACTIONS(1142),
    [anon_sym___fastcall] = ACTIONS(1142),
    [anon_sym___thiscall] = ACTIONS(1142),
    [anon_sym___vectorcall] = ACTIONS(1142),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_signed] = ACTIONS(1142),
    [anon_sym_unsigned] = ACTIONS(1142),
    [anon_sym_long] = ACTIONS(1142),
    [anon_sym_short] = ACTIONS(1142),
    [anon_sym_static] = ACTIONS(1142),
    [anon_sym_auto] = ACTIONS(1142),
    [anon_sym_register] = ACTIONS(1142),
    [anon_sym_inline] = ACTIONS(1142),
    [anon_sym___inline] = ACTIONS(1142),
    [anon_sym___inline__] = ACTIONS(1142),
    [anon_sym___forceinline] = ACTIONS(1142),
    [anon_sym_thread_local] = ACTIONS(1142),
    [anon_sym___thread] = ACTIONS(1142),
    [anon_sym_const] = ACTIONS(1142),
    [anon_sym_constexpr] = ACTIONS(1142),
    [anon_sym_volatile] = ACTIONS(1142),
    [anon_sym_restrict] = ACTIONS(1142),
    [anon_sym___restrict__] = ACTIONS(1142),
    [anon_sym__Atomic] = ACTIONS(1142),
    [anon_sym__Noreturn] = ACTIONS(1142),
    [anon_sym_noreturn] = ACTIONS(1142),
    [anon_sym__Nonnull] = ACTIONS(1142),
    [anon_sym_alignas] = ACTIONS(1142),
    [anon_sym__Alignas] = ACTIONS(1142),
    [sym_primitive_type] = ACTIONS(1142),
    [anon_sym_enum] = ACTIONS(1142),
    [anon_sym_struct] = ACTIONS(1142),
    [anon_sym_union] = ACTIONS(1142),
    [anon_sym_if] = ACTIONS(1142),
    [anon_sym_else] = ACTIONS(1142),
    [anon_sym_switch] = ACTIONS(1142),
    [anon_sym_case] = ACTIONS(1142),
    [anon_sym_default] = ACTIONS(1142),
    [anon_sym_while] = ACTIONS(1142),
    [anon_sym_do] = ACTIONS(1142),
    [anon_sym_for] = ACTIONS(1142),
    [anon_sym_return] = ACTIONS(1142),
    [anon_sym_break] = ACTIONS(1142),
    [anon_sym_continue] = ACTIONS(1142),
    [anon_sym_goto] = ACTIONS(1142),
    [anon_sym___try] = ACTIONS(1142),
    [anon_sym___leave] = ACTIONS(1142),
    [anon_sym_DASH_DASH] = ACTIONS(1144),
    [anon_sym_PLUS_PLUS] = ACTIONS(1144),
    [anon_sym_sizeof] = ACTIONS(1142),
    [anon_sym___alignof__] = ACTIONS(1142),
    [anon_sym___alignof] = ACTIONS(1142),
    [anon_sym__alignof] = ACTIONS(1142),
    [anon_sym_alignof] = ACTIONS(1142),
    [anon_sym__Alignof] = ACTIONS(1142),
    [anon_sym_offsetof] = ACTIONS(1142),
    [anon_sym__Generic] = ACTIONS(1142),
    [anon_sym_asm] = ACTIONS(1142),
    [anon_sym___asm__] = ACTIONS(1142),
    [anon_sym___asm] = ACTIONS(1142),
    [sym_number_literal] = ACTIONS(1144),
    [anon_sym_L_SQUOTE] = ACTIONS(1144),
    [anon_sym_u_SQUOTE] = ACTIONS(1144),
    [anon_sym_U_SQUOTE] = ACTIONS(1144),
    [anon_sym_u8_SQUOTE] = ACTIONS(1144),
    [anon_sym_SQUOTE] = ACTIONS(1144),
    [anon_sym_L_DQUOTE] = ACTIONS(1144),
    [anon_sym_u_DQUOTE] = ACTIONS(1144),
    [anon_sym_U_DQUOTE] = ACTIONS(1144),
    [anon_sym_u8_DQUOTE] = ACTIONS(1144),
    [anon_sym_DQUOTE] = ACTIONS(1144),
    [sym_true] = ACTIONS(1142),
    [sym_false] = ACTIONS(1142),
    [anon_sym_NULL] = ACTIONS(1142),
    [anon_sym_nullptr] = ACTIONS(1142),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(177)] = {
    [sym_identifier] = ACTIONS(1158),
    [aux_sym_preproc_include_token1] = ACTIONS(1158),
    [aux_sym_preproc_def_token1] = ACTIONS(1158),
    [aux_sym_preproc_if_token1] = ACTIONS(1158),
    [aux_sym_preproc_if_token2] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1158),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1158),
    [sym_preproc_directive] = ACTIONS(1158),
    [anon_sym_LPAREN2] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_TILDE] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1158),
    [anon_sym_PLUS] = ACTIONS(1158),
    [anon_sym_STAR] = ACTIONS(1160),
    [anon_sym_AMP] = ACTIONS(1160),
    [anon_sym_SEMI] = ACTIONS(1160),
    [anon_sym___extension__] = ACTIONS(1158),
    [anon_sym_typedef] = ACTIONS(1158),
    [anon_sym_extern] = ACTIONS(1158),
    [anon_sym___attribute__] = ACTIONS(1158),
    [anon_sym___attribute] = ACTIONS(1158),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1160),
    [anon_sym___declspec] = ACTIONS(1158),
    [anon_sym___cdecl] = ACTIONS(1158),
    [anon_sym___clrcall] = ACTIONS(1158),
    [anon_sym___stdcall] = ACTIONS(1158),
    [anon_sym___fastcall] = ACTIONS(1158),
    [anon_sym___thiscall] = ACTIONS(1158),
    [anon_sym___vectorcall] = ACTIONS(1158),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_signed] = ACTIONS(1158),
    [anon_sym_unsigned] = ACTIONS(1158),
    [anon_sym_long] = ACTIONS(1158),
    [anon_sym_short] = ACTIONS(1158),
    [anon_sym_static] = ACTIONS(1158),
    [anon_sym_auto] = ACTIONS(1158),
    [anon_sym_register] = ACTIONS(1158),
    [anon_sym_inline] = ACTIONS(1158),
    [anon_sym___inline] = ACTIONS(1158),
    [anon_sym___inline__] = ACTIONS(1158),
    [anon_sym___forceinline] = ACTIONS(1158),
    [anon_sym_thread_local] = ACTIONS(1158),
    [anon_sym___thread] = ACTIONS(1158),
    [anon_sym_const] = ACTIONS(1158),
    [anon_sym_constexpr] = ACTIONS(1158),
    [anon_sym_volatile] = ACTIONS(1158),
    [anon_sym_restrict] = ACTIONS(1158),
    [anon_sym___restrict__] = ACTIONS(1158),
    [anon_sym__Atomic] = ACTIONS(1158),
    [anon_sym__Noreturn] = ACTIONS(1158),
    [anon_sym_noreturn] = ACTIONS(1158),
    [anon_sym__Nonnull] = ACTIONS(1158),
    [anon_sym_alignas] = ACTIONS(1158),
    [anon_sym__Alignas] = ACTIONS(1158),
    [sym_primitive_type] = ACTIONS(1158),
    [anon_sym_enum] = ACTIONS(1158),
    [anon_sym_struct] = ACTIONS(1158),
    [anon_sym_union] = ACTIONS(1158),
    [anon_sym_if] = ACTIONS(1158),
    [anon_sym_else] = ACTIONS(1158),
    [anon_sym_switch] = ACTIONS(1158),
    [anon_sym_case] = ACTIONS(1158),
    [anon_sym_default] = ACTIONS(1158),
    [anon_sym_while] = ACTIONS(1158),
    [anon_sym_do] = ACTIONS(1158),
    [anon_sym_for] = ACTIONS(1158),
    [anon_sym_return] = ACTIONS(1158),
    [anon_sym_break] = ACTIONS(1158),
    [anon_sym_continue] = ACTIONS(1158),
    [anon_sym_goto] = ACTIONS(1158),
    [anon_sym___try] = ACTIONS(1158),
    [anon_sym___leave] = ACTIONS(1158),
    [anon_sym_DASH_DASH] = ACTIONS(1160),
    [anon_sym_PLUS_PLUS] = ACTIONS(1160),
    [anon_sym_sizeof] = ACTIONS(1158),
    [anon_sym___alignof__] = ACTIONS(1158),
    [anon_sym___alignof] = ACTIONS(1158),
    [anon_sym__alignof] = ACTIONS(1158),
    [anon_sym_alignof] = ACTIONS(1158),
    [anon_sym__Alignof] = ACTIONS(1158),
    [anon_sym_offsetof] = ACTIONS(1158),
    [anon_sym__Generic] = ACTIONS(1158),
    [anon_sym_asm] = ACTIONS(1158),
    [anon_sym___asm__] = ACTIONS(1158),
    [anon_sym___asm] = ACTIONS(1158),
    [sym_number_literal] = ACTIONS(1160),
    [anon_sym_L_SQUOTE] = ACTIONS(1160),
    [anon_sym_u_SQUOTE] = ACTIONS(1160),
    [anon_sym_U_SQUOTE] = ACTIONS(1160),
    [anon_sym_u8_SQUOTE] = ACTIONS(1160),
    [anon_sym_SQUOTE] = ACTIONS(1160),
    [anon_sym_L_DQUOTE] = ACTIONS(1160),
    [anon_sym_u_DQUOTE] = ACTIONS(1160),
    [anon_sym_U_DQUOTE] = ACTIONS(1160),
    [anon_sym_u8_DQUOTE] = ACTIONS(1160),
    [anon_sym_DQUOTE] = ACTIONS(1160),
    [sym_true] = ACTIONS(1158),
    [sym_false] = ACTIONS(1158),
    [anon_sym_NULL] = ACTIONS(1158),
    [anon_sym_nullptr] = ACTIONS(1158),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(178)] = {
    [sym_identifier] = ACTIONS(1134),
    [aux_sym_preproc_include_token1] = ACTIONS(1134),
    [aux_sym_preproc_def_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token2] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1134),
    [sym_preproc_directive] = ACTIONS(1134),
    [anon_sym_LPAREN2] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_TILDE] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1134),
    [anon_sym_PLUS] = ACTIONS(1134),
    [anon_sym_STAR] = ACTIONS(1136),
    [anon_sym_AMP] = ACTIONS(1136),
    [anon_sym_SEMI] = ACTIONS(1136),
    [anon_sym___extension__] = ACTIONS(1134),
    [anon_sym_typedef] = ACTIONS(1134),
    [anon_sym_extern] = ACTIONS(1134),
    [anon_sym___attribute__] = ACTIONS(1134),
    [anon_sym___attribute] = ACTIONS(1134),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1136),
    [anon_sym___declspec] = ACTIONS(1134),
    [anon_sym___cdecl] = ACTIONS(1134),
    [anon_sym___clrcall] = ACTIONS(1134),
    [anon_sym___stdcall] = ACTIONS(1134),
    [anon_sym___fastcall] = ACTIONS(1134),
    [anon_sym___thiscall] = ACTIONS(1134),
    [anon_sym___vectorcall] = ACTIONS(1134),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_signed] = ACTIONS(1134),
    [anon_sym_unsigned] = ACTIONS(1134),
    [anon_sym_long] = ACTIONS(1134),
    [anon_sym_short] = ACTIONS(1134),
    [anon_sym_static] = ACTIONS(1134),
    [anon_sym_auto] = ACTIONS(1134),
    [anon_sym_register] = ACTIONS(1134),
    [anon_sym_inline] = ACTIONS(1134),
    [anon_sym___inline] = ACTIONS(1134),
    [anon_sym___inline__] = ACTIONS(1134),
    [anon_sym___forceinline] = ACTIONS(1134),
    [anon_sym_thread_local] = ACTIONS(1134),
    [anon_sym___thread] = ACTIONS(1134),
    [anon_sym_const] = ACTIONS(1134),
    [anon_sym_constexpr] = ACTIONS(1134),
    [anon_sym_volatile] = ACTIONS(1134),
    [anon_sym_restrict] = ACTIONS(1134),
    [anon_sym___restrict__] = ACTIONS(1134),
    [anon_sym__Atomic] = ACTIONS(1134),
    [anon_sym__Noreturn] = ACTIONS(1134),
    [anon_sym_noreturn] = ACTIONS(1134),
    [anon_sym__Nonnull] = ACTIONS(1134),
    [anon_sym_alignas] = ACTIONS(1134),
    [anon_sym__Alignas] = ACTIONS(1134),
    [sym_primitive_type] = ACTIONS(1134),
    [anon_sym_enum] = ACTIONS(1134),
    [anon_sym_struct] = ACTIONS(1134),
    [anon_sym_union] = ACTIONS(1134),
    [anon_sym_if] = ACTIONS(1134),
    [anon_sym_else] = ACTIONS(1134),
    [anon_sym_switch] = ACTIONS(1134),
    [anon_sym_case] = ACTIONS(1134),
    [anon_sym_default] = ACTIONS(1134),
    [anon_sym_while] = ACTIONS(1134),
    [anon_sym_do] = ACTIONS(1134),
    [anon_sym_for] = ACTIONS(1134),
    [anon_sym_return] = ACTIONS(1134),
    [anon_sym_break] = ACTIONS(1134),
    [anon_sym_continue] = ACTIONS(1134),
    [anon_sym_goto] = ACTIONS(1134),
    [anon_sym___try] = ACTIONS(1134),
    [anon_sym___leave] = ACTIONS(1134),
    [anon_sym_DASH_DASH] = ACTIONS(1136),
    [anon_sym_PLUS_PLUS] = ACTIONS(1136),
    [anon_sym_sizeof] = ACTIONS(1134),
    [anon_sym___alignof__] = ACTIONS(1134),
    [anon_sym___alignof] = ACTIONS(1134),
    [anon_sym__alignof] = ACTIONS(1134),
    [anon_sym_alignof] = ACTIONS(1134),
    [anon_sym__Alignof] = ACTIONS(1134),
    [anon_sym_offsetof] = ACTIONS(1134),
    [anon_sym__Generic] = ACTIONS(1134),
    [anon_sym_asm] = ACTIONS(1134),
    [anon_sym___asm__] = ACTIONS(1134),
    [anon_sym___asm] = ACTIONS(1134),
    [sym_number_literal] = ACTIONS(1136),
    [anon_sym_L_SQUOTE] = ACTIONS(1136),
    [anon_sym_u_SQUOTE] = ACTIONS(1136),
    [anon_sym_U_SQUOTE] = ACTIONS(1136),
    [anon_sym_u8_SQUOTE] = ACTIONS(1136),
    [anon_sym_SQUOTE] = ACTIONS(1136),
    [anon_sym_L_DQUOTE] = ACTIONS(1136),
    [anon_sym_u_DQUOTE] = ACTIONS(1136),
    [anon_sym_U_DQUOTE] = ACTIONS(1136),
    [anon_sym_u8_DQUOTE] = ACTIONS(1136),
    [anon_sym_DQUOTE] = ACTIONS(1136),
    [sym_true] = ACTIONS(1134),
    [sym_false] = ACTIONS(1134),
    [anon_sym_NULL] = ACTIONS(1134),
    [anon_sym_nullptr] = ACTIONS(1134),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(179)] = {
    [sym_identifier] = ACTIONS(1154),
    [aux_sym_preproc_include_token1] = ACTIONS(1154),
    [aux_sym_preproc_def_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1154),
    [sym_preproc_directive] = ACTIONS(1154),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1156),
    [anon_sym_AMP] = ACTIONS(1156),
    [anon_sym_SEMI] = ACTIONS(1156),
    [anon_sym___extension__] = ACTIONS(1154),
    [anon_sym_typedef] = ACTIONS(1154),
    [anon_sym_extern] = ACTIONS(1154),
    [anon_sym___attribute__] = ACTIONS(1154),
    [anon_sym___attribute] = ACTIONS(1154),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1156),
    [anon_sym___declspec] = ACTIONS(1154),
    [anon_sym___cdecl] = ACTIONS(1154),
    [anon_sym___clrcall] = ACTIONS(1154),
    [anon_sym___stdcall] = ACTIONS(1154),
    [anon_sym___fastcall] = ACTIONS(1154),
    [anon_sym___thiscall] = ACTIONS(1154),
    [anon_sym___vectorcall] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_RBRACE] = ACTIONS(1156),
    [anon_sym_signed] = ACTIONS(1154),
    [anon_sym_unsigned] = ACTIONS(1154),
    [anon_sym_long] = ACTIONS(1154),
    [anon_sym_short] = ACTIONS(1154),
    [anon_sym_static] = ACTIONS(1154),
    [anon_sym_auto] = ACTIONS(1154),
    [anon_sym_register] = ACTIONS(1154),
    [anon_sym_inline] = ACTIONS(1154),
    [anon_sym___inline] = ACTIONS(1154),
    [anon_sym___inline__] = ACTIONS(1154),
    [anon_sym___forceinline] = ACTIONS(1154),
    [anon_sym_thread_local] = ACTIONS(1154),
    [anon_sym___thread] = ACTIONS(1154),
    [anon_sym_const] = ACTIONS(1154),
    [anon_sym_constexpr] = ACTIONS(1154),
    [anon_sym_volatile] = ACTIONS(1154),
    [anon_sym_restrict] = ACTIONS(1154),
    [anon_sym___restrict__] = ACTIONS(1154),
    [anon_sym__Atomic] = ACTIONS(1154),
    [anon_sym__Noreturn] = ACTIONS(1154),
    [anon_sym_noreturn] = ACTIONS(1154),
    [anon_sym__Nonnull] = ACTIONS(1154),
    [anon_sym_alignas] = ACTIONS(1154),
    [anon_sym__Alignas] = ACTIONS(1154),
    [sym_primitive_type] = ACTIONS(1154),
    [anon_sym_enum] = ACTIONS(1154),
    [anon_sym_struct] = ACTIONS(1154),
    [anon_sym_union] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_else] = ACTIONS(1154),
    [anon_sym_switch] = ACTIONS(1154),
    [anon_sym_case] = ACTIONS(1154),
    [anon_sym_default] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_break] = ACTIONS(1154),
    [anon_sym_continue] = ACTIONS(1154),
    [anon_sym_goto] = ACTIONS(1154),
    [anon_sym___try] = ACTIONS(1154),
    [anon_sym___leave] = ACTIONS(1154),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [anon_sym_sizeof] = ACTIONS(1154),
    [anon_sym___alignof__] = ACTIONS(1154),
    [anon_sym___alignof] = ACTIONS(1154),
    [anon_sym__alignof] = ACTIONS(1154),
    [anon_sym_alignof] = ACTIONS(1154),
    [anon_sym__Alignof] = ACTIONS(1154),
    [anon_sym_offsetof] = ACTIONS(1154),
    [anon_sym__Generic] = ACTIONS(1154),
    [anon_sym_asm] = ACTIONS(1154),
    [anon_sym___asm__] = ACTIONS(1154),
    [anon_sym___asm] = ACTIONS(1154),
    [sym_number_literal] = ACTIONS(1156),
    [anon_sym_L_SQUOTE] = ACTIONS(1156),
    [anon_sym_u_SQUOTE] = ACTIONS(1156),
    [anon_sym_U_SQUOTE] = ACTIONS(1156),
    [anon_sym_u8_SQUOTE] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_L_DQUOTE] = ACTIONS(1156),
    [anon_sym_u_DQUOTE] = ACTIONS(1156),
    [anon_sym_U_DQUOTE] = ACTIONS(1156),
    [anon_sym_u8_DQUOTE] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [anon_sym_NULL] = ACTIONS(1154),
    [anon_sym_nullptr] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(180)] = {
    [sym_identifier] = ACTIONS(1230),
    [aux_sym_preproc_include_token1] = ACTIONS(1230),
    [aux_sym_preproc_def_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token1] = ACTIONS(1230),
    [aux_sym_preproc_if_token2] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1230),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1230),
    [sym_preproc_directive] = ACTIONS(1230),
    [anon_sym_LPAREN2] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1230),
    [anon_sym_PLUS] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym___extension__] = ACTIONS(1230),
    [anon_sym_typedef] = ACTIONS(1230),
    [anon_sym_extern] = ACTIONS(1230),
    [anon_sym___attribute__] = ACTIONS(1230),
    [anon_sym___attribute] = ACTIONS(1230),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1232),
    [anon_sym___declspec] = ACTIONS(1230),
    [anon_sym___cdecl] = ACTIONS(1230),
    [anon_sym___clrcall] = ACTIONS(1230),
    [anon_sym___stdcall] = ACTIONS(1230),
    [anon_sym___fastcall] = ACTIONS(1230),
    [anon_sym___thiscall] = ACTIONS(1230),
    [anon_sym___vectorcall] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_signed] = ACTIONS(1230),
    [anon_sym_unsigned] = ACTIONS(1230),
    [anon_sym_long] = ACTIONS(1230),
    [anon_sym_short] = ACTIONS(1230),
    [anon_sym_static] = ACTIONS(1230),
    [anon_sym_auto] = ACTIONS(1230),
    [anon_sym_register] = ACTIONS(1230),
    [anon_sym_inline] = ACTIONS(1230),
    [anon_sym___inline] = ACTIONS(1230),
    [anon_sym___inline__] = ACTIONS(1230),
    [anon_sym___forceinline] = ACTIONS(1230),
    [anon_sym_thread_local] = ACTIONS(1230),
    [anon_sym___thread] = ACTIONS(1230),
    [anon_sym_const] = ACTIONS(1230),
    [anon_sym_constexpr] = ACTIONS(1230),
    [anon_sym_volatile] = ACTIONS(1230),
    [anon_sym_restrict] = ACTIONS(1230),
    [anon_sym___restrict__] = ACTIONS(1230),
    [anon_sym__Atomic] = ACTIONS(1230),
    [anon_sym__Noreturn] = ACTIONS(1230),
    [anon_sym_noreturn] = ACTIONS(1230),
    [anon_sym__Nonnull] = ACTIONS(1230),
    [anon_sym_alignas] = ACTIONS(1230),
    [anon_sym__Alignas] = ACTIONS(1230),
    [sym_primitive_type] = ACTIONS(1230),
    [anon_sym_enum] = ACTIONS(1230),
    [anon_sym_struct] = ACTIONS(1230),
    [anon_sym_union] = ACTIONS(1230),
    [anon_sym_if] = ACTIONS(1230),
    [anon_sym_else] = ACTIONS(1230),
    [anon_sym_switch] = ACTIONS(1230),
    [anon_sym_case] = ACTIONS(1230),
    [anon_sym_default] = ACTIONS(1230),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_break] = ACTIONS(1230),
    [anon_sym_continue] = ACTIONS(1230),
    [anon_sym_goto] = ACTIONS(1230),
    [anon_sym___try] = ACTIONS(1230),
    [anon_sym___leave] = ACTIONS(1230),
    [anon_sym_DASH_DASH] = ACTIONS(1232),
    [anon_sym_PLUS_PLUS] = ACTIONS(1232),
    [anon_sym_sizeof] = ACTIONS(1230),
    [anon_sym___alignof__] = ACTIONS(1230),
    [anon_sym___alignof] = ACTIONS(1230),
    [anon_sym__alignof] = ACTIONS(1230),
    [anon_sym_alignof] = ACTIONS(1230),
    [anon_sym__Alignof] = ACTIONS(1230),
    [anon_sym_offsetof] = ACTIONS(1230),
    [anon_sym__Generic] = ACTIONS(1230),
    [anon_sym_asm] = ACTIONS(1230),
    [anon_sym___asm__] = ACTIONS(1230),
    [anon_sym___asm] = ACTIONS(1230),
    [sym_number_literal] = ACTIONS(1232),
    [anon_sym_L_SQUOTE] = ACTIONS(1232),
    [anon_sym_u_SQUOTE] = ACTIONS(1232),
    [anon_sym_U_SQUOTE] = ACTIONS(1232),
    [anon_sym_u8_SQUOTE] = ACTIONS(1232),
    [anon_sym_SQUOTE] = ACTIONS(1232),
    [anon_sym_L_DQUOTE] = ACTIONS(1232),
    [anon_sym_u_DQUOTE] = ACTIONS(1232),
    [anon_sym_U_DQUOTE] = ACTIONS(1232),
    [anon_sym_u8_DQUOTE] = ACTIONS(1232),
    [anon_sym_DQUOTE] = ACTIONS(1232),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [anon_sym_NULL] = ACTIONS(1230),
    [anon_sym_nullptr] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(181)] = {
    [sym_identifier] = ACTIONS(1214),
    [aux_sym_preproc_include_token1] = ACTIONS(1214),
    [aux_sym_preproc_def_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token2] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1214),
    [sym_preproc_directive] = ACTIONS(1214),
    [anon_sym_LPAREN2] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_TILDE] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1214),
    [anon_sym_PLUS] = ACTIONS(1214),
    [anon_sym_STAR] = ACTIONS(1216),
    [anon_sym_AMP] = ACTIONS(1216),
    [anon_sym_SEMI] = ACTIONS(1216),
    [anon_sym___extension__] = ACTIONS(1214),
    [anon_sym_typedef] = ACTIONS(1214),
    [anon_sym_extern] = ACTIONS(1214),
    [anon_sym___attribute__] = ACTIONS(1214),
    [anon_sym___attribute] = ACTIONS(1214),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1216),
    [anon_sym___declspec] = ACTIONS(1214),
    [anon_sym___cdecl] = ACTIONS(1214),
    [anon_sym___clrcall] = ACTIONS(1214),
    [anon_sym___stdcall] = ACTIONS(1214),
    [anon_sym___fastcall] = ACTIONS(1214),
    [anon_sym___thiscall] = ACTIONS(1214),
    [anon_sym___vectorcall] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_signed] = ACTIONS(1214),
    [anon_sym_unsigned] = ACTIONS(1214),
    [anon_sym_long] = ACTIONS(1214),
    [anon_sym_short] = ACTIONS(1214),
    [anon_sym_static] = ACTIONS(1214),
    [anon_sym_auto] = ACTIONS(1214),
    [anon_sym_register] = ACTIONS(1214),
    [anon_sym_inline] = ACTIONS(1214),
    [anon_sym___inline] = ACTIONS(1214),
    [anon_sym___inline__] = ACTIONS(1214),
    [anon_sym___forceinline] = ACTIONS(1214),
    [anon_sym_thread_local] = ACTIONS(1214),
    [anon_sym___thread] = ACTIONS(1214),
    [anon_sym_const] = ACTIONS(1214),
    [anon_sym_constexpr] = ACTIONS(1214),
    [anon_sym_volatile] = ACTIONS(1214),
    [anon_sym_restrict] = ACTIONS(1214),
    [anon_sym___restrict__] = ACTIONS(1214),
    [anon_sym__Atomic] = ACTIONS(1214),
    [anon_sym__Noreturn] = ACTIONS(1214),
    [anon_sym_noreturn] = ACTIONS(1214),
    [anon_sym__Nonnull] = ACTIONS(1214),
    [anon_sym_alignas] = ACTIONS(1214),
    [anon_sym__Alignas] = ACTIONS(1214),
    [sym_primitive_type] = ACTIONS(1214),
    [anon_sym_enum] = ACTIONS(1214),
    [anon_sym_struct] = ACTIONS(1214),
    [anon_sym_union] = ACTIONS(1214),
    [anon_sym_if] = ACTIONS(1214),
    [anon_sym_else] = ACTIONS(1214),
    [anon_sym_switch] = ACTIONS(1214),
    [anon_sym_case] = ACTIONS(1214),
    [anon_sym_default] = ACTIONS(1214),
    [anon_sym_while] = ACTIONS(1214),
    [anon_sym_do] = ACTIONS(1214),
    [anon_sym_for] = ACTIONS(1214),
    [anon_sym_return] = ACTIONS(1214),
    [anon_sym_break] = ACTIONS(1214),
    [anon_sym_continue] = ACTIONS(1214),
    [anon_sym_goto] = ACTIONS(1214),
    [anon_sym___try] = ACTIONS(1214),
    [anon_sym___leave] = ACTIONS(1214),
    [anon_sym_DASH_DASH] = ACTIONS(1216),
    [anon_sym_PLUS_PLUS] = ACTIONS(1216),
    [anon_sym_sizeof] = ACTIONS(1214),
    [anon_sym___alignof__] = ACTIONS(1214),
    [anon_sym___alignof] = ACTIONS(1214),
    [anon_sym__alignof] = ACTIONS(1214),
    [anon_sym_alignof] = ACTIONS(1214),
    [anon_sym__Alignof] = ACTIONS(1214),
    [anon_sym_offsetof] = ACTIONS(1214),
    [anon_sym__Generic] = ACTIONS(1214),
    [anon_sym_asm] = ACTIONS(1214),
    [anon_sym___asm__] = ACTIONS(1214),
    [anon_sym___asm] = ACTIONS(1214),
    [sym_number_literal] = ACTIONS(1216),
    [anon_sym_L_SQUOTE] = ACTIONS(1216),
    [anon_sym_u_SQUOTE] = ACTIONS(1216),
    [anon_sym_U_SQUOTE] = ACTIONS(1216),
    [anon_sym_u8_SQUOTE] = ACTIONS(1216),
    [anon_sym_SQUOTE] = ACTIONS(1216),
    [anon_sym_L_DQUOTE] = ACTIONS(1216),
    [anon_sym_u_DQUOTE] = ACTIONS(1216),
    [anon_sym_U_DQUOTE] = ACTIONS(1216),
    [anon_sym_u8_DQUOTE] = ACTIONS(1216),
    [anon_sym_DQUOTE] = ACTIONS(1216),
    [sym_true] = ACTIONS(1214),
    [sym_false] = ACTIONS(1214),
    [anon_sym_NULL] = ACTIONS(1214),
    [anon_sym_nullptr] = ACTIONS(1214),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(182)] = {
    [sym_identifier] = ACTIONS(1226),
    [aux_sym_preproc_include_token1] = ACTIONS(1226),
    [aux_sym_preproc_def_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token1] = ACTIONS(1226),
    [aux_sym_preproc_if_token2] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1226),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1226),
    [sym_preproc_directive] = ACTIONS(1226),
    [anon_sym_LPAREN2] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1226),
    [anon_sym_PLUS] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym___extension__] = ACTIONS(1226),
    [anon_sym_typedef] = ACTIONS(1226),
    [anon_sym_extern] = ACTIONS(1226),
    [anon_sym___attribute__] = ACTIONS(1226),
    [anon_sym___attribute] = ACTIONS(1226),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1228),
    [anon_sym___declspec] = ACTIONS(1226),
    [anon_sym___cdecl] = ACTIONS(1226),
    [anon_sym___clrcall] = ACTIONS(1226),
    [anon_sym___stdcall] = ACTIONS(1226),
    [anon_sym___fastcall] = ACTIONS(1226),
    [anon_sym___thiscall] = ACTIONS(1226),
    [anon_sym___vectorcall] = ACTIONS(1226),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_signed] = ACTIONS(1226),
    [anon_sym_unsigned] = ACTIONS(1226),
    [anon_sym_long] = ACTIONS(1226),
    [anon_sym_short] = ACTIONS(1226),
    [anon_sym_static] = ACTIONS(1226),
    [anon_sym_auto] = ACTIONS(1226),
    [anon_sym_register] = ACTIONS(1226),
    [anon_sym_inline] = ACTIONS(1226),
    [anon_sym___inline] = ACTIONS(1226),
    [anon_sym___inline__] = ACTIONS(1226),
    [anon_sym___forceinline] = ACTIONS(1226),
    [anon_sym_thread_local] = ACTIONS(1226),
    [anon_sym___thread] = ACTIONS(1226),
    [anon_sym_const] = ACTIONS(1226),
    [anon_sym_constexpr] = ACTIONS(1226),
    [anon_sym_volatile] = ACTIONS(1226),
    [anon_sym_restrict] = ACTIONS(1226),
    [anon_sym___restrict__] = ACTIONS(1226),
    [anon_sym__Atomic] = ACTIONS(1226),
    [anon_sym__Noreturn] = ACTIONS(1226),
    [anon_sym_noreturn] = ACTIONS(1226),
    [anon_sym__Nonnull] = ACTIONS(1226),
    [anon_sym_alignas] = ACTIONS(1226),
    [anon_sym__Alignas] = ACTIONS(1226),
    [sym_primitive_type] = ACTIONS(1226),
    [anon_sym_enum] = ACTIONS(1226),
    [anon_sym_struct] = ACTIONS(1226),
    [anon_sym_union] = ACTIONS(1226),
    [anon_sym_if] = ACTIONS(1226),
    [anon_sym_else] = ACTIONS(1226),
    [anon_sym_switch] = ACTIONS(1226),
    [anon_sym_case] = ACTIONS(1226),
    [anon_sym_default] = ACTIONS(1226),
    [anon_sym_while] = ACTIONS(1226),
    [anon_sym_do] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(1226),
    [anon_sym_return] = ACTIONS(1226),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_goto] = ACTIONS(1226),
    [anon_sym___try] = ACTIONS(1226),
    [anon_sym___leave] = ACTIONS(1226),
    [anon_sym_DASH_DASH] = ACTIONS(1228),
    [anon_sym_PLUS_PLUS] = ACTIONS(1228),
    [anon_sym_sizeof] = ACTIONS(1226),
    [anon_sym___alignof__] = ACTIONS(1226),
    [anon_sym___alignof] = ACTIONS(1226),
    [anon_sym__alignof] = ACTIONS(1226),
    [anon_sym_alignof] = ACTIONS(1226),
    [anon_sym__Alignof] = ACTIONS(1226),
    [anon_sym_offsetof] = ACTIONS(1226),
    [anon_sym__Generic] = ACTIONS(1226),
    [anon_sym_asm] = ACTIONS(1226),
    [anon_sym___asm__] = ACTIONS(1226),
    [anon_sym___asm] = ACTIONS(1226),
    [sym_number_literal] = ACTIONS(1228),
    [anon_sym_L_SQUOTE] = ACTIONS(1228),
    [anon_sym_u_SQUOTE] = ACTIONS(1228),
    [anon_sym_U_SQUOTE] = ACTIONS(1228),
    [anon_sym_u8_SQUOTE] = ACTIONS(1228),
    [anon_sym_SQUOTE] = ACTIONS(1228),
    [anon_sym_L_DQUOTE] = ACTIONS(1228),
    [anon_sym_u_DQUOTE] = ACTIONS(1228),
    [anon_sym_U_DQUOTE] = ACTIONS(1228),
    [anon_sym_u8_DQUOTE] = ACTIONS(1228),
    [anon_sym_DQUOTE] = ACTIONS(1228),
    [sym_true] = ACTIONS(1226),
    [sym_false] = ACTIONS(1226),
    [anon_sym_NULL] = ACTIONS(1226),
    [anon_sym_nullptr] = ACTIONS(1226),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(183)] = {
    [sym_identifier] = ACTIONS(1242),
    [aux_sym_preproc_include_token1] = ACTIONS(1242),
    [aux_sym_preproc_def_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token2] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1242),
    [sym_preproc_directive] = ACTIONS(1242),
    [anon_sym_LPAREN2] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_TILDE] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1242),
    [anon_sym_PLUS] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(1244),
    [anon_sym_AMP] = ACTIONS(1244),
    [anon_sym_SEMI] = ACTIONS(1244),
    [anon_sym___extension__] = ACTIONS(1242),
    [anon_sym_typedef] = ACTIONS(1242),
    [anon_sym_extern] = ACTIONS(1242),
    [anon_sym___attribute__] = ACTIONS(1242),
    [anon_sym___attribute] = ACTIONS(1242),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1244),
    [anon_sym___declspec] = ACTIONS(1242),
    [anon_sym___cdecl] = ACTIONS(1242),
    [anon_sym___clrcall] = ACTIONS(1242),
    [anon_sym___stdcall] = ACTIONS(1242),
    [anon_sym___fastcall] = ACTIONS(1242),
    [anon_sym___thiscall] = ACTIONS(1242),
    [anon_sym___vectorcall] = ACTIONS(1242),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_signed] = ACTIONS(1242),
    [anon_sym_unsigned] = ACTIONS(1242),
    [anon_sym_long] = ACTIONS(1242),
    [anon_sym_short] = ACTIONS(1242),
    [anon_sym_static] = ACTIONS(1242),
    [anon_sym_auto] = ACTIONS(1242),
    [anon_sym_register] = ACTIONS(1242),
    [anon_sym_inline] = ACTIONS(1242),
    [anon_sym___inline] = ACTIONS(1242),
    [anon_sym___inline__] = ACTIONS(1242),
    [anon_sym___forceinline] = ACTIONS(1242),
    [anon_sym_thread_local] = ACTIONS(1242),
    [anon_sym___thread] = ACTIONS(1242),
    [anon_sym_const] = ACTIONS(1242),
    [anon_sym_constexpr] = ACTIONS(1242),
    [anon_sym_volatile] = ACTIONS(1242),
    [anon_sym_restrict] = ACTIONS(1242),
    [anon_sym___restrict__] = ACTIONS(1242),
    [anon_sym__Atomic] = ACTIONS(1242),
    [anon_sym__Noreturn] = ACTIONS(1242),
    [anon_sym_noreturn] = ACTIONS(1242),
    [anon_sym__Nonnull] = ACTIONS(1242),
    [anon_sym_alignas] = ACTIONS(1242),
    [anon_sym__Alignas] = ACTIONS(1242),
    [sym_primitive_type] = ACTIONS(1242),
    [anon_sym_enum] = ACTIONS(1242),
    [anon_sym_struct] = ACTIONS(1242),
    [anon_sym_union] = ACTIONS(1242),
    [anon_sym_if] = ACTIONS(1242),
    [anon_sym_else] = ACTIONS(1242),
    [anon_sym_switch] = ACTIONS(1242),
    [anon_sym_case] = ACTIONS(1242),
    [anon_sym_default] = ACTIONS(1242),
    [anon_sym_while] = ACTIONS(1242),
    [anon_sym_do] = ACTIONS(1242),
    [anon_sym_for] = ACTIONS(1242),
    [anon_sym_return] = ACTIONS(1242),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_goto] = ACTIONS(1242),
    [anon_sym___try] = ACTIONS(1242),
    [anon_sym___leave] = ACTIONS(1242),
    [anon_sym_DASH_DASH] = ACTIONS(1244),
    [anon_sym_PLUS_PLUS] = ACTIONS(1244),
    [anon_sym_sizeof] = ACTIONS(1242),
    [anon_sym___alignof__] = ACTIONS(1242),
    [anon_sym___alignof] = ACTIONS(1242),
    [anon_sym__alignof] = ACTIONS(1242),
    [anon_sym_alignof] = ACTIONS(1242),
    [anon_sym__Alignof] = ACTIONS(1242),
    [anon_sym_offsetof] = ACTIONS(1242),
    [anon_sym__Generic] = ACTIONS(1242),
    [anon_sym_asm] = ACTIONS(1242),
    [anon_sym___asm__] = ACTIONS(1242),
    [anon_sym___asm] = ACTIONS(1242),
    [sym_number_literal] = ACTIONS(1244),
    [anon_sym_L_SQUOTE] = ACTIONS(1244),
    [anon_sym_u_SQUOTE] = ACTIONS(1244),
    [anon_sym_U_SQUOTE] = ACTIONS(1244),
    [anon_sym_u8_SQUOTE] = ACTIONS(1244),
    [anon_sym_SQUOTE] = ACTIONS(1244),
    [anon_sym_L_DQUOTE] = ACTIONS(1244),
    [anon_sym_u_DQUOTE] = ACTIONS(1244),
    [anon_sym_U_DQUOTE] = ACTIONS(1244),
    [anon_sym_u8_DQUOTE] = ACTIONS(1244),
    [anon_sym_DQUOTE] = ACTIONS(1244),
    [sym_true] = ACTIONS(1242),
    [sym_false] = ACTIONS(1242),
    [anon_sym_NULL] = ACTIONS(1242),
    [anon_sym_nullptr] = ACTIONS(1242),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(184)] = {
    [sym_identifier] = ACTIONS(1246),
    [aux_sym_preproc_include_token1] = ACTIONS(1246),
    [aux_sym_preproc_def_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token2] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1246),
    [sym_preproc_directive] = ACTIONS(1246),
    [anon_sym_LPAREN2] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_TILDE] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1246),
    [anon_sym_PLUS] = ACTIONS(1246),
    [anon_sym_STAR] = ACTIONS(1248),
    [anon_sym_AMP] = ACTIONS(1248),
    [anon_sym_SEMI] = ACTIONS(1248),
    [anon_sym___extension__] = ACTIONS(1246),
    [anon_sym_typedef] = ACTIONS(1246),
    [anon_sym_extern] = ACTIONS(1246),
    [anon_sym___attribute__] = ACTIONS(1246),
    [anon_sym___attribute] = ACTIONS(1246),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1248),
    [anon_sym___declspec] = ACTIONS(1246),
    [anon_sym___cdecl] = ACTIONS(1246),
    [anon_sym___clrcall] = ACTIONS(1246),
    [anon_sym___stdcall] = ACTIONS(1246),
    [anon_sym___fastcall] = ACTIONS(1246),
    [anon_sym___thiscall] = ACTIONS(1246),
    [anon_sym___vectorcall] = ACTIONS(1246),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_signed] = ACTIONS(1246),
    [anon_sym_unsigned] = ACTIONS(1246),
    [anon_sym_long] = ACTIONS(1246),
    [anon_sym_short] = ACTIONS(1246),
    [anon_sym_static] = ACTIONS(1246),
    [anon_sym_auto] = ACTIONS(1246),
    [anon_sym_register] = ACTIONS(1246),
    [anon_sym_inline] = ACTIONS(1246),
    [anon_sym___inline] = ACTIONS(1246),
    [anon_sym___inline__] = ACTIONS(1246),
    [anon_sym___forceinline] = ACTIONS(1246),
    [anon_sym_thread_local] = ACTIONS(1246),
    [anon_sym___thread] = ACTIONS(1246),
    [anon_sym_const] = ACTIONS(1246),
    [anon_sym_constexpr] = ACTIONS(1246),
    [anon_sym_volatile] = ACTIONS(1246),
    [anon_sym_restrict] = ACTIONS(1246),
    [anon_sym___restrict__] = ACTIONS(1246),
    [anon_sym__Atomic] = ACTIONS(1246),
    [anon_sym__Noreturn] = ACTIONS(1246),
    [anon_sym_noreturn] = ACTIONS(1246),
    [anon_sym__Nonnull] = ACTIONS(1246),
    [anon_sym_alignas] = ACTIONS(1246),
    [anon_sym__Alignas] = ACTIONS(1246),
    [sym_primitive_type] = ACTIONS(1246),
    [anon_sym_enum] = ACTIONS(1246),
    [anon_sym_struct] = ACTIONS(1246),
    [anon_sym_union] = ACTIONS(1246),
    [anon_sym_if] = ACTIONS(1246),
    [anon_sym_else] = ACTIONS(1246),
    [anon_sym_switch] = ACTIONS(1246),
    [anon_sym_case] = ACTIONS(1246),
    [anon_sym_default] = ACTIONS(1246),
    [anon_sym_while] = ACTIONS(1246),
    [anon_sym_do] = ACTIONS(1246),
    [anon_sym_for] = ACTIONS(1246),
    [anon_sym_return] = ACTIONS(1246),
    [anon_sym_break] = ACTIONS(1246),
    [anon_sym_continue] = ACTIONS(1246),
    [anon_sym_goto] = ACTIONS(1246),
    [anon_sym___try] = ACTIONS(1246),
    [anon_sym___leave] = ACTIONS(1246),
    [anon_sym_DASH_DASH] = ACTIONS(1248),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [anon_sym_sizeof] = ACTIONS(1246),
    [anon_sym___alignof__] = ACTIONS(1246),
    [anon_sym___alignof] = ACTIONS(1246),
    [anon_sym__alignof] = ACTIONS(1246),
    [anon_sym_alignof] = ACTIONS(1246),
    [anon_sym__Alignof] = ACTIONS(1246),
    [anon_sym_offsetof] = ACTIONS(1246),
    [anon_sym__Generic] = ACTIONS(1246),
    [anon_sym_asm] = ACTIONS(1246),
    [anon_sym___asm__] = ACTIONS(1246),
    [anon_sym___asm] = ACTIONS(1246),
    [sym_number_literal] = ACTIONS(1248),
    [anon_sym_L_SQUOTE] = ACTIONS(1248),
    [anon_sym_u_SQUOTE] = ACTIONS(1248),
    [anon_sym_U_SQUOTE] = ACTIONS(1248),
    [anon_sym_u8_SQUOTE] = ACTIONS(1248),
    [anon_sym_SQUOTE] = ACTIONS(1248),
    [anon_sym_L_DQUOTE] = ACTIONS(1248),
    [anon_sym_u_DQUOTE] = ACTIONS(1248),
    [anon_sym_U_DQUOTE] = ACTIONS(1248),
    [anon_sym_u8_DQUOTE] = ACTIONS(1248),
    [anon_sym_DQUOTE] = ACTIONS(1248),
    [sym_true] = ACTIONS(1246),
    [sym_false] = ACTIONS(1246),
    [anon_sym_NULL] = ACTIONS(1246),
    [anon_sym_nullptr] = ACTIONS(1246),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(185)] = {
    [sym_identifier] = ACTIONS(1254),
    [aux_sym_preproc_include_token1] = ACTIONS(1254),
    [aux_sym_preproc_def_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token2] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1254),
    [sym_preproc_directive] = ACTIONS(1254),
    [anon_sym_LPAREN2] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_TILDE] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_STAR] = ACTIONS(1256),
    [anon_sym_AMP] = ACTIONS(1256),
    [anon_sym_SEMI] = ACTIONS(1256),
    [anon_sym___extension__] = ACTIONS(1254),
    [anon_sym_typedef] = ACTIONS(1254),
    [anon_sym_extern] = ACTIONS(1254),
    [anon_sym___attribute__] = ACTIONS(1254),
    [anon_sym___attribute] = ACTIONS(1254),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1256),
    [anon_sym___declspec] = ACTIONS(1254),
    [anon_sym___cdecl] = ACTIONS(1254),
    [anon_sym___clrcall] = ACTIONS(1254),
    [anon_sym___stdcall] = ACTIONS(1254),
    [anon_sym___fastcall] = ACTIONS(1254),
    [anon_sym___thiscall] = ACTIONS(1254),
    [anon_sym___vectorcall] = ACTIONS(1254),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_signed] = ACTIONS(1254),
    [anon_sym_unsigned] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_auto] = ACTIONS(1254),
    [anon_sym_register] = ACTIONS(1254),
    [anon_sym_inline] = ACTIONS(1254),
    [anon_sym___inline] = ACTIONS(1254),
    [anon_sym___inline__] = ACTIONS(1254),
    [anon_sym___forceinline] = ACTIONS(1254),
    [anon_sym_thread_local] = ACTIONS(1254),
    [anon_sym___thread] = ACTIONS(1254),
    [anon_sym_const] = ACTIONS(1254),
    [anon_sym_constexpr] = ACTIONS(1254),
    [anon_sym_volatile] = ACTIONS(1254),
    [anon_sym_restrict] = ACTIONS(1254),
    [anon_sym___restrict__] = ACTIONS(1254),
    [anon_sym__Atomic] = ACTIONS(1254),
    [anon_sym__Noreturn] = ACTIONS(1254),
    [anon_sym_noreturn] = ACTIONS(1254),
    [anon_sym__Nonnull] = ACTIONS(1254),
    [anon_sym_alignas] = ACTIONS(1254),
    [anon_sym__Alignas] = ACTIONS(1254),
    [sym_primitive_type] = ACTIONS(1254),
    [anon_sym_enum] = ACTIONS(1254),
    [anon_sym_struct] = ACTIONS(1254),
    [anon_sym_union] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_goto] = ACTIONS(1254),
    [anon_sym___try] = ACTIONS(1254),
    [anon_sym___leave] = ACTIONS(1254),
    [anon_sym_DASH_DASH] = ACTIONS(1256),
    [anon_sym_PLUS_PLUS] = ACTIONS(1256),
    [anon_sym_sizeof] = ACTIONS(1254),
    [anon_sym___alignof__] = ACTIONS(1254),
    [anon_sym___alignof] = ACTIONS(1254),
    [anon_sym__alignof] = ACTIONS(1254),
    [anon_sym_alignof] = ACTIONS(1254),
    [anon_sym__Alignof] = ACTIONS(1254),
    [anon_sym_offsetof] = ACTIONS(1254),
    [anon_sym__Generic] = ACTIONS(1254),
    [anon_sym_asm] = ACTIONS(1254),
    [anon_sym___asm__] = ACTIONS(1254),
    [anon_sym___asm] = ACTIONS(1254),
    [sym_number_literal] = ACTIONS(1256),
    [anon_sym_L_SQUOTE] = ACTIONS(1256),
    [anon_sym_u_SQUOTE] = ACTIONS(1256),
    [anon_sym_U_SQUOTE] = ACTIONS(1256),
    [anon_sym_u8_SQUOTE] = ACTIONS(1256),
    [anon_sym_SQUOTE] = ACTIONS(1256),
    [anon_sym_L_DQUOTE] = ACTIONS(1256),
    [anon_sym_u_DQUOTE] = ACTIONS(1256),
    [anon_sym_U_DQUOTE] = ACTIONS(1256),
    [anon_sym_u8_DQUOTE] = ACTIONS(1256),
    [anon_sym_DQUOTE] = ACTIONS(1256),
    [sym_true] = ACTIONS(1254),
    [sym_false] = ACTIONS(1254),
    [anon_sym_NULL] = ACTIONS(1254),
    [anon_sym_nullptr] = ACTIONS(1254),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(186)] = {
    [sym_identifier] = ACTIONS(1138),
    [aux_sym_preproc_include_token1] = ACTIONS(1138),
    [aux_sym_preproc_def_token1] = ACTIONS(1138),
    [aux_sym_preproc_if_token1] = ACTIONS(1138),
    [aux_sym_preproc_if_token2] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1138),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1138),
    [sym_preproc_directive] = ACTIONS(1138),
    [anon_sym_LPAREN2] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_TILDE] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1138),
    [anon_sym_PLUS] = ACTIONS(1138),
    [anon_sym_STAR] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(1140),
    [anon_sym_SEMI] = ACTIONS(1140),
    [anon_sym___extension__] = ACTIONS(1138),
    [anon_sym_typedef] = ACTIONS(1138),
    [anon_sym_extern] = ACTIONS(1138),
    [anon_sym___attribute__] = ACTIONS(1138),
    [anon_sym___attribute] = ACTIONS(1138),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1140),
    [anon_sym___declspec] = ACTIONS(1138),
    [anon_sym___cdecl] = ACTIONS(1138),
    [anon_sym___clrcall] = ACTIONS(1138),
    [anon_sym___stdcall] = ACTIONS(1138),
    [anon_sym___fastcall] = ACTIONS(1138),
    [anon_sym___thiscall] = ACTIONS(1138),
    [anon_sym___vectorcall] = ACTIONS(1138),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_signed] = ACTIONS(1138),
    [anon_sym_unsigned] = ACTIONS(1138),
    [anon_sym_long] = ACTIONS(1138),
    [anon_sym_short] = ACTIONS(1138),
    [anon_sym_static] = ACTIONS(1138),
    [anon_sym_auto] = ACTIONS(1138),
    [anon_sym_register] = ACTIONS(1138),
    [anon_sym_inline] = ACTIONS(1138),
    [anon_sym___inline] = ACTIONS(1138),
    [anon_sym___inline__] = ACTIONS(1138),
    [anon_sym___forceinline] = ACTIONS(1138),
    [anon_sym_thread_local] = ACTIONS(1138),
    [anon_sym___thread] = ACTIONS(1138),
    [anon_sym_const] = ACTIONS(1138),
    [anon_sym_constexpr] = ACTIONS(1138),
    [anon_sym_volatile] = ACTIONS(1138),
    [anon_sym_restrict] = ACTIONS(1138),
    [anon_sym___restrict__] = ACTIONS(1138),
    [anon_sym__Atomic] = ACTIONS(1138),
    [anon_sym__Noreturn] = ACTIONS(1138),
    [anon_sym_noreturn] = ACTIONS(1138),
    [anon_sym__Nonnull] = ACTIONS(1138),
    [anon_sym_alignas] = ACTIONS(1138),
    [anon_sym__Alignas] = ACTIONS(1138),
    [sym_primitive_type] = ACTIONS(1138),
    [anon_sym_enum] = ACTIONS(1138),
    [anon_sym_struct] = ACTIONS(1138),
    [anon_sym_union] = ACTIONS(1138),
    [anon_sym_if] = ACTIONS(1138),
    [anon_sym_else] = ACTIONS(1138),
    [anon_sym_switch] = ACTIONS(1138),
    [anon_sym_case] = ACTIONS(1138),
    [anon_sym_default] = ACTIONS(1138),
    [anon_sym_while] = ACTIONS(1138),
    [anon_sym_do] = ACTIONS(1138),
    [anon_sym_for] = ACTIONS(1138),
    [anon_sym_return] = ACTIONS(1138),
    [anon_sym_break] = ACTIONS(1138),
    [anon_sym_continue] = ACTIONS(1138),
    [anon_sym_goto] = ACTIONS(1138),
    [anon_sym___try] = ACTIONS(1138),
    [anon_sym___leave] = ACTIONS(1138),
    [anon_sym_DASH_DASH] = ACTIONS(1140),
    [anon_sym_PLUS_PLUS] = ACTIONS(1140),
    [anon_sym_sizeof] = ACTIONS(1138),
    [anon_sym___alignof__] = ACTIONS(1138),
    [anon_sym___alignof] = ACTIONS(1138),
    [anon_sym__alignof] = ACTIONS(1138),
    [anon_sym_alignof] = ACTIONS(1138),
    [anon_sym__Alignof] = ACTIONS(1138),
    [anon_sym_offsetof] = ACTIONS(1138),
    [anon_sym__Generic] = ACTIONS(1138),
    [anon_sym_asm] = ACTIONS(1138),
    [anon_sym___asm__] = ACTIONS(1138),
    [anon_sym___asm] = ACTIONS(1138),
    [sym_number_literal] = ACTIONS(1140),
    [anon_sym_L_SQUOTE] = ACTIONS(1140),
    [anon_sym_u_SQUOTE] = ACTIONS(1140),
    [anon_sym_U_SQUOTE] = ACTIONS(1140),
    [anon_sym_u8_SQUOTE] = ACTIONS(1140),
    [anon_sym_SQUOTE] = ACTIONS(1140),
    [anon_sym_L_DQUOTE] = ACTIONS(1140),
    [anon_sym_u_DQUOTE] = ACTIONS(1140),
    [anon_sym_U_DQUOTE] = ACTIONS(1140),
    [anon_sym_u8_DQUOTE] = ACTIONS(1140),
    [anon_sym_DQUOTE] = ACTIONS(1140),
    [sym_true] = ACTIONS(1138),
    [sym_false] = ACTIONS(1138),
    [anon_sym_NULL] = ACTIONS(1138),
    [anon_sym_nullptr] = ACTIONS(1138),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(187)] = {
    [ts_builtin_sym_end] = ACTIONS(1180),
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(188)] = {
    [sym_identifier] = ACTIONS(1154),
    [aux_sym_preproc_include_token1] = ACTIONS(1154),
    [aux_sym_preproc_def_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token1] = ACTIONS(1154),
    [aux_sym_preproc_if_token2] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1154),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1154),
    [sym_preproc_directive] = ACTIONS(1154),
    [anon_sym_LPAREN2] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_TILDE] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1156),
    [anon_sym_AMP] = ACTIONS(1156),
    [anon_sym_SEMI] = ACTIONS(1156),
    [anon_sym___extension__] = ACTIONS(1154),
    [anon_sym_typedef] = ACTIONS(1154),
    [anon_sym_extern] = ACTIONS(1154),
    [anon_sym___attribute__] = ACTIONS(1154),
    [anon_sym___attribute] = ACTIONS(1154),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1156),
    [anon_sym___declspec] = ACTIONS(1154),
    [anon_sym___cdecl] = ACTIONS(1154),
    [anon_sym___clrcall] = ACTIONS(1154),
    [anon_sym___stdcall] = ACTIONS(1154),
    [anon_sym___fastcall] = ACTIONS(1154),
    [anon_sym___thiscall] = ACTIONS(1154),
    [anon_sym___vectorcall] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_signed] = ACTIONS(1154),
    [anon_sym_unsigned] = ACTIONS(1154),
    [anon_sym_long] = ACTIONS(1154),
    [anon_sym_short] = ACTIONS(1154),
    [anon_sym_static] = ACTIONS(1154),
    [anon_sym_auto] = ACTIONS(1154),
    [anon_sym_register] = ACTIONS(1154),
    [anon_sym_inline] = ACTIONS(1154),
    [anon_sym___inline] = ACTIONS(1154),
    [anon_sym___inline__] = ACTIONS(1154),
    [anon_sym___forceinline] = ACTIONS(1154),
    [anon_sym_thread_local] = ACTIONS(1154),
    [anon_sym___thread] = ACTIONS(1154),
    [anon_sym_const] = ACTIONS(1154),
    [anon_sym_constexpr] = ACTIONS(1154),
    [anon_sym_volatile] = ACTIONS(1154),
    [anon_sym_restrict] = ACTIONS(1154),
    [anon_sym___restrict__] = ACTIONS(1154),
    [anon_sym__Atomic] = ACTIONS(1154),
    [anon_sym__Noreturn] = ACTIONS(1154),
    [anon_sym_noreturn] = ACTIONS(1154),
    [anon_sym__Nonnull] = ACTIONS(1154),
    [anon_sym_alignas] = ACTIONS(1154),
    [anon_sym__Alignas] = ACTIONS(1154),
    [sym_primitive_type] = ACTIONS(1154),
    [anon_sym_enum] = ACTIONS(1154),
    [anon_sym_struct] = ACTIONS(1154),
    [anon_sym_union] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_else] = ACTIONS(1154),
    [anon_sym_switch] = ACTIONS(1154),
    [anon_sym_case] = ACTIONS(1154),
    [anon_sym_default] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_break] = ACTIONS(1154),
    [anon_sym_continue] = ACTIONS(1154),
    [anon_sym_goto] = ACTIONS(1154),
    [anon_sym___try] = ACTIONS(1154),
    [anon_sym___leave] = ACTIONS(1154),
    [anon_sym_DASH_DASH] = ACTIONS(1156),
    [anon_sym_PLUS_PLUS] = ACTIONS(1156),
    [anon_sym_sizeof] = ACTIONS(1154),
    [anon_sym___alignof__] = ACTIONS(1154),
    [anon_sym___alignof] = ACTIONS(1154),
    [anon_sym__alignof] = ACTIONS(1154),
    [anon_sym_alignof] = ACTIONS(1154),
    [anon_sym__Alignof] = ACTIONS(1154),
    [anon_sym_offsetof] = ACTIONS(1154),
    [anon_sym__Generic] = ACTIONS(1154),
    [anon_sym_asm] = ACTIONS(1154),
    [anon_sym___asm__] = ACTIONS(1154),
    [anon_sym___asm] = ACTIONS(1154),
    [sym_number_literal] = ACTIONS(1156),
    [anon_sym_L_SQUOTE] = ACTIONS(1156),
    [anon_sym_u_SQUOTE] = ACTIONS(1156),
    [anon_sym_U_SQUOTE] = ACTIONS(1156),
    [anon_sym_u8_SQUOTE] = ACTIONS(1156),
    [anon_sym_SQUOTE] = ACTIONS(1156),
    [anon_sym_L_DQUOTE] = ACTIONS(1156),
    [anon_sym_u_DQUOTE] = ACTIONS(1156),
    [anon_sym_U_DQUOTE] = ACTIONS(1156),
    [anon_sym_u8_DQUOTE] = ACTIONS(1156),
    [anon_sym_DQUOTE] = ACTIONS(1156),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [anon_sym_NULL] = ACTIONS(1154),
    [anon_sym_nullptr] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(189)] = {
    [ts_builtin_sym_end] = ACTIONS(1192),
    [sym_identifier] = ACTIONS(1190),
    [aux_sym_preproc_include_token1] = ACTIONS(1190),
    [aux_sym_preproc_def_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1190),
    [sym_preproc_directive] = ACTIONS(1190),
    [anon_sym_LPAREN2] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_TILDE] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1192),
    [anon_sym_AMP] = ACTIONS(1192),
    [anon_sym_SEMI] = ACTIONS(1192),
    [anon_sym___extension__] = ACTIONS(1190),
    [anon_sym_typedef] = ACTIONS(1190),
    [anon_sym_extern] = ACTIONS(1190),
    [anon_sym___attribute__] = ACTIONS(1190),
    [anon_sym___attribute] = ACTIONS(1190),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1192),
    [anon_sym___declspec] = ACTIONS(1190),
    [anon_sym___cdecl] = ACTIONS(1190),
    [anon_sym___clrcall] = ACTIONS(1190),
    [anon_sym___stdcall] = ACTIONS(1190),
    [anon_sym___fastcall] = ACTIONS(1190),
    [anon_sym___thiscall] = ACTIONS(1190),
    [anon_sym___vectorcall] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_signed] = ACTIONS(1190),
    [anon_sym_unsigned] = ACTIONS(1190),
    [anon_sym_long] = ACTIONS(1190),
    [anon_sym_short] = ACTIONS(1190),
    [anon_sym_static] = ACTIONS(1190),
    [anon_sym_auto] = ACTIONS(1190),
    [anon_sym_register] = ACTIONS(1190),
    [anon_sym_inline] = ACTIONS(1190),
    [anon_sym___inline] = ACTIONS(1190),
    [anon_sym___inline__] = ACTIONS(1190),
    [anon_sym___forceinline] = ACTIONS(1190),
    [anon_sym_thread_local] = ACTIONS(1190),
    [anon_sym___thread] = ACTIONS(1190),
    [anon_sym_const] = ACTIONS(1190),
    [anon_sym_constexpr] = ACTIONS(1190),
    [anon_sym_volatile] = ACTIONS(1190),
    [anon_sym_restrict] = ACTIONS(1190),
    [anon_sym___restrict__] = ACTIONS(1190),
    [anon_sym__Atomic] = ACTIONS(1190),
    [anon_sym__Noreturn] = ACTIONS(1190),
    [anon_sym_noreturn] = ACTIONS(1190),
    [anon_sym__Nonnull] = ACTIONS(1190),
    [anon_sym_alignas] = ACTIONS(1190),
    [anon_sym__Alignas] = ACTIONS(1190),
    [sym_primitive_type] = ACTIONS(1190),
    [anon_sym_enum] = ACTIONS(1190),
    [anon_sym_struct] = ACTIONS(1190),
    [anon_sym_union] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_else] = ACTIONS(1190),
    [anon_sym_switch] = ACTIONS(1190),
    [anon_sym_case] = ACTIONS(1190),
    [anon_sym_default] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_break] = ACTIONS(1190),
    [anon_sym_continue] = ACTIONS(1190),
    [anon_sym_goto] = ACTIONS(1190),
    [anon_sym___try] = ACTIONS(1190),
    [anon_sym___leave] = ACTIONS(1190),
    [anon_sym_DASH_DASH] = ACTIONS(1192),
    [anon_sym_PLUS_PLUS] = ACTIONS(1192),
    [anon_sym_sizeof] = ACTIONS(1190),
    [anon_sym___alignof__] = ACTIONS(1190),
    [anon_sym___alignof] = ACTIONS(1190),
    [anon_sym__alignof] = ACTIONS(1190),
    [anon_sym_alignof] = ACTIONS(1190),
    [anon_sym__Alignof] = ACTIONS(1190),
    [anon_sym_offsetof] = ACTIONS(1190),
    [anon_sym__Generic] = ACTIONS(1190),
    [anon_sym_asm] = ACTIONS(1190),
    [anon_sym___asm__] = ACTIONS(1190),
    [anon_sym___asm] = ACTIONS(1190),
    [sym_number_literal] = ACTIONS(1192),
    [anon_sym_L_SQUOTE] = ACTIONS(1192),
    [anon_sym_u_SQUOTE] = ACTIONS(1192),
    [anon_sym_U_SQUOTE] = ACTIONS(1192),
    [anon_sym_u8_SQUOTE] = ACTIONS(1192),
    [anon_sym_SQUOTE] = ACTIONS(1192),
    [anon_sym_L_DQUOTE] = ACTIONS(1192),
    [anon_sym_u_DQUOTE] = ACTIONS(1192),
    [anon_sym_U_DQUOTE] = ACTIONS(1192),
    [anon_sym_u8_DQUOTE] = ACTIONS(1192),
    [anon_sym_DQUOTE] = ACTIONS(1192),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [anon_sym_NULL] = ACTIONS(1190),
    [anon_sym_nullptr] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(190)] = {
    [ts_builtin_sym_end] = ACTIONS(1196),
    [sym_identifier] = ACTIONS(1194),
    [aux_sym_preproc_include_token1] = ACTIONS(1194),
    [aux_sym_preproc_def_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1194),
    [sym_preproc_directive] = ACTIONS(1194),
    [anon_sym_LPAREN2] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_TILDE] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1196),
    [anon_sym_AMP] = ACTIONS(1196),
    [anon_sym_SEMI] = ACTIONS(1196),
    [anon_sym___extension__] = ACTIONS(1194),
    [anon_sym_typedef] = ACTIONS(1194),
    [anon_sym_extern] = ACTIONS(1194),
    [anon_sym___attribute__] = ACTIONS(1194),
    [anon_sym___attribute] = ACTIONS(1194),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1196),
    [anon_sym___declspec] = ACTIONS(1194),
    [anon_sym___cdecl] = ACTIONS(1194),
    [anon_sym___clrcall] = ACTIONS(1194),
    [anon_sym___stdcall] = ACTIONS(1194),
    [anon_sym___fastcall] = ACTIONS(1194),
    [anon_sym___thiscall] = ACTIONS(1194),
    [anon_sym___vectorcall] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_signed] = ACTIONS(1194),
    [anon_sym_unsigned] = ACTIONS(1194),
    [anon_sym_long] = ACTIONS(1194),
    [anon_sym_short] = ACTIONS(1194),
    [anon_sym_static] = ACTIONS(1194),
    [anon_sym_auto] = ACTIONS(1194),
    [anon_sym_register] = ACTIONS(1194),
    [anon_sym_inline] = ACTIONS(1194),
    [anon_sym___inline] = ACTIONS(1194),
    [anon_sym___inline__] = ACTIONS(1194),
    [anon_sym___forceinline] = ACTIONS(1194),
    [anon_sym_thread_local] = ACTIONS(1194),
    [anon_sym___thread] = ACTIONS(1194),
    [anon_sym_const] = ACTIONS(1194),
    [anon_sym_constexpr] = ACTIONS(1194),
    [anon_sym_volatile] = ACTIONS(1194),
    [anon_sym_restrict] = ACTIONS(1194),
    [anon_sym___restrict__] = ACTIONS(1194),
    [anon_sym__Atomic] = ACTIONS(1194),
    [anon_sym__Noreturn] = ACTIONS(1194),
    [anon_sym_noreturn] = ACTIONS(1194),
    [anon_sym__Nonnull] = ACTIONS(1194),
    [anon_sym_alignas] = ACTIONS(1194),
    [anon_sym__Alignas] = ACTIONS(1194),
    [sym_primitive_type] = ACTIONS(1194),
    [anon_sym_enum] = ACTIONS(1194),
    [anon_sym_struct] = ACTIONS(1194),
    [anon_sym_union] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_else] = ACTIONS(1194),
    [anon_sym_switch] = ACTIONS(1194),
    [anon_sym_case] = ACTIONS(1194),
    [anon_sym_default] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_break] = ACTIONS(1194),
    [anon_sym_continue] = ACTIONS(1194),
    [anon_sym_goto] = ACTIONS(1194),
    [anon_sym___try] = ACTIONS(1194),
    [anon_sym___leave] = ACTIONS(1194),
    [anon_sym_DASH_DASH] = ACTIONS(1196),
    [anon_sym_PLUS_PLUS] = ACTIONS(1196),
    [anon_sym_sizeof] = ACTIONS(1194),
    [anon_sym___alignof__] = ACTIONS(1194),
    [anon_sym___alignof] = ACTIONS(1194),
    [anon_sym__alignof] = ACTIONS(1194),
    [anon_sym_alignof] = ACTIONS(1194),
    [anon_sym__Alignof] = ACTIONS(1194),
    [anon_sym_offsetof] = ACTIONS(1194),
    [anon_sym__Generic] = ACTIONS(1194),
    [anon_sym_asm] = ACTIONS(1194),
    [anon_sym___asm__] = ACTIONS(1194),
    [anon_sym___asm] = ACTIONS(1194),
    [sym_number_literal] = ACTIONS(1196),
    [anon_sym_L_SQUOTE] = ACTIONS(1196),
    [anon_sym_u_SQUOTE] = ACTIONS(1196),
    [anon_sym_U_SQUOTE] = ACTIONS(1196),
    [anon_sym_u8_SQUOTE] = ACTIONS(1196),
    [anon_sym_SQUOTE] = ACTIONS(1196),
    [anon_sym_L_DQUOTE] = ACTIONS(1196),
    [anon_sym_u_DQUOTE] = ACTIONS(1196),
    [anon_sym_U_DQUOTE] = ACTIONS(1196),
    [anon_sym_u8_DQUOTE] = ACTIONS(1196),
    [anon_sym_DQUOTE] = ACTIONS(1196),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [anon_sym_NULL] = ACTIONS(1194),
    [anon_sym_nullptr] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(191)] = {
    [ts_builtin_sym_end] = ACTIONS(1216),
    [sym_identifier] = ACTIONS(1214),
    [aux_sym_preproc_include_token1] = ACTIONS(1214),
    [aux_sym_preproc_def_token1] = ACTIONS(1214),
    [aux_sym_preproc_if_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1214),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1214),
    [sym_preproc_directive] = ACTIONS(1214),
    [anon_sym_LPAREN2] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_TILDE] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1214),
    [anon_sym_PLUS] = ACTIONS(1214),
    [anon_sym_STAR] = ACTIONS(1216),
    [anon_sym_AMP] = ACTIONS(1216),
    [anon_sym_SEMI] = ACTIONS(1216),
    [anon_sym___extension__] = ACTIONS(1214),
    [anon_sym_typedef] = ACTIONS(1214),
    [anon_sym_extern] = ACTIONS(1214),
    [anon_sym___attribute__] = ACTIONS(1214),
    [anon_sym___attribute] = ACTIONS(1214),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1216),
    [anon_sym___declspec] = ACTIONS(1214),
    [anon_sym___cdecl] = ACTIONS(1214),
    [anon_sym___clrcall] = ACTIONS(1214),
    [anon_sym___stdcall] = ACTIONS(1214),
    [anon_sym___fastcall] = ACTIONS(1214),
    [anon_sym___thiscall] = ACTIONS(1214),
    [anon_sym___vectorcall] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_signed] = ACTIONS(1214),
    [anon_sym_unsigned] = ACTIONS(1214),
    [anon_sym_long] = ACTIONS(1214),
    [anon_sym_short] = ACTIONS(1214),
    [anon_sym_static] = ACTIONS(1214),
    [anon_sym_auto] = ACTIONS(1214),
    [anon_sym_register] = ACTIONS(1214),
    [anon_sym_inline] = ACTIONS(1214),
    [anon_sym___inline] = ACTIONS(1214),
    [anon_sym___inline__] = ACTIONS(1214),
    [anon_sym___forceinline] = ACTIONS(1214),
    [anon_sym_thread_local] = ACTIONS(1214),
    [anon_sym___thread] = ACTIONS(1214),
    [anon_sym_const] = ACTIONS(1214),
    [anon_sym_constexpr] = ACTIONS(1214),
    [anon_sym_volatile] = ACTIONS(1214),
    [anon_sym_restrict] = ACTIONS(1214),
    [anon_sym___restrict__] = ACTIONS(1214),
    [anon_sym__Atomic] = ACTIONS(1214),
    [anon_sym__Noreturn] = ACTIONS(1214),
    [anon_sym_noreturn] = ACTIONS(1214),
    [anon_sym__Nonnull] = ACTIONS(1214),
    [anon_sym_alignas] = ACTIONS(1214),
    [anon_sym__Alignas] = ACTIONS(1214),
    [sym_primitive_type] = ACTIONS(1214),
    [anon_sym_enum] = ACTIONS(1214),
    [anon_sym_struct] = ACTIONS(1214),
    [anon_sym_union] = ACTIONS(1214),
    [anon_sym_if] = ACTIONS(1214),
    [anon_sym_else] = ACTIONS(1214),
    [anon_sym_switch] = ACTIONS(1214),
    [anon_sym_case] = ACTIONS(1214),
    [anon_sym_default] = ACTIONS(1214),
    [anon_sym_while] = ACTIONS(1214),
    [anon_sym_do] = ACTIONS(1214),
    [anon_sym_for] = ACTIONS(1214),
    [anon_sym_return] = ACTIONS(1214),
    [anon_sym_break] = ACTIONS(1214),
    [anon_sym_continue] = ACTIONS(1214),
    [anon_sym_goto] = ACTIONS(1214),
    [anon_sym___try] = ACTIONS(1214),
    [anon_sym___leave] = ACTIONS(1214),
    [anon_sym_DASH_DASH] = ACTIONS(1216),
    [anon_sym_PLUS_PLUS] = ACTIONS(1216),
    [anon_sym_sizeof] = ACTIONS(1214),
    [anon_sym___alignof__] = ACTIONS(1214),
    [anon_sym___alignof] = ACTIONS(1214),
    [anon_sym__alignof] = ACTIONS(1214),
    [anon_sym_alignof] = ACTIONS(1214),
    [anon_sym__Alignof] = ACTIONS(1214),
    [anon_sym_offsetof] = ACTIONS(1214),
    [anon_sym__Generic] = ACTIONS(1214),
    [anon_sym_asm] = ACTIONS(1214),
    [anon_sym___asm__] = ACTIONS(1214),
    [anon_sym___asm] = ACTIONS(1214),
    [sym_number_literal] = ACTIONS(1216),
    [anon_sym_L_SQUOTE] = ACTIONS(1216),
    [anon_sym_u_SQUOTE] = ACTIONS(1216),
    [anon_sym_U_SQUOTE] = ACTIONS(1216),
    [anon_sym_u8_SQUOTE] = ACTIONS(1216),
    [anon_sym_SQUOTE] = ACTIONS(1216),
    [anon_sym_L_DQUOTE] = ACTIONS(1216),
    [anon_sym_u_DQUOTE] = ACTIONS(1216),
    [anon_sym_U_DQUOTE] = ACTIONS(1216),
    [anon_sym_u8_DQUOTE] = ACTIONS(1216),
    [anon_sym_DQUOTE] = ACTIONS(1216),
    [sym_true] = ACTIONS(1214),
    [sym_false] = ACTIONS(1214),
    [anon_sym_NULL] = ACTIONS(1214),
    [anon_sym_nullptr] = ACTIONS(1214),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(192)] = {
    [sym_identifier] = ACTIONS(1258),
    [aux_sym_preproc_include_token1] = ACTIONS(1258),
    [aux_sym_preproc_def_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token2] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1258),
    [sym_preproc_directive] = ACTIONS(1258),
    [anon_sym_LPAREN2] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1258),
    [anon_sym_PLUS] = ACTIONS(1258),
    [anon_sym_STAR] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym___extension__] = ACTIONS(1258),
    [anon_sym_typedef] = ACTIONS(1258),
    [anon_sym_extern] = ACTIONS(1258),
    [anon_sym___attribute__] = ACTIONS(1258),
    [anon_sym___attribute] = ACTIONS(1258),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1260),
    [anon_sym___declspec] = ACTIONS(1258),
    [anon_sym___cdecl] = ACTIONS(1258),
    [anon_sym___clrcall] = ACTIONS(1258),
    [anon_sym___stdcall] = ACTIONS(1258),
    [anon_sym___fastcall] = ACTIONS(1258),
    [anon_sym___thiscall] = ACTIONS(1258),
    [anon_sym___vectorcall] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_signed] = ACTIONS(1258),
    [anon_sym_unsigned] = ACTIONS(1258),
    [anon_sym_long] = ACTIONS(1258),
    [anon_sym_short] = ACTIONS(1258),
    [anon_sym_static] = ACTIONS(1258),
    [anon_sym_auto] = ACTIONS(1258),
    [anon_sym_register] = ACTIONS(1258),
    [anon_sym_inline] = ACTIONS(1258),
    [anon_sym___inline] = ACTIONS(1258),
    [anon_sym___inline__] = ACTIONS(1258),
    [anon_sym___forceinline] = ACTIONS(1258),
    [anon_sym_thread_local] = ACTIONS(1258),
    [anon_sym___thread] = ACTIONS(1258),
    [anon_sym_const] = ACTIONS(1258),
    [anon_sym_constexpr] = ACTIONS(1258),
    [anon_sym_volatile] = ACTIONS(1258),
    [anon_sym_restrict] = ACTIONS(1258),
    [anon_sym___restrict__] = ACTIONS(1258),
    [anon_sym__Atomic] = ACTIONS(1258),
    [anon_sym__Noreturn] = ACTIONS(1258),
    [anon_sym_noreturn] = ACTIONS(1258),
    [anon_sym__Nonnull] = ACTIONS(1258),
    [anon_sym_alignas] = ACTIONS(1258),
    [anon_sym__Alignas] = ACTIONS(1258),
    [sym_primitive_type] = ACTIONS(1258),
    [anon_sym_enum] = ACTIONS(1258),
    [anon_sym_struct] = ACTIONS(1258),
    [anon_sym_union] = ACTIONS(1258),
    [anon_sym_if] = ACTIONS(1258),
    [anon_sym_else] = ACTIONS(1258),
    [anon_sym_switch] = ACTIONS(1258),
    [anon_sym_case] = ACTIONS(1258),
    [anon_sym_default] = ACTIONS(1258),
    [anon_sym_while] = ACTIONS(1258),
    [anon_sym_do] = ACTIONS(1258),
    [anon_sym_for] = ACTIONS(1258),
    [anon_sym_return] = ACTIONS(1258),
    [anon_sym_break] = ACTIONS(1258),
    [anon_sym_continue] = ACTIONS(1258),
    [anon_sym_goto] = ACTIONS(1258),
    [anon_sym___try] = ACTIONS(1258),
    [anon_sym___leave] = ACTIONS(1258),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_sizeof] = ACTIONS(1258),
    [anon_sym___alignof__] = ACTIONS(1258),
    [anon_sym___alignof] = ACTIONS(1258),
    [anon_sym__alignof] = ACTIONS(1258),
    [anon_sym_alignof] = ACTIONS(1258),
    [anon_sym__Alignof] = ACTIONS(1258),
    [anon_sym_offsetof] = ACTIONS(1258),
    [anon_sym__Generic] = ACTIONS(1258),
    [anon_sym_asm] = ACTIONS(1258),
    [anon_sym___asm__] = ACTIONS(1258),
    [anon_sym___asm] = ACTIONS(1258),
    [sym_number_literal] = ACTIONS(1260),
    [anon_sym_L_SQUOTE] = ACTIONS(1260),
    [anon_sym_u_SQUOTE] = ACTIONS(1260),
    [anon_sym_U_SQUOTE] = ACTIONS(1260),
    [anon_sym_u8_SQUOTE] = ACTIONS(1260),
    [anon_sym_SQUOTE] = ACTIONS(1260),
    [anon_sym_L_DQUOTE] = ACTIONS(1260),
    [anon_sym_u_DQUOTE] = ACTIONS(1260),
    [anon_sym_U_DQUOTE] = ACTIONS(1260),
    [anon_sym_u8_DQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1260),
    [sym_true] = ACTIONS(1258),
    [sym_false] = ACTIONS(1258),
    [anon_sym_NULL] = ACTIONS(1258),
    [anon_sym_nullptr] = ACTIONS(1258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(193)] = {
    [sym_identifier] = ACTIONS(1190),
    [aux_sym_preproc_include_token1] = ACTIONS(1190),
    [aux_sym_preproc_def_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token2] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1190),
    [sym_preproc_directive] = ACTIONS(1190),
    [anon_sym_LPAREN2] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_TILDE] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1192),
    [anon_sym_AMP] = ACTIONS(1192),
    [anon_sym_SEMI] = ACTIONS(1192),
    [anon_sym___extension__] = ACTIONS(1190),
    [anon_sym_typedef] = ACTIONS(1190),
    [anon_sym_extern] = ACTIONS(1190),
    [anon_sym___attribute__] = ACTIONS(1190),
    [anon_sym___attribute] = ACTIONS(1190),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1192),
    [anon_sym___declspec] = ACTIONS(1190),
    [anon_sym___cdecl] = ACTIONS(1190),
    [anon_sym___clrcall] = ACTIONS(1190),
    [anon_sym___stdcall] = ACTIONS(1190),
    [anon_sym___fastcall] = ACTIONS(1190),
    [anon_sym___thiscall] = ACTIONS(1190),
    [anon_sym___vectorcall] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_signed] = ACTIONS(1190),
    [anon_sym_unsigned] = ACTIONS(1190),
    [anon_sym_long] = ACTIONS(1190),
    [anon_sym_short] = ACTIONS(1190),
    [anon_sym_static] = ACTIONS(1190),
    [anon_sym_auto] = ACTIONS(1190),
    [anon_sym_register] = ACTIONS(1190),
    [anon_sym_inline] = ACTIONS(1190),
    [anon_sym___inline] = ACTIONS(1190),
    [anon_sym___inline__] = ACTIONS(1190),
    [anon_sym___forceinline] = ACTIONS(1190),
    [anon_sym_thread_local] = ACTIONS(1190),
    [anon_sym___thread] = ACTIONS(1190),
    [anon_sym_const] = ACTIONS(1190),
    [anon_sym_constexpr] = ACTIONS(1190),
    [anon_sym_volatile] = ACTIONS(1190),
    [anon_sym_restrict] = ACTIONS(1190),
    [anon_sym___restrict__] = ACTIONS(1190),
    [anon_sym__Atomic] = ACTIONS(1190),
    [anon_sym__Noreturn] = ACTIONS(1190),
    [anon_sym_noreturn] = ACTIONS(1190),
    [anon_sym__Nonnull] = ACTIONS(1190),
    [anon_sym_alignas] = ACTIONS(1190),
    [anon_sym__Alignas] = ACTIONS(1190),
    [sym_primitive_type] = ACTIONS(1190),
    [anon_sym_enum] = ACTIONS(1190),
    [anon_sym_struct] = ACTIONS(1190),
    [anon_sym_union] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_else] = ACTIONS(1190),
    [anon_sym_switch] = ACTIONS(1190),
    [anon_sym_case] = ACTIONS(1190),
    [anon_sym_default] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_break] = ACTIONS(1190),
    [anon_sym_continue] = ACTIONS(1190),
    [anon_sym_goto] = ACTIONS(1190),
    [anon_sym___try] = ACTIONS(1190),
    [anon_sym___leave] = ACTIONS(1190),
    [anon_sym_DASH_DASH] = ACTIONS(1192),
    [anon_sym_PLUS_PLUS] = ACTIONS(1192),
    [anon_sym_sizeof] = ACTIONS(1190),
    [anon_sym___alignof__] = ACTIONS(1190),
    [anon_sym___alignof] = ACTIONS(1190),
    [anon_sym__alignof] = ACTIONS(1190),
    [anon_sym_alignof] = ACTIONS(1190),
    [anon_sym__Alignof] = ACTIONS(1190),
    [anon_sym_offsetof] = ACTIONS(1190),
    [anon_sym__Generic] = ACTIONS(1190),
    [anon_sym_asm] = ACTIONS(1190),
    [anon_sym___asm__] = ACTIONS(1190),
    [anon_sym___asm] = ACTIONS(1190),
    [sym_number_literal] = ACTIONS(1192),
    [anon_sym_L_SQUOTE] = ACTIONS(1192),
    [anon_sym_u_SQUOTE] = ACTIONS(1192),
    [anon_sym_U_SQUOTE] = ACTIONS(1192),
    [anon_sym_u8_SQUOTE] = ACTIONS(1192),
    [anon_sym_SQUOTE] = ACTIONS(1192),
    [anon_sym_L_DQUOTE] = ACTIONS(1192),
    [anon_sym_u_DQUOTE] = ACTIONS(1192),
    [anon_sym_U_DQUOTE] = ACTIONS(1192),
    [anon_sym_u8_DQUOTE] = ACTIONS(1192),
    [anon_sym_DQUOTE] = ACTIONS(1192),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [anon_sym_NULL] = ACTIONS(1190),
    [anon_sym_nullptr] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(194)] = {
    [sym_identifier] = ACTIONS(1194),
    [aux_sym_preproc_include_token1] = ACTIONS(1194),
    [aux_sym_preproc_def_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token2] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1194),
    [sym_preproc_directive] = ACTIONS(1194),
    [anon_sym_LPAREN2] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_TILDE] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1196),
    [anon_sym_AMP] = ACTIONS(1196),
    [anon_sym_SEMI] = ACTIONS(1196),
    [anon_sym___extension__] = ACTIONS(1194),
    [anon_sym_typedef] = ACTIONS(1194),
    [anon_sym_extern] = ACTIONS(1194),
    [anon_sym___attribute__] = ACTIONS(1194),
    [anon_sym___attribute] = ACTIONS(1194),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1196),
    [anon_sym___declspec] = ACTIONS(1194),
    [anon_sym___cdecl] = ACTIONS(1194),
    [anon_sym___clrcall] = ACTIONS(1194),
    [anon_sym___stdcall] = ACTIONS(1194),
    [anon_sym___fastcall] = ACTIONS(1194),
    [anon_sym___thiscall] = ACTIONS(1194),
    [anon_sym___vectorcall] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_signed] = ACTIONS(1194),
    [anon_sym_unsigned] = ACTIONS(1194),
    [anon_sym_long] = ACTIONS(1194),
    [anon_sym_short] = ACTIONS(1194),
    [anon_sym_static] = ACTIONS(1194),
    [anon_sym_auto] = ACTIONS(1194),
    [anon_sym_register] = ACTIONS(1194),
    [anon_sym_inline] = ACTIONS(1194),
    [anon_sym___inline] = ACTIONS(1194),
    [anon_sym___inline__] = ACTIONS(1194),
    [anon_sym___forceinline] = ACTIONS(1194),
    [anon_sym_thread_local] = ACTIONS(1194),
    [anon_sym___thread] = ACTIONS(1194),
    [anon_sym_const] = ACTIONS(1194),
    [anon_sym_constexpr] = ACTIONS(1194),
    [anon_sym_volatile] = ACTIONS(1194),
    [anon_sym_restrict] = ACTIONS(1194),
    [anon_sym___restrict__] = ACTIONS(1194),
    [anon_sym__Atomic] = ACTIONS(1194),
    [anon_sym__Noreturn] = ACTIONS(1194),
    [anon_sym_noreturn] = ACTIONS(1194),
    [anon_sym__Nonnull] = ACTIONS(1194),
    [anon_sym_alignas] = ACTIONS(1194),
    [anon_sym__Alignas] = ACTIONS(1194),
    [sym_primitive_type] = ACTIONS(1194),
    [anon_sym_enum] = ACTIONS(1194),
    [anon_sym_struct] = ACTIONS(1194),
    [anon_sym_union] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_else] = ACTIONS(1194),
    [anon_sym_switch] = ACTIONS(1194),
    [anon_sym_case] = ACTIONS(1194),
    [anon_sym_default] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_break] = ACTIONS(1194),
    [anon_sym_continue] = ACTIONS(1194),
    [anon_sym_goto] = ACTIONS(1194),
    [anon_sym___try] = ACTIONS(1194),
    [anon_sym___leave] = ACTIONS(1194),
    [anon_sym_DASH_DASH] = ACTIONS(1196),
    [anon_sym_PLUS_PLUS] = ACTIONS(1196),
    [anon_sym_sizeof] = ACTIONS(1194),
    [anon_sym___alignof__] = ACTIONS(1194),
    [anon_sym___alignof] = ACTIONS(1194),
    [anon_sym__alignof] = ACTIONS(1194),
    [anon_sym_alignof] = ACTIONS(1194),
    [anon_sym__Alignof] = ACTIONS(1194),
    [anon_sym_offsetof] = ACTIONS(1194),
    [anon_sym__Generic] = ACTIONS(1194),
    [anon_sym_asm] = ACTIONS(1194),
    [anon_sym___asm__] = ACTIONS(1194),
    [anon_sym___asm] = ACTIONS(1194),
    [sym_number_literal] = ACTIONS(1196),
    [anon_sym_L_SQUOTE] = ACTIONS(1196),
    [anon_sym_u_SQUOTE] = ACTIONS(1196),
    [anon_sym_U_SQUOTE] = ACTIONS(1196),
    [anon_sym_u8_SQUOTE] = ACTIONS(1196),
    [anon_sym_SQUOTE] = ACTIONS(1196),
    [anon_sym_L_DQUOTE] = ACTIONS(1196),
    [anon_sym_u_DQUOTE] = ACTIONS(1196),
    [anon_sym_U_DQUOTE] = ACTIONS(1196),
    [anon_sym_u8_DQUOTE] = ACTIONS(1196),
    [anon_sym_DQUOTE] = ACTIONS(1196),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [anon_sym_NULL] = ACTIONS(1194),
    [anon_sym_nullptr] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(195)] = {
    [sym_identifier] = ACTIONS(1210),
    [aux_sym_preproc_include_token1] = ACTIONS(1210),
    [aux_sym_preproc_def_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token2] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1210),
    [sym_preproc_directive] = ACTIONS(1210),
    [anon_sym_LPAREN2] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1210),
    [anon_sym_PLUS] = ACTIONS(1210),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym___extension__] = ACTIONS(1210),
    [anon_sym_typedef] = ACTIONS(1210),
    [anon_sym_extern] = ACTIONS(1210),
    [anon_sym___attribute__] = ACTIONS(1210),
    [anon_sym___attribute] = ACTIONS(1210),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1212),
    [anon_sym___declspec] = ACTIONS(1210),
    [anon_sym___cdecl] = ACTIONS(1210),
    [anon_sym___clrcall] = ACTIONS(1210),
    [anon_sym___stdcall] = ACTIONS(1210),
    [anon_sym___fastcall] = ACTIONS(1210),
    [anon_sym___thiscall] = ACTIONS(1210),
    [anon_sym___vectorcall] = ACTIONS(1210),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_signed] = ACTIONS(1210),
    [anon_sym_unsigned] = ACTIONS(1210),
    [anon_sym_long] = ACTIONS(1210),
    [anon_sym_short] = ACTIONS(1210),
    [anon_sym_static] = ACTIONS(1210),
    [anon_sym_auto] = ACTIONS(1210),
    [anon_sym_register] = ACTIONS(1210),
    [anon_sym_inline] = ACTIONS(1210),
    [anon_sym___inline] = ACTIONS(1210),
    [anon_sym___inline__] = ACTIONS(1210),
    [anon_sym___forceinline] = ACTIONS(1210),
    [anon_sym_thread_local] = ACTIONS(1210),
    [anon_sym___thread] = ACTIONS(1210),
    [anon_sym_const] = ACTIONS(1210),
    [anon_sym_constexpr] = ACTIONS(1210),
    [anon_sym_volatile] = ACTIONS(1210),
    [anon_sym_restrict] = ACTIONS(1210),
    [anon_sym___restrict__] = ACTIONS(1210),
    [anon_sym__Atomic] = ACTIONS(1210),
    [anon_sym__Noreturn] = ACTIONS(1210),
    [anon_sym_noreturn] = ACTIONS(1210),
    [anon_sym__Nonnull] = ACTIONS(1210),
    [anon_sym_alignas] = ACTIONS(1210),
    [anon_sym__Alignas] = ACTIONS(1210),
    [sym_primitive_type] = ACTIONS(1210),
    [anon_sym_enum] = ACTIONS(1210),
    [anon_sym_struct] = ACTIONS(1210),
    [anon_sym_union] = ACTIONS(1210),
    [anon_sym_if] = ACTIONS(1210),
    [anon_sym_else] = ACTIONS(1210),
    [anon_sym_switch] = ACTIONS(1210),
    [anon_sym_case] = ACTIONS(1210),
    [anon_sym_default] = ACTIONS(1210),
    [anon_sym_while] = ACTIONS(1210),
    [anon_sym_do] = ACTIONS(1210),
    [anon_sym_for] = ACTIONS(1210),
    [anon_sym_return] = ACTIONS(1210),
    [anon_sym_break] = ACTIONS(1210),
    [anon_sym_continue] = ACTIONS(1210),
    [anon_sym_goto] = ACTIONS(1210),
    [anon_sym___try] = ACTIONS(1210),
    [anon_sym___leave] = ACTIONS(1210),
    [anon_sym_DASH_DASH] = ACTIONS(1212),
    [anon_sym_PLUS_PLUS] = ACTIONS(1212),
    [anon_sym_sizeof] = ACTIONS(1210),
    [anon_sym___alignof__] = ACTIONS(1210),
    [anon_sym___alignof] = ACTIONS(1210),
    [anon_sym__alignof] = ACTIONS(1210),
    [anon_sym_alignof] = ACTIONS(1210),
    [anon_sym__Alignof] = ACTIONS(1210),
    [anon_sym_offsetof] = ACTIONS(1210),
    [anon_sym__Generic] = ACTIONS(1210),
    [anon_sym_asm] = ACTIONS(1210),
    [anon_sym___asm__] = ACTIONS(1210),
    [anon_sym___asm] = ACTIONS(1210),
    [sym_number_literal] = ACTIONS(1212),
    [anon_sym_L_SQUOTE] = ACTIONS(1212),
    [anon_sym_u_SQUOTE] = ACTIONS(1212),
    [anon_sym_U_SQUOTE] = ACTIONS(1212),
    [anon_sym_u8_SQUOTE] = ACTIONS(1212),
    [anon_sym_SQUOTE] = ACTIONS(1212),
    [anon_sym_L_DQUOTE] = ACTIONS(1212),
    [anon_sym_u_DQUOTE] = ACTIONS(1212),
    [anon_sym_U_DQUOTE] = ACTIONS(1212),
    [anon_sym_u8_DQUOTE] = ACTIONS(1212),
    [anon_sym_DQUOTE] = ACTIONS(1212),
    [sym_true] = ACTIONS(1210),
    [sym_false] = ACTIONS(1210),
    [anon_sym_NULL] = ACTIONS(1210),
    [anon_sym_nullptr] = ACTIONS(1210),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(196)] = {
    [sym_identifier] = ACTIONS(1218),
    [aux_sym_preproc_include_token1] = ACTIONS(1218),
    [aux_sym_preproc_def_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token2] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1218),
    [sym_preproc_directive] = ACTIONS(1218),
    [anon_sym_LPAREN2] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_TILDE] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1218),
    [anon_sym_PLUS] = ACTIONS(1218),
    [anon_sym_STAR] = ACTIONS(1220),
    [anon_sym_AMP] = ACTIONS(1220),
    [anon_sym_SEMI] = ACTIONS(1220),
    [anon_sym___extension__] = ACTIONS(1218),
    [anon_sym_typedef] = ACTIONS(1218),
    [anon_sym_extern] = ACTIONS(1218),
    [anon_sym___attribute__] = ACTIONS(1218),
    [anon_sym___attribute] = ACTIONS(1218),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1220),
    [anon_sym___declspec] = ACTIONS(1218),
    [anon_sym___cdecl] = ACTIONS(1218),
    [anon_sym___clrcall] = ACTIONS(1218),
    [anon_sym___stdcall] = ACTIONS(1218),
    [anon_sym___fastcall] = ACTIONS(1218),
    [anon_sym___thiscall] = ACTIONS(1218),
    [anon_sym___vectorcall] = ACTIONS(1218),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_signed] = ACTIONS(1218),
    [anon_sym_unsigned] = ACTIONS(1218),
    [anon_sym_long] = ACTIONS(1218),
    [anon_sym_short] = ACTIONS(1218),
    [anon_sym_static] = ACTIONS(1218),
    [anon_sym_auto] = ACTIONS(1218),
    [anon_sym_register] = ACTIONS(1218),
    [anon_sym_inline] = ACTIONS(1218),
    [anon_sym___inline] = ACTIONS(1218),
    [anon_sym___inline__] = ACTIONS(1218),
    [anon_sym___forceinline] = ACTIONS(1218),
    [anon_sym_thread_local] = ACTIONS(1218),
    [anon_sym___thread] = ACTIONS(1218),
    [anon_sym_const] = ACTIONS(1218),
    [anon_sym_constexpr] = ACTIONS(1218),
    [anon_sym_volatile] = ACTIONS(1218),
    [anon_sym_restrict] = ACTIONS(1218),
    [anon_sym___restrict__] = ACTIONS(1218),
    [anon_sym__Atomic] = ACTIONS(1218),
    [anon_sym__Noreturn] = ACTIONS(1218),
    [anon_sym_noreturn] = ACTIONS(1218),
    [anon_sym__Nonnull] = ACTIONS(1218),
    [anon_sym_alignas] = ACTIONS(1218),
    [anon_sym__Alignas] = ACTIONS(1218),
    [sym_primitive_type] = ACTIONS(1218),
    [anon_sym_enum] = ACTIONS(1218),
    [anon_sym_struct] = ACTIONS(1218),
    [anon_sym_union] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1218),
    [anon_sym_else] = ACTIONS(1218),
    [anon_sym_switch] = ACTIONS(1218),
    [anon_sym_case] = ACTIONS(1218),
    [anon_sym_default] = ACTIONS(1218),
    [anon_sym_while] = ACTIONS(1218),
    [anon_sym_do] = ACTIONS(1218),
    [anon_sym_for] = ACTIONS(1218),
    [anon_sym_return] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1218),
    [anon_sym_continue] = ACTIONS(1218),
    [anon_sym_goto] = ACTIONS(1218),
    [anon_sym___try] = ACTIONS(1218),
    [anon_sym___leave] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_PLUS_PLUS] = ACTIONS(1220),
    [anon_sym_sizeof] = ACTIONS(1218),
    [anon_sym___alignof__] = ACTIONS(1218),
    [anon_sym___alignof] = ACTIONS(1218),
    [anon_sym__alignof] = ACTIONS(1218),
    [anon_sym_alignof] = ACTIONS(1218),
    [anon_sym__Alignof] = ACTIONS(1218),
    [anon_sym_offsetof] = ACTIONS(1218),
    [anon_sym__Generic] = ACTIONS(1218),
    [anon_sym_asm] = ACTIONS(1218),
    [anon_sym___asm__] = ACTIONS(1218),
    [anon_sym___asm] = ACTIONS(1218),
    [sym_number_literal] = ACTIONS(1220),
    [anon_sym_L_SQUOTE] = ACTIONS(1220),
    [anon_sym_u_SQUOTE] = ACTIONS(1220),
    [anon_sym_U_SQUOTE] = ACTIONS(1220),
    [anon_sym_u8_SQUOTE] = ACTIONS(1220),
    [anon_sym_SQUOTE] = ACTIONS(1220),
    [anon_sym_L_DQUOTE] = ACTIONS(1220),
    [anon_sym_u_DQUOTE] = ACTIONS(1220),
    [anon_sym_U_DQUOTE] = ACTIONS(1220),
    [anon_sym_u8_DQUOTE] = ACTIONS(1220),
    [anon_sym_DQUOTE] = ACTIONS(1220),
    [sym_true] = ACTIONS(1218),
    [sym_false] = ACTIONS(1218),
    [anon_sym_NULL] = ACTIONS(1218),
    [anon_sym_nullptr] = ACTIONS(1218),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(197)] = {
    [sym_identifier] = ACTIONS(1222),
    [aux_sym_preproc_include_token1] = ACTIONS(1222),
    [aux_sym_preproc_def_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token2] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1222),
    [sym_preproc_directive] = ACTIONS(1222),
    [anon_sym_LPAREN2] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1222),
    [anon_sym_PLUS] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym___extension__] = ACTIONS(1222),
    [anon_sym_typedef] = ACTIONS(1222),
    [anon_sym_extern] = ACTIONS(1222),
    [anon_sym___attribute__] = ACTIONS(1222),
    [anon_sym___attribute] = ACTIONS(1222),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1224),
    [anon_sym___declspec] = ACTIONS(1222),
    [anon_sym___cdecl] = ACTIONS(1222),
    [anon_sym___clrcall] = ACTIONS(1222),
    [anon_sym___stdcall] = ACTIONS(1222),
    [anon_sym___fastcall] = ACTIONS(1222),
    [anon_sym___thiscall] = ACTIONS(1222),
    [anon_sym___vectorcall] = ACTIONS(1222),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_signed] = ACTIONS(1222),
    [anon_sym_unsigned] = ACTIONS(1222),
    [anon_sym_long] = ACTIONS(1222),
    [anon_sym_short] = ACTIONS(1222),
    [anon_sym_static] = ACTIONS(1222),
    [anon_sym_auto] = ACTIONS(1222),
    [anon_sym_register] = ACTIONS(1222),
    [anon_sym_inline] = ACTIONS(1222),
    [anon_sym___inline] = ACTIONS(1222),
    [anon_sym___inline__] = ACTIONS(1222),
    [anon_sym___forceinline] = ACTIONS(1222),
    [anon_sym_thread_local] = ACTIONS(1222),
    [anon_sym___thread] = ACTIONS(1222),
    [anon_sym_const] = ACTIONS(1222),
    [anon_sym_constexpr] = ACTIONS(1222),
    [anon_sym_volatile] = ACTIONS(1222),
    [anon_sym_restrict] = ACTIONS(1222),
    [anon_sym___restrict__] = ACTIONS(1222),
    [anon_sym__Atomic] = ACTIONS(1222),
    [anon_sym__Noreturn] = ACTIONS(1222),
    [anon_sym_noreturn] = ACTIONS(1222),
    [anon_sym__Nonnull] = ACTIONS(1222),
    [anon_sym_alignas] = ACTIONS(1222),
    [anon_sym__Alignas] = ACTIONS(1222),
    [sym_primitive_type] = ACTIONS(1222),
    [anon_sym_enum] = ACTIONS(1222),
    [anon_sym_struct] = ACTIONS(1222),
    [anon_sym_union] = ACTIONS(1222),
    [anon_sym_if] = ACTIONS(1222),
    [anon_sym_else] = ACTIONS(1222),
    [anon_sym_switch] = ACTIONS(1222),
    [anon_sym_case] = ACTIONS(1222),
    [anon_sym_default] = ACTIONS(1222),
    [anon_sym_while] = ACTIONS(1222),
    [anon_sym_do] = ACTIONS(1222),
    [anon_sym_for] = ACTIONS(1222),
    [anon_sym_return] = ACTIONS(1222),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_goto] = ACTIONS(1222),
    [anon_sym___try] = ACTIONS(1222),
    [anon_sym___leave] = ACTIONS(1222),
    [anon_sym_DASH_DASH] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1224),
    [anon_sym_sizeof] = ACTIONS(1222),
    [anon_sym___alignof__] = ACTIONS(1222),
    [anon_sym___alignof] = ACTIONS(1222),
    [anon_sym__alignof] = ACTIONS(1222),
    [anon_sym_alignof] = ACTIONS(1222),
    [anon_sym__Alignof] = ACTIONS(1222),
    [anon_sym_offsetof] = ACTIONS(1222),
    [anon_sym__Generic] = ACTIONS(1222),
    [anon_sym_asm] = ACTIONS(1222),
    [anon_sym___asm__] = ACTIONS(1222),
    [anon_sym___asm] = ACTIONS(1222),
    [sym_number_literal] = ACTIONS(1224),
    [anon_sym_L_SQUOTE] = ACTIONS(1224),
    [anon_sym_u_SQUOTE] = ACTIONS(1224),
    [anon_sym_U_SQUOTE] = ACTIONS(1224),
    [anon_sym_u8_SQUOTE] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_L_DQUOTE] = ACTIONS(1224),
    [anon_sym_u_DQUOTE] = ACTIONS(1224),
    [anon_sym_U_DQUOTE] = ACTIONS(1224),
    [anon_sym_u8_DQUOTE] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [sym_true] = ACTIONS(1222),
    [sym_false] = ACTIONS(1222),
    [anon_sym_NULL] = ACTIONS(1222),
    [anon_sym_nullptr] = ACTIONS(1222),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(198)] = {
    [sym_identifier] = ACTIONS(1162),
    [aux_sym_preproc_include_token1] = ACTIONS(1162),
    [aux_sym_preproc_def_token1] = ACTIONS(1162),
    [aux_sym_preproc_if_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1162),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1162),
    [sym_preproc_directive] = ACTIONS(1162),
    [anon_sym_LPAREN2] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1162),
    [anon_sym_PLUS] = ACTIONS(1162),
    [anon_sym_STAR] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym___extension__] = ACTIONS(1162),
    [anon_sym_typedef] = ACTIONS(1162),
    [anon_sym_extern] = ACTIONS(1162),
    [anon_sym___attribute__] = ACTIONS(1162),
    [anon_sym___attribute] = ACTIONS(1162),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1164),
    [anon_sym___declspec] = ACTIONS(1162),
    [anon_sym___cdecl] = ACTIONS(1162),
    [anon_sym___clrcall] = ACTIONS(1162),
    [anon_sym___stdcall] = ACTIONS(1162),
    [anon_sym___fastcall] = ACTIONS(1162),
    [anon_sym___thiscall] = ACTIONS(1162),
    [anon_sym___vectorcall] = ACTIONS(1162),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_RBRACE] = ACTIONS(1164),
    [anon_sym_signed] = ACTIONS(1162),
    [anon_sym_unsigned] = ACTIONS(1162),
    [anon_sym_long] = ACTIONS(1162),
    [anon_sym_short] = ACTIONS(1162),
    [anon_sym_static] = ACTIONS(1162),
    [anon_sym_auto] = ACTIONS(1162),
    [anon_sym_register] = ACTIONS(1162),
    [anon_sym_inline] = ACTIONS(1162),
    [anon_sym___inline] = ACTIONS(1162),
    [anon_sym___inline__] = ACTIONS(1162),
    [anon_sym___forceinline] = ACTIONS(1162),
    [anon_sym_thread_local] = ACTIONS(1162),
    [anon_sym___thread] = ACTIONS(1162),
    [anon_sym_const] = ACTIONS(1162),
    [anon_sym_constexpr] = ACTIONS(1162),
    [anon_sym_volatile] = ACTIONS(1162),
    [anon_sym_restrict] = ACTIONS(1162),
    [anon_sym___restrict__] = ACTIONS(1162),
    [anon_sym__Atomic] = ACTIONS(1162),
    [anon_sym__Noreturn] = ACTIONS(1162),
    [anon_sym_noreturn] = ACTIONS(1162),
    [anon_sym__Nonnull] = ACTIONS(1162),
    [anon_sym_alignas] = ACTIONS(1162),
    [anon_sym__Alignas] = ACTIONS(1162),
    [sym_primitive_type] = ACTIONS(1162),
    [anon_sym_enum] = ACTIONS(1162),
    [anon_sym_struct] = ACTIONS(1162),
    [anon_sym_union] = ACTIONS(1162),
    [anon_sym_if] = ACTIONS(1162),
    [anon_sym_else] = ACTIONS(1162),
    [anon_sym_switch] = ACTIONS(1162),
    [anon_sym_case] = ACTIONS(1162),
    [anon_sym_default] = ACTIONS(1162),
    [anon_sym_while] = ACTIONS(1162),
    [anon_sym_do] = ACTIONS(1162),
    [anon_sym_for] = ACTIONS(1162),
    [anon_sym_return] = ACTIONS(1162),
    [anon_sym_break] = ACTIONS(1162),
    [anon_sym_continue] = ACTIONS(1162),
    [anon_sym_goto] = ACTIONS(1162),
    [anon_sym___try] = ACTIONS(1162),
    [anon_sym___leave] = ACTIONS(1162),
    [anon_sym_DASH_DASH] = ACTIONS(1164),
    [anon_sym_PLUS_PLUS] = ACTIONS(1164),
    [anon_sym_sizeof] = ACTIONS(1162),
    [anon_sym___alignof__] = ACTIONS(1162),
    [anon_sym___alignof] = ACTIONS(1162),
    [anon_sym__alignof] = ACTIONS(1162),
    [anon_sym_alignof] = ACTIONS(1162),
    [anon_sym__Alignof] = ACTIONS(1162),
    [anon_sym_offsetof] = ACTIONS(1162),
    [anon_sym__Generic] = ACTIONS(1162),
    [anon_sym_asm] = ACTIONS(1162),
    [anon_sym___asm__] = ACTIONS(1162),
    [anon_sym___asm] = ACTIONS(1162),
    [sym_number_literal] = ACTIONS(1164),
    [anon_sym_L_SQUOTE] = ACTIONS(1164),
    [anon_sym_u_SQUOTE] = ACTIONS(1164),
    [anon_sym_U_SQUOTE] = ACTIONS(1164),
    [anon_sym_u8_SQUOTE] = ACTIONS(1164),
    [anon_sym_SQUOTE] = ACTIONS(1164),
    [anon_sym_L_DQUOTE] = ACTIONS(1164),
    [anon_sym_u_DQUOTE] = ACTIONS(1164),
    [anon_sym_U_DQUOTE] = ACTIONS(1164),
    [anon_sym_u8_DQUOTE] = ACTIONS(1164),
    [anon_sym_DQUOTE] = ACTIONS(1164),
    [sym_true] = ACTIONS(1162),
    [sym_false] = ACTIONS(1162),
    [anon_sym_NULL] = ACTIONS(1162),
    [anon_sym_nullptr] = ACTIONS(1162),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(199)] = {
    [sym_identifier] = ACTIONS(1258),
    [aux_sym_preproc_include_token1] = ACTIONS(1258),
    [aux_sym_preproc_def_token1] = ACTIONS(1258),
    [aux_sym_preproc_if_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1258),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1258),
    [sym_preproc_directive] = ACTIONS(1258),
    [anon_sym_LPAREN2] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1258),
    [anon_sym_PLUS] = ACTIONS(1258),
    [anon_sym_STAR] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym___extension__] = ACTIONS(1258),
    [anon_sym_typedef] = ACTIONS(1258),
    [anon_sym_extern] = ACTIONS(1258),
    [anon_sym___attribute__] = ACTIONS(1258),
    [anon_sym___attribute] = ACTIONS(1258),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1260),
    [anon_sym___declspec] = ACTIONS(1258),
    [anon_sym___cdecl] = ACTIONS(1258),
    [anon_sym___clrcall] = ACTIONS(1258),
    [anon_sym___stdcall] = ACTIONS(1258),
    [anon_sym___fastcall] = ACTIONS(1258),
    [anon_sym___thiscall] = ACTIONS(1258),
    [anon_sym___vectorcall] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [anon_sym_signed] = ACTIONS(1258),
    [anon_sym_unsigned] = ACTIONS(1258),
    [anon_sym_long] = ACTIONS(1258),
    [anon_sym_short] = ACTIONS(1258),
    [anon_sym_static] = ACTIONS(1258),
    [anon_sym_auto] = ACTIONS(1258),
    [anon_sym_register] = ACTIONS(1258),
    [anon_sym_inline] = ACTIONS(1258),
    [anon_sym___inline] = ACTIONS(1258),
    [anon_sym___inline__] = ACTIONS(1258),
    [anon_sym___forceinline] = ACTIONS(1258),
    [anon_sym_thread_local] = ACTIONS(1258),
    [anon_sym___thread] = ACTIONS(1258),
    [anon_sym_const] = ACTIONS(1258),
    [anon_sym_constexpr] = ACTIONS(1258),
    [anon_sym_volatile] = ACTIONS(1258),
    [anon_sym_restrict] = ACTIONS(1258),
    [anon_sym___restrict__] = ACTIONS(1258),
    [anon_sym__Atomic] = ACTIONS(1258),
    [anon_sym__Noreturn] = ACTIONS(1258),
    [anon_sym_noreturn] = ACTIONS(1258),
    [anon_sym__Nonnull] = ACTIONS(1258),
    [anon_sym_alignas] = ACTIONS(1258),
    [anon_sym__Alignas] = ACTIONS(1258),
    [sym_primitive_type] = ACTIONS(1258),
    [anon_sym_enum] = ACTIONS(1258),
    [anon_sym_struct] = ACTIONS(1258),
    [anon_sym_union] = ACTIONS(1258),
    [anon_sym_if] = ACTIONS(1258),
    [anon_sym_else] = ACTIONS(1258),
    [anon_sym_switch] = ACTIONS(1258),
    [anon_sym_case] = ACTIONS(1258),
    [anon_sym_default] = ACTIONS(1258),
    [anon_sym_while] = ACTIONS(1258),
    [anon_sym_do] = ACTIONS(1258),
    [anon_sym_for] = ACTIONS(1258),
    [anon_sym_return] = ACTIONS(1258),
    [anon_sym_break] = ACTIONS(1258),
    [anon_sym_continue] = ACTIONS(1258),
    [anon_sym_goto] = ACTIONS(1258),
    [anon_sym___try] = ACTIONS(1258),
    [anon_sym___leave] = ACTIONS(1258),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_sizeof] = ACTIONS(1258),
    [anon_sym___alignof__] = ACTIONS(1258),
    [anon_sym___alignof] = ACTIONS(1258),
    [anon_sym__alignof] = ACTIONS(1258),
    [anon_sym_alignof] = ACTIONS(1258),
    [anon_sym__Alignof] = ACTIONS(1258),
    [anon_sym_offsetof] = ACTIONS(1258),
    [anon_sym__Generic] = ACTIONS(1258),
    [anon_sym_asm] = ACTIONS(1258),
    [anon_sym___asm__] = ACTIONS(1258),
    [anon_sym___asm] = ACTIONS(1258),
    [sym_number_literal] = ACTIONS(1260),
    [anon_sym_L_SQUOTE] = ACTIONS(1260),
    [anon_sym_u_SQUOTE] = ACTIONS(1260),
    [anon_sym_U_SQUOTE] = ACTIONS(1260),
    [anon_sym_u8_SQUOTE] = ACTIONS(1260),
    [anon_sym_SQUOTE] = ACTIONS(1260),
    [anon_sym_L_DQUOTE] = ACTIONS(1260),
    [anon_sym_u_DQUOTE] = ACTIONS(1260),
    [anon_sym_U_DQUOTE] = ACTIONS(1260),
    [anon_sym_u8_DQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1260),
    [sym_true] = ACTIONS(1258),
    [sym_false] = ACTIONS(1258),
    [anon_sym_NULL] = ACTIONS(1258),
    [anon_sym_nullptr] = ACTIONS(1258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(200)] = {
    [sym_identifier] = ACTIONS(1250),
    [aux_sym_preproc_include_token1] = ACTIONS(1250),
    [aux_sym_preproc_def_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token2] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1250),
    [sym_preproc_directive] = ACTIONS(1250),
    [anon_sym_LPAREN2] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1250),
    [anon_sym_PLUS] = ACTIONS(1250),
    [anon_sym_STAR] = ACTIONS(1252),
    [anon_sym_AMP] = ACTIONS(1252),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym___extension__] = ACTIONS(1250),
    [anon_sym_typedef] = ACTIONS(1250),
    [anon_sym_extern] = ACTIONS(1250),
    [anon_sym___attribute__] = ACTIONS(1250),
    [anon_sym___attribute] = ACTIONS(1250),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1252),
    [anon_sym___declspec] = ACTIONS(1250),
    [anon_sym___cdecl] = ACTIONS(1250),
    [anon_sym___clrcall] = ACTIONS(1250),
    [anon_sym___stdcall] = ACTIONS(1250),
    [anon_sym___fastcall] = ACTIONS(1250),
    [anon_sym___thiscall] = ACTIONS(1250),
    [anon_sym___vectorcall] = ACTIONS(1250),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_signed] = ACTIONS(1250),
    [anon_sym_unsigned] = ACTIONS(1250),
    [anon_sym_long] = ACTIONS(1250),
    [anon_sym_short] = ACTIONS(1250),
    [anon_sym_static] = ACTIONS(1250),
    [anon_sym_auto] = ACTIONS(1250),
    [anon_sym_register] = ACTIONS(1250),
    [anon_sym_inline] = ACTIONS(1250),
    [anon_sym___inline] = ACTIONS(1250),
    [anon_sym___inline__] = ACTIONS(1250),
    [anon_sym___forceinline] = ACTIONS(1250),
    [anon_sym_thread_local] = ACTIONS(1250),
    [anon_sym___thread] = ACTIONS(1250),
    [anon_sym_const] = ACTIONS(1250),
    [anon_sym_constexpr] = ACTIONS(1250),
    [anon_sym_volatile] = ACTIONS(1250),
    [anon_sym_restrict] = ACTIONS(1250),
    [anon_sym___restrict__] = ACTIONS(1250),
    [anon_sym__Atomic] = ACTIONS(1250),
    [anon_sym__Noreturn] = ACTIONS(1250),
    [anon_sym_noreturn] = ACTIONS(1250),
    [anon_sym__Nonnull] = ACTIONS(1250),
    [anon_sym_alignas] = ACTIONS(1250),
    [anon_sym__Alignas] = ACTIONS(1250),
    [sym_primitive_type] = ACTIONS(1250),
    [anon_sym_enum] = ACTIONS(1250),
    [anon_sym_struct] = ACTIONS(1250),
    [anon_sym_union] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(1250),
    [anon_sym_else] = ACTIONS(1250),
    [anon_sym_switch] = ACTIONS(1250),
    [anon_sym_case] = ACTIONS(1250),
    [anon_sym_default] = ACTIONS(1250),
    [anon_sym_while] = ACTIONS(1250),
    [anon_sym_do] = ACTIONS(1250),
    [anon_sym_for] = ACTIONS(1250),
    [anon_sym_return] = ACTIONS(1250),
    [anon_sym_break] = ACTIONS(1250),
    [anon_sym_continue] = ACTIONS(1250),
    [anon_sym_goto] = ACTIONS(1250),
    [anon_sym___try] = ACTIONS(1250),
    [anon_sym___leave] = ACTIONS(1250),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_sizeof] = ACTIONS(1250),
    [anon_sym___alignof__] = ACTIONS(1250),
    [anon_sym___alignof] = ACTIONS(1250),
    [anon_sym__alignof] = ACTIONS(1250),
    [anon_sym_alignof] = ACTIONS(1250),
    [anon_sym__Alignof] = ACTIONS(1250),
    [anon_sym_offsetof] = ACTIONS(1250),
    [anon_sym__Generic] = ACTIONS(1250),
    [anon_sym_asm] = ACTIONS(1250),
    [anon_sym___asm__] = ACTIONS(1250),
    [anon_sym___asm] = ACTIONS(1250),
    [sym_number_literal] = ACTIONS(1252),
    [anon_sym_L_SQUOTE] = ACTIONS(1252),
    [anon_sym_u_SQUOTE] = ACTIONS(1252),
    [anon_sym_U_SQUOTE] = ACTIONS(1252),
    [anon_sym_u8_SQUOTE] = ACTIONS(1252),
    [anon_sym_SQUOTE] = ACTIONS(1252),
    [anon_sym_L_DQUOTE] = ACTIONS(1252),
    [anon_sym_u_DQUOTE] = ACTIONS(1252),
    [anon_sym_U_DQUOTE] = ACTIONS(1252),
    [anon_sym_u8_DQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1252),
    [sym_true] = ACTIONS(1250),
    [sym_false] = ACTIONS(1250),
    [anon_sym_NULL] = ACTIONS(1250),
    [anon_sym_nullptr] = ACTIONS(1250),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(201)] = {
    [sym_identifier] = ACTIONS(1186),
    [aux_sym_preproc_include_token1] = ACTIONS(1186),
    [aux_sym_preproc_def_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token2] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1186),
    [sym_preproc_directive] = ACTIONS(1186),
    [anon_sym_LPAREN2] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1186),
    [anon_sym_PLUS] = ACTIONS(1186),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym___extension__] = ACTIONS(1186),
    [anon_sym_typedef] = ACTIONS(1186),
    [anon_sym_extern] = ACTIONS(1186),
    [anon_sym___attribute__] = ACTIONS(1186),
    [anon_sym___attribute] = ACTIONS(1186),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1188),
    [anon_sym___declspec] = ACTIONS(1186),
    [anon_sym___cdecl] = ACTIONS(1186),
    [anon_sym___clrcall] = ACTIONS(1186),
    [anon_sym___stdcall] = ACTIONS(1186),
    [anon_sym___fastcall] = ACTIONS(1186),
    [anon_sym___thiscall] = ACTIONS(1186),
    [anon_sym___vectorcall] = ACTIONS(1186),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_signed] = ACTIONS(1186),
    [anon_sym_unsigned] = ACTIONS(1186),
    [anon_sym_long] = ACTIONS(1186),
    [anon_sym_short] = ACTIONS(1186),
    [anon_sym_static] = ACTIONS(1186),
    [anon_sym_auto] = ACTIONS(1186),
    [anon_sym_register] = ACTIONS(1186),
    [anon_sym_inline] = ACTIONS(1186),
    [anon_sym___inline] = ACTIONS(1186),
    [anon_sym___inline__] = ACTIONS(1186),
    [anon_sym___forceinline] = ACTIONS(1186),
    [anon_sym_thread_local] = ACTIONS(1186),
    [anon_sym___thread] = ACTIONS(1186),
    [anon_sym_const] = ACTIONS(1186),
    [anon_sym_constexpr] = ACTIONS(1186),
    [anon_sym_volatile] = ACTIONS(1186),
    [anon_sym_restrict] = ACTIONS(1186),
    [anon_sym___restrict__] = ACTIONS(1186),
    [anon_sym__Atomic] = ACTIONS(1186),
    [anon_sym__Noreturn] = ACTIONS(1186),
    [anon_sym_noreturn] = ACTIONS(1186),
    [anon_sym__Nonnull] = ACTIONS(1186),
    [anon_sym_alignas] = ACTIONS(1186),
    [anon_sym__Alignas] = ACTIONS(1186),
    [sym_primitive_type] = ACTIONS(1186),
    [anon_sym_enum] = ACTIONS(1186),
    [anon_sym_struct] = ACTIONS(1186),
    [anon_sym_union] = ACTIONS(1186),
    [anon_sym_if] = ACTIONS(1186),
    [anon_sym_else] = ACTIONS(1186),
    [anon_sym_switch] = ACTIONS(1186),
    [anon_sym_case] = ACTIONS(1186),
    [anon_sym_default] = ACTIONS(1186),
    [anon_sym_while] = ACTIONS(1186),
    [anon_sym_do] = ACTIONS(1186),
    [anon_sym_for] = ACTIONS(1186),
    [anon_sym_return] = ACTIONS(1186),
    [anon_sym_break] = ACTIONS(1186),
    [anon_sym_continue] = ACTIONS(1186),
    [anon_sym_goto] = ACTIONS(1186),
    [anon_sym___try] = ACTIONS(1186),
    [anon_sym___leave] = ACTIONS(1186),
    [anon_sym_DASH_DASH] = ACTIONS(1188),
    [anon_sym_PLUS_PLUS] = ACTIONS(1188),
    [anon_sym_sizeof] = ACTIONS(1186),
    [anon_sym___alignof__] = ACTIONS(1186),
    [anon_sym___alignof] = ACTIONS(1186),
    [anon_sym__alignof] = ACTIONS(1186),
    [anon_sym_alignof] = ACTIONS(1186),
    [anon_sym__Alignof] = ACTIONS(1186),
    [anon_sym_offsetof] = ACTIONS(1186),
    [anon_sym__Generic] = ACTIONS(1186),
    [anon_sym_asm] = ACTIONS(1186),
    [anon_sym___asm__] = ACTIONS(1186),
    [anon_sym___asm] = ACTIONS(1186),
    [sym_number_literal] = ACTIONS(1188),
    [anon_sym_L_SQUOTE] = ACTIONS(1188),
    [anon_sym_u_SQUOTE] = ACTIONS(1188),
    [anon_sym_U_SQUOTE] = ACTIONS(1188),
    [anon_sym_u8_SQUOTE] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_L_DQUOTE] = ACTIONS(1188),
    [anon_sym_u_DQUOTE] = ACTIONS(1188),
    [anon_sym_U_DQUOTE] = ACTIONS(1188),
    [anon_sym_u8_DQUOTE] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [sym_true] = ACTIONS(1186),
    [sym_false] = ACTIONS(1186),
    [anon_sym_NULL] = ACTIONS(1186),
    [anon_sym_nullptr] = ACTIONS(1186),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(202)] = {
    [sym_identifier] = ACTIONS(1146),
    [aux_sym_preproc_include_token1] = ACTIONS(1146),
    [aux_sym_preproc_def_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token2] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1146),
    [sym_preproc_directive] = ACTIONS(1146),
    [anon_sym_LPAREN2] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_TILDE] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1146),
    [anon_sym_PLUS] = ACTIONS(1146),
    [anon_sym_STAR] = ACTIONS(1148),
    [anon_sym_AMP] = ACTIONS(1148),
    [anon_sym_SEMI] = ACTIONS(1148),
    [anon_sym___extension__] = ACTIONS(1146),
    [anon_sym_typedef] = ACTIONS(1146),
    [anon_sym_extern] = ACTIONS(1146),
    [anon_sym___attribute__] = ACTIONS(1146),
    [anon_sym___attribute] = ACTIONS(1146),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1148),
    [anon_sym___declspec] = ACTIONS(1146),
    [anon_sym___cdecl] = ACTIONS(1146),
    [anon_sym___clrcall] = ACTIONS(1146),
    [anon_sym___stdcall] = ACTIONS(1146),
    [anon_sym___fastcall] = ACTIONS(1146),
    [anon_sym___thiscall] = ACTIONS(1146),
    [anon_sym___vectorcall] = ACTIONS(1146),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_signed] = ACTIONS(1146),
    [anon_sym_unsigned] = ACTIONS(1146),
    [anon_sym_long] = ACTIONS(1146),
    [anon_sym_short] = ACTIONS(1146),
    [anon_sym_static] = ACTIONS(1146),
    [anon_sym_auto] = ACTIONS(1146),
    [anon_sym_register] = ACTIONS(1146),
    [anon_sym_inline] = ACTIONS(1146),
    [anon_sym___inline] = ACTIONS(1146),
    [anon_sym___inline__] = ACTIONS(1146),
    [anon_sym___forceinline] = ACTIONS(1146),
    [anon_sym_thread_local] = ACTIONS(1146),
    [anon_sym___thread] = ACTIONS(1146),
    [anon_sym_const] = ACTIONS(1146),
    [anon_sym_constexpr] = ACTIONS(1146),
    [anon_sym_volatile] = ACTIONS(1146),
    [anon_sym_restrict] = ACTIONS(1146),
    [anon_sym___restrict__] = ACTIONS(1146),
    [anon_sym__Atomic] = ACTIONS(1146),
    [anon_sym__Noreturn] = ACTIONS(1146),
    [anon_sym_noreturn] = ACTIONS(1146),
    [anon_sym__Nonnull] = ACTIONS(1146),
    [anon_sym_alignas] = ACTIONS(1146),
    [anon_sym__Alignas] = ACTIONS(1146),
    [sym_primitive_type] = ACTIONS(1146),
    [anon_sym_enum] = ACTIONS(1146),
    [anon_sym_struct] = ACTIONS(1146),
    [anon_sym_union] = ACTIONS(1146),
    [anon_sym_if] = ACTIONS(1146),
    [anon_sym_else] = ACTIONS(1146),
    [anon_sym_switch] = ACTIONS(1146),
    [anon_sym_case] = ACTIONS(1146),
    [anon_sym_default] = ACTIONS(1146),
    [anon_sym_while] = ACTIONS(1146),
    [anon_sym_do] = ACTIONS(1146),
    [anon_sym_for] = ACTIONS(1146),
    [anon_sym_return] = ACTIONS(1146),
    [anon_sym_break] = ACTIONS(1146),
    [anon_sym_continue] = ACTIONS(1146),
    [anon_sym_goto] = ACTIONS(1146),
    [anon_sym___try] = ACTIONS(1146),
    [anon_sym___leave] = ACTIONS(1146),
    [anon_sym_DASH_DASH] = ACTIONS(1148),
    [anon_sym_PLUS_PLUS] = ACTIONS(1148),
    [anon_sym_sizeof] = ACTIONS(1146),
    [anon_sym___alignof__] = ACTIONS(1146),
    [anon_sym___alignof] = ACTIONS(1146),
    [anon_sym__alignof] = ACTIONS(1146),
    [anon_sym_alignof] = ACTIONS(1146),
    [anon_sym__Alignof] = ACTIONS(1146),
    [anon_sym_offsetof] = ACTIONS(1146),
    [anon_sym__Generic] = ACTIONS(1146),
    [anon_sym_asm] = ACTIONS(1146),
    [anon_sym___asm__] = ACTIONS(1146),
    [anon_sym___asm] = ACTIONS(1146),
    [sym_number_literal] = ACTIONS(1148),
    [anon_sym_L_SQUOTE] = ACTIONS(1148),
    [anon_sym_u_SQUOTE] = ACTIONS(1148),
    [anon_sym_U_SQUOTE] = ACTIONS(1148),
    [anon_sym_u8_SQUOTE] = ACTIONS(1148),
    [anon_sym_SQUOTE] = ACTIONS(1148),
    [anon_sym_L_DQUOTE] = ACTIONS(1148),
    [anon_sym_u_DQUOTE] = ACTIONS(1148),
    [anon_sym_U_DQUOTE] = ACTIONS(1148),
    [anon_sym_u8_DQUOTE] = ACTIONS(1148),
    [anon_sym_DQUOTE] = ACTIONS(1148),
    [sym_true] = ACTIONS(1146),
    [sym_false] = ACTIONS(1146),
    [anon_sym_NULL] = ACTIONS(1146),
    [anon_sym_nullptr] = ACTIONS(1146),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(203)] = {
    [sym_identifier] = ACTIONS(1150),
    [aux_sym_preproc_include_token1] = ACTIONS(1150),
    [aux_sym_preproc_def_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token2] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1150),
    [sym_preproc_directive] = ACTIONS(1150),
    [anon_sym_LPAREN2] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_TILDE] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1150),
    [anon_sym_PLUS] = ACTIONS(1150),
    [anon_sym_STAR] = ACTIONS(1152),
    [anon_sym_AMP] = ACTIONS(1152),
    [anon_sym_SEMI] = ACTIONS(1152),
    [anon_sym___extension__] = ACTIONS(1150),
    [anon_sym_typedef] = ACTIONS(1150),
    [anon_sym_extern] = ACTIONS(1150),
    [anon_sym___attribute__] = ACTIONS(1150),
    [anon_sym___attribute] = ACTIONS(1150),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1152),
    [anon_sym___declspec] = ACTIONS(1150),
    [anon_sym___cdecl] = ACTIONS(1150),
    [anon_sym___clrcall] = ACTIONS(1150),
    [anon_sym___stdcall] = ACTIONS(1150),
    [anon_sym___fastcall] = ACTIONS(1150),
    [anon_sym___thiscall] = ACTIONS(1150),
    [anon_sym___vectorcall] = ACTIONS(1150),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_signed] = ACTIONS(1150),
    [anon_sym_unsigned] = ACTIONS(1150),
    [anon_sym_long] = ACTIONS(1150),
    [anon_sym_short] = ACTIONS(1150),
    [anon_sym_static] = ACTIONS(1150),
    [anon_sym_auto] = ACTIONS(1150),
    [anon_sym_register] = ACTIONS(1150),
    [anon_sym_inline] = ACTIONS(1150),
    [anon_sym___inline] = ACTIONS(1150),
    [anon_sym___inline__] = ACTIONS(1150),
    [anon_sym___forceinline] = ACTIONS(1150),
    [anon_sym_thread_local] = ACTIONS(1150),
    [anon_sym___thread] = ACTIONS(1150),
    [anon_sym_const] = ACTIONS(1150),
    [anon_sym_constexpr] = ACTIONS(1150),
    [anon_sym_volatile] = ACTIONS(1150),
    [anon_sym_restrict] = ACTIONS(1150),
    [anon_sym___restrict__] = ACTIONS(1150),
    [anon_sym__Atomic] = ACTIONS(1150),
    [anon_sym__Noreturn] = ACTIONS(1150),
    [anon_sym_noreturn] = ACTIONS(1150),
    [anon_sym__Nonnull] = ACTIONS(1150),
    [anon_sym_alignas] = ACTIONS(1150),
    [anon_sym__Alignas] = ACTIONS(1150),
    [sym_primitive_type] = ACTIONS(1150),
    [anon_sym_enum] = ACTIONS(1150),
    [anon_sym_struct] = ACTIONS(1150),
    [anon_sym_union] = ACTIONS(1150),
    [anon_sym_if] = ACTIONS(1150),
    [anon_sym_else] = ACTIONS(1150),
    [anon_sym_switch] = ACTIONS(1150),
    [anon_sym_case] = ACTIONS(1150),
    [anon_sym_default] = ACTIONS(1150),
    [anon_sym_while] = ACTIONS(1150),
    [anon_sym_do] = ACTIONS(1150),
    [anon_sym_for] = ACTIONS(1150),
    [anon_sym_return] = ACTIONS(1150),
    [anon_sym_break] = ACTIONS(1150),
    [anon_sym_continue] = ACTIONS(1150),
    [anon_sym_goto] = ACTIONS(1150),
    [anon_sym___try] = ACTIONS(1150),
    [anon_sym___leave] = ACTIONS(1150),
    [anon_sym_DASH_DASH] = ACTIONS(1152),
    [anon_sym_PLUS_PLUS] = ACTIONS(1152),
    [anon_sym_sizeof] = ACTIONS(1150),
    [anon_sym___alignof__] = ACTIONS(1150),
    [anon_sym___alignof] = ACTIONS(1150),
    [anon_sym__alignof] = ACTIONS(1150),
    [anon_sym_alignof] = ACTIONS(1150),
    [anon_sym__Alignof] = ACTIONS(1150),
    [anon_sym_offsetof] = ACTIONS(1150),
    [anon_sym__Generic] = ACTIONS(1150),
    [anon_sym_asm] = ACTIONS(1150),
    [anon_sym___asm__] = ACTIONS(1150),
    [anon_sym___asm] = ACTIONS(1150),
    [sym_number_literal] = ACTIONS(1152),
    [anon_sym_L_SQUOTE] = ACTIONS(1152),
    [anon_sym_u_SQUOTE] = ACTIONS(1152),
    [anon_sym_U_SQUOTE] = ACTIONS(1152),
    [anon_sym_u8_SQUOTE] = ACTIONS(1152),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_L_DQUOTE] = ACTIONS(1152),
    [anon_sym_u_DQUOTE] = ACTIONS(1152),
    [anon_sym_U_DQUOTE] = ACTIONS(1152),
    [anon_sym_u8_DQUOTE] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1152),
    [sym_true] = ACTIONS(1150),
    [sym_false] = ACTIONS(1150),
    [anon_sym_NULL] = ACTIONS(1150),
    [anon_sym_nullptr] = ACTIONS(1150),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(204)] = {
    [sym_identifier] = ACTIONS(1166),
    [aux_sym_preproc_include_token1] = ACTIONS(1166),
    [aux_sym_preproc_def_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token2] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1166),
    [sym_preproc_directive] = ACTIONS(1166),
    [anon_sym_LPAREN2] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_TILDE] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1166),
    [anon_sym_PLUS] = ACTIONS(1166),
    [anon_sym_STAR] = ACTIONS(1168),
    [anon_sym_AMP] = ACTIONS(1168),
    [anon_sym_SEMI] = ACTIONS(1168),
    [anon_sym___extension__] = ACTIONS(1166),
    [anon_sym_typedef] = ACTIONS(1166),
    [anon_sym_extern] = ACTIONS(1166),
    [anon_sym___attribute__] = ACTIONS(1166),
    [anon_sym___attribute] = ACTIONS(1166),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1168),
    [anon_sym___declspec] = ACTIONS(1166),
    [anon_sym___cdecl] = ACTIONS(1166),
    [anon_sym___clrcall] = ACTIONS(1166),
    [anon_sym___stdcall] = ACTIONS(1166),
    [anon_sym___fastcall] = ACTIONS(1166),
    [anon_sym___thiscall] = ACTIONS(1166),
    [anon_sym___vectorcall] = ACTIONS(1166),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_signed] = ACTIONS(1166),
    [anon_sym_unsigned] = ACTIONS(1166),
    [anon_sym_long] = ACTIONS(1166),
    [anon_sym_short] = ACTIONS(1166),
    [anon_sym_static] = ACTIONS(1166),
    [anon_sym_auto] = ACTIONS(1166),
    [anon_sym_register] = ACTIONS(1166),
    [anon_sym_inline] = ACTIONS(1166),
    [anon_sym___inline] = ACTIONS(1166),
    [anon_sym___inline__] = ACTIONS(1166),
    [anon_sym___forceinline] = ACTIONS(1166),
    [anon_sym_thread_local] = ACTIONS(1166),
    [anon_sym___thread] = ACTIONS(1166),
    [anon_sym_const] = ACTIONS(1166),
    [anon_sym_constexpr] = ACTIONS(1166),
    [anon_sym_volatile] = ACTIONS(1166),
    [anon_sym_restrict] = ACTIONS(1166),
    [anon_sym___restrict__] = ACTIONS(1166),
    [anon_sym__Atomic] = ACTIONS(1166),
    [anon_sym__Noreturn] = ACTIONS(1166),
    [anon_sym_noreturn] = ACTIONS(1166),
    [anon_sym__Nonnull] = ACTIONS(1166),
    [anon_sym_alignas] = ACTIONS(1166),
    [anon_sym__Alignas] = ACTIONS(1166),
    [sym_primitive_type] = ACTIONS(1166),
    [anon_sym_enum] = ACTIONS(1166),
    [anon_sym_struct] = ACTIONS(1166),
    [anon_sym_union] = ACTIONS(1166),
    [anon_sym_if] = ACTIONS(1166),
    [anon_sym_else] = ACTIONS(1166),
    [anon_sym_switch] = ACTIONS(1166),
    [anon_sym_case] = ACTIONS(1166),
    [anon_sym_default] = ACTIONS(1166),
    [anon_sym_while] = ACTIONS(1166),
    [anon_sym_do] = ACTIONS(1166),
    [anon_sym_for] = ACTIONS(1166),
    [anon_sym_return] = ACTIONS(1166),
    [anon_sym_break] = ACTIONS(1166),
    [anon_sym_continue] = ACTIONS(1166),
    [anon_sym_goto] = ACTIONS(1166),
    [anon_sym___try] = ACTIONS(1166),
    [anon_sym___leave] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_PLUS_PLUS] = ACTIONS(1168),
    [anon_sym_sizeof] = ACTIONS(1166),
    [anon_sym___alignof__] = ACTIONS(1166),
    [anon_sym___alignof] = ACTIONS(1166),
    [anon_sym__alignof] = ACTIONS(1166),
    [anon_sym_alignof] = ACTIONS(1166),
    [anon_sym__Alignof] = ACTIONS(1166),
    [anon_sym_offsetof] = ACTIONS(1166),
    [anon_sym__Generic] = ACTIONS(1166),
    [anon_sym_asm] = ACTIONS(1166),
    [anon_sym___asm__] = ACTIONS(1166),
    [anon_sym___asm] = ACTIONS(1166),
    [sym_number_literal] = ACTIONS(1168),
    [anon_sym_L_SQUOTE] = ACTIONS(1168),
    [anon_sym_u_SQUOTE] = ACTIONS(1168),
    [anon_sym_U_SQUOTE] = ACTIONS(1168),
    [anon_sym_u8_SQUOTE] = ACTIONS(1168),
    [anon_sym_SQUOTE] = ACTIONS(1168),
    [anon_sym_L_DQUOTE] = ACTIONS(1168),
    [anon_sym_u_DQUOTE] = ACTIONS(1168),
    [anon_sym_U_DQUOTE] = ACTIONS(1168),
    [anon_sym_u8_DQUOTE] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1168),
    [sym_true] = ACTIONS(1166),
    [sym_false] = ACTIONS(1166),
    [anon_sym_NULL] = ACTIONS(1166),
    [anon_sym_nullptr] = ACTIONS(1166),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(205)] = {
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1250),
    [aux_sym_preproc_include_token1] = ACTIONS(1250),
    [aux_sym_preproc_def_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1250),
    [sym_preproc_directive] = ACTIONS(1250),
    [anon_sym_LPAREN2] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1250),
    [anon_sym_PLUS] = ACTIONS(1250),
    [anon_sym_STAR] = ACTIONS(1252),
    [anon_sym_AMP] = ACTIONS(1252),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym___extension__] = ACTIONS(1250),
    [anon_sym_typedef] = ACTIONS(1250),
    [anon_sym_extern] = ACTIONS(1250),
    [anon_sym___attribute__] = ACTIONS(1250),
    [anon_sym___attribute] = ACTIONS(1250),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1252),
    [anon_sym___declspec] = ACTIONS(1250),
    [anon_sym___cdecl] = ACTIONS(1250),
    [anon_sym___clrcall] = ACTIONS(1250),
    [anon_sym___stdcall] = ACTIONS(1250),
    [anon_sym___fastcall] = ACTIONS(1250),
    [anon_sym___thiscall] = ACTIONS(1250),
    [anon_sym___vectorcall] = ACTIONS(1250),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_signed] = ACTIONS(1250),
    [anon_sym_unsigned] = ACTIONS(1250),
    [anon_sym_long] = ACTIONS(1250),
    [anon_sym_short] = ACTIONS(1250),
    [anon_sym_static] = ACTIONS(1250),
    [anon_sym_auto] = ACTIONS(1250),
    [anon_sym_register] = ACTIONS(1250),
    [anon_sym_inline] = ACTIONS(1250),
    [anon_sym___inline] = ACTIONS(1250),
    [anon_sym___inline__] = ACTIONS(1250),
    [anon_sym___forceinline] = ACTIONS(1250),
    [anon_sym_thread_local] = ACTIONS(1250),
    [anon_sym___thread] = ACTIONS(1250),
    [anon_sym_const] = ACTIONS(1250),
    [anon_sym_constexpr] = ACTIONS(1250),
    [anon_sym_volatile] = ACTIONS(1250),
    [anon_sym_restrict] = ACTIONS(1250),
    [anon_sym___restrict__] = ACTIONS(1250),
    [anon_sym__Atomic] = ACTIONS(1250),
    [anon_sym__Noreturn] = ACTIONS(1250),
    [anon_sym_noreturn] = ACTIONS(1250),
    [anon_sym__Nonnull] = ACTIONS(1250),
    [anon_sym_alignas] = ACTIONS(1250),
    [anon_sym__Alignas] = ACTIONS(1250),
    [sym_primitive_type] = ACTIONS(1250),
    [anon_sym_enum] = ACTIONS(1250),
    [anon_sym_struct] = ACTIONS(1250),
    [anon_sym_union] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(1250),
    [anon_sym_else] = ACTIONS(1250),
    [anon_sym_switch] = ACTIONS(1250),
    [anon_sym_case] = ACTIONS(1250),
    [anon_sym_default] = ACTIONS(1250),
    [anon_sym_while] = ACTIONS(1250),
    [anon_sym_do] = ACTIONS(1250),
    [anon_sym_for] = ACTIONS(1250),
    [anon_sym_return] = ACTIONS(1250),
    [anon_sym_break] = ACTIONS(1250),
    [anon_sym_continue] = ACTIONS(1250),
    [anon_sym_goto] = ACTIONS(1250),
    [anon_sym___try] = ACTIONS(1250),
    [anon_sym___leave] = ACTIONS(1250),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_sizeof] = ACTIONS(1250),
    [anon_sym___alignof__] = ACTIONS(1250),
    [anon_sym___alignof] = ACTIONS(1250),
    [anon_sym__alignof] = ACTIONS(1250),
    [anon_sym_alignof] = ACTIONS(1250),
    [anon_sym__Alignof] = ACTIONS(1250),
    [anon_sym_offsetof] = ACTIONS(1250),
    [anon_sym__Generic] = ACTIONS(1250),
    [anon_sym_asm] = ACTIONS(1250),
    [anon_sym___asm__] = ACTIONS(1250),
    [anon_sym___asm] = ACTIONS(1250),
    [sym_number_literal] = ACTIONS(1252),
    [anon_sym_L_SQUOTE] = ACTIONS(1252),
    [anon_sym_u_SQUOTE] = ACTIONS(1252),
    [anon_sym_U_SQUOTE] = ACTIONS(1252),
    [anon_sym_u8_SQUOTE] = ACTIONS(1252),
    [anon_sym_SQUOTE] = ACTIONS(1252),
    [anon_sym_L_DQUOTE] = ACTIONS(1252),
    [anon_sym_u_DQUOTE] = ACTIONS(1252),
    [anon_sym_U_DQUOTE] = ACTIONS(1252),
    [anon_sym_u8_DQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1252),
    [sym_true] = ACTIONS(1250),
    [sym_false] = ACTIONS(1250),
    [anon_sym_NULL] = ACTIONS(1250),
    [anon_sym_nullptr] = ACTIONS(1250),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(206)] = {
    [sym_identifier] = ACTIONS(1182),
    [aux_sym_preproc_include_token1] = ACTIONS(1182),
    [aux_sym_preproc_def_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token2] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1182),
    [sym_preproc_directive] = ACTIONS(1182),
    [anon_sym_LPAREN2] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_TILDE] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1182),
    [anon_sym_PLUS] = ACTIONS(1182),
    [anon_sym_STAR] = ACTIONS(1184),
    [anon_sym_AMP] = ACTIONS(1184),
    [anon_sym_SEMI] = ACTIONS(1184),
    [anon_sym___extension__] = ACTIONS(1182),
    [anon_sym_typedef] = ACTIONS(1182),
    [anon_sym_extern] = ACTIONS(1182),
    [anon_sym___attribute__] = ACTIONS(1182),
    [anon_sym___attribute] = ACTIONS(1182),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1184),
    [anon_sym___declspec] = ACTIONS(1182),
    [anon_sym___cdecl] = ACTIONS(1182),
    [anon_sym___clrcall] = ACTIONS(1182),
    [anon_sym___stdcall] = ACTIONS(1182),
    [anon_sym___fastcall] = ACTIONS(1182),
    [anon_sym___thiscall] = ACTIONS(1182),
    [anon_sym___vectorcall] = ACTIONS(1182),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_signed] = ACTIONS(1182),
    [anon_sym_unsigned] = ACTIONS(1182),
    [anon_sym_long] = ACTIONS(1182),
    [anon_sym_short] = ACTIONS(1182),
    [anon_sym_static] = ACTIONS(1182),
    [anon_sym_auto] = ACTIONS(1182),
    [anon_sym_register] = ACTIONS(1182),
    [anon_sym_inline] = ACTIONS(1182),
    [anon_sym___inline] = ACTIONS(1182),
    [anon_sym___inline__] = ACTIONS(1182),
    [anon_sym___forceinline] = ACTIONS(1182),
    [anon_sym_thread_local] = ACTIONS(1182),
    [anon_sym___thread] = ACTIONS(1182),
    [anon_sym_const] = ACTIONS(1182),
    [anon_sym_constexpr] = ACTIONS(1182),
    [anon_sym_volatile] = ACTIONS(1182),
    [anon_sym_restrict] = ACTIONS(1182),
    [anon_sym___restrict__] = ACTIONS(1182),
    [anon_sym__Atomic] = ACTIONS(1182),
    [anon_sym__Noreturn] = ACTIONS(1182),
    [anon_sym_noreturn] = ACTIONS(1182),
    [anon_sym__Nonnull] = ACTIONS(1182),
    [anon_sym_alignas] = ACTIONS(1182),
    [anon_sym__Alignas] = ACTIONS(1182),
    [sym_primitive_type] = ACTIONS(1182),
    [anon_sym_enum] = ACTIONS(1182),
    [anon_sym_struct] = ACTIONS(1182),
    [anon_sym_union] = ACTIONS(1182),
    [anon_sym_if] = ACTIONS(1182),
    [anon_sym_else] = ACTIONS(1182),
    [anon_sym_switch] = ACTIONS(1182),
    [anon_sym_case] = ACTIONS(1182),
    [anon_sym_default] = ACTIONS(1182),
    [anon_sym_while] = ACTIONS(1182),
    [anon_sym_do] = ACTIONS(1182),
    [anon_sym_for] = ACTIONS(1182),
    [anon_sym_return] = ACTIONS(1182),
    [anon_sym_break] = ACTIONS(1182),
    [anon_sym_continue] = ACTIONS(1182),
    [anon_sym_goto] = ACTIONS(1182),
    [anon_sym___try] = ACTIONS(1182),
    [anon_sym___leave] = ACTIONS(1182),
    [anon_sym_DASH_DASH] = ACTIONS(1184),
    [anon_sym_PLUS_PLUS] = ACTIONS(1184),
    [anon_sym_sizeof] = ACTIONS(1182),
    [anon_sym___alignof__] = ACTIONS(1182),
    [anon_sym___alignof] = ACTIONS(1182),
    [anon_sym__alignof] = ACTIONS(1182),
    [anon_sym_alignof] = ACTIONS(1182),
    [anon_sym__Alignof] = ACTIONS(1182),
    [anon_sym_offsetof] = ACTIONS(1182),
    [anon_sym__Generic] = ACTIONS(1182),
    [anon_sym_asm] = ACTIONS(1182),
    [anon_sym___asm__] = ACTIONS(1182),
    [anon_sym___asm] = ACTIONS(1182),
    [sym_number_literal] = ACTIONS(1184),
    [anon_sym_L_SQUOTE] = ACTIONS(1184),
    [anon_sym_u_SQUOTE] = ACTIONS(1184),
    [anon_sym_U_SQUOTE] = ACTIONS(1184),
    [anon_sym_u8_SQUOTE] = ACTIONS(1184),
    [anon_sym_SQUOTE] = ACTIONS(1184),
    [anon_sym_L_DQUOTE] = ACTIONS(1184),
    [anon_sym_u_DQUOTE] = ACTIONS(1184),
    [anon_sym_U_DQUOTE] = ACTIONS(1184),
    [anon_sym_u8_DQUOTE] = ACTIONS(1184),
    [anon_sym_DQUOTE] = ACTIONS(1184),
    [sym_true] = ACTIONS(1182),
    [sym_false] = ACTIONS(1182),
    [anon_sym_NULL] = ACTIONS(1182),
    [anon_sym_nullptr] = ACTIONS(1182),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(207)] = {
    [sym_identifier] = ACTIONS(1198),
    [aux_sym_preproc_include_token1] = ACTIONS(1198),
    [aux_sym_preproc_def_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token2] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1198),
    [sym_preproc_directive] = ACTIONS(1198),
    [anon_sym_LPAREN2] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1198),
    [anon_sym_PLUS] = ACTIONS(1198),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym___extension__] = ACTIONS(1198),
    [anon_sym_typedef] = ACTIONS(1198),
    [anon_sym_extern] = ACTIONS(1198),
    [anon_sym___attribute__] = ACTIONS(1198),
    [anon_sym___attribute] = ACTIONS(1198),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1200),
    [anon_sym___declspec] = ACTIONS(1198),
    [anon_sym___cdecl] = ACTIONS(1198),
    [anon_sym___clrcall] = ACTIONS(1198),
    [anon_sym___stdcall] = ACTIONS(1198),
    [anon_sym___fastcall] = ACTIONS(1198),
    [anon_sym___thiscall] = ACTIONS(1198),
    [anon_sym___vectorcall] = ACTIONS(1198),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_signed] = ACTIONS(1198),
    [anon_sym_unsigned] = ACTIONS(1198),
    [anon_sym_long] = ACTIONS(1198),
    [anon_sym_short] = ACTIONS(1198),
    [anon_sym_static] = ACTIONS(1198),
    [anon_sym_auto] = ACTIONS(1198),
    [anon_sym_register] = ACTIONS(1198),
    [anon_sym_inline] = ACTIONS(1198),
    [anon_sym___inline] = ACTIONS(1198),
    [anon_sym___inline__] = ACTIONS(1198),
    [anon_sym___forceinline] = ACTIONS(1198),
    [anon_sym_thread_local] = ACTIONS(1198),
    [anon_sym___thread] = ACTIONS(1198),
    [anon_sym_const] = ACTIONS(1198),
    [anon_sym_constexpr] = ACTIONS(1198),
    [anon_sym_volatile] = ACTIONS(1198),
    [anon_sym_restrict] = ACTIONS(1198),
    [anon_sym___restrict__] = ACTIONS(1198),
    [anon_sym__Atomic] = ACTIONS(1198),
    [anon_sym__Noreturn] = ACTIONS(1198),
    [anon_sym_noreturn] = ACTIONS(1198),
    [anon_sym__Nonnull] = ACTIONS(1198),
    [anon_sym_alignas] = ACTIONS(1198),
    [anon_sym__Alignas] = ACTIONS(1198),
    [sym_primitive_type] = ACTIONS(1198),
    [anon_sym_enum] = ACTIONS(1198),
    [anon_sym_struct] = ACTIONS(1198),
    [anon_sym_union] = ACTIONS(1198),
    [anon_sym_if] = ACTIONS(1198),
    [anon_sym_else] = ACTIONS(1198),
    [anon_sym_switch] = ACTIONS(1198),
    [anon_sym_case] = ACTIONS(1198),
    [anon_sym_default] = ACTIONS(1198),
    [anon_sym_while] = ACTIONS(1198),
    [anon_sym_do] = ACTIONS(1198),
    [anon_sym_for] = ACTIONS(1198),
    [anon_sym_return] = ACTIONS(1198),
    [anon_sym_break] = ACTIONS(1198),
    [anon_sym_continue] = ACTIONS(1198),
    [anon_sym_goto] = ACTIONS(1198),
    [anon_sym___try] = ACTIONS(1198),
    [anon_sym___leave] = ACTIONS(1198),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_sizeof] = ACTIONS(1198),
    [anon_sym___alignof__] = ACTIONS(1198),
    [anon_sym___alignof] = ACTIONS(1198),
    [anon_sym__alignof] = ACTIONS(1198),
    [anon_sym_alignof] = ACTIONS(1198),
    [anon_sym__Alignof] = ACTIONS(1198),
    [anon_sym_offsetof] = ACTIONS(1198),
    [anon_sym__Generic] = ACTIONS(1198),
    [anon_sym_asm] = ACTIONS(1198),
    [anon_sym___asm__] = ACTIONS(1198),
    [anon_sym___asm] = ACTIONS(1198),
    [sym_number_literal] = ACTIONS(1200),
    [anon_sym_L_SQUOTE] = ACTIONS(1200),
    [anon_sym_u_SQUOTE] = ACTIONS(1200),
    [anon_sym_U_SQUOTE] = ACTIONS(1200),
    [anon_sym_u8_SQUOTE] = ACTIONS(1200),
    [anon_sym_SQUOTE] = ACTIONS(1200),
    [anon_sym_L_DQUOTE] = ACTIONS(1200),
    [anon_sym_u_DQUOTE] = ACTIONS(1200),
    [anon_sym_U_DQUOTE] = ACTIONS(1200),
    [anon_sym_u8_DQUOTE] = ACTIONS(1200),
    [anon_sym_DQUOTE] = ACTIONS(1200),
    [sym_true] = ACTIONS(1198),
    [sym_false] = ACTIONS(1198),
    [anon_sym_NULL] = ACTIONS(1198),
    [anon_sym_nullptr] = ACTIONS(1198),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(208)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token2] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(209)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token2] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(210)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token2] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(211)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token2] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(212)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token2] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(213)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token2] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(214)] = {
    [sym_identifier] = ACTIONS(1174),
    [aux_sym_preproc_include_token1] = ACTIONS(1174),
    [aux_sym_preproc_def_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token2] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1174),
    [sym_preproc_directive] = ACTIONS(1174),
    [anon_sym_LPAREN2] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_TILDE] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1174),
    [anon_sym_PLUS] = ACTIONS(1174),
    [anon_sym_STAR] = ACTIONS(1176),
    [anon_sym_AMP] = ACTIONS(1176),
    [anon_sym_SEMI] = ACTIONS(1176),
    [anon_sym___extension__] = ACTIONS(1174),
    [anon_sym_typedef] = ACTIONS(1174),
    [anon_sym_extern] = ACTIONS(1174),
    [anon_sym___attribute__] = ACTIONS(1174),
    [anon_sym___attribute] = ACTIONS(1174),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1176),
    [anon_sym___declspec] = ACTIONS(1174),
    [anon_sym___cdecl] = ACTIONS(1174),
    [anon_sym___clrcall] = ACTIONS(1174),
    [anon_sym___stdcall] = ACTIONS(1174),
    [anon_sym___fastcall] = ACTIONS(1174),
    [anon_sym___thiscall] = ACTIONS(1174),
    [anon_sym___vectorcall] = ACTIONS(1174),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_signed] = ACTIONS(1174),
    [anon_sym_unsigned] = ACTIONS(1174),
    [anon_sym_long] = ACTIONS(1174),
    [anon_sym_short] = ACTIONS(1174),
    [anon_sym_static] = ACTIONS(1174),
    [anon_sym_auto] = ACTIONS(1174),
    [anon_sym_register] = ACTIONS(1174),
    [anon_sym_inline] = ACTIONS(1174),
    [anon_sym___inline] = ACTIONS(1174),
    [anon_sym___inline__] = ACTIONS(1174),
    [anon_sym___forceinline] = ACTIONS(1174),
    [anon_sym_thread_local] = ACTIONS(1174),
    [anon_sym___thread] = ACTIONS(1174),
    [anon_sym_const] = ACTIONS(1174),
    [anon_sym_constexpr] = ACTIONS(1174),
    [anon_sym_volatile] = ACTIONS(1174),
    [anon_sym_restrict] = ACTIONS(1174),
    [anon_sym___restrict__] = ACTIONS(1174),
    [anon_sym__Atomic] = ACTIONS(1174),
    [anon_sym__Noreturn] = ACTIONS(1174),
    [anon_sym_noreturn] = ACTIONS(1174),
    [anon_sym__Nonnull] = ACTIONS(1174),
    [anon_sym_alignas] = ACTIONS(1174),
    [anon_sym__Alignas] = ACTIONS(1174),
    [sym_primitive_type] = ACTIONS(1174),
    [anon_sym_enum] = ACTIONS(1174),
    [anon_sym_struct] = ACTIONS(1174),
    [anon_sym_union] = ACTIONS(1174),
    [anon_sym_if] = ACTIONS(1174),
    [anon_sym_else] = ACTIONS(1174),
    [anon_sym_switch] = ACTIONS(1174),
    [anon_sym_case] = ACTIONS(1174),
    [anon_sym_default] = ACTIONS(1174),
    [anon_sym_while] = ACTIONS(1174),
    [anon_sym_do] = ACTIONS(1174),
    [anon_sym_for] = ACTIONS(1174),
    [anon_sym_return] = ACTIONS(1174),
    [anon_sym_break] = ACTIONS(1174),
    [anon_sym_continue] = ACTIONS(1174),
    [anon_sym_goto] = ACTIONS(1174),
    [anon_sym___try] = ACTIONS(1174),
    [anon_sym___leave] = ACTIONS(1174),
    [anon_sym_DASH_DASH] = ACTIONS(1176),
    [anon_sym_PLUS_PLUS] = ACTIONS(1176),
    [anon_sym_sizeof] = ACTIONS(1174),
    [anon_sym___alignof__] = ACTIONS(1174),
    [anon_sym___alignof] = ACTIONS(1174),
    [anon_sym__alignof] = ACTIONS(1174),
    [anon_sym_alignof] = ACTIONS(1174),
    [anon_sym__Alignof] = ACTIONS(1174),
    [anon_sym_offsetof] = ACTIONS(1174),
    [anon_sym__Generic] = ACTIONS(1174),
    [anon_sym_asm] = ACTIONS(1174),
    [anon_sym___asm__] = ACTIONS(1174),
    [anon_sym___asm] = ACTIONS(1174),
    [sym_number_literal] = ACTIONS(1176),
    [anon_sym_L_SQUOTE] = ACTIONS(1176),
    [anon_sym_u_SQUOTE] = ACTIONS(1176),
    [anon_sym_U_SQUOTE] = ACTIONS(1176),
    [anon_sym_u8_SQUOTE] = ACTIONS(1176),
    [anon_sym_SQUOTE] = ACTIONS(1176),
    [anon_sym_L_DQUOTE] = ACTIONS(1176),
    [anon_sym_u_DQUOTE] = ACTIONS(1176),
    [anon_sym_U_DQUOTE] = ACTIONS(1176),
    [anon_sym_u8_DQUOTE] = ACTIONS(1176),
    [anon_sym_DQUOTE] = ACTIONS(1176),
    [sym_true] = ACTIONS(1174),
    [sym_false] = ACTIONS(1174),
    [anon_sym_NULL] = ACTIONS(1174),
    [anon_sym_nullptr] = ACTIONS(1174),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(215)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token2] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(216)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token2] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(217)] = {
    [sym_identifier] = ACTIONS(1206),
    [aux_sym_preproc_include_token1] = ACTIONS(1206),
    [aux_sym_preproc_def_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token2] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1206),
    [sym_preproc_directive] = ACTIONS(1206),
    [anon_sym_LPAREN2] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_TILDE] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1208),
    [anon_sym_AMP] = ACTIONS(1208),
    [anon_sym_SEMI] = ACTIONS(1208),
    [anon_sym___extension__] = ACTIONS(1206),
    [anon_sym_typedef] = ACTIONS(1206),
    [anon_sym_extern] = ACTIONS(1206),
    [anon_sym___attribute__] = ACTIONS(1206),
    [anon_sym___attribute] = ACTIONS(1206),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1208),
    [anon_sym___declspec] = ACTIONS(1206),
    [anon_sym___cdecl] = ACTIONS(1206),
    [anon_sym___clrcall] = ACTIONS(1206),
    [anon_sym___stdcall] = ACTIONS(1206),
    [anon_sym___fastcall] = ACTIONS(1206),
    [anon_sym___thiscall] = ACTIONS(1206),
    [anon_sym___vectorcall] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_signed] = ACTIONS(1206),
    [anon_sym_unsigned] = ACTIONS(1206),
    [anon_sym_long] = ACTIONS(1206),
    [anon_sym_short] = ACTIONS(1206),
    [anon_sym_static] = ACTIONS(1206),
    [anon_sym_auto] = ACTIONS(1206),
    [anon_sym_register] = ACTIONS(1206),
    [anon_sym_inline] = ACTIONS(1206),
    [anon_sym___inline] = ACTIONS(1206),
    [anon_sym___inline__] = ACTIONS(1206),
    [anon_sym___forceinline] = ACTIONS(1206),
    [anon_sym_thread_local] = ACTIONS(1206),
    [anon_sym___thread] = ACTIONS(1206),
    [anon_sym_const] = ACTIONS(1206),
    [anon_sym_constexpr] = ACTIONS(1206),
    [anon_sym_volatile] = ACTIONS(1206),
    [anon_sym_restrict] = ACTIONS(1206),
    [anon_sym___restrict__] = ACTIONS(1206),
    [anon_sym__Atomic] = ACTIONS(1206),
    [anon_sym__Noreturn] = ACTIONS(1206),
    [anon_sym_noreturn] = ACTIONS(1206),
    [anon_sym__Nonnull] = ACTIONS(1206),
    [anon_sym_alignas] = ACTIONS(1206),
    [anon_sym__Alignas] = ACTIONS(1206),
    [sym_primitive_type] = ACTIONS(1206),
    [anon_sym_enum] = ACTIONS(1206),
    [anon_sym_struct] = ACTIONS(1206),
    [anon_sym_union] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1206),
    [anon_sym_else] = ACTIONS(1206),
    [anon_sym_switch] = ACTIONS(1206),
    [anon_sym_case] = ACTIONS(1206),
    [anon_sym_default] = ACTIONS(1206),
    [anon_sym_while] = ACTIONS(1206),
    [anon_sym_do] = ACTIONS(1206),
    [anon_sym_for] = ACTIONS(1206),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_break] = ACTIONS(1206),
    [anon_sym_continue] = ACTIONS(1206),
    [anon_sym_goto] = ACTIONS(1206),
    [anon_sym___try] = ACTIONS(1206),
    [anon_sym___leave] = ACTIONS(1206),
    [anon_sym_DASH_DASH] = ACTIONS(1208),
    [anon_sym_PLUS_PLUS] = ACTIONS(1208),
    [anon_sym_sizeof] = ACTIONS(1206),
    [anon_sym___alignof__] = ACTIONS(1206),
    [anon_sym___alignof] = ACTIONS(1206),
    [anon_sym__alignof] = ACTIONS(1206),
    [anon_sym_alignof] = ACTIONS(1206),
    [anon_sym__Alignof] = ACTIONS(1206),
    [anon_sym_offsetof] = ACTIONS(1206),
    [anon_sym__Generic] = ACTIONS(1206),
    [anon_sym_asm] = ACTIONS(1206),
    [anon_sym___asm__] = ACTIONS(1206),
    [anon_sym___asm] = ACTIONS(1206),
    [sym_number_literal] = ACTIONS(1208),
    [anon_sym_L_SQUOTE] = ACTIONS(1208),
    [anon_sym_u_SQUOTE] = ACTIONS(1208),
    [anon_sym_U_SQUOTE] = ACTIONS(1208),
    [anon_sym_u8_SQUOTE] = ACTIONS(1208),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_L_DQUOTE] = ACTIONS(1208),
    [anon_sym_u_DQUOTE] = ACTIONS(1208),
    [anon_sym_U_DQUOTE] = ACTIONS(1208),
    [anon_sym_u8_DQUOTE] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [sym_true] = ACTIONS(1206),
    [sym_false] = ACTIONS(1206),
    [anon_sym_NULL] = ACTIONS(1206),
    [anon_sym_nullptr] = ACTIONS(1206),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(218)] = {
    [sym_identifier] = ACTIONS(1190),
    [aux_sym_preproc_include_token1] = ACTIONS(1190),
    [aux_sym_preproc_def_token1] = ACTIONS(1190),
    [aux_sym_preproc_if_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1190),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1190),
    [sym_preproc_directive] = ACTIONS(1190),
    [anon_sym_LPAREN2] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_TILDE] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1192),
    [anon_sym_AMP] = ACTIONS(1192),
    [anon_sym_SEMI] = ACTIONS(1192),
    [anon_sym___extension__] = ACTIONS(1190),
    [anon_sym_typedef] = ACTIONS(1190),
    [anon_sym_extern] = ACTIONS(1190),
    [anon_sym___attribute__] = ACTIONS(1190),
    [anon_sym___attribute] = ACTIONS(1190),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1192),
    [anon_sym___declspec] = ACTIONS(1190),
    [anon_sym___cdecl] = ACTIONS(1190),
    [anon_sym___clrcall] = ACTIONS(1190),
    [anon_sym___stdcall] = ACTIONS(1190),
    [anon_sym___fastcall] = ACTIONS(1190),
    [anon_sym___thiscall] = ACTIONS(1190),
    [anon_sym___vectorcall] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_RBRACE] = ACTIONS(1192),
    [anon_sym_signed] = ACTIONS(1190),
    [anon_sym_unsigned] = ACTIONS(1190),
    [anon_sym_long] = ACTIONS(1190),
    [anon_sym_short] = ACTIONS(1190),
    [anon_sym_static] = ACTIONS(1190),
    [anon_sym_auto] = ACTIONS(1190),
    [anon_sym_register] = ACTIONS(1190),
    [anon_sym_inline] = ACTIONS(1190),
    [anon_sym___inline] = ACTIONS(1190),
    [anon_sym___inline__] = ACTIONS(1190),
    [anon_sym___forceinline] = ACTIONS(1190),
    [anon_sym_thread_local] = ACTIONS(1190),
    [anon_sym___thread] = ACTIONS(1190),
    [anon_sym_const] = ACTIONS(1190),
    [anon_sym_constexpr] = ACTIONS(1190),
    [anon_sym_volatile] = ACTIONS(1190),
    [anon_sym_restrict] = ACTIONS(1190),
    [anon_sym___restrict__] = ACTIONS(1190),
    [anon_sym__Atomic] = ACTIONS(1190),
    [anon_sym__Noreturn] = ACTIONS(1190),
    [anon_sym_noreturn] = ACTIONS(1190),
    [anon_sym__Nonnull] = ACTIONS(1190),
    [anon_sym_alignas] = ACTIONS(1190),
    [anon_sym__Alignas] = ACTIONS(1190),
    [sym_primitive_type] = ACTIONS(1190),
    [anon_sym_enum] = ACTIONS(1190),
    [anon_sym_struct] = ACTIONS(1190),
    [anon_sym_union] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_else] = ACTIONS(1190),
    [anon_sym_switch] = ACTIONS(1190),
    [anon_sym_case] = ACTIONS(1190),
    [anon_sym_default] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_break] = ACTIONS(1190),
    [anon_sym_continue] = ACTIONS(1190),
    [anon_sym_goto] = ACTIONS(1190),
    [anon_sym___try] = ACTIONS(1190),
    [anon_sym___leave] = ACTIONS(1190),
    [anon_sym_DASH_DASH] = ACTIONS(1192),
    [anon_sym_PLUS_PLUS] = ACTIONS(1192),
    [anon_sym_sizeof] = ACTIONS(1190),
    [anon_sym___alignof__] = ACTIONS(1190),
    [anon_sym___alignof] = ACTIONS(1190),
    [anon_sym__alignof] = ACTIONS(1190),
    [anon_sym_alignof] = ACTIONS(1190),
    [anon_sym__Alignof] = ACTIONS(1190),
    [anon_sym_offsetof] = ACTIONS(1190),
    [anon_sym__Generic] = ACTIONS(1190),
    [anon_sym_asm] = ACTIONS(1190),
    [anon_sym___asm__] = ACTIONS(1190),
    [anon_sym___asm] = ACTIONS(1190),
    [sym_number_literal] = ACTIONS(1192),
    [anon_sym_L_SQUOTE] = ACTIONS(1192),
    [anon_sym_u_SQUOTE] = ACTIONS(1192),
    [anon_sym_U_SQUOTE] = ACTIONS(1192),
    [anon_sym_u8_SQUOTE] = ACTIONS(1192),
    [anon_sym_SQUOTE] = ACTIONS(1192),
    [anon_sym_L_DQUOTE] = ACTIONS(1192),
    [anon_sym_u_DQUOTE] = ACTIONS(1192),
    [anon_sym_U_DQUOTE] = ACTIONS(1192),
    [anon_sym_u8_DQUOTE] = ACTIONS(1192),
    [anon_sym_DQUOTE] = ACTIONS(1192),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [anon_sym_NULL] = ACTIONS(1190),
    [anon_sym_nullptr] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(219)] = {
    [sym_identifier] = ACTIONS(1194),
    [aux_sym_preproc_include_token1] = ACTIONS(1194),
    [aux_sym_preproc_def_token1] = ACTIONS(1194),
    [aux_sym_preproc_if_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1194),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1194),
    [sym_preproc_directive] = ACTIONS(1194),
    [anon_sym_LPAREN2] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_TILDE] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1196),
    [anon_sym_AMP] = ACTIONS(1196),
    [anon_sym_SEMI] = ACTIONS(1196),
    [anon_sym___extension__] = ACTIONS(1194),
    [anon_sym_typedef] = ACTIONS(1194),
    [anon_sym_extern] = ACTIONS(1194),
    [anon_sym___attribute__] = ACTIONS(1194),
    [anon_sym___attribute] = ACTIONS(1194),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1196),
    [anon_sym___declspec] = ACTIONS(1194),
    [anon_sym___cdecl] = ACTIONS(1194),
    [anon_sym___clrcall] = ACTIONS(1194),
    [anon_sym___stdcall] = ACTIONS(1194),
    [anon_sym___fastcall] = ACTIONS(1194),
    [anon_sym___thiscall] = ACTIONS(1194),
    [anon_sym___vectorcall] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_RBRACE] = ACTIONS(1196),
    [anon_sym_signed] = ACTIONS(1194),
    [anon_sym_unsigned] = ACTIONS(1194),
    [anon_sym_long] = ACTIONS(1194),
    [anon_sym_short] = ACTIONS(1194),
    [anon_sym_static] = ACTIONS(1194),
    [anon_sym_auto] = ACTIONS(1194),
    [anon_sym_register] = ACTIONS(1194),
    [anon_sym_inline] = ACTIONS(1194),
    [anon_sym___inline] = ACTIONS(1194),
    [anon_sym___inline__] = ACTIONS(1194),
    [anon_sym___forceinline] = ACTIONS(1194),
    [anon_sym_thread_local] = ACTIONS(1194),
    [anon_sym___thread] = ACTIONS(1194),
    [anon_sym_const] = ACTIONS(1194),
    [anon_sym_constexpr] = ACTIONS(1194),
    [anon_sym_volatile] = ACTIONS(1194),
    [anon_sym_restrict] = ACTIONS(1194),
    [anon_sym___restrict__] = ACTIONS(1194),
    [anon_sym__Atomic] = ACTIONS(1194),
    [anon_sym__Noreturn] = ACTIONS(1194),
    [anon_sym_noreturn] = ACTIONS(1194),
    [anon_sym__Nonnull] = ACTIONS(1194),
    [anon_sym_alignas] = ACTIONS(1194),
    [anon_sym__Alignas] = ACTIONS(1194),
    [sym_primitive_type] = ACTIONS(1194),
    [anon_sym_enum] = ACTIONS(1194),
    [anon_sym_struct] = ACTIONS(1194),
    [anon_sym_union] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_else] = ACTIONS(1194),
    [anon_sym_switch] = ACTIONS(1194),
    [anon_sym_case] = ACTIONS(1194),
    [anon_sym_default] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_break] = ACTIONS(1194),
    [anon_sym_continue] = ACTIONS(1194),
    [anon_sym_goto] = ACTIONS(1194),
    [anon_sym___try] = ACTIONS(1194),
    [anon_sym___leave] = ACTIONS(1194),
    [anon_sym_DASH_DASH] = ACTIONS(1196),
    [anon_sym_PLUS_PLUS] = ACTIONS(1196),
    [anon_sym_sizeof] = ACTIONS(1194),
    [anon_sym___alignof__] = ACTIONS(1194),
    [anon_sym___alignof] = ACTIONS(1194),
    [anon_sym__alignof] = ACTIONS(1194),
    [anon_sym_alignof] = ACTIONS(1194),
    [anon_sym__Alignof] = ACTIONS(1194),
    [anon_sym_offsetof] = ACTIONS(1194),
    [anon_sym__Generic] = ACTIONS(1194),
    [anon_sym_asm] = ACTIONS(1194),
    [anon_sym___asm__] = ACTIONS(1194),
    [anon_sym___asm] = ACTIONS(1194),
    [sym_number_literal] = ACTIONS(1196),
    [anon_sym_L_SQUOTE] = ACTIONS(1196),
    [anon_sym_u_SQUOTE] = ACTIONS(1196),
    [anon_sym_U_SQUOTE] = ACTIONS(1196),
    [anon_sym_u8_SQUOTE] = ACTIONS(1196),
    [anon_sym_SQUOTE] = ACTIONS(1196),
    [anon_sym_L_DQUOTE] = ACTIONS(1196),
    [anon_sym_u_DQUOTE] = ACTIONS(1196),
    [anon_sym_U_DQUOTE] = ACTIONS(1196),
    [anon_sym_u8_DQUOTE] = ACTIONS(1196),
    [anon_sym_DQUOTE] = ACTIONS(1196),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [anon_sym_NULL] = ACTIONS(1194),
    [anon_sym_nullptr] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(220)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_RBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(221)] = {
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_RBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(222)] = {
    [sym_identifier] = ACTIONS(1210),
    [aux_sym_preproc_include_token1] = ACTIONS(1210),
    [aux_sym_preproc_def_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1210),
    [sym_preproc_directive] = ACTIONS(1210),
    [anon_sym_LPAREN2] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1210),
    [anon_sym_PLUS] = ACTIONS(1210),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym___extension__] = ACTIONS(1210),
    [anon_sym_typedef] = ACTIONS(1210),
    [anon_sym_extern] = ACTIONS(1210),
    [anon_sym___attribute__] = ACTIONS(1210),
    [anon_sym___attribute] = ACTIONS(1210),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1212),
    [anon_sym___declspec] = ACTIONS(1210),
    [anon_sym___cdecl] = ACTIONS(1210),
    [anon_sym___clrcall] = ACTIONS(1210),
    [anon_sym___stdcall] = ACTIONS(1210),
    [anon_sym___fastcall] = ACTIONS(1210),
    [anon_sym___thiscall] = ACTIONS(1210),
    [anon_sym___vectorcall] = ACTIONS(1210),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_RBRACE] = ACTIONS(1212),
    [anon_sym_signed] = ACTIONS(1210),
    [anon_sym_unsigned] = ACTIONS(1210),
    [anon_sym_long] = ACTIONS(1210),
    [anon_sym_short] = ACTIONS(1210),
    [anon_sym_static] = ACTIONS(1210),
    [anon_sym_auto] = ACTIONS(1210),
    [anon_sym_register] = ACTIONS(1210),
    [anon_sym_inline] = ACTIONS(1210),
    [anon_sym___inline] = ACTIONS(1210),
    [anon_sym___inline__] = ACTIONS(1210),
    [anon_sym___forceinline] = ACTIONS(1210),
    [anon_sym_thread_local] = ACTIONS(1210),
    [anon_sym___thread] = ACTIONS(1210),
    [anon_sym_const] = ACTIONS(1210),
    [anon_sym_constexpr] = ACTIONS(1210),
    [anon_sym_volatile] = ACTIONS(1210),
    [anon_sym_restrict] = ACTIONS(1210),
    [anon_sym___restrict__] = ACTIONS(1210),
    [anon_sym__Atomic] = ACTIONS(1210),
    [anon_sym__Noreturn] = ACTIONS(1210),
    [anon_sym_noreturn] = ACTIONS(1210),
    [anon_sym__Nonnull] = ACTIONS(1210),
    [anon_sym_alignas] = ACTIONS(1210),
    [anon_sym__Alignas] = ACTIONS(1210),
    [sym_primitive_type] = ACTIONS(1210),
    [anon_sym_enum] = ACTIONS(1210),
    [anon_sym_struct] = ACTIONS(1210),
    [anon_sym_union] = ACTIONS(1210),
    [anon_sym_if] = ACTIONS(1210),
    [anon_sym_else] = ACTIONS(1210),
    [anon_sym_switch] = ACTIONS(1210),
    [anon_sym_case] = ACTIONS(1210),
    [anon_sym_default] = ACTIONS(1210),
    [anon_sym_while] = ACTIONS(1210),
    [anon_sym_do] = ACTIONS(1210),
    [anon_sym_for] = ACTIONS(1210),
    [anon_sym_return] = ACTIONS(1210),
    [anon_sym_break] = ACTIONS(1210),
    [anon_sym_continue] = ACTIONS(1210),
    [anon_sym_goto] = ACTIONS(1210),
    [anon_sym___try] = ACTIONS(1210),
    [anon_sym___leave] = ACTIONS(1210),
    [anon_sym_DASH_DASH] = ACTIONS(1212),
    [anon_sym_PLUS_PLUS] = ACTIONS(1212),
    [anon_sym_sizeof] = ACTIONS(1210),
    [anon_sym___alignof__] = ACTIONS(1210),
    [anon_sym___alignof] = ACTIONS(1210),
    [anon_sym__alignof] = ACTIONS(1210),
    [anon_sym_alignof] = ACTIONS(1210),
    [anon_sym__Alignof] = ACTIONS(1210),
    [anon_sym_offsetof] = ACTIONS(1210),
    [anon_sym__Generic] = ACTIONS(1210),
    [anon_sym_asm] = ACTIONS(1210),
    [anon_sym___asm__] = ACTIONS(1210),
    [anon_sym___asm] = ACTIONS(1210),
    [sym_number_literal] = ACTIONS(1212),
    [anon_sym_L_SQUOTE] = ACTIONS(1212),
    [anon_sym_u_SQUOTE] = ACTIONS(1212),
    [anon_sym_U_SQUOTE] = ACTIONS(1212),
    [anon_sym_u8_SQUOTE] = ACTIONS(1212),
    [anon_sym_SQUOTE] = ACTIONS(1212),
    [anon_sym_L_DQUOTE] = ACTIONS(1212),
    [anon_sym_u_DQUOTE] = ACTIONS(1212),
    [anon_sym_U_DQUOTE] = ACTIONS(1212),
    [anon_sym_u8_DQUOTE] = ACTIONS(1212),
    [anon_sym_DQUOTE] = ACTIONS(1212),
    [sym_true] = ACTIONS(1210),
    [sym_false] = ACTIONS(1210),
    [anon_sym_NULL] = ACTIONS(1210),
    [anon_sym_nullptr] = ACTIONS(1210),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(223)] = {
    [ts_builtin_sym_end] = ACTIONS(1220),
    [sym_identifier] = ACTIONS(1218),
    [aux_sym_preproc_include_token1] = ACTIONS(1218),
    [aux_sym_preproc_def_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1218),
    [sym_preproc_directive] = ACTIONS(1218),
    [anon_sym_LPAREN2] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_TILDE] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1218),
    [anon_sym_PLUS] = ACTIONS(1218),
    [anon_sym_STAR] = ACTIONS(1220),
    [anon_sym_AMP] = ACTIONS(1220),
    [anon_sym_SEMI] = ACTIONS(1220),
    [anon_sym___extension__] = ACTIONS(1218),
    [anon_sym_typedef] = ACTIONS(1218),
    [anon_sym_extern] = ACTIONS(1218),
    [anon_sym___attribute__] = ACTIONS(1218),
    [anon_sym___attribute] = ACTIONS(1218),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1220),
    [anon_sym___declspec] = ACTIONS(1218),
    [anon_sym___cdecl] = ACTIONS(1218),
    [anon_sym___clrcall] = ACTIONS(1218),
    [anon_sym___stdcall] = ACTIONS(1218),
    [anon_sym___fastcall] = ACTIONS(1218),
    [anon_sym___thiscall] = ACTIONS(1218),
    [anon_sym___vectorcall] = ACTIONS(1218),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_signed] = ACTIONS(1218),
    [anon_sym_unsigned] = ACTIONS(1218),
    [anon_sym_long] = ACTIONS(1218),
    [anon_sym_short] = ACTIONS(1218),
    [anon_sym_static] = ACTIONS(1218),
    [anon_sym_auto] = ACTIONS(1218),
    [anon_sym_register] = ACTIONS(1218),
    [anon_sym_inline] = ACTIONS(1218),
    [anon_sym___inline] = ACTIONS(1218),
    [anon_sym___inline__] = ACTIONS(1218),
    [anon_sym___forceinline] = ACTIONS(1218),
    [anon_sym_thread_local] = ACTIONS(1218),
    [anon_sym___thread] = ACTIONS(1218),
    [anon_sym_const] = ACTIONS(1218),
    [anon_sym_constexpr] = ACTIONS(1218),
    [anon_sym_volatile] = ACTIONS(1218),
    [anon_sym_restrict] = ACTIONS(1218),
    [anon_sym___restrict__] = ACTIONS(1218),
    [anon_sym__Atomic] = ACTIONS(1218),
    [anon_sym__Noreturn] = ACTIONS(1218),
    [anon_sym_noreturn] = ACTIONS(1218),
    [anon_sym__Nonnull] = ACTIONS(1218),
    [anon_sym_alignas] = ACTIONS(1218),
    [anon_sym__Alignas] = ACTIONS(1218),
    [sym_primitive_type] = ACTIONS(1218),
    [anon_sym_enum] = ACTIONS(1218),
    [anon_sym_struct] = ACTIONS(1218),
    [anon_sym_union] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1218),
    [anon_sym_else] = ACTIONS(1218),
    [anon_sym_switch] = ACTIONS(1218),
    [anon_sym_case] = ACTIONS(1218),
    [anon_sym_default] = ACTIONS(1218),
    [anon_sym_while] = ACTIONS(1218),
    [anon_sym_do] = ACTIONS(1218),
    [anon_sym_for] = ACTIONS(1218),
    [anon_sym_return] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1218),
    [anon_sym_continue] = ACTIONS(1218),
    [anon_sym_goto] = ACTIONS(1218),
    [anon_sym___try] = ACTIONS(1218),
    [anon_sym___leave] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_PLUS_PLUS] = ACTIONS(1220),
    [anon_sym_sizeof] = ACTIONS(1218),
    [anon_sym___alignof__] = ACTIONS(1218),
    [anon_sym___alignof] = ACTIONS(1218),
    [anon_sym__alignof] = ACTIONS(1218),
    [anon_sym_alignof] = ACTIONS(1218),
    [anon_sym__Alignof] = ACTIONS(1218),
    [anon_sym_offsetof] = ACTIONS(1218),
    [anon_sym__Generic] = ACTIONS(1218),
    [anon_sym_asm] = ACTIONS(1218),
    [anon_sym___asm__] = ACTIONS(1218),
    [anon_sym___asm] = ACTIONS(1218),
    [sym_number_literal] = ACTIONS(1220),
    [anon_sym_L_SQUOTE] = ACTIONS(1220),
    [anon_sym_u_SQUOTE] = ACTIONS(1220),
    [anon_sym_U_SQUOTE] = ACTIONS(1220),
    [anon_sym_u8_SQUOTE] = ACTIONS(1220),
    [anon_sym_SQUOTE] = ACTIONS(1220),
    [anon_sym_L_DQUOTE] = ACTIONS(1220),
    [anon_sym_u_DQUOTE] = ACTIONS(1220),
    [anon_sym_U_DQUOTE] = ACTIONS(1220),
    [anon_sym_u8_DQUOTE] = ACTIONS(1220),
    [anon_sym_DQUOTE] = ACTIONS(1220),
    [sym_true] = ACTIONS(1218),
    [sym_false] = ACTIONS(1218),
    [anon_sym_NULL] = ACTIONS(1218),
    [anon_sym_nullptr] = ACTIONS(1218),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(224)] = {
    [sym_identifier] = ACTIONS(1222),
    [aux_sym_preproc_include_token1] = ACTIONS(1222),
    [aux_sym_preproc_def_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1222),
    [sym_preproc_directive] = ACTIONS(1222),
    [anon_sym_LPAREN2] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1222),
    [anon_sym_PLUS] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym___extension__] = ACTIONS(1222),
    [anon_sym_typedef] = ACTIONS(1222),
    [anon_sym_extern] = ACTIONS(1222),
    [anon_sym___attribute__] = ACTIONS(1222),
    [anon_sym___attribute] = ACTIONS(1222),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1224),
    [anon_sym___declspec] = ACTIONS(1222),
    [anon_sym___cdecl] = ACTIONS(1222),
    [anon_sym___clrcall] = ACTIONS(1222),
    [anon_sym___stdcall] = ACTIONS(1222),
    [anon_sym___fastcall] = ACTIONS(1222),
    [anon_sym___thiscall] = ACTIONS(1222),
    [anon_sym___vectorcall] = ACTIONS(1222),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_signed] = ACTIONS(1222),
    [anon_sym_unsigned] = ACTIONS(1222),
    [anon_sym_long] = ACTIONS(1222),
    [anon_sym_short] = ACTIONS(1222),
    [anon_sym_static] = ACTIONS(1222),
    [anon_sym_auto] = ACTIONS(1222),
    [anon_sym_register] = ACTIONS(1222),
    [anon_sym_inline] = ACTIONS(1222),
    [anon_sym___inline] = ACTIONS(1222),
    [anon_sym___inline__] = ACTIONS(1222),
    [anon_sym___forceinline] = ACTIONS(1222),
    [anon_sym_thread_local] = ACTIONS(1222),
    [anon_sym___thread] = ACTIONS(1222),
    [anon_sym_const] = ACTIONS(1222),
    [anon_sym_constexpr] = ACTIONS(1222),
    [anon_sym_volatile] = ACTIONS(1222),
    [anon_sym_restrict] = ACTIONS(1222),
    [anon_sym___restrict__] = ACTIONS(1222),
    [anon_sym__Atomic] = ACTIONS(1222),
    [anon_sym__Noreturn] = ACTIONS(1222),
    [anon_sym_noreturn] = ACTIONS(1222),
    [anon_sym__Nonnull] = ACTIONS(1222),
    [anon_sym_alignas] = ACTIONS(1222),
    [anon_sym__Alignas] = ACTIONS(1222),
    [sym_primitive_type] = ACTIONS(1222),
    [anon_sym_enum] = ACTIONS(1222),
    [anon_sym_struct] = ACTIONS(1222),
    [anon_sym_union] = ACTIONS(1222),
    [anon_sym_if] = ACTIONS(1222),
    [anon_sym_else] = ACTIONS(1222),
    [anon_sym_switch] = ACTIONS(1222),
    [anon_sym_case] = ACTIONS(1222),
    [anon_sym_default] = ACTIONS(1222),
    [anon_sym_while] = ACTIONS(1222),
    [anon_sym_do] = ACTIONS(1222),
    [anon_sym_for] = ACTIONS(1222),
    [anon_sym_return] = ACTIONS(1222),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_goto] = ACTIONS(1222),
    [anon_sym___try] = ACTIONS(1222),
    [anon_sym___leave] = ACTIONS(1222),
    [anon_sym_DASH_DASH] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1224),
    [anon_sym_sizeof] = ACTIONS(1222),
    [anon_sym___alignof__] = ACTIONS(1222),
    [anon_sym___alignof] = ACTIONS(1222),
    [anon_sym__alignof] = ACTIONS(1222),
    [anon_sym_alignof] = ACTIONS(1222),
    [anon_sym__Alignof] = ACTIONS(1222),
    [anon_sym_offsetof] = ACTIONS(1222),
    [anon_sym__Generic] = ACTIONS(1222),
    [anon_sym_asm] = ACTIONS(1222),
    [anon_sym___asm__] = ACTIONS(1222),
    [anon_sym___asm] = ACTIONS(1222),
    [sym_number_literal] = ACTIONS(1224),
    [anon_sym_L_SQUOTE] = ACTIONS(1224),
    [anon_sym_u_SQUOTE] = ACTIONS(1224),
    [anon_sym_U_SQUOTE] = ACTIONS(1224),
    [anon_sym_u8_SQUOTE] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_L_DQUOTE] = ACTIONS(1224),
    [anon_sym_u_DQUOTE] = ACTIONS(1224),
    [anon_sym_U_DQUOTE] = ACTIONS(1224),
    [anon_sym_u8_DQUOTE] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [sym_true] = ACTIONS(1222),
    [sym_false] = ACTIONS(1222),
    [anon_sym_NULL] = ACTIONS(1222),
    [anon_sym_nullptr] = ACTIONS(1222),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(225)] = {
    [ts_builtin_sym_end] = ACTIONS(1204),
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(226)] = {
    [ts_builtin_sym_end] = ACTIONS(1204),
    [sym_identifier] = ACTIONS(1202),
    [aux_sym_preproc_include_token1] = ACTIONS(1202),
    [aux_sym_preproc_def_token1] = ACTIONS(1202),
    [aux_sym_preproc_if_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1202),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1202),
    [sym_preproc_directive] = ACTIONS(1202),
    [anon_sym_LPAREN2] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1202),
    [anon_sym_PLUS] = ACTIONS(1202),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym___extension__] = ACTIONS(1202),
    [anon_sym_typedef] = ACTIONS(1202),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym___attribute__] = ACTIONS(1202),
    [anon_sym___attribute] = ACTIONS(1202),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1204),
    [anon_sym___declspec] = ACTIONS(1202),
    [anon_sym___cdecl] = ACTIONS(1202),
    [anon_sym___clrcall] = ACTIONS(1202),
    [anon_sym___stdcall] = ACTIONS(1202),
    [anon_sym___fastcall] = ACTIONS(1202),
    [anon_sym___thiscall] = ACTIONS(1202),
    [anon_sym___vectorcall] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_signed] = ACTIONS(1202),
    [anon_sym_unsigned] = ACTIONS(1202),
    [anon_sym_long] = ACTIONS(1202),
    [anon_sym_short] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_auto] = ACTIONS(1202),
    [anon_sym_register] = ACTIONS(1202),
    [anon_sym_inline] = ACTIONS(1202),
    [anon_sym___inline] = ACTIONS(1202),
    [anon_sym___inline__] = ACTIONS(1202),
    [anon_sym___forceinline] = ACTIONS(1202),
    [anon_sym_thread_local] = ACTIONS(1202),
    [anon_sym___thread] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_constexpr] = ACTIONS(1202),
    [anon_sym_volatile] = ACTIONS(1202),
    [anon_sym_restrict] = ACTIONS(1202),
    [anon_sym___restrict__] = ACTIONS(1202),
    [anon_sym__Atomic] = ACTIONS(1202),
    [anon_sym__Noreturn] = ACTIONS(1202),
    [anon_sym_noreturn] = ACTIONS(1202),
    [anon_sym__Nonnull] = ACTIONS(1202),
    [anon_sym_alignas] = ACTIONS(1202),
    [anon_sym__Alignas] = ACTIONS(1202),
    [sym_primitive_type] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_else] = ACTIONS(1202),
    [anon_sym_switch] = ACTIONS(1202),
    [anon_sym_case] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_do] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_goto] = ACTIONS(1202),
    [anon_sym___try] = ACTIONS(1202),
    [anon_sym___leave] = ACTIONS(1202),
    [anon_sym_DASH_DASH] = ACTIONS(1204),
    [anon_sym_PLUS_PLUS] = ACTIONS(1204),
    [anon_sym_sizeof] = ACTIONS(1202),
    [anon_sym___alignof__] = ACTIONS(1202),
    [anon_sym___alignof] = ACTIONS(1202),
    [anon_sym__alignof] = ACTIONS(1202),
    [anon_sym_alignof] = ACTIONS(1202),
    [anon_sym__Alignof] = ACTIONS(1202),
    [anon_sym_offsetof] = ACTIONS(1202),
    [anon_sym__Generic] = ACTIONS(1202),
    [anon_sym_asm] = ACTIONS(1202),
    [anon_sym___asm__] = ACTIONS(1202),
    [anon_sym___asm] = ACTIONS(1202),
    [sym_number_literal] = ACTIONS(1204),
    [anon_sym_L_SQUOTE] = ACTIONS(1204),
    [anon_sym_u_SQUOTE] = ACTIONS(1204),
    [anon_sym_U_SQUOTE] = ACTIONS(1204),
    [anon_sym_u8_SQUOTE] = ACTIONS(1204),
    [anon_sym_SQUOTE] = ACTIONS(1204),
    [anon_sym_L_DQUOTE] = ACTIONS(1204),
    [anon_sym_u_DQUOTE] = ACTIONS(1204),
    [anon_sym_U_DQUOTE] = ACTIONS(1204),
    [anon_sym_u8_DQUOTE] = ACTIONS(1204),
    [anon_sym_DQUOTE] = ACTIONS(1204),
    [sym_true] = ACTIONS(1202),
    [sym_false] = ACTIONS(1202),
    [anon_sym_NULL] = ACTIONS(1202),
    [anon_sym_nullptr] = ACTIONS(1202),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(227)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_RBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(228)] = {
    [sym_identifier] = ACTIONS(1234),
    [aux_sym_preproc_include_token1] = ACTIONS(1234),
    [aux_sym_preproc_def_token1] = ACTIONS(1234),
    [aux_sym_preproc_if_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1234),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1234),
    [sym_preproc_directive] = ACTIONS(1234),
    [anon_sym_LPAREN2] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym___extension__] = ACTIONS(1234),
    [anon_sym_typedef] = ACTIONS(1234),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym___attribute__] = ACTIONS(1234),
    [anon_sym___attribute] = ACTIONS(1234),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1236),
    [anon_sym___declspec] = ACTIONS(1234),
    [anon_sym___cdecl] = ACTIONS(1234),
    [anon_sym___clrcall] = ACTIONS(1234),
    [anon_sym___stdcall] = ACTIONS(1234),
    [anon_sym___fastcall] = ACTIONS(1234),
    [anon_sym___thiscall] = ACTIONS(1234),
    [anon_sym___vectorcall] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_RBRACE] = ACTIONS(1236),
    [anon_sym_signed] = ACTIONS(1234),
    [anon_sym_unsigned] = ACTIONS(1234),
    [anon_sym_long] = ACTIONS(1234),
    [anon_sym_short] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_auto] = ACTIONS(1234),
    [anon_sym_register] = ACTIONS(1234),
    [anon_sym_inline] = ACTIONS(1234),
    [anon_sym___inline] = ACTIONS(1234),
    [anon_sym___inline__] = ACTIONS(1234),
    [anon_sym___forceinline] = ACTIONS(1234),
    [anon_sym_thread_local] = ACTIONS(1234),
    [anon_sym___thread] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_constexpr] = ACTIONS(1234),
    [anon_sym_volatile] = ACTIONS(1234),
    [anon_sym_restrict] = ACTIONS(1234),
    [anon_sym___restrict__] = ACTIONS(1234),
    [anon_sym__Atomic] = ACTIONS(1234),
    [anon_sym__Noreturn] = ACTIONS(1234),
    [anon_sym_noreturn] = ACTIONS(1234),
    [anon_sym__Nonnull] = ACTIONS(1234),
    [anon_sym_alignas] = ACTIONS(1234),
    [anon_sym__Alignas] = ACTIONS(1234),
    [sym_primitive_type] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_else] = ACTIONS(1234),
    [anon_sym_switch] = ACTIONS(1234),
    [anon_sym_case] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_do] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_goto] = ACTIONS(1234),
    [anon_sym___try] = ACTIONS(1234),
    [anon_sym___leave] = ACTIONS(1234),
    [anon_sym_DASH_DASH] = ACTIONS(1236),
    [anon_sym_PLUS_PLUS] = ACTIONS(1236),
    [anon_sym_sizeof] = ACTIONS(1234),
    [anon_sym___alignof__] = ACTIONS(1234),
    [anon_sym___alignof] = ACTIONS(1234),
    [anon_sym__alignof] = ACTIONS(1234),
    [anon_sym_alignof] = ACTIONS(1234),
    [anon_sym__Alignof] = ACTIONS(1234),
    [anon_sym_offsetof] = ACTIONS(1234),
    [anon_sym__Generic] = ACTIONS(1234),
    [anon_sym_asm] = ACTIONS(1234),
    [anon_sym___asm__] = ACTIONS(1234),
    [anon_sym___asm] = ACTIONS(1234),
    [sym_number_literal] = ACTIONS(1236),
    [anon_sym_L_SQUOTE] = ACTIONS(1236),
    [anon_sym_u_SQUOTE] = ACTIONS(1236),
    [anon_sym_U_SQUOTE] = ACTIONS(1236),
    [anon_sym_u8_SQUOTE] = ACTIONS(1236),
    [anon_sym_SQUOTE] = ACTIONS(1236),
    [anon_sym_L_DQUOTE] = ACTIONS(1236),
    [anon_sym_u_DQUOTE] = ACTIONS(1236),
    [anon_sym_U_DQUOTE] = ACTIONS(1236),
    [anon_sym_u8_DQUOTE] = ACTIONS(1236),
    [anon_sym_DQUOTE] = ACTIONS(1236),
    [sym_true] = ACTIONS(1234),
    [sym_false] = ACTIONS(1234),
    [anon_sym_NULL] = ACTIONS(1234),
    [anon_sym_nullptr] = ACTIONS(1234),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(229)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(230)] = {
    [sym_identifier] = ACTIONS(1238),
    [aux_sym_preproc_include_token1] = ACTIONS(1238),
    [aux_sym_preproc_def_token1] = ACTIONS(1238),
    [aux_sym_preproc_if_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1238),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1238),
    [sym_preproc_directive] = ACTIONS(1238),
    [anon_sym_LPAREN2] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1238),
    [anon_sym_PLUS] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym___extension__] = ACTIONS(1238),
    [anon_sym_typedef] = ACTIONS(1238),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym___attribute__] = ACTIONS(1238),
    [anon_sym___attribute] = ACTIONS(1238),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1240),
    [anon_sym___declspec] = ACTIONS(1238),
    [anon_sym___cdecl] = ACTIONS(1238),
    [anon_sym___clrcall] = ACTIONS(1238),
    [anon_sym___stdcall] = ACTIONS(1238),
    [anon_sym___fastcall] = ACTIONS(1238),
    [anon_sym___thiscall] = ACTIONS(1238),
    [anon_sym___vectorcall] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_signed] = ACTIONS(1238),
    [anon_sym_unsigned] = ACTIONS(1238),
    [anon_sym_long] = ACTIONS(1238),
    [anon_sym_short] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_auto] = ACTIONS(1238),
    [anon_sym_register] = ACTIONS(1238),
    [anon_sym_inline] = ACTIONS(1238),
    [anon_sym___inline] = ACTIONS(1238),
    [anon_sym___inline__] = ACTIONS(1238),
    [anon_sym___forceinline] = ACTIONS(1238),
    [anon_sym_thread_local] = ACTIONS(1238),
    [anon_sym___thread] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_constexpr] = ACTIONS(1238),
    [anon_sym_volatile] = ACTIONS(1238),
    [anon_sym_restrict] = ACTIONS(1238),
    [anon_sym___restrict__] = ACTIONS(1238),
    [anon_sym__Atomic] = ACTIONS(1238),
    [anon_sym__Noreturn] = ACTIONS(1238),
    [anon_sym_noreturn] = ACTIONS(1238),
    [anon_sym__Nonnull] = ACTIONS(1238),
    [anon_sym_alignas] = ACTIONS(1238),
    [anon_sym__Alignas] = ACTIONS(1238),
    [sym_primitive_type] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_else] = ACTIONS(1238),
    [anon_sym_switch] = ACTIONS(1238),
    [anon_sym_case] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_do] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_goto] = ACTIONS(1238),
    [anon_sym___try] = ACTIONS(1238),
    [anon_sym___leave] = ACTIONS(1238),
    [anon_sym_DASH_DASH] = ACTIONS(1240),
    [anon_sym_PLUS_PLUS] = ACTIONS(1240),
    [anon_sym_sizeof] = ACTIONS(1238),
    [anon_sym___alignof__] = ACTIONS(1238),
    [anon_sym___alignof] = ACTIONS(1238),
    [anon_sym__alignof] = ACTIONS(1238),
    [anon_sym_alignof] = ACTIONS(1238),
    [anon_sym__Alignof] = ACTIONS(1238),
    [anon_sym_offsetof] = ACTIONS(1238),
    [anon_sym__Generic] = ACTIONS(1238),
    [anon_sym_asm] = ACTIONS(1238),
    [anon_sym___asm__] = ACTIONS(1238),
    [anon_sym___asm] = ACTIONS(1238),
    [sym_number_literal] = ACTIONS(1240),
    [anon_sym_L_SQUOTE] = ACTIONS(1240),
    [anon_sym_u_SQUOTE] = ACTIONS(1240),
    [anon_sym_U_SQUOTE] = ACTIONS(1240),
    [anon_sym_u8_SQUOTE] = ACTIONS(1240),
    [anon_sym_SQUOTE] = ACTIONS(1240),
    [anon_sym_L_DQUOTE] = ACTIONS(1240),
    [anon_sym_u_DQUOTE] = ACTIONS(1240),
    [anon_sym_U_DQUOTE] = ACTIONS(1240),
    [anon_sym_u8_DQUOTE] = ACTIONS(1240),
    [anon_sym_DQUOTE] = ACTIONS(1240),
    [sym_true] = ACTIONS(1238),
    [sym_false] = ACTIONS(1238),
    [anon_sym_NULL] = ACTIONS(1238),
    [anon_sym_nullptr] = ACTIONS(1238),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(231)] = {
    [sym_identifier] = ACTIONS(1250),
    [aux_sym_preproc_include_token1] = ACTIONS(1250),
    [aux_sym_preproc_def_token1] = ACTIONS(1250),
    [aux_sym_preproc_if_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1250),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1250),
    [sym_preproc_directive] = ACTIONS(1250),
    [anon_sym_LPAREN2] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1250),
    [anon_sym_PLUS] = ACTIONS(1250),
    [anon_sym_STAR] = ACTIONS(1252),
    [anon_sym_AMP] = ACTIONS(1252),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym___extension__] = ACTIONS(1250),
    [anon_sym_typedef] = ACTIONS(1250),
    [anon_sym_extern] = ACTIONS(1250),
    [anon_sym___attribute__] = ACTIONS(1250),
    [anon_sym___attribute] = ACTIONS(1250),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1252),
    [anon_sym___declspec] = ACTIONS(1250),
    [anon_sym___cdecl] = ACTIONS(1250),
    [anon_sym___clrcall] = ACTIONS(1250),
    [anon_sym___stdcall] = ACTIONS(1250),
    [anon_sym___fastcall] = ACTIONS(1250),
    [anon_sym___thiscall] = ACTIONS(1250),
    [anon_sym___vectorcall] = ACTIONS(1250),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [anon_sym_signed] = ACTIONS(1250),
    [anon_sym_unsigned] = ACTIONS(1250),
    [anon_sym_long] = ACTIONS(1250),
    [anon_sym_short] = ACTIONS(1250),
    [anon_sym_static] = ACTIONS(1250),
    [anon_sym_auto] = ACTIONS(1250),
    [anon_sym_register] = ACTIONS(1250),
    [anon_sym_inline] = ACTIONS(1250),
    [anon_sym___inline] = ACTIONS(1250),
    [anon_sym___inline__] = ACTIONS(1250),
    [anon_sym___forceinline] = ACTIONS(1250),
    [anon_sym_thread_local] = ACTIONS(1250),
    [anon_sym___thread] = ACTIONS(1250),
    [anon_sym_const] = ACTIONS(1250),
    [anon_sym_constexpr] = ACTIONS(1250),
    [anon_sym_volatile] = ACTIONS(1250),
    [anon_sym_restrict] = ACTIONS(1250),
    [anon_sym___restrict__] = ACTIONS(1250),
    [anon_sym__Atomic] = ACTIONS(1250),
    [anon_sym__Noreturn] = ACTIONS(1250),
    [anon_sym_noreturn] = ACTIONS(1250),
    [anon_sym__Nonnull] = ACTIONS(1250),
    [anon_sym_alignas] = ACTIONS(1250),
    [anon_sym__Alignas] = ACTIONS(1250),
    [sym_primitive_type] = ACTIONS(1250),
    [anon_sym_enum] = ACTIONS(1250),
    [anon_sym_struct] = ACTIONS(1250),
    [anon_sym_union] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(1250),
    [anon_sym_else] = ACTIONS(1250),
    [anon_sym_switch] = ACTIONS(1250),
    [anon_sym_case] = ACTIONS(1250),
    [anon_sym_default] = ACTIONS(1250),
    [anon_sym_while] = ACTIONS(1250),
    [anon_sym_do] = ACTIONS(1250),
    [anon_sym_for] = ACTIONS(1250),
    [anon_sym_return] = ACTIONS(1250),
    [anon_sym_break] = ACTIONS(1250),
    [anon_sym_continue] = ACTIONS(1250),
    [anon_sym_goto] = ACTIONS(1250),
    [anon_sym___try] = ACTIONS(1250),
    [anon_sym___leave] = ACTIONS(1250),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_sizeof] = ACTIONS(1250),
    [anon_sym___alignof__] = ACTIONS(1250),
    [anon_sym___alignof] = ACTIONS(1250),
    [anon_sym__alignof] = ACTIONS(1250),
    [anon_sym_alignof] = ACTIONS(1250),
    [anon_sym__Alignof] = ACTIONS(1250),
    [anon_sym_offsetof] = ACTIONS(1250),
    [anon_sym__Generic] = ACTIONS(1250),
    [anon_sym_asm] = ACTIONS(1250),
    [anon_sym___asm__] = ACTIONS(1250),
    [anon_sym___asm] = ACTIONS(1250),
    [sym_number_literal] = ACTIONS(1252),
    [anon_sym_L_SQUOTE] = ACTIONS(1252),
    [anon_sym_u_SQUOTE] = ACTIONS(1252),
    [anon_sym_U_SQUOTE] = ACTIONS(1252),
    [anon_sym_u8_SQUOTE] = ACTIONS(1252),
    [anon_sym_SQUOTE] = ACTIONS(1252),
    [anon_sym_L_DQUOTE] = ACTIONS(1252),
    [anon_sym_u_DQUOTE] = ACTIONS(1252),
    [anon_sym_U_DQUOTE] = ACTIONS(1252),
    [anon_sym_u8_DQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1252),
    [sym_true] = ACTIONS(1250),
    [sym_false] = ACTIONS(1250),
    [anon_sym_NULL] = ACTIONS(1250),
    [anon_sym_nullptr] = ACTIONS(1250),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(232)] = {
    [sym_identifier] = ACTIONS(1186),
    [aux_sym_preproc_include_token1] = ACTIONS(1186),
    [aux_sym_preproc_def_token1] = ACTIONS(1186),
    [aux_sym_preproc_if_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1186),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1186),
    [sym_preproc_directive] = ACTIONS(1186),
    [anon_sym_LPAREN2] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1186),
    [anon_sym_PLUS] = ACTIONS(1186),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym___extension__] = ACTIONS(1186),
    [anon_sym_typedef] = ACTIONS(1186),
    [anon_sym_extern] = ACTIONS(1186),
    [anon_sym___attribute__] = ACTIONS(1186),
    [anon_sym___attribute] = ACTIONS(1186),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1188),
    [anon_sym___declspec] = ACTIONS(1186),
    [anon_sym___cdecl] = ACTIONS(1186),
    [anon_sym___clrcall] = ACTIONS(1186),
    [anon_sym___stdcall] = ACTIONS(1186),
    [anon_sym___fastcall] = ACTIONS(1186),
    [anon_sym___thiscall] = ACTIONS(1186),
    [anon_sym___vectorcall] = ACTIONS(1186),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_signed] = ACTIONS(1186),
    [anon_sym_unsigned] = ACTIONS(1186),
    [anon_sym_long] = ACTIONS(1186),
    [anon_sym_short] = ACTIONS(1186),
    [anon_sym_static] = ACTIONS(1186),
    [anon_sym_auto] = ACTIONS(1186),
    [anon_sym_register] = ACTIONS(1186),
    [anon_sym_inline] = ACTIONS(1186),
    [anon_sym___inline] = ACTIONS(1186),
    [anon_sym___inline__] = ACTIONS(1186),
    [anon_sym___forceinline] = ACTIONS(1186),
    [anon_sym_thread_local] = ACTIONS(1186),
    [anon_sym___thread] = ACTIONS(1186),
    [anon_sym_const] = ACTIONS(1186),
    [anon_sym_constexpr] = ACTIONS(1186),
    [anon_sym_volatile] = ACTIONS(1186),
    [anon_sym_restrict] = ACTIONS(1186),
    [anon_sym___restrict__] = ACTIONS(1186),
    [anon_sym__Atomic] = ACTIONS(1186),
    [anon_sym__Noreturn] = ACTIONS(1186),
    [anon_sym_noreturn] = ACTIONS(1186),
    [anon_sym__Nonnull] = ACTIONS(1186),
    [anon_sym_alignas] = ACTIONS(1186),
    [anon_sym__Alignas] = ACTIONS(1186),
    [sym_primitive_type] = ACTIONS(1186),
    [anon_sym_enum] = ACTIONS(1186),
    [anon_sym_struct] = ACTIONS(1186),
    [anon_sym_union] = ACTIONS(1186),
    [anon_sym_if] = ACTIONS(1186),
    [anon_sym_else] = ACTIONS(1186),
    [anon_sym_switch] = ACTIONS(1186),
    [anon_sym_case] = ACTIONS(1186),
    [anon_sym_default] = ACTIONS(1186),
    [anon_sym_while] = ACTIONS(1186),
    [anon_sym_do] = ACTIONS(1186),
    [anon_sym_for] = ACTIONS(1186),
    [anon_sym_return] = ACTIONS(1186),
    [anon_sym_break] = ACTIONS(1186),
    [anon_sym_continue] = ACTIONS(1186),
    [anon_sym_goto] = ACTIONS(1186),
    [anon_sym___try] = ACTIONS(1186),
    [anon_sym___leave] = ACTIONS(1186),
    [anon_sym_DASH_DASH] = ACTIONS(1188),
    [anon_sym_PLUS_PLUS] = ACTIONS(1188),
    [anon_sym_sizeof] = ACTIONS(1186),
    [anon_sym___alignof__] = ACTIONS(1186),
    [anon_sym___alignof] = ACTIONS(1186),
    [anon_sym__alignof] = ACTIONS(1186),
    [anon_sym_alignof] = ACTIONS(1186),
    [anon_sym__Alignof] = ACTIONS(1186),
    [anon_sym_offsetof] = ACTIONS(1186),
    [anon_sym__Generic] = ACTIONS(1186),
    [anon_sym_asm] = ACTIONS(1186),
    [anon_sym___asm__] = ACTIONS(1186),
    [anon_sym___asm] = ACTIONS(1186),
    [sym_number_literal] = ACTIONS(1188),
    [anon_sym_L_SQUOTE] = ACTIONS(1188),
    [anon_sym_u_SQUOTE] = ACTIONS(1188),
    [anon_sym_U_SQUOTE] = ACTIONS(1188),
    [anon_sym_u8_SQUOTE] = ACTIONS(1188),
    [anon_sym_SQUOTE] = ACTIONS(1188),
    [anon_sym_L_DQUOTE] = ACTIONS(1188),
    [anon_sym_u_DQUOTE] = ACTIONS(1188),
    [anon_sym_U_DQUOTE] = ACTIONS(1188),
    [anon_sym_u8_DQUOTE] = ACTIONS(1188),
    [anon_sym_DQUOTE] = ACTIONS(1188),
    [sym_true] = ACTIONS(1186),
    [sym_false] = ACTIONS(1186),
    [anon_sym_NULL] = ACTIONS(1186),
    [anon_sym_nullptr] = ACTIONS(1186),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(233)] = {
    [ts_builtin_sym_end] = ACTIONS(1148),
    [sym_identifier] = ACTIONS(1146),
    [aux_sym_preproc_include_token1] = ACTIONS(1146),
    [aux_sym_preproc_def_token1] = ACTIONS(1146),
    [aux_sym_preproc_if_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1146),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1146),
    [sym_preproc_directive] = ACTIONS(1146),
    [anon_sym_LPAREN2] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_TILDE] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1146),
    [anon_sym_PLUS] = ACTIONS(1146),
    [anon_sym_STAR] = ACTIONS(1148),
    [anon_sym_AMP] = ACTIONS(1148),
    [anon_sym_SEMI] = ACTIONS(1148),
    [anon_sym___extension__] = ACTIONS(1146),
    [anon_sym_typedef] = ACTIONS(1146),
    [anon_sym_extern] = ACTIONS(1146),
    [anon_sym___attribute__] = ACTIONS(1146),
    [anon_sym___attribute] = ACTIONS(1146),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1148),
    [anon_sym___declspec] = ACTIONS(1146),
    [anon_sym___cdecl] = ACTIONS(1146),
    [anon_sym___clrcall] = ACTIONS(1146),
    [anon_sym___stdcall] = ACTIONS(1146),
    [anon_sym___fastcall] = ACTIONS(1146),
    [anon_sym___thiscall] = ACTIONS(1146),
    [anon_sym___vectorcall] = ACTIONS(1146),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_signed] = ACTIONS(1146),
    [anon_sym_unsigned] = ACTIONS(1146),
    [anon_sym_long] = ACTIONS(1146),
    [anon_sym_short] = ACTIONS(1146),
    [anon_sym_static] = ACTIONS(1146),
    [anon_sym_auto] = ACTIONS(1146),
    [anon_sym_register] = ACTIONS(1146),
    [anon_sym_inline] = ACTIONS(1146),
    [anon_sym___inline] = ACTIONS(1146),
    [anon_sym___inline__] = ACTIONS(1146),
    [anon_sym___forceinline] = ACTIONS(1146),
    [anon_sym_thread_local] = ACTIONS(1146),
    [anon_sym___thread] = ACTIONS(1146),
    [anon_sym_const] = ACTIONS(1146),
    [anon_sym_constexpr] = ACTIONS(1146),
    [anon_sym_volatile] = ACTIONS(1146),
    [anon_sym_restrict] = ACTIONS(1146),
    [anon_sym___restrict__] = ACTIONS(1146),
    [anon_sym__Atomic] = ACTIONS(1146),
    [anon_sym__Noreturn] = ACTIONS(1146),
    [anon_sym_noreturn] = ACTIONS(1146),
    [anon_sym__Nonnull] = ACTIONS(1146),
    [anon_sym_alignas] = ACTIONS(1146),
    [anon_sym__Alignas] = ACTIONS(1146),
    [sym_primitive_type] = ACTIONS(1146),
    [anon_sym_enum] = ACTIONS(1146),
    [anon_sym_struct] = ACTIONS(1146),
    [anon_sym_union] = ACTIONS(1146),
    [anon_sym_if] = ACTIONS(1146),
    [anon_sym_else] = ACTIONS(1146),
    [anon_sym_switch] = ACTIONS(1146),
    [anon_sym_case] = ACTIONS(1146),
    [anon_sym_default] = ACTIONS(1146),
    [anon_sym_while] = ACTIONS(1146),
    [anon_sym_do] = ACTIONS(1146),
    [anon_sym_for] = ACTIONS(1146),
    [anon_sym_return] = ACTIONS(1146),
    [anon_sym_break] = ACTIONS(1146),
    [anon_sym_continue] = ACTIONS(1146),
    [anon_sym_goto] = ACTIONS(1146),
    [anon_sym___try] = ACTIONS(1146),
    [anon_sym___leave] = ACTIONS(1146),
    [anon_sym_DASH_DASH] = ACTIONS(1148),
    [anon_sym_PLUS_PLUS] = ACTIONS(1148),
    [anon_sym_sizeof] = ACTIONS(1146),
    [anon_sym___alignof__] = ACTIONS(1146),
    [anon_sym___alignof] = ACTIONS(1146),
    [anon_sym__alignof] = ACTIONS(1146),
    [anon_sym_alignof] = ACTIONS(1146),
    [anon_sym__Alignof] = ACTIONS(1146),
    [anon_sym_offsetof] = ACTIONS(1146),
    [anon_sym__Generic] = ACTIONS(1146),
    [anon_sym_asm] = ACTIONS(1146),
    [anon_sym___asm__] = ACTIONS(1146),
    [anon_sym___asm] = ACTIONS(1146),
    [sym_number_literal] = ACTIONS(1148),
    [anon_sym_L_SQUOTE] = ACTIONS(1148),
    [anon_sym_u_SQUOTE] = ACTIONS(1148),
    [anon_sym_U_SQUOTE] = ACTIONS(1148),
    [anon_sym_u8_SQUOTE] = ACTIONS(1148),
    [anon_sym_SQUOTE] = ACTIONS(1148),
    [anon_sym_L_DQUOTE] = ACTIONS(1148),
    [anon_sym_u_DQUOTE] = ACTIONS(1148),
    [anon_sym_U_DQUOTE] = ACTIONS(1148),
    [anon_sym_u8_DQUOTE] = ACTIONS(1148),
    [anon_sym_DQUOTE] = ACTIONS(1148),
    [sym_true] = ACTIONS(1146),
    [sym_false] = ACTIONS(1146),
    [anon_sym_NULL] = ACTIONS(1146),
    [anon_sym_nullptr] = ACTIONS(1146),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(234)] = {
    [sym_identifier] = ACTIONS(1150),
    [aux_sym_preproc_include_token1] = ACTIONS(1150),
    [aux_sym_preproc_def_token1] = ACTIONS(1150),
    [aux_sym_preproc_if_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1150),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1150),
    [sym_preproc_directive] = ACTIONS(1150),
    [anon_sym_LPAREN2] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_TILDE] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1150),
    [anon_sym_PLUS] = ACTIONS(1150),
    [anon_sym_STAR] = ACTIONS(1152),
    [anon_sym_AMP] = ACTIONS(1152),
    [anon_sym_SEMI] = ACTIONS(1152),
    [anon_sym___extension__] = ACTIONS(1150),
    [anon_sym_typedef] = ACTIONS(1150),
    [anon_sym_extern] = ACTIONS(1150),
    [anon_sym___attribute__] = ACTIONS(1150),
    [anon_sym___attribute] = ACTIONS(1150),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1152),
    [anon_sym___declspec] = ACTIONS(1150),
    [anon_sym___cdecl] = ACTIONS(1150),
    [anon_sym___clrcall] = ACTIONS(1150),
    [anon_sym___stdcall] = ACTIONS(1150),
    [anon_sym___fastcall] = ACTIONS(1150),
    [anon_sym___thiscall] = ACTIONS(1150),
    [anon_sym___vectorcall] = ACTIONS(1150),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_RBRACE] = ACTIONS(1152),
    [anon_sym_signed] = ACTIONS(1150),
    [anon_sym_unsigned] = ACTIONS(1150),
    [anon_sym_long] = ACTIONS(1150),
    [anon_sym_short] = ACTIONS(1150),
    [anon_sym_static] = ACTIONS(1150),
    [anon_sym_auto] = ACTIONS(1150),
    [anon_sym_register] = ACTIONS(1150),
    [anon_sym_inline] = ACTIONS(1150),
    [anon_sym___inline] = ACTIONS(1150),
    [anon_sym___inline__] = ACTIONS(1150),
    [anon_sym___forceinline] = ACTIONS(1150),
    [anon_sym_thread_local] = ACTIONS(1150),
    [anon_sym___thread] = ACTIONS(1150),
    [anon_sym_const] = ACTIONS(1150),
    [anon_sym_constexpr] = ACTIONS(1150),
    [anon_sym_volatile] = ACTIONS(1150),
    [anon_sym_restrict] = ACTIONS(1150),
    [anon_sym___restrict__] = ACTIONS(1150),
    [anon_sym__Atomic] = ACTIONS(1150),
    [anon_sym__Noreturn] = ACTIONS(1150),
    [anon_sym_noreturn] = ACTIONS(1150),
    [anon_sym__Nonnull] = ACTIONS(1150),
    [anon_sym_alignas] = ACTIONS(1150),
    [anon_sym__Alignas] = ACTIONS(1150),
    [sym_primitive_type] = ACTIONS(1150),
    [anon_sym_enum] = ACTIONS(1150),
    [anon_sym_struct] = ACTIONS(1150),
    [anon_sym_union] = ACTIONS(1150),
    [anon_sym_if] = ACTIONS(1150),
    [anon_sym_else] = ACTIONS(1150),
    [anon_sym_switch] = ACTIONS(1150),
    [anon_sym_case] = ACTIONS(1150),
    [anon_sym_default] = ACTIONS(1150),
    [anon_sym_while] = ACTIONS(1150),
    [anon_sym_do] = ACTIONS(1150),
    [anon_sym_for] = ACTIONS(1150),
    [anon_sym_return] = ACTIONS(1150),
    [anon_sym_break] = ACTIONS(1150),
    [anon_sym_continue] = ACTIONS(1150),
    [anon_sym_goto] = ACTIONS(1150),
    [anon_sym___try] = ACTIONS(1150),
    [anon_sym___leave] = ACTIONS(1150),
    [anon_sym_DASH_DASH] = ACTIONS(1152),
    [anon_sym_PLUS_PLUS] = ACTIONS(1152),
    [anon_sym_sizeof] = ACTIONS(1150),
    [anon_sym___alignof__] = ACTIONS(1150),
    [anon_sym___alignof] = ACTIONS(1150),
    [anon_sym__alignof] = ACTIONS(1150),
    [anon_sym_alignof] = ACTIONS(1150),
    [anon_sym__Alignof] = ACTIONS(1150),
    [anon_sym_offsetof] = ACTIONS(1150),
    [anon_sym__Generic] = ACTIONS(1150),
    [anon_sym_asm] = ACTIONS(1150),
    [anon_sym___asm__] = ACTIONS(1150),
    [anon_sym___asm] = ACTIONS(1150),
    [sym_number_literal] = ACTIONS(1152),
    [anon_sym_L_SQUOTE] = ACTIONS(1152),
    [anon_sym_u_SQUOTE] = ACTIONS(1152),
    [anon_sym_U_SQUOTE] = ACTIONS(1152),
    [anon_sym_u8_SQUOTE] = ACTIONS(1152),
    [anon_sym_SQUOTE] = ACTIONS(1152),
    [anon_sym_L_DQUOTE] = ACTIONS(1152),
    [anon_sym_u_DQUOTE] = ACTIONS(1152),
    [anon_sym_U_DQUOTE] = ACTIONS(1152),
    [anon_sym_u8_DQUOTE] = ACTIONS(1152),
    [anon_sym_DQUOTE] = ACTIONS(1152),
    [sym_true] = ACTIONS(1150),
    [sym_false] = ACTIONS(1150),
    [anon_sym_NULL] = ACTIONS(1150),
    [anon_sym_nullptr] = ACTIONS(1150),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(235)] = {
    [sym_identifier] = ACTIONS(1166),
    [aux_sym_preproc_include_token1] = ACTIONS(1166),
    [aux_sym_preproc_def_token1] = ACTIONS(1166),
    [aux_sym_preproc_if_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1166),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1166),
    [sym_preproc_directive] = ACTIONS(1166),
    [anon_sym_LPAREN2] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_TILDE] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1166),
    [anon_sym_PLUS] = ACTIONS(1166),
    [anon_sym_STAR] = ACTIONS(1168),
    [anon_sym_AMP] = ACTIONS(1168),
    [anon_sym_SEMI] = ACTIONS(1168),
    [anon_sym___extension__] = ACTIONS(1166),
    [anon_sym_typedef] = ACTIONS(1166),
    [anon_sym_extern] = ACTIONS(1166),
    [anon_sym___attribute__] = ACTIONS(1166),
    [anon_sym___attribute] = ACTIONS(1166),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1168),
    [anon_sym___declspec] = ACTIONS(1166),
    [anon_sym___cdecl] = ACTIONS(1166),
    [anon_sym___clrcall] = ACTIONS(1166),
    [anon_sym___stdcall] = ACTIONS(1166),
    [anon_sym___fastcall] = ACTIONS(1166),
    [anon_sym___thiscall] = ACTIONS(1166),
    [anon_sym___vectorcall] = ACTIONS(1166),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_RBRACE] = ACTIONS(1168),
    [anon_sym_signed] = ACTIONS(1166),
    [anon_sym_unsigned] = ACTIONS(1166),
    [anon_sym_long] = ACTIONS(1166),
    [anon_sym_short] = ACTIONS(1166),
    [anon_sym_static] = ACTIONS(1166),
    [anon_sym_auto] = ACTIONS(1166),
    [anon_sym_register] = ACTIONS(1166),
    [anon_sym_inline] = ACTIONS(1166),
    [anon_sym___inline] = ACTIONS(1166),
    [anon_sym___inline__] = ACTIONS(1166),
    [anon_sym___forceinline] = ACTIONS(1166),
    [anon_sym_thread_local] = ACTIONS(1166),
    [anon_sym___thread] = ACTIONS(1166),
    [anon_sym_const] = ACTIONS(1166),
    [anon_sym_constexpr] = ACTIONS(1166),
    [anon_sym_volatile] = ACTIONS(1166),
    [anon_sym_restrict] = ACTIONS(1166),
    [anon_sym___restrict__] = ACTIONS(1166),
    [anon_sym__Atomic] = ACTIONS(1166),
    [anon_sym__Noreturn] = ACTIONS(1166),
    [anon_sym_noreturn] = ACTIONS(1166),
    [anon_sym__Nonnull] = ACTIONS(1166),
    [anon_sym_alignas] = ACTIONS(1166),
    [anon_sym__Alignas] = ACTIONS(1166),
    [sym_primitive_type] = ACTIONS(1166),
    [anon_sym_enum] = ACTIONS(1166),
    [anon_sym_struct] = ACTIONS(1166),
    [anon_sym_union] = ACTIONS(1166),
    [anon_sym_if] = ACTIONS(1166),
    [anon_sym_else] = ACTIONS(1166),
    [anon_sym_switch] = ACTIONS(1166),
    [anon_sym_case] = ACTIONS(1166),
    [anon_sym_default] = ACTIONS(1166),
    [anon_sym_while] = ACTIONS(1166),
    [anon_sym_do] = ACTIONS(1166),
    [anon_sym_for] = ACTIONS(1166),
    [anon_sym_return] = ACTIONS(1166),
    [anon_sym_break] = ACTIONS(1166),
    [anon_sym_continue] = ACTIONS(1166),
    [anon_sym_goto] = ACTIONS(1166),
    [anon_sym___try] = ACTIONS(1166),
    [anon_sym___leave] = ACTIONS(1166),
    [anon_sym_DASH_DASH] = ACTIONS(1168),
    [anon_sym_PLUS_PLUS] = ACTIONS(1168),
    [anon_sym_sizeof] = ACTIONS(1166),
    [anon_sym___alignof__] = ACTIONS(1166),
    [anon_sym___alignof] = ACTIONS(1166),
    [anon_sym__alignof] = ACTIONS(1166),
    [anon_sym_alignof] = ACTIONS(1166),
    [anon_sym__Alignof] = ACTIONS(1166),
    [anon_sym_offsetof] = ACTIONS(1166),
    [anon_sym__Generic] = ACTIONS(1166),
    [anon_sym_asm] = ACTIONS(1166),
    [anon_sym___asm__] = ACTIONS(1166),
    [anon_sym___asm] = ACTIONS(1166),
    [sym_number_literal] = ACTIONS(1168),
    [anon_sym_L_SQUOTE] = ACTIONS(1168),
    [anon_sym_u_SQUOTE] = ACTIONS(1168),
    [anon_sym_U_SQUOTE] = ACTIONS(1168),
    [anon_sym_u8_SQUOTE] = ACTIONS(1168),
    [anon_sym_SQUOTE] = ACTIONS(1168),
    [anon_sym_L_DQUOTE] = ACTIONS(1168),
    [anon_sym_u_DQUOTE] = ACTIONS(1168),
    [anon_sym_U_DQUOTE] = ACTIONS(1168),
    [anon_sym_u8_DQUOTE] = ACTIONS(1168),
    [anon_sym_DQUOTE] = ACTIONS(1168),
    [sym_true] = ACTIONS(1166),
    [sym_false] = ACTIONS(1166),
    [anon_sym_NULL] = ACTIONS(1166),
    [anon_sym_nullptr] = ACTIONS(1166),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(236)] = {
    [sym_identifier] = ACTIONS(1170),
    [aux_sym_preproc_include_token1] = ACTIONS(1170),
    [aux_sym_preproc_def_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1170),
    [sym_preproc_directive] = ACTIONS(1170),
    [anon_sym_LPAREN2] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_TILDE] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1172),
    [anon_sym_AMP] = ACTIONS(1172),
    [anon_sym_SEMI] = ACTIONS(1172),
    [anon_sym___extension__] = ACTIONS(1170),
    [anon_sym_typedef] = ACTIONS(1170),
    [anon_sym_extern] = ACTIONS(1170),
    [anon_sym___attribute__] = ACTIONS(1170),
    [anon_sym___attribute] = ACTIONS(1170),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1172),
    [anon_sym___declspec] = ACTIONS(1170),
    [anon_sym___cdecl] = ACTIONS(1170),
    [anon_sym___clrcall] = ACTIONS(1170),
    [anon_sym___stdcall] = ACTIONS(1170),
    [anon_sym___fastcall] = ACTIONS(1170),
    [anon_sym___thiscall] = ACTIONS(1170),
    [anon_sym___vectorcall] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_RBRACE] = ACTIONS(1172),
    [anon_sym_signed] = ACTIONS(1170),
    [anon_sym_unsigned] = ACTIONS(1170),
    [anon_sym_long] = ACTIONS(1170),
    [anon_sym_short] = ACTIONS(1170),
    [anon_sym_static] = ACTIONS(1170),
    [anon_sym_auto] = ACTIONS(1170),
    [anon_sym_register] = ACTIONS(1170),
    [anon_sym_inline] = ACTIONS(1170),
    [anon_sym___inline] = ACTIONS(1170),
    [anon_sym___inline__] = ACTIONS(1170),
    [anon_sym___forceinline] = ACTIONS(1170),
    [anon_sym_thread_local] = ACTIONS(1170),
    [anon_sym___thread] = ACTIONS(1170),
    [anon_sym_const] = ACTIONS(1170),
    [anon_sym_constexpr] = ACTIONS(1170),
    [anon_sym_volatile] = ACTIONS(1170),
    [anon_sym_restrict] = ACTIONS(1170),
    [anon_sym___restrict__] = ACTIONS(1170),
    [anon_sym__Atomic] = ACTIONS(1170),
    [anon_sym__Noreturn] = ACTIONS(1170),
    [anon_sym_noreturn] = ACTIONS(1170),
    [anon_sym__Nonnull] = ACTIONS(1170),
    [anon_sym_alignas] = ACTIONS(1170),
    [anon_sym__Alignas] = ACTIONS(1170),
    [sym_primitive_type] = ACTIONS(1170),
    [anon_sym_enum] = ACTIONS(1170),
    [anon_sym_struct] = ACTIONS(1170),
    [anon_sym_union] = ACTIONS(1170),
    [anon_sym_if] = ACTIONS(1170),
    [anon_sym_else] = ACTIONS(1170),
    [anon_sym_switch] = ACTIONS(1170),
    [anon_sym_case] = ACTIONS(1170),
    [anon_sym_default] = ACTIONS(1170),
    [anon_sym_while] = ACTIONS(1170),
    [anon_sym_do] = ACTIONS(1170),
    [anon_sym_for] = ACTIONS(1170),
    [anon_sym_return] = ACTIONS(1170),
    [anon_sym_break] = ACTIONS(1170),
    [anon_sym_continue] = ACTIONS(1170),
    [anon_sym_goto] = ACTIONS(1170),
    [anon_sym___try] = ACTIONS(1170),
    [anon_sym___leave] = ACTIONS(1170),
    [anon_sym_DASH_DASH] = ACTIONS(1172),
    [anon_sym_PLUS_PLUS] = ACTIONS(1172),
    [anon_sym_sizeof] = ACTIONS(1170),
    [anon_sym___alignof__] = ACTIONS(1170),
    [anon_sym___alignof] = ACTIONS(1170),
    [anon_sym__alignof] = ACTIONS(1170),
    [anon_sym_alignof] = ACTIONS(1170),
    [anon_sym__Alignof] = ACTIONS(1170),
    [anon_sym_offsetof] = ACTIONS(1170),
    [anon_sym__Generic] = ACTIONS(1170),
    [anon_sym_asm] = ACTIONS(1170),
    [anon_sym___asm__] = ACTIONS(1170),
    [anon_sym___asm] = ACTIONS(1170),
    [sym_number_literal] = ACTIONS(1172),
    [anon_sym_L_SQUOTE] = ACTIONS(1172),
    [anon_sym_u_SQUOTE] = ACTIONS(1172),
    [anon_sym_U_SQUOTE] = ACTIONS(1172),
    [anon_sym_u8_SQUOTE] = ACTIONS(1172),
    [anon_sym_SQUOTE] = ACTIONS(1172),
    [anon_sym_L_DQUOTE] = ACTIONS(1172),
    [anon_sym_u_DQUOTE] = ACTIONS(1172),
    [anon_sym_U_DQUOTE] = ACTIONS(1172),
    [anon_sym_u8_DQUOTE] = ACTIONS(1172),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [sym_true] = ACTIONS(1170),
    [sym_false] = ACTIONS(1170),
    [anon_sym_NULL] = ACTIONS(1170),
    [anon_sym_nullptr] = ACTIONS(1170),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(237)] = {
    [sym_identifier] = ACTIONS(1174),
    [aux_sym_preproc_include_token1] = ACTIONS(1174),
    [aux_sym_preproc_def_token1] = ACTIONS(1174),
    [aux_sym_preproc_if_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1174),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1174),
    [sym_preproc_directive] = ACTIONS(1174),
    [anon_sym_LPAREN2] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_TILDE] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1174),
    [anon_sym_PLUS] = ACTIONS(1174),
    [anon_sym_STAR] = ACTIONS(1176),
    [anon_sym_AMP] = ACTIONS(1176),
    [anon_sym_SEMI] = ACTIONS(1176),
    [anon_sym___extension__] = ACTIONS(1174),
    [anon_sym_typedef] = ACTIONS(1174),
    [anon_sym_extern] = ACTIONS(1174),
    [anon_sym___attribute__] = ACTIONS(1174),
    [anon_sym___attribute] = ACTIONS(1174),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1176),
    [anon_sym___declspec] = ACTIONS(1174),
    [anon_sym___cdecl] = ACTIONS(1174),
    [anon_sym___clrcall] = ACTIONS(1174),
    [anon_sym___stdcall] = ACTIONS(1174),
    [anon_sym___fastcall] = ACTIONS(1174),
    [anon_sym___thiscall] = ACTIONS(1174),
    [anon_sym___vectorcall] = ACTIONS(1174),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_RBRACE] = ACTIONS(1176),
    [anon_sym_signed] = ACTIONS(1174),
    [anon_sym_unsigned] = ACTIONS(1174),
    [anon_sym_long] = ACTIONS(1174),
    [anon_sym_short] = ACTIONS(1174),
    [anon_sym_static] = ACTIONS(1174),
    [anon_sym_auto] = ACTIONS(1174),
    [anon_sym_register] = ACTIONS(1174),
    [anon_sym_inline] = ACTIONS(1174),
    [anon_sym___inline] = ACTIONS(1174),
    [anon_sym___inline__] = ACTIONS(1174),
    [anon_sym___forceinline] = ACTIONS(1174),
    [anon_sym_thread_local] = ACTIONS(1174),
    [anon_sym___thread] = ACTIONS(1174),
    [anon_sym_const] = ACTIONS(1174),
    [anon_sym_constexpr] = ACTIONS(1174),
    [anon_sym_volatile] = ACTIONS(1174),
    [anon_sym_restrict] = ACTIONS(1174),
    [anon_sym___restrict__] = ACTIONS(1174),
    [anon_sym__Atomic] = ACTIONS(1174),
    [anon_sym__Noreturn] = ACTIONS(1174),
    [anon_sym_noreturn] = ACTIONS(1174),
    [anon_sym__Nonnull] = ACTIONS(1174),
    [anon_sym_alignas] = ACTIONS(1174),
    [anon_sym__Alignas] = ACTIONS(1174),
    [sym_primitive_type] = ACTIONS(1174),
    [anon_sym_enum] = ACTIONS(1174),
    [anon_sym_struct] = ACTIONS(1174),
    [anon_sym_union] = ACTIONS(1174),
    [anon_sym_if] = ACTIONS(1174),
    [anon_sym_else] = ACTIONS(1174),
    [anon_sym_switch] = ACTIONS(1174),
    [anon_sym_case] = ACTIONS(1174),
    [anon_sym_default] = ACTIONS(1174),
    [anon_sym_while] = ACTIONS(1174),
    [anon_sym_do] = ACTIONS(1174),
    [anon_sym_for] = ACTIONS(1174),
    [anon_sym_return] = ACTIONS(1174),
    [anon_sym_break] = ACTIONS(1174),
    [anon_sym_continue] = ACTIONS(1174),
    [anon_sym_goto] = ACTIONS(1174),
    [anon_sym___try] = ACTIONS(1174),
    [anon_sym___leave] = ACTIONS(1174),
    [anon_sym_DASH_DASH] = ACTIONS(1176),
    [anon_sym_PLUS_PLUS] = ACTIONS(1176),
    [anon_sym_sizeof] = ACTIONS(1174),
    [anon_sym___alignof__] = ACTIONS(1174),
    [anon_sym___alignof] = ACTIONS(1174),
    [anon_sym__alignof] = ACTIONS(1174),
    [anon_sym_alignof] = ACTIONS(1174),
    [anon_sym__Alignof] = ACTIONS(1174),
    [anon_sym_offsetof] = ACTIONS(1174),
    [anon_sym__Generic] = ACTIONS(1174),
    [anon_sym_asm] = ACTIONS(1174),
    [anon_sym___asm__] = ACTIONS(1174),
    [anon_sym___asm] = ACTIONS(1174),
    [sym_number_literal] = ACTIONS(1176),
    [anon_sym_L_SQUOTE] = ACTIONS(1176),
    [anon_sym_u_SQUOTE] = ACTIONS(1176),
    [anon_sym_U_SQUOTE] = ACTIONS(1176),
    [anon_sym_u8_SQUOTE] = ACTIONS(1176),
    [anon_sym_SQUOTE] = ACTIONS(1176),
    [anon_sym_L_DQUOTE] = ACTIONS(1176),
    [anon_sym_u_DQUOTE] = ACTIONS(1176),
    [anon_sym_U_DQUOTE] = ACTIONS(1176),
    [anon_sym_u8_DQUOTE] = ACTIONS(1176),
    [anon_sym_DQUOTE] = ACTIONS(1176),
    [sym_true] = ACTIONS(1174),
    [sym_false] = ACTIONS(1174),
    [anon_sym_NULL] = ACTIONS(1174),
    [anon_sym_nullptr] = ACTIONS(1174),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(238)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_RBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(239)] = {
    [sym_identifier] = ACTIONS(1178),
    [aux_sym_preproc_include_token1] = ACTIONS(1178),
    [aux_sym_preproc_def_token1] = ACTIONS(1178),
    [aux_sym_preproc_if_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1178),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1178),
    [sym_preproc_directive] = ACTIONS(1178),
    [anon_sym_LPAREN2] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1178),
    [anon_sym_PLUS] = ACTIONS(1178),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym___extension__] = ACTIONS(1178),
    [anon_sym_typedef] = ACTIONS(1178),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym___attribute__] = ACTIONS(1178),
    [anon_sym___attribute] = ACTIONS(1178),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1180),
    [anon_sym___declspec] = ACTIONS(1178),
    [anon_sym___cdecl] = ACTIONS(1178),
    [anon_sym___clrcall] = ACTIONS(1178),
    [anon_sym___stdcall] = ACTIONS(1178),
    [anon_sym___fastcall] = ACTIONS(1178),
    [anon_sym___thiscall] = ACTIONS(1178),
    [anon_sym___vectorcall] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_RBRACE] = ACTIONS(1180),
    [anon_sym_signed] = ACTIONS(1178),
    [anon_sym_unsigned] = ACTIONS(1178),
    [anon_sym_long] = ACTIONS(1178),
    [anon_sym_short] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_auto] = ACTIONS(1178),
    [anon_sym_register] = ACTIONS(1178),
    [anon_sym_inline] = ACTIONS(1178),
    [anon_sym___inline] = ACTIONS(1178),
    [anon_sym___inline__] = ACTIONS(1178),
    [anon_sym___forceinline] = ACTIONS(1178),
    [anon_sym_thread_local] = ACTIONS(1178),
    [anon_sym___thread] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_constexpr] = ACTIONS(1178),
    [anon_sym_volatile] = ACTIONS(1178),
    [anon_sym_restrict] = ACTIONS(1178),
    [anon_sym___restrict__] = ACTIONS(1178),
    [anon_sym__Atomic] = ACTIONS(1178),
    [anon_sym__Noreturn] = ACTIONS(1178),
    [anon_sym_noreturn] = ACTIONS(1178),
    [anon_sym__Nonnull] = ACTIONS(1178),
    [anon_sym_alignas] = ACTIONS(1178),
    [anon_sym__Alignas] = ACTIONS(1178),
    [sym_primitive_type] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_else] = ACTIONS(1178),
    [anon_sym_switch] = ACTIONS(1178),
    [anon_sym_case] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_do] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_goto] = ACTIONS(1178),
    [anon_sym___try] = ACTIONS(1178),
    [anon_sym___leave] = ACTIONS(1178),
    [anon_sym_DASH_DASH] = ACTIONS(1180),
    [anon_sym_PLUS_PLUS] = ACTIONS(1180),
    [anon_sym_sizeof] = ACTIONS(1178),
    [anon_sym___alignof__] = ACTIONS(1178),
    [anon_sym___alignof] = ACTIONS(1178),
    [anon_sym__alignof] = ACTIONS(1178),
    [anon_sym_alignof] = ACTIONS(1178),
    [anon_sym__Alignof] = ACTIONS(1178),
    [anon_sym_offsetof] = ACTIONS(1178),
    [anon_sym__Generic] = ACTIONS(1178),
    [anon_sym_asm] = ACTIONS(1178),
    [anon_sym___asm__] = ACTIONS(1178),
    [anon_sym___asm] = ACTIONS(1178),
    [sym_number_literal] = ACTIONS(1180),
    [anon_sym_L_SQUOTE] = ACTIONS(1180),
    [anon_sym_u_SQUOTE] = ACTIONS(1180),
    [anon_sym_U_SQUOTE] = ACTIONS(1180),
    [anon_sym_u8_SQUOTE] = ACTIONS(1180),
    [anon_sym_SQUOTE] = ACTIONS(1180),
    [anon_sym_L_DQUOTE] = ACTIONS(1180),
    [anon_sym_u_DQUOTE] = ACTIONS(1180),
    [anon_sym_U_DQUOTE] = ACTIONS(1180),
    [anon_sym_u8_DQUOTE] = ACTIONS(1180),
    [anon_sym_DQUOTE] = ACTIONS(1180),
    [sym_true] = ACTIONS(1178),
    [sym_false] = ACTIONS(1178),
    [anon_sym_NULL] = ACTIONS(1178),
    [anon_sym_nullptr] = ACTIONS(1178),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(240)] = {
    [sym_identifier] = ACTIONS(1182),
    [aux_sym_preproc_include_token1] = ACTIONS(1182),
    [aux_sym_preproc_def_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1182),
    [sym_preproc_directive] = ACTIONS(1182),
    [anon_sym_LPAREN2] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_TILDE] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1182),
    [anon_sym_PLUS] = ACTIONS(1182),
    [anon_sym_STAR] = ACTIONS(1184),
    [anon_sym_AMP] = ACTIONS(1184),
    [anon_sym_SEMI] = ACTIONS(1184),
    [anon_sym___extension__] = ACTIONS(1182),
    [anon_sym_typedef] = ACTIONS(1182),
    [anon_sym_extern] = ACTIONS(1182),
    [anon_sym___attribute__] = ACTIONS(1182),
    [anon_sym___attribute] = ACTIONS(1182),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1184),
    [anon_sym___declspec] = ACTIONS(1182),
    [anon_sym___cdecl] = ACTIONS(1182),
    [anon_sym___clrcall] = ACTIONS(1182),
    [anon_sym___stdcall] = ACTIONS(1182),
    [anon_sym___fastcall] = ACTIONS(1182),
    [anon_sym___thiscall] = ACTIONS(1182),
    [anon_sym___vectorcall] = ACTIONS(1182),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_RBRACE] = ACTIONS(1184),
    [anon_sym_signed] = ACTIONS(1182),
    [anon_sym_unsigned] = ACTIONS(1182),
    [anon_sym_long] = ACTIONS(1182),
    [anon_sym_short] = ACTIONS(1182),
    [anon_sym_static] = ACTIONS(1182),
    [anon_sym_auto] = ACTIONS(1182),
    [anon_sym_register] = ACTIONS(1182),
    [anon_sym_inline] = ACTIONS(1182),
    [anon_sym___inline] = ACTIONS(1182),
    [anon_sym___inline__] = ACTIONS(1182),
    [anon_sym___forceinline] = ACTIONS(1182),
    [anon_sym_thread_local] = ACTIONS(1182),
    [anon_sym___thread] = ACTIONS(1182),
    [anon_sym_const] = ACTIONS(1182),
    [anon_sym_constexpr] = ACTIONS(1182),
    [anon_sym_volatile] = ACTIONS(1182),
    [anon_sym_restrict] = ACTIONS(1182),
    [anon_sym___restrict__] = ACTIONS(1182),
    [anon_sym__Atomic] = ACTIONS(1182),
    [anon_sym__Noreturn] = ACTIONS(1182),
    [anon_sym_noreturn] = ACTIONS(1182),
    [anon_sym__Nonnull] = ACTIONS(1182),
    [anon_sym_alignas] = ACTIONS(1182),
    [anon_sym__Alignas] = ACTIONS(1182),
    [sym_primitive_type] = ACTIONS(1182),
    [anon_sym_enum] = ACTIONS(1182),
    [anon_sym_struct] = ACTIONS(1182),
    [anon_sym_union] = ACTIONS(1182),
    [anon_sym_if] = ACTIONS(1182),
    [anon_sym_else] = ACTIONS(1182),
    [anon_sym_switch] = ACTIONS(1182),
    [anon_sym_case] = ACTIONS(1182),
    [anon_sym_default] = ACTIONS(1182),
    [anon_sym_while] = ACTIONS(1182),
    [anon_sym_do] = ACTIONS(1182),
    [anon_sym_for] = ACTIONS(1182),
    [anon_sym_return] = ACTIONS(1182),
    [anon_sym_break] = ACTIONS(1182),
    [anon_sym_continue] = ACTIONS(1182),
    [anon_sym_goto] = ACTIONS(1182),
    [anon_sym___try] = ACTIONS(1182),
    [anon_sym___leave] = ACTIONS(1182),
    [anon_sym_DASH_DASH] = ACTIONS(1184),
    [anon_sym_PLUS_PLUS] = ACTIONS(1184),
    [anon_sym_sizeof] = ACTIONS(1182),
    [anon_sym___alignof__] = ACTIONS(1182),
    [anon_sym___alignof] = ACTIONS(1182),
    [anon_sym__alignof] = ACTIONS(1182),
    [anon_sym_alignof] = ACTIONS(1182),
    [anon_sym__Alignof] = ACTIONS(1182),
    [anon_sym_offsetof] = ACTIONS(1182),
    [anon_sym__Generic] = ACTIONS(1182),
    [anon_sym_asm] = ACTIONS(1182),
    [anon_sym___asm__] = ACTIONS(1182),
    [anon_sym___asm] = ACTIONS(1182),
    [sym_number_literal] = ACTIONS(1184),
    [anon_sym_L_SQUOTE] = ACTIONS(1184),
    [anon_sym_u_SQUOTE] = ACTIONS(1184),
    [anon_sym_U_SQUOTE] = ACTIONS(1184),
    [anon_sym_u8_SQUOTE] = ACTIONS(1184),
    [anon_sym_SQUOTE] = ACTIONS(1184),
    [anon_sym_L_DQUOTE] = ACTIONS(1184),
    [anon_sym_u_DQUOTE] = ACTIONS(1184),
    [anon_sym_U_DQUOTE] = ACTIONS(1184),
    [anon_sym_u8_DQUOTE] = ACTIONS(1184),
    [anon_sym_DQUOTE] = ACTIONS(1184),
    [sym_true] = ACTIONS(1182),
    [sym_false] = ACTIONS(1182),
    [anon_sym_NULL] = ACTIONS(1182),
    [anon_sym_nullptr] = ACTIONS(1182),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(241)] = {
    [ts_builtin_sym_end] = ACTIONS(1200),
    [sym_identifier] = ACTIONS(1198),
    [aux_sym_preproc_include_token1] = ACTIONS(1198),
    [aux_sym_preproc_def_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1198),
    [sym_preproc_directive] = ACTIONS(1198),
    [anon_sym_LPAREN2] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1198),
    [anon_sym_PLUS] = ACTIONS(1198),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym___extension__] = ACTIONS(1198),
    [anon_sym_typedef] = ACTIONS(1198),
    [anon_sym_extern] = ACTIONS(1198),
    [anon_sym___attribute__] = ACTIONS(1198),
    [anon_sym___attribute] = ACTIONS(1198),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1200),
    [anon_sym___declspec] = ACTIONS(1198),
    [anon_sym___cdecl] = ACTIONS(1198),
    [anon_sym___clrcall] = ACTIONS(1198),
    [anon_sym___stdcall] = ACTIONS(1198),
    [anon_sym___fastcall] = ACTIONS(1198),
    [anon_sym___thiscall] = ACTIONS(1198),
    [anon_sym___vectorcall] = ACTIONS(1198),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_signed] = ACTIONS(1198),
    [anon_sym_unsigned] = ACTIONS(1198),
    [anon_sym_long] = ACTIONS(1198),
    [anon_sym_short] = ACTIONS(1198),
    [anon_sym_static] = ACTIONS(1198),
    [anon_sym_auto] = ACTIONS(1198),
    [anon_sym_register] = ACTIONS(1198),
    [anon_sym_inline] = ACTIONS(1198),
    [anon_sym___inline] = ACTIONS(1198),
    [anon_sym___inline__] = ACTIONS(1198),
    [anon_sym___forceinline] = ACTIONS(1198),
    [anon_sym_thread_local] = ACTIONS(1198),
    [anon_sym___thread] = ACTIONS(1198),
    [anon_sym_const] = ACTIONS(1198),
    [anon_sym_constexpr] = ACTIONS(1198),
    [anon_sym_volatile] = ACTIONS(1198),
    [anon_sym_restrict] = ACTIONS(1198),
    [anon_sym___restrict__] = ACTIONS(1198),
    [anon_sym__Atomic] = ACTIONS(1198),
    [anon_sym__Noreturn] = ACTIONS(1198),
    [anon_sym_noreturn] = ACTIONS(1198),
    [anon_sym__Nonnull] = ACTIONS(1198),
    [anon_sym_alignas] = ACTIONS(1198),
    [anon_sym__Alignas] = ACTIONS(1198),
    [sym_primitive_type] = ACTIONS(1198),
    [anon_sym_enum] = ACTIONS(1198),
    [anon_sym_struct] = ACTIONS(1198),
    [anon_sym_union] = ACTIONS(1198),
    [anon_sym_if] = ACTIONS(1198),
    [anon_sym_else] = ACTIONS(1198),
    [anon_sym_switch] = ACTIONS(1198),
    [anon_sym_case] = ACTIONS(1198),
    [anon_sym_default] = ACTIONS(1198),
    [anon_sym_while] = ACTIONS(1198),
    [anon_sym_do] = ACTIONS(1198),
    [anon_sym_for] = ACTIONS(1198),
    [anon_sym_return] = ACTIONS(1198),
    [anon_sym_break] = ACTIONS(1198),
    [anon_sym_continue] = ACTIONS(1198),
    [anon_sym_goto] = ACTIONS(1198),
    [anon_sym___try] = ACTIONS(1198),
    [anon_sym___leave] = ACTIONS(1198),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_sizeof] = ACTIONS(1198),
    [anon_sym___alignof__] = ACTIONS(1198),
    [anon_sym___alignof] = ACTIONS(1198),
    [anon_sym__alignof] = ACTIONS(1198),
    [anon_sym_alignof] = ACTIONS(1198),
    [anon_sym__Alignof] = ACTIONS(1198),
    [anon_sym_offsetof] = ACTIONS(1198),
    [anon_sym__Generic] = ACTIONS(1198),
    [anon_sym_asm] = ACTIONS(1198),
    [anon_sym___asm__] = ACTIONS(1198),
    [anon_sym___asm] = ACTIONS(1198),
    [sym_number_literal] = ACTIONS(1200),
    [anon_sym_L_SQUOTE] = ACTIONS(1200),
    [anon_sym_u_SQUOTE] = ACTIONS(1200),
    [anon_sym_U_SQUOTE] = ACTIONS(1200),
    [anon_sym_u8_SQUOTE] = ACTIONS(1200),
    [anon_sym_SQUOTE] = ACTIONS(1200),
    [anon_sym_L_DQUOTE] = ACTIONS(1200),
    [anon_sym_u_DQUOTE] = ACTIONS(1200),
    [anon_sym_U_DQUOTE] = ACTIONS(1200),
    [anon_sym_u8_DQUOTE] = ACTIONS(1200),
    [anon_sym_DQUOTE] = ACTIONS(1200),
    [sym_true] = ACTIONS(1198),
    [sym_false] = ACTIONS(1198),
    [anon_sym_NULL] = ACTIONS(1198),
    [anon_sym_nullptr] = ACTIONS(1198),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(242)] = {
    [sym_identifier] = ACTIONS(1206),
    [aux_sym_preproc_include_token1] = ACTIONS(1206),
    [aux_sym_preproc_def_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1206),
    [sym_preproc_directive] = ACTIONS(1206),
    [anon_sym_LPAREN2] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_TILDE] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1208),
    [anon_sym_AMP] = ACTIONS(1208),
    [anon_sym_SEMI] = ACTIONS(1208),
    [anon_sym___extension__] = ACTIONS(1206),
    [anon_sym_typedef] = ACTIONS(1206),
    [anon_sym_extern] = ACTIONS(1206),
    [anon_sym___attribute__] = ACTIONS(1206),
    [anon_sym___attribute] = ACTIONS(1206),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1208),
    [anon_sym___declspec] = ACTIONS(1206),
    [anon_sym___cdecl] = ACTIONS(1206),
    [anon_sym___clrcall] = ACTIONS(1206),
    [anon_sym___stdcall] = ACTIONS(1206),
    [anon_sym___fastcall] = ACTIONS(1206),
    [anon_sym___thiscall] = ACTIONS(1206),
    [anon_sym___vectorcall] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_RBRACE] = ACTIONS(1208),
    [anon_sym_signed] = ACTIONS(1206),
    [anon_sym_unsigned] = ACTIONS(1206),
    [anon_sym_long] = ACTIONS(1206),
    [anon_sym_short] = ACTIONS(1206),
    [anon_sym_static] = ACTIONS(1206),
    [anon_sym_auto] = ACTIONS(1206),
    [anon_sym_register] = ACTIONS(1206),
    [anon_sym_inline] = ACTIONS(1206),
    [anon_sym___inline] = ACTIONS(1206),
    [anon_sym___inline__] = ACTIONS(1206),
    [anon_sym___forceinline] = ACTIONS(1206),
    [anon_sym_thread_local] = ACTIONS(1206),
    [anon_sym___thread] = ACTIONS(1206),
    [anon_sym_const] = ACTIONS(1206),
    [anon_sym_constexpr] = ACTIONS(1206),
    [anon_sym_volatile] = ACTIONS(1206),
    [anon_sym_restrict] = ACTIONS(1206),
    [anon_sym___restrict__] = ACTIONS(1206),
    [anon_sym__Atomic] = ACTIONS(1206),
    [anon_sym__Noreturn] = ACTIONS(1206),
    [anon_sym_noreturn] = ACTIONS(1206),
    [anon_sym__Nonnull] = ACTIONS(1206),
    [anon_sym_alignas] = ACTIONS(1206),
    [anon_sym__Alignas] = ACTIONS(1206),
    [sym_primitive_type] = ACTIONS(1206),
    [anon_sym_enum] = ACTIONS(1206),
    [anon_sym_struct] = ACTIONS(1206),
    [anon_sym_union] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1206),
    [anon_sym_else] = ACTIONS(1206),
    [anon_sym_switch] = ACTIONS(1206),
    [anon_sym_case] = ACTIONS(1206),
    [anon_sym_default] = ACTIONS(1206),
    [anon_sym_while] = ACTIONS(1206),
    [anon_sym_do] = ACTIONS(1206),
    [anon_sym_for] = ACTIONS(1206),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_break] = ACTIONS(1206),
    [anon_sym_continue] = ACTIONS(1206),
    [anon_sym_goto] = ACTIONS(1206),
    [anon_sym___try] = ACTIONS(1206),
    [anon_sym___leave] = ACTIONS(1206),
    [anon_sym_DASH_DASH] = ACTIONS(1208),
    [anon_sym_PLUS_PLUS] = ACTIONS(1208),
    [anon_sym_sizeof] = ACTIONS(1206),
    [anon_sym___alignof__] = ACTIONS(1206),
    [anon_sym___alignof] = ACTIONS(1206),
    [anon_sym__alignof] = ACTIONS(1206),
    [anon_sym_alignof] = ACTIONS(1206),
    [anon_sym__Alignof] = ACTIONS(1206),
    [anon_sym_offsetof] = ACTIONS(1206),
    [anon_sym__Generic] = ACTIONS(1206),
    [anon_sym_asm] = ACTIONS(1206),
    [anon_sym___asm__] = ACTIONS(1206),
    [anon_sym___asm] = ACTIONS(1206),
    [sym_number_literal] = ACTIONS(1208),
    [anon_sym_L_SQUOTE] = ACTIONS(1208),
    [anon_sym_u_SQUOTE] = ACTIONS(1208),
    [anon_sym_U_SQUOTE] = ACTIONS(1208),
    [anon_sym_u8_SQUOTE] = ACTIONS(1208),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_L_DQUOTE] = ACTIONS(1208),
    [anon_sym_u_DQUOTE] = ACTIONS(1208),
    [anon_sym_U_DQUOTE] = ACTIONS(1208),
    [anon_sym_u8_DQUOTE] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [sym_true] = ACTIONS(1206),
    [sym_false] = ACTIONS(1206),
    [anon_sym_NULL] = ACTIONS(1206),
    [anon_sym_nullptr] = ACTIONS(1206),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(243)] = {
    [ts_builtin_sym_end] = ACTIONS(1184),
    [sym_identifier] = ACTIONS(1182),
    [aux_sym_preproc_include_token1] = ACTIONS(1182),
    [aux_sym_preproc_def_token1] = ACTIONS(1182),
    [aux_sym_preproc_if_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1182),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1182),
    [sym_preproc_directive] = ACTIONS(1182),
    [anon_sym_LPAREN2] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_TILDE] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1182),
    [anon_sym_PLUS] = ACTIONS(1182),
    [anon_sym_STAR] = ACTIONS(1184),
    [anon_sym_AMP] = ACTIONS(1184),
    [anon_sym_SEMI] = ACTIONS(1184),
    [anon_sym___extension__] = ACTIONS(1182),
    [anon_sym_typedef] = ACTIONS(1182),
    [anon_sym_extern] = ACTIONS(1182),
    [anon_sym___attribute__] = ACTIONS(1182),
    [anon_sym___attribute] = ACTIONS(1182),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1184),
    [anon_sym___declspec] = ACTIONS(1182),
    [anon_sym___cdecl] = ACTIONS(1182),
    [anon_sym___clrcall] = ACTIONS(1182),
    [anon_sym___stdcall] = ACTIONS(1182),
    [anon_sym___fastcall] = ACTIONS(1182),
    [anon_sym___thiscall] = ACTIONS(1182),
    [anon_sym___vectorcall] = ACTIONS(1182),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_signed] = ACTIONS(1182),
    [anon_sym_unsigned] = ACTIONS(1182),
    [anon_sym_long] = ACTIONS(1182),
    [anon_sym_short] = ACTIONS(1182),
    [anon_sym_static] = ACTIONS(1182),
    [anon_sym_auto] = ACTIONS(1182),
    [anon_sym_register] = ACTIONS(1182),
    [anon_sym_inline] = ACTIONS(1182),
    [anon_sym___inline] = ACTIONS(1182),
    [anon_sym___inline__] = ACTIONS(1182),
    [anon_sym___forceinline] = ACTIONS(1182),
    [anon_sym_thread_local] = ACTIONS(1182),
    [anon_sym___thread] = ACTIONS(1182),
    [anon_sym_const] = ACTIONS(1182),
    [anon_sym_constexpr] = ACTIONS(1182),
    [anon_sym_volatile] = ACTIONS(1182),
    [anon_sym_restrict] = ACTIONS(1182),
    [anon_sym___restrict__] = ACTIONS(1182),
    [anon_sym__Atomic] = ACTIONS(1182),
    [anon_sym__Noreturn] = ACTIONS(1182),
    [anon_sym_noreturn] = ACTIONS(1182),
    [anon_sym__Nonnull] = ACTIONS(1182),
    [anon_sym_alignas] = ACTIONS(1182),
    [anon_sym__Alignas] = ACTIONS(1182),
    [sym_primitive_type] = ACTIONS(1182),
    [anon_sym_enum] = ACTIONS(1182),
    [anon_sym_struct] = ACTIONS(1182),
    [anon_sym_union] = ACTIONS(1182),
    [anon_sym_if] = ACTIONS(1182),
    [anon_sym_else] = ACTIONS(1182),
    [anon_sym_switch] = ACTIONS(1182),
    [anon_sym_case] = ACTIONS(1182),
    [anon_sym_default] = ACTIONS(1182),
    [anon_sym_while] = ACTIONS(1182),
    [anon_sym_do] = ACTIONS(1182),
    [anon_sym_for] = ACTIONS(1182),
    [anon_sym_return] = ACTIONS(1182),
    [anon_sym_break] = ACTIONS(1182),
    [anon_sym_continue] = ACTIONS(1182),
    [anon_sym_goto] = ACTIONS(1182),
    [anon_sym___try] = ACTIONS(1182),
    [anon_sym___leave] = ACTIONS(1182),
    [anon_sym_DASH_DASH] = ACTIONS(1184),
    [anon_sym_PLUS_PLUS] = ACTIONS(1184),
    [anon_sym_sizeof] = ACTIONS(1182),
    [anon_sym___alignof__] = ACTIONS(1182),
    [anon_sym___alignof] = ACTIONS(1182),
    [anon_sym__alignof] = ACTIONS(1182),
    [anon_sym_alignof] = ACTIONS(1182),
    [anon_sym__Alignof] = ACTIONS(1182),
    [anon_sym_offsetof] = ACTIONS(1182),
    [anon_sym__Generic] = ACTIONS(1182),
    [anon_sym_asm] = ACTIONS(1182),
    [anon_sym___asm__] = ACTIONS(1182),
    [anon_sym___asm] = ACTIONS(1182),
    [sym_number_literal] = ACTIONS(1184),
    [anon_sym_L_SQUOTE] = ACTIONS(1184),
    [anon_sym_u_SQUOTE] = ACTIONS(1184),
    [anon_sym_U_SQUOTE] = ACTIONS(1184),
    [anon_sym_u8_SQUOTE] = ACTIONS(1184),
    [anon_sym_SQUOTE] = ACTIONS(1184),
    [anon_sym_L_DQUOTE] = ACTIONS(1184),
    [anon_sym_u_DQUOTE] = ACTIONS(1184),
    [anon_sym_U_DQUOTE] = ACTIONS(1184),
    [anon_sym_u8_DQUOTE] = ACTIONS(1184),
    [anon_sym_DQUOTE] = ACTIONS(1184),
    [sym_true] = ACTIONS(1182),
    [sym_false] = ACTIONS(1182),
    [anon_sym_NULL] = ACTIONS(1182),
    [anon_sym_nullptr] = ACTIONS(1182),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(244)] = {
    [sym_identifier] = ACTIONS(1134),
    [aux_sym_preproc_include_token1] = ACTIONS(1134),
    [aux_sym_preproc_def_token1] = ACTIONS(1134),
    [aux_sym_preproc_if_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1134),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1134),
    [sym_preproc_directive] = ACTIONS(1134),
    [anon_sym_LPAREN2] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_TILDE] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1134),
    [anon_sym_PLUS] = ACTIONS(1134),
    [anon_sym_STAR] = ACTIONS(1136),
    [anon_sym_AMP] = ACTIONS(1136),
    [anon_sym_SEMI] = ACTIONS(1136),
    [anon_sym___extension__] = ACTIONS(1134),
    [anon_sym_typedef] = ACTIONS(1134),
    [anon_sym_extern] = ACTIONS(1134),
    [anon_sym___attribute__] = ACTIONS(1134),
    [anon_sym___attribute] = ACTIONS(1134),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1136),
    [anon_sym___declspec] = ACTIONS(1134),
    [anon_sym___cdecl] = ACTIONS(1134),
    [anon_sym___clrcall] = ACTIONS(1134),
    [anon_sym___stdcall] = ACTIONS(1134),
    [anon_sym___fastcall] = ACTIONS(1134),
    [anon_sym___thiscall] = ACTIONS(1134),
    [anon_sym___vectorcall] = ACTIONS(1134),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_RBRACE] = ACTIONS(1136),
    [anon_sym_signed] = ACTIONS(1134),
    [anon_sym_unsigned] = ACTIONS(1134),
    [anon_sym_long] = ACTIONS(1134),
    [anon_sym_short] = ACTIONS(1134),
    [anon_sym_static] = ACTIONS(1134),
    [anon_sym_auto] = ACTIONS(1134),
    [anon_sym_register] = ACTIONS(1134),
    [anon_sym_inline] = ACTIONS(1134),
    [anon_sym___inline] = ACTIONS(1134),
    [anon_sym___inline__] = ACTIONS(1134),
    [anon_sym___forceinline] = ACTIONS(1134),
    [anon_sym_thread_local] = ACTIONS(1134),
    [anon_sym___thread] = ACTIONS(1134),
    [anon_sym_const] = ACTIONS(1134),
    [anon_sym_constexpr] = ACTIONS(1134),
    [anon_sym_volatile] = ACTIONS(1134),
    [anon_sym_restrict] = ACTIONS(1134),
    [anon_sym___restrict__] = ACTIONS(1134),
    [anon_sym__Atomic] = ACTIONS(1134),
    [anon_sym__Noreturn] = ACTIONS(1134),
    [anon_sym_noreturn] = ACTIONS(1134),
    [anon_sym__Nonnull] = ACTIONS(1134),
    [anon_sym_alignas] = ACTIONS(1134),
    [anon_sym__Alignas] = ACTIONS(1134),
    [sym_primitive_type] = ACTIONS(1134),
    [anon_sym_enum] = ACTIONS(1134),
    [anon_sym_struct] = ACTIONS(1134),
    [anon_sym_union] = ACTIONS(1134),
    [anon_sym_if] = ACTIONS(1134),
    [anon_sym_else] = ACTIONS(1134),
    [anon_sym_switch] = ACTIONS(1134),
    [anon_sym_case] = ACTIONS(1134),
    [anon_sym_default] = ACTIONS(1134),
    [anon_sym_while] = ACTIONS(1134),
    [anon_sym_do] = ACTIONS(1134),
    [anon_sym_for] = ACTIONS(1134),
    [anon_sym_return] = ACTIONS(1134),
    [anon_sym_break] = ACTIONS(1134),
    [anon_sym_continue] = ACTIONS(1134),
    [anon_sym_goto] = ACTIONS(1134),
    [anon_sym___try] = ACTIONS(1134),
    [anon_sym___leave] = ACTIONS(1134),
    [anon_sym_DASH_DASH] = ACTIONS(1136),
    [anon_sym_PLUS_PLUS] = ACTIONS(1136),
    [anon_sym_sizeof] = ACTIONS(1134),
    [anon_sym___alignof__] = ACTIONS(1134),
    [anon_sym___alignof] = ACTIONS(1134),
    [anon_sym__alignof] = ACTIONS(1134),
    [anon_sym_alignof] = ACTIONS(1134),
    [anon_sym__Alignof] = ACTIONS(1134),
    [anon_sym_offsetof] = ACTIONS(1134),
    [anon_sym__Generic] = ACTIONS(1134),
    [anon_sym_asm] = ACTIONS(1134),
    [anon_sym___asm__] = ACTIONS(1134),
    [anon_sym___asm] = ACTIONS(1134),
    [sym_number_literal] = ACTIONS(1136),
    [anon_sym_L_SQUOTE] = ACTIONS(1136),
    [anon_sym_u_SQUOTE] = ACTIONS(1136),
    [anon_sym_U_SQUOTE] = ACTIONS(1136),
    [anon_sym_u8_SQUOTE] = ACTIONS(1136),
    [anon_sym_SQUOTE] = ACTIONS(1136),
    [anon_sym_L_DQUOTE] = ACTIONS(1136),
    [anon_sym_u_DQUOTE] = ACTIONS(1136),
    [anon_sym_U_DQUOTE] = ACTIONS(1136),
    [anon_sym_u8_DQUOTE] = ACTIONS(1136),
    [anon_sym_DQUOTE] = ACTIONS(1136),
    [sym_true] = ACTIONS(1134),
    [sym_false] = ACTIONS(1134),
    [anon_sym_NULL] = ACTIONS(1134),
    [anon_sym_nullptr] = ACTIONS(1134),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(245)] = {
    [sym_identifier] = ACTIONS(1142),
    [aux_sym_preproc_include_token1] = ACTIONS(1142),
    [aux_sym_preproc_def_token1] = ACTIONS(1142),
    [aux_sym_preproc_if_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1142),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1142),
    [sym_preproc_directive] = ACTIONS(1142),
    [anon_sym_LPAREN2] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_TILDE] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1142),
    [anon_sym_PLUS] = ACTIONS(1142),
    [anon_sym_STAR] = ACTIONS(1144),
    [anon_sym_AMP] = ACTIONS(1144),
    [anon_sym_SEMI] = ACTIONS(1144),
    [anon_sym___extension__] = ACTIONS(1142),
    [anon_sym_typedef] = ACTIONS(1142),
    [anon_sym_extern] = ACTIONS(1142),
    [anon_sym___attribute__] = ACTIONS(1142),
    [anon_sym___attribute] = ACTIONS(1142),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1144),
    [anon_sym___declspec] = ACTIONS(1142),
    [anon_sym___cdecl] = ACTIONS(1142),
    [anon_sym___clrcall] = ACTIONS(1142),
    [anon_sym___stdcall] = ACTIONS(1142),
    [anon_sym___fastcall] = ACTIONS(1142),
    [anon_sym___thiscall] = ACTIONS(1142),
    [anon_sym___vectorcall] = ACTIONS(1142),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_RBRACE] = ACTIONS(1144),
    [anon_sym_signed] = ACTIONS(1142),
    [anon_sym_unsigned] = ACTIONS(1142),
    [anon_sym_long] = ACTIONS(1142),
    [anon_sym_short] = ACTIONS(1142),
    [anon_sym_static] = ACTIONS(1142),
    [anon_sym_auto] = ACTIONS(1142),
    [anon_sym_register] = ACTIONS(1142),
    [anon_sym_inline] = ACTIONS(1142),
    [anon_sym___inline] = ACTIONS(1142),
    [anon_sym___inline__] = ACTIONS(1142),
    [anon_sym___forceinline] = ACTIONS(1142),
    [anon_sym_thread_local] = ACTIONS(1142),
    [anon_sym___thread] = ACTIONS(1142),
    [anon_sym_const] = ACTIONS(1142),
    [anon_sym_constexpr] = ACTIONS(1142),
    [anon_sym_volatile] = ACTIONS(1142),
    [anon_sym_restrict] = ACTIONS(1142),
    [anon_sym___restrict__] = ACTIONS(1142),
    [anon_sym__Atomic] = ACTIONS(1142),
    [anon_sym__Noreturn] = ACTIONS(1142),
    [anon_sym_noreturn] = ACTIONS(1142),
    [anon_sym__Nonnull] = ACTIONS(1142),
    [anon_sym_alignas] = ACTIONS(1142),
    [anon_sym__Alignas] = ACTIONS(1142),
    [sym_primitive_type] = ACTIONS(1142),
    [anon_sym_enum] = ACTIONS(1142),
    [anon_sym_struct] = ACTIONS(1142),
    [anon_sym_union] = ACTIONS(1142),
    [anon_sym_if] = ACTIONS(1142),
    [anon_sym_else] = ACTIONS(1142),
    [anon_sym_switch] = ACTIONS(1142),
    [anon_sym_case] = ACTIONS(1142),
    [anon_sym_default] = ACTIONS(1142),
    [anon_sym_while] = ACTIONS(1142),
    [anon_sym_do] = ACTIONS(1142),
    [anon_sym_for] = ACTIONS(1142),
    [anon_sym_return] = ACTIONS(1142),
    [anon_sym_break] = ACTIONS(1142),
    [anon_sym_continue] = ACTIONS(1142),
    [anon_sym_goto] = ACTIONS(1142),
    [anon_sym___try] = ACTIONS(1142),
    [anon_sym___leave] = ACTIONS(1142),
    [anon_sym_DASH_DASH] = ACTIONS(1144),
    [anon_sym_PLUS_PLUS] = ACTIONS(1144),
    [anon_sym_sizeof] = ACTIONS(1142),
    [anon_sym___alignof__] = ACTIONS(1142),
    [anon_sym___alignof] = ACTIONS(1142),
    [anon_sym__alignof] = ACTIONS(1142),
    [anon_sym_alignof] = ACTIONS(1142),
    [anon_sym__Alignof] = ACTIONS(1142),
    [anon_sym_offsetof] = ACTIONS(1142),
    [anon_sym__Generic] = ACTIONS(1142),
    [anon_sym_asm] = ACTIONS(1142),
    [anon_sym___asm__] = ACTIONS(1142),
    [anon_sym___asm] = ACTIONS(1142),
    [sym_number_literal] = ACTIONS(1144),
    [anon_sym_L_SQUOTE] = ACTIONS(1144),
    [anon_sym_u_SQUOTE] = ACTIONS(1144),
    [anon_sym_U_SQUOTE] = ACTIONS(1144),
    [anon_sym_u8_SQUOTE] = ACTIONS(1144),
    [anon_sym_SQUOTE] = ACTIONS(1144),
    [anon_sym_L_DQUOTE] = ACTIONS(1144),
    [anon_sym_u_DQUOTE] = ACTIONS(1144),
    [anon_sym_U_DQUOTE] = ACTIONS(1144),
    [anon_sym_u8_DQUOTE] = ACTIONS(1144),
    [anon_sym_DQUOTE] = ACTIONS(1144),
    [sym_true] = ACTIONS(1142),
    [sym_false] = ACTIONS(1142),
    [anon_sym_NULL] = ACTIONS(1142),
    [anon_sym_nullptr] = ACTIONS(1142),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(246)] = {
    [sym_identifier] = ACTIONS(1246),
    [aux_sym_preproc_include_token1] = ACTIONS(1246),
    [aux_sym_preproc_def_token1] = ACTIONS(1246),
    [aux_sym_preproc_if_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1246),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1246),
    [sym_preproc_directive] = ACTIONS(1246),
    [anon_sym_LPAREN2] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_TILDE] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1246),
    [anon_sym_PLUS] = ACTIONS(1246),
    [anon_sym_STAR] = ACTIONS(1248),
    [anon_sym_AMP] = ACTIONS(1248),
    [anon_sym_SEMI] = ACTIONS(1248),
    [anon_sym___extension__] = ACTIONS(1246),
    [anon_sym_typedef] = ACTIONS(1246),
    [anon_sym_extern] = ACTIONS(1246),
    [anon_sym___attribute__] = ACTIONS(1246),
    [anon_sym___attribute] = ACTIONS(1246),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1248),
    [anon_sym___declspec] = ACTIONS(1246),
    [anon_sym___cdecl] = ACTIONS(1246),
    [anon_sym___clrcall] = ACTIONS(1246),
    [anon_sym___stdcall] = ACTIONS(1246),
    [anon_sym___fastcall] = ACTIONS(1246),
    [anon_sym___thiscall] = ACTIONS(1246),
    [anon_sym___vectorcall] = ACTIONS(1246),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_RBRACE] = ACTIONS(1248),
    [anon_sym_signed] = ACTIONS(1246),
    [anon_sym_unsigned] = ACTIONS(1246),
    [anon_sym_long] = ACTIONS(1246),
    [anon_sym_short] = ACTIONS(1246),
    [anon_sym_static] = ACTIONS(1246),
    [anon_sym_auto] = ACTIONS(1246),
    [anon_sym_register] = ACTIONS(1246),
    [anon_sym_inline] = ACTIONS(1246),
    [anon_sym___inline] = ACTIONS(1246),
    [anon_sym___inline__] = ACTIONS(1246),
    [anon_sym___forceinline] = ACTIONS(1246),
    [anon_sym_thread_local] = ACTIONS(1246),
    [anon_sym___thread] = ACTIONS(1246),
    [anon_sym_const] = ACTIONS(1246),
    [anon_sym_constexpr] = ACTIONS(1246),
    [anon_sym_volatile] = ACTIONS(1246),
    [anon_sym_restrict] = ACTIONS(1246),
    [anon_sym___restrict__] = ACTIONS(1246),
    [anon_sym__Atomic] = ACTIONS(1246),
    [anon_sym__Noreturn] = ACTIONS(1246),
    [anon_sym_noreturn] = ACTIONS(1246),
    [anon_sym__Nonnull] = ACTIONS(1246),
    [anon_sym_alignas] = ACTIONS(1246),
    [anon_sym__Alignas] = ACTIONS(1246),
    [sym_primitive_type] = ACTIONS(1246),
    [anon_sym_enum] = ACTIONS(1246),
    [anon_sym_struct] = ACTIONS(1246),
    [anon_sym_union] = ACTIONS(1246),
    [anon_sym_if] = ACTIONS(1246),
    [anon_sym_else] = ACTIONS(1246),
    [anon_sym_switch] = ACTIONS(1246),
    [anon_sym_case] = ACTIONS(1246),
    [anon_sym_default] = ACTIONS(1246),
    [anon_sym_while] = ACTIONS(1246),
    [anon_sym_do] = ACTIONS(1246),
    [anon_sym_for] = ACTIONS(1246),
    [anon_sym_return] = ACTIONS(1246),
    [anon_sym_break] = ACTIONS(1246),
    [anon_sym_continue] = ACTIONS(1246),
    [anon_sym_goto] = ACTIONS(1246),
    [anon_sym___try] = ACTIONS(1246),
    [anon_sym___leave] = ACTIONS(1246),
    [anon_sym_DASH_DASH] = ACTIONS(1248),
    [anon_sym_PLUS_PLUS] = ACTIONS(1248),
    [anon_sym_sizeof] = ACTIONS(1246),
    [anon_sym___alignof__] = ACTIONS(1246),
    [anon_sym___alignof] = ACTIONS(1246),
    [anon_sym__alignof] = ACTIONS(1246),
    [anon_sym_alignof] = ACTIONS(1246),
    [anon_sym__Alignof] = ACTIONS(1246),
    [anon_sym_offsetof] = ACTIONS(1246),
    [anon_sym__Generic] = ACTIONS(1246),
    [anon_sym_asm] = ACTIONS(1246),
    [anon_sym___asm__] = ACTIONS(1246),
    [anon_sym___asm] = ACTIONS(1246),
    [sym_number_literal] = ACTIONS(1248),
    [anon_sym_L_SQUOTE] = ACTIONS(1248),
    [anon_sym_u_SQUOTE] = ACTIONS(1248),
    [anon_sym_U_SQUOTE] = ACTIONS(1248),
    [anon_sym_u8_SQUOTE] = ACTIONS(1248),
    [anon_sym_SQUOTE] = ACTIONS(1248),
    [anon_sym_L_DQUOTE] = ACTIONS(1248),
    [anon_sym_u_DQUOTE] = ACTIONS(1248),
    [anon_sym_U_DQUOTE] = ACTIONS(1248),
    [anon_sym_u8_DQUOTE] = ACTIONS(1248),
    [anon_sym_DQUOTE] = ACTIONS(1248),
    [sym_true] = ACTIONS(1246),
    [sym_false] = ACTIONS(1246),
    [anon_sym_NULL] = ACTIONS(1246),
    [anon_sym_nullptr] = ACTIONS(1246),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(247)] = {
    [ts_builtin_sym_end] = ACTIONS(1212),
    [sym_identifier] = ACTIONS(1210),
    [aux_sym_preproc_include_token1] = ACTIONS(1210),
    [aux_sym_preproc_def_token1] = ACTIONS(1210),
    [aux_sym_preproc_if_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1210),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1210),
    [sym_preproc_directive] = ACTIONS(1210),
    [anon_sym_LPAREN2] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1210),
    [anon_sym_PLUS] = ACTIONS(1210),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym___extension__] = ACTIONS(1210),
    [anon_sym_typedef] = ACTIONS(1210),
    [anon_sym_extern] = ACTIONS(1210),
    [anon_sym___attribute__] = ACTIONS(1210),
    [anon_sym___attribute] = ACTIONS(1210),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1212),
    [anon_sym___declspec] = ACTIONS(1210),
    [anon_sym___cdecl] = ACTIONS(1210),
    [anon_sym___clrcall] = ACTIONS(1210),
    [anon_sym___stdcall] = ACTIONS(1210),
    [anon_sym___fastcall] = ACTIONS(1210),
    [anon_sym___thiscall] = ACTIONS(1210),
    [anon_sym___vectorcall] = ACTIONS(1210),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_signed] = ACTIONS(1210),
    [anon_sym_unsigned] = ACTIONS(1210),
    [anon_sym_long] = ACTIONS(1210),
    [anon_sym_short] = ACTIONS(1210),
    [anon_sym_static] = ACTIONS(1210),
    [anon_sym_auto] = ACTIONS(1210),
    [anon_sym_register] = ACTIONS(1210),
    [anon_sym_inline] = ACTIONS(1210),
    [anon_sym___inline] = ACTIONS(1210),
    [anon_sym___inline__] = ACTIONS(1210),
    [anon_sym___forceinline] = ACTIONS(1210),
    [anon_sym_thread_local] = ACTIONS(1210),
    [anon_sym___thread] = ACTIONS(1210),
    [anon_sym_const] = ACTIONS(1210),
    [anon_sym_constexpr] = ACTIONS(1210),
    [anon_sym_volatile] = ACTIONS(1210),
    [anon_sym_restrict] = ACTIONS(1210),
    [anon_sym___restrict__] = ACTIONS(1210),
    [anon_sym__Atomic] = ACTIONS(1210),
    [anon_sym__Noreturn] = ACTIONS(1210),
    [anon_sym_noreturn] = ACTIONS(1210),
    [anon_sym__Nonnull] = ACTIONS(1210),
    [anon_sym_alignas] = ACTIONS(1210),
    [anon_sym__Alignas] = ACTIONS(1210),
    [sym_primitive_type] = ACTIONS(1210),
    [anon_sym_enum] = ACTIONS(1210),
    [anon_sym_struct] = ACTIONS(1210),
    [anon_sym_union] = ACTIONS(1210),
    [anon_sym_if] = ACTIONS(1210),
    [anon_sym_else] = ACTIONS(1210),
    [anon_sym_switch] = ACTIONS(1210),
    [anon_sym_case] = ACTIONS(1210),
    [anon_sym_default] = ACTIONS(1210),
    [anon_sym_while] = ACTIONS(1210),
    [anon_sym_do] = ACTIONS(1210),
    [anon_sym_for] = ACTIONS(1210),
    [anon_sym_return] = ACTIONS(1210),
    [anon_sym_break] = ACTIONS(1210),
    [anon_sym_continue] = ACTIONS(1210),
    [anon_sym_goto] = ACTIONS(1210),
    [anon_sym___try] = ACTIONS(1210),
    [anon_sym___leave] = ACTIONS(1210),
    [anon_sym_DASH_DASH] = ACTIONS(1212),
    [anon_sym_PLUS_PLUS] = ACTIONS(1212),
    [anon_sym_sizeof] = ACTIONS(1210),
    [anon_sym___alignof__] = ACTIONS(1210),
    [anon_sym___alignof] = ACTIONS(1210),
    [anon_sym__alignof] = ACTIONS(1210),
    [anon_sym_alignof] = ACTIONS(1210),
    [anon_sym__Alignof] = ACTIONS(1210),
    [anon_sym_offsetof] = ACTIONS(1210),
    [anon_sym__Generic] = ACTIONS(1210),
    [anon_sym_asm] = ACTIONS(1210),
    [anon_sym___asm__] = ACTIONS(1210),
    [anon_sym___asm] = ACTIONS(1210),
    [sym_number_literal] = ACTIONS(1212),
    [anon_sym_L_SQUOTE] = ACTIONS(1212),
    [anon_sym_u_SQUOTE] = ACTIONS(1212),
    [anon_sym_U_SQUOTE] = ACTIONS(1212),
    [anon_sym_u8_SQUOTE] = ACTIONS(1212),
    [anon_sym_SQUOTE] = ACTIONS(1212),
    [anon_sym_L_DQUOTE] = ACTIONS(1212),
    [anon_sym_u_DQUOTE] = ACTIONS(1212),
    [anon_sym_U_DQUOTE] = ACTIONS(1212),
    [anon_sym_u8_DQUOTE] = ACTIONS(1212),
    [anon_sym_DQUOTE] = ACTIONS(1212),
    [sym_true] = ACTIONS(1210),
    [sym_false] = ACTIONS(1210),
    [anon_sym_NULL] = ACTIONS(1210),
    [anon_sym_nullptr] = ACTIONS(1210),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(248)] = {
    [sym_identifier] = ACTIONS(1254),
    [aux_sym_preproc_include_token1] = ACTIONS(1254),
    [aux_sym_preproc_def_token1] = ACTIONS(1254),
    [aux_sym_preproc_if_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1254),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1254),
    [sym_preproc_directive] = ACTIONS(1254),
    [anon_sym_LPAREN2] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_TILDE] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_STAR] = ACTIONS(1256),
    [anon_sym_AMP] = ACTIONS(1256),
    [anon_sym_SEMI] = ACTIONS(1256),
    [anon_sym___extension__] = ACTIONS(1254),
    [anon_sym_typedef] = ACTIONS(1254),
    [anon_sym_extern] = ACTIONS(1254),
    [anon_sym___attribute__] = ACTIONS(1254),
    [anon_sym___attribute] = ACTIONS(1254),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1256),
    [anon_sym___declspec] = ACTIONS(1254),
    [anon_sym___cdecl] = ACTIONS(1254),
    [anon_sym___clrcall] = ACTIONS(1254),
    [anon_sym___stdcall] = ACTIONS(1254),
    [anon_sym___fastcall] = ACTIONS(1254),
    [anon_sym___thiscall] = ACTIONS(1254),
    [anon_sym___vectorcall] = ACTIONS(1254),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_RBRACE] = ACTIONS(1256),
    [anon_sym_signed] = ACTIONS(1254),
    [anon_sym_unsigned] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_auto] = ACTIONS(1254),
    [anon_sym_register] = ACTIONS(1254),
    [anon_sym_inline] = ACTIONS(1254),
    [anon_sym___inline] = ACTIONS(1254),
    [anon_sym___inline__] = ACTIONS(1254),
    [anon_sym___forceinline] = ACTIONS(1254),
    [anon_sym_thread_local] = ACTIONS(1254),
    [anon_sym___thread] = ACTIONS(1254),
    [anon_sym_const] = ACTIONS(1254),
    [anon_sym_constexpr] = ACTIONS(1254),
    [anon_sym_volatile] = ACTIONS(1254),
    [anon_sym_restrict] = ACTIONS(1254),
    [anon_sym___restrict__] = ACTIONS(1254),
    [anon_sym__Atomic] = ACTIONS(1254),
    [anon_sym__Noreturn] = ACTIONS(1254),
    [anon_sym_noreturn] = ACTIONS(1254),
    [anon_sym__Nonnull] = ACTIONS(1254),
    [anon_sym_alignas] = ACTIONS(1254),
    [anon_sym__Alignas] = ACTIONS(1254),
    [sym_primitive_type] = ACTIONS(1254),
    [anon_sym_enum] = ACTIONS(1254),
    [anon_sym_struct] = ACTIONS(1254),
    [anon_sym_union] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_goto] = ACTIONS(1254),
    [anon_sym___try] = ACTIONS(1254),
    [anon_sym___leave] = ACTIONS(1254),
    [anon_sym_DASH_DASH] = ACTIONS(1256),
    [anon_sym_PLUS_PLUS] = ACTIONS(1256),
    [anon_sym_sizeof] = ACTIONS(1254),
    [anon_sym___alignof__] = ACTIONS(1254),
    [anon_sym___alignof] = ACTIONS(1254),
    [anon_sym__alignof] = ACTIONS(1254),
    [anon_sym_alignof] = ACTIONS(1254),
    [anon_sym__Alignof] = ACTIONS(1254),
    [anon_sym_offsetof] = ACTIONS(1254),
    [anon_sym__Generic] = ACTIONS(1254),
    [anon_sym_asm] = ACTIONS(1254),
    [anon_sym___asm__] = ACTIONS(1254),
    [anon_sym___asm] = ACTIONS(1254),
    [sym_number_literal] = ACTIONS(1256),
    [anon_sym_L_SQUOTE] = ACTIONS(1256),
    [anon_sym_u_SQUOTE] = ACTIONS(1256),
    [anon_sym_U_SQUOTE] = ACTIONS(1256),
    [anon_sym_u8_SQUOTE] = ACTIONS(1256),
    [anon_sym_SQUOTE] = ACTIONS(1256),
    [anon_sym_L_DQUOTE] = ACTIONS(1256),
    [anon_sym_u_DQUOTE] = ACTIONS(1256),
    [anon_sym_U_DQUOTE] = ACTIONS(1256),
    [anon_sym_u8_DQUOTE] = ACTIONS(1256),
    [anon_sym_DQUOTE] = ACTIONS(1256),
    [sym_true] = ACTIONS(1254),
    [sym_false] = ACTIONS(1254),
    [anon_sym_NULL] = ACTIONS(1254),
    [anon_sym_nullptr] = ACTIONS(1254),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(249)] = {
    [sym_identifier] = ACTIONS(1218),
    [aux_sym_preproc_include_token1] = ACTIONS(1218),
    [aux_sym_preproc_def_token1] = ACTIONS(1218),
    [aux_sym_preproc_if_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1218),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1218),
    [sym_preproc_directive] = ACTIONS(1218),
    [anon_sym_LPAREN2] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_TILDE] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1218),
    [anon_sym_PLUS] = ACTIONS(1218),
    [anon_sym_STAR] = ACTIONS(1220),
    [anon_sym_AMP] = ACTIONS(1220),
    [anon_sym_SEMI] = ACTIONS(1220),
    [anon_sym___extension__] = ACTIONS(1218),
    [anon_sym_typedef] = ACTIONS(1218),
    [anon_sym_extern] = ACTIONS(1218),
    [anon_sym___attribute__] = ACTIONS(1218),
    [anon_sym___attribute] = ACTIONS(1218),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1220),
    [anon_sym___declspec] = ACTIONS(1218),
    [anon_sym___cdecl] = ACTIONS(1218),
    [anon_sym___clrcall] = ACTIONS(1218),
    [anon_sym___stdcall] = ACTIONS(1218),
    [anon_sym___fastcall] = ACTIONS(1218),
    [anon_sym___thiscall] = ACTIONS(1218),
    [anon_sym___vectorcall] = ACTIONS(1218),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_RBRACE] = ACTIONS(1220),
    [anon_sym_signed] = ACTIONS(1218),
    [anon_sym_unsigned] = ACTIONS(1218),
    [anon_sym_long] = ACTIONS(1218),
    [anon_sym_short] = ACTIONS(1218),
    [anon_sym_static] = ACTIONS(1218),
    [anon_sym_auto] = ACTIONS(1218),
    [anon_sym_register] = ACTIONS(1218),
    [anon_sym_inline] = ACTIONS(1218),
    [anon_sym___inline] = ACTIONS(1218),
    [anon_sym___inline__] = ACTIONS(1218),
    [anon_sym___forceinline] = ACTIONS(1218),
    [anon_sym_thread_local] = ACTIONS(1218),
    [anon_sym___thread] = ACTIONS(1218),
    [anon_sym_const] = ACTIONS(1218),
    [anon_sym_constexpr] = ACTIONS(1218),
    [anon_sym_volatile] = ACTIONS(1218),
    [anon_sym_restrict] = ACTIONS(1218),
    [anon_sym___restrict__] = ACTIONS(1218),
    [anon_sym__Atomic] = ACTIONS(1218),
    [anon_sym__Noreturn] = ACTIONS(1218),
    [anon_sym_noreturn] = ACTIONS(1218),
    [anon_sym__Nonnull] = ACTIONS(1218),
    [anon_sym_alignas] = ACTIONS(1218),
    [anon_sym__Alignas] = ACTIONS(1218),
    [sym_primitive_type] = ACTIONS(1218),
    [anon_sym_enum] = ACTIONS(1218),
    [anon_sym_struct] = ACTIONS(1218),
    [anon_sym_union] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1218),
    [anon_sym_else] = ACTIONS(1218),
    [anon_sym_switch] = ACTIONS(1218),
    [anon_sym_case] = ACTIONS(1218),
    [anon_sym_default] = ACTIONS(1218),
    [anon_sym_while] = ACTIONS(1218),
    [anon_sym_do] = ACTIONS(1218),
    [anon_sym_for] = ACTIONS(1218),
    [anon_sym_return] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1218),
    [anon_sym_continue] = ACTIONS(1218),
    [anon_sym_goto] = ACTIONS(1218),
    [anon_sym___try] = ACTIONS(1218),
    [anon_sym___leave] = ACTIONS(1218),
    [anon_sym_DASH_DASH] = ACTIONS(1220),
    [anon_sym_PLUS_PLUS] = ACTIONS(1220),
    [anon_sym_sizeof] = ACTIONS(1218),
    [anon_sym___alignof__] = ACTIONS(1218),
    [anon_sym___alignof] = ACTIONS(1218),
    [anon_sym__alignof] = ACTIONS(1218),
    [anon_sym_alignof] = ACTIONS(1218),
    [anon_sym__Alignof] = ACTIONS(1218),
    [anon_sym_offsetof] = ACTIONS(1218),
    [anon_sym__Generic] = ACTIONS(1218),
    [anon_sym_asm] = ACTIONS(1218),
    [anon_sym___asm__] = ACTIONS(1218),
    [anon_sym___asm] = ACTIONS(1218),
    [sym_number_literal] = ACTIONS(1220),
    [anon_sym_L_SQUOTE] = ACTIONS(1220),
    [anon_sym_u_SQUOTE] = ACTIONS(1220),
    [anon_sym_U_SQUOTE] = ACTIONS(1220),
    [anon_sym_u8_SQUOTE] = ACTIONS(1220),
    [anon_sym_SQUOTE] = ACTIONS(1220),
    [anon_sym_L_DQUOTE] = ACTIONS(1220),
    [anon_sym_u_DQUOTE] = ACTIONS(1220),
    [anon_sym_U_DQUOTE] = ACTIONS(1220),
    [anon_sym_u8_DQUOTE] = ACTIONS(1220),
    [anon_sym_DQUOTE] = ACTIONS(1220),
    [sym_true] = ACTIONS(1218),
    [sym_false] = ACTIONS(1218),
    [anon_sym_NULL] = ACTIONS(1218),
    [anon_sym_nullptr] = ACTIONS(1218),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(250)] = {
    [ts_builtin_sym_end] = ACTIONS(1224),
    [sym_identifier] = ACTIONS(1222),
    [aux_sym_preproc_include_token1] = ACTIONS(1222),
    [aux_sym_preproc_def_token1] = ACTIONS(1222),
    [aux_sym_preproc_if_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1222),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1222),
    [sym_preproc_directive] = ACTIONS(1222),
    [anon_sym_LPAREN2] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1222),
    [anon_sym_PLUS] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym___extension__] = ACTIONS(1222),
    [anon_sym_typedef] = ACTIONS(1222),
    [anon_sym_extern] = ACTIONS(1222),
    [anon_sym___attribute__] = ACTIONS(1222),
    [anon_sym___attribute] = ACTIONS(1222),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1224),
    [anon_sym___declspec] = ACTIONS(1222),
    [anon_sym___cdecl] = ACTIONS(1222),
    [anon_sym___clrcall] = ACTIONS(1222),
    [anon_sym___stdcall] = ACTIONS(1222),
    [anon_sym___fastcall] = ACTIONS(1222),
    [anon_sym___thiscall] = ACTIONS(1222),
    [anon_sym___vectorcall] = ACTIONS(1222),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_signed] = ACTIONS(1222),
    [anon_sym_unsigned] = ACTIONS(1222),
    [anon_sym_long] = ACTIONS(1222),
    [anon_sym_short] = ACTIONS(1222),
    [anon_sym_static] = ACTIONS(1222),
    [anon_sym_auto] = ACTIONS(1222),
    [anon_sym_register] = ACTIONS(1222),
    [anon_sym_inline] = ACTIONS(1222),
    [anon_sym___inline] = ACTIONS(1222),
    [anon_sym___inline__] = ACTIONS(1222),
    [anon_sym___forceinline] = ACTIONS(1222),
    [anon_sym_thread_local] = ACTIONS(1222),
    [anon_sym___thread] = ACTIONS(1222),
    [anon_sym_const] = ACTIONS(1222),
    [anon_sym_constexpr] = ACTIONS(1222),
    [anon_sym_volatile] = ACTIONS(1222),
    [anon_sym_restrict] = ACTIONS(1222),
    [anon_sym___restrict__] = ACTIONS(1222),
    [anon_sym__Atomic] = ACTIONS(1222),
    [anon_sym__Noreturn] = ACTIONS(1222),
    [anon_sym_noreturn] = ACTIONS(1222),
    [anon_sym__Nonnull] = ACTIONS(1222),
    [anon_sym_alignas] = ACTIONS(1222),
    [anon_sym__Alignas] = ACTIONS(1222),
    [sym_primitive_type] = ACTIONS(1222),
    [anon_sym_enum] = ACTIONS(1222),
    [anon_sym_struct] = ACTIONS(1222),
    [anon_sym_union] = ACTIONS(1222),
    [anon_sym_if] = ACTIONS(1222),
    [anon_sym_else] = ACTIONS(1222),
    [anon_sym_switch] = ACTIONS(1222),
    [anon_sym_case] = ACTIONS(1222),
    [anon_sym_default] = ACTIONS(1222),
    [anon_sym_while] = ACTIONS(1222),
    [anon_sym_do] = ACTIONS(1222),
    [anon_sym_for] = ACTIONS(1222),
    [anon_sym_return] = ACTIONS(1222),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_goto] = ACTIONS(1222),
    [anon_sym___try] = ACTIONS(1222),
    [anon_sym___leave] = ACTIONS(1222),
    [anon_sym_DASH_DASH] = ACTIONS(1224),
    [anon_sym_PLUS_PLUS] = ACTIONS(1224),
    [anon_sym_sizeof] = ACTIONS(1222),
    [anon_sym___alignof__] = ACTIONS(1222),
    [anon_sym___alignof] = ACTIONS(1222),
    [anon_sym__alignof] = ACTIONS(1222),
    [anon_sym_alignof] = ACTIONS(1222),
    [anon_sym__Alignof] = ACTIONS(1222),
    [anon_sym_offsetof] = ACTIONS(1222),
    [anon_sym__Generic] = ACTIONS(1222),
    [anon_sym_asm] = ACTIONS(1222),
    [anon_sym___asm__] = ACTIONS(1222),
    [anon_sym___asm] = ACTIONS(1222),
    [sym_number_literal] = ACTIONS(1224),
    [anon_sym_L_SQUOTE] = ACTIONS(1224),
    [anon_sym_u_SQUOTE] = ACTIONS(1224),
    [anon_sym_U_SQUOTE] = ACTIONS(1224),
    [anon_sym_u8_SQUOTE] = ACTIONS(1224),
    [anon_sym_SQUOTE] = ACTIONS(1224),
    [anon_sym_L_DQUOTE] = ACTIONS(1224),
    [anon_sym_u_DQUOTE] = ACTIONS(1224),
    [anon_sym_U_DQUOTE] = ACTIONS(1224),
    [anon_sym_u8_DQUOTE] = ACTIONS(1224),
    [anon_sym_DQUOTE] = ACTIONS(1224),
    [sym_true] = ACTIONS(1222),
    [sym_false] = ACTIONS(1222),
    [anon_sym_NULL] = ACTIONS(1222),
    [anon_sym_nullptr] = ACTIONS(1222),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(251)] = {
    [sym_identifier] = ACTIONS(1198),
    [aux_sym_preproc_include_token1] = ACTIONS(1198),
    [aux_sym_preproc_def_token1] = ACTIONS(1198),
    [aux_sym_preproc_if_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1198),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1198),
    [sym_preproc_directive] = ACTIONS(1198),
    [anon_sym_LPAREN2] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1198),
    [anon_sym_PLUS] = ACTIONS(1198),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym___extension__] = ACTIONS(1198),
    [anon_sym_typedef] = ACTIONS(1198),
    [anon_sym_extern] = ACTIONS(1198),
    [anon_sym___attribute__] = ACTIONS(1198),
    [anon_sym___attribute] = ACTIONS(1198),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1200),
    [anon_sym___declspec] = ACTIONS(1198),
    [anon_sym___cdecl] = ACTIONS(1198),
    [anon_sym___clrcall] = ACTIONS(1198),
    [anon_sym___stdcall] = ACTIONS(1198),
    [anon_sym___fastcall] = ACTIONS(1198),
    [anon_sym___thiscall] = ACTIONS(1198),
    [anon_sym___vectorcall] = ACTIONS(1198),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_RBRACE] = ACTIONS(1200),
    [anon_sym_signed] = ACTIONS(1198),
    [anon_sym_unsigned] = ACTIONS(1198),
    [anon_sym_long] = ACTIONS(1198),
    [anon_sym_short] = ACTIONS(1198),
    [anon_sym_static] = ACTIONS(1198),
    [anon_sym_auto] = ACTIONS(1198),
    [anon_sym_register] = ACTIONS(1198),
    [anon_sym_inline] = ACTIONS(1198),
    [anon_sym___inline] = ACTIONS(1198),
    [anon_sym___inline__] = ACTIONS(1198),
    [anon_sym___forceinline] = ACTIONS(1198),
    [anon_sym_thread_local] = ACTIONS(1198),
    [anon_sym___thread] = ACTIONS(1198),
    [anon_sym_const] = ACTIONS(1198),
    [anon_sym_constexpr] = ACTIONS(1198),
    [anon_sym_volatile] = ACTIONS(1198),
    [anon_sym_restrict] = ACTIONS(1198),
    [anon_sym___restrict__] = ACTIONS(1198),
    [anon_sym__Atomic] = ACTIONS(1198),
    [anon_sym__Noreturn] = ACTIONS(1198),
    [anon_sym_noreturn] = ACTIONS(1198),
    [anon_sym__Nonnull] = ACTIONS(1198),
    [anon_sym_alignas] = ACTIONS(1198),
    [anon_sym__Alignas] = ACTIONS(1198),
    [sym_primitive_type] = ACTIONS(1198),
    [anon_sym_enum] = ACTIONS(1198),
    [anon_sym_struct] = ACTIONS(1198),
    [anon_sym_union] = ACTIONS(1198),
    [anon_sym_if] = ACTIONS(1198),
    [anon_sym_else] = ACTIONS(1198),
    [anon_sym_switch] = ACTIONS(1198),
    [anon_sym_case] = ACTIONS(1198),
    [anon_sym_default] = ACTIONS(1198),
    [anon_sym_while] = ACTIONS(1198),
    [anon_sym_do] = ACTIONS(1198),
    [anon_sym_for] = ACTIONS(1198),
    [anon_sym_return] = ACTIONS(1198),
    [anon_sym_break] = ACTIONS(1198),
    [anon_sym_continue] = ACTIONS(1198),
    [anon_sym_goto] = ACTIONS(1198),
    [anon_sym___try] = ACTIONS(1198),
    [anon_sym___leave] = ACTIONS(1198),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_sizeof] = ACTIONS(1198),
    [anon_sym___alignof__] = ACTIONS(1198),
    [anon_sym___alignof] = ACTIONS(1198),
    [anon_sym__alignof] = ACTIONS(1198),
    [anon_sym_alignof] = ACTIONS(1198),
    [anon_sym__Alignof] = ACTIONS(1198),
    [anon_sym_offsetof] = ACTIONS(1198),
    [anon_sym__Generic] = ACTIONS(1198),
    [anon_sym_asm] = ACTIONS(1198),
    [anon_sym___asm__] = ACTIONS(1198),
    [anon_sym___asm] = ACTIONS(1198),
    [sym_number_literal] = ACTIONS(1200),
    [anon_sym_L_SQUOTE] = ACTIONS(1200),
    [anon_sym_u_SQUOTE] = ACTIONS(1200),
    [anon_sym_U_SQUOTE] = ACTIONS(1200),
    [anon_sym_u8_SQUOTE] = ACTIONS(1200),
    [anon_sym_SQUOTE] = ACTIONS(1200),
    [anon_sym_L_DQUOTE] = ACTIONS(1200),
    [anon_sym_u_DQUOTE] = ACTIONS(1200),
    [anon_sym_U_DQUOTE] = ACTIONS(1200),
    [anon_sym_u8_DQUOTE] = ACTIONS(1200),
    [anon_sym_DQUOTE] = ACTIONS(1200),
    [sym_true] = ACTIONS(1198),
    [sym_false] = ACTIONS(1198),
    [anon_sym_NULL] = ACTIONS(1198),
    [anon_sym_nullptr] = ACTIONS(1198),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(252)] = {
    [ts_builtin_sym_end] = ACTIONS(1244),
    [sym_identifier] = ACTIONS(1242),
    [aux_sym_preproc_include_token1] = ACTIONS(1242),
    [aux_sym_preproc_def_token1] = ACTIONS(1242),
    [aux_sym_preproc_if_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1242),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1242),
    [sym_preproc_directive] = ACTIONS(1242),
    [anon_sym_LPAREN2] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_TILDE] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1242),
    [anon_sym_PLUS] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(1244),
    [anon_sym_AMP] = ACTIONS(1244),
    [anon_sym_SEMI] = ACTIONS(1244),
    [anon_sym___extension__] = ACTIONS(1242),
    [anon_sym_typedef] = ACTIONS(1242),
    [anon_sym_extern] = ACTIONS(1242),
    [anon_sym___attribute__] = ACTIONS(1242),
    [anon_sym___attribute] = ACTIONS(1242),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1244),
    [anon_sym___declspec] = ACTIONS(1242),
    [anon_sym___cdecl] = ACTIONS(1242),
    [anon_sym___clrcall] = ACTIONS(1242),
    [anon_sym___stdcall] = ACTIONS(1242),
    [anon_sym___fastcall] = ACTIONS(1242),
    [anon_sym___thiscall] = ACTIONS(1242),
    [anon_sym___vectorcall] = ACTIONS(1242),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_signed] = ACTIONS(1242),
    [anon_sym_unsigned] = ACTIONS(1242),
    [anon_sym_long] = ACTIONS(1242),
    [anon_sym_short] = ACTIONS(1242),
    [anon_sym_static] = ACTIONS(1242),
    [anon_sym_auto] = ACTIONS(1242),
    [anon_sym_register] = ACTIONS(1242),
    [anon_sym_inline] = ACTIONS(1242),
    [anon_sym___inline] = ACTIONS(1242),
    [anon_sym___inline__] = ACTIONS(1242),
    [anon_sym___forceinline] = ACTIONS(1242),
    [anon_sym_thread_local] = ACTIONS(1242),
    [anon_sym___thread] = ACTIONS(1242),
    [anon_sym_const] = ACTIONS(1242),
    [anon_sym_constexpr] = ACTIONS(1242),
    [anon_sym_volatile] = ACTIONS(1242),
    [anon_sym_restrict] = ACTIONS(1242),
    [anon_sym___restrict__] = ACTIONS(1242),
    [anon_sym__Atomic] = ACTIONS(1242),
    [anon_sym__Noreturn] = ACTIONS(1242),
    [anon_sym_noreturn] = ACTIONS(1242),
    [anon_sym__Nonnull] = ACTIONS(1242),
    [anon_sym_alignas] = ACTIONS(1242),
    [anon_sym__Alignas] = ACTIONS(1242),
    [sym_primitive_type] = ACTIONS(1242),
    [anon_sym_enum] = ACTIONS(1242),
    [anon_sym_struct] = ACTIONS(1242),
    [anon_sym_union] = ACTIONS(1242),
    [anon_sym_if] = ACTIONS(1242),
    [anon_sym_else] = ACTIONS(1242),
    [anon_sym_switch] = ACTIONS(1242),
    [anon_sym_case] = ACTIONS(1242),
    [anon_sym_default] = ACTIONS(1242),
    [anon_sym_while] = ACTIONS(1242),
    [anon_sym_do] = ACTIONS(1242),
    [anon_sym_for] = ACTIONS(1242),
    [anon_sym_return] = ACTIONS(1242),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_goto] = ACTIONS(1242),
    [anon_sym___try] = ACTIONS(1242),
    [anon_sym___leave] = ACTIONS(1242),
    [anon_sym_DASH_DASH] = ACTIONS(1244),
    [anon_sym_PLUS_PLUS] = ACTIONS(1244),
    [anon_sym_sizeof] = ACTIONS(1242),
    [anon_sym___alignof__] = ACTIONS(1242),
    [anon_sym___alignof] = ACTIONS(1242),
    [anon_sym__alignof] = ACTIONS(1242),
    [anon_sym_alignof] = ACTIONS(1242),
    [anon_sym__Alignof] = ACTIONS(1242),
    [anon_sym_offsetof] = ACTIONS(1242),
    [anon_sym__Generic] = ACTIONS(1242),
    [anon_sym_asm] = ACTIONS(1242),
    [anon_sym___asm__] = ACTIONS(1242),
    [anon_sym___asm] = ACTIONS(1242),
    [sym_number_literal] = ACTIONS(1244),
    [anon_sym_L_SQUOTE] = ACTIONS(1244),
    [anon_sym_u_SQUOTE] = ACTIONS(1244),
    [anon_sym_U_SQUOTE] = ACTIONS(1244),
    [anon_sym_u8_SQUOTE] = ACTIONS(1244),
    [anon_sym_SQUOTE] = ACTIONS(1244),
    [anon_sym_L_DQUOTE] = ACTIONS(1244),
    [anon_sym_u_DQUOTE] = ACTIONS(1244),
    [anon_sym_U_DQUOTE] = ACTIONS(1244),
    [anon_sym_u8_DQUOTE] = ACTIONS(1244),
    [anon_sym_DQUOTE] = ACTIONS(1244),
    [sym_true] = ACTIONS(1242),
    [sym_false] = ACTIONS(1242),
    [anon_sym_NULL] = ACTIONS(1242),
    [anon_sym_nullptr] = ACTIONS(1242),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(253)] = {
    [ts_builtin_sym_end] = ACTIONS(1208),
    [sym_identifier] = ACTIONS(1206),
    [aux_sym_preproc_include_token1] = ACTIONS(1206),
    [aux_sym_preproc_def_token1] = ACTIONS(1206),
    [aux_sym_preproc_if_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1206),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1206),
    [sym_preproc_directive] = ACTIONS(1206),
    [anon_sym_LPAREN2] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_TILDE] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_STAR] = ACTIONS(1208),
    [anon_sym_AMP] = ACTIONS(1208),
    [anon_sym_SEMI] = ACTIONS(1208),
    [anon_sym___extension__] = ACTIONS(1206),
    [anon_sym_typedef] = ACTIONS(1206),
    [anon_sym_extern] = ACTIONS(1206),
    [anon_sym___attribute__] = ACTIONS(1206),
    [anon_sym___attribute] = ACTIONS(1206),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1208),
    [anon_sym___declspec] = ACTIONS(1206),
    [anon_sym___cdecl] = ACTIONS(1206),
    [anon_sym___clrcall] = ACTIONS(1206),
    [anon_sym___stdcall] = ACTIONS(1206),
    [anon_sym___fastcall] = ACTIONS(1206),
    [anon_sym___thiscall] = ACTIONS(1206),
    [anon_sym___vectorcall] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_signed] = ACTIONS(1206),
    [anon_sym_unsigned] = ACTIONS(1206),
    [anon_sym_long] = ACTIONS(1206),
    [anon_sym_short] = ACTIONS(1206),
    [anon_sym_static] = ACTIONS(1206),
    [anon_sym_auto] = ACTIONS(1206),
    [anon_sym_register] = ACTIONS(1206),
    [anon_sym_inline] = ACTIONS(1206),
    [anon_sym___inline] = ACTIONS(1206),
    [anon_sym___inline__] = ACTIONS(1206),
    [anon_sym___forceinline] = ACTIONS(1206),
    [anon_sym_thread_local] = ACTIONS(1206),
    [anon_sym___thread] = ACTIONS(1206),
    [anon_sym_const] = ACTIONS(1206),
    [anon_sym_constexpr] = ACTIONS(1206),
    [anon_sym_volatile] = ACTIONS(1206),
    [anon_sym_restrict] = ACTIONS(1206),
    [anon_sym___restrict__] = ACTIONS(1206),
    [anon_sym__Atomic] = ACTIONS(1206),
    [anon_sym__Noreturn] = ACTIONS(1206),
    [anon_sym_noreturn] = ACTIONS(1206),
    [anon_sym__Nonnull] = ACTIONS(1206),
    [anon_sym_alignas] = ACTIONS(1206),
    [anon_sym__Alignas] = ACTIONS(1206),
    [sym_primitive_type] = ACTIONS(1206),
    [anon_sym_enum] = ACTIONS(1206),
    [anon_sym_struct] = ACTIONS(1206),
    [anon_sym_union] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1206),
    [anon_sym_else] = ACTIONS(1206),
    [anon_sym_switch] = ACTIONS(1206),
    [anon_sym_case] = ACTIONS(1206),
    [anon_sym_default] = ACTIONS(1206),
    [anon_sym_while] = ACTIONS(1206),
    [anon_sym_do] = ACTIONS(1206),
    [anon_sym_for] = ACTIONS(1206),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_break] = ACTIONS(1206),
    [anon_sym_continue] = ACTIONS(1206),
    [anon_sym_goto] = ACTIONS(1206),
    [anon_sym___try] = ACTIONS(1206),
    [anon_sym___leave] = ACTIONS(1206),
    [anon_sym_DASH_DASH] = ACTIONS(1208),
    [anon_sym_PLUS_PLUS] = ACTIONS(1208),
    [anon_sym_sizeof] = ACTIONS(1206),
    [anon_sym___alignof__] = ACTIONS(1206),
    [anon_sym___alignof] = ACTIONS(1206),
    [anon_sym__alignof] = ACTIONS(1206),
    [anon_sym_alignof] = ACTIONS(1206),
    [anon_sym__Alignof] = ACTIONS(1206),
    [anon_sym_offsetof] = ACTIONS(1206),
    [anon_sym__Generic] = ACTIONS(1206),
    [anon_sym_asm] = ACTIONS(1206),
    [anon_sym___asm__] = ACTIONS(1206),
    [anon_sym___asm] = ACTIONS(1206),
    [sym_number_literal] = ACTIONS(1208),
    [anon_sym_L_SQUOTE] = ACTIONS(1208),
    [anon_sym_u_SQUOTE] = ACTIONS(1208),
    [anon_sym_U_SQUOTE] = ACTIONS(1208),
    [anon_sym_u8_SQUOTE] = ACTIONS(1208),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_L_DQUOTE] = ACTIONS(1208),
    [anon_sym_u_DQUOTE] = ACTIONS(1208),
    [anon_sym_U_DQUOTE] = ACTIONS(1208),
    [anon_sym_u8_DQUOTE] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1208),
    [sym_true] = ACTIONS(1206),
    [sym_false] = ACTIONS(1206),
    [anon_sym_NULL] = ACTIONS(1206),
    [anon_sym_nullptr] = ACTIONS(1206),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(254)] = {
    [sym_identifier] = ACTIONS(1170),
    [aux_sym_preproc_include_token1] = ACTIONS(1170),
    [aux_sym_preproc_def_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token1] = ACTIONS(1170),
    [aux_sym_preproc_if_token2] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1170),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1170),
    [sym_preproc_directive] = ACTIONS(1170),
    [anon_sym_LPAREN2] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_TILDE] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1170),
    [anon_sym_PLUS] = ACTIONS(1170),
    [anon_sym_STAR] = ACTIONS(1172),
    [anon_sym_AMP] = ACTIONS(1172),
    [anon_sym_SEMI] = ACTIONS(1172),
    [anon_sym___extension__] = ACTIONS(1170),
    [anon_sym_typedef] = ACTIONS(1170),
    [anon_sym_extern] = ACTIONS(1170),
    [anon_sym___attribute__] = ACTIONS(1170),
    [anon_sym___attribute] = ACTIONS(1170),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1172),
    [anon_sym___declspec] = ACTIONS(1170),
    [anon_sym___cdecl] = ACTIONS(1170),
    [anon_sym___clrcall] = ACTIONS(1170),
    [anon_sym___stdcall] = ACTIONS(1170),
    [anon_sym___fastcall] = ACTIONS(1170),
    [anon_sym___thiscall] = ACTIONS(1170),
    [anon_sym___vectorcall] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_signed] = ACTIONS(1170),
    [anon_sym_unsigned] = ACTIONS(1170),
    [anon_sym_long] = ACTIONS(1170),
    [anon_sym_short] = ACTIONS(1170),
    [anon_sym_static] = ACTIONS(1170),
    [anon_sym_auto] = ACTIONS(1170),
    [anon_sym_register] = ACTIONS(1170),
    [anon_sym_inline] = ACTIONS(1170),
    [anon_sym___inline] = ACTIONS(1170),
    [anon_sym___inline__] = ACTIONS(1170),
    [anon_sym___forceinline] = ACTIONS(1170),
    [anon_sym_thread_local] = ACTIONS(1170),
    [anon_sym___thread] = ACTIONS(1170),
    [anon_sym_const] = ACTIONS(1170),
    [anon_sym_constexpr] = ACTIONS(1170),
    [anon_sym_volatile] = ACTIONS(1170),
    [anon_sym_restrict] = ACTIONS(1170),
    [anon_sym___restrict__] = ACTIONS(1170),
    [anon_sym__Atomic] = ACTIONS(1170),
    [anon_sym__Noreturn] = ACTIONS(1170),
    [anon_sym_noreturn] = ACTIONS(1170),
    [anon_sym__Nonnull] = ACTIONS(1170),
    [anon_sym_alignas] = ACTIONS(1170),
    [anon_sym__Alignas] = ACTIONS(1170),
    [sym_primitive_type] = ACTIONS(1170),
    [anon_sym_enum] = ACTIONS(1170),
    [anon_sym_struct] = ACTIONS(1170),
    [anon_sym_union] = ACTIONS(1170),
    [anon_sym_if] = ACTIONS(1170),
    [anon_sym_else] = ACTIONS(1170),
    [anon_sym_switch] = ACTIONS(1170),
    [anon_sym_case] = ACTIONS(1170),
    [anon_sym_default] = ACTIONS(1170),
    [anon_sym_while] = ACTIONS(1170),
    [anon_sym_do] = ACTIONS(1170),
    [anon_sym_for] = ACTIONS(1170),
    [anon_sym_return] = ACTIONS(1170),
    [anon_sym_break] = ACTIONS(1170),
    [anon_sym_continue] = ACTIONS(1170),
    [anon_sym_goto] = ACTIONS(1170),
    [anon_sym___try] = ACTIONS(1170),
    [anon_sym___leave] = ACTIONS(1170),
    [anon_sym_DASH_DASH] = ACTIONS(1172),
    [anon_sym_PLUS_PLUS] = ACTIONS(1172),
    [anon_sym_sizeof] = ACTIONS(1170),
    [anon_sym___alignof__] = ACTIONS(1170),
    [anon_sym___alignof] = ACTIONS(1170),
    [anon_sym__alignof] = ACTIONS(1170),
    [anon_sym_alignof] = ACTIONS(1170),
    [anon_sym__Alignof] = ACTIONS(1170),
    [anon_sym_offsetof] = ACTIONS(1170),
    [anon_sym__Generic] = ACTIONS(1170),
    [anon_sym_asm] = ACTIONS(1170),
    [anon_sym___asm__] = ACTIONS(1170),
    [anon_sym___asm] = ACTIONS(1170),
    [sym_number_literal] = ACTIONS(1172),
    [anon_sym_L_SQUOTE] = ACTIONS(1172),
    [anon_sym_u_SQUOTE] = ACTIONS(1172),
    [anon_sym_U_SQUOTE] = ACTIONS(1172),
    [anon_sym_u8_SQUOTE] = ACTIONS(1172),
    [anon_sym_SQUOTE] = ACTIONS(1172),
    [anon_sym_L_DQUOTE] = ACTIONS(1172),
    [anon_sym_u_DQUOTE] = ACTIONS(1172),
    [anon_sym_U_DQUOTE] = ACTIONS(1172),
    [anon_sym_u8_DQUOTE] = ACTIONS(1172),
    [anon_sym_DQUOTE] = ACTIONS(1172),
    [sym_true] = ACTIONS(1170),
    [sym_false] = ACTIONS(1170),
    [anon_sym_NULL] = ACTIONS(1170),
    [anon_sym_nullptr] = ACTIONS(1170),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(255)] = {
    [ts_builtin_sym_end] = ACTIONS(1272),
    [sym_identifier] = ACTIONS(1270),
    [aux_sym_preproc_include_token1] = ACTIONS(1270),
    [aux_sym_preproc_def_token1] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1272),
    [aux_sym_preproc_if_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1270),
    [sym_preproc_directive] = ACTIONS(1270),
    [anon_sym_LPAREN2] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_TILDE] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_AMP] = ACTIONS(1272),
    [anon_sym_SEMI] = ACTIONS(1272),
    [anon_sym___extension__] = ACTIONS(1270),
    [anon_sym_typedef] = ACTIONS(1270),
    [anon_sym_extern] = ACTIONS(1270),
    [anon_sym___attribute__] = ACTIONS(1270),
    [anon_sym___attribute] = ACTIONS(1270),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1272),
    [anon_sym___declspec] = ACTIONS(1270),
    [anon_sym___cdecl] = ACTIONS(1270),
    [anon_sym___clrcall] = ACTIONS(1270),
    [anon_sym___stdcall] = ACTIONS(1270),
    [anon_sym___fastcall] = ACTIONS(1270),
    [anon_sym___thiscall] = ACTIONS(1270),
    [anon_sym___vectorcall] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_RBRACE] = ACTIONS(1272),
    [anon_sym_signed] = ACTIONS(1270),
    [anon_sym_unsigned] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_auto] = ACTIONS(1270),
    [anon_sym_register] = ACTIONS(1270),
    [anon_sym_inline] = ACTIONS(1270),
    [anon_sym___inline] = ACTIONS(1270),
    [anon_sym___inline__] = ACTIONS(1270),
    [anon_sym___forceinline] = ACTIONS(1270),
    [anon_sym_thread_local] = ACTIONS(1270),
    [anon_sym___thread] = ACTIONS(1270),
    [anon_sym_const] = ACTIONS(1270),
    [anon_sym_constexpr] = ACTIONS(1270),
    [anon_sym_volatile] = ACTIONS(1270),
    [anon_sym_restrict] = ACTIONS(1270),
    [anon_sym___restrict__] = ACTIONS(1270),
    [anon_sym__Atomic] = ACTIONS(1270),
    [anon_sym__Noreturn] = ACTIONS(1270),
    [anon_sym_noreturn] = ACTIONS(1270),
    [anon_sym__Nonnull] = ACTIONS(1270),
    [anon_sym_alignas] = ACTIONS(1270),
    [anon_sym__Alignas] = ACTIONS(1270),
    [sym_primitive_type] = ACTIONS(1270),
    [anon_sym_enum] = ACTIONS(1270),
    [anon_sym_struct] = ACTIONS(1270),
    [anon_sym_union] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_goto] = ACTIONS(1270),
    [anon_sym_DASH_DASH] = ACTIONS(1272),
    [anon_sym_PLUS_PLUS] = ACTIONS(1272),
    [anon_sym_sizeof] = ACTIONS(1270),
    [anon_sym___alignof__] = ACTIONS(1270),
    [anon_sym___alignof] = ACTIONS(1270),
    [anon_sym__alignof] = ACTIONS(1270),
    [anon_sym_alignof] = ACTIONS(1270),
    [anon_sym__Alignof] = ACTIONS(1270),
    [anon_sym_offsetof] = ACTIONS(1270),
    [anon_sym__Generic] = ACTIONS(1270),
    [anon_sym_asm] = ACTIONS(1270),
    [anon_sym___asm__] = ACTIONS(1270),
    [anon_sym___asm] = ACTIONS(1270),
    [sym_number_literal] = ACTIONS(1272),
    [anon_sym_L_SQUOTE] = ACTIONS(1272),
    [anon_sym_u_SQUOTE] = ACTIONS(1272),
    [anon_sym_U_SQUOTE] = ACTIONS(1272),
    [anon_sym_u8_SQUOTE] = ACTIONS(1272),
    [anon_sym_SQUOTE] = ACTIONS(1272),
    [anon_sym_L_DQUOTE] = ACTIONS(1272),
    [anon_sym_u_DQUOTE] = ACTIONS(1272),
    [anon_sym_U_DQUOTE] = ACTIONS(1272),
    [anon_sym_u8_DQUOTE] = ACTIONS(1272),
    [anon_sym_DQUOTE] = ACTIONS(1272),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [anon_sym_NULL] = ACTIONS(1270),
    [anon_sym_nullptr] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(256)] = {
    [sym_identifier] = ACTIONS(1368),
    [aux_sym_preproc_include_token1] = ACTIONS(1368),
    [aux_sym_preproc_def_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1368),
    [sym_preproc_directive] = ACTIONS(1368),
    [anon_sym_LPAREN2] = ACTIONS(1370),
    [anon_sym_BANG] = ACTIONS(1370),
    [anon_sym_TILDE] = ACTIONS(1370),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_PLUS] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1370),
    [anon_sym_AMP] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1370),
    [anon_sym___extension__] = ACTIONS(1368),
    [anon_sym_typedef] = ACTIONS(1368),
    [anon_sym_extern] = ACTIONS(1368),
    [anon_sym___attribute__] = ACTIONS(1368),
    [anon_sym___attribute] = ACTIONS(1368),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1370),
    [anon_sym___declspec] = ACTIONS(1368),
    [anon_sym___cdecl] = ACTIONS(1368),
    [anon_sym___clrcall] = ACTIONS(1368),
    [anon_sym___stdcall] = ACTIONS(1368),
    [anon_sym___fastcall] = ACTIONS(1368),
    [anon_sym___thiscall] = ACTIONS(1368),
    [anon_sym___vectorcall] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1370),
    [anon_sym_RBRACE] = ACTIONS(1370),
    [anon_sym_signed] = ACTIONS(1368),
    [anon_sym_unsigned] = ACTIONS(1368),
    [anon_sym_long] = ACTIONS(1368),
    [anon_sym_short] = ACTIONS(1368),
    [anon_sym_static] = ACTIONS(1368),
    [anon_sym_auto] = ACTIONS(1368),
    [anon_sym_register] = ACTIONS(1368),
    [anon_sym_inline] = ACTIONS(1368),
    [anon_sym___inline] = ACTIONS(1368),
    [anon_sym___inline__] = ACTIONS(1368),
    [anon_sym___forceinline] = ACTIONS(1368),
    [anon_sym_thread_local] = ACTIONS(1368),
    [anon_sym___thread] = ACTIONS(1368),
    [anon_sym_const] = ACTIONS(1368),
    [anon_sym_constexpr] = ACTIONS(1368),
    [anon_sym_volatile] = ACTIONS(1368),
    [anon_sym_restrict] = ACTIONS(1368),
    [anon_sym___restrict__] = ACTIONS(1368),
    [anon_sym__Atomic] = ACTIONS(1368),
    [anon_sym__Noreturn] = ACTIONS(1368),
    [anon_sym_noreturn] = ACTIONS(1368),
    [anon_sym__Nonnull] = ACTIONS(1368),
    [anon_sym_alignas] = ACTIONS(1368),
    [anon_sym__Alignas] = ACTIONS(1368),
    [sym_primitive_type] = ACTIONS(1368),
    [anon_sym_enum] = ACTIONS(1368),
    [anon_sym_struct] = ACTIONS(1368),
    [anon_sym_union] = ACTIONS(1368),
    [anon_sym_if] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1368),
    [anon_sym_case] = ACTIONS(1368),
    [anon_sym_default] = ACTIONS(1368),
    [anon_sym_while] = ACTIONS(1368),
    [anon_sym_do] = ACTIONS(1368),
    [anon_sym_for] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1368),
    [anon_sym_break] = ACTIONS(1368),
    [anon_sym_continue] = ACTIONS(1368),
    [anon_sym_goto] = ACTIONS(1368),
    [anon_sym___try] = ACTIONS(1368),
    [anon_sym___leave] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1370),
    [anon_sym_PLUS_PLUS] = ACTIONS(1370),
    [anon_sym_sizeof] = ACTIONS(1368),
    [anon_sym___alignof__] = ACTIONS(1368),
    [anon_sym___alignof] = ACTIONS(1368),
    [anon_sym__alignof] = ACTIONS(1368),
    [anon_sym_alignof] = ACTIONS(1368),
    [anon_sym__Alignof] = ACTIONS(1368),
    [anon_sym_offsetof] = ACTIONS(1368),
    [anon_sym__Generic] = ACTIONS(1368),
    [anon_sym_asm] = ACTIONS(1368),
    [anon_sym___asm__] = ACTIONS(1368),
    [anon_sym___asm] = ACTIONS(1368),
    [sym_number_literal] = ACTIONS(1370),
    [anon_sym_L_SQUOTE] = ACTIONS(1370),
    [anon_sym_u_SQUOTE] = ACTIONS(1370),
    [anon_sym_U_SQUOTE] = ACTIONS(1370),
    [anon_sym_u8_SQUOTE] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_L_DQUOTE] = ACTIONS(1370),
    [anon_sym_u_DQUOTE] = ACTIONS(1370),
    [anon_sym_U_DQUOTE] = ACTIONS(1370),
    [anon_sym_u8_DQUOTE] = ACTIONS(1370),
    [anon_sym_DQUOTE] = ACTIONS(1370),
    [sym_true] = ACTIONS(1368),
    [sym_false] = ACTIONS(1368),
    [anon_sym_NULL] = ACTIONS(1368),
    [anon_sym_nullptr] = ACTIONS(1368),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(257)] = {
    [sym_identifier] = ACTIONS(1360),
    [aux_sym_preproc_include_token1] = ACTIONS(1360),
    [aux_sym_preproc_def_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1360),
    [sym_preproc_directive] = ACTIONS(1360),
    [anon_sym_LPAREN2] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_PLUS] = ACTIONS(1360),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym___extension__] = ACTIONS(1360),
    [anon_sym_typedef] = ACTIONS(1360),
    [anon_sym_extern] = ACTIONS(1360),
    [anon_sym___attribute__] = ACTIONS(1360),
    [anon_sym___attribute] = ACTIONS(1360),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1362),
    [anon_sym___declspec] = ACTIONS(1360),
    [anon_sym___cdecl] = ACTIONS(1360),
    [anon_sym___clrcall] = ACTIONS(1360),
    [anon_sym___stdcall] = ACTIONS(1360),
    [anon_sym___fastcall] = ACTIONS(1360),
    [anon_sym___thiscall] = ACTIONS(1360),
    [anon_sym___vectorcall] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [anon_sym_signed] = ACTIONS(1360),
    [anon_sym_unsigned] = ACTIONS(1360),
    [anon_sym_long] = ACTIONS(1360),
    [anon_sym_short] = ACTIONS(1360),
    [anon_sym_static] = ACTIONS(1360),
    [anon_sym_auto] = ACTIONS(1360),
    [anon_sym_register] = ACTIONS(1360),
    [anon_sym_inline] = ACTIONS(1360),
    [anon_sym___inline] = ACTIONS(1360),
    [anon_sym___inline__] = ACTIONS(1360),
    [anon_sym___forceinline] = ACTIONS(1360),
    [anon_sym_thread_local] = ACTIONS(1360),
    [anon_sym___thread] = ACTIONS(1360),
    [anon_sym_const] = ACTIONS(1360),
    [anon_sym_constexpr] = ACTIONS(1360),
    [anon_sym_volatile] = ACTIONS(1360),
    [anon_sym_restrict] = ACTIONS(1360),
    [anon_sym___restrict__] = ACTIONS(1360),
    [anon_sym__Atomic] = ACTIONS(1360),
    [anon_sym__Noreturn] = ACTIONS(1360),
    [anon_sym_noreturn] = ACTIONS(1360),
    [anon_sym__Nonnull] = ACTIONS(1360),
    [anon_sym_alignas] = ACTIONS(1360),
    [anon_sym__Alignas] = ACTIONS(1360),
    [sym_primitive_type] = ACTIONS(1360),
    [anon_sym_enum] = ACTIONS(1360),
    [anon_sym_struct] = ACTIONS(1360),
    [anon_sym_union] = ACTIONS(1360),
    [anon_sym_if] = ACTIONS(1360),
    [anon_sym_switch] = ACTIONS(1360),
    [anon_sym_case] = ACTIONS(1360),
    [anon_sym_default] = ACTIONS(1360),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_do] = ACTIONS(1360),
    [anon_sym_for] = ACTIONS(1360),
    [anon_sym_return] = ACTIONS(1360),
    [anon_sym_break] = ACTIONS(1360),
    [anon_sym_continue] = ACTIONS(1360),
    [anon_sym_goto] = ACTIONS(1360),
    [anon_sym___try] = ACTIONS(1360),
    [anon_sym___leave] = ACTIONS(1360),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_sizeof] = ACTIONS(1360),
    [anon_sym___alignof__] = ACTIONS(1360),
    [anon_sym___alignof] = ACTIONS(1360),
    [anon_sym__alignof] = ACTIONS(1360),
    [anon_sym_alignof] = ACTIONS(1360),
    [anon_sym__Alignof] = ACTIONS(1360),
    [anon_sym_offsetof] = ACTIONS(1360),
    [anon_sym__Generic] = ACTIONS(1360),
    [anon_sym_asm] = ACTIONS(1360),
    [anon_sym___asm__] = ACTIONS(1360),
    [anon_sym___asm] = ACTIONS(1360),
    [sym_number_literal] = ACTIONS(1362),
    [anon_sym_L_SQUOTE] = ACTIONS(1362),
    [anon_sym_u_SQUOTE] = ACTIONS(1362),
    [anon_sym_U_SQUOTE] = ACTIONS(1362),
    [anon_sym_u8_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_L_DQUOTE] = ACTIONS(1362),
    [anon_sym_u_DQUOTE] = ACTIONS(1362),
    [anon_sym_U_DQUOTE] = ACTIONS(1362),
    [anon_sym_u8_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [sym_true] = ACTIONS(1360),
    [sym_false] = ACTIONS(1360),
    [anon_sym_NULL] = ACTIONS(1360),
    [anon_sym_nullptr] = ACTIONS(1360),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(258)] = {
    [sym_identifier] = ACTIONS(1266),
    [aux_sym_preproc_include_token1] = ACTIONS(1266),
    [aux_sym_preproc_def_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1266),
    [sym_preproc_directive] = ACTIONS(1266),
    [anon_sym_LPAREN2] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1266),
    [anon_sym_PLUS] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1268),
    [anon_sym_AMP] = ACTIONS(1268),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym___extension__] = ACTIONS(1266),
    [anon_sym_typedef] = ACTIONS(1266),
    [anon_sym_extern] = ACTIONS(1266),
    [anon_sym___attribute__] = ACTIONS(1266),
    [anon_sym___attribute] = ACTIONS(1266),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1268),
    [anon_sym___declspec] = ACTIONS(1266),
    [anon_sym___cdecl] = ACTIONS(1266),
    [anon_sym___clrcall] = ACTIONS(1266),
    [anon_sym___stdcall] = ACTIONS(1266),
    [anon_sym___fastcall] = ACTIONS(1266),
    [anon_sym___thiscall] = ACTIONS(1266),
    [anon_sym___vectorcall] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [anon_sym_signed] = ACTIONS(1266),
    [anon_sym_unsigned] = ACTIONS(1266),
    [anon_sym_long] = ACTIONS(1266),
    [anon_sym_short] = ACTIONS(1266),
    [anon_sym_static] = ACTIONS(1266),
    [anon_sym_auto] = ACTIONS(1266),
    [anon_sym_register] = ACTIONS(1266),
    [anon_sym_inline] = ACTIONS(1266),
    [anon_sym___inline] = ACTIONS(1266),
    [anon_sym___inline__] = ACTIONS(1266),
    [anon_sym___forceinline] = ACTIONS(1266),
    [anon_sym_thread_local] = ACTIONS(1266),
    [anon_sym___thread] = ACTIONS(1266),
    [anon_sym_const] = ACTIONS(1266),
    [anon_sym_constexpr] = ACTIONS(1266),
    [anon_sym_volatile] = ACTIONS(1266),
    [anon_sym_restrict] = ACTIONS(1266),
    [anon_sym___restrict__] = ACTIONS(1266),
    [anon_sym__Atomic] = ACTIONS(1266),
    [anon_sym__Noreturn] = ACTIONS(1266),
    [anon_sym_noreturn] = ACTIONS(1266),
    [anon_sym__Nonnull] = ACTIONS(1266),
    [anon_sym_alignas] = ACTIONS(1266),
    [anon_sym__Alignas] = ACTIONS(1266),
    [sym_primitive_type] = ACTIONS(1266),
    [anon_sym_enum] = ACTIONS(1266),
    [anon_sym_struct] = ACTIONS(1266),
    [anon_sym_union] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_switch] = ACTIONS(1266),
    [anon_sym_case] = ACTIONS(1266),
    [anon_sym_default] = ACTIONS(1266),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_break] = ACTIONS(1266),
    [anon_sym_continue] = ACTIONS(1266),
    [anon_sym_goto] = ACTIONS(1266),
    [anon_sym___try] = ACTIONS(1266),
    [anon_sym___leave] = ACTIONS(1266),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_sizeof] = ACTIONS(1266),
    [anon_sym___alignof__] = ACTIONS(1266),
    [anon_sym___alignof] = ACTIONS(1266),
    [anon_sym__alignof] = ACTIONS(1266),
    [anon_sym_alignof] = ACTIONS(1266),
    [anon_sym__Alignof] = ACTIONS(1266),
    [anon_sym_offsetof] = ACTIONS(1266),
    [anon_sym__Generic] = ACTIONS(1266),
    [anon_sym_asm] = ACTIONS(1266),
    [anon_sym___asm__] = ACTIONS(1266),
    [anon_sym___asm] = ACTIONS(1266),
    [sym_number_literal] = ACTIONS(1268),
    [anon_sym_L_SQUOTE] = ACTIONS(1268),
    [anon_sym_u_SQUOTE] = ACTIONS(1268),
    [anon_sym_U_SQUOTE] = ACTIONS(1268),
    [anon_sym_u8_SQUOTE] = ACTIONS(1268),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_L_DQUOTE] = ACTIONS(1268),
    [anon_sym_u_DQUOTE] = ACTIONS(1268),
    [anon_sym_U_DQUOTE] = ACTIONS(1268),
    [anon_sym_u8_DQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1268),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [anon_sym_NULL] = ACTIONS(1266),
    [anon_sym_nullptr] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(259)] = {
    [sym_identifier] = ACTIONS(1270),
    [aux_sym_preproc_include_token1] = ACTIONS(1270),
    [aux_sym_preproc_def_token1] = ACTIONS(1270),
    [aux_sym_preproc_if_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1270),
    [sym_preproc_directive] = ACTIONS(1270),
    [anon_sym_LPAREN2] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_TILDE] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_AMP] = ACTIONS(1272),
    [anon_sym_SEMI] = ACTIONS(1272),
    [anon_sym___extension__] = ACTIONS(1270),
    [anon_sym_typedef] = ACTIONS(1270),
    [anon_sym_extern] = ACTIONS(1270),
    [anon_sym___attribute__] = ACTIONS(1270),
    [anon_sym___attribute] = ACTIONS(1270),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1272),
    [anon_sym___declspec] = ACTIONS(1270),
    [anon_sym___cdecl] = ACTIONS(1270),
    [anon_sym___clrcall] = ACTIONS(1270),
    [anon_sym___stdcall] = ACTIONS(1270),
    [anon_sym___fastcall] = ACTIONS(1270),
    [anon_sym___thiscall] = ACTIONS(1270),
    [anon_sym___vectorcall] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_RBRACE] = ACTIONS(1272),
    [anon_sym_signed] = ACTIONS(1270),
    [anon_sym_unsigned] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_auto] = ACTIONS(1270),
    [anon_sym_register] = ACTIONS(1270),
    [anon_sym_inline] = ACTIONS(1270),
    [anon_sym___inline] = ACTIONS(1270),
    [anon_sym___inline__] = ACTIONS(1270),
    [anon_sym___forceinline] = ACTIONS(1270),
    [anon_sym_thread_local] = ACTIONS(1270),
    [anon_sym___thread] = ACTIONS(1270),
    [anon_sym_const] = ACTIONS(1270),
    [anon_sym_constexpr] = ACTIONS(1270),
    [anon_sym_volatile] = ACTIONS(1270),
    [anon_sym_restrict] = ACTIONS(1270),
    [anon_sym___restrict__] = ACTIONS(1270),
    [anon_sym__Atomic] = ACTIONS(1270),
    [anon_sym__Noreturn] = ACTIONS(1270),
    [anon_sym_noreturn] = ACTIONS(1270),
    [anon_sym__Nonnull] = ACTIONS(1270),
    [anon_sym_alignas] = ACTIONS(1270),
    [anon_sym__Alignas] = ACTIONS(1270),
    [sym_primitive_type] = ACTIONS(1270),
    [anon_sym_enum] = ACTIONS(1270),
    [anon_sym_struct] = ACTIONS(1270),
    [anon_sym_union] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_goto] = ACTIONS(1270),
    [anon_sym___try] = ACTIONS(1270),
    [anon_sym___leave] = ACTIONS(1270),
    [anon_sym_DASH_DASH] = ACTIONS(1272),
    [anon_sym_PLUS_PLUS] = ACTIONS(1272),
    [anon_sym_sizeof] = ACTIONS(1270),
    [anon_sym___alignof__] = ACTIONS(1270),
    [anon_sym___alignof] = ACTIONS(1270),
    [anon_sym__alignof] = ACTIONS(1270),
    [anon_sym_alignof] = ACTIONS(1270),
    [anon_sym__Alignof] = ACTIONS(1270),
    [anon_sym_offsetof] = ACTIONS(1270),
    [anon_sym__Generic] = ACTIONS(1270),
    [anon_sym_asm] = ACTIONS(1270),
    [anon_sym___asm__] = ACTIONS(1270),
    [anon_sym___asm] = ACTIONS(1270),
    [sym_number_literal] = ACTIONS(1272),
    [anon_sym_L_SQUOTE] = ACTIONS(1272),
    [anon_sym_u_SQUOTE] = ACTIONS(1272),
    [anon_sym_U_SQUOTE] = ACTIONS(1272),
    [anon_sym_u8_SQUOTE] = ACTIONS(1272),
    [anon_sym_SQUOTE] = ACTIONS(1272),
    [anon_sym_L_DQUOTE] = ACTIONS(1272),
    [anon_sym_u_DQUOTE] = ACTIONS(1272),
    [anon_sym_U_DQUOTE] = ACTIONS(1272),
    [anon_sym_u8_DQUOTE] = ACTIONS(1272),
    [anon_sym_DQUOTE] = ACTIONS(1272),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [anon_sym_NULL] = ACTIONS(1270),
    [anon_sym_nullptr] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(260)] = {
    [sym_identifier] = ACTIONS(1274),
    [aux_sym_preproc_include_token1] = ACTIONS(1274),
    [aux_sym_preproc_def_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1274),
    [sym_preproc_directive] = ACTIONS(1274),
    [anon_sym_LPAREN2] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_TILDE] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1274),
    [anon_sym_PLUS] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1276),
    [anon_sym_AMP] = ACTIONS(1276),
    [anon_sym_SEMI] = ACTIONS(1276),
    [anon_sym___extension__] = ACTIONS(1274),
    [anon_sym_typedef] = ACTIONS(1274),
    [anon_sym_extern] = ACTIONS(1274),
    [anon_sym___attribute__] = ACTIONS(1274),
    [anon_sym___attribute] = ACTIONS(1274),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1276),
    [anon_sym___declspec] = ACTIONS(1274),
    [anon_sym___cdecl] = ACTIONS(1274),
    [anon_sym___clrcall] = ACTIONS(1274),
    [anon_sym___stdcall] = ACTIONS(1274),
    [anon_sym___fastcall] = ACTIONS(1274),
    [anon_sym___thiscall] = ACTIONS(1274),
    [anon_sym___vectorcall] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_RBRACE] = ACTIONS(1276),
    [anon_sym_signed] = ACTIONS(1274),
    [anon_sym_unsigned] = ACTIONS(1274),
    [anon_sym_long] = ACTIONS(1274),
    [anon_sym_short] = ACTIONS(1274),
    [anon_sym_static] = ACTIONS(1274),
    [anon_sym_auto] = ACTIONS(1274),
    [anon_sym_register] = ACTIONS(1274),
    [anon_sym_inline] = ACTIONS(1274),
    [anon_sym___inline] = ACTIONS(1274),
    [anon_sym___inline__] = ACTIONS(1274),
    [anon_sym___forceinline] = ACTIONS(1274),
    [anon_sym_thread_local] = ACTIONS(1274),
    [anon_sym___thread] = ACTIONS(1274),
    [anon_sym_const] = ACTIONS(1274),
    [anon_sym_constexpr] = ACTIONS(1274),
    [anon_sym_volatile] = ACTIONS(1274),
    [anon_sym_restrict] = ACTIONS(1274),
    [anon_sym___restrict__] = ACTIONS(1274),
    [anon_sym__Atomic] = ACTIONS(1274),
    [anon_sym__Noreturn] = ACTIONS(1274),
    [anon_sym_noreturn] = ACTIONS(1274),
    [anon_sym__Nonnull] = ACTIONS(1274),
    [anon_sym_alignas] = ACTIONS(1274),
    [anon_sym__Alignas] = ACTIONS(1274),
    [sym_primitive_type] = ACTIONS(1274),
    [anon_sym_enum] = ACTIONS(1274),
    [anon_sym_struct] = ACTIONS(1274),
    [anon_sym_union] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1274),
    [anon_sym_switch] = ACTIONS(1274),
    [anon_sym_case] = ACTIONS(1274),
    [anon_sym_default] = ACTIONS(1274),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_break] = ACTIONS(1274),
    [anon_sym_continue] = ACTIONS(1274),
    [anon_sym_goto] = ACTIONS(1274),
    [anon_sym___try] = ACTIONS(1274),
    [anon_sym___leave] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_PLUS_PLUS] = ACTIONS(1276),
    [anon_sym_sizeof] = ACTIONS(1274),
    [anon_sym___alignof__] = ACTIONS(1274),
    [anon_sym___alignof] = ACTIONS(1274),
    [anon_sym__alignof] = ACTIONS(1274),
    [anon_sym_alignof] = ACTIONS(1274),
    [anon_sym__Alignof] = ACTIONS(1274),
    [anon_sym_offsetof] = ACTIONS(1274),
    [anon_sym__Generic] = ACTIONS(1274),
    [anon_sym_asm] = ACTIONS(1274),
    [anon_sym___asm__] = ACTIONS(1274),
    [anon_sym___asm] = ACTIONS(1274),
    [sym_number_literal] = ACTIONS(1276),
    [anon_sym_L_SQUOTE] = ACTIONS(1276),
    [anon_sym_u_SQUOTE] = ACTIONS(1276),
    [anon_sym_U_SQUOTE] = ACTIONS(1276),
    [anon_sym_u8_SQUOTE] = ACTIONS(1276),
    [anon_sym_SQUOTE] = ACTIONS(1276),
    [anon_sym_L_DQUOTE] = ACTIONS(1276),
    [anon_sym_u_DQUOTE] = ACTIONS(1276),
    [anon_sym_U_DQUOTE] = ACTIONS(1276),
    [anon_sym_u8_DQUOTE] = ACTIONS(1276),
    [anon_sym_DQUOTE] = ACTIONS(1276),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [anon_sym_NULL] = ACTIONS(1274),
    [anon_sym_nullptr] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(261)] = {
    [sym_identifier] = ACTIONS(1282),
    [aux_sym_preproc_include_token1] = ACTIONS(1282),
    [aux_sym_preproc_def_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1282),
    [sym_preproc_directive] = ACTIONS(1282),
    [anon_sym_LPAREN2] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_TILDE] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1282),
    [anon_sym_PLUS] = ACTIONS(1282),
    [anon_sym_STAR] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym___extension__] = ACTIONS(1282),
    [anon_sym_typedef] = ACTIONS(1282),
    [anon_sym_extern] = ACTIONS(1282),
    [anon_sym___attribute__] = ACTIONS(1282),
    [anon_sym___attribute] = ACTIONS(1282),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1284),
    [anon_sym___declspec] = ACTIONS(1282),
    [anon_sym___cdecl] = ACTIONS(1282),
    [anon_sym___clrcall] = ACTIONS(1282),
    [anon_sym___stdcall] = ACTIONS(1282),
    [anon_sym___fastcall] = ACTIONS(1282),
    [anon_sym___thiscall] = ACTIONS(1282),
    [anon_sym___vectorcall] = ACTIONS(1282),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_RBRACE] = ACTIONS(1284),
    [anon_sym_signed] = ACTIONS(1282),
    [anon_sym_unsigned] = ACTIONS(1282),
    [anon_sym_long] = ACTIONS(1282),
    [anon_sym_short] = ACTIONS(1282),
    [anon_sym_static] = ACTIONS(1282),
    [anon_sym_auto] = ACTIONS(1282),
    [anon_sym_register] = ACTIONS(1282),
    [anon_sym_inline] = ACTIONS(1282),
    [anon_sym___inline] = ACTIONS(1282),
    [anon_sym___inline__] = ACTIONS(1282),
    [anon_sym___forceinline] = ACTIONS(1282),
    [anon_sym_thread_local] = ACTIONS(1282),
    [anon_sym___thread] = ACTIONS(1282),
    [anon_sym_const] = ACTIONS(1282),
    [anon_sym_constexpr] = ACTIONS(1282),
    [anon_sym_volatile] = ACTIONS(1282),
    [anon_sym_restrict] = ACTIONS(1282),
    [anon_sym___restrict__] = ACTIONS(1282),
    [anon_sym__Atomic] = ACTIONS(1282),
    [anon_sym__Noreturn] = ACTIONS(1282),
    [anon_sym_noreturn] = ACTIONS(1282),
    [anon_sym__Nonnull] = ACTIONS(1282),
    [anon_sym_alignas] = ACTIONS(1282),
    [anon_sym__Alignas] = ACTIONS(1282),
    [sym_primitive_type] = ACTIONS(1282),
    [anon_sym_enum] = ACTIONS(1282),
    [anon_sym_struct] = ACTIONS(1282),
    [anon_sym_union] = ACTIONS(1282),
    [anon_sym_if] = ACTIONS(1282),
    [anon_sym_switch] = ACTIONS(1282),
    [anon_sym_case] = ACTIONS(1282),
    [anon_sym_default] = ACTIONS(1282),
    [anon_sym_while] = ACTIONS(1282),
    [anon_sym_do] = ACTIONS(1282),
    [anon_sym_for] = ACTIONS(1282),
    [anon_sym_return] = ACTIONS(1282),
    [anon_sym_break] = ACTIONS(1282),
    [anon_sym_continue] = ACTIONS(1282),
    [anon_sym_goto] = ACTIONS(1282),
    [anon_sym___try] = ACTIONS(1282),
    [anon_sym___leave] = ACTIONS(1282),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_sizeof] = ACTIONS(1282),
    [anon_sym___alignof__] = ACTIONS(1282),
    [anon_sym___alignof] = ACTIONS(1282),
    [anon_sym__alignof] = ACTIONS(1282),
    [anon_sym_alignof] = ACTIONS(1282),
    [anon_sym__Alignof] = ACTIONS(1282),
    [anon_sym_offsetof] = ACTIONS(1282),
    [anon_sym__Generic] = ACTIONS(1282),
    [anon_sym_asm] = ACTIONS(1282),
    [anon_sym___asm__] = ACTIONS(1282),
    [anon_sym___asm] = ACTIONS(1282),
    [sym_number_literal] = ACTIONS(1284),
    [anon_sym_L_SQUOTE] = ACTIONS(1284),
    [anon_sym_u_SQUOTE] = ACTIONS(1284),
    [anon_sym_U_SQUOTE] = ACTIONS(1284),
    [anon_sym_u8_SQUOTE] = ACTIONS(1284),
    [anon_sym_SQUOTE] = ACTIONS(1284),
    [anon_sym_L_DQUOTE] = ACTIONS(1284),
    [anon_sym_u_DQUOTE] = ACTIONS(1284),
    [anon_sym_U_DQUOTE] = ACTIONS(1284),
    [anon_sym_u8_DQUOTE] = ACTIONS(1284),
    [anon_sym_DQUOTE] = ACTIONS(1284),
    [sym_true] = ACTIONS(1282),
    [sym_false] = ACTIONS(1282),
    [anon_sym_NULL] = ACTIONS(1282),
    [anon_sym_nullptr] = ACTIONS(1282),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(262)] = {
    [sym_identifier] = ACTIONS(1286),
    [aux_sym_preproc_include_token1] = ACTIONS(1286),
    [aux_sym_preproc_def_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1286),
    [sym_preproc_directive] = ACTIONS(1286),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_TILDE] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1288),
    [anon_sym_AMP] = ACTIONS(1288),
    [anon_sym_SEMI] = ACTIONS(1288),
    [anon_sym___extension__] = ACTIONS(1286),
    [anon_sym_typedef] = ACTIONS(1286),
    [anon_sym_extern] = ACTIONS(1286),
    [anon_sym___attribute__] = ACTIONS(1286),
    [anon_sym___attribute] = ACTIONS(1286),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1288),
    [anon_sym___declspec] = ACTIONS(1286),
    [anon_sym___cdecl] = ACTIONS(1286),
    [anon_sym___clrcall] = ACTIONS(1286),
    [anon_sym___stdcall] = ACTIONS(1286),
    [anon_sym___fastcall] = ACTIONS(1286),
    [anon_sym___thiscall] = ACTIONS(1286),
    [anon_sym___vectorcall] = ACTIONS(1286),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_RBRACE] = ACTIONS(1288),
    [anon_sym_signed] = ACTIONS(1286),
    [anon_sym_unsigned] = ACTIONS(1286),
    [anon_sym_long] = ACTIONS(1286),
    [anon_sym_short] = ACTIONS(1286),
    [anon_sym_static] = ACTIONS(1286),
    [anon_sym_auto] = ACTIONS(1286),
    [anon_sym_register] = ACTIONS(1286),
    [anon_sym_inline] = ACTIONS(1286),
    [anon_sym___inline] = ACTIONS(1286),
    [anon_sym___inline__] = ACTIONS(1286),
    [anon_sym___forceinline] = ACTIONS(1286),
    [anon_sym_thread_local] = ACTIONS(1286),
    [anon_sym___thread] = ACTIONS(1286),
    [anon_sym_const] = ACTIONS(1286),
    [anon_sym_constexpr] = ACTIONS(1286),
    [anon_sym_volatile] = ACTIONS(1286),
    [anon_sym_restrict] = ACTIONS(1286),
    [anon_sym___restrict__] = ACTIONS(1286),
    [anon_sym__Atomic] = ACTIONS(1286),
    [anon_sym__Noreturn] = ACTIONS(1286),
    [anon_sym_noreturn] = ACTIONS(1286),
    [anon_sym__Nonnull] = ACTIONS(1286),
    [anon_sym_alignas] = ACTIONS(1286),
    [anon_sym__Alignas] = ACTIONS(1286),
    [sym_primitive_type] = ACTIONS(1286),
    [anon_sym_enum] = ACTIONS(1286),
    [anon_sym_struct] = ACTIONS(1286),
    [anon_sym_union] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(1286),
    [anon_sym_switch] = ACTIONS(1286),
    [anon_sym_case] = ACTIONS(1286),
    [anon_sym_default] = ACTIONS(1286),
    [anon_sym_while] = ACTIONS(1286),
    [anon_sym_do] = ACTIONS(1286),
    [anon_sym_for] = ACTIONS(1286),
    [anon_sym_return] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(1286),
    [anon_sym_continue] = ACTIONS(1286),
    [anon_sym_goto] = ACTIONS(1286),
    [anon_sym___try] = ACTIONS(1286),
    [anon_sym___leave] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1288),
    [anon_sym_PLUS_PLUS] = ACTIONS(1288),
    [anon_sym_sizeof] = ACTIONS(1286),
    [anon_sym___alignof__] = ACTIONS(1286),
    [anon_sym___alignof] = ACTIONS(1286),
    [anon_sym__alignof] = ACTIONS(1286),
    [anon_sym_alignof] = ACTIONS(1286),
    [anon_sym__Alignof] = ACTIONS(1286),
    [anon_sym_offsetof] = ACTIONS(1286),
    [anon_sym__Generic] = ACTIONS(1286),
    [anon_sym_asm] = ACTIONS(1286),
    [anon_sym___asm__] = ACTIONS(1286),
    [anon_sym___asm] = ACTIONS(1286),
    [sym_number_literal] = ACTIONS(1288),
    [anon_sym_L_SQUOTE] = ACTIONS(1288),
    [anon_sym_u_SQUOTE] = ACTIONS(1288),
    [anon_sym_U_SQUOTE] = ACTIONS(1288),
    [anon_sym_u8_SQUOTE] = ACTIONS(1288),
    [anon_sym_SQUOTE] = ACTIONS(1288),
    [anon_sym_L_DQUOTE] = ACTIONS(1288),
    [anon_sym_u_DQUOTE] = ACTIONS(1288),
    [anon_sym_U_DQUOTE] = ACTIONS(1288),
    [anon_sym_u8_DQUOTE] = ACTIONS(1288),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [sym_true] = ACTIONS(1286),
    [sym_false] = ACTIONS(1286),
    [anon_sym_NULL] = ACTIONS(1286),
    [anon_sym_nullptr] = ACTIONS(1286),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(263)] = {
    [sym_identifier] = ACTIONS(1290),
    [aux_sym_preproc_include_token1] = ACTIONS(1290),
    [aux_sym_preproc_def_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1290),
    [sym_preproc_directive] = ACTIONS(1290),
    [anon_sym_LPAREN2] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym___extension__] = ACTIONS(1290),
    [anon_sym_typedef] = ACTIONS(1290),
    [anon_sym_extern] = ACTIONS(1290),
    [anon_sym___attribute__] = ACTIONS(1290),
    [anon_sym___attribute] = ACTIONS(1290),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1292),
    [anon_sym___declspec] = ACTIONS(1290),
    [anon_sym___cdecl] = ACTIONS(1290),
    [anon_sym___clrcall] = ACTIONS(1290),
    [anon_sym___stdcall] = ACTIONS(1290),
    [anon_sym___fastcall] = ACTIONS(1290),
    [anon_sym___thiscall] = ACTIONS(1290),
    [anon_sym___vectorcall] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_RBRACE] = ACTIONS(1292),
    [anon_sym_signed] = ACTIONS(1290),
    [anon_sym_unsigned] = ACTIONS(1290),
    [anon_sym_long] = ACTIONS(1290),
    [anon_sym_short] = ACTIONS(1290),
    [anon_sym_static] = ACTIONS(1290),
    [anon_sym_auto] = ACTIONS(1290),
    [anon_sym_register] = ACTIONS(1290),
    [anon_sym_inline] = ACTIONS(1290),
    [anon_sym___inline] = ACTIONS(1290),
    [anon_sym___inline__] = ACTIONS(1290),
    [anon_sym___forceinline] = ACTIONS(1290),
    [anon_sym_thread_local] = ACTIONS(1290),
    [anon_sym___thread] = ACTIONS(1290),
    [anon_sym_const] = ACTIONS(1290),
    [anon_sym_constexpr] = ACTIONS(1290),
    [anon_sym_volatile] = ACTIONS(1290),
    [anon_sym_restrict] = ACTIONS(1290),
    [anon_sym___restrict__] = ACTIONS(1290),
    [anon_sym__Atomic] = ACTIONS(1290),
    [anon_sym__Noreturn] = ACTIONS(1290),
    [anon_sym_noreturn] = ACTIONS(1290),
    [anon_sym__Nonnull] = ACTIONS(1290),
    [anon_sym_alignas] = ACTIONS(1290),
    [anon_sym__Alignas] = ACTIONS(1290),
    [sym_primitive_type] = ACTIONS(1290),
    [anon_sym_enum] = ACTIONS(1290),
    [anon_sym_struct] = ACTIONS(1290),
    [anon_sym_union] = ACTIONS(1290),
    [anon_sym_if] = ACTIONS(1290),
    [anon_sym_switch] = ACTIONS(1290),
    [anon_sym_case] = ACTIONS(1290),
    [anon_sym_default] = ACTIONS(1290),
    [anon_sym_while] = ACTIONS(1290),
    [anon_sym_do] = ACTIONS(1290),
    [anon_sym_for] = ACTIONS(1290),
    [anon_sym_return] = ACTIONS(1290),
    [anon_sym_break] = ACTIONS(1290),
    [anon_sym_continue] = ACTIONS(1290),
    [anon_sym_goto] = ACTIONS(1290),
    [anon_sym___try] = ACTIONS(1290),
    [anon_sym___leave] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_sizeof] = ACTIONS(1290),
    [anon_sym___alignof__] = ACTIONS(1290),
    [anon_sym___alignof] = ACTIONS(1290),
    [anon_sym__alignof] = ACTIONS(1290),
    [anon_sym_alignof] = ACTIONS(1290),
    [anon_sym__Alignof] = ACTIONS(1290),
    [anon_sym_offsetof] = ACTIONS(1290),
    [anon_sym__Generic] = ACTIONS(1290),
    [anon_sym_asm] = ACTIONS(1290),
    [anon_sym___asm__] = ACTIONS(1290),
    [anon_sym___asm] = ACTIONS(1290),
    [sym_number_literal] = ACTIONS(1292),
    [anon_sym_L_SQUOTE] = ACTIONS(1292),
    [anon_sym_u_SQUOTE] = ACTIONS(1292),
    [anon_sym_U_SQUOTE] = ACTIONS(1292),
    [anon_sym_u8_SQUOTE] = ACTIONS(1292),
    [anon_sym_SQUOTE] = ACTIONS(1292),
    [anon_sym_L_DQUOTE] = ACTIONS(1292),
    [anon_sym_u_DQUOTE] = ACTIONS(1292),
    [anon_sym_U_DQUOTE] = ACTIONS(1292),
    [anon_sym_u8_DQUOTE] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [sym_true] = ACTIONS(1290),
    [sym_false] = ACTIONS(1290),
    [anon_sym_NULL] = ACTIONS(1290),
    [anon_sym_nullptr] = ACTIONS(1290),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(264)] = {
    [sym_identifier] = ACTIONS(1278),
    [aux_sym_preproc_include_token1] = ACTIONS(1278),
    [aux_sym_preproc_def_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1278),
    [sym_preproc_directive] = ACTIONS(1278),
    [anon_sym_LPAREN2] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1278),
    [anon_sym_PLUS] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1280),
    [anon_sym___extension__] = ACTIONS(1278),
    [anon_sym_typedef] = ACTIONS(1278),
    [anon_sym_extern] = ACTIONS(1278),
    [anon_sym___attribute__] = ACTIONS(1278),
    [anon_sym___attribute] = ACTIONS(1278),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1280),
    [anon_sym___declspec] = ACTIONS(1278),
    [anon_sym___cdecl] = ACTIONS(1278),
    [anon_sym___clrcall] = ACTIONS(1278),
    [anon_sym___stdcall] = ACTIONS(1278),
    [anon_sym___fastcall] = ACTIONS(1278),
    [anon_sym___thiscall] = ACTIONS(1278),
    [anon_sym___vectorcall] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_RBRACE] = ACTIONS(1280),
    [anon_sym_signed] = ACTIONS(1278),
    [anon_sym_unsigned] = ACTIONS(1278),
    [anon_sym_long] = ACTIONS(1278),
    [anon_sym_short] = ACTIONS(1278),
    [anon_sym_static] = ACTIONS(1278),
    [anon_sym_auto] = ACTIONS(1278),
    [anon_sym_register] = ACTIONS(1278),
    [anon_sym_inline] = ACTIONS(1278),
    [anon_sym___inline] = ACTIONS(1278),
    [anon_sym___inline__] = ACTIONS(1278),
    [anon_sym___forceinline] = ACTIONS(1278),
    [anon_sym_thread_local] = ACTIONS(1278),
    [anon_sym___thread] = ACTIONS(1278),
    [anon_sym_const] = ACTIONS(1278),
    [anon_sym_constexpr] = ACTIONS(1278),
    [anon_sym_volatile] = ACTIONS(1278),
    [anon_sym_restrict] = ACTIONS(1278),
    [anon_sym___restrict__] = ACTIONS(1278),
    [anon_sym__Atomic] = ACTIONS(1278),
    [anon_sym__Noreturn] = ACTIONS(1278),
    [anon_sym_noreturn] = ACTIONS(1278),
    [anon_sym__Nonnull] = ACTIONS(1278),
    [anon_sym_alignas] = ACTIONS(1278),
    [anon_sym__Alignas] = ACTIONS(1278),
    [sym_primitive_type] = ACTIONS(1278),
    [anon_sym_enum] = ACTIONS(1278),
    [anon_sym_struct] = ACTIONS(1278),
    [anon_sym_union] = ACTIONS(1278),
    [anon_sym_if] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1278),
    [anon_sym_case] = ACTIONS(1278),
    [anon_sym_default] = ACTIONS(1278),
    [anon_sym_while] = ACTIONS(1278),
    [anon_sym_do] = ACTIONS(1278),
    [anon_sym_for] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1278),
    [anon_sym_break] = ACTIONS(1278),
    [anon_sym_continue] = ACTIONS(1278),
    [anon_sym_goto] = ACTIONS(1278),
    [anon_sym___try] = ACTIONS(1278),
    [anon_sym___leave] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1280),
    [anon_sym_sizeof] = ACTIONS(1278),
    [anon_sym___alignof__] = ACTIONS(1278),
    [anon_sym___alignof] = ACTIONS(1278),
    [anon_sym__alignof] = ACTIONS(1278),
    [anon_sym_alignof] = ACTIONS(1278),
    [anon_sym__Alignof] = ACTIONS(1278),
    [anon_sym_offsetof] = ACTIONS(1278),
    [anon_sym__Generic] = ACTIONS(1278),
    [anon_sym_asm] = ACTIONS(1278),
    [anon_sym___asm__] = ACTIONS(1278),
    [anon_sym___asm] = ACTIONS(1278),
    [sym_number_literal] = ACTIONS(1280),
    [anon_sym_L_SQUOTE] = ACTIONS(1280),
    [anon_sym_u_SQUOTE] = ACTIONS(1280),
    [anon_sym_U_SQUOTE] = ACTIONS(1280),
    [anon_sym_u8_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_L_DQUOTE] = ACTIONS(1280),
    [anon_sym_u_DQUOTE] = ACTIONS(1280),
    [anon_sym_U_DQUOTE] = ACTIONS(1280),
    [anon_sym_u8_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [sym_true] = ACTIONS(1278),
    [sym_false] = ACTIONS(1278),
    [anon_sym_NULL] = ACTIONS(1278),
    [anon_sym_nullptr] = ACTIONS(1278),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(265)] = {
    [sym_identifier] = ACTIONS(1298),
    [aux_sym_preproc_include_token1] = ACTIONS(1298),
    [aux_sym_preproc_def_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1298),
    [sym_preproc_directive] = ACTIONS(1298),
    [anon_sym_LPAREN2] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1298),
    [anon_sym_PLUS] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym___extension__] = ACTIONS(1298),
    [anon_sym_typedef] = ACTIONS(1298),
    [anon_sym_extern] = ACTIONS(1298),
    [anon_sym___attribute__] = ACTIONS(1298),
    [anon_sym___attribute] = ACTIONS(1298),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1300),
    [anon_sym___declspec] = ACTIONS(1298),
    [anon_sym___cdecl] = ACTIONS(1298),
    [anon_sym___clrcall] = ACTIONS(1298),
    [anon_sym___stdcall] = ACTIONS(1298),
    [anon_sym___fastcall] = ACTIONS(1298),
    [anon_sym___thiscall] = ACTIONS(1298),
    [anon_sym___vectorcall] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_signed] = ACTIONS(1298),
    [anon_sym_unsigned] = ACTIONS(1298),
    [anon_sym_long] = ACTIONS(1298),
    [anon_sym_short] = ACTIONS(1298),
    [anon_sym_static] = ACTIONS(1298),
    [anon_sym_auto] = ACTIONS(1298),
    [anon_sym_register] = ACTIONS(1298),
    [anon_sym_inline] = ACTIONS(1298),
    [anon_sym___inline] = ACTIONS(1298),
    [anon_sym___inline__] = ACTIONS(1298),
    [anon_sym___forceinline] = ACTIONS(1298),
    [anon_sym_thread_local] = ACTIONS(1298),
    [anon_sym___thread] = ACTIONS(1298),
    [anon_sym_const] = ACTIONS(1298),
    [anon_sym_constexpr] = ACTIONS(1298),
    [anon_sym_volatile] = ACTIONS(1298),
    [anon_sym_restrict] = ACTIONS(1298),
    [anon_sym___restrict__] = ACTIONS(1298),
    [anon_sym__Atomic] = ACTIONS(1298),
    [anon_sym__Noreturn] = ACTIONS(1298),
    [anon_sym_noreturn] = ACTIONS(1298),
    [anon_sym__Nonnull] = ACTIONS(1298),
    [anon_sym_alignas] = ACTIONS(1298),
    [anon_sym__Alignas] = ACTIONS(1298),
    [sym_primitive_type] = ACTIONS(1298),
    [anon_sym_enum] = ACTIONS(1298),
    [anon_sym_struct] = ACTIONS(1298),
    [anon_sym_union] = ACTIONS(1298),
    [anon_sym_if] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1298),
    [anon_sym_case] = ACTIONS(1298),
    [anon_sym_default] = ACTIONS(1298),
    [anon_sym_while] = ACTIONS(1298),
    [anon_sym_do] = ACTIONS(1298),
    [anon_sym_for] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1298),
    [anon_sym_break] = ACTIONS(1298),
    [anon_sym_continue] = ACTIONS(1298),
    [anon_sym_goto] = ACTIONS(1298),
    [anon_sym___try] = ACTIONS(1298),
    [anon_sym___leave] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_sizeof] = ACTIONS(1298),
    [anon_sym___alignof__] = ACTIONS(1298),
    [anon_sym___alignof] = ACTIONS(1298),
    [anon_sym__alignof] = ACTIONS(1298),
    [anon_sym_alignof] = ACTIONS(1298),
    [anon_sym__Alignof] = ACTIONS(1298),
    [anon_sym_offsetof] = ACTIONS(1298),
    [anon_sym__Generic] = ACTIONS(1298),
    [anon_sym_asm] = ACTIONS(1298),
    [anon_sym___asm__] = ACTIONS(1298),
    [anon_sym___asm] = ACTIONS(1298),
    [sym_number_literal] = ACTIONS(1300),
    [anon_sym_L_SQUOTE] = ACTIONS(1300),
    [anon_sym_u_SQUOTE] = ACTIONS(1300),
    [anon_sym_U_SQUOTE] = ACTIONS(1300),
    [anon_sym_u8_SQUOTE] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_L_DQUOTE] = ACTIONS(1300),
    [anon_sym_u_DQUOTE] = ACTIONS(1300),
    [anon_sym_U_DQUOTE] = ACTIONS(1300),
    [anon_sym_u8_DQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [sym_true] = ACTIONS(1298),
    [sym_false] = ACTIONS(1298),
    [anon_sym_NULL] = ACTIONS(1298),
    [anon_sym_nullptr] = ACTIONS(1298),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(266)] = {
    [sym_identifier] = ACTIONS(1294),
    [aux_sym_preproc_include_token1] = ACTIONS(1294),
    [aux_sym_preproc_def_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1294),
    [sym_preproc_directive] = ACTIONS(1294),
    [anon_sym_LPAREN2] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym___extension__] = ACTIONS(1294),
    [anon_sym_typedef] = ACTIONS(1294),
    [anon_sym_extern] = ACTIONS(1294),
    [anon_sym___attribute__] = ACTIONS(1294),
    [anon_sym___attribute] = ACTIONS(1294),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1296),
    [anon_sym___declspec] = ACTIONS(1294),
    [anon_sym___cdecl] = ACTIONS(1294),
    [anon_sym___clrcall] = ACTIONS(1294),
    [anon_sym___stdcall] = ACTIONS(1294),
    [anon_sym___fastcall] = ACTIONS(1294),
    [anon_sym___thiscall] = ACTIONS(1294),
    [anon_sym___vectorcall] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_RBRACE] = ACTIONS(1296),
    [anon_sym_signed] = ACTIONS(1294),
    [anon_sym_unsigned] = ACTIONS(1294),
    [anon_sym_long] = ACTIONS(1294),
    [anon_sym_short] = ACTIONS(1294),
    [anon_sym_static] = ACTIONS(1294),
    [anon_sym_auto] = ACTIONS(1294),
    [anon_sym_register] = ACTIONS(1294),
    [anon_sym_inline] = ACTIONS(1294),
    [anon_sym___inline] = ACTIONS(1294),
    [anon_sym___inline__] = ACTIONS(1294),
    [anon_sym___forceinline] = ACTIONS(1294),
    [anon_sym_thread_local] = ACTIONS(1294),
    [anon_sym___thread] = ACTIONS(1294),
    [anon_sym_const] = ACTIONS(1294),
    [anon_sym_constexpr] = ACTIONS(1294),
    [anon_sym_volatile] = ACTIONS(1294),
    [anon_sym_restrict] = ACTIONS(1294),
    [anon_sym___restrict__] = ACTIONS(1294),
    [anon_sym__Atomic] = ACTIONS(1294),
    [anon_sym__Noreturn] = ACTIONS(1294),
    [anon_sym_noreturn] = ACTIONS(1294),
    [anon_sym__Nonnull] = ACTIONS(1294),
    [anon_sym_alignas] = ACTIONS(1294),
    [anon_sym__Alignas] = ACTIONS(1294),
    [sym_primitive_type] = ACTIONS(1294),
    [anon_sym_enum] = ACTIONS(1294),
    [anon_sym_struct] = ACTIONS(1294),
    [anon_sym_union] = ACTIONS(1294),
    [anon_sym_if] = ACTIONS(1294),
    [anon_sym_switch] = ACTIONS(1294),
    [anon_sym_case] = ACTIONS(1294),
    [anon_sym_default] = ACTIONS(1294),
    [anon_sym_while] = ACTIONS(1294),
    [anon_sym_do] = ACTIONS(1294),
    [anon_sym_for] = ACTIONS(1294),
    [anon_sym_return] = ACTIONS(1294),
    [anon_sym_break] = ACTIONS(1294),
    [anon_sym_continue] = ACTIONS(1294),
    [anon_sym_goto] = ACTIONS(1294),
    [anon_sym___try] = ACTIONS(1294),
    [anon_sym___leave] = ACTIONS(1294),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_sizeof] = ACTIONS(1294),
    [anon_sym___alignof__] = ACTIONS(1294),
    [anon_sym___alignof] = ACTIONS(1294),
    [anon_sym__alignof] = ACTIONS(1294),
    [anon_sym_alignof] = ACTIONS(1294),
    [anon_sym__Alignof] = ACTIONS(1294),
    [anon_sym_offsetof] = ACTIONS(1294),
    [anon_sym__Generic] = ACTIONS(1294),
    [anon_sym_asm] = ACTIONS(1294),
    [anon_sym___asm__] = ACTIONS(1294),
    [anon_sym___asm] = ACTIONS(1294),
    [sym_number_literal] = ACTIONS(1296),
    [anon_sym_L_SQUOTE] = ACTIONS(1296),
    [anon_sym_u_SQUOTE] = ACTIONS(1296),
    [anon_sym_U_SQUOTE] = ACTIONS(1296),
    [anon_sym_u8_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_L_DQUOTE] = ACTIONS(1296),
    [anon_sym_u_DQUOTE] = ACTIONS(1296),
    [anon_sym_U_DQUOTE] = ACTIONS(1296),
    [anon_sym_u8_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [sym_true] = ACTIONS(1294),
    [sym_false] = ACTIONS(1294),
    [anon_sym_NULL] = ACTIONS(1294),
    [anon_sym_nullptr] = ACTIONS(1294),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(267)] = {
    [sym_identifier] = ACTIONS(1262),
    [aux_sym_preproc_include_token1] = ACTIONS(1262),
    [aux_sym_preproc_def_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1262),
    [sym_preproc_directive] = ACTIONS(1262),
    [anon_sym_LPAREN2] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_TILDE] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_STAR] = ACTIONS(1264),
    [anon_sym_AMP] = ACTIONS(1264),
    [anon_sym_SEMI] = ACTIONS(1264),
    [anon_sym___extension__] = ACTIONS(1262),
    [anon_sym_typedef] = ACTIONS(1262),
    [anon_sym_extern] = ACTIONS(1262),
    [anon_sym___attribute__] = ACTIONS(1262),
    [anon_sym___attribute] = ACTIONS(1262),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1264),
    [anon_sym___declspec] = ACTIONS(1262),
    [anon_sym___cdecl] = ACTIONS(1262),
    [anon_sym___clrcall] = ACTIONS(1262),
    [anon_sym___stdcall] = ACTIONS(1262),
    [anon_sym___fastcall] = ACTIONS(1262),
    [anon_sym___thiscall] = ACTIONS(1262),
    [anon_sym___vectorcall] = ACTIONS(1262),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_RBRACE] = ACTIONS(1264),
    [anon_sym_signed] = ACTIONS(1262),
    [anon_sym_unsigned] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_auto] = ACTIONS(1262),
    [anon_sym_register] = ACTIONS(1262),
    [anon_sym_inline] = ACTIONS(1262),
    [anon_sym___inline] = ACTIONS(1262),
    [anon_sym___inline__] = ACTIONS(1262),
    [anon_sym___forceinline] = ACTIONS(1262),
    [anon_sym_thread_local] = ACTIONS(1262),
    [anon_sym___thread] = ACTIONS(1262),
    [anon_sym_const] = ACTIONS(1262),
    [anon_sym_constexpr] = ACTIONS(1262),
    [anon_sym_volatile] = ACTIONS(1262),
    [anon_sym_restrict] = ACTIONS(1262),
    [anon_sym___restrict__] = ACTIONS(1262),
    [anon_sym__Atomic] = ACTIONS(1262),
    [anon_sym__Noreturn] = ACTIONS(1262),
    [anon_sym_noreturn] = ACTIONS(1262),
    [anon_sym__Nonnull] = ACTIONS(1262),
    [anon_sym_alignas] = ACTIONS(1262),
    [anon_sym__Alignas] = ACTIONS(1262),
    [sym_primitive_type] = ACTIONS(1262),
    [anon_sym_enum] = ACTIONS(1262),
    [anon_sym_struct] = ACTIONS(1262),
    [anon_sym_union] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_goto] = ACTIONS(1262),
    [anon_sym___try] = ACTIONS(1262),
    [anon_sym___leave] = ACTIONS(1262),
    [anon_sym_DASH_DASH] = ACTIONS(1264),
    [anon_sym_PLUS_PLUS] = ACTIONS(1264),
    [anon_sym_sizeof] = ACTIONS(1262),
    [anon_sym___alignof__] = ACTIONS(1262),
    [anon_sym___alignof] = ACTIONS(1262),
    [anon_sym__alignof] = ACTIONS(1262),
    [anon_sym_alignof] = ACTIONS(1262),
    [anon_sym__Alignof] = ACTIONS(1262),
    [anon_sym_offsetof] = ACTIONS(1262),
    [anon_sym__Generic] = ACTIONS(1262),
    [anon_sym_asm] = ACTIONS(1262),
    [anon_sym___asm__] = ACTIONS(1262),
    [anon_sym___asm] = ACTIONS(1262),
    [sym_number_literal] = ACTIONS(1264),
    [anon_sym_L_SQUOTE] = ACTIONS(1264),
    [anon_sym_u_SQUOTE] = ACTIONS(1264),
    [anon_sym_U_SQUOTE] = ACTIONS(1264),
    [anon_sym_u8_SQUOTE] = ACTIONS(1264),
    [anon_sym_SQUOTE] = ACTIONS(1264),
    [anon_sym_L_DQUOTE] = ACTIONS(1264),
    [anon_sym_u_DQUOTE] = ACTIONS(1264),
    [anon_sym_U_DQUOTE] = ACTIONS(1264),
    [anon_sym_u8_DQUOTE] = ACTIONS(1264),
    [anon_sym_DQUOTE] = ACTIONS(1264),
    [sym_true] = ACTIONS(1262),
    [sym_false] = ACTIONS(1262),
    [anon_sym_NULL] = ACTIONS(1262),
    [anon_sym_nullptr] = ACTIONS(1262),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(268)] = {
    [sym_identifier] = ACTIONS(1310),
    [aux_sym_preproc_include_token1] = ACTIONS(1310),
    [aux_sym_preproc_def_token1] = ACTIONS(1310),
    [aux_sym_preproc_if_token1] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1310),
    [sym_preproc_directive] = ACTIONS(1310),
    [anon_sym_LPAREN2] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym___extension__] = ACTIONS(1310),
    [anon_sym_typedef] = ACTIONS(1310),
    [anon_sym_extern] = ACTIONS(1310),
    [anon_sym___attribute__] = ACTIONS(1310),
    [anon_sym___attribute] = ACTIONS(1310),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1312),
    [anon_sym___declspec] = ACTIONS(1310),
    [anon_sym___cdecl] = ACTIONS(1310),
    [anon_sym___clrcall] = ACTIONS(1310),
    [anon_sym___stdcall] = ACTIONS(1310),
    [anon_sym___fastcall] = ACTIONS(1310),
    [anon_sym___thiscall] = ACTIONS(1310),
    [anon_sym___vectorcall] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_RBRACE] = ACTIONS(1312),
    [anon_sym_signed] = ACTIONS(1310),
    [anon_sym_unsigned] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_auto] = ACTIONS(1310),
    [anon_sym_register] = ACTIONS(1310),
    [anon_sym_inline] = ACTIONS(1310),
    [anon_sym___inline] = ACTIONS(1310),
    [anon_sym___inline__] = ACTIONS(1310),
    [anon_sym___forceinline] = ACTIONS(1310),
    [anon_sym_thread_local] = ACTIONS(1310),
    [anon_sym___thread] = ACTIONS(1310),
    [anon_sym_const] = ACTIONS(1310),
    [anon_sym_constexpr] = ACTIONS(1310),
    [anon_sym_volatile] = ACTIONS(1310),
    [anon_sym_restrict] = ACTIONS(1310),
    [anon_sym___restrict__] = ACTIONS(1310),
    [anon_sym__Atomic] = ACTIONS(1310),
    [anon_sym__Noreturn] = ACTIONS(1310),
    [anon_sym_noreturn] = ACTIONS(1310),
    [anon_sym__Nonnull] = ACTIONS(1310),
    [anon_sym_alignas] = ACTIONS(1310),
    [anon_sym__Alignas] = ACTIONS(1310),
    [sym_primitive_type] = ACTIONS(1310),
    [anon_sym_enum] = ACTIONS(1310),
    [anon_sym_struct] = ACTIONS(1310),
    [anon_sym_union] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_goto] = ACTIONS(1310),
    [anon_sym___try] = ACTIONS(1310),
    [anon_sym___leave] = ACTIONS(1310),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_sizeof] = ACTIONS(1310),
    [anon_sym___alignof__] = ACTIONS(1310),
    [anon_sym___alignof] = ACTIONS(1310),
    [anon_sym__alignof] = ACTIONS(1310),
    [anon_sym_alignof] = ACTIONS(1310),
    [anon_sym__Alignof] = ACTIONS(1310),
    [anon_sym_offsetof] = ACTIONS(1310),
    [anon_sym__Generic] = ACTIONS(1310),
    [anon_sym_asm] = ACTIONS(1310),
    [anon_sym___asm__] = ACTIONS(1310),
    [anon_sym___asm] = ACTIONS(1310),
    [sym_number_literal] = ACTIONS(1312),
    [anon_sym_L_SQUOTE] = ACTIONS(1312),
    [anon_sym_u_SQUOTE] = ACTIONS(1312),
    [anon_sym_U_SQUOTE] = ACTIONS(1312),
    [anon_sym_u8_SQUOTE] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_L_DQUOTE] = ACTIONS(1312),
    [anon_sym_u_DQUOTE] = ACTIONS(1312),
    [anon_sym_U_DQUOTE] = ACTIONS(1312),
    [anon_sym_u8_DQUOTE] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [sym_true] = ACTIONS(1310),
    [sym_false] = ACTIONS(1310),
    [anon_sym_NULL] = ACTIONS(1310),
    [anon_sym_nullptr] = ACTIONS(1310),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(269)] = {
    [sym_identifier] = ACTIONS(1334),
    [aux_sym_preproc_include_token1] = ACTIONS(1334),
    [aux_sym_preproc_def_token1] = ACTIONS(1334),
    [aux_sym_preproc_if_token1] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1334),
    [sym_preproc_directive] = ACTIONS(1334),
    [anon_sym_LPAREN2] = ACTIONS(1337),
    [anon_sym_BANG] = ACTIONS(1337),
    [anon_sym_TILDE] = ACTIONS(1337),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_STAR] = ACTIONS(1337),
    [anon_sym_AMP] = ACTIONS(1337),
    [anon_sym_SEMI] = ACTIONS(1337),
    [anon_sym___extension__] = ACTIONS(1334),
    [anon_sym_typedef] = ACTIONS(1334),
    [anon_sym_extern] = ACTIONS(1334),
    [anon_sym___attribute__] = ACTIONS(1334),
    [anon_sym___attribute] = ACTIONS(1334),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1337),
    [anon_sym___declspec] = ACTIONS(1334),
    [anon_sym___cdecl] = ACTIONS(1334),
    [anon_sym___clrcall] = ACTIONS(1334),
    [anon_sym___stdcall] = ACTIONS(1334),
    [anon_sym___fastcall] = ACTIONS(1334),
    [anon_sym___thiscall] = ACTIONS(1334),
    [anon_sym___vectorcall] = ACTIONS(1334),
    [anon_sym_LBRACE] = ACTIONS(1337),
    [anon_sym_RBRACE] = ACTIONS(1337),
    [anon_sym_signed] = ACTIONS(1334),
    [anon_sym_unsigned] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_auto] = ACTIONS(1334),
    [anon_sym_register] = ACTIONS(1334),
    [anon_sym_inline] = ACTIONS(1334),
    [anon_sym___inline] = ACTIONS(1334),
    [anon_sym___inline__] = ACTIONS(1334),
    [anon_sym___forceinline] = ACTIONS(1334),
    [anon_sym_thread_local] = ACTIONS(1334),
    [anon_sym___thread] = ACTIONS(1334),
    [anon_sym_const] = ACTIONS(1334),
    [anon_sym_constexpr] = ACTIONS(1334),
    [anon_sym_volatile] = ACTIONS(1334),
    [anon_sym_restrict] = ACTIONS(1334),
    [anon_sym___restrict__] = ACTIONS(1334),
    [anon_sym__Atomic] = ACTIONS(1334),
    [anon_sym__Noreturn] = ACTIONS(1334),
    [anon_sym_noreturn] = ACTIONS(1334),
    [anon_sym__Nonnull] = ACTIONS(1334),
    [anon_sym_alignas] = ACTIONS(1334),
    [anon_sym__Alignas] = ACTIONS(1334),
    [sym_primitive_type] = ACTIONS(1334),
    [anon_sym_enum] = ACTIONS(1334),
    [anon_sym_struct] = ACTIONS(1334),
    [anon_sym_union] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_goto] = ACTIONS(1334),
    [anon_sym___try] = ACTIONS(1334),
    [anon_sym___leave] = ACTIONS(1334),
    [anon_sym_DASH_DASH] = ACTIONS(1337),
    [anon_sym_PLUS_PLUS] = ACTIONS(1337),
    [anon_sym_sizeof] = ACTIONS(1334),
    [anon_sym___alignof__] = ACTIONS(1334),
    [anon_sym___alignof] = ACTIONS(1334),
    [anon_sym__alignof] = ACTIONS(1334),
    [anon_sym_alignof] = ACTIONS(1334),
    [anon_sym__Alignof] = ACTIONS(1334),
    [anon_sym_offsetof] = ACTIONS(1334),
    [anon_sym__Generic] = ACTIONS(1334),
    [anon_sym_asm] = ACTIONS(1334),
    [anon_sym___asm__] = ACTIONS(1334),
    [anon_sym___asm] = ACTIONS(1334),
    [sym_number_literal] = ACTIONS(1337),
    [anon_sym_L_SQUOTE] = ACTIONS(1337),
    [anon_sym_u_SQUOTE] = ACTIONS(1337),
    [anon_sym_U_SQUOTE] = ACTIONS(1337),
    [anon_sym_u8_SQUOTE] = ACTIONS(1337),
    [anon_sym_SQUOTE] = ACTIONS(1337),
    [anon_sym_L_DQUOTE] = ACTIONS(1337),
    [anon_sym_u_DQUOTE] = ACTIONS(1337),
    [anon_sym_U_DQUOTE] = ACTIONS(1337),
    [anon_sym_u8_DQUOTE] = ACTIONS(1337),
    [anon_sym_DQUOTE] = ACTIONS(1337),
    [sym_true] = ACTIONS(1334),
    [sym_false] = ACTIONS(1334),
    [anon_sym_NULL] = ACTIONS(1334),
    [anon_sym_nullptr] = ACTIONS(1334),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(270)] = {
    [sym_expression] = STATE(699),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(675),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(675),
    [sym_call_expression] = STATE(675),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(675),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(675),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1378),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1382),
    [anon_sym_TILDE] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1386),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym___extension__] = ACTIONS(1388),
    [anon_sym___attribute__] = ACTIONS(1386),
    [anon_sym___attribute] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1386),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_STAR_EQ] = ACTIONS(1380),
    [anon_sym_SLASH_EQ] = ACTIONS(1380),
    [anon_sym_PERCENT_EQ] = ACTIONS(1380),
    [anon_sym_PLUS_EQ] = ACTIONS(1380),
    [anon_sym_DASH_EQ] = ACTIONS(1380),
    [anon_sym_LT_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT_GT_EQ] = ACTIONS(1380),
    [anon_sym_AMP_EQ] = ACTIONS(1380),
    [anon_sym_CARET_EQ] = ACTIONS(1380),
    [anon_sym_PIPE_EQ] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(1392),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(271)] = {
    [sym_identifier] = ACTIONS(1302),
    [aux_sym_preproc_include_token1] = ACTIONS(1302),
    [aux_sym_preproc_def_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1302),
    [sym_preproc_directive] = ACTIONS(1302),
    [anon_sym_LPAREN2] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym___extension__] = ACTIONS(1302),
    [anon_sym_typedef] = ACTIONS(1302),
    [anon_sym_extern] = ACTIONS(1302),
    [anon_sym___attribute__] = ACTIONS(1302),
    [anon_sym___attribute] = ACTIONS(1302),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1304),
    [anon_sym___declspec] = ACTIONS(1302),
    [anon_sym___cdecl] = ACTIONS(1302),
    [anon_sym___clrcall] = ACTIONS(1302),
    [anon_sym___stdcall] = ACTIONS(1302),
    [anon_sym___fastcall] = ACTIONS(1302),
    [anon_sym___thiscall] = ACTIONS(1302),
    [anon_sym___vectorcall] = ACTIONS(1302),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [anon_sym_signed] = ACTIONS(1302),
    [anon_sym_unsigned] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_auto] = ACTIONS(1302),
    [anon_sym_register] = ACTIONS(1302),
    [anon_sym_inline] = ACTIONS(1302),
    [anon_sym___inline] = ACTIONS(1302),
    [anon_sym___inline__] = ACTIONS(1302),
    [anon_sym___forceinline] = ACTIONS(1302),
    [anon_sym_thread_local] = ACTIONS(1302),
    [anon_sym___thread] = ACTIONS(1302),
    [anon_sym_const] = ACTIONS(1302),
    [anon_sym_constexpr] = ACTIONS(1302),
    [anon_sym_volatile] = ACTIONS(1302),
    [anon_sym_restrict] = ACTIONS(1302),
    [anon_sym___restrict__] = ACTIONS(1302),
    [anon_sym__Atomic] = ACTIONS(1302),
    [anon_sym__Noreturn] = ACTIONS(1302),
    [anon_sym_noreturn] = ACTIONS(1302),
    [anon_sym__Nonnull] = ACTIONS(1302),
    [anon_sym_alignas] = ACTIONS(1302),
    [anon_sym__Alignas] = ACTIONS(1302),
    [sym_primitive_type] = ACTIONS(1302),
    [anon_sym_enum] = ACTIONS(1302),
    [anon_sym_struct] = ACTIONS(1302),
    [anon_sym_union] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_goto] = ACTIONS(1302),
    [anon_sym___try] = ACTIONS(1302),
    [anon_sym___leave] = ACTIONS(1302),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_sizeof] = ACTIONS(1302),
    [anon_sym___alignof__] = ACTIONS(1302),
    [anon_sym___alignof] = ACTIONS(1302),
    [anon_sym__alignof] = ACTIONS(1302),
    [anon_sym_alignof] = ACTIONS(1302),
    [anon_sym__Alignof] = ACTIONS(1302),
    [anon_sym_offsetof] = ACTIONS(1302),
    [anon_sym__Generic] = ACTIONS(1302),
    [anon_sym_asm] = ACTIONS(1302),
    [anon_sym___asm__] = ACTIONS(1302),
    [anon_sym___asm] = ACTIONS(1302),
    [sym_number_literal] = ACTIONS(1304),
    [anon_sym_L_SQUOTE] = ACTIONS(1304),
    [anon_sym_u_SQUOTE] = ACTIONS(1304),
    [anon_sym_U_SQUOTE] = ACTIONS(1304),
    [anon_sym_u8_SQUOTE] = ACTIONS(1304),
    [anon_sym_SQUOTE] = ACTIONS(1304),
    [anon_sym_L_DQUOTE] = ACTIONS(1304),
    [anon_sym_u_DQUOTE] = ACTIONS(1304),
    [anon_sym_U_DQUOTE] = ACTIONS(1304),
    [anon_sym_u8_DQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [sym_true] = ACTIONS(1302),
    [sym_false] = ACTIONS(1302),
    [anon_sym_NULL] = ACTIONS(1302),
    [anon_sym_nullptr] = ACTIONS(1302),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(272)] = {
    [sym_identifier] = ACTIONS(1306),
    [aux_sym_preproc_include_token1] = ACTIONS(1306),
    [aux_sym_preproc_def_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1306),
    [sym_preproc_directive] = ACTIONS(1306),
    [anon_sym_LPAREN2] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym___extension__] = ACTIONS(1306),
    [anon_sym_typedef] = ACTIONS(1306),
    [anon_sym_extern] = ACTIONS(1306),
    [anon_sym___attribute__] = ACTIONS(1306),
    [anon_sym___attribute] = ACTIONS(1306),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1308),
    [anon_sym___declspec] = ACTIONS(1306),
    [anon_sym___cdecl] = ACTIONS(1306),
    [anon_sym___clrcall] = ACTIONS(1306),
    [anon_sym___stdcall] = ACTIONS(1306),
    [anon_sym___fastcall] = ACTIONS(1306),
    [anon_sym___thiscall] = ACTIONS(1306),
    [anon_sym___vectorcall] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [anon_sym_signed] = ACTIONS(1306),
    [anon_sym_unsigned] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_auto] = ACTIONS(1306),
    [anon_sym_register] = ACTIONS(1306),
    [anon_sym_inline] = ACTIONS(1306),
    [anon_sym___inline] = ACTIONS(1306),
    [anon_sym___inline__] = ACTIONS(1306),
    [anon_sym___forceinline] = ACTIONS(1306),
    [anon_sym_thread_local] = ACTIONS(1306),
    [anon_sym___thread] = ACTIONS(1306),
    [anon_sym_const] = ACTIONS(1306),
    [anon_sym_constexpr] = ACTIONS(1306),
    [anon_sym_volatile] = ACTIONS(1306),
    [anon_sym_restrict] = ACTIONS(1306),
    [anon_sym___restrict__] = ACTIONS(1306),
    [anon_sym__Atomic] = ACTIONS(1306),
    [anon_sym__Noreturn] = ACTIONS(1306),
    [anon_sym_noreturn] = ACTIONS(1306),
    [anon_sym__Nonnull] = ACTIONS(1306),
    [anon_sym_alignas] = ACTIONS(1306),
    [anon_sym__Alignas] = ACTIONS(1306),
    [sym_primitive_type] = ACTIONS(1306),
    [anon_sym_enum] = ACTIONS(1306),
    [anon_sym_struct] = ACTIONS(1306),
    [anon_sym_union] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_goto] = ACTIONS(1306),
    [anon_sym___try] = ACTIONS(1306),
    [anon_sym___leave] = ACTIONS(1306),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_sizeof] = ACTIONS(1306),
    [anon_sym___alignof__] = ACTIONS(1306),
    [anon_sym___alignof] = ACTIONS(1306),
    [anon_sym__alignof] = ACTIONS(1306),
    [anon_sym_alignof] = ACTIONS(1306),
    [anon_sym__Alignof] = ACTIONS(1306),
    [anon_sym_offsetof] = ACTIONS(1306),
    [anon_sym__Generic] = ACTIONS(1306),
    [anon_sym_asm] = ACTIONS(1306),
    [anon_sym___asm__] = ACTIONS(1306),
    [anon_sym___asm] = ACTIONS(1306),
    [sym_number_literal] = ACTIONS(1308),
    [anon_sym_L_SQUOTE] = ACTIONS(1308),
    [anon_sym_u_SQUOTE] = ACTIONS(1308),
    [anon_sym_U_SQUOTE] = ACTIONS(1308),
    [anon_sym_u8_SQUOTE] = ACTIONS(1308),
    [anon_sym_SQUOTE] = ACTIONS(1308),
    [anon_sym_L_DQUOTE] = ACTIONS(1308),
    [anon_sym_u_DQUOTE] = ACTIONS(1308),
    [anon_sym_U_DQUOTE] = ACTIONS(1308),
    [anon_sym_u8_DQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_true] = ACTIONS(1306),
    [sym_false] = ACTIONS(1306),
    [anon_sym_NULL] = ACTIONS(1306),
    [anon_sym_nullptr] = ACTIONS(1306),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(273)] = {
    [sym_identifier] = ACTIONS(1314),
    [aux_sym_preproc_include_token1] = ACTIONS(1314),
    [aux_sym_preproc_def_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1314),
    [sym_preproc_directive] = ACTIONS(1314),
    [anon_sym_LPAREN2] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1314),
    [anon_sym_PLUS] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym___extension__] = ACTIONS(1314),
    [anon_sym_typedef] = ACTIONS(1314),
    [anon_sym_extern] = ACTIONS(1314),
    [anon_sym___attribute__] = ACTIONS(1314),
    [anon_sym___attribute] = ACTIONS(1314),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1316),
    [anon_sym___declspec] = ACTIONS(1314),
    [anon_sym___cdecl] = ACTIONS(1314),
    [anon_sym___clrcall] = ACTIONS(1314),
    [anon_sym___stdcall] = ACTIONS(1314),
    [anon_sym___fastcall] = ACTIONS(1314),
    [anon_sym___thiscall] = ACTIONS(1314),
    [anon_sym___vectorcall] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [anon_sym_signed] = ACTIONS(1314),
    [anon_sym_unsigned] = ACTIONS(1314),
    [anon_sym_long] = ACTIONS(1314),
    [anon_sym_short] = ACTIONS(1314),
    [anon_sym_static] = ACTIONS(1314),
    [anon_sym_auto] = ACTIONS(1314),
    [anon_sym_register] = ACTIONS(1314),
    [anon_sym_inline] = ACTIONS(1314),
    [anon_sym___inline] = ACTIONS(1314),
    [anon_sym___inline__] = ACTIONS(1314),
    [anon_sym___forceinline] = ACTIONS(1314),
    [anon_sym_thread_local] = ACTIONS(1314),
    [anon_sym___thread] = ACTIONS(1314),
    [anon_sym_const] = ACTIONS(1314),
    [anon_sym_constexpr] = ACTIONS(1314),
    [anon_sym_volatile] = ACTIONS(1314),
    [anon_sym_restrict] = ACTIONS(1314),
    [anon_sym___restrict__] = ACTIONS(1314),
    [anon_sym__Atomic] = ACTIONS(1314),
    [anon_sym__Noreturn] = ACTIONS(1314),
    [anon_sym_noreturn] = ACTIONS(1314),
    [anon_sym__Nonnull] = ACTIONS(1314),
    [anon_sym_alignas] = ACTIONS(1314),
    [anon_sym__Alignas] = ACTIONS(1314),
    [sym_primitive_type] = ACTIONS(1314),
    [anon_sym_enum] = ACTIONS(1314),
    [anon_sym_struct] = ACTIONS(1314),
    [anon_sym_union] = ACTIONS(1314),
    [anon_sym_if] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1314),
    [anon_sym_case] = ACTIONS(1314),
    [anon_sym_default] = ACTIONS(1314),
    [anon_sym_while] = ACTIONS(1314),
    [anon_sym_do] = ACTIONS(1314),
    [anon_sym_for] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1314),
    [anon_sym_break] = ACTIONS(1314),
    [anon_sym_continue] = ACTIONS(1314),
    [anon_sym_goto] = ACTIONS(1314),
    [anon_sym___try] = ACTIONS(1314),
    [anon_sym___leave] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_sizeof] = ACTIONS(1314),
    [anon_sym___alignof__] = ACTIONS(1314),
    [anon_sym___alignof] = ACTIONS(1314),
    [anon_sym__alignof] = ACTIONS(1314),
    [anon_sym_alignof] = ACTIONS(1314),
    [anon_sym__Alignof] = ACTIONS(1314),
    [anon_sym_offsetof] = ACTIONS(1314),
    [anon_sym__Generic] = ACTIONS(1314),
    [anon_sym_asm] = ACTIONS(1314),
    [anon_sym___asm__] = ACTIONS(1314),
    [anon_sym___asm] = ACTIONS(1314),
    [sym_number_literal] = ACTIONS(1316),
    [anon_sym_L_SQUOTE] = ACTIONS(1316),
    [anon_sym_u_SQUOTE] = ACTIONS(1316),
    [anon_sym_U_SQUOTE] = ACTIONS(1316),
    [anon_sym_u8_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_L_DQUOTE] = ACTIONS(1316),
    [anon_sym_u_DQUOTE] = ACTIONS(1316),
    [anon_sym_U_DQUOTE] = ACTIONS(1316),
    [anon_sym_u8_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [sym_true] = ACTIONS(1314),
    [sym_false] = ACTIONS(1314),
    [anon_sym_NULL] = ACTIONS(1314),
    [anon_sym_nullptr] = ACTIONS(1314),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(274)] = {
    [sym_identifier] = ACTIONS(1318),
    [aux_sym_preproc_include_token1] = ACTIONS(1318),
    [aux_sym_preproc_def_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1318),
    [sym_preproc_directive] = ACTIONS(1318),
    [anon_sym_LPAREN2] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym___extension__] = ACTIONS(1318),
    [anon_sym_typedef] = ACTIONS(1318),
    [anon_sym_extern] = ACTIONS(1318),
    [anon_sym___attribute__] = ACTIONS(1318),
    [anon_sym___attribute] = ACTIONS(1318),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1320),
    [anon_sym___declspec] = ACTIONS(1318),
    [anon_sym___cdecl] = ACTIONS(1318),
    [anon_sym___clrcall] = ACTIONS(1318),
    [anon_sym___stdcall] = ACTIONS(1318),
    [anon_sym___fastcall] = ACTIONS(1318),
    [anon_sym___thiscall] = ACTIONS(1318),
    [anon_sym___vectorcall] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_RBRACE] = ACTIONS(1320),
    [anon_sym_signed] = ACTIONS(1318),
    [anon_sym_unsigned] = ACTIONS(1318),
    [anon_sym_long] = ACTIONS(1318),
    [anon_sym_short] = ACTIONS(1318),
    [anon_sym_static] = ACTIONS(1318),
    [anon_sym_auto] = ACTIONS(1318),
    [anon_sym_register] = ACTIONS(1318),
    [anon_sym_inline] = ACTIONS(1318),
    [anon_sym___inline] = ACTIONS(1318),
    [anon_sym___inline__] = ACTIONS(1318),
    [anon_sym___forceinline] = ACTIONS(1318),
    [anon_sym_thread_local] = ACTIONS(1318),
    [anon_sym___thread] = ACTIONS(1318),
    [anon_sym_const] = ACTIONS(1318),
    [anon_sym_constexpr] = ACTIONS(1318),
    [anon_sym_volatile] = ACTIONS(1318),
    [anon_sym_restrict] = ACTIONS(1318),
    [anon_sym___restrict__] = ACTIONS(1318),
    [anon_sym__Atomic] = ACTIONS(1318),
    [anon_sym__Noreturn] = ACTIONS(1318),
    [anon_sym_noreturn] = ACTIONS(1318),
    [anon_sym__Nonnull] = ACTIONS(1318),
    [anon_sym_alignas] = ACTIONS(1318),
    [anon_sym__Alignas] = ACTIONS(1318),
    [sym_primitive_type] = ACTIONS(1318),
    [anon_sym_enum] = ACTIONS(1318),
    [anon_sym_struct] = ACTIONS(1318),
    [anon_sym_union] = ACTIONS(1318),
    [anon_sym_if] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1318),
    [anon_sym_case] = ACTIONS(1318),
    [anon_sym_default] = ACTIONS(1318),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_break] = ACTIONS(1318),
    [anon_sym_continue] = ACTIONS(1318),
    [anon_sym_goto] = ACTIONS(1318),
    [anon_sym___try] = ACTIONS(1318),
    [anon_sym___leave] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_sizeof] = ACTIONS(1318),
    [anon_sym___alignof__] = ACTIONS(1318),
    [anon_sym___alignof] = ACTIONS(1318),
    [anon_sym__alignof] = ACTIONS(1318),
    [anon_sym_alignof] = ACTIONS(1318),
    [anon_sym__Alignof] = ACTIONS(1318),
    [anon_sym_offsetof] = ACTIONS(1318),
    [anon_sym__Generic] = ACTIONS(1318),
    [anon_sym_asm] = ACTIONS(1318),
    [anon_sym___asm__] = ACTIONS(1318),
    [anon_sym___asm] = ACTIONS(1318),
    [sym_number_literal] = ACTIONS(1320),
    [anon_sym_L_SQUOTE] = ACTIONS(1320),
    [anon_sym_u_SQUOTE] = ACTIONS(1320),
    [anon_sym_U_SQUOTE] = ACTIONS(1320),
    [anon_sym_u8_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_L_DQUOTE] = ACTIONS(1320),
    [anon_sym_u_DQUOTE] = ACTIONS(1320),
    [anon_sym_U_DQUOTE] = ACTIONS(1320),
    [anon_sym_u8_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [anon_sym_NULL] = ACTIONS(1318),
    [anon_sym_nullptr] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(275)] = {
    [sym_identifier] = ACTIONS(1322),
    [aux_sym_preproc_include_token1] = ACTIONS(1322),
    [aux_sym_preproc_def_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1322),
    [sym_preproc_directive] = ACTIONS(1322),
    [anon_sym_LPAREN2] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1322),
    [anon_sym_PLUS] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym___extension__] = ACTIONS(1322),
    [anon_sym_typedef] = ACTIONS(1322),
    [anon_sym_extern] = ACTIONS(1322),
    [anon_sym___attribute__] = ACTIONS(1322),
    [anon_sym___attribute] = ACTIONS(1322),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1324),
    [anon_sym___declspec] = ACTIONS(1322),
    [anon_sym___cdecl] = ACTIONS(1322),
    [anon_sym___clrcall] = ACTIONS(1322),
    [anon_sym___stdcall] = ACTIONS(1322),
    [anon_sym___fastcall] = ACTIONS(1322),
    [anon_sym___thiscall] = ACTIONS(1322),
    [anon_sym___vectorcall] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_signed] = ACTIONS(1322),
    [anon_sym_unsigned] = ACTIONS(1322),
    [anon_sym_long] = ACTIONS(1322),
    [anon_sym_short] = ACTIONS(1322),
    [anon_sym_static] = ACTIONS(1322),
    [anon_sym_auto] = ACTIONS(1322),
    [anon_sym_register] = ACTIONS(1322),
    [anon_sym_inline] = ACTIONS(1322),
    [anon_sym___inline] = ACTIONS(1322),
    [anon_sym___inline__] = ACTIONS(1322),
    [anon_sym___forceinline] = ACTIONS(1322),
    [anon_sym_thread_local] = ACTIONS(1322),
    [anon_sym___thread] = ACTIONS(1322),
    [anon_sym_const] = ACTIONS(1322),
    [anon_sym_constexpr] = ACTIONS(1322),
    [anon_sym_volatile] = ACTIONS(1322),
    [anon_sym_restrict] = ACTIONS(1322),
    [anon_sym___restrict__] = ACTIONS(1322),
    [anon_sym__Atomic] = ACTIONS(1322),
    [anon_sym__Noreturn] = ACTIONS(1322),
    [anon_sym_noreturn] = ACTIONS(1322),
    [anon_sym__Nonnull] = ACTIONS(1322),
    [anon_sym_alignas] = ACTIONS(1322),
    [anon_sym__Alignas] = ACTIONS(1322),
    [sym_primitive_type] = ACTIONS(1322),
    [anon_sym_enum] = ACTIONS(1322),
    [anon_sym_struct] = ACTIONS(1322),
    [anon_sym_union] = ACTIONS(1322),
    [anon_sym_if] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1322),
    [anon_sym_case] = ACTIONS(1322),
    [anon_sym_default] = ACTIONS(1322),
    [anon_sym_while] = ACTIONS(1322),
    [anon_sym_do] = ACTIONS(1322),
    [anon_sym_for] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1322),
    [anon_sym_break] = ACTIONS(1322),
    [anon_sym_continue] = ACTIONS(1322),
    [anon_sym_goto] = ACTIONS(1322),
    [anon_sym___try] = ACTIONS(1322),
    [anon_sym___leave] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_sizeof] = ACTIONS(1322),
    [anon_sym___alignof__] = ACTIONS(1322),
    [anon_sym___alignof] = ACTIONS(1322),
    [anon_sym__alignof] = ACTIONS(1322),
    [anon_sym_alignof] = ACTIONS(1322),
    [anon_sym__Alignof] = ACTIONS(1322),
    [anon_sym_offsetof] = ACTIONS(1322),
    [anon_sym__Generic] = ACTIONS(1322),
    [anon_sym_asm] = ACTIONS(1322),
    [anon_sym___asm__] = ACTIONS(1322),
    [anon_sym___asm] = ACTIONS(1322),
    [sym_number_literal] = ACTIONS(1324),
    [anon_sym_L_SQUOTE] = ACTIONS(1324),
    [anon_sym_u_SQUOTE] = ACTIONS(1324),
    [anon_sym_U_SQUOTE] = ACTIONS(1324),
    [anon_sym_u8_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_L_DQUOTE] = ACTIONS(1324),
    [anon_sym_u_DQUOTE] = ACTIONS(1324),
    [anon_sym_U_DQUOTE] = ACTIONS(1324),
    [anon_sym_u8_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [sym_true] = ACTIONS(1322),
    [sym_false] = ACTIONS(1322),
    [anon_sym_NULL] = ACTIONS(1322),
    [anon_sym_nullptr] = ACTIONS(1322),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(276)] = {
    [sym_identifier] = ACTIONS(1326),
    [aux_sym_preproc_include_token1] = ACTIONS(1326),
    [aux_sym_preproc_def_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1326),
    [sym_preproc_directive] = ACTIONS(1326),
    [anon_sym_LPAREN2] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1326),
    [anon_sym_PLUS] = ACTIONS(1326),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym___extension__] = ACTIONS(1326),
    [anon_sym_typedef] = ACTIONS(1326),
    [anon_sym_extern] = ACTIONS(1326),
    [anon_sym___attribute__] = ACTIONS(1326),
    [anon_sym___attribute] = ACTIONS(1326),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1328),
    [anon_sym___declspec] = ACTIONS(1326),
    [anon_sym___cdecl] = ACTIONS(1326),
    [anon_sym___clrcall] = ACTIONS(1326),
    [anon_sym___stdcall] = ACTIONS(1326),
    [anon_sym___fastcall] = ACTIONS(1326),
    [anon_sym___thiscall] = ACTIONS(1326),
    [anon_sym___vectorcall] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_RBRACE] = ACTIONS(1328),
    [anon_sym_signed] = ACTIONS(1326),
    [anon_sym_unsigned] = ACTIONS(1326),
    [anon_sym_long] = ACTIONS(1326),
    [anon_sym_short] = ACTIONS(1326),
    [anon_sym_static] = ACTIONS(1326),
    [anon_sym_auto] = ACTIONS(1326),
    [anon_sym_register] = ACTIONS(1326),
    [anon_sym_inline] = ACTIONS(1326),
    [anon_sym___inline] = ACTIONS(1326),
    [anon_sym___inline__] = ACTIONS(1326),
    [anon_sym___forceinline] = ACTIONS(1326),
    [anon_sym_thread_local] = ACTIONS(1326),
    [anon_sym___thread] = ACTIONS(1326),
    [anon_sym_const] = ACTIONS(1326),
    [anon_sym_constexpr] = ACTIONS(1326),
    [anon_sym_volatile] = ACTIONS(1326),
    [anon_sym_restrict] = ACTIONS(1326),
    [anon_sym___restrict__] = ACTIONS(1326),
    [anon_sym__Atomic] = ACTIONS(1326),
    [anon_sym__Noreturn] = ACTIONS(1326),
    [anon_sym_noreturn] = ACTIONS(1326),
    [anon_sym__Nonnull] = ACTIONS(1326),
    [anon_sym_alignas] = ACTIONS(1326),
    [anon_sym__Alignas] = ACTIONS(1326),
    [sym_primitive_type] = ACTIONS(1326),
    [anon_sym_enum] = ACTIONS(1326),
    [anon_sym_struct] = ACTIONS(1326),
    [anon_sym_union] = ACTIONS(1326),
    [anon_sym_if] = ACTIONS(1326),
    [anon_sym_switch] = ACTIONS(1326),
    [anon_sym_case] = ACTIONS(1326),
    [anon_sym_default] = ACTIONS(1326),
    [anon_sym_while] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1326),
    [anon_sym_for] = ACTIONS(1326),
    [anon_sym_return] = ACTIONS(1326),
    [anon_sym_break] = ACTIONS(1326),
    [anon_sym_continue] = ACTIONS(1326),
    [anon_sym_goto] = ACTIONS(1326),
    [anon_sym___try] = ACTIONS(1326),
    [anon_sym___leave] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_sizeof] = ACTIONS(1326),
    [anon_sym___alignof__] = ACTIONS(1326),
    [anon_sym___alignof] = ACTIONS(1326),
    [anon_sym__alignof] = ACTIONS(1326),
    [anon_sym_alignof] = ACTIONS(1326),
    [anon_sym__Alignof] = ACTIONS(1326),
    [anon_sym_offsetof] = ACTIONS(1326),
    [anon_sym__Generic] = ACTIONS(1326),
    [anon_sym_asm] = ACTIONS(1326),
    [anon_sym___asm__] = ACTIONS(1326),
    [anon_sym___asm] = ACTIONS(1326),
    [sym_number_literal] = ACTIONS(1328),
    [anon_sym_L_SQUOTE] = ACTIONS(1328),
    [anon_sym_u_SQUOTE] = ACTIONS(1328),
    [anon_sym_U_SQUOTE] = ACTIONS(1328),
    [anon_sym_u8_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_L_DQUOTE] = ACTIONS(1328),
    [anon_sym_u_DQUOTE] = ACTIONS(1328),
    [anon_sym_U_DQUOTE] = ACTIONS(1328),
    [anon_sym_u8_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [sym_true] = ACTIONS(1326),
    [sym_false] = ACTIONS(1326),
    [anon_sym_NULL] = ACTIONS(1326),
    [anon_sym_nullptr] = ACTIONS(1326),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(277)] = {
    [sym_identifier] = ACTIONS(1330),
    [aux_sym_preproc_include_token1] = ACTIONS(1330),
    [aux_sym_preproc_def_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1330),
    [sym_preproc_directive] = ACTIONS(1330),
    [anon_sym_LPAREN2] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1330),
    [anon_sym_PLUS] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym___extension__] = ACTIONS(1330),
    [anon_sym_typedef] = ACTIONS(1330),
    [anon_sym_extern] = ACTIONS(1330),
    [anon_sym___attribute__] = ACTIONS(1330),
    [anon_sym___attribute] = ACTIONS(1330),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1332),
    [anon_sym___declspec] = ACTIONS(1330),
    [anon_sym___cdecl] = ACTIONS(1330),
    [anon_sym___clrcall] = ACTIONS(1330),
    [anon_sym___stdcall] = ACTIONS(1330),
    [anon_sym___fastcall] = ACTIONS(1330),
    [anon_sym___thiscall] = ACTIONS(1330),
    [anon_sym___vectorcall] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_signed] = ACTIONS(1330),
    [anon_sym_unsigned] = ACTIONS(1330),
    [anon_sym_long] = ACTIONS(1330),
    [anon_sym_short] = ACTIONS(1330),
    [anon_sym_static] = ACTIONS(1330),
    [anon_sym_auto] = ACTIONS(1330),
    [anon_sym_register] = ACTIONS(1330),
    [anon_sym_inline] = ACTIONS(1330),
    [anon_sym___inline] = ACTIONS(1330),
    [anon_sym___inline__] = ACTIONS(1330),
    [anon_sym___forceinline] = ACTIONS(1330),
    [anon_sym_thread_local] = ACTIONS(1330),
    [anon_sym___thread] = ACTIONS(1330),
    [anon_sym_const] = ACTIONS(1330),
    [anon_sym_constexpr] = ACTIONS(1330),
    [anon_sym_volatile] = ACTIONS(1330),
    [anon_sym_restrict] = ACTIONS(1330),
    [anon_sym___restrict__] = ACTIONS(1330),
    [anon_sym__Atomic] = ACTIONS(1330),
    [anon_sym__Noreturn] = ACTIONS(1330),
    [anon_sym_noreturn] = ACTIONS(1330),
    [anon_sym__Nonnull] = ACTIONS(1330),
    [anon_sym_alignas] = ACTIONS(1330),
    [anon_sym__Alignas] = ACTIONS(1330),
    [sym_primitive_type] = ACTIONS(1330),
    [anon_sym_enum] = ACTIONS(1330),
    [anon_sym_struct] = ACTIONS(1330),
    [anon_sym_union] = ACTIONS(1330),
    [anon_sym_if] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1330),
    [anon_sym_case] = ACTIONS(1330),
    [anon_sym_default] = ACTIONS(1330),
    [anon_sym_while] = ACTIONS(1330),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_for] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1330),
    [anon_sym_break] = ACTIONS(1330),
    [anon_sym_continue] = ACTIONS(1330),
    [anon_sym_goto] = ACTIONS(1330),
    [anon_sym___try] = ACTIONS(1330),
    [anon_sym___leave] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_sizeof] = ACTIONS(1330),
    [anon_sym___alignof__] = ACTIONS(1330),
    [anon_sym___alignof] = ACTIONS(1330),
    [anon_sym__alignof] = ACTIONS(1330),
    [anon_sym_alignof] = ACTIONS(1330),
    [anon_sym__Alignof] = ACTIONS(1330),
    [anon_sym_offsetof] = ACTIONS(1330),
    [anon_sym__Generic] = ACTIONS(1330),
    [anon_sym_asm] = ACTIONS(1330),
    [anon_sym___asm__] = ACTIONS(1330),
    [anon_sym___asm] = ACTIONS(1330),
    [sym_number_literal] = ACTIONS(1332),
    [anon_sym_L_SQUOTE] = ACTIONS(1332),
    [anon_sym_u_SQUOTE] = ACTIONS(1332),
    [anon_sym_U_SQUOTE] = ACTIONS(1332),
    [anon_sym_u8_SQUOTE] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_L_DQUOTE] = ACTIONS(1332),
    [anon_sym_u_DQUOTE] = ACTIONS(1332),
    [anon_sym_U_DQUOTE] = ACTIONS(1332),
    [anon_sym_u8_DQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [sym_true] = ACTIONS(1330),
    [sym_false] = ACTIONS(1330),
    [anon_sym_NULL] = ACTIONS(1330),
    [anon_sym_nullptr] = ACTIONS(1330),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(278)] = {
    [sym_identifier] = ACTIONS(1340),
    [aux_sym_preproc_include_token1] = ACTIONS(1340),
    [aux_sym_preproc_def_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1340),
    [sym_preproc_directive] = ACTIONS(1340),
    [anon_sym_LPAREN2] = ACTIONS(1342),
    [anon_sym_BANG] = ACTIONS(1342),
    [anon_sym_TILDE] = ACTIONS(1342),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1342),
    [anon_sym_AMP] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1342),
    [anon_sym___extension__] = ACTIONS(1340),
    [anon_sym_typedef] = ACTIONS(1340),
    [anon_sym_extern] = ACTIONS(1340),
    [anon_sym___attribute__] = ACTIONS(1340),
    [anon_sym___attribute] = ACTIONS(1340),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1342),
    [anon_sym___declspec] = ACTIONS(1340),
    [anon_sym___cdecl] = ACTIONS(1340),
    [anon_sym___clrcall] = ACTIONS(1340),
    [anon_sym___stdcall] = ACTIONS(1340),
    [anon_sym___fastcall] = ACTIONS(1340),
    [anon_sym___thiscall] = ACTIONS(1340),
    [anon_sym___vectorcall] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1342),
    [anon_sym_RBRACE] = ACTIONS(1342),
    [anon_sym_signed] = ACTIONS(1340),
    [anon_sym_unsigned] = ACTIONS(1340),
    [anon_sym_long] = ACTIONS(1340),
    [anon_sym_short] = ACTIONS(1340),
    [anon_sym_static] = ACTIONS(1340),
    [anon_sym_auto] = ACTIONS(1340),
    [anon_sym_register] = ACTIONS(1340),
    [anon_sym_inline] = ACTIONS(1340),
    [anon_sym___inline] = ACTIONS(1340),
    [anon_sym___inline__] = ACTIONS(1340),
    [anon_sym___forceinline] = ACTIONS(1340),
    [anon_sym_thread_local] = ACTIONS(1340),
    [anon_sym___thread] = ACTIONS(1340),
    [anon_sym_const] = ACTIONS(1340),
    [anon_sym_constexpr] = ACTIONS(1340),
    [anon_sym_volatile] = ACTIONS(1340),
    [anon_sym_restrict] = ACTIONS(1340),
    [anon_sym___restrict__] = ACTIONS(1340),
    [anon_sym__Atomic] = ACTIONS(1340),
    [anon_sym__Noreturn] = ACTIONS(1340),
    [anon_sym_noreturn] = ACTIONS(1340),
    [anon_sym__Nonnull] = ACTIONS(1340),
    [anon_sym_alignas] = ACTIONS(1340),
    [anon_sym__Alignas] = ACTIONS(1340),
    [sym_primitive_type] = ACTIONS(1340),
    [anon_sym_enum] = ACTIONS(1340),
    [anon_sym_struct] = ACTIONS(1340),
    [anon_sym_union] = ACTIONS(1340),
    [anon_sym_if] = ACTIONS(1340),
    [anon_sym_switch] = ACTIONS(1340),
    [anon_sym_case] = ACTIONS(1340),
    [anon_sym_default] = ACTIONS(1340),
    [anon_sym_while] = ACTIONS(1340),
    [anon_sym_do] = ACTIONS(1340),
    [anon_sym_for] = ACTIONS(1340),
    [anon_sym_return] = ACTIONS(1340),
    [anon_sym_break] = ACTIONS(1340),
    [anon_sym_continue] = ACTIONS(1340),
    [anon_sym_goto] = ACTIONS(1340),
    [anon_sym___try] = ACTIONS(1340),
    [anon_sym___leave] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1342),
    [anon_sym_PLUS_PLUS] = ACTIONS(1342),
    [anon_sym_sizeof] = ACTIONS(1340),
    [anon_sym___alignof__] = ACTIONS(1340),
    [anon_sym___alignof] = ACTIONS(1340),
    [anon_sym__alignof] = ACTIONS(1340),
    [anon_sym_alignof] = ACTIONS(1340),
    [anon_sym__Alignof] = ACTIONS(1340),
    [anon_sym_offsetof] = ACTIONS(1340),
    [anon_sym__Generic] = ACTIONS(1340),
    [anon_sym_asm] = ACTIONS(1340),
    [anon_sym___asm__] = ACTIONS(1340),
    [anon_sym___asm] = ACTIONS(1340),
    [sym_number_literal] = ACTIONS(1342),
    [anon_sym_L_SQUOTE] = ACTIONS(1342),
    [anon_sym_u_SQUOTE] = ACTIONS(1342),
    [anon_sym_U_SQUOTE] = ACTIONS(1342),
    [anon_sym_u8_SQUOTE] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_L_DQUOTE] = ACTIONS(1342),
    [anon_sym_u_DQUOTE] = ACTIONS(1342),
    [anon_sym_U_DQUOTE] = ACTIONS(1342),
    [anon_sym_u8_DQUOTE] = ACTIONS(1342),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [sym_true] = ACTIONS(1340),
    [sym_false] = ACTIONS(1340),
    [anon_sym_NULL] = ACTIONS(1340),
    [anon_sym_nullptr] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(279)] = {
    [sym_identifier] = ACTIONS(1344),
    [aux_sym_preproc_include_token1] = ACTIONS(1344),
    [aux_sym_preproc_def_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1344),
    [sym_preproc_directive] = ACTIONS(1344),
    [anon_sym_LPAREN2] = ACTIONS(1346),
    [anon_sym_BANG] = ACTIONS(1346),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1346),
    [anon_sym_AMP] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1346),
    [anon_sym___extension__] = ACTIONS(1344),
    [anon_sym_typedef] = ACTIONS(1344),
    [anon_sym_extern] = ACTIONS(1344),
    [anon_sym___attribute__] = ACTIONS(1344),
    [anon_sym___attribute] = ACTIONS(1344),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1346),
    [anon_sym___declspec] = ACTIONS(1344),
    [anon_sym___cdecl] = ACTIONS(1344),
    [anon_sym___clrcall] = ACTIONS(1344),
    [anon_sym___stdcall] = ACTIONS(1344),
    [anon_sym___fastcall] = ACTIONS(1344),
    [anon_sym___thiscall] = ACTIONS(1344),
    [anon_sym___vectorcall] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [anon_sym_RBRACE] = ACTIONS(1346),
    [anon_sym_signed] = ACTIONS(1344),
    [anon_sym_unsigned] = ACTIONS(1344),
    [anon_sym_long] = ACTIONS(1344),
    [anon_sym_short] = ACTIONS(1344),
    [anon_sym_static] = ACTIONS(1344),
    [anon_sym_auto] = ACTIONS(1344),
    [anon_sym_register] = ACTIONS(1344),
    [anon_sym_inline] = ACTIONS(1344),
    [anon_sym___inline] = ACTIONS(1344),
    [anon_sym___inline__] = ACTIONS(1344),
    [anon_sym___forceinline] = ACTIONS(1344),
    [anon_sym_thread_local] = ACTIONS(1344),
    [anon_sym___thread] = ACTIONS(1344),
    [anon_sym_const] = ACTIONS(1344),
    [anon_sym_constexpr] = ACTIONS(1344),
    [anon_sym_volatile] = ACTIONS(1344),
    [anon_sym_restrict] = ACTIONS(1344),
    [anon_sym___restrict__] = ACTIONS(1344),
    [anon_sym__Atomic] = ACTIONS(1344),
    [anon_sym__Noreturn] = ACTIONS(1344),
    [anon_sym_noreturn] = ACTIONS(1344),
    [anon_sym__Nonnull] = ACTIONS(1344),
    [anon_sym_alignas] = ACTIONS(1344),
    [anon_sym__Alignas] = ACTIONS(1344),
    [sym_primitive_type] = ACTIONS(1344),
    [anon_sym_enum] = ACTIONS(1344),
    [anon_sym_struct] = ACTIONS(1344),
    [anon_sym_union] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_switch] = ACTIONS(1344),
    [anon_sym_case] = ACTIONS(1344),
    [anon_sym_default] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_break] = ACTIONS(1344),
    [anon_sym_continue] = ACTIONS(1344),
    [anon_sym_goto] = ACTIONS(1344),
    [anon_sym___try] = ACTIONS(1344),
    [anon_sym___leave] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1346),
    [anon_sym_PLUS_PLUS] = ACTIONS(1346),
    [anon_sym_sizeof] = ACTIONS(1344),
    [anon_sym___alignof__] = ACTIONS(1344),
    [anon_sym___alignof] = ACTIONS(1344),
    [anon_sym__alignof] = ACTIONS(1344),
    [anon_sym_alignof] = ACTIONS(1344),
    [anon_sym__Alignof] = ACTIONS(1344),
    [anon_sym_offsetof] = ACTIONS(1344),
    [anon_sym__Generic] = ACTIONS(1344),
    [anon_sym_asm] = ACTIONS(1344),
    [anon_sym___asm__] = ACTIONS(1344),
    [anon_sym___asm] = ACTIONS(1344),
    [sym_number_literal] = ACTIONS(1346),
    [anon_sym_L_SQUOTE] = ACTIONS(1346),
    [anon_sym_u_SQUOTE] = ACTIONS(1346),
    [anon_sym_U_SQUOTE] = ACTIONS(1346),
    [anon_sym_u8_SQUOTE] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_L_DQUOTE] = ACTIONS(1346),
    [anon_sym_u_DQUOTE] = ACTIONS(1346),
    [anon_sym_U_DQUOTE] = ACTIONS(1346),
    [anon_sym_u8_DQUOTE] = ACTIONS(1346),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [anon_sym_NULL] = ACTIONS(1344),
    [anon_sym_nullptr] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(280)] = {
    [sym_identifier] = ACTIONS(1348),
    [aux_sym_preproc_include_token1] = ACTIONS(1348),
    [aux_sym_preproc_def_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1348),
    [sym_preproc_directive] = ACTIONS(1348),
    [anon_sym_LPAREN2] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym___extension__] = ACTIONS(1348),
    [anon_sym_typedef] = ACTIONS(1348),
    [anon_sym_extern] = ACTIONS(1348),
    [anon_sym___attribute__] = ACTIONS(1348),
    [anon_sym___attribute] = ACTIONS(1348),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1350),
    [anon_sym___declspec] = ACTIONS(1348),
    [anon_sym___cdecl] = ACTIONS(1348),
    [anon_sym___clrcall] = ACTIONS(1348),
    [anon_sym___stdcall] = ACTIONS(1348),
    [anon_sym___fastcall] = ACTIONS(1348),
    [anon_sym___thiscall] = ACTIONS(1348),
    [anon_sym___vectorcall] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [anon_sym_signed] = ACTIONS(1348),
    [anon_sym_unsigned] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_auto] = ACTIONS(1348),
    [anon_sym_register] = ACTIONS(1348),
    [anon_sym_inline] = ACTIONS(1348),
    [anon_sym___inline] = ACTIONS(1348),
    [anon_sym___inline__] = ACTIONS(1348),
    [anon_sym___forceinline] = ACTIONS(1348),
    [anon_sym_thread_local] = ACTIONS(1348),
    [anon_sym___thread] = ACTIONS(1348),
    [anon_sym_const] = ACTIONS(1348),
    [anon_sym_constexpr] = ACTIONS(1348),
    [anon_sym_volatile] = ACTIONS(1348),
    [anon_sym_restrict] = ACTIONS(1348),
    [anon_sym___restrict__] = ACTIONS(1348),
    [anon_sym__Atomic] = ACTIONS(1348),
    [anon_sym__Noreturn] = ACTIONS(1348),
    [anon_sym_noreturn] = ACTIONS(1348),
    [anon_sym__Nonnull] = ACTIONS(1348),
    [anon_sym_alignas] = ACTIONS(1348),
    [anon_sym__Alignas] = ACTIONS(1348),
    [sym_primitive_type] = ACTIONS(1348),
    [anon_sym_enum] = ACTIONS(1348),
    [anon_sym_struct] = ACTIONS(1348),
    [anon_sym_union] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_goto] = ACTIONS(1348),
    [anon_sym___try] = ACTIONS(1348),
    [anon_sym___leave] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_sizeof] = ACTIONS(1348),
    [anon_sym___alignof__] = ACTIONS(1348),
    [anon_sym___alignof] = ACTIONS(1348),
    [anon_sym__alignof] = ACTIONS(1348),
    [anon_sym_alignof] = ACTIONS(1348),
    [anon_sym__Alignof] = ACTIONS(1348),
    [anon_sym_offsetof] = ACTIONS(1348),
    [anon_sym__Generic] = ACTIONS(1348),
    [anon_sym_asm] = ACTIONS(1348),
    [anon_sym___asm__] = ACTIONS(1348),
    [anon_sym___asm] = ACTIONS(1348),
    [sym_number_literal] = ACTIONS(1350),
    [anon_sym_L_SQUOTE] = ACTIONS(1350),
    [anon_sym_u_SQUOTE] = ACTIONS(1350),
    [anon_sym_U_SQUOTE] = ACTIONS(1350),
    [anon_sym_u8_SQUOTE] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_L_DQUOTE] = ACTIONS(1350),
    [anon_sym_u_DQUOTE] = ACTIONS(1350),
    [anon_sym_U_DQUOTE] = ACTIONS(1350),
    [anon_sym_u8_DQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [sym_true] = ACTIONS(1348),
    [sym_false] = ACTIONS(1348),
    [anon_sym_NULL] = ACTIONS(1348),
    [anon_sym_nullptr] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(281)] = {
    [sym_identifier] = ACTIONS(1352),
    [aux_sym_preproc_include_token1] = ACTIONS(1352),
    [aux_sym_preproc_def_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1352),
    [sym_preproc_directive] = ACTIONS(1352),
    [anon_sym_LPAREN2] = ACTIONS(1354),
    [anon_sym_BANG] = ACTIONS(1354),
    [anon_sym_TILDE] = ACTIONS(1354),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1354),
    [anon_sym_AMP] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1354),
    [anon_sym___extension__] = ACTIONS(1352),
    [anon_sym_typedef] = ACTIONS(1352),
    [anon_sym_extern] = ACTIONS(1352),
    [anon_sym___attribute__] = ACTIONS(1352),
    [anon_sym___attribute] = ACTIONS(1352),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1354),
    [anon_sym___declspec] = ACTIONS(1352),
    [anon_sym___cdecl] = ACTIONS(1352),
    [anon_sym___clrcall] = ACTIONS(1352),
    [anon_sym___stdcall] = ACTIONS(1352),
    [anon_sym___fastcall] = ACTIONS(1352),
    [anon_sym___thiscall] = ACTIONS(1352),
    [anon_sym___vectorcall] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1354),
    [anon_sym_RBRACE] = ACTIONS(1354),
    [anon_sym_signed] = ACTIONS(1352),
    [anon_sym_unsigned] = ACTIONS(1352),
    [anon_sym_long] = ACTIONS(1352),
    [anon_sym_short] = ACTIONS(1352),
    [anon_sym_static] = ACTIONS(1352),
    [anon_sym_auto] = ACTIONS(1352),
    [anon_sym_register] = ACTIONS(1352),
    [anon_sym_inline] = ACTIONS(1352),
    [anon_sym___inline] = ACTIONS(1352),
    [anon_sym___inline__] = ACTIONS(1352),
    [anon_sym___forceinline] = ACTIONS(1352),
    [anon_sym_thread_local] = ACTIONS(1352),
    [anon_sym___thread] = ACTIONS(1352),
    [anon_sym_const] = ACTIONS(1352),
    [anon_sym_constexpr] = ACTIONS(1352),
    [anon_sym_volatile] = ACTIONS(1352),
    [anon_sym_restrict] = ACTIONS(1352),
    [anon_sym___restrict__] = ACTIONS(1352),
    [anon_sym__Atomic] = ACTIONS(1352),
    [anon_sym__Noreturn] = ACTIONS(1352),
    [anon_sym_noreturn] = ACTIONS(1352),
    [anon_sym__Nonnull] = ACTIONS(1352),
    [anon_sym_alignas] = ACTIONS(1352),
    [anon_sym__Alignas] = ACTIONS(1352),
    [sym_primitive_type] = ACTIONS(1352),
    [anon_sym_enum] = ACTIONS(1352),
    [anon_sym_struct] = ACTIONS(1352),
    [anon_sym_union] = ACTIONS(1352),
    [anon_sym_if] = ACTIONS(1352),
    [anon_sym_switch] = ACTIONS(1352),
    [anon_sym_case] = ACTIONS(1352),
    [anon_sym_default] = ACTIONS(1352),
    [anon_sym_while] = ACTIONS(1352),
    [anon_sym_do] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_break] = ACTIONS(1352),
    [anon_sym_continue] = ACTIONS(1352),
    [anon_sym_goto] = ACTIONS(1352),
    [anon_sym___try] = ACTIONS(1352),
    [anon_sym___leave] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1354),
    [anon_sym_PLUS_PLUS] = ACTIONS(1354),
    [anon_sym_sizeof] = ACTIONS(1352),
    [anon_sym___alignof__] = ACTIONS(1352),
    [anon_sym___alignof] = ACTIONS(1352),
    [anon_sym__alignof] = ACTIONS(1352),
    [anon_sym_alignof] = ACTIONS(1352),
    [anon_sym__Alignof] = ACTIONS(1352),
    [anon_sym_offsetof] = ACTIONS(1352),
    [anon_sym__Generic] = ACTIONS(1352),
    [anon_sym_asm] = ACTIONS(1352),
    [anon_sym___asm__] = ACTIONS(1352),
    [anon_sym___asm] = ACTIONS(1352),
    [sym_number_literal] = ACTIONS(1354),
    [anon_sym_L_SQUOTE] = ACTIONS(1354),
    [anon_sym_u_SQUOTE] = ACTIONS(1354),
    [anon_sym_U_SQUOTE] = ACTIONS(1354),
    [anon_sym_u8_SQUOTE] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_L_DQUOTE] = ACTIONS(1354),
    [anon_sym_u_DQUOTE] = ACTIONS(1354),
    [anon_sym_U_DQUOTE] = ACTIONS(1354),
    [anon_sym_u8_DQUOTE] = ACTIONS(1354),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [sym_true] = ACTIONS(1352),
    [sym_false] = ACTIONS(1352),
    [anon_sym_NULL] = ACTIONS(1352),
    [anon_sym_nullptr] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(282)] = {
    [sym_identifier] = ACTIONS(1310),
    [aux_sym_preproc_include_token1] = ACTIONS(1310),
    [aux_sym_preproc_def_token1] = ACTIONS(1310),
    [aux_sym_preproc_if_token1] = ACTIONS(1310),
    [aux_sym_preproc_if_token2] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1310),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1310),
    [sym_preproc_directive] = ACTIONS(1310),
    [anon_sym_LPAREN2] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym___extension__] = ACTIONS(1310),
    [anon_sym_typedef] = ACTIONS(1310),
    [anon_sym_extern] = ACTIONS(1310),
    [anon_sym___attribute__] = ACTIONS(1310),
    [anon_sym___attribute] = ACTIONS(1310),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1312),
    [anon_sym___declspec] = ACTIONS(1310),
    [anon_sym___cdecl] = ACTIONS(1310),
    [anon_sym___clrcall] = ACTIONS(1310),
    [anon_sym___stdcall] = ACTIONS(1310),
    [anon_sym___fastcall] = ACTIONS(1310),
    [anon_sym___thiscall] = ACTIONS(1310),
    [anon_sym___vectorcall] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_signed] = ACTIONS(1310),
    [anon_sym_unsigned] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_auto] = ACTIONS(1310),
    [anon_sym_register] = ACTIONS(1310),
    [anon_sym_inline] = ACTIONS(1310),
    [anon_sym___inline] = ACTIONS(1310),
    [anon_sym___inline__] = ACTIONS(1310),
    [anon_sym___forceinline] = ACTIONS(1310),
    [anon_sym_thread_local] = ACTIONS(1310),
    [anon_sym___thread] = ACTIONS(1310),
    [anon_sym_const] = ACTIONS(1310),
    [anon_sym_constexpr] = ACTIONS(1310),
    [anon_sym_volatile] = ACTIONS(1310),
    [anon_sym_restrict] = ACTIONS(1310),
    [anon_sym___restrict__] = ACTIONS(1310),
    [anon_sym__Atomic] = ACTIONS(1310),
    [anon_sym__Noreturn] = ACTIONS(1310),
    [anon_sym_noreturn] = ACTIONS(1310),
    [anon_sym__Nonnull] = ACTIONS(1310),
    [anon_sym_alignas] = ACTIONS(1310),
    [anon_sym__Alignas] = ACTIONS(1310),
    [sym_primitive_type] = ACTIONS(1310),
    [anon_sym_enum] = ACTIONS(1310),
    [anon_sym_struct] = ACTIONS(1310),
    [anon_sym_union] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_goto] = ACTIONS(1310),
    [anon_sym___try] = ACTIONS(1310),
    [anon_sym___leave] = ACTIONS(1310),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_sizeof] = ACTIONS(1310),
    [anon_sym___alignof__] = ACTIONS(1310),
    [anon_sym___alignof] = ACTIONS(1310),
    [anon_sym__alignof] = ACTIONS(1310),
    [anon_sym_alignof] = ACTIONS(1310),
    [anon_sym__Alignof] = ACTIONS(1310),
    [anon_sym_offsetof] = ACTIONS(1310),
    [anon_sym__Generic] = ACTIONS(1310),
    [anon_sym_asm] = ACTIONS(1310),
    [anon_sym___asm__] = ACTIONS(1310),
    [anon_sym___asm] = ACTIONS(1310),
    [sym_number_literal] = ACTIONS(1312),
    [anon_sym_L_SQUOTE] = ACTIONS(1312),
    [anon_sym_u_SQUOTE] = ACTIONS(1312),
    [anon_sym_U_SQUOTE] = ACTIONS(1312),
    [anon_sym_u8_SQUOTE] = ACTIONS(1312),
    [anon_sym_SQUOTE] = ACTIONS(1312),
    [anon_sym_L_DQUOTE] = ACTIONS(1312),
    [anon_sym_u_DQUOTE] = ACTIONS(1312),
    [anon_sym_U_DQUOTE] = ACTIONS(1312),
    [anon_sym_u8_DQUOTE] = ACTIONS(1312),
    [anon_sym_DQUOTE] = ACTIONS(1312),
    [sym_true] = ACTIONS(1310),
    [sym_false] = ACTIONS(1310),
    [anon_sym_NULL] = ACTIONS(1310),
    [anon_sym_nullptr] = ACTIONS(1310),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(283)] = {
    [ts_builtin_sym_end] = ACTIONS(1358),
    [sym_identifier] = ACTIONS(1356),
    [aux_sym_preproc_include_token1] = ACTIONS(1356),
    [aux_sym_preproc_def_token1] = ACTIONS(1356),
    [anon_sym_COMMA] = ACTIONS(1358),
    [aux_sym_preproc_if_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1356),
    [sym_preproc_directive] = ACTIONS(1356),
    [anon_sym_LPAREN2] = ACTIONS(1358),
    [anon_sym_BANG] = ACTIONS(1358),
    [anon_sym_TILDE] = ACTIONS(1358),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1358),
    [anon_sym_AMP] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1358),
    [anon_sym___extension__] = ACTIONS(1356),
    [anon_sym_typedef] = ACTIONS(1356),
    [anon_sym_extern] = ACTIONS(1356),
    [anon_sym___attribute__] = ACTIONS(1356),
    [anon_sym___attribute] = ACTIONS(1356),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1358),
    [anon_sym___declspec] = ACTIONS(1356),
    [anon_sym___cdecl] = ACTIONS(1356),
    [anon_sym___clrcall] = ACTIONS(1356),
    [anon_sym___stdcall] = ACTIONS(1356),
    [anon_sym___fastcall] = ACTIONS(1356),
    [anon_sym___thiscall] = ACTIONS(1356),
    [anon_sym___vectorcall] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1358),
    [anon_sym_RBRACE] = ACTIONS(1358),
    [anon_sym_signed] = ACTIONS(1356),
    [anon_sym_unsigned] = ACTIONS(1356),
    [anon_sym_long] = ACTIONS(1356),
    [anon_sym_short] = ACTIONS(1356),
    [anon_sym_static] = ACTIONS(1356),
    [anon_sym_auto] = ACTIONS(1356),
    [anon_sym_register] = ACTIONS(1356),
    [anon_sym_inline] = ACTIONS(1356),
    [anon_sym___inline] = ACTIONS(1356),
    [anon_sym___inline__] = ACTIONS(1356),
    [anon_sym___forceinline] = ACTIONS(1356),
    [anon_sym_thread_local] = ACTIONS(1356),
    [anon_sym___thread] = ACTIONS(1356),
    [anon_sym_const] = ACTIONS(1356),
    [anon_sym_constexpr] = ACTIONS(1356),
    [anon_sym_volatile] = ACTIONS(1356),
    [anon_sym_restrict] = ACTIONS(1356),
    [anon_sym___restrict__] = ACTIONS(1356),
    [anon_sym__Atomic] = ACTIONS(1356),
    [anon_sym__Noreturn] = ACTIONS(1356),
    [anon_sym_noreturn] = ACTIONS(1356),
    [anon_sym__Nonnull] = ACTIONS(1356),
    [anon_sym_alignas] = ACTIONS(1356),
    [anon_sym__Alignas] = ACTIONS(1356),
    [sym_primitive_type] = ACTIONS(1356),
    [anon_sym_enum] = ACTIONS(1356),
    [anon_sym_struct] = ACTIONS(1356),
    [anon_sym_union] = ACTIONS(1356),
    [anon_sym_if] = ACTIONS(1356),
    [anon_sym_switch] = ACTIONS(1356),
    [anon_sym_case] = ACTIONS(1356),
    [anon_sym_default] = ACTIONS(1356),
    [anon_sym_while] = ACTIONS(1356),
    [anon_sym_do] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1356),
    [anon_sym_return] = ACTIONS(1356),
    [anon_sym_break] = ACTIONS(1356),
    [anon_sym_continue] = ACTIONS(1356),
    [anon_sym_goto] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1358),
    [anon_sym_PLUS_PLUS] = ACTIONS(1358),
    [anon_sym_sizeof] = ACTIONS(1356),
    [anon_sym___alignof__] = ACTIONS(1356),
    [anon_sym___alignof] = ACTIONS(1356),
    [anon_sym__alignof] = ACTIONS(1356),
    [anon_sym_alignof] = ACTIONS(1356),
    [anon_sym__Alignof] = ACTIONS(1356),
    [anon_sym_offsetof] = ACTIONS(1356),
    [anon_sym__Generic] = ACTIONS(1356),
    [anon_sym_asm] = ACTIONS(1356),
    [anon_sym___asm__] = ACTIONS(1356),
    [anon_sym___asm] = ACTIONS(1356),
    [sym_number_literal] = ACTIONS(1358),
    [anon_sym_L_SQUOTE] = ACTIONS(1358),
    [anon_sym_u_SQUOTE] = ACTIONS(1358),
    [anon_sym_U_SQUOTE] = ACTIONS(1358),
    [anon_sym_u8_SQUOTE] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_L_DQUOTE] = ACTIONS(1358),
    [anon_sym_u_DQUOTE] = ACTIONS(1358),
    [anon_sym_U_DQUOTE] = ACTIONS(1358),
    [anon_sym_u8_DQUOTE] = ACTIONS(1358),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [sym_true] = ACTIONS(1356),
    [sym_false] = ACTIONS(1356),
    [anon_sym_NULL] = ACTIONS(1356),
    [anon_sym_nullptr] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(284)] = {
    [sym_identifier] = ACTIONS(1364),
    [aux_sym_preproc_include_token1] = ACTIONS(1364),
    [aux_sym_preproc_def_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token2] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1364),
    [sym_preproc_directive] = ACTIONS(1364),
    [anon_sym_LPAREN2] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_PLUS] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym___extension__] = ACTIONS(1364),
    [anon_sym_typedef] = ACTIONS(1364),
    [anon_sym_extern] = ACTIONS(1364),
    [anon_sym___attribute__] = ACTIONS(1364),
    [anon_sym___attribute] = ACTIONS(1364),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1366),
    [anon_sym___declspec] = ACTIONS(1364),
    [anon_sym___cdecl] = ACTIONS(1364),
    [anon_sym___clrcall] = ACTIONS(1364),
    [anon_sym___stdcall] = ACTIONS(1364),
    [anon_sym___fastcall] = ACTIONS(1364),
    [anon_sym___thiscall] = ACTIONS(1364),
    [anon_sym___vectorcall] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_signed] = ACTIONS(1364),
    [anon_sym_unsigned] = ACTIONS(1364),
    [anon_sym_long] = ACTIONS(1364),
    [anon_sym_short] = ACTIONS(1364),
    [anon_sym_static] = ACTIONS(1364),
    [anon_sym_auto] = ACTIONS(1364),
    [anon_sym_register] = ACTIONS(1364),
    [anon_sym_inline] = ACTIONS(1364),
    [anon_sym___inline] = ACTIONS(1364),
    [anon_sym___inline__] = ACTIONS(1364),
    [anon_sym___forceinline] = ACTIONS(1364),
    [anon_sym_thread_local] = ACTIONS(1364),
    [anon_sym___thread] = ACTIONS(1364),
    [anon_sym_const] = ACTIONS(1364),
    [anon_sym_constexpr] = ACTIONS(1364),
    [anon_sym_volatile] = ACTIONS(1364),
    [anon_sym_restrict] = ACTIONS(1364),
    [anon_sym___restrict__] = ACTIONS(1364),
    [anon_sym__Atomic] = ACTIONS(1364),
    [anon_sym__Noreturn] = ACTIONS(1364),
    [anon_sym_noreturn] = ACTIONS(1364),
    [anon_sym__Nonnull] = ACTIONS(1364),
    [anon_sym_alignas] = ACTIONS(1364),
    [anon_sym__Alignas] = ACTIONS(1364),
    [sym_primitive_type] = ACTIONS(1364),
    [anon_sym_enum] = ACTIONS(1364),
    [anon_sym_struct] = ACTIONS(1364),
    [anon_sym_union] = ACTIONS(1364),
    [anon_sym_if] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1364),
    [anon_sym_case] = ACTIONS(1364),
    [anon_sym_default] = ACTIONS(1364),
    [anon_sym_while] = ACTIONS(1364),
    [anon_sym_do] = ACTIONS(1364),
    [anon_sym_for] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1364),
    [anon_sym_break] = ACTIONS(1364),
    [anon_sym_continue] = ACTIONS(1364),
    [anon_sym_goto] = ACTIONS(1364),
    [anon_sym___try] = ACTIONS(1364),
    [anon_sym___leave] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1366),
    [anon_sym_sizeof] = ACTIONS(1364),
    [anon_sym___alignof__] = ACTIONS(1364),
    [anon_sym___alignof] = ACTIONS(1364),
    [anon_sym__alignof] = ACTIONS(1364),
    [anon_sym_alignof] = ACTIONS(1364),
    [anon_sym__Alignof] = ACTIONS(1364),
    [anon_sym_offsetof] = ACTIONS(1364),
    [anon_sym__Generic] = ACTIONS(1364),
    [anon_sym_asm] = ACTIONS(1364),
    [anon_sym___asm__] = ACTIONS(1364),
    [anon_sym___asm] = ACTIONS(1364),
    [sym_number_literal] = ACTIONS(1366),
    [anon_sym_L_SQUOTE] = ACTIONS(1366),
    [anon_sym_u_SQUOTE] = ACTIONS(1366),
    [anon_sym_U_SQUOTE] = ACTIONS(1366),
    [anon_sym_u8_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_L_DQUOTE] = ACTIONS(1366),
    [anon_sym_u_DQUOTE] = ACTIONS(1366),
    [anon_sym_U_DQUOTE] = ACTIONS(1366),
    [anon_sym_u8_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [sym_true] = ACTIONS(1364),
    [sym_false] = ACTIONS(1364),
    [anon_sym_NULL] = ACTIONS(1364),
    [anon_sym_nullptr] = ACTIONS(1364),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(285)] = {
    [sym_identifier] = ACTIONS(1368),
    [aux_sym_preproc_include_token1] = ACTIONS(1368),
    [aux_sym_preproc_def_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token2] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1368),
    [sym_preproc_directive] = ACTIONS(1368),
    [anon_sym_LPAREN2] = ACTIONS(1370),
    [anon_sym_BANG] = ACTIONS(1370),
    [anon_sym_TILDE] = ACTIONS(1370),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_PLUS] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1370),
    [anon_sym_AMP] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1370),
    [anon_sym___extension__] = ACTIONS(1368),
    [anon_sym_typedef] = ACTIONS(1368),
    [anon_sym_extern] = ACTIONS(1368),
    [anon_sym___attribute__] = ACTIONS(1368),
    [anon_sym___attribute] = ACTIONS(1368),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1370),
    [anon_sym___declspec] = ACTIONS(1368),
    [anon_sym___cdecl] = ACTIONS(1368),
    [anon_sym___clrcall] = ACTIONS(1368),
    [anon_sym___stdcall] = ACTIONS(1368),
    [anon_sym___fastcall] = ACTIONS(1368),
    [anon_sym___thiscall] = ACTIONS(1368),
    [anon_sym___vectorcall] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1370),
    [anon_sym_signed] = ACTIONS(1368),
    [anon_sym_unsigned] = ACTIONS(1368),
    [anon_sym_long] = ACTIONS(1368),
    [anon_sym_short] = ACTIONS(1368),
    [anon_sym_static] = ACTIONS(1368),
    [anon_sym_auto] = ACTIONS(1368),
    [anon_sym_register] = ACTIONS(1368),
    [anon_sym_inline] = ACTIONS(1368),
    [anon_sym___inline] = ACTIONS(1368),
    [anon_sym___inline__] = ACTIONS(1368),
    [anon_sym___forceinline] = ACTIONS(1368),
    [anon_sym_thread_local] = ACTIONS(1368),
    [anon_sym___thread] = ACTIONS(1368),
    [anon_sym_const] = ACTIONS(1368),
    [anon_sym_constexpr] = ACTIONS(1368),
    [anon_sym_volatile] = ACTIONS(1368),
    [anon_sym_restrict] = ACTIONS(1368),
    [anon_sym___restrict__] = ACTIONS(1368),
    [anon_sym__Atomic] = ACTIONS(1368),
    [anon_sym__Noreturn] = ACTIONS(1368),
    [anon_sym_noreturn] = ACTIONS(1368),
    [anon_sym__Nonnull] = ACTIONS(1368),
    [anon_sym_alignas] = ACTIONS(1368),
    [anon_sym__Alignas] = ACTIONS(1368),
    [sym_primitive_type] = ACTIONS(1368),
    [anon_sym_enum] = ACTIONS(1368),
    [anon_sym_struct] = ACTIONS(1368),
    [anon_sym_union] = ACTIONS(1368),
    [anon_sym_if] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1368),
    [anon_sym_case] = ACTIONS(1368),
    [anon_sym_default] = ACTIONS(1368),
    [anon_sym_while] = ACTIONS(1368),
    [anon_sym_do] = ACTIONS(1368),
    [anon_sym_for] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1368),
    [anon_sym_break] = ACTIONS(1368),
    [anon_sym_continue] = ACTIONS(1368),
    [anon_sym_goto] = ACTIONS(1368),
    [anon_sym___try] = ACTIONS(1368),
    [anon_sym___leave] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1370),
    [anon_sym_PLUS_PLUS] = ACTIONS(1370),
    [anon_sym_sizeof] = ACTIONS(1368),
    [anon_sym___alignof__] = ACTIONS(1368),
    [anon_sym___alignof] = ACTIONS(1368),
    [anon_sym__alignof] = ACTIONS(1368),
    [anon_sym_alignof] = ACTIONS(1368),
    [anon_sym__Alignof] = ACTIONS(1368),
    [anon_sym_offsetof] = ACTIONS(1368),
    [anon_sym__Generic] = ACTIONS(1368),
    [anon_sym_asm] = ACTIONS(1368),
    [anon_sym___asm__] = ACTIONS(1368),
    [anon_sym___asm] = ACTIONS(1368),
    [sym_number_literal] = ACTIONS(1370),
    [anon_sym_L_SQUOTE] = ACTIONS(1370),
    [anon_sym_u_SQUOTE] = ACTIONS(1370),
    [anon_sym_U_SQUOTE] = ACTIONS(1370),
    [anon_sym_u8_SQUOTE] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_L_DQUOTE] = ACTIONS(1370),
    [anon_sym_u_DQUOTE] = ACTIONS(1370),
    [anon_sym_U_DQUOTE] = ACTIONS(1370),
    [anon_sym_u8_DQUOTE] = ACTIONS(1370),
    [anon_sym_DQUOTE] = ACTIONS(1370),
    [sym_true] = ACTIONS(1368),
    [sym_false] = ACTIONS(1368),
    [anon_sym_NULL] = ACTIONS(1368),
    [anon_sym_nullptr] = ACTIONS(1368),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(286)] = {
    [sym_identifier] = ACTIONS(1266),
    [aux_sym_preproc_include_token1] = ACTIONS(1266),
    [aux_sym_preproc_def_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token2] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1266),
    [sym_preproc_directive] = ACTIONS(1266),
    [anon_sym_LPAREN2] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1266),
    [anon_sym_PLUS] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1268),
    [anon_sym_AMP] = ACTIONS(1268),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym___extension__] = ACTIONS(1266),
    [anon_sym_typedef] = ACTIONS(1266),
    [anon_sym_extern] = ACTIONS(1266),
    [anon_sym___attribute__] = ACTIONS(1266),
    [anon_sym___attribute] = ACTIONS(1266),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1268),
    [anon_sym___declspec] = ACTIONS(1266),
    [anon_sym___cdecl] = ACTIONS(1266),
    [anon_sym___clrcall] = ACTIONS(1266),
    [anon_sym___stdcall] = ACTIONS(1266),
    [anon_sym___fastcall] = ACTIONS(1266),
    [anon_sym___thiscall] = ACTIONS(1266),
    [anon_sym___vectorcall] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_signed] = ACTIONS(1266),
    [anon_sym_unsigned] = ACTIONS(1266),
    [anon_sym_long] = ACTIONS(1266),
    [anon_sym_short] = ACTIONS(1266),
    [anon_sym_static] = ACTIONS(1266),
    [anon_sym_auto] = ACTIONS(1266),
    [anon_sym_register] = ACTIONS(1266),
    [anon_sym_inline] = ACTIONS(1266),
    [anon_sym___inline] = ACTIONS(1266),
    [anon_sym___inline__] = ACTIONS(1266),
    [anon_sym___forceinline] = ACTIONS(1266),
    [anon_sym_thread_local] = ACTIONS(1266),
    [anon_sym___thread] = ACTIONS(1266),
    [anon_sym_const] = ACTIONS(1266),
    [anon_sym_constexpr] = ACTIONS(1266),
    [anon_sym_volatile] = ACTIONS(1266),
    [anon_sym_restrict] = ACTIONS(1266),
    [anon_sym___restrict__] = ACTIONS(1266),
    [anon_sym__Atomic] = ACTIONS(1266),
    [anon_sym__Noreturn] = ACTIONS(1266),
    [anon_sym_noreturn] = ACTIONS(1266),
    [anon_sym__Nonnull] = ACTIONS(1266),
    [anon_sym_alignas] = ACTIONS(1266),
    [anon_sym__Alignas] = ACTIONS(1266),
    [sym_primitive_type] = ACTIONS(1266),
    [anon_sym_enum] = ACTIONS(1266),
    [anon_sym_struct] = ACTIONS(1266),
    [anon_sym_union] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_switch] = ACTIONS(1266),
    [anon_sym_case] = ACTIONS(1266),
    [anon_sym_default] = ACTIONS(1266),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_break] = ACTIONS(1266),
    [anon_sym_continue] = ACTIONS(1266),
    [anon_sym_goto] = ACTIONS(1266),
    [anon_sym___try] = ACTIONS(1266),
    [anon_sym___leave] = ACTIONS(1266),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_sizeof] = ACTIONS(1266),
    [anon_sym___alignof__] = ACTIONS(1266),
    [anon_sym___alignof] = ACTIONS(1266),
    [anon_sym__alignof] = ACTIONS(1266),
    [anon_sym_alignof] = ACTIONS(1266),
    [anon_sym__Alignof] = ACTIONS(1266),
    [anon_sym_offsetof] = ACTIONS(1266),
    [anon_sym__Generic] = ACTIONS(1266),
    [anon_sym_asm] = ACTIONS(1266),
    [anon_sym___asm__] = ACTIONS(1266),
    [anon_sym___asm] = ACTIONS(1266),
    [sym_number_literal] = ACTIONS(1268),
    [anon_sym_L_SQUOTE] = ACTIONS(1268),
    [anon_sym_u_SQUOTE] = ACTIONS(1268),
    [anon_sym_U_SQUOTE] = ACTIONS(1268),
    [anon_sym_u8_SQUOTE] = ACTIONS(1268),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_L_DQUOTE] = ACTIONS(1268),
    [anon_sym_u_DQUOTE] = ACTIONS(1268),
    [anon_sym_U_DQUOTE] = ACTIONS(1268),
    [anon_sym_u8_DQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1268),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [anon_sym_NULL] = ACTIONS(1266),
    [anon_sym_nullptr] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(287)] = {
    [sym_identifier] = ACTIONS(1274),
    [aux_sym_preproc_include_token1] = ACTIONS(1274),
    [aux_sym_preproc_def_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token2] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1274),
    [sym_preproc_directive] = ACTIONS(1274),
    [anon_sym_LPAREN2] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_TILDE] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1274),
    [anon_sym_PLUS] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1276),
    [anon_sym_AMP] = ACTIONS(1276),
    [anon_sym_SEMI] = ACTIONS(1276),
    [anon_sym___extension__] = ACTIONS(1274),
    [anon_sym_typedef] = ACTIONS(1274),
    [anon_sym_extern] = ACTIONS(1274),
    [anon_sym___attribute__] = ACTIONS(1274),
    [anon_sym___attribute] = ACTIONS(1274),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1276),
    [anon_sym___declspec] = ACTIONS(1274),
    [anon_sym___cdecl] = ACTIONS(1274),
    [anon_sym___clrcall] = ACTIONS(1274),
    [anon_sym___stdcall] = ACTIONS(1274),
    [anon_sym___fastcall] = ACTIONS(1274),
    [anon_sym___thiscall] = ACTIONS(1274),
    [anon_sym___vectorcall] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_signed] = ACTIONS(1274),
    [anon_sym_unsigned] = ACTIONS(1274),
    [anon_sym_long] = ACTIONS(1274),
    [anon_sym_short] = ACTIONS(1274),
    [anon_sym_static] = ACTIONS(1274),
    [anon_sym_auto] = ACTIONS(1274),
    [anon_sym_register] = ACTIONS(1274),
    [anon_sym_inline] = ACTIONS(1274),
    [anon_sym___inline] = ACTIONS(1274),
    [anon_sym___inline__] = ACTIONS(1274),
    [anon_sym___forceinline] = ACTIONS(1274),
    [anon_sym_thread_local] = ACTIONS(1274),
    [anon_sym___thread] = ACTIONS(1274),
    [anon_sym_const] = ACTIONS(1274),
    [anon_sym_constexpr] = ACTIONS(1274),
    [anon_sym_volatile] = ACTIONS(1274),
    [anon_sym_restrict] = ACTIONS(1274),
    [anon_sym___restrict__] = ACTIONS(1274),
    [anon_sym__Atomic] = ACTIONS(1274),
    [anon_sym__Noreturn] = ACTIONS(1274),
    [anon_sym_noreturn] = ACTIONS(1274),
    [anon_sym__Nonnull] = ACTIONS(1274),
    [anon_sym_alignas] = ACTIONS(1274),
    [anon_sym__Alignas] = ACTIONS(1274),
    [sym_primitive_type] = ACTIONS(1274),
    [anon_sym_enum] = ACTIONS(1274),
    [anon_sym_struct] = ACTIONS(1274),
    [anon_sym_union] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1274),
    [anon_sym_switch] = ACTIONS(1274),
    [anon_sym_case] = ACTIONS(1274),
    [anon_sym_default] = ACTIONS(1274),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_break] = ACTIONS(1274),
    [anon_sym_continue] = ACTIONS(1274),
    [anon_sym_goto] = ACTIONS(1274),
    [anon_sym___try] = ACTIONS(1274),
    [anon_sym___leave] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_PLUS_PLUS] = ACTIONS(1276),
    [anon_sym_sizeof] = ACTIONS(1274),
    [anon_sym___alignof__] = ACTIONS(1274),
    [anon_sym___alignof] = ACTIONS(1274),
    [anon_sym__alignof] = ACTIONS(1274),
    [anon_sym_alignof] = ACTIONS(1274),
    [anon_sym__Alignof] = ACTIONS(1274),
    [anon_sym_offsetof] = ACTIONS(1274),
    [anon_sym__Generic] = ACTIONS(1274),
    [anon_sym_asm] = ACTIONS(1274),
    [anon_sym___asm__] = ACTIONS(1274),
    [anon_sym___asm] = ACTIONS(1274),
    [sym_number_literal] = ACTIONS(1276),
    [anon_sym_L_SQUOTE] = ACTIONS(1276),
    [anon_sym_u_SQUOTE] = ACTIONS(1276),
    [anon_sym_U_SQUOTE] = ACTIONS(1276),
    [anon_sym_u8_SQUOTE] = ACTIONS(1276),
    [anon_sym_SQUOTE] = ACTIONS(1276),
    [anon_sym_L_DQUOTE] = ACTIONS(1276),
    [anon_sym_u_DQUOTE] = ACTIONS(1276),
    [anon_sym_U_DQUOTE] = ACTIONS(1276),
    [anon_sym_u8_DQUOTE] = ACTIONS(1276),
    [anon_sym_DQUOTE] = ACTIONS(1276),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [anon_sym_NULL] = ACTIONS(1274),
    [anon_sym_nullptr] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(288)] = {
    [sym_identifier] = ACTIONS(1282),
    [aux_sym_preproc_include_token1] = ACTIONS(1282),
    [aux_sym_preproc_def_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token2] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1282),
    [sym_preproc_directive] = ACTIONS(1282),
    [anon_sym_LPAREN2] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_TILDE] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1282),
    [anon_sym_PLUS] = ACTIONS(1282),
    [anon_sym_STAR] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym___extension__] = ACTIONS(1282),
    [anon_sym_typedef] = ACTIONS(1282),
    [anon_sym_extern] = ACTIONS(1282),
    [anon_sym___attribute__] = ACTIONS(1282),
    [anon_sym___attribute] = ACTIONS(1282),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1284),
    [anon_sym___declspec] = ACTIONS(1282),
    [anon_sym___cdecl] = ACTIONS(1282),
    [anon_sym___clrcall] = ACTIONS(1282),
    [anon_sym___stdcall] = ACTIONS(1282),
    [anon_sym___fastcall] = ACTIONS(1282),
    [anon_sym___thiscall] = ACTIONS(1282),
    [anon_sym___vectorcall] = ACTIONS(1282),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_signed] = ACTIONS(1282),
    [anon_sym_unsigned] = ACTIONS(1282),
    [anon_sym_long] = ACTIONS(1282),
    [anon_sym_short] = ACTIONS(1282),
    [anon_sym_static] = ACTIONS(1282),
    [anon_sym_auto] = ACTIONS(1282),
    [anon_sym_register] = ACTIONS(1282),
    [anon_sym_inline] = ACTIONS(1282),
    [anon_sym___inline] = ACTIONS(1282),
    [anon_sym___inline__] = ACTIONS(1282),
    [anon_sym___forceinline] = ACTIONS(1282),
    [anon_sym_thread_local] = ACTIONS(1282),
    [anon_sym___thread] = ACTIONS(1282),
    [anon_sym_const] = ACTIONS(1282),
    [anon_sym_constexpr] = ACTIONS(1282),
    [anon_sym_volatile] = ACTIONS(1282),
    [anon_sym_restrict] = ACTIONS(1282),
    [anon_sym___restrict__] = ACTIONS(1282),
    [anon_sym__Atomic] = ACTIONS(1282),
    [anon_sym__Noreturn] = ACTIONS(1282),
    [anon_sym_noreturn] = ACTIONS(1282),
    [anon_sym__Nonnull] = ACTIONS(1282),
    [anon_sym_alignas] = ACTIONS(1282),
    [anon_sym__Alignas] = ACTIONS(1282),
    [sym_primitive_type] = ACTIONS(1282),
    [anon_sym_enum] = ACTIONS(1282),
    [anon_sym_struct] = ACTIONS(1282),
    [anon_sym_union] = ACTIONS(1282),
    [anon_sym_if] = ACTIONS(1282),
    [anon_sym_switch] = ACTIONS(1282),
    [anon_sym_case] = ACTIONS(1282),
    [anon_sym_default] = ACTIONS(1282),
    [anon_sym_while] = ACTIONS(1282),
    [anon_sym_do] = ACTIONS(1282),
    [anon_sym_for] = ACTIONS(1282),
    [anon_sym_return] = ACTIONS(1282),
    [anon_sym_break] = ACTIONS(1282),
    [anon_sym_continue] = ACTIONS(1282),
    [anon_sym_goto] = ACTIONS(1282),
    [anon_sym___try] = ACTIONS(1282),
    [anon_sym___leave] = ACTIONS(1282),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_sizeof] = ACTIONS(1282),
    [anon_sym___alignof__] = ACTIONS(1282),
    [anon_sym___alignof] = ACTIONS(1282),
    [anon_sym__alignof] = ACTIONS(1282),
    [anon_sym_alignof] = ACTIONS(1282),
    [anon_sym__Alignof] = ACTIONS(1282),
    [anon_sym_offsetof] = ACTIONS(1282),
    [anon_sym__Generic] = ACTIONS(1282),
    [anon_sym_asm] = ACTIONS(1282),
    [anon_sym___asm__] = ACTIONS(1282),
    [anon_sym___asm] = ACTIONS(1282),
    [sym_number_literal] = ACTIONS(1284),
    [anon_sym_L_SQUOTE] = ACTIONS(1284),
    [anon_sym_u_SQUOTE] = ACTIONS(1284),
    [anon_sym_U_SQUOTE] = ACTIONS(1284),
    [anon_sym_u8_SQUOTE] = ACTIONS(1284),
    [anon_sym_SQUOTE] = ACTIONS(1284),
    [anon_sym_L_DQUOTE] = ACTIONS(1284),
    [anon_sym_u_DQUOTE] = ACTIONS(1284),
    [anon_sym_U_DQUOTE] = ACTIONS(1284),
    [anon_sym_u8_DQUOTE] = ACTIONS(1284),
    [anon_sym_DQUOTE] = ACTIONS(1284),
    [sym_true] = ACTIONS(1282),
    [sym_false] = ACTIONS(1282),
    [anon_sym_NULL] = ACTIONS(1282),
    [anon_sym_nullptr] = ACTIONS(1282),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(289)] = {
    [sym_identifier] = ACTIONS(1278),
    [aux_sym_preproc_include_token1] = ACTIONS(1278),
    [aux_sym_preproc_def_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token2] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1278),
    [sym_preproc_directive] = ACTIONS(1278),
    [anon_sym_LPAREN2] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1278),
    [anon_sym_PLUS] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1280),
    [anon_sym___extension__] = ACTIONS(1278),
    [anon_sym_typedef] = ACTIONS(1278),
    [anon_sym_extern] = ACTIONS(1278),
    [anon_sym___attribute__] = ACTIONS(1278),
    [anon_sym___attribute] = ACTIONS(1278),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1280),
    [anon_sym___declspec] = ACTIONS(1278),
    [anon_sym___cdecl] = ACTIONS(1278),
    [anon_sym___clrcall] = ACTIONS(1278),
    [anon_sym___stdcall] = ACTIONS(1278),
    [anon_sym___fastcall] = ACTIONS(1278),
    [anon_sym___thiscall] = ACTIONS(1278),
    [anon_sym___vectorcall] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_signed] = ACTIONS(1278),
    [anon_sym_unsigned] = ACTIONS(1278),
    [anon_sym_long] = ACTIONS(1278),
    [anon_sym_short] = ACTIONS(1278),
    [anon_sym_static] = ACTIONS(1278),
    [anon_sym_auto] = ACTIONS(1278),
    [anon_sym_register] = ACTIONS(1278),
    [anon_sym_inline] = ACTIONS(1278),
    [anon_sym___inline] = ACTIONS(1278),
    [anon_sym___inline__] = ACTIONS(1278),
    [anon_sym___forceinline] = ACTIONS(1278),
    [anon_sym_thread_local] = ACTIONS(1278),
    [anon_sym___thread] = ACTIONS(1278),
    [anon_sym_const] = ACTIONS(1278),
    [anon_sym_constexpr] = ACTIONS(1278),
    [anon_sym_volatile] = ACTIONS(1278),
    [anon_sym_restrict] = ACTIONS(1278),
    [anon_sym___restrict__] = ACTIONS(1278),
    [anon_sym__Atomic] = ACTIONS(1278),
    [anon_sym__Noreturn] = ACTIONS(1278),
    [anon_sym_noreturn] = ACTIONS(1278),
    [anon_sym__Nonnull] = ACTIONS(1278),
    [anon_sym_alignas] = ACTIONS(1278),
    [anon_sym__Alignas] = ACTIONS(1278),
    [sym_primitive_type] = ACTIONS(1278),
    [anon_sym_enum] = ACTIONS(1278),
    [anon_sym_struct] = ACTIONS(1278),
    [anon_sym_union] = ACTIONS(1278),
    [anon_sym_if] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1278),
    [anon_sym_case] = ACTIONS(1278),
    [anon_sym_default] = ACTIONS(1278),
    [anon_sym_while] = ACTIONS(1278),
    [anon_sym_do] = ACTIONS(1278),
    [anon_sym_for] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1278),
    [anon_sym_break] = ACTIONS(1278),
    [anon_sym_continue] = ACTIONS(1278),
    [anon_sym_goto] = ACTIONS(1278),
    [anon_sym___try] = ACTIONS(1278),
    [anon_sym___leave] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1280),
    [anon_sym_sizeof] = ACTIONS(1278),
    [anon_sym___alignof__] = ACTIONS(1278),
    [anon_sym___alignof] = ACTIONS(1278),
    [anon_sym__alignof] = ACTIONS(1278),
    [anon_sym_alignof] = ACTIONS(1278),
    [anon_sym__Alignof] = ACTIONS(1278),
    [anon_sym_offsetof] = ACTIONS(1278),
    [anon_sym__Generic] = ACTIONS(1278),
    [anon_sym_asm] = ACTIONS(1278),
    [anon_sym___asm__] = ACTIONS(1278),
    [anon_sym___asm] = ACTIONS(1278),
    [sym_number_literal] = ACTIONS(1280),
    [anon_sym_L_SQUOTE] = ACTIONS(1280),
    [anon_sym_u_SQUOTE] = ACTIONS(1280),
    [anon_sym_U_SQUOTE] = ACTIONS(1280),
    [anon_sym_u8_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_L_DQUOTE] = ACTIONS(1280),
    [anon_sym_u_DQUOTE] = ACTIONS(1280),
    [anon_sym_U_DQUOTE] = ACTIONS(1280),
    [anon_sym_u8_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [sym_true] = ACTIONS(1278),
    [sym_false] = ACTIONS(1278),
    [anon_sym_NULL] = ACTIONS(1278),
    [anon_sym_nullptr] = ACTIONS(1278),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(290)] = {
    [sym_identifier] = ACTIONS(1298),
    [aux_sym_preproc_include_token1] = ACTIONS(1298),
    [aux_sym_preproc_def_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token2] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1298),
    [sym_preproc_directive] = ACTIONS(1298),
    [anon_sym_LPAREN2] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1298),
    [anon_sym_PLUS] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym___extension__] = ACTIONS(1298),
    [anon_sym_typedef] = ACTIONS(1298),
    [anon_sym_extern] = ACTIONS(1298),
    [anon_sym___attribute__] = ACTIONS(1298),
    [anon_sym___attribute] = ACTIONS(1298),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1300),
    [anon_sym___declspec] = ACTIONS(1298),
    [anon_sym___cdecl] = ACTIONS(1298),
    [anon_sym___clrcall] = ACTIONS(1298),
    [anon_sym___stdcall] = ACTIONS(1298),
    [anon_sym___fastcall] = ACTIONS(1298),
    [anon_sym___thiscall] = ACTIONS(1298),
    [anon_sym___vectorcall] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_signed] = ACTIONS(1298),
    [anon_sym_unsigned] = ACTIONS(1298),
    [anon_sym_long] = ACTIONS(1298),
    [anon_sym_short] = ACTIONS(1298),
    [anon_sym_static] = ACTIONS(1298),
    [anon_sym_auto] = ACTIONS(1298),
    [anon_sym_register] = ACTIONS(1298),
    [anon_sym_inline] = ACTIONS(1298),
    [anon_sym___inline] = ACTIONS(1298),
    [anon_sym___inline__] = ACTIONS(1298),
    [anon_sym___forceinline] = ACTIONS(1298),
    [anon_sym_thread_local] = ACTIONS(1298),
    [anon_sym___thread] = ACTIONS(1298),
    [anon_sym_const] = ACTIONS(1298),
    [anon_sym_constexpr] = ACTIONS(1298),
    [anon_sym_volatile] = ACTIONS(1298),
    [anon_sym_restrict] = ACTIONS(1298),
    [anon_sym___restrict__] = ACTIONS(1298),
    [anon_sym__Atomic] = ACTIONS(1298),
    [anon_sym__Noreturn] = ACTIONS(1298),
    [anon_sym_noreturn] = ACTIONS(1298),
    [anon_sym__Nonnull] = ACTIONS(1298),
    [anon_sym_alignas] = ACTIONS(1298),
    [anon_sym__Alignas] = ACTIONS(1298),
    [sym_primitive_type] = ACTIONS(1298),
    [anon_sym_enum] = ACTIONS(1298),
    [anon_sym_struct] = ACTIONS(1298),
    [anon_sym_union] = ACTIONS(1298),
    [anon_sym_if] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1298),
    [anon_sym_case] = ACTIONS(1298),
    [anon_sym_default] = ACTIONS(1298),
    [anon_sym_while] = ACTIONS(1298),
    [anon_sym_do] = ACTIONS(1298),
    [anon_sym_for] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1298),
    [anon_sym_break] = ACTIONS(1298),
    [anon_sym_continue] = ACTIONS(1298),
    [anon_sym_goto] = ACTIONS(1298),
    [anon_sym___try] = ACTIONS(1298),
    [anon_sym___leave] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_sizeof] = ACTIONS(1298),
    [anon_sym___alignof__] = ACTIONS(1298),
    [anon_sym___alignof] = ACTIONS(1298),
    [anon_sym__alignof] = ACTIONS(1298),
    [anon_sym_alignof] = ACTIONS(1298),
    [anon_sym__Alignof] = ACTIONS(1298),
    [anon_sym_offsetof] = ACTIONS(1298),
    [anon_sym__Generic] = ACTIONS(1298),
    [anon_sym_asm] = ACTIONS(1298),
    [anon_sym___asm__] = ACTIONS(1298),
    [anon_sym___asm] = ACTIONS(1298),
    [sym_number_literal] = ACTIONS(1300),
    [anon_sym_L_SQUOTE] = ACTIONS(1300),
    [anon_sym_u_SQUOTE] = ACTIONS(1300),
    [anon_sym_U_SQUOTE] = ACTIONS(1300),
    [anon_sym_u8_SQUOTE] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_L_DQUOTE] = ACTIONS(1300),
    [anon_sym_u_DQUOTE] = ACTIONS(1300),
    [anon_sym_U_DQUOTE] = ACTIONS(1300),
    [anon_sym_u8_DQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [sym_true] = ACTIONS(1298),
    [sym_false] = ACTIONS(1298),
    [anon_sym_NULL] = ACTIONS(1298),
    [anon_sym_nullptr] = ACTIONS(1298),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(291)] = {
    [sym_identifier] = ACTIONS(1294),
    [aux_sym_preproc_include_token1] = ACTIONS(1294),
    [aux_sym_preproc_def_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token2] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1294),
    [sym_preproc_directive] = ACTIONS(1294),
    [anon_sym_LPAREN2] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym___extension__] = ACTIONS(1294),
    [anon_sym_typedef] = ACTIONS(1294),
    [anon_sym_extern] = ACTIONS(1294),
    [anon_sym___attribute__] = ACTIONS(1294),
    [anon_sym___attribute] = ACTIONS(1294),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1296),
    [anon_sym___declspec] = ACTIONS(1294),
    [anon_sym___cdecl] = ACTIONS(1294),
    [anon_sym___clrcall] = ACTIONS(1294),
    [anon_sym___stdcall] = ACTIONS(1294),
    [anon_sym___fastcall] = ACTIONS(1294),
    [anon_sym___thiscall] = ACTIONS(1294),
    [anon_sym___vectorcall] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_signed] = ACTIONS(1294),
    [anon_sym_unsigned] = ACTIONS(1294),
    [anon_sym_long] = ACTIONS(1294),
    [anon_sym_short] = ACTIONS(1294),
    [anon_sym_static] = ACTIONS(1294),
    [anon_sym_auto] = ACTIONS(1294),
    [anon_sym_register] = ACTIONS(1294),
    [anon_sym_inline] = ACTIONS(1294),
    [anon_sym___inline] = ACTIONS(1294),
    [anon_sym___inline__] = ACTIONS(1294),
    [anon_sym___forceinline] = ACTIONS(1294),
    [anon_sym_thread_local] = ACTIONS(1294),
    [anon_sym___thread] = ACTIONS(1294),
    [anon_sym_const] = ACTIONS(1294),
    [anon_sym_constexpr] = ACTIONS(1294),
    [anon_sym_volatile] = ACTIONS(1294),
    [anon_sym_restrict] = ACTIONS(1294),
    [anon_sym___restrict__] = ACTIONS(1294),
    [anon_sym__Atomic] = ACTIONS(1294),
    [anon_sym__Noreturn] = ACTIONS(1294),
    [anon_sym_noreturn] = ACTIONS(1294),
    [anon_sym__Nonnull] = ACTIONS(1294),
    [anon_sym_alignas] = ACTIONS(1294),
    [anon_sym__Alignas] = ACTIONS(1294),
    [sym_primitive_type] = ACTIONS(1294),
    [anon_sym_enum] = ACTIONS(1294),
    [anon_sym_struct] = ACTIONS(1294),
    [anon_sym_union] = ACTIONS(1294),
    [anon_sym_if] = ACTIONS(1294),
    [anon_sym_switch] = ACTIONS(1294),
    [anon_sym_case] = ACTIONS(1294),
    [anon_sym_default] = ACTIONS(1294),
    [anon_sym_while] = ACTIONS(1294),
    [anon_sym_do] = ACTIONS(1294),
    [anon_sym_for] = ACTIONS(1294),
    [anon_sym_return] = ACTIONS(1294),
    [anon_sym_break] = ACTIONS(1294),
    [anon_sym_continue] = ACTIONS(1294),
    [anon_sym_goto] = ACTIONS(1294),
    [anon_sym___try] = ACTIONS(1294),
    [anon_sym___leave] = ACTIONS(1294),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_sizeof] = ACTIONS(1294),
    [anon_sym___alignof__] = ACTIONS(1294),
    [anon_sym___alignof] = ACTIONS(1294),
    [anon_sym__alignof] = ACTIONS(1294),
    [anon_sym_alignof] = ACTIONS(1294),
    [anon_sym__Alignof] = ACTIONS(1294),
    [anon_sym_offsetof] = ACTIONS(1294),
    [anon_sym__Generic] = ACTIONS(1294),
    [anon_sym_asm] = ACTIONS(1294),
    [anon_sym___asm__] = ACTIONS(1294),
    [anon_sym___asm] = ACTIONS(1294),
    [sym_number_literal] = ACTIONS(1296),
    [anon_sym_L_SQUOTE] = ACTIONS(1296),
    [anon_sym_u_SQUOTE] = ACTIONS(1296),
    [anon_sym_U_SQUOTE] = ACTIONS(1296),
    [anon_sym_u8_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_L_DQUOTE] = ACTIONS(1296),
    [anon_sym_u_DQUOTE] = ACTIONS(1296),
    [anon_sym_U_DQUOTE] = ACTIONS(1296),
    [anon_sym_u8_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [sym_true] = ACTIONS(1294),
    [sym_false] = ACTIONS(1294),
    [anon_sym_NULL] = ACTIONS(1294),
    [anon_sym_nullptr] = ACTIONS(1294),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(292)] = {
    [sym_identifier] = ACTIONS(1262),
    [aux_sym_preproc_include_token1] = ACTIONS(1262),
    [aux_sym_preproc_def_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token2] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1262),
    [sym_preproc_directive] = ACTIONS(1262),
    [anon_sym_LPAREN2] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_TILDE] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_STAR] = ACTIONS(1264),
    [anon_sym_AMP] = ACTIONS(1264),
    [anon_sym_SEMI] = ACTIONS(1264),
    [anon_sym___extension__] = ACTIONS(1262),
    [anon_sym_typedef] = ACTIONS(1262),
    [anon_sym_extern] = ACTIONS(1262),
    [anon_sym___attribute__] = ACTIONS(1262),
    [anon_sym___attribute] = ACTIONS(1262),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1264),
    [anon_sym___declspec] = ACTIONS(1262),
    [anon_sym___cdecl] = ACTIONS(1262),
    [anon_sym___clrcall] = ACTIONS(1262),
    [anon_sym___stdcall] = ACTIONS(1262),
    [anon_sym___fastcall] = ACTIONS(1262),
    [anon_sym___thiscall] = ACTIONS(1262),
    [anon_sym___vectorcall] = ACTIONS(1262),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_signed] = ACTIONS(1262),
    [anon_sym_unsigned] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_auto] = ACTIONS(1262),
    [anon_sym_register] = ACTIONS(1262),
    [anon_sym_inline] = ACTIONS(1262),
    [anon_sym___inline] = ACTIONS(1262),
    [anon_sym___inline__] = ACTIONS(1262),
    [anon_sym___forceinline] = ACTIONS(1262),
    [anon_sym_thread_local] = ACTIONS(1262),
    [anon_sym___thread] = ACTIONS(1262),
    [anon_sym_const] = ACTIONS(1262),
    [anon_sym_constexpr] = ACTIONS(1262),
    [anon_sym_volatile] = ACTIONS(1262),
    [anon_sym_restrict] = ACTIONS(1262),
    [anon_sym___restrict__] = ACTIONS(1262),
    [anon_sym__Atomic] = ACTIONS(1262),
    [anon_sym__Noreturn] = ACTIONS(1262),
    [anon_sym_noreturn] = ACTIONS(1262),
    [anon_sym__Nonnull] = ACTIONS(1262),
    [anon_sym_alignas] = ACTIONS(1262),
    [anon_sym__Alignas] = ACTIONS(1262),
    [sym_primitive_type] = ACTIONS(1262),
    [anon_sym_enum] = ACTIONS(1262),
    [anon_sym_struct] = ACTIONS(1262),
    [anon_sym_union] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_goto] = ACTIONS(1262),
    [anon_sym___try] = ACTIONS(1262),
    [anon_sym___leave] = ACTIONS(1262),
    [anon_sym_DASH_DASH] = ACTIONS(1264),
    [anon_sym_PLUS_PLUS] = ACTIONS(1264),
    [anon_sym_sizeof] = ACTIONS(1262),
    [anon_sym___alignof__] = ACTIONS(1262),
    [anon_sym___alignof] = ACTIONS(1262),
    [anon_sym__alignof] = ACTIONS(1262),
    [anon_sym_alignof] = ACTIONS(1262),
    [anon_sym__Alignof] = ACTIONS(1262),
    [anon_sym_offsetof] = ACTIONS(1262),
    [anon_sym__Generic] = ACTIONS(1262),
    [anon_sym_asm] = ACTIONS(1262),
    [anon_sym___asm__] = ACTIONS(1262),
    [anon_sym___asm] = ACTIONS(1262),
    [sym_number_literal] = ACTIONS(1264),
    [anon_sym_L_SQUOTE] = ACTIONS(1264),
    [anon_sym_u_SQUOTE] = ACTIONS(1264),
    [anon_sym_U_SQUOTE] = ACTIONS(1264),
    [anon_sym_u8_SQUOTE] = ACTIONS(1264),
    [anon_sym_SQUOTE] = ACTIONS(1264),
    [anon_sym_L_DQUOTE] = ACTIONS(1264),
    [anon_sym_u_DQUOTE] = ACTIONS(1264),
    [anon_sym_U_DQUOTE] = ACTIONS(1264),
    [anon_sym_u8_DQUOTE] = ACTIONS(1264),
    [anon_sym_DQUOTE] = ACTIONS(1264),
    [sym_true] = ACTIONS(1262),
    [sym_false] = ACTIONS(1262),
    [anon_sym_NULL] = ACTIONS(1262),
    [anon_sym_nullptr] = ACTIONS(1262),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(293)] = {
    [sym_identifier] = ACTIONS(1302),
    [aux_sym_preproc_include_token1] = ACTIONS(1302),
    [aux_sym_preproc_def_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token2] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1302),
    [sym_preproc_directive] = ACTIONS(1302),
    [anon_sym_LPAREN2] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym___extension__] = ACTIONS(1302),
    [anon_sym_typedef] = ACTIONS(1302),
    [anon_sym_extern] = ACTIONS(1302),
    [anon_sym___attribute__] = ACTIONS(1302),
    [anon_sym___attribute] = ACTIONS(1302),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1304),
    [anon_sym___declspec] = ACTIONS(1302),
    [anon_sym___cdecl] = ACTIONS(1302),
    [anon_sym___clrcall] = ACTIONS(1302),
    [anon_sym___stdcall] = ACTIONS(1302),
    [anon_sym___fastcall] = ACTIONS(1302),
    [anon_sym___thiscall] = ACTIONS(1302),
    [anon_sym___vectorcall] = ACTIONS(1302),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_signed] = ACTIONS(1302),
    [anon_sym_unsigned] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_auto] = ACTIONS(1302),
    [anon_sym_register] = ACTIONS(1302),
    [anon_sym_inline] = ACTIONS(1302),
    [anon_sym___inline] = ACTIONS(1302),
    [anon_sym___inline__] = ACTIONS(1302),
    [anon_sym___forceinline] = ACTIONS(1302),
    [anon_sym_thread_local] = ACTIONS(1302),
    [anon_sym___thread] = ACTIONS(1302),
    [anon_sym_const] = ACTIONS(1302),
    [anon_sym_constexpr] = ACTIONS(1302),
    [anon_sym_volatile] = ACTIONS(1302),
    [anon_sym_restrict] = ACTIONS(1302),
    [anon_sym___restrict__] = ACTIONS(1302),
    [anon_sym__Atomic] = ACTIONS(1302),
    [anon_sym__Noreturn] = ACTIONS(1302),
    [anon_sym_noreturn] = ACTIONS(1302),
    [anon_sym__Nonnull] = ACTIONS(1302),
    [anon_sym_alignas] = ACTIONS(1302),
    [anon_sym__Alignas] = ACTIONS(1302),
    [sym_primitive_type] = ACTIONS(1302),
    [anon_sym_enum] = ACTIONS(1302),
    [anon_sym_struct] = ACTIONS(1302),
    [anon_sym_union] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_goto] = ACTIONS(1302),
    [anon_sym___try] = ACTIONS(1302),
    [anon_sym___leave] = ACTIONS(1302),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_sizeof] = ACTIONS(1302),
    [anon_sym___alignof__] = ACTIONS(1302),
    [anon_sym___alignof] = ACTIONS(1302),
    [anon_sym__alignof] = ACTIONS(1302),
    [anon_sym_alignof] = ACTIONS(1302),
    [anon_sym__Alignof] = ACTIONS(1302),
    [anon_sym_offsetof] = ACTIONS(1302),
    [anon_sym__Generic] = ACTIONS(1302),
    [anon_sym_asm] = ACTIONS(1302),
    [anon_sym___asm__] = ACTIONS(1302),
    [anon_sym___asm] = ACTIONS(1302),
    [sym_number_literal] = ACTIONS(1304),
    [anon_sym_L_SQUOTE] = ACTIONS(1304),
    [anon_sym_u_SQUOTE] = ACTIONS(1304),
    [anon_sym_U_SQUOTE] = ACTIONS(1304),
    [anon_sym_u8_SQUOTE] = ACTIONS(1304),
    [anon_sym_SQUOTE] = ACTIONS(1304),
    [anon_sym_L_DQUOTE] = ACTIONS(1304),
    [anon_sym_u_DQUOTE] = ACTIONS(1304),
    [anon_sym_U_DQUOTE] = ACTIONS(1304),
    [anon_sym_u8_DQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [sym_true] = ACTIONS(1302),
    [sym_false] = ACTIONS(1302),
    [anon_sym_NULL] = ACTIONS(1302),
    [anon_sym_nullptr] = ACTIONS(1302),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(294)] = {
    [sym_identifier] = ACTIONS(1306),
    [aux_sym_preproc_include_token1] = ACTIONS(1306),
    [aux_sym_preproc_def_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token2] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1306),
    [sym_preproc_directive] = ACTIONS(1306),
    [anon_sym_LPAREN2] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym___extension__] = ACTIONS(1306),
    [anon_sym_typedef] = ACTIONS(1306),
    [anon_sym_extern] = ACTIONS(1306),
    [anon_sym___attribute__] = ACTIONS(1306),
    [anon_sym___attribute] = ACTIONS(1306),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1308),
    [anon_sym___declspec] = ACTIONS(1306),
    [anon_sym___cdecl] = ACTIONS(1306),
    [anon_sym___clrcall] = ACTIONS(1306),
    [anon_sym___stdcall] = ACTIONS(1306),
    [anon_sym___fastcall] = ACTIONS(1306),
    [anon_sym___thiscall] = ACTIONS(1306),
    [anon_sym___vectorcall] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_signed] = ACTIONS(1306),
    [anon_sym_unsigned] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_auto] = ACTIONS(1306),
    [anon_sym_register] = ACTIONS(1306),
    [anon_sym_inline] = ACTIONS(1306),
    [anon_sym___inline] = ACTIONS(1306),
    [anon_sym___inline__] = ACTIONS(1306),
    [anon_sym___forceinline] = ACTIONS(1306),
    [anon_sym_thread_local] = ACTIONS(1306),
    [anon_sym___thread] = ACTIONS(1306),
    [anon_sym_const] = ACTIONS(1306),
    [anon_sym_constexpr] = ACTIONS(1306),
    [anon_sym_volatile] = ACTIONS(1306),
    [anon_sym_restrict] = ACTIONS(1306),
    [anon_sym___restrict__] = ACTIONS(1306),
    [anon_sym__Atomic] = ACTIONS(1306),
    [anon_sym__Noreturn] = ACTIONS(1306),
    [anon_sym_noreturn] = ACTIONS(1306),
    [anon_sym__Nonnull] = ACTIONS(1306),
    [anon_sym_alignas] = ACTIONS(1306),
    [anon_sym__Alignas] = ACTIONS(1306),
    [sym_primitive_type] = ACTIONS(1306),
    [anon_sym_enum] = ACTIONS(1306),
    [anon_sym_struct] = ACTIONS(1306),
    [anon_sym_union] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_goto] = ACTIONS(1306),
    [anon_sym___try] = ACTIONS(1306),
    [anon_sym___leave] = ACTIONS(1306),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_sizeof] = ACTIONS(1306),
    [anon_sym___alignof__] = ACTIONS(1306),
    [anon_sym___alignof] = ACTIONS(1306),
    [anon_sym__alignof] = ACTIONS(1306),
    [anon_sym_alignof] = ACTIONS(1306),
    [anon_sym__Alignof] = ACTIONS(1306),
    [anon_sym_offsetof] = ACTIONS(1306),
    [anon_sym__Generic] = ACTIONS(1306),
    [anon_sym_asm] = ACTIONS(1306),
    [anon_sym___asm__] = ACTIONS(1306),
    [anon_sym___asm] = ACTIONS(1306),
    [sym_number_literal] = ACTIONS(1308),
    [anon_sym_L_SQUOTE] = ACTIONS(1308),
    [anon_sym_u_SQUOTE] = ACTIONS(1308),
    [anon_sym_U_SQUOTE] = ACTIONS(1308),
    [anon_sym_u8_SQUOTE] = ACTIONS(1308),
    [anon_sym_SQUOTE] = ACTIONS(1308),
    [anon_sym_L_DQUOTE] = ACTIONS(1308),
    [anon_sym_u_DQUOTE] = ACTIONS(1308),
    [anon_sym_U_DQUOTE] = ACTIONS(1308),
    [anon_sym_u8_DQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_true] = ACTIONS(1306),
    [sym_false] = ACTIONS(1306),
    [anon_sym_NULL] = ACTIONS(1306),
    [anon_sym_nullptr] = ACTIONS(1306),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(295)] = {
    [sym_identifier] = ACTIONS(1314),
    [aux_sym_preproc_include_token1] = ACTIONS(1314),
    [aux_sym_preproc_def_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token2] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1314),
    [sym_preproc_directive] = ACTIONS(1314),
    [anon_sym_LPAREN2] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1314),
    [anon_sym_PLUS] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym___extension__] = ACTIONS(1314),
    [anon_sym_typedef] = ACTIONS(1314),
    [anon_sym_extern] = ACTIONS(1314),
    [anon_sym___attribute__] = ACTIONS(1314),
    [anon_sym___attribute] = ACTIONS(1314),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1316),
    [anon_sym___declspec] = ACTIONS(1314),
    [anon_sym___cdecl] = ACTIONS(1314),
    [anon_sym___clrcall] = ACTIONS(1314),
    [anon_sym___stdcall] = ACTIONS(1314),
    [anon_sym___fastcall] = ACTIONS(1314),
    [anon_sym___thiscall] = ACTIONS(1314),
    [anon_sym___vectorcall] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_signed] = ACTIONS(1314),
    [anon_sym_unsigned] = ACTIONS(1314),
    [anon_sym_long] = ACTIONS(1314),
    [anon_sym_short] = ACTIONS(1314),
    [anon_sym_static] = ACTIONS(1314),
    [anon_sym_auto] = ACTIONS(1314),
    [anon_sym_register] = ACTIONS(1314),
    [anon_sym_inline] = ACTIONS(1314),
    [anon_sym___inline] = ACTIONS(1314),
    [anon_sym___inline__] = ACTIONS(1314),
    [anon_sym___forceinline] = ACTIONS(1314),
    [anon_sym_thread_local] = ACTIONS(1314),
    [anon_sym___thread] = ACTIONS(1314),
    [anon_sym_const] = ACTIONS(1314),
    [anon_sym_constexpr] = ACTIONS(1314),
    [anon_sym_volatile] = ACTIONS(1314),
    [anon_sym_restrict] = ACTIONS(1314),
    [anon_sym___restrict__] = ACTIONS(1314),
    [anon_sym__Atomic] = ACTIONS(1314),
    [anon_sym__Noreturn] = ACTIONS(1314),
    [anon_sym_noreturn] = ACTIONS(1314),
    [anon_sym__Nonnull] = ACTIONS(1314),
    [anon_sym_alignas] = ACTIONS(1314),
    [anon_sym__Alignas] = ACTIONS(1314),
    [sym_primitive_type] = ACTIONS(1314),
    [anon_sym_enum] = ACTIONS(1314),
    [anon_sym_struct] = ACTIONS(1314),
    [anon_sym_union] = ACTIONS(1314),
    [anon_sym_if] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1314),
    [anon_sym_case] = ACTIONS(1314),
    [anon_sym_default] = ACTIONS(1314),
    [anon_sym_while] = ACTIONS(1314),
    [anon_sym_do] = ACTIONS(1314),
    [anon_sym_for] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1314),
    [anon_sym_break] = ACTIONS(1314),
    [anon_sym_continue] = ACTIONS(1314),
    [anon_sym_goto] = ACTIONS(1314),
    [anon_sym___try] = ACTIONS(1314),
    [anon_sym___leave] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_sizeof] = ACTIONS(1314),
    [anon_sym___alignof__] = ACTIONS(1314),
    [anon_sym___alignof] = ACTIONS(1314),
    [anon_sym__alignof] = ACTIONS(1314),
    [anon_sym_alignof] = ACTIONS(1314),
    [anon_sym__Alignof] = ACTIONS(1314),
    [anon_sym_offsetof] = ACTIONS(1314),
    [anon_sym__Generic] = ACTIONS(1314),
    [anon_sym_asm] = ACTIONS(1314),
    [anon_sym___asm__] = ACTIONS(1314),
    [anon_sym___asm] = ACTIONS(1314),
    [sym_number_literal] = ACTIONS(1316),
    [anon_sym_L_SQUOTE] = ACTIONS(1316),
    [anon_sym_u_SQUOTE] = ACTIONS(1316),
    [anon_sym_U_SQUOTE] = ACTIONS(1316),
    [anon_sym_u8_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_L_DQUOTE] = ACTIONS(1316),
    [anon_sym_u_DQUOTE] = ACTIONS(1316),
    [anon_sym_U_DQUOTE] = ACTIONS(1316),
    [anon_sym_u8_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [sym_true] = ACTIONS(1314),
    [sym_false] = ACTIONS(1314),
    [anon_sym_NULL] = ACTIONS(1314),
    [anon_sym_nullptr] = ACTIONS(1314),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(296)] = {
    [sym_identifier] = ACTIONS(1322),
    [aux_sym_preproc_include_token1] = ACTIONS(1322),
    [aux_sym_preproc_def_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token2] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1322),
    [sym_preproc_directive] = ACTIONS(1322),
    [anon_sym_LPAREN2] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1322),
    [anon_sym_PLUS] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym___extension__] = ACTIONS(1322),
    [anon_sym_typedef] = ACTIONS(1322),
    [anon_sym_extern] = ACTIONS(1322),
    [anon_sym___attribute__] = ACTIONS(1322),
    [anon_sym___attribute] = ACTIONS(1322),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1324),
    [anon_sym___declspec] = ACTIONS(1322),
    [anon_sym___cdecl] = ACTIONS(1322),
    [anon_sym___clrcall] = ACTIONS(1322),
    [anon_sym___stdcall] = ACTIONS(1322),
    [anon_sym___fastcall] = ACTIONS(1322),
    [anon_sym___thiscall] = ACTIONS(1322),
    [anon_sym___vectorcall] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_signed] = ACTIONS(1322),
    [anon_sym_unsigned] = ACTIONS(1322),
    [anon_sym_long] = ACTIONS(1322),
    [anon_sym_short] = ACTIONS(1322),
    [anon_sym_static] = ACTIONS(1322),
    [anon_sym_auto] = ACTIONS(1322),
    [anon_sym_register] = ACTIONS(1322),
    [anon_sym_inline] = ACTIONS(1322),
    [anon_sym___inline] = ACTIONS(1322),
    [anon_sym___inline__] = ACTIONS(1322),
    [anon_sym___forceinline] = ACTIONS(1322),
    [anon_sym_thread_local] = ACTIONS(1322),
    [anon_sym___thread] = ACTIONS(1322),
    [anon_sym_const] = ACTIONS(1322),
    [anon_sym_constexpr] = ACTIONS(1322),
    [anon_sym_volatile] = ACTIONS(1322),
    [anon_sym_restrict] = ACTIONS(1322),
    [anon_sym___restrict__] = ACTIONS(1322),
    [anon_sym__Atomic] = ACTIONS(1322),
    [anon_sym__Noreturn] = ACTIONS(1322),
    [anon_sym_noreturn] = ACTIONS(1322),
    [anon_sym__Nonnull] = ACTIONS(1322),
    [anon_sym_alignas] = ACTIONS(1322),
    [anon_sym__Alignas] = ACTIONS(1322),
    [sym_primitive_type] = ACTIONS(1322),
    [anon_sym_enum] = ACTIONS(1322),
    [anon_sym_struct] = ACTIONS(1322),
    [anon_sym_union] = ACTIONS(1322),
    [anon_sym_if] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1322),
    [anon_sym_case] = ACTIONS(1322),
    [anon_sym_default] = ACTIONS(1322),
    [anon_sym_while] = ACTIONS(1322),
    [anon_sym_do] = ACTIONS(1322),
    [anon_sym_for] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1322),
    [anon_sym_break] = ACTIONS(1322),
    [anon_sym_continue] = ACTIONS(1322),
    [anon_sym_goto] = ACTIONS(1322),
    [anon_sym___try] = ACTIONS(1322),
    [anon_sym___leave] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_sizeof] = ACTIONS(1322),
    [anon_sym___alignof__] = ACTIONS(1322),
    [anon_sym___alignof] = ACTIONS(1322),
    [anon_sym__alignof] = ACTIONS(1322),
    [anon_sym_alignof] = ACTIONS(1322),
    [anon_sym__Alignof] = ACTIONS(1322),
    [anon_sym_offsetof] = ACTIONS(1322),
    [anon_sym__Generic] = ACTIONS(1322),
    [anon_sym_asm] = ACTIONS(1322),
    [anon_sym___asm__] = ACTIONS(1322),
    [anon_sym___asm] = ACTIONS(1322),
    [sym_number_literal] = ACTIONS(1324),
    [anon_sym_L_SQUOTE] = ACTIONS(1324),
    [anon_sym_u_SQUOTE] = ACTIONS(1324),
    [anon_sym_U_SQUOTE] = ACTIONS(1324),
    [anon_sym_u8_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_L_DQUOTE] = ACTIONS(1324),
    [anon_sym_u_DQUOTE] = ACTIONS(1324),
    [anon_sym_U_DQUOTE] = ACTIONS(1324),
    [anon_sym_u8_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [sym_true] = ACTIONS(1322),
    [sym_false] = ACTIONS(1322),
    [anon_sym_NULL] = ACTIONS(1322),
    [anon_sym_nullptr] = ACTIONS(1322),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(297)] = {
    [sym_identifier] = ACTIONS(1326),
    [aux_sym_preproc_include_token1] = ACTIONS(1326),
    [aux_sym_preproc_def_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token2] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1326),
    [sym_preproc_directive] = ACTIONS(1326),
    [anon_sym_LPAREN2] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1326),
    [anon_sym_PLUS] = ACTIONS(1326),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym___extension__] = ACTIONS(1326),
    [anon_sym_typedef] = ACTIONS(1326),
    [anon_sym_extern] = ACTIONS(1326),
    [anon_sym___attribute__] = ACTIONS(1326),
    [anon_sym___attribute] = ACTIONS(1326),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1328),
    [anon_sym___declspec] = ACTIONS(1326),
    [anon_sym___cdecl] = ACTIONS(1326),
    [anon_sym___clrcall] = ACTIONS(1326),
    [anon_sym___stdcall] = ACTIONS(1326),
    [anon_sym___fastcall] = ACTIONS(1326),
    [anon_sym___thiscall] = ACTIONS(1326),
    [anon_sym___vectorcall] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_signed] = ACTIONS(1326),
    [anon_sym_unsigned] = ACTIONS(1326),
    [anon_sym_long] = ACTIONS(1326),
    [anon_sym_short] = ACTIONS(1326),
    [anon_sym_static] = ACTIONS(1326),
    [anon_sym_auto] = ACTIONS(1326),
    [anon_sym_register] = ACTIONS(1326),
    [anon_sym_inline] = ACTIONS(1326),
    [anon_sym___inline] = ACTIONS(1326),
    [anon_sym___inline__] = ACTIONS(1326),
    [anon_sym___forceinline] = ACTIONS(1326),
    [anon_sym_thread_local] = ACTIONS(1326),
    [anon_sym___thread] = ACTIONS(1326),
    [anon_sym_const] = ACTIONS(1326),
    [anon_sym_constexpr] = ACTIONS(1326),
    [anon_sym_volatile] = ACTIONS(1326),
    [anon_sym_restrict] = ACTIONS(1326),
    [anon_sym___restrict__] = ACTIONS(1326),
    [anon_sym__Atomic] = ACTIONS(1326),
    [anon_sym__Noreturn] = ACTIONS(1326),
    [anon_sym_noreturn] = ACTIONS(1326),
    [anon_sym__Nonnull] = ACTIONS(1326),
    [anon_sym_alignas] = ACTIONS(1326),
    [anon_sym__Alignas] = ACTIONS(1326),
    [sym_primitive_type] = ACTIONS(1326),
    [anon_sym_enum] = ACTIONS(1326),
    [anon_sym_struct] = ACTIONS(1326),
    [anon_sym_union] = ACTIONS(1326),
    [anon_sym_if] = ACTIONS(1326),
    [anon_sym_switch] = ACTIONS(1326),
    [anon_sym_case] = ACTIONS(1326),
    [anon_sym_default] = ACTIONS(1326),
    [anon_sym_while] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1326),
    [anon_sym_for] = ACTIONS(1326),
    [anon_sym_return] = ACTIONS(1326),
    [anon_sym_break] = ACTIONS(1326),
    [anon_sym_continue] = ACTIONS(1326),
    [anon_sym_goto] = ACTIONS(1326),
    [anon_sym___try] = ACTIONS(1326),
    [anon_sym___leave] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_sizeof] = ACTIONS(1326),
    [anon_sym___alignof__] = ACTIONS(1326),
    [anon_sym___alignof] = ACTIONS(1326),
    [anon_sym__alignof] = ACTIONS(1326),
    [anon_sym_alignof] = ACTIONS(1326),
    [anon_sym__Alignof] = ACTIONS(1326),
    [anon_sym_offsetof] = ACTIONS(1326),
    [anon_sym__Generic] = ACTIONS(1326),
    [anon_sym_asm] = ACTIONS(1326),
    [anon_sym___asm__] = ACTIONS(1326),
    [anon_sym___asm] = ACTIONS(1326),
    [sym_number_literal] = ACTIONS(1328),
    [anon_sym_L_SQUOTE] = ACTIONS(1328),
    [anon_sym_u_SQUOTE] = ACTIONS(1328),
    [anon_sym_U_SQUOTE] = ACTIONS(1328),
    [anon_sym_u8_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_L_DQUOTE] = ACTIONS(1328),
    [anon_sym_u_DQUOTE] = ACTIONS(1328),
    [anon_sym_U_DQUOTE] = ACTIONS(1328),
    [anon_sym_u8_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [sym_true] = ACTIONS(1326),
    [sym_false] = ACTIONS(1326),
    [anon_sym_NULL] = ACTIONS(1326),
    [anon_sym_nullptr] = ACTIONS(1326),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(298)] = {
    [sym_identifier] = ACTIONS(1330),
    [aux_sym_preproc_include_token1] = ACTIONS(1330),
    [aux_sym_preproc_def_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token2] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1330),
    [sym_preproc_directive] = ACTIONS(1330),
    [anon_sym_LPAREN2] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1330),
    [anon_sym_PLUS] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym___extension__] = ACTIONS(1330),
    [anon_sym_typedef] = ACTIONS(1330),
    [anon_sym_extern] = ACTIONS(1330),
    [anon_sym___attribute__] = ACTIONS(1330),
    [anon_sym___attribute] = ACTIONS(1330),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1332),
    [anon_sym___declspec] = ACTIONS(1330),
    [anon_sym___cdecl] = ACTIONS(1330),
    [anon_sym___clrcall] = ACTIONS(1330),
    [anon_sym___stdcall] = ACTIONS(1330),
    [anon_sym___fastcall] = ACTIONS(1330),
    [anon_sym___thiscall] = ACTIONS(1330),
    [anon_sym___vectorcall] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_signed] = ACTIONS(1330),
    [anon_sym_unsigned] = ACTIONS(1330),
    [anon_sym_long] = ACTIONS(1330),
    [anon_sym_short] = ACTIONS(1330),
    [anon_sym_static] = ACTIONS(1330),
    [anon_sym_auto] = ACTIONS(1330),
    [anon_sym_register] = ACTIONS(1330),
    [anon_sym_inline] = ACTIONS(1330),
    [anon_sym___inline] = ACTIONS(1330),
    [anon_sym___inline__] = ACTIONS(1330),
    [anon_sym___forceinline] = ACTIONS(1330),
    [anon_sym_thread_local] = ACTIONS(1330),
    [anon_sym___thread] = ACTIONS(1330),
    [anon_sym_const] = ACTIONS(1330),
    [anon_sym_constexpr] = ACTIONS(1330),
    [anon_sym_volatile] = ACTIONS(1330),
    [anon_sym_restrict] = ACTIONS(1330),
    [anon_sym___restrict__] = ACTIONS(1330),
    [anon_sym__Atomic] = ACTIONS(1330),
    [anon_sym__Noreturn] = ACTIONS(1330),
    [anon_sym_noreturn] = ACTIONS(1330),
    [anon_sym__Nonnull] = ACTIONS(1330),
    [anon_sym_alignas] = ACTIONS(1330),
    [anon_sym__Alignas] = ACTIONS(1330),
    [sym_primitive_type] = ACTIONS(1330),
    [anon_sym_enum] = ACTIONS(1330),
    [anon_sym_struct] = ACTIONS(1330),
    [anon_sym_union] = ACTIONS(1330),
    [anon_sym_if] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1330),
    [anon_sym_case] = ACTIONS(1330),
    [anon_sym_default] = ACTIONS(1330),
    [anon_sym_while] = ACTIONS(1330),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_for] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1330),
    [anon_sym_break] = ACTIONS(1330),
    [anon_sym_continue] = ACTIONS(1330),
    [anon_sym_goto] = ACTIONS(1330),
    [anon_sym___try] = ACTIONS(1330),
    [anon_sym___leave] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_sizeof] = ACTIONS(1330),
    [anon_sym___alignof__] = ACTIONS(1330),
    [anon_sym___alignof] = ACTIONS(1330),
    [anon_sym__alignof] = ACTIONS(1330),
    [anon_sym_alignof] = ACTIONS(1330),
    [anon_sym__Alignof] = ACTIONS(1330),
    [anon_sym_offsetof] = ACTIONS(1330),
    [anon_sym__Generic] = ACTIONS(1330),
    [anon_sym_asm] = ACTIONS(1330),
    [anon_sym___asm__] = ACTIONS(1330),
    [anon_sym___asm] = ACTIONS(1330),
    [sym_number_literal] = ACTIONS(1332),
    [anon_sym_L_SQUOTE] = ACTIONS(1332),
    [anon_sym_u_SQUOTE] = ACTIONS(1332),
    [anon_sym_U_SQUOTE] = ACTIONS(1332),
    [anon_sym_u8_SQUOTE] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_L_DQUOTE] = ACTIONS(1332),
    [anon_sym_u_DQUOTE] = ACTIONS(1332),
    [anon_sym_U_DQUOTE] = ACTIONS(1332),
    [anon_sym_u8_DQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [sym_true] = ACTIONS(1330),
    [sym_false] = ACTIONS(1330),
    [anon_sym_NULL] = ACTIONS(1330),
    [anon_sym_nullptr] = ACTIONS(1330),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(299)] = {
    [sym_identifier] = ACTIONS(1340),
    [aux_sym_preproc_include_token1] = ACTIONS(1340),
    [aux_sym_preproc_def_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token2] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1340),
    [sym_preproc_directive] = ACTIONS(1340),
    [anon_sym_LPAREN2] = ACTIONS(1342),
    [anon_sym_BANG] = ACTIONS(1342),
    [anon_sym_TILDE] = ACTIONS(1342),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1342),
    [anon_sym_AMP] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1342),
    [anon_sym___extension__] = ACTIONS(1340),
    [anon_sym_typedef] = ACTIONS(1340),
    [anon_sym_extern] = ACTIONS(1340),
    [anon_sym___attribute__] = ACTIONS(1340),
    [anon_sym___attribute] = ACTIONS(1340),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1342),
    [anon_sym___declspec] = ACTIONS(1340),
    [anon_sym___cdecl] = ACTIONS(1340),
    [anon_sym___clrcall] = ACTIONS(1340),
    [anon_sym___stdcall] = ACTIONS(1340),
    [anon_sym___fastcall] = ACTIONS(1340),
    [anon_sym___thiscall] = ACTIONS(1340),
    [anon_sym___vectorcall] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1342),
    [anon_sym_signed] = ACTIONS(1340),
    [anon_sym_unsigned] = ACTIONS(1340),
    [anon_sym_long] = ACTIONS(1340),
    [anon_sym_short] = ACTIONS(1340),
    [anon_sym_static] = ACTIONS(1340),
    [anon_sym_auto] = ACTIONS(1340),
    [anon_sym_register] = ACTIONS(1340),
    [anon_sym_inline] = ACTIONS(1340),
    [anon_sym___inline] = ACTIONS(1340),
    [anon_sym___inline__] = ACTIONS(1340),
    [anon_sym___forceinline] = ACTIONS(1340),
    [anon_sym_thread_local] = ACTIONS(1340),
    [anon_sym___thread] = ACTIONS(1340),
    [anon_sym_const] = ACTIONS(1340),
    [anon_sym_constexpr] = ACTIONS(1340),
    [anon_sym_volatile] = ACTIONS(1340),
    [anon_sym_restrict] = ACTIONS(1340),
    [anon_sym___restrict__] = ACTIONS(1340),
    [anon_sym__Atomic] = ACTIONS(1340),
    [anon_sym__Noreturn] = ACTIONS(1340),
    [anon_sym_noreturn] = ACTIONS(1340),
    [anon_sym__Nonnull] = ACTIONS(1340),
    [anon_sym_alignas] = ACTIONS(1340),
    [anon_sym__Alignas] = ACTIONS(1340),
    [sym_primitive_type] = ACTIONS(1340),
    [anon_sym_enum] = ACTIONS(1340),
    [anon_sym_struct] = ACTIONS(1340),
    [anon_sym_union] = ACTIONS(1340),
    [anon_sym_if] = ACTIONS(1340),
    [anon_sym_switch] = ACTIONS(1340),
    [anon_sym_case] = ACTIONS(1340),
    [anon_sym_default] = ACTIONS(1340),
    [anon_sym_while] = ACTIONS(1340),
    [anon_sym_do] = ACTIONS(1340),
    [anon_sym_for] = ACTIONS(1340),
    [anon_sym_return] = ACTIONS(1340),
    [anon_sym_break] = ACTIONS(1340),
    [anon_sym_continue] = ACTIONS(1340),
    [anon_sym_goto] = ACTIONS(1340),
    [anon_sym___try] = ACTIONS(1340),
    [anon_sym___leave] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1342),
    [anon_sym_PLUS_PLUS] = ACTIONS(1342),
    [anon_sym_sizeof] = ACTIONS(1340),
    [anon_sym___alignof__] = ACTIONS(1340),
    [anon_sym___alignof] = ACTIONS(1340),
    [anon_sym__alignof] = ACTIONS(1340),
    [anon_sym_alignof] = ACTIONS(1340),
    [anon_sym__Alignof] = ACTIONS(1340),
    [anon_sym_offsetof] = ACTIONS(1340),
    [anon_sym__Generic] = ACTIONS(1340),
    [anon_sym_asm] = ACTIONS(1340),
    [anon_sym___asm__] = ACTIONS(1340),
    [anon_sym___asm] = ACTIONS(1340),
    [sym_number_literal] = ACTIONS(1342),
    [anon_sym_L_SQUOTE] = ACTIONS(1342),
    [anon_sym_u_SQUOTE] = ACTIONS(1342),
    [anon_sym_U_SQUOTE] = ACTIONS(1342),
    [anon_sym_u8_SQUOTE] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_L_DQUOTE] = ACTIONS(1342),
    [anon_sym_u_DQUOTE] = ACTIONS(1342),
    [anon_sym_U_DQUOTE] = ACTIONS(1342),
    [anon_sym_u8_DQUOTE] = ACTIONS(1342),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [sym_true] = ACTIONS(1340),
    [sym_false] = ACTIONS(1340),
    [anon_sym_NULL] = ACTIONS(1340),
    [anon_sym_nullptr] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(300)] = {
    [sym_identifier] = ACTIONS(1344),
    [aux_sym_preproc_include_token1] = ACTIONS(1344),
    [aux_sym_preproc_def_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token2] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1344),
    [sym_preproc_directive] = ACTIONS(1344),
    [anon_sym_LPAREN2] = ACTIONS(1346),
    [anon_sym_BANG] = ACTIONS(1346),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1346),
    [anon_sym_AMP] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1346),
    [anon_sym___extension__] = ACTIONS(1344),
    [anon_sym_typedef] = ACTIONS(1344),
    [anon_sym_extern] = ACTIONS(1344),
    [anon_sym___attribute__] = ACTIONS(1344),
    [anon_sym___attribute] = ACTIONS(1344),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1346),
    [anon_sym___declspec] = ACTIONS(1344),
    [anon_sym___cdecl] = ACTIONS(1344),
    [anon_sym___clrcall] = ACTIONS(1344),
    [anon_sym___stdcall] = ACTIONS(1344),
    [anon_sym___fastcall] = ACTIONS(1344),
    [anon_sym___thiscall] = ACTIONS(1344),
    [anon_sym___vectorcall] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [anon_sym_signed] = ACTIONS(1344),
    [anon_sym_unsigned] = ACTIONS(1344),
    [anon_sym_long] = ACTIONS(1344),
    [anon_sym_short] = ACTIONS(1344),
    [anon_sym_static] = ACTIONS(1344),
    [anon_sym_auto] = ACTIONS(1344),
    [anon_sym_register] = ACTIONS(1344),
    [anon_sym_inline] = ACTIONS(1344),
    [anon_sym___inline] = ACTIONS(1344),
    [anon_sym___inline__] = ACTIONS(1344),
    [anon_sym___forceinline] = ACTIONS(1344),
    [anon_sym_thread_local] = ACTIONS(1344),
    [anon_sym___thread] = ACTIONS(1344),
    [anon_sym_const] = ACTIONS(1344),
    [anon_sym_constexpr] = ACTIONS(1344),
    [anon_sym_volatile] = ACTIONS(1344),
    [anon_sym_restrict] = ACTIONS(1344),
    [anon_sym___restrict__] = ACTIONS(1344),
    [anon_sym__Atomic] = ACTIONS(1344),
    [anon_sym__Noreturn] = ACTIONS(1344),
    [anon_sym_noreturn] = ACTIONS(1344),
    [anon_sym__Nonnull] = ACTIONS(1344),
    [anon_sym_alignas] = ACTIONS(1344),
    [anon_sym__Alignas] = ACTIONS(1344),
    [sym_primitive_type] = ACTIONS(1344),
    [anon_sym_enum] = ACTIONS(1344),
    [anon_sym_struct] = ACTIONS(1344),
    [anon_sym_union] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_switch] = ACTIONS(1344),
    [anon_sym_case] = ACTIONS(1344),
    [anon_sym_default] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_break] = ACTIONS(1344),
    [anon_sym_continue] = ACTIONS(1344),
    [anon_sym_goto] = ACTIONS(1344),
    [anon_sym___try] = ACTIONS(1344),
    [anon_sym___leave] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1346),
    [anon_sym_PLUS_PLUS] = ACTIONS(1346),
    [anon_sym_sizeof] = ACTIONS(1344),
    [anon_sym___alignof__] = ACTIONS(1344),
    [anon_sym___alignof] = ACTIONS(1344),
    [anon_sym__alignof] = ACTIONS(1344),
    [anon_sym_alignof] = ACTIONS(1344),
    [anon_sym__Alignof] = ACTIONS(1344),
    [anon_sym_offsetof] = ACTIONS(1344),
    [anon_sym__Generic] = ACTIONS(1344),
    [anon_sym_asm] = ACTIONS(1344),
    [anon_sym___asm__] = ACTIONS(1344),
    [anon_sym___asm] = ACTIONS(1344),
    [sym_number_literal] = ACTIONS(1346),
    [anon_sym_L_SQUOTE] = ACTIONS(1346),
    [anon_sym_u_SQUOTE] = ACTIONS(1346),
    [anon_sym_U_SQUOTE] = ACTIONS(1346),
    [anon_sym_u8_SQUOTE] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_L_DQUOTE] = ACTIONS(1346),
    [anon_sym_u_DQUOTE] = ACTIONS(1346),
    [anon_sym_U_DQUOTE] = ACTIONS(1346),
    [anon_sym_u8_DQUOTE] = ACTIONS(1346),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [anon_sym_NULL] = ACTIONS(1344),
    [anon_sym_nullptr] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(301)] = {
    [sym_identifier] = ACTIONS(1348),
    [aux_sym_preproc_include_token1] = ACTIONS(1348),
    [aux_sym_preproc_def_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token2] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1348),
    [sym_preproc_directive] = ACTIONS(1348),
    [anon_sym_LPAREN2] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym___extension__] = ACTIONS(1348),
    [anon_sym_typedef] = ACTIONS(1348),
    [anon_sym_extern] = ACTIONS(1348),
    [anon_sym___attribute__] = ACTIONS(1348),
    [anon_sym___attribute] = ACTIONS(1348),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1350),
    [anon_sym___declspec] = ACTIONS(1348),
    [anon_sym___cdecl] = ACTIONS(1348),
    [anon_sym___clrcall] = ACTIONS(1348),
    [anon_sym___stdcall] = ACTIONS(1348),
    [anon_sym___fastcall] = ACTIONS(1348),
    [anon_sym___thiscall] = ACTIONS(1348),
    [anon_sym___vectorcall] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_signed] = ACTIONS(1348),
    [anon_sym_unsigned] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_auto] = ACTIONS(1348),
    [anon_sym_register] = ACTIONS(1348),
    [anon_sym_inline] = ACTIONS(1348),
    [anon_sym___inline] = ACTIONS(1348),
    [anon_sym___inline__] = ACTIONS(1348),
    [anon_sym___forceinline] = ACTIONS(1348),
    [anon_sym_thread_local] = ACTIONS(1348),
    [anon_sym___thread] = ACTIONS(1348),
    [anon_sym_const] = ACTIONS(1348),
    [anon_sym_constexpr] = ACTIONS(1348),
    [anon_sym_volatile] = ACTIONS(1348),
    [anon_sym_restrict] = ACTIONS(1348),
    [anon_sym___restrict__] = ACTIONS(1348),
    [anon_sym__Atomic] = ACTIONS(1348),
    [anon_sym__Noreturn] = ACTIONS(1348),
    [anon_sym_noreturn] = ACTIONS(1348),
    [anon_sym__Nonnull] = ACTIONS(1348),
    [anon_sym_alignas] = ACTIONS(1348),
    [anon_sym__Alignas] = ACTIONS(1348),
    [sym_primitive_type] = ACTIONS(1348),
    [anon_sym_enum] = ACTIONS(1348),
    [anon_sym_struct] = ACTIONS(1348),
    [anon_sym_union] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_goto] = ACTIONS(1348),
    [anon_sym___try] = ACTIONS(1348),
    [anon_sym___leave] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_sizeof] = ACTIONS(1348),
    [anon_sym___alignof__] = ACTIONS(1348),
    [anon_sym___alignof] = ACTIONS(1348),
    [anon_sym__alignof] = ACTIONS(1348),
    [anon_sym_alignof] = ACTIONS(1348),
    [anon_sym__Alignof] = ACTIONS(1348),
    [anon_sym_offsetof] = ACTIONS(1348),
    [anon_sym__Generic] = ACTIONS(1348),
    [anon_sym_asm] = ACTIONS(1348),
    [anon_sym___asm__] = ACTIONS(1348),
    [anon_sym___asm] = ACTIONS(1348),
    [sym_number_literal] = ACTIONS(1350),
    [anon_sym_L_SQUOTE] = ACTIONS(1350),
    [anon_sym_u_SQUOTE] = ACTIONS(1350),
    [anon_sym_U_SQUOTE] = ACTIONS(1350),
    [anon_sym_u8_SQUOTE] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_L_DQUOTE] = ACTIONS(1350),
    [anon_sym_u_DQUOTE] = ACTIONS(1350),
    [anon_sym_U_DQUOTE] = ACTIONS(1350),
    [anon_sym_u8_DQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [sym_true] = ACTIONS(1348),
    [sym_false] = ACTIONS(1348),
    [anon_sym_NULL] = ACTIONS(1348),
    [anon_sym_nullptr] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(302)] = {
    [sym_identifier] = ACTIONS(1352),
    [aux_sym_preproc_include_token1] = ACTIONS(1352),
    [aux_sym_preproc_def_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token2] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1352),
    [sym_preproc_directive] = ACTIONS(1352),
    [anon_sym_LPAREN2] = ACTIONS(1354),
    [anon_sym_BANG] = ACTIONS(1354),
    [anon_sym_TILDE] = ACTIONS(1354),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1354),
    [anon_sym_AMP] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1354),
    [anon_sym___extension__] = ACTIONS(1352),
    [anon_sym_typedef] = ACTIONS(1352),
    [anon_sym_extern] = ACTIONS(1352),
    [anon_sym___attribute__] = ACTIONS(1352),
    [anon_sym___attribute] = ACTIONS(1352),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1354),
    [anon_sym___declspec] = ACTIONS(1352),
    [anon_sym___cdecl] = ACTIONS(1352),
    [anon_sym___clrcall] = ACTIONS(1352),
    [anon_sym___stdcall] = ACTIONS(1352),
    [anon_sym___fastcall] = ACTIONS(1352),
    [anon_sym___thiscall] = ACTIONS(1352),
    [anon_sym___vectorcall] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1354),
    [anon_sym_signed] = ACTIONS(1352),
    [anon_sym_unsigned] = ACTIONS(1352),
    [anon_sym_long] = ACTIONS(1352),
    [anon_sym_short] = ACTIONS(1352),
    [anon_sym_static] = ACTIONS(1352),
    [anon_sym_auto] = ACTIONS(1352),
    [anon_sym_register] = ACTIONS(1352),
    [anon_sym_inline] = ACTIONS(1352),
    [anon_sym___inline] = ACTIONS(1352),
    [anon_sym___inline__] = ACTIONS(1352),
    [anon_sym___forceinline] = ACTIONS(1352),
    [anon_sym_thread_local] = ACTIONS(1352),
    [anon_sym___thread] = ACTIONS(1352),
    [anon_sym_const] = ACTIONS(1352),
    [anon_sym_constexpr] = ACTIONS(1352),
    [anon_sym_volatile] = ACTIONS(1352),
    [anon_sym_restrict] = ACTIONS(1352),
    [anon_sym___restrict__] = ACTIONS(1352),
    [anon_sym__Atomic] = ACTIONS(1352),
    [anon_sym__Noreturn] = ACTIONS(1352),
    [anon_sym_noreturn] = ACTIONS(1352),
    [anon_sym__Nonnull] = ACTIONS(1352),
    [anon_sym_alignas] = ACTIONS(1352),
    [anon_sym__Alignas] = ACTIONS(1352),
    [sym_primitive_type] = ACTIONS(1352),
    [anon_sym_enum] = ACTIONS(1352),
    [anon_sym_struct] = ACTIONS(1352),
    [anon_sym_union] = ACTIONS(1352),
    [anon_sym_if] = ACTIONS(1352),
    [anon_sym_switch] = ACTIONS(1352),
    [anon_sym_case] = ACTIONS(1352),
    [anon_sym_default] = ACTIONS(1352),
    [anon_sym_while] = ACTIONS(1352),
    [anon_sym_do] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_break] = ACTIONS(1352),
    [anon_sym_continue] = ACTIONS(1352),
    [anon_sym_goto] = ACTIONS(1352),
    [anon_sym___try] = ACTIONS(1352),
    [anon_sym___leave] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1354),
    [anon_sym_PLUS_PLUS] = ACTIONS(1354),
    [anon_sym_sizeof] = ACTIONS(1352),
    [anon_sym___alignof__] = ACTIONS(1352),
    [anon_sym___alignof] = ACTIONS(1352),
    [anon_sym__alignof] = ACTIONS(1352),
    [anon_sym_alignof] = ACTIONS(1352),
    [anon_sym__Alignof] = ACTIONS(1352),
    [anon_sym_offsetof] = ACTIONS(1352),
    [anon_sym__Generic] = ACTIONS(1352),
    [anon_sym_asm] = ACTIONS(1352),
    [anon_sym___asm__] = ACTIONS(1352),
    [anon_sym___asm] = ACTIONS(1352),
    [sym_number_literal] = ACTIONS(1354),
    [anon_sym_L_SQUOTE] = ACTIONS(1354),
    [anon_sym_u_SQUOTE] = ACTIONS(1354),
    [anon_sym_U_SQUOTE] = ACTIONS(1354),
    [anon_sym_u8_SQUOTE] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_L_DQUOTE] = ACTIONS(1354),
    [anon_sym_u_DQUOTE] = ACTIONS(1354),
    [anon_sym_U_DQUOTE] = ACTIONS(1354),
    [anon_sym_u8_DQUOTE] = ACTIONS(1354),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [sym_true] = ACTIONS(1352),
    [sym_false] = ACTIONS(1352),
    [anon_sym_NULL] = ACTIONS(1352),
    [anon_sym_nullptr] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(303)] = {
    [sym_identifier] = ACTIONS(1356),
    [aux_sym_preproc_include_token1] = ACTIONS(1356),
    [aux_sym_preproc_def_token1] = ACTIONS(1356),
    [aux_sym_preproc_if_token1] = ACTIONS(1356),
    [aux_sym_preproc_if_token2] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1356),
    [sym_preproc_directive] = ACTIONS(1356),
    [anon_sym_LPAREN2] = ACTIONS(1358),
    [anon_sym_BANG] = ACTIONS(1358),
    [anon_sym_TILDE] = ACTIONS(1358),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1358),
    [anon_sym_AMP] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1358),
    [anon_sym___extension__] = ACTIONS(1356),
    [anon_sym_typedef] = ACTIONS(1356),
    [anon_sym_extern] = ACTIONS(1356),
    [anon_sym___attribute__] = ACTIONS(1356),
    [anon_sym___attribute] = ACTIONS(1356),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1358),
    [anon_sym___declspec] = ACTIONS(1356),
    [anon_sym___cdecl] = ACTIONS(1356),
    [anon_sym___clrcall] = ACTIONS(1356),
    [anon_sym___stdcall] = ACTIONS(1356),
    [anon_sym___fastcall] = ACTIONS(1356),
    [anon_sym___thiscall] = ACTIONS(1356),
    [anon_sym___vectorcall] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1358),
    [anon_sym_signed] = ACTIONS(1356),
    [anon_sym_unsigned] = ACTIONS(1356),
    [anon_sym_long] = ACTIONS(1356),
    [anon_sym_short] = ACTIONS(1356),
    [anon_sym_static] = ACTIONS(1356),
    [anon_sym_auto] = ACTIONS(1356),
    [anon_sym_register] = ACTIONS(1356),
    [anon_sym_inline] = ACTIONS(1356),
    [anon_sym___inline] = ACTIONS(1356),
    [anon_sym___inline__] = ACTIONS(1356),
    [anon_sym___forceinline] = ACTIONS(1356),
    [anon_sym_thread_local] = ACTIONS(1356),
    [anon_sym___thread] = ACTIONS(1356),
    [anon_sym_const] = ACTIONS(1356),
    [anon_sym_constexpr] = ACTIONS(1356),
    [anon_sym_volatile] = ACTIONS(1356),
    [anon_sym_restrict] = ACTIONS(1356),
    [anon_sym___restrict__] = ACTIONS(1356),
    [anon_sym__Atomic] = ACTIONS(1356),
    [anon_sym__Noreturn] = ACTIONS(1356),
    [anon_sym_noreturn] = ACTIONS(1356),
    [anon_sym__Nonnull] = ACTIONS(1356),
    [anon_sym_alignas] = ACTIONS(1356),
    [anon_sym__Alignas] = ACTIONS(1356),
    [sym_primitive_type] = ACTIONS(1356),
    [anon_sym_enum] = ACTIONS(1356),
    [anon_sym_struct] = ACTIONS(1356),
    [anon_sym_union] = ACTIONS(1356),
    [anon_sym_if] = ACTIONS(1356),
    [anon_sym_switch] = ACTIONS(1356),
    [anon_sym_case] = ACTIONS(1356),
    [anon_sym_default] = ACTIONS(1356),
    [anon_sym_while] = ACTIONS(1356),
    [anon_sym_do] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1356),
    [anon_sym_return] = ACTIONS(1356),
    [anon_sym_break] = ACTIONS(1356),
    [anon_sym_continue] = ACTIONS(1356),
    [anon_sym_goto] = ACTIONS(1356),
    [anon_sym___try] = ACTIONS(1356),
    [anon_sym___leave] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1358),
    [anon_sym_PLUS_PLUS] = ACTIONS(1358),
    [anon_sym_sizeof] = ACTIONS(1356),
    [anon_sym___alignof__] = ACTIONS(1356),
    [anon_sym___alignof] = ACTIONS(1356),
    [anon_sym__alignof] = ACTIONS(1356),
    [anon_sym_alignof] = ACTIONS(1356),
    [anon_sym__Alignof] = ACTIONS(1356),
    [anon_sym_offsetof] = ACTIONS(1356),
    [anon_sym__Generic] = ACTIONS(1356),
    [anon_sym_asm] = ACTIONS(1356),
    [anon_sym___asm__] = ACTIONS(1356),
    [anon_sym___asm] = ACTIONS(1356),
    [sym_number_literal] = ACTIONS(1358),
    [anon_sym_L_SQUOTE] = ACTIONS(1358),
    [anon_sym_u_SQUOTE] = ACTIONS(1358),
    [anon_sym_U_SQUOTE] = ACTIONS(1358),
    [anon_sym_u8_SQUOTE] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_L_DQUOTE] = ACTIONS(1358),
    [anon_sym_u_DQUOTE] = ACTIONS(1358),
    [anon_sym_U_DQUOTE] = ACTIONS(1358),
    [anon_sym_u8_DQUOTE] = ACTIONS(1358),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [sym_true] = ACTIONS(1356),
    [sym_false] = ACTIONS(1356),
    [anon_sym_NULL] = ACTIONS(1356),
    [anon_sym_nullptr] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(304)] = {
    [sym_identifier] = ACTIONS(1360),
    [aux_sym_preproc_include_token1] = ACTIONS(1360),
    [aux_sym_preproc_def_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token2] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1360),
    [sym_preproc_directive] = ACTIONS(1360),
    [anon_sym_LPAREN2] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_PLUS] = ACTIONS(1360),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym___extension__] = ACTIONS(1360),
    [anon_sym_typedef] = ACTIONS(1360),
    [anon_sym_extern] = ACTIONS(1360),
    [anon_sym___attribute__] = ACTIONS(1360),
    [anon_sym___attribute] = ACTIONS(1360),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1362),
    [anon_sym___declspec] = ACTIONS(1360),
    [anon_sym___cdecl] = ACTIONS(1360),
    [anon_sym___clrcall] = ACTIONS(1360),
    [anon_sym___stdcall] = ACTIONS(1360),
    [anon_sym___fastcall] = ACTIONS(1360),
    [anon_sym___thiscall] = ACTIONS(1360),
    [anon_sym___vectorcall] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_signed] = ACTIONS(1360),
    [anon_sym_unsigned] = ACTIONS(1360),
    [anon_sym_long] = ACTIONS(1360),
    [anon_sym_short] = ACTIONS(1360),
    [anon_sym_static] = ACTIONS(1360),
    [anon_sym_auto] = ACTIONS(1360),
    [anon_sym_register] = ACTIONS(1360),
    [anon_sym_inline] = ACTIONS(1360),
    [anon_sym___inline] = ACTIONS(1360),
    [anon_sym___inline__] = ACTIONS(1360),
    [anon_sym___forceinline] = ACTIONS(1360),
    [anon_sym_thread_local] = ACTIONS(1360),
    [anon_sym___thread] = ACTIONS(1360),
    [anon_sym_const] = ACTIONS(1360),
    [anon_sym_constexpr] = ACTIONS(1360),
    [anon_sym_volatile] = ACTIONS(1360),
    [anon_sym_restrict] = ACTIONS(1360),
    [anon_sym___restrict__] = ACTIONS(1360),
    [anon_sym__Atomic] = ACTIONS(1360),
    [anon_sym__Noreturn] = ACTIONS(1360),
    [anon_sym_noreturn] = ACTIONS(1360),
    [anon_sym__Nonnull] = ACTIONS(1360),
    [anon_sym_alignas] = ACTIONS(1360),
    [anon_sym__Alignas] = ACTIONS(1360),
    [sym_primitive_type] = ACTIONS(1360),
    [anon_sym_enum] = ACTIONS(1360),
    [anon_sym_struct] = ACTIONS(1360),
    [anon_sym_union] = ACTIONS(1360),
    [anon_sym_if] = ACTIONS(1360),
    [anon_sym_switch] = ACTIONS(1360),
    [anon_sym_case] = ACTIONS(1360),
    [anon_sym_default] = ACTIONS(1360),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_do] = ACTIONS(1360),
    [anon_sym_for] = ACTIONS(1360),
    [anon_sym_return] = ACTIONS(1360),
    [anon_sym_break] = ACTIONS(1360),
    [anon_sym_continue] = ACTIONS(1360),
    [anon_sym_goto] = ACTIONS(1360),
    [anon_sym___try] = ACTIONS(1360),
    [anon_sym___leave] = ACTIONS(1360),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_sizeof] = ACTIONS(1360),
    [anon_sym___alignof__] = ACTIONS(1360),
    [anon_sym___alignof] = ACTIONS(1360),
    [anon_sym__alignof] = ACTIONS(1360),
    [anon_sym_alignof] = ACTIONS(1360),
    [anon_sym__Alignof] = ACTIONS(1360),
    [anon_sym_offsetof] = ACTIONS(1360),
    [anon_sym__Generic] = ACTIONS(1360),
    [anon_sym_asm] = ACTIONS(1360),
    [anon_sym___asm__] = ACTIONS(1360),
    [anon_sym___asm] = ACTIONS(1360),
    [sym_number_literal] = ACTIONS(1362),
    [anon_sym_L_SQUOTE] = ACTIONS(1362),
    [anon_sym_u_SQUOTE] = ACTIONS(1362),
    [anon_sym_U_SQUOTE] = ACTIONS(1362),
    [anon_sym_u8_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_L_DQUOTE] = ACTIONS(1362),
    [anon_sym_u_DQUOTE] = ACTIONS(1362),
    [anon_sym_U_DQUOTE] = ACTIONS(1362),
    [anon_sym_u8_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [sym_true] = ACTIONS(1360),
    [sym_false] = ACTIONS(1360),
    [anon_sym_NULL] = ACTIONS(1360),
    [anon_sym_nullptr] = ACTIONS(1360),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(305)] = {
    [sym_identifier] = ACTIONS(1270),
    [aux_sym_preproc_include_token1] = ACTIONS(1270),
    [aux_sym_preproc_def_token1] = ACTIONS(1270),
    [aux_sym_preproc_if_token1] = ACTIONS(1270),
    [aux_sym_preproc_if_token2] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1270),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1270),
    [sym_preproc_directive] = ACTIONS(1270),
    [anon_sym_LPAREN2] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_TILDE] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_AMP] = ACTIONS(1272),
    [anon_sym_SEMI] = ACTIONS(1272),
    [anon_sym___extension__] = ACTIONS(1270),
    [anon_sym_typedef] = ACTIONS(1270),
    [anon_sym_extern] = ACTIONS(1270),
    [anon_sym___attribute__] = ACTIONS(1270),
    [anon_sym___attribute] = ACTIONS(1270),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1272),
    [anon_sym___declspec] = ACTIONS(1270),
    [anon_sym___cdecl] = ACTIONS(1270),
    [anon_sym___clrcall] = ACTIONS(1270),
    [anon_sym___stdcall] = ACTIONS(1270),
    [anon_sym___fastcall] = ACTIONS(1270),
    [anon_sym___thiscall] = ACTIONS(1270),
    [anon_sym___vectorcall] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_signed] = ACTIONS(1270),
    [anon_sym_unsigned] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_auto] = ACTIONS(1270),
    [anon_sym_register] = ACTIONS(1270),
    [anon_sym_inline] = ACTIONS(1270),
    [anon_sym___inline] = ACTIONS(1270),
    [anon_sym___inline__] = ACTIONS(1270),
    [anon_sym___forceinline] = ACTIONS(1270),
    [anon_sym_thread_local] = ACTIONS(1270),
    [anon_sym___thread] = ACTIONS(1270),
    [anon_sym_const] = ACTIONS(1270),
    [anon_sym_constexpr] = ACTIONS(1270),
    [anon_sym_volatile] = ACTIONS(1270),
    [anon_sym_restrict] = ACTIONS(1270),
    [anon_sym___restrict__] = ACTIONS(1270),
    [anon_sym__Atomic] = ACTIONS(1270),
    [anon_sym__Noreturn] = ACTIONS(1270),
    [anon_sym_noreturn] = ACTIONS(1270),
    [anon_sym__Nonnull] = ACTIONS(1270),
    [anon_sym_alignas] = ACTIONS(1270),
    [anon_sym__Alignas] = ACTIONS(1270),
    [sym_primitive_type] = ACTIONS(1270),
    [anon_sym_enum] = ACTIONS(1270),
    [anon_sym_struct] = ACTIONS(1270),
    [anon_sym_union] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_goto] = ACTIONS(1270),
    [anon_sym___try] = ACTIONS(1270),
    [anon_sym___leave] = ACTIONS(1270),
    [anon_sym_DASH_DASH] = ACTIONS(1272),
    [anon_sym_PLUS_PLUS] = ACTIONS(1272),
    [anon_sym_sizeof] = ACTIONS(1270),
    [anon_sym___alignof__] = ACTIONS(1270),
    [anon_sym___alignof] = ACTIONS(1270),
    [anon_sym__alignof] = ACTIONS(1270),
    [anon_sym_alignof] = ACTIONS(1270),
    [anon_sym__Alignof] = ACTIONS(1270),
    [anon_sym_offsetof] = ACTIONS(1270),
    [anon_sym__Generic] = ACTIONS(1270),
    [anon_sym_asm] = ACTIONS(1270),
    [anon_sym___asm__] = ACTIONS(1270),
    [anon_sym___asm] = ACTIONS(1270),
    [sym_number_literal] = ACTIONS(1272),
    [anon_sym_L_SQUOTE] = ACTIONS(1272),
    [anon_sym_u_SQUOTE] = ACTIONS(1272),
    [anon_sym_U_SQUOTE] = ACTIONS(1272),
    [anon_sym_u8_SQUOTE] = ACTIONS(1272),
    [anon_sym_SQUOTE] = ACTIONS(1272),
    [anon_sym_L_DQUOTE] = ACTIONS(1272),
    [anon_sym_u_DQUOTE] = ACTIONS(1272),
    [anon_sym_U_DQUOTE] = ACTIONS(1272),
    [anon_sym_u8_DQUOTE] = ACTIONS(1272),
    [anon_sym_DQUOTE] = ACTIONS(1272),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [anon_sym_NULL] = ACTIONS(1270),
    [anon_sym_nullptr] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(306)] = {
    [sym_identifier] = ACTIONS(1286),
    [aux_sym_preproc_include_token1] = ACTIONS(1286),
    [aux_sym_preproc_def_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token2] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1286),
    [sym_preproc_directive] = ACTIONS(1286),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_TILDE] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1288),
    [anon_sym_AMP] = ACTIONS(1288),
    [anon_sym_SEMI] = ACTIONS(1288),
    [anon_sym___extension__] = ACTIONS(1286),
    [anon_sym_typedef] = ACTIONS(1286),
    [anon_sym_extern] = ACTIONS(1286),
    [anon_sym___attribute__] = ACTIONS(1286),
    [anon_sym___attribute] = ACTIONS(1286),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1288),
    [anon_sym___declspec] = ACTIONS(1286),
    [anon_sym___cdecl] = ACTIONS(1286),
    [anon_sym___clrcall] = ACTIONS(1286),
    [anon_sym___stdcall] = ACTIONS(1286),
    [anon_sym___fastcall] = ACTIONS(1286),
    [anon_sym___thiscall] = ACTIONS(1286),
    [anon_sym___vectorcall] = ACTIONS(1286),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_signed] = ACTIONS(1286),
    [anon_sym_unsigned] = ACTIONS(1286),
    [anon_sym_long] = ACTIONS(1286),
    [anon_sym_short] = ACTIONS(1286),
    [anon_sym_static] = ACTIONS(1286),
    [anon_sym_auto] = ACTIONS(1286),
    [anon_sym_register] = ACTIONS(1286),
    [anon_sym_inline] = ACTIONS(1286),
    [anon_sym___inline] = ACTIONS(1286),
    [anon_sym___inline__] = ACTIONS(1286),
    [anon_sym___forceinline] = ACTIONS(1286),
    [anon_sym_thread_local] = ACTIONS(1286),
    [anon_sym___thread] = ACTIONS(1286),
    [anon_sym_const] = ACTIONS(1286),
    [anon_sym_constexpr] = ACTIONS(1286),
    [anon_sym_volatile] = ACTIONS(1286),
    [anon_sym_restrict] = ACTIONS(1286),
    [anon_sym___restrict__] = ACTIONS(1286),
    [anon_sym__Atomic] = ACTIONS(1286),
    [anon_sym__Noreturn] = ACTIONS(1286),
    [anon_sym_noreturn] = ACTIONS(1286),
    [anon_sym__Nonnull] = ACTIONS(1286),
    [anon_sym_alignas] = ACTIONS(1286),
    [anon_sym__Alignas] = ACTIONS(1286),
    [sym_primitive_type] = ACTIONS(1286),
    [anon_sym_enum] = ACTIONS(1286),
    [anon_sym_struct] = ACTIONS(1286),
    [anon_sym_union] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(1286),
    [anon_sym_switch] = ACTIONS(1286),
    [anon_sym_case] = ACTIONS(1286),
    [anon_sym_default] = ACTIONS(1286),
    [anon_sym_while] = ACTIONS(1286),
    [anon_sym_do] = ACTIONS(1286),
    [anon_sym_for] = ACTIONS(1286),
    [anon_sym_return] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(1286),
    [anon_sym_continue] = ACTIONS(1286),
    [anon_sym_goto] = ACTIONS(1286),
    [anon_sym___try] = ACTIONS(1286),
    [anon_sym___leave] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1288),
    [anon_sym_PLUS_PLUS] = ACTIONS(1288),
    [anon_sym_sizeof] = ACTIONS(1286),
    [anon_sym___alignof__] = ACTIONS(1286),
    [anon_sym___alignof] = ACTIONS(1286),
    [anon_sym__alignof] = ACTIONS(1286),
    [anon_sym_alignof] = ACTIONS(1286),
    [anon_sym__Alignof] = ACTIONS(1286),
    [anon_sym_offsetof] = ACTIONS(1286),
    [anon_sym__Generic] = ACTIONS(1286),
    [anon_sym_asm] = ACTIONS(1286),
    [anon_sym___asm__] = ACTIONS(1286),
    [anon_sym___asm] = ACTIONS(1286),
    [sym_number_literal] = ACTIONS(1288),
    [anon_sym_L_SQUOTE] = ACTIONS(1288),
    [anon_sym_u_SQUOTE] = ACTIONS(1288),
    [anon_sym_U_SQUOTE] = ACTIONS(1288),
    [anon_sym_u8_SQUOTE] = ACTIONS(1288),
    [anon_sym_SQUOTE] = ACTIONS(1288),
    [anon_sym_L_DQUOTE] = ACTIONS(1288),
    [anon_sym_u_DQUOTE] = ACTIONS(1288),
    [anon_sym_U_DQUOTE] = ACTIONS(1288),
    [anon_sym_u8_DQUOTE] = ACTIONS(1288),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [sym_true] = ACTIONS(1286),
    [sym_false] = ACTIONS(1286),
    [anon_sym_NULL] = ACTIONS(1286),
    [anon_sym_nullptr] = ACTIONS(1286),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(307)] = {
    [sym_identifier] = ACTIONS(1290),
    [aux_sym_preproc_include_token1] = ACTIONS(1290),
    [aux_sym_preproc_def_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token2] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1290),
    [sym_preproc_directive] = ACTIONS(1290),
    [anon_sym_LPAREN2] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym___extension__] = ACTIONS(1290),
    [anon_sym_typedef] = ACTIONS(1290),
    [anon_sym_extern] = ACTIONS(1290),
    [anon_sym___attribute__] = ACTIONS(1290),
    [anon_sym___attribute] = ACTIONS(1290),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1292),
    [anon_sym___declspec] = ACTIONS(1290),
    [anon_sym___cdecl] = ACTIONS(1290),
    [anon_sym___clrcall] = ACTIONS(1290),
    [anon_sym___stdcall] = ACTIONS(1290),
    [anon_sym___fastcall] = ACTIONS(1290),
    [anon_sym___thiscall] = ACTIONS(1290),
    [anon_sym___vectorcall] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_signed] = ACTIONS(1290),
    [anon_sym_unsigned] = ACTIONS(1290),
    [anon_sym_long] = ACTIONS(1290),
    [anon_sym_short] = ACTIONS(1290),
    [anon_sym_static] = ACTIONS(1290),
    [anon_sym_auto] = ACTIONS(1290),
    [anon_sym_register] = ACTIONS(1290),
    [anon_sym_inline] = ACTIONS(1290),
    [anon_sym___inline] = ACTIONS(1290),
    [anon_sym___inline__] = ACTIONS(1290),
    [anon_sym___forceinline] = ACTIONS(1290),
    [anon_sym_thread_local] = ACTIONS(1290),
    [anon_sym___thread] = ACTIONS(1290),
    [anon_sym_const] = ACTIONS(1290),
    [anon_sym_constexpr] = ACTIONS(1290),
    [anon_sym_volatile] = ACTIONS(1290),
    [anon_sym_restrict] = ACTIONS(1290),
    [anon_sym___restrict__] = ACTIONS(1290),
    [anon_sym__Atomic] = ACTIONS(1290),
    [anon_sym__Noreturn] = ACTIONS(1290),
    [anon_sym_noreturn] = ACTIONS(1290),
    [anon_sym__Nonnull] = ACTIONS(1290),
    [anon_sym_alignas] = ACTIONS(1290),
    [anon_sym__Alignas] = ACTIONS(1290),
    [sym_primitive_type] = ACTIONS(1290),
    [anon_sym_enum] = ACTIONS(1290),
    [anon_sym_struct] = ACTIONS(1290),
    [anon_sym_union] = ACTIONS(1290),
    [anon_sym_if] = ACTIONS(1290),
    [anon_sym_switch] = ACTIONS(1290),
    [anon_sym_case] = ACTIONS(1290),
    [anon_sym_default] = ACTIONS(1290),
    [anon_sym_while] = ACTIONS(1290),
    [anon_sym_do] = ACTIONS(1290),
    [anon_sym_for] = ACTIONS(1290),
    [anon_sym_return] = ACTIONS(1290),
    [anon_sym_break] = ACTIONS(1290),
    [anon_sym_continue] = ACTIONS(1290),
    [anon_sym_goto] = ACTIONS(1290),
    [anon_sym___try] = ACTIONS(1290),
    [anon_sym___leave] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_sizeof] = ACTIONS(1290),
    [anon_sym___alignof__] = ACTIONS(1290),
    [anon_sym___alignof] = ACTIONS(1290),
    [anon_sym__alignof] = ACTIONS(1290),
    [anon_sym_alignof] = ACTIONS(1290),
    [anon_sym__Alignof] = ACTIONS(1290),
    [anon_sym_offsetof] = ACTIONS(1290),
    [anon_sym__Generic] = ACTIONS(1290),
    [anon_sym_asm] = ACTIONS(1290),
    [anon_sym___asm__] = ACTIONS(1290),
    [anon_sym___asm] = ACTIONS(1290),
    [sym_number_literal] = ACTIONS(1292),
    [anon_sym_L_SQUOTE] = ACTIONS(1292),
    [anon_sym_u_SQUOTE] = ACTIONS(1292),
    [anon_sym_U_SQUOTE] = ACTIONS(1292),
    [anon_sym_u8_SQUOTE] = ACTIONS(1292),
    [anon_sym_SQUOTE] = ACTIONS(1292),
    [anon_sym_L_DQUOTE] = ACTIONS(1292),
    [anon_sym_u_DQUOTE] = ACTIONS(1292),
    [anon_sym_U_DQUOTE] = ACTIONS(1292),
    [anon_sym_u8_DQUOTE] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [sym_true] = ACTIONS(1290),
    [sym_false] = ACTIONS(1290),
    [anon_sym_NULL] = ACTIONS(1290),
    [anon_sym_nullptr] = ACTIONS(1290),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(308)] = {
    [sym_identifier] = ACTIONS(1318),
    [aux_sym_preproc_include_token1] = ACTIONS(1318),
    [aux_sym_preproc_def_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token2] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1318),
    [sym_preproc_directive] = ACTIONS(1318),
    [anon_sym_LPAREN2] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym___extension__] = ACTIONS(1318),
    [anon_sym_typedef] = ACTIONS(1318),
    [anon_sym_extern] = ACTIONS(1318),
    [anon_sym___attribute__] = ACTIONS(1318),
    [anon_sym___attribute] = ACTIONS(1318),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1320),
    [anon_sym___declspec] = ACTIONS(1318),
    [anon_sym___cdecl] = ACTIONS(1318),
    [anon_sym___clrcall] = ACTIONS(1318),
    [anon_sym___stdcall] = ACTIONS(1318),
    [anon_sym___fastcall] = ACTIONS(1318),
    [anon_sym___thiscall] = ACTIONS(1318),
    [anon_sym___vectorcall] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_signed] = ACTIONS(1318),
    [anon_sym_unsigned] = ACTIONS(1318),
    [anon_sym_long] = ACTIONS(1318),
    [anon_sym_short] = ACTIONS(1318),
    [anon_sym_static] = ACTIONS(1318),
    [anon_sym_auto] = ACTIONS(1318),
    [anon_sym_register] = ACTIONS(1318),
    [anon_sym_inline] = ACTIONS(1318),
    [anon_sym___inline] = ACTIONS(1318),
    [anon_sym___inline__] = ACTIONS(1318),
    [anon_sym___forceinline] = ACTIONS(1318),
    [anon_sym_thread_local] = ACTIONS(1318),
    [anon_sym___thread] = ACTIONS(1318),
    [anon_sym_const] = ACTIONS(1318),
    [anon_sym_constexpr] = ACTIONS(1318),
    [anon_sym_volatile] = ACTIONS(1318),
    [anon_sym_restrict] = ACTIONS(1318),
    [anon_sym___restrict__] = ACTIONS(1318),
    [anon_sym__Atomic] = ACTIONS(1318),
    [anon_sym__Noreturn] = ACTIONS(1318),
    [anon_sym_noreturn] = ACTIONS(1318),
    [anon_sym__Nonnull] = ACTIONS(1318),
    [anon_sym_alignas] = ACTIONS(1318),
    [anon_sym__Alignas] = ACTIONS(1318),
    [sym_primitive_type] = ACTIONS(1318),
    [anon_sym_enum] = ACTIONS(1318),
    [anon_sym_struct] = ACTIONS(1318),
    [anon_sym_union] = ACTIONS(1318),
    [anon_sym_if] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1318),
    [anon_sym_case] = ACTIONS(1318),
    [anon_sym_default] = ACTIONS(1318),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_break] = ACTIONS(1318),
    [anon_sym_continue] = ACTIONS(1318),
    [anon_sym_goto] = ACTIONS(1318),
    [anon_sym___try] = ACTIONS(1318),
    [anon_sym___leave] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_sizeof] = ACTIONS(1318),
    [anon_sym___alignof__] = ACTIONS(1318),
    [anon_sym___alignof] = ACTIONS(1318),
    [anon_sym__alignof] = ACTIONS(1318),
    [anon_sym_alignof] = ACTIONS(1318),
    [anon_sym__Alignof] = ACTIONS(1318),
    [anon_sym_offsetof] = ACTIONS(1318),
    [anon_sym__Generic] = ACTIONS(1318),
    [anon_sym_asm] = ACTIONS(1318),
    [anon_sym___asm__] = ACTIONS(1318),
    [anon_sym___asm] = ACTIONS(1318),
    [sym_number_literal] = ACTIONS(1320),
    [anon_sym_L_SQUOTE] = ACTIONS(1320),
    [anon_sym_u_SQUOTE] = ACTIONS(1320),
    [anon_sym_U_SQUOTE] = ACTIONS(1320),
    [anon_sym_u8_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_L_DQUOTE] = ACTIONS(1320),
    [anon_sym_u_DQUOTE] = ACTIONS(1320),
    [anon_sym_U_DQUOTE] = ACTIONS(1320),
    [anon_sym_u8_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [anon_sym_NULL] = ACTIONS(1318),
    [anon_sym_nullptr] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(309)] = {
    [sym_identifier] = ACTIONS(1356),
    [aux_sym_preproc_include_token1] = ACTIONS(1356),
    [aux_sym_preproc_def_token1] = ACTIONS(1356),
    [aux_sym_preproc_if_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1356),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1356),
    [sym_preproc_directive] = ACTIONS(1356),
    [anon_sym_LPAREN2] = ACTIONS(1358),
    [anon_sym_BANG] = ACTIONS(1358),
    [anon_sym_TILDE] = ACTIONS(1358),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1358),
    [anon_sym_AMP] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1358),
    [anon_sym___extension__] = ACTIONS(1356),
    [anon_sym_typedef] = ACTIONS(1356),
    [anon_sym_extern] = ACTIONS(1356),
    [anon_sym___attribute__] = ACTIONS(1356),
    [anon_sym___attribute] = ACTIONS(1356),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1358),
    [anon_sym___declspec] = ACTIONS(1356),
    [anon_sym___cdecl] = ACTIONS(1356),
    [anon_sym___clrcall] = ACTIONS(1356),
    [anon_sym___stdcall] = ACTIONS(1356),
    [anon_sym___fastcall] = ACTIONS(1356),
    [anon_sym___thiscall] = ACTIONS(1356),
    [anon_sym___vectorcall] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1358),
    [anon_sym_RBRACE] = ACTIONS(1358),
    [anon_sym_signed] = ACTIONS(1356),
    [anon_sym_unsigned] = ACTIONS(1356),
    [anon_sym_long] = ACTIONS(1356),
    [anon_sym_short] = ACTIONS(1356),
    [anon_sym_static] = ACTIONS(1356),
    [anon_sym_auto] = ACTIONS(1356),
    [anon_sym_register] = ACTIONS(1356),
    [anon_sym_inline] = ACTIONS(1356),
    [anon_sym___inline] = ACTIONS(1356),
    [anon_sym___inline__] = ACTIONS(1356),
    [anon_sym___forceinline] = ACTIONS(1356),
    [anon_sym_thread_local] = ACTIONS(1356),
    [anon_sym___thread] = ACTIONS(1356),
    [anon_sym_const] = ACTIONS(1356),
    [anon_sym_constexpr] = ACTIONS(1356),
    [anon_sym_volatile] = ACTIONS(1356),
    [anon_sym_restrict] = ACTIONS(1356),
    [anon_sym___restrict__] = ACTIONS(1356),
    [anon_sym__Atomic] = ACTIONS(1356),
    [anon_sym__Noreturn] = ACTIONS(1356),
    [anon_sym_noreturn] = ACTIONS(1356),
    [anon_sym__Nonnull] = ACTIONS(1356),
    [anon_sym_alignas] = ACTIONS(1356),
    [anon_sym__Alignas] = ACTIONS(1356),
    [sym_primitive_type] = ACTIONS(1356),
    [anon_sym_enum] = ACTIONS(1356),
    [anon_sym_struct] = ACTIONS(1356),
    [anon_sym_union] = ACTIONS(1356),
    [anon_sym_if] = ACTIONS(1356),
    [anon_sym_switch] = ACTIONS(1356),
    [anon_sym_case] = ACTIONS(1356),
    [anon_sym_default] = ACTIONS(1356),
    [anon_sym_while] = ACTIONS(1356),
    [anon_sym_do] = ACTIONS(1356),
    [anon_sym_for] = ACTIONS(1356),
    [anon_sym_return] = ACTIONS(1356),
    [anon_sym_break] = ACTIONS(1356),
    [anon_sym_continue] = ACTIONS(1356),
    [anon_sym_goto] = ACTIONS(1356),
    [anon_sym___try] = ACTIONS(1356),
    [anon_sym___leave] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1358),
    [anon_sym_PLUS_PLUS] = ACTIONS(1358),
    [anon_sym_sizeof] = ACTIONS(1356),
    [anon_sym___alignof__] = ACTIONS(1356),
    [anon_sym___alignof] = ACTIONS(1356),
    [anon_sym__alignof] = ACTIONS(1356),
    [anon_sym_alignof] = ACTIONS(1356),
    [anon_sym__Alignof] = ACTIONS(1356),
    [anon_sym_offsetof] = ACTIONS(1356),
    [anon_sym__Generic] = ACTIONS(1356),
    [anon_sym_asm] = ACTIONS(1356),
    [anon_sym___asm__] = ACTIONS(1356),
    [anon_sym___asm] = ACTIONS(1356),
    [sym_number_literal] = ACTIONS(1358),
    [anon_sym_L_SQUOTE] = ACTIONS(1358),
    [anon_sym_u_SQUOTE] = ACTIONS(1358),
    [anon_sym_U_SQUOTE] = ACTIONS(1358),
    [anon_sym_u8_SQUOTE] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_L_DQUOTE] = ACTIONS(1358),
    [anon_sym_u_DQUOTE] = ACTIONS(1358),
    [anon_sym_U_DQUOTE] = ACTIONS(1358),
    [anon_sym_u8_DQUOTE] = ACTIONS(1358),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [sym_true] = ACTIONS(1356),
    [sym_false] = ACTIONS(1356),
    [anon_sym_NULL] = ACTIONS(1356),
    [anon_sym_nullptr] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(310)] = {
    [sym_identifier] = ACTIONS(1364),
    [aux_sym_preproc_include_token1] = ACTIONS(1364),
    [aux_sym_preproc_def_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1364),
    [sym_preproc_directive] = ACTIONS(1364),
    [anon_sym_LPAREN2] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_PLUS] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym___extension__] = ACTIONS(1364),
    [anon_sym_typedef] = ACTIONS(1364),
    [anon_sym_extern] = ACTIONS(1364),
    [anon_sym___attribute__] = ACTIONS(1364),
    [anon_sym___attribute] = ACTIONS(1364),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1366),
    [anon_sym___declspec] = ACTIONS(1364),
    [anon_sym___cdecl] = ACTIONS(1364),
    [anon_sym___clrcall] = ACTIONS(1364),
    [anon_sym___stdcall] = ACTIONS(1364),
    [anon_sym___fastcall] = ACTIONS(1364),
    [anon_sym___thiscall] = ACTIONS(1364),
    [anon_sym___vectorcall] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_RBRACE] = ACTIONS(1366),
    [anon_sym_signed] = ACTIONS(1364),
    [anon_sym_unsigned] = ACTIONS(1364),
    [anon_sym_long] = ACTIONS(1364),
    [anon_sym_short] = ACTIONS(1364),
    [anon_sym_static] = ACTIONS(1364),
    [anon_sym_auto] = ACTIONS(1364),
    [anon_sym_register] = ACTIONS(1364),
    [anon_sym_inline] = ACTIONS(1364),
    [anon_sym___inline] = ACTIONS(1364),
    [anon_sym___inline__] = ACTIONS(1364),
    [anon_sym___forceinline] = ACTIONS(1364),
    [anon_sym_thread_local] = ACTIONS(1364),
    [anon_sym___thread] = ACTIONS(1364),
    [anon_sym_const] = ACTIONS(1364),
    [anon_sym_constexpr] = ACTIONS(1364),
    [anon_sym_volatile] = ACTIONS(1364),
    [anon_sym_restrict] = ACTIONS(1364),
    [anon_sym___restrict__] = ACTIONS(1364),
    [anon_sym__Atomic] = ACTIONS(1364),
    [anon_sym__Noreturn] = ACTIONS(1364),
    [anon_sym_noreturn] = ACTIONS(1364),
    [anon_sym__Nonnull] = ACTIONS(1364),
    [anon_sym_alignas] = ACTIONS(1364),
    [anon_sym__Alignas] = ACTIONS(1364),
    [sym_primitive_type] = ACTIONS(1364),
    [anon_sym_enum] = ACTIONS(1364),
    [anon_sym_struct] = ACTIONS(1364),
    [anon_sym_union] = ACTIONS(1364),
    [anon_sym_if] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1364),
    [anon_sym_case] = ACTIONS(1364),
    [anon_sym_default] = ACTIONS(1364),
    [anon_sym_while] = ACTIONS(1364),
    [anon_sym_do] = ACTIONS(1364),
    [anon_sym_for] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1364),
    [anon_sym_break] = ACTIONS(1364),
    [anon_sym_continue] = ACTIONS(1364),
    [anon_sym_goto] = ACTIONS(1364),
    [anon_sym___try] = ACTIONS(1364),
    [anon_sym___leave] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1366),
    [anon_sym_sizeof] = ACTIONS(1364),
    [anon_sym___alignof__] = ACTIONS(1364),
    [anon_sym___alignof] = ACTIONS(1364),
    [anon_sym__alignof] = ACTIONS(1364),
    [anon_sym_alignof] = ACTIONS(1364),
    [anon_sym__Alignof] = ACTIONS(1364),
    [anon_sym_offsetof] = ACTIONS(1364),
    [anon_sym__Generic] = ACTIONS(1364),
    [anon_sym_asm] = ACTIONS(1364),
    [anon_sym___asm__] = ACTIONS(1364),
    [anon_sym___asm] = ACTIONS(1364),
    [sym_number_literal] = ACTIONS(1366),
    [anon_sym_L_SQUOTE] = ACTIONS(1366),
    [anon_sym_u_SQUOTE] = ACTIONS(1366),
    [anon_sym_U_SQUOTE] = ACTIONS(1366),
    [anon_sym_u8_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_L_DQUOTE] = ACTIONS(1366),
    [anon_sym_u_DQUOTE] = ACTIONS(1366),
    [anon_sym_U_DQUOTE] = ACTIONS(1366),
    [anon_sym_u8_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [sym_true] = ACTIONS(1364),
    [sym_false] = ACTIONS(1364),
    [anon_sym_NULL] = ACTIONS(1364),
    [anon_sym_nullptr] = ACTIONS(1364),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(311)] = {
    [sym_identifier] = ACTIONS(1334),
    [aux_sym_preproc_include_token1] = ACTIONS(1334),
    [aux_sym_preproc_def_token1] = ACTIONS(1334),
    [aux_sym_preproc_if_token1] = ACTIONS(1334),
    [aux_sym_preproc_if_token2] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1334),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1334),
    [sym_preproc_directive] = ACTIONS(1334),
    [anon_sym_LPAREN2] = ACTIONS(1337),
    [anon_sym_BANG] = ACTIONS(1337),
    [anon_sym_TILDE] = ACTIONS(1337),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_STAR] = ACTIONS(1337),
    [anon_sym_AMP] = ACTIONS(1337),
    [anon_sym_SEMI] = ACTIONS(1337),
    [anon_sym___extension__] = ACTIONS(1334),
    [anon_sym_typedef] = ACTIONS(1334),
    [anon_sym_extern] = ACTIONS(1334),
    [anon_sym___attribute__] = ACTIONS(1334),
    [anon_sym___attribute] = ACTIONS(1334),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1337),
    [anon_sym___declspec] = ACTIONS(1334),
    [anon_sym___cdecl] = ACTIONS(1334),
    [anon_sym___clrcall] = ACTIONS(1334),
    [anon_sym___stdcall] = ACTIONS(1334),
    [anon_sym___fastcall] = ACTIONS(1334),
    [anon_sym___thiscall] = ACTIONS(1334),
    [anon_sym___vectorcall] = ACTIONS(1334),
    [anon_sym_LBRACE] = ACTIONS(1337),
    [anon_sym_signed] = ACTIONS(1334),
    [anon_sym_unsigned] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_auto] = ACTIONS(1334),
    [anon_sym_register] = ACTIONS(1334),
    [anon_sym_inline] = ACTIONS(1334),
    [anon_sym___inline] = ACTIONS(1334),
    [anon_sym___inline__] = ACTIONS(1334),
    [anon_sym___forceinline] = ACTIONS(1334),
    [anon_sym_thread_local] = ACTIONS(1334),
    [anon_sym___thread] = ACTIONS(1334),
    [anon_sym_const] = ACTIONS(1334),
    [anon_sym_constexpr] = ACTIONS(1334),
    [anon_sym_volatile] = ACTIONS(1334),
    [anon_sym_restrict] = ACTIONS(1334),
    [anon_sym___restrict__] = ACTIONS(1334),
    [anon_sym__Atomic] = ACTIONS(1334),
    [anon_sym__Noreturn] = ACTIONS(1334),
    [anon_sym_noreturn] = ACTIONS(1334),
    [anon_sym__Nonnull] = ACTIONS(1334),
    [anon_sym_alignas] = ACTIONS(1334),
    [anon_sym__Alignas] = ACTIONS(1334),
    [sym_primitive_type] = ACTIONS(1334),
    [anon_sym_enum] = ACTIONS(1334),
    [anon_sym_struct] = ACTIONS(1334),
    [anon_sym_union] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_goto] = ACTIONS(1334),
    [anon_sym___try] = ACTIONS(1334),
    [anon_sym___leave] = ACTIONS(1334),
    [anon_sym_DASH_DASH] = ACTIONS(1337),
    [anon_sym_PLUS_PLUS] = ACTIONS(1337),
    [anon_sym_sizeof] = ACTIONS(1334),
    [anon_sym___alignof__] = ACTIONS(1334),
    [anon_sym___alignof] = ACTIONS(1334),
    [anon_sym__alignof] = ACTIONS(1334),
    [anon_sym_alignof] = ACTIONS(1334),
    [anon_sym__Alignof] = ACTIONS(1334),
    [anon_sym_offsetof] = ACTIONS(1334),
    [anon_sym__Generic] = ACTIONS(1334),
    [anon_sym_asm] = ACTIONS(1334),
    [anon_sym___asm__] = ACTIONS(1334),
    [anon_sym___asm] = ACTIONS(1334),
    [sym_number_literal] = ACTIONS(1337),
    [anon_sym_L_SQUOTE] = ACTIONS(1337),
    [anon_sym_u_SQUOTE] = ACTIONS(1337),
    [anon_sym_U_SQUOTE] = ACTIONS(1337),
    [anon_sym_u8_SQUOTE] = ACTIONS(1337),
    [anon_sym_SQUOTE] = ACTIONS(1337),
    [anon_sym_L_DQUOTE] = ACTIONS(1337),
    [anon_sym_u_DQUOTE] = ACTIONS(1337),
    [anon_sym_U_DQUOTE] = ACTIONS(1337),
    [anon_sym_u8_DQUOTE] = ACTIONS(1337),
    [anon_sym_DQUOTE] = ACTIONS(1337),
    [sym_true] = ACTIONS(1334),
    [sym_false] = ACTIONS(1334),
    [anon_sym_NULL] = ACTIONS(1334),
    [anon_sym_nullptr] = ACTIONS(1334),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(312)] = {
    [sym_expression] = STATE(699),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(675),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(675),
    [sym_call_expression] = STATE(675),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(675),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(675),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(713),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1386),
    [anon_sym_COMMA] = ACTIONS(1380),
    [aux_sym_preproc_if_token2] = ACTIONS(1380),
    [aux_sym_preproc_else_token1] = ACTIONS(1380),
    [aux_sym_preproc_elif_token1] = ACTIONS(1386),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1380),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1394),
    [anon_sym_TILDE] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1386),
    [anon_sym___extension__] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1386),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_STAR_EQ] = ACTIONS(1380),
    [anon_sym_SLASH_EQ] = ACTIONS(1380),
    [anon_sym_PERCENT_EQ] = ACTIONS(1380),
    [anon_sym_PLUS_EQ] = ACTIONS(1380),
    [anon_sym_DASH_EQ] = ACTIONS(1380),
    [anon_sym_LT_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT_GT_EQ] = ACTIONS(1380),
    [anon_sym_AMP_EQ] = ACTIONS(1380),
    [anon_sym_CARET_EQ] = ACTIONS(1380),
    [anon_sym_PIPE_EQ] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(1400),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(313)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(2003),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(314)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(1995),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(315)] = {
    [ts_builtin_sym_end] = ACTIONS(1276),
    [sym_identifier] = ACTIONS(1274),
    [aux_sym_preproc_include_token1] = ACTIONS(1274),
    [aux_sym_preproc_def_token1] = ACTIONS(1274),
    [aux_sym_preproc_if_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1274),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1274),
    [sym_preproc_directive] = ACTIONS(1274),
    [anon_sym_LPAREN2] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_TILDE] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1274),
    [anon_sym_PLUS] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1276),
    [anon_sym_AMP] = ACTIONS(1276),
    [anon_sym_SEMI] = ACTIONS(1276),
    [anon_sym___extension__] = ACTIONS(1274),
    [anon_sym_typedef] = ACTIONS(1274),
    [anon_sym_extern] = ACTIONS(1274),
    [anon_sym___attribute__] = ACTIONS(1274),
    [anon_sym___attribute] = ACTIONS(1274),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1276),
    [anon_sym___declspec] = ACTIONS(1274),
    [anon_sym___cdecl] = ACTIONS(1274),
    [anon_sym___clrcall] = ACTIONS(1274),
    [anon_sym___stdcall] = ACTIONS(1274),
    [anon_sym___fastcall] = ACTIONS(1274),
    [anon_sym___thiscall] = ACTIONS(1274),
    [anon_sym___vectorcall] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_signed] = ACTIONS(1274),
    [anon_sym_unsigned] = ACTIONS(1274),
    [anon_sym_long] = ACTIONS(1274),
    [anon_sym_short] = ACTIONS(1274),
    [anon_sym_static] = ACTIONS(1274),
    [anon_sym_auto] = ACTIONS(1274),
    [anon_sym_register] = ACTIONS(1274),
    [anon_sym_inline] = ACTIONS(1274),
    [anon_sym___inline] = ACTIONS(1274),
    [anon_sym___inline__] = ACTIONS(1274),
    [anon_sym___forceinline] = ACTIONS(1274),
    [anon_sym_thread_local] = ACTIONS(1274),
    [anon_sym___thread] = ACTIONS(1274),
    [anon_sym_const] = ACTIONS(1274),
    [anon_sym_constexpr] = ACTIONS(1274),
    [anon_sym_volatile] = ACTIONS(1274),
    [anon_sym_restrict] = ACTIONS(1274),
    [anon_sym___restrict__] = ACTIONS(1274),
    [anon_sym__Atomic] = ACTIONS(1274),
    [anon_sym__Noreturn] = ACTIONS(1274),
    [anon_sym_noreturn] = ACTIONS(1274),
    [anon_sym__Nonnull] = ACTIONS(1274),
    [anon_sym_alignas] = ACTIONS(1274),
    [anon_sym__Alignas] = ACTIONS(1274),
    [sym_primitive_type] = ACTIONS(1274),
    [anon_sym_enum] = ACTIONS(1274),
    [anon_sym_struct] = ACTIONS(1274),
    [anon_sym_union] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1274),
    [anon_sym_switch] = ACTIONS(1274),
    [anon_sym_case] = ACTIONS(1274),
    [anon_sym_default] = ACTIONS(1274),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_break] = ACTIONS(1274),
    [anon_sym_continue] = ACTIONS(1274),
    [anon_sym_goto] = ACTIONS(1274),
    [anon_sym_DASH_DASH] = ACTIONS(1276),
    [anon_sym_PLUS_PLUS] = ACTIONS(1276),
    [anon_sym_sizeof] = ACTIONS(1274),
    [anon_sym___alignof__] = ACTIONS(1274),
    [anon_sym___alignof] = ACTIONS(1274),
    [anon_sym__alignof] = ACTIONS(1274),
    [anon_sym_alignof] = ACTIONS(1274),
    [anon_sym__Alignof] = ACTIONS(1274),
    [anon_sym_offsetof] = ACTIONS(1274),
    [anon_sym__Generic] = ACTIONS(1274),
    [anon_sym_asm] = ACTIONS(1274),
    [anon_sym___asm__] = ACTIONS(1274),
    [anon_sym___asm] = ACTIONS(1274),
    [sym_number_literal] = ACTIONS(1276),
    [anon_sym_L_SQUOTE] = ACTIONS(1276),
    [anon_sym_u_SQUOTE] = ACTIONS(1276),
    [anon_sym_U_SQUOTE] = ACTIONS(1276),
    [anon_sym_u8_SQUOTE] = ACTIONS(1276),
    [anon_sym_SQUOTE] = ACTIONS(1276),
    [anon_sym_L_DQUOTE] = ACTIONS(1276),
    [anon_sym_u_DQUOTE] = ACTIONS(1276),
    [anon_sym_U_DQUOTE] = ACTIONS(1276),
    [anon_sym_u8_DQUOTE] = ACTIONS(1276),
    [anon_sym_DQUOTE] = ACTIONS(1276),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [anon_sym_NULL] = ACTIONS(1274),
    [anon_sym_nullptr] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(316)] = {
    [sym_attribute_declaration] = STATE(369),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(166),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [sym_identifier] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(317)] = {
    [ts_builtin_sym_end] = ACTIONS(1300),
    [sym_identifier] = ACTIONS(1298),
    [aux_sym_preproc_include_token1] = ACTIONS(1298),
    [aux_sym_preproc_def_token1] = ACTIONS(1298),
    [aux_sym_preproc_if_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1298),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1298),
    [sym_preproc_directive] = ACTIONS(1298),
    [anon_sym_LPAREN2] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1298),
    [anon_sym_PLUS] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym___extension__] = ACTIONS(1298),
    [anon_sym_typedef] = ACTIONS(1298),
    [anon_sym_extern] = ACTIONS(1298),
    [anon_sym___attribute__] = ACTIONS(1298),
    [anon_sym___attribute] = ACTIONS(1298),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1300),
    [anon_sym___declspec] = ACTIONS(1298),
    [anon_sym___cdecl] = ACTIONS(1298),
    [anon_sym___clrcall] = ACTIONS(1298),
    [anon_sym___stdcall] = ACTIONS(1298),
    [anon_sym___fastcall] = ACTIONS(1298),
    [anon_sym___thiscall] = ACTIONS(1298),
    [anon_sym___vectorcall] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_signed] = ACTIONS(1298),
    [anon_sym_unsigned] = ACTIONS(1298),
    [anon_sym_long] = ACTIONS(1298),
    [anon_sym_short] = ACTIONS(1298),
    [anon_sym_static] = ACTIONS(1298),
    [anon_sym_auto] = ACTIONS(1298),
    [anon_sym_register] = ACTIONS(1298),
    [anon_sym_inline] = ACTIONS(1298),
    [anon_sym___inline] = ACTIONS(1298),
    [anon_sym___inline__] = ACTIONS(1298),
    [anon_sym___forceinline] = ACTIONS(1298),
    [anon_sym_thread_local] = ACTIONS(1298),
    [anon_sym___thread] = ACTIONS(1298),
    [anon_sym_const] = ACTIONS(1298),
    [anon_sym_constexpr] = ACTIONS(1298),
    [anon_sym_volatile] = ACTIONS(1298),
    [anon_sym_restrict] = ACTIONS(1298),
    [anon_sym___restrict__] = ACTIONS(1298),
    [anon_sym__Atomic] = ACTIONS(1298),
    [anon_sym__Noreturn] = ACTIONS(1298),
    [anon_sym_noreturn] = ACTIONS(1298),
    [anon_sym__Nonnull] = ACTIONS(1298),
    [anon_sym_alignas] = ACTIONS(1298),
    [anon_sym__Alignas] = ACTIONS(1298),
    [sym_primitive_type] = ACTIONS(1298),
    [anon_sym_enum] = ACTIONS(1298),
    [anon_sym_struct] = ACTIONS(1298),
    [anon_sym_union] = ACTIONS(1298),
    [anon_sym_if] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1298),
    [anon_sym_case] = ACTIONS(1298),
    [anon_sym_default] = ACTIONS(1298),
    [anon_sym_while] = ACTIONS(1298),
    [anon_sym_do] = ACTIONS(1298),
    [anon_sym_for] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1298),
    [anon_sym_break] = ACTIONS(1298),
    [anon_sym_continue] = ACTIONS(1298),
    [anon_sym_goto] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_sizeof] = ACTIONS(1298),
    [anon_sym___alignof__] = ACTIONS(1298),
    [anon_sym___alignof] = ACTIONS(1298),
    [anon_sym__alignof] = ACTIONS(1298),
    [anon_sym_alignof] = ACTIONS(1298),
    [anon_sym__Alignof] = ACTIONS(1298),
    [anon_sym_offsetof] = ACTIONS(1298),
    [anon_sym__Generic] = ACTIONS(1298),
    [anon_sym_asm] = ACTIONS(1298),
    [anon_sym___asm__] = ACTIONS(1298),
    [anon_sym___asm] = ACTIONS(1298),
    [sym_number_literal] = ACTIONS(1300),
    [anon_sym_L_SQUOTE] = ACTIONS(1300),
    [anon_sym_u_SQUOTE] = ACTIONS(1300),
    [anon_sym_U_SQUOTE] = ACTIONS(1300),
    [anon_sym_u8_SQUOTE] = ACTIONS(1300),
    [anon_sym_SQUOTE] = ACTIONS(1300),
    [anon_sym_L_DQUOTE] = ACTIONS(1300),
    [anon_sym_u_DQUOTE] = ACTIONS(1300),
    [anon_sym_U_DQUOTE] = ACTIONS(1300),
    [anon_sym_u8_DQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1300),
    [sym_true] = ACTIONS(1298),
    [sym_false] = ACTIONS(1298),
    [anon_sym_NULL] = ACTIONS(1298),
    [anon_sym_nullptr] = ACTIONS(1298),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(318)] = {
    [ts_builtin_sym_end] = ACTIONS(1296),
    [sym_identifier] = ACTIONS(1294),
    [aux_sym_preproc_include_token1] = ACTIONS(1294),
    [aux_sym_preproc_def_token1] = ACTIONS(1294),
    [aux_sym_preproc_if_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1294),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1294),
    [sym_preproc_directive] = ACTIONS(1294),
    [anon_sym_LPAREN2] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1294),
    [anon_sym_PLUS] = ACTIONS(1294),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym___extension__] = ACTIONS(1294),
    [anon_sym_typedef] = ACTIONS(1294),
    [anon_sym_extern] = ACTIONS(1294),
    [anon_sym___attribute__] = ACTIONS(1294),
    [anon_sym___attribute] = ACTIONS(1294),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1296),
    [anon_sym___declspec] = ACTIONS(1294),
    [anon_sym___cdecl] = ACTIONS(1294),
    [anon_sym___clrcall] = ACTIONS(1294),
    [anon_sym___stdcall] = ACTIONS(1294),
    [anon_sym___fastcall] = ACTIONS(1294),
    [anon_sym___thiscall] = ACTIONS(1294),
    [anon_sym___vectorcall] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_signed] = ACTIONS(1294),
    [anon_sym_unsigned] = ACTIONS(1294),
    [anon_sym_long] = ACTIONS(1294),
    [anon_sym_short] = ACTIONS(1294),
    [anon_sym_static] = ACTIONS(1294),
    [anon_sym_auto] = ACTIONS(1294),
    [anon_sym_register] = ACTIONS(1294),
    [anon_sym_inline] = ACTIONS(1294),
    [anon_sym___inline] = ACTIONS(1294),
    [anon_sym___inline__] = ACTIONS(1294),
    [anon_sym___forceinline] = ACTIONS(1294),
    [anon_sym_thread_local] = ACTIONS(1294),
    [anon_sym___thread] = ACTIONS(1294),
    [anon_sym_const] = ACTIONS(1294),
    [anon_sym_constexpr] = ACTIONS(1294),
    [anon_sym_volatile] = ACTIONS(1294),
    [anon_sym_restrict] = ACTIONS(1294),
    [anon_sym___restrict__] = ACTIONS(1294),
    [anon_sym__Atomic] = ACTIONS(1294),
    [anon_sym__Noreturn] = ACTIONS(1294),
    [anon_sym_noreturn] = ACTIONS(1294),
    [anon_sym__Nonnull] = ACTIONS(1294),
    [anon_sym_alignas] = ACTIONS(1294),
    [anon_sym__Alignas] = ACTIONS(1294),
    [sym_primitive_type] = ACTIONS(1294),
    [anon_sym_enum] = ACTIONS(1294),
    [anon_sym_struct] = ACTIONS(1294),
    [anon_sym_union] = ACTIONS(1294),
    [anon_sym_if] = ACTIONS(1294),
    [anon_sym_switch] = ACTIONS(1294),
    [anon_sym_case] = ACTIONS(1294),
    [anon_sym_default] = ACTIONS(1294),
    [anon_sym_while] = ACTIONS(1294),
    [anon_sym_do] = ACTIONS(1294),
    [anon_sym_for] = ACTIONS(1294),
    [anon_sym_return] = ACTIONS(1294),
    [anon_sym_break] = ACTIONS(1294),
    [anon_sym_continue] = ACTIONS(1294),
    [anon_sym_goto] = ACTIONS(1294),
    [anon_sym_DASH_DASH] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1296),
    [anon_sym_sizeof] = ACTIONS(1294),
    [anon_sym___alignof__] = ACTIONS(1294),
    [anon_sym___alignof] = ACTIONS(1294),
    [anon_sym__alignof] = ACTIONS(1294),
    [anon_sym_alignof] = ACTIONS(1294),
    [anon_sym__Alignof] = ACTIONS(1294),
    [anon_sym_offsetof] = ACTIONS(1294),
    [anon_sym__Generic] = ACTIONS(1294),
    [anon_sym_asm] = ACTIONS(1294),
    [anon_sym___asm__] = ACTIONS(1294),
    [anon_sym___asm] = ACTIONS(1294),
    [sym_number_literal] = ACTIONS(1296),
    [anon_sym_L_SQUOTE] = ACTIONS(1296),
    [anon_sym_u_SQUOTE] = ACTIONS(1296),
    [anon_sym_U_SQUOTE] = ACTIONS(1296),
    [anon_sym_u8_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_L_DQUOTE] = ACTIONS(1296),
    [anon_sym_u_DQUOTE] = ACTIONS(1296),
    [anon_sym_U_DQUOTE] = ACTIONS(1296),
    [anon_sym_u8_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [sym_true] = ACTIONS(1294),
    [sym_false] = ACTIONS(1294),
    [anon_sym_NULL] = ACTIONS(1294),
    [anon_sym_nullptr] = ACTIONS(1294),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(319)] = {
    [ts_builtin_sym_end] = ACTIONS(1264),
    [sym_identifier] = ACTIONS(1262),
    [aux_sym_preproc_include_token1] = ACTIONS(1262),
    [aux_sym_preproc_def_token1] = ACTIONS(1262),
    [aux_sym_preproc_if_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1262),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1262),
    [sym_preproc_directive] = ACTIONS(1262),
    [anon_sym_LPAREN2] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_TILDE] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_STAR] = ACTIONS(1264),
    [anon_sym_AMP] = ACTIONS(1264),
    [anon_sym_SEMI] = ACTIONS(1264),
    [anon_sym___extension__] = ACTIONS(1262),
    [anon_sym_typedef] = ACTIONS(1262),
    [anon_sym_extern] = ACTIONS(1262),
    [anon_sym___attribute__] = ACTIONS(1262),
    [anon_sym___attribute] = ACTIONS(1262),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1264),
    [anon_sym___declspec] = ACTIONS(1262),
    [anon_sym___cdecl] = ACTIONS(1262),
    [anon_sym___clrcall] = ACTIONS(1262),
    [anon_sym___stdcall] = ACTIONS(1262),
    [anon_sym___fastcall] = ACTIONS(1262),
    [anon_sym___thiscall] = ACTIONS(1262),
    [anon_sym___vectorcall] = ACTIONS(1262),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_signed] = ACTIONS(1262),
    [anon_sym_unsigned] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_auto] = ACTIONS(1262),
    [anon_sym_register] = ACTIONS(1262),
    [anon_sym_inline] = ACTIONS(1262),
    [anon_sym___inline] = ACTIONS(1262),
    [anon_sym___inline__] = ACTIONS(1262),
    [anon_sym___forceinline] = ACTIONS(1262),
    [anon_sym_thread_local] = ACTIONS(1262),
    [anon_sym___thread] = ACTIONS(1262),
    [anon_sym_const] = ACTIONS(1262),
    [anon_sym_constexpr] = ACTIONS(1262),
    [anon_sym_volatile] = ACTIONS(1262),
    [anon_sym_restrict] = ACTIONS(1262),
    [anon_sym___restrict__] = ACTIONS(1262),
    [anon_sym__Atomic] = ACTIONS(1262),
    [anon_sym__Noreturn] = ACTIONS(1262),
    [anon_sym_noreturn] = ACTIONS(1262),
    [anon_sym__Nonnull] = ACTIONS(1262),
    [anon_sym_alignas] = ACTIONS(1262),
    [anon_sym__Alignas] = ACTIONS(1262),
    [sym_primitive_type] = ACTIONS(1262),
    [anon_sym_enum] = ACTIONS(1262),
    [anon_sym_struct] = ACTIONS(1262),
    [anon_sym_union] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_goto] = ACTIONS(1262),
    [anon_sym_DASH_DASH] = ACTIONS(1264),
    [anon_sym_PLUS_PLUS] = ACTIONS(1264),
    [anon_sym_sizeof] = ACTIONS(1262),
    [anon_sym___alignof__] = ACTIONS(1262),
    [anon_sym___alignof] = ACTIONS(1262),
    [anon_sym__alignof] = ACTIONS(1262),
    [anon_sym_alignof] = ACTIONS(1262),
    [anon_sym__Alignof] = ACTIONS(1262),
    [anon_sym_offsetof] = ACTIONS(1262),
    [anon_sym__Generic] = ACTIONS(1262),
    [anon_sym_asm] = ACTIONS(1262),
    [anon_sym___asm__] = ACTIONS(1262),
    [anon_sym___asm] = ACTIONS(1262),
    [sym_number_literal] = ACTIONS(1264),
    [anon_sym_L_SQUOTE] = ACTIONS(1264),
    [anon_sym_u_SQUOTE] = ACTIONS(1264),
    [anon_sym_U_SQUOTE] = ACTIONS(1264),
    [anon_sym_u8_SQUOTE] = ACTIONS(1264),
    [anon_sym_SQUOTE] = ACTIONS(1264),
    [anon_sym_L_DQUOTE] = ACTIONS(1264),
    [anon_sym_u_DQUOTE] = ACTIONS(1264),
    [anon_sym_U_DQUOTE] = ACTIONS(1264),
    [anon_sym_u8_DQUOTE] = ACTIONS(1264),
    [anon_sym_DQUOTE] = ACTIONS(1264),
    [sym_true] = ACTIONS(1262),
    [sym_false] = ACTIONS(1262),
    [anon_sym_NULL] = ACTIONS(1262),
    [anon_sym_nullptr] = ACTIONS(1262),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(320)] = {
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_identifier] = ACTIONS(1412),
    [aux_sym_preproc_include_token1] = ACTIONS(1412),
    [aux_sym_preproc_def_token1] = ACTIONS(1412),
    [aux_sym_preproc_if_token1] = ACTIONS(1412),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1412),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1412),
    [sym_preproc_directive] = ACTIONS(1412),
    [anon_sym_LPAREN2] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_STAR] = ACTIONS(1410),
    [anon_sym_AMP] = ACTIONS(1410),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym___extension__] = ACTIONS(1412),
    [anon_sym_typedef] = ACTIONS(1412),
    [anon_sym_extern] = ACTIONS(1412),
    [anon_sym___attribute__] = ACTIONS(1412),
    [anon_sym___attribute] = ACTIONS(1412),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1410),
    [anon_sym___declspec] = ACTIONS(1412),
    [anon_sym___cdecl] = ACTIONS(1412),
    [anon_sym___clrcall] = ACTIONS(1412),
    [anon_sym___stdcall] = ACTIONS(1412),
    [anon_sym___fastcall] = ACTIONS(1412),
    [anon_sym___thiscall] = ACTIONS(1412),
    [anon_sym___vectorcall] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_signed] = ACTIONS(1412),
    [anon_sym_unsigned] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_auto] = ACTIONS(1412),
    [anon_sym_register] = ACTIONS(1412),
    [anon_sym_inline] = ACTIONS(1412),
    [anon_sym___inline] = ACTIONS(1412),
    [anon_sym___inline__] = ACTIONS(1412),
    [anon_sym___forceinline] = ACTIONS(1412),
    [anon_sym_thread_local] = ACTIONS(1412),
    [anon_sym___thread] = ACTIONS(1412),
    [anon_sym_const] = ACTIONS(1412),
    [anon_sym_constexpr] = ACTIONS(1412),
    [anon_sym_volatile] = ACTIONS(1412),
    [anon_sym_restrict] = ACTIONS(1412),
    [anon_sym___restrict__] = ACTIONS(1412),
    [anon_sym__Atomic] = ACTIONS(1412),
    [anon_sym__Noreturn] = ACTIONS(1412),
    [anon_sym_noreturn] = ACTIONS(1412),
    [anon_sym__Nonnull] = ACTIONS(1412),
    [anon_sym_alignas] = ACTIONS(1412),
    [anon_sym__Alignas] = ACTIONS(1412),
    [sym_primitive_type] = ACTIONS(1412),
    [anon_sym_enum] = ACTIONS(1412),
    [anon_sym_struct] = ACTIONS(1412),
    [anon_sym_union] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_case] = ACTIONS(1412),
    [anon_sym_default] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_goto] = ACTIONS(1412),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_sizeof] = ACTIONS(1412),
    [anon_sym___alignof__] = ACTIONS(1412),
    [anon_sym___alignof] = ACTIONS(1412),
    [anon_sym__alignof] = ACTIONS(1412),
    [anon_sym_alignof] = ACTIONS(1412),
    [anon_sym__Alignof] = ACTIONS(1412),
    [anon_sym_offsetof] = ACTIONS(1412),
    [anon_sym__Generic] = ACTIONS(1412),
    [anon_sym_asm] = ACTIONS(1412),
    [anon_sym___asm__] = ACTIONS(1412),
    [anon_sym___asm] = ACTIONS(1412),
    [sym_number_literal] = ACTIONS(1410),
    [anon_sym_L_SQUOTE] = ACTIONS(1410),
    [anon_sym_u_SQUOTE] = ACTIONS(1410),
    [anon_sym_U_SQUOTE] = ACTIONS(1410),
    [anon_sym_u8_SQUOTE] = ACTIONS(1410),
    [anon_sym_SQUOTE] = ACTIONS(1410),
    [anon_sym_L_DQUOTE] = ACTIONS(1410),
    [anon_sym_u_DQUOTE] = ACTIONS(1410),
    [anon_sym_U_DQUOTE] = ACTIONS(1410),
    [anon_sym_u8_DQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1410),
    [sym_true] = ACTIONS(1412),
    [sym_false] = ACTIONS(1412),
    [anon_sym_NULL] = ACTIONS(1412),
    [anon_sym_nullptr] = ACTIONS(1412),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(321)] = {
    [sym_attribute_declaration] = STATE(369),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(155),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [sym_identifier] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(322)] = {
    [ts_builtin_sym_end] = ACTIONS(1414),
    [sym_identifier] = ACTIONS(1416),
    [aux_sym_preproc_include_token1] = ACTIONS(1416),
    [aux_sym_preproc_def_token1] = ACTIONS(1416),
    [aux_sym_preproc_if_token1] = ACTIONS(1416),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1416),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1416),
    [sym_preproc_directive] = ACTIONS(1416),
    [anon_sym_LPAREN2] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1414),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_STAR] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1414),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym___extension__] = ACTIONS(1416),
    [anon_sym_typedef] = ACTIONS(1416),
    [anon_sym_extern] = ACTIONS(1416),
    [anon_sym___attribute__] = ACTIONS(1416),
    [anon_sym___attribute] = ACTIONS(1416),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1414),
    [anon_sym___declspec] = ACTIONS(1416),
    [anon_sym___cdecl] = ACTIONS(1416),
    [anon_sym___clrcall] = ACTIONS(1416),
    [anon_sym___stdcall] = ACTIONS(1416),
    [anon_sym___fastcall] = ACTIONS(1416),
    [anon_sym___thiscall] = ACTIONS(1416),
    [anon_sym___vectorcall] = ACTIONS(1416),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_signed] = ACTIONS(1416),
    [anon_sym_unsigned] = ACTIONS(1416),
    [anon_sym_long] = ACTIONS(1416),
    [anon_sym_short] = ACTIONS(1416),
    [anon_sym_static] = ACTIONS(1416),
    [anon_sym_auto] = ACTIONS(1416),
    [anon_sym_register] = ACTIONS(1416),
    [anon_sym_inline] = ACTIONS(1416),
    [anon_sym___inline] = ACTIONS(1416),
    [anon_sym___inline__] = ACTIONS(1416),
    [anon_sym___forceinline] = ACTIONS(1416),
    [anon_sym_thread_local] = ACTIONS(1416),
    [anon_sym___thread] = ACTIONS(1416),
    [anon_sym_const] = ACTIONS(1416),
    [anon_sym_constexpr] = ACTIONS(1416),
    [anon_sym_volatile] = ACTIONS(1416),
    [anon_sym_restrict] = ACTIONS(1416),
    [anon_sym___restrict__] = ACTIONS(1416),
    [anon_sym__Atomic] = ACTIONS(1416),
    [anon_sym__Noreturn] = ACTIONS(1416),
    [anon_sym_noreturn] = ACTIONS(1416),
    [anon_sym__Nonnull] = ACTIONS(1416),
    [anon_sym_alignas] = ACTIONS(1416),
    [anon_sym__Alignas] = ACTIONS(1416),
    [sym_primitive_type] = ACTIONS(1416),
    [anon_sym_enum] = ACTIONS(1416),
    [anon_sym_struct] = ACTIONS(1416),
    [anon_sym_union] = ACTIONS(1416),
    [anon_sym_if] = ACTIONS(1416),
    [anon_sym_switch] = ACTIONS(1416),
    [anon_sym_case] = ACTIONS(1416),
    [anon_sym_default] = ACTIONS(1416),
    [anon_sym_while] = ACTIONS(1416),
    [anon_sym_do] = ACTIONS(1416),
    [anon_sym_for] = ACTIONS(1416),
    [anon_sym_return] = ACTIONS(1416),
    [anon_sym_break] = ACTIONS(1416),
    [anon_sym_continue] = ACTIONS(1416),
    [anon_sym_goto] = ACTIONS(1416),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [anon_sym_sizeof] = ACTIONS(1416),
    [anon_sym___alignof__] = ACTIONS(1416),
    [anon_sym___alignof] = ACTIONS(1416),
    [anon_sym__alignof] = ACTIONS(1416),
    [anon_sym_alignof] = ACTIONS(1416),
    [anon_sym__Alignof] = ACTIONS(1416),
    [anon_sym_offsetof] = ACTIONS(1416),
    [anon_sym__Generic] = ACTIONS(1416),
    [anon_sym_asm] = ACTIONS(1416),
    [anon_sym___asm__] = ACTIONS(1416),
    [anon_sym___asm] = ACTIONS(1416),
    [sym_number_literal] = ACTIONS(1414),
    [anon_sym_L_SQUOTE] = ACTIONS(1414),
    [anon_sym_u_SQUOTE] = ACTIONS(1414),
    [anon_sym_U_SQUOTE] = ACTIONS(1414),
    [anon_sym_u8_SQUOTE] = ACTIONS(1414),
    [anon_sym_SQUOTE] = ACTIONS(1414),
    [anon_sym_L_DQUOTE] = ACTIONS(1414),
    [anon_sym_u_DQUOTE] = ACTIONS(1414),
    [anon_sym_U_DQUOTE] = ACTIONS(1414),
    [anon_sym_u8_DQUOTE] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1414),
    [sym_true] = ACTIONS(1416),
    [sym_false] = ACTIONS(1416),
    [anon_sym_NULL] = ACTIONS(1416),
    [anon_sym_nullptr] = ACTIONS(1416),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(323)] = {
    [sym_attribute_declaration] = STATE(369),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(158),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [sym_identifier] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(324)] = {
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1302),
    [aux_sym_preproc_include_token1] = ACTIONS(1302),
    [aux_sym_preproc_def_token1] = ACTIONS(1302),
    [aux_sym_preproc_if_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1302),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1302),
    [sym_preproc_directive] = ACTIONS(1302),
    [anon_sym_LPAREN2] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym___extension__] = ACTIONS(1302),
    [anon_sym_typedef] = ACTIONS(1302),
    [anon_sym_extern] = ACTIONS(1302),
    [anon_sym___attribute__] = ACTIONS(1302),
    [anon_sym___attribute] = ACTIONS(1302),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1304),
    [anon_sym___declspec] = ACTIONS(1302),
    [anon_sym___cdecl] = ACTIONS(1302),
    [anon_sym___clrcall] = ACTIONS(1302),
    [anon_sym___stdcall] = ACTIONS(1302),
    [anon_sym___fastcall] = ACTIONS(1302),
    [anon_sym___thiscall] = ACTIONS(1302),
    [anon_sym___vectorcall] = ACTIONS(1302),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_signed] = ACTIONS(1302),
    [anon_sym_unsigned] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_auto] = ACTIONS(1302),
    [anon_sym_register] = ACTIONS(1302),
    [anon_sym_inline] = ACTIONS(1302),
    [anon_sym___inline] = ACTIONS(1302),
    [anon_sym___inline__] = ACTIONS(1302),
    [anon_sym___forceinline] = ACTIONS(1302),
    [anon_sym_thread_local] = ACTIONS(1302),
    [anon_sym___thread] = ACTIONS(1302),
    [anon_sym_const] = ACTIONS(1302),
    [anon_sym_constexpr] = ACTIONS(1302),
    [anon_sym_volatile] = ACTIONS(1302),
    [anon_sym_restrict] = ACTIONS(1302),
    [anon_sym___restrict__] = ACTIONS(1302),
    [anon_sym__Atomic] = ACTIONS(1302),
    [anon_sym__Noreturn] = ACTIONS(1302),
    [anon_sym_noreturn] = ACTIONS(1302),
    [anon_sym__Nonnull] = ACTIONS(1302),
    [anon_sym_alignas] = ACTIONS(1302),
    [anon_sym__Alignas] = ACTIONS(1302),
    [sym_primitive_type] = ACTIONS(1302),
    [anon_sym_enum] = ACTIONS(1302),
    [anon_sym_struct] = ACTIONS(1302),
    [anon_sym_union] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_goto] = ACTIONS(1302),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_sizeof] = ACTIONS(1302),
    [anon_sym___alignof__] = ACTIONS(1302),
    [anon_sym___alignof] = ACTIONS(1302),
    [anon_sym__alignof] = ACTIONS(1302),
    [anon_sym_alignof] = ACTIONS(1302),
    [anon_sym__Alignof] = ACTIONS(1302),
    [anon_sym_offsetof] = ACTIONS(1302),
    [anon_sym__Generic] = ACTIONS(1302),
    [anon_sym_asm] = ACTIONS(1302),
    [anon_sym___asm__] = ACTIONS(1302),
    [anon_sym___asm] = ACTIONS(1302),
    [sym_number_literal] = ACTIONS(1304),
    [anon_sym_L_SQUOTE] = ACTIONS(1304),
    [anon_sym_u_SQUOTE] = ACTIONS(1304),
    [anon_sym_U_SQUOTE] = ACTIONS(1304),
    [anon_sym_u8_SQUOTE] = ACTIONS(1304),
    [anon_sym_SQUOTE] = ACTIONS(1304),
    [anon_sym_L_DQUOTE] = ACTIONS(1304),
    [anon_sym_u_DQUOTE] = ACTIONS(1304),
    [anon_sym_U_DQUOTE] = ACTIONS(1304),
    [anon_sym_u8_DQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1304),
    [sym_true] = ACTIONS(1302),
    [sym_false] = ACTIONS(1302),
    [anon_sym_NULL] = ACTIONS(1302),
    [anon_sym_nullptr] = ACTIONS(1302),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(325)] = {
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1306),
    [aux_sym_preproc_include_token1] = ACTIONS(1306),
    [aux_sym_preproc_def_token1] = ACTIONS(1306),
    [aux_sym_preproc_if_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1306),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1306),
    [sym_preproc_directive] = ACTIONS(1306),
    [anon_sym_LPAREN2] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym___extension__] = ACTIONS(1306),
    [anon_sym_typedef] = ACTIONS(1306),
    [anon_sym_extern] = ACTIONS(1306),
    [anon_sym___attribute__] = ACTIONS(1306),
    [anon_sym___attribute] = ACTIONS(1306),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1308),
    [anon_sym___declspec] = ACTIONS(1306),
    [anon_sym___cdecl] = ACTIONS(1306),
    [anon_sym___clrcall] = ACTIONS(1306),
    [anon_sym___stdcall] = ACTIONS(1306),
    [anon_sym___fastcall] = ACTIONS(1306),
    [anon_sym___thiscall] = ACTIONS(1306),
    [anon_sym___vectorcall] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_signed] = ACTIONS(1306),
    [anon_sym_unsigned] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_auto] = ACTIONS(1306),
    [anon_sym_register] = ACTIONS(1306),
    [anon_sym_inline] = ACTIONS(1306),
    [anon_sym___inline] = ACTIONS(1306),
    [anon_sym___inline__] = ACTIONS(1306),
    [anon_sym___forceinline] = ACTIONS(1306),
    [anon_sym_thread_local] = ACTIONS(1306),
    [anon_sym___thread] = ACTIONS(1306),
    [anon_sym_const] = ACTIONS(1306),
    [anon_sym_constexpr] = ACTIONS(1306),
    [anon_sym_volatile] = ACTIONS(1306),
    [anon_sym_restrict] = ACTIONS(1306),
    [anon_sym___restrict__] = ACTIONS(1306),
    [anon_sym__Atomic] = ACTIONS(1306),
    [anon_sym__Noreturn] = ACTIONS(1306),
    [anon_sym_noreturn] = ACTIONS(1306),
    [anon_sym__Nonnull] = ACTIONS(1306),
    [anon_sym_alignas] = ACTIONS(1306),
    [anon_sym__Alignas] = ACTIONS(1306),
    [sym_primitive_type] = ACTIONS(1306),
    [anon_sym_enum] = ACTIONS(1306),
    [anon_sym_struct] = ACTIONS(1306),
    [anon_sym_union] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_goto] = ACTIONS(1306),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_sizeof] = ACTIONS(1306),
    [anon_sym___alignof__] = ACTIONS(1306),
    [anon_sym___alignof] = ACTIONS(1306),
    [anon_sym__alignof] = ACTIONS(1306),
    [anon_sym_alignof] = ACTIONS(1306),
    [anon_sym__Alignof] = ACTIONS(1306),
    [anon_sym_offsetof] = ACTIONS(1306),
    [anon_sym__Generic] = ACTIONS(1306),
    [anon_sym_asm] = ACTIONS(1306),
    [anon_sym___asm__] = ACTIONS(1306),
    [anon_sym___asm] = ACTIONS(1306),
    [sym_number_literal] = ACTIONS(1308),
    [anon_sym_L_SQUOTE] = ACTIONS(1308),
    [anon_sym_u_SQUOTE] = ACTIONS(1308),
    [anon_sym_U_SQUOTE] = ACTIONS(1308),
    [anon_sym_u8_SQUOTE] = ACTIONS(1308),
    [anon_sym_SQUOTE] = ACTIONS(1308),
    [anon_sym_L_DQUOTE] = ACTIONS(1308),
    [anon_sym_u_DQUOTE] = ACTIONS(1308),
    [anon_sym_U_DQUOTE] = ACTIONS(1308),
    [anon_sym_u8_DQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1308),
    [sym_true] = ACTIONS(1306),
    [sym_false] = ACTIONS(1306),
    [anon_sym_NULL] = ACTIONS(1306),
    [anon_sym_nullptr] = ACTIONS(1306),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(326)] = {
    [ts_builtin_sym_end] = ACTIONS(1316),
    [sym_identifier] = ACTIONS(1314),
    [aux_sym_preproc_include_token1] = ACTIONS(1314),
    [aux_sym_preproc_def_token1] = ACTIONS(1314),
    [aux_sym_preproc_if_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1314),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1314),
    [sym_preproc_directive] = ACTIONS(1314),
    [anon_sym_LPAREN2] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1314),
    [anon_sym_PLUS] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym___extension__] = ACTIONS(1314),
    [anon_sym_typedef] = ACTIONS(1314),
    [anon_sym_extern] = ACTIONS(1314),
    [anon_sym___attribute__] = ACTIONS(1314),
    [anon_sym___attribute] = ACTIONS(1314),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1316),
    [anon_sym___declspec] = ACTIONS(1314),
    [anon_sym___cdecl] = ACTIONS(1314),
    [anon_sym___clrcall] = ACTIONS(1314),
    [anon_sym___stdcall] = ACTIONS(1314),
    [anon_sym___fastcall] = ACTIONS(1314),
    [anon_sym___thiscall] = ACTIONS(1314),
    [anon_sym___vectorcall] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_signed] = ACTIONS(1314),
    [anon_sym_unsigned] = ACTIONS(1314),
    [anon_sym_long] = ACTIONS(1314),
    [anon_sym_short] = ACTIONS(1314),
    [anon_sym_static] = ACTIONS(1314),
    [anon_sym_auto] = ACTIONS(1314),
    [anon_sym_register] = ACTIONS(1314),
    [anon_sym_inline] = ACTIONS(1314),
    [anon_sym___inline] = ACTIONS(1314),
    [anon_sym___inline__] = ACTIONS(1314),
    [anon_sym___forceinline] = ACTIONS(1314),
    [anon_sym_thread_local] = ACTIONS(1314),
    [anon_sym___thread] = ACTIONS(1314),
    [anon_sym_const] = ACTIONS(1314),
    [anon_sym_constexpr] = ACTIONS(1314),
    [anon_sym_volatile] = ACTIONS(1314),
    [anon_sym_restrict] = ACTIONS(1314),
    [anon_sym___restrict__] = ACTIONS(1314),
    [anon_sym__Atomic] = ACTIONS(1314),
    [anon_sym__Noreturn] = ACTIONS(1314),
    [anon_sym_noreturn] = ACTIONS(1314),
    [anon_sym__Nonnull] = ACTIONS(1314),
    [anon_sym_alignas] = ACTIONS(1314),
    [anon_sym__Alignas] = ACTIONS(1314),
    [sym_primitive_type] = ACTIONS(1314),
    [anon_sym_enum] = ACTIONS(1314),
    [anon_sym_struct] = ACTIONS(1314),
    [anon_sym_union] = ACTIONS(1314),
    [anon_sym_if] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1314),
    [anon_sym_case] = ACTIONS(1314),
    [anon_sym_default] = ACTIONS(1314),
    [anon_sym_while] = ACTIONS(1314),
    [anon_sym_do] = ACTIONS(1314),
    [anon_sym_for] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1314),
    [anon_sym_break] = ACTIONS(1314),
    [anon_sym_continue] = ACTIONS(1314),
    [anon_sym_goto] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1316),
    [anon_sym_sizeof] = ACTIONS(1314),
    [anon_sym___alignof__] = ACTIONS(1314),
    [anon_sym___alignof] = ACTIONS(1314),
    [anon_sym__alignof] = ACTIONS(1314),
    [anon_sym_alignof] = ACTIONS(1314),
    [anon_sym__Alignof] = ACTIONS(1314),
    [anon_sym_offsetof] = ACTIONS(1314),
    [anon_sym__Generic] = ACTIONS(1314),
    [anon_sym_asm] = ACTIONS(1314),
    [anon_sym___asm__] = ACTIONS(1314),
    [anon_sym___asm] = ACTIONS(1314),
    [sym_number_literal] = ACTIONS(1316),
    [anon_sym_L_SQUOTE] = ACTIONS(1316),
    [anon_sym_u_SQUOTE] = ACTIONS(1316),
    [anon_sym_U_SQUOTE] = ACTIONS(1316),
    [anon_sym_u8_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_L_DQUOTE] = ACTIONS(1316),
    [anon_sym_u_DQUOTE] = ACTIONS(1316),
    [anon_sym_U_DQUOTE] = ACTIONS(1316),
    [anon_sym_u8_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [sym_true] = ACTIONS(1314),
    [sym_false] = ACTIONS(1314),
    [anon_sym_NULL] = ACTIONS(1314),
    [anon_sym_nullptr] = ACTIONS(1314),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(327)] = {
    [ts_builtin_sym_end] = ACTIONS(1320),
    [sym_identifier] = ACTIONS(1318),
    [aux_sym_preproc_include_token1] = ACTIONS(1318),
    [aux_sym_preproc_def_token1] = ACTIONS(1318),
    [aux_sym_preproc_if_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1318),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1318),
    [sym_preproc_directive] = ACTIONS(1318),
    [anon_sym_LPAREN2] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1318),
    [anon_sym_PLUS] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym___extension__] = ACTIONS(1318),
    [anon_sym_typedef] = ACTIONS(1318),
    [anon_sym_extern] = ACTIONS(1318),
    [anon_sym___attribute__] = ACTIONS(1318),
    [anon_sym___attribute] = ACTIONS(1318),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1320),
    [anon_sym___declspec] = ACTIONS(1318),
    [anon_sym___cdecl] = ACTIONS(1318),
    [anon_sym___clrcall] = ACTIONS(1318),
    [anon_sym___stdcall] = ACTIONS(1318),
    [anon_sym___fastcall] = ACTIONS(1318),
    [anon_sym___thiscall] = ACTIONS(1318),
    [anon_sym___vectorcall] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_signed] = ACTIONS(1318),
    [anon_sym_unsigned] = ACTIONS(1318),
    [anon_sym_long] = ACTIONS(1318),
    [anon_sym_short] = ACTIONS(1318),
    [anon_sym_static] = ACTIONS(1318),
    [anon_sym_auto] = ACTIONS(1318),
    [anon_sym_register] = ACTIONS(1318),
    [anon_sym_inline] = ACTIONS(1318),
    [anon_sym___inline] = ACTIONS(1318),
    [anon_sym___inline__] = ACTIONS(1318),
    [anon_sym___forceinline] = ACTIONS(1318),
    [anon_sym_thread_local] = ACTIONS(1318),
    [anon_sym___thread] = ACTIONS(1318),
    [anon_sym_const] = ACTIONS(1318),
    [anon_sym_constexpr] = ACTIONS(1318),
    [anon_sym_volatile] = ACTIONS(1318),
    [anon_sym_restrict] = ACTIONS(1318),
    [anon_sym___restrict__] = ACTIONS(1318),
    [anon_sym__Atomic] = ACTIONS(1318),
    [anon_sym__Noreturn] = ACTIONS(1318),
    [anon_sym_noreturn] = ACTIONS(1318),
    [anon_sym__Nonnull] = ACTIONS(1318),
    [anon_sym_alignas] = ACTIONS(1318),
    [anon_sym__Alignas] = ACTIONS(1318),
    [sym_primitive_type] = ACTIONS(1318),
    [anon_sym_enum] = ACTIONS(1318),
    [anon_sym_struct] = ACTIONS(1318),
    [anon_sym_union] = ACTIONS(1318),
    [anon_sym_if] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1318),
    [anon_sym_case] = ACTIONS(1318),
    [anon_sym_default] = ACTIONS(1318),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_break] = ACTIONS(1318),
    [anon_sym_continue] = ACTIONS(1318),
    [anon_sym_goto] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1320),
    [anon_sym_sizeof] = ACTIONS(1318),
    [anon_sym___alignof__] = ACTIONS(1318),
    [anon_sym___alignof] = ACTIONS(1318),
    [anon_sym__alignof] = ACTIONS(1318),
    [anon_sym_alignof] = ACTIONS(1318),
    [anon_sym__Alignof] = ACTIONS(1318),
    [anon_sym_offsetof] = ACTIONS(1318),
    [anon_sym__Generic] = ACTIONS(1318),
    [anon_sym_asm] = ACTIONS(1318),
    [anon_sym___asm__] = ACTIONS(1318),
    [anon_sym___asm] = ACTIONS(1318),
    [sym_number_literal] = ACTIONS(1320),
    [anon_sym_L_SQUOTE] = ACTIONS(1320),
    [anon_sym_u_SQUOTE] = ACTIONS(1320),
    [anon_sym_U_SQUOTE] = ACTIONS(1320),
    [anon_sym_u8_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_L_DQUOTE] = ACTIONS(1320),
    [anon_sym_u_DQUOTE] = ACTIONS(1320),
    [anon_sym_U_DQUOTE] = ACTIONS(1320),
    [anon_sym_u8_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [anon_sym_NULL] = ACTIONS(1318),
    [anon_sym_nullptr] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(328)] = {
    [ts_builtin_sym_end] = ACTIONS(1324),
    [sym_identifier] = ACTIONS(1322),
    [aux_sym_preproc_include_token1] = ACTIONS(1322),
    [aux_sym_preproc_def_token1] = ACTIONS(1322),
    [aux_sym_preproc_if_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1322),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1322),
    [sym_preproc_directive] = ACTIONS(1322),
    [anon_sym_LPAREN2] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1322),
    [anon_sym_PLUS] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym___extension__] = ACTIONS(1322),
    [anon_sym_typedef] = ACTIONS(1322),
    [anon_sym_extern] = ACTIONS(1322),
    [anon_sym___attribute__] = ACTIONS(1322),
    [anon_sym___attribute] = ACTIONS(1322),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1324),
    [anon_sym___declspec] = ACTIONS(1322),
    [anon_sym___cdecl] = ACTIONS(1322),
    [anon_sym___clrcall] = ACTIONS(1322),
    [anon_sym___stdcall] = ACTIONS(1322),
    [anon_sym___fastcall] = ACTIONS(1322),
    [anon_sym___thiscall] = ACTIONS(1322),
    [anon_sym___vectorcall] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_signed] = ACTIONS(1322),
    [anon_sym_unsigned] = ACTIONS(1322),
    [anon_sym_long] = ACTIONS(1322),
    [anon_sym_short] = ACTIONS(1322),
    [anon_sym_static] = ACTIONS(1322),
    [anon_sym_auto] = ACTIONS(1322),
    [anon_sym_register] = ACTIONS(1322),
    [anon_sym_inline] = ACTIONS(1322),
    [anon_sym___inline] = ACTIONS(1322),
    [anon_sym___inline__] = ACTIONS(1322),
    [anon_sym___forceinline] = ACTIONS(1322),
    [anon_sym_thread_local] = ACTIONS(1322),
    [anon_sym___thread] = ACTIONS(1322),
    [anon_sym_const] = ACTIONS(1322),
    [anon_sym_constexpr] = ACTIONS(1322),
    [anon_sym_volatile] = ACTIONS(1322),
    [anon_sym_restrict] = ACTIONS(1322),
    [anon_sym___restrict__] = ACTIONS(1322),
    [anon_sym__Atomic] = ACTIONS(1322),
    [anon_sym__Noreturn] = ACTIONS(1322),
    [anon_sym_noreturn] = ACTIONS(1322),
    [anon_sym__Nonnull] = ACTIONS(1322),
    [anon_sym_alignas] = ACTIONS(1322),
    [anon_sym__Alignas] = ACTIONS(1322),
    [sym_primitive_type] = ACTIONS(1322),
    [anon_sym_enum] = ACTIONS(1322),
    [anon_sym_struct] = ACTIONS(1322),
    [anon_sym_union] = ACTIONS(1322),
    [anon_sym_if] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1322),
    [anon_sym_case] = ACTIONS(1322),
    [anon_sym_default] = ACTIONS(1322),
    [anon_sym_while] = ACTIONS(1322),
    [anon_sym_do] = ACTIONS(1322),
    [anon_sym_for] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1322),
    [anon_sym_break] = ACTIONS(1322),
    [anon_sym_continue] = ACTIONS(1322),
    [anon_sym_goto] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1324),
    [anon_sym_sizeof] = ACTIONS(1322),
    [anon_sym___alignof__] = ACTIONS(1322),
    [anon_sym___alignof] = ACTIONS(1322),
    [anon_sym__alignof] = ACTIONS(1322),
    [anon_sym_alignof] = ACTIONS(1322),
    [anon_sym__Alignof] = ACTIONS(1322),
    [anon_sym_offsetof] = ACTIONS(1322),
    [anon_sym__Generic] = ACTIONS(1322),
    [anon_sym_asm] = ACTIONS(1322),
    [anon_sym___asm__] = ACTIONS(1322),
    [anon_sym___asm] = ACTIONS(1322),
    [sym_number_literal] = ACTIONS(1324),
    [anon_sym_L_SQUOTE] = ACTIONS(1324),
    [anon_sym_u_SQUOTE] = ACTIONS(1324),
    [anon_sym_U_SQUOTE] = ACTIONS(1324),
    [anon_sym_u8_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_L_DQUOTE] = ACTIONS(1324),
    [anon_sym_u_DQUOTE] = ACTIONS(1324),
    [anon_sym_U_DQUOTE] = ACTIONS(1324),
    [anon_sym_u8_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [sym_true] = ACTIONS(1322),
    [sym_false] = ACTIONS(1322),
    [anon_sym_NULL] = ACTIONS(1322),
    [anon_sym_nullptr] = ACTIONS(1322),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(329)] = {
    [ts_builtin_sym_end] = ACTIONS(1328),
    [sym_identifier] = ACTIONS(1326),
    [aux_sym_preproc_include_token1] = ACTIONS(1326),
    [aux_sym_preproc_def_token1] = ACTIONS(1326),
    [aux_sym_preproc_if_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1326),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1326),
    [sym_preproc_directive] = ACTIONS(1326),
    [anon_sym_LPAREN2] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1326),
    [anon_sym_PLUS] = ACTIONS(1326),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym___extension__] = ACTIONS(1326),
    [anon_sym_typedef] = ACTIONS(1326),
    [anon_sym_extern] = ACTIONS(1326),
    [anon_sym___attribute__] = ACTIONS(1326),
    [anon_sym___attribute] = ACTIONS(1326),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1328),
    [anon_sym___declspec] = ACTIONS(1326),
    [anon_sym___cdecl] = ACTIONS(1326),
    [anon_sym___clrcall] = ACTIONS(1326),
    [anon_sym___stdcall] = ACTIONS(1326),
    [anon_sym___fastcall] = ACTIONS(1326),
    [anon_sym___thiscall] = ACTIONS(1326),
    [anon_sym___vectorcall] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_signed] = ACTIONS(1326),
    [anon_sym_unsigned] = ACTIONS(1326),
    [anon_sym_long] = ACTIONS(1326),
    [anon_sym_short] = ACTIONS(1326),
    [anon_sym_static] = ACTIONS(1326),
    [anon_sym_auto] = ACTIONS(1326),
    [anon_sym_register] = ACTIONS(1326),
    [anon_sym_inline] = ACTIONS(1326),
    [anon_sym___inline] = ACTIONS(1326),
    [anon_sym___inline__] = ACTIONS(1326),
    [anon_sym___forceinline] = ACTIONS(1326),
    [anon_sym_thread_local] = ACTIONS(1326),
    [anon_sym___thread] = ACTIONS(1326),
    [anon_sym_const] = ACTIONS(1326),
    [anon_sym_constexpr] = ACTIONS(1326),
    [anon_sym_volatile] = ACTIONS(1326),
    [anon_sym_restrict] = ACTIONS(1326),
    [anon_sym___restrict__] = ACTIONS(1326),
    [anon_sym__Atomic] = ACTIONS(1326),
    [anon_sym__Noreturn] = ACTIONS(1326),
    [anon_sym_noreturn] = ACTIONS(1326),
    [anon_sym__Nonnull] = ACTIONS(1326),
    [anon_sym_alignas] = ACTIONS(1326),
    [anon_sym__Alignas] = ACTIONS(1326),
    [sym_primitive_type] = ACTIONS(1326),
    [anon_sym_enum] = ACTIONS(1326),
    [anon_sym_struct] = ACTIONS(1326),
    [anon_sym_union] = ACTIONS(1326),
    [anon_sym_if] = ACTIONS(1326),
    [anon_sym_switch] = ACTIONS(1326),
    [anon_sym_case] = ACTIONS(1326),
    [anon_sym_default] = ACTIONS(1326),
    [anon_sym_while] = ACTIONS(1326),
    [anon_sym_do] = ACTIONS(1326),
    [anon_sym_for] = ACTIONS(1326),
    [anon_sym_return] = ACTIONS(1326),
    [anon_sym_break] = ACTIONS(1326),
    [anon_sym_continue] = ACTIONS(1326),
    [anon_sym_goto] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1328),
    [anon_sym_sizeof] = ACTIONS(1326),
    [anon_sym___alignof__] = ACTIONS(1326),
    [anon_sym___alignof] = ACTIONS(1326),
    [anon_sym__alignof] = ACTIONS(1326),
    [anon_sym_alignof] = ACTIONS(1326),
    [anon_sym__Alignof] = ACTIONS(1326),
    [anon_sym_offsetof] = ACTIONS(1326),
    [anon_sym__Generic] = ACTIONS(1326),
    [anon_sym_asm] = ACTIONS(1326),
    [anon_sym___asm__] = ACTIONS(1326),
    [anon_sym___asm] = ACTIONS(1326),
    [sym_number_literal] = ACTIONS(1328),
    [anon_sym_L_SQUOTE] = ACTIONS(1328),
    [anon_sym_u_SQUOTE] = ACTIONS(1328),
    [anon_sym_U_SQUOTE] = ACTIONS(1328),
    [anon_sym_u8_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_L_DQUOTE] = ACTIONS(1328),
    [anon_sym_u_DQUOTE] = ACTIONS(1328),
    [anon_sym_U_DQUOTE] = ACTIONS(1328),
    [anon_sym_u8_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [sym_true] = ACTIONS(1326),
    [sym_false] = ACTIONS(1326),
    [anon_sym_NULL] = ACTIONS(1326),
    [anon_sym_nullptr] = ACTIONS(1326),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(330)] = {
    [ts_builtin_sym_end] = ACTIONS(1366),
    [sym_identifier] = ACTIONS(1364),
    [aux_sym_preproc_include_token1] = ACTIONS(1364),
    [aux_sym_preproc_def_token1] = ACTIONS(1364),
    [aux_sym_preproc_if_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1364),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1364),
    [sym_preproc_directive] = ACTIONS(1364),
    [anon_sym_LPAREN2] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_PLUS] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [anon_sym___extension__] = ACTIONS(1364),
    [anon_sym_typedef] = ACTIONS(1364),
    [anon_sym_extern] = ACTIONS(1364),
    [anon_sym___attribute__] = ACTIONS(1364),
    [anon_sym___attribute] = ACTIONS(1364),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1366),
    [anon_sym___declspec] = ACTIONS(1364),
    [anon_sym___cdecl] = ACTIONS(1364),
    [anon_sym___clrcall] = ACTIONS(1364),
    [anon_sym___stdcall] = ACTIONS(1364),
    [anon_sym___fastcall] = ACTIONS(1364),
    [anon_sym___thiscall] = ACTIONS(1364),
    [anon_sym___vectorcall] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_signed] = ACTIONS(1364),
    [anon_sym_unsigned] = ACTIONS(1364),
    [anon_sym_long] = ACTIONS(1364),
    [anon_sym_short] = ACTIONS(1364),
    [anon_sym_static] = ACTIONS(1364),
    [anon_sym_auto] = ACTIONS(1364),
    [anon_sym_register] = ACTIONS(1364),
    [anon_sym_inline] = ACTIONS(1364),
    [anon_sym___inline] = ACTIONS(1364),
    [anon_sym___inline__] = ACTIONS(1364),
    [anon_sym___forceinline] = ACTIONS(1364),
    [anon_sym_thread_local] = ACTIONS(1364),
    [anon_sym___thread] = ACTIONS(1364),
    [anon_sym_const] = ACTIONS(1364),
    [anon_sym_constexpr] = ACTIONS(1364),
    [anon_sym_volatile] = ACTIONS(1364),
    [anon_sym_restrict] = ACTIONS(1364),
    [anon_sym___restrict__] = ACTIONS(1364),
    [anon_sym__Atomic] = ACTIONS(1364),
    [anon_sym__Noreturn] = ACTIONS(1364),
    [anon_sym_noreturn] = ACTIONS(1364),
    [anon_sym__Nonnull] = ACTIONS(1364),
    [anon_sym_alignas] = ACTIONS(1364),
    [anon_sym__Alignas] = ACTIONS(1364),
    [sym_primitive_type] = ACTIONS(1364),
    [anon_sym_enum] = ACTIONS(1364),
    [anon_sym_struct] = ACTIONS(1364),
    [anon_sym_union] = ACTIONS(1364),
    [anon_sym_if] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1364),
    [anon_sym_case] = ACTIONS(1364),
    [anon_sym_default] = ACTIONS(1364),
    [anon_sym_while] = ACTIONS(1364),
    [anon_sym_do] = ACTIONS(1364),
    [anon_sym_for] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1364),
    [anon_sym_break] = ACTIONS(1364),
    [anon_sym_continue] = ACTIONS(1364),
    [anon_sym_goto] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1366),
    [anon_sym_sizeof] = ACTIONS(1364),
    [anon_sym___alignof__] = ACTIONS(1364),
    [anon_sym___alignof] = ACTIONS(1364),
    [anon_sym__alignof] = ACTIONS(1364),
    [anon_sym_alignof] = ACTIONS(1364),
    [anon_sym__Alignof] = ACTIONS(1364),
    [anon_sym_offsetof] = ACTIONS(1364),
    [anon_sym__Generic] = ACTIONS(1364),
    [anon_sym_asm] = ACTIONS(1364),
    [anon_sym___asm__] = ACTIONS(1364),
    [anon_sym___asm] = ACTIONS(1364),
    [sym_number_literal] = ACTIONS(1366),
    [anon_sym_L_SQUOTE] = ACTIONS(1366),
    [anon_sym_u_SQUOTE] = ACTIONS(1366),
    [anon_sym_U_SQUOTE] = ACTIONS(1366),
    [anon_sym_u8_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_L_DQUOTE] = ACTIONS(1366),
    [anon_sym_u_DQUOTE] = ACTIONS(1366),
    [anon_sym_U_DQUOTE] = ACTIONS(1366),
    [anon_sym_u8_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [sym_true] = ACTIONS(1364),
    [sym_false] = ACTIONS(1364),
    [anon_sym_NULL] = ACTIONS(1364),
    [anon_sym_nullptr] = ACTIONS(1364),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(331)] = {
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1330),
    [aux_sym_preproc_include_token1] = ACTIONS(1330),
    [aux_sym_preproc_def_token1] = ACTIONS(1330),
    [aux_sym_preproc_if_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1330),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1330),
    [sym_preproc_directive] = ACTIONS(1330),
    [anon_sym_LPAREN2] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1330),
    [anon_sym_PLUS] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym___extension__] = ACTIONS(1330),
    [anon_sym_typedef] = ACTIONS(1330),
    [anon_sym_extern] = ACTIONS(1330),
    [anon_sym___attribute__] = ACTIONS(1330),
    [anon_sym___attribute] = ACTIONS(1330),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1332),
    [anon_sym___declspec] = ACTIONS(1330),
    [anon_sym___cdecl] = ACTIONS(1330),
    [anon_sym___clrcall] = ACTIONS(1330),
    [anon_sym___stdcall] = ACTIONS(1330),
    [anon_sym___fastcall] = ACTIONS(1330),
    [anon_sym___thiscall] = ACTIONS(1330),
    [anon_sym___vectorcall] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_signed] = ACTIONS(1330),
    [anon_sym_unsigned] = ACTIONS(1330),
    [anon_sym_long] = ACTIONS(1330),
    [anon_sym_short] = ACTIONS(1330),
    [anon_sym_static] = ACTIONS(1330),
    [anon_sym_auto] = ACTIONS(1330),
    [anon_sym_register] = ACTIONS(1330),
    [anon_sym_inline] = ACTIONS(1330),
    [anon_sym___inline] = ACTIONS(1330),
    [anon_sym___inline__] = ACTIONS(1330),
    [anon_sym___forceinline] = ACTIONS(1330),
    [anon_sym_thread_local] = ACTIONS(1330),
    [anon_sym___thread] = ACTIONS(1330),
    [anon_sym_const] = ACTIONS(1330),
    [anon_sym_constexpr] = ACTIONS(1330),
    [anon_sym_volatile] = ACTIONS(1330),
    [anon_sym_restrict] = ACTIONS(1330),
    [anon_sym___restrict__] = ACTIONS(1330),
    [anon_sym__Atomic] = ACTIONS(1330),
    [anon_sym__Noreturn] = ACTIONS(1330),
    [anon_sym_noreturn] = ACTIONS(1330),
    [anon_sym__Nonnull] = ACTIONS(1330),
    [anon_sym_alignas] = ACTIONS(1330),
    [anon_sym__Alignas] = ACTIONS(1330),
    [sym_primitive_type] = ACTIONS(1330),
    [anon_sym_enum] = ACTIONS(1330),
    [anon_sym_struct] = ACTIONS(1330),
    [anon_sym_union] = ACTIONS(1330),
    [anon_sym_if] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1330),
    [anon_sym_case] = ACTIONS(1330),
    [anon_sym_default] = ACTIONS(1330),
    [anon_sym_while] = ACTIONS(1330),
    [anon_sym_do] = ACTIONS(1330),
    [anon_sym_for] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1330),
    [anon_sym_break] = ACTIONS(1330),
    [anon_sym_continue] = ACTIONS(1330),
    [anon_sym_goto] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_sizeof] = ACTIONS(1330),
    [anon_sym___alignof__] = ACTIONS(1330),
    [anon_sym___alignof] = ACTIONS(1330),
    [anon_sym__alignof] = ACTIONS(1330),
    [anon_sym_alignof] = ACTIONS(1330),
    [anon_sym__Alignof] = ACTIONS(1330),
    [anon_sym_offsetof] = ACTIONS(1330),
    [anon_sym__Generic] = ACTIONS(1330),
    [anon_sym_asm] = ACTIONS(1330),
    [anon_sym___asm__] = ACTIONS(1330),
    [anon_sym___asm] = ACTIONS(1330),
    [sym_number_literal] = ACTIONS(1332),
    [anon_sym_L_SQUOTE] = ACTIONS(1332),
    [anon_sym_u_SQUOTE] = ACTIONS(1332),
    [anon_sym_U_SQUOTE] = ACTIONS(1332),
    [anon_sym_u8_SQUOTE] = ACTIONS(1332),
    [anon_sym_SQUOTE] = ACTIONS(1332),
    [anon_sym_L_DQUOTE] = ACTIONS(1332),
    [anon_sym_u_DQUOTE] = ACTIONS(1332),
    [anon_sym_U_DQUOTE] = ACTIONS(1332),
    [anon_sym_u8_DQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1332),
    [sym_true] = ACTIONS(1330),
    [sym_false] = ACTIONS(1330),
    [anon_sym_NULL] = ACTIONS(1330),
    [anon_sym_nullptr] = ACTIONS(1330),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(332)] = {
    [ts_builtin_sym_end] = ACTIONS(1342),
    [sym_identifier] = ACTIONS(1340),
    [aux_sym_preproc_include_token1] = ACTIONS(1340),
    [aux_sym_preproc_def_token1] = ACTIONS(1340),
    [aux_sym_preproc_if_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1340),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1340),
    [sym_preproc_directive] = ACTIONS(1340),
    [anon_sym_LPAREN2] = ACTIONS(1342),
    [anon_sym_BANG] = ACTIONS(1342),
    [anon_sym_TILDE] = ACTIONS(1342),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1342),
    [anon_sym_AMP] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1342),
    [anon_sym___extension__] = ACTIONS(1340),
    [anon_sym_typedef] = ACTIONS(1340),
    [anon_sym_extern] = ACTIONS(1340),
    [anon_sym___attribute__] = ACTIONS(1340),
    [anon_sym___attribute] = ACTIONS(1340),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1342),
    [anon_sym___declspec] = ACTIONS(1340),
    [anon_sym___cdecl] = ACTIONS(1340),
    [anon_sym___clrcall] = ACTIONS(1340),
    [anon_sym___stdcall] = ACTIONS(1340),
    [anon_sym___fastcall] = ACTIONS(1340),
    [anon_sym___thiscall] = ACTIONS(1340),
    [anon_sym___vectorcall] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1342),
    [anon_sym_signed] = ACTIONS(1340),
    [anon_sym_unsigned] = ACTIONS(1340),
    [anon_sym_long] = ACTIONS(1340),
    [anon_sym_short] = ACTIONS(1340),
    [anon_sym_static] = ACTIONS(1340),
    [anon_sym_auto] = ACTIONS(1340),
    [anon_sym_register] = ACTIONS(1340),
    [anon_sym_inline] = ACTIONS(1340),
    [anon_sym___inline] = ACTIONS(1340),
    [anon_sym___inline__] = ACTIONS(1340),
    [anon_sym___forceinline] = ACTIONS(1340),
    [anon_sym_thread_local] = ACTIONS(1340),
    [anon_sym___thread] = ACTIONS(1340),
    [anon_sym_const] = ACTIONS(1340),
    [anon_sym_constexpr] = ACTIONS(1340),
    [anon_sym_volatile] = ACTIONS(1340),
    [anon_sym_restrict] = ACTIONS(1340),
    [anon_sym___restrict__] = ACTIONS(1340),
    [anon_sym__Atomic] = ACTIONS(1340),
    [anon_sym__Noreturn] = ACTIONS(1340),
    [anon_sym_noreturn] = ACTIONS(1340),
    [anon_sym__Nonnull] = ACTIONS(1340),
    [anon_sym_alignas] = ACTIONS(1340),
    [anon_sym__Alignas] = ACTIONS(1340),
    [sym_primitive_type] = ACTIONS(1340),
    [anon_sym_enum] = ACTIONS(1340),
    [anon_sym_struct] = ACTIONS(1340),
    [anon_sym_union] = ACTIONS(1340),
    [anon_sym_if] = ACTIONS(1340),
    [anon_sym_switch] = ACTIONS(1340),
    [anon_sym_case] = ACTIONS(1340),
    [anon_sym_default] = ACTIONS(1340),
    [anon_sym_while] = ACTIONS(1340),
    [anon_sym_do] = ACTIONS(1340),
    [anon_sym_for] = ACTIONS(1340),
    [anon_sym_return] = ACTIONS(1340),
    [anon_sym_break] = ACTIONS(1340),
    [anon_sym_continue] = ACTIONS(1340),
    [anon_sym_goto] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1342),
    [anon_sym_PLUS_PLUS] = ACTIONS(1342),
    [anon_sym_sizeof] = ACTIONS(1340),
    [anon_sym___alignof__] = ACTIONS(1340),
    [anon_sym___alignof] = ACTIONS(1340),
    [anon_sym__alignof] = ACTIONS(1340),
    [anon_sym_alignof] = ACTIONS(1340),
    [anon_sym__Alignof] = ACTIONS(1340),
    [anon_sym_offsetof] = ACTIONS(1340),
    [anon_sym__Generic] = ACTIONS(1340),
    [anon_sym_asm] = ACTIONS(1340),
    [anon_sym___asm__] = ACTIONS(1340),
    [anon_sym___asm] = ACTIONS(1340),
    [sym_number_literal] = ACTIONS(1342),
    [anon_sym_L_SQUOTE] = ACTIONS(1342),
    [anon_sym_u_SQUOTE] = ACTIONS(1342),
    [anon_sym_U_SQUOTE] = ACTIONS(1342),
    [anon_sym_u8_SQUOTE] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_L_DQUOTE] = ACTIONS(1342),
    [anon_sym_u_DQUOTE] = ACTIONS(1342),
    [anon_sym_U_DQUOTE] = ACTIONS(1342),
    [anon_sym_u8_DQUOTE] = ACTIONS(1342),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [sym_true] = ACTIONS(1340),
    [sym_false] = ACTIONS(1340),
    [anon_sym_NULL] = ACTIONS(1340),
    [anon_sym_nullptr] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(333)] = {
    [sym_attribute_declaration] = STATE(342),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(236),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(334)] = {
    [ts_builtin_sym_end] = ACTIONS(1346),
    [sym_identifier] = ACTIONS(1344),
    [aux_sym_preproc_include_token1] = ACTIONS(1344),
    [aux_sym_preproc_def_token1] = ACTIONS(1344),
    [aux_sym_preproc_if_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1344),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1344),
    [sym_preproc_directive] = ACTIONS(1344),
    [anon_sym_LPAREN2] = ACTIONS(1346),
    [anon_sym_BANG] = ACTIONS(1346),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1346),
    [anon_sym_AMP] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1346),
    [anon_sym___extension__] = ACTIONS(1344),
    [anon_sym_typedef] = ACTIONS(1344),
    [anon_sym_extern] = ACTIONS(1344),
    [anon_sym___attribute__] = ACTIONS(1344),
    [anon_sym___attribute] = ACTIONS(1344),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1346),
    [anon_sym___declspec] = ACTIONS(1344),
    [anon_sym___cdecl] = ACTIONS(1344),
    [anon_sym___clrcall] = ACTIONS(1344),
    [anon_sym___stdcall] = ACTIONS(1344),
    [anon_sym___fastcall] = ACTIONS(1344),
    [anon_sym___thiscall] = ACTIONS(1344),
    [anon_sym___vectorcall] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [anon_sym_signed] = ACTIONS(1344),
    [anon_sym_unsigned] = ACTIONS(1344),
    [anon_sym_long] = ACTIONS(1344),
    [anon_sym_short] = ACTIONS(1344),
    [anon_sym_static] = ACTIONS(1344),
    [anon_sym_auto] = ACTIONS(1344),
    [anon_sym_register] = ACTIONS(1344),
    [anon_sym_inline] = ACTIONS(1344),
    [anon_sym___inline] = ACTIONS(1344),
    [anon_sym___inline__] = ACTIONS(1344),
    [anon_sym___forceinline] = ACTIONS(1344),
    [anon_sym_thread_local] = ACTIONS(1344),
    [anon_sym___thread] = ACTIONS(1344),
    [anon_sym_const] = ACTIONS(1344),
    [anon_sym_constexpr] = ACTIONS(1344),
    [anon_sym_volatile] = ACTIONS(1344),
    [anon_sym_restrict] = ACTIONS(1344),
    [anon_sym___restrict__] = ACTIONS(1344),
    [anon_sym__Atomic] = ACTIONS(1344),
    [anon_sym__Noreturn] = ACTIONS(1344),
    [anon_sym_noreturn] = ACTIONS(1344),
    [anon_sym__Nonnull] = ACTIONS(1344),
    [anon_sym_alignas] = ACTIONS(1344),
    [anon_sym__Alignas] = ACTIONS(1344),
    [sym_primitive_type] = ACTIONS(1344),
    [anon_sym_enum] = ACTIONS(1344),
    [anon_sym_struct] = ACTIONS(1344),
    [anon_sym_union] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_switch] = ACTIONS(1344),
    [anon_sym_case] = ACTIONS(1344),
    [anon_sym_default] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_break] = ACTIONS(1344),
    [anon_sym_continue] = ACTIONS(1344),
    [anon_sym_goto] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1346),
    [anon_sym_PLUS_PLUS] = ACTIONS(1346),
    [anon_sym_sizeof] = ACTIONS(1344),
    [anon_sym___alignof__] = ACTIONS(1344),
    [anon_sym___alignof] = ACTIONS(1344),
    [anon_sym__alignof] = ACTIONS(1344),
    [anon_sym_alignof] = ACTIONS(1344),
    [anon_sym__Alignof] = ACTIONS(1344),
    [anon_sym_offsetof] = ACTIONS(1344),
    [anon_sym__Generic] = ACTIONS(1344),
    [anon_sym_asm] = ACTIONS(1344),
    [anon_sym___asm__] = ACTIONS(1344),
    [anon_sym___asm] = ACTIONS(1344),
    [sym_number_literal] = ACTIONS(1346),
    [anon_sym_L_SQUOTE] = ACTIONS(1346),
    [anon_sym_u_SQUOTE] = ACTIONS(1346),
    [anon_sym_U_SQUOTE] = ACTIONS(1346),
    [anon_sym_u8_SQUOTE] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_L_DQUOTE] = ACTIONS(1346),
    [anon_sym_u_DQUOTE] = ACTIONS(1346),
    [anon_sym_U_DQUOTE] = ACTIONS(1346),
    [anon_sym_u8_DQUOTE] = ACTIONS(1346),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [anon_sym_NULL] = ACTIONS(1344),
    [anon_sym_nullptr] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(335)] = {
    [ts_builtin_sym_end] = ACTIONS(1350),
    [sym_identifier] = ACTIONS(1348),
    [aux_sym_preproc_include_token1] = ACTIONS(1348),
    [aux_sym_preproc_def_token1] = ACTIONS(1348),
    [aux_sym_preproc_if_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1348),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1348),
    [sym_preproc_directive] = ACTIONS(1348),
    [anon_sym_LPAREN2] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym___extension__] = ACTIONS(1348),
    [anon_sym_typedef] = ACTIONS(1348),
    [anon_sym_extern] = ACTIONS(1348),
    [anon_sym___attribute__] = ACTIONS(1348),
    [anon_sym___attribute] = ACTIONS(1348),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1350),
    [anon_sym___declspec] = ACTIONS(1348),
    [anon_sym___cdecl] = ACTIONS(1348),
    [anon_sym___clrcall] = ACTIONS(1348),
    [anon_sym___stdcall] = ACTIONS(1348),
    [anon_sym___fastcall] = ACTIONS(1348),
    [anon_sym___thiscall] = ACTIONS(1348),
    [anon_sym___vectorcall] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_signed] = ACTIONS(1348),
    [anon_sym_unsigned] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_auto] = ACTIONS(1348),
    [anon_sym_register] = ACTIONS(1348),
    [anon_sym_inline] = ACTIONS(1348),
    [anon_sym___inline] = ACTIONS(1348),
    [anon_sym___inline__] = ACTIONS(1348),
    [anon_sym___forceinline] = ACTIONS(1348),
    [anon_sym_thread_local] = ACTIONS(1348),
    [anon_sym___thread] = ACTIONS(1348),
    [anon_sym_const] = ACTIONS(1348),
    [anon_sym_constexpr] = ACTIONS(1348),
    [anon_sym_volatile] = ACTIONS(1348),
    [anon_sym_restrict] = ACTIONS(1348),
    [anon_sym___restrict__] = ACTIONS(1348),
    [anon_sym__Atomic] = ACTIONS(1348),
    [anon_sym__Noreturn] = ACTIONS(1348),
    [anon_sym_noreturn] = ACTIONS(1348),
    [anon_sym__Nonnull] = ACTIONS(1348),
    [anon_sym_alignas] = ACTIONS(1348),
    [anon_sym__Alignas] = ACTIONS(1348),
    [sym_primitive_type] = ACTIONS(1348),
    [anon_sym_enum] = ACTIONS(1348),
    [anon_sym_struct] = ACTIONS(1348),
    [anon_sym_union] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_goto] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_sizeof] = ACTIONS(1348),
    [anon_sym___alignof__] = ACTIONS(1348),
    [anon_sym___alignof] = ACTIONS(1348),
    [anon_sym__alignof] = ACTIONS(1348),
    [anon_sym_alignof] = ACTIONS(1348),
    [anon_sym__Alignof] = ACTIONS(1348),
    [anon_sym_offsetof] = ACTIONS(1348),
    [anon_sym__Generic] = ACTIONS(1348),
    [anon_sym_asm] = ACTIONS(1348),
    [anon_sym___asm__] = ACTIONS(1348),
    [anon_sym___asm] = ACTIONS(1348),
    [sym_number_literal] = ACTIONS(1350),
    [anon_sym_L_SQUOTE] = ACTIONS(1350),
    [anon_sym_u_SQUOTE] = ACTIONS(1350),
    [anon_sym_U_SQUOTE] = ACTIONS(1350),
    [anon_sym_u8_SQUOTE] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_L_DQUOTE] = ACTIONS(1350),
    [anon_sym_u_DQUOTE] = ACTIONS(1350),
    [anon_sym_U_DQUOTE] = ACTIONS(1350),
    [anon_sym_u8_DQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [sym_true] = ACTIONS(1348),
    [sym_false] = ACTIONS(1348),
    [anon_sym_NULL] = ACTIONS(1348),
    [anon_sym_nullptr] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(336)] = {
    [ts_builtin_sym_end] = ACTIONS(1354),
    [sym_identifier] = ACTIONS(1352),
    [aux_sym_preproc_include_token1] = ACTIONS(1352),
    [aux_sym_preproc_def_token1] = ACTIONS(1352),
    [aux_sym_preproc_if_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1352),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1352),
    [sym_preproc_directive] = ACTIONS(1352),
    [anon_sym_LPAREN2] = ACTIONS(1354),
    [anon_sym_BANG] = ACTIONS(1354),
    [anon_sym_TILDE] = ACTIONS(1354),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1354),
    [anon_sym_AMP] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1354),
    [anon_sym___extension__] = ACTIONS(1352),
    [anon_sym_typedef] = ACTIONS(1352),
    [anon_sym_extern] = ACTIONS(1352),
    [anon_sym___attribute__] = ACTIONS(1352),
    [anon_sym___attribute] = ACTIONS(1352),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1354),
    [anon_sym___declspec] = ACTIONS(1352),
    [anon_sym___cdecl] = ACTIONS(1352),
    [anon_sym___clrcall] = ACTIONS(1352),
    [anon_sym___stdcall] = ACTIONS(1352),
    [anon_sym___fastcall] = ACTIONS(1352),
    [anon_sym___thiscall] = ACTIONS(1352),
    [anon_sym___vectorcall] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1354),
    [anon_sym_signed] = ACTIONS(1352),
    [anon_sym_unsigned] = ACTIONS(1352),
    [anon_sym_long] = ACTIONS(1352),
    [anon_sym_short] = ACTIONS(1352),
    [anon_sym_static] = ACTIONS(1352),
    [anon_sym_auto] = ACTIONS(1352),
    [anon_sym_register] = ACTIONS(1352),
    [anon_sym_inline] = ACTIONS(1352),
    [anon_sym___inline] = ACTIONS(1352),
    [anon_sym___inline__] = ACTIONS(1352),
    [anon_sym___forceinline] = ACTIONS(1352),
    [anon_sym_thread_local] = ACTIONS(1352),
    [anon_sym___thread] = ACTIONS(1352),
    [anon_sym_const] = ACTIONS(1352),
    [anon_sym_constexpr] = ACTIONS(1352),
    [anon_sym_volatile] = ACTIONS(1352),
    [anon_sym_restrict] = ACTIONS(1352),
    [anon_sym___restrict__] = ACTIONS(1352),
    [anon_sym__Atomic] = ACTIONS(1352),
    [anon_sym__Noreturn] = ACTIONS(1352),
    [anon_sym_noreturn] = ACTIONS(1352),
    [anon_sym__Nonnull] = ACTIONS(1352),
    [anon_sym_alignas] = ACTIONS(1352),
    [anon_sym__Alignas] = ACTIONS(1352),
    [sym_primitive_type] = ACTIONS(1352),
    [anon_sym_enum] = ACTIONS(1352),
    [anon_sym_struct] = ACTIONS(1352),
    [anon_sym_union] = ACTIONS(1352),
    [anon_sym_if] = ACTIONS(1352),
    [anon_sym_switch] = ACTIONS(1352),
    [anon_sym_case] = ACTIONS(1352),
    [anon_sym_default] = ACTIONS(1352),
    [anon_sym_while] = ACTIONS(1352),
    [anon_sym_do] = ACTIONS(1352),
    [anon_sym_for] = ACTIONS(1352),
    [anon_sym_return] = ACTIONS(1352),
    [anon_sym_break] = ACTIONS(1352),
    [anon_sym_continue] = ACTIONS(1352),
    [anon_sym_goto] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1354),
    [anon_sym_PLUS_PLUS] = ACTIONS(1354),
    [anon_sym_sizeof] = ACTIONS(1352),
    [anon_sym___alignof__] = ACTIONS(1352),
    [anon_sym___alignof] = ACTIONS(1352),
    [anon_sym__alignof] = ACTIONS(1352),
    [anon_sym_alignof] = ACTIONS(1352),
    [anon_sym__Alignof] = ACTIONS(1352),
    [anon_sym_offsetof] = ACTIONS(1352),
    [anon_sym__Generic] = ACTIONS(1352),
    [anon_sym_asm] = ACTIONS(1352),
    [anon_sym___asm__] = ACTIONS(1352),
    [anon_sym___asm] = ACTIONS(1352),
    [sym_number_literal] = ACTIONS(1354),
    [anon_sym_L_SQUOTE] = ACTIONS(1354),
    [anon_sym_u_SQUOTE] = ACTIONS(1354),
    [anon_sym_U_SQUOTE] = ACTIONS(1354),
    [anon_sym_u8_SQUOTE] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_L_DQUOTE] = ACTIONS(1354),
    [anon_sym_u_DQUOTE] = ACTIONS(1354),
    [anon_sym_U_DQUOTE] = ACTIONS(1354),
    [anon_sym_u8_DQUOTE] = ACTIONS(1354),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [sym_true] = ACTIONS(1352),
    [sym_false] = ACTIONS(1352),
    [anon_sym_NULL] = ACTIONS(1352),
    [anon_sym_nullptr] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(337)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(1888),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(338)] = {
    [sym_attribute_declaration] = STATE(338),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(167),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(338),
    [sym_identifier] = ACTIONS(1420),
    [anon_sym_LPAREN2] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1435),
    [anon_sym___extension__] = ACTIONS(1438),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1444),
    [anon_sym_if] = ACTIONS(1447),
    [anon_sym_switch] = ACTIONS(1450),
    [anon_sym_case] = ACTIONS(1453),
    [anon_sym_default] = ACTIONS(1456),
    [anon_sym_while] = ACTIONS(1459),
    [anon_sym_do] = ACTIONS(1462),
    [anon_sym_for] = ACTIONS(1465),
    [anon_sym_return] = ACTIONS(1468),
    [anon_sym_break] = ACTIONS(1471),
    [anon_sym_continue] = ACTIONS(1474),
    [anon_sym_goto] = ACTIONS(1477),
    [anon_sym___try] = ACTIONS(1480),
    [anon_sym___leave] = ACTIONS(1483),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [anon_sym_sizeof] = ACTIONS(1489),
    [anon_sym___alignof__] = ACTIONS(1492),
    [anon_sym___alignof] = ACTIONS(1492),
    [anon_sym__alignof] = ACTIONS(1492),
    [anon_sym_alignof] = ACTIONS(1492),
    [anon_sym__Alignof] = ACTIONS(1492),
    [anon_sym_offsetof] = ACTIONS(1495),
    [anon_sym__Generic] = ACTIONS(1498),
    [anon_sym_asm] = ACTIONS(1501),
    [anon_sym___asm__] = ACTIONS(1501),
    [anon_sym___asm] = ACTIONS(1501),
    [sym_number_literal] = ACTIONS(1504),
    [anon_sym_L_SQUOTE] = ACTIONS(1507),
    [anon_sym_u_SQUOTE] = ACTIONS(1507),
    [anon_sym_U_SQUOTE] = ACTIONS(1507),
    [anon_sym_u8_SQUOTE] = ACTIONS(1507),
    [anon_sym_SQUOTE] = ACTIONS(1507),
    [anon_sym_L_DQUOTE] = ACTIONS(1510),
    [anon_sym_u_DQUOTE] = ACTIONS(1510),
    [anon_sym_U_DQUOTE] = ACTIONS(1510),
    [anon_sym_u8_DQUOTE] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [sym_true] = ACTIONS(1513),
    [sym_false] = ACTIONS(1513),
    [anon_sym_NULL] = ACTIONS(1516),
    [anon_sym_nullptr] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(339)] = {
    [ts_builtin_sym_end] = ACTIONS(1284),
    [sym_identifier] = ACTIONS(1282),
    [aux_sym_preproc_include_token1] = ACTIONS(1282),
    [aux_sym_preproc_def_token1] = ACTIONS(1282),
    [aux_sym_preproc_if_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1282),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1282),
    [sym_preproc_directive] = ACTIONS(1282),
    [anon_sym_LPAREN2] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_TILDE] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1282),
    [anon_sym_PLUS] = ACTIONS(1282),
    [anon_sym_STAR] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym___extension__] = ACTIONS(1282),
    [anon_sym_typedef] = ACTIONS(1282),
    [anon_sym_extern] = ACTIONS(1282),
    [anon_sym___attribute__] = ACTIONS(1282),
    [anon_sym___attribute] = ACTIONS(1282),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1284),
    [anon_sym___declspec] = ACTIONS(1282),
    [anon_sym___cdecl] = ACTIONS(1282),
    [anon_sym___clrcall] = ACTIONS(1282),
    [anon_sym___stdcall] = ACTIONS(1282),
    [anon_sym___fastcall] = ACTIONS(1282),
    [anon_sym___thiscall] = ACTIONS(1282),
    [anon_sym___vectorcall] = ACTIONS(1282),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_signed] = ACTIONS(1282),
    [anon_sym_unsigned] = ACTIONS(1282),
    [anon_sym_long] = ACTIONS(1282),
    [anon_sym_short] = ACTIONS(1282),
    [anon_sym_static] = ACTIONS(1282),
    [anon_sym_auto] = ACTIONS(1282),
    [anon_sym_register] = ACTIONS(1282),
    [anon_sym_inline] = ACTIONS(1282),
    [anon_sym___inline] = ACTIONS(1282),
    [anon_sym___inline__] = ACTIONS(1282),
    [anon_sym___forceinline] = ACTIONS(1282),
    [anon_sym_thread_local] = ACTIONS(1282),
    [anon_sym___thread] = ACTIONS(1282),
    [anon_sym_const] = ACTIONS(1282),
    [anon_sym_constexpr] = ACTIONS(1282),
    [anon_sym_volatile] = ACTIONS(1282),
    [anon_sym_restrict] = ACTIONS(1282),
    [anon_sym___restrict__] = ACTIONS(1282),
    [anon_sym__Atomic] = ACTIONS(1282),
    [anon_sym__Noreturn] = ACTIONS(1282),
    [anon_sym_noreturn] = ACTIONS(1282),
    [anon_sym__Nonnull] = ACTIONS(1282),
    [anon_sym_alignas] = ACTIONS(1282),
    [anon_sym__Alignas] = ACTIONS(1282),
    [sym_primitive_type] = ACTIONS(1282),
    [anon_sym_enum] = ACTIONS(1282),
    [anon_sym_struct] = ACTIONS(1282),
    [anon_sym_union] = ACTIONS(1282),
    [anon_sym_if] = ACTIONS(1282),
    [anon_sym_switch] = ACTIONS(1282),
    [anon_sym_case] = ACTIONS(1282),
    [anon_sym_default] = ACTIONS(1282),
    [anon_sym_while] = ACTIONS(1282),
    [anon_sym_do] = ACTIONS(1282),
    [anon_sym_for] = ACTIONS(1282),
    [anon_sym_return] = ACTIONS(1282),
    [anon_sym_break] = ACTIONS(1282),
    [anon_sym_continue] = ACTIONS(1282),
    [anon_sym_goto] = ACTIONS(1282),
    [anon_sym_DASH_DASH] = ACTIONS(1284),
    [anon_sym_PLUS_PLUS] = ACTIONS(1284),
    [anon_sym_sizeof] = ACTIONS(1282),
    [anon_sym___alignof__] = ACTIONS(1282),
    [anon_sym___alignof] = ACTIONS(1282),
    [anon_sym__alignof] = ACTIONS(1282),
    [anon_sym_alignof] = ACTIONS(1282),
    [anon_sym__Alignof] = ACTIONS(1282),
    [anon_sym_offsetof] = ACTIONS(1282),
    [anon_sym__Generic] = ACTIONS(1282),
    [anon_sym_asm] = ACTIONS(1282),
    [anon_sym___asm__] = ACTIONS(1282),
    [anon_sym___asm] = ACTIONS(1282),
    [sym_number_literal] = ACTIONS(1284),
    [anon_sym_L_SQUOTE] = ACTIONS(1284),
    [anon_sym_u_SQUOTE] = ACTIONS(1284),
    [anon_sym_U_SQUOTE] = ACTIONS(1284),
    [anon_sym_u8_SQUOTE] = ACTIONS(1284),
    [anon_sym_SQUOTE] = ACTIONS(1284),
    [anon_sym_L_DQUOTE] = ACTIONS(1284),
    [anon_sym_u_DQUOTE] = ACTIONS(1284),
    [anon_sym_U_DQUOTE] = ACTIONS(1284),
    [anon_sym_u8_DQUOTE] = ACTIONS(1284),
    [anon_sym_DQUOTE] = ACTIONS(1284),
    [sym_true] = ACTIONS(1282),
    [sym_false] = ACTIONS(1282),
    [anon_sym_NULL] = ACTIONS(1282),
    [anon_sym_nullptr] = ACTIONS(1282),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(340)] = {
    [sym_attribute_declaration] = STATE(369),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(146),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(369),
    [sym_identifier] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(341)] = {
    [ts_builtin_sym_end] = ACTIONS(1288),
    [sym_identifier] = ACTIONS(1286),
    [aux_sym_preproc_include_token1] = ACTIONS(1286),
    [aux_sym_preproc_def_token1] = ACTIONS(1286),
    [aux_sym_preproc_if_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1286),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1286),
    [sym_preproc_directive] = ACTIONS(1286),
    [anon_sym_LPAREN2] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_TILDE] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1286),
    [anon_sym_PLUS] = ACTIONS(1286),
    [anon_sym_STAR] = ACTIONS(1288),
    [anon_sym_AMP] = ACTIONS(1288),
    [anon_sym_SEMI] = ACTIONS(1288),
    [anon_sym___extension__] = ACTIONS(1286),
    [anon_sym_typedef] = ACTIONS(1286),
    [anon_sym_extern] = ACTIONS(1286),
    [anon_sym___attribute__] = ACTIONS(1286),
    [anon_sym___attribute] = ACTIONS(1286),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1288),
    [anon_sym___declspec] = ACTIONS(1286),
    [anon_sym___cdecl] = ACTIONS(1286),
    [anon_sym___clrcall] = ACTIONS(1286),
    [anon_sym___stdcall] = ACTIONS(1286),
    [anon_sym___fastcall] = ACTIONS(1286),
    [anon_sym___thiscall] = ACTIONS(1286),
    [anon_sym___vectorcall] = ACTIONS(1286),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_signed] = ACTIONS(1286),
    [anon_sym_unsigned] = ACTIONS(1286),
    [anon_sym_long] = ACTIONS(1286),
    [anon_sym_short] = ACTIONS(1286),
    [anon_sym_static] = ACTIONS(1286),
    [anon_sym_auto] = ACTIONS(1286),
    [anon_sym_register] = ACTIONS(1286),
    [anon_sym_inline] = ACTIONS(1286),
    [anon_sym___inline] = ACTIONS(1286),
    [anon_sym___inline__] = ACTIONS(1286),
    [anon_sym___forceinline] = ACTIONS(1286),
    [anon_sym_thread_local] = ACTIONS(1286),
    [anon_sym___thread] = ACTIONS(1286),
    [anon_sym_const] = ACTIONS(1286),
    [anon_sym_constexpr] = ACTIONS(1286),
    [anon_sym_volatile] = ACTIONS(1286),
    [anon_sym_restrict] = ACTIONS(1286),
    [anon_sym___restrict__] = ACTIONS(1286),
    [anon_sym__Atomic] = ACTIONS(1286),
    [anon_sym__Noreturn] = ACTIONS(1286),
    [anon_sym_noreturn] = ACTIONS(1286),
    [anon_sym__Nonnull] = ACTIONS(1286),
    [anon_sym_alignas] = ACTIONS(1286),
    [anon_sym__Alignas] = ACTIONS(1286),
    [sym_primitive_type] = ACTIONS(1286),
    [anon_sym_enum] = ACTIONS(1286),
    [anon_sym_struct] = ACTIONS(1286),
    [anon_sym_union] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(1286),
    [anon_sym_switch] = ACTIONS(1286),
    [anon_sym_case] = ACTIONS(1286),
    [anon_sym_default] = ACTIONS(1286),
    [anon_sym_while] = ACTIONS(1286),
    [anon_sym_do] = ACTIONS(1286),
    [anon_sym_for] = ACTIONS(1286),
    [anon_sym_return] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(1286),
    [anon_sym_continue] = ACTIONS(1286),
    [anon_sym_goto] = ACTIONS(1286),
    [anon_sym_DASH_DASH] = ACTIONS(1288),
    [anon_sym_PLUS_PLUS] = ACTIONS(1288),
    [anon_sym_sizeof] = ACTIONS(1286),
    [anon_sym___alignof__] = ACTIONS(1286),
    [anon_sym___alignof] = ACTIONS(1286),
    [anon_sym__alignof] = ACTIONS(1286),
    [anon_sym_alignof] = ACTIONS(1286),
    [anon_sym__Alignof] = ACTIONS(1286),
    [anon_sym_offsetof] = ACTIONS(1286),
    [anon_sym__Generic] = ACTIONS(1286),
    [anon_sym_asm] = ACTIONS(1286),
    [anon_sym___asm__] = ACTIONS(1286),
    [anon_sym___asm] = ACTIONS(1286),
    [sym_number_literal] = ACTIONS(1288),
    [anon_sym_L_SQUOTE] = ACTIONS(1288),
    [anon_sym_u_SQUOTE] = ACTIONS(1288),
    [anon_sym_U_SQUOTE] = ACTIONS(1288),
    [anon_sym_u8_SQUOTE] = ACTIONS(1288),
    [anon_sym_SQUOTE] = ACTIONS(1288),
    [anon_sym_L_DQUOTE] = ACTIONS(1288),
    [anon_sym_u_DQUOTE] = ACTIONS(1288),
    [anon_sym_U_DQUOTE] = ACTIONS(1288),
    [anon_sym_u8_DQUOTE] = ACTIONS(1288),
    [anon_sym_DQUOTE] = ACTIONS(1288),
    [sym_true] = ACTIONS(1286),
    [sym_false] = ACTIONS(1286),
    [anon_sym_NULL] = ACTIONS(1286),
    [anon_sym_nullptr] = ACTIONS(1286),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(342)] = {
    [sym_attribute_declaration] = STATE(338),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(167),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(338),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(343)] = {
    [sym_attribute_declaration] = STATE(367),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(80),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [sym_identifier] = ACTIONS(1519),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(344)] = {
    [sym_attribute_declaration] = STATE(367),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(90),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [sym_identifier] = ACTIONS(1519),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(345)] = {
    [sym_attribute_declaration] = STATE(345),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(114),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(345),
    [sym_identifier] = ACTIONS(1521),
    [anon_sym_LPAREN2] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1524),
    [anon_sym___extension__] = ACTIONS(1438),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1527),
    [anon_sym_if] = ACTIONS(1530),
    [anon_sym_switch] = ACTIONS(1533),
    [anon_sym_case] = ACTIONS(1536),
    [anon_sym_default] = ACTIONS(1539),
    [anon_sym_while] = ACTIONS(1542),
    [anon_sym_do] = ACTIONS(1545),
    [anon_sym_for] = ACTIONS(1548),
    [anon_sym_return] = ACTIONS(1551),
    [anon_sym_break] = ACTIONS(1554),
    [anon_sym_continue] = ACTIONS(1557),
    [anon_sym_goto] = ACTIONS(1560),
    [anon_sym___try] = ACTIONS(1563),
    [anon_sym___leave] = ACTIONS(1566),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [anon_sym_sizeof] = ACTIONS(1489),
    [anon_sym___alignof__] = ACTIONS(1492),
    [anon_sym___alignof] = ACTIONS(1492),
    [anon_sym__alignof] = ACTIONS(1492),
    [anon_sym_alignof] = ACTIONS(1492),
    [anon_sym__Alignof] = ACTIONS(1492),
    [anon_sym_offsetof] = ACTIONS(1495),
    [anon_sym__Generic] = ACTIONS(1498),
    [anon_sym_asm] = ACTIONS(1501),
    [anon_sym___asm__] = ACTIONS(1501),
    [anon_sym___asm] = ACTIONS(1501),
    [sym_number_literal] = ACTIONS(1504),
    [anon_sym_L_SQUOTE] = ACTIONS(1507),
    [anon_sym_u_SQUOTE] = ACTIONS(1507),
    [anon_sym_U_SQUOTE] = ACTIONS(1507),
    [anon_sym_u8_SQUOTE] = ACTIONS(1507),
    [anon_sym_SQUOTE] = ACTIONS(1507),
    [anon_sym_L_DQUOTE] = ACTIONS(1510),
    [anon_sym_u_DQUOTE] = ACTIONS(1510),
    [anon_sym_U_DQUOTE] = ACTIONS(1510),
    [anon_sym_u8_DQUOTE] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [sym_true] = ACTIONS(1513),
    [sym_false] = ACTIONS(1513),
    [anon_sym_NULL] = ACTIONS(1516),
    [anon_sym_nullptr] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(346)] = {
    [sym_attribute_declaration] = STATE(367),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(85),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [sym_identifier] = ACTIONS(1519),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(347)] = {
    [sym_attribute_declaration] = STATE(367),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(92),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(367),
    [sym_identifier] = ACTIONS(1519),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(348)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(398),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(349)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(166),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(350)] = {
    [sym_attribute_declaration] = STATE(350),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(252),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(350),
    [sym_identifier] = ACTIONS(1569),
    [anon_sym_LPAREN2] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1435),
    [anon_sym___extension__] = ACTIONS(1438),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1572),
    [anon_sym_if] = ACTIONS(1575),
    [anon_sym_switch] = ACTIONS(1578),
    [anon_sym_case] = ACTIONS(1581),
    [anon_sym_default] = ACTIONS(1584),
    [anon_sym_while] = ACTIONS(1587),
    [anon_sym_do] = ACTIONS(1590),
    [anon_sym_for] = ACTIONS(1593),
    [anon_sym_return] = ACTIONS(1596),
    [anon_sym_break] = ACTIONS(1599),
    [anon_sym_continue] = ACTIONS(1602),
    [anon_sym_goto] = ACTIONS(1605),
    [anon_sym___try] = ACTIONS(1608),
    [anon_sym___leave] = ACTIONS(1483),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [anon_sym_sizeof] = ACTIONS(1489),
    [anon_sym___alignof__] = ACTIONS(1492),
    [anon_sym___alignof] = ACTIONS(1492),
    [anon_sym__alignof] = ACTIONS(1492),
    [anon_sym_alignof] = ACTIONS(1492),
    [anon_sym__Alignof] = ACTIONS(1492),
    [anon_sym_offsetof] = ACTIONS(1495),
    [anon_sym__Generic] = ACTIONS(1498),
    [anon_sym_asm] = ACTIONS(1501),
    [anon_sym___asm__] = ACTIONS(1501),
    [anon_sym___asm] = ACTIONS(1501),
    [sym_number_literal] = ACTIONS(1504),
    [anon_sym_L_SQUOTE] = ACTIONS(1507),
    [anon_sym_u_SQUOTE] = ACTIONS(1507),
    [anon_sym_U_SQUOTE] = ACTIONS(1507),
    [anon_sym_u8_SQUOTE] = ACTIONS(1507),
    [anon_sym_SQUOTE] = ACTIONS(1507),
    [anon_sym_L_DQUOTE] = ACTIONS(1510),
    [anon_sym_u_DQUOTE] = ACTIONS(1510),
    [anon_sym_U_DQUOTE] = ACTIONS(1510),
    [anon_sym_u8_DQUOTE] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [sym_true] = ACTIONS(1513),
    [sym_false] = ACTIONS(1513),
    [anon_sym_NULL] = ACTIONS(1516),
    [anon_sym_nullptr] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(351)] = {
    [ts_builtin_sym_end] = ACTIONS(1292),
    [sym_identifier] = ACTIONS(1290),
    [aux_sym_preproc_include_token1] = ACTIONS(1290),
    [aux_sym_preproc_def_token1] = ACTIONS(1290),
    [aux_sym_preproc_if_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1290),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1290),
    [sym_preproc_directive] = ACTIONS(1290),
    [anon_sym_LPAREN2] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1290),
    [anon_sym_PLUS] = ACTIONS(1290),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym___extension__] = ACTIONS(1290),
    [anon_sym_typedef] = ACTIONS(1290),
    [anon_sym_extern] = ACTIONS(1290),
    [anon_sym___attribute__] = ACTIONS(1290),
    [anon_sym___attribute] = ACTIONS(1290),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1292),
    [anon_sym___declspec] = ACTIONS(1290),
    [anon_sym___cdecl] = ACTIONS(1290),
    [anon_sym___clrcall] = ACTIONS(1290),
    [anon_sym___stdcall] = ACTIONS(1290),
    [anon_sym___fastcall] = ACTIONS(1290),
    [anon_sym___thiscall] = ACTIONS(1290),
    [anon_sym___vectorcall] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_signed] = ACTIONS(1290),
    [anon_sym_unsigned] = ACTIONS(1290),
    [anon_sym_long] = ACTIONS(1290),
    [anon_sym_short] = ACTIONS(1290),
    [anon_sym_static] = ACTIONS(1290),
    [anon_sym_auto] = ACTIONS(1290),
    [anon_sym_register] = ACTIONS(1290),
    [anon_sym_inline] = ACTIONS(1290),
    [anon_sym___inline] = ACTIONS(1290),
    [anon_sym___inline__] = ACTIONS(1290),
    [anon_sym___forceinline] = ACTIONS(1290),
    [anon_sym_thread_local] = ACTIONS(1290),
    [anon_sym___thread] = ACTIONS(1290),
    [anon_sym_const] = ACTIONS(1290),
    [anon_sym_constexpr] = ACTIONS(1290),
    [anon_sym_volatile] = ACTIONS(1290),
    [anon_sym_restrict] = ACTIONS(1290),
    [anon_sym___restrict__] = ACTIONS(1290),
    [anon_sym__Atomic] = ACTIONS(1290),
    [anon_sym__Noreturn] = ACTIONS(1290),
    [anon_sym_noreturn] = ACTIONS(1290),
    [anon_sym__Nonnull] = ACTIONS(1290),
    [anon_sym_alignas] = ACTIONS(1290),
    [anon_sym__Alignas] = ACTIONS(1290),
    [sym_primitive_type] = ACTIONS(1290),
    [anon_sym_enum] = ACTIONS(1290),
    [anon_sym_struct] = ACTIONS(1290),
    [anon_sym_union] = ACTIONS(1290),
    [anon_sym_if] = ACTIONS(1290),
    [anon_sym_switch] = ACTIONS(1290),
    [anon_sym_case] = ACTIONS(1290),
    [anon_sym_default] = ACTIONS(1290),
    [anon_sym_while] = ACTIONS(1290),
    [anon_sym_do] = ACTIONS(1290),
    [anon_sym_for] = ACTIONS(1290),
    [anon_sym_return] = ACTIONS(1290),
    [anon_sym_break] = ACTIONS(1290),
    [anon_sym_continue] = ACTIONS(1290),
    [anon_sym_goto] = ACTIONS(1290),
    [anon_sym_DASH_DASH] = ACTIONS(1292),
    [anon_sym_PLUS_PLUS] = ACTIONS(1292),
    [anon_sym_sizeof] = ACTIONS(1290),
    [anon_sym___alignof__] = ACTIONS(1290),
    [anon_sym___alignof] = ACTIONS(1290),
    [anon_sym__alignof] = ACTIONS(1290),
    [anon_sym_alignof] = ACTIONS(1290),
    [anon_sym__Alignof] = ACTIONS(1290),
    [anon_sym_offsetof] = ACTIONS(1290),
    [anon_sym__Generic] = ACTIONS(1290),
    [anon_sym_asm] = ACTIONS(1290),
    [anon_sym___asm__] = ACTIONS(1290),
    [anon_sym___asm] = ACTIONS(1290),
    [sym_number_literal] = ACTIONS(1292),
    [anon_sym_L_SQUOTE] = ACTIONS(1292),
    [anon_sym_u_SQUOTE] = ACTIONS(1292),
    [anon_sym_U_SQUOTE] = ACTIONS(1292),
    [anon_sym_u8_SQUOTE] = ACTIONS(1292),
    [anon_sym_SQUOTE] = ACTIONS(1292),
    [anon_sym_L_DQUOTE] = ACTIONS(1292),
    [anon_sym_u_DQUOTE] = ACTIONS(1292),
    [anon_sym_U_DQUOTE] = ACTIONS(1292),
    [anon_sym_u8_DQUOTE] = ACTIONS(1292),
    [anon_sym_DQUOTE] = ACTIONS(1292),
    [sym_true] = ACTIONS(1290),
    [sym_false] = ACTIONS(1290),
    [anon_sym_NULL] = ACTIONS(1290),
    [anon_sym_nullptr] = ACTIONS(1290),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(352)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(155),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(353)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(158),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(354)] = {
    [ts_builtin_sym_end] = ACTIONS(1280),
    [sym_identifier] = ACTIONS(1278),
    [aux_sym_preproc_include_token1] = ACTIONS(1278),
    [aux_sym_preproc_def_token1] = ACTIONS(1278),
    [aux_sym_preproc_if_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1278),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1278),
    [sym_preproc_directive] = ACTIONS(1278),
    [anon_sym_LPAREN2] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1278),
    [anon_sym_PLUS] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1280),
    [anon_sym___extension__] = ACTIONS(1278),
    [anon_sym_typedef] = ACTIONS(1278),
    [anon_sym_extern] = ACTIONS(1278),
    [anon_sym___attribute__] = ACTIONS(1278),
    [anon_sym___attribute] = ACTIONS(1278),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1280),
    [anon_sym___declspec] = ACTIONS(1278),
    [anon_sym___cdecl] = ACTIONS(1278),
    [anon_sym___clrcall] = ACTIONS(1278),
    [anon_sym___stdcall] = ACTIONS(1278),
    [anon_sym___fastcall] = ACTIONS(1278),
    [anon_sym___thiscall] = ACTIONS(1278),
    [anon_sym___vectorcall] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_signed] = ACTIONS(1278),
    [anon_sym_unsigned] = ACTIONS(1278),
    [anon_sym_long] = ACTIONS(1278),
    [anon_sym_short] = ACTIONS(1278),
    [anon_sym_static] = ACTIONS(1278),
    [anon_sym_auto] = ACTIONS(1278),
    [anon_sym_register] = ACTIONS(1278),
    [anon_sym_inline] = ACTIONS(1278),
    [anon_sym___inline] = ACTIONS(1278),
    [anon_sym___inline__] = ACTIONS(1278),
    [anon_sym___forceinline] = ACTIONS(1278),
    [anon_sym_thread_local] = ACTIONS(1278),
    [anon_sym___thread] = ACTIONS(1278),
    [anon_sym_const] = ACTIONS(1278),
    [anon_sym_constexpr] = ACTIONS(1278),
    [anon_sym_volatile] = ACTIONS(1278),
    [anon_sym_restrict] = ACTIONS(1278),
    [anon_sym___restrict__] = ACTIONS(1278),
    [anon_sym__Atomic] = ACTIONS(1278),
    [anon_sym__Noreturn] = ACTIONS(1278),
    [anon_sym_noreturn] = ACTIONS(1278),
    [anon_sym__Nonnull] = ACTIONS(1278),
    [anon_sym_alignas] = ACTIONS(1278),
    [anon_sym__Alignas] = ACTIONS(1278),
    [sym_primitive_type] = ACTIONS(1278),
    [anon_sym_enum] = ACTIONS(1278),
    [anon_sym_struct] = ACTIONS(1278),
    [anon_sym_union] = ACTIONS(1278),
    [anon_sym_if] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1278),
    [anon_sym_case] = ACTIONS(1278),
    [anon_sym_default] = ACTIONS(1278),
    [anon_sym_while] = ACTIONS(1278),
    [anon_sym_do] = ACTIONS(1278),
    [anon_sym_for] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1278),
    [anon_sym_break] = ACTIONS(1278),
    [anon_sym_continue] = ACTIONS(1278),
    [anon_sym_goto] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1280),
    [anon_sym_sizeof] = ACTIONS(1278),
    [anon_sym___alignof__] = ACTIONS(1278),
    [anon_sym___alignof] = ACTIONS(1278),
    [anon_sym__alignof] = ACTIONS(1278),
    [anon_sym_alignof] = ACTIONS(1278),
    [anon_sym__Alignof] = ACTIONS(1278),
    [anon_sym_offsetof] = ACTIONS(1278),
    [anon_sym__Generic] = ACTIONS(1278),
    [anon_sym_asm] = ACTIONS(1278),
    [anon_sym___asm__] = ACTIONS(1278),
    [anon_sym___asm] = ACTIONS(1278),
    [sym_number_literal] = ACTIONS(1280),
    [anon_sym_L_SQUOTE] = ACTIONS(1280),
    [anon_sym_u_SQUOTE] = ACTIONS(1280),
    [anon_sym_U_SQUOTE] = ACTIONS(1280),
    [anon_sym_u8_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_L_DQUOTE] = ACTIONS(1280),
    [anon_sym_u_DQUOTE] = ACTIONS(1280),
    [anon_sym_U_DQUOTE] = ACTIONS(1280),
    [anon_sym_u8_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [sym_true] = ACTIONS(1278),
    [sym_false] = ACTIONS(1278),
    [anon_sym_NULL] = ACTIONS(1278),
    [anon_sym_nullptr] = ACTIONS(1278),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(355)] = {
    [sym_attribute_declaration] = STATE(350),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(252),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(350),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(356)] = {
    [ts_builtin_sym_end] = ACTIONS(1611),
    [sym_identifier] = ACTIONS(1614),
    [aux_sym_preproc_include_token1] = ACTIONS(1614),
    [aux_sym_preproc_def_token1] = ACTIONS(1614),
    [aux_sym_preproc_if_token1] = ACTIONS(1614),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1614),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1614),
    [sym_preproc_directive] = ACTIONS(1614),
    [anon_sym_LPAREN2] = ACTIONS(1611),
    [anon_sym_BANG] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_DASH] = ACTIONS(1614),
    [anon_sym_PLUS] = ACTIONS(1614),
    [anon_sym_STAR] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1611),
    [anon_sym___extension__] = ACTIONS(1614),
    [anon_sym_typedef] = ACTIONS(1614),
    [anon_sym_extern] = ACTIONS(1614),
    [anon_sym___attribute__] = ACTIONS(1614),
    [anon_sym___attribute] = ACTIONS(1614),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1611),
    [anon_sym___declspec] = ACTIONS(1614),
    [anon_sym___cdecl] = ACTIONS(1614),
    [anon_sym___clrcall] = ACTIONS(1614),
    [anon_sym___stdcall] = ACTIONS(1614),
    [anon_sym___fastcall] = ACTIONS(1614),
    [anon_sym___thiscall] = ACTIONS(1614),
    [anon_sym___vectorcall] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [anon_sym_signed] = ACTIONS(1614),
    [anon_sym_unsigned] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_auto] = ACTIONS(1614),
    [anon_sym_register] = ACTIONS(1614),
    [anon_sym_inline] = ACTIONS(1614),
    [anon_sym___inline] = ACTIONS(1614),
    [anon_sym___inline__] = ACTIONS(1614),
    [anon_sym___forceinline] = ACTIONS(1614),
    [anon_sym_thread_local] = ACTIONS(1614),
    [anon_sym___thread] = ACTIONS(1614),
    [anon_sym_const] = ACTIONS(1614),
    [anon_sym_constexpr] = ACTIONS(1614),
    [anon_sym_volatile] = ACTIONS(1614),
    [anon_sym_restrict] = ACTIONS(1614),
    [anon_sym___restrict__] = ACTIONS(1614),
    [anon_sym__Atomic] = ACTIONS(1614),
    [anon_sym__Noreturn] = ACTIONS(1614),
    [anon_sym_noreturn] = ACTIONS(1614),
    [anon_sym__Nonnull] = ACTIONS(1614),
    [anon_sym_alignas] = ACTIONS(1614),
    [anon_sym__Alignas] = ACTIONS(1614),
    [sym_primitive_type] = ACTIONS(1614),
    [anon_sym_enum] = ACTIONS(1614),
    [anon_sym_struct] = ACTIONS(1614),
    [anon_sym_union] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_case] = ACTIONS(1614),
    [anon_sym_default] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_goto] = ACTIONS(1614),
    [anon_sym_DASH_DASH] = ACTIONS(1611),
    [anon_sym_PLUS_PLUS] = ACTIONS(1611),
    [anon_sym_sizeof] = ACTIONS(1614),
    [anon_sym___alignof__] = ACTIONS(1614),
    [anon_sym___alignof] = ACTIONS(1614),
    [anon_sym__alignof] = ACTIONS(1614),
    [anon_sym_alignof] = ACTIONS(1614),
    [anon_sym__Alignof] = ACTIONS(1614),
    [anon_sym_offsetof] = ACTIONS(1614),
    [anon_sym__Generic] = ACTIONS(1614),
    [anon_sym_asm] = ACTIONS(1614),
    [anon_sym___asm__] = ACTIONS(1614),
    [anon_sym___asm] = ACTIONS(1614),
    [sym_number_literal] = ACTIONS(1611),
    [anon_sym_L_SQUOTE] = ACTIONS(1611),
    [anon_sym_u_SQUOTE] = ACTIONS(1611),
    [anon_sym_U_SQUOTE] = ACTIONS(1611),
    [anon_sym_u8_SQUOTE] = ACTIONS(1611),
    [anon_sym_SQUOTE] = ACTIONS(1611),
    [anon_sym_L_DQUOTE] = ACTIONS(1611),
    [anon_sym_u_DQUOTE] = ACTIONS(1611),
    [anon_sym_U_DQUOTE] = ACTIONS(1611),
    [anon_sym_u8_DQUOTE] = ACTIONS(1611),
    [anon_sym_DQUOTE] = ACTIONS(1611),
    [sym_true] = ACTIONS(1614),
    [sym_false] = ACTIONS(1614),
    [anon_sym_NULL] = ACTIONS(1614),
    [anon_sym_nullptr] = ACTIONS(1614),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(357)] = {
    [sym_attribute_declaration] = STATE(371),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(148),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(358)] = {
    [sym_attribute_declaration] = STATE(371),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(149),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(359)] = {
    [sym_attribute_declaration] = STATE(359),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(183),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(359),
    [sym_identifier] = ACTIONS(1619),
    [anon_sym_LPAREN2] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym___extension__] = ACTIONS(1438),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1625),
    [anon_sym_if] = ACTIONS(1628),
    [anon_sym_switch] = ACTIONS(1631),
    [anon_sym_case] = ACTIONS(1634),
    [anon_sym_default] = ACTIONS(1637),
    [anon_sym_while] = ACTIONS(1640),
    [anon_sym_do] = ACTIONS(1643),
    [anon_sym_for] = ACTIONS(1646),
    [anon_sym_return] = ACTIONS(1649),
    [anon_sym_break] = ACTIONS(1652),
    [anon_sym_continue] = ACTIONS(1655),
    [anon_sym_goto] = ACTIONS(1658),
    [anon_sym___try] = ACTIONS(1661),
    [anon_sym___leave] = ACTIONS(1664),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [anon_sym_sizeof] = ACTIONS(1489),
    [anon_sym___alignof__] = ACTIONS(1492),
    [anon_sym___alignof] = ACTIONS(1492),
    [anon_sym__alignof] = ACTIONS(1492),
    [anon_sym_alignof] = ACTIONS(1492),
    [anon_sym__Alignof] = ACTIONS(1492),
    [anon_sym_offsetof] = ACTIONS(1495),
    [anon_sym__Generic] = ACTIONS(1498),
    [anon_sym_asm] = ACTIONS(1501),
    [anon_sym___asm__] = ACTIONS(1501),
    [anon_sym___asm] = ACTIONS(1501),
    [sym_number_literal] = ACTIONS(1504),
    [anon_sym_L_SQUOTE] = ACTIONS(1507),
    [anon_sym_u_SQUOTE] = ACTIONS(1507),
    [anon_sym_U_SQUOTE] = ACTIONS(1507),
    [anon_sym_u8_SQUOTE] = ACTIONS(1507),
    [anon_sym_SQUOTE] = ACTIONS(1507),
    [anon_sym_L_DQUOTE] = ACTIONS(1510),
    [anon_sym_u_DQUOTE] = ACTIONS(1510),
    [anon_sym_U_DQUOTE] = ACTIONS(1510),
    [anon_sym_u8_DQUOTE] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [sym_true] = ACTIONS(1513),
    [sym_false] = ACTIONS(1513),
    [anon_sym_NULL] = ACTIONS(1516),
    [anon_sym_nullptr] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(360)] = {
    [sym_attribute_declaration] = STATE(360),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(252),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(360),
    [sym_identifier] = ACTIONS(1667),
    [anon_sym_LPAREN2] = ACTIONS(1423),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1670),
    [anon_sym___extension__] = ACTIONS(1438),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1441),
    [anon_sym_LBRACE] = ACTIONS(1572),
    [anon_sym_if] = ACTIONS(1673),
    [anon_sym_switch] = ACTIONS(1578),
    [anon_sym_case] = ACTIONS(1676),
    [anon_sym_default] = ACTIONS(1679),
    [anon_sym_while] = ACTIONS(1682),
    [anon_sym_do] = ACTIONS(1590),
    [anon_sym_for] = ACTIONS(1685),
    [anon_sym_return] = ACTIONS(1596),
    [anon_sym_break] = ACTIONS(1599),
    [anon_sym_continue] = ACTIONS(1602),
    [anon_sym_goto] = ACTIONS(1605),
    [anon_sym___try] = ACTIONS(1688),
    [anon_sym___leave] = ACTIONS(1691),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [anon_sym_sizeof] = ACTIONS(1489),
    [anon_sym___alignof__] = ACTIONS(1492),
    [anon_sym___alignof] = ACTIONS(1492),
    [anon_sym__alignof] = ACTIONS(1492),
    [anon_sym_alignof] = ACTIONS(1492),
    [anon_sym__Alignof] = ACTIONS(1492),
    [anon_sym_offsetof] = ACTIONS(1495),
    [anon_sym__Generic] = ACTIONS(1498),
    [anon_sym_asm] = ACTIONS(1501),
    [anon_sym___asm__] = ACTIONS(1501),
    [anon_sym___asm] = ACTIONS(1501),
    [sym_number_literal] = ACTIONS(1504),
    [anon_sym_L_SQUOTE] = ACTIONS(1507),
    [anon_sym_u_SQUOTE] = ACTIONS(1507),
    [anon_sym_U_SQUOTE] = ACTIONS(1507),
    [anon_sym_u8_SQUOTE] = ACTIONS(1507),
    [anon_sym_SQUOTE] = ACTIONS(1507),
    [anon_sym_L_DQUOTE] = ACTIONS(1510),
    [anon_sym_u_DQUOTE] = ACTIONS(1510),
    [anon_sym_U_DQUOTE] = ACTIONS(1510),
    [anon_sym_u8_DQUOTE] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [sym_true] = ACTIONS(1513),
    [sym_false] = ACTIONS(1513),
    [anon_sym_NULL] = ACTIONS(1516),
    [anon_sym_nullptr] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(361)] = {
    [ts_builtin_sym_end] = ACTIONS(1370),
    [sym_identifier] = ACTIONS(1368),
    [aux_sym_preproc_include_token1] = ACTIONS(1368),
    [aux_sym_preproc_def_token1] = ACTIONS(1368),
    [aux_sym_preproc_if_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1368),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1368),
    [sym_preproc_directive] = ACTIONS(1368),
    [anon_sym_LPAREN2] = ACTIONS(1370),
    [anon_sym_BANG] = ACTIONS(1370),
    [anon_sym_TILDE] = ACTIONS(1370),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_PLUS] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1370),
    [anon_sym_AMP] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1370),
    [anon_sym___extension__] = ACTIONS(1368),
    [anon_sym_typedef] = ACTIONS(1368),
    [anon_sym_extern] = ACTIONS(1368),
    [anon_sym___attribute__] = ACTIONS(1368),
    [anon_sym___attribute] = ACTIONS(1368),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1370),
    [anon_sym___declspec] = ACTIONS(1368),
    [anon_sym___cdecl] = ACTIONS(1368),
    [anon_sym___clrcall] = ACTIONS(1368),
    [anon_sym___stdcall] = ACTIONS(1368),
    [anon_sym___fastcall] = ACTIONS(1368),
    [anon_sym___thiscall] = ACTIONS(1368),
    [anon_sym___vectorcall] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1370),
    [anon_sym_signed] = ACTIONS(1368),
    [anon_sym_unsigned] = ACTIONS(1368),
    [anon_sym_long] = ACTIONS(1368),
    [anon_sym_short] = ACTIONS(1368),
    [anon_sym_static] = ACTIONS(1368),
    [anon_sym_auto] = ACTIONS(1368),
    [anon_sym_register] = ACTIONS(1368),
    [anon_sym_inline] = ACTIONS(1368),
    [anon_sym___inline] = ACTIONS(1368),
    [anon_sym___inline__] = ACTIONS(1368),
    [anon_sym___forceinline] = ACTIONS(1368),
    [anon_sym_thread_local] = ACTIONS(1368),
    [anon_sym___thread] = ACTIONS(1368),
    [anon_sym_const] = ACTIONS(1368),
    [anon_sym_constexpr] = ACTIONS(1368),
    [anon_sym_volatile] = ACTIONS(1368),
    [anon_sym_restrict] = ACTIONS(1368),
    [anon_sym___restrict__] = ACTIONS(1368),
    [anon_sym__Atomic] = ACTIONS(1368),
    [anon_sym__Noreturn] = ACTIONS(1368),
    [anon_sym_noreturn] = ACTIONS(1368),
    [anon_sym__Nonnull] = ACTIONS(1368),
    [anon_sym_alignas] = ACTIONS(1368),
    [anon_sym__Alignas] = ACTIONS(1368),
    [sym_primitive_type] = ACTIONS(1368),
    [anon_sym_enum] = ACTIONS(1368),
    [anon_sym_struct] = ACTIONS(1368),
    [anon_sym_union] = ACTIONS(1368),
    [anon_sym_if] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1368),
    [anon_sym_case] = ACTIONS(1368),
    [anon_sym_default] = ACTIONS(1368),
    [anon_sym_while] = ACTIONS(1368),
    [anon_sym_do] = ACTIONS(1368),
    [anon_sym_for] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1368),
    [anon_sym_break] = ACTIONS(1368),
    [anon_sym_continue] = ACTIONS(1368),
    [anon_sym_goto] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1370),
    [anon_sym_PLUS_PLUS] = ACTIONS(1370),
    [anon_sym_sizeof] = ACTIONS(1368),
    [anon_sym___alignof__] = ACTIONS(1368),
    [anon_sym___alignof] = ACTIONS(1368),
    [anon_sym__alignof] = ACTIONS(1368),
    [anon_sym_alignof] = ACTIONS(1368),
    [anon_sym__Alignof] = ACTIONS(1368),
    [anon_sym_offsetof] = ACTIONS(1368),
    [anon_sym__Generic] = ACTIONS(1368),
    [anon_sym_asm] = ACTIONS(1368),
    [anon_sym___asm__] = ACTIONS(1368),
    [anon_sym___asm] = ACTIONS(1368),
    [sym_number_literal] = ACTIONS(1370),
    [anon_sym_L_SQUOTE] = ACTIONS(1370),
    [anon_sym_u_SQUOTE] = ACTIONS(1370),
    [anon_sym_U_SQUOTE] = ACTIONS(1370),
    [anon_sym_u8_SQUOTE] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_L_DQUOTE] = ACTIONS(1370),
    [anon_sym_u_DQUOTE] = ACTIONS(1370),
    [anon_sym_U_DQUOTE] = ACTIONS(1370),
    [anon_sym_u8_DQUOTE] = ACTIONS(1370),
    [anon_sym_DQUOTE] = ACTIONS(1370),
    [sym_true] = ACTIONS(1368),
    [sym_false] = ACTIONS(1368),
    [anon_sym_NULL] = ACTIONS(1368),
    [anon_sym_nullptr] = ACTIONS(1368),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(362)] = {
    [sym_attribute_declaration] = STATE(342),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(147),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(363)] = {
    [ts_builtin_sym_end] = ACTIONS(1362),
    [sym_identifier] = ACTIONS(1360),
    [aux_sym_preproc_include_token1] = ACTIONS(1360),
    [aux_sym_preproc_def_token1] = ACTIONS(1360),
    [aux_sym_preproc_if_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1360),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1360),
    [sym_preproc_directive] = ACTIONS(1360),
    [anon_sym_LPAREN2] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_PLUS] = ACTIONS(1360),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [anon_sym___extension__] = ACTIONS(1360),
    [anon_sym_typedef] = ACTIONS(1360),
    [anon_sym_extern] = ACTIONS(1360),
    [anon_sym___attribute__] = ACTIONS(1360),
    [anon_sym___attribute] = ACTIONS(1360),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1362),
    [anon_sym___declspec] = ACTIONS(1360),
    [anon_sym___cdecl] = ACTIONS(1360),
    [anon_sym___clrcall] = ACTIONS(1360),
    [anon_sym___stdcall] = ACTIONS(1360),
    [anon_sym___fastcall] = ACTIONS(1360),
    [anon_sym___thiscall] = ACTIONS(1360),
    [anon_sym___vectorcall] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_signed] = ACTIONS(1360),
    [anon_sym_unsigned] = ACTIONS(1360),
    [anon_sym_long] = ACTIONS(1360),
    [anon_sym_short] = ACTIONS(1360),
    [anon_sym_static] = ACTIONS(1360),
    [anon_sym_auto] = ACTIONS(1360),
    [anon_sym_register] = ACTIONS(1360),
    [anon_sym_inline] = ACTIONS(1360),
    [anon_sym___inline] = ACTIONS(1360),
    [anon_sym___inline__] = ACTIONS(1360),
    [anon_sym___forceinline] = ACTIONS(1360),
    [anon_sym_thread_local] = ACTIONS(1360),
    [anon_sym___thread] = ACTIONS(1360),
    [anon_sym_const] = ACTIONS(1360),
    [anon_sym_constexpr] = ACTIONS(1360),
    [anon_sym_volatile] = ACTIONS(1360),
    [anon_sym_restrict] = ACTIONS(1360),
    [anon_sym___restrict__] = ACTIONS(1360),
    [anon_sym__Atomic] = ACTIONS(1360),
    [anon_sym__Noreturn] = ACTIONS(1360),
    [anon_sym_noreturn] = ACTIONS(1360),
    [anon_sym__Nonnull] = ACTIONS(1360),
    [anon_sym_alignas] = ACTIONS(1360),
    [anon_sym__Alignas] = ACTIONS(1360),
    [sym_primitive_type] = ACTIONS(1360),
    [anon_sym_enum] = ACTIONS(1360),
    [anon_sym_struct] = ACTIONS(1360),
    [anon_sym_union] = ACTIONS(1360),
    [anon_sym_if] = ACTIONS(1360),
    [anon_sym_switch] = ACTIONS(1360),
    [anon_sym_case] = ACTIONS(1360),
    [anon_sym_default] = ACTIONS(1360),
    [anon_sym_while] = ACTIONS(1360),
    [anon_sym_do] = ACTIONS(1360),
    [anon_sym_for] = ACTIONS(1360),
    [anon_sym_return] = ACTIONS(1360),
    [anon_sym_break] = ACTIONS(1360),
    [anon_sym_continue] = ACTIONS(1360),
    [anon_sym_goto] = ACTIONS(1360),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [anon_sym_sizeof] = ACTIONS(1360),
    [anon_sym___alignof__] = ACTIONS(1360),
    [anon_sym___alignof] = ACTIONS(1360),
    [anon_sym__alignof] = ACTIONS(1360),
    [anon_sym_alignof] = ACTIONS(1360),
    [anon_sym__Alignof] = ACTIONS(1360),
    [anon_sym_offsetof] = ACTIONS(1360),
    [anon_sym__Generic] = ACTIONS(1360),
    [anon_sym_asm] = ACTIONS(1360),
    [anon_sym___asm__] = ACTIONS(1360),
    [anon_sym___asm] = ACTIONS(1360),
    [sym_number_literal] = ACTIONS(1362),
    [anon_sym_L_SQUOTE] = ACTIONS(1362),
    [anon_sym_u_SQUOTE] = ACTIONS(1362),
    [anon_sym_U_SQUOTE] = ACTIONS(1362),
    [anon_sym_u8_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_L_DQUOTE] = ACTIONS(1362),
    [anon_sym_u_DQUOTE] = ACTIONS(1362),
    [anon_sym_U_DQUOTE] = ACTIONS(1362),
    [anon_sym_u8_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [sym_true] = ACTIONS(1360),
    [sym_false] = ACTIONS(1360),
    [anon_sym_NULL] = ACTIONS(1360),
    [anon_sym_nullptr] = ACTIONS(1360),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(364)] = {
    [sym_attribute_declaration] = STATE(371),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(203),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(365)] = {
    [sym_attribute_declaration] = STATE(371),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(254),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(371),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(366)] = {
    [sym_attribute_declaration] = STATE(342),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(198),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(367)] = {
    [sym_attribute_declaration] = STATE(345),
    [sym_compound_statement] = STATE(81),
    [sym_attributed_statement] = STATE(81),
    [sym_statement] = STATE(114),
    [sym_labeled_statement] = STATE(81),
    [sym_expression_statement] = STATE(81),
    [sym_if_statement] = STATE(81),
    [sym_switch_statement] = STATE(81),
    [sym_case_statement] = STATE(81),
    [sym_while_statement] = STATE(81),
    [sym_do_statement] = STATE(81),
    [sym_for_statement] = STATE(81),
    [sym_return_statement] = STATE(81),
    [sym_break_statement] = STATE(81),
    [sym_continue_statement] = STATE(81),
    [sym_goto_statement] = STATE(81),
    [sym_seh_try_statement] = STATE(81),
    [sym_seh_leave_statement] = STATE(81),
    [sym_expression] = STATE(1055),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1904),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(345),
    [sym_identifier] = ACTIONS(1519),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(125),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(133),
    [anon_sym_if] = ACTIONS(135),
    [anon_sym_switch] = ACTIONS(137),
    [anon_sym_case] = ACTIONS(139),
    [anon_sym_default] = ACTIONS(141),
    [anon_sym_while] = ACTIONS(143),
    [anon_sym_do] = ACTIONS(145),
    [anon_sym_for] = ACTIONS(147),
    [anon_sym_return] = ACTIONS(149),
    [anon_sym_break] = ACTIONS(151),
    [anon_sym_continue] = ACTIONS(153),
    [anon_sym_goto] = ACTIONS(155),
    [anon_sym___try] = ACTIONS(157),
    [anon_sym___leave] = ACTIONS(159),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(368)] = {
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1266),
    [aux_sym_preproc_include_token1] = ACTIONS(1266),
    [aux_sym_preproc_def_token1] = ACTIONS(1266),
    [aux_sym_preproc_if_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1266),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1266),
    [sym_preproc_directive] = ACTIONS(1266),
    [anon_sym_LPAREN2] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1266),
    [anon_sym_PLUS] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1268),
    [anon_sym_AMP] = ACTIONS(1268),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym___extension__] = ACTIONS(1266),
    [anon_sym_typedef] = ACTIONS(1266),
    [anon_sym_extern] = ACTIONS(1266),
    [anon_sym___attribute__] = ACTIONS(1266),
    [anon_sym___attribute] = ACTIONS(1266),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1268),
    [anon_sym___declspec] = ACTIONS(1266),
    [anon_sym___cdecl] = ACTIONS(1266),
    [anon_sym___clrcall] = ACTIONS(1266),
    [anon_sym___stdcall] = ACTIONS(1266),
    [anon_sym___fastcall] = ACTIONS(1266),
    [anon_sym___thiscall] = ACTIONS(1266),
    [anon_sym___vectorcall] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_signed] = ACTIONS(1266),
    [anon_sym_unsigned] = ACTIONS(1266),
    [anon_sym_long] = ACTIONS(1266),
    [anon_sym_short] = ACTIONS(1266),
    [anon_sym_static] = ACTIONS(1266),
    [anon_sym_auto] = ACTIONS(1266),
    [anon_sym_register] = ACTIONS(1266),
    [anon_sym_inline] = ACTIONS(1266),
    [anon_sym___inline] = ACTIONS(1266),
    [anon_sym___inline__] = ACTIONS(1266),
    [anon_sym___forceinline] = ACTIONS(1266),
    [anon_sym_thread_local] = ACTIONS(1266),
    [anon_sym___thread] = ACTIONS(1266),
    [anon_sym_const] = ACTIONS(1266),
    [anon_sym_constexpr] = ACTIONS(1266),
    [anon_sym_volatile] = ACTIONS(1266),
    [anon_sym_restrict] = ACTIONS(1266),
    [anon_sym___restrict__] = ACTIONS(1266),
    [anon_sym__Atomic] = ACTIONS(1266),
    [anon_sym__Noreturn] = ACTIONS(1266),
    [anon_sym_noreturn] = ACTIONS(1266),
    [anon_sym__Nonnull] = ACTIONS(1266),
    [anon_sym_alignas] = ACTIONS(1266),
    [anon_sym__Alignas] = ACTIONS(1266),
    [sym_primitive_type] = ACTIONS(1266),
    [anon_sym_enum] = ACTIONS(1266),
    [anon_sym_struct] = ACTIONS(1266),
    [anon_sym_union] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_switch] = ACTIONS(1266),
    [anon_sym_case] = ACTIONS(1266),
    [anon_sym_default] = ACTIONS(1266),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_break] = ACTIONS(1266),
    [anon_sym_continue] = ACTIONS(1266),
    [anon_sym_goto] = ACTIONS(1266),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_sizeof] = ACTIONS(1266),
    [anon_sym___alignof__] = ACTIONS(1266),
    [anon_sym___alignof] = ACTIONS(1266),
    [anon_sym__alignof] = ACTIONS(1266),
    [anon_sym_alignof] = ACTIONS(1266),
    [anon_sym__Alignof] = ACTIONS(1266),
    [anon_sym_offsetof] = ACTIONS(1266),
    [anon_sym__Generic] = ACTIONS(1266),
    [anon_sym_asm] = ACTIONS(1266),
    [anon_sym___asm__] = ACTIONS(1266),
    [anon_sym___asm] = ACTIONS(1266),
    [sym_number_literal] = ACTIONS(1268),
    [anon_sym_L_SQUOTE] = ACTIONS(1268),
    [anon_sym_u_SQUOTE] = ACTIONS(1268),
    [anon_sym_U_SQUOTE] = ACTIONS(1268),
    [anon_sym_u8_SQUOTE] = ACTIONS(1268),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_L_DQUOTE] = ACTIONS(1268),
    [anon_sym_u_DQUOTE] = ACTIONS(1268),
    [anon_sym_U_DQUOTE] = ACTIONS(1268),
    [anon_sym_u8_DQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1268),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [anon_sym_NULL] = ACTIONS(1266),
    [anon_sym_nullptr] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(369)] = {
    [sym_attribute_declaration] = STATE(360),
    [sym_compound_statement] = STATE(154),
    [sym_attributed_statement] = STATE(154),
    [sym_statement] = STATE(252),
    [sym_labeled_statement] = STATE(154),
    [sym_expression_statement] = STATE(154),
    [sym_if_statement] = STATE(154),
    [sym_switch_statement] = STATE(154),
    [sym_case_statement] = STATE(154),
    [sym_while_statement] = STATE(154),
    [sym_do_statement] = STATE(154),
    [sym_for_statement] = STATE(154),
    [sym_return_statement] = STATE(154),
    [sym_break_statement] = STATE(154),
    [sym_continue_statement] = STATE(154),
    [sym_goto_statement] = STATE(154),
    [sym_seh_try_statement] = STATE(154),
    [sym_seh_leave_statement] = STATE(154),
    [sym_expression] = STATE(1035),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1977),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(360),
    [sym_identifier] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(65),
    [anon_sym_default] = ACTIONS(67),
    [anon_sym_while] = ACTIONS(69),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(935),
    [anon_sym___leave] = ACTIONS(937),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(370)] = {
    [ts_builtin_sym_end] = ACTIONS(1694),
    [sym_identifier] = ACTIONS(1696),
    [aux_sym_preproc_include_token1] = ACTIONS(1696),
    [aux_sym_preproc_def_token1] = ACTIONS(1696),
    [aux_sym_preproc_if_token1] = ACTIONS(1696),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1696),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1696),
    [sym_preproc_directive] = ACTIONS(1696),
    [anon_sym_LPAREN2] = ACTIONS(1694),
    [anon_sym_BANG] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_DASH] = ACTIONS(1696),
    [anon_sym_PLUS] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1694),
    [anon_sym___extension__] = ACTIONS(1696),
    [anon_sym_typedef] = ACTIONS(1696),
    [anon_sym_extern] = ACTIONS(1696),
    [anon_sym___attribute__] = ACTIONS(1696),
    [anon_sym___attribute] = ACTIONS(1696),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1694),
    [anon_sym___declspec] = ACTIONS(1696),
    [anon_sym___cdecl] = ACTIONS(1696),
    [anon_sym___clrcall] = ACTIONS(1696),
    [anon_sym___stdcall] = ACTIONS(1696),
    [anon_sym___fastcall] = ACTIONS(1696),
    [anon_sym___thiscall] = ACTIONS(1696),
    [anon_sym___vectorcall] = ACTIONS(1696),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_signed] = ACTIONS(1696),
    [anon_sym_unsigned] = ACTIONS(1696),
    [anon_sym_long] = ACTIONS(1696),
    [anon_sym_short] = ACTIONS(1696),
    [anon_sym_static] = ACTIONS(1696),
    [anon_sym_auto] = ACTIONS(1696),
    [anon_sym_register] = ACTIONS(1696),
    [anon_sym_inline] = ACTIONS(1696),
    [anon_sym___inline] = ACTIONS(1696),
    [anon_sym___inline__] = ACTIONS(1696),
    [anon_sym___forceinline] = ACTIONS(1696),
    [anon_sym_thread_local] = ACTIONS(1696),
    [anon_sym___thread] = ACTIONS(1696),
    [anon_sym_const] = ACTIONS(1696),
    [anon_sym_constexpr] = ACTIONS(1696),
    [anon_sym_volatile] = ACTIONS(1696),
    [anon_sym_restrict] = ACTIONS(1696),
    [anon_sym___restrict__] = ACTIONS(1696),
    [anon_sym__Atomic] = ACTIONS(1696),
    [anon_sym__Noreturn] = ACTIONS(1696),
    [anon_sym_noreturn] = ACTIONS(1696),
    [anon_sym__Nonnull] = ACTIONS(1696),
    [anon_sym_alignas] = ACTIONS(1696),
    [anon_sym__Alignas] = ACTIONS(1696),
    [sym_primitive_type] = ACTIONS(1696),
    [anon_sym_enum] = ACTIONS(1696),
    [anon_sym_struct] = ACTIONS(1696),
    [anon_sym_union] = ACTIONS(1696),
    [anon_sym_if] = ACTIONS(1696),
    [anon_sym_switch] = ACTIONS(1696),
    [anon_sym_case] = ACTIONS(1696),
    [anon_sym_default] = ACTIONS(1696),
    [anon_sym_while] = ACTIONS(1696),
    [anon_sym_do] = ACTIONS(1696),
    [anon_sym_for] = ACTIONS(1696),
    [anon_sym_return] = ACTIONS(1696),
    [anon_sym_break] = ACTIONS(1696),
    [anon_sym_continue] = ACTIONS(1696),
    [anon_sym_goto] = ACTIONS(1696),
    [anon_sym_DASH_DASH] = ACTIONS(1694),
    [anon_sym_PLUS_PLUS] = ACTIONS(1694),
    [anon_sym_sizeof] = ACTIONS(1696),
    [anon_sym___alignof__] = ACTIONS(1696),
    [anon_sym___alignof] = ACTIONS(1696),
    [anon_sym__alignof] = ACTIONS(1696),
    [anon_sym_alignof] = ACTIONS(1696),
    [anon_sym__Alignof] = ACTIONS(1696),
    [anon_sym_offsetof] = ACTIONS(1696),
    [anon_sym__Generic] = ACTIONS(1696),
    [anon_sym_asm] = ACTIONS(1696),
    [anon_sym___asm__] = ACTIONS(1696),
    [anon_sym___asm] = ACTIONS(1696),
    [sym_number_literal] = ACTIONS(1694),
    [anon_sym_L_SQUOTE] = ACTIONS(1694),
    [anon_sym_u_SQUOTE] = ACTIONS(1694),
    [anon_sym_U_SQUOTE] = ACTIONS(1694),
    [anon_sym_u8_SQUOTE] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1694),
    [anon_sym_L_DQUOTE] = ACTIONS(1694),
    [anon_sym_u_DQUOTE] = ACTIONS(1694),
    [anon_sym_U_DQUOTE] = ACTIONS(1694),
    [anon_sym_u8_DQUOTE] = ACTIONS(1694),
    [anon_sym_DQUOTE] = ACTIONS(1694),
    [sym_true] = ACTIONS(1696),
    [sym_false] = ACTIONS(1696),
    [anon_sym_NULL] = ACTIONS(1696),
    [anon_sym_nullptr] = ACTIONS(1696),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(371)] = {
    [sym_attribute_declaration] = STATE(359),
    [sym_compound_statement] = STATE(178),
    [sym_attributed_statement] = STATE(178),
    [sym_statement] = STATE(183),
    [sym_labeled_statement] = STATE(178),
    [sym_expression_statement] = STATE(178),
    [sym_if_statement] = STATE(178),
    [sym_switch_statement] = STATE(178),
    [sym_case_statement] = STATE(178),
    [sym_while_statement] = STATE(178),
    [sym_do_statement] = STATE(178),
    [sym_for_statement] = STATE(178),
    [sym_return_statement] = STATE(178),
    [sym_break_statement] = STATE(178),
    [sym_continue_statement] = STATE(178),
    [sym_goto_statement] = STATE(178),
    [sym_seh_try_statement] = STATE(178),
    [sym_seh_leave_statement] = STATE(178),
    [sym_expression] = STATE(1028),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1804),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(359),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_switch] = ACTIONS(434),
    [anon_sym_case] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(438),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_do] = ACTIONS(442),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(446),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(450),
    [anon_sym_goto] = ACTIONS(452),
    [anon_sym___try] = ACTIONS(454),
    [anon_sym___leave] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(372)] = {
    [sym_attribute_declaration] = STATE(355),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(1980),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(355),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_if] = ACTIONS(1095),
    [anon_sym_switch] = ACTIONS(63),
    [anon_sym_case] = ACTIONS(1103),
    [anon_sym_default] = ACTIONS(1105),
    [anon_sym_while] = ACTIONS(1097),
    [anon_sym_do] = ACTIONS(71),
    [anon_sym_for] = ACTIONS(1099),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_break] = ACTIONS(77),
    [anon_sym_continue] = ACTIONS(79),
    [anon_sym_goto] = ACTIONS(81),
    [anon_sym___try] = ACTIONS(1101),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(373)] = {
    [sym_attribute_declaration] = STATE(342),
    [sym_compound_statement] = STATE(244),
    [sym_attributed_statement] = STATE(244),
    [sym_statement] = STATE(234),
    [sym_labeled_statement] = STATE(244),
    [sym_expression_statement] = STATE(244),
    [sym_if_statement] = STATE(244),
    [sym_switch_statement] = STATE(244),
    [sym_case_statement] = STATE(244),
    [sym_while_statement] = STATE(244),
    [sym_do_statement] = STATE(244),
    [sym_for_statement] = STATE(244),
    [sym_return_statement] = STATE(244),
    [sym_break_statement] = STATE(244),
    [sym_continue_statement] = STATE(244),
    [sym_goto_statement] = STATE(244),
    [sym_seh_try_statement] = STATE(244),
    [sym_seh_leave_statement] = STATE(244),
    [sym_expression] = STATE(1056),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1942),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_attributed_declarator_repeat1] = STATE(342),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym_SEMI] = ACTIONS(370),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(378),
    [anon_sym_if] = ACTIONS(382),
    [anon_sym_switch] = ACTIONS(384),
    [anon_sym_case] = ACTIONS(386),
    [anon_sym_default] = ACTIONS(388),
    [anon_sym_while] = ACTIONS(390),
    [anon_sym_do] = ACTIONS(392),
    [anon_sym_for] = ACTIONS(394),
    [anon_sym_return] = ACTIONS(396),
    [anon_sym_break] = ACTIONS(398),
    [anon_sym_continue] = ACTIONS(400),
    [anon_sym_goto] = ACTIONS(402),
    [anon_sym___try] = ACTIONS(404),
    [anon_sym___leave] = ACTIONS(406),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(374)] = {
    [sym_expression] = STATE(980),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1701),
    [anon_sym_typedef] = ACTIONS(1704),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym___attribute__] = ACTIONS(1706),
    [anon_sym___attribute] = ACTIONS(1706),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1708),
    [anon_sym___declspec] = ACTIONS(1706),
    [anon_sym_signed] = ACTIONS(1706),
    [anon_sym_unsigned] = ACTIONS(1706),
    [anon_sym_long] = ACTIONS(1706),
    [anon_sym_short] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_auto] = ACTIONS(1706),
    [anon_sym_register] = ACTIONS(1706),
    [anon_sym_inline] = ACTIONS(1706),
    [anon_sym___inline] = ACTIONS(1706),
    [anon_sym___inline__] = ACTIONS(1706),
    [anon_sym___forceinline] = ACTIONS(1706),
    [anon_sym_thread_local] = ACTIONS(1706),
    [anon_sym___thread] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [sym_primitive_type] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(375)] = {
    [sym_expression] = STATE(980),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1701),
    [anon_sym_typedef] = ACTIONS(1710),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym___attribute__] = ACTIONS(1706),
    [anon_sym___attribute] = ACTIONS(1706),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1708),
    [anon_sym___declspec] = ACTIONS(1706),
    [anon_sym_signed] = ACTIONS(1706),
    [anon_sym_unsigned] = ACTIONS(1706),
    [anon_sym_long] = ACTIONS(1706),
    [anon_sym_short] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_auto] = ACTIONS(1706),
    [anon_sym_register] = ACTIONS(1706),
    [anon_sym_inline] = ACTIONS(1706),
    [anon_sym___inline] = ACTIONS(1706),
    [anon_sym___inline__] = ACTIONS(1706),
    [anon_sym___forceinline] = ACTIONS(1706),
    [anon_sym_thread_local] = ACTIONS(1706),
    [anon_sym___thread] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [sym_primitive_type] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(376)] = {
    [sym_expression] = STATE(980),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1701),
    [anon_sym_typedef] = ACTIONS(1712),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym___attribute__] = ACTIONS(1706),
    [anon_sym___attribute] = ACTIONS(1706),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1708),
    [anon_sym___declspec] = ACTIONS(1706),
    [anon_sym_signed] = ACTIONS(1706),
    [anon_sym_unsigned] = ACTIONS(1706),
    [anon_sym_long] = ACTIONS(1706),
    [anon_sym_short] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_auto] = ACTIONS(1706),
    [anon_sym_register] = ACTIONS(1706),
    [anon_sym_inline] = ACTIONS(1706),
    [anon_sym___inline] = ACTIONS(1706),
    [anon_sym___inline__] = ACTIONS(1706),
    [anon_sym___forceinline] = ACTIONS(1706),
    [anon_sym_thread_local] = ACTIONS(1706),
    [anon_sym___thread] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [sym_primitive_type] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(377)] = {
    [sym_expression] = STATE(980),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1701),
    [anon_sym_typedef] = ACTIONS(1714),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym___attribute__] = ACTIONS(1706),
    [anon_sym___attribute] = ACTIONS(1706),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1708),
    [anon_sym___declspec] = ACTIONS(1706),
    [anon_sym_signed] = ACTIONS(1706),
    [anon_sym_unsigned] = ACTIONS(1706),
    [anon_sym_long] = ACTIONS(1706),
    [anon_sym_short] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_auto] = ACTIONS(1706),
    [anon_sym_register] = ACTIONS(1706),
    [anon_sym_inline] = ACTIONS(1706),
    [anon_sym___inline] = ACTIONS(1706),
    [anon_sym___inline__] = ACTIONS(1706),
    [anon_sym___forceinline] = ACTIONS(1706),
    [anon_sym_thread_local] = ACTIONS(1706),
    [anon_sym___thread] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [sym_primitive_type] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(378)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1856),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(379)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1975),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(380)] = {
    [sym_expression] = STATE(980),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1701),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym___attribute__] = ACTIONS(1706),
    [anon_sym___attribute] = ACTIONS(1706),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1708),
    [anon_sym___declspec] = ACTIONS(1706),
    [anon_sym_signed] = ACTIONS(1706),
    [anon_sym_unsigned] = ACTIONS(1706),
    [anon_sym_long] = ACTIONS(1706),
    [anon_sym_short] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_auto] = ACTIONS(1706),
    [anon_sym_register] = ACTIONS(1706),
    [anon_sym_inline] = ACTIONS(1706),
    [anon_sym___inline] = ACTIONS(1706),
    [anon_sym___inline__] = ACTIONS(1706),
    [anon_sym___forceinline] = ACTIONS(1706),
    [anon_sym_thread_local] = ACTIONS(1706),
    [anon_sym___thread] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [sym_primitive_type] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(381)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1779),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(382)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1996),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(383)] = {
    [sym_expression] = STATE(889),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(675),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(675),
    [sym_call_expression] = STATE(675),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(675),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(675),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1378),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1722),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1386),
    [anon_sym___extension__] = ACTIONS(1726),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_EQ] = ACTIONS(1386),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_STAR_EQ] = ACTIONS(1380),
    [anon_sym_SLASH_EQ] = ACTIONS(1380),
    [anon_sym_PERCENT_EQ] = ACTIONS(1380),
    [anon_sym_PLUS_EQ] = ACTIONS(1380),
    [anon_sym_DASH_EQ] = ACTIONS(1380),
    [anon_sym_LT_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT_GT_EQ] = ACTIONS(1380),
    [anon_sym_AMP_EQ] = ACTIONS(1380),
    [anon_sym_CARET_EQ] = ACTIONS(1380),
    [anon_sym_PIPE_EQ] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(1728),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(384)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1793),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(385)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1940),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(386)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1930),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(387)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1887),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(388)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1768),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(389)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1775),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(390)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1825),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(391)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1862),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(392)] = {
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1084),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1885),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(393)] = {
    [sym_type_qualifier] = STATE(1006),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(1077),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_expression] = STATE(1080),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_type_descriptor] = STATE(1776),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__type_definition_type_repeat1] = STATE(1006),
    [aux_sym_sized_type_specifier_repeat1] = STATE(1026),
    [sym_identifier] = ACTIONS(1716),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1091),
    [anon_sym_signed] = ACTIONS(1718),
    [anon_sym_unsigned] = ACTIONS(1718),
    [anon_sym_long] = ACTIONS(1718),
    [anon_sym_short] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(1720),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(394)] = {
    [sym_identifier] = ACTIONS(1730),
    [anon_sym_COMMA] = ACTIONS(1732),
    [anon_sym_RPAREN] = ACTIONS(1732),
    [anon_sym_LPAREN2] = ACTIONS(1732),
    [anon_sym_BANG] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_DASH] = ACTIONS(1730),
    [anon_sym_PLUS] = ACTIONS(1730),
    [anon_sym_STAR] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym___extension__] = ACTIONS(1730),
    [anon_sym_extern] = ACTIONS(1730),
    [anon_sym___attribute__] = ACTIONS(1730),
    [anon_sym___attribute] = ACTIONS(1730),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1732),
    [anon_sym___declspec] = ACTIONS(1730),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_signed] = ACTIONS(1730),
    [anon_sym_unsigned] = ACTIONS(1730),
    [anon_sym_long] = ACTIONS(1730),
    [anon_sym_short] = ACTIONS(1730),
    [anon_sym_LBRACK] = ACTIONS(1730),
    [anon_sym_static] = ACTIONS(1730),
    [anon_sym_EQ] = ACTIONS(1732),
    [anon_sym_auto] = ACTIONS(1730),
    [anon_sym_register] = ACTIONS(1730),
    [anon_sym_inline] = ACTIONS(1730),
    [anon_sym___inline] = ACTIONS(1730),
    [anon_sym___inline__] = ACTIONS(1730),
    [anon_sym___forceinline] = ACTIONS(1730),
    [anon_sym_thread_local] = ACTIONS(1730),
    [anon_sym___thread] = ACTIONS(1730),
    [anon_sym_const] = ACTIONS(1730),
    [anon_sym_constexpr] = ACTIONS(1730),
    [anon_sym_volatile] = ACTIONS(1730),
    [anon_sym_restrict] = ACTIONS(1730),
    [anon_sym___restrict__] = ACTIONS(1730),
    [anon_sym__Atomic] = ACTIONS(1730),
    [anon_sym__Noreturn] = ACTIONS(1730),
    [anon_sym_noreturn] = ACTIONS(1730),
    [anon_sym__Nonnull] = ACTIONS(1730),
    [anon_sym_alignas] = ACTIONS(1730),
    [anon_sym__Alignas] = ACTIONS(1730),
    [sym_primitive_type] = ACTIONS(1730),
    [anon_sym_enum] = ACTIONS(1730),
    [anon_sym_COLON] = ACTIONS(1732),
    [anon_sym_struct] = ACTIONS(1730),
    [anon_sym_union] = ACTIONS(1730),
    [anon_sym_if] = ACTIONS(1730),
    [anon_sym_switch] = ACTIONS(1730),
    [anon_sym_case] = ACTIONS(1730),
    [anon_sym_default] = ACTIONS(1730),
    [anon_sym_while] = ACTIONS(1730),
    [anon_sym_do] = ACTIONS(1730),
    [anon_sym_for] = ACTIONS(1730),
    [anon_sym_return] = ACTIONS(1730),
    [anon_sym_break] = ACTIONS(1730),
    [anon_sym_continue] = ACTIONS(1730),
    [anon_sym_goto] = ACTIONS(1730),
    [anon_sym___try] = ACTIONS(1730),
    [anon_sym___leave] = ACTIONS(1730),
    [anon_sym_DASH_DASH] = ACTIONS(1732),
    [anon_sym_PLUS_PLUS] = ACTIONS(1732),
    [anon_sym_sizeof] = ACTIONS(1730),
    [anon_sym___alignof__] = ACTIONS(1730),
    [anon_sym___alignof] = ACTIONS(1730),
    [anon_sym__alignof] = ACTIONS(1730),
    [anon_sym_alignof] = ACTIONS(1730),
    [anon_sym__Alignof] = ACTIONS(1730),
    [anon_sym_offsetof] = ACTIONS(1730),
    [anon_sym__Generic] = ACTIONS(1730),
    [anon_sym_asm] = ACTIONS(1730),
    [anon_sym___asm__] = ACTIONS(1730),
    [anon_sym___asm] = ACTIONS(1730),
    [sym_number_literal] = ACTIONS(1732),
    [anon_sym_L_SQUOTE] = ACTIONS(1732),
    [anon_sym_u_SQUOTE] = ACTIONS(1732),
    [anon_sym_U_SQUOTE] = ACTIONS(1732),
    [anon_sym_u8_SQUOTE] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1732),
    [anon_sym_L_DQUOTE] = ACTIONS(1732),
    [anon_sym_u_DQUOTE] = ACTIONS(1732),
    [anon_sym_U_DQUOTE] = ACTIONS(1732),
    [anon_sym_u8_DQUOTE] = ACTIONS(1732),
    [anon_sym_DQUOTE] = ACTIONS(1732),
    [sym_true] = ACTIONS(1730),
    [sym_false] = ACTIONS(1730),
    [anon_sym_NULL] = ACTIONS(1730),
    [anon_sym_nullptr] = ACTIONS(1730),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(395)] = {
    [sym_identifier] = ACTIONS(1734),
    [anon_sym_COMMA] = ACTIONS(1736),
    [anon_sym_RPAREN] = ACTIONS(1736),
    [anon_sym_LPAREN2] = ACTIONS(1736),
    [anon_sym_BANG] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_DASH] = ACTIONS(1734),
    [anon_sym_PLUS] = ACTIONS(1734),
    [anon_sym_STAR] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym___extension__] = ACTIONS(1734),
    [anon_sym_extern] = ACTIONS(1734),
    [anon_sym___attribute__] = ACTIONS(1734),
    [anon_sym___attribute] = ACTIONS(1734),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1736),
    [anon_sym___declspec] = ACTIONS(1734),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_signed] = ACTIONS(1734),
    [anon_sym_unsigned] = ACTIONS(1734),
    [anon_sym_long] = ACTIONS(1734),
    [anon_sym_short] = ACTIONS(1734),
    [anon_sym_LBRACK] = ACTIONS(1734),
    [anon_sym_static] = ACTIONS(1734),
    [anon_sym_EQ] = ACTIONS(1736),
    [anon_sym_auto] = ACTIONS(1734),
    [anon_sym_register] = ACTIONS(1734),
    [anon_sym_inline] = ACTIONS(1734),
    [anon_sym___inline] = ACTIONS(1734),
    [anon_sym___inline__] = ACTIONS(1734),
    [anon_sym___forceinline] = ACTIONS(1734),
    [anon_sym_thread_local] = ACTIONS(1734),
    [anon_sym___thread] = ACTIONS(1734),
    [anon_sym_const] = ACTIONS(1734),
    [anon_sym_constexpr] = ACTIONS(1734),
    [anon_sym_volatile] = ACTIONS(1734),
    [anon_sym_restrict] = ACTIONS(1734),
    [anon_sym___restrict__] = ACTIONS(1734),
    [anon_sym__Atomic] = ACTIONS(1734),
    [anon_sym__Noreturn] = ACTIONS(1734),
    [anon_sym_noreturn] = ACTIONS(1734),
    [anon_sym__Nonnull] = ACTIONS(1734),
    [anon_sym_alignas] = ACTIONS(1734),
    [anon_sym__Alignas] = ACTIONS(1734),
    [sym_primitive_type] = ACTIONS(1734),
    [anon_sym_enum] = ACTIONS(1734),
    [anon_sym_COLON] = ACTIONS(1736),
    [anon_sym_struct] = ACTIONS(1734),
    [anon_sym_union] = ACTIONS(1734),
    [anon_sym_if] = ACTIONS(1734),
    [anon_sym_switch] = ACTIONS(1734),
    [anon_sym_case] = ACTIONS(1734),
    [anon_sym_default] = ACTIONS(1734),
    [anon_sym_while] = ACTIONS(1734),
    [anon_sym_do] = ACTIONS(1734),
    [anon_sym_for] = ACTIONS(1734),
    [anon_sym_return] = ACTIONS(1734),
    [anon_sym_break] = ACTIONS(1734),
    [anon_sym_continue] = ACTIONS(1734),
    [anon_sym_goto] = ACTIONS(1734),
    [anon_sym___try] = ACTIONS(1734),
    [anon_sym___leave] = ACTIONS(1734),
    [anon_sym_DASH_DASH] = ACTIONS(1736),
    [anon_sym_PLUS_PLUS] = ACTIONS(1736),
    [anon_sym_sizeof] = ACTIONS(1734),
    [anon_sym___alignof__] = ACTIONS(1734),
    [anon_sym___alignof] = ACTIONS(1734),
    [anon_sym__alignof] = ACTIONS(1734),
    [anon_sym_alignof] = ACTIONS(1734),
    [anon_sym__Alignof] = ACTIONS(1734),
    [anon_sym_offsetof] = ACTIONS(1734),
    [anon_sym__Generic] = ACTIONS(1734),
    [anon_sym_asm] = ACTIONS(1734),
    [anon_sym___asm__] = ACTIONS(1734),
    [anon_sym___asm] = ACTIONS(1734),
    [sym_number_literal] = ACTIONS(1736),
    [anon_sym_L_SQUOTE] = ACTIONS(1736),
    [anon_sym_u_SQUOTE] = ACTIONS(1736),
    [anon_sym_U_SQUOTE] = ACTIONS(1736),
    [anon_sym_u8_SQUOTE] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1736),
    [anon_sym_L_DQUOTE] = ACTIONS(1736),
    [anon_sym_u_DQUOTE] = ACTIONS(1736),
    [anon_sym_U_DQUOTE] = ACTIONS(1736),
    [anon_sym_u8_DQUOTE] = ACTIONS(1736),
    [anon_sym_DQUOTE] = ACTIONS(1736),
    [sym_true] = ACTIONS(1734),
    [sym_false] = ACTIONS(1734),
    [anon_sym_NULL] = ACTIONS(1734),
    [anon_sym_nullptr] = ACTIONS(1734),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(396)] = {
    [sym_expression] = STATE(699),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_COMMA] = ACTIONS(1380),
    [anon_sym_RPAREN] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(23),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1380),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym___attribute__] = ACTIONS(1386),
    [anon_sym___attribute] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_COLON] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(397)] = {
    [sym_expression] = STATE(699),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(835),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(835),
    [sym_call_expression] = STATE(835),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(835),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(835),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(713),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1386),
    [anon_sym_COMMA] = ACTIONS(1380),
    [aux_sym_preproc_if_token2] = ACTIONS(1380),
    [aux_sym_preproc_else_token1] = ACTIONS(1380),
    [aux_sym_preproc_elif_token1] = ACTIONS(1386),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1380),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1742),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1380),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym___extension__] = ACTIONS(1744),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(1746),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(398)] = {
    [sym_else_clause] = STATE(250),
    [sym_identifier] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1130),
    [anon_sym_AMP] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [anon_sym___extension__] = ACTIONS(1128),
    [anon_sym_typedef] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1128),
    [anon_sym___attribute__] = ACTIONS(1128),
    [anon_sym___attribute] = ACTIONS(1128),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1130),
    [anon_sym___declspec] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_signed] = ACTIONS(1128),
    [anon_sym_unsigned] = ACTIONS(1128),
    [anon_sym_long] = ACTIONS(1128),
    [anon_sym_short] = ACTIONS(1128),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_auto] = ACTIONS(1128),
    [anon_sym_register] = ACTIONS(1128),
    [anon_sym_inline] = ACTIONS(1128),
    [anon_sym___inline] = ACTIONS(1128),
    [anon_sym___inline__] = ACTIONS(1128),
    [anon_sym___forceinline] = ACTIONS(1128),
    [anon_sym_thread_local] = ACTIONS(1128),
    [anon_sym___thread] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_constexpr] = ACTIONS(1128),
    [anon_sym_volatile] = ACTIONS(1128),
    [anon_sym_restrict] = ACTIONS(1128),
    [anon_sym___restrict__] = ACTIONS(1128),
    [anon_sym__Atomic] = ACTIONS(1128),
    [anon_sym__Noreturn] = ACTIONS(1128),
    [anon_sym_noreturn] = ACTIONS(1128),
    [anon_sym__Nonnull] = ACTIONS(1128),
    [anon_sym_alignas] = ACTIONS(1128),
    [anon_sym__Alignas] = ACTIONS(1128),
    [sym_primitive_type] = ACTIONS(1128),
    [anon_sym_enum] = ACTIONS(1128),
    [anon_sym_struct] = ACTIONS(1128),
    [anon_sym_union] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1748),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_goto] = ACTIONS(1128),
    [anon_sym___try] = ACTIONS(1128),
    [anon_sym___leave] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [anon_sym_sizeof] = ACTIONS(1128),
    [anon_sym___alignof__] = ACTIONS(1128),
    [anon_sym___alignof] = ACTIONS(1128),
    [anon_sym__alignof] = ACTIONS(1128),
    [anon_sym_alignof] = ACTIONS(1128),
    [anon_sym__Alignof] = ACTIONS(1128),
    [anon_sym_offsetof] = ACTIONS(1128),
    [anon_sym__Generic] = ACTIONS(1128),
    [anon_sym_asm] = ACTIONS(1128),
    [anon_sym___asm__] = ACTIONS(1128),
    [anon_sym___asm] = ACTIONS(1128),
    [sym_number_literal] = ACTIONS(1130),
    [anon_sym_L_SQUOTE] = ACTIONS(1130),
    [anon_sym_u_SQUOTE] = ACTIONS(1130),
    [anon_sym_U_SQUOTE] = ACTIONS(1130),
    [anon_sym_u8_SQUOTE] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_L_DQUOTE] = ACTIONS(1130),
    [anon_sym_u_DQUOTE] = ACTIONS(1130),
    [anon_sym_U_DQUOTE] = ACTIONS(1130),
    [anon_sym_u8_DQUOTE] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [anon_sym_NULL] = ACTIONS(1128),
    [anon_sym_nullptr] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(399)] = {
    [sym_identifier] = ACTIONS(1750),
    [anon_sym_LPAREN2] = ACTIONS(1753),
    [anon_sym_BANG] = ACTIONS(1753),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_DASH] = ACTIONS(1755),
    [anon_sym_PLUS] = ACTIONS(1755),
    [anon_sym_STAR] = ACTIONS(1753),
    [anon_sym_AMP] = ACTIONS(1753),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym___extension__] = ACTIONS(1750),
    [anon_sym_extern] = ACTIONS(1757),
    [anon_sym___attribute__] = ACTIONS(1757),
    [anon_sym___attribute] = ACTIONS(1757),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1759),
    [anon_sym___declspec] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_signed] = ACTIONS(1757),
    [anon_sym_unsigned] = ACTIONS(1757),
    [anon_sym_long] = ACTIONS(1757),
    [anon_sym_short] = ACTIONS(1757),
    [anon_sym_static] = ACTIONS(1757),
    [anon_sym_auto] = ACTIONS(1757),
    [anon_sym_register] = ACTIONS(1757),
    [anon_sym_inline] = ACTIONS(1757),
    [anon_sym___inline] = ACTIONS(1757),
    [anon_sym___inline__] = ACTIONS(1757),
    [anon_sym___forceinline] = ACTIONS(1757),
    [anon_sym_thread_local] = ACTIONS(1757),
    [anon_sym___thread] = ACTIONS(1757),
    [anon_sym_const] = ACTIONS(1757),
    [anon_sym_constexpr] = ACTIONS(1757),
    [anon_sym_volatile] = ACTIONS(1757),
    [anon_sym_restrict] = ACTIONS(1757),
    [anon_sym___restrict__] = ACTIONS(1757),
    [anon_sym__Atomic] = ACTIONS(1757),
    [anon_sym__Noreturn] = ACTIONS(1757),
    [anon_sym_noreturn] = ACTIONS(1757),
    [anon_sym__Nonnull] = ACTIONS(1757),
    [anon_sym_alignas] = ACTIONS(1757),
    [anon_sym__Alignas] = ACTIONS(1757),
    [sym_primitive_type] = ACTIONS(1757),
    [anon_sym_enum] = ACTIONS(1757),
    [anon_sym_struct] = ACTIONS(1757),
    [anon_sym_union] = ACTIONS(1757),
    [anon_sym_if] = ACTIONS(1755),
    [anon_sym_switch] = ACTIONS(1755),
    [anon_sym_case] = ACTIONS(1755),
    [anon_sym_default] = ACTIONS(1755),
    [anon_sym_while] = ACTIONS(1755),
    [anon_sym_do] = ACTIONS(1755),
    [anon_sym_for] = ACTIONS(1755),
    [anon_sym_return] = ACTIONS(1755),
    [anon_sym_break] = ACTIONS(1755),
    [anon_sym_continue] = ACTIONS(1755),
    [anon_sym_goto] = ACTIONS(1755),
    [anon_sym___try] = ACTIONS(1755),
    [anon_sym___leave] = ACTIONS(1755),
    [anon_sym_DASH_DASH] = ACTIONS(1753),
    [anon_sym_PLUS_PLUS] = ACTIONS(1753),
    [anon_sym_sizeof] = ACTIONS(1755),
    [anon_sym___alignof__] = ACTIONS(1755),
    [anon_sym___alignof] = ACTIONS(1755),
    [anon_sym__alignof] = ACTIONS(1755),
    [anon_sym_alignof] = ACTIONS(1755),
    [anon_sym__Alignof] = ACTIONS(1755),
    [anon_sym_offsetof] = ACTIONS(1755),
    [anon_sym__Generic] = ACTIONS(1755),
    [anon_sym_asm] = ACTIONS(1755),
    [anon_sym___asm__] = ACTIONS(1755),
    [anon_sym___asm] = ACTIONS(1755),
    [sym_number_literal] = ACTIONS(1753),
    [anon_sym_L_SQUOTE] = ACTIONS(1753),
    [anon_sym_u_SQUOTE] = ACTIONS(1753),
    [anon_sym_U_SQUOTE] = ACTIONS(1753),
    [anon_sym_u8_SQUOTE] = ACTIONS(1753),
    [anon_sym_SQUOTE] = ACTIONS(1753),
    [anon_sym_L_DQUOTE] = ACTIONS(1753),
    [anon_sym_u_DQUOTE] = ACTIONS(1753),
    [anon_sym_U_DQUOTE] = ACTIONS(1753),
    [anon_sym_u8_DQUOTE] = ACTIONS(1753),
    [anon_sym_DQUOTE] = ACTIONS(1753),
    [sym_true] = ACTIONS(1755),
    [sym_false] = ACTIONS(1755),
    [anon_sym_NULL] = ACTIONS(1755),
    [anon_sym_nullptr] = ACTIONS(1755),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(400)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1775),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1784),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(401)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1788),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(402)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1790),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(403)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1792),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(404)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1794),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(405)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1775),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1792),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(406)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1775),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1794),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(407)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1784),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(408)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_COLON] = ACTIONS(1790),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(409)] = {
    [sym_expression] = STATE(889),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_initializer_list] = STATE(678),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1801),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1386),
    [anon_sym_PLUS] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_SLASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1380),
    [anon_sym_PIPE_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1386),
    [anon_sym_CARET] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1386),
    [anon_sym_GT_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1386),
    [anon_sym_LT_LT] = ACTIONS(1380),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym___extension__] = ACTIONS(1805),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_RBRACK] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1380),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1380),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(410)] = {
    [sym_string_literal] = STATE(621),
    [aux_sym_sized_type_specifier_repeat1] = STATE(762),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_COMMA] = ACTIONS(1764),
    [anon_sym_LPAREN2] = ACTIONS(1766),
    [anon_sym_DASH] = ACTIONS(1770),
    [anon_sym_PLUS] = ACTIONS(1770),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_SLASH] = ACTIONS(1770),
    [anon_sym_PERCENT] = ACTIONS(1770),
    [anon_sym_PIPE_PIPE] = ACTIONS(1764),
    [anon_sym_AMP_AMP] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1770),
    [anon_sym_CARET] = ACTIONS(1770),
    [anon_sym_AMP] = ACTIONS(1770),
    [anon_sym_EQ_EQ] = ACTIONS(1764),
    [anon_sym_BANG_EQ] = ACTIONS(1764),
    [anon_sym_GT] = ACTIONS(1770),
    [anon_sym_GT_EQ] = ACTIONS(1764),
    [anon_sym_LT_EQ] = ACTIONS(1764),
    [anon_sym_LT] = ACTIONS(1770),
    [anon_sym_LT_LT] = ACTIONS(1770),
    [anon_sym_GT_GT] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym___extension__] = ACTIONS(1762),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym___attribute__] = ACTIONS(1762),
    [anon_sym___attribute] = ACTIONS(1762),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1778),
    [anon_sym___declspec] = ACTIONS(1762),
    [anon_sym___based] = ACTIONS(1762),
    [anon_sym___cdecl] = ACTIONS(1762),
    [anon_sym___clrcall] = ACTIONS(1762),
    [anon_sym___stdcall] = ACTIONS(1762),
    [anon_sym___fastcall] = ACTIONS(1762),
    [anon_sym___thiscall] = ACTIONS(1762),
    [anon_sym___vectorcall] = ACTIONS(1762),
    [anon_sym_signed] = ACTIONS(1780),
    [anon_sym_unsigned] = ACTIONS(1780),
    [anon_sym_long] = ACTIONS(1780),
    [anon_sym_short] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_EQ] = ACTIONS(1782),
    [anon_sym_auto] = ACTIONS(1762),
    [anon_sym_register] = ACTIONS(1762),
    [anon_sym_inline] = ACTIONS(1762),
    [anon_sym___inline] = ACTIONS(1762),
    [anon_sym___inline__] = ACTIONS(1762),
    [anon_sym___forceinline] = ACTIONS(1762),
    [anon_sym_thread_local] = ACTIONS(1762),
    [anon_sym___thread] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_constexpr] = ACTIONS(1762),
    [anon_sym_volatile] = ACTIONS(1762),
    [anon_sym_restrict] = ACTIONS(1762),
    [anon_sym___restrict__] = ACTIONS(1762),
    [anon_sym__Atomic] = ACTIONS(1762),
    [anon_sym__Noreturn] = ACTIONS(1762),
    [anon_sym_noreturn] = ACTIONS(1762),
    [anon_sym__Nonnull] = ACTIONS(1762),
    [anon_sym_alignas] = ACTIONS(1762),
    [anon_sym__Alignas] = ACTIONS(1762),
    [anon_sym_QMARK] = ACTIONS(1764),
    [anon_sym_STAR_EQ] = ACTIONS(1786),
    [anon_sym_SLASH_EQ] = ACTIONS(1786),
    [anon_sym_PERCENT_EQ] = ACTIONS(1786),
    [anon_sym_PLUS_EQ] = ACTIONS(1786),
    [anon_sym_DASH_EQ] = ACTIONS(1786),
    [anon_sym_LT_LT_EQ] = ACTIONS(1786),
    [anon_sym_GT_GT_EQ] = ACTIONS(1786),
    [anon_sym_AMP_EQ] = ACTIONS(1786),
    [anon_sym_CARET_EQ] = ACTIONS(1786),
    [anon_sym_PIPE_EQ] = ACTIONS(1786),
    [anon_sym_DASH_DASH] = ACTIONS(1764),
    [anon_sym_PLUS_PLUS] = ACTIONS(1764),
    [anon_sym_DOT] = ACTIONS(1764),
    [anon_sym_DASH_GT] = ACTIONS(1764),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(411)] = {
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1118),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_based_modifier] = STATE(1900),
    [sym_ms_call_modifier] = STATE(1257),
    [sym__declarator] = STATE(1442),
    [sym__abstract_declarator] = STATE(1527),
    [sym_parenthesized_declarator] = STATE(1300),
    [sym_abstract_parenthesized_declarator] = STATE(1443),
    [sym_attributed_declarator] = STATE(1300),
    [sym_pointer_declarator] = STATE(1300),
    [sym_abstract_pointer_declarator] = STATE(1443),
    [sym_function_declarator] = STATE(1300),
    [sym_abstract_function_declarator] = STATE(1443),
    [sym_array_declarator] = STATE(1300),
    [sym_abstract_array_declarator] = STATE(1443),
    [sym_compound_statement] = STATE(1943),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_variadic_parameter] = STATE(1598),
    [sym_parameter_list] = STATE(1452),
    [sym_parameter_declaration] = STATE(1598),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1809),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1811),
    [anon_sym_RPAREN] = ACTIONS(1813),
    [anon_sym_LPAREN2] = ACTIONS(1815),
    [anon_sym_STAR] = ACTIONS(1817),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___based] = ACTIONS(1819),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(412)] = {
    [sym_type_qualifier] = STATE(662),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1097),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(662),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1825),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1831),
    [anon_sym_RBRACK] = ACTIONS(1833),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(413)] = {
    [sym_type_qualifier] = STATE(414),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1083),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(414),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1841),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1843),
    [anon_sym_RBRACK] = ACTIONS(1845),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(414)] = {
    [sym_type_qualifier] = STATE(662),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1099),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(662),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1847),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1831),
    [anon_sym_RBRACK] = ACTIONS(1849),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(415)] = {
    [sym_type_qualifier] = STATE(662),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1075),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(662),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1851),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1831),
    [anon_sym_RBRACK] = ACTIONS(1853),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(416)] = {
    [sym_type_qualifier] = STATE(412),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1089),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(412),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1855),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1857),
    [anon_sym_RBRACK] = ACTIONS(1859),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(417)] = {
    [sym_type_qualifier] = STATE(662),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1081),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(662),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1861),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1831),
    [anon_sym_RBRACK] = ACTIONS(1863),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(418)] = {
    [sym_type_qualifier] = STATE(662),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1087),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(662),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1865),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1831),
    [anon_sym_RBRACK] = ACTIONS(1867),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(419)] = {
    [sym_type_qualifier] = STATE(418),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1100),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(418),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1869),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1871),
    [anon_sym_RBRACK] = ACTIONS(1873),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(420)] = {
    [sym_type_qualifier] = STATE(417),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1071),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(417),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1875),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1877),
    [anon_sym_RBRACK] = ACTIONS(1879),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(421)] = {
    [sym_type_qualifier] = STATE(415),
    [sym_alignas_qualifier] = STATE(700),
    [sym_expression] = STATE(1096),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_array_declarator_repeat1] = STATE(415),
    [sym_identifier] = ACTIONS(1799),
    [anon_sym_LPAREN2] = ACTIONS(1823),
    [anon_sym_BANG] = ACTIONS(1803),
    [anon_sym_TILDE] = ACTIONS(1803),
    [anon_sym_DASH] = ACTIONS(1801),
    [anon_sym_PLUS] = ACTIONS(1801),
    [anon_sym_STAR] = ACTIONS(1881),
    [anon_sym_AMP] = ACTIONS(1827),
    [anon_sym___extension__] = ACTIONS(1829),
    [anon_sym_static] = ACTIONS(1883),
    [anon_sym_RBRACK] = ACTIONS(1885),
    [anon_sym_const] = ACTIONS(1835),
    [anon_sym_constexpr] = ACTIONS(1835),
    [anon_sym_volatile] = ACTIONS(1835),
    [anon_sym_restrict] = ACTIONS(1835),
    [anon_sym___restrict__] = ACTIONS(1835),
    [anon_sym__Atomic] = ACTIONS(1835),
    [anon_sym__Noreturn] = ACTIONS(1835),
    [anon_sym_noreturn] = ACTIONS(1835),
    [anon_sym__Nonnull] = ACTIONS(1835),
    [anon_sym_alignas] = ACTIONS(1837),
    [anon_sym__Alignas] = ACTIONS(1837),
    [anon_sym_DASH_DASH] = ACTIONS(1839),
    [anon_sym_PLUS_PLUS] = ACTIONS(1839),
    [anon_sym_sizeof] = ACTIONS(1807),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(422)] = {
    [sym_expression] = STATE(1036),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(911),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(911),
    [sym_call_expression] = STATE(911),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(911),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(911),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1887),
    [anon_sym_LPAREN2] = ACTIONS(1890),
    [anon_sym_BANG] = ACTIONS(1893),
    [anon_sym_TILDE] = ACTIONS(1893),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_PLUS] = ACTIONS(1896),
    [anon_sym_STAR] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym___extension__] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_RBRACK] = ACTIONS(1708),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_constexpr] = ACTIONS(1706),
    [anon_sym_volatile] = ACTIONS(1706),
    [anon_sym_restrict] = ACTIONS(1706),
    [anon_sym___restrict__] = ACTIONS(1706),
    [anon_sym__Atomic] = ACTIONS(1706),
    [anon_sym__Noreturn] = ACTIONS(1706),
    [anon_sym_noreturn] = ACTIONS(1706),
    [anon_sym__Nonnull] = ACTIONS(1706),
    [anon_sym_alignas] = ACTIONS(1706),
    [anon_sym__Alignas] = ACTIONS(1706),
    [anon_sym_DASH_DASH] = ACTIONS(1905),
    [anon_sym_PLUS_PLUS] = ACTIONS(1905),
    [anon_sym_sizeof] = ACTIONS(1908),
    [anon_sym___alignof__] = ACTIONS(1911),
    [anon_sym___alignof] = ACTIONS(1911),
    [anon_sym__alignof] = ACTIONS(1911),
    [anon_sym_alignof] = ACTIONS(1911),
    [anon_sym__Alignof] = ACTIONS(1911),
    [anon_sym_offsetof] = ACTIONS(1914),
    [anon_sym__Generic] = ACTIONS(1917),
    [anon_sym_asm] = ACTIONS(1920),
    [anon_sym___asm__] = ACTIONS(1920),
    [anon_sym___asm] = ACTIONS(1920),
    [sym_number_literal] = ACTIONS(1923),
    [anon_sym_L_SQUOTE] = ACTIONS(1926),
    [anon_sym_u_SQUOTE] = ACTIONS(1926),
    [anon_sym_U_SQUOTE] = ACTIONS(1926),
    [anon_sym_u8_SQUOTE] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_L_DQUOTE] = ACTIONS(1929),
    [anon_sym_u_DQUOTE] = ACTIONS(1929),
    [anon_sym_U_DQUOTE] = ACTIONS(1929),
    [anon_sym_u8_DQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_true] = ACTIONS(1932),
    [sym_false] = ACTIONS(1932),
    [anon_sym_NULL] = ACTIONS(1935),
    [anon_sym_nullptr] = ACTIONS(1935),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(423)] = {
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1118),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_ms_call_modifier] = STATE(1375),
    [sym__abstract_declarator] = STATE(1527),
    [sym_abstract_parenthesized_declarator] = STATE(1443),
    [sym_abstract_pointer_declarator] = STATE(1443),
    [sym_abstract_function_declarator] = STATE(1443),
    [sym_abstract_array_declarator] = STATE(1443),
    [sym_compound_statement] = STATE(1943),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym_variadic_parameter] = STATE(1598),
    [sym_parameter_list] = STATE(1452),
    [sym_parameter_declaration] = STATE(1598),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1811),
    [anon_sym_RPAREN] = ACTIONS(1813),
    [anon_sym_LPAREN2] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1942),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym___cdecl] = ACTIONS(41),
    [anon_sym___clrcall] = ACTIONS(41),
    [anon_sym___stdcall] = ACTIONS(41),
    [anon_sym___fastcall] = ACTIONS(41),
    [anon_sym___thiscall] = ACTIONS(41),
    [anon_sym___vectorcall] = ACTIONS(41),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(424)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1916),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1916),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1916),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1948),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(425)] = {
    [sym_preproc_def] = STATE(435),
    [sym_preproc_function_def] = STATE(435),
    [sym_preproc_call] = STATE(435),
    [sym_preproc_if_in_field_declaration_list] = STATE(435),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(435),
    [sym_preproc_else_in_field_declaration_list] = STATE(1913),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1913),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1913),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(435),
    [sym_field_declaration] = STATE(435),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(435),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1960),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(426)] = {
    [sym_preproc_def] = STATE(428),
    [sym_preproc_function_def] = STATE(428),
    [sym_preproc_call] = STATE(428),
    [sym_preproc_if_in_field_declaration_list] = STATE(428),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(428),
    [sym_preproc_else_in_field_declaration_list] = STATE(1891),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1891),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1891),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(428),
    [sym_field_declaration] = STATE(428),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(428),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1962),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(427)] = {
    [sym_preproc_def] = STATE(429),
    [sym_preproc_function_def] = STATE(429),
    [sym_preproc_call] = STATE(429),
    [sym_preproc_if_in_field_declaration_list] = STATE(429),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(429),
    [sym_preproc_else_in_field_declaration_list] = STATE(1915),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1915),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1915),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(429),
    [sym_field_declaration] = STATE(429),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(429),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1964),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(428)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1918),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1918),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1918),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1966),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(429)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1921),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1921),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1921),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1968),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(430)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1892),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1892),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1892),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1970),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(431)] = {
    [sym_preproc_def] = STATE(434),
    [sym_preproc_function_def] = STATE(434),
    [sym_preproc_call] = STATE(434),
    [sym_preproc_if_in_field_declaration_list] = STATE(434),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(434),
    [sym_preproc_else_in_field_declaration_list] = STATE(1966),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1966),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1966),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(434),
    [sym_field_declaration] = STATE(434),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(434),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1972),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(432)] = {
    [sym_expression] = STATE(1015),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_initializer_list] = STATE(1643),
    [sym_initializer_pair] = STATE(1643),
    [sym_subscript_designator] = STATE(1445),
    [sym_subscript_range_designator] = STATE(1445),
    [sym_field_designator] = STATE(1445),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_initializer_pair_repeat1] = STATE(1445),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_COMMA] = ACTIONS(1976),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1978),
    [anon_sym_LBRACK] = ACTIONS(1980),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1982),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(433)] = {
    [sym_preproc_def] = STATE(430),
    [sym_preproc_function_def] = STATE(430),
    [sym_preproc_call] = STATE(430),
    [sym_preproc_if_in_field_declaration_list] = STATE(430),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(430),
    [sym_preproc_else_in_field_declaration_list] = STATE(1822),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1822),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1822),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(430),
    [sym_field_declaration] = STATE(430),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(430),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1984),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(434)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1788),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1788),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1788),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1986),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(435)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1927),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1927),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1927),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1988),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(436)] = {
    [sym_preproc_def] = STATE(438),
    [sym_preproc_function_def] = STATE(438),
    [sym_preproc_call] = STATE(438),
    [sym_preproc_if_in_field_declaration_list] = STATE(438),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(438),
    [sym_preproc_else_in_field_declaration_list] = STATE(1982),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1982),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1982),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(438),
    [sym_field_declaration] = STATE(438),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(438),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1990),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(437)] = {
    [sym_preproc_def] = STATE(439),
    [sym_preproc_function_def] = STATE(439),
    [sym_preproc_call] = STATE(439),
    [sym_preproc_if_in_field_declaration_list] = STATE(439),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(439),
    [sym_preproc_else_in_field_declaration_list] = STATE(2007),
    [sym_preproc_elif_in_field_declaration_list] = STATE(2007),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(2007),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(439),
    [sym_field_declaration] = STATE(439),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(439),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1992),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(438)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(2010),
    [sym_preproc_elif_in_field_declaration_list] = STATE(2010),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(2010),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1994),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(439)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym_preproc_else_in_field_declaration_list] = STATE(1863),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1863),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1863),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1996),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(440)] = {
    [sym_preproc_def] = STATE(424),
    [sym_preproc_function_def] = STATE(424),
    [sym_preproc_call] = STATE(424),
    [sym_preproc_if_in_field_declaration_list] = STATE(424),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(424),
    [sym_preproc_else_in_field_declaration_list] = STATE(1872),
    [sym_preproc_elif_in_field_declaration_list] = STATE(1872),
    [sym_preproc_elifdef_in_field_declaration_list] = STATE(1872),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(424),
    [sym_field_declaration] = STATE(424),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(424),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(1944),
    [aux_sym_preproc_if_token1] = ACTIONS(1946),
    [aux_sym_preproc_if_token2] = ACTIONS(1998),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(1950),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(1950),
    [aux_sym_preproc_else_token1] = ACTIONS(1952),
    [aux_sym_preproc_elif_token1] = ACTIONS(1954),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(1956),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(1956),
    [sym_preproc_directive] = ACTIONS(1958),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(441)] = {
    [sym_expression] = STATE(1030),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_initializer_list] = STATE(1683),
    [sym_initializer_pair] = STATE(1683),
    [sym_subscript_designator] = STATE(1445),
    [sym_subscript_range_designator] = STATE(1445),
    [sym_field_designator] = STATE(1445),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_initializer_pair_repeat1] = STATE(1445),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(2000),
    [anon_sym_LBRACK] = ACTIONS(1980),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1982),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(442)] = {
    [sym_expression] = STATE(1030),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_initializer_list] = STATE(1683),
    [sym_initializer_pair] = STATE(1683),
    [sym_subscript_designator] = STATE(1445),
    [sym_subscript_range_designator] = STATE(1445),
    [sym_field_designator] = STATE(1445),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_initializer_pair_repeat1] = STATE(1445),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(2002),
    [anon_sym_LBRACK] = ACTIONS(1980),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1982),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(443)] = {
    [sym_expression] = STATE(1030),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_initializer_list] = STATE(1683),
    [sym_initializer_pair] = STATE(1683),
    [sym_subscript_designator] = STATE(1445),
    [sym_subscript_range_designator] = STATE(1445),
    [sym_field_designator] = STATE(1445),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [aux_sym_initializer_pair_repeat1] = STATE(1445),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_LBRACK] = ACTIONS(1980),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [anon_sym_DOT] = ACTIONS(1982),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(444)] = {
    [sym_preproc_def] = STATE(444),
    [sym_preproc_function_def] = STATE(444),
    [sym_preproc_call] = STATE(444),
    [sym_preproc_if_in_field_declaration_list] = STATE(444),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(444),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1279),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(444),
    [sym_field_declaration] = STATE(444),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(444),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(2004),
    [aux_sym_preproc_def_token1] = ACTIONS(2007),
    [aux_sym_preproc_if_token1] = ACTIONS(2010),
    [aux_sym_preproc_if_token2] = ACTIONS(2013),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2015),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2015),
    [aux_sym_preproc_else_token1] = ACTIONS(2013),
    [aux_sym_preproc_elif_token1] = ACTIONS(2013),
    [aux_sym_preproc_elifdef_token1] = ACTIONS(2013),
    [aux_sym_preproc_elifdef_token2] = ACTIONS(2013),
    [sym_preproc_directive] = ACTIONS(2018),
    [anon_sym___extension__] = ACTIONS(2021),
    [anon_sym_extern] = ACTIONS(2024),
    [anon_sym___attribute__] = ACTIONS(2027),
    [anon_sym___attribute] = ACTIONS(2027),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(2030),
    [anon_sym___declspec] = ACTIONS(2033),
    [anon_sym_signed] = ACTIONS(2036),
    [anon_sym_unsigned] = ACTIONS(2036),
    [anon_sym_long] = ACTIONS(2036),
    [anon_sym_short] = ACTIONS(2036),
    [anon_sym_static] = ACTIONS(2024),
    [anon_sym_auto] = ACTIONS(2024),
    [anon_sym_register] = ACTIONS(2024),
    [anon_sym_inline] = ACTIONS(2024),
    [anon_sym___inline] = ACTIONS(2024),
    [anon_sym___inline__] = ACTIONS(2024),
    [anon_sym___forceinline] = ACTIONS(2024),
    [anon_sym_thread_local] = ACTIONS(2024),
    [anon_sym___thread] = ACTIONS(2024),
    [anon_sym_const] = ACTIONS(2021),
    [anon_sym_constexpr] = ACTIONS(2021),
    [anon_sym_volatile] = ACTIONS(2021),
    [anon_sym_restrict] = ACTIONS(2021),
    [anon_sym___restrict__] = ACTIONS(2021),
    [anon_sym__Atomic] = ACTIONS(2021),
    [anon_sym__Noreturn] = ACTIONS(2021),
    [anon_sym_noreturn] = ACTIONS(2021),
    [anon_sym__Nonnull] = ACTIONS(2021),
    [anon_sym_alignas] = ACTIONS(2039),
    [anon_sym__Alignas] = ACTIONS(2039),
    [sym_primitive_type] = ACTIONS(2042),
    [anon_sym_enum] = ACTIONS(2045),
    [anon_sym_struct] = ACTIONS(2048),
    [anon_sym_union] = ACTIONS(2051),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(445)] = {
    [sym_preproc_def] = STATE(445),
    [sym_preproc_function_def] = STATE(445),
    [sym_preproc_call] = STATE(445),
    [sym_preproc_if_in_field_declaration_list] = STATE(445),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(445),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1275),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(445),
    [sym_field_declaration] = STATE(445),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(445),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(2004),
    [aux_sym_preproc_def_token1] = ACTIONS(2054),
    [aux_sym_preproc_if_token1] = ACTIONS(2057),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2060),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2060),
    [sym_preproc_directive] = ACTIONS(2063),
    [anon_sym___extension__] = ACTIONS(2021),
    [anon_sym_extern] = ACTIONS(2024),
    [anon_sym___attribute__] = ACTIONS(2027),
    [anon_sym___attribute] = ACTIONS(2027),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(2030),
    [anon_sym___declspec] = ACTIONS(2033),
    [anon_sym_RBRACE] = ACTIONS(2066),
    [anon_sym_signed] = ACTIONS(2036),
    [anon_sym_unsigned] = ACTIONS(2036),
    [anon_sym_long] = ACTIONS(2036),
    [anon_sym_short] = ACTIONS(2036),
    [anon_sym_static] = ACTIONS(2024),
    [anon_sym_auto] = ACTIONS(2024),
    [anon_sym_register] = ACTIONS(2024),
    [anon_sym_inline] = ACTIONS(2024),
    [anon_sym___inline] = ACTIONS(2024),
    [anon_sym___inline__] = ACTIONS(2024),
    [anon_sym___forceinline] = ACTIONS(2024),
    [anon_sym_thread_local] = ACTIONS(2024),
    [anon_sym___thread] = ACTIONS(2024),
    [anon_sym_const] = ACTIONS(2021),
    [anon_sym_constexpr] = ACTIONS(2021),
    [anon_sym_volatile] = ACTIONS(2021),
    [anon_sym_restrict] = ACTIONS(2021),
    [anon_sym___restrict__] = ACTIONS(2021),
    [anon_sym__Atomic] = ACTIONS(2021),
    [anon_sym__Noreturn] = ACTIONS(2021),
    [anon_sym_noreturn] = ACTIONS(2021),
    [anon_sym__Nonnull] = ACTIONS(2021),
    [anon_sym_alignas] = ACTIONS(2039),
    [anon_sym__Alignas] = ACTIONS(2039),
    [sym_primitive_type] = ACTIONS(2042),
    [anon_sym_enum] = ACTIONS(2045),
    [anon_sym_struct] = ACTIONS(2048),
    [anon_sym_union] = ACTIONS(2051),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(446)] = {
    [sym_preproc_def] = STATE(449),
    [sym_preproc_function_def] = STATE(449),
    [sym_preproc_call] = STATE(449),
    [sym_preproc_if_in_field_declaration_list] = STATE(449),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(449),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1278),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(449),
    [sym_field_declaration] = STATE(449),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(449),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(2068),
    [aux_sym_preproc_if_token1] = ACTIONS(2070),
    [aux_sym_preproc_if_token2] = ACTIONS(2072),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2074),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2074),
    [sym_preproc_directive] = ACTIONS(2076),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(447)] = {
    [sym_preproc_def] = STATE(450),
    [sym_preproc_function_def] = STATE(450),
    [sym_preproc_call] = STATE(450),
    [sym_preproc_if_in_field_declaration_list] = STATE(450),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(450),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1275),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(450),
    [sym_field_declaration] = STATE(450),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(450),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(2078),
    [aux_sym_preproc_if_token1] = ACTIONS(2080),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2082),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2082),
    [sym_preproc_directive] = ACTIONS(2084),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_RBRACE] = ACTIONS(2086),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(448)] = {
    [sym_preproc_def] = STATE(446),
    [sym_preproc_function_def] = STATE(446),
    [sym_preproc_call] = STATE(446),
    [sym_preproc_if_in_field_declaration_list] = STATE(446),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(446),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1278),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(446),
    [sym_field_declaration] = STATE(446),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(446),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(2068),
    [aux_sym_preproc_if_token1] = ACTIONS(2070),
    [aux_sym_preproc_if_token2] = ACTIONS(2088),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2074),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2074),
    [sym_preproc_directive] = ACTIONS(2076),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(449)] = {
    [sym_preproc_def] = STATE(449),
    [sym_preproc_function_def] = STATE(449),
    [sym_preproc_call] = STATE(449),
    [sym_preproc_if_in_field_declaration_list] = STATE(449),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(449),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1278),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(449),
    [sym_field_declaration] = STATE(449),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(449),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(2004),
    [aux_sym_preproc_def_token1] = ACTIONS(2090),
    [aux_sym_preproc_if_token1] = ACTIONS(2093),
    [aux_sym_preproc_if_token2] = ACTIONS(2013),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2096),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2096),
    [sym_preproc_directive] = ACTIONS(2099),
    [anon_sym___extension__] = ACTIONS(2021),
    [anon_sym_extern] = ACTIONS(2024),
    [anon_sym___attribute__] = ACTIONS(2027),
    [anon_sym___attribute] = ACTIONS(2027),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(2030),
    [anon_sym___declspec] = ACTIONS(2033),
    [anon_sym_signed] = ACTIONS(2036),
    [anon_sym_unsigned] = ACTIONS(2036),
    [anon_sym_long] = ACTIONS(2036),
    [anon_sym_short] = ACTIONS(2036),
    [anon_sym_static] = ACTIONS(2024),
    [anon_sym_auto] = ACTIONS(2024),
    [anon_sym_register] = ACTIONS(2024),
    [anon_sym_inline] = ACTIONS(2024),
    [anon_sym___inline] = ACTIONS(2024),
    [anon_sym___inline__] = ACTIONS(2024),
    [anon_sym___forceinline] = ACTIONS(2024),
    [anon_sym_thread_local] = ACTIONS(2024),
    [anon_sym___thread] = ACTIONS(2024),
    [anon_sym_const] = ACTIONS(2021),
    [anon_sym_constexpr] = ACTIONS(2021),
    [anon_sym_volatile] = ACTIONS(2021),
    [anon_sym_restrict] = ACTIONS(2021),
    [anon_sym___restrict__] = ACTIONS(2021),
    [anon_sym__Atomic] = ACTIONS(2021),
    [anon_sym__Noreturn] = ACTIONS(2021),
    [anon_sym_noreturn] = ACTIONS(2021),
    [anon_sym__Nonnull] = ACTIONS(2021),
    [anon_sym_alignas] = ACTIONS(2039),
    [anon_sym__Alignas] = ACTIONS(2039),
    [sym_primitive_type] = ACTIONS(2042),
    [anon_sym_enum] = ACTIONS(2045),
    [anon_sym_struct] = ACTIONS(2048),
    [anon_sym_union] = ACTIONS(2051),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(450)] = {
    [sym_preproc_def] = STATE(445),
    [sym_preproc_function_def] = STATE(445),
    [sym_preproc_call] = STATE(445),
    [sym_preproc_if_in_field_declaration_list] = STATE(445),
    [sym_preproc_ifdef_in_field_declaration_list] = STATE(445),
    [sym__declaration_modifiers] = STATE(695),
    [sym__declaration_specifiers] = STATE(1275),
    [sym_attribute_specifier] = STATE(695),
    [sym_attribute_declaration] = STATE(695),
    [sym_ms_declspec_modifier] = STATE(695),
    [sym_storage_class_specifier] = STATE(695),
    [sym_type_qualifier] = STATE(695),
    [sym_alignas_qualifier] = STATE(711),
    [sym_type_specifier] = STATE(720),
    [sym_sized_type_specifier] = STATE(770),
    [sym_enum_specifier] = STATE(770),
    [sym_struct_specifier] = STATE(770),
    [sym_union_specifier] = STATE(770),
    [sym__field_declaration_list_item] = STATE(445),
    [sym_field_declaration] = STATE(445),
    [sym_macro_type_specifier] = STATE(770),
    [aux_sym_preproc_if_in_field_declaration_list_repeat1] = STATE(445),
    [aux_sym__declaration_specifiers_repeat1] = STATE(695),
    [aux_sym_sized_type_specifier_repeat1] = STATE(704),
    [sym_identifier] = ACTIONS(1938),
    [aux_sym_preproc_def_token1] = ACTIONS(2078),
    [aux_sym_preproc_if_token1] = ACTIONS(2080),
    [aux_sym_preproc_ifdef_token1] = ACTIONS(2082),
    [aux_sym_preproc_ifdef_token2] = ACTIONS(2082),
    [sym_preproc_directive] = ACTIONS(2084),
    [anon_sym___extension__] = ACTIONS(49),
    [anon_sym_extern] = ACTIONS(47),
    [anon_sym___attribute__] = ACTIONS(35),
    [anon_sym___attribute] = ACTIONS(35),
    [anon_sym_LBRACK_LBRACK] = ACTIONS(1126),
    [anon_sym___declspec] = ACTIONS(39),
    [anon_sym_RBRACE] = ACTIONS(2102),
    [anon_sym_signed] = ACTIONS(45),
    [anon_sym_unsigned] = ACTIONS(45),
    [anon_sym_long] = ACTIONS(45),
    [anon_sym_short] = ACTIONS(45),
    [anon_sym_static] = ACTIONS(47),
    [anon_sym_auto] = ACTIONS(47),
    [anon_sym_register] = ACTIONS(47),
    [anon_sym_inline] = ACTIONS(47),
    [anon_sym___inline] = ACTIONS(47),
    [anon_sym___inline__] = ACTIONS(47),
    [anon_sym___forceinline] = ACTIONS(47),
    [anon_sym_thread_local] = ACTIONS(47),
    [anon_sym___thread] = ACTIONS(47),
    [anon_sym_const] = ACTIONS(49),
    [anon_sym_constexpr] = ACTIONS(49),
    [anon_sym_volatile] = ACTIONS(49),
    [anon_sym_restrict] = ACTIONS(49),
    [anon_sym___restrict__] = ACTIONS(49),
    [anon_sym__Atomic] = ACTIONS(49),
    [anon_sym__Noreturn] = ACTIONS(49),
    [anon_sym_noreturn] = ACTIONS(49),
    [anon_sym__Nonnull] = ACTIONS(49),
    [anon_sym_alignas] = ACTIONS(51),
    [anon_sym__Alignas] = ACTIONS(51),
    [sym_primitive_type] = ACTIONS(53),
    [anon_sym_enum] = ACTIONS(55),
    [anon_sym_struct] = ACTIONS(57),
    [anon_sym_union] = ACTIONS(59),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(451)] = {
    [sym_compound_statement] = STATE(1667),
    [sym_expression] = STATE(1018),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_RPAREN] = ACTIONS(2104),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(452)] = {
    [sym_compound_statement] = STATE(1772),
    [sym_expression] = STATE(1033),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1772),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(453)] = {
    [sym_compound_statement] = STATE(1901),
    [sym_expression] = STATE(1031),
    [sym__string] = STATE(684),
    [sym_comma_expression] = STATE(1901),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(454)] = {
    [sym_compound_statement] = STATE(1621),
    [sym_expression] = STATE(1017),
    [sym__string] = STATE(684),
    [sym_conditional_expression] = STATE(684),
    [sym_assignment_expression] = STATE(684),
    [sym_pointer_expression] = STATE(837),
    [sym_unary_expression] = STATE(684),
    [sym_binary_expression] = STATE(684),
    [sym_update_expression] = STATE(684),
    [sym_cast_expression] = STATE(684),
    [sym_sizeof_expression] = STATE(684),
    [sym_alignof_expression] = STATE(684),
    [sym_offsetof_expression] = STATE(684),
    [sym_generic_expression] = STATE(684),
    [sym_subscript_expression] = STATE(837),
    [sym_call_expression] = STATE(837),
    [sym_gnu_asm_expression] = STATE(684),
    [sym_extension_expression] = STATE(684),
    [sym_field_expression] = STATE(837),
    [sym_compound_literal_expression] = STATE(684),
    [sym_parenthesized_expression] = STATE(837),
    [sym_char_literal] = STATE(684),
    [sym_concatenated_string] = STATE(684),
    [sym_string_literal] = STATE(663),
    [sym_null] = STATE(684),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_RPAREN] = ACTIONS(2106),
    [anon_sym_LPAREN2] = ACTIONS(19),
    [anon_sym_BANG] = ACTIONS(21),
    [anon_sym_TILDE] = ACTIONS(21),
    [anon_sym_DASH] = ACTIONS(23),
    [anon_sym_PLUS] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(25),
    [anon_sym_AMP] = ACTIONS(25),
    [anon_sym___extension__] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(43),
    [anon_sym_DASH_DASH] = ACTIONS(83),
    [anon_sym_PLUS_PLUS] = ACTIONS(83),
    [anon_sym_sizeof] = ACTIONS(85),
    [anon_sym___alignof__] = ACTIONS(87),
    [anon_sym___alignof] = ACTIONS(87),
    [anon_sym__alignof] = ACTIONS(87),
    [anon_sym_alignof] = ACTIONS(87),
    [anon_sym__Alignof] = ACTIONS(87),
    [anon_sym_offsetof] = ACTIONS(89),
    [anon_sym__Generic] = ACTIONS(91),
    [anon_sym_asm] = ACTIONS(93),
    [anon_sym___asm__] = ACTIONS(93),
    [anon_sym___asm] = ACTIONS(93),
    [sym_number_literal] = ACTIONS(161),
    [anon_sym_L_SQUOTE] = ACTIONS(97),
    [anon_sym_u_SQUOTE] = ACTIONS(97),
    [anon_sym_U_SQUOTE] = ACTIONS(97),
    [anon_sym_u8_SQUOTE] = ACTIONS(97),
    [anon_sym_SQUOTE] = ACTIONS(97),
    [anon_sym_L_DQUOTE] = ACTIONS(99),
    [anon_sym_u_DQUOTE] = ACTIONS(99),
    [anon_sym_U_DQUOTE] = ACTIONS(99),
    [anon_sym_u8_DQUOTE] = ACTIONS(99),
    [anon_sym_DQUOTE] = ACTIONS(99),
    [sym_true] = ACTIONS(163),
    [sym_false] = ACTIONS(163),
    [anon_sym_NULL] = ACTIONS(103),
    [anon_sym_nullptr] = ACTIONS(103),
    [sym_comment] = ACTIONS(3),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2108), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1048), 1,
      sym_expression,
    STATE(1814), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [113] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(678), 1,
      sym_initializer_list,
    STATE(699), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [226] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2110), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1046), 1,
      sym_expression,
    STATE(1957), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [339] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2112), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1052), 1,
      sym_expression,
    STATE(1807), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [452] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1032), 1,
      sym_expression,
    STATE(1754), 1,
      sym_initializer_list,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [565] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2114), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1037), 1,
      sym_expression,
    STATE(1954), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [678] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2116), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1038), 1,
      sym_expression,
    STATE(1971), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [791] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(678), 1,
      sym_initializer_list,
    STATE(699), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [904] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1041), 1,
      sym_expression,
    STATE(1686), 1,
      sym_initializer_list,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1017] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2124), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1029), 1,
      sym_expression,
    STATE(1817), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1130] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2126), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1049), 1,
      sym_expression,
    STATE(1771), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1243] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(678), 1,
      sym_initializer_list,
    STATE(699), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1356] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2132), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1040), 1,
      sym_expression,
    STATE(1973), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1469] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2134), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1057), 1,
      sym_expression,
    STATE(1849), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1582] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2136), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1027), 1,
      sym_expression,
    STATE(1981), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1695] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2138), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1045), 1,
      sym_expression,
    STATE(1999), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1808] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2140), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1047), 1,
      sym_expression,
    STATE(1871), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [1921] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(678), 1,
      sym_initializer_list,
    STATE(889), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2034] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(678), 1,
      sym_initializer_list,
    STATE(699), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2147] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2148), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1019), 1,
      sym_expression,
    STATE(2014), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2260] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2150), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1070), 1,
      sym_expression,
    STATE(1851), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2373] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1025), 1,
      sym_expression,
    STATE(1758), 1,
      sym_compound_statement,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2486] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1043), 1,
      sym_expression,
    STATE(1712), 1,
      sym_initializer_list,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2599] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2152), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1065), 1,
      sym_expression,
    STATE(1989), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2712] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2154), 1,
      anon_sym_SEMI,
    STATE(663), 1,
      sym_string_literal,
    STATE(1034), 1,
      sym_expression,
    STATE(1770), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2825] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2156), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1069), 1,
      sym_expression,
    STATE(1809), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [2938] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2158), 1,
      anon_sym_RPAREN,
    STATE(663), 1,
      sym_string_literal,
    STATE(1022), 1,
      sym_expression,
    STATE(1843), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3051] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2160), 1,
      anon_sym_COLON,
    STATE(663), 1,
      sym_string_literal,
    STATE(1067), 1,
      sym_expression,
    STATE(1831), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3164] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1390), 1,
      anon_sym_LBRACE,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(678), 1,
      sym_initializer_list,
    STATE(889), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3277] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(2166), 1,
      anon_sym_LBRACE,
    STATE(687), 1,
      sym_ms_call_modifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1143), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(122), 3,
      sym_function_definition,
      sym_declaration,
      sym_declaration_list,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [3387] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2168), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3497] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2170), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3607] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2172), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3717] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(2174), 1,
      anon_sym_LBRACE,
    STATE(685), 1,
      sym_ms_call_modifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1144), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(260), 3,
      sym_function_definition,
      sym_declaration,
      sym_declaration_list,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [3827] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2176), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [3937] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2178), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4047] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(2180), 1,
      anon_sym_LBRACE,
    STATE(689), 1,
      sym_ms_call_modifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1145), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(287), 3,
      sym_function_definition,
      sym_declaration,
      sym_declaration_list,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [4157] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2182), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4267] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(2184), 1,
      anon_sym_LBRACE,
    STATE(676), 1,
      sym_ms_call_modifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1154), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(315), 3,
      sym_function_definition,
      sym_declaration,
      sym_declaration_list,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [4377] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2186), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4487] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2188), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4597] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2190), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4707] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1012), 1,
      sym_expression,
    STATE(1656), 1,
      sym_comma_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4817] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    ACTIONS(2192), 1,
      anon_sym_RBRACK,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [4927] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(831), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5034] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    ACTIONS(2194), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(702), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5141] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(829), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5248] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(973), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5355] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1078), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5462] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(983), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5569] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(834), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5676] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(985), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5783] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(978), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5890] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(974), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [5997] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(962), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6104] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1053), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6211] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1023), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6318] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(833), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6425] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(888), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6532] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(805), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6639] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(806), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6746] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(807), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6853] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(808), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [6960] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(809), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7067] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(810), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7174] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(811), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7281] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(812), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7388] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1082), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7495] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1092), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7602] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(819), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7709] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(813), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7816] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(961), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [7923] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1762), 1,
      anon_sym_const,
    ACTIONS(1766), 1,
      anon_sym_LPAREN2,
    ACTIONS(1772), 1,
      anon_sym_STAR,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    STATE(621), 1,
      sym_string_literal,
    STATE(762), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(1775), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
    ACTIONS(2196), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 11,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1778), 11,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(1764), 12,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [8014] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1090), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8121] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(814), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8228] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1039), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8335] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1095), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8442] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(691), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8549] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(975), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8656] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(701), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8763] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2198), 1,
      anon_sym_LPAREN2,
    STATE(702), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8870] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(980), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [8977] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(822), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9084] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(981), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9191] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(697), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9298] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(701), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9405] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(691), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9512] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1013), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9619] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(959), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9726] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(976), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9833] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1016), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [9940] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1091), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10047] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1094), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10154] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(697), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10261] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(977), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10368] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(691), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10475] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(818), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10582] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(701), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10689] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2200), 1,
      anon_sym_LPAREN2,
    STATE(702), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10796] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(692), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [10903] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(692), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11010] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1398), 1,
      anon_sym___extension__,
    ACTIONS(1400), 1,
      anon_sym_sizeof,
    ACTIONS(2118), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(817), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1394), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1396), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2122), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11117] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(701), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11224] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2202), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(702), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11331] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(982), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11438] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(830), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11545] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(697), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11652] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1073), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11759] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(825), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11866] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1076), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [11973] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(826), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12080] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1051), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12187] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1054), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12294] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(888), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12401] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1058), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12508] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1059), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12615] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1060), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12722] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1061), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12829] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1062), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [12936] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1063), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13043] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1064), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13150] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1066), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13257] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(979), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13364] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1020), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13471] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1024), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13578] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(892), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13685] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(893), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13792] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(909), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [13899] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(894), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14006] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(895), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14113] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(896), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14220] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(897), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14327] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(898), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14434] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(899), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14541] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(900), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14648] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(901), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14755] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(905), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14862] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(907), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [14969] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2204), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(885), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15076] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(902), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15183] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(984), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15290] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(963), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15397] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1072), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15504] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(815), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15611] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1085), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15718] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(964), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15825] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1088), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [15932] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(697), 1,
      sym_expression,
    STATE(713), 1,
      sym_string_literal,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16039] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(965), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16146] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(905), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16253] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1726), 1,
      anon_sym___extension__,
    ACTIONS(1728), 1,
      anon_sym_sizeof,
    ACTIONS(2162), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(906), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1722), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1724), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2164), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16360] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(907), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16467] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(2206), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(885), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16574] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(823), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16681] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1799), 1,
      sym_identifier,
    ACTIONS(1805), 1,
      anon_sym___extension__,
    ACTIONS(1807), 1,
      anon_sym_sizeof,
    ACTIONS(1823), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(1036), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1801), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1803), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(1827), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(1839), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(911), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16788] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(966), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [16895] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(967), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17002] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(968), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17109] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(816), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17216] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(969), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17323] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(970), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17430] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(971), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17537] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1744), 1,
      anon_sym___extension__,
    ACTIONS(1746), 1,
      anon_sym_sizeof,
    ACTIONS(2142), 1,
      sym_identifier,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(713), 1,
      sym_string_literal,
    STATE(972), 1,
      sym_expression,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1740), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1742), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2120), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2146), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(835), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17644] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(19), 1,
      anon_sym_LPAREN2,
    ACTIONS(85), 1,
      anon_sym_sizeof,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1404), 1,
      anon_sym___extension__,
    ACTIONS(1738), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(960), 1,
      sym_expression,
    ACTIONS(21), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(23), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(83), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(837), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17751] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(89), 1,
      anon_sym_offsetof,
    ACTIONS(91), 1,
      anon_sym__Generic,
    ACTIONS(161), 1,
      sym_number_literal,
    ACTIONS(1378), 1,
      sym_identifier,
    ACTIONS(1388), 1,
      anon_sym___extension__,
    ACTIONS(1392), 1,
      anon_sym_sizeof,
    ACTIONS(2128), 1,
      anon_sym_LPAREN2,
    STATE(663), 1,
      sym_string_literal,
    STATE(691), 1,
      sym_expression,
    ACTIONS(25), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(103), 2,
      anon_sym_NULL,
      anon_sym_nullptr,
    ACTIONS(163), 2,
      sym_true,
      sym_false,
    ACTIONS(1382), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1384), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(2130), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(87), 5,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
    ACTIONS(97), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(675), 5,
      sym_pointer_expression,
      sym_subscript_expression,
      sym_call_expression,
      sym_field_expression,
      sym_parenthesized_expression,
    STATE(684), 17,
      sym__string,
      sym_conditional_expression,
      sym_assignment_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_update_expression,
      sym_cast_expression,
      sym_sizeof_expression,
      sym_alignof_expression,
      sym_offsetof_expression,
      sym_generic_expression,
      sym_gnu_asm_expression,
      sym_extension_expression,
      sym_compound_literal_expression,
      sym_char_literal,
      sym_concatenated_string,
      sym_null,
  [17858] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2208), 1,
      sym_identifier,
    STATE(620), 2,
      sym_string_literal,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(2215), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2213), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2211), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [17930] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2218), 1,
      sym_identifier,
    STATE(622), 2,
      sym_string_literal,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2222), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2220), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18002] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2224), 1,
      sym_identifier,
    STATE(620), 2,
      sym_string_literal,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2228), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2226), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18074] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1811), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1813), 1,
      anon_sym_RPAREN,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1118), 1,
      sym__declaration_specifiers,
    STATE(1943), 1,
      sym_compound_statement,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1598), 2,
      sym_variadic_parameter,
      sym_parameter_declaration,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [18181] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1811), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(2230), 1,
      sym_identifier,
    ACTIONS(2232), 1,
      anon_sym_RPAREN,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1118), 1,
      sym__declaration_specifiers,
    STATE(1537), 1,
      sym_variadic_parameter,
    STATE(1598), 1,
      sym_parameter_declaration,
    STATE(1943), 1,
      sym_compound_statement,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [18290] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1770), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(1764), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18359] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2247), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2250), 1,
      anon_sym___declspec,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(2244), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2253), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2236), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
    STATE(626), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2238), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2241), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(2234), 17,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [18439] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2256), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2258), 38,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [18503] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2260), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2262), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18567] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2264), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2266), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18631] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2268), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2270), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18695] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2272), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2274), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18759] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2276), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2278), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18823] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2280), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2282), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18887] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2284), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2286), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [18951] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2288), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2290), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [19015] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2292), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2294), 38,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [19079] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2296), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2298), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [19143] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2300), 22,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2302), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [19207] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(294), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19308] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(335), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19409] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(301), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19510] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(130), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19611] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(325), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19712] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(140), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19813] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(280), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [19914] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(272), 1,
      sym_compound_statement,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20015] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2304), 1,
      sym_identifier,
    ACTIONS(2316), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2319), 1,
      anon_sym___declspec,
    ACTIONS(2322), 1,
      anon_sym_LBRACE,
    ACTIONS(2330), 1,
      sym_primitive_type,
    ACTIONS(2333), 1,
      anon_sym_enum,
    ACTIONS(2336), 1,
      anon_sym_struct,
    ACTIONS(2339), 1,
      anon_sym_union,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(2313), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2327), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(647), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(2324), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2307), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2310), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
  [20113] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2344), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(2342), 45,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
  [20175] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1811), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1118), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1748), 2,
      sym_variadic_parameter,
      sym_parameter_declaration,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20273] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1811), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(2346), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1118), 1,
      sym__declaration_specifiers,
    STATE(1695), 1,
      sym_variadic_parameter,
    STATE(1748), 1,
      sym_parameter_declaration,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20373] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(643), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20468] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(642), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20563] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2350), 21,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2348), 32,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_case,
      anon_sym_default,
      anon_sym_while,
      anon_sym_do,
      anon_sym_for,
      anon_sym_return,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_goto,
      anon_sym___try,
      anon_sym___leave,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
  [20624] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(644), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20719] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(640), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20814] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1736), 21,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1734), 32,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_case,
      anon_sym_default,
      anon_sym_while,
      anon_sym_do,
      anon_sym_for,
      anon_sym_return,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_goto,
      anon_sym___try,
      anon_sym___leave,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
  [20875] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(639), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [20970] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(645), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [21065] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1732), 21,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1730), 32,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_case,
      anon_sym_default,
      anon_sym_while,
      anon_sym_do,
      anon_sym_for,
      anon_sym_return,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_goto,
      anon_sym___try,
      anon_sym___leave,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
  [21126] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(641), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [21221] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1156), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(646), 2,
      sym_declaration,
      aux_sym__old_style_function_definition_repeat1,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [21316] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2359), 1,
      anon_sym_static,
    STATE(700), 1,
      sym_alignas_qualifier,
    ACTIONS(2362), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(662), 2,
      sym_type_qualifier,
      aux_sym_array_declarator_repeat1,
    ACTIONS(2356), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2352), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(2354), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [21387] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2365), 1,
      sym_identifier,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2369), 16,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2367), 29,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21453] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2348), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2350), 34,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21513] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(2378), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2373), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2375), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2371), 30,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [21581] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1243), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [21672] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2381), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2383), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21731] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2385), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2387), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21790] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2389), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2391), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21849] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2393), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2395), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21908] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2397), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2399), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [21967] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2401), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2403), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22026] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2405), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2407), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22085] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2409), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2411), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22144] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1770), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(1764), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22203] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1258), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [22294] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2413), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2415), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22353] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2417), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2419), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22412] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2421), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2423), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22471] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2425), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2427), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22530] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2429), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2431), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22589] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2433), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2435), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22648] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1240), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [22739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1770), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(1764), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [22798] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1259), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [22889] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1248), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [22980] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1261), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [23071] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1251), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [23162] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(720), 1,
      sym_type_specifier,
    STATE(1253), 1,
      sym__declaration_specifiers,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(695), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [23253] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2437), 18,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      anon_sym_DOT,
      sym_identifier,
    ACTIONS(2439), 33,
      anon_sym_DOT_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [23312] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2441), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2443), 28,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
  [23378] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2451), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2453), 26,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [23446] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2457), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2459), 33,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACK_RBRACK,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [23504] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2463), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2461), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_static,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
  [23562] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(55), 1,
      anon_sym_enum,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(704), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(719), 1,
      sym_type_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(45), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    STATE(626), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [23650] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2465), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2467), 33,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACK_RBRACK,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [23708] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2471), 26,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [23776] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2473), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2475), 33,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACK_RBRACK,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [23834] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2477), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2479), 26,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [23902] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1708), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1706), 31,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_static,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
  [23960] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2481), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2483), 26,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [24028] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2485), 17,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2487), 26,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [24096] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    STATE(728), 1,
      sym_field_declaration_list,
    STATE(749), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2491), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2489), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [24161] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2495), 1,
      sym_identifier,
    ACTIONS(2510), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(750), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2507), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(712), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2505), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2498), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2500), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2503), 21,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
  [24234] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    STATE(730), 1,
      sym_field_declaration_list,
    STATE(766), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2514), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2512), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [24299] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(2516), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [24356] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2463), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2461), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [24413] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    STATE(725), 1,
      sym_field_declaration_list,
    STATE(756), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2522), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2520), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [24478] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    STATE(732), 1,
      sym_field_declaration_list,
    STATE(734), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2526), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2524), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [24543] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    STATE(727), 1,
      sym_field_declaration_list,
    STATE(759), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2530), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2528), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [24608] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1708), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(1706), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [24665] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2532), 1,
      sym_identifier,
    ACTIONS(2547), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(745), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2544), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2542), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2535), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2537), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
    ACTIONS(2540), 21,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
  [24738] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2369), 15,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
      sym_identifier,
    ACTIONS(2367), 28,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [24799] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2549), 1,
      anon_sym_EQ,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2551), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 14,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      sym_identifier,
    ACTIONS(1764), 18,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [24864] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
    ACTIONS(1764), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [24929] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2555), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
    STATE(626), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2553), 9,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [25001] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1732), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1730), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [25057] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1736), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1734), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [25113] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2559), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
    STATE(721), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2557), 9,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [25185] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2563), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
    STATE(716), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2561), 9,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [25257] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2567), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
    STATE(626), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2565), 9,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [25329] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2571), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2569), 42,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [25385] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2577), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2575), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2573), 35,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25444] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(769), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2582), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2580), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25503] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(773), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2586), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2584), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25562] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(738), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2590), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2588), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25621] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(751), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2594), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2592), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25680] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(755), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2598), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2596), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25739] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(757), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2602), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2600), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25798] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(758), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2606), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2604), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25857] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(760), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2610), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2608), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25916] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(735), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2614), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2612), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [25975] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(763), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2618), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2616), 37,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26034] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2622), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2620), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26088] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2626), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2624), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26142] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(742), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2632), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2630), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2628), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26200] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2636), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2634), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26258] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2642), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2640), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26312] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(737), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2648), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2646), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2644), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26370] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2652), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2650), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26424] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2656), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2654), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26482] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2660), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2658), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26540] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(764), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2666), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2664), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2662), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26598] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(765), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2672), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2670), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2668), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26656] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2676), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2674), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26714] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2680), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2678), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26768] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2684), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2682), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26822] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2688), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2686), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26876] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2692), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2690), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [26930] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2573), 1,
      sym_primitive_type,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2577), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2697), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2694), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [26990] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2702), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2700), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27044] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2706), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2704), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27098] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2710), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2708), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27152] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2712), 1,
      anon_sym_LPAREN2,
    STATE(762), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(1780), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(1778), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(1762), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [27212] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2717), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2715), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27266] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2721), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2719), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27320] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2725), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2723), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27374] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2729), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2727), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27428] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2733), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2731), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27482] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2737), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2735), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27536] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2741), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2739), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27590] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2745), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2743), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [27648] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2749), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2747), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27702] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2753), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2751), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [27760] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2638), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2757), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2755), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [27818] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2761), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2759), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27872] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(741), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2767), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2765), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2763), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [27930] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2771), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2769), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [27984] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2775), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2773), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28038] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2646), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2644), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28092] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2779), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2777), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28146] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2783), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2781), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28200] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2787), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2785), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28254] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2791), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
    ACTIONS(2789), 39,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [28308] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2795), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2793), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28361] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2799), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2797), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28414] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2803), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2801), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28467] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1790), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [28530] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2807), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2805), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28583] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2811), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2809), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28636] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2813), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28689] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2817), 1,
      anon_sym_SEMI,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2563), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(716), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2561), 8,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [28760] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1788), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [28823] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2823), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1367), 1,
      sym__declarator,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2821), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2825), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(926), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [28914] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2835), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2833), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [28967] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1784), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [29030] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1356), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29083] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2823), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1366), 1,
      sym__declarator,
    STATE(1423), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    ACTIONS(2837), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2839), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(784), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(927), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [29174] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2841), 1,
      anon_sym_SEMI,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2563), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(716), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2561), 8,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [29245] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1794), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [29308] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1362), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1360), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29361] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1272), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1270), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29414] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1792), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [29477] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(2843), 1,
      anon_sym_COLON,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 15,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [29540] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1288), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1286), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29593] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2845), 1,
      anon_sym_SEMI,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2563), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(716), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2561), 8,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [29664] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1292), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1290), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29717] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1320), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1318), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29770] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2849), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2847), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29823] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2853), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2851), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [29876] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(1126), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2855), 1,
      anon_sym_SEMI,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2563), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(716), 7,
      sym__declaration_modifiers,
      sym_attribute_specifier,
      sym_attribute_declaration,
      sym_ms_declspec_modifier,
      sym_storage_class_specifier,
      sym_type_qualifier,
      aux_sym__declaration_specifiers_repeat1,
    ACTIONS(2561), 8,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      sym_identifier,
    ACTIONS(47), 10,
      anon_sym_extern,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [29947] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2859), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2857), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [30000] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2861), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [30053] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2865), 44,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elif_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [30106] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2873), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 18,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30188] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30268] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_EQ,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30346] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 4,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30422] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30496] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(2471), 21,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30568] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 7,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30636] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 9,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30702] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2873), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    ACTIONS(2893), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2895), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2891), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2889), 16,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30788] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2873), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    ACTIONS(2893), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2895), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2899), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2897), 16,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [30874] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [30954] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31028] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2873), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    ACTIONS(2893), 1,
      anon_sym_PIPE_PIPE,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2921), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2919), 17,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [31112] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 21,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31184] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 7,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31252] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2930), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2923), 3,
      anon_sym___attribute__,
      anon_sym___attribute,
      sym_identifier,
    ACTIONS(2933), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(2926), 5,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
    ACTIONS(2928), 31,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
  [31310] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2942), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2935), 3,
      anon_sym___attribute__,
      anon_sym___attribute,
      sym_identifier,
    ACTIONS(2945), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(2938), 5,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
    ACTIONS(2940), 31,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
  [31368] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 9,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31434] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      anon_sym_PIPE,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31512] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(2956), 1,
      anon_sym_COLON,
    STATE(746), 1,
      sym_attribute_specifier,
    STATE(874), 1,
      sym_enumerator_list,
    ACTIONS(2951), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2949), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2947), 32,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [31574] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    ACTIONS(2958), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2960), 1,
      anon_sym_AMP_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2921), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2919), 17,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31658] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    ACTIONS(2960), 1,
      anon_sym_AMP_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 18,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31740] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2962), 1,
      anon_sym_EQ,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2964), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(1764), 14,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [31800] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    ACTIONS(2958), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2960), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2970), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2968), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2966), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31886] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    ACTIONS(2958), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2960), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2970), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2891), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2889), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [31972] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 11,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [32036] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [32112] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2873), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2875), 1,
      anon_sym_PIPE,
    ACTIONS(2877), 1,
      anon_sym_CARET,
    ACTIONS(2879), 1,
      anon_sym_AMP,
    ACTIONS(2893), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2895), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2869), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2881), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2883), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2885), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2887), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2968), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_EQ,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2966), 16,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [32198] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2871), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 11,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2471), 23,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_identifier,
  [32262] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2905), 1,
      anon_sym_PIPE,
    ACTIONS(2907), 1,
      anon_sym_CARET,
    ACTIONS(2909), 1,
      anon_sym_AMP,
    ACTIONS(2958), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2960), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2970), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2899), 2,
      anon_sym___attribute,
      anon_sym_EQ,
    ACTIONS(2901), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2911), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2913), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2915), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2917), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2903), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2897), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [32348] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2549), 1,
      anon_sym_EQ,
    ACTIONS(2551), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 13,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 19,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      sym_identifier,
  [32403] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    STATE(772), 1,
      sym_attribute_specifier,
    STATE(883), 1,
      sym_enumerator_list,
    ACTIONS(2976), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2974), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2972), 32,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [32462] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym___attribute,
    ACTIONS(1764), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [32517] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1817), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    ACTIONS(2837), 1,
      anon_sym_RPAREN,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1291), 1,
      sym__declarator,
    STATE(1423), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(840), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(946), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [32603] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2294), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(2292), 40,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32653] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1817), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2821), 1,
      anon_sym_RPAREN,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1293), 1,
      sym__declarator,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(947), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [32739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2258), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(2256), 40,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32789] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2803), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2801), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32838] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2795), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2793), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32887] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2807), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2805), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32936] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2811), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2809), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [32985] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2813), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33034] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2853), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2851), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33083] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(747), 1,
      sym_attribute_specifier,
    ACTIONS(2983), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2981), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2979), 32,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [33136] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2859), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2857), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33185] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2861), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33234] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2799), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2797), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33283] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2803), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2801), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33332] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2865), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33381] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2835), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2833), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33430] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2849), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2847), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33479] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2795), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2793), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33528] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2807), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2805), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33577] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2811), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2809), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33626] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2813), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33675] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1356), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33724] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1362), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1360), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33773] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2859), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2857), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33822] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1288), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1286), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33871] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1292), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1290), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33920] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1320), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1318), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [33969] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1356), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34018] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1362), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1360), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34067] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1272), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1270), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34116] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1288), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1286), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34165] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1292), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1290), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1320), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1318), 40,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34263] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2853), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2851), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34312] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2571), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(493), 1,
      sym_string_literal,
    ACTIONS(2986), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2569), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34365] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(753), 1,
      sym_attribute_specifier,
    ACTIONS(2992), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2990), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2988), 32,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [34418] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2865), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34467] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2835), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2833), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34516] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2571), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(488), 1,
      sym_string_literal,
    ACTIONS(2986), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2569), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34569] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2849), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2847), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34618] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2863), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2861), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34667] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2799), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(2797), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34716] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2571), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(484), 1,
      sym_string_literal,
    ACTIONS(2986), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2569), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34769] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2571), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(491), 1,
      sym_string_literal,
    ACTIONS(2986), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(2569), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34822] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(768), 1,
      sym_attribute_specifier,
    ACTIONS(2999), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2997), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2995), 32,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [34875] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1272), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_RBRACE,
    ACTIONS(1270), 39,
      aux_sym_preproc_def_token1,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [34924] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2485), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2487), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [34983] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3008), 3,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3006), 36,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [35030] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1335), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(903), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1005), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [35107] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35166] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2477), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2479), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35225] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2712), 1,
      anon_sym_LPAREN2,
    ACTIONS(3020), 1,
      anon_sym_COMMA,
    ACTIONS(3023), 1,
      anon_sym_RPAREN,
    STATE(762), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1594), 1,
      aux_sym__old_style_parameter_list_repeat1,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1780), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(1762), 28,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [35284] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2968), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3032), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    ACTIONS(3048), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2966), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35367] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 10,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35428] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2469), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 14,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35507] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym_PIPE,
      anon_sym_EQ,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35582] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_EQ,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35655] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(2471), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35726] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 4,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ,
    ACTIONS(2471), 17,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35795] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 6,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35860] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2469), 8,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2471), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [35923] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2891), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3032), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    ACTIONS(3048), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2889), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36006] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2899), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3032), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    ACTIONS(3048), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2897), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36089] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2921), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3032), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2919), 13,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36170] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1341), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1008), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [36247] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1341), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(908), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1008), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [36324] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2441), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2443), 21,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
  [36381] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2451), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2453), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36440] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2481), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
    ACTIONS(2483), 19,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36499] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1326), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1009), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [36576] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2469), 1,
      anon_sym_EQ,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3034), 1,
      anon_sym_PIPE,
    ACTIONS(3036), 1,
      anon_sym_CARET,
    ACTIONS(3038), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3026), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3040), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3042), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3044), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3046), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3028), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2471), 15,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [36653] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2457), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2459), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [36699] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2962), 1,
      anon_sym_EQ,
    ACTIONS(2964), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 13,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
    ACTIONS(1764), 14,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [36749] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2712), 1,
      anon_sym_LPAREN2,
    STATE(762), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3050), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(1780), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(1762), 28,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_LBRACK,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [36803] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1206), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1208), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [36849] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1178), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1180), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [36895] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1202), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1204), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [36941] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2473), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2475), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [36987] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2465), 14,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_EQ,
      anon_sym_DOT,
    ACTIONS(2467), 24,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
  [37033] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1236), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37079] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1202), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1204), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37125] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1234), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1236), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37171] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1178), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1180), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37217] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1238), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1240), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37263] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1238), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1240), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37309] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3056), 1,
      anon_sym_LPAREN2,
    ACTIONS(3060), 1,
      anon_sym_LBRACK,
    STATE(762), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(1778), 2,
      anon_sym_COMMA,
      anon_sym_STAR,
    ACTIONS(3053), 2,
      anon_sym_RPAREN,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(1780), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(1762), 27,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym___based,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_identifier,
  [37365] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1174), 19,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym___extension__,
      anon_sym_sizeof,
      anon_sym___alignof__,
      anon_sym___alignof,
      anon_sym__alignof,
      anon_sym_alignof,
      anon_sym__Alignof,
      anon_sym_offsetof,
      anon_sym__Generic,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_true,
      sym_false,
      anon_sym_NULL,
      anon_sym_nullptr,
      sym_identifier,
    ACTIONS(1176), 19,
      anon_sym_LPAREN2,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SEMI,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      sym_number_literal,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [37411] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2823), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1355), 1,
      sym__declarator,
    STATE(1424), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(3063), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(3065), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [37486] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2823), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1367), 1,
      sym__declarator,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2821), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2825), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [37561] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1782), 1,
      anon_sym_EQ,
    ACTIONS(3067), 1,
      anon_sym_SEMI,
    ACTIONS(1786), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(1770), 12,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(1764), 13,
      anon_sym_LPAREN2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [37612] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1236), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1234), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37656] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1240), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1238), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37700] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1240), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1238), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37744] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1204), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1202), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37788] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1204), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1202), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37832] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1180), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1178), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37876] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1180), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1178), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37920] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1236), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1234), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [37964] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1208), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1206), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [38008] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1176), 2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
    ACTIONS(1174), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [38052] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    ACTIONS(3076), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    ACTIONS(3073), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3071), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(3069), 21,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [38103] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3081), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3079), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [38146] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2827), 1,
      sym_ms_restrict_modifier,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    ACTIONS(3085), 2,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3087), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1086), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2821), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38219] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3091), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(2928), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [38262] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2827), 1,
      sym_ms_restrict_modifier,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1423), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    ACTIONS(3085), 2,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    ACTIONS(3087), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(941), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1098), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2837), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38335] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3095), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3093), 34,
      anon_sym___extension__,
      anon_sym_extern,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___declspec,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_static,
      anon_sym_auto,
      anon_sym_register,
      anon_sym_inline,
      anon_sym___inline,
      anon_sym___inline__,
      anon_sym___forceinline,
      anon_sym_thread_local,
      anon_sym___thread,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      anon_sym_enum,
      anon_sym_struct,
      anon_sym_union,
      sym_identifier,
  [38378] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1319), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1106), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38446] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1817), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2821), 1,
      anon_sym_RPAREN,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1293), 1,
      sym__declarator,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38516] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1817), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    ACTIONS(3063), 1,
      anon_sym_RPAREN,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1294), 1,
      sym__declarator,
    STATE(1424), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38586] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1294), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1111), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38654] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1291), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(950), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1107), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38722] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1293), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1109), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38790] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1322), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1110), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38858] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1293), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(948), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1109), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38926] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1318), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(945), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1108), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [38994] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1367), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1104), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [39062] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1367), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(956), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1104), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [39130] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1355), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(939), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1105), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [39198] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1366), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(954), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1102), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [39266] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(989), 1,
      sym_ms_unaligned_ptr_modifier,
    STATE(1319), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(2829), 2,
      anon_sym__unaligned,
      anon_sym___unaligned,
    STATE(951), 2,
      sym_ms_pointer_modifier,
      aux_sym_pointer_declarator_repeat1,
    STATE(1106), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2827), 3,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [39334] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2469), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 8,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [39407] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym___attribute,
    ACTIONS(2471), 12,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [39470] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(2471), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [39527] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2891), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2889), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [39604] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2968), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3141), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    ACTIONS(3159), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2966), 6,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [39681] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 7,
      aux_sym_preproc_elif_token1,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 16,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_QMARK,
      sym_identifier,
  [39736] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2469), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 8,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      sym_identifier,
  [39809] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2469), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 9,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_identifier,
  [39880] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 9,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_identifier,
  [39949] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 10,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_QMARK,
      sym_identifier,
  [40016] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2471), 10,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_QMARK,
      sym_identifier,
  [40081] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2471), 12,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK,
      sym_identifier,
  [40144] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 5,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 14,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_QMARK,
      sym_identifier,
  [40203] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 5,
      aux_sym_preproc_elif_token1,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 16,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_QMARK,
      sym_identifier,
  [40260] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2891), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3141), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    ACTIONS(3159), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2889), 6,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [40337] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2899), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3141), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    ACTIONS(3159), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2897), 6,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [40414] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 7,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(2471), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [40469] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2469), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [40540] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym_PIPE,
      anon_sym___attribute,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [40609] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2899), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2897), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [40686] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym_PIPE,
      anon_sym___attribute,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [40753] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2921), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2919), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [40828] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2921), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3141), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2919), 7,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      anon_sym_QMARK,
      sym_identifier,
  [40903] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2968), 1,
      anon_sym___attribute,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2966), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [40980] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3139), 1,
      anon_sym_SLASH,
    ACTIONS(3141), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3143), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3145), 1,
      anon_sym_PIPE,
    ACTIONS(3147), 1,
      anon_sym_CARET,
    ACTIONS(3149), 1,
      anon_sym_AMP,
    ACTIONS(3159), 1,
      anon_sym_QMARK,
    ACTIONS(3163), 1,
      aux_sym_preproc_elif_token1,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3135), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3137), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3151), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3153), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3155), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3157), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3161), 6,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [41057] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym___attribute,
    ACTIONS(2471), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [41122] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute,
    ACTIONS(2471), 14,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_QMARK,
  [41181] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3165), 13,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(3167), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
  [41221] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(3169), 1,
      anon_sym_enum,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1077), 1,
      sym_type_specifier,
    STATE(1919), 1,
      sym_type_descriptor,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1007), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41285] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1077), 1,
      sym_type_specifier,
    STATE(1813), 1,
      sym_type_descriptor,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1006), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41349] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3173), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(3171), 26,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___based,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
      anon_sym__unaligned,
      anon_sym___unaligned,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [41389] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3177), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(3175), 26,
      anon_sym___extension__,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym___based,
      sym_ms_restrict_modifier,
      sym_ms_unsigned_ptr_modifier,
      sym_ms_signed_ptr_modifier,
      anon_sym__unaligned,
      anon_sym___unaligned,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [41429] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3179), 13,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(3181), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
  [41469] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1270), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41533] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1265), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41597] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1273), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41661] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1077), 1,
      sym_type_specifier,
    STATE(1782), 1,
      sym_type_descriptor,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1006), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41725] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1077), 1,
      sym_type_specifier,
    STATE(1812), 1,
      sym_type_descriptor,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1006), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41789] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1267), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41853] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3189), 13,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(3191), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
  [41893] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1266), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [41957] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(3169), 1,
      anon_sym_enum,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1077), 1,
      sym_type_specifier,
    STATE(1944), 1,
      sym_type_descriptor,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1007), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42021] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3193), 13,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(3195), 19,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
  [42061] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1271), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42125] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1274), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42189] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1115), 1,
      sym_type_specifier,
    STATE(1272), 1,
      sym__type_definition_type,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1010), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42253] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1341), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42314] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(1938), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1093), 1,
      sym_type_specifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42375] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(53), 1,
      sym_primitive_type,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1938), 1,
      sym_identifier,
    ACTIONS(3169), 1,
      anon_sym_enum,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1026), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1093), 1,
      sym_type_specifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(1718), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42436] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1326), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42497] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1323), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42558] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_struct,
    ACTIONS(59), 1,
      anon_sym_union,
    ACTIONS(1720), 1,
      anon_sym_enum,
    ACTIONS(3183), 1,
      sym_identifier,
    ACTIONS(3187), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1114), 1,
      aux_sym_sized_type_specifier_repeat1,
    STATE(1116), 1,
      sym_type_specifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3185), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(770), 5,
      sym_sized_type_specifier,
      sym_enum_specifier,
      sym_struct_specifier,
      sym_union_specifier,
      sym_macro_type_specifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [42619] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    STATE(772), 1,
      sym_attribute_specifier,
    STATE(1068), 1,
      sym_enumerator_list,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2974), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2972), 19,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [42665] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3199), 3,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_COLON,
  [42739] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3203), 1,
      anon_sym___attribute,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [42813] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(3205), 1,
      anon_sym_COLON,
    STATE(746), 1,
      sym_attribute_specifier,
    STATE(1021), 1,
      sym_enumerator_list,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2949), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
    ACTIONS(2947), 19,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [42861] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3207), 1,
      anon_sym_COMMA,
    ACTIONS(3209), 1,
      anon_sym_RBRACE,
    STATE(668), 1,
      sym_argument_list,
    STATE(1660), 1,
      aux_sym_initializer_list_repeat1,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [42936] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3211), 1,
      anon_sym_COMMA,
    ACTIONS(3213), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    STATE(1568), 1,
      aux_sym_generic_expression_repeat1,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43011] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(3217), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    STATE(1654), 1,
      aux_sym_argument_list_repeat1,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43086] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(3219), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    STATE(1572), 1,
      aux_sym_argument_list_repeat1,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43161] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3221), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43233] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2889), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43305] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(753), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2990), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2988), 19,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [43345] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3249), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43417] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3251), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [43487] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2897), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43559] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3253), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [43629] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2510), 1,
      sym_primitive_type,
    ACTIONS(3255), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1113), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2507), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1044), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3257), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2498), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2500), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [43679] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3259), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43751] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3261), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43823] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3263), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [43895] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3265), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [43965] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3267), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44037] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3269), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [44107] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3271), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44179] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3273), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44251] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3275), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44323] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2919), 3,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [44393] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3277), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44465] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3279), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44537] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    ACTIONS(3281), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(3283), 1,
      anon_sym_RBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44611] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3285), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44683] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3287), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [44753] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3289), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [44825] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3291), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [44895] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2547), 1,
      sym_primitive_type,
    ACTIONS(3293), 1,
      sym_identifier,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(745), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2544), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2542), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2535), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2537), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [44945] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3295), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45017] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3297), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45089] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3299), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45161] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3301), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45233] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3303), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45305] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(747), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2981), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2979), 19,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [45345] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2966), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_RBRACK,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45417] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3305), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45489] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3161), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [45559] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 6,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [45611] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3307), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45683] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3309), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45755] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3311), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [45827] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 4,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [45895] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 5,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [45961] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 5,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46027] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(2469), 1,
      anon_sym_PIPE,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 6,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46091] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 6,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46153] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(2469), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2471), 8,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46213] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2469), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 10,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46269] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3313), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46341] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(2449), 1,
      anon_sym_DASH_GT,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3004), 1,
      anon_sym_DOT,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2469), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2471), 12,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
  [46395] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3315), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46467] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(768), 1,
      sym_attribute_specifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(2997), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2995), 19,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [46507] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3317), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46579] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    ACTIONS(3319), 1,
      anon_sym_SEMI,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46651] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2188), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46720] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3321), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46789] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3323), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46858] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1469), 1,
      sym__abstract_declarator,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3325), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [46913] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2192), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [46982] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3327), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47051] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1474), 1,
      sym__abstract_declarator,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1079), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3329), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [47106] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3331), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47175] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1483), 1,
      sym__abstract_declarator,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3333), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [47230] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3335), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47299] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2170), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47368] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3337), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47437] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2190), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47506] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3339), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47575] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3341), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47644] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1424), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3063), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [47699] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2186), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47768] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3343), 1,
      anon_sym_COLON,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47837] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2178), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47906] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    ACTIONS(3345), 1,
      anon_sym_RBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [47975] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3347), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48044] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3349), 1,
      anon_sym_RPAREN,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48113] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1470), 1,
      sym__abstract_declarator,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1074), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3351), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48168] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    ACTIONS(3353), 1,
      anon_sym_RBRACK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48237] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    ACTIONS(3355), 1,
      anon_sym_COMMA,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48306] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2172), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48375] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2182), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48444] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(49), 1,
      anon_sym_const,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1398), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    ACTIONS(3089), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(2821), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    ACTIONS(3083), 9,
      anon_sym___extension__,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48499] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2168), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48568] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2176), 1,
      anon_sym_RBRACK,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3002), 1,
      anon_sym_LPAREN2,
    ACTIONS(3227), 1,
      anon_sym_SLASH,
    ACTIONS(3229), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3231), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3233), 1,
      anon_sym_PIPE,
    ACTIONS(3235), 1,
      anon_sym_CARET,
    ACTIONS(3237), 1,
      anon_sym_AMP,
    ACTIONS(3247), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3223), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3225), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3239), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3241), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3243), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3245), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48637] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(2447), 1,
      anon_sym_LBRACK,
    ACTIONS(3113), 1,
      anon_sym_SLASH,
    ACTIONS(3115), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3117), 1,
      anon_sym_PIPE,
    ACTIONS(3119), 1,
      anon_sym_CARET,
    ACTIONS(3121), 1,
      anon_sym_AMP,
    ACTIONS(3131), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3133), 1,
      anon_sym_QMARK,
    STATE(668), 1,
      sym_argument_list,
    ACTIONS(2449), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2455), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(3109), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3111), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3123), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3125), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3127), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3129), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [48703] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1367), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48755] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3067), 1,
      anon_sym_SEMI,
    ACTIONS(1770), 7,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1764), 18,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DOT,
      anon_sym_DASH_GT,
  [48791] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1355), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48843] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3107), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1359), 1,
      sym__declarator,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48895] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1322), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48947] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1293), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [48999] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1319), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49051] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1294), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49103] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1317), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49155] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1292), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49207] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2540), 1,
      anon_sym___based,
    ACTIONS(3357), 1,
      sym_identifier,
    ACTIONS(3363), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1130), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2535), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(2544), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3360), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2537), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49256] = 6,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2573), 2,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(2577), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2697), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_COLON,
    ACTIONS(2694), 12,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
  [49295] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2503), 1,
      anon_sym___based,
    ACTIONS(3366), 1,
      sym_identifier,
    ACTIONS(3372), 1,
      sym_primitive_type,
    STATE(711), 1,
      sym_alignas_qualifier,
    STATE(1136), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2498), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(2507), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    STATE(1112), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3369), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2500), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49344] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(3377), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(1119), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3375), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49384] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(3381), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(1117), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3379), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49424] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(3385), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3383), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49464] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2823), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(1321), 1,
      sym__declarator,
    STATE(1393), 1,
      sym__abstract_declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1984), 1,
      sym_ms_based_modifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(3387), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1489), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [49520] = 7,
    ACTIONS(3), 1,
      sym_comment,
    STATE(711), 1,
      sym_alignas_qualifier,
    ACTIONS(51), 2,
      anon_sym_alignas,
      anon_sym__Alignas,
    ACTIONS(3391), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    STATE(665), 2,
      sym_type_qualifier,
      aux_sym__type_definition_type_repeat1,
    ACTIONS(3389), 7,
      anon_sym___based,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_primitive_type,
      sym_identifier,
    ACTIONS(49), 10,
      anon_sym___extension__,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
  [49560] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1280), 1,
      sym_ms_call_modifier,
    STATE(1430), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [49609] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3395), 1,
      anon_sym_LPAREN2,
    STATE(991), 1,
      sym_preproc_argument_list,
    ACTIONS(3397), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3393), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [49643] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2753), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3399), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2751), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49677] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1124), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2765), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3402), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2763), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49711] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2656), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3405), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2654), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49745] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2757), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3408), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2755), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49779] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2660), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3411), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2658), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49813] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1122), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2664), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3414), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2662), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49847] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(651), 1,
      sym__old_style_function_declarator,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1316), 1,
      sym_ms_call_modifier,
    STATE(1356), 1,
      sym__declarator,
    STATE(1458), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1606), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [49901] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1125), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2670), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3419), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2668), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49935] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2676), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3422), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2674), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [49969] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1126), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2630), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3425), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2628), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50003] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(652), 1,
      sym__old_style_function_declarator,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1302), 1,
      sym_ms_call_modifier,
    STATE(1362), 1,
      sym__declarator,
    STATE(1447), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1575), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50057] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(657), 1,
      sym__old_style_function_declarator,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1310), 1,
      sym_ms_call_modifier,
    STATE(1364), 1,
      sym__declarator,
    STATE(1439), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1632), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50111] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2745), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3428), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2743), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50145] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2636), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3431), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2634), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50179] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(723), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2697), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3434), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2694), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50213] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1778), 1,
      anon_sym_STAR,
    ACTIONS(2712), 1,
      anon_sym_LPAREN2,
    STATE(1134), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(3438), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(1762), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50249] = 5,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1135), 1,
      aux_sym_sized_type_specifier_repeat1,
    ACTIONS(2646), 2,
      anon_sym_LPAREN2,
      anon_sym_STAR,
    ACTIONS(3441), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    ACTIONS(2644), 15,
      anon_sym___extension__,
      anon_sym___based,
      anon_sym_const,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
      sym_primitive_type,
      sym_identifier,
  [50283] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(661), 1,
      sym__old_style_function_declarator,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1312), 1,
      sym_ms_call_modifier,
    STATE(1361), 1,
      sym__declarator,
    STATE(1455), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1592), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50337] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1315), 1,
      sym_ms_call_modifier,
    STATE(1455), 1,
      sym__declaration_declarator,
    STATE(1457), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1592), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50388] = 5,
    ACTIONS(3393), 1,
      anon_sym_LF,
    ACTIONS(3444), 1,
      anon_sym_LPAREN2,
    ACTIONS(3446), 1,
      sym_comment,
    STATE(1230), 1,
      sym_preproc_argument_list,
    ACTIONS(3397), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [50421] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1307), 1,
      sym_ms_call_modifier,
    STATE(1447), 1,
      sym__declaration_declarator,
    STATE(1457), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1575), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50472] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1302), 1,
      sym_ms_call_modifier,
    STATE(1380), 1,
      sym__declarator,
    STATE(1447), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1575), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50523] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1312), 1,
      sym_ms_call_modifier,
    STATE(1376), 1,
      sym__declarator,
    STATE(1455), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1592), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50574] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1310), 1,
      sym_ms_call_modifier,
    STATE(1372), 1,
      sym__declarator,
    STATE(1439), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1632), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50625] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3450), 1,
      anon_sym_RPAREN,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3460), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1149), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [50668] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1314), 1,
      sym_ms_call_modifier,
    STATE(1439), 1,
      sym__declaration_declarator,
    STATE(1457), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1632), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50719] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1305), 1,
      sym_ms_call_modifier,
    STATE(1457), 1,
      sym__declarator,
    STATE(1477), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1765), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50770] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(3466), 1,
      anon_sym_RPAREN,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    STATE(1655), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [50825] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3492), 1,
      anon_sym_RPAREN,
    ACTIONS(3494), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1152), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [50868] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1311), 1,
      sym_ms_call_modifier,
    STATE(1457), 1,
      sym__declarator,
    STATE(1465), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1605), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [50919] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3496), 1,
      anon_sym_RPAREN,
    STATE(1593), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [50974] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3498), 1,
      anon_sym_RPAREN,
    STATE(1578), 1,
      aux_sym_preproc_argument_list_repeat1,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [51029] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1316), 1,
      sym_ms_call_modifier,
    STATE(1387), 1,
      sym__declarator,
    STATE(1458), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1606), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [51080] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1298), 1,
      sym_ms_call_modifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1457), 1,
      sym__declarator,
    STATE(1458), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1606), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [51131] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1295), 1,
      sym_ms_call_modifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1437), 1,
      sym__declaration_declarator,
    STATE(1457), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1678), 1,
      sym_init_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [51182] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3500), 1,
      anon_sym_RPAREN,
    ACTIONS(3502), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1153), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51225] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(2947), 1,
      anon_sym_const,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(3506), 1,
      anon_sym_COLON,
    STATE(746), 1,
      sym_attribute_specifier,
    STATE(1021), 1,
      sym_enumerator_list,
    ACTIONS(2949), 14,
      anon_sym_LPAREN2,
      anon_sym_STAR,
      anon_sym___extension__,
      anon_sym_LBRACK,
      anon_sym_constexpr,
      anon_sym_volatile,
      anon_sym_restrict,
      anon_sym___restrict__,
      anon_sym__Atomic,
      anon_sym__Noreturn,
      anon_sym_noreturn,
      anon_sym__Nonnull,
      anon_sym_alignas,
      anon_sym__Alignas,
  [51266] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3519), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1249), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51306] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3523), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1201), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51346] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3525), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1228), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51386] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3527), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1223), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51426] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3529), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1216), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51466] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3531), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1232), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51506] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3535), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3533), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [51534] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3537), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1225), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51574] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3539), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1229), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51614] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3541), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1210), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51654] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3543), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1224), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51694] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3545), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1220), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51734] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3547), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1179), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51774] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3549), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1211), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51814] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3551), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1203), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51854] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3553), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1218), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51894] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3555), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1231), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51934] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3557), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1244), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [51974] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3559), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1226), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52014] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3561), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1214), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52054] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3565), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3563), 11,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [52088] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3567), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1217), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3571), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3569), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [52156] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3573), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [52206] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3575), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1165), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52246] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3577), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1233), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52286] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3579), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1235), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52326] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3581), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1241), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52366] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3583), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1207), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52406] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3585), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1191), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52446] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3587), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1252), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52486] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3589), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1219), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52526] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3565), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(3563), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [52566] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3591), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1209), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52606] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3595), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3593), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [52634] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3597), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1196), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52674] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3599), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1245), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52714] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3565), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3563), 13,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [52746] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3601), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1246), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52786] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3603), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1222), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52826] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3605), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1247), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52866] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3607), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1234), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52906] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3565), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3563), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [52934] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3609), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1204), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [52974] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3563), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
  [53022] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3565), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3563), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [53058] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3611), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1250), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [53098] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3509), 1,
      sym_identifier,
    ACTIONS(3511), 1,
      anon_sym_LPAREN2,
    ACTIONS(3513), 1,
      anon_sym_defined,
    ACTIONS(3613), 1,
      sym_number_literal,
    ACTIONS(3515), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3517), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3521), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1221), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [53138] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3563), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [53184] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3617), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3615), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53212] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3565), 1,
      anon_sym_PIPE,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3563), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [53258] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3565), 1,
      anon_sym_PIPE,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3563), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [53302] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3565), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(3563), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [53344] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2437), 5,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2439), 15,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53372] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3448), 1,
      sym_identifier,
    ACTIONS(3452), 1,
      anon_sym_LPAREN2,
    ACTIONS(3454), 1,
      anon_sym_defined,
    ACTIONS(3619), 1,
      sym_number_literal,
    ACTIONS(3456), 2,
      anon_sym_BANG,
      anon_sym_TILDE,
    ACTIONS(3458), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3462), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    STATE(1182), 7,
      sym__preproc_expression,
      sym_preproc_parenthesized_expression,
      sym_preproc_defined,
      sym_preproc_unary_expression,
      sym_preproc_call_expression,
      sym_preproc_binary_expression,
      sym_char_literal,
  [53412] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3621), 1,
      anon_sym_RPAREN,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53461] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3569), 1,
      anon_sym_LF,
    ACTIONS(3571), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53488] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3565), 15,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53517] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3565), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53544] = 11,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3565), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53587] = 10,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3565), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53628] = 9,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3565), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53667] = 8,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
    ACTIONS(3565), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
  [53704] = 7,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
    ACTIONS(3565), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [53739] = 6,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3565), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53772] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3565), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53803] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3639), 1,
      anon_sym_LF,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53848] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3645), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53893] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_LF,
    ACTIONS(3595), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [53920] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3647), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [53965] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3649), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54010] = 3,
    ACTIONS(3181), 1,
      anon_sym_LF,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3179), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54037] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3651), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54082] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3472), 1,
      anon_sym_SLASH,
    ACTIONS(3474), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3476), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    ACTIONS(3480), 1,
      anon_sym_CARET,
    ACTIONS(3482), 1,
      anon_sym_AMP,
    ACTIONS(3653), 1,
      anon_sym_RPAREN,
    ACTIONS(3468), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3470), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3484), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3486), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3488), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3490), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54131] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3533), 1,
      anon_sym_LF,
    ACTIONS(3535), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54158] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3655), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54203] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3657), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54248] = 3,
    ACTIONS(2439), 1,
      anon_sym_LF,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(2437), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54275] = 3,
    ACTIONS(3167), 1,
      anon_sym_LF,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3165), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54302] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3615), 1,
      anon_sym_LF,
    ACTIONS(3617), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54329] = 3,
    ACTIONS(3195), 1,
      anon_sym_LF,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3193), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54356] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(655), 1,
      sym__old_style_function_declarator,
    STATE(1350), 1,
      sym_ms_call_modifier,
    STATE(1390), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [54399] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3659), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54444] = 3,
    ACTIONS(3191), 1,
      anon_sym_LF,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3189), 18,
      anon_sym_DASH,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [54471] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(658), 1,
      sym__old_style_function_declarator,
    STATE(1344), 1,
      sym_ms_call_modifier,
    STATE(1384), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [54514] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3661), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54559] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3663), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54604] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3665), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54649] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3667), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54694] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(654), 1,
      sym__old_style_function_declarator,
    STATE(1348), 1,
      sym_ms_call_modifier,
    STATE(1379), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [54737] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3669), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54782] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3641), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3671), 1,
      anon_sym_LF,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54827] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(660), 1,
      sym__old_style_function_declarator,
    STATE(1349), 1,
      sym_ms_call_modifier,
    STATE(1371), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [54870] = 12,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(3563), 1,
      anon_sym_LF,
    ACTIONS(3565), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3627), 1,
      anon_sym_PIPE,
    ACTIONS(3629), 1,
      anon_sym_CARET,
    ACTIONS(3631), 1,
      anon_sym_AMP,
    ACTIONS(3643), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3625), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(3633), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3637), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3623), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3635), 4,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT,
  [54915] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1349), 1,
      sym_ms_call_modifier,
    STATE(1413), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [54955] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3677), 1,
      anon_sym_LBRACK,
    STATE(1268), 1,
      sym_gnu_asm_expression,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    STATE(1264), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3675), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [54991] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(1352), 1,
      sym_ms_call_modifier,
    STATE(1460), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [55031] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1337), 1,
      sym_ms_call_modifier,
    STATE(1442), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [55071] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1815), 1,
      anon_sym_LPAREN2,
    ACTIONS(1817), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(1433), 1,
      sym__declarator,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1551), 1,
      sym__abstract_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [55115] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1350), 1,
      sym_ms_call_modifier,
    STATE(1403), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [55155] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1344), 1,
      sym_ms_call_modifier,
    STATE(1422), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [55195] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3677), 1,
      anon_sym_LBRACK,
    STATE(1263), 1,
      sym_gnu_asm_expression,
    STATE(1313), 1,
      sym_attribute_specifier,
    STATE(1370), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(3679), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(1264), 2,
      sym_preproc_call_expression,
      aux_sym_function_declarator_repeat1,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3675), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55237] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1348), 1,
      sym_ms_call_modifier,
    STATE(1406), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
    ACTIONS(41), 6,
      anon_sym___cdecl,
      anon_sym___clrcall,
      anon_sym___stdcall,
      anon_sym___fastcall,
      anon_sym___thiscall,
      anon_sym___vectorcall,
  [55277] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1269), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3683), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3681), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55308] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3689), 1,
      anon_sym_LBRACK,
    STATE(1313), 1,
      sym_attribute_specifier,
    STATE(1369), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(3685), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    STATE(1262), 2,
      sym_preproc_call_expression,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3691), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3687), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55347] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1269), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3689), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3687), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55378] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1503), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55419] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1513), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55460] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1510), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55501] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1262), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3689), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3687), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55532] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3693), 1,
      sym_identifier,
    ACTIONS(3698), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1269), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3701), 4,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3696), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [55563] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1493), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55604] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1515), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55645] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1502), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55686] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1496), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55727] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1338), 1,
      sym__type_declarator,
    STATE(1508), 1,
      sym__type_definition_declarators,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55768] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    ACTIONS(3703), 1,
      anon_sym_SEMI,
    STATE(1304), 1,
      sym__field_declarator,
    STATE(1536), 1,
      sym__field_declaration_declarator,
    STATE(1976), 1,
      sym_attribute_specifier,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
  [55810] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1357), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55848] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3709), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(1277), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3707), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3705), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_asm,
      anon_sym___asm__,
  [55876] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    ACTIONS(3712), 1,
      anon_sym_SEMI,
    STATE(1304), 1,
      sym__field_declarator,
    STATE(1561), 1,
      sym__field_declaration_declarator,
    STATE(1869), 1,
      sym_attribute_specifier,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
  [55918] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    ACTIONS(3714), 1,
      anon_sym_SEMI,
    STATE(1304), 1,
      sym__field_declarator,
    STATE(1521), 1,
      sym__field_declaration_declarator,
    STATE(1852), 1,
      sym_attribute_specifier,
    STATE(1990), 1,
      sym_ms_based_modifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
  [55960] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3010), 1,
      sym_identifier,
    ACTIONS(3012), 1,
      anon_sym_LPAREN2,
    ACTIONS(3014), 1,
      anon_sym_STAR,
    ACTIONS(3018), 1,
      sym_primitive_type,
    STATE(1446), 1,
      sym__type_declarator,
    STATE(1920), 1,
      sym_ms_based_modifier,
    ACTIONS(3016), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1415), 5,
      sym_parenthesized_type_declarator,
      sym_attributed_type_declarator,
      sym_pointer_type_declarator,
      sym_function_type_declarator,
      sym_array_type_declarator,
  [55998] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3677), 1,
      anon_sym_LBRACK,
    STATE(1325), 1,
      sym_gnu_asm_expression,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(93), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    STATE(1331), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3675), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
  [56031] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2945), 7,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(2938), 8,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_COLON,
  [56054] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3718), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3720), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3722), 1,
      aux_sym_preproc_elif_token1,
    STATE(1324), 1,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1334), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1451), 1,
      sym_enumerator,
    ACTIONS(3724), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1787), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
    STATE(1860), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [56093] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3395), 1,
      anon_sym_LPAREN2,
    STATE(991), 1,
      sym_preproc_argument_list,
    ACTIONS(3728), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
    ACTIONS(3726), 7,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
  [56120] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3720), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3722), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3730), 1,
      aux_sym_preproc_if_token2,
    STATE(1346), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1347), 1,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1451), 1,
      sym_enumerator,
    ACTIONS(3724), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1894), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
    STATE(1895), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [56159] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3720), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3722), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3732), 1,
      aux_sym_preproc_if_token2,
    STATE(1329), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1330), 1,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1451), 1,
      sym_enumerator,
    ACTIONS(3724), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1767), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
    STATE(1853), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [56198] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3720), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3722), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3734), 1,
      aux_sym_preproc_if_token2,
    STATE(1339), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1340), 1,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1451), 1,
      sym_enumerator,
    ACTIONS(3724), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1836), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
    STATE(1838), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [56237] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3736), 7,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(3738), 8,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_COLON,
  [56260] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    STATE(1277), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3742), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3740), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56287] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2933), 7,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
      sym_identifier,
    ACTIONS(2926), 8,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_COLON,
  [56310] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3750), 1,
      anon_sym___asm,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3744), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56342] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3754), 1,
      anon_sym___asm,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3752), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56374] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3758), 1,
      anon_sym___asm,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3756), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56406] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3762), 1,
      anon_sym___asm,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3760), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56438] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1438), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56475] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(17), 1,
      sym_preproc_directive,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      aux_sym_preproc_if_token1,
    ACTIONS(3770), 1,
      anon_sym_RBRACE,
    ACTIONS(3768), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(1718), 2,
      sym_preproc_call,
      sym_enumerator,
    STATE(1823), 2,
      sym_preproc_if_in_enumerator_list_no_comma,
      sym_preproc_ifdef_in_enumerator_list_no_comma,
    STATE(1308), 3,
      sym_preproc_if_in_enumerator_list,
      sym_preproc_ifdef_in_enumerator_list,
      aux_sym_enumerator_list_repeat1,
  [56508] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3774), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3772), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56529] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1450), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56566] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3778), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3776), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3782), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3780), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56608] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3786), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3784), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56629] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1418), 1,
      sym__declarator,
    STATE(1448), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56666] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3790), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3788), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56687] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3792), 1,
      anon_sym_COMMA,
    ACTIONS(3796), 1,
      anon_sym___attribute,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3800), 1,
      anon_sym_COLON,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1514), 1,
      aux_sym__field_declaration_declarator_repeat1,
    STATE(1518), 1,
      sym_bitfield_clause,
    ACTIONS(3794), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [56726] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1484), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56763] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3804), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3802), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56784] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1448), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56821] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(17), 1,
      sym_preproc_directive,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      aux_sym_preproc_if_token1,
    ACTIONS(3806), 1,
      anon_sym_RBRACE,
    ACTIONS(3768), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(1734), 2,
      sym_preproc_call,
      sym_enumerator,
    STATE(1889), 2,
      sym_preproc_if_in_enumerator_list_no_comma,
      sym_preproc_ifdef_in_enumerator_list_no_comma,
    STATE(1332), 3,
      sym_preproc_if_in_enumerator_list,
      sym_preproc_ifdef_in_enumerator_list,
      aux_sym_enumerator_list_repeat1,
  [56854] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3810), 3,
      anon_sym___attribute,
      anon_sym_LBRACK,
      anon_sym___asm,
    ACTIONS(3808), 10,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_asm,
      anon_sym___asm__,
  [56875] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1419), 1,
      sym__declarator,
    STATE(1440), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56912] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1434), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56949] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1405), 1,
      sym__declarator,
    STATE(1444), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [56986] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3726), 2,
      anon_sym_LBRACK,
      sym_identifier,
    ACTIONS(3812), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    ACTIONS(3814), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    ACTIONS(3817), 3,
      anon_sym_asm,
      anon_sym___asm__,
      anon_sym___asm,
    ACTIONS(3728), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
  [57013] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1440), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [57050] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1444), 1,
      sym__declaration_declarator,
    STATE(1487), 1,
      sym__declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [57087] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    ACTIONS(3417), 1,
      sym_identifier,
    STATE(1300), 1,
      sym_function_declarator,
    STATE(1401), 1,
      sym__declarator,
    STATE(1450), 1,
      sym__declaration_declarator,
    STATE(1511), 1,
      sym__function_declaration_declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1363), 4,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_array_declarator,
  [57124] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3821), 1,
      anon_sym___attribute,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3819), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [57154] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3825), 1,
      anon_sym___attribute,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3823), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [57184] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3829), 1,
      anon_sym___attribute,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3827), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [57214] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3800), 1,
      anon_sym_COLON,
    ACTIONS(3833), 1,
      anon_sym___attribute,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1535), 1,
      sym_bitfield_clause,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3831), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [57248] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(1281), 1,
      sym_parameter_list,
    ACTIONS(3835), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    STATE(1478), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [57282] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(3839), 1,
      anon_sym___attribute,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3837), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [57312] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3843), 1,
      anon_sym___attribute,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3841), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [57341] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3847), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1909), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [57370] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3689), 1,
      anon_sym_LBRACK,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
    STATE(1336), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3687), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
  [57395] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(3857), 1,
      anon_sym___attribute,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3855), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [57424] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3718), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    STATE(1334), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1787), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57455] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3865), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    ACTIONS(3867), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(1553), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [57480] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3869), 1,
      aux_sym_preproc_if_token2,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1821), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57511] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3871), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1830), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [57540] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3689), 3,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
    STATE(1269), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3687), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
  [57563] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3873), 1,
      sym_identifier,
    ACTIONS(3876), 1,
      aux_sym_preproc_if_token1,
    ACTIONS(3882), 1,
      sym_preproc_directive,
    ACTIONS(3885), 1,
      anon_sym_RBRACE,
    ACTIONS(3879), 2,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
    STATE(1937), 2,
      sym_preproc_call,
      sym_enumerator,
    STATE(1332), 3,
      sym_preproc_if_in_enumerator_list,
      sym_preproc_ifdef_in_enumerator_list,
      aux_sym_enumerator_list_repeat1,
  [57592] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3889), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    STATE(1277), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3887), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_COLON,
  [57615] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3891), 1,
      aux_sym_preproc_if_token2,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1905), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57646] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(3895), 1,
      anon_sym___attribute,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3893), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [57675] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3673), 1,
      sym_identifier,
    ACTIONS(3683), 3,
      anon_sym___attribute__,
      anon_sym___attribute,
      anon_sym_LBRACK,
    STATE(1269), 3,
      sym_preproc_call_expression,
      sym_attribute_specifier,
      aux_sym_function_declarator_repeat1,
    ACTIONS(3681), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
  [57698] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1433), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [57727] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(3897), 1,
      anon_sym_COMMA,
    ACTIONS(3901), 1,
      anon_sym___attribute,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1500), 1,
      aux_sym__type_definition_declarators_repeat1,
    ACTIONS(3899), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [57760] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3903), 1,
      aux_sym_preproc_if_token2,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1949), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57791] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3905), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1828), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [57820] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(3909), 1,
      anon_sym___attribute,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3907), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [57849] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3734), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    STATE(1339), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1836), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57880] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3911), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1340), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1838), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [57909] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1416), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [57938] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3913), 1,
      aux_sym_preproc_if_token2,
    STATE(1329), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1853), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [57969] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3915), 1,
      aux_sym_preproc_if_token2,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1799), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [58000] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3917), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1908), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [58029] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1399), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [58058] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1412), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [58087] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(2819), 1,
      sym_identifier,
    ACTIONS(3103), 1,
      anon_sym_LPAREN2,
    ACTIONS(3105), 1,
      anon_sym_STAR,
    STATE(1427), 1,
      sym__declarator,
    STATE(1900), 1,
      sym_ms_based_modifier,
    STATE(1300), 5,
      sym_parenthesized_declarator,
      sym_attributed_declarator,
      sym_pointer_declarator,
      sym_function_declarator,
      sym_array_declarator,
  [58116] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(3859), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3861), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3919), 1,
      aux_sym_preproc_if_token2,
    STATE(1346), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(3863), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1894), 3,
      sym_preproc_else_in_enumerator_list,
      sym_preproc_elif_in_enumerator_list,
      sym_preproc_elifdef_in_enumerator_list,
  [58147] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(1441), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
  [58176] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1819), 1,
      anon_sym___based,
    ACTIONS(3097), 1,
      sym_identifier,
    ACTIONS(3099), 1,
      anon_sym_LPAREN2,
    ACTIONS(3101), 1,
      anon_sym_STAR,
    STATE(1320), 1,
      sym__field_declarator,
    STATE(1990), 1,
      sym_ms_based_modifier,
    STATE(1382), 5,
      sym_parenthesized_field_declarator,
      sym_attributed_field_declarator,
      sym_pointer_field_declarator,
      sym_function_field_declarator,
      sym_array_field_declarator,
  [58205] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(3849), 1,
      aux_sym_preproc_else_token1,
    ACTIONS(3851), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(3921), 1,
      aux_sym_preproc_if_token2,
    ACTIONS(3853), 2,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
    STATE(1324), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1860), 3,
      sym_preproc_else_in_enumerator_list_no_comma,
      sym_preproc_elif_in_enumerator_list_no_comma,
      sym_preproc_elifdef_in_enumerator_list_no_comma,
  [58234] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3762), 1,
      anon_sym___attribute,
    STATE(1281), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3760), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
  [58262] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(339), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58294] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(3929), 1,
      anon_sym___attribute,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3927), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [58322] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3933), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    STATE(1277), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3931), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [58344] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3754), 1,
      anon_sym___attribute,
    STATE(1281), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3752), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
  [58372] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3937), 1,
      anon_sym_LBRACK,
    STATE(1541), 1,
      sym_gnu_asm_output_operand,
    STATE(1972), 1,
      sym_string_literal,
    ACTIONS(3935), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [58396] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(261), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58428] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(124), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58460] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3782), 1,
      anon_sym_LBRACK,
    ACTIONS(3941), 1,
      anon_sym___asm,
    ACTIONS(3780), 4,
      anon_sym_LPAREN2,
      anon_sym_LBRACK_LBRACK,
      anon_sym_LBRACE,
      anon_sym_EQ,
    ACTIONS(3939), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_asm,
      anon_sym___asm__,
  [58482] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(288), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58514] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3945), 1,
      anon_sym___attribute__,
    ACTIONS(3948), 1,
      anon_sym___attribute,
    ACTIONS(3951), 1,
      anon_sym___asm,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
    ACTIONS(3943), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_asm,
      anon_sym___asm__,
  [58538] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3750), 1,
      anon_sym___attribute,
    STATE(1281), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3744), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
  [58566] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3758), 1,
      anon_sym___attribute,
    STATE(1281), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
    ACTIONS(3756), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
  [58594] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3955), 1,
      anon_sym_LBRACK,
    STATE(1539), 1,
      sym_gnu_asm_input_operand,
    STATE(1846), 1,
      sym_string_literal,
    ACTIONS(3953), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [58618] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(3959), 1,
      anon_sym___asm,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
    ACTIONS(3957), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_asm,
      anon_sym___asm__,
  [58641] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(3691), 1,
      anon_sym___asm,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
    ACTIONS(3685), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_asm,
      anon_sym___asm__,
  [58664] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    STATE(295), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58693] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(288), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58722] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3961), 1,
      sym_identifier,
    ACTIONS(3965), 1,
      sym_system_lib_string,
    STATE(2005), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(3963), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [58743] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3969), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3967), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [58760] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1940), 1,
      anon_sym_LPAREN2,
    ACTIONS(1942), 1,
      anon_sym_STAR,
    ACTIONS(2831), 1,
      anon_sym_LBRACK,
    STATE(1452), 1,
      sym_parameter_list,
    STATE(1551), 1,
      sym__abstract_declarator,
    STATE(1443), 4,
      sym_abstract_parenthesized_declarator,
      sym_abstract_pointer_declarator,
      sym_abstract_function_declarator,
      sym_abstract_array_declarator,
  [58785] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(261), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58814] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3971), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [58831] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3977), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3975), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [58848] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    STATE(132), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58877] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(124), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58906] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3865), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1600), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [58927] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3981), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3979), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [58944] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3983), 1,
      sym_identifier,
    ACTIONS(3985), 1,
      sym_system_lib_string,
    STATE(1884), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(3963), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [58965] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    STATE(273), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [58994] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3989), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3987), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [59011] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3993), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3991), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [59028] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(339), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59057] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3995), 1,
      sym_identifier,
    ACTIONS(3997), 1,
      sym_system_lib_string,
    STATE(1774), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(3963), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59078] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3865), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1566), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59099] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3923), 1,
      anon_sym_LPAREN2,
    STATE(326), 1,
      sym_compound_statement,
    STATE(940), 1,
      sym__old_style_parameter_list,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4001), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(3999), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [59145] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4003), 1,
      sym_identifier,
    ACTIONS(4005), 1,
      sym_system_lib_string,
    STATE(1816), 2,
      sym_preproc_call_expression,
      sym_string_literal,
    ACTIONS(3963), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59166] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(3835), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1481), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [59193] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4011), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4009), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [59210] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3865), 1,
      sym_identifier,
    STATE(663), 1,
      sym_string_literal,
    STATE(1648), 2,
      sym__string,
      sym_concatenated_string,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59231] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4015), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4013), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
      anon_sym_COLON,
  [59248] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4019), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4017), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59264] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4023), 1,
      anon_sym___attribute,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4021), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
      anon_sym_COLON,
  [59286] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(139), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59312] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4027), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4025), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59328] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(324), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59354] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4031), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4029), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59370] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(326), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59396] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4035), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4033), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59412] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(271), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59438] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(132), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59464] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4039), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4037), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59480] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4041), 1,
      sym_identifier,
    STATE(733), 1,
      sym_field_declaration_list,
    STATE(1516), 1,
      sym_attribute_specifier,
    STATE(1653), 1,
      sym_ms_declspec_modifier,
    ACTIONS(35), 2,
      anon_sym___attribute__,
      anon_sym___attribute,
  [59506] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4043), 1,
      sym_identifier,
    ACTIONS(4048), 1,
      aux_sym_preproc_elif_token1,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
    ACTIONS(4046), 4,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
  [59528] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4052), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4050), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59544] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4056), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4054), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59560] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(300), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59586] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(295), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59612] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4058), 1,
      sym_identifier,
    ACTIONS(4063), 1,
      aux_sym_preproc_elif_token1,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    ACTIONS(4061), 4,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
  [59632] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4067), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4065), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59648] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(279), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59674] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4071), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4069), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59690] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(129), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59716] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(293), 1,
      sym_compound_statement,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59742] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4075), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(4077), 1,
      anon_sym_EQ,
    ACTIONS(4073), 6,
      anon_sym_COMMA,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [59760] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3937), 1,
      anon_sym_LBRACK,
    STATE(1671), 1,
      sym_gnu_asm_output_operand,
    STATE(1972), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59780] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(273), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59806] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4081), 1,
      anon_sym___attribute,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4079), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
      anon_sym_COLON,
  [59828] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4085), 1,
      anon_sym___attribute,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4083), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym___attribute__,
      anon_sym_COLON,
  [59850] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3955), 1,
      anon_sym_LBRACK,
    STATE(1612), 1,
      sym_gnu_asm_input_operand,
    STATE(1846), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [59870] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4089), 2,
      anon_sym___attribute,
      anon_sym_LBRACK,
    ACTIONS(4087), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym_SEMI,
      anon_sym___attribute__,
      anon_sym_LBRACK_LBRACK,
  [59886] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(334), 1,
      sym_compound_statement,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59912] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4093), 1,
      anon_sym___attribute,
    ACTIONS(4091), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [59927] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4097), 1,
      anon_sym___attribute,
    ACTIONS(4095), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [59942] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(4099), 1,
      anon_sym_RPAREN,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [59965] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4101), 1,
      anon_sym_LPAREN2,
    STATE(1431), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(4103), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [59982] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4108), 1,
      anon_sym___attribute,
    ACTIONS(4106), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [59997] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(4110), 1,
      anon_sym_RPAREN,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60020] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4114), 1,
      anon_sym_SEMI,
    STATE(1614), 1,
      sym_gnu_asm_expression,
    STATE(1615), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60043] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4120), 1,
      anon_sym___attribute,
    ACTIONS(4118), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60058] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4124), 1,
      anon_sym___attribute,
    ACTIONS(4122), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60073] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4126), 1,
      anon_sym_SEMI,
    STATE(1624), 1,
      sym_gnu_asm_expression,
    STATE(1625), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60096] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4128), 1,
      anon_sym_SEMI,
    STATE(1627), 1,
      sym_gnu_asm_expression,
    STATE(1628), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60119] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_SEMI,
    STATE(1633), 1,
      sym_gnu_asm_expression,
    STATE(1634), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60142] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4132), 1,
      anon_sym_SEMI,
    STATE(1636), 1,
      sym_gnu_asm_expression,
    STATE(1637), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60165] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(4134), 1,
      anon_sym_RPAREN,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60188] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(4136), 1,
      anon_sym_RPAREN,
    STATE(1254), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60211] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4140), 1,
      anon_sym___attribute,
    ACTIONS(4138), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60226] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4142), 1,
      anon_sym_SEMI,
    STATE(1650), 1,
      sym_gnu_asm_expression,
    STATE(1651), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60249] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1980), 1,
      anon_sym_LBRACK,
    ACTIONS(4144), 1,
      anon_sym_EQ,
    ACTIONS(4146), 1,
      anon_sym_DOT,
    STATE(1461), 4,
      sym_subscript_designator,
      sym_subscript_range_designator,
      sym_field_designator,
      aux_sym_initializer_pair_repeat1,
  [60268] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3845), 1,
      anon_sym_LBRACK,
    ACTIONS(4148), 1,
      anon_sym_RPAREN,
    STATE(1397), 1,
      sym_parameter_list,
    STATE(1358), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60291] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4150), 1,
      anon_sym_SEMI,
    STATE(1581), 1,
      sym_gnu_asm_expression,
    STATE(1582), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60314] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4152), 1,
      anon_sym_SEMI,
    STATE(1589), 1,
      sym_gnu_asm_expression,
    STATE(1590), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60337] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4156), 1,
      anon_sym___attribute,
    ACTIONS(4154), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60352] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4158), 1,
      anon_sym_SEMI,
    STATE(1662), 1,
      sym_gnu_asm_expression,
    STATE(1663), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60375] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_COMMA,
    ACTIONS(4164), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(4160), 5,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [60392] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4168), 1,
      anon_sym___attribute,
    ACTIONS(4166), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4172), 1,
      anon_sym___attribute,
    ACTIONS(4170), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60422] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4174), 1,
      anon_sym_LPAREN2,
    STATE(1431), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(4176), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [60439] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4178), 1,
      anon_sym_SEMI,
    STATE(1613), 1,
      sym_gnu_asm_expression,
    STATE(1617), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60462] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4180), 1,
      anon_sym_LPAREN2,
    STATE(1454), 2,
      sym_gnu_asm_qualifier,
      aux_sym_gnu_asm_expression_repeat1,
    ACTIONS(4176), 4,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [60479] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60502] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4182), 1,
      anon_sym_SEMI,
    STATE(1661), 1,
      sym_gnu_asm_expression,
    STATE(1664), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60525] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4186), 1,
      anon_sym___attribute,
    ACTIONS(4184), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60540] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3798), 1,
      anon_sym_LBRACK,
    ACTIONS(4188), 1,
      anon_sym_RPAREN,
    STATE(1396), 1,
      sym_parameter_list,
    STATE(1333), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60563] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4190), 1,
      anon_sym_LBRACK,
    ACTIONS(4193), 1,
      anon_sym_EQ,
    ACTIONS(4195), 1,
      anon_sym_DOT,
    STATE(1461), 4,
      sym_subscript_designator,
      sym_subscript_range_designator,
      sym_field_designator,
      aux_sym_initializer_pair_repeat1,
  [60582] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4200), 1,
      anon_sym___attribute,
    ACTIONS(4198), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60597] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4204), 1,
      anon_sym___attribute,
    ACTIONS(4202), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60612] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4208), 1,
      anon_sym___attribute,
    ACTIONS(4206), 6,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LPAREN2,
      anon_sym___attribute__,
      anon_sym_LBRACK,
      anon_sym_COLON,
  [60627] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4210), 1,
      anon_sym_SEMI,
    STATE(1608), 1,
      sym_gnu_asm_expression,
    STATE(1609), 1,
      aux_sym_declaration_repeat1,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
  [60650] = 3,
    ACTIONS(3), 1,
      sym_comment,
    STATE(2012), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [60664] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4212), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4214), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60678] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4216), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4218), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60692] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4220), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [60710] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4222), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [60728] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4224), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4226), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60742] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4228), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4230), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60756] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4232), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4234), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60770] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4236), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [60788] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4238), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4240), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60802] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4048), 1,
      aux_sym_preproc_elif_token1,
    ACTIONS(4046), 5,
      aux_sym_preproc_if_token2,
      aux_sym_preproc_else_token1,
      aux_sym_preproc_elifdef_token1,
      aux_sym_preproc_elifdef_token2,
      sym_identifier,
  [60816] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    STATE(1744), 1,
      sym_gnu_asm_expression,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
    ACTIONS(4242), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [60834] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4244), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [60852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4238), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4240), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60866] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4246), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4248), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60880] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4244), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [60898] = 3,
    ACTIONS(3), 1,
      sym_comment,
    STATE(621), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [60912] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    STATE(1432), 1,
      sym_parameter_list,
    ACTIONS(4250), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [60930] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(93), 1,
      anon_sym___asm,
    STATE(1701), 1,
      sym_gnu_asm_expression,
    ACTIONS(4116), 2,
      anon_sym_asm,
      anon_sym___asm__,
    ACTIONS(4252), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [60948] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3885), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4254), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60962] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4256), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4258), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [60976] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(37), 1,
      anon_sym_LBRACK_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(3748), 1,
      anon_sym_LBRACK,
    STATE(1260), 1,
      sym_parameter_list,
    STATE(1289), 2,
      sym_attribute_declaration,
      aux_sym_attributed_declarator_repeat1,
  [60996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4232), 2,
      anon_sym_RBRACE,
      sym_identifier,
    ACTIONS(4234), 4,
      aux_sym_preproc_if_token1,
      aux_sym_preproc_ifdef_token1,
      aux_sym_preproc_ifdef_token2,
      sym_preproc_directive,
  [61010] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4260), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61028] = 3,
    ACTIONS(3), 1,
      sym_comment,
    STATE(1786), 1,
      sym_string_literal,
    ACTIONS(99), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [61042] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4262), 1,
      anon_sym_COMMA,
    ACTIONS(4267), 1,
      anon_sym___attribute,
    STATE(1491), 1,
      aux_sym__type_definition_declarators_repeat1,
    ACTIONS(4265), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61059] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4269), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61076] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4271), 1,
      anon_sym_SEMI,
    STATE(1501), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61093] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(4273), 1,
      aux_sym_preproc_if_token2,
    STATE(1451), 1,
      sym_enumerator,
    STATE(1523), 1,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
    STATE(1560), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
  [61112] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4275), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61129] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4277), 1,
      anon_sym_SEMI,
    STATE(1497), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61146] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4279), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61163] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4281), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61180] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4283), 5,
      anon_sym_LPAREN2,
      anon_sym_inline,
      anon_sym_volatile,
      anon_sym_goto,
      anon_sym___volatile__,
  [61191] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3897), 1,
      anon_sym_COMMA,
    ACTIONS(4287), 1,
      anon_sym___attribute,
    STATE(1491), 1,
      aux_sym__type_definition_declarators_repeat1,
    ACTIONS(4285), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61208] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4289), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61225] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4291), 1,
      anon_sym_SEMI,
    STATE(1495), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61242] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4293), 1,
      anon_sym_SEMI,
    STATE(1498), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61259] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3792), 1,
      anon_sym_COMMA,
    ACTIONS(4297), 1,
      anon_sym___attribute,
    STATE(1509), 1,
      aux_sym__field_declaration_declarator_repeat1,
    ACTIONS(4295), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61276] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4299), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61293] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4301), 1,
      sym_identifier,
    STATE(724), 1,
      sym_field_declaration_list,
    STATE(1646), 1,
      sym_ms_declspec_modifier,
  [61312] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4303), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61329] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4305), 1,
      anon_sym_SEMI,
    STATE(1505), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61346] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4307), 1,
      anon_sym_COMMA,
    ACTIONS(4312), 1,
      anon_sym___attribute,
    STATE(1509), 1,
      aux_sym__field_declaration_declarator_repeat1,
    ACTIONS(4310), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61363] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4314), 1,
      anon_sym_SEMI,
    STATE(1517), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61380] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4318), 1,
      anon_sym___asm,
    ACTIONS(4316), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_asm,
      anon_sym___asm__,
  [61393] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    ACTIONS(4322), 1,
      anon_sym_COLON_COLON,
    STATE(1699), 1,
      sym_argument_list,
    ACTIONS(4320), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK_RBRACK,
  [61410] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4324), 1,
      anon_sym_SEMI,
    STATE(1492), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61427] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3792), 1,
      anon_sym_COMMA,
    ACTIONS(4328), 1,
      anon_sym___attribute,
    STATE(1509), 1,
      aux_sym__field_declaration_declarator_repeat1,
    ACTIONS(4326), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61444] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4330), 1,
      anon_sym_SEMI,
    STATE(1507), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61461] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      anon_sym___declspec,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4332), 1,
      sym_identifier,
    STATE(729), 1,
      sym_field_declaration_list,
    STATE(1670), 1,
      sym_ms_declspec_modifier,
  [61480] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4334), 1,
      anon_sym_SEMI,
    STATE(1365), 2,
      sym_attribute_specifier,
      aux_sym_type_definition_repeat1,
  [61497] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3792), 1,
      anon_sym_COMMA,
    ACTIONS(4338), 1,
      anon_sym___attribute,
    STATE(1504), 1,
      aux_sym__field_declaration_declarator_repeat1,
    ACTIONS(4336), 2,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61514] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4340), 1,
      anon_sym_SQUOTE,
    STATE(1530), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4342), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [61528] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4344), 1,
      anon_sym_DQUOTE,
    ACTIONS(4346), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4348), 1,
      sym_escape_sequence,
    STATE(1554), 1,
      aux_sym_string_literal_repeat1,
  [61544] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4350), 1,
      anon_sym_SEMI,
    STATE(1897), 1,
      sym_attribute_specifier,
  [61560] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4352), 1,
      anon_sym_SQUOTE,
    STATE(1530), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4342), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [61574] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(4354), 1,
      aux_sym_preproc_if_token2,
    STATE(1414), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
  [61588] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4356), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4360), 1,
      sym_preproc_arg,
    STATE(1705), 1,
      sym_preproc_params,
  [61604] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4362), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4364), 1,
      sym_preproc_arg,
    STATE(1687), 1,
      sym_preproc_params,
  [61620] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3716), 1,
      sym_identifier,
    ACTIONS(4366), 1,
      aux_sym_preproc_if_token2,
    STATE(1523), 2,
      sym_enumerator,
      aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1,
  [61634] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4368), 1,
      anon_sym_RPAREN,
    STATE(1432), 1,
      sym_parameter_list,
  [61650] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4370), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4372), 1,
      sym_preproc_arg,
    STATE(1684), 1,
      sym_preproc_params,
  [61666] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4374), 1,
      anon_sym_COMMA,
    STATE(1529), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(4377), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61680] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4379), 1,
      anon_sym_SQUOTE,
    STATE(1530), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4381), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [61694] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4384), 1,
      anon_sym_DQUOTE,
    ACTIONS(4386), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4388), 1,
      sym_escape_sequence,
    STATE(1547), 1,
      aux_sym_string_literal_repeat1,
  [61710] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4390), 1,
      anon_sym_DQUOTE,
    ACTIONS(4392), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4394), 1,
      sym_escape_sequence,
    STATE(1520), 1,
      aux_sym_string_literal_repeat1,
  [61726] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4396), 1,
      anon_sym_COMMA,
    STATE(1540), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(4398), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61740] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4400), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4402), 1,
      sym_preproc_arg,
    STATE(1710), 1,
      sym_preproc_params,
  [61756] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4406), 1,
      anon_sym___attribute,
    ACTIONS(4404), 3,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym___attribute__,
  [61768] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4408), 1,
      anon_sym_SEMI,
    STATE(1873), 1,
      sym_attribute_specifier,
  [61784] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4410), 1,
      anon_sym_COMMA,
    ACTIONS(4412), 1,
      anon_sym_RPAREN,
    STATE(1594), 1,
      aux_sym__old_style_parameter_list_repeat1,
    STATE(1601), 1,
      aux_sym_parameter_list_repeat1,
  [61800] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4346), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4348), 1,
      sym_escape_sequence,
    ACTIONS(4414), 1,
      anon_sym_DQUOTE,
    STATE(1554), 1,
      aux_sym_string_literal_repeat1,
  [61816] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4416), 1,
      anon_sym_COMMA,
    STATE(1552), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(4418), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61830] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4420), 1,
      anon_sym_COMMA,
    STATE(1540), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(4423), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61844] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4425), 1,
      anon_sym_COMMA,
    STATE(1555), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(4427), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61858] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4429), 1,
      anon_sym___except,
    ACTIONS(4431), 1,
      anon_sym___finally,
    STATE(249), 2,
      sym_seh_except_clause,
      sym_seh_finally_clause,
  [61872] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4433), 1,
      anon_sym___except,
    ACTIONS(4435), 1,
      anon_sym___finally,
    STATE(196), 2,
      sym_seh_except_clause,
      sym_seh_finally_clause,
  [61886] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4437), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4439), 1,
      sym_preproc_arg,
    STATE(1717), 1,
      sym_preproc_params,
  [61902] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    STATE(1735), 1,
      sym_argument_list,
    ACTIONS(4441), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK_RBRACK,
  [61916] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(4273), 1,
      aux_sym_preproc_if_token2,
    STATE(1560), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
  [61932] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4346), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4348), 1,
      sym_escape_sequence,
    ACTIONS(4443), 1,
      anon_sym_DQUOTE,
    STATE(1554), 1,
      aux_sym_string_literal_repeat1,
  [61948] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4445), 1,
      anon_sym___except,
    ACTIONS(4447), 1,
      anon_sym___finally,
    STATE(249), 2,
      sym_seh_except_clause,
      sym_seh_finally_clause,
  [61962] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4449), 1,
      anon_sym_COMMA,
    STATE(1549), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(4452), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [61976] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4454), 1,
      anon_sym___except,
    ACTIONS(4456), 1,
      anon_sym___finally,
    STATE(106), 2,
      sym_seh_except_clause,
      sym_seh_finally_clause,
  [61990] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3746), 1,
      anon_sym_LPAREN2,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4458), 1,
      anon_sym_RPAREN,
    STATE(1432), 1,
      sym_parameter_list,
  [62006] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4416), 1,
      anon_sym_COMMA,
    STATE(1529), 1,
      aux_sym_gnu_asm_input_operand_list_repeat1,
    ACTIONS(4460), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62020] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4396), 1,
      anon_sym_COMMA,
    STATE(1533), 1,
      aux_sym_gnu_asm_clobber_list_repeat1,
    ACTIONS(4462), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62034] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4464), 1,
      anon_sym_DQUOTE,
    ACTIONS(4466), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4469), 1,
      sym_escape_sequence,
    STATE(1554), 1,
      aux_sym_string_literal_repeat1,
  [62050] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4425), 1,
      anon_sym_COMMA,
    STATE(1549), 1,
      aux_sym_gnu_asm_output_operand_list_repeat1,
    ACTIONS(4472), 2,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62064] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4474), 1,
      anon_sym___except,
    ACTIONS(4476), 1,
      anon_sym___finally,
    STATE(223), 2,
      sym_seh_except_clause,
      sym_seh_finally_clause,
  [62078] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4478), 1,
      anon_sym_DQUOTE,
    ACTIONS(4480), 1,
      aux_sym_string_literal_token1,
    ACTIONS(4482), 1,
      sym_escape_sequence,
    STATE(1538), 1,
      aux_sym_string_literal_repeat1,
  [62094] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4484), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4486), 1,
      sym_preproc_arg,
    STATE(1696), 1,
      sym_preproc_params,
  [62110] = 5,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4358), 1,
      anon_sym_LPAREN,
    ACTIONS(4488), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4490), 1,
      sym_preproc_arg,
    STATE(1692), 1,
      sym_preproc_params,
  [62126] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3764), 1,
      sym_identifier,
    ACTIONS(4492), 1,
      aux_sym_preproc_if_token2,
    STATE(1409), 1,
      aux_sym_preproc_if_in_enumerator_list_repeat1,
    STATE(1802), 1,
      sym_enumerator,
  [62142] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(35), 1,
      anon_sym___attribute,
    ACTIONS(3504), 1,
      anon_sym___attribute__,
    ACTIONS(4494), 1,
      anon_sym_SEMI,
    STATE(1991), 1,
      sym_attribute_specifier,
  [62158] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4496), 1,
      anon_sym_SQUOTE,
    STATE(1530), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4342), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [62172] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4498), 1,
      anon_sym_RPAREN,
    ACTIONS(4500), 1,
      anon_sym_COLON,
    STATE(1896), 1,
      sym_gnu_asm_goto_list,
  [62185] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(4502), 1,
      sym_identifier,
    STATE(1050), 1,
      sym_enumerator_list,
  [62198] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4504), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62207] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4506), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62216] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4508), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4510), 1,
      anon_sym_LPAREN2,
    STATE(1796), 1,
      sym_preproc_argument_list,
  [62229] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3211), 1,
      anon_sym_COMMA,
    ACTIONS(4512), 1,
      anon_sym_RPAREN,
    STATE(1672), 1,
      aux_sym_generic_expression_repeat1,
  [62242] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4500), 1,
      anon_sym_COLON,
    ACTIONS(4514), 1,
      anon_sym_RPAREN,
    STATE(2000), 1,
      sym_gnu_asm_goto_list,
  [62255] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4518), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1579), 1,
      aux_sym_attribute_declaration_repeat1,
  [62268] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3573), 1,
      anon_sym_RPAREN,
    ACTIONS(4520), 1,
      anon_sym_COMMA,
    STATE(1571), 1,
      aux_sym_preproc_argument_list_repeat1,
  [62281] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(4523), 1,
      anon_sym_RPAREN,
    STATE(1668), 1,
      aux_sym_argument_list_repeat1,
  [62294] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4525), 1,
      anon_sym_RPAREN,
    ACTIONS(4527), 1,
      anon_sym_COLON,
    STATE(1563), 1,
      sym_gnu_asm_clobber_list,
  [62307] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4529), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1604), 1,
      aux_sym_attribute_declaration_repeat1,
  [62320] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4531), 1,
      anon_sym_SEMI,
    STATE(1583), 1,
      aux_sym_declaration_repeat1,
  [62333] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4533), 1,
      anon_sym_COMMA,
    ACTIONS(4536), 1,
      anon_sym_RPAREN,
    STATE(1576), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [62346] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4538), 1,
      anon_sym_COMMA,
    ACTIONS(4541), 1,
      anon_sym_RPAREN,
    STATE(1577), 1,
      aux_sym_preproc_params_repeat1,
  [62359] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(4543), 1,
      anon_sym_RPAREN,
    STATE(1571), 1,
      aux_sym_preproc_argument_list_repeat1,
  [62372] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4545), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1607), 1,
      aux_sym_attribute_declaration_repeat1,
  [62385] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4547), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1607), 1,
      aux_sym_attribute_declaration_repeat1,
  [62398] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4549), 1,
      anon_sym_SEMI,
    STATE(1591), 1,
      aux_sym_declaration_repeat1,
  [62411] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4551), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62424] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4553), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62437] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4555), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [62446] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4557), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1580), 1,
      aux_sym_attribute_declaration_repeat1,
  [62459] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4559), 1,
      anon_sym_COMMA,
    ACTIONS(4561), 1,
      anon_sym_RPAREN,
    STATE(1577), 1,
      aux_sym_preproc_params_repeat1,
  [62472] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4563), 1,
      anon_sym_COMMA,
    ACTIONS(4565), 1,
      anon_sym_RPAREN,
    STATE(1603), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [62485] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4567), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62494] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4569), 1,
      anon_sym_SEMI,
    STATE(1595), 1,
      aux_sym_declaration_repeat1,
  [62507] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4571), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62520] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4573), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62533] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4575), 1,
      anon_sym_SEMI,
    STATE(1619), 1,
      aux_sym_declaration_repeat1,
  [62546] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(4577), 1,
      anon_sym_RPAREN,
    STATE(1571), 1,
      aux_sym_preproc_argument_list_repeat1,
  [62559] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4579), 1,
      anon_sym_COMMA,
    ACTIONS(4581), 1,
      anon_sym_RPAREN,
    STATE(1576), 1,
      aux_sym__old_style_parameter_list_repeat1,
  [62572] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4583), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62585] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4585), 1,
      anon_sym_RPAREN,
    ACTIONS(4587), 1,
      anon_sym_COLON,
    STATE(1573), 1,
      sym_gnu_asm_input_operand_list,
  [62598] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3265), 1,
      anon_sym_RBRACE,
    ACTIONS(4589), 1,
      anon_sym_COMMA,
    STATE(1597), 1,
      aux_sym_initializer_list_repeat1,
  [62611] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4592), 1,
      anon_sym_COMMA,
    ACTIONS(4594), 1,
      anon_sym_RPAREN,
    STATE(1601), 1,
      aux_sym_parameter_list_repeat1,
  [62624] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4510), 1,
      anon_sym_LPAREN2,
    ACTIONS(4596), 1,
      aux_sym_preproc_include_token2,
    STATE(1796), 1,
      sym_preproc_argument_list,
  [62637] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4598), 1,
      anon_sym_RPAREN,
    ACTIONS(4600), 1,
      anon_sym_COLON,
    STATE(1647), 1,
      sym_gnu_asm_output_operand_list,
  [62650] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4592), 1,
      anon_sym_COMMA,
    ACTIONS(4602), 1,
      anon_sym_RPAREN,
    STATE(1669), 1,
      aux_sym_parameter_list_repeat1,
  [62663] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4510), 1,
      anon_sym_LPAREN2,
    ACTIONS(4604), 1,
      aux_sym_preproc_include_token2,
    STATE(1796), 1,
      sym_preproc_argument_list,
  [62676] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4606), 1,
      anon_sym_COMMA,
    ACTIONS(4609), 1,
      anon_sym_RPAREN,
    STATE(1603), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [62689] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4516), 1,
      anon_sym_COMMA,
    ACTIONS(4611), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1607), 1,
      aux_sym_attribute_declaration_repeat1,
  [62702] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4613), 1,
      anon_sym_SEMI,
    STATE(1611), 1,
      aux_sym_declaration_repeat1,
  [62715] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4615), 1,
      anon_sym_SEMI,
    STATE(1665), 1,
      aux_sym_declaration_repeat1,
  [62728] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4617), 1,
      anon_sym_COMMA,
    ACTIONS(4620), 1,
      anon_sym_RBRACK_RBRACK,
    STATE(1607), 1,
      aux_sym_attribute_declaration_repeat1,
  [62741] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4622), 1,
      anon_sym_SEMI,
    STATE(1616), 1,
      aux_sym_declaration_repeat1,
  [62754] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4624), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62767] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4626), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62776] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4628), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62789] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4630), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62798] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4632), 1,
      anon_sym_SEMI,
    STATE(1652), 1,
      aux_sym_declaration_repeat1,
  [62811] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4634), 1,
      anon_sym_SEMI,
    STATE(1620), 1,
      aux_sym_declaration_repeat1,
  [62824] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4636), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62837] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4638), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62850] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4640), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62863] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    STATE(1562), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4642), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [62874] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4644), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62887] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4646), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62900] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(3217), 1,
      anon_sym_RPAREN,
    STATE(1654), 1,
      aux_sym_argument_list_repeat1,
  [62913] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4648), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [62922] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4653), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62935] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4655), 1,
      anon_sym_SEMI,
    STATE(1629), 1,
      aux_sym_declaration_repeat1,
  [62948] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4657), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62961] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4659), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [62974] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4661), 1,
      anon_sym_SEMI,
    STATE(1630), 1,
      aux_sym_declaration_repeat1,
  [62987] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4663), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63000] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4665), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63013] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4667), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63026] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4559), 1,
      anon_sym_COMMA,
    ACTIONS(4669), 1,
      anon_sym_RPAREN,
    STATE(1586), 1,
      aux_sym_preproc_params_repeat1,
  [63039] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4671), 1,
      anon_sym_SEMI,
    STATE(1635), 1,
      aux_sym_declaration_repeat1,
  [63052] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4673), 1,
      anon_sym_SEMI,
    STATE(1638), 1,
      aux_sym_declaration_repeat1,
  [63065] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4675), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63078] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4677), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63091] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4679), 1,
      anon_sym_SEMI,
    STATE(1639), 1,
      aux_sym_declaration_repeat1,
  [63104] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4681), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63117] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4683), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63130] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4685), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63143] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4687), 1,
      anon_sym_EQ,
    ACTIONS(4073), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [63154] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    STATE(1519), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4689), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [63165] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4691), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [63174] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3207), 1,
      anon_sym_COMMA,
    ACTIONS(3209), 1,
      anon_sym_RBRACE,
    STATE(1660), 1,
      aux_sym_initializer_list_repeat1,
  [63187] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(4693), 1,
      sym_identifier,
    STATE(1764), 1,
      sym_variadic_parameter,
  [63200] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4563), 1,
      anon_sym_COMMA,
    ACTIONS(4695), 1,
      anon_sym_RPAREN,
    STATE(1587), 1,
      aux_sym_gnu_asm_goto_list_repeat1,
  [63213] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4697), 1,
      sym_identifier,
    STATE(731), 1,
      sym_field_declaration_list,
  [63226] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4587), 1,
      anon_sym_COLON,
    ACTIONS(4699), 1,
      anon_sym_RPAREN,
    STATE(1676), 1,
      sym_gnu_asm_input_operand_list,
  [63239] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4600), 1,
      anon_sym_COLON,
    ACTIONS(4701), 1,
      anon_sym_RPAREN,
    STATE(1596), 1,
      sym_gnu_asm_output_operand_list,
  [63252] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4705), 1,
      anon_sym_RPAREN,
    ACTIONS(4703), 2,
      anon_sym_DOT_DOT_DOT,
      sym_identifier,
  [63263] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4707), 1,
      anon_sym_SEMI,
    STATE(1658), 1,
      aux_sym_declaration_repeat1,
  [63276] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4709), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63289] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4711), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63302] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4713), 1,
      sym_identifier,
    STATE(729), 1,
      sym_field_declaration_list,
  [63315] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(4715), 1,
      anon_sym_RPAREN,
    STATE(1668), 1,
      aux_sym_argument_list_repeat1,
  [63328] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3464), 1,
      anon_sym_COMMA,
    ACTIONS(4717), 1,
      anon_sym_RPAREN,
    STATE(1571), 1,
      aux_sym_preproc_argument_list_repeat1,
  [63341] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3199), 3,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_COLON,
  [63350] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4719), 3,
      anon_sym_LBRACK,
      anon_sym_EQ,
      anon_sym_DOT,
  [63359] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4721), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63372] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    STATE(1522), 1,
      aux_sym_char_literal_repeat1,
    ACTIONS(4723), 2,
      aux_sym_char_literal_token1,
      sym_escape_sequence,
  [63383] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2002), 1,
      anon_sym_RBRACE,
    ACTIONS(4725), 1,
      anon_sym_COMMA,
    STATE(1597), 1,
      aux_sym_initializer_list_repeat1,
  [63396] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4727), 1,
      anon_sym_SEMI,
    STATE(1677), 1,
      aux_sym_declaration_repeat1,
  [63409] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4729), 1,
      anon_sym_SEMI,
    STATE(1673), 1,
      aux_sym_declaration_repeat1,
  [63422] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4731), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63435] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4733), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63448] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4735), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63461] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(4737), 1,
      sym_identifier,
    STATE(1050), 1,
      sym_enumerator_list,
  [63474] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3215), 1,
      anon_sym_COMMA,
    ACTIONS(3219), 1,
      anon_sym_RPAREN,
    STATE(1572), 1,
      aux_sym_argument_list_repeat1,
  [63487] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3253), 1,
      anon_sym_RPAREN,
    ACTIONS(4739), 1,
      anon_sym_COMMA,
    STATE(1668), 1,
      aux_sym_argument_list_repeat1,
  [63500] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4742), 1,
      anon_sym_COMMA,
    ACTIONS(4745), 1,
      anon_sym_RPAREN,
    STATE(1669), 1,
      aux_sym_parameter_list_repeat1,
  [63513] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 1,
      anon_sym_LBRACE,
    ACTIONS(4747), 1,
      sym_identifier,
    STATE(726), 1,
      sym_field_declaration_list,
  [63526] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4749), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [63535] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4751), 1,
      anon_sym_COMMA,
    ACTIONS(4754), 1,
      anon_sym_RPAREN,
    STATE(1672), 1,
      aux_sym_generic_expression_repeat1,
  [63548] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4756), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63561] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2954), 1,
      anon_sym_LBRACE,
    ACTIONS(4758), 1,
      sym_identifier,
    STATE(848), 1,
      sym_enumerator_list,
  [63574] = 4,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4510), 1,
      anon_sym_LPAREN2,
    ACTIONS(4760), 1,
      aux_sym_preproc_include_token2,
    STATE(1796), 1,
      sym_preproc_argument_list,
  [63587] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4527), 1,
      anon_sym_COLON,
    ACTIONS(4762), 1,
      anon_sym_RPAREN,
    STATE(1569), 1,
      sym_gnu_asm_clobber_list,
  [63600] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4764), 1,
      anon_sym_SEMI,
    STATE(1623), 1,
      aux_sym_declaration_repeat1,
  [63613] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4112), 1,
      anon_sym_COMMA,
    ACTIONS(4766), 1,
      anon_sym_SEMI,
    STATE(1626), 1,
      aux_sym_declaration_repeat1,
  [63626] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4768), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4770), 1,
      sym_preproc_arg,
  [63636] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4772), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4774), 1,
      sym_preproc_arg,
  [63646] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(340), 1,
      sym_parenthesized_expression,
  [63656] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4778), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4780), 1,
      sym_preproc_arg,
  [63666] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3265), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [63674] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4782), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4784), 1,
      sym_preproc_arg,
  [63684] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4786), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4788), 1,
      sym_preproc_arg,
  [63694] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3287), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [63702] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4790), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4792), 1,
      sym_preproc_arg,
  [63712] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4794), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [63720] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    STATE(1875), 1,
      sym_argument_list,
  [63730] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    STATE(188), 1,
      sym_compound_statement,
  [63740] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(348), 1,
      sym_parenthesized_expression,
  [63750] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4796), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4798), 1,
      sym_preproc_arg,
  [63760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1762), 1,
      sym_parenthesized_expression,
  [63770] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4620), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK_RBRACK,
  [63778] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4802), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [63786] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4805), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4807), 1,
      sym_preproc_arg,
  [63796] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1736), 1,
      sym_parenthesized_expression,
  [63806] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym_compound_statement,
  [63816] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4809), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK_RBRACK,
  [63824] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    STATE(84), 1,
      sym_compound_statement,
  [63834] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4811), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [63842] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4813), 1,
      sym_identifier,
    STATE(1694), 1,
      sym_attribute,
  [63852] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    STATE(100), 1,
      sym_compound_statement,
  [63862] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4815), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4817), 1,
      sym_preproc_arg,
  [63872] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4819), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4821), 1,
      sym_preproc_arg,
  [63882] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    STATE(202), 1,
      sym_compound_statement,
  [63892] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4823), 2,
      anon_sym_DOT_DOT_DOT,
      sym_identifier,
  [63900] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    STATE(179), 1,
      sym_compound_statement,
  [63910] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(430), 1,
      anon_sym_LBRACE,
    STATE(207), 1,
      sym_compound_statement,
  [63920] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4825), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4827), 1,
      sym_preproc_arg,
  [63930] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(349), 1,
      sym_parenthesized_expression,
  [63940] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3291), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [63948] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2445), 1,
      anon_sym_LPAREN2,
    STATE(1781), 1,
      sym_argument_list,
  [63958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1906), 1,
      sym_parenthesized_expression,
  [63968] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4829), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4831), 1,
      sym_preproc_arg,
  [63978] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4833), 1,
      sym_identifier,
    ACTIONS(4835), 1,
      anon_sym_LPAREN2,
  [63988] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4837), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4839), 1,
      sym_preproc_arg,
  [63998] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3806), 1,
      anon_sym_RBRACE,
    ACTIONS(4841), 1,
      anon_sym_COMMA,
  [64008] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4813), 1,
      sym_identifier,
    STATE(1574), 1,
      sym_attribute,
  [64018] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(343), 1,
      sym_parenthesized_expression,
  [64028] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1708), 1,
      sym_parenthesized_expression,
  [64038] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4813), 1,
      sym_identifier,
    STATE(1585), 1,
      sym_attribute,
  [64048] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4843), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4845), 1,
      sym_preproc_arg,
  [64058] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(1548), 1,
      sym_compound_statement,
  [64068] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(1556), 1,
      sym_compound_statement,
  [64078] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4847), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4849), 1,
      sym_preproc_arg,
  [64088] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4851), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4853), 1,
      sym_preproc_arg,
  [64098] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(153), 1,
      sym_compound_statement,
  [64108] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1857), 1,
      sym_parenthesized_expression,
  [64118] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(233), 1,
      sym_compound_statement,
  [64128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1737), 1,
      sym_parenthesized_expression,
  [64138] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4855), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [64146] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(316), 1,
      sym_parenthesized_expression,
  [64156] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4841), 1,
      anon_sym_COMMA,
    ACTIONS(4857), 1,
      anon_sym_RBRACE,
  [64166] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4859), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK_RBRACK,
  [64174] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(378), 1,
      anon_sym_LBRACE,
    STATE(251), 1,
      sym_compound_statement,
  [64184] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(251), 1,
      sym_compound_statement,
  [64194] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4813), 1,
      sym_identifier,
    STATE(1570), 1,
      sym_attribute,
  [64204] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(357), 1,
      sym_parenthesized_expression,
  [64214] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1747), 1,
      sym_parenthesized_expression,
  [64224] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(358), 1,
      sym_parenthesized_expression,
  [64234] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(1550), 1,
      sym_compound_statement,
  [64244] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4861), 1,
      sym_identifier,
    ACTIONS(4863), 1,
      anon_sym_LPAREN2,
  [64254] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4252), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [64262] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1883), 1,
      sym_parenthesized_expression,
  [64272] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1760), 1,
      sym_parenthesized_expression,
  [64282] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(133), 1,
      anon_sym_LBRACE,
    STATE(86), 1,
      sym_compound_statement,
  [64292] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4745), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [64300] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1690), 1,
      sym_parenthesized_expression,
  [64310] = 3,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4865), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(4867), 1,
      sym_preproc_arg,
  [64320] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(1543), 1,
      sym_compound_statement,
  [64330] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1955), 1,
      sym_parenthesized_expression,
  [64340] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1703), 1,
      sym_parenthesized_expression,
  [64350] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3269), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [64358] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(362), 1,
      sym_parenthesized_expression,
  [64368] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_LPAREN2,
    STATE(1709), 1,
      sym_parenthesized_expression,
  [64378] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(366), 1,
      sym_parenthesized_expression,
  [64388] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3253), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [64396] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4541), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [64404] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(241), 1,
      sym_compound_statement,
  [64414] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(1542), 1,
      sym_compound_statement,
  [64424] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(43), 1,
      anon_sym_LBRACE,
    STATE(172), 1,
      sym_compound_statement,
  [64434] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4869), 1,
      sym_identifier,
    ACTIONS(4871), 1,
      anon_sym_RPAREN,
  [64444] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4536), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [64452] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4242), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [64460] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
    STATE(344), 1,
      sym_parenthesized_expression,
  [64470] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4873), 1,
      aux_sym_preproc_if_token2,
  [64477] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4875), 1,
      anon_sym_RPAREN,
  [64484] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4877), 1,
      aux_sym_preproc_include_token2,
  [64491] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3273), 1,
      anon_sym_SEMI,
  [64498] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_RPAREN,
  [64505] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3271), 1,
      anon_sym_RPAREN,
  [64512] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4879), 1,
      aux_sym_preproc_include_token2,
  [64519] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4596), 1,
      aux_sym_preproc_include_token2,
  [64526] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4881), 1,
      anon_sym_RPAREN,
  [64533] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3335), 1,
      anon_sym_RPAREN,
  [64540] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4883), 1,
      anon_sym_RPAREN,
  [64547] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4885), 1,
      aux_sym_preproc_include_token2,
  [64554] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4887), 1,
      anon_sym_RPAREN,
  [64561] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4889), 1,
      aux_sym_preproc_include_token2,
  [64568] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4891), 1,
      anon_sym_RPAREN,
  [64575] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4893), 1,
      anon_sym_RPAREN,
  [64582] = 2,
    ACTIONS(3191), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [64589] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4895), 1,
      aux_sym_preproc_include_token2,
  [64596] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4897), 1,
      aux_sym_preproc_include_token2,
  [64603] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4899), 1,
      anon_sym_LPAREN2,
  [64610] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4901), 1,
      aux_sym_preproc_if_token2,
  [64617] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4903), 1,
      aux_sym_preproc_if_token2,
  [64624] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4905), 1,
      aux_sym_preproc_include_token2,
  [64631] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4907), 1,
      sym_identifier,
  [64638] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4909), 1,
      aux_sym_preproc_if_token2,
  [64645] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4911), 1,
      aux_sym_preproc_include_token2,
  [64652] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4913), 1,
      anon_sym_RPAREN,
  [64659] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4915), 1,
      sym_identifier,
  [64666] = 2,
    ACTIONS(3167), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [64673] = 2,
    ACTIONS(3181), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [64680] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4917), 1,
      sym_identifier,
  [64687] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4919), 1,
      aux_sym_preproc_include_token2,
  [64694] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4921), 1,
      aux_sym_preproc_if_token2,
  [64701] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3289), 1,
      anon_sym_SEMI,
  [64708] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4923), 1,
      anon_sym_SEMI,
  [64715] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4162), 1,
      anon_sym_COMMA,
  [64722] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4925), 1,
      aux_sym_preproc_if_token2,
  [64729] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3261), 1,
      anon_sym_SEMI,
  [64736] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4927), 1,
      aux_sym_preproc_include_token2,
  [64743] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4929), 1,
      aux_sym_preproc_include_token2,
  [64750] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3305), 1,
      anon_sym_SEMI,
  [64757] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4931), 1,
      aux_sym_preproc_include_token2,
  [64764] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3317), 1,
      anon_sym_COLON,
  [64771] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4933), 1,
      sym_identifier,
  [64778] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4935), 1,
      sym_identifier,
  [64785] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4937), 1,
      anon_sym_COMMA,
  [64792] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4939), 1,
      anon_sym_RPAREN,
  [64799] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3301), 1,
      anon_sym_SEMI,
  [64806] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4941), 1,
      sym_identifier,
  [64813] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4604), 1,
      aux_sym_preproc_include_token2,
  [64820] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3263), 1,
      anon_sym_SEMI,
  [64827] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4943), 1,
      anon_sym_SEMI,
  [64834] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4945), 1,
      sym_identifier,
  [64841] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4947), 1,
      ts_builtin_sym_end,
  [64848] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4949), 1,
      aux_sym_preproc_if_token2,
  [64855] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4951), 1,
      aux_sym_preproc_if_token2,
  [64862] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3806), 1,
      anon_sym_RBRACE,
  [64869] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4953), 1,
      anon_sym_SEMI,
  [64876] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4955), 1,
      anon_sym_RPAREN,
  [64883] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4957), 1,
      aux_sym_preproc_include_token2,
  [64890] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4959), 1,
      sym_primitive_type,
  [64897] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4961), 1,
      aux_sym_preproc_if_token2,
  [64904] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4963), 1,
      anon_sym_RBRACE,
  [64911] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4965), 1,
      aux_sym_preproc_if_token2,
  [64918] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3315), 1,
      anon_sym_COLON,
  [64925] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4967), 1,
      sym_identifier,
  [64932] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3209), 1,
      anon_sym_RBRACE,
  [64939] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4969), 1,
      anon_sym_RPAREN,
  [64946] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4971), 1,
      aux_sym_preproc_include_token2,
  [64953] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4973), 1,
      aux_sym_preproc_if_token2,
  [64960] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4975), 1,
      anon_sym_RPAREN,
  [64967] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4977), 1,
      aux_sym_preproc_if_token2,
  [64974] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4979), 1,
      sym_identifier,
  [64981] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4981), 1,
      aux_sym_preproc_if_token2,
  [64988] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4983), 1,
      sym_identifier,
  [64995] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4985), 1,
      aux_sym_preproc_if_token2,
  [65002] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3249), 1,
      anon_sym_RPAREN,
  [65009] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4987), 1,
      aux_sym_preproc_if_token2,
  [65016] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4989), 1,
      sym_identifier,
  [65023] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4991), 1,
      anon_sym_LPAREN2,
  [65030] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4993), 1,
      aux_sym_preproc_include_token2,
  [65037] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4995), 1,
      anon_sym_LPAREN2,
  [65044] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3311), 1,
      anon_sym_SEMI,
  [65051] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4997), 1,
      aux_sym_preproc_if_token2,
  [65058] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3319), 1,
      anon_sym_SEMI,
  [65065] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4999), 1,
      anon_sym_SEMI,
  [65072] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5001), 1,
      aux_sym_preproc_if_token2,
  [65079] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5003), 1,
      sym_identifier,
  [65086] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5005), 1,
      anon_sym_SEMI,
  [65093] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5007), 1,
      anon_sym_RPAREN,
  [65100] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5009), 1,
      anon_sym_SEMI,
  [65107] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5011), 1,
      anon_sym_SEMI,
  [65114] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5013), 1,
      sym_identifier,
  [65121] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5015), 1,
      aux_sym_preproc_if_token2,
  [65128] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5017), 1,
      anon_sym_RPAREN,
  [65135] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5019), 1,
      anon_sym_RPAREN,
  [65142] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5021), 1,
      aux_sym_preproc_if_token2,
  [65149] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5023), 1,
      sym_identifier,
  [65156] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5025), 1,
      anon_sym_SEMI,
  [65163] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5027), 1,
      anon_sym_RBRACE,
  [65170] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5029), 1,
      aux_sym_preproc_if_token2,
  [65177] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5031), 1,
      anon_sym_SEMI,
  [65184] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5033), 1,
      anon_sym_SEMI,
  [65191] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5035), 1,
      anon_sym_SEMI,
  [65198] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3299), 1,
      anon_sym_COLON,
  [65205] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5037), 1,
      aux_sym_preproc_if_token2,
  [65212] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5039), 1,
      anon_sym_SEMI,
  [65219] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5041), 1,
      anon_sym_RBRACE,
  [65226] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5043), 1,
      anon_sym_STAR,
  [65233] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5045), 1,
      sym_identifier,
  [65240] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5047), 1,
      aux_sym_preproc_include_token2,
  [65247] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5049), 1,
      anon_sym_SEMI,
  [65254] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5051), 1,
      sym_identifier,
  [65261] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5053), 1,
      aux_sym_preproc_if_token2,
  [65268] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5055), 1,
      anon_sym_COLON,
  [65275] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5057), 1,
      anon_sym_LPAREN2,
  [65282] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5059), 1,
      anon_sym_SEMI,
  [65289] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4508), 1,
      aux_sym_preproc_include_token2,
  [65296] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3339), 1,
      anon_sym_RPAREN,
  [65303] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5061), 1,
      sym_primitive_type,
  [65310] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5063), 1,
      anon_sym_RPAREN,
  [65317] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5065), 1,
      anon_sym_while,
  [65324] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4857), 1,
      anon_sym_RBRACE,
  [65331] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5067), 1,
      sym_identifier,
  [65338] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5069), 1,
      aux_sym_preproc_if_token2,
  [65345] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5071), 1,
      aux_sym_preproc_if_token2,
  [65352] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5073), 1,
      sym_identifier,
  [65359] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5075), 1,
      aux_sym_preproc_if_token2,
  [65366] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5077), 1,
      aux_sym_preproc_if_token2,
  [65373] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5079), 1,
      anon_sym_RPAREN,
  [65380] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5081), 1,
      anon_sym_SEMI,
  [65387] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5083), 1,
      anon_sym_LPAREN2,
  [65394] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5085), 1,
      anon_sym_LPAREN2,
  [65401] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5087), 1,
      anon_sym_STAR,
  [65408] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3267), 1,
      anon_sym_RPAREN,
  [65415] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5089), 1,
      sym_identifier,
  [65422] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5091), 1,
      sym_identifier,
  [65429] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3307), 1,
      anon_sym_SEMI,
  [65436] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5093), 1,
      aux_sym_preproc_if_token2,
  [65443] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5095), 1,
      anon_sym_SEMI,
  [65450] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5097), 1,
      aux_sym_preproc_include_token2,
  [65457] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5099), 1,
      aux_sym_preproc_if_token2,
  [65464] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5101), 1,
      aux_sym_preproc_if_token2,
  [65471] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5103), 1,
      anon_sym_SEMI,
  [65478] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5105), 1,
      sym_identifier,
  [65485] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5107), 1,
      anon_sym_LPAREN2,
  [65492] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5109), 1,
      aux_sym_preproc_if_token2,
  [65499] = 2,
    ACTIONS(2258), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [65506] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5111), 1,
      aux_sym_preproc_if_token2,
  [65513] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5113), 1,
      aux_sym_preproc_if_token2,
  [65520] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5115), 1,
      aux_sym_preproc_if_token2,
  [65527] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5117), 1,
      aux_sym_preproc_if_token2,
  [65534] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5119), 1,
      anon_sym_COLON,
  [65541] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5121), 1,
      anon_sym_STAR,
  [65548] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5123), 1,
      aux_sym_preproc_if_token2,
  [65555] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5125), 1,
      anon_sym_RPAREN,
  [65562] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5127), 1,
      aux_sym_preproc_include_token2,
  [65569] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5129), 1,
      sym_identifier,
  [65576] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5131), 1,
      aux_sym_preproc_if_token2,
  [65583] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5133), 1,
      anon_sym_RPAREN,
  [65590] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5135), 1,
      aux_sym_preproc_if_token2,
  [65597] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5137), 1,
      sym_identifier,
  [65604] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5139), 1,
      anon_sym_RPAREN,
  [65611] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5141), 1,
      anon_sym_RPAREN,
  [65618] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5143), 1,
      anon_sym_RBRACE,
  [65625] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5145), 1,
      aux_sym_preproc_if_token2,
  [65632] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5147), 1,
      aux_sym_preproc_if_token2,
  [65639] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5149), 1,
      anon_sym_RPAREN,
  [65646] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5151), 1,
      sym_identifier,
  [65653] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5153), 1,
      sym_identifier,
  [65660] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4841), 1,
      anon_sym_COMMA,
  [65667] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5155), 1,
      anon_sym_COLON,
  [65674] = 2,
    ACTIONS(3195), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [65681] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5157), 1,
      anon_sym_RPAREN,
  [65688] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5159), 1,
      sym_identifier,
  [65695] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3309), 1,
      anon_sym_SEMI,
  [65702] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4594), 1,
      anon_sym_RPAREN,
  [65709] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5161), 1,
      anon_sym_COLON,
  [65716] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5163), 1,
      anon_sym_SEMI,
  [65723] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5165), 1,
      anon_sym_LPAREN2,
  [65730] = 2,
    ACTIONS(2294), 1,
      aux_sym_preproc_include_token2,
    ACTIONS(3446), 1,
      sym_comment,
  [65737] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5167), 1,
      aux_sym_preproc_if_token2,
  [65744] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5169), 1,
      aux_sym_preproc_if_token2,
  [65751] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5171), 1,
      aux_sym_preproc_if_token2,
  [65758] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5173), 1,
      anon_sym_SEMI,
  [65765] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5175), 1,
      sym_identifier,
  [65772] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5177), 1,
      anon_sym_SEMI,
  [65779] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3277), 1,
      anon_sym_RPAREN,
  [65786] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5179), 1,
      anon_sym_SEMI,
  [65793] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5181), 1,
      anon_sym_COLON,
  [65800] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3297), 1,
      anon_sym_SEMI,
  [65807] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5183), 1,
      sym_identifier,
  [65814] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5185), 1,
      anon_sym_COLON,
  [65821] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5187), 1,
      sym_identifier,
  [65828] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5189), 1,
      aux_sym_preproc_include_token2,
  [65835] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5191), 1,
      sym_identifier,
  [65842] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5193), 1,
      sym_identifier,
  [65849] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5195), 1,
      anon_sym_SEMI,
  [65856] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5197), 1,
      sym_identifier,
  [65863] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5199), 1,
      aux_sym_preproc_if_token2,
  [65870] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5201), 1,
      anon_sym_RBRACE,
  [65877] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5203), 1,
      aux_sym_preproc_if_token2,
  [65884] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5205), 1,
      anon_sym_LPAREN2,
  [65891] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5207), 1,
      anon_sym_LPAREN2,
  [65898] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3279), 1,
      anon_sym_RPAREN,
  [65905] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5209), 1,
      anon_sym_LPAREN2,
  [65912] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3285), 1,
      anon_sym_RPAREN,
  [65919] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5211), 1,
      sym_identifier,
  [65926] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5213), 1,
      anon_sym_RPAREN,
  [65933] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5215), 1,
      anon_sym_SEMI,
  [65940] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3275), 1,
      anon_sym_SEMI,
  [65947] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5217), 1,
      aux_sym_preproc_if_token2,
  [65954] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5219), 1,
      aux_sym_preproc_if_token2,
  [65961] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5221), 1,
      anon_sym_while,
  [65968] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3259), 1,
      anon_sym_RPAREN,
  [65975] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5223), 1,
      aux_sym_preproc_if_token2,
  [65982] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5225), 1,
      anon_sym_SEMI,
  [65989] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5227), 1,
      anon_sym_STAR,
  [65996] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5229), 1,
      sym_identifier,
  [66003] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5231), 1,
      anon_sym_RBRACK,
  [66010] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5233), 1,
      anon_sym_SEMI,
  [66017] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5235), 1,
      anon_sym_RBRACK,
  [66024] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3313), 1,
      anon_sym_COLON,
  [66031] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5237), 1,
      anon_sym_STAR,
  [66038] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5239), 1,
      anon_sym_SEMI,
  [66045] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5241), 1,
      anon_sym_LPAREN2,
  [66052] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5243), 1,
      aux_sym_preproc_if_token2,
  [66059] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5245), 1,
      sym_identifier,
  [66066] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5247), 1,
      anon_sym_while,
  [66073] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5249), 1,
      anon_sym_RPAREN,
  [66080] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5251), 1,
      aux_sym_preproc_include_token2,
  [66087] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5253), 1,
      anon_sym_RPAREN,
  [66094] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3295), 1,
      anon_sym_COLON,
  [66101] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5255), 1,
      anon_sym_RPAREN,
  [66108] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(5257), 1,
      aux_sym_preproc_include_token2,
  [66115] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5259), 1,
      anon_sym_RBRACE,
  [66122] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5261), 1,
      anon_sym_while,
  [66129] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5263), 1,
      anon_sym_COLON,
  [66136] = 2,
    ACTIONS(3446), 1,
      sym_comment,
    ACTIONS(4760), 1,
      aux_sym_preproc_include_token2,
  [66143] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5265), 1,
      aux_sym_preproc_if_token2,
  [66150] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5267), 1,
      aux_sym_preproc_if_token2,
  [66157] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5269), 1,
      anon_sym_LPAREN2,
  [66164] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5271), 1,
      aux_sym_preproc_if_token2,
  [66171] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5273), 1,
      aux_sym_preproc_if_token2,
  [66178] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5275), 1,
      anon_sym_LPAREN2,
  [66185] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5277), 1,
      anon_sym_LPAREN2,
  [66192] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5279), 1,
      anon_sym_LPAREN2,
  [66199] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3221), 1,
      anon_sym_COLON,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(455)] = 0,
  [SMALL_STATE(456)] = 113,
  [SMALL_STATE(457)] = 226,
  [SMALL_STATE(458)] = 339,
  [SMALL_STATE(459)] = 452,
  [SMALL_STATE(460)] = 565,
  [SMALL_STATE(461)] = 678,
  [SMALL_STATE(462)] = 791,
  [SMALL_STATE(463)] = 904,
  [SMALL_STATE(464)] = 1017,
  [SMALL_STATE(465)] = 1130,
  [SMALL_STATE(466)] = 1243,
  [SMALL_STATE(467)] = 1356,
  [SMALL_STATE(468)] = 1469,
  [SMALL_STATE(469)] = 1582,
  [SMALL_STATE(470)] = 1695,
  [SMALL_STATE(471)] = 1808,
  [SMALL_STATE(472)] = 1921,
  [SMALL_STATE(473)] = 2034,
  [SMALL_STATE(474)] = 2147,
  [SMALL_STATE(475)] = 2260,
  [SMALL_STATE(476)] = 2373,
  [SMALL_STATE(477)] = 2486,
  [SMALL_STATE(478)] = 2599,
  [SMALL_STATE(479)] = 2712,
  [SMALL_STATE(480)] = 2825,
  [SMALL_STATE(481)] = 2938,
  [SMALL_STATE(482)] = 3051,
  [SMALL_STATE(483)] = 3164,
  [SMALL_STATE(484)] = 3277,
  [SMALL_STATE(485)] = 3387,
  [SMALL_STATE(486)] = 3497,
  [SMALL_STATE(487)] = 3607,
  [SMALL_STATE(488)] = 3717,
  [SMALL_STATE(489)] = 3827,
  [SMALL_STATE(490)] = 3937,
  [SMALL_STATE(491)] = 4047,
  [SMALL_STATE(492)] = 4157,
  [SMALL_STATE(493)] = 4267,
  [SMALL_STATE(494)] = 4377,
  [SMALL_STATE(495)] = 4487,
  [SMALL_STATE(496)] = 4597,
  [SMALL_STATE(497)] = 4707,
  [SMALL_STATE(498)] = 4817,
  [SMALL_STATE(499)] = 4927,
  [SMALL_STATE(500)] = 5034,
  [SMALL_STATE(501)] = 5141,
  [SMALL_STATE(502)] = 5248,
  [SMALL_STATE(503)] = 5355,
  [SMALL_STATE(504)] = 5462,
  [SMALL_STATE(505)] = 5569,
  [SMALL_STATE(506)] = 5676,
  [SMALL_STATE(507)] = 5783,
  [SMALL_STATE(508)] = 5890,
  [SMALL_STATE(509)] = 5997,
  [SMALL_STATE(510)] = 6104,
  [SMALL_STATE(511)] = 6211,
  [SMALL_STATE(512)] = 6318,
  [SMALL_STATE(513)] = 6425,
  [SMALL_STATE(514)] = 6532,
  [SMALL_STATE(515)] = 6639,
  [SMALL_STATE(516)] = 6746,
  [SMALL_STATE(517)] = 6853,
  [SMALL_STATE(518)] = 6960,
  [SMALL_STATE(519)] = 7067,
  [SMALL_STATE(520)] = 7174,
  [SMALL_STATE(521)] = 7281,
  [SMALL_STATE(522)] = 7388,
  [SMALL_STATE(523)] = 7495,
  [SMALL_STATE(524)] = 7602,
  [SMALL_STATE(525)] = 7709,
  [SMALL_STATE(526)] = 7816,
  [SMALL_STATE(527)] = 7923,
  [SMALL_STATE(528)] = 8014,
  [SMALL_STATE(529)] = 8121,
  [SMALL_STATE(530)] = 8228,
  [SMALL_STATE(531)] = 8335,
  [SMALL_STATE(532)] = 8442,
  [SMALL_STATE(533)] = 8549,
  [SMALL_STATE(534)] = 8656,
  [SMALL_STATE(535)] = 8763,
  [SMALL_STATE(536)] = 8870,
  [SMALL_STATE(537)] = 8977,
  [SMALL_STATE(538)] = 9084,
  [SMALL_STATE(539)] = 9191,
  [SMALL_STATE(540)] = 9298,
  [SMALL_STATE(541)] = 9405,
  [SMALL_STATE(542)] = 9512,
  [SMALL_STATE(543)] = 9619,
  [SMALL_STATE(544)] = 9726,
  [SMALL_STATE(545)] = 9833,
  [SMALL_STATE(546)] = 9940,
  [SMALL_STATE(547)] = 10047,
  [SMALL_STATE(548)] = 10154,
  [SMALL_STATE(549)] = 10261,
  [SMALL_STATE(550)] = 10368,
  [SMALL_STATE(551)] = 10475,
  [SMALL_STATE(552)] = 10582,
  [SMALL_STATE(553)] = 10689,
  [SMALL_STATE(554)] = 10796,
  [SMALL_STATE(555)] = 10903,
  [SMALL_STATE(556)] = 11010,
  [SMALL_STATE(557)] = 11117,
  [SMALL_STATE(558)] = 11224,
  [SMALL_STATE(559)] = 11331,
  [SMALL_STATE(560)] = 11438,
  [SMALL_STATE(561)] = 11545,
  [SMALL_STATE(562)] = 11652,
  [SMALL_STATE(563)] = 11759,
  [SMALL_STATE(564)] = 11866,
  [SMALL_STATE(565)] = 11973,
  [SMALL_STATE(566)] = 12080,
  [SMALL_STATE(567)] = 12187,
  [SMALL_STATE(568)] = 12294,
  [SMALL_STATE(569)] = 12401,
  [SMALL_STATE(570)] = 12508,
  [SMALL_STATE(571)] = 12615,
  [SMALL_STATE(572)] = 12722,
  [SMALL_STATE(573)] = 12829,
  [SMALL_STATE(574)] = 12936,
  [SMALL_STATE(575)] = 13043,
  [SMALL_STATE(576)] = 13150,
  [SMALL_STATE(577)] = 13257,
  [SMALL_STATE(578)] = 13364,
  [SMALL_STATE(579)] = 13471,
  [SMALL_STATE(580)] = 13578,
  [SMALL_STATE(581)] = 13685,
  [SMALL_STATE(582)] = 13792,
  [SMALL_STATE(583)] = 13899,
  [SMALL_STATE(584)] = 14006,
  [SMALL_STATE(585)] = 14113,
  [SMALL_STATE(586)] = 14220,
  [SMALL_STATE(587)] = 14327,
  [SMALL_STATE(588)] = 14434,
  [SMALL_STATE(589)] = 14541,
  [SMALL_STATE(590)] = 14648,
  [SMALL_STATE(591)] = 14755,
  [SMALL_STATE(592)] = 14862,
  [SMALL_STATE(593)] = 14969,
  [SMALL_STATE(594)] = 15076,
  [SMALL_STATE(595)] = 15183,
  [SMALL_STATE(596)] = 15290,
  [SMALL_STATE(597)] = 15397,
  [SMALL_STATE(598)] = 15504,
  [SMALL_STATE(599)] = 15611,
  [SMALL_STATE(600)] = 15718,
  [SMALL_STATE(601)] = 15825,
  [SMALL_STATE(602)] = 15932,
  [SMALL_STATE(603)] = 16039,
  [SMALL_STATE(604)] = 16146,
  [SMALL_STATE(605)] = 16253,
  [SMALL_STATE(606)] = 16360,
  [SMALL_STATE(607)] = 16467,
  [SMALL_STATE(608)] = 16574,
  [SMALL_STATE(609)] = 16681,
  [SMALL_STATE(610)] = 16788,
  [SMALL_STATE(611)] = 16895,
  [SMALL_STATE(612)] = 17002,
  [SMALL_STATE(613)] = 17109,
  [SMALL_STATE(614)] = 17216,
  [SMALL_STATE(615)] = 17323,
  [SMALL_STATE(616)] = 17430,
  [SMALL_STATE(617)] = 17537,
  [SMALL_STATE(618)] = 17644,
  [SMALL_STATE(619)] = 17751,
  [SMALL_STATE(620)] = 17858,
  [SMALL_STATE(621)] = 17930,
  [SMALL_STATE(622)] = 18002,
  [SMALL_STATE(623)] = 18074,
  [SMALL_STATE(624)] = 18181,
  [SMALL_STATE(625)] = 18290,
  [SMALL_STATE(626)] = 18359,
  [SMALL_STATE(627)] = 18439,
  [SMALL_STATE(628)] = 18503,
  [SMALL_STATE(629)] = 18567,
  [SMALL_STATE(630)] = 18631,
  [SMALL_STATE(631)] = 18695,
  [SMALL_STATE(632)] = 18759,
  [SMALL_STATE(633)] = 18823,
  [SMALL_STATE(634)] = 18887,
  [SMALL_STATE(635)] = 18951,
  [SMALL_STATE(636)] = 19015,
  [SMALL_STATE(637)] = 19079,
  [SMALL_STATE(638)] = 19143,
  [SMALL_STATE(639)] = 19207,
  [SMALL_STATE(640)] = 19308,
  [SMALL_STATE(641)] = 19409,
  [SMALL_STATE(642)] = 19510,
  [SMALL_STATE(643)] = 19611,
  [SMALL_STATE(644)] = 19712,
  [SMALL_STATE(645)] = 19813,
  [SMALL_STATE(646)] = 19914,
  [SMALL_STATE(647)] = 20015,
  [SMALL_STATE(648)] = 20113,
  [SMALL_STATE(649)] = 20175,
  [SMALL_STATE(650)] = 20273,
  [SMALL_STATE(651)] = 20373,
  [SMALL_STATE(652)] = 20468,
  [SMALL_STATE(653)] = 20563,
  [SMALL_STATE(654)] = 20624,
  [SMALL_STATE(655)] = 20719,
  [SMALL_STATE(656)] = 20814,
  [SMALL_STATE(657)] = 20875,
  [SMALL_STATE(658)] = 20970,
  [SMALL_STATE(659)] = 21065,
  [SMALL_STATE(660)] = 21126,
  [SMALL_STATE(661)] = 21221,
  [SMALL_STATE(662)] = 21316,
  [SMALL_STATE(663)] = 21387,
  [SMALL_STATE(664)] = 21453,
  [SMALL_STATE(665)] = 21513,
  [SMALL_STATE(666)] = 21581,
  [SMALL_STATE(667)] = 21672,
  [SMALL_STATE(668)] = 21731,
  [SMALL_STATE(669)] = 21790,
  [SMALL_STATE(670)] = 21849,
  [SMALL_STATE(671)] = 21908,
  [SMALL_STATE(672)] = 21967,
  [SMALL_STATE(673)] = 22026,
  [SMALL_STATE(674)] = 22085,
  [SMALL_STATE(675)] = 22144,
  [SMALL_STATE(676)] = 22203,
  [SMALL_STATE(677)] = 22294,
  [SMALL_STATE(678)] = 22353,
  [SMALL_STATE(679)] = 22412,
  [SMALL_STATE(680)] = 22471,
  [SMALL_STATE(681)] = 22530,
  [SMALL_STATE(682)] = 22589,
  [SMALL_STATE(683)] = 22648,
  [SMALL_STATE(684)] = 22739,
  [SMALL_STATE(685)] = 22798,
  [SMALL_STATE(686)] = 22889,
  [SMALL_STATE(687)] = 22980,
  [SMALL_STATE(688)] = 23071,
  [SMALL_STATE(689)] = 23162,
  [SMALL_STATE(690)] = 23253,
  [SMALL_STATE(691)] = 23312,
  [SMALL_STATE(692)] = 23378,
  [SMALL_STATE(693)] = 23446,
  [SMALL_STATE(694)] = 23504,
  [SMALL_STATE(695)] = 23562,
  [SMALL_STATE(696)] = 23650,
  [SMALL_STATE(697)] = 23708,
  [SMALL_STATE(698)] = 23776,
  [SMALL_STATE(699)] = 23834,
  [SMALL_STATE(700)] = 23902,
  [SMALL_STATE(701)] = 23960,
  [SMALL_STATE(702)] = 24028,
  [SMALL_STATE(703)] = 24096,
  [SMALL_STATE(704)] = 24161,
  [SMALL_STATE(705)] = 24234,
  [SMALL_STATE(706)] = 24299,
  [SMALL_STATE(707)] = 24356,
  [SMALL_STATE(708)] = 24413,
  [SMALL_STATE(709)] = 24478,
  [SMALL_STATE(710)] = 24543,
  [SMALL_STATE(711)] = 24608,
  [SMALL_STATE(712)] = 24665,
  [SMALL_STATE(713)] = 24738,
  [SMALL_STATE(714)] = 24799,
  [SMALL_STATE(715)] = 24864,
  [SMALL_STATE(716)] = 24929,
  [SMALL_STATE(717)] = 25001,
  [SMALL_STATE(718)] = 25057,
  [SMALL_STATE(719)] = 25113,
  [SMALL_STATE(720)] = 25185,
  [SMALL_STATE(721)] = 25257,
  [SMALL_STATE(722)] = 25329,
  [SMALL_STATE(723)] = 25385,
  [SMALL_STATE(724)] = 25444,
  [SMALL_STATE(725)] = 25503,
  [SMALL_STATE(726)] = 25562,
  [SMALL_STATE(727)] = 25621,
  [SMALL_STATE(728)] = 25680,
  [SMALL_STATE(729)] = 25739,
  [SMALL_STATE(730)] = 25798,
  [SMALL_STATE(731)] = 25857,
  [SMALL_STATE(732)] = 25916,
  [SMALL_STATE(733)] = 25975,
  [SMALL_STATE(734)] = 26034,
  [SMALL_STATE(735)] = 26088,
  [SMALL_STATE(736)] = 26142,
  [SMALL_STATE(737)] = 26200,
  [SMALL_STATE(738)] = 26258,
  [SMALL_STATE(739)] = 26312,
  [SMALL_STATE(740)] = 26370,
  [SMALL_STATE(741)] = 26424,
  [SMALL_STATE(742)] = 26482,
  [SMALL_STATE(743)] = 26540,
  [SMALL_STATE(744)] = 26598,
  [SMALL_STATE(745)] = 26656,
  [SMALL_STATE(746)] = 26714,
  [SMALL_STATE(747)] = 26768,
  [SMALL_STATE(748)] = 26822,
  [SMALL_STATE(749)] = 26876,
  [SMALL_STATE(750)] = 26930,
  [SMALL_STATE(751)] = 26990,
  [SMALL_STATE(752)] = 27044,
  [SMALL_STATE(753)] = 27098,
  [SMALL_STATE(754)] = 27152,
  [SMALL_STATE(755)] = 27212,
  [SMALL_STATE(756)] = 27266,
  [SMALL_STATE(757)] = 27320,
  [SMALL_STATE(758)] = 27374,
  [SMALL_STATE(759)] = 27428,
  [SMALL_STATE(760)] = 27482,
  [SMALL_STATE(761)] = 27536,
  [SMALL_STATE(762)] = 27590,
  [SMALL_STATE(763)] = 27648,
  [SMALL_STATE(764)] = 27702,
  [SMALL_STATE(765)] = 27760,
  [SMALL_STATE(766)] = 27818,
  [SMALL_STATE(767)] = 27872,
  [SMALL_STATE(768)] = 27930,
  [SMALL_STATE(769)] = 27984,
  [SMALL_STATE(770)] = 28038,
  [SMALL_STATE(771)] = 28092,
  [SMALL_STATE(772)] = 28146,
  [SMALL_STATE(773)] = 28200,
  [SMALL_STATE(774)] = 28254,
  [SMALL_STATE(775)] = 28308,
  [SMALL_STATE(776)] = 28361,
  [SMALL_STATE(777)] = 28414,
  [SMALL_STATE(778)] = 28467,
  [SMALL_STATE(779)] = 28530,
  [SMALL_STATE(780)] = 28583,
  [SMALL_STATE(781)] = 28636,
  [SMALL_STATE(782)] = 28689,
  [SMALL_STATE(783)] = 28760,
  [SMALL_STATE(784)] = 28823,
  [SMALL_STATE(785)] = 28914,
  [SMALL_STATE(786)] = 28967,
  [SMALL_STATE(787)] = 29030,
  [SMALL_STATE(788)] = 29083,
  [SMALL_STATE(789)] = 29174,
  [SMALL_STATE(790)] = 29245,
  [SMALL_STATE(791)] = 29308,
  [SMALL_STATE(792)] = 29361,
  [SMALL_STATE(793)] = 29414,
  [SMALL_STATE(794)] = 29477,
  [SMALL_STATE(795)] = 29540,
  [SMALL_STATE(796)] = 29593,
  [SMALL_STATE(797)] = 29664,
  [SMALL_STATE(798)] = 29717,
  [SMALL_STATE(799)] = 29770,
  [SMALL_STATE(800)] = 29823,
  [SMALL_STATE(801)] = 29876,
  [SMALL_STATE(802)] = 29947,
  [SMALL_STATE(803)] = 30000,
  [SMALL_STATE(804)] = 30053,
  [SMALL_STATE(805)] = 30106,
  [SMALL_STATE(806)] = 30188,
  [SMALL_STATE(807)] = 30268,
  [SMALL_STATE(808)] = 30346,
  [SMALL_STATE(809)] = 30422,
  [SMALL_STATE(810)] = 30496,
  [SMALL_STATE(811)] = 30568,
  [SMALL_STATE(812)] = 30636,
  [SMALL_STATE(813)] = 30702,
  [SMALL_STATE(814)] = 30788,
  [SMALL_STATE(815)] = 30874,
  [SMALL_STATE(816)] = 30954,
  [SMALL_STATE(817)] = 31028,
  [SMALL_STATE(818)] = 31112,
  [SMALL_STATE(819)] = 31184,
  [SMALL_STATE(820)] = 31252,
  [SMALL_STATE(821)] = 31310,
  [SMALL_STATE(822)] = 31368,
  [SMALL_STATE(823)] = 31434,
  [SMALL_STATE(824)] = 31512,
  [SMALL_STATE(825)] = 31574,
  [SMALL_STATE(826)] = 31658,
  [SMALL_STATE(827)] = 31740,
  [SMALL_STATE(828)] = 31800,
  [SMALL_STATE(829)] = 31886,
  [SMALL_STATE(830)] = 31972,
  [SMALL_STATE(831)] = 32036,
  [SMALL_STATE(832)] = 32112,
  [SMALL_STATE(833)] = 32198,
  [SMALL_STATE(834)] = 32262,
  [SMALL_STATE(835)] = 32348,
  [SMALL_STATE(836)] = 32403,
  [SMALL_STATE(837)] = 32462,
  [SMALL_STATE(838)] = 32517,
  [SMALL_STATE(839)] = 32603,
  [SMALL_STATE(840)] = 32653,
  [SMALL_STATE(841)] = 32739,
  [SMALL_STATE(842)] = 32789,
  [SMALL_STATE(843)] = 32838,
  [SMALL_STATE(844)] = 32887,
  [SMALL_STATE(845)] = 32936,
  [SMALL_STATE(846)] = 32985,
  [SMALL_STATE(847)] = 33034,
  [SMALL_STATE(848)] = 33083,
  [SMALL_STATE(849)] = 33136,
  [SMALL_STATE(850)] = 33185,
  [SMALL_STATE(851)] = 33234,
  [SMALL_STATE(852)] = 33283,
  [SMALL_STATE(853)] = 33332,
  [SMALL_STATE(854)] = 33381,
  [SMALL_STATE(855)] = 33430,
  [SMALL_STATE(856)] = 33479,
  [SMALL_STATE(857)] = 33528,
  [SMALL_STATE(858)] = 33577,
  [SMALL_STATE(859)] = 33626,
  [SMALL_STATE(860)] = 33675,
  [SMALL_STATE(861)] = 33724,
  [SMALL_STATE(862)] = 33773,
  [SMALL_STATE(863)] = 33822,
  [SMALL_STATE(864)] = 33871,
  [SMALL_STATE(865)] = 33920,
  [SMALL_STATE(866)] = 33969,
  [SMALL_STATE(867)] = 34018,
  [SMALL_STATE(868)] = 34067,
  [SMALL_STATE(869)] = 34116,
  [SMALL_STATE(870)] = 34165,
  [SMALL_STATE(871)] = 34214,
  [SMALL_STATE(872)] = 34263,
  [SMALL_STATE(873)] = 34312,
  [SMALL_STATE(874)] = 34365,
  [SMALL_STATE(875)] = 34418,
  [SMALL_STATE(876)] = 34467,
  [SMALL_STATE(877)] = 34516,
  [SMALL_STATE(878)] = 34569,
  [SMALL_STATE(879)] = 34618,
  [SMALL_STATE(880)] = 34667,
  [SMALL_STATE(881)] = 34716,
  [SMALL_STATE(882)] = 34769,
  [SMALL_STATE(883)] = 34822,
  [SMALL_STATE(884)] = 34875,
  [SMALL_STATE(885)] = 34924,
  [SMALL_STATE(886)] = 34983,
  [SMALL_STATE(887)] = 35030,
  [SMALL_STATE(888)] = 35107,
  [SMALL_STATE(889)] = 35166,
  [SMALL_STATE(890)] = 35225,
  [SMALL_STATE(891)] = 35284,
  [SMALL_STATE(892)] = 35367,
  [SMALL_STATE(893)] = 35428,
  [SMALL_STATE(894)] = 35507,
  [SMALL_STATE(895)] = 35582,
  [SMALL_STATE(896)] = 35655,
  [SMALL_STATE(897)] = 35726,
  [SMALL_STATE(898)] = 35795,
  [SMALL_STATE(899)] = 35860,
  [SMALL_STATE(900)] = 35923,
  [SMALL_STATE(901)] = 36006,
  [SMALL_STATE(902)] = 36089,
  [SMALL_STATE(903)] = 36170,
  [SMALL_STATE(904)] = 36247,
  [SMALL_STATE(905)] = 36324,
  [SMALL_STATE(906)] = 36381,
  [SMALL_STATE(907)] = 36440,
  [SMALL_STATE(908)] = 36499,
  [SMALL_STATE(909)] = 36576,
  [SMALL_STATE(910)] = 36653,
  [SMALL_STATE(911)] = 36699,
  [SMALL_STATE(912)] = 36749,
  [SMALL_STATE(913)] = 36803,
  [SMALL_STATE(914)] = 36849,
  [SMALL_STATE(915)] = 36895,
  [SMALL_STATE(916)] = 36941,
  [SMALL_STATE(917)] = 36987,
  [SMALL_STATE(918)] = 37033,
  [SMALL_STATE(919)] = 37079,
  [SMALL_STATE(920)] = 37125,
  [SMALL_STATE(921)] = 37171,
  [SMALL_STATE(922)] = 37217,
  [SMALL_STATE(923)] = 37263,
  [SMALL_STATE(924)] = 37309,
  [SMALL_STATE(925)] = 37365,
  [SMALL_STATE(926)] = 37411,
  [SMALL_STATE(927)] = 37486,
  [SMALL_STATE(928)] = 37561,
  [SMALL_STATE(929)] = 37612,
  [SMALL_STATE(930)] = 37656,
  [SMALL_STATE(931)] = 37700,
  [SMALL_STATE(932)] = 37744,
  [SMALL_STATE(933)] = 37788,
  [SMALL_STATE(934)] = 37832,
  [SMALL_STATE(935)] = 37876,
  [SMALL_STATE(936)] = 37920,
  [SMALL_STATE(937)] = 37964,
  [SMALL_STATE(938)] = 38008,
  [SMALL_STATE(939)] = 38052,
  [SMALL_STATE(940)] = 38103,
  [SMALL_STATE(941)] = 38146,
  [SMALL_STATE(942)] = 38219,
  [SMALL_STATE(943)] = 38262,
  [SMALL_STATE(944)] = 38335,
  [SMALL_STATE(945)] = 38378,
  [SMALL_STATE(946)] = 38446,
  [SMALL_STATE(947)] = 38516,
  [SMALL_STATE(948)] = 38586,
  [SMALL_STATE(949)] = 38654,
  [SMALL_STATE(950)] = 38722,
  [SMALL_STATE(951)] = 38790,
  [SMALL_STATE(952)] = 38858,
  [SMALL_STATE(953)] = 38926,
  [SMALL_STATE(954)] = 38994,
  [SMALL_STATE(955)] = 39062,
  [SMALL_STATE(956)] = 39130,
  [SMALL_STATE(957)] = 39198,
  [SMALL_STATE(958)] = 39266,
  [SMALL_STATE(959)] = 39334,
  [SMALL_STATE(960)] = 39407,
  [SMALL_STATE(961)] = 39470,
  [SMALL_STATE(962)] = 39527,
  [SMALL_STATE(963)] = 39604,
  [SMALL_STATE(964)] = 39681,
  [SMALL_STATE(965)] = 39736,
  [SMALL_STATE(966)] = 39809,
  [SMALL_STATE(967)] = 39880,
  [SMALL_STATE(968)] = 39949,
  [SMALL_STATE(969)] = 40016,
  [SMALL_STATE(970)] = 40081,
  [SMALL_STATE(971)] = 40144,
  [SMALL_STATE(972)] = 40203,
  [SMALL_STATE(973)] = 40260,
  [SMALL_STATE(974)] = 40337,
  [SMALL_STATE(975)] = 40414,
  [SMALL_STATE(976)] = 40469,
  [SMALL_STATE(977)] = 40540,
  [SMALL_STATE(978)] = 40609,
  [SMALL_STATE(979)] = 40686,
  [SMALL_STATE(980)] = 40753,
  [SMALL_STATE(981)] = 40828,
  [SMALL_STATE(982)] = 40903,
  [SMALL_STATE(983)] = 40980,
  [SMALL_STATE(984)] = 41057,
  [SMALL_STATE(985)] = 41122,
  [SMALL_STATE(986)] = 41181,
  [SMALL_STATE(987)] = 41221,
  [SMALL_STATE(988)] = 41285,
  [SMALL_STATE(989)] = 41349,
  [SMALL_STATE(990)] = 41389,
  [SMALL_STATE(991)] = 41429,
  [SMALL_STATE(992)] = 41469,
  [SMALL_STATE(993)] = 41533,
  [SMALL_STATE(994)] = 41597,
  [SMALL_STATE(995)] = 41661,
  [SMALL_STATE(996)] = 41725,
  [SMALL_STATE(997)] = 41789,
  [SMALL_STATE(998)] = 41853,
  [SMALL_STATE(999)] = 41893,
  [SMALL_STATE(1000)] = 41957,
  [SMALL_STATE(1001)] = 42021,
  [SMALL_STATE(1002)] = 42061,
  [SMALL_STATE(1003)] = 42125,
  [SMALL_STATE(1004)] = 42189,
  [SMALL_STATE(1005)] = 42253,
  [SMALL_STATE(1006)] = 42314,
  [SMALL_STATE(1007)] = 42375,
  [SMALL_STATE(1008)] = 42436,
  [SMALL_STATE(1009)] = 42497,
  [SMALL_STATE(1010)] = 42558,
  [SMALL_STATE(1011)] = 42619,
  [SMALL_STATE(1012)] = 42665,
  [SMALL_STATE(1013)] = 42739,
  [SMALL_STATE(1014)] = 42813,
  [SMALL_STATE(1015)] = 42861,
  [SMALL_STATE(1016)] = 42936,
  [SMALL_STATE(1017)] = 43011,
  [SMALL_STATE(1018)] = 43086,
  [SMALL_STATE(1019)] = 43161,
  [SMALL_STATE(1020)] = 43233,
  [SMALL_STATE(1021)] = 43305,
  [SMALL_STATE(1022)] = 43345,
  [SMALL_STATE(1023)] = 43417,
  [SMALL_STATE(1024)] = 43487,
  [SMALL_STATE(1025)] = 43559,
  [SMALL_STATE(1026)] = 43629,
  [SMALL_STATE(1027)] = 43679,
  [SMALL_STATE(1028)] = 43751,
  [SMALL_STATE(1029)] = 43823,
  [SMALL_STATE(1030)] = 43895,
  [SMALL_STATE(1031)] = 43965,
  [SMALL_STATE(1032)] = 44037,
  [SMALL_STATE(1033)] = 44107,
  [SMALL_STATE(1034)] = 44179,
  [SMALL_STATE(1035)] = 44251,
  [SMALL_STATE(1036)] = 44323,
  [SMALL_STATE(1037)] = 44393,
  [SMALL_STATE(1038)] = 44465,
  [SMALL_STATE(1039)] = 44537,
  [SMALL_STATE(1040)] = 44611,
  [SMALL_STATE(1041)] = 44683,
  [SMALL_STATE(1042)] = 44753,
  [SMALL_STATE(1043)] = 44825,
  [SMALL_STATE(1044)] = 44895,
  [SMALL_STATE(1045)] = 44945,
  [SMALL_STATE(1046)] = 45017,
  [SMALL_STATE(1047)] = 45089,
  [SMALL_STATE(1048)] = 45161,
  [SMALL_STATE(1049)] = 45233,
  [SMALL_STATE(1050)] = 45305,
  [SMALL_STATE(1051)] = 45345,
  [SMALL_STATE(1052)] = 45417,
  [SMALL_STATE(1053)] = 45489,
  [SMALL_STATE(1054)] = 45559,
  [SMALL_STATE(1055)] = 45611,
  [SMALL_STATE(1056)] = 45683,
  [SMALL_STATE(1057)] = 45755,
  [SMALL_STATE(1058)] = 45827,
  [SMALL_STATE(1059)] = 45895,
  [SMALL_STATE(1060)] = 45961,
  [SMALL_STATE(1061)] = 46027,
  [SMALL_STATE(1062)] = 46091,
  [SMALL_STATE(1063)] = 46153,
  [SMALL_STATE(1064)] = 46213,
  [SMALL_STATE(1065)] = 46269,
  [SMALL_STATE(1066)] = 46341,
  [SMALL_STATE(1067)] = 46395,
  [SMALL_STATE(1068)] = 46467,
  [SMALL_STATE(1069)] = 46507,
  [SMALL_STATE(1070)] = 46579,
  [SMALL_STATE(1071)] = 46651,
  [SMALL_STATE(1072)] = 46720,
  [SMALL_STATE(1073)] = 46789,
  [SMALL_STATE(1074)] = 46858,
  [SMALL_STATE(1075)] = 46913,
  [SMALL_STATE(1076)] = 46982,
  [SMALL_STATE(1077)] = 47051,
  [SMALL_STATE(1078)] = 47106,
  [SMALL_STATE(1079)] = 47175,
  [SMALL_STATE(1080)] = 47230,
  [SMALL_STATE(1081)] = 47299,
  [SMALL_STATE(1082)] = 47368,
  [SMALL_STATE(1083)] = 47437,
  [SMALL_STATE(1084)] = 47506,
  [SMALL_STATE(1085)] = 47575,
  [SMALL_STATE(1086)] = 47644,
  [SMALL_STATE(1087)] = 47699,
  [SMALL_STATE(1088)] = 47768,
  [SMALL_STATE(1089)] = 47837,
  [SMALL_STATE(1090)] = 47906,
  [SMALL_STATE(1091)] = 47975,
  [SMALL_STATE(1092)] = 48044,
  [SMALL_STATE(1093)] = 48113,
  [SMALL_STATE(1094)] = 48168,
  [SMALL_STATE(1095)] = 48237,
  [SMALL_STATE(1096)] = 48306,
  [SMALL_STATE(1097)] = 48375,
  [SMALL_STATE(1098)] = 48444,
  [SMALL_STATE(1099)] = 48499,
  [SMALL_STATE(1100)] = 48568,
  [SMALL_STATE(1101)] = 48637,
  [SMALL_STATE(1102)] = 48703,
  [SMALL_STATE(1103)] = 48755,
  [SMALL_STATE(1104)] = 48791,
  [SMALL_STATE(1105)] = 48843,
  [SMALL_STATE(1106)] = 48895,
  [SMALL_STATE(1107)] = 48947,
  [SMALL_STATE(1108)] = 48999,
  [SMALL_STATE(1109)] = 49051,
  [SMALL_STATE(1110)] = 49103,
  [SMALL_STATE(1111)] = 49155,
  [SMALL_STATE(1112)] = 49207,
  [SMALL_STATE(1113)] = 49256,
  [SMALL_STATE(1114)] = 49295,
  [SMALL_STATE(1115)] = 49344,
  [SMALL_STATE(1116)] = 49384,
  [SMALL_STATE(1117)] = 49424,
  [SMALL_STATE(1118)] = 49464,
  [SMALL_STATE(1119)] = 49520,
  [SMALL_STATE(1120)] = 49560,
  [SMALL_STATE(1121)] = 49609,
  [SMALL_STATE(1122)] = 49643,
  [SMALL_STATE(1123)] = 49677,
  [SMALL_STATE(1124)] = 49711,
  [SMALL_STATE(1125)] = 49745,
  [SMALL_STATE(1126)] = 49779,
  [SMALL_STATE(1127)] = 49813,
  [SMALL_STATE(1128)] = 49847,
  [SMALL_STATE(1129)] = 49901,
  [SMALL_STATE(1130)] = 49935,
  [SMALL_STATE(1131)] = 49969,
  [SMALL_STATE(1132)] = 50003,
  [SMALL_STATE(1133)] = 50057,
  [SMALL_STATE(1134)] = 50111,
  [SMALL_STATE(1135)] = 50145,
  [SMALL_STATE(1136)] = 50179,
  [SMALL_STATE(1137)] = 50213,
  [SMALL_STATE(1138)] = 50249,
  [SMALL_STATE(1139)] = 50283,
  [SMALL_STATE(1140)] = 50337,
  [SMALL_STATE(1141)] = 50388,
  [SMALL_STATE(1142)] = 50421,
  [SMALL_STATE(1143)] = 50472,
  [SMALL_STATE(1144)] = 50523,
  [SMALL_STATE(1145)] = 50574,
  [SMALL_STATE(1146)] = 50625,
  [SMALL_STATE(1147)] = 50668,
  [SMALL_STATE(1148)] = 50719,
  [SMALL_STATE(1149)] = 50770,
  [SMALL_STATE(1150)] = 50825,
  [SMALL_STATE(1151)] = 50868,
  [SMALL_STATE(1152)] = 50919,
  [SMALL_STATE(1153)] = 50974,
  [SMALL_STATE(1154)] = 51029,
  [SMALL_STATE(1155)] = 51080,
  [SMALL_STATE(1156)] = 51131,
  [SMALL_STATE(1157)] = 51182,
  [SMALL_STATE(1158)] = 51225,
  [SMALL_STATE(1159)] = 51266,
  [SMALL_STATE(1160)] = 51306,
  [SMALL_STATE(1161)] = 51346,
  [SMALL_STATE(1162)] = 51386,
  [SMALL_STATE(1163)] = 51426,
  [SMALL_STATE(1164)] = 51466,
  [SMALL_STATE(1165)] = 51506,
  [SMALL_STATE(1166)] = 51534,
  [SMALL_STATE(1167)] = 51574,
  [SMALL_STATE(1168)] = 51614,
  [SMALL_STATE(1169)] = 51654,
  [SMALL_STATE(1170)] = 51694,
  [SMALL_STATE(1171)] = 51734,
  [SMALL_STATE(1172)] = 51774,
  [SMALL_STATE(1173)] = 51814,
  [SMALL_STATE(1174)] = 51854,
  [SMALL_STATE(1175)] = 51894,
  [SMALL_STATE(1176)] = 51934,
  [SMALL_STATE(1177)] = 51974,
  [SMALL_STATE(1178)] = 52014,
  [SMALL_STATE(1179)] = 52054,
  [SMALL_STATE(1180)] = 52088,
  [SMALL_STATE(1181)] = 52128,
  [SMALL_STATE(1182)] = 52156,
  [SMALL_STATE(1183)] = 52206,
  [SMALL_STATE(1184)] = 52246,
  [SMALL_STATE(1185)] = 52286,
  [SMALL_STATE(1186)] = 52326,
  [SMALL_STATE(1187)] = 52366,
  [SMALL_STATE(1188)] = 52406,
  [SMALL_STATE(1189)] = 52446,
  [SMALL_STATE(1190)] = 52486,
  [SMALL_STATE(1191)] = 52526,
  [SMALL_STATE(1192)] = 52566,
  [SMALL_STATE(1193)] = 52606,
  [SMALL_STATE(1194)] = 52634,
  [SMALL_STATE(1195)] = 52674,
  [SMALL_STATE(1196)] = 52714,
  [SMALL_STATE(1197)] = 52746,
  [SMALL_STATE(1198)] = 52786,
  [SMALL_STATE(1199)] = 52826,
  [SMALL_STATE(1200)] = 52866,
  [SMALL_STATE(1201)] = 52906,
  [SMALL_STATE(1202)] = 52934,
  [SMALL_STATE(1203)] = 52974,
  [SMALL_STATE(1204)] = 53022,
  [SMALL_STATE(1205)] = 53058,
  [SMALL_STATE(1206)] = 53098,
  [SMALL_STATE(1207)] = 53138,
  [SMALL_STATE(1208)] = 53184,
  [SMALL_STATE(1209)] = 53212,
  [SMALL_STATE(1210)] = 53258,
  [SMALL_STATE(1211)] = 53302,
  [SMALL_STATE(1212)] = 53344,
  [SMALL_STATE(1213)] = 53372,
  [SMALL_STATE(1214)] = 53412,
  [SMALL_STATE(1215)] = 53461,
  [SMALL_STATE(1216)] = 53488,
  [SMALL_STATE(1217)] = 53517,
  [SMALL_STATE(1218)] = 53544,
  [SMALL_STATE(1219)] = 53587,
  [SMALL_STATE(1220)] = 53628,
  [SMALL_STATE(1221)] = 53667,
  [SMALL_STATE(1222)] = 53704,
  [SMALL_STATE(1223)] = 53739,
  [SMALL_STATE(1224)] = 53772,
  [SMALL_STATE(1225)] = 53803,
  [SMALL_STATE(1226)] = 53848,
  [SMALL_STATE(1227)] = 53893,
  [SMALL_STATE(1228)] = 53920,
  [SMALL_STATE(1229)] = 53965,
  [SMALL_STATE(1230)] = 54010,
  [SMALL_STATE(1231)] = 54037,
  [SMALL_STATE(1232)] = 54082,
  [SMALL_STATE(1233)] = 54131,
  [SMALL_STATE(1234)] = 54158,
  [SMALL_STATE(1235)] = 54203,
  [SMALL_STATE(1236)] = 54248,
  [SMALL_STATE(1237)] = 54275,
  [SMALL_STATE(1238)] = 54302,
  [SMALL_STATE(1239)] = 54329,
  [SMALL_STATE(1240)] = 54356,
  [SMALL_STATE(1241)] = 54399,
  [SMALL_STATE(1242)] = 54444,
  [SMALL_STATE(1243)] = 54471,
  [SMALL_STATE(1244)] = 54514,
  [SMALL_STATE(1245)] = 54559,
  [SMALL_STATE(1246)] = 54604,
  [SMALL_STATE(1247)] = 54649,
  [SMALL_STATE(1248)] = 54694,
  [SMALL_STATE(1249)] = 54737,
  [SMALL_STATE(1250)] = 54782,
  [SMALL_STATE(1251)] = 54827,
  [SMALL_STATE(1252)] = 54870,
  [SMALL_STATE(1253)] = 54915,
  [SMALL_STATE(1254)] = 54955,
  [SMALL_STATE(1255)] = 54991,
  [SMALL_STATE(1256)] = 55031,
  [SMALL_STATE(1257)] = 55071,
  [SMALL_STATE(1258)] = 55115,
  [SMALL_STATE(1259)] = 55155,
  [SMALL_STATE(1260)] = 55195,
  [SMALL_STATE(1261)] = 55237,
  [SMALL_STATE(1262)] = 55277,
  [SMALL_STATE(1263)] = 55308,
  [SMALL_STATE(1264)] = 55347,
  [SMALL_STATE(1265)] = 55378,
  [SMALL_STATE(1266)] = 55419,
  [SMALL_STATE(1267)] = 55460,
  [SMALL_STATE(1268)] = 55501,
  [SMALL_STATE(1269)] = 55532,
  [SMALL_STATE(1270)] = 55563,
  [SMALL_STATE(1271)] = 55604,
  [SMALL_STATE(1272)] = 55645,
  [SMALL_STATE(1273)] = 55686,
  [SMALL_STATE(1274)] = 55727,
  [SMALL_STATE(1275)] = 55768,
  [SMALL_STATE(1276)] = 55810,
  [SMALL_STATE(1277)] = 55848,
  [SMALL_STATE(1278)] = 55876,
  [SMALL_STATE(1279)] = 55918,
  [SMALL_STATE(1280)] = 55960,
  [SMALL_STATE(1281)] = 55998,
  [SMALL_STATE(1282)] = 56031,
  [SMALL_STATE(1283)] = 56054,
  [SMALL_STATE(1284)] = 56093,
  [SMALL_STATE(1285)] = 56120,
  [SMALL_STATE(1286)] = 56159,
  [SMALL_STATE(1287)] = 56198,
  [SMALL_STATE(1288)] = 56237,
  [SMALL_STATE(1289)] = 56260,
  [SMALL_STATE(1290)] = 56287,
  [SMALL_STATE(1291)] = 56310,
  [SMALL_STATE(1292)] = 56342,
  [SMALL_STATE(1293)] = 56374,
  [SMALL_STATE(1294)] = 56406,
  [SMALL_STATE(1295)] = 56438,
  [SMALL_STATE(1296)] = 56475,
  [SMALL_STATE(1297)] = 56508,
  [SMALL_STATE(1298)] = 56529,
  [SMALL_STATE(1299)] = 56566,
  [SMALL_STATE(1300)] = 56587,
  [SMALL_STATE(1301)] = 56608,
  [SMALL_STATE(1302)] = 56629,
  [SMALL_STATE(1303)] = 56666,
  [SMALL_STATE(1304)] = 56687,
  [SMALL_STATE(1305)] = 56726,
  [SMALL_STATE(1306)] = 56763,
  [SMALL_STATE(1307)] = 56784,
  [SMALL_STATE(1308)] = 56821,
  [SMALL_STATE(1309)] = 56854,
  [SMALL_STATE(1310)] = 56875,
  [SMALL_STATE(1311)] = 56912,
  [SMALL_STATE(1312)] = 56949,
  [SMALL_STATE(1313)] = 56986,
  [SMALL_STATE(1314)] = 57013,
  [SMALL_STATE(1315)] = 57050,
  [SMALL_STATE(1316)] = 57087,
  [SMALL_STATE(1317)] = 57124,
  [SMALL_STATE(1318)] = 57154,
  [SMALL_STATE(1319)] = 57184,
  [SMALL_STATE(1320)] = 57214,
  [SMALL_STATE(1321)] = 57248,
  [SMALL_STATE(1322)] = 57282,
  [SMALL_STATE(1323)] = 57312,
  [SMALL_STATE(1324)] = 57341,
  [SMALL_STATE(1325)] = 57370,
  [SMALL_STATE(1326)] = 57395,
  [SMALL_STATE(1327)] = 57424,
  [SMALL_STATE(1328)] = 57455,
  [SMALL_STATE(1329)] = 57480,
  [SMALL_STATE(1330)] = 57511,
  [SMALL_STATE(1331)] = 57540,
  [SMALL_STATE(1332)] = 57563,
  [SMALL_STATE(1333)] = 57592,
  [SMALL_STATE(1334)] = 57615,
  [SMALL_STATE(1335)] = 57646,
  [SMALL_STATE(1336)] = 57675,
  [SMALL_STATE(1337)] = 57698,
  [SMALL_STATE(1338)] = 57727,
  [SMALL_STATE(1339)] = 57760,
  [SMALL_STATE(1340)] = 57791,
  [SMALL_STATE(1341)] = 57820,
  [SMALL_STATE(1342)] = 57849,
  [SMALL_STATE(1343)] = 57880,
  [SMALL_STATE(1344)] = 57909,
  [SMALL_STATE(1345)] = 57938,
  [SMALL_STATE(1346)] = 57969,
  [SMALL_STATE(1347)] = 58000,
  [SMALL_STATE(1348)] = 58029,
  [SMALL_STATE(1349)] = 58058,
  [SMALL_STATE(1350)] = 58087,
  [SMALL_STATE(1351)] = 58116,
  [SMALL_STATE(1352)] = 58147,
  [SMALL_STATE(1353)] = 58176,
  [SMALL_STATE(1354)] = 58205,
  [SMALL_STATE(1355)] = 58234,
  [SMALL_STATE(1356)] = 58262,
  [SMALL_STATE(1357)] = 58294,
  [SMALL_STATE(1358)] = 58322,
  [SMALL_STATE(1359)] = 58344,
  [SMALL_STATE(1360)] = 58372,
  [SMALL_STATE(1361)] = 58396,
  [SMALL_STATE(1362)] = 58428,
  [SMALL_STATE(1363)] = 58460,
  [SMALL_STATE(1364)] = 58482,
  [SMALL_STATE(1365)] = 58514,
  [SMALL_STATE(1366)] = 58538,
  [SMALL_STATE(1367)] = 58566,
  [SMALL_STATE(1368)] = 58594,
  [SMALL_STATE(1369)] = 58618,
  [SMALL_STATE(1370)] = 58641,
  [SMALL_STATE(1371)] = 58664,
  [SMALL_STATE(1372)] = 58693,
  [SMALL_STATE(1373)] = 58722,
  [SMALL_STATE(1374)] = 58743,
  [SMALL_STATE(1375)] = 58760,
  [SMALL_STATE(1376)] = 58785,
  [SMALL_STATE(1377)] = 58814,
  [SMALL_STATE(1378)] = 58831,
  [SMALL_STATE(1379)] = 58848,
  [SMALL_STATE(1380)] = 58877,
  [SMALL_STATE(1381)] = 58906,
  [SMALL_STATE(1382)] = 58927,
  [SMALL_STATE(1383)] = 58944,
  [SMALL_STATE(1384)] = 58965,
  [SMALL_STATE(1385)] = 58994,
  [SMALL_STATE(1386)] = 59011,
  [SMALL_STATE(1387)] = 59028,
  [SMALL_STATE(1388)] = 59057,
  [SMALL_STATE(1389)] = 59078,
  [SMALL_STATE(1390)] = 59099,
  [SMALL_STATE(1391)] = 59128,
  [SMALL_STATE(1392)] = 59145,
  [SMALL_STATE(1393)] = 59166,
  [SMALL_STATE(1394)] = 59193,
  [SMALL_STATE(1395)] = 59210,
  [SMALL_STATE(1396)] = 59231,
  [SMALL_STATE(1397)] = 59248,
  [SMALL_STATE(1398)] = 59264,
  [SMALL_STATE(1399)] = 59286,
  [SMALL_STATE(1400)] = 59312,
  [SMALL_STATE(1401)] = 59328,
  [SMALL_STATE(1402)] = 59354,
  [SMALL_STATE(1403)] = 59370,
  [SMALL_STATE(1404)] = 59396,
  [SMALL_STATE(1405)] = 59412,
  [SMALL_STATE(1406)] = 59438,
  [SMALL_STATE(1407)] = 59464,
  [SMALL_STATE(1408)] = 59480,
  [SMALL_STATE(1409)] = 59506,
  [SMALL_STATE(1410)] = 59528,
  [SMALL_STATE(1411)] = 59544,
  [SMALL_STATE(1412)] = 59560,
  [SMALL_STATE(1413)] = 59586,
  [SMALL_STATE(1414)] = 59612,
  [SMALL_STATE(1415)] = 59632,
  [SMALL_STATE(1416)] = 59648,
  [SMALL_STATE(1417)] = 59674,
  [SMALL_STATE(1418)] = 59690,
  [SMALL_STATE(1419)] = 59716,
  [SMALL_STATE(1420)] = 59742,
  [SMALL_STATE(1421)] = 59760,
  [SMALL_STATE(1422)] = 59780,
  [SMALL_STATE(1423)] = 59806,
  [SMALL_STATE(1424)] = 59828,
  [SMALL_STATE(1425)] = 59850,
  [SMALL_STATE(1426)] = 59870,
  [SMALL_STATE(1427)] = 59886,
  [SMALL_STATE(1428)] = 59912,
  [SMALL_STATE(1429)] = 59927,
  [SMALL_STATE(1430)] = 59942,
  [SMALL_STATE(1431)] = 59965,
  [SMALL_STATE(1432)] = 59982,
  [SMALL_STATE(1433)] = 59997,
  [SMALL_STATE(1434)] = 60020,
  [SMALL_STATE(1435)] = 60043,
  [SMALL_STATE(1436)] = 60058,
  [SMALL_STATE(1437)] = 60073,
  [SMALL_STATE(1438)] = 60096,
  [SMALL_STATE(1439)] = 60119,
  [SMALL_STATE(1440)] = 60142,
  [SMALL_STATE(1441)] = 60165,
  [SMALL_STATE(1442)] = 60188,
  [SMALL_STATE(1443)] = 60211,
  [SMALL_STATE(1444)] = 60226,
  [SMALL_STATE(1445)] = 60249,
  [SMALL_STATE(1446)] = 60268,
  [SMALL_STATE(1447)] = 60291,
  [SMALL_STATE(1448)] = 60314,
  [SMALL_STATE(1449)] = 60337,
  [SMALL_STATE(1450)] = 60352,
  [SMALL_STATE(1451)] = 60375,
  [SMALL_STATE(1452)] = 60392,
  [SMALL_STATE(1453)] = 60407,
  [SMALL_STATE(1454)] = 60422,
  [SMALL_STATE(1455)] = 60439,
  [SMALL_STATE(1456)] = 60462,
  [SMALL_STATE(1457)] = 60479,
  [SMALL_STATE(1458)] = 60502,
  [SMALL_STATE(1459)] = 60525,
  [SMALL_STATE(1460)] = 60540,
  [SMALL_STATE(1461)] = 60563,
  [SMALL_STATE(1462)] = 60582,
  [SMALL_STATE(1463)] = 60597,
  [SMALL_STATE(1464)] = 60612,
  [SMALL_STATE(1465)] = 60627,
  [SMALL_STATE(1466)] = 60650,
  [SMALL_STATE(1467)] = 60664,
  [SMALL_STATE(1468)] = 60678,
  [SMALL_STATE(1469)] = 60692,
  [SMALL_STATE(1470)] = 60710,
  [SMALL_STATE(1471)] = 60728,
  [SMALL_STATE(1472)] = 60742,
  [SMALL_STATE(1473)] = 60756,
  [SMALL_STATE(1474)] = 60770,
  [SMALL_STATE(1475)] = 60788,
  [SMALL_STATE(1476)] = 60802,
  [SMALL_STATE(1477)] = 60816,
  [SMALL_STATE(1478)] = 60834,
  [SMALL_STATE(1479)] = 60852,
  [SMALL_STATE(1480)] = 60866,
  [SMALL_STATE(1481)] = 60880,
  [SMALL_STATE(1482)] = 60898,
  [SMALL_STATE(1483)] = 60912,
  [SMALL_STATE(1484)] = 60930,
  [SMALL_STATE(1485)] = 60948,
  [SMALL_STATE(1486)] = 60962,
  [SMALL_STATE(1487)] = 60976,
  [SMALL_STATE(1488)] = 60996,
  [SMALL_STATE(1489)] = 61010,
  [SMALL_STATE(1490)] = 61028,
  [SMALL_STATE(1491)] = 61042,
  [SMALL_STATE(1492)] = 61059,
  [SMALL_STATE(1493)] = 61076,
  [SMALL_STATE(1494)] = 61093,
  [SMALL_STATE(1495)] = 61112,
  [SMALL_STATE(1496)] = 61129,
  [SMALL_STATE(1497)] = 61146,
  [SMALL_STATE(1498)] = 61163,
  [SMALL_STATE(1499)] = 61180,
  [SMALL_STATE(1500)] = 61191,
  [SMALL_STATE(1501)] = 61208,
  [SMALL_STATE(1502)] = 61225,
  [SMALL_STATE(1503)] = 61242,
  [SMALL_STATE(1504)] = 61259,
  [SMALL_STATE(1505)] = 61276,
  [SMALL_STATE(1506)] = 61293,
  [SMALL_STATE(1507)] = 61312,
  [SMALL_STATE(1508)] = 61329,
  [SMALL_STATE(1509)] = 61346,
  [SMALL_STATE(1510)] = 61363,
  [SMALL_STATE(1511)] = 61380,
  [SMALL_STATE(1512)] = 61393,
  [SMALL_STATE(1513)] = 61410,
  [SMALL_STATE(1514)] = 61427,
  [SMALL_STATE(1515)] = 61444,
  [SMALL_STATE(1516)] = 61461,
  [SMALL_STATE(1517)] = 61480,
  [SMALL_STATE(1518)] = 61497,
  [SMALL_STATE(1519)] = 61514,
  [SMALL_STATE(1520)] = 61528,
  [SMALL_STATE(1521)] = 61544,
  [SMALL_STATE(1522)] = 61560,
  [SMALL_STATE(1523)] = 61574,
  [SMALL_STATE(1524)] = 61588,
  [SMALL_STATE(1525)] = 61604,
  [SMALL_STATE(1526)] = 61620,
  [SMALL_STATE(1527)] = 61634,
  [SMALL_STATE(1528)] = 61650,
  [SMALL_STATE(1529)] = 61666,
  [SMALL_STATE(1530)] = 61680,
  [SMALL_STATE(1531)] = 61694,
  [SMALL_STATE(1532)] = 61710,
  [SMALL_STATE(1533)] = 61726,
  [SMALL_STATE(1534)] = 61740,
  [SMALL_STATE(1535)] = 61756,
  [SMALL_STATE(1536)] = 61768,
  [SMALL_STATE(1537)] = 61784,
  [SMALL_STATE(1538)] = 61800,
  [SMALL_STATE(1539)] = 61816,
  [SMALL_STATE(1540)] = 61830,
  [SMALL_STATE(1541)] = 61844,
  [SMALL_STATE(1542)] = 61858,
  [SMALL_STATE(1543)] = 61872,
  [SMALL_STATE(1544)] = 61886,
  [SMALL_STATE(1545)] = 61902,
  [SMALL_STATE(1546)] = 61916,
  [SMALL_STATE(1547)] = 61932,
  [SMALL_STATE(1548)] = 61948,
  [SMALL_STATE(1549)] = 61962,
  [SMALL_STATE(1550)] = 61976,
  [SMALL_STATE(1551)] = 61990,
  [SMALL_STATE(1552)] = 62006,
  [SMALL_STATE(1553)] = 62020,
  [SMALL_STATE(1554)] = 62034,
  [SMALL_STATE(1555)] = 62050,
  [SMALL_STATE(1556)] = 62064,
  [SMALL_STATE(1557)] = 62078,
  [SMALL_STATE(1558)] = 62094,
  [SMALL_STATE(1559)] = 62110,
  [SMALL_STATE(1560)] = 62126,
  [SMALL_STATE(1561)] = 62142,
  [SMALL_STATE(1562)] = 62158,
  [SMALL_STATE(1563)] = 62172,
  [SMALL_STATE(1564)] = 62185,
  [SMALL_STATE(1565)] = 62198,
  [SMALL_STATE(1566)] = 62207,
  [SMALL_STATE(1567)] = 62216,
  [SMALL_STATE(1568)] = 62229,
  [SMALL_STATE(1569)] = 62242,
  [SMALL_STATE(1570)] = 62255,
  [SMALL_STATE(1571)] = 62268,
  [SMALL_STATE(1572)] = 62281,
  [SMALL_STATE(1573)] = 62294,
  [SMALL_STATE(1574)] = 62307,
  [SMALL_STATE(1575)] = 62320,
  [SMALL_STATE(1576)] = 62333,
  [SMALL_STATE(1577)] = 62346,
  [SMALL_STATE(1578)] = 62359,
  [SMALL_STATE(1579)] = 62372,
  [SMALL_STATE(1580)] = 62385,
  [SMALL_STATE(1581)] = 62398,
  [SMALL_STATE(1582)] = 62411,
  [SMALL_STATE(1583)] = 62424,
  [SMALL_STATE(1584)] = 62437,
  [SMALL_STATE(1585)] = 62446,
  [SMALL_STATE(1586)] = 62459,
  [SMALL_STATE(1587)] = 62472,
  [SMALL_STATE(1588)] = 62485,
  [SMALL_STATE(1589)] = 62494,
  [SMALL_STATE(1590)] = 62507,
  [SMALL_STATE(1591)] = 62520,
  [SMALL_STATE(1592)] = 62533,
  [SMALL_STATE(1593)] = 62546,
  [SMALL_STATE(1594)] = 62559,
  [SMALL_STATE(1595)] = 62572,
  [SMALL_STATE(1596)] = 62585,
  [SMALL_STATE(1597)] = 62598,
  [SMALL_STATE(1598)] = 62611,
  [SMALL_STATE(1599)] = 62624,
  [SMALL_STATE(1600)] = 62637,
  [SMALL_STATE(1601)] = 62650,
  [SMALL_STATE(1602)] = 62663,
  [SMALL_STATE(1603)] = 62676,
  [SMALL_STATE(1604)] = 62689,
  [SMALL_STATE(1605)] = 62702,
  [SMALL_STATE(1606)] = 62715,
  [SMALL_STATE(1607)] = 62728,
  [SMALL_STATE(1608)] = 62741,
  [SMALL_STATE(1609)] = 62754,
  [SMALL_STATE(1610)] = 62767,
  [SMALL_STATE(1611)] = 62776,
  [SMALL_STATE(1612)] = 62789,
  [SMALL_STATE(1613)] = 62798,
  [SMALL_STATE(1614)] = 62811,
  [SMALL_STATE(1615)] = 62824,
  [SMALL_STATE(1616)] = 62837,
  [SMALL_STATE(1617)] = 62850,
  [SMALL_STATE(1618)] = 62863,
  [SMALL_STATE(1619)] = 62874,
  [SMALL_STATE(1620)] = 62887,
  [SMALL_STATE(1621)] = 62900,
  [SMALL_STATE(1622)] = 62913,
  [SMALL_STATE(1623)] = 62922,
  [SMALL_STATE(1624)] = 62935,
  [SMALL_STATE(1625)] = 62948,
  [SMALL_STATE(1626)] = 62961,
  [SMALL_STATE(1627)] = 62974,
  [SMALL_STATE(1628)] = 62987,
  [SMALL_STATE(1629)] = 63000,
  [SMALL_STATE(1630)] = 63013,
  [SMALL_STATE(1631)] = 63026,
  [SMALL_STATE(1632)] = 63039,
  [SMALL_STATE(1633)] = 63052,
  [SMALL_STATE(1634)] = 63065,
  [SMALL_STATE(1635)] = 63078,
  [SMALL_STATE(1636)] = 63091,
  [SMALL_STATE(1637)] = 63104,
  [SMALL_STATE(1638)] = 63117,
  [SMALL_STATE(1639)] = 63130,
  [SMALL_STATE(1640)] = 63143,
  [SMALL_STATE(1641)] = 63154,
  [SMALL_STATE(1642)] = 63165,
  [SMALL_STATE(1643)] = 63174,
  [SMALL_STATE(1644)] = 63187,
  [SMALL_STATE(1645)] = 63200,
  [SMALL_STATE(1646)] = 63213,
  [SMALL_STATE(1647)] = 63226,
  [SMALL_STATE(1648)] = 63239,
  [SMALL_STATE(1649)] = 63252,
  [SMALL_STATE(1650)] = 63263,
  [SMALL_STATE(1651)] = 63276,
  [SMALL_STATE(1652)] = 63289,
  [SMALL_STATE(1653)] = 63302,
  [SMALL_STATE(1654)] = 63315,
  [SMALL_STATE(1655)] = 63328,
  [SMALL_STATE(1656)] = 63341,
  [SMALL_STATE(1657)] = 63350,
  [SMALL_STATE(1658)] = 63359,
  [SMALL_STATE(1659)] = 63372,
  [SMALL_STATE(1660)] = 63383,
  [SMALL_STATE(1661)] = 63396,
  [SMALL_STATE(1662)] = 63409,
  [SMALL_STATE(1663)] = 63422,
  [SMALL_STATE(1664)] = 63435,
  [SMALL_STATE(1665)] = 63448,
  [SMALL_STATE(1666)] = 63461,
  [SMALL_STATE(1667)] = 63474,
  [SMALL_STATE(1668)] = 63487,
  [SMALL_STATE(1669)] = 63500,
  [SMALL_STATE(1670)] = 63513,
  [SMALL_STATE(1671)] = 63526,
  [SMALL_STATE(1672)] = 63535,
  [SMALL_STATE(1673)] = 63548,
  [SMALL_STATE(1674)] = 63561,
  [SMALL_STATE(1675)] = 63574,
  [SMALL_STATE(1676)] = 63587,
  [SMALL_STATE(1677)] = 63600,
  [SMALL_STATE(1678)] = 63613,
  [SMALL_STATE(1679)] = 63626,
  [SMALL_STATE(1680)] = 63636,
  [SMALL_STATE(1681)] = 63646,
  [SMALL_STATE(1682)] = 63656,
  [SMALL_STATE(1683)] = 63666,
  [SMALL_STATE(1684)] = 63674,
  [SMALL_STATE(1685)] = 63684,
  [SMALL_STATE(1686)] = 63694,
  [SMALL_STATE(1687)] = 63702,
  [SMALL_STATE(1688)] = 63712,
  [SMALL_STATE(1689)] = 63720,
  [SMALL_STATE(1690)] = 63730,
  [SMALL_STATE(1691)] = 63740,
  [SMALL_STATE(1692)] = 63750,
  [SMALL_STATE(1693)] = 63760,
  [SMALL_STATE(1694)] = 63770,
  [SMALL_STATE(1695)] = 63778,
  [SMALL_STATE(1696)] = 63786,
  [SMALL_STATE(1697)] = 63796,
  [SMALL_STATE(1698)] = 63806,
  [SMALL_STATE(1699)] = 63816,
  [SMALL_STATE(1700)] = 63824,
  [SMALL_STATE(1701)] = 63834,
  [SMALL_STATE(1702)] = 63842,
  [SMALL_STATE(1703)] = 63852,
  [SMALL_STATE(1704)] = 63862,
  [SMALL_STATE(1705)] = 63872,
  [SMALL_STATE(1706)] = 63882,
  [SMALL_STATE(1707)] = 63892,
  [SMALL_STATE(1708)] = 63900,
  [SMALL_STATE(1709)] = 63910,
  [SMALL_STATE(1710)] = 63920,
  [SMALL_STATE(1711)] = 63930,
  [SMALL_STATE(1712)] = 63940,
  [SMALL_STATE(1713)] = 63948,
  [SMALL_STATE(1714)] = 63958,
  [SMALL_STATE(1715)] = 63968,
  [SMALL_STATE(1716)] = 63978,
  [SMALL_STATE(1717)] = 63988,
  [SMALL_STATE(1718)] = 63998,
  [SMALL_STATE(1719)] = 64008,
  [SMALL_STATE(1720)] = 64018,
  [SMALL_STATE(1721)] = 64028,
  [SMALL_STATE(1722)] = 64038,
  [SMALL_STATE(1723)] = 64048,
  [SMALL_STATE(1724)] = 64058,
  [SMALL_STATE(1725)] = 64068,
  [SMALL_STATE(1726)] = 64078,
  [SMALL_STATE(1727)] = 64088,
  [SMALL_STATE(1728)] = 64098,
  [SMALL_STATE(1729)] = 64108,
  [SMALL_STATE(1730)] = 64118,
  [SMALL_STATE(1731)] = 64128,
  [SMALL_STATE(1732)] = 64138,
  [SMALL_STATE(1733)] = 64146,
  [SMALL_STATE(1734)] = 64156,
  [SMALL_STATE(1735)] = 64166,
  [SMALL_STATE(1736)] = 64174,
  [SMALL_STATE(1737)] = 64184,
  [SMALL_STATE(1738)] = 64194,
  [SMALL_STATE(1739)] = 64204,
  [SMALL_STATE(1740)] = 64214,
  [SMALL_STATE(1741)] = 64224,
  [SMALL_STATE(1742)] = 64234,
  [SMALL_STATE(1743)] = 64244,
  [SMALL_STATE(1744)] = 64254,
  [SMALL_STATE(1745)] = 64262,
  [SMALL_STATE(1746)] = 64272,
  [SMALL_STATE(1747)] = 64282,
  [SMALL_STATE(1748)] = 64292,
  [SMALL_STATE(1749)] = 64300,
  [SMALL_STATE(1750)] = 64310,
  [SMALL_STATE(1751)] = 64320,
  [SMALL_STATE(1752)] = 64330,
  [SMALL_STATE(1753)] = 64340,
  [SMALL_STATE(1754)] = 64350,
  [SMALL_STATE(1755)] = 64358,
  [SMALL_STATE(1756)] = 64368,
  [SMALL_STATE(1757)] = 64378,
  [SMALL_STATE(1758)] = 64388,
  [SMALL_STATE(1759)] = 64396,
  [SMALL_STATE(1760)] = 64404,
  [SMALL_STATE(1761)] = 64414,
  [SMALL_STATE(1762)] = 64424,
  [SMALL_STATE(1763)] = 64434,
  [SMALL_STATE(1764)] = 64444,
  [SMALL_STATE(1765)] = 64452,
  [SMALL_STATE(1766)] = 64460,
  [SMALL_STATE(1767)] = 64470,
  [SMALL_STATE(1768)] = 64477,
  [SMALL_STATE(1769)] = 64484,
  [SMALL_STATE(1770)] = 64491,
  [SMALL_STATE(1771)] = 64498,
  [SMALL_STATE(1772)] = 64505,
  [SMALL_STATE(1773)] = 64512,
  [SMALL_STATE(1774)] = 64519,
  [SMALL_STATE(1775)] = 64526,
  [SMALL_STATE(1776)] = 64533,
  [SMALL_STATE(1777)] = 64540,
  [SMALL_STATE(1778)] = 64547,
  [SMALL_STATE(1779)] = 64554,
  [SMALL_STATE(1780)] = 64561,
  [SMALL_STATE(1781)] = 64568,
  [SMALL_STATE(1782)] = 64575,
  [SMALL_STATE(1783)] = 64582,
  [SMALL_STATE(1784)] = 64589,
  [SMALL_STATE(1785)] = 64596,
  [SMALL_STATE(1786)] = 64603,
  [SMALL_STATE(1787)] = 64610,
  [SMALL_STATE(1788)] = 64617,
  [SMALL_STATE(1789)] = 64624,
  [SMALL_STATE(1790)] = 64631,
  [SMALL_STATE(1791)] = 64638,
  [SMALL_STATE(1792)] = 64645,
  [SMALL_STATE(1793)] = 64652,
  [SMALL_STATE(1794)] = 64659,
  [SMALL_STATE(1795)] = 64666,
  [SMALL_STATE(1796)] = 64673,
  [SMALL_STATE(1797)] = 64680,
  [SMALL_STATE(1798)] = 64687,
  [SMALL_STATE(1799)] = 64694,
  [SMALL_STATE(1800)] = 64701,
  [SMALL_STATE(1801)] = 64708,
  [SMALL_STATE(1802)] = 64715,
  [SMALL_STATE(1803)] = 64722,
  [SMALL_STATE(1804)] = 64729,
  [SMALL_STATE(1805)] = 64736,
  [SMALL_STATE(1806)] = 64743,
  [SMALL_STATE(1807)] = 64750,
  [SMALL_STATE(1808)] = 64757,
  [SMALL_STATE(1809)] = 64764,
  [SMALL_STATE(1810)] = 64771,
  [SMALL_STATE(1811)] = 64778,
  [SMALL_STATE(1812)] = 64785,
  [SMALL_STATE(1813)] = 64792,
  [SMALL_STATE(1814)] = 64799,
  [SMALL_STATE(1815)] = 64806,
  [SMALL_STATE(1816)] = 64813,
  [SMALL_STATE(1817)] = 64820,
  [SMALL_STATE(1818)] = 64827,
  [SMALL_STATE(1819)] = 64834,
  [SMALL_STATE(1820)] = 64841,
  [SMALL_STATE(1821)] = 64848,
  [SMALL_STATE(1822)] = 64855,
  [SMALL_STATE(1823)] = 64862,
  [SMALL_STATE(1824)] = 64869,
  [SMALL_STATE(1825)] = 64876,
  [SMALL_STATE(1826)] = 64883,
  [SMALL_STATE(1827)] = 64890,
  [SMALL_STATE(1828)] = 64897,
  [SMALL_STATE(1829)] = 64904,
  [SMALL_STATE(1830)] = 64911,
  [SMALL_STATE(1831)] = 64918,
  [SMALL_STATE(1832)] = 64925,
  [SMALL_STATE(1833)] = 64932,
  [SMALL_STATE(1834)] = 64939,
  [SMALL_STATE(1835)] = 64946,
  [SMALL_STATE(1836)] = 64953,
  [SMALL_STATE(1837)] = 64960,
  [SMALL_STATE(1838)] = 64967,
  [SMALL_STATE(1839)] = 64974,
  [SMALL_STATE(1840)] = 64981,
  [SMALL_STATE(1841)] = 64988,
  [SMALL_STATE(1842)] = 64995,
  [SMALL_STATE(1843)] = 65002,
  [SMALL_STATE(1844)] = 65009,
  [SMALL_STATE(1845)] = 65016,
  [SMALL_STATE(1846)] = 65023,
  [SMALL_STATE(1847)] = 65030,
  [SMALL_STATE(1848)] = 65037,
  [SMALL_STATE(1849)] = 65044,
  [SMALL_STATE(1850)] = 65051,
  [SMALL_STATE(1851)] = 65058,
  [SMALL_STATE(1852)] = 65065,
  [SMALL_STATE(1853)] = 65072,
  [SMALL_STATE(1854)] = 65079,
  [SMALL_STATE(1855)] = 65086,
  [SMALL_STATE(1856)] = 65093,
  [SMALL_STATE(1857)] = 65100,
  [SMALL_STATE(1858)] = 65107,
  [SMALL_STATE(1859)] = 65114,
  [SMALL_STATE(1860)] = 65121,
  [SMALL_STATE(1861)] = 65128,
  [SMALL_STATE(1862)] = 65135,
  [SMALL_STATE(1863)] = 65142,
  [SMALL_STATE(1864)] = 65149,
  [SMALL_STATE(1865)] = 65156,
  [SMALL_STATE(1866)] = 65163,
  [SMALL_STATE(1867)] = 65170,
  [SMALL_STATE(1868)] = 65177,
  [SMALL_STATE(1869)] = 65184,
  [SMALL_STATE(1870)] = 65191,
  [SMALL_STATE(1871)] = 65198,
  [SMALL_STATE(1872)] = 65205,
  [SMALL_STATE(1873)] = 65212,
  [SMALL_STATE(1874)] = 65219,
  [SMALL_STATE(1875)] = 65226,
  [SMALL_STATE(1876)] = 65233,
  [SMALL_STATE(1877)] = 65240,
  [SMALL_STATE(1878)] = 65247,
  [SMALL_STATE(1879)] = 65254,
  [SMALL_STATE(1880)] = 65261,
  [SMALL_STATE(1881)] = 65268,
  [SMALL_STATE(1882)] = 65275,
  [SMALL_STATE(1883)] = 65282,
  [SMALL_STATE(1884)] = 65289,
  [SMALL_STATE(1885)] = 65296,
  [SMALL_STATE(1886)] = 65303,
  [SMALL_STATE(1887)] = 65310,
  [SMALL_STATE(1888)] = 65317,
  [SMALL_STATE(1889)] = 65324,
  [SMALL_STATE(1890)] = 65331,
  [SMALL_STATE(1891)] = 65338,
  [SMALL_STATE(1892)] = 65345,
  [SMALL_STATE(1893)] = 65352,
  [SMALL_STATE(1894)] = 65359,
  [SMALL_STATE(1895)] = 65366,
  [SMALL_STATE(1896)] = 65373,
  [SMALL_STATE(1897)] = 65380,
  [SMALL_STATE(1898)] = 65387,
  [SMALL_STATE(1899)] = 65394,
  [SMALL_STATE(1900)] = 65401,
  [SMALL_STATE(1901)] = 65408,
  [SMALL_STATE(1902)] = 65415,
  [SMALL_STATE(1903)] = 65422,
  [SMALL_STATE(1904)] = 65429,
  [SMALL_STATE(1905)] = 65436,
  [SMALL_STATE(1906)] = 65443,
  [SMALL_STATE(1907)] = 65450,
  [SMALL_STATE(1908)] = 65457,
  [SMALL_STATE(1909)] = 65464,
  [SMALL_STATE(1910)] = 65471,
  [SMALL_STATE(1911)] = 65478,
  [SMALL_STATE(1912)] = 65485,
  [SMALL_STATE(1913)] = 65492,
  [SMALL_STATE(1914)] = 65499,
  [SMALL_STATE(1915)] = 65506,
  [SMALL_STATE(1916)] = 65513,
  [SMALL_STATE(1917)] = 65520,
  [SMALL_STATE(1918)] = 65527,
  [SMALL_STATE(1919)] = 65534,
  [SMALL_STATE(1920)] = 65541,
  [SMALL_STATE(1921)] = 65548,
  [SMALL_STATE(1922)] = 65555,
  [SMALL_STATE(1923)] = 65562,
  [SMALL_STATE(1924)] = 65569,
  [SMALL_STATE(1925)] = 65576,
  [SMALL_STATE(1926)] = 65583,
  [SMALL_STATE(1927)] = 65590,
  [SMALL_STATE(1928)] = 65597,
  [SMALL_STATE(1929)] = 65604,
  [SMALL_STATE(1930)] = 65611,
  [SMALL_STATE(1931)] = 65618,
  [SMALL_STATE(1932)] = 65625,
  [SMALL_STATE(1933)] = 65632,
  [SMALL_STATE(1934)] = 65639,
  [SMALL_STATE(1935)] = 65646,
  [SMALL_STATE(1936)] = 65653,
  [SMALL_STATE(1937)] = 65660,
  [SMALL_STATE(1938)] = 65667,
  [SMALL_STATE(1939)] = 65674,
  [SMALL_STATE(1940)] = 65681,
  [SMALL_STATE(1941)] = 65688,
  [SMALL_STATE(1942)] = 65695,
  [SMALL_STATE(1943)] = 65702,
  [SMALL_STATE(1944)] = 65709,
  [SMALL_STATE(1945)] = 65716,
  [SMALL_STATE(1946)] = 65723,
  [SMALL_STATE(1947)] = 65730,
  [SMALL_STATE(1948)] = 65737,
  [SMALL_STATE(1949)] = 65744,
  [SMALL_STATE(1950)] = 65751,
  [SMALL_STATE(1951)] = 65758,
  [SMALL_STATE(1952)] = 65765,
  [SMALL_STATE(1953)] = 65772,
  [SMALL_STATE(1954)] = 65779,
  [SMALL_STATE(1955)] = 65786,
  [SMALL_STATE(1956)] = 65793,
  [SMALL_STATE(1957)] = 65800,
  [SMALL_STATE(1958)] = 65807,
  [SMALL_STATE(1959)] = 65814,
  [SMALL_STATE(1960)] = 65821,
  [SMALL_STATE(1961)] = 65828,
  [SMALL_STATE(1962)] = 65835,
  [SMALL_STATE(1963)] = 65842,
  [SMALL_STATE(1964)] = 65849,
  [SMALL_STATE(1965)] = 65856,
  [SMALL_STATE(1966)] = 65863,
  [SMALL_STATE(1967)] = 65870,
  [SMALL_STATE(1968)] = 65877,
  [SMALL_STATE(1969)] = 65884,
  [SMALL_STATE(1970)] = 65891,
  [SMALL_STATE(1971)] = 65898,
  [SMALL_STATE(1972)] = 65905,
  [SMALL_STATE(1973)] = 65912,
  [SMALL_STATE(1974)] = 65919,
  [SMALL_STATE(1975)] = 65926,
  [SMALL_STATE(1976)] = 65933,
  [SMALL_STATE(1977)] = 65940,
  [SMALL_STATE(1978)] = 65947,
  [SMALL_STATE(1979)] = 65954,
  [SMALL_STATE(1980)] = 65961,
  [SMALL_STATE(1981)] = 65968,
  [SMALL_STATE(1982)] = 65975,
  [SMALL_STATE(1983)] = 65982,
  [SMALL_STATE(1984)] = 65989,
  [SMALL_STATE(1985)] = 65996,
  [SMALL_STATE(1986)] = 66003,
  [SMALL_STATE(1987)] = 66010,
  [SMALL_STATE(1988)] = 66017,
  [SMALL_STATE(1989)] = 66024,
  [SMALL_STATE(1990)] = 66031,
  [SMALL_STATE(1991)] = 66038,
  [SMALL_STATE(1992)] = 66045,
  [SMALL_STATE(1993)] = 66052,
  [SMALL_STATE(1994)] = 66059,
  [SMALL_STATE(1995)] = 66066,
  [SMALL_STATE(1996)] = 66073,
  [SMALL_STATE(1997)] = 66080,
  [SMALL_STATE(1998)] = 66087,
  [SMALL_STATE(1999)] = 66094,
  [SMALL_STATE(2000)] = 66101,
  [SMALL_STATE(2001)] = 66108,
  [SMALL_STATE(2002)] = 66115,
  [SMALL_STATE(2003)] = 66122,
  [SMALL_STATE(2004)] = 66129,
  [SMALL_STATE(2005)] = 66136,
  [SMALL_STATE(2006)] = 66143,
  [SMALL_STATE(2007)] = 66150,
  [SMALL_STATE(2008)] = 66157,
  [SMALL_STATE(2009)] = 66164,
  [SMALL_STATE(2010)] = 66171,
  [SMALL_STATE(2011)] = 66178,
  [SMALL_STATE(2012)] = 66185,
  [SMALL_STATE(2013)] = 66192,
  [SMALL_STATE(2014)] = 66199,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_translation_unit, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1392),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1879),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1200),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1960),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1680),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(541),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(873),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1912),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1722),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1882),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(886),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(722),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(711),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1898),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1674),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1506),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1681),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1693),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1881),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1946),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1953),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1865),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1890),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2013),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1848),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1899),
  [93] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1456),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1618),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1103),
  [103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(680),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1902),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1903),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(24),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1974),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1723),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(374),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(994),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1720),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1740),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1938),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1766),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(314),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2008),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(468),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1858),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1870),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1911),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1742),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1868),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(684),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(318),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(354),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef, 2, 0, 17),
  [173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif, 3, 0, 41),
  [175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef, 3, 0, 17),
  [177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif, 4, 0, 41),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(264),
  [183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(291),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(297),
  [203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(400),
  [206] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1383),
  [209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1902),
  [212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1199),
  [215] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0),
  [217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1903),
  [220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1723),
  [223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(378),
  [226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [229] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(555),
  [235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(83),
  [238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(374),
  [241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(994),
  [244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(881),
  [247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1722),
  [253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [259] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(33),
  [262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [265] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [271] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [277] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1674),
  [280] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1408),
  [283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1720),
  [289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1740),
  [292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(599),
  [295] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1938),
  [298] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1766),
  [301] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(314),
  [304] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2008),
  [307] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(468),
  [310] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1858),
  [313] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1870),
  [316] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1911),
  [319] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1742),
  [322] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1868),
  [325] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(540),
  [328] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [331] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2013),
  [334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1848),
  [337] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1899),
  [340] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1456),
  [343] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(684),
  [346] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1618),
  [349] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(684),
  [355] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(680),
  [358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(405),
  [360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1811),
  [368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1750),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1755),
  [384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1721),
  [386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(564),
  [388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1959),
  [390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1757),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1970),
  [396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(475),
  [398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1983),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1987),
  [402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1819),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1724),
  [406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1824),
  [408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(406),
  [410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1388),
  [412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1958),
  [414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1205),
  [416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else, 1, 0, 0),
  [418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1936),
  [420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1704),
  [422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1003),
  [428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(882),
  [430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1749),
  [436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(601),
  [438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2004),
  [440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1741),
  [442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2011),
  [446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(464),
  [448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1945),
  [450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1951),
  [452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1941),
  [454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1751),
  [456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1801),
  [458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [460] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(406),
  [463] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1388),
  [466] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1958),
  [469] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1205),
  [472] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1936),
  [475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1704),
  [478] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(176),
  [481] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(376),
  [484] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1003),
  [487] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(882),
  [490] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(38),
  [493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1739),
  [496] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1749),
  [499] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(601),
  [502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2004),
  [505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1741),
  [508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(313),
  [511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(2011),
  [514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(464),
  [517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1945),
  [520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1951),
  [523] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1941),
  [526] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1751),
  [529] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1801),
  [532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [540] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(405),
  [543] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1373),
  [546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1810),
  [549] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1186),
  [552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1811),
  [555] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1750),
  [558] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(245),
  [561] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(377),
  [564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(999),
  [567] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(877),
  [570] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(27),
  [573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0),
  [575] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1755),
  [578] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1721),
  [581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(564),
  [584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1959),
  [587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1757),
  [590] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(372),
  [593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1970),
  [596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(475),
  [599] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1983),
  [602] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1987),
  [605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1819),
  [608] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1724),
  [611] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_repeat1, 2, 0, 0), SHIFT_REPEAT(1824),
  [614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else, 2, 0, 0),
  [624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0),
  [638] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(408),
  [641] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1392),
  [644] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1879),
  [647] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1200),
  [650] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1960),
  [653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1680),
  [656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(378),
  [659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(555),
  [668] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(370),
  [671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(375),
  [674] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(993),
  [677] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(873),
  [680] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [683] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1722),
  [686] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [689] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(886),
  [692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [695] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [701] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [707] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [710] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1674),
  [713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1408),
  [716] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [719] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1681),
  [722] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1693),
  [725] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(522),
  [728] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1881),
  [731] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1733),
  [734] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(337),
  [737] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1946),
  [740] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(458),
  [743] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1953),
  [746] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1865),
  [749] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1890),
  [752] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(540),
  [755] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [758] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(2013),
  [761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1848),
  [764] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1899),
  [767] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1456),
  [770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1103),
  [773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1618),
  [776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [779] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(1103),
  [782] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_translation_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(680),
  [785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_translation_unit, 1, 0, 0),
  [787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(407),
  [789] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 2, 0, 0),
  [791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 0),
  [793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 9),
  [795] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(407),
  [798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0),
  [800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(378),
  [803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [806] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(555),
  [812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(83),
  [815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(374),
  [818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(994),
  [821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [824] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1722),
  [830] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(33),
  [836] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [839] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [842] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1674),
  [851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1408),
  [854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [857] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1720),
  [860] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1740),
  [863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1766),
  [866] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(314),
  [869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2008),
  [872] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(468),
  [875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1858),
  [878] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1870),
  [881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1911),
  [884] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1742),
  [887] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1868),
  [890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(540),
  [893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2013),
  [899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1848),
  [902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1899),
  [905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1456),
  [908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(684),
  [911] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1618),
  [914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(684),
  [920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(680),
  [923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, 0, 9),
  [925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 0),
  [929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 2, 0, 0),
  [931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1725),
  [937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1910),
  [939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, 0, 9),
  [941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 9),
  [943] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(403),
  [946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(245),
  [949] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(377),
  [952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(999),
  [955] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(27),
  [958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0),
  [960] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1755),
  [963] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1721),
  [966] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1757),
  [969] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(372),
  [972] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1970),
  [975] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(475),
  [978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1983),
  [981] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1987),
  [984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1819),
  [987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1724),
  [990] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1824),
  [993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [995] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(404),
  [998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(176),
  [1001] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(376),
  [1004] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1003),
  [1007] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(38),
  [1010] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1739),
  [1013] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1749),
  [1016] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1741),
  [1019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(313),
  [1022] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2011),
  [1025] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(464),
  [1028] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1945),
  [1031] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1951),
  [1034] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1941),
  [1037] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1751),
  [1040] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1801),
  [1043] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(402),
  [1046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(150),
  [1049] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(375),
  [1052] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(993),
  [1055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [1058] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1681),
  [1061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1693),
  [1064] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1733),
  [1067] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(337),
  [1070] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1946),
  [1073] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(458),
  [1076] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1953),
  [1079] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1865),
  [1082] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1890),
  [1085] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1725),
  [1088] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1910),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(380),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [1095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1691),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1711),
  [1099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1992),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1761),
  [1103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [1105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1956),
  [1107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(401),
  [1110] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1691),
  [1113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1711),
  [1116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1992),
  [1119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1761),
  [1122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(410),
  [1124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [1126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [1128] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 27),
  [1130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 27),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [1134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement, 1, 0, 0),
  [1136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement, 1, 0, 0),
  [1138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [1144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [1146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_seh_finally_clause, 2, 0, 8),
  [1148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_seh_finally_clause, 2, 0, 8),
  [1150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 28),
  [1156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 28),
  [1158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, 0, 28),
  [1164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, 0, 28),
  [1166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 5, 0, 80),
  [1168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 5, 0, 80),
  [1170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 5, 0, 84),
  [1172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 5, 0, 84),
  [1174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 89),
  [1176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 89),
  [1178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 90),
  [1180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 90),
  [1182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 6, 0, 75),
  [1184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 6, 0, 75),
  [1186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 48),
  [1188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 48),
  [1190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_statement, 3, 0, 29),
  [1192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_statement, 3, 0, 29),
  [1194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_statement, 3, 0, 31),
  [1196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_statement, 3, 0, 31),
  [1198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_seh_except_clause, 3, 0, 99),
  [1200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_seh_except_clause, 3, 0, 99),
  [1202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 3, 0, 35),
  [1204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 3, 0, 35),
  [1206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 6, 0, 110),
  [1208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 6, 0, 110),
  [1210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 4, 0, 48),
  [1212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 4, 0, 48),
  [1214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_seh_try_statement, 3, 0, 8),
  [1220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_seh_try_statement, 3, 0, 8),
  [1222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 56),
  [1224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 56),
  [1226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 62),
  [1236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 62),
  [1238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 64),
  [1240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 64),
  [1242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_statement, 2, 0, 0),
  [1244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_statement, 2, 0, 0),
  [1246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_seh_leave_statement, 2, 0, 0),
  [1248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_seh_leave_statement, 2, 0, 0),
  [1250] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 75),
  [1252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 75),
  [1254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1258] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 3, 0, 17),
  [1268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 3, 0, 17),
  [1270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call, 3, 0, 18),
  [1272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call, 3, 0, 18),
  [1274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_linkage_specification, 3, 0, 23),
  [1276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_linkage_specification, 3, 0, 23),
  [1278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 4, 0, 41),
  [1280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 4, 0, 41),
  [1282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_def, 4, 0, 39),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_def, 4, 0, 39),
  [1290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_function_def, 4, 0, 40),
  [1292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_function_def, 4, 0, 40),
  [1294] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 4, 0, 17),
  [1296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 4, 0, 17),
  [1298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 4, 0, 42),
  [1300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 4, 0, 42),
  [1302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, 0, 61),
  [1304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, 0, 61),
  [1306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_definition, 4, 0, 66),
  [1308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_definition, 4, 0, 66),
  [1310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_item, 1, 0, 2),
  [1312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_item, 1, 0, 2),
  [1314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, 0, 67),
  [1316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, 0, 67),
  [1318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_function_def, 5, 0, 70),
  [1320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_function_def, 5, 0, 70),
  [1322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 5, 0, 71),
  [1324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 5, 0, 71),
  [1326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 5, 0, 41),
  [1328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 5, 0, 41),
  [1330] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef, 5, 0, 72),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef, 5, 0, 72),
  [1334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__block_item, 1, 0, 0), REDUCE(sym_statement, 1, 0, 0),
  [1337] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__block_item, 1, 0, 0), REDUCE(sym_statement, 1, 0, 0),
  [1340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 5, 0, 93),
  [1346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 5, 0, 93),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_definition, 5, 0, 94),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_definition, 5, 0, 94),
  [1352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if, 6, 0, 96),
  [1354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if, 6, 0, 96),
  [1356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call, 2, 0, 4),
  [1358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call, 2, 0, 4),
  [1360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_def, 3, 0, 17),
  [1362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_def, 3, 0, 17),
  [1364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__empty_declaration, 2, 0, 0),
  [1366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__empty_declaration, 2, 0, 0),
  [1368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_include, 3, 0, 16),
  [1370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_include, 3, 0, 16),
  [1372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [1374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [1376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [1378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(625),
  [1380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sizeof_expression, 4, 0, 37),
  [1382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(619),
  [1384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [1386] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sizeof_expression, 4, 0, 37),
  [1388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [1390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [1392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(558),
  [1394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [1396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [1398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(556),
  [1400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [1402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [1404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [1406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1719),
  [1408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(778),
  [1410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_item, 1, 0, 2),
  [1412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_item, 1, 0, 2),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_expression_statement, 2, 0, 0),
  [1416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_expression_statement, 2, 0, 0),
  [1418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [1420] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(793),
  [1423] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(378),
  [1426] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(541),
  [1429] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(541),
  [1432] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(555),
  [1435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(245),
  [1438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(536),
  [1441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1719),
  [1444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(27),
  [1447] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1755),
  [1450] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1721),
  [1453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(564),
  [1456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1959),
  [1459] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1757),
  [1462] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(372),
  [1465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1970),
  [1468] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(475),
  [1471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1983),
  [1474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1987),
  [1477] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1819),
  [1480] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1724),
  [1483] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1824),
  [1486] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(540),
  [1489] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(500),
  [1492] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(2013),
  [1495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1848),
  [1498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1899),
  [1501] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1456),
  [1504] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(684),
  [1507] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1618),
  [1510] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1531),
  [1513] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(684),
  [1516] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(680),
  [1519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(786),
  [1521] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(786),
  [1524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(83),
  [1527] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(33),
  [1530] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1720),
  [1533] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1740),
  [1536] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(599),
  [1539] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1938),
  [1542] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1766),
  [1545] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(314),
  [1548] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(2008),
  [1551] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(468),
  [1554] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1858),
  [1557] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1870),
  [1560] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1911),
  [1563] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1742),
  [1566] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1868),
  [1569] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(783),
  [1572] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(23),
  [1575] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1691),
  [1578] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1693),
  [1581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(597),
  [1584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1956),
  [1587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1711),
  [1590] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(337),
  [1593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1992),
  [1596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(458),
  [1599] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1953),
  [1602] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1865),
  [1605] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1890),
  [1608] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1761),
  [1611] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__top_level_item, 1, 0, 0), REDUCE(sym__top_level_statement, 1, 0, 0),
  [1614] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__top_level_item, 1, 0, 0), REDUCE(sym__top_level_statement, 1, 0, 0),
  [1617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(790),
  [1619] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(790),
  [1622] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(176),
  [1625] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(38),
  [1628] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1739),
  [1631] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1749),
  [1634] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(601),
  [1637] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(2004),
  [1640] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1741),
  [1643] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(313),
  [1646] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(2011),
  [1649] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(464),
  [1652] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1945),
  [1655] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1951),
  [1658] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1941),
  [1661] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1751),
  [1664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1801),
  [1667] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(778),
  [1670] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(150),
  [1673] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1681),
  [1676] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(522),
  [1679] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1881),
  [1682] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1733),
  [1685] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1946),
  [1688] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1725),
  [1691] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT(1910),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__top_level_expression_statement, 1, 0, 0),
  [1696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__top_level_expression_statement, 1, 0, 0),
  [1698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(715),
  [1701] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(536),
  [1704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [1706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0),
  [1708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0),
  [1710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [1712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [1714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [1716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [1718] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1026),
  [1720] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1666),
  [1722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(591),
  [1724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [1726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [1728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(593),
  [1730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_declaration, 4, 0, 0),
  [1732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_declaration, 4, 0, 0),
  [1734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_declaration, 3, 0, 0),
  [1736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_declaration, 3, 0, 0),
  [1738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [1740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [1742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [1744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [1746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [1748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [1750] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__declaration_modifiers, 1, 0, 0), REDUCE(aux_sym_attributed_declarator_repeat1, 1, 0, 0),
  [1753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 1, 0, 0),
  [1755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 1, 0, 0),
  [1757] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_modifiers, 1, 0, 0),
  [1759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__declaration_modifiers, 1, 0, 0), REDUCE(aux_sym_attributed_declarator_repeat1, 1, 0, 0),
  [1762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 1),
  [1764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [1766] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), REDUCE(sym_expression, 1, 0, 0), SHIFT(988),
  [1770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [1772] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 1), REDUCE(sym_expression, 1, 0, 0),
  [1775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), REDUCE(sym_expression, 1, 0, 0),
  [1778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1),
  [1780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(762),
  [1782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(559),
  [1784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [1786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [1788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [1790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [1792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [1794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [1796] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(322),
  [1799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(827),
  [1801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(604),
  [1803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [1805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(609),
  [1807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [1809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [1811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [1813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1282),
  [1815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [1817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1689),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(419),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [1825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [1829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(422),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(662),
  [1833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1429),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1969),
  [1839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [1841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [1845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [1847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [1849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [1851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [1853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [1855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1436),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [1863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [1865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [1867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [1869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [1871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(418),
  [1873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [1875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [1877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(417),
  [1879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [1881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [1883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(415),
  [1885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [1887] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(827),
  [1890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(379),
  [1893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(604),
  [1896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(604),
  [1899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(605),
  [1902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(609),
  [1905] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(606),
  [1908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(607),
  [1911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(2013),
  [1914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(1848),
  [1917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(1899),
  [1920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(1456),
  [1923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(684),
  [1926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(1618),
  [1929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(1531),
  [1932] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(684),
  [1935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_qualifier, 1, 0, 0), SHIFT(680),
  [1938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [1940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [1942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [1944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1962),
  [1946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [1948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 4, 0, 41),
  [1950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1859),
  [1952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(448),
  [1954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1185),
  [1956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [1958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1679),
  [1960] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 2, 0, 17),
  [1962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(802),
  [1964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [1966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [1968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(779),
  [1970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(844),
  [1972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(862),
  [1974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [1976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [1978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [1980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [1982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1839),
  [1984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(842),
  [1986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [1988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 3, 0, 17),
  [1990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(849),
  [1992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [1994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [1996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(857),
  [1998] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 3, 0, 41),
  [2000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [2004] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(754),
  [2007] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1962),
  [2010] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1197),
  [2013] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0),
  [2015] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1859),
  [2018] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1679),
  [2021] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [2024] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [2027] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [2030] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1738),
  [2033] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [2036] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [2039] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [2042] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [2045] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1674),
  [2048] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1408),
  [2051] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [2054] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1935),
  [2057] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1161),
  [2060] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1876),
  [2063] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1685),
  [2066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0),
  [2068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1963),
  [2070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [2072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_field_declaration_list, 2, 0, 0),
  [2074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1924),
  [2076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1715),
  [2078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1935),
  [2080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [2082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1876),
  [2084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1685),
  [2086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [2088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_else_in_field_declaration_list, 1, 0, 0),
  [2090] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1963),
  [2093] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1159),
  [2096] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1924),
  [2099] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1715),
  [2102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [2104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [2106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [2108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [2110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [2112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [2114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 41),
  [2116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 83),
  [2118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [2120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [2122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [2124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [2126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 2, 0, 0),
  [2128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [2130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [2132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 57),
  [2134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [2136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 104),
  [2138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [2140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [2142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [2144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [2146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [2150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [2152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [2154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [2156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [2158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 2, 0, 57),
  [2160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [2162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [2164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [2166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [2168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [2170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [2172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [2174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [2176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [2178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [2180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [2182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1449),
  [2184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [2186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [2188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [2190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [2192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [2194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [2196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [2198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [2200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [2202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [2204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [2206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [2208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(620),
  [2211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [2213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [2215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1531),
  [2218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [2220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [2222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [2224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(620),
  [2226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenated_string, 3, 0, 0),
  [2228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenated_string, 3, 0, 0),
  [2230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [2232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [2234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0),
  [2236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0),
  [2238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [2241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [2244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [2247] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(1738),
  [2250] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [2253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [2256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 3, 0, 0),
  [2258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 3, 0, 0),
  [2260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 5, 0, 86),
  [2262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 5, 0, 86),
  [2264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 4, 0, 58),
  [2266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 4, 0, 58),
  [2268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 5, 0, 87),
  [2270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 5, 0, 87),
  [2272] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 7, 0, 119),
  [2274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 7, 0, 119),
  [2276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 8, 0, 123),
  [2278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 8, 0, 123),
  [2280] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 6, 0, 107),
  [2282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 6, 0, 107),
  [2284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 6, 0, 108),
  [2286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 6, 0, 108),
  [2288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 7, 0, 118),
  [2290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 7, 0, 118),
  [2292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 2, 0, 0),
  [2294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 2, 0, 0),
  [2296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 8, 0, 124),
  [2298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 8, 0, 124),
  [2300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_gnu_asm_expression, 9, 0, 128),
  [2302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_expression, 9, 0, 128),
  [2304] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(754),
  [2307] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [2310] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(722),
  [2313] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [2316] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1738),
  [2319] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1882),
  [2322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0),
  [2324] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [2327] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [2330] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [2333] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1674),
  [2336] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1408),
  [2339] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__old_style_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [2342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_specifier, 4, 0, 0),
  [2344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_specifier, 4, 0, 0),
  [2346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [2348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [2350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [2352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0),
  [2354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0),
  [2356] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(700),
  [2359] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(662),
  [2362] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_array_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(1969),
  [2365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(621),
  [2367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__string, 1, 0, 0),
  [2369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__string, 1, 0, 0),
  [2371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0),
  [2373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0),
  [2375] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(711),
  [2378] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__type_definition_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [2381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 12),
  [2383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 12),
  [2385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 13),
  [2387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 13),
  [2389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 3, 0, 0),
  [2391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 3, 0, 0),
  [2393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_offsetof_expression, 6, 0, 105),
  [2395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_offsetof_expression, 6, 0, 105),
  [2397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 4, 0, 0),
  [2399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 4, 0, 0),
  [2401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 5, 0, 0),
  [2403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 5, 0, 0),
  [2405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_expression, 8, 0, 0),
  [2407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_expression, 8, 0, 0),
  [2409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_expression, 9, 0, 0),
  [2411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_expression, 9, 0, 0),
  [2413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_expression, 3, 0, 36),
  [2415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, 0, 36),
  [2417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_literal_expression, 4, 0, 46),
  [2419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_literal_expression, 4, 0, 46),
  [2421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alignof_expression, 4, 0, 37),
  [2423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alignof_expression, 4, 0, 37),
  [2425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_null, 1, 0, 0),
  [2427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null, 1, 0, 0),
  [2429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript_expression, 4, 0, 68),
  [2431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_expression, 4, 0, 68),
  [2433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_initializer_list, 2, 0, 0),
  [2435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_list, 2, 0, 0),
  [2437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_char_literal, 3, 0, 0),
  [2439] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_char_literal, 3, 0, 0),
  [2441] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, 0, 5),
  [2443] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, 0, 5),
  [2445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [2447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [2449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1994),
  [2451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_expression, 2, 0, 5),
  [2453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_expression, 2, 0, 5),
  [2455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [2457] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [2459] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [2461] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alignas_qualifier, 4, 0, 0),
  [2463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alignas_qualifier, 4, 0, 0),
  [2465] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [2467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [2469] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 30),
  [2471] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 30),
  [2473] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [2475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [2477] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 4, 0, 46),
  [2479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 4, 0, 46),
  [2481] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 5),
  [2483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 5),
  [2485] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sizeof_expression, 2, 0, 9),
  [2487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sizeof_expression, 2, 0, 9),
  [2489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 2, 0, 7),
  [2491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 2, 0, 7),
  [2493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [2495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(736),
  [2498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 1, 0, 0),
  [2500] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(711),
  [2503] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0),
  [2505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(750),
  [2507] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(1898),
  [2510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [2512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 2, 0, 7),
  [2514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 2, 0, 7),
  [2516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_declspec_modifier, 4, 0, 0),
  [2518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_declspec_modifier, 4, 0, 0),
  [2520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 25),
  [2522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 25),
  [2524] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 54),
  [2526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 54),
  [2528] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 25),
  [2530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 25),
  [2532] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(744),
  [2535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 0),
  [2537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(711),
  [2540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0),
  [2542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [2544] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(1898),
  [2547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [2549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(596),
  [2551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [2553] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 2, 0, 3),
  [2555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 2, 0, 3),
  [2557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 2, 0, 14),
  [2559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 2, 0, 14),
  [2561] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 1, 0, 3),
  [2563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 1, 0, 3),
  [2565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_specifiers, 3, 0, 14),
  [2567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_specifiers, 3, 0, 14),
  [2569] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_storage_class_specifier, 1, 0, 0),
  [2571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_storage_class_specifier, 1, 0, 0),
  [2573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [2575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [2577] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(723),
  [2580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 2, 0, 8),
  [2582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 2, 0, 8),
  [2584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 53),
  [2586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 53),
  [2588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 55),
  [2590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 55),
  [2592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 53),
  [2594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 53),
  [2596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 24),
  [2598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 24),
  [2600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 26),
  [2602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 26),
  [2604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 24),
  [2606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 24),
  [2608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 26),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 26),
  [2612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 79),
  [2614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 79),
  [2616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 2, 0, 8),
  [2618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 2, 0, 8),
  [2620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 54),
  [2622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 54),
  [2624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 6, 0, 79),
  [2626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 6, 0, 79),
  [2628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 15),
  [2630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, -1, 15),
  [2632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [2634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 3),
  [2636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 3),
  [2638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [2640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 55),
  [2642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 55),
  [2644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 0),
  [2646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 0),
  [2648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [2650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 2, 0, 0),
  [2652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 2, 0, 0),
  [2654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 14),
  [2656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 14),
  [2658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 15),
  [2660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, -1, 15),
  [2662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 37),
  [2664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 37),
  [2666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [2668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 38),
  [2670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, -1, 38),
  [2672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [2674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 0),
  [2676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 3, 0, 0),
  [2678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 7),
  [2680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 7),
  [2682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 8),
  [2684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 8),
  [2686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [2688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [2690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 7),
  [2692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 7),
  [2694] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [2697] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0),
  [2700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 5, 0, 53),
  [2702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 5, 0, 53),
  [2704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 3, 0, 0),
  [2706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 3, 0, 0),
  [2708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 24),
  [2710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 24),
  [2712] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(988),
  [2715] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 24),
  [2717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 24),
  [2719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 25),
  [2721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 25),
  [2723] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 4, 0, 26),
  [2725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 4, 0, 26),
  [2727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 24),
  [2729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 24),
  [2731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 25),
  [2733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 25),
  [2735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 4, 0, 26),
  [2737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 4, 0, 26),
  [2739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_type_specifier, 4, -1, 59),
  [2741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_type_specifier, 4, -1, 59),
  [2743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 10),
  [2745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, -1, 10),
  [2747] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 3, 0, 8),
  [2749] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 3, 0, 8),
  [2751] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, 0, 37),
  [2753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 4, 0, 37),
  [2755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, -1, 38),
  [2757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 4, -1, 38),
  [2759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 7),
  [2761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 7),
  [2763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 14),
  [2765] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sized_type_specifier, 2, 0, 14),
  [2767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(741),
  [2769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 6, 0, 77),
  [2771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 6, 0, 77),
  [2773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_specifier, 3, 0, 8),
  [2775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_specifier, 3, 0, 8),
  [2777] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator_list, 4, 0, 0),
  [2779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator_list, 4, 0, 0),
  [2781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 51),
  [2783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 51),
  [2785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_specifier, 5, 0, 53),
  [2787] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_specifier, 5, 0, 53),
  [2789] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [2791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [2793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 71),
  [2795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 71),
  [2797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 78),
  [2799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 78),
  [2801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 4, 0, 41),
  [2803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 4, 0, 41),
  [2805] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 41),
  [2807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 5, 0, 41),
  [2809] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 5, 0, 72),
  [2811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 5, 0, 72),
  [2813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_field_declaration_list, 6, 0, 96),
  [2815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_field_declaration_list, 6, 0, 96),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [2819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1300),
  [2821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 0),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [2825] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 0),
  [2827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(989),
  [2829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(990),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [2833] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 17),
  [2835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 17),
  [2837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 1, 1, 0),
  [2839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 1, 1, 0),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [2843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [2847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 4, 0, 78),
  [2849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, 0, 78),
  [2851] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 43),
  [2853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 43),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [2857] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 3, 0, 17),
  [2859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 3, 0, 17),
  [2861] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 43),
  [2863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 43),
  [2865] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 42),
  [2867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_field_declaration_list, 4, 0, 42),
  [2869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(512),
  [2871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(548),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [2875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(516),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [2879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(518),
  [2881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [2887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(521),
  [2889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 4, 0, 69),
  [2891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 4, 0, 69),
  [2893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [2895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [2897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 95),
  [2899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 95),
  [2901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [2903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(561),
  [2905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [2907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [2909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(613),
  [2911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [2913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(524),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [2919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension_expression, 2, 0, 0),
  [2921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension_expression, 2, 0, 0),
  [2923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0), REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [2926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [2928] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [2930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0), REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [2933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [2935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0), REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [2938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [2940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [2942] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0), REDUCE(sym__old_style_parameter_list, 2, 0, 0),
  [2945] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [2947] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 7),
  [2949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 7),
  [2951] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 7), SHIFT(1912),
  [2954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [2956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1827),
  [2958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [2960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [2962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(566),
  [2964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [2966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 30),
  [2968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 3, 0, 30),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [2972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 51),
  [2974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 4, 0, 51),
  [2976] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 4, 0, 51), SHIFT(1912),
  [2979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 8),
  [2981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 8),
  [2983] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 2, 0, 8), SHIFT(1912),
  [2986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [2988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 24),
  [2990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 3, 0, 24),
  [2992] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 3, 0, 24), SHIFT(1912),
  [2995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 77),
  [2997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_specifier, 5, 0, 77),
  [2999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_enum_specifier, 5, 0, 77), SHIFT(1912),
  [3002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1994),
  [3006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_call_modifier, 1, 0, 0),
  [3008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_call_modifier, 1, 0, 0),
  [3010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [3012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [3016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [3018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [3020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(1644),
  [3023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(942),
  [3026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(580),
  [3028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [3034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(583),
  [3036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(584),
  [3038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(585),
  [3040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [3042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(587),
  [3044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [3046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(588),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [3050] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_type_specifier, 1, 0, 1), REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [3053] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 1),
  [3056] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(988),
  [3060] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__declarator, 1, 0, 0), REDUCE(sym_type_specifier, 1, 0, 1),
  [3063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 0),
  [3065] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 0),
  [3067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [3069] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0),
  [3071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0),
  [3073] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(989),
  [3076] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_pointer_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(990),
  [3079] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_function_declarator, 2, 0, 34),
  [3081] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_function_declarator, 2, 0, 34),
  [3083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [3085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [3087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [3089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [3091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_parameter_list, 3, 0, 0),
  [3093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__old_style_parameter_list, 4, 0, 0),
  [3095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__old_style_parameter_list, 4, 0, 0),
  [3097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1378),
  [3099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [3101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [3103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [3105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [3107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [3109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(533),
  [3111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [3113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [3115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [3117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(549),
  [3119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [3121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(595),
  [3123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [3125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [3127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [3129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [3131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [3133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [3135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [3137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [3139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(602),
  [3141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [3143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [3145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [3147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [3149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(614),
  [3151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [3153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(616),
  [3155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [3157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [3159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [3161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator, 3, 0, 76),
  [3163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator, 3, 0, 76),
  [3165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 2, 0, 0),
  [3167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 2, 0, 0),
  [3169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1564),
  [3171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_pointer_modifier, 1, 0, 0),
  [3173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_pointer_modifier, 1, 0, 0),
  [3175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ms_unaligned_ptr_modifier, 1, 0, 0),
  [3177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_unaligned_ptr_modifier, 1, 0, 0),
  [3179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_call_expression, 2, 0, 13),
  [3181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_call_expression, 2, 0, 13),
  [3183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [3185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [3187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1138),
  [3189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 4, 0, 0),
  [3191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 4, 0, 0),
  [3193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_argument_list, 3, 0, 0),
  [3195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_argument_list, 3, 0, 0),
  [3197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [3199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comma_expression, 3, 0, 45),
  [3201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitfield_clause, 2, 0, 0),
  [3203] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitfield_clause, 2, 0, 0),
  [3205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [3207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [3209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [3211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [3213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [3215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [3217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [3219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [3221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [3223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [3225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [3227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(568),
  [3229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [3231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [3233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(571),
  [3235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [3237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(573),
  [3239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [3241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(575),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [3245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [3247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [3249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 82),
  [3251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 4, 0, 0),
  [3253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [3255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(736),
  [3257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [3259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 5, 0, 115),
  [3261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [3265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_initializer_list_repeat1, 2, 0, 0),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [3269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_init_declarator, 3, 0, 63),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [3273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [3277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 101),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 102),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [3283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [3285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 4, 0, 103),
  [3287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_pair, 3, 0, 114),
  [3289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [3291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initializer_pair, 3, 0, 113),
  [3293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [3303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_statement_body, 3, 0, 81),
  [3305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [3307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [3309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [3313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [3315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [3317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [3319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [3321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [3323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [3325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 14),
  [3327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [3329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 1, 0, 3),
  [3331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [3333] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 3),
  [3335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [3337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [3341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [3343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [3345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [3347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [3349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1622),
  [3351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 14),
  [3353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1657),
  [3355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [3357] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(1129),
  [3360] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(1130),
  [3363] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), SHIFT(1127),
  [3366] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(1131),
  [3369] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(1136),
  [3372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 1, 0, 0), SHIFT(1123),
  [3375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 1, 0, 3),
  [3377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 1, 0, 3),
  [3379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 2, 0, 14),
  [3381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 2, 0, 14),
  [3383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 3, 0, 14),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 3, 0, 14),
  [3387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 1, 0, 43),
  [3389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_type, 2, 0, 3),
  [3391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_type, 2, 0, 3),
  [3393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__preproc_expression, 1, 0, 0),
  [3395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [3397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__preproc_expression, 1, 0, 0),
  [3399] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, 0, 37), SHIFT(723),
  [3402] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 14), SHIFT(1124),
  [3405] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 14), SHIFT(723),
  [3408] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 4, -1, 38), SHIFT(723),
  [3411] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 15), SHIFT(723),
  [3414] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 37), SHIFT(1122),
  [3417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [3419] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, -1, 38), SHIFT(1125),
  [3422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 3, 0, 0), SHIFT(723),
  [3425] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 15), SHIFT(1126),
  [3428] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, -1, 10), SHIFT(723),
  [3431] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 3), SHIFT(723),
  [3434] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_sized_type_specifier, 2, 0, 0), REDUCE(aux_sym_sized_type_specifier_repeat1, 2, 0, 0), SHIFT(723),
  [3438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 1), SHIFT(1134),
  [3441] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_type_specifier, 1, 0, 0), SHIFT(1135),
  [3444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1146),
  [3446] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [3448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1121),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [3454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1743),
  [3456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [3458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [3462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1659),
  [3464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [3466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [3468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [3470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [3474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [3478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [3480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [3482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [3486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1202),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [3492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [3494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [3496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [3498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1795),
  [3502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [3504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [3506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_enum_specifier, 2, 0, 7), SHIFT(1886),
  [3509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [3511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [3513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1716),
  [3515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [3517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1184),
  [3519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [3521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [3523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [3525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [3527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [3529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [3531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [3533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_unary_expression, 2, 0, 5),
  [3535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_unary_expression, 2, 0, 5),
  [3537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [3539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [3541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [3543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [3545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [3547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [3549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [3551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [3553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [3555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [3557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [3559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [3561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [3563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_binary_expression, 3, 0, 30),
  [3565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_binary_expression, 3, 0, 30),
  [3567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [3569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_parenthesized_expression, 3, 0, 0),
  [3571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_parenthesized_expression, 3, 0, 0),
  [3573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_argument_list_repeat1, 2, 0, 0),
  [3575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [3577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [3579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1235),
  [3581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [3583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [3585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [3587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [3589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [3591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [3593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_defined, 2, 0, 0),
  [3595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_defined, 2, 0, 0),
  [3597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [3599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [3601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [3603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [3605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [3607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [3609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [3611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [3613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [3615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_defined, 4, 0, 0),
  [3617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_defined, 4, 0, 0),
  [3619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [3621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [3623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1180),
  [3625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1163),
  [3627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [3629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [3631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [3633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1198),
  [3635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [3637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [3639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [3641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1189),
  [3643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [3645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [3647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [3649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1283),
  [3651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [3653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [3655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [3657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [3659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [3661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [3663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [3665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [3669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [3671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [3673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [3675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 2, 0, 34),
  [3677] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 2, 0, 34),
  [3679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_declaration_declarator, 2, 0, 34),
  [3681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 4, 0, 34),
  [3683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 4, 0, 34),
  [3685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_declaration_declarator, 3, 0, 34),
  [3687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declarator, 3, 0, 34),
  [3689] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declarator, 3, 0, 34),
  [3691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__function_declaration_declarator, 3, 0, 34),
  [3693] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(1284),
  [3696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0),
  [3698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [3701] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 2, 0, 0),
  [3703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [3705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0),
  [3707] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0),
  [3709] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attributed_declarator_repeat1, 2, 0, 0), SHIFT_REPEAT(1722),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [3716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [3718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list, 3, 0, 41),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [3722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1167),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1893),
  [3726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_declarator_repeat1, 1, 0, 0),
  [3728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_declarator_repeat1, 1, 0, 0),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [3734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list, 2, 0, 17),
  [3736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [3738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [3740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_declarator, 2, 0, 0),
  [3742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_declarator, 2, 0, 0),
  [3744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 2, 1, 32),
  [3746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [3748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [3750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 2, 1, 32),
  [3752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 5, 1, 109),
  [3754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 5, 1, 109),
  [3756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 3, 1, 60),
  [3758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 3, 1, 60),
  [3760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_declarator, 4, 1, 88),
  [3762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_declarator, 4, 1, 88),
  [3764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1640),
  [3766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [3768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1794),
  [3770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [3772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 3, 0, 22),
  [3774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 3, 0, 22),
  [3776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 5, 0, 111),
  [3778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 5, 0, 111),
  [3780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declarator, 1, 0, 0),
  [3782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declarator, 1, 0, 0),
  [3784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_declarator, 3, -10, 0),
  [3786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_declarator, 3, -10, 0),
  [3788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 4, 0, 91),
  [3790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 4, 0, 91),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declaration_declarator, 1, 0, 22),
  [3796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declaration_declarator, 1, 0, 22),
  [3798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(413),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [3802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_declarator, 4, 0, 22),
  [3804] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_declarator, 4, 0, 22),
  [3806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [3808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_declarator, 4, -10, 0),
  [3810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_declarator, 4, -10, 0),
  [3812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 1, 0, 0),
  [3814] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 1, 0, 0), REDUCE(aux_sym_function_declarator_repeat1, 1, 0, 0),
  [3817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 1, 0, 0),
  [3819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 5, 1, 109),
  [3821] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 5, 1, 109),
  [3823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 2, 1, 32),
  [3825] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 2, 1, 32),
  [3827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 3, 1, 60),
  [3829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 3, 1, 60),
  [3831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 2, 0, 32),
  [3833] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 2, 0, 32),
  [3835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 2, 0, 35),
  [3837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_field_declarator, 4, 1, 88),
  [3839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_field_declarator, 4, 1, 88),
  [3841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 5, 1, 109),
  [3843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 5, 1, 109),
  [3845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [3847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list_no_comma, 4, 0, 41),
  [3849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1526),
  [3851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [3853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [3855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 4, 1, 88),
  [3857] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 4, 1, 88),
  [3859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [3861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [3863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1790),
  [3865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1482),
  [3867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 1, 0, 0),
  [3869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [3871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [3873] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1640),
  [3876] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1175),
  [3879] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1928),
  [3882] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1680),
  [3885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0),
  [3887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_field_declarator, 2, 0, 0),
  [3889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_field_declarator, 2, 0, 0),
  [3891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list, 4, 0, 41),
  [3893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 2, 1, 32),
  [3895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 2, 1, 32),
  [3897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [3899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 1, 0, 22),
  [3901] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_declarators, 1, 0, 22),
  [3903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list, 3, 0, 17),
  [3905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list_no_comma, 3, 0, 17),
  [3907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type_declarator, 3, 1, 60),
  [3909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type_declarator, 3, 1, 60),
  [3911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list_no_comma, 2, 0, 17),
  [3913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [3915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [3917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [3919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [3921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list_no_comma, 3, 0, 41),
  [3923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [3925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [3927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 32),
  [3929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 32),
  [3931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attributed_type_declarator, 2, 0, 0),
  [3933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attributed_type_declarator, 2, 0, 0),
  [3935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 1, 0, 0),
  [3937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1952),
  [3939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_declarator, 1, 0, 0),
  [3941] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_declarator, 1, 0, 0),
  [3943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [3945] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [3948] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [3951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 1, 0, 0),
  [3955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [3957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_declaration_declarator, 4, 0, 34),
  [3959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__function_declaration_declarator, 4, 0, 34),
  [3961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1675),
  [3963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [3965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [3967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 3, 0, 22),
  [3969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 3, 0, 22),
  [3971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 5, 0, 111),
  [3973] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 5, 0, 111),
  [3975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declarator, 1, 0, 52),
  [3977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declarator, 1, 0, 52),
  [3979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declarator, 1, 0, 0),
  [3981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declarator, 1, 0, 0),
  [3983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1567),
  [3985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_field_declarator, 4, -10, 0),
  [3989] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_field_declarator, 4, -10, 0),
  [3991] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_field_declarator, 3, -10, 0),
  [3993] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_field_declarator, 3, -10, 0),
  [3995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1599),
  [3997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1774),
  [3999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 4, 0, 91),
  [4001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 4, 0, 91),
  [4003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1602),
  [4005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1816),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [4009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_field_declarator, 4, 0, 22),
  [4011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_field_declarator, 4, 0, 22),
  [4013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_field_declarator, 2, 0, 34),
  [4015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_field_declarator, 2, 0, 34),
  [4017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type_declarator, 2, 0, 34),
  [4019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type_declarator, 2, 0, 34),
  [4021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 60),
  [4023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 3, 1, 60),
  [4025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 5, 0, 111),
  [4027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 5, 0, 111),
  [4029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 3, 0, 22),
  [4031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 3, 0, 22),
  [4033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 4, 0, 91),
  [4035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 4, 0, 91),
  [4037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_declarator, 4, 0, 22),
  [4039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type_declarator, 4, 0, 22),
  [4041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [4043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1640),
  [4046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_repeat1, 2, 0, 0),
  [4048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_repeat1, 2, 0, 0),
  [4050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 21),
  [4052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 21),
  [4054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type_declarator, 3, -10, 0),
  [4056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type_declarator, 3, -10, 0),
  [4058] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1, 2, 0, 0), SHIFT_REPEAT(1420),
  [4061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1, 2, 0, 0),
  [4063] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1, 2, 0, 0),
  [4065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 0),
  [4067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 0),
  [4069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_declarator, 1, 0, 1),
  [4071] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_declarator, 1, 0, 1),
  [4073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumerator, 1, 0, 6),
  [4075] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enumerator, 1, 0, 6),
  [4077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [4079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 32),
  [4081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 2, 1, 32),
  [4083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_pointer_declarator, 4, 1, 88),
  [4085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_pointer_declarator, 4, 1, 88),
  [4087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type_declarator, 4, -10, 0),
  [4089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type_declarator, 4, -10, 0),
  [4091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 73),
  [4093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 73),
  [4095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 22),
  [4097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 22),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [4101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_expression_repeat1, 2, 0, 0),
  [4103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(1499),
  [4106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 2, 0, 34),
  [4108] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_function_declarator, 2, 0, 34),
  [4110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [4112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [4118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 0),
  [4120] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 0),
  [4122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 3, 0, 22),
  [4124] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 3, 0, 22),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [4138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_declarator, 1, 0, 0),
  [4140] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__abstract_declarator, 1, 0, 0),
  [4142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [4144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [4146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1839),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [4150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [4152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [4154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 5, 0, 111),
  [4156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 5, 0, 111),
  [4158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [4160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1, 1, 0, 0),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [4164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_preproc_if_in_enumerator_list_no_comma_repeat1, 1, 0, 0),
  [4166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_function_declarator, 1, 0, 20),
  [4168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_function_declarator, 1, 0, 20),
  [4170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_parenthesized_declarator, 4, 0, 0),
  [4172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_parenthesized_declarator, 4, 0, 0),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [4184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 2, 0, 0),
  [4186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 2, 0, 0),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1386),
  [4190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0), SHIFT_REPEAT(530),
  [4193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0),
  [4195] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_pair_repeat1, 2, 0, 0), SHIFT_REPEAT(1839),
  [4198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 97),
  [4200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 97),
  [4202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_parenthesized_declarator, 3, 0, 0),
  [4204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_parenthesized_declarator, 3, 0, 0),
  [4206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_array_declarator, 4, 0, 91),
  [4208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_array_declarator, 4, 0, 91),
  [4210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [4212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 4, 0, 42),
  [4214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 4, 0, 42),
  [4216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list, 5, 0, 71),
  [4218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_enumerator_list, 5, 0, 71),
  [4220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 4, 0, 74),
  [4222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 47),
  [4224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 5, 0, 72),
  [4226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 5, 0, 72),
  [4228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list, 6, 0, 96),
  [4230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_enumerator_list, 6, 0, 96),
  [4232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list, 4, 0, 41),
  [4234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_enumerator_list, 4, 0, 41),
  [4236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 2, 0, 19),
  [4238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 3, 0, 17),
  [4240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 3, 0, 17),
  [4242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 32),
  [4244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 3, 0, 35),
  [4246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 4, 0, 17),
  [4248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_ifdef_in_enumerator_list, 4, 0, 17),
  [4250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_descriptor, 3, 0, 44),
  [4252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 3, 0, 92),
  [4254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enumerator_list_repeat1, 2, 0, 0),
  [4256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list, 5, 0, 41),
  [4258] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_if_in_enumerator_list, 5, 0, 41),
  [4260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 2, 0, 43),
  [4262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 65), SHIFT_REPEAT(1276),
  [4265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 65),
  [4267] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__type_definition_declarators_repeat1, 2, 0, 65),
  [4269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [4271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [4273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator_list, 1, 0, 0),
  [4275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [4279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [4283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_qualifier, 1, 0, 0),
  [4285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition_declarators, 2, 0, 49),
  [4287] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_definition_declarators, 2, 0, 49),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [4293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [4295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declaration_declarator, 3, 0, 100),
  [4297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declaration_declarator, 3, 0, 100),
  [4299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [4301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(705),
  [4303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [4305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [4307] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 2, 0, 65), SHIFT_REPEAT(1353),
  [4310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 2, 0, 65),
  [4312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 2, 0, 65),
  [4314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [4316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__declaration_declarator, 1, 0, 11),
  [4318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__declaration_declarator, 1, 0, 11),
  [4320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 6),
  [4322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [4324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [4326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declaration_declarator, 2, 0, 49),
  [4328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declaration_declarator, 2, 0, 49),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [4332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [4336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__field_declaration_declarator, 2, 0, 22),
  [4338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__field_declaration_declarator, 2, 0, 22),
  [4340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [4342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1530),
  [4344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1914),
  [4346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [4348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1554),
  [4350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [4352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [4354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator_list_no_comma, 2, 0, 0),
  [4356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [4358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [4360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1773),
  [4362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [4364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2001),
  [4366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator_list_no_comma, 1, 0, 0),
  [4368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [4370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [4372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1877),
  [4374] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 116), SHIFT_REPEAT(1425),
  [4377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 116),
  [4379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_char_literal_repeat1, 2, 0, 0),
  [4381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_char_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1530),
  [4384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [4386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1547),
  [4388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1547),
  [4390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [4392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [4394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1520),
  [4396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [4398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 3, 0, 121),
  [4400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [4402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1792),
  [4404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 3, 0, 32),
  [4406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__field_declaration_declarator_repeat1, 3, 0, 32),
  [4408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [4412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [4418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 2, 0, 85),
  [4420] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 126), SHIFT_REPEAT(1389),
  [4423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 126),
  [4425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [4427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 2, 0, 85),
  [4429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [4431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [4433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [4435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1706),
  [4437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [4439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1806),
  [4441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 50),
  [4443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(627),
  [4445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [4447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1698),
  [4449] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 116), SHIFT_REPEAT(1421),
  [4452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 116),
  [4454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [4456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1700),
  [4458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [4460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand_list, 3, 0, 106),
  [4462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_clobber_list, 2, 0, 117),
  [4464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0),
  [4466] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1554),
  [4469] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1554),
  [4472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand_list, 3, 0, 106),
  [4474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [4476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [4478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(839),
  [4480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [4482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1538),
  [4484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [4486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1907),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [4490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1847),
  [4492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_else_in_enumerator_list, 2, 0, 0),
  [4494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [4496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(690),
  [4498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [4504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand, 4, 0, 120),
  [4506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_clobber_list_repeat1, 2, 0, 117),
  [4508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [4510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [4512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [4514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [4516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1702),
  [4518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [4520] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1213),
  [4523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [4525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [4527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [4529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [4531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [4533] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1644),
  [4536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [4538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_preproc_params_repeat1, 2, 0, 0), SHIFT_REPEAT(1707),
  [4541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_preproc_params_repeat1, 2, 0, 0),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [4549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [4551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [4555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_designator, 3, 0, 0),
  [4557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1707),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [4563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [4565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 3, 0, 127),
  [4567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand, 4, 0, 120),
  [4569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [4573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [4575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [4577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [4579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [4581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [4585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [4587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [4589] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_initializer_list_repeat1, 2, 0, 0), SHIFT_REPEAT(443),
  [4592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [4594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [4596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [4598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [4600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [4602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1288),
  [4604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [4606] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 129), SHIFT_REPEAT(1864),
  [4609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 129),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [4615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [4617] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1702),
  [4620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_declaration_repeat1, 2, 0, 0),
  [4622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [4624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [4626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_output_operand, 7, 0, 130),
  [4628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [4630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_input_operand_list_repeat1, 2, 0, 85),
  [4632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [4634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [4636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [4638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [4640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [4642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1562),
  [4644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [4646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [4648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_input_operand, 7, 0, 130),
  [4650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 65), SHIFT_REPEAT(1148),
  [4653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 2, 0, 65),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1727),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [4675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [4681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [4683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [4685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [4687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [4689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [4691] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_designator, 2, 0, 98),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [4695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 2, 0, 122),
  [4697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [4699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1631),
  [4705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [4713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [4717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [4719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_range_designator, 5, 0, 125),
  [4721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [4723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1522),
  [4725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [4727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [4729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [4731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [4735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [4737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [4739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(476),
  [4742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(649),
  [4745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [4749] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_output_operand_list_repeat1, 2, 0, 85),
  [4751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(987),
  [4754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_expression_repeat1, 2, 0, 0),
  [4756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [4758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [4762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [4766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [4768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [4770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1789),
  [4772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [4774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1997),
  [4776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [4778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 4, 0, 0),
  [4780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 4, 0, 0),
  [4782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [4784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1784),
  [4786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [4788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1785),
  [4790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [4792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1835),
  [4794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_gnu_asm_goto_list_repeat1, 2, 0, 122),
  [4796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [4798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1923),
  [4800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [4802] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), REDUCE(aux_sym__old_style_parameter_list_repeat1, 2, 0, 0),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [4807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1826),
  [4809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, 0, 6),
  [4811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_repeat1, 4, 0, 112),
  [4813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [4815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [4817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [4819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [4821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1780),
  [4823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [4825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [4827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1798),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [4831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [4833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [4835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1845),
  [4837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [4839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [4845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1778),
  [4847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 2, 0, 0),
  [4849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 2, 0, 0),
  [4851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_params, 3, 0, 0),
  [4853] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_preproc_params, 3, 0, 0),
  [4855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 1, 0, 0),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [4859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 50),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [4863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [4865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [4867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1961),
  [4869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1645),
  [4871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_gnu_asm_goto_list, 1, 0, 0),
  [4873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [4875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [4883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [4885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [4887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [4889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [4891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [4893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [4895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [4897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [4899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [4901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list, 4, 0, 71),
  [4903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [4905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [4907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [4909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif, 4, 0, 71),
  [4911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [4913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [4915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [4919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [4921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [4923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [4925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [4927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [4929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [4931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [4933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [4935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [4937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [4941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [4947] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [4949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [4951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [4955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [4957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [4959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [4961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list_no_comma, 4, 0, 72),
  [4963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list_no_comma, 5, 0, 41),
  [4965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [4967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [4969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [4973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list, 3, 0, 42),
  [4975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [4977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list_no_comma, 3, 0, 42),
  [4979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [4981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [4983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1834),
  [4985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [4987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [4989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [4991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [4993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [4995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [4997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [4999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [5001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [5003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1837),
  [5005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [5007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [5009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [5011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [5013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [5015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list_no_comma, 4, 0, 71),
  [5017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [5019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [5021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [5023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1688),
  [5025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [5027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list_no_comma, 5, 0, 72),
  [5029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [5031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [5037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 4, 0, 71),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [5041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list_no_comma, 4, 0, 17),
  [5043] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ms_based_modifier, 2, 0, 0),
  [5045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [5047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [5049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [5053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [5055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [5057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [5059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [5061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [5063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [5065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [5067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [5069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [5071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [5075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [5077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1931),
  [5079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [5081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [5083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [5085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [5087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [5089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [5091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [5093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list, 5, 0, 96),
  [5095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [5097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [5099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [5101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_enumerator_list_no_comma, 5, 0, 96),
  [5103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [5105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [5107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [5109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 3, 0, 42),
  [5111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [5113] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif_in_field_declaration_list, 5, 0, 96),
  [5115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [5117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [5119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [5121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [5133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [5135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_field_declaration_list, 4, 0, 72),
  [5137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [5139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [5143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_ifdef_in_enumerator_list_no_comma, 4, 0, 42),
  [5145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [5147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [5149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [5151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [5153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [5155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [5157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [5159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1818),
  [5161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [5163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [5165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [5167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [5169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef_in_enumerator_list, 4, 0, 72),
  [5171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [5173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [5175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [5177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [5179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [5181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [5183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [5185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [5187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [5189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [5191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [5193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [5195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [5197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [5199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [5201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list_no_comma, 5, 0, 71),
  [5203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [5205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [5207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [5209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [5211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [5213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [5215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [5217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [5219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elif, 5, 0, 96),
  [5221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1729),
  [5223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [5225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [5227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [5229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [5231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [5233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [5235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [5237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [5239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [5241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [5243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [5245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [5247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [5249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [5251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [5253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [5255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [5257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [5259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_if_in_enumerator_list_no_comma, 6, 0, 96),
  [5261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [5263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [5265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef, 4, 0, 72),
  [5267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [5269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [5271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_preproc_elifdef, 3, 0, 42),
  [5273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [5275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [5277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [5279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
};

#ifdef __cplusplus
extern "C" {
#endif
#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_c(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .primary_state_ids = ts_primary_state_ids,
    .name = "c",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 24,
      .patch_version = 1,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
